;; Adaptive Laguerre Filter moving average.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart adaptive-laguerre)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart laguerre)
  #:use-module (chart median)
  #:use-module (chart misc)
  #:use-module (chart williams-r))


;;http://www.mesasoftware.com/Papers/Time%20Warp%20Without%20Space%20Travel.exe
;;     Paper by John Elhers (self-extracting zip archive of MS word).
;;


;; Return a procedure which calculates an adaptive laguerre filter over a
;; window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the adaptive laguerre up to (and including) that value, or #f if not
;; enough points yet.
;;
;; An adaptive laguerre filter is in theory influenced by all preceding
;; data, but `adaptive-laguerre-warmup-count' below is designed to determine
;; a warmup count.  After calling PROC with (adaptive-laguerre-warmup-count
;; COUNT) many values, the next call will have an omitted weight of no more
;; than 0.1% of the total.  Omitting 0.1% should be negligable, unless past
;; values are ridiculously bigger than recent ones.
;;
(define-public (adaptive-laguerre-calc-proc count)
  (let ((a-sum       0.0)
	(a-div       0.0)
	(a-prev      #f)
	(a-pdiv      #f)
	(b-sum       0.0)
	(b-div       0.0)
	(b-prev      #f)
	(b-pdiv      #f)
	(c-sum       0.0)
	(c-div       0.0)
	(c-prev      #f)
	(c-pdiv      #f)
	(d-sum       0.0)
	(d-div       0.0)
	(fac         (/ 6.0))
	(prev-filt   #f)
	(alpha       .2)
	(f           .8)
	(will-proc   (williams-%R-calc-proc count))
	(median-proc (median-calc-proc 5)))

    (lambda (x)
      (and-let* (prev-filt
		 (diff      (abs (- x prev-filt)))
		 (w         (will-proc #f #f diff))
		 (new-alpha (median-proc (* 0.01 (+ 100 w)))))
	(set! alpha new-alpha)
	(set! f (- 1 alpha)))

      (set! a-prev a-sum)
      (set! a-pdiv a-div)
      (set! a-sum (+ (* x alpha) (* a-sum f)))
      (set! a-div (+ alpha       (* a-div f)))

      (set! b-prev b-sum)
      (set! b-pdiv b-div)
      (set! b-sum (+ (- (* a-sum f)) a-prev (* b-sum f)))
      (set! b-div (+ (- (* a-div f)) a-pdiv (* b-div f)))

      (set! c-prev c-sum)
      (set! c-pdiv c-div)
      (set! c-sum (+ (- (* b-sum f)) b-prev (* c-sum f)))
      (set! c-div (+ (- (* b-div f)) b-pdiv (* c-div f)))

      (set! d-sum (+ (- (* c-sum f)) c-prev (* d-sum f)))
      (set! d-div (+ (- (* c-div f)) c-pdiv (* d-div f)))

      (and (not (zero? a-div))
	   (begin
	     (set! prev-filt (* fac (+ (/ a-sum a-div)
				       (* 2 (/ b-sum b-div))
				       (* 2 (/ c-sum c-div))
				       (/ d-sum d-div))))
	     (list prev-filt alpha))))))

(define-public (adaptive-laguerre-warmup-count count)
  (laguerre-warmup-count 0.01))


;;-----------------------------------------------------------------------------
;; filter series

(define (adaptive-laguerre-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (adaptive-laguerre-calc-proc count)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (first (proc (array-ref array i)))
			i))))))

(define-public (adaptive-laguerre-series series count)
  (let ((new-series (series-copy series))
	(warmup     (adaptive-laguerre-warmup-count count)))
    (series-name!     new-series (format #f "~a ~a"
					 (_ "Adaptive Laguerre") count))
    (series-datatype! new-series adaptive-laguerre-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count warmup))
    (series-require!  new-series (indicator-require-parts
				  adaptive-laguerre-fill warmup))
    (series-array!    new-series #f)
    new-series))

(define adaptive-laguerre-datatype
  (datatype-new!
   #:key    'adaptive-laguerre
   #:name   (_ "Adaptive Laguerre Filter")
   #:manual (_ "CONTEXT(manual-node): Laguerre Filter")
   #:units  #f
   #:proc   adaptive-laguerre-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'adaptive-laguerre-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))


;;-----------------------------------------------------------------------------
;; alpha series

(define (adaptive-laguerre-alpha-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (adaptive-laguerre-calc-proc count)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (second (proc (array-ref array i)))
			i))))))

(define-public (adaptive-laguerre-alpha-series series count)
  (let ((new-series (series-copy series))
	(warmup     (adaptive-laguerre-warmup-count count)))
    (series-name!      new-series (format #f "~a ~a"
					  (_ "Adaptive Laguerre Alpha") count))
    (series-datatype!  new-series adaptive-laguerre-alpha-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list count warmup))
    (series-require!   new-series (indicator-require-parts
				   adaptive-laguerre-alpha-fill warmup))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define adaptive-laguerre-alpha-datatype
  (datatype-new!
   #:key      'adaptive-laguerre-alpha
   #:name     (_ "Adaptive Laguerre Alpha")
   #:priority -10
   #:manual   (_ "CONTEXT(manual-node): Adaptive Laguerre Filter")
   #:minimum  0
   #:maximum  1
   #:units    'ema-alpha
   #:proc     adaptive-laguerre-alpha-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'adaptive-laguerre-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))

