;; Adjust series for splits, dividends and rollovers.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart adjust)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; misc

(define (frac->inexact x)
  (if (integer? x)
      x
      (exact->inexact x)))


;;-----------------------------------------------------------------------------
;; adjustment

(define (elem-selector-proc adj-splits adj-dividends adj-rollovers)
  (lambda (elem)
    (and (pair? (cdr elem))
	 (or (and adj-splits
		  (eq? (second elem) 'split))
	     (and adj-dividends
		  (eq? (second elem) 'amount) ;; "amount", not string type
		  (number? (third elem)))     ;; number, not 'unknown etc
	     (and adj-rollovers
		  (eq? (second elem) 'month)
		  (list>=? elem 4))))))       ;; new format with amount

(define (amount-factor parent tdate amount decimals imputation)
  (let ((lst (series-prev-n-closes parent #:before tdate #:count 1)))
    (if (null? lst)
	(values 1 1) ;; no closes prior to TDATE
	(let ((close (first lst)))
	  (or (number? imputation)
	      (set! imputation 0))

	  ;; adjust to common decimals basis by cross multiplying
	  (set! close (* close (integer-expt 10 (or decimals 0))))
	  (let ((factor (integer-expt 10 (series-decimals parent))))
	    (set! amount (* (+ amount imputation) factor))
	    (set! imputation (* imputation factor)))

	  (values (- close amount)
		  close)))))

(define (elem-factor parent adj-imputation elem)
  (case (second elem)
    ((amount) ;; (TDATE . (amount X (decimals . N) (imputation . N)))
     (amount-factor parent (first elem) (third elem)
		    (assq-ref (cdddr elem) 'decimals)
		    (and adj-imputation
			 (assq-ref (cdddr elem) 'imputation))))
    ((split)  ;; (TDATE . (split NEW OLD))
     (values (fourth elem)
	     (third elem)))
    ((month)  ;; (TDATE . (month MDATE STRINGNUMBER))
     (amount-factor parent (first elem)
		    (- (string->number (fourth elem)))
		    0 #f))
    (else
     (error "elem-factor: unrecognised element:" elem))))

(define (adjust-fill series new-array lo hi)
  (receive-list (adj-splits adj-dividends adj-imputation adj-rollovers)
      (series-data series)

    (let* ((parent         (series-parent series))
	   (array          (series-array parent lo hi))
	   (open-interest? (ohlcv-open-interest? series))
	   (factor         1)
	   (inv-factor     1)
	   (divs           (series-dividends series)))

      ;; FACTOR and INV-FACTOR are kept as exact integers if possible,
      ;; otherwise flonums; in practice INV-FACTOR is an integer after
      ;; splits like "2 for 1", but otherwise both end up with flonums
      (define (process-elem elem)
	(receive (num den)
	    (elem-factor parent adj-imputation elem)
	  (set! factor     (frac->inexact (/ (* factor num) den)))
	  (set! inv-factor (frac->inexact (/ (* inv-factor den) num)))))


      ;; pick out relevant elements
      ;;    - the desired and recognised adjusts per `elem-selector-proc'
      ;;    - nothing beyond `series-hi', such as future dividends
      ;;    - nothing below LO as it has no effect
      (let ((selector (elem-selector-proc adj-splits adj-dividends
					  adj-rollovers)))
	(set! divs (filter (lambda (elem)
			     (and (selector elem)
				  (<= (car elem) (series-hi series))
				  (> (car elem) lo)))
			   divs)))

      ;; accumulate elements above HI, keep remaining sorted high to low
      (receive (before after)
	  (partition (lambda (elem)
		       (<= (first elem) hi))
		     divs)
	(for-each process-elem after)
	(set! divs (sort before car>)))


      ;; points hi to lo inclusive
      (do ((i hi (1- i)))
	  ((< i lo))

	(if (array-ref array i 0)  ;; open
	    (array-set! new-array (* factor (array-ref array i 0))
			i 0))
	(if (array-ref array i 1)  ;; high
	    (array-set! new-array (* factor (array-ref array i 1))
			i 1))
	(if (array-ref array i 2)  ;; low
	    (array-set! new-array (* factor (array-ref array i 2))
			i 2))
	(if (array-ref array i 3)  ;; close
	    (array-set! new-array (* factor (array-ref array i 3))
			i 3))

	(if (array-ref array i 4)  ;; volume
	    (array-set! new-array (* inv-factor (array-ref array i 4))
			i 4))
	(and open-interest?
	     (array-ref array i 5) ;; open interest
	     (array-set! new-array (* inv-factor (array-ref array i 5))
			 i 5))

	(do ()
	    ((or (null? divs)
		 (not (= i (car (first divs))))))
	  (process-elem (first divs))
	  (set! divs (cdr divs)))))))

(define-public (adjust-series series
			      adj-splits adj-dividends adj-imputation
			      adj-rollovers)
  (ohlcv-required "Adjust Splits/Dividends/Rollovers" series)
  (if (not (any (elem-selector-proc adj-splits adj-dividends adj-rollovers)
		(series-dividends series)))
      series

      (let ((new-series (series-copy series)))
	(series-parent!  new-series series)
	(series-data!    new-series (list adj-splits adj-dividends
					  adj-imputation adj-rollovers))
	(series-require! new-series (indicator-require-parts
				     adjust-fill 0
				     (if (ohlcv-open-interest? series)
					 '(0 5)
					 '(0 4))))
	(series-array!   new-series #f)
	new-series)))

