;; Average Directional Index (ADX).

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart adx)
  #:use-module (ice-9 receive)
  #:use-module (chart dmi)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.stockcharts.com/education/IndicatorAnalysis/indic_ADX.html


;; Return a procedure which calculates the Average Directional Index over an
;; accumulated window of days.
;;
;; Each call (PROC high low close) enters a new day into the calculation,
;; and the return is the ADX for that day.
;;
;; To prime initially, call PROC with (dmi-warmup-count COUNT) many points
;; (from dmi.scm), to get the next call based on a full window.
;;
(define (adx-calc-proc count)
  (let ((dmi-proc (dmi-calc-proc count)))

    (lambda (high low close)
      (receive (di+ di-)
	  (dmi-proc high low close)
	(and di+
	     (let ((sum (+ di+ di-)))
	       (and (not (zero? sum))
		    (* 100.0 (/ (abs (- di+ di-)) sum)))))))))


(define (adx-fill series new-array lo hi)
  (let* ((count    (series-data series))
	 (parent   (series-parent series))
	 (start    (series-prev-n-closes-pos
		    parent #:before lo #:count (dmi-warmup-count count)))
	 (array    (series-array parent start hi))
	 (adx      (adx-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (adx (array-ref array i 1)    ;; high
	       (array-ref array i 2)    ;; low
	       (array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (adx (array-ref array i 1)  ;; high
				     (array-ref array i 2)  ;; low
				     (array-ref array i 3)) ;; close
		      i)))))

(define-public (adx-series series count)
  (ohlcv-required "ADX" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "ADX") count))
    (series-datatype!  new-series adx-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   adx-fill
				   (dmi-warmup-count count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public adx-datatype
  (datatype-new!
   #:key     'adx
   #:name    (_ "ADX - Average Directional Index")
   #:input   datatype-ohlcv
   #:units   'percentage
   #:minimum 0
   #:maximum 100
   #:manual  (_ "CONTEXT(manual-node): Average Directional Index")
   #:proc    adx-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'dmi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
