;; Annotations edit dialog.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ann-edit)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (chart database)
  #:use-module (chart date)
  #:use-module (chart extra-gtk)
  #:use-module (chart gui-common)
  #:use-module (chart main)
  #:use-module (chart misc)
  #:use-module (chart view))



;;-----------------------------------------------------------------------------
;; annotations funcs

;; most recent first
(define (annotation<? x y)
  (< (second y) (second x)))

(define (series-annotation-add! series tdate text)
  (series-notes! series (cons (list 'annotation tdate text)
			      (series-notes series))))

(define (series-annotation-delete! series tdate)
  (series-notes! series (remove (lambda (elem)
				  (and (annotation? elem)
				       (= tdate (second elem))))
				(series-notes series))))

(define (series-annotation-replace! series tdate text)
  (series-annotation-delete! series tdate)
  (series-annotation-add! series tdate text))

(define-public (series-notes-elem-del! series elem)
  (series-notes! series (delq elem (series-notes series))))


;;-----------------------------------------------------------------------------
;; annotations dialog

(define dialog            #f)
(define heading-label 	  #f)
(define list-widget   	  #f)
(define item-to-elem-prop (primitive-make-property #f))
(define current-item      #f)
(define current-symbol    #f)

(define (init)
  (let ((date-widget (date-widget-new))
	(entry       (gtk-entry-new)))

    (define (entry-activate)
      (let* ((adate (date-widget-adate date-widget)))
	(if (adate-weekend? adate)
	    (error-dialog-popup dialog
				(format #f (_ "~a is not a weekday")
					(strftime-utf8 (locale-d-fmt)
						       (adate->tm adate))))
	    (begin
	      (database-call-modify-notes current-symbol
		(lambda (series)
		  (series-annotation-replace!
		   series (adate->tdate adate)
		   (gtk-entry-get-text entry))))
	      (fill)))))

    (set! dialog (gtk-dialog-new-with-buttons "" #f '()))
    (gtk-dialog-add-buttons dialog
			    "gtk-delete" 0
			    "gtk-close"  gtk-response-close
			    "gtk-help"   gtk-response-help)
    (gtk-dialog-set-response-sensitive dialog 0 #f)

    (c-gtk-dialog-hide-on-delete dialog)
    (gtk-signal-connect dialog "response"
      (lambda (id)
	(cond
	 ((= id 0)
	  (database-call-modify-notes current-symbol
	    (lambda (series)
	      (series-notes-elem-del! series
				      (primitive-property-ref
				       item-to-elem-prop current-item))))
	  (fill))

	 ((= id gtk-response-close)
	  (gtk-widget-hide dialog))

	 ((= id gtk-response-help)
	  (manual-popup (_ "CONTEXT(manual-node): Annotations") dialog)))))

    (set! heading-label (gtk-label-new ""))
    (gtk-box-pack-start (gtk-dialog-vbox dialog) heading-label #f #f 0)

    ;; scrolled
    (let ((scrolled (gtk-scrolled-window-new)))
      (gtk-box-pack-start (gtk-dialog-vbox dialog) scrolled #t #t 0)
      (gtk-scrolled-window-set-policy scrolled 'automatic 'always)
      (c-gtk-widget-set-usize-units scrolled '(em . 30) '(height . 15))

      (set! list-widget (gtk-list-new))
      (gtk-scrolled-window-add-with-viewport scrolled list-widget)
      (gtk-signal-connect list-widget "select_child"
	(lambda (item)
	  (set! current-item item)
	  (gtk-dialog-set-response-sensitive dialog 0 #t) ;; delete
	  (let ((ann (primitive-property-ref item-to-elem-prop item)))
	    (date-widget-tdate! date-widget (second ann))
	    (gtk-entry-set-text entry (third ann)))))

      (gtk-signal-connect list-widget "unselect_child"
	(lambda (item)
	  (set! current-item #f)
	  (gtk-dialog-set-response-sensitive dialog 0 #f))))

    (gtk-box-pack-start (gtk-dialog-vbox dialog) date-widget #f #f 0)

    (let ((hbox (gtk-hbox-new #f 4)))
      (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0)
      (gtk-box-pack-start hbox (gtk-label-new (_ "Text")) #f #f 0)
      (gtk-signal-connect entry "activate" entry-activate)
      (gtk-box-pack-start hbox entry  #t #t 0)
      (gtk-widget-grab-focus entry)

      (gtk-signal-connect dialog "map_event"
	(lambda (event)
	  (c-gtk-editable-select-region-noclip entry 0 -1)))

      (let ((button (gtk-button-new-with-label (_ "Enter"))))
	(gtk-box-pack-start hbox button #f #f 0)
	(gtk-signal-connect button "clicked" entry-activate))))

  (gtk-widget-show-all dialog)

  ;; once only, and gc this code
  (set! init noop))

(define (fill)
  (gtk-dialog-set-response-sensitive dialog 0 #f) ;; delete
  (let ((annotes (sort (filter annotation?
			       (series-notes (database-read-series
					      current-symbol)))
		       annotation<?)))
    (c-gtk-list-require-n-labels list-widget (length annotes))
    (for-each (lambda (note item)
		(primitive-property-set! item-to-elem-prop item note)
		(gtk-label-set-text
		 (gtk-bin-get-child item)
		 (string-append (strftime-utf8 (locale-d-fmt)
					       (tdate->tm (second note)))
				"  "
				(third note))))
	      annotes (gtk-container-children list-widget))))

(define-public (annotations-edit-popup)
  (let ((symbol (view-symbol outer)))
    (if (not symbol)
	(error-dialog-popup app-widget (_ "No chart being displayed"))

	(begin
	  (init)
	  (set! current-symbol symbol)
	  (let* ((str (format #f (_ "~a Annotations") current-symbol)))
	    (gtk-label-set-text heading-label str)
	    (gtk-window-set-title dialog (c-make-window-title str)))
	  (fill)
	  (c-gtk-widget-show-or-raise dialog)))))
