;; Aroon up/down and oscillator.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart aroon)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))



;; Return a procedure which calculates Aroon up, down and oscillator over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC x) enters a new value into the window, and the return is
;; three (values UP DOWN OSCILLATOR), or (values #f #f #f) if not enough
;; points yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from start to end).  COUNT is
;; enough to ensure the first PROC call is based on a full window.
;;
(define*-public (aroon-calc-proc count #:optional (data '()))
  (let* ((want-points (1+ count))
	 (q           (list->q (take-right-maybe data want-points)))
	 (points      (q-length q)))
    (set! data #f) ;; gc orig list

    (lambda (x)

      ;; drop old point if full
      (if (>= points want-points)
	  (deq! q)
	  (set! points (1+ points)))

      ;; add new point
      (enq! q x)

      (if (< points want-points)
	  (values #f #f #f)

	  ;; find highest and lowest positions
	  ;;
	  ;; for equal highs or equal lows, the oldest is taken, on the
	  ;; theory that a new high/low has not been established but rather
	  ;; the old one is still in force
	  ;;
	  (let ((high-val (q-front q))
		(high-pos 0)
		(low-val  (q-front q))
		(low-pos  0))
	    (do ((lst (cdr (q-list q)) (cdr lst))
		 (pos 1 (1+ pos)))
		((null? lst))
	      (let ((val (first lst)))
		(if (>= val high-val)
		    (begin
		      (set! high-val val)
		      (set! high-pos pos)))
		(if (< val low-val)
		    (begin
		      (set! low-val val)
		      (set! low-pos pos)))))

	    (set! high-pos (* 100.0 (/ high-pos count)))
	    (set! low-pos  (* 100.0 (/ low-pos  count)))
	    (values high-pos low-pos (- high-pos low-pos)))))))


(define (aroon-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count count))
	 (array  (series-array-closes parent start hi))
	 (aroon  (aroon-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((this (array-ref array i)))
	(aroon this)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(receive (up down osc)
	    (aroon this)
	  (array-set! new-array up   i 0)
	  (array-set! new-array down i 1)
	  (array-set! new-array osc  i 2))))))

(define-public (aroon-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Aroon") count))
    (series-datatype!  new-series aroon-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   aroon-fill count '(0 2)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public aroon-datatype
  (datatype-new!
   #:key       'aroon
   #:name      (_ "Aroon")
   #:units     'percentage-plusorminus100
   #:linestyle 'aroon
   #:minimum   -100
   #:maximum   100
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Aroon")
   #:proc      aroon-series
   #:closes-apply-arg '(2)
   #:parameter-list
   (list (datatype-param-new
	  #:key     'aroon-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 25))))
