;; Accumulative Swing Index (ASI).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart asi)
  #:use-module (ice-9 optargs)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.ensignsoftware.com/espl/espl71.htm
;;     Code for swing index, ASI part confusing though.
;;
;; http://www.investopedia.com/articles/technical/02/100702.asp
;;     Sample chart of Apple (AAPL).
;;


;;-----------------------------------------------------------------------------
;; swing index calc

(define*-public (swing-index-calc prev-open prev-close
				  open high low close
				  #:optional (limit 10000))
  (or prev-open (set! prev-open prev-close))

  (or open (set! open close))
  (or high (set! high close))
  (or low  (set! low  close))

  (let* ((r1 (abs (- high prev-close)))
	 (r2 (abs (- low  prev-close)))
	 (r3 (- high low))
	 (r4 (abs (- prev-close prev-open)))
	 (k  (max r1 r2))

	 (r  (cond ((>= r1 (max r2 r3))
		    (+ r1 (* -0.5 r2) (* 0.25 r4)))
		   ((>= r2 (max r1 r3))
		    (+ r2 (* -0.5 r1) (* 0.25 r4)))
		   (else
		    (+ r3 (* 0.25 r4))))))

    (if (zero? r)
	0
	(* k
	   (/ (+ (- close prev-close)
		 (* 0.5 (- close open))
		 (* 0.25 (- prev-close prev-open)))
	      r)))))


;; Return a procedure which calculates Wilder's swing index over successive
;; values.
;;
;; Each call (PROC open high low close) enters a new value into the
;; calculation, and the return is the swing index at that value, or #f if
;; there hasn't been a previous call yet.
;;
(define*-public (swing-index-calc-proc count #:optional (limit 10000))

  (define prev-open  #f)
  (define prev-close #f)

  (lambda (open high low close)
    (let ((result (and prev-close
		       (swing-index-calc prev-open prev-close
					 open high low close))))
      (set! prev-open  open)
      (set! prev-close close)
      result)))


;;-----------------------------------------------------------------------------
;; swing index series

(define (swing-index-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count 1))
	 (array  (series-array parent start hi))
	 (proc   (swing-index-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 0)    ;; open
		(array-ref array i 1)    ;; high
		(array-ref array i 2)    ;; low
		(array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (proc (array-ref array i 0)  ;; open
				      (array-ref array i 1)  ;; high
				      (array-ref array i 2)  ;; low
				      (array-ref array i 3)) ;; close
		      i)))))

(define-public (swing-index-series series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "Swing"))
    (series-datatype!  new-series swing-index-datatype)
    (series-parent!    new-series series)
    (series-require!   new-series (indicator-require-parts swing-index-fill 1))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())

    new-series))

(define-public swing-index-datatype
  (datatype-new!
   #:key       'swing-index
   #:name      (_ "Swing Index")
   #:priority  -10
   #:units     'swing-index
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Accumulative Swing Index")
   #:proc      swing-index-series))


;;-----------------------------------------------------------------------------
;; accumulative swing index series

(define (asi-func array prev this)
  (swing-index-calc (array-ref array prev 0)   ;; open
		    (array-ref array prev 3)   ;; close
		    (array-ref array this 0)   ;; open
		    (array-ref array this 1)   ;; high
		    (array-ref array this 2)   ;; low
		    (array-ref array this 3))) ;; close

(define-public (asi-series series)
  (accumulation-series series (_ "ASI") asi-datatype asi-func))

(define-public asi-datatype
  (datatype-new! #:key    'asi
		 #:name   (_ "ASI - Accumulative Swing Index")
		 #:units  'accumulative-swing-index
		 #:manual (_ "CONTEXT(manual-node): Accumulative Swing Index")
		 #:proc   asi-series))
