;; Athens Stock Exchange (ATHEX) setups.       -*- coding: iso-8859-7 -*-

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart athens)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format) (format)
  #:use-module (ice-9 regex)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart down-cost)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public timezone-athens
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Athens")
		      '("Europe/Athens"
			;; EU: last Sun in Mar, to last Sun in Oct
			"EET-2EEST,M3.5.0/3,M10.5.0/4")
		      "EET-2"))))

(define-public (athens-symbol? symbol)
  (string-suffix? ".ATH" symbol))

(symbol-timezone! athens-symbol? (_ "Athens") timezone-athens)
(source-help! athens-symbol?
	      (_ "CONTEXT(manual-node): Athens Stock Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - company info

;; the greek pages "/gr/" need greek symbols, the english doesn't work,
;; hence only an english link here
(weblink-handler! athens-symbol?
  (lambda (symbol)
    (list (_ "ATHEX _Company Information")
	  (_ "Open web browser at the Athens Stock Exchange page for this company")
	  (string-append "http://www.ase.gr/content/en/Companies/ListedCo/Profiles/Profile.asp?name=" (chart-symbol-sans-dot symbol)))))


;;-----------------------------------------------------------------------------
;; misc

;; the following transliteration doesn't try very hard at conversion, it's
;; only enough to make sense of some greek chars in english names, in
;; particular  for BANK in ALPHA.ATH.
;;
(let-toplevel
    ((table
      (string-append
       (string-tabulate integer->char 128)
       (make-string 32 #\_) ;; 80 to 9F
       " " ;; A0  NO-BREAK SPACE
       "`" ;; A1  MODIFIER LETTER REVERSED COMMA
       "'" ;; A2  MODIFIER LETTER APOSTROPHE
       "_" ;; A3  POUND SIGN
       "_" ;; A4
       "_" ;; A5
       "|" ;; A6  BROKEN BAR
       "_" ;; A7  SECTION SIGN
       "_" ;; A8  DIAERESIS
       "_" ;; A9
       "_" ;; AA
       "_" ;; AB  LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
       "_" ;; AC  NOT SIGN
       "-" ;; AD  SOFT HYPHEN
       "_" ;; AE
       "-" ;; AF  HORIZONTAL BAR
       "_" ;; B0  DEGREE SIGN
       "_" ;; B1  PLUS-MINUS SIGN
       "_" ;; B2  SUPERSCRIPT TWO
       "_" ;; B3  SUPERSCRIPT THREE
       "_" ;; B4  GREEK TONOS
       "_" ;; B5  GREEK DIALYTIKA TONOS
       "A" ;; B6  GREEK CAPITAL LETTER ALPHA WITH TONOS
       "." ;; B7  MIDDLE DOT
       "E" ;; B8  GREEK CAPITAL LETTER EPSILON WITH TONOS
       "H" ;; B9  GREEK CAPITAL LETTER ETA WITH TONOS
       "I" ;; BA  GREEK CAPITAL LETTER IOTA WITH TONOS
       "_" ;; BB  RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
       "_" ;; BC  GREEK CAPITAL LETTER OMICRON WITH TONOS
       "_" ;; BD  VULGAR FRACTION ONE HALF
       "_" ;; BE  GREEK CAPITAL LETTER UPSILON WITH TONOS
       "O" ;; BF  GREEK CAPITAL LETTER OMEGA WITH TONOS
       "_" ;; C0  GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
       "A" ;; C1  GREEK CAPITAL LETTER ALPHA
       "B" ;; C2  GREEK CAPITAL LETTER BETA
       "G" ;; C3  GREEK CAPITAL LETTER GAMMA
       "D" ;; C4  GREEK CAPITAL LETTER DELTA
       "E" ;; C5  GREEK CAPITAL LETTER EPSILON
       "Z" ;; C6  GREEK CAPITAL LETTER ZETA
       "H" ;; C7  GREEK CAPITAL LETTER ETA
       "_" ;; C8  GREEK CAPITAL LETTER THETA
       "I" ;; C9  GREEK CAPITAL LETTER IOTA
       "K" ;; CA  GREEK CAPITAL LETTER KAPPA
       "L" ;; CB  GREEK CAPITAL LETTER LAMDA
       "M" ;; CC  GREEK CAPITAL LETTER MU
       "N" ;; CD  GREEK CAPITAL LETTER NU
       "X" ;; CE  GREEK CAPITAL LETTER XI
       "_" ;; CF  GREEK CAPITAL LETTER OMICRON
       "P" ;; D0  GREEK CAPITAL LETTER PI
       "R" ;; D1  GREEK CAPITAL LETTER RHO
       "_" ;; D2
       "S" ;; D3  GREEK CAPITAL LETTER SIGMA
       "T" ;; D4  GREEK CAPITAL LETTER TAU
       "_" ;; D5  GREEK CAPITAL LETTER UPSILON
       "_" ;; D6  GREEK CAPITAL LETTER PHI
       "_" ;; D7  GREEK CAPITAL LETTER CHI
       "_" ;; D8  GREEK CAPITAL LETTER PSI
       "O" ;; D9  GREEK CAPITAL LETTER OMEGA
       "_" ;; DA  GREEK CAPITAL LETTER IOTA WITH DIALYTIKA
       "_" ;; DB  GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
       "a" ;; DC  GREEK SMALL LETTER ALPHA WITH TONOS
       "e" ;; DD  GREEK SMALL LETTER EPSILON WITH TONOS
       "_" ;; DE  GREEK SMALL LETTER ETA WITH TONOS
       "i" ;; DF  GREEK SMALL LETTER IOTA WITH TONOS
       "_" ;; E0  GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS
       "a" ;; E1  GREEK SMALL LETTER ALPHA
       "b" ;; E2  GREEK SMALL LETTER BETA
       "g" ;; E3  GREEK SMALL LETTER GAMMA
       "d" ;; E4  GREEK SMALL LETTER DELTA
       "e" ;; E5  GREEK SMALL LETTER EPSILON
       "z" ;; E6  GREEK SMALL LETTER ZETA
       "_" ;; E7  GREEK SMALL LETTER ETA
       "_" ;; E8  GREEK SMALL LETTER THETA
       "i" ;; E9  GREEK SMALL LETTER IOTA
       "k" ;; EA  GREEK SMALL LETTER KAPPA
       "l" ;; EB  GREEK SMALL LETTER LAMDA
       "m" ;; EC  GREEK SMALL LETTER MU
       "n" ;; ED  GREEK SMALL LETTER NU
       "_" ;; EE  GREEK SMALL LETTER XI
       "_" ;; EF  GREEK SMALL LETTER OMICRON
       "p" ;; F0  GREEK SMALL LETTER PI
       "r" ;; F1  GREEK SMALL LETTER RHO
       "s" ;; F2  GREEK SMALL LETTER FINAL SIGMA
       "s" ;; F3  GREEK SMALL LETTER SIGMA
       "t" ;; F4  GREEK SMALL LETTER TAU
       "_" ;; F5  GREEK SMALL LETTER UPSILON
       "_" ;; F6  GREEK SMALL LETTER PHI
       "_" ;; F7  GREEK SMALL LETTER CHI
       "_" ;; F8  GREEK SMALL LETTER PSI
       "o" ;; F9  GREEK SMALL LETTER OMEGA
       "i" ;; FA  GREEK SMALL LETTER IOTA WITH DIALYTIKA
       "_" ;; FB  GREEK SMALL LETTER UPSILON WITH DIALYTIKA
       "_" ;; FC  GREEK SMALL LETTER OMICRON WITH TONOS
       "_" ;; FD  GREEK SMALL LETTER UPSILON WITH TONOS
       "o" ;; FE  GREEK SMALL LETTER OMEGA WITH TONOS
       "_" ;; FF
       )))

  (or (= 256 (string-length table))
      (error "oops, wrong table length"))

  ;; STR contains ISO-8859-7 bytes, return it with those transliterated to
  ;; ASCII, either close-enough chars or "_"s for untranslatable bits
  (define-toplevel (string-iso-8859-7->ascii str)
    (string-map (lambda (c)
		  (string-ref table (char->integer c)))
		str)))


;;-----------------------------------------------------------------------------
;; download - dividends
;;
;; This uses the dividends page at
;;
;;   http://www.ase.gr/content/en/announcements/dailypress/Daily_Dividends.asp
;;

;; eg. "0.120"
(define (athens-dividend-parse str)
  (define orig str)

  ;; most amounts shown in 3 decimals, but trim to 2 if trailing zeros
  (set! str (string-trim-to-2-decimals str))

  (or (string->dividend-amount str)
      orig))

(define (dividends-process symbol-list body)
  (let* ((offpos (string-contains-ci body "Daily Official List Announcements"))
	 (t1pos  (string-contains-after-ci body "<table" offpos))
	 (row-list (html-table-rows body t1pos)))

    ;; blanks and sub-headings
    (set! row-list (filter! (lambda (row)
			      (list>=? row 2))
			    row-list))

    (let ((headings (first row-list))
	  (want-headings '("Company Name" "Symbol" "Price in &euro;"
			   "Type" "Ex-Dividend Date" "Start Payment Date")))
      (or (equal? headings want-headings)
	  (error "ATHEX: unrecognised dividend columns:" headings))

      ;; drop initial headings, and also second set of headings under
      ;; "Pre-paid Dividends"
      (set! row-list (delete! want-headings (cddr row-list))))

    ;; record dividends, but only for symbols already in the database
    (for-each
     (lambda (row)
       ;; ignore section rows like "ADVERTISING", and empty spacer rows
       (if (list>=? row 2)
	   (receive-list (name symbol amount type ex-date pay-date)
	       row
	     (set! symbol (string-append symbol ".ATH"))
	     (if (member symbol symbol-list)
		 (begin
		   (download-status (_ "ATHEX") (_ "dividends") symbol)

		   (set! symbol-list (delete symbol symbol-list))
		   (set! ex-date (d/m/y-str->tdate ex-date))

		   (c-idle) ;; pause between writes
		   (database-call-modify-data symbol
		     (lambda (series)
		       (series-dividend-add! series ex-date
					     (athens-dividend-parse amount))
		       (series-info-add! series 'athens-dividends-time
					 (current-time)))))))))
     row-list)

    ;; record dividend coverage date for remaining symbols in the database
    (for-each (lambda (symbol)
		(download-status (_ "ATHEX") (_ "dividends") symbol)
		(c-idle) ;; pause between writes
		(database-call-modify-data symbol
		  (lambda (series)
		    (series-info-add! series 'athens-dividends-time
				      (current-time)))))
	      symbol-list)))

(define (dividends-download symbol-list)

  ;; update every 6 days, or if not previously fetched
  (define target-time (- (current-time) (* 6 86400)))
  (define (want-update? symbol)
    (let ((last-time (database-read-info symbol 'athens-dividends-time)))
      (or (not last-time)
	  (< last-time target-time))))

  (if (any want-update? symbol-list)
      (begin
	(download-status (_ "ATHEX") (_ "dividends"))

	(receive (headers body)
	    (http-request "http://www.ase.gr/content/en/announcements/dailypress/Daily_Dividends.asp"
			  #:want-ok #t)
	  (dividends-process (download-also symbol-list
					    #:selector athens-symbol?)
			     body)))))


;;-----------------------------------------------------------------------------
;; download - last 30 days by symbol
;;
;; This uses the prices pages like
;;
;;     http://www.ase.gr/content/en/marketdata/stocks/prices/Share_SearchResults.asp?share=HTO
;;
;; Various places link to those price pages using a "SID" id number, but the
;; symbol works too.
;;
;; There's no ETag or Last-Modified to save re-downloading if our idea of
;; what should be available is a bit out.


(define (last30-process headers body)
  (download-cost-store 'athens-last30 body)

  (let ((charset (html-charset headers body)))
    (or (string-ci=? charset "iso-8859-7")
	(error "ATHEX unrecognised charset:" charset)))

  ;; Eg.
  ;; <table width="100%" border="0" cellpadding="3" cellspacing="0" class="border">
  ;; <TR>
  ;; <TD colspan="9" class="bodyText" bgcolor="#003399" style="border-bottom-style: solid; border-width:1; border-color:#000000">
  ;; <STRONG><font color="#ffffff">
  ;; Share Closing Prices: HTO&nbsp;- &nbsp;
  ;; <a href="/content/en/Companies/ListedCo/Profiles/Profile.asp?name=HTO">HELLENIC TELECOM. ORG.</a>
  ;;
  ;; In the regexp, the first ".*" is greedy so it gives the last "<table"
  ;; before "Share Closing Prices".  That heading (and symbol and name) are
  ;; the first row of the table, but let the regexp pick out the parts.
  ;;
  (let* ((m         (must-match
		     (string-match-C
		      ".*(<table).*Share Closing Prices: ([A-Z]+)[^-]*-[^>]*>([^<]+)</a>" body)))
	 (commodity (match:substring m 2))
	 (symbol    (string-append commodity ".ATH"))
	 ;; some names on the english pages have greek 8859-7 capitals, mung
	 ;; those to plain ascii
	 (name      (string-iso-8859-7->ascii (match:substring m 3)))
	 (row-list  (html-table-rows body (match:start m 1))))

    (let ((headings (second row-list)))
      (or (equal? headings
		  '("Date" "Price" "%Change" "Volume" "Max" "Min" "Value"
		    "Trades" "Open"))
	  (error "ATHEX: unrecognised data columns:" headings)))
    (set! row-list (cddr row-list))

    (download-process
     #:module        (_ "ATHEX")
     #:symbol-list   (list symbol)
     #:currency      "EUR"   ;; what currency ?
     #:last-download #t
     #:row-list
     (map-rows! (lambda (tdate close percent-change volume high low
			       volume-value volume-trades open)
		  (list #:tdate     (d/m/y-str->tdate tdate)
			#:commodity commodity
			#:name      name
			#:open      open
			#:high      high
			#:low       low
			#:close     close
			#:volume    volume))
		row-list))))

(define (last30-download symbol-list)
  (for-each (lambda (symbol)
	      (download-status (_ "ATHEX") (_ "30 days data") symbol)
	      (receive (headers body)
		  (http-request (string-append "http://www.ase.gr/content/en/marketdata/stocks/prices/Share_SearchResults.asp?share="
					       (chart-symbol-sans-dot symbol))
				#:want-ok #t)
		;; message in page if bad symbol
		(if (not (string-contains
			  body "Your search didn't return any results"))
		    (last30-process headers body))))
	    symbol-list))


;;-----------------------------------------------------------------------------
;; download - by whole day files
;;
;; This uses the daily closing prices like
;;
;; http://www.ase.gr/content/en/MarketData/Stocks/Prices/Files/txt/20050318.txt
;;
;; This is about 50kbytes, and that's smaller than the same data available
;; in xls (about 103k) or HTML (about 265k).
;;
;; There's an ETag / Last-Modified on the files, but that's hardly needed
;; with the date in the filename.


;; Lines like the following, fixed width fields, CRLF line endings:
;; AAAK      12/07/2005     6.42    -          75      6.42      5.40           430.50         2    5.40                                                 
;;
;; returned as a list of lists
;; (("AAAK" "12/07/2005" ...) ...)
;;
(define (txt->row-list body)
  (let ((row-list (map! (lambda (row-str)
			  (string-split-widths row-str
					       '(10 10 9 7 10 10 10 17 10 8)))
			(string-split-lines body))))

    (let ((headings (first row-list)))
      (or (equal? headings '("SYMBOL" "DATE" "CLOSE" "B/S" "VOLUME"
			     "HIGH" "LOW" "VALUE" "TRANS" "OPEN"))
	  (error "ATHEX unrecognised daily txt headings:" headings)))
    (set! row-list (cdr row-list))

    row-list))

(define (daily-process symbol-list tdate row-list)
  (download-process
   #:module        (_ "ATHEX")
   #:symbol-list   symbol-list
   #:currency      "EUR"
   #:lo            tdate
   #:hi            tdate
   #:row-list
   (map-rows! (lambda (commodity tdate close b/s volume
				 high low volume-value volume-trades open)
		(list #:tdate     (d/m/y-str->tdate tdate)
		      #:commodity commodity
		      #:open      open
		      #:high      high
		      #:low       low
		      #:close     close
		      #:volume    volume))
	      (list-copy row-list))))

(define (athens-available-tdate)
  (tdate-today-after 18 0 (timezone-athens)))

(define (daily-download start-tdate avail-tdate)
  (define symbol-list '())

  (do ((t start-tdate (1+ t)))
      ((> t avail-tdate))
    (download-status (_ "ATHEX") (_ "data") (tdate-range-string t t))

    (receive (headers body)
	(http-request (daily-url t))
      (let ((status-code (http-headers-status-code headers)))
	(cond ((http-status-code-ok? status-code)
	       (download-cost-store 'athens-daily body)
	       (set! symbol-list (download-also symbol-list
						#:selector    athens-symbol?
						#:start-tdate t
						#:end-tdate   t))
	       (daily-process symbol-list t (txt->row-list body)))

	      ((not (= 404 status-code))
	       ;; 404 if data for date, anything else is bad
	       (http-headers-throw headers)))))))


;;-----------------------------------------------------------------------------
;; download
;;
;; The download chooses between the whole-day update files and individual
;; last 30 pages.
;;
;; The whole day is about 50k and the last 30 page is about 48k, so
;; generally if you've got more than 1 ATHEX symbol then the whole day is
;; best (or more symbols than days to download).
;;
;; A mixture is used too, if there's a few symbols quite a bit behind then
;; they'll be done individually, and the balance with whole-day.


(define (athens-download symbol-list)
  (define avail-tdate (athens-available-tdate))
  (define last30-cost (download-cost-get 'athens-last30 48800))  ;; as at Mar06

  (receive (use-hist start-tdate)
      (download-by-day-or-by-symbol
       avail-tdate symbol-list
       (lambda (tdate) last30-cost)
       (download-cost-get 'athens-daily 51000))  ;; as at Mar06

    (last30-download use-hist)
    (daily-download start-tdate avail-tdate)))

(define (athens-backto want-tdate symbol-list)
  (download-message
   (string-append (_ "ATHEX") ": "
		  (_ "no further historical data available"))))

(download-handler! #:selector   athens-symbol?
		   #:update     (lambda (symbol-list)
				  (athens-download symbol-list)
				  (dividends-download symbol-list))
		   #:avail-proc athens-available-tdate
		   #:backto     athens-backto)


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the daily closing prices like
;;
;; http://www.ase.gr/content/en/MarketData/Stocks/Prices/Files/txt/20050318.txt
;;
;; There's no change column, so the latest day and previous day (skipping
;; public holidays) are compared.  At about 53kbytes each this means 106k
;; total.
;;
;; The same data is available in xls format, but those files at about 103k
;; are bigger than the txt form.  HTML is also available, but at 265k it's
;; even bigger still.


;; PREV-LIST is a list from txt->row-list, or #t to use the database, or #f
;; for no prev data available
;;
;; records for SYMBOL-LIST are always produced, others in LAST-LIST are only
;; produced when prev data is available, ie. either PREV-LIST is a list or
;; it's #t and there's a database record for PREV-ADATE
;;
(define (txt->latest-list symbol-list
			  prev-tdate prev-list last-adate last-list)
  (define ret '())

  (for-each
   (lambda (last-row)
     (receive-list (symbol date close b/s volume
			   high low value trans open)
	 last-row
       (define decimals #f)
       (define prev     #f)

       (cond ((eq? #f prev-list)
	      ;; no prev data at all
	      (set! decimals 0))

	     ((eq? #t prev-list)
	      ;; use database
	      (let ((symbol (string-append symbol ".ATH")))
		(or (and-let* ((series (database-read-series symbol))
			       (       (<= (series-lo series) prev-tdate))
			       (       (<= prev-tdate (series-hi series))))
		      (set! prev (array-ref (series-array
					     series prev-tdate prev-tdate)
					    prev-tdate 3))
		      (set! decimals (series-decimals series))
		      #t)

		    ;; allow missing in database if explicitly in SYMBOL-LIST
		    (and (member symbol symbol-list)
			 (set! decimals 0)))))

	     (else
	      ;; get string value from PREV-LIST
	      (set! prev (any (lambda (prev-row)
				(and (string=? symbol (first prev-row))
				     (third prev-row)))
			      prev-list))
	      (set! decimals (if prev
				 (string-decimals prev)
				 0))))

       ;; skip LAST-LIST records when we're using the database for prev
       ;; values and there's nothing there for this SYMBOL
       (if decimals
	   (begin
	     (set! decimals (max (string-decimals open)
				 (string-decimals high)
				 (string-decimals low)
				 (string-decimals close)))
	     (set! open  (string->number+decimals open  decimals))
	     (set! high  (string->number+decimals high  decimals))
	     (set! low   (string->number+decimals low   decimals))
	     (set! close (string->number+decimals close decimals))
	     (if (string? prev)
		 (set! prev (string->number+decimals prev decimals)))

	     (set! ret
		   (cons (latest-new #:symbol     (string-append symbol ".ATH")
				     #:open       open
				     #:low        low
				     #:high       high
				     #:last       close
				     #:last-adate last-adate
				     #:prev       prev
				     #:volume     volume
				     #:decimals   decimals
				     #:source     'athens) ret))))))
   last-list)
  ret)

(define (daily-url tdate)
  (apply format #f "http://www.ase.gr/content/en/MarketData/Stocks/Prices/Files/txt/~a~2,'0d~2,'0d.txt"
	 (tdate->ymd tdate)))

;; return values (TDATE BODY) of daily txt file on or before START-TDATE
;; (TEST TDATE) is called before doing a download, if it's true then the
;; return is (TDATE #t) instead
;; only 4 attempts are made (START-TDATE through START-TDATE - 3), if
;; there's no download available for any then the return is (#f #f)
;;
(define (latest-txt start-tdate test)
  (let more ((attempt 0))
    (let ((tdate (- start-tdate attempt)))
      (if (test tdate)
	  (values tdate #f)
	  (receive (headers body)
	      (http-request (daily-url tdate))
	    (let ((status-code (http-headers-status-code headers)))
	      (cond ((http-status-code-ok? status-code)
		     (values tdate body))

		    ((= 404 status-code) ;; no data for date
		     (if (< attempt 3)
			 (more (1+ attempt))
			 (values #f #f)))

		    (else ;; anything else is bad
		     (http-headers-throw headers)))))))))

(define (athens-latest-get symbol-list extra-list proc)
  ;; latest data
  (receive (last-tdate last-body)
      (latest-txt (athens-available-tdate) (lambda (tdate) #f))
    (if last-tdate

	;; previous data
	(receive (prev-tdate prev-body)
	    (latest-txt (1- last-tdate)
			;; use database if it covers all in SYMBOL-LIST,
			;; doesn't have to cover EXTRA-LIST
			(lambda (tdate)
			  (every (lambda (symbol)
				   (< tdate (download-start-tdate symbol)))
				 symbol-list)))

	  (let ((prev-list (and prev-body
				(txt->row-list prev-body)))
		(last-list (txt->row-list last-body)))


	    (if prev-list
		(daily-process (download-also '()
					      #:selector    athens-symbol?
					      #:start-tdate prev-tdate
					      #:end-tdate   prev-tdate)
			       prev-tdate prev-list))
	    (daily-process (download-also '()
					  #:selector    athens-symbol?
					  #:start-tdate (if prev-tdate
							    (1+ prev-tdate)
							    last-tdate)
					  #:end-tdate   last-tdate)
			   last-tdate last-list)

	    (proc (txt->latest-list symbol-list
				    prev-tdate prev-list
				    (tdate->adate last-tdate) last-list)))))))

(latest-handler! #:selector   athens-symbol?
		 #:handler    athens-latest-get
		 #:adate-time (lambda (symbol)
				(list (tdate->adate (athens-available-tdate))
				      #f)))
