;; Average true range (ATR).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart atr)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.incrediblecharts.com/technical/average_true_range.htm
;; http://www.incrediblecharts.com/technical/atr_construction.htm
;;
;; http://www.trade2win.com/knowledge/articles/general_articles/average-true-range-indicator/
;;     Sample S&P 500 Sep2005 to Feb2006, seems to be 14-day smoothing.
;;


;; Return a procedure which calculates the "true range" for successive days.
;;
;; Each call (PROC high low close) enters a new day into the window, and the
;; return is the true range for that day, or #f on the first (since there's
;; no previous close yet).
;;
;; HIGH and/or LOW can be #f in each call, in which case the CLOSE is used
;; (CLOSE cannot be #f) and the true range will be just close-to-close
;; changes (as positive values).
;;
(define-public (true-range-calc-proc)
  (let ((prev-close #f))
    (lambda (high low close)
      (set! high (or high close))
      (set! low  (or low  close))
      (let ((tr (and prev-close
		     (- (max high prev-close)
			(min low  prev-close)))))
	(set! prev-close close)
	tr))))


;; Return a procedure which calculates the average true range over an
;; accumulated exponential moving average of COUNT days.
;;
;; Each call (PROC high low close) enters a new day into the window, and the
;; return is the ATR for that day, or #f if not enough points yet.
;;
;; HIGH and/or LOW can be #f in each call, in which case the CLOSE is used.
;; (CLOSE cannot be #f.)
;;
;; An ATR is in theory influenced by all preceding data, but
;; `atr-warmup-count' below is designed to determine a warmup count.
;;
(define-public (atr-calc-proc count)
  (set! count (ema-count-wilder count))
  (let ((tr-proc  (true-range-calc-proc))
	(ema-proc (ema-calc-proc count)))
    (lambda (high low close)
      (and-let* ((tr (tr-proc high low close)))
	(ema-proc tr)))))

(define-public (atr-warmup-count count)
  (1+ (ema-warmup-count (ema-count-wilder count))))


(define (atr-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (warmup (atr-warmup-count count))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo
					   #:count  warmup))
	 (array  (series-array parent start hi))
	 (atr    (atr-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (atr (array-ref array i 1)    ;; high
	       (array-ref array i 2)    ;; low
	       (array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array
		      (atr (array-ref array i 1)  ;; high
			   (array-ref array i 2)  ;; low
			   (array-ref array i 3)) ;; close
		      i)))))

(define-public (atr-series series count)
  (ohlcv-required "ATR" series)
  (let ((new-series (series-copy series)))
    (series-datatype!  new-series atr-datatype)
    (series-name!      new-series (format #f "~a ~a" (_ "ATR") count))
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   atr-fill (atr-warmup-count count)))
    (series-parent!    new-series series)
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public atr-datatype
  (datatype-new!
   #:key     'atr
   #:name    (_ "ATR - Average True Range")
   #:input   datatype-ohlcv
   #:units   'price
   #:minimum 0
   #:manual  (_ "CONTEXT(manual-node): Average True Range")
   #:proc    atr-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'atr-days
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
