;; Australian Stock Exchange (ASX) setups.

;; Copyright 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart australia)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart http) ;; url-quote
  #:use-module (chart misc)
  #:use-module (chart yahoo))


(define-public (australia-symbol? symbol)
  (string-suffix? ".AX" symbol))

;; http://au.finance.yahoo.com/indices
;; for the ^AT symbols must avoid ^ATX in vienna.scm
;;
(define (yahoo-index-symbol-australia? symbol)
  (or (string-prefix? "^AX" symbol) ;; several, inc ^AXJO, ^AXPJ
      (member symbol '("^AORD"      ;; all ords
		       "^ATLI"      ;; asx 20
		       "^AFLI"      ;; asx 50
		       "^ATOI"))))  ;; asx 100

(symbol-timezone! australia-symbol? (_ "Sydney") timezone-sydney)
(symbol-timezone! yahoo-index-symbol-australia? (_ "Sydney") timezone-sydney)
(source-help! (list australia-symbol? yahoo-index-symbol-australia?)
	      (_ "CONTEXT(manual-node): Australian Stock Exchange"))

;; See http://www.asx.com.au/markets/l4/MarketPhases_AM4.shtm
;; SEATS takes changes only between 7am and 7pm
;; 5pm to 7pm is only amend/cancel, no new orders, which is unlikely to be
;; too interesting, but update during that anyway
(yahoo-quote-lock! australia-symbol?
		   #,(hms->seconds 7 0 0) #,(hms->seconds 19 0 0))

;; Yahoo index values are based on last trades, so they only update during
;; trading 10:00 to 16:06.  Not sure if they update for after hours
;; broker-to-broker trades 16:06 to 17:00, allow for that by locking only
;; after 17:00.
(yahoo-quote-lock! yahoo-index-symbol-australia?
		   #,(hms->seconds 10 0 0) #,(hms->seconds 17 0 0))

(yahoo-quote-delay! australia-symbol? 20)


;;-----------------------------------------------------------------------------
;; weblink - ASX company info
;;
;; Eg. http://www.asx.com.au/asx/research/companyInfo.do?by=asxCode&asxCode=ANZ

;; `string-take' but LEN chopped to the length of STR
(define (string-take-maybe str len)
  (string-take str (min len (string-length str))))

(weblink-handler! australia-symbol?
  (lambda (symbol)
    ;; the info search doesn't accept a pref or convertible note symbol like
    ;; "NABHA", only the issuing company "NAB", so prune accordingly
    (set! symbol (string-take-maybe (chart-symbol-sans-dot symbol) 3))

    (list (_ "ASX _Company Information")
	  (_ "Open web browser at the Australian Stock Exchange page for this company")
	  (string-append "http://www.asx.com.au/asx/research/companyInfo.do?by=asxCode&asxCode=" (url-quote symbol)))))


;;-----------------------------------------------------------------------------
;; weblink - ASX interest rate market info
;;
;; Eg. http://www.asx.com.au/investor/irs/irsdescriptionNABHA.htm

;; Alas, these pages seem to have gone away ...
;;
;; ;; `australia-irm-symbol?' returns true if SYMBOL is an Australian Stock
;; ;; Exchange Interest Rate Market stock.  For a full current list see
;; ;; http://www.asx.com.au/asx/markets/displayInterestRateSecurities.do
;; ;;
;; ;; The pattern is a fourth letter
;; ;;     G = convertible bonds
;; ;;     H = corporate bonds and floating rate notes
;; ;;     P = hybrid prefs
;; ;;
;; (define (australia-irm-symbol? symbol)
;;   (regexp-exec #,(regexp "^...[GHP].*\\.AX$") symbol))

;; ;; The company information page doesn't have a good link across to the IRM
;; ;; description page, so having this separately is important.
;; ;;
;; (weblink-handler! australia-irm-symbol?
;;   (lambda (symbol)
;;     (list (_ "ASX IRM _Stock Information")
;; 	  (_ "Open web browser at the Australian Stock Exchange Interest Rate Market information page for this stock")
;; 	  (string-append "http://www.asx.com.au/investor/irs/irsdescription"
;; 			 (chart-symbol-sans-dot symbol) ".htm"))))
