;; barchart.com setups and data downloading.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart barchart)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 regex)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define (liffe-symbol? symbol) (string-suffix? ".LIFFE" symbol))

(define-public barchart-latest-symbol-predlist '())
(define-public barchart-download-symbol-predlist '())

(define (barchart-latest-symbol? symbol)
  (or (any-pred barchart-latest-symbol-predlist symbol)
      (liffe-symbol? symbol)))

(define-public (barchart-download-symbol? symbol)
  (or (any-pred barchart-download-symbol-predlist symbol)
      (liffe-symbol? symbol)))

(source-help! barchart-latest-symbol?
	      (_ "CONTEXT(manual-node): Barchart")
	      -10) ;; below others like WCE

(symbol-timezone! liffe-symbol? (_ "London") timezone-london)


;; (PRED . ALIST) where ALIST is commodities (EXCHANGE . BARCHART)
(define-public barchart-commodity-mung-alist
  (list (cons liffe-symbol?  ;; LIFFE part of euronext ...
	      '(
		;; (? . "CA")     ;; cocoa
		;; (? . "LW")     ;; wheat (405)
		;; (? . "RC")     ;; robusta
		;; (? . "SW")     ;; white sugar
		;; ("FAP" . ?)    ;; potatos
		;; (? . "XR")     ;; rapeseed - paris ?
		;; "Y"            ;; ftse 250
		("Z" . "X")       ;; ftse 100 - lon

		;; MATIF
		("FCE" . "MX")     ;; cac 40 - par
		;; ("FTI" . "")    ;; aex - ams
		;; ("FTL" . "")    ;; aex light - ams
		;; ("BXF" . "")    ;; bel 20 - bru
		;; ("PSI" . "")    ;; psi 20 - lis
		;; (? . "ML")      ;; milling wheat
		;; (? . "XB")      ;; corn
		))

	;; WCE - no transforms
	;; "RS" canola
	;; "WW" feed wheat
	;; "WF" flax
	;; "AB" western barley
	))

(define (symbol->barchart-commodity symbol)
  (let* ((alist     (or (sympred-assoc-ref barchart-commodity-mung-alist
					   symbol)
			'()))
	 (commodity (chart-symbol-commodity symbol)))
    (or (assoc-ref alist commodity)
	commodity)))


;;-----------------------------------------------------------------------------
;; latest
;;
;; The intraday commodity quotes pages are used, like oats
;;
;;     http://www2.barchart.com/ifutpage.asp?code=BSTK&sym=O
;;
;; which is about 35 kbytes each.  An alternative would be the combined
;; pages like all grains
;;
;;     http://www2.barchart.com/mktcom.asp?code=BSTK&section=grains
;;
;; which has the front month or two of various at about 50kbytes the lot.

;; return a substring of STR, with START and END measured relative to the
;; end of the string, with 1 begin the last character
;;
(define (substring-end str start end)
  (substring str
	     (- (string-length str) start)
	     (- (string-length str) end)))

;; convert number like "99-1" with dash fraction to decimals like "99.125"
;; single dash digit is 1/8s
;; three dash digits -xxy is xx 1/32s and y is 0,2,5,7 for further 1/4, 2/4,
;; or 3/4 of 1/32
;;
(define-public (barchart-dash-frac->decimals str)
  (let ((pos (string-rindex str #\-)))
    (if (and pos
	     (>= pos 1))  ;; not a leading "-" sign
	(let ((frac #f))
	  (case (- (string-length str) pos)
	    ((2) ;; 99-1
	     ;; only 2 decimals for 1/4s, since for various commodities
	     ;; that's the minimum tick
	     (set! frac
		   (assv-ref '((#\0 . "00")
			       (#\1 . "125")
			       (#\2 . "25")
			       (#\3 . "375")
			       (#\4 . "50")
			       (#\5 . "625")
			       (#\6 . "75")
			       (#\7 . "875"))
			     (string-ref str (1- (string-length str))))))
	    ((3) ;; 109-30, in 1/32nds
	     (let ((th (string->number-err (string-take-right str 2))))
	       (if (<= th 31)
		   (begin
		     (set! frac (string-drop (number->string
					      (+ 100000
						 (quotient (* 100000 th)
							   32)))
					     1))
		     (set! frac (string-trim-right frac #\0))))))
	    ((4) ;; 99-130
	     ;; TH is 1/32s, QU is further 1/4s of that
	     (let ((th (string->number-err (substring-end str 3 1)))
		   (qu (assv-ref '((#\0 . 0)
				   (#\2 . 1)
				   (#\5 . 2)
				   (#\7 . 3))
				 (string-ref str (1- (string-length str))))))
	       (if (and (<= th 31)
			qu)
		   (let ((ht (+ (ash th 2) qu)))  ;; 1/128s
		     (set! frac (string-drop (number->string
					      (+ 10000000
						 (quotient (* 10000000 ht)
							   128)))
					     1))
		     ;; 1/128 is 0.0078125, so up to 7 decimals, but trim
		     ;; trailing zeros so prices all in 1/32s get just 5
		     (set! frac (string-trim-right frac #\0)))))))
	  (or frac
	      (error "Barchart: unrecognised fraction:" str))
	  (set! str (string-append (string-take str pos)
				   "."
				   frac)))))
  str)

;; barchart.com doesn't say what the delay for LIFFE is, but www.liffe.com
;; says the quotes there are at least 15 minutes, so assume that.
;;
(define-public barchart-suffix-delay-alist '((".LIFFE" . 15)))

;; return the delay in seconds for SYMBOL from barchart
(define (barchart-symbol-delay symbol)
  (* 60 (or (assoc-ref barchart-suffix-delay-alist
		       (chart-symbol-suffix symbol))
	    0)))

;; return list (TDATE TIME) for quote on SYMBOL
;;
;; this includes the exchange delay, and knows the ifut pages only update on
;; the 5 minutes, so rounds back accordingly
;;
;; the new 5-minute time is only returned at 5 seconds past, in case our
;; system time is a bit off
;;
(define (barchart-quote-adate-time symbol)
  (tm->adate-time
   (localtime (round-down-multiple (- (current-time)
				      (barchart-symbol-delay symbol) 5)
				   #,(hms->seconds 0 5 0))
	      (symbol-timezone symbol))))

;; return a list (ADATE TIME) which is the given ADATE/TIME converted from
;; Chicago (US Central) to the timezone of SYMBOL, if that's different
;; TIME can be #f, in which case no conversion is made
;;
(define (adate-time-chicago->symbol symbol adate time)
  (if (and time
	   (not (equal? (timezone-chicago)
			(symbol-timezone symbol))))

      ;; tm from adate/time and isdst for that time
      (let ((tm (adate->tm adate)))
	(set-tm:sec tm time)
	(set-tm:isdst tm (tm:isdst
			  (cdr (mktime tm (timezone-chicago)))))

	;; now convert to symbol zone
	(set! tm (localtime (car (mktime tm (timezone-chicago)))
			    (symbol-timezone symbol)))
	(set! adate (tm->adate tm))
	(set! time  (tm->time tm))))
  (list adate time))

;; return list of `latest' records
;; SYMBOL is used for the commodity and suffix
(define (ifutpage->latest-list symbol body)
  (let* ((commodity   (chart-symbol-commodity symbol))
	 (suffix      (chart-symbol-suffix symbol))
	 ;; look for heading date/time row with "Delayed Futures", the
	 ;; leading ".*" in the regexp ensures we get the last "<TABLE"
	 ;; before that point
	 ;;
	 ;; eg. "   <B>CRUDE OIL</B> Delayed Futures -20:10 - Sunday, 19 June"
	 ;;     "   <B>SIMEX NIKKEI 225</B> Delayed Futures -18:20 - Tuesday, 12 December</td>"
	 ;;     "   <B>OATS   </B> Daily Futures -     Friday, 20 April                 </td>
	 ;;                                                    1         2                                         3                              4
	 (m           (must-match (regexp-exec #,(regexp ".*<B>([^<\r\n]+( *[^<\r\n]+)*) *</B> Delayed Futures *- *([0-9]+:[0-9]+) *- *[A-Za-z]+, ([0-9]+ [A-Za-z]+)" regexp/icase) body)))
	 (name        (match:substring m 1))
	 ;; date in the heading
	 (head-time   (h:m:s-str->seconds (match:substring m 3)))
	 (head-adate  (d/m-str->adate-nearest (match:substring m 4)))

	 (row-list    (html-table-rows body (string-contains body "Go to Daily Price")))
	 (quote-adate head-adate)
	 (quote-time  head-time))

    (let ((headings (first row-list)))
      (or (equal? headings '("Contract" "Last" "Change" "Open" "High" "Low"
			     "Previous" "Time" "Links"))
	  (error "barchart.com ifutpage: unrecognised data columns:"
		 headings)))
    (set! row-list (cdr row-list))

    ;; subtract normal exchange delay to get quote time
    (set! quote-time (- quote-time (barchart-symbol-delay symbol)))
    (if (negative? quote-time)
	(begin
	  (set! quote-time (+ quote-time 86400))
	  (set! quote-adate (1- quote-adate))))

    ;; convert quote adate/time into symbol timezone
    (receive-list (adate time)
	(adate-time-chicago->symbol symbol quote-adate quote-time)
      (set! quote-adate adate)
      (set! quote-time time))

    (latest-list-add-front-months
     (map-rows! (lambda (contract last change open high low prev time link)

		  ;; eg. "August '05 ( CLQ05 )"
		  (let* ((m          (must-match (regexp-exec #,(regexp "^([^ ]+( +[^ ]+)*) *\\( *([^ ]+) *\\)$") contract)))
			 (month-name (match:substring m 1)) ;; "August '05"
			 (bar-symbol (match:substring m 3)) ;; "CLQ05"
			 (mdate      (chart-symbol-mdate bar-symbol))
			 (last-adate #f)
			 (last-time  #f))

		    ;; trailing "s" on last for settlement price
		    ;; have also seen "c", maybe for close
		    (cond ((string-suffix? "s" last)
			   (set! last (string-drop-right last 1)))
			  ((string-suffix? "c" last)
			   (set! last (string-drop-right last 1))))

		    (if (string=? change "unch")
			(set! change "0"))

		    (if (string-any #\: time)
			(begin
			  ;; time is HH:MM on same day as the quote
			  (set! last-adate head-adate)
			  (set! last-time (h:m:s-str->seconds time)))
			(begin
			  ;; time is a date MM/DD/YY later (on the weekend)
			  (set! last-adate (m/d/y-str->adate time))
			  (set! last-time #f)))

		    ;; convert last adate/time into symbol timezone
		    (receive-list (adate time)
			(adate-time-chicago->symbol symbol last-adate last-time)
		      (set! last-adate adate)
		      (set! last-time time))

		    ;; dash is frac in various CBOT
		    (if (string-any #\- last)
			(begin
			  (set! last   (barchart-dash-frac->decimals last))
			  (set! change (barchart-dash-frac->decimals change))
			  (set! open   (barchart-dash-frac->decimals open))
			  (set! high   (barchart-dash-frac->decimals high))
			  (set! low    (barchart-dash-frac->decimals low))))

		    (latest-new
		     ;; our COMMODITY instead of what BAR-SYMBOL has
		     #:symbol  (string-append commodity
					      (mdate->MYY-str mdate)
					      suffix)
		     #:name    (string-append name " " month-name)
		     #:quote-adate    quote-adate
		     #:quote-time     quote-time
		     #:last-adate     last-adate
		     #:last-time      last-time
		     #:open           open
		     #:high           high
		     #:low            low
		     #:last           last
		     #:change         change
		     #:contract-mdate mdate
		     #:source         'barchart)))
		row-list))))

(define (barchart-latest-get symbol-list extra-list proc)
  (for-each (lambda (symbol-list) ;; of same commodity+suffix
	      (receive (headers body)
		  (http-request
		   (string-append
		    "http://www2.barchart.com/ifutpage.asp?code=BSTK&sym="
		    (symbol->barchart-commodity (first symbol-list)))
		   #:want-ok #t)
		;; "does not exist" message in the page when unknown symbol
		(if (not (string-contains body "does not exist"))
		    (let ((latest-list (ifutpage->latest-list
					(first symbol-list) body)))
		      (proc latest-list)
		      (set! symbol-list (lset-difference string=? symbol-list
							 (map latest-symbol
							      latest-list)))))
		;; any not covered by page
		(latest-fetch-fail symbol-list)))

	    (partition-equal chart-symbol-commodity+suffix symbol-list)))

(latest-handler! #:selector barchart-latest-symbol?
		 #:handler  barchart-latest-get
		 #:priority -20)  ;; below yahoo


;;-----------------------------------------------------------------------------
;; intraday
;;
;; There's no direct image url to download, instead a html page like
;;
;;     http://charts.barchart.com/chart.asp?data=Z05&sly=N&sym=CLQ06
;;
;; must be fetched (about 50kbytes unfortunately), and it has a gif url.
;; That url is some generated 4-digit number, apparently different each
;; time, even outside trading hours.  The server doesn't give an etag or
;; last-modified to avoid re-downloading.
;;
;; The form fields are
;;
;;     sym=
;;     date=
;;     size=A      504 by 288
;;          E      576 by 360
;;          B      612 by 360
;;          C      720 by 432
;;          D      864 by 504
;;     data=Z05    minutes
;;          Z10     "
;;          Z15     "
;;          Z30     "
;;          Z45     "
;;          Z60     "
;;          Z90     "
;;          A      daily
;;          D      weekly
;;          G      monthly
;;     den=HIGH    density
;;         MEDHI
;;         MED     [default]
;;         MEDLO
;;         LOW
;;         this combines with data period for how much is shown
;;                  5min    10min   15min
;;         HIGH    3.5day  7day    10day  ... etc
;;         MEDHI   2day    4.5day  7day
;;         MED     1.5day  3day    4day
;;         MEDLO   1day    2day    3day
;;         LOW     0.5day  1.5day  2day
;;     evnt=ADV    events
;;          off
;;     grid=Y/N    background green grid
;;     sky=Y/N     fine grid lines
;;     jav=ADV     prices on, gif file
;;                 [other options for subscribers]
;;     size=A      504 by 288
;;          E      576 by 360
;;          B      612 by 360
;;          C      720 by 432
;;          D      864 by 504
;;     sly=N       linear
;;         L       log
;;         Y       fit available space
;;     late=
;;     ch1=011     OHLC [default]
;;         012     close-only
;;         013     candlestick
;;         ...
;;     ov1=
;;     ch2=
;;     ov2=
;;     code=BSTKIC IC for interactive chart
;;     vol=y/n     volume (not avail for 5min)


(define mode-seven (_ "_7 Days"))
(define mode-daily (_ "_Daily"))

(define-public (barchart-intraday-url symbol mode)
  (and-let* ((mdate (latest-symbol-mdate
		     symbol (adate->tdate
			     (first (barchart-quote-adate-time symbol))))))
    (receive (headers body)
	(http-request (string-append
		       "http://charts.barchart.com/chart.asp?sym="
		       (symbol->barchart-commodity symbol)
		       (mdate->MYY-str mdate)
		       (cond ((string=? mode mode-daily)
			      "&data=A")
			     ((string=? mode mode-seven)
			      "&data=Z60&den=MEDHIGH")
			     (else ;; 1 1/2 Days
			      ;; 5 minute, linear scale
			      "&data=Z05"))))

      (and-let* ((  (http-headers-ok? headers))
		 (m (must-match
		     (string-match "<img src=\"(http://charts.barchart.com/custom/stocks/[^\"]*)\"" body))))
	(match:substring m 1)))))

(define-public (barchart-intraday-modes)
  (list (_ "_1 1/2 Days")
	mode-seven
	mode-daily))

(define (wce-symbol? symbol)
  (string-suffix? ".WCE" symbol))

(intraday-url-proc! #:symbol-pred (lambda (symbol)
				    (or (barchart-download-symbol? symbol)
					(wce-symbol?  symbol)))
		    #:url-proc    barchart-intraday-url
		    #:modes-proc  barchart-intraday-modes)


;;-----------------------------------------------------------------------------
;; download - 5 day end-of-day
;;
;; This uses the rolling 5-day quote pages like
;;
;;     http://quote.barchart.com/quote.asp?sym=CLM08
;;
;; which has daily open, high, low and close, but no volume figure.
;;
;; Going back only five days isn't good for much, but it's better than
;; nothing and regular updates accumulate enough for a short-term picture.


;; process BODY into the database for SYMBOL
;; MDATE is the month of the data in BODY
(define (five-process symbol-list commodity mdate body)
  (let* ((m          (must-match
		      (string-match "<br /><p align=\"center\" class=\"bcText\">[^\r\n]*<b>(([^\r\n]*)[A-Za-z]+ [0-9]+)</b>" body)))
	 ;; month/year included when specific symbol
	 (name       (match:substring m (if mdate 1 2)))

	 ;; headings and latest day in one table, then following four days
	 ;; in another
	 (table1-pos (string-contains-ci body "<table width=\"100%\""
					 (match:end m)))
	 (row-list   (append (html-table-rows body table1-pos)
			     (html-table-rows body (1+ table1-pos)))))

    (let ((headings (first row-list)))
      (or (equal? headings '("Date" "Open" "High" "Low" "Last" "Change"))
	  (error "Barchart 5-day data unrecognised headings:" headings)))
    (set! row-list (cdr row-list))

    ;; message in table when nothing info for given symbol (eg. an old
    ;; month/year)
    (if (string-prefix? "In order to form an extended quote"
			(first (first row-list)))
	(set! row-list '()))

    (download-process
     #:module        (_ "Barchart")
     #:symbol-list   symbol-list
     #:name          name
     #:last-download #t
     #:row-list
     (map-rows!
      (lambda (tdate open high low last change)

	;; dash for eighths in various agriculturals, eg. O.CBOT
	(set! open (barchart-dash-frac->decimals open))
	(set! high (barchart-dash-frac->decimals high))
	(set! low  (barchart-dash-frac->decimals low))
	(set! last (barchart-dash-frac->decimals last))

	(list #:tdate     (m/d/y-str->tdate tdate)
	      #:mdate     mdate
	      #:commodity commodity
	      #:open      open
	      #:high      high
	      #:low       low
	      #:close     last))

      row-list))))

;; latest data available from barchart 5-day quote page
;; the support page says the daily pages update at 5pm US central, try 5pm
;; local for the 5-day
;;
(define-public (barchart-five-available-tdate symbol)
  (tdate-today-after 17 0 (symbol-timezone symbol)))

;; for symbol like "NKZ05.SIMEX", return barchart symbol like "NXZ05"
(define (symbol->barchart-symbol symbol)
  (string-append (symbol->barchart-commodity symbol)
		 ;; month/year part of SYMBOL
		 (substring symbol
			    (string-length (chart-symbol-commodity symbol))
			    (- (string-length symbol)
			       (string-length (chart-symbol-suffix symbol))))))


(define (barchart-five-download-one symbol-list commodity mdate)
  (define specific (string-append commodity
				  (mdate->MYY-str mdate)
				  (chart-symbol-suffix (first symbol-list))))
  (download-status (_ "Barchart") (_ "data") specific)
  (receive (headers body)
      (http-request (string-append "http://quote.barchart.com/quote.asp?sym="
				   (symbol->barchart-symbol specific))
		    #:follow #f)
    (let ((status-code (http-headers-status-code headers)))
      (cond ((http-status-code-ok? status-code)
	     (five-process symbol-list commodity mdate body)
	     #t)
	    ((= 302 status-code)
	     ;; "Moved Temporarily" for unknown symbol or month/year
	     #f)
	    (else
	     ;; anything else is bad
	     (http-headers-throw headers))))))

(define-public (barchart-five-download symbol-list)
  (futures-download-by-specific
   symbol-list
   #:download-proc    barchart-five-download-one
   #:avail-tdate-proc barchart-five-available-tdate
   ;; only expecting 5 days data, pretend 10 just in case we're a little late
   #:initial-proc     (lambda (symbol) 10)))

(define-public (barchart-backto want-tdate symbol-list)
  (download-message
   (string-append (_ "Barchart") ": "
		  (_ "no further historical data available"))))

(download-handler! #:selector barchart-download-symbol?
		   #:update   barchart-five-download
		   #:backto   barchart-backto)
