;; Bendigo Stock Exchange (BSX) setups.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart bendigo)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart pagebits)
  #:use-module (chart misc))


(define-public (bendigo-symbol? symbol)
  (string-suffix? ".BEN" symbol))

;; Victoria is the same as NSW these days, but let's not assume this rare
;; outbreak of interstate cooperation will be permanent (in the past the
;; dates to begin or end daylight savings have varied)
;;
(define-public timezone-melbourne
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Melbourne")
		      '("Australia/Melbourne"
			;; DST first Sun in Oct, to first Sun in Apr, for
			;; 2008 onwards
			"EST-10EDT,M10.1.0,M4.1.0/3")
		      "EST-10"))))

(symbol-timezone! bendigo-symbol? (_ "Bendigo") timezone-melbourne)
(source-help! bendigo-symbol?
	      (_ "CONTEXT(manual-node): Bendigo Stock Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - company info

(weblink-handler! bendigo-symbol?
  (lambda (symbol)
    (list (_ "BSX _Company Information")
	  (_ "Open web browser at the Bendigo Stock Exchange information page for this company")
	  (lambda ()
	    (and-let* ((option (or (symbol->option symbol #f)
				   (begin
				     (weblink-message
				      (_ "Fetching BSX symbol codes ..."))
				     (symbol->option symbol #t)))))
	      (string-append
	       "http://www.bsx.com.au/markets_pricesresearch_com.asp?security="
	       option))))))


;;-----------------------------------------------------------------------------
;; misc

;; Partition LST into groups of elements according to what (PROC elem) returns.
;; The returns from PROC are compared with `equal?'.
;; Equal elements must be adjacent, there's no re-ordering of LST.
;; Each group becomes a list of elements, and a list of all those lists is
;; returned.
;;
(define (partition-equal-adjacent proc lst)
  (let more ((lst lst)
	     (ret '()))
    (if (null? lst)
	(reverse! ret)

	(let ((val (proc (first lst))))
	  (receive (this-list tail-list)
	      (span (lambda (elem)
		      (equal? val (proc elem)))
		    lst)
	    (more tail-list
		  (cons this-list ret)))))))


;;-----------------------------------------------------------------------------
;; option codes
;;
;; The home page
;;
;;     http://www.bsx.com.au
;;
;; has a pull-down menu of symbols, with the code numbers to be used for
;; quotes and trade history.
;;
;; The cache here saves the code numbers for one day (the current trading
;; day, bendigo time).  This ensures we don't download every time, but still
;; have a fresh set of codes, in case there's additions or perhaps changes.
;;
;; Thought was given to a more sophisticated caching scheme, for instance
;; use an old code and if the quote turns out to be for something else and
;; the cache isn't fresh then refill it and get the quote again.  But for
;; the data download it'd be a case of checking the name against the
;; database name, since there's no symbol on that page, except that for the
;; initial download when there's no database name yet the cache would have
;; to be pre-freshened.  This seems unnecessarily complicated to merely
;; avoid downloading one page, especially when it's compressed to about
;; 3kbytes if we've got gzip or zlib.


;; return options alist ((SYMBOL . CODE-STRING) ...)
(define (option-codes-parse body)
  (map! (lambda (m)
	  (cons (string-append (match:substring m 2) ".BEN")
		(match:substring m 1)))
	(must-match
	 (list-matches "<option value=\"([0-9]+)\">([A-Z]+)</option>" body))))

;; return option code string, or #f if unknown
(define* (symbol->option symbol #:optional (want-download #t))
  (assoc-ref (pagebits-read #:filename "bendigo-codes"
			    #:url      "http://www.bsx.com.au"
			    #:timezone (timezone-melbourne)
			    #:parse    option-codes-parse
			    #:want-download want-download)
	     symbol))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the quotes pages like
;;
;;     http://www.bsx.com.au/markets_pricesresearch_pri.asp?security=20
;;
;; with the security option code from the home page menu (cached above).
;;
;; www.bsx.com.au sends gzipped data, which is good since it compresses
;; about 19kbytes of javascript junk down to about 3kbytes sent.
;;
;; The trade history pages like
;;
;;     http://www.bsx.com.au/markets_pricesresearch_tra.asp?security=1
;;
;; also have a current bid/offer and show the date/time of the last trade.
;; Not sure if there's much value showing open/high/low/volume from trading
;; that could be weeks ago.


;; return list (ADATE TIME) for latest quote
;;
;; the BSX system takes changes only during 9:30am-11:30am weekdays (order
;; entry 9:30 to 11, then trading 11 to 11:30), so outside those hours lock
;; to 11:30am
;;
;; http://www.bsx.com.au/markets_aboutbsxmarkets_tra.asp
;;
(define (bendigo-quote-adate-time . dummy-symbol)
  (tm->adate-time-within (localtime (current-time) (timezone-melbourne))
			 #,(hms->seconds 9 30 0)
			 #,(hms->seconds 11 30 0)))

(define (body->latest-list body)
  ;; second <table> after LISTING SELECTOR ENDS
  (define row-list (html-table-rows-raw
		    body
		    (string-contains-after-ci
		     body "<table"
		     (string-contains body "LISTING SELECTOR ENDS"))))

  (let ((spacer (first (first row-list))))
    (or (string-contains spacer "spacer.gif")
	(error "BSX unrecognised quote format")))
  (set! row-list (cdr row-list))

  (set! row-list (html-rows-text-extract row-list))

  (receive-list (quote-adate quote-time)
      (bendigo-quote-adate-time)

    ;; there's only one row, but use `map' since we've got a list and want
    ;; to return one too
    (map-rows! (lambda (name+symbol bid offer last)
		 (let* ((m      (must-match (regexp-exec #,(regexp "([^ ]+( +[^ ]+)*) *\\(([^)]+)\\)$") name+symbol)))
			(name   (match:substring m 1))
			(symbol (match:substring m 3))

			;; eg: <span class="Blue"><br><br><b>Capilano Honey Limited</b><br><font color=red>TRADING HALT</font><br></span><br>
			(m      (regexp-exec #,(regexp "Blue.*<b>.*red[^>]*>([^<]+)" regexp/icase regexp/newline) body))
			(note   (and m (match:substring m 1))))

		   (latest-new #:symbol      (string-append symbol ".BEN")
			       #:name        name
			       #:quote-adate quote-adate
			       #:quote-time  quote-time
			       #:bid         bid
			       #:offer       offer
			       #:last        last
			       #:note        note
			       #:source      'bendigo)))
	       row-list)))

(define (bendigo-latest-get symbol-list extra-list proc)
  (for-each (lambda (symbol)
	      (define option (symbol->option symbol))

	      (if (not option)
		  (proc (list (latest-new-unknown symbol)))

		  (receive (headers body)
		      (http-request (string-append "http://www.bsx.com.au/markets_pricesresearch_pri.asp?security=" option)
				    #:want-ok #t)
		    (proc (body->latest-list body)))))

	    symbol-list))

(latest-handler! #:selector   bendigo-symbol?  
		 #:handler    bendigo-latest-get
		 #:adate-time bendigo-quote-adate-time)



;;-----------------------------------------------------------------------------
;; download
;;
;; This uses the "view all" trade history pages like
;;
;;     http://www.bsx.com.au/markets_pricesresearch_tra_popup.asp?security=20
;;
;; with the security number from the home page menu (cached above).
;;
;; This history has each trade individually, but we collapse that to daily
;; open/high/low/close.


;; Eg:
;; <span class="Blue"><b>Capilano Honey Limited</b></span><br>
;; <span class="Blue"><br>Trade History<br></span><br>
;; <table ...
;;
;; Or when ex dividend:
;; <span class="Blue"><b>Capilano Honey Limited</b><br><font color=red>XD</font></span><br>
;; <span class="Blue"><br>Trade History<br></span><br>
;;
;; Or when halted:
;; <span class="Blue"><b>Capilano Honey Limited</b><br><font color=red>TRADING HALT</font></span><br>
;; <span class="Blue"><br>Trade History<br></span><br>
;;
;; match 1 is company name
;;
(define history-regexp
  "<b>([^<]+)[^\n]*\n[^\n]*<br>Trade History")

(define (bendigo-process-download symbol url headers body)
  (define commodity (chart-symbol-sans-dot symbol))

  (let* ((m        (must-match (string-match history-regexp body)))
	 (name     (match:substring m 1))
	 (row-list (html-table-rows body
				    ;; skip empty table, the second is wanted
				    (string-contains-after-ci body "<table"
							      (match:end m)))))

    (let ((headings (map! string-trim-right (first row-list))))
      (or (equal? headings '("Qty" "Price" "Date" "Time"))
	  (error "BSX: unrecognised data columns:" headings)))
    (set! row-list (cdr row-list))

    (download-process
     #:module          (_ "BSX")
     #:symbol-list     (list symbol)
     #:name            name
     #:url             url
     #:headers         headers
     #:currency        "AUD"
     #:hi              (bendigo-available-tdate)
     #:last-download   #t
     #:prefer-decimals 2
     #:row-list
     (map (lambda (row-list)
	    (receive-list (volume-1 price-1 date time-1)
		(first row-list)
	      ;; most recent first in list, so reverse for sessions
	      (list #:tdate     (d/m/y-str->tdate date)
		    #:commodity commodity
		    #:sessions  (reverse! (map second row-list))
		    #:volume    (apply + (map! string->number-err
					       (map first row-list))))))
	  (partition-equal-adjacent third row-list)))))

;; latest download data available
;; guess today available from 6pm
;;
(define (bendigo-available-tdate)
  (tdate-today-after 18 0 (timezone-melbourne)))

(define (bendigo-download symbol-list)
  (for-each
   (lambda (symbol)
     (download-status (_ "BSX") (_ "data") symbol)

     (let ((option (symbol->option symbol)))
       (if (not option)
	   (download-message
	    (string-append (_ "BSX") ": " (_ "unknown symbol") " " symbol))

	   (let ((url (string-append "http://www.bsx.com.au/markets_pricesresearch_tra_popup.asp?security=" option)))
	     (receive (headers body)
		 (http-request url #:want-ok #t)
	       (bendigo-process-download symbol url headers body))))))
   symbol-list))

(define (bendigo-backto want-tdate symbol-list)
  (download-message
   (string-append (_ "BSX") ": " (_ "no further historical data available"))))

(download-handler! #:selector   bendigo-symbol?
		   #:update     bendigo-download
		   #:avail-proc bendigo-available-tdate
		   #:backto     bendigo-backto)
