;; Bollinger bands.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart bollinger)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a simple moving average and standard
;; deviation over an accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is (values SMA STDDEV) which are the calculated results up to (and
;; including) that VAL, or #f if not enough points yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (sma/stddev-calc-proc count)
  
  ;; If the inputs X are all exact then TOTAL and TOTAL-SQUARES are exact
  ;; and the only rounding is at the final sqrt and division.  But when
  ;; they're inexact the totals can round a little and there's a danger of
  ;; the subtraction making a negative sqrt, hence the max 0 there.
  ;;
  (let ((data          (make-circular-list count #f))
	(total         0)
	(total-squares 0)
	(points        0)
	(min-points    (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (x)

      ;; drop oldest point
      (if (car data)
	  ;; drop oldest point
	  (begin
	    (set! total         (- total (car data)))
	    (set! total-squares (- total-squares (* (car data) (car data)))))
	  ;; introducing new point
	  (set! points (1+ points)))

      ;; replace oldest with new x
      (set-car! data x)
      (set! data (cdr data))

      ;; add new x
      (set! total         (+ total x))
      (set! total-squares (+ total-squares (* x x)))

      (values (and (>= points min-points)   ;; sma
		   (exact->inexact (/ total points)))

	      (and (>= points min-points)   ;; stddev
		   (/ (sqrt (max 0 (- (* total-squares points)
				      (square total))))
		      points))))))


(define (bollinger-fill series new-array lo hi)
  (receive-list (count stddev-factor)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (sma/stddev-calc-proc count)))

      ;; warmup
      (for-each proc (series-prev-n-closes parent
					   #:before lo
					   #:count  (1- count)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (receive (sma stddev)
		(proc (array-ref array i))
	      (if sma
		  (begin
		    (set! stddev (* stddev-factor stddev))
		    (array-set! new-array (- sma stddev) i 0)
		    (array-set! new-array sma            i 1)
		    (array-set! new-array (+ sma stddev) i 2)))))))))

(define-public (bollinger-series series count stddev-factor)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f (_ "Bollinger ~a, ~a")
					 count (locale-number-min-decimals
						stddev-factor 1)))
    (series-datatype! new-series bollinger-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count stddev-factor))
    (series-require!  new-series (indicator-require-parts bollinger-fill
							  (1- count)
							  '(0 2)))
    (series-array!    new-series #f)

    new-series))

(define-public bollinger-datatype
  (datatype-new!
   #:key       'bollinger
   #:name      (_ "Bollinger Bands")
   #:units     #f
   #:linestyle 'bands
   #:manual    (_ "CONTEXT(manual-node): Bollinger Bands")
   #:proc      bollinger-series
   #:closes-apply-arg '(1)  ;; middle MA line
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "Days")
	  #:key      'bollinger-days
	  #:type     'integer
	  #:default  20
	  #:minimum  1)
	 (datatype-param-new
	  #:name     (_ "Stddevs")
	  #:key      'stddevs
	  #:type     'float
	  #:decimals 2
	  #:default  1
	  #:step     0.5
	  #:minimum  0
	  #:maximum  9))))
