;; Casablanca Stock Exchange setups.   -*- coding: latin-1 -*-

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart casablanca)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart pagebits))


(define-public (casablanca-symbol? symbol)
  (string-suffix? ".CAS" symbol))

;; `timezone-choose' can't handle named zones which are GMT, like Casablanca
(define-public timezone-casablanca
  (make-procedure-with-cache-0
   (lambda ()
     "Africa/Casablanca"

     ;; no daylight savings post 1978
     ;; "WET+0"
     )))

(symbol-timezone! casablanca-symbol? (_ "Casablanca") timezone-casablanca)
(source-help! casablanca-symbol?
	      (_ "CONTEXT(manual-node): Casablanca Stock Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - company info
;;
;; Eg. in French and English
;;     http://www.casablanca-bourse.com/cgi/ASP/Fiches/fiche.asp?TICKER=NEJ
;;     http://www.casablanca-bourse.com/cgi/ASP/Fiches/anglais/fiche.asp?TICKER=NEJ

(weblink-handler! casablanca-symbol?
  (lambda (symbol)
    (list (_ "Casablanca _Company Information")
	  (_ "Open web browser at the Casablanca Stock Exchange page for this company")
	  (string-append "http://www.casablanca-bourse.com/cgi/ASP/Fiches"
			 (lang-select '(("fr" "")
					("en" "/anglais")))
			 "/fiche.asp?TICKER="
			 (chart-symbol-sans-dot symbol)))))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the "key indicators" stock data pages (about 9 kbytes each) like
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/Donnees_Valeur/anglais/Donnes_valeurs.asp?ticker_valeur=NEJ
;;
;; Similar information is on the stocks-by-sector page
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/Marche_Central/sectors_en.asp
;;
;; but there's no bid/offer there, so the individual pages are used.
;;
;; The "Reference price" shown on both seems to be the previous close.  The
;; volume is a dollar volume, "Quantity exchanged" is the number of shares
;; (we use the latter).
;;
;; The individual data pages don't show the quantity, so we derive something
;; close enough from the dollar volume and last price.
;;
;; The ex-dividend date field is empty when no dividends, eg. ACRED
;;     http://www.casablanca-bourse.com/cgi/ASP/Donnees_Valeur/anglais/Donnes_valeurs.asp?ticker_valeur=ACR
;;


(define char-set:whitespace+latin1nbsp
  (char-set-adjoin char-set:whitespace (integer->char #xA0)))

;; return list of positions in STR which are the non-overlapping occurrances
;; of PART
(define (list-string-contains-ci str part)
  (let more ((upto 0)
	     (ret  '()))
    (let ((pos (string-contains-ci str part upto)))
      (if pos
	  (more (+ pos (string-length part))
		(cons pos ret))
	  (reverse! ret)))))

;; turn LST like (A B C D) into pairs like ((A . B) (C . D))
(define (lst->pairs lst)
  (let more ((lst lst)
	     (ret '()))
    (if (length>=? lst 2)
	(more (cddr lst)
	      (acons (first lst) (second lst) ret))
	(reverse! ret))))

(define (donnes-body->latest symbol headers body)

  ;; actually there's no meta charset as of Apr 2007, the default latin-1
  ;; from html-charset is right though
  (define charset (html-charset headers body))

  (receive-list (quote-adate quote-time)
      (casablanca-quote-adate-time symbol)

    ;; Pick out the name, eg.
    ;; <center>
    ;;   <p>&nbsp;</p>
    ;;   <p><font color="#004496" face="Verdana" size="-1"><b>AUTO NEJMA                                        </b></font><br>
    (let* ((m        (must-match
		      (regexp-exec-C
		       #,(regexp "<[bB]>([A-Z][^<\r\n]+)")
		       body)))
	   (name     (charset->utf8 charset
				    (string-trim-right (match:substring m 1))))
	   (row-list (append-map (lambda (pos)
				   (html-table-rows body pos))
				 (list-string-contains-ci body "<table")))
	   (alist    (append-map lst->pairs row-list)))

      (define (get field)
	(let ((value (or (assoc-ref alist field)
			 (error "Casablanca field missing:" field))))
	  ;; latin1 0xA0 nbsp in volume and big prices
	  (string-delete value char-set:whitespace+latin1nbsp)))

      ;; spaces and CRLF in fields, drop that to make field name comparisons
      ;; easier
      (for-each (lambda (pair)
		  (set-car! pair (string-delete (car pair)
						char-set:whitespace)))
		alist)

      (let ((current       (get "Currentprice:"))
	    (open          (get "Opening:"))
	    (prev          (get "Referenceprice:"))
	    (high          (get "Today'shigh:"))
	    (low           (get "Today'slow:"))
	    (bid           (get "Bestbuyer'sask:"))
	    (offer         (get "Bestseller'sbid:"))
	    (volume-dollar (get "Volume:"))
	    (div-amount    (get "Dividend:"))
	    (ex-div        (get "Ex-datedividend:"))
	    (volume        #f)
	    (note          #f))

	;; commas for decimal point, even on english page
	(set! current       (string-comma->dot current))
	(set! open          (string-comma->dot open))
	(set! prev          (string-comma->dot prev))
	(set! high          (string-comma->dot high))
	(set! low           (string-comma->dot low))
	(set! offer         (string-comma->dot offer))
	(set! bid           (string-comma->dot bid))
	(set! volume-dollar (string-comma->dot volume-dollar))

	;; ex-dividend date field empty when no dividend
	(if (not (string-null? ex-div))
	    (let ((ex-tdate   (d/m/y-str->tdate ex-div))
		  (last-tdate (adate->tdate quote-adate)))
	      (if (= ex-tdate last-tdate)
		  (set! note (format #f (_ "ex ~a") div-amount)))))

	;; volume-dollar/current-price is not right for the volume of
	;; course, since it isn't all at the last price, but it should be
	;; close
	;;
	;; volume-dollar seems to have a factor of 2 in it, compared to
	;; QUANTITE_ECHANGE in the historical data, dunno why, but strip
	;; it out
	;;
	(and-let* ((dvol  (string->number volume-dollar))
		   (price (string->number current)))
	  (set! volume (round-exact (* 0.5 (/ dvol price)))))

	(latest-new #:symbol      symbol
		    #:name        name
		    #:quote-adate quote-adate
		    #:quote-time  quote-time
		    #:bid         bid
		    #:offer       offer
		    #:last-adate  quote-adate
		    #:open        open
		    #:high        high
		    #:low         low
		    #:last        current
		    #:prev        prev
		    #:volume      volume
		    #:note        note
		    #:source      'casablanca)))))

(define (casablanca-latest-get symbol-list extra-list proc)
  (for-each (lambda (symbol)
	      (receive (headers body)
		  (http-request (string-append "http://www.casablanca-bourse.com/cgi/ASP/Donnees_Valeur/anglais/Donnes_valeurs.asp?ticker_valeur="
					       (chart-symbol-sans-dot symbol))
				#:want-ok #t)
		(proc (list (donnes-body->latest symbol headers body)))))
	    symbol-list))

;; return current Casablanca adate/time, with 15-minute delay
;; pre-open 9:00am through to closure 15:30
;;
(define (casablanca-quote-adate-time symbol)
  (tm->adate-time-within (localtime (- (current-time) #,(hms->seconds 0 15 0))
				    (timezone-casablanca))
			 #,(hms->seconds 9 0 0)
			 #,(hms->seconds 15 30 0)))

(latest-handler! #:selector   casablanca-symbol?
		 #:handler    casablanca-latest-get
		 #:adate-time casablanca-quote-adate-time)


;;-----------------------------------------------------------------------------
;; option codes
;;
;; The historical data page
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/histo_valeurs/anglais/histo_valeur.asp
;;
;; has a pull-down menu of symbols, with the "valeur" code number to be used
;; for historical data for a symbol.
;;
;; The cache here essentially only saves the code numbers for one day (the
;; current trading day, Casablanca time).  This ensures we don't download
;; every time, but also ensures we're on a fresh set of codes, in case
;; there's additions or perhaps changes.
;;
;; Refreshing the cache every day seems best, since the actual data file
;; doesn't have a symbol, which means if the code number changes we won't
;; detect it.


;; return valeur alist ((SYMBOL . CODE-STRING) ...)
(define (histo-valeur-parse body)
  (map! (lambda (m)
	  (cons (string-append (match:substring m 2) ".CAS")
		(match:substring m 1)))

	;; eg: <option value="001046">ACR  </option>
	;; symbols are all alphabetic currently, but allow numbers
	;; use "-C" since there's some latin-1 chars in the page,
	;; though none in the bits we extract
	(must-match
	 (list-matches-C "<option value=\"([0-9]+)\">([A-Z0-9]+) *</option>"
			 body))))

;; return valeur code as a string, or #f if unknown
(define (symbol->valeur symbol)
  (assoc-ref (pagebits-read #:filename "casablanca-codes"
			    #:status   (list (_ "Casablanca")
					     (_ "symbol code numbers"))
			    #:url "http://www.casablanca-bourse.com/cgi/ASP/histo_valeurs/histo_valeur.asp"
			    #:timezone (timezone-casablanca)
			    #:parse    histo-valeur-parse)
	     symbol))


;;-----------------------------------------------------------------------------
;; download - data
;;
;; This uses the historical data download at (normally shown in a frame),
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/histo_valeurs/anglais/histo_valeur.asp
;;
;; which ends up with a link
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/histo_valeurs/anglais/TelechargerDonnees.asp?code_valeur=001076&D1=01/01/2007&D2=16/04/2007
;;
;; or from the French pages
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/histo_valeurs/TelechargerDonnees.asp?code_valeur=001076&D1=01/01/2007&D2=16/04/2007
;;
;; The headings are in French, even from the English link.
;;
;; The "code_valeur" for the symbol is extracted from the download page with
;; the symbol->valeur code above.
;;

(define (casablanca-download-process symbol lo hi body)
  (define commodity (chart-symbol-commodity symbol))

  (define row-list (map! (lambda (row-str)
			   (string-split row-str #\tab))
			 (string-split-lines body)))
  (let ((headings (first row-list)))
    (or (equal? headings '("SEANCE" "COURS_CLOTURE" "EVOLUTION" "QUANTITE_ECHANGE" "VOLUME"))
	(error "Casablanca data unrecognised headings:" headings)))
  (set! row-list (cdr row-list))

  (download-process
   #:module        (_ "Casablanca")
   #:symbol-list   (list symbol)
   #:currency      "MAD"  ;; Moroccan Dirham
   #:lo            lo
   #:hi            hi
   #:row-list
   (map-rows! (lambda (tdate close change volume volume-dollar)
		(list #:tdate     (d/m/y-str->tdate tdate)
		      #:commodity commodity
		      #:close     close
		      #:volume    volume))
	      row-list)))

;; latest download data available
;; today is available from 6pm casablanca time (weekdays)
;;
(define (casablanca-available-tdate)
  (tdate-today-after 18 0 (timezone-casablanca)))

(define (casablanca-download-url valeur start-tdate end-tdate)
  (receive-list (s-year s-month s-day)
      (tdate->ymd start-tdate)
    (receive-list (e-year e-month e-day)
	(tdate->ymd end-tdate)
      (format #f "http://www.casablanca-bourse.com/cgi/ASP/histo_valeurs/TelechargerDonnees.asp?code_valeur=~a&D1=~a/~a/~a&D2=~a/~a/~a"
		     valeur
		     s-day s-month s-year
		     e-day e-month e-year))))

(define (casablanca-download-range symbol lo hi)
  (let ((valeur (symbol->valeur symbol)))
    (if (not valeur)
	(download-message
	 (string-append (_ "Casablanca") ": " (_ "unknown symbol") " " symbol))

	(begin
	  (download-status (_ "Casablanca") (_ "data")
			   symbol (tdate-range-string lo hi))
	  (receive (headers body)
	      (http-request (casablanca-download-url valeur lo hi)
			    #:want-ok #t)
	    (casablanca-download-process symbol lo hi body))))))

(define (casablanca-download symbol-list)
  (for-each (lambda (symbol)
	      (casablanca-download-range symbol
					 (download-start-tdate symbol)
					 (tdate-today (timezone-casablanca))))
	    symbol-list))

(define (casablanca-backto want-tdate symbol-list)
  (for-each (lambda (symbol)
	      (casablanca-download-range symbol
					 want-tdate
					 (download-back-tdate symbol)))
	    symbol-list))


;;-----------------------------------------------------------------------------
;; download - names
;;
;; This uses the list of ticker symbols which comes up if you don't enter
;; one for the "All About" button on the home page,
;;
;;     http://www.casablanca-bourse.com/cgi/ASP/pts/pts_en.asp
;;
;; There's also names on the all-quotes page, or the individual symbol data
;; quote pages, but the names in this list seem a little fuller.


(define (process-names symbol-list body)
  ;; no charset as a html meta or in the http headers, but it's latin1
  (define charset "ISO-8859-1")

  (define row-list (html-table-rows body (string-contains body "Choose")))

  ;; return utf8 name for SYMBOL, or #f
  (define (find-name symbol)
    (set! symbol (chart-symbol-sans-dot symbol))
    (any (lambda (row)
	   (and (equal? (first row) symbol)
		(charset->utf8 charset (second row))))
	 row-list))

  (for-each (lambda (symbol)
	      (download-status (_ "Casablanca") (_ "name") symbol)
  	      (database-call-modify-data symbol
  		(lambda (series)
  		  (series-name! series (find-name symbol))
		  (series-info-add! series 'name-time (current-time)))))
	    symbol-list))

(define (casablanca-download-names symbol-list)
  (if (any download-want-name? symbol-list)
      (begin
	(download-status (_ "Casablanca") (_ "names"))
	(receive (headers body)
	    (http-request
	     "http://www.casablanca-bourse.com/cgi/ASP/pts/pts_en.asp"
	     #:want-ok #t)
	  ;; process all names while we're at it
	  (process-names (download-also symbol-list
					#:selector casablanca-symbol?)
			 body)))))


;;-----------------------------------------------------------------------------
;; download - dividends
;;
;; This uses the Market Transaction (OST in French) pages like
;;
;;     http://www.casablanca-bourse.com/fiches/valeurs/MNG/fr/ost.html
;;
;; The english pages for these are in-progress, apparently, as of Jan 2007,
;; so the French pages are used.
;;
;; The server provides an ETag and Last-Modified which we use to avoid
;; getting another copy of data we've already seen and processed.  The
;; download is about 11k so this is a worthwhile saving.


;; return tdate for STR like "26.05.00"
(define (d.m.y-str->tdate str)
  (receive-list (day month year)
      (string-split str #\.)
    (ymd->tdate (yyy-str->year year)
		(string->number-err month)
		(string->number-err day))))

;; eg. "22,00"
(define (casablanca-dividend-parse str)
  (or (string->dividend-amount str #:decimal-point ",")
      str))

(define (process-dividends symbol url headers body)
  (if (http-headers-ok? headers)
      (let ((row-list (html-table-rows body (string-contains
					     body "name=\"div\""))))

	;; mung out various spaces and CRLF for ease of comparison
	(let ((headings (map (lambda (field)
			       (string-delete field char-set:whitespace))
			     (first row-list))))
	  (or (equal? headings '(""
				 "DividendesBrut(enMAD)"
				 "Num&eacute;roCoupon"
				 "D&eacute;tachement&agrave;laBourse"))
	      (error "Casablanca field missing:" headings)))
	(set! row-list (cdr row-list))

	(database-call-modify-data symbol
	  (lambda (series)
	    (for-each (lambda (row)
			(receive-list (year amount coupon tdate)
			    row
			  ;; "NEANT" (nothing) when current year has no
			  ;; dividends yet
			  (if (string<> "NEANT" tdate)
			      (series-dividend-add!
			       series
			       (d.m.y-str->tdate tdate)
			       (casablanca-dividend-parse amount)))))
		      row-list)
	    (series-info-add! series 'dividends-time (current-time))
	    (download-series-tag-info! series url headers))))

      ;; 302 Not Modified, no new data, just update time
      (database-call-modify-data symbol
	(lambda (series)
	  (series-info-add! series 'dividends-time (current-time))))))

(define (casablanca-download-dividends symbol-list)

  ;; recheck every 30 days
  (define target-time (- (current-time) (* 30 86400)))

  (define (want-dividends? symbol)
    (let ((t (database-read-info symbol 'dividends-time)))
      (or (not t)
	  (< t target-time))))

  (for-each
   (lambda (symbol)
     (if (want-dividends? symbol)
	 (let ((url (string-append
		     "http://www.casablanca-bourse.com/fiches/valeurs/"
		     (chart-symbol-sans-dot symbol)
		     "/fr/ost.html")))
	   (download-status (_ "Casablanca") (_ "dividends") symbol)
	   (receive (headers body)
	       (apply http-request url #:want-ok #t
		      (download-tag-args url symbol-list))
	     (process-dividends symbol url headers body)))))
   symbol-list))


;;-----------------------------------------------------------------------------
;; download

(download-handler! #:selector casablanca-symbol?
		   #:update   (lambda (symbol-list)
				(casablanca-download symbol-list)
				(casablanca-download-names symbol-list)
				(casablanca-download-dividends symbol-list))
		   #:avail-proc casablanca-available-tdate
		   #:backto   casablanca-backto)
