;; Commodity Channel Index (CCI).

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart cci)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (chart indicator)
  #:use-module (chart misc))

;; http://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:commodity_channel_index_cci
;;
;; http://www.incrediblecharts.com/technical/commodity_channel_index.php
;;


;; Return a procedure which calculates a Commodity Channel Index over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC tp) enters a new typical price into the window, and the
;; return is the CCI up to (and including) that value, or #f if not enough
;; values yet.
;;
;; DATA is a list of initial typical price values to enter into the window,
;; the same as if PROC were called on them successively (from start to end).
;; COUNT-1 is enough to ensure the first PROC call is based on a full
;; window.
;;
(define*-public (cci-calc-proc count #:optional (data '()))
  (set! data (take-right-maybe data count))

  ;; DATA is an "ice-9 q" of typical prices
  ;; POINTS is the number of non-#f points in DATA
  ;; TOTAL is the sum of the point values in DATA

  (let ((min-points (ceiling-exact (* indicator-minimum-fraction count)))
	(points     (length data))
	(total      (apply + data)))
    (set! data (list->q data))

    (lambda (x)

      ;; drop oldest point
      (if (>= points count)
	  (set! total (- total (deq! data)))
	  (set! points (1+ points)))

      ;; add this point
      (enq! data x)
      (set! total (+ total x))

      (and (>= points min-points)
	   (let* ((smatp   (/ total points))
		  (meandev (apply arithmetic-mean (map (lambda (tp)
							 (abs (- tp smatp)))
						       (q-list data)))))
	     (and (not (zero? meandev))
		  (/ (- x smatp)
		     (* 0.015 meandev))))))))

(define (cci-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count (1- count)))
	 (array  (series-array parent lo hi))
	 (cci    (cci-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(cci (arithmetic-mean
	      (or (array-ref array i 1) close)  ;; high
	      (or (array-ref array i 2) close)  ;; low
	      close))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array
		    (cci (arithmetic-mean
			  (or (array-ref array i 1) close)  ;; high
			  (or (array-ref array i 2) close)  ;; low
			  close))
		    i)))))

(define-public (cci-series series count)
  (ohlcv-required "CCI" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "CCI") count))
    (series-datatype!  new-series cci-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts cci-fill
							   count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public cci-datatype
  (datatype-new!
   #:key    'cci
   #:name   (_ "CCI - Commodity Channel Index")
   #:input   datatype-ohlcv
   #:units  'cci
   #:hlines '(-100 100)
   #:manual (_ "CONTEXT(manual-node): Commodity Channel Index")
   #:proc   cci-series
   #:parameter-list (list (datatype-param-new
			   #:name    (_ "Days")
			   #:key     'cci-days
			   #:type    'integer
			   #:minimum 1
			   #:default 20))))
