;; Central Japan Commodity Exchange setups. -*- coding: shift_jis -*-

;; Copyright 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ccom)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (ccom-symbol? symbol)
  (string-suffix? ".CCOM" symbol))

(symbol-timezone! ccom-symbol? (_ "Japan") timezone-tokyo)
(source-help! ccom-symbol?
	      (_ "CONTEXT(manual-node): Central Japan Commodity Exchange"))

;; eg. http://www.c-com.or.jp/public_html_e/guide/kerosene.php
(weblink-handler! ccom-symbol?
  (lambda (symbol)
    (let ((name (string-downcase
		 (string-delete (chart-symbol-commodity symbol)
				#\space)))
	  (lang (lang-select '(("ja" "")
			       ("en" "_e")))))
      (if (string=? name "ferrousscrap")
	  (set! name "scrap"))
      (if (string=? name "rubberindex")
	  (set! name "rubberIndex"))

      ;; more munging for japanese links
      (if (string=? lang "")
	  (set! name (or (assoc-ref '(("gasoline" . "gas")
				      ("kerosene" . "heating")
				      ("gasoil"   . "diesel")
				      ("eggs"     . "egg"))
				    name)
			 name)))

      (list (_ "C-COM _Contract Specifications")
	    (_ "Open web browser at the Central Japan Commodities Exchange contract specifications for this commodity")
	    (string-append "http://www.c-com.or.jp/public_html" lang "/guide/"
			   name ".php")))))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the quotes page
;;
;;     http://www.c-com.or.jp/public_html/cce-01_e.htm
;;
;; which has all commodities.


;; return list (ADATE TIME) for latest quote
;; trading by sessions 9:00 to 15:30, not sure when last will appear, so
;; keep looking until 16:00
;; reference: http://www.c-com.or.jp/public_html/cce-09rd01_e.htm
;;
(define (ccom-quote-adate-time . dummy-symbol)
  (tm->adate-time-within (localtime (current-time) (timezone-tokyo))
			 #,(hms->seconds 9 0 0)
			 #,(hms->seconds 16 0 0)))

;; return mdate for STR like "2005SEP"
(define (yyyyMMM-str->mdate str)
  (ymd->mdate (string->number-err (string-take str 4))
	      (Mmm-str->month (string-drop str 4))
	      1))

(define (ccom-body->latest-list headers body)
  (define char-set:ABC (char-set #\A #\B #\C))
  (define char-set:^v_ (char-set #\^ #\v #\space))

  ;; the name heading for each table is a gif, so instead extract text from
  ;; the links that each table has to other tables, like
  ;;         <a id="comm" href="#407">Kerosene</a>
  ;; form an alist of (("405" . "Eggs") ...)
  (let* ((num->name-alist
	  (map (lambda (m)
		 (let ((num  (match:substring m 1))
		       (name (match:substring m 2)))
		   ;; apparent misspelling
		   (if (string=? name "Aluminum")
		       (set! name "Aluminium"))
		   (cons num name)))
	       (must-match
		(list-matches #,(regexp "<a id=\"comm\" href=\"#([^\"]+)\">([^<]+)</a>") body))))
	 (ret   '()))

    ;; anchor for commodity sections:
    ;; <TD colspan="2" align="left"><a name="406">
    (fold-matches
     #,(regexp "name=\"([0-9]+)\"")
     body #f
     (lambda (m prev)
       (let* ((num        (match:substring m 1))
	      (name       (or (assoc-ref num->name-alist num)
			      (error "C-COM: name not found for quote")))
	      (commodity  name)

	      ;; date+time like
	      ;; <TD align="left">2005/07/29 17:04</TD>
	      (m          (must-match (regexp-exec #,(regexp "([0-9][0-9][0-9][0-9]/[0-9]+/[0-9]+) ([0-9]+:[0-9]+)") body (match:start m))))
	      (adate      (y/m/d-str->adate (match:substring m 1)))
	      (time       (h:m:s-str->seconds (match:substring m 2)))

	      (row-list   (html-table-rows body (match:end m))))

	 (set! row-list (apply zip row-list))

	 (let ((headings (first row-list)))
	   ;; lose date from "Open Interest(10/03)"
	   (set! headings (map (lambda (heading)
				 (let ((pos (string-index heading #\()))
				   (if pos
				       (string-take heading pos)
				       heading)))
			       headings))
	   (or (equal? (cdr headings)
		       '("Previous Closing Price"
			 "Morning Session 1st"
			 "Morning Session 2nd"
			 "Morning Session 3rd"
			 "Afternoon Session 1st"
			 "Afternoon Session 2nd"
			 "Afternoon Session 3rd"
			 "Change from prev. day"
			 "Settlement Price"
			 "Volume"
			 "Open Interest"))
	       (error "C-COM: unrecognised quote headings:" headings)))
	 (set! row-list (cdr row-list))


	 ;; total volumes by session
	 (if (equal? "Volume" (first (last row-list)))
	     (set! row-list (drop-right row-list 1)))

	 ;; empty month (U+3000 space char) in some Nickel from Feb 2009
	 (set! row-list (remove (lambda (row)
				  (equal? "@" (car row)))
				row-list))

	 (for-each
	  (lambda (row)
	    (receive-list (month prev m1 m2 m3 a1 a2 a3 change settle volume
				 prev-int)
		row

	      ;; blank (shift-jis char) when no trade in session
	      (if (equal? m1 	 "@") (set! m1     ""))
	      (if (equal? m2 	 "@") (set! m2     ""))
	      (if (equal? m3 	 "@") (set! m3     ""))
	      (if (equal? a1 	 "@") (set! a1     ""))
	      (if (equal? a2 	 "@") (set! a2     ""))
	      (if (equal? a3 	 "@") (set! a3     ""))
	      (if (equal? settle "@") (set! settle ""))
	      (if (equal? volume "@") (set! volume "0"))

	      ;; commas in prices and volume
	      (set! m1     (string-delete m1 	 #\,))
	      (set! m2     (string-delete m2 	 #\,))
	      (set! m3     (string-delete m3 	 #\,))
	      (set! a1     (string-delete a1 	 #\,))
	      (set! a2     (string-delete a2 	 #\,))
	      (set! a3     (string-delete a3     #\,))
	      (set! change (string-delete change #\,))
	      (set! settle (string-delete settle #\,))

	      ;; leading A=ask, B=bid, C=close on intraday (and prev actually)
	      (set! m1     (string-trim m1 char-set:ABC))
	      (set! m2     (string-trim m2 char-set:ABC))
	      (set! m3     (string-trim m3 char-set:ABC))
	      (set! a1     (string-trim a1 char-set:ABC))
	      (set! a2     (string-trim a2 char-set:ABC))
	      (set! a3     (string-trim a3 char-set:ABC))

	      (let* (;; month was eg. "2005<BR>DEC" but <BR> stripped
		     (mdate (yyyyMMM-str->mdate month))
		     (note  (cond ((string-prefix? "^" a3)
				   (_ "limit up"))
				  ((string-prefix? "v" a3)
				   (_ "limit down"))
				  (else
				   #f))))

		;; leading ^=limit up, v=limit down, or space filler
		(set! m1 (string-trim m1 char-set:^v_))
		(set! m2 (string-trim m2 char-set:^v_))
		(set! m3 (string-trim m3 char-set:^v_))
		(set! a1 (string-trim a1 char-set:^v_))
		(set! a2 (string-trim a2 char-set:^v_))
		(set! a3 (string-trim a3 char-set:^v_))

		;; limit shown on a3, drop it if the settle is not the same
		(if (not (string=? a3 settle))
		    (set! note #f))

		(receive-list (decimals change . seq)
		    (strings->numbers+decimals change m1 m2 m3 a1 a2 a3 settle)

		  (let ((open  (and (not (null? seq)) (first seq))))
		    (set! seq (delq #f seq))
		    (let ((high  (and (not (null? seq)) (apply max seq)))
			  (low   (and (not (null? seq)) (apply min seq)))
			  (final (and (not (null? seq)) (last seq))))

		      (set! ret (cons
				 (latest-new
				  #:symbol (string-append
					    commodity " "
					    (mdate->MMM_YY-str mdate)
					    ".CCOM")
				  #:name   (string-append
					    name " "
					    (string-drop month 4)
					    " "
					    (string-take month 4))
				  #:last-adate     adate
				  #:last-time      time
				  #:open           open
				  #:high           high
				  #:low            low
				  #:last           final
				  #:change         change
				  #:volume         volume
				  #:decimals       decimals
				  #:contract-mdate mdate
				  #:note           note
				  #:headers        headers
				  #:source         'ccom)
				 ret))))))))

	  row-list))))
    (latest-list-add-front-months ret)))

(define (ccom-latest-get symbol-list extra-list proc)
  (receive (headers body)
      (apply http-request "http://www.c-com.or.jp/public_html/cce-01_e.htm"
	     #:want-ok #t
	     (latest-tag-args 'ccom symbol-list))
    (proc (if (http-headers-ok? headers) ;; can be 304 not modified
	      (ccom-body->latest-list headers body)
	      '()))))

(latest-handler! #:selector   ccom-symbol?  
		 #:handler    ccom-latest-get
		 #:adate-time ccom-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - data
;;
;; This uses the downloads at
;;
;;     http://www.c-com.or.jp/public_html_e/historical/index.php
;;
;; The current year file goes up to the day before yesterday, and the
;; yesterday file gives the latest.


(define (ccom-line->tdate line)
  (yyyymmdd-str->tdate (string-take-to-char line #\,)))

(define (ccom-process-download symbol-list url headers body)

  (define row-list (string-split-lines body))

  (let ((headings      (first row-list))
	(want-headings "yr_mo_dy,contract,contract_month,price_type_m1,contract_price_m1,price_type_m2,contract_price_m2,price_type_m3,contract_price_m3,price_type_a1,contract_price_a1,price_type_a2,contract_price_a2,price_type_a3,contract_price_a3,sett_price,volume,open_int(Previous Day)"))
    (or (string=? headings want-headings)
	(error "C-COM: data unrecognised headings:" headings))

    ;; header line is repeated within data in prior years files
    (set! row-list (delete! want-headings row-list)))

  (set! row-list (map! (lambda (line)
			 (string-split line #\,))
		       row-list))

  (download-process
   #:module        (_ "C-COM")
   #:symbol-list   symbol-list
   #:currency      "JPY"
   #:url           url
   #:headers       headers
   #:row-list
   (map-rows! (lambda (tdate commodity mdate
			     type-m1 m1 type-m2 m2 type-m3 m3
			     type-a1 a1 type-a2 a2 type-a3 a3
			     settle volume openint)

		;; drop the "Shell" in "Shell Eggs"
		(if (string-prefix? "Shell " commodity)
		    (set! commodity (string-drop commodity 6)))

		;; an apparent misspelling in TYRHST file as of Jan 2007
		(if (string=? commodity "Aluminum")
		    (set! commodity "Aluminium"))

		;; dummy rows in BYRHST_405_E.csv near end of 2005 have just
		;; "20" for the contract month
		(and (string<> mdate "20")

		     (list #:tdate         (yyyymmdd-str->tdate tdate)
			   #:mdate         (yyyymm-str->mdate mdate)
			   #:commodity     commodity
			   #:sessions      (list m1 m2 m3 a1 a2 a3 settle)
			   #:volume        volume
			   #:open-interest openint)))
	      row-list)))

(define (ccom-commodity->code commodity)
  (or (assoc-ref '(("Gasoline" 	    . "406")
		   ("Kerosene" 	    . "407")
		   ("Gas Oil"  	    . "409")
		   ("Eggs"     	    . "405")
		   ("Ferrous Scrap" . "410")

		   ;; former OME commodities
		   ("RSS3"          . "411")
		   ("TSR20"         . "412")
		   ("Rubber Index"  . "415")
		   ("Aluminium"     . "414")
		   ("Nickel"        . "413")

		   ) commodity)
      (error "C-COM: unrecognised commodity:" commodity)))

(define (year-available-tdate)
  (year->tdate (today-year (timezone-tokyo))))

;; Last-Modified seen on year files
;;     Fri, 02 Sep 2005 14:30:01 GMT   = 23:30:01 tokyo
;; and on day files
;;     Fri, 02 Sep 2005 14:30:20 GMT
;; so try for new data at midnight
;;
(define (day-available-tdate)
  (1- (tdate-today-after 0 0 (timezone-tokyo))))

;; download previous years file, perhaps
(define (prev-download symbol-list) ;; of one commodity
  (define avail-tdate (year-available-tdate))
  (if (any (lambda (symbol)
	     (< (download-start-tdate symbol) avail-tdate))
	   symbol-list)

      (let* ((commodity (chart-symbol-commodity (first symbol-list)))
	     (url (simple-format
		   #f "http://www.c-com.or.jp/public_html/csv/BYRHST_~a_E.csv"
		   (ccom-commodity->code commodity))))
	(download-status (_ "C-COM") (_ "data") commodity (_ "previous years"))

	(receive (headers body)
	    (apply http-request url
		   (download-tag-args url symbol-list))
	  (let ((status-code (http-headers-status-code headers)))
	    (cond ((http-status-code-ok? status-code)
		   (set! symbol-list
			 (download-also symbol-list
					#:selector  ccom-symbol?
					#:end-tdate (ccom-line->tdate
						     (string-last-line body))
					#:commodity #t))
		   (ccom-process-download symbol-list url headers body))

		  ((= 404 status-code)
		   ;; 404 for no BYR format for old Osaka commodities
		   )
		  ((= 304 status-code)
		   ;; nothing new
		   )
		  (else
		   ;; anything else is bad
		   (http-headers-throw headers))))))))

;; download this year file
(define (year-download symbol-list) ;; of one commodity
  (define avail-tdate (1- (day-available-tdate)))
  (if (any (lambda (symbol)
	     (<= (download-start-tdate symbol) avail-tdate))
	   symbol-list)
      (let* ((commodity (chart-symbol-commodity (first symbol-list)))
	     (url (simple-format #f "http://www.c-com.or.jp/public_html/csv/TYRHST_~a_E.csv" (ccom-commodity->code commodity))))
	(download-status (_ "C-COM") (_ "data") commodity (_ "this year"))
	(receive (headers body)
	    (http-request url #:want-ok #t)
	  (set! symbol-list
		(download-also symbol-list
			       #:selector  ccom-symbol?
			       #:end-tdate (ccom-line->tdate
					    (string-last-line body))
			       #:commodity #t))

	  (ccom-process-download symbol-list url headers body)))))

;; download day file, if it will update all of SYMBOL-LIST
;; return #t if updated successful or #f if not (should use year file)
;; 
(define (day-download symbol-list) ;; of one commodity
  (define avail-tdate (day-available-tdate))
  ;; use csv if all of SYMBOL-LIST just wanting AVAIL-TDATE
  (and (every (lambda (symbol)
		(= avail-tdate (download-start-tdate symbol)))
	      symbol-list)
       (let* ((commodity (chart-symbol-commodity (first symbol-list)))
	      (url (simple-format
		    #f "http://www.c-com.or.jp/public_html/csv/YSDHST_~a_E.csv"
		    (ccom-commodity->code commodity))))
	 (download-status (_ "C-COM") (_ "data") commodity (_ "yesterday"))

	 (receive (headers body)
	     (apply http-request url #:want-ok #t
		    (download-tag-args url symbol-list))
	   (if (http-headers-ok? headers)
	       (let ((got-tdate (ccom-line->tdate (string-last-line body))))
		 (cond ((= got-tdate avail-tdate)
			;; got the expected data, process it
			;; also processs for any others covered by this date
			(set! symbol-list
			      (download-also symbol-list
					     #:selector    ccom-symbol?
					     #:commodity   #t
					     #:start-tdate avail-tdate
					     #:end-tdate   avail-tdate))
			(ccom-process-download symbol-list url headers body)
			#t)
		       ((< got-tdate avail-tdate)
			;; got something older, there's no new data
			#t)
		       (else
			#f)))
	       ;; 304 Not Modified, no new data (yet)
	       #t)))))

(define (ccom-download symbol-list)
  (for-each (lambda (symbol-list) ;; of the same commodity
	      (prev-download symbol-list)
	      (year-download symbol-list)
	      (day-download symbol-list))

	    (partition-equal chart-symbol-commodity symbol-list))

  (download-name-from-latest (_ "C-COM") symbol-list))

(define (ccom-backto want-tdate symbol-list)
  (download-message
   (string-append
    (_ "C-COM") ": "
    (_ "no further historical data available"))))


;;-----------------------------------------------------------------------------
;; download

(download-handler! #:selector ccom-symbol?
		   #:update   ccom-download
		   #:backto   ccom-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any ccom-symbol? symbol-list)
			      (download-now-commodity-all-months))))
