;; Chaikin Money Flow.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart chaikin-mf)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a Chaikin Money Flow index over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC high low close volume) enters a new value into the
;; window, and the return is the Chaiking Money Flow up to (and including)
;; that value.  The values can be #f when data is not available.  If there's
;; not enough valid points the return will be #f.
;;
;; If all the VOLUME amounts in the window are zero, then the return is 0
;; too.
;;
;; This windowed parts/total calculation is similar to
;; `percent-posneg-calc-proc', but we scale to -1 to +1.
;;
(define-public (chaikin-mf-calc-proc count)
  (let ((data       (make-circular-list (* 2 count) #f))
	(ad         0)
	(total      0)
	(points     0)
	;; at least 3/4 of points in window to give a value
	(min-points (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (high low close volume)

      ;; dummy out if not high/low and/or volume
      (set! high   (or high close))
      (set! low    (or low  close))
      (set! volume (or volume 0))

      (let* ((diff (- high low))
	     (x    (if (zero? diff)
		       0
		       ;; 2.0 forces to a flonum, not a frac in guile 1.8
		       (* volume (1- (* 2.0 (/ (- close low)
					       diff)))))))

	;; subtract old point from values
	(if (first data)
	    (begin
	      (set! ad    (- ad    (first data)))
	      (set! total (- total (second data))))

	    ;; otherwise a new point
	    (set! points (1+ points)))

	;; add this point to totals
	(set! ad    (+ ad    x))
	(set! total (+ total volume))

	;; enter this point into window, replacing old
	(set-first!  data x)
	(set-second! data volume)
	(set! data (cddr data))

	(and (>= points min-points)
	     (if (= 0 total)
		 0
		 (/ ad total)))))))

(define (chaikin-mf-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo
					   #:count  count))
	 (array  (series-array parent start hi))
	 (cmf    (chaikin-mf-calc-proc count)))

    ;; warmup points preceding lo, if available
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (cmf (array-ref array i 1)    ;; high
	       (array-ref array i 2)    ;; low
	       (array-ref array i 3)    ;; close
	       (array-ref array i 4)))) ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array
		      (cmf (array-ref array i 1)   ;; high
			   (array-ref array i 2)   ;; low
			   (array-ref array i 3)   ;; close
			   (array-ref array i 4))  ;; volume
		      i)))))

(define-public (chaikin-mf-series series count)
  (ohlcv-required "Chaikin MF" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Chaikin MF") count))
    (series-datatype!  new-series chaikin-mf-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-array!     new-series #f)
    (series-require!   new-series (indicator-require-parts chaikin-mf-fill
							   count))
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public chaikin-mf-datatype
  (datatype-new!
   #:key       'chaikin-mf
   #:name      (_ "Chaikin Money Flow")
   #:input     datatype-ohlcv
   #:units     'plus-or-minus-one
   #:linestyle 'line
   #:minimum   -1
   #:maximum   1
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Chaikin Money Flow")
   #:proc      chaikin-mf-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'chaikin-mf-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 21))))
