/* Chart gnutls interface program.

   Copyright 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/* Usage: chart-gnutls --gnutlsversion
          chart-gnutls hostname portnum

   With --gnutlsversion, print the GnuTLS library version number to stdout
   and exit 0.

   With hostname and portnum, establish a connection to that destination and
   then copy data back and forward between it and stdin/stdout.  Any errors
   are printed to stderr.

   This is similar to "gnutls-cli", or "openssl s_client -quiet".
   But gnutls-cli (as of version 1.6.2) is not usable because it prints
   various informational messages to stdout, and shutdown messages could be
   misinterpreted as part of a http body.  It also uses fgets() and
   therefore hangs waiting for a newline after http POST data.  */


#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include <gnutls/gnutls.h>


/* compile-time options for development and testing */
const int option_verbose = 0;


#if ! HAVE_HSTRERROR
char *
hstrerror (int err)
{
  switch (err) {
  case HOST_NOT_FOUND: return "Unknown host";
  case TRY_AGAIN:      return "Host name lookup failure";
  case NO_RECOVERY:    return "Unknown server error";
  case NO_DATA:        return "No address associated with name";
  default:             return "Unknown hostname resolver error";
  }
}
#endif

#define MIN(x,y) ((x) < (y) ? (x) : (y))


#define TLSERR(proc, args)  tlserr_proc (#proc, proc args)
int
tlserr_proc (const char *func, int ret)
{
  if (ret < 0)
    {
      fprintf (stderr, "%s: %s\n", func, gnutls_strerror (ret));
      exit (1);
    }
  return ret;
}

int
die_errno (const char *func)
{
  perror (func);
  exit (1);
}

void
log_to_stderr (int level, const char *msg)
{
  fputs (msg, stderr);
}

#define toprint(c)  (isprint(c) ? c : '.')

static void
print_bytes (const char *buf, size_t len)
{
  size_t i;
  while (len > 0)
    {
      size_t this_len = MIN (len, 16);
      for (i = 0; i < this_len; i++)
        fprintf (stderr, " %2X", (int) (unsigned char) buf[i]);
      fprintf (stderr, "%*s", (int) (3 * (16 - i) + 2), "");
      for (i = 0; i < this_len; i++)
        fputc (toprint ((int) (unsigned char) buf[i]), stderr);
      fputc ('\n', stderr);
      buf += this_len;
      len -= this_len;
    }
}

static ssize_t
read_with_trace (gnutls_transport_ptr_t ptr, void *buf, size_t size)
{
  int sock = (int) ptr;
  ssize_t ret;

  fprintf (stderr, "socket read() %u ... ", size);
  ret = read (sock, buf, size);

  fprintf (stderr, "got %d\n", ret);
  if (ret > 0)
    print_bytes (buf, ret);

  return ret;
}

static ssize_t
write_with_trace (gnutls_transport_ptr_t ptr, const void *buf, size_t size)
{
  int sock = (int) ptr;
  ssize_t ret;

  fprintf (stderr, "socket write() %u\n", size);
  print_bytes (buf, size);

  ret = write (sock, buf, size);
  fprintf (stderr, " wrote %d%s\n", ret, (ret == size ? "(ok)" : ""));
  return ret;
}


int
main (int argc, char *argv[])
{
  static fd_set fds; /* initially zeros */
  static char buf[128000];

  int sfamily, sstyle, sprotocol;
  struct sockaddr *sa;
  size_t salen;
  char *hoststr, *portstr;
  gnutls_session_t session;
  int sock;

  setbuf (stdin, NULL);
  setbuf (stdout, NULL);
  setbuf (stderr, NULL);

  if (argc >= 2 && strcmp (argv[1], "--gnutlsversion") == 0)
    {
      puts (gnutls_check_version (NULL));
      exit (0);
    }

  if (argc != 3)
    {
      fprintf (stderr, "usage: chart-gnutls host port\n");
      exit (1);
    }

  hoststr = argv[1];
  portstr = argv[2];

  TLSERR (gnutls_global_init, ());
  if (option_verbose)
    {
      gnutls_global_set_log_function (log_to_stderr);
      gnutls_global_set_log_level (option_verbose >= 2 ? 1000 : 9);
    }

  TLSERR (gnutls_init, (&session, GNUTLS_CLIENT));
  TLSERR (gnutls_set_default_priority, (session));

  if (option_verbose >= 2)
    {
      gnutls_transport_set_pull_function (session, read_with_trace);
      gnutls_transport_set_push_function (session, write_with_trace);
    }


#if 0
  {
    static gnutls_anon_client_credentials_t anoncred;
    static const int prios[] = { GNUTLS_KX_ANON_DH, 0 };
    TLSERR (gnutls_anon_allocate_client_credentials, (&anoncred));
    TLSERR (gnutls_credentials_set, (session, GNUTLS_CRD_ANON, anoncred));
    TLSERR (gnutls_kx_set_priority, (session, prios));
  }
#endif

#if 1
  {
    static gnutls_certificate_credentials_t xcred;
    static const int types[] = { GNUTLS_CRT_X509, GNUTLS_CRT_OPENPGP, 0 };
    TLSERR (gnutls_certificate_allocate_credentials, (&xcred));
    TLSERR (gnutls_credentials_set, (session, GNUTLS_CRD_CERTIFICATE, &xcred));
    TLSERR (gnutls_certificate_type_set_priority, (session, types));
  }
#endif

#if HAVE_GETADDRINFO
  {
    static struct addrinfo hints; /* zeros in other fields */
    struct addrinfo *ai;
    int ret;

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = 0;
    hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICSERV;

    ret = getaddrinfo (hoststr, portstr, &hints, &ai);
    if (ret == EAI_SYSTEM)
      die_errno ("getaddrinfo");
    if (ret != 0)
      {
        fprintf (stderr, "getaddrinfo: %s: %s\n", hoststr, gai_strerror (ret));
        exit (1);
      }

    sfamily = ai->ai_family;
    sstyle = ai->ai_socktype;
    sprotocol = ai->ai_protocol;
    sa = ai->ai_addr;
    salen = ai->ai_addrlen;
  }
#else
  {
    static struct sockaddr_in sabuf;
    struct hostent *hent;

    hent = gethostbyname (hoststr);
    if (hent == NULL)
      {
        fprintf (stderr, "gethostbyname: %s: %s\n",
                 hoststr, hstrerror (h_errno));
        exit (1);
      }

    /* normally gethostbyname doesn't give AF_INET6 anyway; in glibc can ask
       for it with _res.options, but getaddrinfo() is available there of
       course */
    if (hent->h_addrtype != AF_INET)
      {
        fprintf (stderr, "chart-gnutls: only AF_INET supported, got %d\n",
                 hent->h_addrtype);
        exit (1);
      }

    sfamily = PF_INET;
    sstyle = SOCK_STREAM;
    sprotocol = 0;

    sabuf.sin_family = AF_INET;
    sabuf.sin_addr = * (struct in_addr *) hent->h_addr;
    sabuf.sin_port = htons (atoi (portstr));

    sa = (struct sockaddr *) &sabuf;
    salen = sizeof (sabuf);
  }
#endif

  if (option_verbose)
    {
      fprintf (stderr, "sock family %d %s\n", sfamily,
               (sfamily == PF_INET ? "PF_INET"
                : sfamily == PF_INET6 ? "PF_INET6"
                : "??"));
      fprintf (stderr, "sock style %d %s\n", sstyle,
               (sstyle == SOCK_STREAM ? "SOCK_STREAM"
                : "??"));
      fprintf (stderr, "sock protocol %d\n", sprotocol);
      fprintf (stderr, "sock addr family %d %s", sa->sa_family,
               (sa->sa_family == AF_INET ? "AF_INET"
                : sa->sa_family == AF_INET6 ? "AF_INET6"
                : "??"));
      if (sa->sa_family == AF_INET)
        {
          struct sockaddr_in *in = (struct sockaddr_in *) sa;
          fprintf (stderr, " addr %s port %d\n",
                   inet_ntoa (in->sin_addr),
                   in->sin_port);
        }
#if HAVE_GETADDRINFO
      else if (sa->sa_family == AF_INET6)
        {
          struct sockaddr_in6 *in6 = (struct sockaddr_in6 *) sa;
          char text[128];
          fprintf (stderr, " addr %s port %d\n",
                   inet_ntop (sa->sa_family, sa->sa_data,
                              text, sizeof (text)),
                   in6->sin6_port);
        }
#endif
    }


  sock = socket (sfamily, sstyle, sprotocol);
  if (sock < 0)
    die_errno ("socket");

  /* Linux kernel default for keepalive is 9 tries 75 seconds apart, which
     is fully 12.5 minutes, making this option fairly useless; the timeouts
     in http.scm are shorter.  */
  {
    int val = 1;
    if (setsockopt (sock, SOL_SOCKET, SO_KEEPALIVE, &val, sizeof (val)) < 0)
      die_errno ("setsockopt SO_KEEPALIVE");
  }

#ifdef SO_RCVBUF
  if (option_verbose)
    {
      int val;
      socklen_t size = sizeof (val);
      if (getsockopt (sock, SOL_SOCKET, SO_RCVBUF, &val, &size) != 0)
        die_errno ("getsockopt SO_RCVBUF");
      fprintf (stderr, "socket SO_RCVBUF %d\n", val);
    }
#endif

  if (connect (sock, sa, salen) != 0)
    die_errno ("connect");

  gnutls_transport_set_ptr (session, (gnutls_transport_ptr_t) sock);

  TLSERR (gnutls_handshake, (session));


  for (;;)
    {
      FD_SET (0, &fds);
      FD_SET (sock, &fds);
      if (option_verbose) fprintf (stderr, "select()\n");
      if (select (FD_SETSIZE, &fds, NULL, NULL, NULL) < 0)
        die_errno ("select");
      if (option_verbose) fprintf (stderr, "  done\n");

      if (FD_ISSET (0, &fds))
        {
          ssize_t r, w;

          if (option_verbose) fprintf (stderr, "stdin read\n");
          r = read (0, buf, sizeof (buf));
          if (option_verbose) fprintf (stderr, "  done %d\n", r);
          if (r < 0)
            die_errno ("read");

          if (r == 0)  /* EOF */
            goto done;

          if (option_verbose >= 2 && r > 0) print_bytes (buf, r);

          w = gnutls_record_send (session, buf, r);
          if (w < 0)
            {
              fprintf (stderr, "gnutls_record_send: %s\n", gnutls_strerror (w));
              exit (1);
            }
          if (w != r)
            {
              fprintf (stderr, "gnutls_record_send: wrote only %d of %d\n", w, r);
              exit (1);
            }
        }


      /* only one look for socket readability, but repeat as much as
         gnutls_record_check_pending() says */
      while (FD_ISSET (sock, &fds)
             || gnutls_record_check_pending (session))
        {
          ssize_t r, w;
          FD_CLR (sock, &fds);

          if (option_verbose) fprintf (stderr, "gnutls_record_recv()\n");
          r = gnutls_record_recv (session, buf, sizeof (buf));
          if (option_verbose) fprintf (stderr, "  done %d\n", r);
          if (r < 0)
            {
              fprintf (stderr, "gnutls_record_recv: %s\n",
                       gnutls_strerror (r));
              exit (1);
            }
          if (r == 0)  /* EOF */
            goto done;

          if (option_verbose >= 2) print_bytes (buf, r);

          w = write (1, buf, r);
          if (w < 0)
            die_errno ("write");
          if (w != r)
            {
              fprintf (stderr, "stdout: wrote only %d of %d\n", w, r);
              exit (1);
            }
        }

    }

 done:
  TLSERR (gnutls_bye, (session, GNUTLS_SHUT_RDWR));
  if (close (sock) != 0)
    die_errno ("close");

  gnutls_deinit (session);

  return 0;
}
