# Chart support for bash command line completions.

# Copyright 2006, 2007 Kevin Ryde

# This file is part of Chart.
#
# Chart is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3, or (at your option) any later version.
#
# Chart is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# This script sets up command line completions in bash for the chart
# program.  See "Programmable Completion" and "Commands For Completion" in
# the bash manual for how that works.
#
# Stock symbols from the Chart database are completed, as are the various
# chart command line options.
#
# This script can be used standalone, or with the bash_completion project
#
#     http://freshmeat.net/projects/bashcompletion/
#

# To install you can either
#
# 1) Source from your .bashrc,
#
#        . /my-install-dir/share/chart/chart.bash
#
# 2) Or load it with bash_completion, to have it only when using that, either
#
#    2a) Source from your ~/.bash_completion file,
#
#            . /my/location/share/chart/chart.bash
#
#    2b) Or put it in the $BASH_COMPLETION_DIR directory (which might be for
#        instance "/etc/bash_completion.d", and bash_completion will source
#        it (along with everything else in that directory).
#


# $_chart_completions__data_directory is per `data-directory' in misc.scm.
# If you point that directory somewhere new in your ~/Chart/init.scm then
# you can set _chart_completions__data_directory in your .bashrc to match.
#
# It'd be possible to run up a chart process to query it, but that'd be a
# bit slow for interactive use.
#
if [ -z "$_chart_completions__data_directory" ]; then
  _chart_completions__data_directory=~/Chart/data
fi


_chart_completions()
{
  # $1 is command being completed, ie. "chart"
  # $2 is current word, ie. the one to complete
  # $3 is the preceding word

  # names $cur and $prev are needed by _filedir() if using the bash_completion
  # package
  local cur=$2 prev=$3
  COMPREPLY=()

  if [ "$prev" = --display ]; then
    # a display name, but alas there's no easy way to find out what's available
    return 0
  fi

  # -s <filename>, to run a Chart script, allow any filenames at all, and
  # also any further argument after -s completes as a filename, being an
  # argument to the script
  local i=1
  while [ $i -le $COMP_CWORD ]; do
    if [ "${COMP_WORDS[i]}" = -s ]; then

      if type -t _filedir >/dev/null; then
        # using bash_completion
        _filedir
      else
        # on our own
        local IFS=$'\n'
        COMPREPLY=( $(compgen -f -- $cur) )
      fi
      return 0
    fi
    i=`expr $i + 1`
  done

  # command line option
  case "$cur" in
    -*)
      # a command line option
      #
      local options='
    	--alerts
    	--all
    	--backto
    	--debug
    	--display
    	--download
    	--favourites
    	--help -h
    	-s
    	--ticker
    	--version -v'
      local IFS=$' \n\t'
      COMPREPLY=( $( compgen -W "$options" -- $cur ) )
      return 0
      ;;
  esac

  # otherwise a Chart symbol, being the filenames in the database directory
  # $_chart_completions__data_directory
  #
  # shopt -s nocaseglob gives case-insensitive matching, and the ( )
  # subshell restricts that to the maching here, leaving any global setting
  # alone
  #
  # -X '.*' excludes ".", "..", ".lockfile" and any ".temp-XXXXXX" temporaries 
  #
  local IFS=$'\n'
  COMPREPLY=( $(cd $_chart_completions__data_directory;
                shopt -s nocaseglob;
		compgen -X '.*' -G "$cur*") )
  return 0
}

# "-o filenames" ensures spaces in symbols are escaped.  However that's a
# feature introduced in bash 2.05, which dates from early 2001, so if you're
# using a very ancient bash you might have to take it out.
#
complete -o filenames -F _chart_completions chart



# Local variables:
# mode: sh
# sh-shell: bash
# End:
