;;; chart.el --- stock quotes using Chart.

;; Copyright 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; See "Emacs" in the Chart manual for usage.


(require 'cl)

;;-----------------------------------------------------------------------------
;; customizations

(defgroup chart nil
  "Chart."
  :prefix "chart-"
  :group 'applications
  :link '(custom-manual "(chart)Emacs"))

(defface chart-up
  `(;; plain "green" is too light to see against a white background
    (((class color) (background light))
     (:foreground "green4"))
    (((class color))
     (:foreground "green")))
  "Face for a Chart quote which is up."
  :group 'chart)

(defface chart-down
  `((((class color))
     (:foreground "red")))
  "Face for a Chart quote which is down."
  :group 'chart)

(defface chart-in-progress
  `((((class color) (background dark))
     (:foreground "cyan"))
    (((class color) (background light))
     (:foreground "blue")))
  "Face for Chart quote fetch in progress."
  :group 'chart)

(defcustom chart-watchlist-hook nil
  "*Hook called by `chart-watchlist'."
  :type  'hook
  :group 'chart)

(defcustom chart-process-debug nil
  "Enable debugging on the Chart subprocess.
This runs chart with the \"--debug\" option, providing a Guile backtrace if
an error occurs, though it runs a little slower.  The backtrace is put into
a buffer \"*chart-process-backtrace*\".

This variable is used when the subprocess is started, so if you change it
you must eval `(chart-process-kill)' to force a restart.  The `customize'
interface does this automatically."
  :type  'boolean
  :set   (lambda (symbol value)
           ;; restart if changed
           (if (and (boundp 'chart-process-debug) ;; unbound when first loading
                    (not (eq value chart-process-debug)))
               (chart-process-kill))
           (set-default symbol value))
  :group 'chart)


;;-----------------------------------------------------------------------------
;; xemacs workarounds

;; eval at compile time to avoid warnings about unbound procs, eval when
;; executing the byte code to get the actual definitions
(eval-and-compile
  (when (featurep 'xemacs)
    (require 'timer)

    (unless (fboundp 'propertize)
      (defun propertize (str &rest properties)
        "Return a copy of STR with PROPERTIES added.
PROPERTIES is successive arguments PROPERTY VALUE PROPERTY VALUE ..."
        (setq str (copy-sequence str))
        (add-text-properties 0 (length str) properties str)
        str))))


;;-----------------------------------------------------------------------------
;; misc

(defmacro chart-with-temp-message (message &rest body)
  "Display MESSAGE temporarily while evaluating BODY.
This is the same as `with-temp-message' but has a workaround for a bug in
Emacs 21.4 where the temporary message isn't erased if there was no previous
message."
  (if (eval-when-compile (featurep 'xemacs))
      ;; one key for each macro usage, which means each usage is not reentrant
      (let ((key (gensym "chart-with-temp-message--")))
        `(unwind-protect
             (progn
               (display-message ',key ,message)
               (prog1 (progn ,@body)
                 (clear-message ',key)))
           (clear-message ',key)))

    `(let* ((chart-with-temp-message--oldmsg (current-message)))
       (unwind-protect
           (prog1 (with-temp-message ,message ,@body)
             (or chart-with-temp-message--oldmsg (message nil)))
         (or chart-with-temp-message--oldmsg (message nil))))))

(put 'chart-with-temp-message 'lisp-indent-function 1)

(defmacro chart-save-row-col (&rest body)
  "Evaluate BODY, preserving point+mark row/col and window start positions.
This is a bit like `save-excursion', but working with row+column rather than
a point position."
  `(let* ((point-row (count-lines (point-min) (point-at-bol)))
          (point-col (current-column))
          (mark-pos  (mark t)))
     (and mark-pos
          (goto-char mark-pos))
     (let* ((mark-row    (count-lines (point-min) (point-at-bol)))
            (mark-col    (current-column))
            ;; list of pairs (WINDOW . ROW)
            (window-rows (mapcar (lambda (window)
                                   (cons window
                                         (count-lines (point-min)
                                                      (window-start window))))
                                 (get-buffer-window-list (current-buffer)))))
       (prog1 (progn ,@body)
         (dolist (pair window-rows)
           (goto-char (point-min))
           (forward-line (cdr pair))
           (set-window-start (car pair) (point)))

         (goto-char (point-min))
         (forward-line mark-row)
         (move-to-column mark-col)
         (and mark-pos
              (set-marker (mark-marker) (point)))

         (goto-char (point-min))
         (forward-line point-row)
         (move-to-column point-col)))))

(put 'chart-save-row-col lisp-indent-function 0)

(defun chart-intersection (x y)
  "Return the intersection of lists X and Y, ie. elements common to both.
Elements are compared with `equal' and returned in the same order as they
appear in X.

This differs from the cl package `intersection' in preserving the order of
elements for the return, the cl package doesn't preserve the order."

  (remove* nil x :test-not (lambda (dummy xelem)
                             (member xelem y))))

(defvar chart-symbol-history nil
  "Interactive history list of Chart symbols.")

(defun chart-copy-tree-no-properties (obj)
  "Return a copy of OBJ with no text properties on strings.
OBJ can be a list or other nested structure."
  (cond ((stringp obj)
         (setq obj (copy-sequence obj))
         (set-text-properties 0 (length obj) nil obj)
         obj)
        ((sequencep obj)
         (mapcar 'chart-copy-tree-no-properties obj))
        (t
         obj)))


;;-----------------------------------------------------------------------------
;; subprocess

(defconst chart-protocol-version 11)  ;; see emacs.scm

(defvar chart-process nil
  "The running chart subprocess, or nil if not running.")
(defvar chart-process-timer nil
  "Idle timer to kill chart subprocess when it's unused for a while.")

;; forward references
(defvar chart-completion-symbols-alist)
(defvar chart-latest-cache)
(defvar chart-symlist-alist)
(defvar chart-watchlist-map)
(defvar chart-watchlist-menu)

(defun chart-exec (proc &rest args)
  "Call chart PROC (a symbol) with ARGS (lists, strings, whatever)."

  ;; startup subprocess if not already running
  (unless chart-process
    (when (get-buffer " *chart subprocess*") ;; possible old buffer
      (kill-buffer " *chart subprocess*"))

    (setq chart-process
          (let ((process-connection-type nil)) ;; pipe
            (apply 'start-process
                   "chart" (get-buffer-create " *chart subprocess*")
                   "chart" (if chart-process-debug
                               '("--debug" "--emacs")
                             '("--emacs")))))
    (set-process-coding-system chart-process 'utf-8 'utf-8)
    (set-process-filter chart-process 'chart-process-filter)
    (set-process-sentinel chart-process 'chart-process-sentinel)
    ;; `process-kill-without-query' is "obsolete" in emacs 22 (superceded by
    ;; `set-process-query-on-exit-flag'), but keep using it for emacs 21
    ;; compatibility
    (process-kill-without-query chart-process)
    (buffer-disable-undo (process-buffer chart-process)))

  ;; send this command
  (let ((form (cons proc args)))
    (process-send-string chart-process
                         (concat (prin1-to-string
                                  (chart-copy-tree-no-properties form)) "\n")))
  ;; start or restart idle timer
  (when chart-process-timer
    (cancel-timer chart-process-timer))
  (setq chart-process-timer (run-at-time "5 min" nil 'chart-process-kill)))

(defun chart-incoming-init (codeset protocol-version)
  "Handle chart subprocess init message.
CODESET is always \"UTF-8\".
PROTOCOL-VERSION is the protocol number the suprocess is speaking, to be
matched against `chart-protocol-version'."

  (unless (= protocol-version chart-protocol-version)
    (when (get-buffer "*chart-watchlist*") ;; ignore if gone
      (with-current-buffer "*chart-watchlist*"
        (let ((buffer-read-only nil))
          (erase-buffer)
          (insert (format "Chart program doesn't match this chart.el.

  chart.el protocol:      %s
  chart program protocol: %s

Check your installation.
" chart-protocol-version protocol-version)))))

    (chart-process-kill)
    (error "Chart subprocess protocol version mismatch, got %s want %s"
           chart-protocol-version protocol-version)))

(defun chart-process-filter (proc str)
  "Handle chart PROC subprocess output STR."
  (with-current-buffer (process-buffer proc)
    (goto-char (point-max))
    (insert str)
    (while (progn
             ;; form begins with "(", ignore other diagnostics or whatever
             (goto-char (point-min))
             (skip-chars-forward "^(")
             (delete-region (point-min) (point))

             ;; see if a complete form has arrived
             (let ((form (condition-case nil
                             (read (process-buffer proc))
                           (error nil))))
               (when form
                 (delete-region (point-min) (point))
                 (apply (intern (concat "chart-incoming-"
                                        (symbol-name (car form))))
                        (cdr form)))

               ;; no more processing after `synchronous', let the result get
               ;; back to the caller before further asynch stuff is
               ;; processed (that further stuff deferred under a timer)
               (when (eq (first form) 'synchronous)
                 (run-at-time 0.0000001 nil
                              (lambda ()
                                (chart-process-filter chart-process "")))
                 (setq form nil))

               ;; process another form, perhaps
               form)))))

(defun chart-process-sentinel (proc event)
  "Handle chart PROC subprocess termination, per EVENT string."
  (when (get-buffer "*chart-watchlist*")
    (with-current-buffer "*chart-watchlist*"
      (let ((buffer-read-only nil))
        (save-excursion
          (goto-char (point-min))
          (when (looking-at "\\s-*Starting")
            (erase-buffer))
          (insert (format "\nSubprocess died: %s\n\n" event))))))
  (chart-process-kill)
  (message "Chart subprocess died: %s" event))

(defun chart-process-kill ()
  "Kill chart subprocess."
  (when chart-process-timer
    (cancel-timer chart-process-timer)
    (setq chart-process-timer nil))
  (when chart-process
    ;; clear chart-process variable immediately, xemacs recurses to here
    (let ((p chart-process))
      (setq chart-process nil)
      (set-process-sentinel p nil)
      (set-process-filter p nil)
      (delete-process p)
      (kill-buffer (process-buffer p)))
    ;; go back to uninitialized to force a re-read on the contents if the
    ;; subprocess is restarted, that way any additions while we were away
    ;; will appear
    (setq chart-completion-symbols-alist 'uninitialized)
    (setq chart-symlist-alist 'uninitialized)
    (setq chart-latest-cache (make-hash-table :test 'equal))))


;;-----------------------------------------------------------------------------
;; synchronous commands

(defvar chart-exec-synchronous-seq 0)
(defvar chart-exec-synchronous-got 0)
(defvar chart-exec-synchronous-result nil)

(defun chart-incoming-synchronous (got result)
  "Receive synchronize number GOT from Chart subprocess."
  (setq chart-exec-synchronous-got got)
  (setq chart-exec-synchronous-result result))
  
(defun chart-exec-synchronous (proc &rest args)
  "Call chart PROC (a symbol) with ARGS (lists, strings, whatever).
Return the return value from that call, when it completes."

  (setq chart-exec-synchronous-seq (1+ chart-exec-synchronous-seq))
  (apply 'chart-exec 'synchronous chart-exec-synchronous-seq proc args)

  (while (not (= chart-exec-synchronous-seq
                 chart-exec-synchronous-got)) ;; ignore old abandoned calls
    (if (not (eq 'run (process-status chart-process)))
        (error "Chart process died"))
    (accept-process-output chart-process))

  chart-exec-synchronous-result)


;;-----------------------------------------------------------------------------
;; incoming from subprocess

(defun chart-incoming-update (symbol-list)
  "Receive advice from Chart subprocess that SYMBOL-LIST have updated.
Any in the watchlist are reread, any cached data for `chart-latest' is
discarded."
  (dolist (symbol symbol-list)
    (remhash symbol chart-latest-cache))
  (let ((want-list (chart-intersection (chart-watchlist-symbol-list)
                                       symbol-list)))
    (if want-list
        (chart-exec 'latest-get-list want-list))))

(defun chart-incoming-message (str)
  "Receive a free-form message STR from the Chart subprocess."
  (message "%s" str))

(defun chart-incoming-error (errstr backtrace)
  "Receive an error message from the Chart subprocess.
ERRSTR is a string, BACKTRACE is either a string or nil."
  (when backtrace
    (with-current-buffer (get-buffer-create "*chart-process-backtrace*")
      (let ((follow (= (point) (point-max))))
        (save-excursion
          (goto-char (point-max))
          (insert "-------------------------------------------------------------------------------\n")
          (insert backtrace))
        (when follow
          (goto-char (point-max))))))
  (message "%s" errstr))


;;-----------------------------------------------------------------------------
;; symbols completion

(defconst chart-minibuffer-local-completion-map
  (let ((m (make-sparse-keymap)))
    (set-keymap-parent m minibuffer-local-completion-map)
    (define-key m " " 'self-insert-command)
    m)
  "Inherited `minibuffer-local-completion-map' but with <SPACE> self-inserting.
Chart symbols can contain spaces, so <SPACE> should be an ordinary insert,
not a completion like the default in `minibuffer-local-completion-map'.")

(defvar chart-completion-symbols-alist 'uninitialized
  "Alist of Chart symbols for completing read, or 'uninitialized.
Call function `chart-completion-symbols-alist' instead of reading this
variable, the function gets the list from the Chart subprocess when
'uninitialized.")

(defun chart-completion-symbols-alist ()
  "Return an alist of Chart symbols for completing read.
Currently there's nothing in the `cdr's, it's just ((SYMBOL) (SYMBOL) ...)."
  (when (eq 'uninitialized chart-completion-symbols-alist)
    (chart-with-temp-message "Receiving database symbols ..."
      (setq chart-completion-symbols-alist
            (chart-exec-synchronous 'get-completion-symbols))))
  chart-completion-symbols-alist)

(defun chart-incoming-completion-symbols-update ()
  "Receive advice from Chart subprocess that completion symbols have changed."
  (setq chart-completion-symbols-alist 'uninitialized))

;; emacs 22 has `dynamic-completion-table' to construct a function like
;; this, but emacs 21 and xemacs 21 don't
(defun chart-symbol-completion (str pred all)
  "Chart symbol completion handler, for `completing-read'."
  (cond ((null all)
         (try-completion str (chart-completion-symbols-alist) pred))
        ((eq all t)
         (all-completions str (chart-completion-symbols-alist) pred))
        ((eq all 'lambda)
         (test-completion str (chart-completion-symbols-alist) pred))))

(defun chart-completing-read-symbol (&optional default)
  "Read a Chart symbol using `completing-read'.
Optional DEFAULT is a string."
  (let ((minibuffer-local-completion-map chart-minibuffer-local-completion-map)
        (completion-ignore-case t))
    (if (equal "" default) ;; allow for empty from thing-at-point
        (setq default nil))
    (completing-read (if default
                         (format "Symbol (%s): " default)
                       "Symbol: ")
                     'chart-symbol-completion
                     nil  ;; pred
                     nil  ;; require-match
                     nil  ;; initial-input
                     'chart-symbol-history
                     default)))


;;-----------------------------------------------------------------------------
;; symlist stuff

(defvar chart-symlist-history nil
  "Interactive history list of Chart symlists.")

(defvar chart-symlist-alist 'uninitialized
  "Alist of symlists in Chart.
Each element is (NAME KEY EDITABLE), where NAME is a string, KEY is a
symbol, and EDITABLE is t or nil.  NAME is first so the list can be used
with `completing-read'.

This is the symbol `uninitialized' when data hasn't yet been read.  See the
function `chart-symlist-alist' for reading and initializing.")

(defun chart-symlist-alist ()
  "Return the variable `chart-symlist-alist', initializing it if necessary."
  (when (eq 'uninitialized chart-symlist-alist)
    (chart-with-temp-message "Receiving symlist info ..."
      (chart-exec-synchronous 'get-symlist-alist)))
  chart-symlist-alist)

(defun chart-incoming-symlist-alist (alist)
  "Receive the `chart-symlist-alist' data from Chart subprocess."
  (setq chart-symlist-alist alist)

  ;; freshen name in watchlist, if in use
  (when (get-buffer "*chart-watchlist*")
    (chart-watchlist-update-symlist-name))

  ;; fill Chart menu for watchlist
  (mapcar (lambda (elem)
            (let ((name (first elem))
                  (key  (second elem)))
              (define-key chart-watchlist-menu
                (vector key)
                (cons name `(lambda ()
                              (interactive)
                              (chart-watchlist-symlist ',key))))))
          (reverse chart-symlist-alist)))

(defun chart-symlist-editable-p (key)
  "Return true if symlist KEY (a Lisp symbol) is editable."
  (third (find key (chart-symlist-alist) :key 'second)))


;;-----------------------------------------------------------------------------
;; symlist name completion

;; emacs 22 has `dynamic-completion-table' to construct a function like
;; this, but emacs 21 and xemacs 21 don't
(defun chart-symlist-completion (str pred all)
  "Chart symlist completion handler, for `completing-read'."
  (cond ((null all)
         (try-completion str (chart-symlist-alist) pred))
        ((eq all t)
         (all-completions str (chart-symlist-alist) pred))
        ((eq all 'lambda)
         (test-completion str (chart-completion-symbols-alist) pred))))

;; emacs has `compare-strings' to do this, but xemacs doesn't
(defun chart-string-prefix-ci-p (part str)
  "Return t if PART is a prefix of STR, case insensitive."
  (and (>= (length str) (length part))
       (string-equal (upcase part)
                     (upcase (substring str 0 (length part))))))

(defun chart-completing-read-symlist ()
  "Read a Chart symlist using `completing-read'.
The return is the symlist key (a symbol), eg. 'favourites."
  (let ((minibuffer-local-completion-map chart-minibuffer-local-completion-map)
        (completion-ignore-case t))
    (let ((name (completing-read "Symlist: "
                                 'chart-symlist-completion
                                 nil  ;; pred
                                 t    ;; require-match
                                 nil  ;; initial-input
                                 'chart-symlist-history)))
      ;; completing-read with require-match will return with just a prefix
      ;; of one or more names, go with the first
      (second (assoc* name (chart-symlist-alist)
                      :test 'chart-string-prefix-ci-p)))))


;;-----------------------------------------------------------------------------
;; watchlist funcs

(defvar chart-watchlist-current-symlist 'favourites
  "Current symlist being displayed (it's Scheme `symlist-key').")

(defun chart-watchlist-find (symbol)
  "Move point to line for SYMBOL.
Return true if found, or return nil and leave point unchanged if not found."
  (let ((oldpos (point))
        found)
    (goto-char (point-min))
    ;; continue while not found and can move forward a line
    (while (and (not (setq found (equal symbol (chart-watchlist-symbol))))
                (= 0 (forward-line))))
    (unless found
      (goto-char oldpos))
    found))

(defun chart-watchlist-symbol ()
  "Return symbol on current watchlist line, or nil if none."
  (get-text-property (point-at-bol) 'chart-symbol))

(defun chart-watchlist-symbol-list ()
  "Return list of symbols in watchlist buffer."
  (and (get-buffer "*chart-watchlist*") ;; ignore if gone
       (with-current-buffer "*chart-watchlist*"
         (let (lst)
           (save-excursion
             (goto-char (point-min))
             (while (let ((symbol (chart-watchlist-symbol)))
                      (if symbol
                          (setq lst (cons symbol lst)))
                      (= 0 (forward-line)))))
           (nreverse lst)))))


;;-----------------------------------------------------------------------------
;; watchlist display

(defun chart-incoming-symlist-update (key-list)
  "Receive advice from Chart subprocess that symlists KEY-LIST have updated."
  (if (and (get-buffer "*chart-watchlist*") ;; ignore if gone
           (memq chart-watchlist-current-symlist key-list))
      (chart-exec 'get-symlist chart-watchlist-current-symlist)))

(defun chart-incoming-latest-line-list (lst)
  "Receive LST of latest elements (SYMBOL STR FACE HELP)."
  (when (get-buffer "*chart-watchlist*") ;; ignore if gone
    (with-current-buffer "*chart-watchlist*"
      (chart-save-row-col
        (let ((buffer-read-only nil))
          (dolist (elem lst) ;; elements (SYMBOL STR FACE)
            (when (chart-watchlist-find (first elem))
              (delete-region (point-at-bol) (point-at-eol))
              (insert (propertize (second elem)
                                  'chart-symbol (first elem)
                                  'face (third elem)
                                  'help-echo (fourth elem))))))))))

(defun chart-incoming-symlist-list (symlist symbol-list)
  "Receive SYMLIST contents SYMBOL-LIST from Chart subprocess."
  (when (and (get-buffer "*chart-watchlist*")              ;; ignore if gone
             (eq symlist chart-watchlist-current-symlist)) ;; stray response
    (with-current-buffer "*chart-watchlist*"
      (let (alst need)

        ;; build alst (SYMBOL . LINE-STRING) for existing lines
        (save-excursion
          (goto-char (point-min))
          (while (let ((symbol (chart-watchlist-symbol)))
                   (when symbol
                     (setq alst
                           (acons symbol
                                  (buffer-substring (point)
                                                    (1+ (point-at-eol)))
                                  alst)))
                   (= 0 (forward-line)))))

        ;; fill buffer, and use existing lines from alst
        (let ((buffer-read-only nil))
          (chart-save-row-col
            (erase-buffer)
            (dolist (symbol symbol-list)
              (insert (or (cdr (assoc symbol alst))
                          (progn
                            (setq need (cons symbol need))
                            (propertize (concat symbol "\n")
                                        'chart-symbol symbol)))))
            (unless symbol-list
              (insert (format "\n\n(Empty list, use `%s' to add.)"
                              (key-description
                               (car (where-is-internal
                                     'chart-watchlist-add
                                     chart-watchlist-map))))))))
        (if need
            (chart-exec 'latest-get-list (nreverse need)))))))


;;-----------------------------------------------------------------------------
;; header-line-format hscrolling

(defconst chart-header-line-scrolling-align0
  (propertize " " 'display '((space :align-to 0)))
  "An align-to 0 space string.")

(defvar chart-header-line-scrolling-str nil
  "Buffer local full `header-line-format' string to be hscrolled.")

(defun chart-header-line-scrolling-align ()
  "Return a string which will align to column 0 in a `header-line-format'."
  (if (string-match "^21\\." emacs-version)
      (and (display-graphic-p)
           (concat " "  ;; the fringe
                   (and (eq 'left (frame-parameter nil 'vertical-scroll-bars))
                        "  ")))  ;; left scrollbar
    ;; in emacs 22 and up align-to understands fringe and scrollbar
    chart-header-line-scrolling-align0))

(defun chart-header-line-scrolling-eval ()
  "Install hscrolling header line updates on the windows of the current frame."
  (concat (chart-header-line-scrolling-align)
          (substring chart-header-line-scrolling-str
                     (min (length chart-header-line-scrolling-str)
                          (window-hscroll)))))

(defun chart-header-line-scrolling (str)
  "Set STR as `header-line-format' and make it follow any hscrolling."
  (set (make-local-variable 'chart-header-line-scrolling-str) str)
  (set (make-local-variable 'header-line-format)
       '(:eval (chart-header-line-scrolling-eval))))


;;-----------------------------------------------------------------------------
;; watchlist commands

(defvar chart-watchlist-menu (make-sparse-keymap "Chart")
  "Menu for Chart watchlist.")

(defvar chart-watchlist-map
  (let ((m (make-sparse-keymap)))
    (define-key m "\C-k" 'chart-watchlist-kill-line)
    (define-key m "\C-w" 'chart-watchlist-kill-region)
    (define-key m "\C-y" 'chart-watchlist-yank)
    (define-key m "\C-_" 'chart-watchlist-undo)
    (define-key m "a"    'chart-watchlist-add)
    (define-key m "g"    'chart-watchlist-refresh)
    (define-key m "n"    'next-line)
    (define-key m "q"    'chart-watchlist-quit)
    (define-key m "p"    'previous-line)
    (define-key m "L"    'chart-watchlist-symlist)
    (define-key m "?"    'chart-watchlist-detail)
    (define-key m [menu-bar chart] (cons "Chart" chart-watchlist-menu))
    m)
  "Keymap for Chart watchlist.")

(defun chart-watchlist-want-edit ()
  "Check that the watchlist being displayed is editable."
  (or (chart-symlist-editable-p chart-watchlist-current-symlist)
      (error "This list is not editable")))

(defun chart-watchlist-add (symbol)
  "Add a symbol to the watchlist (after the current one).
SYMBOL is read from the minibuffer, with completion from the database symbols."
  (interactive (progn (chart-watchlist-want-edit)
                      (list (chart-completing-read-symbol))))
  (chart-watchlist-want-edit)
  (unless (chart-watchlist-find symbol)
    (beginning-of-line)
    (forward-line)
    (chart-exec 'symlist-insert chart-watchlist-current-symlist
                (count-lines (point-min) (point))
                (list symbol))
    (let ((buffer-read-only nil))
      (insert (propertize (concat symbol "\n") 'chart-symbol symbol)))
    (forward-line -1)
    (chart-exec 'latest-get-list (list symbol))
    (chart-exec 'request-symbols (list symbol))))

(defun chart-watchlist-detail ()
  "Show detail for this line (stock name and full quote/last times)."
  (interactive)
  (let ((str (get-text-property (point-at-bol) 'help-echo)))
    (if str
        (message "%s" str))))

(defun chart-watchlist-kill-line ()
  "Kill watchlist line into the kill ring."
  (interactive)
  (chart-watchlist-want-edit)
  (let ((buffer-read-only nil))
    (save-excursion
      (beginning-of-line)
      (kill-line 1)))
  (chart-exec 'symlist-delete chart-watchlist-current-symlist
              (count-lines (point-min) (point-at-bol)) 1))

(defun chart-watchlist-kill-region ()
  "Kill watchlist region between point and mark into the kill ring."
  (interactive)
  (chart-watchlist-want-edit)
  (beginning-of-line)
  (let* ((point-row (count-lines (point-min) (point)))
         (mark-bol  (save-excursion (goto-char (mark)) (point-at-bol)))
         (mark-row   (count-lines (point-min) mark-bol)))
    (let ((buffer-read-only nil))
      (kill-region mark-bol (point)))
    (chart-exec 'symlist-delete chart-watchlist-current-symlist
                (min point-row mark-row)
                (abs (- point-row mark-row)))))

(defun chart-watchlist-update-symlist-name ()
  "Update the symlist name shown in the mode line."
  (with-current-buffer "*chart-watchlist*"
    (setq mode-name (concat "Watchlist - "
                            (first (find chart-watchlist-current-symlist
                                         (chart-symlist-alist)
                                         :key 'second))))))

(defun chart-watchlist-symlist (symlist)
  "Select SYMLIST to view."
  (interactive (list (chart-completing-read-symlist)))
  (unless (eq symlist chart-watchlist-current-symlist)
    (chart-exec 'get-symlist symlist))
  (setq chart-watchlist-current-symlist symlist)
  (chart-watchlist-update-symlist-name))

(defun chart-watchlist-yank ()
  "Yank watchlist line."
  (interactive)
  (chart-watchlist-want-edit)
  (let ((symbol-list (mapcar (lambda (line)
                               (get-text-property 0 'chart-symbol line))
                             (split-string (current-kill 0 t) "\n"))))
    (if (memq nil symbol-list)
        (error "Can only yank killed watchlist line(s)"))
    (beginning-of-line)
    (chart-exec 'symlist-insert chart-watchlist-current-symlist
                (count-lines (point-min) (point)) symbol-list)
    (let ((buffer-read-only nil))
      (yank))))

(defun chart-watchlist-undo ()
  "Undo last edit in the watchlist."
  (interactive)
  (chart-watchlist-want-edit)
  (let ((buffer-read-only nil))
    (undo)))

(defun chart-watchlist-refresh (arg)
  "Refresh watchlist quotes.
With a prefix ARG (\\[universal-argument]), refresh only current line.

For the Alerts list, all symbols with alert levels are refreshed, and the
list contents updated accordingly.  (It's not merely those already showing
which are refreshed.)"
  (interactive "P")
  ;; updates from the subprocess with come with an in-progress face, but
  ;; apply that here explicitly to get it shown immediately
  (if arg
      (progn ;; one symbol
        (let ((buffer-read-only nil))
          (add-text-properties (point-at-bol) (point-at-eol)
                               (list 'face 'chart-in-progress)))
        (chart-exec 'request-explicit (list (chart-watchlist-symbol))))
    (progn ;; whole list
      (let ((buffer-read-only nil))
        (add-text-properties (point-min) (point-max)
                             (list 'face 'chart-in-progress)))
      (chart-exec 'request-explicit-symlist chart-watchlist-current-symlist))))

(defun chart-watchlist-quit ()
  "Quit from the watchlist display."
  (interactive)
  (chart-process-kill)
  (kill-buffer (current-buffer)))

;;;###autoload
(defun chart-watchlist ()
  "Chart watchlist display.

\\{chart-watchlist-map}
On a colour screen, faces `chart-up' and `chart-down' show each line in
green or red according to whether the last trade was higher or lower than
the previous close (ie. the change column positive or negative).

Stock name and quote/last-trade times can be seen in a tooltip by moving the
mouse over each line.  The same can be seen on a text terminal with `\\[chart-watchlist-detail]'."

  (interactive)
  (switch-to-buffer (get-buffer-create "*chart-watchlist*"))
  (when (save-excursion
          (goto-char (point-min))
          (or (looking-at "Subprocess died")
              (eq (point-min) (point-max))))

    (setq buffer-read-only nil)
    (erase-buffer)
    (insert "\nStarting chart subprocess ...\n")
    (sit-for 0)
    (kill-all-local-variables)
    (use-local-map chart-watchlist-map)
    (setq major-mode 'chart-watchlist
          mode-name  "Watchlist"
          truncate-lines t
          buffer-read-only t
          chart-watchlist-current-symlist 'favourites
          ;; header-line-format "Symbol       bid/offer     last  change    low    high    volume   when   note"
          )
    (chart-header-line-scrolling "Symbol       bid/offer     last  change    low    high    volume   when   note")
    (add-hook 'kill-buffer-hook 'chart-process-kill t t)
    (make-local-variable 'window-scroll-functions)
    (chart-exec 'get-symlist chart-watchlist-current-symlist)
    (chart-exec 'request-symlist chart-watchlist-current-symlist)
    (chart-exec 'get-symlist-alist)
    (run-hooks 'chart-watchlist-hook)))


;;-----------------------------------------------------------------------------
;; quotes

;; `thing-at-point' setups
;; chart-symbol is alphanumerics plus .,^- and ending in an alphanumeric
;; note no [:alnum:] in xemacs 21
(put 'chart-symbol 'beginning-op
     (lambda ()
       (if (re-search-backward "[^A-Za-z0-9.,^-]" nil t)
           (forward-char)
         (goto-char (point-min)))))
(put 'chart-symbol 'end-op
     (lambda ()
       (unless (re-search-forward "\\=[A-Za-z0-9.,^-]*[A-Za-z0-9]" nil t)
         (goto-char (point-max)))))
(put 'chart-symbol 'thing-at-point
     (lambda ()
       (let ((bounds (bounds-of-thing-at-point 'chart-symbol)))
         (and bounds
              (let ((str (buffer-substring (car bounds) (cdr bounds))))
                ;; suffix in upper case, yahoo reuters news often has lower
                (if (string-match "\\(.*\\)\\([.][^.]*\\)" str)
                    (setq str (concat (match-string 1 str)
                                      (upcase (match-string 2 str)))))
                ;; leading "." assumed to be "^", often seen in yahoo reuters
                (if (string-match "^\\." str)
                    (setq str (concat "^" (substring str 1))))
                str)))))

;;;###autoload
(defun chart-quote (symbol)
  "Show a quote in the message area for the Chart stock SYMBOL.
Interactively SYMBOL is read from the minibuffer, the default is the symbol
at point in the current buffer."
  (interactive (list (chart-completing-read-symbol (thing-at-point
                                                    'chart-symbol))))
  (message "Fetching ...")
  (let ((elem (chart-exec-synchronous 'quote-one symbol)))
    (message "%s" (propertize (second elem) 'face (third elem)))))

;;;###autoload
(defun chart-quote-at-point ()
  "Show a quote in the message area for the Chart stock symbol at point."
  (interactive)
  (chart-quote (thing-at-point 'chart-symbol)))


;;-----------------------------------------------------------------------------
;; latest from elisp

(defvar chart-latest-cache
  (make-hash-table :test 'equal :weakness 'value)
  "Hash table, key is a chart symbol (a string), value is a latest record.")

;;;###autoload
(defun chart-latest (symbol &optional field scale)
  "Return the latest price for SYMBOL (a string) from Chart.
If there's no information available (an unknown stock, not online and
nothing cached, or whatever) the return is nil.

FIELD is a symbol (a Lisp symbol) for what data to return.  The default is
`last' which is the last price, other possibilities are: name, bid, offer,
open, high, low, change, volume, decimals, note.  Which fields actually have
data depends on the data source.

SCALE is a power of 10 to apply to prices.  For example if SCALE is 2 then a
price 1.23 is returned as 123.  This can be useful for instance if you want
to work in cents but quotes are in dollars.  The default is 0, for no
scaling.

FIELD `decimals' is how many decimal places Chart is using for prices
internally.  Internally prices are kept as an integer and count of decimals.
Using this value for SCALE will ensure an integer return.

FIELD `name' is the stock or commodity name as a string, or nil.  FIELD
`note' is a string with extra notes, like ex-dividend or limit up, or
nil."

  (unless (stringp symbol)
    (error "Not a chart symbol (should be a string)"))
  (unless field
    (setq field 'last))
  (let ((latest (gethash symbol chart-latest-cache)))
    (when (or (not latest)
              (plist-get latest 'dirty))
      (setq latest (chart-exec-synchronous 'get-latest-record symbol))
      (puthash symbol latest chart-latest-cache))

    (let ((value (plist-get latest field)))
      (when (and value
                 (memq field '(bid offer open high low last change)))
        (let ((factor (- (or scale 0) (plist-get latest 'decimals))))
          (if (/= 0 factor)
              (setq value (* value (expt (if (< factor 0) 10.0 10)
                                         factor))))))
      value)))
(put 'chart-latest 'safe-function t)


;;-----------------------------------------------------------------------------
;; ses.el additions

;;;###autoload
(defun chart-ses-refresh ()
  "Refresh Chart prices in a SES spreadsheet.
`ses-recalculate-all' is run first to find what `chart-latest' prices are
required, then it's run again after downloading new data for those.

If the second recalculate uses prices for further symbols (perhaps through
tricky conditionals), then those are downloaded and the recalculate done
again.  This is repeated until all prices used have been downloaded."

  (interactive)
  (let (chart-ses-recalculate--fetched)
    (while
        (let ((chart-ses-recalculate--orig-chart-latest
               (symbol-function 'chart-latest))
              chart-ses-recalculate--need)
          (flet ((chart-latest
                  (symbol &rest args)
                  (unless (member symbol chart-ses-recalculate--fetched)
                    (add-to-list 'chart-ses-recalculate--need symbol))
                  (apply chart-ses-recalculate--orig-chart-latest symbol
                         args)))
            (ses-recalculate-all))
          (and chart-ses-recalculate--need
               (progn
                 (chart-with-temp-message "Downloading quotes ..."
                   (chart-exec-synchronous 'request-symbols-synchronous
                                           chart-ses-recalculate--need))
                 (setq chart-ses-recalculate--fetched
                       (nconc chart-ses-recalculate--need
                              chart-ses-recalculate--fetched))
                 t))))))


;;-----------------------------------------------------------------------------

(provide 'chart)

;;; chart.el ends here
