;; Centre of Gravity Oscillator (COG).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart cog)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.mesasoftware.com/technicalpapers.htm
;; http://www.mesasoftware.com/Papers/The%20CG%20Oscillator.exe
;;     Paper by John Elhers, self-extracting zip archive of MS word.
;;
;; http://www.linnsoft.com/tour/techind/cog.htm
;;     Sample Intel (INTC) 2001/2.
;;
;; http://www.working-money.com/Documentation/FEEDbk_docs/Archive/082002/TradersTips/TradersTips.html
;;     Trader's tips August 2002, formulas from May 2002.
;;


;; Return a procedure which calculates the centre of gravity oscillator over
;; an accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the COG up to (and including) that point, or #f if not enough values
;; yet.
;;
;; The window can be primed by calling with COUNT-1 many values preceding
;; the first desired.  `cog-warmup-count' below returns that value.
;;
(define-public (cog-calc-proc count)

  ;; DATA is current points accumulated, or #f in empty positions
  ;;
  ;; DEN is the sum of the values in DATA.  This is exact if all the X
  ;; inputs are exact.
  ;;
  ;; NUM is the sum D[1]*1 + D[2]*2 + ... + D[POINTS]*POINTS, where D[1] is
  ;; the most recent value which has come in.  This is exact if all the X
  ;; inputs are exact.
  ;;
  (let ((data       (make-circular-list count #f))
	(den        0)
	(num        0)
	(points     0)
	(min-points (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (x)

      (if (car data)
	  ;; drop oldest point
	  (begin
	    (set! num (- num (* count (car data))))
	    (set! den (- den (car data))))

	  ;; otherwise gaining a point
	  (set! points (1+ points)))

      ;; replace oldest with new point
      (set-car! data x)
      (set! data (cdr data))

      ;; step existing points in num
      (set! num (+ num den))

      ;; add new point
      (set! num (+ num x))
      (set! den (+ den x))

      (and (>= points min-points)
	   (if (zero? den)
	       1
	       ;; no fracs in the output
	       (- (/ (exact->inexact num) den)))))))

(define-public (cog-warmup-count count)
  (1- count))


(define (cog-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (warmup (cog-warmup-count count))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count warmup))
	 (array  (series-array parent lo hi))
	 (proc   (cog-calc-proc count)))

    ;; use high/low midpoints

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(proc (arithmetic-mean (or (array-ref array i 1) close)     ;; high
			       (or (array-ref array i 2) close))))) ;; low

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array (proc (arithmetic-mean
				     (or (array-ref array i 1) close)   ;; high
				     (or (array-ref array i 2) close))) ;; low
		    i)))))

(define-public (cog-series series count)
  (ohlcv-required "COG" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "CG") count))
    (series-datatype!  new-series cog-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   cog-fill (cog-warmup-count count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public cog-datatype
  (datatype-new!
   #:key    'cog
   #:name   (_ "Centre of Gravity")
   #:input  datatype-ohlcv
   #:units  'cog
   #:manual (_ "CONTEXT(manual-node): Centre of Gravity Oscillator")
   #:proc   cog-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'cog-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
