;; HTTP cookie functions.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart cookies)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart http)
  #:use-module (chart misc))


;; Crib notes:
;;
;; http://wp.netscape.com/newsref/std/cookie_spec.html
;;     Netscape cookies draft.
;; RFC-2109
;;     Cookies version 1.
;; RFC-2965
;;     Cookies version 1 revised, Set-Cookie2 header.



;;-----------------------------------------------------------------------------
;; misc

(define (ymdhms-gmt->time_t year month day hour minute second)
  (let ((tm (ymd->tm year month day)))
    (set-tm:hms tm hour minute second)
    (car (mktime tm "GMT"))))

(define (string-match-ci pattern str)
  (regexp-exec (make-regexp pattern regexp/icase) str))


;;-----------------------------------------------------------------------------
;; cookie record

(let-toplevel ((rec-type (make-record-type "http-cookie"
					   '(name value
						  expires
						  domain
						  path
						  ctime))))
  (define rec-new (record-constructor rec-type))

  (define-public cookie-name     (rec-accessor rec-type 'name))
  (define-public cookie-value    (rec-accessor rec-type 'value))
  (define-public cookie-expires  (rec-accessor rec-type 'expires))
  (define-public cookie-domain   (rec-accessor rec-type 'domain))
  (define-public cookie-path     (rec-accessor rec-type 'path))
  (define-public cookie-ctime    (rec-accessor rec-type 'ctime))

  (define*-public (cookie-new #:key
			      name value expires domain
			      (path "/")
			      (ctime (current-time)))
    (rec-new name value expires domain path ctime)))

;; STR is the value part of a Set-Cookie header line
;; Eg. "foo=bar; path=/zz; domain=somewhere.com"
;; return a cookie record
;;
(define-public (cookie-parse str url)
  (let ((name    #f)
	(value   #f)
	(expires #f)
	(domain  #f)
	(path    (http-url-path url)))

    (let ((m (must-match (string-match "[ \t]*([^=;, \t]+)=([^;, \t]+)(;(.*))?" str))))
      (set! name  (match:substring m 1))
      (set! value (match:substring m 2))
      (set! str   (or (match:substring m 4) "")))

    (for-each
     (lambda (part)
       (set! part (string-trim-both part))

       (cond
	;; old netscape spec allows 2-digit years
	((string-match-ci "^expires=[A-Za-z]+, ([0-9]+)-([A-Za-z]+)-([0-9][0-9]([0-9][0-9])?) ([0-9]+):([0-9]+):([0-9]+) GMT" part)
	 => (lambda (m)
	      (set! expires
		    (ymdhms-gmt->time_t
		     (yyy-str->year (match:substring m 3))
		     (Mmm-str->month (match:substring m 2))
		     (string->number (match:substring m 1))
		     (string->number (match:substring m 4))
		     (string->number (match:substring m 5))
		     (string->number (match:substring m 6))))))

	((string-match-ci "^path=(.*)" part)
	 => (lambda (m)
	      (set! path (match:substring m 1))))))

     (string-split str #\;))

    (cookie-new #:name    name
		#:value   value
		#:expires expires
		#:domain  domain
		#:path    path)))


;;-----------------------------------------------------------------------------
;; cookie jar files

;; return a list of cookie records
(define-public (cookiejar-read filename)
  (map! (lambda (lst)
	  (apply cookie-new lst))
	(call-with-input-file-or '() filename read)))

;; write a list LST of cookie records to FILENAME
(define (cookiejar-write filename lst)
  (database-call-with-output-file filename
    (lambda (port)
      (write-list-lines (map (lambda (cookie)
			       (list #:name    (cookie-name    cookie)
				     #:value   (cookie-value   cookie)
				     #:expires (cookie-expires cookie)
				     #:domain  (cookie-domain  cookie)
				     #:path    (cookie-path    cookie)
				     #:ctime   (cookie-ctime   cookie)))
			     lst)
			port))))

;; return a http "Cookie:" header line, including CRLF, based on the cookies
;; in FILENAME
;; URL is the destination, its path restricts the cookies sent
;;
(define-public (cookiejar-get-line filename url)
  (let* ((lst (cookiejar-read filename)))

    (let ((path (http-url-path url)))
      (set! lst (filter (lambda (cookie)
			  (string-prefix? (cookie-path cookie) path))
			lst)))

    (set! lst (sort lst cookie-send<?))

    (if (null? lst)
	""
	(string-append "Cookie: "
		       (string-join
			(map (lambda (cookie)
			       (string-append (cookie-name cookie)
					      "="
					      (cookie-value cookie)))
			     lst)
			"; ")
		       "\r\n"))))

;; return true if cookie X should be sent before cookie Y
;; a cookie with a more specific path (ie. longer one) must go out before a
;; less specific
;; (per RFC-2109 4.3.4 Sending Cookies to the Origin Server)
;;
(define (cookie-send<? x y)
  (> (string-length (cookie-path x))
     (string-length (cookie-path y))))

;; add any cookies in HEADERS, which comes from URL, to the given FILENAME
(define-public (cookiejar-add-headers filename url headers)
  (let* ((str-list    (http-headers-get-list headers "Set-Cookie"))
	 (cookie-list (map! (lambda (str)
			      (cookie-parse str url))
			    str-list)))
    (cookiejar-add-cookies filename cookie-list)))

;; add all cookie records in COOKIE-LIST to FILENAME
(define-public (cookiejar-add-cookies filename cookie-list)
  (if (not (null? cookie-list))
      (database-call-with-lock (dirname filename)
	(lambda ()
	  (let ((lst (cookiejar-read filename)))
	    (for-each (lambda (cookie)
			(set! lst (cookiejar-list-add-cookie lst cookie)))
		      cookie-list)
	    (cookiejar-write filename lst))))))

;; LST is a list of cookies in a jar, add COOKIE to it, replacing any
;; existing cookie with the same name+path
(define (cookiejar-list-add-cookie lst cookie)
  (set! lst (remove! (lambda (c)
		       (and (equal? (cookie-name cookie) (cookie-name c))
			    (equal? (cookie-path cookie) (cookie-path c))))
		     lst))
  (if #t ; (> (cookie-expires cookie) (current-time))
      (set! lst (cons cookie lst)))
  lst)

;; look at HEADERS for any fields called NAME, and return a list of the
;; value parts of those fields (an empty list if there's no NAME fields)
;; each value has leading and trailing whitespace stripped, but CRLFs
;; between continuation lines remain
;;
;; the regexp matches the value part of a header field up to CRLF, plus
;; continuation on further lines beginning with space or tab
;;
(define (http-headers-get-list headers name)
  (map (lambda (m)
	 (string-trim-both (match:substring m 1)))
       (list-matches (make-regexp (string-append "^" (regexp-quote name)
						 ":((.*\n[ \t])*.*)\n")
				  regexp/newline)
		     headers)))
	       

;; delete any cookies from FILENAME which are SECONDS or longer old
(define-public (cookiejar-expire filename seconds)
  (if (not (null? (cookiejar-read filename))) ;; no write if empty already
      (database-call-with-lock (dirname filename)
	(lambda ()
	  (let ((lst    (cookiejar-read filename))
		(target (- (current-time) seconds)))
	    (set! lst (remove! (lambda (cookie)
				 (< (cookie-ctime cookie) target))
			       lst))
	    (cookiejar-write filename lst))))))

