;; Cross drawing.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart cross)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (srfi srfi-1)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart keymap)
  #:use-module (chart main)
  #:use-module (chart misc)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; GtkWidget extras

;; return true if WIDGET contains the pointer
(define (c-gtk-widget-contains-pointer? widget)
  (receive-list (x y)
      (gtk-widget-get-pointer widget)
    (and (>= x 0)
	 (>= y 0)
	 (let ((size (gdk-drawable-get-size (gtk-widget-window widget))))
	   (and (< x (car size))
		(< y (cdr size)))))))


;;-----------------------------------------------------------------------------
;; misc

;; Return the number of extra decimal places needed to get to pixel
;; resolution on GRAPH.  This extra is on top of the decimals in the series
;; in GRAPH.  It can be 0 if the series decimals are enough already.
;;
(define (graph-pixel-extra-decimals graph)
  (max 0
       (ceiling-exact (- (log10 (- (graph-y->price-inexact graph 0)
				   (graph-y->price-inexact graph 1)))))))


;;-----------------------------------------------------------------------------
;; cross
;;
;; Done without grabs, so the chart can be changed or scaled or whatever
;; with normal events and menus.
;;
;; pointer-motion-hint-mask is used to keep down network traffic.  The
;; drawing itself seems fast enough not to need any special attention, even
;; on a slow video card.
;;
;; It's assumed a "y" value in one graph suits all in that view.  Could go
;; via x->tdate if that might not be true one day.

(define-public cross-state  #f)
(define-public cross-cursor #f)
(define-public cross-colour "orange")


(define-public cross-status-date-alist
  `((,timebase-days  . ,(locale-wd-fmt))
    (,timebase-weeks . ,(locale-d-fmt))
    (#f              . "%b-%Y")))

(define-public (cross-status-format graph t price)
  (let ((series (graph-series graph)))
    (if series
	(let ((timebase (series-timebase series))
	      (decimals (series-decimals (graph-series graph))))

	  (let ((extra-decimals (graph-pixel-extra-decimals graph)))
	    (set! decimals (+ decimals extra-decimals))
	    (set! price    (* price (integer-expt 10 extra-decimals))))
	  (set! price (round-exact price))

	  (format #f "~a  ~a"
		  (strftime-utf8
		   (or (assq-ref cross-status-date-alist timebase)
		       (assq-ref cross-status-date-alist #f)
		       (locale-d-fmt))
		   (timebase-t->tm timebase t))
		  (locale-number+decimals price decimals)))
	"")))


(define x       #f) ;; currently drawn x/y coords
(define y       #f)
(define y-graph #f) ;; graph containing y (and hence the horiz line)
(define xy-view #f) ;; view widget containing these
(define cross-gc      #f)
(define button-number #f)

(define (cross-xor-one graph)
  (if x
      (let* ((win    (gtk-widget-window graph))
	     (width  (c-gdk-drawable-get-width win))
	     (height (c-gdk-drawable-get-height win)))

	(if (not cross-gc)
	    (let ((c  (gdk-color-new))
		  (bg (gtk-style-bg (gtk-widget-get-style graph) 'normal)))
	      (set! cross-colour (gdk-color-intern cross-colour))
	      (gdk-color-set-pixel c (logxor (gdk-color-pixel cross-colour)
					     (gdk-color-pixel bg)))
	      (set! cross-gc (c-gdk-gc-get win
					   #:foreground c
					   #:dashed #t
					   #:function 'xor))))

	(gdk-draw-line win cross-gc x 0 x height)
	(if (eq? graph y-graph)
	    (gdk-draw-line win cross-gc 0 y width y)))))

(define (cross-xor)
  (if x
      (for-each cross-xor-one (view-graph-list xy-view))))

(define (cross-drag-release event)
  (if (and button-number
	   (= button-number (gdk-event-button event)))
      (begin
	(set! button-number #f)
	(cross-disable))))

(define (cross-move-or-enter graph)
  ;; must call gdk-window-get-pointer on any motion_notify event since we're
  ;; using hints
  (let ((pointer (gdk-window-get-pointer (gtk-widget-window graph))))
    ;; but ignore events queued after a disable
    (if cross-state
	(begin
	  (cross-xor)  ;; undraw
	  (set! y-graph graph)
	  (set! x (second pointer))
	  (set! y (third  pointer))
	  ;; draw
	  (cross-xor)
	  (view-message "~a" (cross-status-format
			      graph
			      (graph-x->date y-graph x)
			      (graph-y->price-inexact y-graph y)))))))


(define (init)
  (set! xy-view outer)
  (for-each (lambda (graph)
	      (define (leave event)
		(if cross-state ;; ignore events queued after a disable
		    (begin
		      (cross-xor)
		      (view-message #f)
		      (set! x #f))))

	      (define (move-or-enter event)
		(cross-move-or-enter graph))
	      (gtk-signal-connect graph "motion_notify_event" move-or-enter)
	      (gtk-signal-connect graph "enter_notify_event"  move-or-enter)
	      (gtk-signal-connect graph "leave_notify_event"  leave)
	      (gtk-signal-connect graph "button_release_event"
		cross-drag-release))

	    (view-graph-list outer))

  ;; once only, and gc this code
  (set! init noop))

(define-public (cross-drag-button)
  (init)
  (set! button-number (gdk-event-button keymap-event))
  (cross-enable))

(graph-element-new #:xor #t
		   #:proc (lambda (graph)
			    (if cross-state
				(cross-xor-one graph))))



(define-public (cross-set-state! state)
  (if (not (eq? state cross-state))
      (begin
	(init)
	(set! cross-state state)
	(if cross-state
	    (begin
	      ;; enable
	      (for-each
	       (lambda (graph)
		 (c-gtk-widget-add-events graph 'cross
					  '(pointer-motion-mask
					    pointer-motion-hint-mask
					    button-motion-mask
					    enter-notify-mask
					    leave-notify-mask))

		 ;; cursor is 1x1 of `cross-colour', which in effect means
		 ;; no cursor shown
		 (or cross-cursor
		     (let ((cursor-pixmap (gdk-pixmap-new
					   (gtk-widget-window graph) 1 1 1)))
		       (set! cross-cursor
			     (gdk-cursor-new-from-pixmap
			      cursor-pixmap cursor-pixmap
			      cross-colour
			      (gtk-style-bg (gtk-widget-get-style graph)
					    'normal)
			      0 0))))

		 (gdk-window-set-cursor (gtk-widget-window graph)
					cross-cursor))
	       (view-graph-list xy-view))
	      (set! x #f)
	      (cond ((find c-gtk-widget-contains-pointer?
			   (view-graph-list xy-view))
		     => cross-move-or-enter)))
	    (begin
	      ;; disable
	      (cross-xor)
	      (view-message #f)
	      (for-each
	       (lambda (graph)
		 (c-gtk-widget-remove-events graph 'cross)
		 (gdk-window-set-cursor (gtk-widget-window graph) #f))
	       (view-graph-list xy-view))))

	(notify-send-locally 'cross-update cross-state))))

(define-public (cross-enable)
  (cross-set-state! #t))

(define-public (cross-disable)
  (cross-set-state! #f))

(define-public (cross-toggle)
  (cross-set-state! (not cross-state)))

(define-public cross-keymap (keymap-new (_ "While cross is active")
					(lambda () cross-state)))


(keymap-set! cross-keymap "Escape" cross-disable)
