/* Chart curses interface C code.

   Copyright 2006, 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"

#include <stdio.h>     /* for NULL */
#include <string.h>

#include <libguile.h>

#if HAVE_CURSES

#include <curses.h>
#include <term.h>

#include "compat.h"


int setupterm_done;

SCM_SYMBOL (sym_curses, "curses");


SCM_DEFINE_PUBLIC (chart_tput_available_p, "tput-available?", 0, 0, 0,
            (),
	    "Return true if @code{tput} to print curses control strings is\n"
	    "available.")
#define FUNC_NAME s_chart_tput_available_p
{
  return SCM_BOOL_T;
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_setupterm, "setupterm", 0, 1, 0,
            (SCM term),
	    "Initialize the curses library for output to a terminal of type\n"
	    "@var{term} (a string) on the current output port.  The default\n"
	    "for @var{term} is the @env{TERM} environment variable.\n"
	    "\n"
	    "The return value is @code{#f} if successful, or a string\n"
	    "describing the problem if not.")
#define FUNC_NAME s_chart_setupterm
{
  char *c_term;
  int c_fd, ret, err;

  if (setupterm_done)
    {
      del_curterm (cur_term);
      setupterm_done = 0;
    }

  c_fd = scm_to_int (scm_fileno (scm_current_output_port ()));

  if (SCM_UNBNDP (term) || scm_is_false (term))
    c_term = NULL;
  else
    c_term = scm_to_locale_string (term);

  ret = setupterm (c_term, c_fd, &err);

  if (c_term)
    free (c_term);

  if (ret == ERR)
    {
      switch (err) {
      case 1:  return scm_from_locale_string ("hardcopy terminal");
      case 0:  return scm_from_locale_string ("unknown terminal type");
      case -1: return scm_from_locale_string ("no terminal database");
      }
      return scm_list_2 (scm_from_locale_string ("Cannot initialize terminal"),
                         scm_from_int (err));
    }

  setupterm_done = 1;
  return SCM_BOOL_F;
}
#undef FUNC_NAME


static int
chart_tput_putc (int c)
{
  scm_display (SCM_MAKE_CHAR (c), SCM_UNDEFINED);
  scm_force_output (scm_current_output_port ());
  return c;
}

SCM_DEFINE_PUBLIC (chart_tput, "tput", 1, 1, 0,
            (SCM capname, SCM arg),
	    "Print a curses control string for capability @var{cap} (a\n"
	    "string) to the current output port.  @var{arg} is an optional\n"
	    "numeric argument, for those @var{cap} capilities which require\n"
	    "an argument.")
#define FUNC_NAME s_chart_tput
{
  char *c_capname, *tistr, *outstr;
  SCM column;

  if (! setupterm_done)
    if (scm_is_true (chart_setupterm (SCM_UNDEFINED)))
      return SCM_BOOL_F;

  c_capname = scm_to_locale_string (capname);
  tistr = tigetstr (c_capname);
  free (c_capname);

  if (tistr == (char *) -1)
    scm_throw (sym_curses, scm_list_2 (scm_from_locale_string
                                       ("tigetstr not a string capability"),
                                       capname));

  if (tistr == NULL)  /* cancelled or absent from info */
    return SCM_BOOL_F;

  if (SCM_UNBNDP (arg))
    outstr = tparm (tistr);
  else
    outstr = tparm (tistr, scm_to_int (arg));
  if (outstr == NULL)
    scm_throw (sym_curses, scm_list_1 (scm_from_locale_string
                                       ("tparm cannot instantiate")));


  column = scm_port_column (scm_current_output_port ());
  scm_force_output (scm_current_output_port ());
  if (tputs (outstr, 1, chart_tput_putc) == ERR)
    scm_throw (sym_curses, scm_list_1 (scm_from_locale_string
                                       ("Cannot print with tputs")));
  scm_set_port_column_x (scm_current_output_port (), column);

  return SCM_BOOL_T;
}
#undef FUNC_NAME


#endif /* HAVE_CURSES */


/*---------------------------------------------------------------------------*/

void
chart_curses_init (void)
{
#ifndef SCM_MAGIC_SNARFER
#include "curses.x"
#endif
}
