;; Darvas boxes.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart darvas)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.gerryco.com/tech/darvas.html
;;     State diagram.
;;
;; http://www.linnsoft.com/tour/techind/darvas.htm
;;     Sample MSFT August 2002, using close-must-penetrate for breakouts.
;;
;; http://www.guppytraders.com/gup206.htm
;;


;; Darryl Guppy's modified Darvas (not implemented here):
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/052005/Abstracts_new/Guppy/guppy.html
;;     TASC May 2005 intro, rest for sale.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/062005/TradersTips/TradersTips.html
;;     Trader's Tips June 2005.



;; A darvas series has two columns, which are high and low of the box, or #f
;; if no box at that point.
;;
(define (darvas-fill series lo hi)

  ;; do everything
  (set! lo (series-lo series))
  (set! hi (series-hi series))

  (let* ((new-array (make-array #f (list lo hi) '(0 1)))
	 (parent    (series-parent series))
	 (array     (series-array parent lo hi))
	 (box-start #f)  ;; date pos
	 (box-high  #f)
	 (box-low   #f)
	 (state     0))

    (define (fill-box box-end)
      (do ((i box-start (1+ i)))
	  ((> i box-end))
	(array-set! new-array box-high i 0)
	(array-set! new-array box-low  i 1)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(let ((high (or (array-ref array i 1) close))
	      (low  (or (array-ref array i 2) close)))

	  (case state
	    ((0)  ;; initial
	     (set! box-high high)
	     (set! box-start i)
	     (set! state 1))

	    ((1)
	     (if (> high box-high)
		 (begin                   ;; broke, stay here
		   (set! box-high high)
		   (set! box-start i))
		 (set! state 2)))         ;; held, go onwards

	    ((2)
	     (if (> high box-high)
		 (begin                   ;; broke, back to 1
		   (set! box-high high)
		   (set! box-start i)
		   (set! state 1))
		 (begin                   ;; held, go onwards
		   (set! box-low low)
		   (set! state 3))))

	    ((3)
	     (cond ((> high box-high)     ;; broke high, back to 1
		    (set! box-high high)
		    (set! box-start i)
		    (set! state 1))
		   ((< low box-low)       ;; broke low, stay here
		    (set! box-low low))
		   (else                  ;; held, go onwards
		    (set! state 4))))

	    ((4)
	     (cond ((> high box-high)     ;; broke high, back to 1
		    (set! box-high high)
		    (set! box-start i)
		    (set! state 1))
		   ((< low box-low)       ;; broke low, back to 3
		    (set! box-low low)
		    (set! state 3))
		   (else                  ;; held, go onwards
		    (set! state 5))))

	    ((5)
	     (if (or (> high box-high)
		     (< low box-low))
		 ;; (or (> close box-high)  ;; close-must-penetrate style
		 ;;     (< close box-low))
		 (begin                   ;; break either way, stop box
		   (fill-box i)
		   (set! box-high high)
		   (set! box-start i)
		   (set! state 1))))))))

    (if (= 5 state)
	(fill-box hi))

    new-array))

(define-public (darvas-series series)
  (ohlcv-required "Darvas" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "Darvas"))
    (series-datatype!  new-series darvas-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series darvas-fill)
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public darvas-datatype
  (datatype-new!
   #:key       'darvas
   #:name      (_ "Darvas Boxes")
   #:units     #f
   #:input     datatype-ohlcv
   #:linestyle 'stops
   #:manual    (_ "CONTEXT(manual-node): Darvas Boxes")
   #:proc      darvas-series
   #:closes-apply-arg '(0)))
