;; Data editing.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart data-edit)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:autoload   (chart cross)      (cross-disable)
  #:autoload   (chart database)   (database-add-series)
  #:use-module (chart graph)
  #:use-module (chart keymap)
  #:use-module (chart misc)
  #:use-module (chart view))


;; This is some rough code for editing price data in the database.  It's
;; only meant for cooking up some likely shapes to screenshots of or
;; whatever.  There's no button bindings for this by default, but it can be
;; put for instance on button 1 with the following in your ~/Chart/gui.scm
;;
;;     (use-modules (chart data-edit))
;;     (keymap-set! app-keymap "Pointer_Button1" data-edit-button)
;;
;; It's worth increasing the horizontal scale to make it big enough to see
;; while dragging, with say
;;
;;     (set! scale-x-step 24)
;;
;; Editing is meant to be done in OHLC figures mode.  Press the bound mouse
;; button and drag to move the open or close notch, or the high or low end
;; of the vertical bar.


;;-----------------------------------------------------------------------------
;; cross state

(let-toplevel ((state #f))
  (notify-connect 'cross-update (lambda (new-state)
				  (set! state new-state)))
  ;; do a `cross-disable', but avoid loading the cross module if it isn't
  ;; already in use
  (define-toplevel (data-edit-cross-disable)
    (if state
	(cross-disable))))


;;-----------------------------------------------------------------------------
;; data editing

(define data-edit-cursor (gdk-cursor-new 'left-ptr))

(define data-edit-state #f)
(define x-edit   #f)
(define t-edit   #f)
(define t-idx    #f)
(define p-new    #f)
(define x-edit   #f)  ;; x,y offset applied for diff between
(define y-offset #f)  ;;   initial pointer and initial t,p

(define data-edit-gc #f)
(define data-edit-colour "orange")

(define (data-edit-xor)
  (let* ((win (gtk-widget-window draw)))

    (if (not data-edit-gc)
	(let ((c  (gdk-color-new))
	      (bg (gtk-style-bg (gtk-widget-get-style draw) 'normal)))

	  (set! data-edit-colour (gdk-color-intern data-edit-colour))
	  (gdk-color-set-pixel c (logxor (gdk-color-pixel data-edit-colour)
					 (gdk-color-pixel bg)))
	  (set! data-edit-gc (gdk-gc-new-with-values
			      win
			      #:foreground c
			      #:graphics-exposures #f
			      #:function 'xor))))

    (let ((y-new (graph-scale-y draw p-new)))
      (gdk-draw-line win data-edit-gc (- x-edit 2) y-new (+ x-edit 2) y-new))))

(define (data-edit-move x y)
  (if data-edit-state
      (begin
	(data-edit-xor)
	(set! p-new (round-exact (graph-y->price draw (+ y-offset y))))
	(data-edit-xor))))


(define (init)
  (gtk-widget-add-events draw '(button-motion-mask
				pointer-motion-hint-mask))
  (gtk-signal-connect draw "motion_notify_event"
    (lambda (event)
      (let ((values (gdk-window-get-pointer
		     (gtk-widget-window draw))))
	(data-edit-move (inexact->exact (second values))
			(inexact->exact (third  values))))))
  (gtk-widget-add-events draw '(button-release-mask))
  (gtk-signal-connect draw "button_release_event"
    (lambda (event)
      (data-edit-move (inexact->exact (gdk-event-x event))
		      (inexact->exact (gdk-event-y event)))
      (data-edit-done)))

  ;; once only, and gc code
  (set! init noop))

(define (series-within-range? series t)
  (within-range? (series-lo series) (series-hi series) t))

(define-public (data-edit-button)
  (init)
  (data-edit-cross-disable)
  (gdk-window-set-cursor (gtk-widget-window draw) data-edit-cursor)

  (set! x-edit (inexact->exact (gdk-event-x keymap-event)))
  (let* ((series   (graph-series draw))
	 (y        (inexact->exact (gdk-event-y keymap-event)))
	 (scale-x  (graph-scale-x-proc draw))
	 (scale-y  (graph-scale-y-proc draw)))
    (set! data-edit-state #t)

    (set! t-edit   (graph-x->tdate draw x-edit))

    (set! t-idx (round-exact (* 3.0 (/ (- x-edit (scale-x t-edit))
				       scale-x-step))))
    (case t-idx
      ((1)
       (and-let* ((      (series-within-range? series t-edit))
		  (array (series-array series t-edit t-edit)))
	 (let ((h (array-ref array t-edit 1))
	       (l (array-ref array t-edit 2))
	       (p (graph-y->price draw y)))
	   (cond ((and h l (within-range? l h p))
		  (set! t-idx #f))
		 (else
		  (and-let* ((mid (or (and h l (/ (+ h l) 2.0))
				      h
				      l
				      (array-ref array t-edit 3)
				      (array-ref array t-edit 0))))
		    (if (< p mid)
			(set! t-idx 2))))))))
      ((2)
       (set! t-idx 3)))

    (let ((p-old (and (series-within-range? series t-edit)
		      (array-ref (series-array series t-edit t-edit)
				 t-edit (or t-idx 1)))))
      (set! y-offset (- (scale-y p-old) y))
      (set! p-new    p-old))))

(define-public (data-edit-done)
  (if data-edit-state
      (let ((array  (make-array #f (list t-edit t-edit) '(0 4)))
	    (series (series-parent-top (graph-series draw))))

	(if t-idx
	    (array-set! array p-new t-edit t-idx)

	    (let* ((old-array (series-array series t-edit t-edit))
		   (add       (- p-new (array-ref old-array t-edit 1))))
	      (do ((i 0 (1+ i)))
		  ((> i 3))
		(array-set! array
			    (and (array-ref old-array t-edit i)
				 (+ add (array-ref old-array t-edit i)))
			    t-edit i))))

	(database-add-series
	 (series-new #:symbol   (series-symbol series)
		     #:decimals (series-decimals series)
		     #:array array
		     #:datatype datatype-ohlcv))
	(gdk-window-set-cursor (gtk-widget-window draw) #f)
	(set! data-edit-state #f))))

(define-public (data-edit-abort)
  '(if data-edit-state
      (begin
	(data-edit-xor)
	(set! new-line data-edit-state)
	(if new-line
	    (data-edit-xor))
	(gdk-window-set-cursor (gtk-widget-window draw) #f)
	(set! data-edit-state #f))))

(define-public data-edit-keymap (keymap-new (_ "While moving a line")
					    (lambda () data-edit-state)))
(keymap-set! data-edit-keymap "Escape" data-edit-abort)
