;; Database functions.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart database)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart extra-glib)
  #:use-module (chart misc)
  #:use-module (chart symlists))


;; initial 5 years, as tdate count of weekdays
(define-public initial-download-days 1300)


;;-----------------------------------------------------------------------------
;; locations

;; eg. "/home/username/Chart/data"
(define-public data-directory
  (g-build-filename database-directory "data"))

;; eg. "/home/username/Chart/notes"
(define-public user-notes-directory
  (g-build-filename user-directory "notes"))


;;-----------------------------------------------------------------------------
;; file stuff

;; Make a directory, and all components leading to it, if it doesn't already
;; exist.
(define (mkdir-all-ensure dir)
  (if (not (access? dir F_OK))
      (mkdir-all dir)))


;;-----------------------------------------------------------------------------
;; misc

(define-public eof-object (read (open-input-string "")))

(define-public (write-nl obj port)
  (write obj port)
  (newline port))

;; return identifier list (DEVNO INODE MTIME SIZE), or #f if no such file
(define-public (file-ident port-or-name)
  (and-let* ((st (false-if-exception (stat port-or-name))))
    (list (stat:dev st)
	  (stat:ino st)
	  (stat:mtime st)
	  (stat:size st))))

;; return true if file identifiers X and Y are equal
;; #f for no such file is considered never equal
(define-public (file-ident-equal? x y)
  (and x y (equal? x y)))


;;-----------------------------------------------------------------------------
;; series merging

(define (first-eqv? x y)
  (eqv? (first x) (first y)))

(define (third= x y)
  (= (third x) (third y)))

;; return LST with just the first among runs of equal elements kept
(define (uniq = lst)
  (if (null? lst)
      '()
      (reverse! (fold (lambda (elem ret)
			(if (= (first ret) elem)
			    ret
			    (cons elem ret)))
		      (list (first lst))
		      (cdr lst)))))

;; return true if dividend element (TDATE . VALUE) has value (month MDATE)
(define (dividend-month? elem)
  (and (pair? (cdr elem))
       (eq? 'month (first (cdr elem)))))

(define-public (notes-change-decimals notes old new)
  (let ((factor (integer-expt 10 (- new old))))
    (map (lambda (note)
	   (case (first note)
	     ((line hline)  ;; (line <t1> <p1> <t2> <p2>)
	      (set! note (list-copy note))
	      (set-third! note (* factor (third note)))
	      (set-fifth! note (* factor (fifth note))))

	     ((alert-above alert-below)  ;; (alert <p>)
	      (set! note (list-copy note))
	      (set-second! note (* factor (second note)))))
	   note)

	 notes)))

(define (dividend-elem-type? elem)
  (if (pair? (cdr elem))
      (second elem)
      'dividend))

(define (dividend-elem-eqv? x y)
  (and (= (car x) (car y))
       (eq? (dividend-elem-type? x) (dividend-elem-type? y))))

;; remove repeated (TDATE . (month MDATE)) month entries in DIVIDENDS
(define (dividend-months-flatten dividends)
  (receive (months other)
      (partition dividend-month? dividends)
    (append (uniq third= (sort months car<)) other)))

;; copy source array SA with decimals SD, to destination array DA and change
;; the data to DD decimals while copying
;;
(define (array-copy-ohlcv-plus-decimals! sa sd da dd)
  (let ((factor         (integer-expt 10 (- dd sd)))
	(hi             (array-last-index sa))
	(open-interest? (= 5 (array-last-column sa))))
    (do ((i (array-first-index sa) (1+ i)))
	((> i hi))
      (and-let* ((o (array-ref sa i 0)))
	(array-set! da (* factor o) i 0))
      (and-let* ((h (array-ref sa i 1)))
	(array-set! da (* factor h) i 1))
      (and-let* ((l (array-ref sa i 2)))
	(array-set! da (* factor l) i 2))
      (and-let* ((c (array-ref sa i 3)))
	(array-set! da (* factor c) i 3))
      (and-let* ((v (array-ref sa i 4)))
	(array-set! da v i 4))
      (and-let* (open-interest?
		 (int (array-ref sa i 5)))
	(array-set! da int i 5)))))

;; merge series X and Y, with Y overwriting
(define-public (series-merge-ohlcv x y)
  (or (eq? (series-timebase x) (series-timebase y))
      (error "series-merge: different timebases"))
  (let* ((xlo (series-lo x))
	 (xhi (series-hi x))
	 (xa  (series-array x xlo xhi))
	 (xd  (series-decimals x))

	 (ylo (series-lo y))
	 (yhi (series-hi y))
	 (ya  (series-array y ylo yhi))
	 (yd  (series-decimals y))

	 (sd  (max xd yd))
	 (s   (series-copy x))
	 (slo (min xlo ylo))
	 (shi (max xhi yhi))
	 (sa  (make-array #f (list slo shi)
			  (list 0 (max (array-last-column xa)
				       (array-last-column ya))))))

    (array-copy-ohlcv-plus-decimals! xa xd sa sd)
    (array-copy-ohlcv-plus-decimals! ya yd sa sd)
    (series-array! s sa)
    (series-decimals! s sd)
    (series-lo! s slo)
    (series-hi! s shi)

    (and-let* ((name (series-name y)))
      (series-name! s name))

    (series-dividends! s (dividend-months-flatten
			  ;; Y replaces X on same tdate
			  (lset-union dividend-elem-eqv?
				      (series-dividends y)
				      (series-dividends x))))
    ;; Y preferred over X of same key
    (series-info! s (lset-union first-eqv?
				(series-info y)
				(series-info x)))
    ;; except max for `download-done'
    (and-let* ((done (max-maybe (series-info-get x 'download-done)
				(series-info-get y 'download-done))))
      (series-info-add! s 'download-done done))

    (series-notes! s (append (notes-change-decimals (series-notes x) xd sd)
			     (notes-change-decimals (series-notes y) yd sd)))

    (series-require! s noop)

    s))


;;-----------------------------------------------------------------------------
;; historical list

(define historical-filename (g-build-filename cache-directory
					      "historical-symbols"))

(define (series-actual-hi series)
  (let more ((hi (series-hi series)))
    (cond ((<= hi (series-lo series))
	   ;; no data at all, fake up a very old tdate
	   (- (series-lo series) 9999))
	  ((array-ref (series-array series hi hi) hi 3)
	   hi)
	  (else
	   (more (1- hi))))))

(define (series-any-data-within? series lo hi)
  (set! lo (max lo (series-lo series)))
  (set! hi (min hi (series-hi series)))
  (and (>= hi lo)  ;; must overlap in part with SERIES range
       (let ((array (series-array series lo hi)))
	 (set! array (make-shared-array-restrict array lo hi))
	 (set! array (make-shared-array-column array 3)) ;; closes
	 (any noop (array->list array)))))

;; return true if SERIES should be considered historical
;;
(define (historical-series? series)
  (and (eq? timebase-days (series-timebase series)) ;; only on daily series
       (or

	;; no new data for 2 weeks, ie. last-download more than 2 weeks
	;; after series-hi
	;;
	;; this is for sources like yahoo which don't say for certain what
	;; dates are covered, merely stop supplying data
	;;
	(and-let* ((last-download (series-info-get series 'last-download)))
	  (> (- last-download (series-actual-hi series))
	     10))

	;; no data in the latest 2 weeks covered (either from series-hi or
	;; from recorded download-done covered date)
	;;
	;; this is for sources which give a whole daily file or similar so
	;; it's certain what dates are supposed to be covered
	;;
	(let ((done (max-maybe (series-info-get series 'download-done)
			       (series-hi series))))
	  (not (series-any-data-within? series (- done 9) done)))

	;; 1 week after a known expiry date
	(and-let* ((expiry-date (series-info-get series 'expiry-date)))
	  (> (- (tdate-today (symbol-timezone (series-symbol series)))
		expiry-date)
	     5)))))

(define (historical-symbol? symbol)
  (historical-series? (database-read-series symbol)))

(define (historical-write symbol-list)
  (database-call-with-output-file historical-filename
    (lambda (port)
      (display ";; Historical symbols list, generated by Chart\n" port)
      (write-list-lines symbol-list port)
      (newline port)))
  (notify-send 'historical-cache-update #f))

;; return list of symbols
(define-public (historical-read)
  (or (call-with-input-file-or #f historical-filename read)
      (let ((symbol-list (filter historical-symbol? (database-symbols))))
	(historical-write symbol-list)
	symbol-list)))

(define (historical-list-op proc symbol-list)
  (database-call-with-lock (dirname historical-filename)
    (lambda ()
      (let* ((old (historical-read))
	     (new (proc string=? old symbol-list)))
	(if (not (equal? old new))
	    (historical-write new))))))

;; remove all in SYMBOL-LIST from the historical list
(define-public (historical-remove symbol-list)
  (historical-list-op lset-difference symbol-list))

;; check if any of SYMBOL-LIST should become historical
(define-public (historical-check symbol-list)
  (historical-list-op lset-union (filter historical-symbol? symbol-list)))


;;-----------------------------------------------------------------------------
;; alert list

(define alert-interest-filename (g-build-filename cache-directory
						  "alert-interest-symbols"))

;; return true if SYMBOL has any alerts set
(define (alert-interest-symbol? symbol)
  (and (not (member symbol (symlist-list historical-symlist)))
       (database-read-note symbol alert-key?)))

;; return true if SERIES has any alerts set
(define (alert-interest-series? series)
  (any (lambda (elem)
	 (alert-key? (first elem)))
       (series-notes series)))

(define (alert-interest-write symbol-list)
  (database-call-with-output-file alert-interest-filename
    (lambda (port)
      (display ";; Alert symbols list, generated by Chart\n" port)
      (write-list-lines symbol-list port)
      (newline port)))
  (notify-send 'alert-cache-update #f))

;; return list of symbols
(define-public (alert-interest-read)
  (or (call-with-input-file-or #f alert-interest-filename read)
      (let ((symbol-list (filter alert-interest-symbol? (database-symbols))))
	(alert-interest-write symbol-list)
	symbol-list)))


;;-----------------------------------------------------------------------------
;; common

(define-public (database-new . symbol-list)
  (database-call-with-lock data-directory
    (lambda ()
      ;; those which don't already exist
      (set! symbol-list (lset-difference string=? symbol-list
					 (database-symbols)))
      (for-each (lambda (symbol)
		  (database-write-series
		   (series-new #:symbol   symbol
			       #:datatype datatype-ohlcv
			       #:array    dummy-ohlcv-array)))
		symbol-list)))
  (if (not (null? symbol-list))
      (notify-send 'database-new symbol-list)))

(define-public (database-delete . symbol-list)
  (define notes-too #f)

  ;; optional initial #:notes-too BOOL args
  (if (and (not (null? symbol-list))
	   (eq? #:notes-too (first symbol-list)))
      (begin
	(set! notes-too (second symbol-list))
	(set! symbol-list (cddr symbol-list))))

  (database-call-with-lock data-directory
    (lambda ()
      (for-each (lambda (symbol)
		  (hash-remove! database-series-hash symbol)
		  (delete-file-maybe (data-filename symbol))
		  (if notes-too
		      (delete-file-maybe (notes-filename symbol))))
		symbol-list)))
  (if (not (null? symbol-list))
      (begin
	(historical-remove symbol-list)
	(symlist-delete-symbols all-symlist symbol-list)

	(database-call-with-lock (dirname alert-interest-filename)
	  (lambda ()
	    (let* ((old (alert-interest-read))
		   (new (lset-difference string=? old symbol-list)))
	      (and (not (equal? old new))
		   (alert-interest-write new)))))

	(notify-send 'database-delete symbol-list)
	(notify-send 'database-update symbol-list))))


;; check for valid filename (ie. the filename part, no directory)
(define-public (filename-valid? filename)
  
  (and (not (string-any #\/ filename))
       (string<> "." filename)
       (string<> ".." filename)))

;; call (THUNK) with position information disabled on `read'
(define (call-without-read-positions thunk)
  (let ((old #f))
    (dynamic-wind
	(lambda ()
	  (set! old (read-options-interface))
	  (read-disable 'positions))
	thunk
	(lambda ()
	  (read-options-interface old)))))


;;-----------------------------------------------------------------------------
;; writing and locking

;; call (THUNK) with an `flock' of the given OPERATION on PORT/FD
(define (with-flock port/fd operation thunk)
  (dynamic-wind
      (lambda () (flock port/fd operation))
      thunk
      (lambda () (flock port/fd LOCK_UN))))

(let-toplevel ((lock-list '()))

  ;; Call (THUNK), with the given directory locked.
  ;; .lockfile is never deleted, mainly because there's no good way to do
  ;; some sort of age-check-and-delete atomically
  ;;
  ;; A second flock on .lockfile from within the same process but using a
  ;; different open file will block.  So we keep a lock-list of those we
  ;; already hold.
  ;;
  (define-public (database-call-with-lock dir thunk)
    (mkdir-all-ensure dir)
    (if (member dir lock-list)
	(thunk)
	(dynamic-wind
	    (lambda ()
	      (set! lock-list (cons dir lock-list)))
	    (lambda ()
	      (call-with-output-file (g-build-filename dir ".lockfile")
		(lambda (port)
		  (with-flock port LOCK_EX thunk))))
	    (lambda ()
	      (set! lock-list (delete! dir lock-list)))))))

;; Call (PROC port) where PORT is setup to write to FILENAME
;;
;; This is like `call-with-output-file' but it locks the directory and it
;; writes first to a temporary file then renames if PROC is successful, so
;; as to be safe to concurrent reads (ie. if someone else opens to read then
;; they won't see a half-written file).
;; 
(define-public (database-call-with-output-file filename proc)
  (let ((dir (dirname filename)))
    (database-call-with-lock dir
      (lambda ()
	(let* ((tempname (g-build-filename dir ".temp-XXXXXX"))
	       (port     (mkstemp! tempname)))
	  (chmod port (logand #o666 (lognot (umask)))) ;; usual umask
	  (proc port)
	  (close port)
	  (rename-file tempname filename))))))


;;-----------------------------------------------------------------------------
;; notes

(define (notes-filename symbol)
  (or (filename-valid? symbol)
      (error "bad symbol:" symbol))
  (g-build-filename user-notes-directory symbol))

;; LST should be (YEAR MONTH DAY)
(define (ymdlst->tdate lst)
  (apply ymd->tdate lst))

;; mung an old notes file format element into internal format
(define (notes-mung-old elem)
  (cond ((and (memq (first elem) '(line hline))
	      (list? (second elem)))
	 ;; used to have (y m d) for dates in line and hline
	 (list (first elem)
	       (ymdlst->tdate (second elem)) (third elem)
	       (ymdlst->tdate (fourth elem)) (fifth elem)))
	((and (eq? (first elem) 'annotation)
	      (list? (second elem)))
	 ;; used to have (y m d) for date in annotation
	 (list (first elem) (ymdlst->tdate (second elem)) (third elem)))
	(else
	 elem)))

(define (database-read-notes series)
  (series-notes! series
		 (call-with-input-file-or '()
		     (notes-filename (series-symbol series))
		   (lambda (port)
		     (call-without-read-positions
		      (lambda ()
			(set! (notes-file-ident-prop series) (file-ident port))
			(map! notes-mung-old (list-read-to-eof port))))))))

(define-public (database-write-notes series)
  (hash-set! database-series-hash (series-symbol series) series)
  (let* ((symbol   (series-symbol series))
	 (notes    (series-notes series))
	 (filename (notes-filename symbol)))
    (if (null? notes)
	(delete-file-maybe filename)
	(begin
	  (mkdir-all-ensure user-notes-directory)
	  (database-call-with-output-file filename
	    (lambda (port)
	      (format port ";; ~a notes, generated by Chart  -*- coding: utf-8 -*-\n\n" symbol)
	      (for-each (lambda (elem)
			  (write-nl elem port))
			notes)
	      (force-output port)
	      (set! (notes-file-ident-prop series) (file-ident port))))))

    (database-call-with-lock (dirname alert-interest-filename)
      (lambda ()
	(let* ((old (alert-interest-read))
	       (new (if (alert-interest-series? series)
			(lset-adjoin string=? old symbol)
			(delete! symbol old))))
	  (and (not (equal? old new))
	       (alert-interest-write new)))))

    (notify-send 'database-update (list (series-symbol series)))))


;; return notes value or #f
;;
(define-public (database-read-note symbol key)
  (cond
   ((hash-ref database-series-hash symbol)
    => (lambda (series)
	 (if (procedure? key)
	     (find (lambda (elem)
		     (key (first elem)))
		   (series-notes series))
	     (assq key (series-notes series)))))

   (else
    (call-with-input-file-or #f (notes-filename symbol)
      (lambda (port)
	(call-without-read-positions
	 (lambda ()
	   (if (symbol? key)
	       (let ((sym key))
		 (set! key (lambda (key)
			     (eq? sym key)))))
	   (let more ((elem (read port)))
	     (cond ((eof-object? elem)
		    #f)
		   ((key (first elem))
		    (notes-mung-old elem))
		   (else
		    (more (read port))))))))))))


;;-----------------------------------------------------------------------------
;; data stuff

;; the filename for SYMBOL
;; eg. "/home/username/Chart/data/FOO.NYBOT"
(define (data-filename symbol)
  (or (filename-valid? symbol)
      (error "bad symbol:" symbol))
  (g-build-filename data-directory symbol))

;; key SYMBOL, value SERIES
(define database-series-hash (make-weak-value-hash-table 61))

(define data-file-ident-prop  (make-object-property))
(define notes-file-ident-prop (make-object-property))

(define (database-series-hash-ref symbol)
  (and-let* ((series (hash-ref database-series-hash symbol))
	     (       (file-ident-equal? (file-ident (data-filename symbol))
					(data-file-ident-prop series)))
	     (       (file-ident-equal? (file-ident (notes-filename symbol))
					(notes-file-ident-prop series))))
    series))


;;-----------------------------------------------------------------------------
;; read series

(define dummy-tdate (- (tdate-today) 2))
(define dummy-ohlcv-array (make-array #f (list dummy-tdate dummy-tdate)
				      '(0 4)))
(define dummy-ohlcvi-array (make-array #f (list dummy-tdate dummy-tdate)
				       '(0 5)))

;; return series, or #f if not in database
;; if SYMBOL is in the database but there's no data, the series returned has
;; a single entry for today, all #f's
;;
(define-public (database-read-series symbol)
  (or (database-series-hash-ref symbol)
      (call-with-input-file-or #f (data-filename symbol)
	(lambda (port)
	  (call-without-read-positions
	   (lambda ()
	     (let ((series (series-new #:symbol symbol
				       #:datatype datatype-ohlcv
				       #:array dummy-ohlcv-array
				       #:require database-read-require)))

	       (do ((elem (read port) (read port)))
		   ((eof-object? elem))
		 (let ((key (first elem)))
		   (cond ((eq? key 'name)
			  (series-name! series (second elem)))
			 ((eq? key 'format)
			  (case (second elem)
			    ((ohlcvi)
			     (series-array! series dummy-ohlcvi-array)
			     (series-info-add! series 'format 'ohlcvi))
			    (else
			     (error "database-read-series: unrecognised format:" elem))))
			 ((eq? key 'decimals)
			  (series-decimals! series (second elem)))
			 ((eq? key 'dividend)
			  (series-dividend-add! series
						(second elem) (third elem)))
			 ((number? key) ;; data
			  (series-array! series #f)
			  (series-lo! series key)
			  ;; database-last-tdate-port moves port to eof, so
			  ;; loop stops
			  (series-hi! series (database-last-tdate-port port)))
			 (else
			  (series-info-add! series key (second elem))))))

	       (set! (data-file-ident-prop series) (file-ident port))
	       (database-read-notes series)
	       (hash-set! database-series-hash (series-symbol series) series)
	       series)))))))

(define (database-read-require series lo hi)
  (let* ((array          #f)
	 (old-array      (series-held-array series))
	 (old-lo         (1+ hi))  ;; fake to induce read of whole lo to hi
	 (old-hi         hi)
	 (slo            (series-lo series))
	 (shi            (series-hi series))
	 (open-interest? (ohlcv-open-interest? series)))

    ;; extend lo,hi to ends if nearly there anyway
    (if (<= (- lo 30) slo)
	(set! lo slo))
    (if (>= (+ hi 30) shi)
	(set! hi shi))

    (if old-array
	(begin
	  (set! old-lo (array-first-index old-array))
	  (set! old-hi (array-last-index old-array))))

    (or (call-with-input-file-or #f (data-filename (series-symbol series))
	  (lambda (port)

	    (define (read-fill rlo rhi)

	      ;; estimate location in proportion to tdates
	      (seek port (floor-exact (* (seek port 0 SEEK_END)
					 (/ (max 0 (- rlo slo 50))
					    (max 1 (- shi slo)))))
		    SEEK_SET)

	      (let ((elem (read-to-data port)))

		;; if not in right spot, go back to start
		(if (or (not elem)
			(> (first elem) rlo))
		    (begin
		      '(format (current-error-port)
			       "database lookup: ~a estimate gave ~a wanting ~a\n"
			       (series-symbol series) (first elem) rlo)
		      (seek port 0 SEEK_SET)
		      (set! elem (read-to-data port))))

		;; store all of what we read
		(let ((t (first elem)))

		  (set! lo (min lo t))
		  (set! rlo (min rlo t))
		  (if (not array)
		      (begin
			(set! array (make-array #f (list lo hi)
						(if open-interest?
						    '(0 5)
						    '(0 4))))
			(if old-array
			    (array-copy! old-array array)))))

		(let more ((elem elem))
		  (if (not (eof-object? elem))
		      (let ((t (first elem)))
			(if (<= t rhi)
			    (begin
			      (array-set! array (second elem) t 0) ;; O
			      (array-set! array (third  elem) t 1) ;; H
			      (array-set! array (fourth elem) t 2) ;; L
			      (array-set! array (fifth  elem) t 3) ;; C
			      (array-set! array (sixth  elem) t 4) ;; V
			      (if open-interest?
				  (array-set! array (seventh elem) t 5)) ;; I
			      (more (read port)))))))))

	    (if (< lo old-lo)
		(read-fill lo (1- old-lo)))
	    (if (> hi old-hi)
		(read-fill old-hi hi))))

	;; file has gone away, keep old data but #f for rest
	(begin
	  (set! array (make-array #f (list lo hi) '(0 4)))
	  (if old-array
	      (array-copy! old-array array))))

    array))

;; return data element, or #f
(define (read-to-data port)
  (read-skip-to-next-line port)
  (let more ((elem (read port)))
    (cond ((eof-object? elem)
	   #f)   ;; bad
	  ((number? (first elem))
	   elem) ;; good
	  (else
	   (more (read port))))))

;; read and discard characters from PORT until the next line (or eof) is
;; reached
(define (read-skip-to-next-line port)
  (let ((c (read-char port)))
    (or (eof-object? c)
	(char=? c #\newline)
	(read-skip-to-next-line port))))

;; `read' last object before EOF from PORT
(define (read-last port)
  (do ((prev eof-object this)
       (this (read port) (read port)))
      ((eof-object? this)
       prev)))

;; last data element from PORT, or eof-object if none
(define (database-last-elem-port port)
  (seek port (- (min 128 (seek port 0 SEEK_END))) SEEK_END)
  (read-skip-to-next-line port)
  (read-last port))

;; last tdate of data from PORT, or #f if no data
(define (database-last-tdate-port port)
  (let ((elem (database-last-elem-port port)))
    (and (pair? elem)
	 (number? (first elem))
	 (first elem))))

;; return last tdate recorded for SYMBOL, or #f if none at all
(define-public (database-last-tdate symbol)
  (call-with-input-file-or #f (data-filename symbol)
    (lambda (port)
      (call-without-read-positions
       (lambda ()
	 (database-last-tdate-port port))))))

;; return info value or #f
;; KEY can be for instance 'name or 'stock-exchange
;;
(define-public (database-read-info symbol key)
  (cond
   ((hash-ref database-series-hash symbol)
    => (lambda (series)
	 (case key
	   ((name)
	    (series-name series))
	   (else
	    (assq-ref (series-info series) key)))))

   (else
    (call-with-input-file-or #f (data-filename symbol)
      (lambda (port)
	(call-without-read-positions
	 (lambda ()
	   (let more ((elem (read port)))
	     (cond ((or (eof-object? elem)
			(number? (first elem))) ;; stop at data section
		    #f)
		   ((eq? key (first elem))
		    (second elem))
		   (else
		    (more (read port))))))))))))


;;-----------------------------------------------------------------------------
;; write series

;; compare symbols in the car of X and Y, by name
(define (car-symbol-string<? x y)
  (string<? (symbol->string (car x)) (symbol->string (car y))))

;; check some aspects of SERIES before writing, to avoid bad stuff getting
;; into the database
(define-public (database-write-validate series)
  (if (series-name series)
      (utf8-validate (series-name series)))
  (for-each (lambda (pair)
	      (if (string? (cdr pair))
		  (utf8-validate (cdr pair))))
	    (series-dividends series))
  (for-each (lambda (note)
	      (if (annotation? note)
		  (utf8-validate (third note))))
	    (series-notes series)))

(define-public (database-write-series series)
  (database-write-validate series)

  (let* ((symbol         (series-symbol series))
	 (filename       (data-filename symbol))
	 (new            (not (file-exists? filename)))
	 (lo             (series-lo series))
	 (hi             (series-hi series))
	 (array          (series-array series lo hi))
	 (open-interest? (= 5 (array-last-column array))))
    (hash-set! database-series-hash symbol series)
    (mkdir-all-ensure data-directory)

    (if open-interest?
	(series-info-add! series 'format 'ohlcvi)
	(series-info-del! series 'format))

    (database-call-with-output-file filename
      (lambda (port)
	(format port ";; ~a data, generated by Chart  -*- coding: utf-8 -*-\n\n" symbol)
	(and-let* ((name (series-name series))) ;; if name known
	  (write-nl (list 'name name) port))
	(write-nl (list 'decimals (series-decimals series)) port)

	(for-each (lambda (pair)
		    (write-nl (list (car pair) (cdr pair)) port))
		  (sort (series-info series) car-symbol-string<?)) ;; by key

	(for-each (lambda (pair)
		    (write-nl (list 'dividend (car pair) (cdr pair)) port))
		  (sort (series-dividends series) car<))  ;; by date
	(do ((t lo (1+ t)))
	    ((> t hi))
	  (let ((o (array-ref array t 0))  ;; O
		(h (array-ref array t 1))  ;; H
		(l (array-ref array t 2))  ;; L
		(c (array-ref array t 3))  ;; C
		(v (array-ref array t 4))  ;; V
		(i (and open-interest?
			(array-ref array t 5)))) ;; I
	    (if (or o h l c v i) ;; suppress all-#f rows
		(write-nl (if open-interest?
			      (list t o h l c v i)
			      (list t o h l c v))
			  port))))
	(set! (notes-file-ident-prop series) (file-ident port))))

    (if new
	(notify-send 'database-new (list symbol)))
    (notify-send 'database-update (list symbol))))

(define-public (database-add-series series)
  (database-call-with-lock data-directory
    (lambda ()
      (let ((old-series (database-read-series (series-symbol series))))
	(if old-series
	    (set! series (series-merge-ohlcv old-series series))
	    ;; multiple month specs from futures.scm
	    (series-dividends! series (dividend-months-flatten
				       (series-dividends series))))

	(database-write-series series)

	;; series-merge-ohlcv may have increased the decimals and consequently
	;; changed the notes data
	(if (and old-series
		 (not (equal? (series-notes series)
			      (series-notes old-series))))
	    (database-write-notes series))))))

;; calling (PROC series) with series for SYMBOL, rewriting afterwards
(define-public (database-call-modify-data symbol proc)
  (database-call-with-lock data-directory
    (lambda ()
      (let ((series (database-read-series symbol)))
	(if series
	    (begin
	      (proc series)
	      (database-write-series series)))))))

;; calling (PROC series) with series for SYMBOL, rewriting notes afterwards
(define-public (database-call-modify-notes symbol proc)
  (database-call-with-lock user-notes-directory
    (lambda ()
      (let ((series (database-read-series symbol)))
	(if series
	    (proc series))
	(database-write-notes series)))))


;;-----------------------------------------------------------------------------
;; available symbols

;; data files are everything except
;;     ".", ".."       standard directories
;;     ".lockfile"     database-call-with-lock
;;     ".temp-XXXXXX"  temp file
;;     "#foo"          emacs locks
;;     "foo~"          emacs backups
(define-public (data-filename? filename)
  (not (or (string-prefix? "." filename)
	   (string-prefix? "#" filename)
	   (string-suffix? "~" filename))))

;; Database symbols are cached mainly to make `database-symbol?' fast, so
;; that for instance commsec.scm doesn't read the whole directory every time
;; when deciding whether a symbol in the downloaded data is wanted.
;;
;; If there's two changes to the directory within 1 second, then freshen
;; might only see the first, and believe mtime etc is unchanged by the
;; second.

(let-toplevel ((symbol-list '())
	       (dir-ident   #f))

  ;; update symbol-list if the data directory has changed
  (define (freshen)
    (let ((ident (file-ident data-directory)))
      (if (not (file-ident-equal? ident dir-ident))
	  (begin
	    (set! symbol-list
		  (sort! (or (directory-files data-directory
					      #f data-filename?) '())
			 chart-symbol<?))
	    (set! dir-ident ident)))))

  ;; return list of available symbols
  (define-public (database-symbols)
    (freshen)
    symbol-list)

  ;; return true if SYMBOL is in the database
  (define-public (database-symbol? symbol)
    (freshen)
    (member symbol symbol-list)))
