;; DEMA Double Exponential Moving Average.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart dema)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.stockworm.com/help/manual/double-ema.html
;;     Example graph PFE (what year?).
;;
;; http://trader.online.pl/ELZ/t-i-Double_Smoothed_Exponential_Moving_Average.html
;;     Tradestation example ^SPC 2001.
;;
;; http://store.traders.com/-v12-c01-smoothi-pdf.html
;;     Intro of TA S&C magazine, describing as faster.
;;
;; http://www.fmlabs.com/reference/DEMA.htm
;;     Formula only.
;;


;; Return a procedure which calculates a Double EMA with given COUNT period
;; smoothing.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the Double EMA up to (and including) that value, or #f if not enough
;; points yet.
;;
;; A Double EMA is in theory influenced by all preceding data, but
;; `dema-warmup-count' below is designed to determine a warmup count.
;;
;;
;; A direct implementation in terms of `ema-calc-proc' and `ema-2-calc-proc'
;; would be as follows,
;;
;;     (define-public (dema-calc-proc count)
;;       (let ((ema-proc   (ema-calc-proc count))
;;     	       (ema-2-proc (ema-2-calc-proc count)))
;;         (lambda (x)
;;           (let ((e  (ema-proc x))
;;     		   (ee (ema-2-proc eA)))
;;             (and e ee (- (* 2.0 e) ee)))))
;;
;; but at the start of data there's no need to wait for ema-2 proc to start
;; giving values, the weights are faster than that.  With the combined
;; function the weight of a partial value can be carried through.
;;
(define-public (dema-calc-proc count)
  (let* ((f          (ema-count->f count))
	 (alpha      (ema-count->alpha count))
	 (min-points (dema-omitted-search count
					  (- 1 indicator-minimum-fraction)))
	 (points     1)
	 (a-sum      0.0)
	 (a-div      0.0)
	 (b-sum      0.0)
	 (b-div      0.0))

    (lambda (x)
      ;; reduce past data
      (set! a-sum (* a-sum f))
      (set! a-div (* a-div f))
      ;; add new data
      (set! a-sum (+ a-sum (* x alpha)))
      (set! a-div (+ a-div alpha))

      (set! b-sum (* b-sum f))
      (set! b-div (* b-div f))
      (set! b-sum (+ b-sum (* a-sum alpha)))
      (set! b-div (+ b-div (* a-div alpha)))

      (if (< points min-points)
	  (begin
	    (set! points (1+ points))
	    #f)
	  (- (* 2.0 (/ a-sum a-div)) (/ b-sum b-div))))))



;; By priming a DEMA accumulator PROC above with (dema-warmup-count COUNT)
;; many values, the next call will have an omitted weight of no more than
;; 0.1% of the total.  Omitting 0.1% should be negligable, unless past
;; values are ridiculously bigger than recent ones.
;;
;; The implementation here does a binary search for the first i satisfying
;; Omitted(i)<=0.001, so it's not very fast.
;;
(define-public (dema-warmup-count count)
  (dema-omitted-search count 0.001))

(define-public (dema-omitted-search count target)
  (define f (ema-count->f count))
  (bsearch-first-true (lambda (i)
			(<= (dema-omitted count f i) target))
		      count))

;; `dema-omitted' returns the fraction (between 0 and 1) of absolute weight
;; omitted by stopping a DEMA at the f^k term, which means the first k+1
;; terms.
;;
;; The EMA and EMAofEMA omitted totals are
;;
;;     Q(k) = f^(k+1)
;;     R(k) = f^(k+1) * (k+2 - f * (k+1))
;;
;; thus for the DEMA the net signed amount, implemented in
;; `dema-omitted-signed', is
;;
;;     S(k) = 2*Q(k) - R(k)
;;          = f^(k+1) * (f*(k+1) - k)
;;
;; This grows above 1 up to the f^N term, then the terms go negative and it
;; decreases towards 1.
;;
;; The position of the negative/positive transition is always at f^N.  The
;; coefficient of that f^N term is
;;
;;     2 - (1-f)*(N+1) = 2 - (1-(N-1)/(N+1))*(N+1)
;;                     = 2 - (N+1-(N-1))
;;                     = 2 - N - 1 + N - 1
;;                     = 0
;;
;; An absolute omitted weight is calculated from the signed omitted amount.
;; When k>N we can just negate the signed omitted.  When k<N we add the
;; negative terms past N in twice, first to cancel the negative then to add
;; in as positive.
;;
;; The total of all the negatives beyond N is -S(N),
;;
;;     tail = -S(N) = f^(N+1) * (f*(N+1) - N)
;;                  = f^(N+1) * ((N-1)/(N+1) * (N+1) - N)
;;                  = f^(N+1) * (N-1 - N)
;;                  = f^(N+1)
;;
;; Thus the absolute omitted,
;;
;;     T(k) = / - S(k)             if k >= N
;;            \ S(k) + 2*f^(N+1)   if k < N
;;
;; This is out of a total which is the positive and negative parts added as
;; abolute values.  Knowing pos+neg=1 and neg=-f^(N+1),
;;
;;     total absolute weight = 1 + 2*f(N+1)
;;
;; And that total is applied as a divisor, so the return from `dema-omitted'
;; is between 0 and 1.  (It works to call it with k=-1 for no terms omitted,
;; the result is 1.0.)
;;

(define-public (dema-omitted count f k)
  (let ((tail (expt f (1+ count))))
    (/ (if (>= k count)
	   (- (dema-omitted-signed f k))
	   (+ (dema-omitted-signed f k)
	      (* 2 tail)))
       (1+ (* 2 tail)))))

(define (dema-omitted-signed f k)
  (* (expt f (1+ k))
     (- (* f (1+ k)) k)))



(define (dema-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (dema-calc-proc count)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (proc this) i))))))

(define-public (dema-series series count)
  (let ((new-series (series-copy series))
	(warmup     (dema-warmup-count count)))
    (series-name!     new-series (format #f "~a ~a" (_ "DEMA") count))
    (series-datatype! new-series dema-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count warmup))
    (series-require!  new-series (indicator-require-parts dema-fill warmup))
    (series-array!    new-series #f)
    new-series))


(define dema-datatype
  (datatype-new!
   #:key      'dema
   #:name     (_ "DEMA - Double EMA")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Double and Triple Exponential Moving Average")
   #:proc     dema-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'dema-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
