#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Run Hurst R/S calculation on brown72.h test file.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: chart -s brown72.scm [/my/path/to/brown72.h]
;;
;; This program runs the hurst-rs module on the brown72.h test file by Edgar
;; Peters, as described on Ian Kaplan's page
;;
;;     http://www.bearcave.com/misl/misl_tech/wavelets/hurst/
;;
;; The individual pass results are (or should be) the same as shown in table
;; 4 on Ian Kaplan's page, namely
;;
;;     1024  96.4451
;;     512   55.7367
;;     256   30.2581
;;     128   20.9820
;;     64    12.6513
;;     32    7.2883
;;     16    4.4608
;;     8     2.7399
;;
;; And the final Hurst exponent estimate for the whole file 0.7270 (hence
;; the name of the file "72" in brown72.h).
;;
;; The brown72.h file should be given on the command line, or the default is
;; to look for it in the current directory.


(use-modules (ice-9 rdelim)
	     (srfi srfi-1)
	     (srfi srfi-13)
	     (chart hurst)
	     (chart misc))


(define filename (if (length>=? (program-arguments) 2)
		     (second (program-arguments))
		     "brown72.h"))

(define diff-lst '())
(call-with-input-file filename
  (lambda (port)
    (do ((line (read-line port) (read-line port)))
	((eof-object? line))
      ;; one number per line, with a comma at the end (for use as C data)
      (set! line (string-delete line #\,))
      (set! line (string-trim-both line))
      (set! diff-lst (cons (string->number line) diff-lst)))))
(set! diff-lst (reverse! diff-lst))

(define cumul-lst (let ((sum 0))
		    (cons 0 (map (lambda (x)
				   (set! sum (+ sum x))
				   sum)
				 diff-lst))))
(format #t "length ~a\n" (length cumul-lst))

(format #t "1024  ~a\n" (hurst-r/s-pass cumul-lst 1024 1 1024))
(format #t "512   ~a\n" (hurst-r/s-pass cumul-lst 1024 2 512))
(format #t "256   ~a\n" (hurst-r/s-pass cumul-lst 1024 4 256))
(format #t "128   ~a\n" (hurst-r/s-pass cumul-lst 1024 8 128))
(format #t "64    ~a\n" (hurst-r/s-pass cumul-lst 1024 16 64))
(format #t "32    ~a\n" (hurst-r/s-pass cumul-lst 1024 32 32))
(format #t "16    ~a\n" (hurst-r/s-pass cumul-lst 1024 64 16))
(format #t "8     ~a\n" (hurst-r/s-pass cumul-lst 1024 128 8))

(exit 0)


;; Local variables:
;; mode: scheme
;; End:
