#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Chicago Board of Trade link digging.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: ./cbot-dig.scm [--verbose] >cbot-info.scm
;;
;; Look around the CBOT web site and generate the cbot-info.scm file of
;; links and info.
;;
;; Pages downloaded during the digging are saved under ~/Chart/cache/dig and
;; are re-used on a re-run, to save repeated downloads during development.
;; Delete them to run on freshly downloaded pages.
;;

(use-modules (ice-9 and-let-star)
	     (ice-9 optargs)
	     (ice-9 receive)
	     (ice-9 regex)
	     (srfi srfi-1)
	     (srfi srfi-13)
	     (srfi srfi-14)
	     (chart database)
	     (chart download)
	     (chart extra-glib)
	     (chart futures)
	     (chart html)
	     (chart http)
	     (chart misc))


(define home-url "http://www.cbot.com")

(define option-verbose #f)

;; eg. "/home/username/Chart/cache/dig"
(define dig-directory (g-build-filename cache-directory "dig"))


(for-each (lambda (arg)
	    (cond ((member arg '("--verbose" "-v"))
		   (set! option-verbose #t))))
	  (program-arguments))


;;-----------------------------------------------------------------------------
;; misc

(define (html-link-extract str)
  (let* ((m   (must-match
	       (regexp-exec #,(regexp "<a href=['\"]?([^'\">]*)[^>]*>"
				      regexp/icase) str)))
	 (end (string-contains str "</a>" (match:end m))))
    (values (match:substring m 1)
	    (substring str (match:end m) end))))

;; (define (html-link-target str)
;;   (receive (link text)
;;       (html-link-extract str)
;;     link))

(define (slurp filename)
  (call-with-input-file filename read-string-to-eof))


;;-----------------------------------------------------------------------------
;; commodity info

;; alist of (COMMODITY . ALIST) with ALIST elements various (KEY . VALUE)
(define info-alist '())

(define (add-commodity commodity)
  (or (assoc-ref info-alist commodity)
      (set! info-alist (acons commodity '() info-alist))))

(define (commodity-list)
  (map car info-alist))

(define (assq-set-end! alist key value)
  (if (assq-ref alist key)
      (assq-set! alist key value)
      (append alist (list (cons key value)))))

(define* (put-info commodity key value #:key allow-override)
  (let ((alist (assoc-ref info-alist commodity)))
    (or alist
	(error "put-info: unknown commodity" commodity key value))

    (let ((old-value (assq-ref alist key)))
      (if (and old-value
	       (not allow-override)
	       (not (equal? old-value value)))
	  (error "Different values:" commodity key old-value value)))

    (set! alist (assq-set-end! alist key value))
    (assoc-set! info-alist commodity alist)))

(define* (get-info commodity key #:key allow-missing)
  (let ((alist (assoc-ref info-alist commodity)))
    (or alist
	(error "get-info: unknown commodity" commodity key))
    (let ((elem (assq key alist)))
      (cond (elem
	     (cdr elem))
	    (allow-missing
	     #f)
	    (else
	     (error "get-info: unknown key" commodity key))))))
      
(define quotes-url-list '())

(define (add-quotes-url name url)
  (if (member url quotes-url-list)
      #f ;; not added
      (begin
	(if option-verbose
	    (format (current-error-port) "add quotes page ~s: ~a\n" name url))
	(set! quotes-url-list (cons url quotes-url-list))
	#t))) ;; added


;;-----------------------------------------------------------------------------
;; download and cache

(define (url->filename url)
  (g-build-filename dig-directory
		    (regexp-substitute/global #f "/" url 'pre "_" 'post)))

(define (download-url url)
  (let ((filename (url->filename url)))
    (or (file-exists? filename)
	(begin
	  (if option-verbose
	      (format (current-error-port) "download ~a\n" url))
	  (receive (headers body)
	      (http-request url #:want-ok #f)
	    (database-call-with-output-file filename
	      (lambda (port)
		(display body port))))))))

(define (read-url url)
  (download-url url)
  (let ((filename (url->filename url)))
    (if option-verbose
	(format (current-error-port) "looking at ~a\n" filename))
    (slurp filename)))


;;-----------------------------------------------------------------------------
;; output heading

(display ";; cbot-info.scm - generated by cbot-dig.scm - DO NOT EDIT\n")
(let ((body (slurp (port-filename (current-load-port)))))
  (display (substring body
		      (match:end (string-match ";;[^\n]+\n" body))
		      (+ 2 (string-contains body "\n\n\n")))))


;;-----------------------------------------------------------------------------
;; "Trading Hours / Symbols" page (linked from home page)
;;

;; normally http://www.cbot.com/cbot/pub/page/0,3181,932,00.html
(define symbols-url
  (let* ((body (read-url home-url))
	 (m    (must-match (regexp-exec #,(regexp "<a href=['\"]?([^'\">]+)['\"]?[^>]*>Trading Hours / Symbols</a>" regexp/icase) body)))
	 (link (match:substring m 1)))
    (http-url-relative home-url link)))
(format #t ";; Symbols page: ~a\n" symbols-url)

(let* ((body (read-url symbols-url)))
  (for-each
   (lambda (m)
     (let ((row-list (html-table-rows-raw body (match:start m)))
	   (name     #f)
	   (url      #f))

       (let ((headings (map (lambda (str)
			      (string-delete (html-text-extract str)
					     char-set:whitespace))
			    (first row-list))))
	 (or (equal? headings
		     '("Product" "" "FuturesSymbol" "FuturesHours"
		       "OptionsSymbol" "OptionsHours" "ClearingCode"))
	     (error "Symbols page unrecognised headings:" headings)))
       (set! row-list (cdr row-list))

       (for-each
	(lambda (row)
	  (receive-list (this-name type symbol hours
				   options-symbol options-hours clearing-code)
	      row

	    ;; first row (open auction) gives name and quotes url
	    (if (not (string-every char-set:whitespace
				   (html-text-extract this-name)))
		(receive (link text)
		    (html-link-extract this-name)
		  (set! name (html-text-extract text))
		  (set! url (http-url-relative symbols-url link))))

	    (set! symbol (html-text-extract symbol))
	    (set! hours (html-text-extract hours))

	    ;; various rows with "n/a" when there's only one of
	    ;; electronic/open-auction
	    (if (string-ci<> symbol "n/a")
		(begin
		  (if url
		      (begin
			(add-quotes-url name url)
			(set! url #f))) ;; once only

		  (add-commodity symbol)
		  (put-info symbol 'name name)

		  (put-info symbol 'timezone
			    (if (string-contains-ci hours
						    "Central European time")
				'cet 'chicago))))))

	row-list)))

   (must-match (list-matches "<p class=\"dataHeaderProductName\"><a name="
			     body))))

(set! info-alist (sort! info-alist (lambda (x y)
				     (string<? (car x) (car y)))))

(format #t ";; Symbols page: ~a commodities, ~a quotes links\n"
	(length info-alist) (length quotes-url-list))


;;-----------------------------------------------------------------------------
;; Quotes pages linked from home page

(let ((body         (read-url home-url))
      (total-quotes 0)
      (new-quotes   0))
  (for-each
   (lambda (m)
     (let ((row-list (html-table-rows-raw body (match:start m))))

       ;; first row "All | Ag | Interest Rate | ..." links
       (if (string-contains (first (first row-list)) "Interest Rate")
	   (set! row-list (cdr row-list)))

       ;; blank line
       (if (string-every char-set:whitespace (html-text-extract
					      (first (first row-list))))
	   (set! row-list (cdr row-list)))

       (let ((headings (map! string-trim-both
			     (map! html-text-extract (first row-list)))))
	 (or (equal? headings '("Product" "Month" "Price" "Change"))
	     (error "CBOT quotes links headings unrecognised:" headings)))
       (set! row-list (cdr row-list))

       (for-each
	(lambda (row)
	  (and-let* ((     (= 4 (length row)))
		     (m    (must-match
			    (regexp-exec #,(regexp "<a .*href=['\"]?([^'\">]+)['\"]?>([^<]+)")
					 (first row))))
		     (link (match:substring m 1))
		     (name (match:substring m 2))
		     (url  (http-url-relative home-url link)))
	    (set! total-quotes (1+ total-quotes))
	    (if (add-quotes-url name url)
		(set! new-quotes (1+ new-quotes)))))
	row-list)))
   (must-match
    (list-matches "<table[^\n]*homepage_quotetable" body)))

  (format #t ";; Home page: ~a quotes links (~a new)\n"
	  total-quotes new-quotes))


;;-----------------------------------------------------------------------------
;; electronic quotes

;; look through all the quotes-url-list and get further links to electronic
;; traded pages
;;
(let ((total 0)
      (added 0))
  (if option-verbose
      (display "Electronic quotes subpages:\n" (current-error-port)))
  (for-each (lambda (url)
	      (let ((body (read-url url)))
		(and-let* ((m    (regexp-exec #,(regexp "<a href=['\"]?([^'\">]+)['\"]?[^>]*><img src='/cbot/pub/siteimagesv4/electronic" regexp/icase) body))
			   (link (match:substring m 1))
			   (url  (http-url-relative url link)))
		  (set! total (1+ total))
		  (if (add-quotes-url #f url)
		      (set! added (1+ added))))))
	    quotes-url-list)
  (format #t ";; Quotes pages: ~a electronic (~a new)\n"
	  total added))


;;-----------------------------------------------------------------------------
;; open auction quotes

;; look through all the quotes-url-list and get further links to
;; open auction traded pages
;;
(let ((total 0)
      (added 0))
  (if option-verbose
      (display "Open auction quotes subpages:\n" (current-error-port)))
  (for-each (lambda (url)
	      (let ((body (read-url url)))
		(and-let* ((m    (regexp-exec #,(regexp "<a href=['\"]?([^'\">]+)['\"]?[^>]*><img src='/cbot/pub/siteimagesv4/openauction" regexp/icase) body))
			   (link (match:substring m 1))
			   (url  (http-url-relative url link)))
		  (set! total (1+ total))
		  (if option-verbose
		      (format (current-error-port) "Open Auction: ~a\n" link))
		  (if (add-quotes-url #f url)
		      (set! added (1+ added))))))
	    quotes-url-list)
  (format #t ";; Quotes pages: ~a open auction (~a new)\n"
	  total added))


;;-----------------------------------------------------------------------------
;; examine quotes pages for contract specs links

(for-each 
 (lambda (url)
   (let* ((body       (read-url url))

	  ;;                                              1                                                      2          3
	  (m          (must-match (or (regexp-exec #,(regexp "(dataHeaderProductName|content_storytitle)\">[ \t\r\n]*([^\n]*)\\(([^)]+)\\)") body)
				      (regexp-exec #,(regexp "(dataHeaderProductName|content_storytitle)\">[ \t\r\n]*([^\n]*)(nomatchhere)?") body))))
	  (name       (match:substring m 2))
	  (commodity  (or (match:substring m 3)
			  ;; footnote too hard to parse
			  (and (string-prefix-ci? "Synthetic Soybean Crush" name)
			       "BCX")
			  (error "Commodity symbol not found in quotes page" url)))

	  ;; Note regexp "Futures +", multiple spaces for instance
	  ;; on ZO page
	  ;; "http://www.cbot.com/cbot/pub/page/0,3181,1416,00.html"
	  (m          (must-match (or (regexp-exec #,(regexp "<a href=\"([^\"]+)\"[^>]*>Futures +Contract Specs</a>" regexp/icase) body)
				      (regexp-exec #,(regexp "<a href=\"([^\"]+)\"[^>]*>Contract Specs</a>" regexp/icase) body)
				      (regexp-exec #,(regexp "<a href=\"([^\"]+)\"[^>]*>OTC Contract Specs</a>" regexp/icase) body)
				      (regexp-exec #,(regexp "<a href=\"([^\"]+)\"[^>]*>Synthetic Crush Definition</a>" regexp/icase) body))))
	  (specs-link (match:substring m 1))
	  (specs-url  (http-url-relative url specs-link)))

     (or (assoc commodity info-alist)
	 (format #t ";; ~s has quotes page, but not in symbols page\n"
		 commodity))

     (add-commodity commodity)
     (put-info commodity 'name (html-text-extract name)
	       #:allow-override #t)
     (put-info commodity 'specs-url specs-url)))

	  quotes-url-list)


;;-----------------------------------------------------------------------------
;; examine specs pages

;; forms: "Open Auction:MB\nElectronic: ZU"
;;        "AC for Open Auction\nZE for Electronic"
;;        "GBL"
;;
;; unused for now
(define (tickers->symbol-list str)
  (cond ((regexp-exec #,(regexp "^(Open Auction|Open [Oo]utcry|Electronic|Electronic Trading) *[:-] *([A-Z]+) *(\n(Open Auction|Electronic)[:-] *([A-Z]+))?$") str)
	 => (lambda (m)
	      (if (match:start m 5)
		  (list (match:substring m 2) (match:substring m 5))
		  (list (match:substring m 2)))))

	((regexp-exec #,(regexp "^([A-Z]+) for Open Auction\n([A-Z]+) for Electronic$") str)
	 => (lambda (m)
	      (list (match:substring m 1) (match:substring m 2))))

	((string-every char-set:upper-case str)
	 (list str))

	(else
	 (format #t "ERROR: Contract specs page symbols format unrecognised: ~s\n"
		 str)
	 (exit 1))))

(define specs-url-list
  (filter-map (lambda (commodity)
		(get-info commodity 'specs-url #:allow-missing #t))
	      (commodity-list)))

;; TURNED OFF:
(set! specs-url-list '())

(for-each (lambda (url)
	    (let* ((body  (string-delete (read-url url) #\cr))

		   (table (html-table-rows
			   body  (or (string-contains body "storyBody")
				     (error "Oops, specs page storyBody not found"))))
		   (tail    (or (member '("Ticker Symbol") table)
				(member '("Ticker Symbols") table)
				(error "Oops, \"Ticker Symbols\" not found in specs"
				       url)))
		   (tickers (first (second tail)))
		   (commodity-list (tickers->symbol-list tickers)))

	      ;; any mentioned on the pages and not found through the quotes
	      (for-each (lambda (commodity)
			  (put-info commodity 'specs-url url))
			commodity-list)))
	  specs-url-list)


;;-----------------------------------------------------------------------------
;; print specs weblinks

(newline)
(let* ((alist (filter-map
	       (lambda (commodity)
		 (let ((url (get-info commodity 'specs-url #:allow-missing #t)))
		   (if url
		       (cons commodity url)
		       (begin
			 (format #t ";; No contract specs found: ~s ~s\n"
				 commodity
				 (get-info commodity 'name #:allow-missing #t))
			 #f))))
	       (commodity-list)))
       (cbot-specs-list
	(map (lambda (alist)  ;; of same target url
	       (cons (map car alist)
		     (cdr (first alist))))
	     (partition-equal cdr alist))))

  (display "(define cbot-specs-list\n")
  (let ((indent "  '("))
    (for-each (lambda (elem)
		(display indent)
		(set! indent "    ")
		(let ((inner-indent "(("))
		  (for-each (lambda (commodity)
			      (display inner-indent)
			      (set! inner-indent (string-append indent "  "))
			      (write commodity)
			      (if (string=? commodity (last (car elem)))
				  (display ")"))
			      (and-let* ((name (get-info commodity 'name
							 #:allow-missing #t)))
				(format #t "  ;; ~a" name))
			      (newline))
			    (car elem)))
		(format #t "~a . ~s)\n" indent (cdr elem)))
	      cbot-specs-list))
  (display "  ))\n"))


;; Local variables:
;; mode: scheme
;; compile-command: "chart -s cbot-dig.scm"
;; End:
