#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Closed form polynomials for tricky sums.

;; Copyright 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: ./closed-form.scm
;;
;; This started as a spot of code for turning sums into closed forms, for
;; example the triangular function
;;
;;     1 + 2 + ... + n
;;
;; can be turned into a quadratic
;;
;;     1/2 * n + 1/2 * n^2.
;;
;; Triangular numbers are easy, but trickier bits like sums of squares or
;; whatever can be written in a function in an obvious way, and this program
;; used to simplify them.
;;
;; The algorithm is simple, and quite well known.  The function value at a
;; set of test points is taken.  The differences between those values are
;; calculated, then the differences between those differences, etc, until
;; reaching differences which are all the same, which gives the coefficient
;; of the highest power of n.  Then with that term subtracted out the
;; process repeats for the next highest power.
;;
;; The program has since expanded to related stuff that arises for weights
;; of coefficients in EMA style moving averages, including doubly-smoothed
;; averages.  There's no command line interface for any of this, just
;; printouts sprayed through the file.  Guile 1.8 is recommended because
;; it's got exact fractions.
;;


(use-modules (ice-9 format)
	     (ice-9 optargs)
	     (ice-9 receive)
	     (srfi srfi-1)
	     (chart misc))


;;-----------------------------------------------------------------------------
;; misc

(define (cached func)
  (let ((table (make-hash-table 201)))
    (lambda (n)
      (or (hashv-ref table n)
	  (let ((val (func n)))
	    (hashv-set! table n val)
	    val)))))

;; (iota-cached N) returns a list of N integers 0 to N-1
(define iota-cached (cached iota))

(define (drop-while-right pred lst)
  (if (pred (last lst))
      (reverse (drop-while pred (reverse lst)))
      lst))

(define (not-zero? x)
  (not (zero? x)))

(define-macro (assert expr)
  `(or ,expr
       (error "assert failed" ,expr)))

;; return the index of the first element of ARRAY satisfying (PRED elem), or
;; #f if none
;; ARRAY must be 1-dimensional
(define-public (array-find-pos array pred)
  (let ((hi (array-last-index array)))
    (let more ((i (array-first-index array)))
      (and (<= i hi)
	   (if (pred (array-ref array i))
	       i
	       (more (1+ i)))))))

;; return the index of the last element of ARRAY satisfying (PRED elem), or
;; #f if none
;; ARRAY must be 1-dimensional
(define-public (array-find-last-pos array pred)
  (let ((lo (array-first-index array)))
    (let more ((i (array-last-index array)))
      (and (>= i lo)
	   (if (pred (array-ref array i))
	       i
	       (more (1- i)))))))


;;-----------------------------------------------------------------------------
;; polynomial stuff
;;
;; A polynomial is a list of coefficients starting from x^0, so say a list
;; (9 8 7) is 9 + 8*x + 7*x^2.

(define (poly-coeff poly i)
  (cond ((negative? i)
	 0)
	((length>=? poly (1+ i))
	 (list-ref poly i))
	(else
	 0)))

;; drop any high zero coefficients from POLY
(define (poly-trim poly)
  (drop-while-right zero? poly))

(define (poly-neg poly)
  (map - poly))

(define (poly-sub a b)
  (poly-trim (poly-add a (poly-neg b))))

(define (poly-mul-scalar poly x)
  (map (lambda (coeff)
	 (* coeff x))
       poly))

(define (poly-mul a b)
  (apply poly-add
	 (map (lambda (coeff pos)
		(poly-shift (poly-mul-scalar b coeff)
			    pos))
	      a (iota-cached (length a)))))

(define (poly-add . lst)
  (if (null? lst)
      '()
      (let ((new (make-list (apply max (map length lst)) 0)))
	(for-each (lambda (poly)
		    (for-each (lambda (coeff pos)
				(list-set! new pos
					   (+ coeff (list-ref new pos))))
			      poly (iota-cached (length poly))))
		  lst)
	new)))

;; return poly*x^k, ie. shift all the coeffs along k places
(define (poly-shift poly k)
  (append (make-list k 0) poly))

;; return two (values Q R) which satisfy N = Q*D + R, with the highest term
;; (ie. highest power of x) in R begin less than the highest term in D
(define (poly-qr n d)
  (set! d (poly-trim d))
  (if (poly-zero? d)
      (error "poly-qr: divide by zero" n d))
  (let* ((nlen  (length n))
	 (dlen  (length d))
	 (dhigh (last d))
	 (qlen  (- nlen dlen -1))
	 (q     (make-list qlen 0))
	 (r     n))
    (do ((rpos (1- nlen)     (1- rpos))
	 (qpos (- nlen dlen) (1- qpos)))
	((negative? qpos))
      (let ((q-scalar (/ (poly-coeff r rpos) dhigh)))
	(set! r (poly-sub r (poly-shift (poly-mul-scalar d q-scalar)
					qpos)))
	(list-set! q qpos q-scalar)))
    (assert (poly-equal? n (poly-add r (poly-mul q d))))
    (values q r)))

(define (poly-zero? poly)
  (every zero? poly))

(define (poly-not-zero? poly)
  (not (poly-zero? poly)))

(define (poly-div n d)
  (receive (q r)
      (poly-qr n d)
    (or (poly-zero? r)
	(begin
	  (format (current-error-port)
		  "n ~a\nd ~a\nq ~a\nr ~a\n" n d q r)
	  (error "remainder")))
    q))

(define* (poly-print poly #:optional (var "k"))
  (define plus "")
  (define minus "-")
  (for-each (lambda (coeff power)
	      (if (not (zero? coeff))
		  (begin
		    (if (positive? coeff)
			(display plus)
			(begin
			  (display minus)
			  (set! coeff (abs coeff))))
		    (if (or (= 0 power)
			    (not (= 1 coeff)))
			(begin
			  (format #t "~a" coeff)
			  (if (>= power 1)
			      (display "*"))))
		    (if (>= power 1)
			(display var))
		    (if (>= power 2)
			(format #t "^~a" power))
		    (set! plus  " + ")
		    (set! minus " - "))))
	    poly (iota-cached (length poly)))
  (if (string=? "" plus)
      (display "0")))
;; (reverse poly) (iota-range (1- (length poly)) 0 -1))

;; print with parens if more than one term
(define (poly-print-parens poly . rest)
  (if (>= (count not-zero? poly) 2)
      (begin
	(display "(")
	(apply poly-print poly rest)
	(display ")"))
      (apply poly-print poly rest)))

(define (poly-equal? a b)
  (equal? (poly-trim a) (poly-trim b)))

;; return true if any coefficient in POLY is a fraction, ie. not an integer
(define (poly-frac-coeff? poly)
  (any (lambda (coeff)
	 (not (integer? coeff)))
       poly))


;;-----------------------------------------------------------------------------
;; polynomials as scheme expressions to eval

;; return a scheme expression which evaluates POLY, in a free variable `x'
;; eg. poly (1 2 3) is 1 + 2*x + 3*x^2 and gives a return `(+ 1 (* x (+ 2 x)))'
(define (poly->expr poly)
  (expr-simplify (poly->expr-raw (drop-while-right zero? poly))))

(define (poly->expr-raw poly)
  (if (null? poly)
      0
      (list '+ (first poly) (list '* 'k (poly->expr-raw (cdr poly))))))

(define (expr-simplify expr)
  (let ((new-expr (expr-simplify-one expr)))
    (if (equal? expr new-expr)
	expr
	(expr-simplify new-expr))))

(define (expr-simplify-one expr)
  (cond ((and (list? expr)       ;; (+ consts...)
	      (eq? '+ (first expr))
	      (every number? (cdr expr)))
	 (apply + (cdr expr)))
	((and (list? expr)       ;; (+ x 0)
	      (eq? '+ (first expr))
	      (memv 0 expr))
	 (expr-simplify-one (delv 0 expr)))

	((and (list? expr)       ;; (+ 1 x)
	      (= 3 (length expr))
	      (eq? '+ (first expr))
	      (eqv? 1 (second expr)))
	 (list '1+ (expr-simplify-one (third expr))))
	((and (list? expr)       ;; (+ -1 x)
	      (= 3 (length expr))
	      (eq? '+ (first expr))
	      (eqv? -1 (second expr)))
	 (list '1- (expr-simplify-one (third expr))))

	((and (list? expr)       ;; (+ x (- y))
	      (= 3 (length expr))
	      (eq? '+ (first expr))
	      (list? (third expr))
	      (= 2 (length (third expr)))
	      (eq? '- (first (third expr))))
	 (list '-
	       (expr-simplify-one (second expr))
	       (second (third expr))))

	((and (list? expr)       ;; (1- (- x))
	      (= 2 (length expr))
	      (eq? '1- (first expr))
	      (list? (second expr))
	      (= 2 (length (second expr)))
	      (eq? '- (first (second expr))))
	 (list '- -1 (second (second expr))))

	((and (list? expr)       ;; (+ x) or (* x)
	      (= 2 (length expr))
	      (memq (first expr) '(+ *)))
	 (expr-simplify-one (second expr)))

	((and (list? expr)       ;; (1+ const)
	      (= 2 (length expr))
	      (eq? '1+ (first expr))
	      (number? (second expr)))
	 (1+ (second expr)))
	((and (list? expr)       ;; (1- const)
	      (= 2 (length expr))
	      (eq? '1- (first expr))
	      (number? (second expr)))
	 (1- (second expr)))

	((and (list? expr)       ;; (* x 1)
	      (eq? '* (first expr))
	      (memv 1 expr))
	 (expr-simplify-one (delv 1 expr)))
	((and (list? expr)       ;; (* x 0)
	      (eq? '* (first expr))
	      (memv 0 expr))
	 0)
	((and (list? expr)       ;; (* x -1)
	      (= 3 (length expr))
	      (eqv? -1 (third expr)))
	 (list '- (expr-simplify-one (second expr))))

	((and (list? expr)       ;; (* x (* ...)) -> (* x ...)
	      (length>=? expr 3)
	      (eq? '* (first expr))
	      (list? (last expr))
	      (eq? '* (first (last expr))))
	 (expr-simplify-one
	  (append (drop-right expr 1)
		  (drop (last expr) 1))))

	((list? expr)
	 (map expr-simplify-one expr))
	(else
	 expr)))


;;-----------------------------------------------------------------------------

;; PROC is a procedure to be called (PROC x).  Return a polynomial
;; representing its calculation.  The polynomial is a list of
;; coefficient/exponent pairs such as ((2 . 3) (4 . 5)) for 2*x^3 + 4*x^5.
;;
(define-macro (closed-form proc . rest)
  `(closed-form-func ',proc ,proc ,@rest))
(define* (closed-form-func name proc #:key (den 1))
  (format #t "~a\n" name)
  (let* ((poly (closed-poly proc))
	 (expr (poly->expr poly))
	 (int? (poly-integer? poly)))
    (check proc poly)
    (check (expr->proc expr) poly)

    (display "  ") (poly-print poly) (newline)
    (format #t "  ~s\n" expr)

    (if (and int?
	     (poly-frac-coeff? poly))
	(display "  always an integer\n"))

    (if (or (> den 1)
	    (poly-frac-coeff? poly))
	(let* ((den     (lcm den (poly-denominator poly)))
	       (intpoly (poly-mul-scalar poly den))
	       (expr    (list (if int? 'quotient '/)
			      (poly->expr intpoly) den)))
	  (check (expr->proc expr) poly)
	  (display "  (")
	  (poly-print intpoly)
	  (format #t ") / ~a\n" den)
	  (format #t "  ~s\n" expr)))))

(define (expr->proc expr)
  (eval `(lambda (k) ,expr)
	(interaction-environment)))



;; Check that PROC and POLY really do evaluate to the same values on all the
;; `i-points' test arguments.
(define (check proc poly)
  (define poly-proc (make-poly-proc poly))
  (for-each (lambda (n)
	      (or (eqv? (proc n) (poly-proc n))
		  (begin
		    (format (current-error-port)
			    "proc ~a\n" proc)
		    (format (current-error-port)
			    "poly ~a\n" poly)
		    (format (current-error-port)
			    "n    ~a\n" n)
		    (format (current-error-port)
			    "proc got ~a\n" (proc n))
		    (format (current-error-port)
			    "poly got ~a\n" (poly-proc n))
		    (error "different"))))
	    i-points))

;; A list '(1 2 ... 99 100) which are the test points for the closed form
;; determinations.
(define num-points 30)
(define i-points #f)

;; Return a procedure to be called (PROC x) which evaluates POLY at X.
(define (make-poly-proc poly)
  (define power-list (iota-cached (length poly)))
  (lambda (n)
    (apply + (map (lambda (coeff power)
		    (* coeff (integer-expt n power)))
		  poly power-list))))

(define* (closed-poly proc #:optional (start 1))
  (set! i-points (iota num-points start))
  (closed-data (map proc i-points)))

;; DATA is a list of values which are some function f(x) for x values from
;; the `i-points' test list.  Return a polynomial which represents that
;; function.
;;
(define (closed-data data)
  (define poly (make-list 30 0))
  (let more ()
    (receive (coeff power)
	(high-power data)
      (list-set! poly power coeff)

      (set! data (map (lambda (x n)
			(- x (* coeff (integer-expt n power))))
		      data i-points))
      (if (every zero? data)
	  poly
	  (more)))))

(define (high-power data)
  (let more ((data  data)
	     (power 0)
	     (deriv 1))
    (if (apply eqv? data)
	(values (/ (first data) deriv) power)

	(more (map - (cdr data) data)
	      (1+ power)
	      (* deriv (1+ power))))))

(define (poly-denominator poly)
  (apply lcm (map denominator poly)))

;; return true if POLY is an integer for any integer value X
(define (poly-integer? poly)
  (let ((den  (poly-denominator poly))
	(proc (make-poly-proc poly)))
    (every (lambda (n)
	     (integer? (proc n)))
	   (iota-cached (* 3 den)))))


;;-----------------------------------------------------------------------------

(display "\nMisc closed\n")

(define (triangular n)
  (ash (* n (1+ n)) -1))

(define (ref:yfactor n)
  (- (* n (apply + (map square (iota n 1))))
     (square (triangular n))))

;; sum 1^2 + 2^2 + 3^2 + ... + n^2
(define (sum-squares n)
  (apply + (map square (iota n 1))))

(define (triang-squared n)
  (square (triangular n)))

(define (epma-total-weight n)
  (apply + (iota n (+ (- n) 2) 3)))

(define (mma-total-weight-A n)
  (let ((slopes (iota n (/ (- n 1) 2) -1)))
    (+ 1 (* 6 (/ (apply + slopes)
		 n
		 (1+ n))))))

(define (mma-total-weight-B n)
  (apply + (iota n (- (* 4 n) 2) -6)))

;; denominator of factor used in linreg-stderr.scm
(define (xfactor-den-old n)
  (let ((x-total (apply + (iota n 1)))
	(xx-total (apply + (map square (iota n 1)))))
    (- (* n xx-total)
       (* x-total x-total))))

;; denominator of factor used in linreg-stderr.scm
(define (xfactor-den n)
  (* n (apply + (map square (iota n (* 1/2 (- (1- n))))))))

;; sum used in `xx2-calc' in linreg.scm
;; eg. n==4 is (-1.5)^2 + (-0.5)^2 + (0.5)^2 + (1.5)^2
(define (xx-calc n)
  (apply + (map square (iota n (* -1/2 (1- n))))))

(closed-form triangular)
(closed-form ref:yfactor)
(closed-form sum-squares)
(closed-form triang-squared)
(closed-form epma-total-weight)
(closed-form mma-total-weight-A)
(closed-form mma-total-weight-B)
(closed-form xfactor-den-old)
(closed-form xfactor-den)
(closed-form xx-calc)

;; terms in `ema-3-omitted'
(define (ema3-a k)
  (* 2 (- (+ (* -3 (triangular (1+ k)))
	     (*  3 (triangular k))
	     (* -1 (triangular (1- k)))))))
(define (ema3-b k)
  (* 2 (- (+ (*  3 (triangular (1+ k)))
	     (* -1 (triangular k))))))
(define (ema3-c k)
  (* 2 (- (+ (* -1 (triangular (1+ k)))))))
(closed-form ema3-a)
(closed-form ema3-b)
(closed-form ema3-c)

(define (ema4-term k)
  (let ((sum 0))
    (do ((i 1 (1+ i)))
	((> i (1+ k)))
      (set! sum (+ sum (triangular i))))
     sum))
(dv (ema4-term 0) (ema4-term 1) (ema4-term 2))
(closed-form ema4-term)



;;-----------------------------------------------------------------------------
;; ends up as successively broader triangular moving averages

(display "\nBandini weights\n")
(let* ((prev (list 1024 1024))
       (lst  (cons prev
		   (map (lambda (n)
			  (set! prev
				(map (lambda (x y)
				       (/ (+ x y) 2))
				     (append '(0) prev)
				     (append prev '(0))))
			  prev)
			(iota 10)))))
  (for-each (lambda (elem)
	      (dv elem))
	    lst))       


;;-----------------------------------------------------------------------------
;; ema weights
;;

(display "\nEMA\n")

(define (print-ma-terms name term-proc)
  (format #t "~a\n" name)
  (for-each (lambda (k)
	      (format #t "  p[~a] * " k)
	      (poly-print-parens (term-proc k) "f")
	      (newline))
	    (iota-cached 6)))

;; `ema-term' returns a polynomial in f=1-alpha which is the coefficient of
;; the f^k term.  This is simply (1-f)*f^k
(define ema-term
  (cached (lambda (k)
	    (if (negative? k)
		'()
		(let ((poly (make-list (+ k 2) 0)))
		  (list-set! poly k      1)
		  (list-set! poly (1+ k) -1)
		  poly)))))
(print-ma-terms "ema" ema-term)
(assert (equal? (ema-term 0) '(1 -1)))
(assert (equal? (ema-term 1) '(0 1 -1)))
(assert (equal? (ema-term 2) '(0 0 1 -1)))
(assert (equal? (ema-term 3) '(0 0 0 1 -1)))

;; return a procedure WEIGHTTO which called (WEIGHTTO k) calculates the
;; total weight of the terms up to and including the p[k] input price term.
;;
;; (TERM-PROC i) should return a polynomial in f which is the coefficient of
;; the f^i term.  The created WEIGHTTO function simply adds them up.
;;
(define (make-weight-to-terms term-proc)
  (letrec ((weight-proc (cached
			 (lambda (k)
			   (if (zero? k)
			       (term-proc k)
			       (poly-add (term-proc k)
					 (weight-proc (1- k))))))))
    weight-proc))

;; (ema-weight-to k) returns a polynomial in f which is the total weight of
;; terms up to and including the p[k] input.
(define ema-weight-to (make-weight-to-terms ema-term))
(dv (ema-weight-to 4))

;; (ema-omitted-after k) returns a polynomial in f which is the total weight
;; omitted by stopping at (and including) the p[k] input.
(define ema-omitted-after
  (cached (lambda (k)
	    (poly-sub '(1) (ema-weight-to k)))))
(dv (ema-omitted-after 4))


;;-----------------------------------------------------------------------------
;; ema-2 weights

(display "\nEMA-2 -- EMA of EMA\n")

;; (TERM-PROC i) should return a polynomial in f which is the weight of the
;; p[i] input term in some weighted mean series.
;; `make-ema-on-terms' returns a new procedure SMOOTHED-PROC which called
;; (SMOOTHED-PROC i) similarly returns a polynomial in f, but representing
;; the result of smoothing TERM-PROC with an EMA.
;;
(define (make-ema-on-terms term-proc)
  (cached (lambda (k)
	    (apply poly-add (map (lambda (i)
				   (poly-mul (ema-term i)
					     (term-proc (- k i))))
				 (iota-range 0 k))))))
(define ema2-term (make-ema-on-terms ema-term))
(print-ma-terms "ema-2" ema2-term)
(dv (ema2-term 4))

(define ema2-weight-to (make-weight-to-terms ema2-term))
(dv (ema2-weight-to 4))
(define ema2-omitted-after
  (cached (lambda (k)
	    (poly-sub '(1) (ema2-weight-to k)))))
(dv (ema2-omitted-after 4))


;; OMITTED-POLY-PROC returns a polynomial in f which is the weight omitted
;; from some weighted average by stopping at (and including) the p[k] term.
;;
;; `omitted-fk' returns the coefficient (a number) of the f^(k+offset) term
;; in the p[k] omitted polynomial of OMITTED-POLY-PROC.  This is meant for
;; finding closed-form expressions (a poly in k) for that f^(k+offset)
;; coefficient.
;;
(define (omitted-fk omitted-poly-proc offset)
  (cached (lambda (k)
	    (poly-coeff (omitted-poly-proc k)
			(+ k offset)))))

(defmacro* omitted-expr (omitted-poly-proc #:optional (den 1))
  `(omitted-expr-func ',omitted-poly-proc ,omitted-poly-proc ,den))
(define (omitted-expr-func name omitted-poly-proc den)
  (let ((array (make-array #f '(-5 20))))
    (array-index-map!
     array (lambda (offset)
	     (let ((poly (closed-poly (omitted-fk omitted-poly-proc
						  offset))))
	       (if (not (poly-zero? poly))
		   (begin
		     (format #t "f^(k~@d) * " offset)
		     (poly-print-parens poly "k")
		     (newline)))
	       poly)))

    (let ((start (array-find-pos array poly-not-zero?))
	  (end   (array-find-last-pos array poly-not-zero?))
	  (level 1))
      (format #t "(* (expt f ~a)\n" (offset-k-form start))
      (do ((offset start (1+ offset)))
	  ((> offset end))
	(if (not (= offset start))
	    (begin
	      (format #t "~a  (* f             ;; f^(k~@d)\n"
		      (make-string level #\space)
		      offset)
	      (set! level (1+ level))))
	(format #t "~a  (+ ~a\n"
		(make-string level #\space)
		(poly->expr (array-ref array offset)))
	(set! level (1+ level)))
      (format #t "~a~a\n"
	      (make-string level #\space)
	      (make-string level #\) )))))

(define (offset-k-form offset)
  (case offset
    ((-1) '(1- k))
    ((0)  'k)
    ((1)  '(1+ k))
    (else
     (list (if (positive? offset) '+ '-) 'k (abs offset)))))

(closed-form (omitted-fk ema2-omitted-after 0))
(closed-form (omitted-fk ema2-omitted-after 1))
(closed-form (omitted-fk ema2-omitted-after 2))
(closed-form (omitted-fk ema2-omitted-after 3))
(omitted-expr ema2-omitted-after)


;;-----------------------------------------------------------------------------
;; ema-3 weights

(display "\nEMA-3 -- EMA of EMA of EMA\n")

(define ema3-term (make-ema-on-terms ema2-term))
(dv (ema3-term 4))
(print-ma-terms "ema-3" ema3-term)

(define ema3-weight-to (make-weight-to-terms ema3-term))
(dv (ema3-weight-to 4))
(define ema3-omitted-after
  (cached (lambda (k)
	    (poly-sub '(1) (ema3-weight-to k)))))
(dv (ema3-omitted-after 4))

(closed-form (omitted-fk ema3-omitted-after 0) #:den 2)
(closed-form (omitted-fk ema3-omitted-after 1) #:den 2)
(closed-form (omitted-fk ema3-omitted-after 2) #:den 2)
(closed-form (omitted-fk ema3-omitted-after 3) #:den 2)
(closed-form (omitted-fk ema3-omitted-after 4) #:den 2)
(omitted-expr ema3-omitted-after)


;;-----------------------------------------------------------------------------
;; laguerre filter weights

(display "\nLaguerre filter\n")

(define L0-term ema-term)
(dv (L0-term 6))

;; From
;;
;;     L1 = -(1-alpha)*L0 + L0prev + (1-alpha)*L1prev
;;
;; rearrange to show it's an EMA,
;;
;;                  L0prev - f*L0
;;     L1 = alpha * ------------- + (1-alpha)*L1prev
;;                      (1-f)
;;
;; of the term
;;
;;     L0prev - f*L0
;;     -------------
;;         (1-f)
;;

;; (define (L1-pre-ema-term k)
;;   (poly-div (poly-sub (if (zero? k)
;; 			  '()
;; 			  (L0-term (1- k)))
;; 		      (poly-shift (L0-term k) 1))
;; 	    '(1 -1)))

(define (make-laguerre-for-ema term-proc)
  (cached (lambda (k)
	    (poly-div (poly-sub (if (zero? k)
				    '()
				    (term-proc (1- k)))
				(poly-shift (term-proc k) 1))
		      '(1 -1)))))

(define L1-pre-ema-term (make-laguerre-for-ema L0-term))
(print-ma-terms "L1-pre-ema" L1-pre-ema-term)
(define L1-term (make-ema-on-terms L1-pre-ema-term))
(dv (L1-term 6))
(print-ma-terms "L1" L1-term)

(define L2-term
  (make-ema-on-terms (make-laguerre-for-ema L1-term)))
(dv (L2-term 6))
(print-ma-terms "L2" L2-term)

(define L3-term
  (make-ema-on-terms (make-laguerre-for-ema L2-term)))
(dv (L3-term 6))
(print-ma-terms "L3" L3-term)

(define laguerre-term
  (cached (lambda (k)
	    (poly-mul-scalar (poly-add (L0-term k)
				       (poly-mul-scalar (L1-term k) 2)
				       (poly-mul-scalar (L2-term k) 2)
				       (L3-term k))
			     1/6))))
(dv (laguerre-term 6))
(print-ma-terms "Laguerre" laguerre-term)

(define laguerre-weight-to (make-weight-to-terms laguerre-term))
(dv (laguerre-weight-to 10))
(define laguerre-omitted-after
  (cached (lambda (k)
	    (poly-sub '(1) (laguerre-weight-to k)))))
(dv (laguerre-omitted-after 10))


(set! i-points (iota 20 7))
(closed-form (omitted-fk laguerre-omitted-after -5) #:den 36)
(closed-form (omitted-fk laguerre-omitted-after -4) #:den 36)
(closed-form (omitted-fk laguerre-omitted-after -3) #:den 36)
(closed-form (omitted-fk laguerre-omitted-after -2) #:den 36)
(closed-form (omitted-fk laguerre-omitted-after -1) #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 0)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 1)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 2)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 3)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 4)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 5)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 6)  #:den 36)
(closed-form (omitted-fk laguerre-omitted-after 7)  #:den 36)
(omitted-expr laguerre-omitted-after 36)

(exit 0)


;; Local variables:
;; mode: scheme
;; End:
