#!/bin/sh
exec guile -s "$0" "$@"
!#

;; Check defines provided by configure vs what the sources use.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: guile -s configure-defs.scm [sourcedir [builddir]]
;;

(use-modules (ice-9 and-let-star)
	     (ice-9 rdelim)
	     (ice-9 regex)
	     (srfi srfi-1)
	     (srfi srfi-13)
	     (srfi srfi-14)
	     (chart misc))


(define srcdir ".")
(if (length>=? (program-arguments) 2)
    (set! srcdir (second (program-arguments))))

(define builddir ".")
(if (length>=? (program-arguments) 3)
    (set! builddir (third (program-arguments))))


;;-----------------------------------------------------------------------------
;; misc

;; return the number of entries in the given hash TABLE
(define (hash-count-entries table)
  (hash-fold (lambda (key value prev)
	       (1+ prev))
	     0 table))

(define (c-or-h-filename? filename)
  (or (string-suffix? ".c" filename)
      (string-suffix? ".h" filename)))

(define (preprocessor-def? str)
  (or (string-prefix? "HAVE_" str)
      (string-prefix? "SIZEOF_" str)
      (string-prefix? "ENABLE_" str)
      (hash-get-handle config-hash str))) ;; anything in config.h

;; the characters valid in a C symbol, per C99 section 6.4.2.1
(define char-set:C-symbol
  (char-set-adjoin char-set:letter+digit #\_))



;;-----------------------------------------------------------------------------

(define config-hash (make-hash-table 129))

(call-with-input-file (string-append srcdir "/config.h.in")
  (lambda (port)
    (do ((line (read-line port) (read-line port)))
	((eof-object? line))
      (and-let* ((m (regexp-exec #,(regexp "#undef ([^ ]+)") line)))
	(hash-set! config-hash (match:substring m 1) #f)))))
(format #t "config.h.in ~a defines\n"
	(hash-count-entries config-hash))

;; autoconf builtins marked as used
(for-each (lambda (str)
	    (and-let* ((handle (hash-get-handle config-hash str)))
	      (set-cdr! handle #t)))
	  '("PACKAGE"
	    "PACKAGE_BUGREPORT"
	    "PACKAGE_NAME"
	    "PACKAGE_STRING"
	    "PACKAGE_TARNAME"
	    "PACKAGE_VERSION"
	    "STDC_HEADERS"
	    "VERSION"))


;;-----------------------------------------------------------------------------

(define source-files-count 0)
(define sources-count 0)

(define (check-token port token)
  (if (preprocessor-def? token)
      (begin
	(set! sources-count (1+ sources-count))
	(let ((handle (hash-get-handle config-hash token)))
	  (if handle
	      (set-cdr! handle #t)
	      (format #t "~a:~a: ~a unknown\n"
		      (port-filename port)
		      (port-line port)
		      token))))))

(define (check-file filename)
  (set! source-files-count (1+ source-files-count))
  (call-with-input-file filename
    (lambda (port)
      (do ((line (read-line port) (read-line port)))
	  ((eof-object? line))
	(for-each (lambda (token)
		    (check-token port token))
		  (string-tokenize line char-set:C-symbol))))))


(for-each check-file
	  (directory-files srcdir #t c-or-h-filename?))

(let ((unused-lst (hash-fold (lambda (key value prev)
			       (if value
				   prev
				   (cons key prev)))
			     '() config-hash)))
  (for-each (lambda (str)
	      (format #t "config.h.in: unused: ~a\n" str))
	    (sort! unused-lst string<?)))

(format #t "sources ~a files using ~a defines\n"
	source-files-count sources-count)

(exit 0)


;; Local variables:
;; mode: scheme
;; End:
