#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Mung database contents, for development.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(use-modules (ice-9 and-let-star)
	     (srfi srfi-1)
	     (chart australia)
	     (chart database)
	     (chart download)
	     (chart misc))

;; as per Guile 1.8, made available in 1.6
(if (not (defined? '@@))
    (define-macro (@@ mod-name var-name)
      (let ((var (module-variable (resolve-module mod-name) var-name)))
	(if (not var)
	    (error "no such variable" (list '@@ mod-name var-name)))
	(list variable-ref var))))

(define (mung-series proc symlist)
  (for-each (lambda (sym)
	      (and-let* ((series (database-read-series sym)))
		(proc series)
		(database-write-series series)))
	    symlist))


;;-----------------------------------------------------------------------------
;; various series munging

(define-public (series-prune-hi series)
  (let ((i (series-hi series)))
    (let ((array (series-array series i i)))
      (if (and (eq? #f (array-ref array i 0))
	       (eq? #f (array-ref array i 1))
	       (eq? #f (array-ref array i 2))
	       (eq? #f (array-ref array i 3))
	       (eq? #f (array-ref array i 4)))
	  (begin
	    (series-hi! series (1- i))
	    (series-prune-hi series))))))

(define-public (series-zap-zero-vol series)
  (let ((array (series-array series (series-lo series) (series-hi series))))
    (do ((i (array-first-index array) (1+ i)))
	((> i (array-last-index array)))
      (if (and (eqv? 0 (array-ref array i 4))  ;; no volume
	       (eqv? (array-ref array i 0)     ;; and ohlc all equal
		     (array-ref array i 1)
		     (array-ref array i 2)
		     (array-ref array i 3)))
	  (begin
	    (array-set! array #f i 0)
	    (array-set! array #f i 1)
	    (array-set! array #f i 2)
	    (array-set! array #f i 3)
	    (array-set! array #f i 4))))))

(define-public (series-close-only series)
  (let ((array (series-array series (series-lo series) (series-hi series))))
    (do ((i (array-first-index array) (1+ i)))
	((> i (array-last-index array)))
      (if (and (eqv? 0 (array-ref array i 4))  ;; no volume
	       (eqv? (array-ref array i 0)     ;; and ohlc all equal
		     (array-ref array i 1)
		     (array-ref array i 2)
		     (array-ref array i 3)))
	  (begin
	    (array-set! array #f i 0) ;; o
	    (array-set! array #f i 1) ;; h
	    (array-set! array #f i 2) ;; l
	    ;; leave close only
	    (array-set! array #f i 4)))))) ;; v

(define-public (Mmm_yyyy-str->mdate str)
  (receive-list (month year)
      (string-split str #\space)
    (ymd->mdate (string->number year)
		(Mmm-str->month month)
		1)))		

(define-public (series-no-file-date series)
  (series-info! series (assq-remove! (series-info series) 'file-date)))

(define-public (series-kill-dividends series)
  (series-dividends! series '()))

;; dividends "May 2005" etc convert to (month MDATE)
(define-public (sfe-series-rollovers series)
  (for-each (lambda (pair)
	      (set-cdr! pair (list 'month (Mmm_yyyy-str->mdate (cdr pair)))))
	    (series-dividends series)))

;; dividends "ex" changed to "unknown"
(define-public (series-dividends-ex->unknown series)
  (for-each (lambda (pair)
	      (if (equal? (cdr pair) "ex")
		  (set-cdr! pair "unknown")))
	    (series-dividends series)))

;; dividends strip any " D.R.P" suffix
(define-public (series-dividends-lose-drp series)
  (for-each (lambda (pair)
	      (let (;; (tdate  (car pair))
		    (amount (cdr pair)))
		(if (string? amount)
		    (begin
		      (set! amount (string-trim-right-str amount " D.R.P"))
		      (set! amount (string-trim-right-str amount " D.R.P."))
		      (set! amount (string-trim-right-str amount " US 15% W/H TAX"))
		      (set! amount (string-trim-right-str amount " DRP SUSPENDED"))
		      ;; become "unknown" if then empty
		      (if (string-null? amount)
			  (set! amount "unknown"))))
		(set-cdr! pair amount)))
	    (series-dividends series)))

;; convert series dividends from strings to amount objects
(define-public (series-dividends-parse series)
  (define yahoo-dividend-parse (@@ (chart yahoo) yahoo-dividend-parse))
  (for-each (lambda (pair)
	      (let (;; (tdate  (car pair))
		    (amount (cdr pair)))
		(if (string? amount)
		    (set! amount (yahoo-dividend-parse amount)))
		(set-cdr! pair amount)))
	    (series-dividends series)))

;; delete download-done date
(define-public (series-delete-download-done series)
  (series-info! series (assq-remove! (series-info series) 'download-done)))

(define-public (array-contents->list array)
  (let ((lst (array->list array)))
    (do ((i (array-rank array) (1- i)))
	((<= i 1))
      (set! lst (concatenate! lst)))
    lst))

(define-public (array-every pred array)
  (every pred (array-contents->list array)))

(define-public (multiple-of-10? n)
  (= 0 (modulo n 10)))



(define-public (series-zap-9758 series)
  (let ((t  9758)
	(hi (series-hi series)))
    (if (>= hi t)
	(let* ((array (series-array series (series-lo series) hi))
	       (part  (make-shared-array-restrict array t hi)))
	  (dv (series-symbol series) part)
	  (array-fill! part #f))))
  (series-delete-download-done series))

(define-public (series-info-delete-possible-rollover series)
  (series-info! series (alist-delete 'possible-rollover (series-info series))))

(define-public (series-rollovers-to-strings series)
  (for-each (lambda (elem)
	      ;; wanting (TDATE . (month MDATE STRINGNUMBER (decimals N)))
	      (if (and (list? (cdr elem))
		       (eq? 'month (second elem))
		       (length>=? elem 4)
		       (number? (fourth elem)))
		  (let ((tail (cdddr elem)))
		    (dv elem)
		    (set-car! tail (number+decimals->string
				    (fourth elem)
				    (assq-ref (cddddr elem) 'decimals)))
		    (set-cdr! tail '())
		    (dv elem))))
	    (series-dividends series)))

;; Reduce the number of decimals in SERIES, to the value of TARGET.
;; Decimal places are stripped only if all open/high/low/close have a
;; trailing zero there.
;; This is good if a data source seems to have given some unnecessary
;; trailing zeros.  The data source code is supposed to cope with that, but
;; sometimes it slips though.
;;
(define-public (series-reduce-decimals series)
  (let* ((target 2)
	 (lo     (series-lo series))
	 (hi     (series-hi series))
	 (array  (series-array series lo hi)))
    (set! array (make-shared-array-columns array 0 3))

    (let more ()
      (if (> (series-decimals series) target)
	  (let ((lst (delq #f (array-contents->list array))))
	    (cond ((null? lst)
		   #f)
		  ((every multiple-of-10? lst)
		   (dv "reduce" (series-symbol series))
		   (array-map! array (lambda (n)
				       (and n
					    (quotient n 10)))
			       array)

		   (series-notes! series (notes-change-decimals
					  (series-notes series)
					  (series-decimals series)
					  (1- (series-decimals series))))
		   (series-decimals! series (1- (series-decimals series)))
		   (database-write-notes series)
		   (more))

		  (else
		   (let ((example (find (lambda (price)
					  (not (multiple-of-10? price)))
					lst)))
		     (format #t "~a: ~a decimals, eg. ~a\n"
			     (series-symbol series)
			     (series-decimals series)
			     example)))))))))


;;-----------------------------------------------------------------------------

(mung-series series-delete-download-done
	     (filter australia-symbol? (database-symbols)))

;;	     (database-symbols)



;; Local variables:
;; mode: scheme
;; compile-command: "./mung.scm"
;; End:
