#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Exercise intraday graph downloading.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: chart -s online-intraday.scm
;;
;; Check intraday URLs for some symbols and sources.  The URLs are validated
;; by making a "HEAD" request, so there's no great amount downloaded, except
;; for some sources, in particular barchart, where getting the image URL
;; needs some other downloading too.
;;
;; Because this program does online interaction it's not run by "make
;; check".
;;

(use-modules (ice-9 receive)
	     (srfi srfi-1)
	     (chart download)
	     (chart http)
	     (chart intraday)
	     (chart india)
	     (chart misc)
	     (chart toronto))


(define good #f)

(define (attempt symbol)
  (define mode-list (intraday-modes symbol))
  (if (null? mode-list)
      (format #t "no modes for: ~a\n" symbol)

      (every
       (lambda (mode)
	 (let ((url (intraday-url symbol mode)))
	   (if (not url)
	       (begin
		 (format #t "no intraday for: ~a\n" symbol)
		 (set! good #f))

	       (receive (headers body)
		   (http-request url #:method "HEAD" #:follow #f)
		 (let ((status-code (http-headers-status-code headers)))
		   (cond ((http-status-code-ok? status-code)
			  #t)
			 (else
			  (format #t "intraday error: ~a\n" symbol)
			  (format #t "  ~a\n" url)
			  (format #t "  ~a\n" (http-headers-error-str headers))
			  (set! good #f)
			  #f)))))))
       mode-list)))


(define fake-preference-alist '())
(define (fake-preference-set key value)
  (set! fake-preference-alist (assq-set! fake-preference-alist key value)))
(define (fake-preference-get key default)
  (or (assq-ref fake-preference-alist key)
      default))
(set! preference-get fake-preference-get)


(define (all)

  ;; wtb
  (for-each attempt
	    '("W.WTB"
	      "B.WTB"
	      "BX07.WTB"  ;;
	      "LP.WTB"
	      "LPJ08.WTB" ;; Apr08
	      "H.WTB"
	      "F.WTB"
	      "P.WTB"
	      "XP.WTB"
	      "POTX.WTB"
	      "TAPX.WTB"
	      "WTBBPC.WTB"))


  ;; kcbt
  (for-each attempt
	    '("KW.KCBT"
	      "MV.KCBT"
	      "MVM06.KCBT"
	      "MVM06.KCBT"))

  (for-each-two (lambda (host timezone)
		  (fake-preference-set 'yahoo-quote-host     host )
		  (fake-preference-set 'yahoo-quote-timezone timezone)

		  (format #t "from ~a  ~a\n" host timezone)

		  ;; yahoo
		  (for-each attempt
			    '("^SPX"
			      "^NZ50"
			      "^AORD"
			      "AAPL"
			      "GM"
			      "BHP.AX"
			      "LCA.DE"
			      "TSCO.L")))

		(list "finance.yahoo.com"    (timezone-newyork)
		      "au.finance.yahoo.com" (timezone-sydney)
		      "ca.finance.yahoo.com" (timezone-toronto)
		      "in.finance.yahoo.com" (timezone-calcutta)
		      "sg.finance.yahoo.com" (timezone-singapore)
		      "uk.finance.yahoo.com" (timezone-london)
		      ))

  ;; sfe
  (for-each attempt
	    '("IR.SFE"
	      "BB.SFE"
	      "FW.SFE"))

  ;; mammon
  (attempt "BILLION.MAM")

  ;; fatdog
  (attempt "Mark Webber.FAT")

  ;; kex
  (for-each attempt
	    '("CF.KEX"
	      "EBI.KEX"
	      "KI.KEX"
	      "CF.KEX"
	      "N.KEX"
	      "RB.KEX"))

  ;; mgex
  (for-each attempt
	    '("MW.MGEX"
	      "IC.MGEX"
	      "ICN06.MGEX"
	      "ICU06.MGEX"))

  ;; barchart
  (for-each attempt
	    '("O.CBOT"
	      "GC.COMEX"
	      ;; "NK.SIMEX"
	      "RS.WCE"))

  )


(c-main-enq! all)
(c-main)
(exit (if good 0 1))


;; Local variables:
;; mode: scheme
;; compile-command: "chart -s online-intraday.scm"
;; End:
