#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Check external URLs in the chart.html manual.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: chart -s online-manual.scm [chart.html]
;;
;; Check external href URL links in the Chart manual.  The chart.html file
;; can be given on the command line, otherwise "../doc/chart.html" is used
;; (for when this program is run from the devel subdir).
;;
;; Links are checked with a HEAD request, so the amount downloaded is
;; minimal.  301 and 302 redirects are warned about, since usually prefer
;; the full correct target in the manual in the first place.
;;
;; Because this program does some online interaction it's not run by "make
;; check".
;;

(use-modules (ice-9 receive)
	     (ice-9 regex)
	     (srfi srfi-1)
	     (chart extra-glib)
	     (chart http)
	     (chart misc))

(define verbose #f)
(define good-count 0)

(define filename
  (cond ((length>=? (program-arguments) 2)
	 (second (program-arguments)))
	((getenv "srcdir")
	 => (lambda (dir)
	      (g-build-filename dir "../doc/chart.html")))
	(else
	 (display "online-manual.scm: no filename or srcdir specified\n"
		  (current-error-port))
	 (exit 1))))


;; Copied from http.scm:
;; return error message string from gethost error ARGLIST
;; this will be locale encoded
(define gethost-error-message third)

;; Copied from http.scm:
;; return the error message from a 'system-error ARGLIST
;; this will be locale encoded
(define (system-error-message arglist)
  (apply simple-format #f (third arglist) (fourth arglist)))


;; key is a url which has been checked already
(define done-hash (make-hash-table 997))

(define (linkcheck where url)
  (if (not (hash-ref done-hash url))
      (begin
	(hash-set! done-hash url #t)

	(if verbose
	    (format #t "~a: ~a\n" where url))
	(c-lazy-catch #t
	  (lambda ()
	    (receive (headers body)
		(http-request url #:method "HEAD" #:follow #f)
	      (if (http-headers-ok? headers)
		  (set! good-count (1+ good-count))

		  (begin
		    (if (not verbose)
			(format #t "~a: ~a\n" where url))
		    (format #t "    ~a\n" (http-headers-error-str headers))
		    (case (http-headers-status-code headers)
		      ((301 302)
		       (format #t "    to ~a\n" (http-headers-get headers "Location"))))))))

	  (lambda args
	    (if (not verbose)
		(format #t "~a: ~a\n" where url))
	    ;; http protocol or connection problems produce messages,
	    ;; anything else is probably a code error so re-throw it
	    (cond ((eq? 'http (first args))
		   (format #t "    ~a\n" (second args)))
		  ((gethost-error-try-again? args)
		   (format #t "    ~a\n" (gethost-error-message args)))
		  ((system-error-econn? args)
		   (format #t "    ~a\n" (system-error-message args)))
		  (else
		   (apply throw args))))))))


(define (slurp filename)
  (call-with-input-file filename read-string-to-eof))

(format #t "checking ~a\n" filename)
(fold-matches "<a[ \t]+href=\"(http:[^#\"]+)(#[^\"]*)?\"[^>]*>([^<]+)</a>"
	      (slurp filename) #f
	      (lambda (m prev)
		(let ((link (match:substring m 1))
		      (text (match:substring m 3)))
		  (linkcheck (if (string=? link text)
				 filename
				 (format #f "~a ~s" filename text))
			     link))))

(format #t "~a good links\n" good-count)
(exit 0)


;; Local variables:
;; mode: scheme
;; compile-command: "chart -s online-manual.scm"
;; End:
