#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Check web links for selected symbols.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: chart -s online-weblinks.scm
;;
;; Check `weblink-handler!' URLs for various symbols.  The URLs are checked
;; by making a "HEAD" request, so there's no great amount downloaded, except
;; fro a couple of data sources which may do extra downloading to find the
;; link (eg. bendigo.scm).
;;
;; Because this program does online interaction it's not run by "make
;; check".
;;
;; Misfeatures:
;;
;; finance.yahoo.com gives a page with a message for an unknown symbol, so a
;; HEAD is not much good there.
;;


(use-modules (ice-9 and-let-star)
	     (ice-9 receive)
	     (srfi srfi-1)
	     (srfi srfi-13)
	     (chart http)
	     (chart misc))

(define verbose #t)

(define good-count 0)


;;-----------------------------------------------------------------------------

;; as per Guile 1.8, made available in 1.6
(if (not (defined? '@@))
    (define-macro (@@ mod-name var-name)
      (let ((var (module-variable (resolve-module mod-name) var-name)))
	(if (not var)
	    (error "no such variable" (list '@@ mod-name var-name)))
	(list variable-ref var))))

;; Copied from main.scm:
;; return a list of web links for SYMBOL, possibly empty
;; each list element is itself a list, of (MENUNAME TOOLTIP URL)
(define (symbol-weblinks symbol)
  (symbol-setups symbol)
  (filter-map (lambda (elem)
		(and (sympred-test (car elem) symbol)
		     ((cdr elem) symbol)))
	      weblink-handler-list))

(define yahoo-weblink-prefix  ;; per yahoo.scm
  (string-append "http://" (preference-get 'yahoo-quote-host
					   "finance.yahoo.com")
		 "/q?s="))
(define (yahoo-weblink? url)
  (string-prefix? yahoo-weblink-prefix url ))
	    

(define attempted-urls-hash (make-hash-table 211))

(define (attempt-one symbol menuname url)
  (or (hash-ref attempted-urls-hash url)

      ;; finance.yahoo.com gives a page with a message for an unknown
      ;; symbol, so no point in HEAD there
      (yahoo-weblink? url)

      (begin
	(hash-set! attempted-urls-hash url #t)

	(if verbose
	    (format #t "~a ~s\n    ~a\n" symbol menuname url))
	(receive (headers body)
	    (http-request url #:method "HEAD" #:follow #f)
	  (let ((status-code (http-headers-status-code headers)))
	    (cond ((http-status-code-ok? status-code)
		   (set! good-count (1+ good-count))
		   #t)
		  (else
		   (format #t "Web link URL error: ~a ~s\n" symbol menuname)
		   (format #t "  ~a\n" url)
		   (format #t "  ~a\n" (http-headers-error-str headers))
		   (and-let* ((location (http-headers-get headers "Location")))
		     (format #t "  to ~a\n" location))
		   #f)))))))

(define (attempt symbol)
  (let ((lst (symbol-weblinks symbol)))
    (if (null? lst)
	(format #t "No web link URLs for: ~a\n" symbol)

	(for-each (lambda (elem)
		    (receive-list (menuname tooltip url)
			elem
		      (if (procedure? url)
			  (set! url (url)))
		      (if url
			  (attempt-one symbol menuname url)
			  (format #t "No URL (#f) for: ~a\n" symbol))))
		  lst))))

(define (call-with-each-lang lang-list thunk)
  (for-each (lambda (lang)
	      (setenv "LANGUAGE" lang)
	      (thunk))
	    lang-list))

(set! weblink-message
      (lambda (fmt . args)
	(apply format #t fmt args)))


;;-----------------------------------------------------------------------------

(define (all)

  ;; london.scm
  (for-each attempt
	    '("TSCO.L"
	      "BLT.L"
	      "TPSD.IL"))

  ;; nybot-info.scm
  ;;   (let ((nybot-specs-list (@@ (chart nybot) nybot-specs-list)))
  ;;     (for-each (lambda (elem)
  ;; 		(attempt (string-append (first (car elem)) ".NYBOT")))
  ;; 	      nybot-specs-list))
  
  ;; nybot.scm
  (for-each attempt
	    '("CC.NYBOT"
	      ))

  ;; cbot-info.scm
  (let ((cbot-specs-list (@@ (chart cbot) cbot-specs-list)))
    (for-each (lambda (elem)
		(attempt (string-append (first (car elem)) ".CBOT")))
	      cbot-specs-list))

  ;; cbot.scm
  (for-each attempt
	    '("O.CBOT"
	      ))

  ;; casablanca.scm
  (call-with-each-lang '("fr" "en")
    (lambda ()
      (for-each attempt
		'("MNG.CAS"))))

  ;; cme.scm
  (for-each attempt
	    '("SP.CME"
	      ))

  ;; thailand.scm
  (for-each attempt
	    '("^THDOW"
	      "^DJTH"
	      ))

  ;; santiago.scm
  (call-with-each-lang '("en" "es")
    (lambda ()
      (for-each attempt
		'("^CLDOW"
		  ))))

  ;; mgex.scm
  (for-each attempt
	    '("MW.MGEX"
	      "IC.MGEX"
	      ))

  ;; amsterdam.scm
  (for-each attempt
	    '("^AEX"
	      ))

  ;; cairo.scm
  (call-with-each-lang '("ar" "en")
    (lambda ()
      (for-each attempt
		'("^CCSI"
		  ))))

  ;; brussels.scm
  (call-with-each-lang '("en" "fr")
    (lambda ()
      (for-each attempt
		'("^BFX"
		  ))))

  ;; vienna.scm
  (call-with-each-lang '("de" "en")
    (lambda ()
      (for-each attempt
		'("WST.VI"
		  "^ATX"
		  ))))

  ;; colombo.scm
  (for-each attempt
	    '("^CSE"))

  ;; virtx.scm
  (for-each attempt
	    '("BAER.VX"    ;; virt-x
	      ))

  ;; bilbao.scm
  (call-with-each-lang '("en" "eu" "es" "fr")
    (lambda ()
      (for-each attempt
		'("20.BI"
		  ))))

  ;; karachi.scm
  (for-each attempt
	    '("^KSE"))

  ;; phillipine.scm
  (for-each attempt
	    '("^PSI"))

  ;; prague.scm
  (call-with-each-lang '("cs" "en")
    (lambda ()
      (for-each attempt
		'("^PX50"))))

  ;; australia.scm
  (for-each attempt
	    '("NAB.AX"
	      "PCAPA.AX"
	      "NABHA.AX"
	      "AEQCA.AX"))

  ;; oslo.scm
  (call-with-each-lang '("no" "en")
    (lambda ()
      (for-each attempt
		'("PLUS.OL"
		  ))))

  ;; tokyo.scm
  (call-with-each-lang '("ja" "en")
    (lambda ()
      (for-each attempt
		'("^N225"  ;; Nikkei 225
		  ))))

  ;; korea.scm
  (call-with-each-lang '("en" "ko")
    (lambda ()
      (for-each attempt
		'("052300.KQ"  ;; Digilant FEF (DFEF)
		  "003660.KS"  ;; Korea Cement
		  ))))

  ;; india.scm
  (for-each attempt
	    '("SESAGOA.NS"  ;; Sesa Goa
	      ))

  ;; bombay.scm
  (for-each attempt
	    '("532401.BO"  ;; Vijaya Bank
	      "^BSESN"     ;; SENSEX index
	      ))

  ;; nymex.scm
  (for-each attempt
	    '("GC.COMEX"  ;; gold
	      "CL.NYMEX"  ;; crude
	      "CLZ07.NYMEX" ;; light sweet crude
	      "HU.NYMEX"    ;; gasoline
	      "HO.NYMEX"    ;; heating oil
	      "NG.NYMEX"    ;; henry hub natural gas
	      "PA.NYMEX"    ;; palladium
	      "PL.NYMEX"    ;; platinum
	      "PN.NYMEX"    ;; propane
	      "QG.NYMEX"    ;; e-miNY henry hub natural gas
	      "QL.NYMEX"    ;; CAPP central appalacian coal
	      "QM.NYMEX"    ;; e-miNY light sweet crude
	      "BB.NYMEX"    ;; brent bullet swap - london
	      ))

  ;; sandp.scm, all links
  (call-with-each-lang '("es" "en" "pt")
    (lambda ()
      (for-each (lambda (elem)
		  (attempt (car elem)))
		(@@ (chart sandp) standard-and-poors-weblink-alist))))

  ;; shanghai.scm
  (for-each attempt
	    '("000010.SS"  ;; SSE 180 index
	      ))

  ;; mx.scm
  (call-with-each-lang '("en" "fr")
    (lambda ()
      (for-each attempt
		'(;; indexes
		  "SXFZ06.MON"
		  "SXA.MON"
		  "SXB.MON"
		  "SXH.MON"
		  "SXY.MON"
		  ;; interest rates
		  "BAX.MON"
		  "OBX.MON"
		  "ONX.MON"
		  "CGZ.MON"
		  "CGB.MON"
		  "OGB.MON"))))

  ;; ccom.scm
  (call-with-each-lang '("ja" "en")
    (lambda ()
      (for-each attempt
		'("Gasoline.CCOM"
		  "Kerosene.CCOM"
		  "Gas Oil.CCOM"
		  "Eggs DEC 06.CCOM"
		  "Ferrous Scrap.CCOM"
		  "RSS3.CCOM"
		  "TSR20.CCOM"
		  "Nickel.CCOM"
		  "Aluminium.CCOM"
		  "Rubber Index.CCOM"
		  ))))

  ;; lme.scm
  (for-each attempt
	    '("COPPER 3.LME"
	      "ALUMINIUM.LME"
	      "ALUMINIUM ALLOY.LME"
	      "LEAD.LME"
	      "LMEX.LME"
	      "NASAAC.LME"
	      "NICKEL.LME"
	      "TIN.LME"
	      "ZINC.LME"
	      "PP.LME"
	      "LL APR 07.LME"))

  ;; kex.scm
  (call-with-each-lang '("en" "ja")
    (lambda ()
      (for-each attempt
		'("CF.KEX"
		  "EBI.KEX"
		  "KI JAN 06.KEX"
		  "N.KEX"
		  "RB.KEX"
		  "RS.KEX"
		  "SG.KEX"
		  "BR.KEX"
		  "CO.KEX"
		  ;; "SM.KEX" ;; delisted Jul07
		  ))))

  ;; athens.scm
  (for-each attempt
	    '("HTO.ATH"))

  ;; bendigo.scm
  (for-each attempt
	    '("BTT.BEN" "CAP.BEN"))

  ;; barchart.scm comex
  (for-each attempt
	    '("HG.COMEX"     ;; high grade copper
	      "GCZ07.COMEX"  ;; gold
	      "SI.COMEX"     ;; silver
	      "AL.COMEX"))   ;; aluminium

  ;; fatdog.scm
  (for-each attempt
	    '("Lindsay Davenport.FAT"
	      "Mark Webber.FAT"))

  ;; kcbt.scm
  (for-each attempt
	    '("KW.KCBT"
	      "KE.KCBT"
	      "MV.KCBT"
	      "MVM06.KCBT"))

  ;; ljubljana.scm
  (call-with-each-lang '("en" "sl")
    (lambda ()
      (for-each attempt
		;; kcbt
		'("DRKR.LJ"))))

  ;; newzealand.scm
  (for-each attempt
	    '("TEL.NZ"))


  ;; sfe.scm
  (for-each attempt
	    '("BB.SFE"
	      "YT MAR 07.SFE"))

  ;; sicom.scm
  (for-each attempt
	    '("CF.SICOM"
	      "RS MAY 07.SICOM"
	      "RT.SICOM"
	      "RI.SICOM"
	      "TF AMJ 07.SICOM"))

  ;; tge.scm
  (call-with-each-lang '("en" "ja")
    (lambda ()
      (for-each attempt
		'("CO.TGE"
		  "SM.TGE"
		  "SB.TGE"
		  "NG.TGE"
		  "RB OCT 06.TGE"
		  "AC.TGE"
		  "RC.TGE"
		  "SG.TGE"
		  "SL.TGE"
		  "VG.TGE"))))

  ;; tocom.scm
  (call-with-each-lang '("en" "ja")
    (lambda ()
      (for-each attempt
		;; tocom
		'("Gold.TOCOM"
		  "Silver.TOCOM"
		  "Platinum.TOCOM"
		  "Palladium.TOCOM"
		  "Aluminum.TOCOM"
		  "Gasoline.TOCOM"
		  "Kerosene.TOCOM"
		  "Crude Oil.TOCOM"
		  "Gas Oil.TOCOM"
		  "Rubber Aug 2006.TOCOM"))))

  ;; wce.scm
  (for-each attempt
	    '("RS.WCE"
	      "WWK06.WCE"
	      "AB.WCE"
	      ;; delisted
	      "RM.WCE"
	      "WF.WCE"
	      "WO.WCE"
	      "WP.WCE"
	      "WQ.WCE"))

  ;; wtb.scm
  (call-with-each-lang '("en" "de")
    (lambda ()
      (for-each attempt
		'("H.WTB"
		  "F.WTB"
		  "P.WTB"
		  "XP.WTB"
		  "LPJ06.WTB"
		  "W.WTB"
		  "BX06.WTB"
		  "POTX.WTB"
		  "TAPX.WTB"
		  "WTBBPC.WTB"))))

  ;; hongkong.scm
  (for-each attempt
	    '("0013.HK"))

  ;; otcbb.scm
  (for-each attempt
	    '("CNES.OB"
	      "PCLO.OB"
	      "USXP.OB"))

  ;; pinksheets.scm
  (for-each attempt
	    '("BOREF.PK"
	      "PCMC.PK"))

  ;; usa.scm
  (for-each attempt
	    '("GM"     ;; NYSE
	      "AUY"    ;; AMEX
	      "AAPL")) ;; Nasdaq

  ;; zagreb.scm
  (call-with-each-lang '("en" "hr")
    (lambda ()
      (for-each attempt
		'("PLVA-R-A.ZAG"))))

  (format #t "~a checks ok\n" good-count))



(c-main-enq! all)
(c-main)


;; Local variables:
;; mode: scheme
;; compile-command: "chart -s online-weblinks.scm"
;; End:
