;; Diagnostics dialog.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart diagnostics)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart database)
  #:use-module (chart extra-gtk)
  #:use-module (chart latest)
  #:use-module (chart misc))


;; This is a dialog box showing some diagnostic information about memory and
;; sockets use.  It's bound to a secret key "<Alt><Shift>D" (see main.scm)
;; and is mainly for development use.


;;-----------------------------------------------------------------------------
;; misc

;; like Guile 1.8, made available in 1.6, but only for fetching
(if (not (defined? '@@))
    (define-macro (@@ mod-name var-name)
      (let ((var (module-variable (resolve-module mod-name) var-name)))
	(if (not var)
	    (error "no such variable" (list '@@ mod-name var-name)))
	(list variable-ref var))))


;; return the number of entries in the given hash TABLE
(define (hash-count table)
  (hash-fold (lambda (key value prev)
	       (1+ prev))
	     0 table))

;; return the number of latest quotes cached on disk
(define (latest-cache-count)
  (length (directory-files
	   latest-directory #f (lambda (filename)
				 (not (string-prefix? "." filename))))))
  
;; return true if module M is only a directory entry, like "(ice-9)", as
;; opposed to an actual module with bindings
(define (module-directory-only? m)
  (let ((ret #t))
    (module-for-each (lambda (sym var)
		       (if (and var
				(variable-bound? var)
				(not (module? (variable-ref var))))
			   (set! ret #f)))
		     m)
    ret))

;; call (PROC m) for all currently loaded modules M
(define (traverse-modules proc)
  (define seen '(duplicate-handlers))
  (define (recurse m)
    (or (module-directory-only? m)
	(proc m))
    (set! seen (cons (module-name m) seen))
    (module-for-each (lambda (sym var)
		       (and-let* (var
				  (  (variable-bound? var))
				  (m (variable-ref var))
				  (  (module? m)))
			 (or (member (module-name m) seen)
			     (recurse m))))
		     m))
  (recurse (or (nested-ref the-root-module '(%app modules))    ;; guile 1.8
	       (nested-ref the-root-module '(app modules)))))  ;; guile 1.6

;; compare symbols in the car of X and Y, by name (per database.scm)
(define (car-symbol-string<? x y)
  (string<? (symbol->string (car x)) (symbol->string (car y))))

(define (car-string<? x y)
  (string<? (car x) (car y)))

;; return true if OBJ is an output string port
(define (output-string-port? obj)
  (and (port? obj)
       (false-if-exception (get-output-string obj))))

;; return true if OBJ is a file port open on a network socket
(define (socket-port? obj)
  (and (file-port? obj)
       (eq? 'socket (stat:type (stat obj)))))

;; return true if OBJ is a file port open on a regular file, ie. a disk file
(define (regular-file-port? obj)
  (and (file-port? obj)
       (eq? 'regular (stat:type (stat obj)))))

;; return true if PORT is one of the current input/output/error ports
(define (port-current-stdio? port)
  (or (eq? port (current-input-port))
      (eq? port (current-output-port))
      (eq? port (current-error-port))))


;;-----------------------------------------------------------------------------
;; diagnostics

;; return a string of diagnostic information
(define (diagnostics-str)
  ;; mallinfo retrieved first since rest may use up some memory, in
  ;; particular gc-live-object-stats does
  (define mall (mallinfo))

  (let ((gcs  (gc-stats))
	(live (and (defined? 'gc-live-object-stats) ;; new in guile 1.8
		   (gc-live-object-stats))))

    (call-with-output-string
     (lambda (port)

       (display "database:\n" port)
       (let ((database-series-hash (@@ (chart database) database-series-hash))
	     (latest-cache         (@@ (chart latest) cache)))
	 (format port "disk database symbols: ~a\n"
		 (length (database-symbols)))
	 (format port "disk latest quotes: ~a\n"
		 (latest-cache-count))
	 (format port "memory series cache: ~a\n"
		 (hash-count database-series-hash))
	 (format port "memory latest quotes: ~a\n"
		 (hash-count latest-cache))
	 (format port "latest download in progress: ~a\n"
		 (length (@@ (chart latest) latest-in-progress-list))))
       (let ((regular 0)
	     (sock    0)
	     (outstr  0)
	     (other   0))
	 (port-for-each (lambda (port)
			  (cond ((port-current-stdio? port)) ;; ignore
				((output-string-port? port)
				 (set! outstr (1+ outstr)))
				((socket-port? port)
				 (set! sock (1+ sock)))
				((regular-file-port? port)
				 (set! regular (1+ regular)))
				(else
				 (set! other (1+ other))))))
	 (format port "ports: file ~a, socket ~a, outstr ~a, other ~a (plus stdio)\n"
		 regular sock outstr other))

       (display "\nsockpool-list:\n" port)
       (let ((sockpool-list  	 (@@ (chart http) sockpool-list))
	     (sockpool-inuse 	 (@@ (chart http) sockpool-inuse))
	     (sockpool-not-inuse (@@ (chart http) sockpool-not-inuse))
	     (sockpool-host  	 (@@ (chart http) sockpool-host))
	     (sockpool-port  	 (@@ (chart http) sockpool-port)))
	 (if (null? sockpool-list)
	     (display "empty\n" port)
	     (begin
	       (for-each (lambda (rec)
			   (format port "~a:~a inuse\n"
				   (sockpool-host rec)
				   (sockpool-port rec)))
			 (filter sockpool-inuse sockpool-list))
	       (for-each (lambda (rec)
			   (format port "~a:~a free\n"
				   (sockpool-host rec)
				   (sockpool-port rec)))
			 (filter sockpool-not-inuse sockpool-list)))))
       ;; gethost
       (let ((cache               (@@ (chart gethost) cache))
	     (child-id            (@@ (chart gethost) child-id))
	     (subproc-use-threads (@@ (chart gethost) subproc-use-threads)))
	 (format port "gethost cache: ~a\n" (length cache))
	 (format port "lookups by ~a, current: ~a\n"
		 (if subproc-use-threads "thread" "child process")
		 child-id))

       (display "\nc-waitfor-queue:\n" port)
       (let* ((q   (@@ (chart http) c-waitfor-queue))
	      (lst (q-list q)))
	 (if (null? lst)
	     (display "empty\n" port)
	     (for-each (lambda (pair)
			 (format port "~a\n" (car pair)))
		       lst)))

       (display "\nmallinfo:\n" port)
       (for-each (lambda (elem)
		   (format port "~a ~a\n"
			   (string-pad-right
			    (symbol->string (car elem)) 20)
			   (locale-number+decimals (cdr elem) 0)))
		 mall)

       (let ((segs (assoc-ref gcs 'cell-heap-segments))
	     (total 0))
	 (display "\ngc-stats:\n" port)
	 (for-each (lambda (elem)
		     (format port "~a ~a\n"
			     (string-pad-right
			      (symbol->string (car elem)) 20)
			     (if (number? (cdr elem))
				 (locale-number+decimals (cdr elem) 0)
				 (cdr elem))))
		   (assoc-remove! gcs 'cell-heap-segments))
	 (format port "cell-heap-segments (kbytes): ~a\n"
		 (map (lambda (seg)
			;; "abs" because it's (START . END) in guile 1.8,
			;; but (END . START) in 1.6
			(let ((this (abs (- (cdr seg) (car seg)))))
			  (set! total (+ total this))
			  (quotient-ceiling this 1024)))
		      segs))
	 (format port "cell-heap-segments total: ~a kbytes\n"
		 (quotient-ceiling total 1024)))

       (if live
	   (let ((total 0))
	     (display "\ngc-live-object-stats:\n" port)
	     (for-each (lambda (elem)
			 (format port "~a ~a\n"
				 (string-pad-right (car elem) 20)
				 (if (number? (cdr elem))
				     (locale-number+decimals (cdr elem) 0)
				     (cdr elem)))
			 (if (number? (cdr elem))
			     (set! total (+ total (cdr elem)))))
		       (sort live car-string<?))
	     (format port "total: ~a\n" (locale-number+decimals total 0))))

       (and-let* (c-main-goto-top  ;; when c-main running
		  (stack (make-stack c-main-goto-top))) ;; when debug
	 (display "\nmain loop\n" port)
	 (format port "c-main-goto-top depth ~a\n" (stack-length stack)))

       (let ((alist '()))
	 (traverse-modules (lambda (m)
			     (let* ((name (module-name m))
				    (top  (first name)))
			       (set! alist
				     (assq-set! alist top
						(1+ (or (assq-ref alist top)
							0)))))))
	 (display "\nloaded modules:\n" port)
	 (for-each (lambda (elem)
		     (format port "~a: ~a\n" (car elem) (cdr elem)))
		   (sort alist car-symbol-string<?)))))))


(define-public (diagnostics-dialog-popup)
  (let* ((str      (diagnostics-str))
	 (dialog   (gtk-dialog-new-with-buttons
		    (c-make-window-title (_ "Diagnostics")) #f '()
		    "gtk-ok" gtk-response-ok
		    "gtk-refresh" 0
		    "_Force GC" 1))
	 (label    (gtk-label-new str))
	 (scrolled (gtk-scrolled-window-new)))

    (gtk-widget-set label #:wrap #t)
    (c-gtk-widget-set-usize-units scrolled '(em . 40)
				  (round-exact (* 0.8 (gdk-screen-height))))
    (gtk-scrolled-window-set-policy scrolled 'never 'automatic)
    (gtk-scrolled-window-add-with-viewport scrolled label)
    (gtk-box-pack-start (gtk-dialog-vbox dialog) scrolled #t #t 0)

    (gtk-signal-connect dialog "response"
      (lambda (id)
	(cond ((memv id '(0 1)) ;; refresh or force-gc
	       (if (= 1 id) ;; force-gc
		   (gc))
	       (gtk-label-set-text label (diagnostics-str)))
	      (else ;; gtk-response-ok, or gtk-response-delete-event
	       (gtk-widget-destroy dialog)))))

    (gtk-widget-show-all dialog)))
