;; Dividends edit dialog.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart div-edit)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 regex)
  #:use-module (gtk gtk)
  #:use-module (chart database)
  #:use-module (chart date)
  #:use-module (chart extra-gtk)
  #:use-module (chart indicator)
  #:use-module (chart gui-common)
  #:use-module (chart main)
  #:use-module (chart misc)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; misc

;; use `delete' here in case SERIES has been re-read from the database, so
;; old elements are no longer `eq?'
(define-public (series-dividend-del! series div)
  (series-dividends! series (delete div (series-dividends series))))

(define (tm->mdate tm)
  (apply ymd->mdate (tm->ymd tm)))

;; `strptime' but with FMT and input as utf8 strings
(define-public (strptime-utf8 fmt str)
  (strptime (utf8->locale fmt) (utf8->locale str)))


;;-----------------------------------------------------------------------------

;; created on demand, then re-used
(define dialog         #f)

(define heading-label  #f)
(define list-widget    #f)
(define item->div-prop (primitive-make-property #f))
(define current-item   #f)
(define current-symbol #f)

;; return a string for series-dividend value VAL
;; VAL is utf-8 encoded if it's a string, the result is utf8 too
(define (value->text val)
  (or (cond
       ((string? val)
	val)
       ((and (pair? val)
	     (eq? 'month (first val)))
	(strftime-utf8 (_ "month: %b %Y") (mdate->tm (second val))))
       ((and (pair? val)
	     (eq? 'split (first val)))
	(format #f (_ "split ~a for ~a") (second val) (third val)))
       ((and (pair? val)
	     (eq? 'amount (first val)))
	(let ((amount (second val))
	      (alist  (cddr val)))
	  (cond ((eq? amount 'unknown)
		 (_ "unknown"))
		((eq? amount 'to-be-advised)
		 (_ "to-be-advised"))
		((number? amount)
		 (let* ((decimals (or (assq-ref alist 'decimals) 0))
			(str      (locale-number+decimals amount decimals)))
		   (if (assq-ref alist 'estimated)
		       (set! str (string-append (_ "estimated ") str)))
		   (and-let* ((imputation (assq-ref alist 'imputation)))
		     (if (number? imputation)
			 (set! imputation
			       (locale-number+decimals imputation decimals)))
		     (set! str (format #f "~a + ~s" str imputation)))
		   str))
		(else
		 #f))))
       (else
	#f))

      ;; some unrecognised object
      ;; showing it as a `write' like this won't come back as the
      ;; original object after editing, but least it'll show up
      (format #f "~s" val)))

;; return a series-dividend value from STR
;; STR is locale encoded, the result (if it's a string) will be utf-8 
(define (text->value str)
  (set! str (string-trim-both str))
  (cond ((or (false-if-exception (strptime-utf8 (_ "month: %b %Y") str))
	     (false-if-exception (strptime-utf8    "month: %b %Y"  str)))
	 => (lambda (pair)
	      (let ((tm (car pair)))
		(set-tm:mday tm 1) ;; satisfy range check in ymd->mdate
		(list 'month (tm->mdate tm)))))
	((or (string-match (_ "^split +([0-9.]+) +for +([0-9.]+)$") str)
	     (string-match    "^split +([0-9.]+) +for +([0-9.]+)$" str))
	 => (lambda (m)
	      (list 'split
		    (string->number (match:substring m 1))
		    (string->number (match:substring m 2)))))
	((or (string-ci=? str (_ "unknown"))
	     (string-ci=? str    "unknown"))
	 '(amount unknown))
	((or (string-ci=? str (_ "to-be-advised"))
	     (string-ci=? str    "to-be-advised"))
	 '(amount to-be-advised))
	((or (string-match (string-append (_ "^(estimated )?(([0-9]*)(")
					  (locale-decimal-point)
					  "([0-9]+))?|([0-9]+)("
					  (locale-decimal-point)
					  "([0-9]*)))$")
			   str)
	     (string-match (string-append "^(estimated )?(([0-9]*)("
					  (locale-decimal-point)
					  "([0-9]+))?|([0-9]+)("
					  (locale-decimal-point)
					  "([0-9]*)))$")
			   str))
	 => (lambda (m)
	      (let* ((amount-str (match:substring m 2))
		     (decimals   (string-decimals amount-str))
		     (amount     (string->number+decimals amount-str decimals))
		     (alist      '()))
		(if (match:start m 1)
		    (set! alist (acons 'estimated #t alist)))
		`(amount ,amount (decimals . ,decimals) ,@alist))))
	(else
	 str)))

(define (init)
  (set! dialog (gtk-dialog-new-with-buttons "" #f '()))
  (gtk-dialog-add-buttons dialog
			  "gtk-delete" 0
			  "gtk-close"  gtk-response-close
			  "gtk-help"   gtk-response-help)
  (gtk-dialog-set-response-sensitive dialog 0 #f)

  (let ((date-widget (date-widget-new))
	(entry       (gtk-entry-new)))

    (define (entry-activate)
      (let* ((adate (date-widget-adate date-widget)))
	(if (adate-weekend? adate)
	    (error-dialog-popup dialog
				(format #f (_ "~a is not a weekday")
					(strftime-utf8 (locale-d-fmt)
						       (adate->tm adate))))
	    (begin
	      (database-call-modify-data current-symbol
		(lambda (series)
		  (series-dividend-add! series (adate->tdate adate)
					(text->value (gtk-entry-get-text
						      entry)))))
	      (fill)))))

    (c-gtk-dialog-hide-on-delete dialog)
    (gtk-signal-connect dialog "response"
      (lambda (id)
	(cond
	 ((= id 0) ;; delete
	  (database-call-modify-data current-symbol
	    (lambda (series)
	      (series-dividend-del! series (primitive-property-ref
					    item->div-prop current-item))))
	  (fill))

	 ((= id gtk-response-close)
	  (gtk-widget-hide dialog))

	 ((= id gtk-response-help)
	  (manual-popup (_ "CONTEXT(manual-node): Dividends") dialog)))))

    (set! heading-label (gtk-label-new ""))
    (gtk-box-pack-start (gtk-dialog-vbox dialog) heading-label #f #f 0)

    ;; scrolled
    (let ((scrolled (gtk-scrolled-window-new)))
      (gtk-box-pack-start (gtk-dialog-vbox dialog) scrolled #t #t 0)
      (gtk-scrolled-window-set-policy scrolled 'automatic 'always)
      (c-gtk-widget-set-usize-units scrolled '(em . 30) '(height . 15))

      (set! list-widget (gtk-list-new))
      (gtk-scrolled-window-add-with-viewport scrolled list-widget)
      (gtk-signal-connect list-widget "select_child"
	(lambda (item)
	  (set! current-item item)
	  (gtk-dialog-set-response-sensitive dialog 0 #t) ;; delete
	  (let* ((div (primitive-property-ref item->div-prop item)))
	    (date-widget-tdate! date-widget (car div))
	    (gtk-entry-set-text entry (value->text (cdr div))))))

      (gtk-signal-connect list-widget "unselect_child"
	(lambda (item)
	  (set! current-item #f)
	  (gtk-dialog-set-response-sensitive dialog 0 #f)))) ;; delete

    (gtk-box-pack-start (gtk-dialog-vbox dialog) date-widget #f #f 0)

    (let ((hbox (gtk-hbox-new #f 4)))
      (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0)
      (gtk-box-pack-start hbox (gtk-label-new (_ "Dividend")) #f #f 0)
      (gtk-signal-connect entry "activate" entry-activate)
      (c-gtk-widget-set-usize-units entry '(digit . 10) '(height . 1.5))
      (gtk-box-pack-start hbox entry  #t #t 0)
      (gtk-widget-grab-focus entry)

      (gtk-signal-connect dialog "map_event"
	(lambda (event)
	  (c-gtk-editable-select-region-noclip entry 0 -1)))

      (let ((button (gtk-button-new-with-label (_ "Enter"))))
	(gtk-box-pack-start hbox button #f #f 0)
	(gtk-signal-connect button "clicked" entry-activate))))

  (gtk-widget-show-all dialog)

  ;; once only, and GC this code
  (set! init noop))

(define (fill)
  (gtk-dialog-set-response-sensitive dialog 0 #f) ;; delete
  (let ((divlst (sort (series-dividends (database-read-series current-symbol))
		      car>)))
    (c-gtk-list-require-n-labels list-widget (length divlst))
    (for-each (lambda (div item)
		(primitive-property-set! item->div-prop item div)
		(gtk-label-set-text
		 (gtk-bin-get-child item)
		 (string-append (strftime-utf8 (locale-d-fmt)
					       (tdate->tm (car div)))
				"  "
				(value->text (cdr div)))))
	      divlst (gtk-container-children list-widget))))

(define-public (dividends-edit-popup)
  (let ((symbol (view-symbol outer)))
    (if (not symbol)
	(error-dialog-popup app-widget (_ "No current chart"))

	(begin
	  (init)
	  (set! current-symbol symbol)
	  (let* ((str (format #f (_ "~a Dividends") current-symbol)))
	    (gtk-label-set-text heading-label str)
	    (gtk-window-set-title dialog (c-make-window-title str)))
	  (fill)
	  (c-gtk-widget-show-or-raise dialog)))))


