;; Directional Movement Index (DMI).

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart dmi)
  #:use-module (ice-9 receive)
  #:use-module (chart atr)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.incrediblecharts.com/technical/directional_movement.htm
;; http://www.incrediblecharts.com/technical/dms_construction.htm


;; Return a procedure which calculates directional movement over successive
;; days.
;;
;; Each call (PROC high low) enters a new day into the calculation, and the
;; return is (values DM+ DM-), the directional movement for the day.
;;
;; To prime initially, call PROC with the high/low of the day before, to
;; give it those values ready for the next call.
;;
(define-public (dm-calc-proc)

  (let ((prev-high #f)
	(prev-low  #f))

    (lambda (high low)
      (define dm+ #f)
      (define dm- #f)

      (if prev-high
	  (begin
	    (set! dm+ (max 0 (- high prev-high)))
	    (set! dm- (max 0 (- prev-low low)))

	    ;; zap the smaller of the two, or if equal zap both
	    (cond ((> dm+ dm-)
		   (set! dm- 0))
		  ((< dm+ dm-)
		   (set! dm+ 0))
		  (else
		   (set! dm+ 0)
		   (set! dm- 0)))))

      (set! prev-high high)
      (set! prev-low  low)
      (values dm+ dm-))))


;; Return a procedure which calculates directional movement indexes over an
;; accumulated window of days.
;;
;; Each call (PROC high low close) enters a new day into the calculation,
;; and the return is (values DMI+ DMI-) the directional movement indexes.
;;
;; A DMI is in theory influenced by all preceding data, but
;; `dmi-warmup-count' below is designed to determine a warmup count.
;;
(define-public (dmi-calc-proc count)

  (let ((dm-proc  (dm-calc-proc))
	(ema+proc (ema-calc-proc (ema-count-wilder count)))
	(ema-proc (ema-calc-proc (ema-count-wilder count)))
	(atr-proc (atr-calc-proc count)))

    (lambda (high low close)
      (define atr (atr-proc high low close))

      (set! high (or high close))
      (set! low  (or low  close))
      (receive (dm+ dm-)
	  (dm-proc high low)
	(if (not dm+)
	    (values #f #f)

	    (let ((di+ (ema+proc dm+))
		  (di- (ema-proc dm-)))
	      (if (and di+ atr (not (zero? atr)))
		  (values (* 100.0 (/ di+ atr))
			  (* 100.0 (/ di- atr)))
		  (values #f #f))))))))

;; same EMA+1
(define-public dmi-warmup-count atr-warmup-count)
      

(define (dmi-fill series new-array lo hi)
  (let* ((count    (series-data series))
	 (parent   (series-parent series))
	 (start    (series-prev-n-closes-pos
		    parent #:before lo #:count (dmi-warmup-count count)))
	 (array    (series-array parent start hi))
	 (dmi      (dmi-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (dmi (array-ref array i 1)    ;; high
	       (array-ref array i 2)    ;; low
	       (array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (receive (di+ di-)
	      (dmi (array-ref array i 1)  ;; high
		   (array-ref array i 2)  ;; low
		   (array-ref array i 3)) ;; close
	    (array-set! new-array di+ i 0)
	    (array-set! new-array di- i 1))))))

(define-public (dmi-series series count)
  (ohlcv-required "DMI" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "DMI") count))
    (series-datatype!  new-series dmi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   dmi-fill
				   (dmi-warmup-count count)
				   '(0 1)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public dmi-datatype
  (datatype-new!
   #:key       'dmi
   #:name      (_ "DMI - Directional Movement Index")
   #:input     datatype-ohlcv
   #:units     'dmi
   #:linestyle 'dmi
   #:minimum   0
   #:manual    (_ "CONTEXT(manual-node): Directional Movement Index")
   #:proc      dmi-series
   #:closes-apply-arg '(0)
   #:parameter-list
   (list (datatype-param-new
	  #:key     'dmi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))



