#!/bin/sh
exec guile -s "$0" "$@"
!#

;; HTML manual href link munging.

;; Copyright 2005, 2006, 2007 Kevin Ryde
;;
;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: guile -s mung-html.scm filename.html ...
;;
;; This program modifies the given "filename.html" etc input files, changing
;; links like href="../foo.html#Node", as generated by makeinfo, into links
;; to the GNU web site, or wherever.
;;
;; This is used to create the distributed chart.html file.
;;

(use-modules (ice-9 optargs)
	     (ice-9 regex)
	     (ice-9 rw)
	     (srfi srfi-1)
	     (srfi srfi-13))

(define everything-good #t)


(define (slurp filename)
  (call-with-input-file filename read-string-to-eof))

;; return string read from PORT up to eof
;; (copied from misc.scm)
(define (read-string-to-eof port)
  (let ((buf (make-string 4096)))
    (do ((lst   '()
		(cons (substring buf 0 nread) lst))
	 (nread (read-string!/partial buf port)
		(read-string!/partial buf port)))
	((not nread)                       ;; stop at #f
	 (string-concatenate-reverse lst)) ;; join parts
      ;; allow for 0 meaning EWOULDBLOCK
      ;; if 0 is in fact from a signal then this select does no harm
      (if (= 0 nread)
	  (select (list port) '() '())))))


;;-----------------------------------------------------------------------------
;; href external links

(define external-count 0)

(define (mung-external-links body)

  ;; no-split style:
  ;;     foo.html#Top
  ;;     foo.html#Node-Name
  ;;
  ;; split style:
  ;;    ../foo/index.html#Top
  ;;    ../foo/Node-Name.html#Node-Name
  ;;
  (regexp-substitute/global
   #f "\
href=\"([a-z-]+)\\.html#([^\"]+)|\
href=\"\\.\\./([a-z-]+)/index\\.html#(Top)|\
href=\"\\.\\./([a-z-]+)/([^\".]+)\\.html#\\6"
   body
   'pre
   (lambda (m)
     (let ((manual (or (match:substring m 1)
		       (match:substring m 3)
		       (match:substring m 5)))
	   (node   (or (match:substring m 2)
		       (match:substring m 4)
		       (match:substring m 6))))
       (cond ((and (string=? manual "index")
		   (string=? node "Top"))
	      ;; top of this manual, no change
	      (match:substring m))

	     ((mung-find-replacement manual node)
	      => (lambda (repl)
		   (or (string? repl)
		       (error "replacement not a string"
			      mung-filename manual node repl))
		   (set! external-count (1+ external-count))
		   ;; (format #t "~a\n" repl)
		   (string-append "href=\"" repl)))
	     (else
	      (format #t "~a: unmunged ~a\n"
		      mung-filename (match:substring m))))))
   'post))

(define (mung-find-replacement manual node)
  (any (lambda (handler)
	 (handler manual node))
       external-handler-list))

(define external-handler-list '())

(define (add-external-handler proc)
  (set! external-handler-list (cons proc external-handler-list)))

;; MANUAL is at www.gnu.org.
;; SUBDIRNAME is the subdirectory of http://www.gnu.org/software/ to find
;; it, defaulting to
;;     http://www.gnu.org/software/PACKAGE/manual/index.html
;;     http://www.gnu.org/software/PACKAGE/manual/html_node/Foo.html
;;
(define* (external-gnu manual
		       #:optional
		       (subdirname (string-append manual "/manual"))
		       #:key
		       (html_node "/html_node"))
  (add-external-handler
   (lambda (found-manual found-node)
     (and (string=? manual found-manual)
	  (if (string=? "Top" found-node)
	      (string-append "http://www.gnu.org/software/" subdirname
			     "/index.html")
	      (string-append "http://www.gnu.org/software/" subdirname
			     html_node "/"
			     found-node ".html"))))))

;; TARGET is a url used for MANUAL+NODE
(define (external-one manual node target)
  (add-external-handler
   (lambda (found-manual found-node)
     (and (string=? manual found-manual)
	  (string=? node found-node)
	  target))))

;; TARGET is a url used for any link out of MANUAL, irrespective of the
;; specified node
(define (external-any manual target)
  (add-external-handler
   (lambda (found-manual found-node)
     (and (string=? manual found-manual)
	  target))))

;; MANUAL is at www.gnu.org, but only in single-page html form
;;     http://www.gnu.org/software/PACKAGE/MANUAL.html
;;
(define* (external-gnu-onepage manual
			       #:optional
			       (subdirname manual))
  (add-external-handler
   (lambda (found-manual found-node)
     (and (string=? manual found-manual)
	  (if (string=? "Top" found-node)
	      (string-append "http://www.gnu.org/software/" subdirname
			     "/" manual ".html")
	      (string-append "http://www.gnu.org/software/" subdirname
			     "/" manual ".html#" found-node))))))


(external-gnu "autoconf")
(external-one "automake" "Top" "http://sourceware.org/automake/automake.html")
(external-gnu-onepage "bashref" "bash")
(external-gnu "elisp"    "emacs/elisp-manual")
(external-gnu "emacs")
(external-gnu "gettext")

(external-gnu "guile"     "guile/docs/docs-1.8/guile-ref" #:html_node "")
(external-gnu "guile-tut" "guile/docs/guile-tut"          #:html_node "")
(external-gnu "guile-gtk" "guile-gtk/docs/guile-gtk"      #:html_node "")
(external-gnu "gzip")
(external-gnu "info-stnd" "texinfo/manual/info-stnd")
(external-gnu "libc")

;; no online ses manual yet ...
(external-any "ses" "http://home.comcast.net/~jyavner/ses/")

(external-gnu "texinfo" "texinfo/manual/texinfo")


;;-----------------------------------------------------------------------------
;; png sizes

(define image-count 0)

;; return a long (4-byte) signed integer from the bytes in STR at OFFSET
(define (bigendian-bytes->ulong str offset)
  (+ (* #x1000000 (char->integer (string-ref str offset)))
     (* #x10000   (char->integer (string-ref str (1+ offset))))
     (* #x100     (char->integer (string-ref str (+ 2 offset))))
     (* #x1       (char->integer (string-ref str (+ 3 offset))))))

;; like `string' but taking integer char values
(define-public (string-ints . int-list)
  (list->string (map! integer->char int-list)))
 
(define png-image-header-string (string-ints 137 80 78 71 13 10 26 10))

;; return list (WIDTH HEIGHT) which is the size of the png image in FILENAME
(define (png-image-size filename)
  (let ((data (slurp filename)))
    (or (string-prefix? png-image-header-string data)
	(error "missing PNG signature:" filename))
    (or (string=? (substring data 12 16) "IHDR")
	(error "PNG doesn't begin with IHDR:" filename))
    (list (bigendian-bytes->ulong data 16)
	  (bigendian-bytes->ulong data 20))))

(define (mung-img-sizes body)

  (set! body (regexp-substitute/global
	      #f "<img src=\"([^\"]+)\"" body
	      'pre
	      (lambda (m)
		(set! image-count (1+ image-count))
		(let ((filename (match:substring m 1)))
		  (apply format #f "~a width=~a height=~a"
			 (match:substring m) (png-image-size filename))))
	      'post))
  body)


;;-----------------------------------------------------------------------------

(define mung-filename #f)

(define (mung-file filename)
  (set! mung-filename filename)

  (let ((body (slurp filename)))

    (set! body (mung-external-links body))
    (set! body (mung-img-sizes body))

    (call-with-output-file filename
      (lambda (port)
	(display body port)))))

(for-each mung-file (cdr (program-arguments)))

(if (not everything-good)
    (exit 1))

(format #t "munged ~a external links, ~a image sizes\n"
	external-count image-count)

(exit 0)


;; Local variables:
;; mode: scheme
;; End:
