/* PNG file manipulations for distribution.

   Copyright 2006, 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>

#include <png.h>


/* Usage: mung-png filename.png title-text

   Insert some comment strings into the given png file, including the given
   title-text as the "Title" element.

   The copyright string inserted is deliberately not compressed, so it's
   visible in a hex dump or similar.

   Crib notes: The png spec at http://www.w3.org/TR/PNG section 11.3.4.2 has
   the defined keywords.  The "Homepage" added is not among them.  */


void
error (png_structp png_ptr, png_const_charp str)
{
  printf ("libpng error: %s\n", str);
  exit (1);
}

void
warning (png_structp png_ptr, png_const_charp str)
{
  printf ("libpng warning: %s\n", str);
  exit (1);
}

void
pngtextadd (const char *filename, const char *keyword, const char *text)
{
  png_infop info_ptr;

  {
    FILE *input_fp;
    png_structp png_ptr;
    png_text *text_ptr, textnew;
    int i, num_text;

    input_fp = fopen(filename, "rb");
    if (input_fp == NULL)
      {
        fprintf (stderr, "Cannot open %s\n", filename);
        exit (1);
      }

    png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING,
                                      NULL, error, warning);
    if (png_ptr == NULL)
      {
        fprintf (stderr, "Cannot create png structure\n");
        exit (1);
      }

    info_ptr = png_create_info_struct (png_ptr);
    if (info_ptr == NULL)
      {
        fprintf (stderr, "Cannot create png info structure\n");
        exit (1);
      }

    if (setjmp (png_jmpbuf (png_ptr)))
      {
      aborting:
        fprintf (stderr, "aborting\n");
        exit (1);
      }

    png_init_io (png_ptr, input_fp);

    png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, png_voidp_NULL);

    if (fclose (input_fp) != 0)
      {
        fprintf (stderr, "Error closing input\n");
        exit (1);
      }

    num_text = png_get_text (png_ptr, info_ptr, &text_ptr, NULL);

    for (i = 0; i < num_text; i++)
      {
        if (strcmp (text_ptr[i].key, keyword) == 0)
          {
            /* fprintf (stderr, "Replacing keyword %s in %s\n",
               keyword, filename); */
            text_ptr[i].key = (char *) keyword;
            text_ptr[i].text = (char *) text;
            text_ptr[i].text_length = strlen (text);
            text_ptr[i].compression = PNG_TEXT_COMPRESSION_NONE;
            goto done_text;
          }
      }

    memset (&textnew, '\0', sizeof (textnew));
    textnew.key = (char *) keyword;
    textnew.text = (char *) text;
    textnew.text_length = strlen (text);
    textnew.compression = PNG_TEXT_COMPRESSION_NONE;
    png_set_text (png_ptr, info_ptr, &textnew, 1);
  done_text:

    /* update last-mod time */
    {
      png_time mod_time;
      png_convert_from_time_t (&mod_time, time (NULL));
      png_set_tIME (png_ptr, info_ptr, &mod_time);
    }
  }

  {
    FILE *output_fp;
    png_structp png_ptr;

    output_fp = fopen (filename, "wb");
    if (output_fp == NULL)
      {
        fprintf (stderr, "Cannot rewrite file\n");
        exit (1);
      }

    png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
                                       NULL, error, warning);
    if (png_ptr == NULL)
      {
        fprintf (stderr, "Cannot create png structure\n");
        exit (1);
      }

    if (setjmp (png_jmpbuf (png_ptr)))
      goto aborting;

    png_init_io(png_ptr, output_fp);

    png_write_png (png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

    if (fclose (output_fp) != 0)
      {
        fprintf (stderr, "Error closing output\n");
        exit (1);
      }
  }
}

int
main (int argc, char **argv)
{
  const char *filename;
  const char *title;

  if (argc != 3)
    {
      fprintf (stderr, "Usage: mung-png filename.png title-text\n");
      exit (1);
    }

  filename = argv[1];
  title = argv[2];

  pngtextadd (filename, "Title", title);
  pngtextadd (filename, "Author", "Kevin Ryde");
  pngtextadd (filename, "Copyright", "\
Copyright 2007 Kevin Ryde\n\
\n\
This file is part of Chart.\n\
\n\
Chart is free software; you can redistribute it and/or modify it under\n\
the terms of the GNU General Public License as published by the Free\n\
Software Foundation; either version 3, or (at your option) any later\n\
version.\n\
\n\
Chart is distributed in the hope that it will be useful, but WITHOUT ANY\n\
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n\
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more\n\
details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program.  If not, see <http://www.gnu.org/licenses/>.\n\
");
  pngtextadd (filename, "Software", "Chart `weights.scm', and gnuplot");

  /* RFC 822 style date (format string per glibc manual on strftime). */
  {
    char buf[128];
    time_t t = time (NULL);
    if (strftime (buf, sizeof (buf), "%a, %d %b %Y %H:%M:%S %z",
                  localtime (&t)) == 0)
      {
        perror ("strftime");
        exit (1);
      }
    pngtextadd (filename, "Creation Time", buf);
  }

  pngtextadd (filename, "Homepage", "http://www.geocities.com/user42_kevin/chart/index.html");

  return 0;
}
