;; Donchian Channels.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart donchian)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.linnsoft.com/tour/techind/donch.htm
;;     Sample Intel (INTC) July-September 2001 (year not marked, but is 2001).


;; Return a procedure which calculates a Donchian channel over a window of
;; COUNT many points.
;;
;; Each call (PROC high low) enters a new point into the window, and the
;; return is multiple (values LOWER MID UPPER) for the three lines.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (donchian-calc-proc count)

  ;; HIGH-Q is the last COUNT many HIGH values
  ;; LOW-Q is the last COUNT many LOW values
  ;; POINTS is how many entries in HIGH-Q, and likewise in LOW-Q
  ;;
  (let ((high-q     (make-q))
	(low-q      (make-q))
	(points     0)
	(points-min (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (high low)
      (let ((upper #f)
	    (lower #f)
	    (mid   #f))

	;; don't use latest day in calculation
	(if (>= points points-min)
	    (begin
	      (set! upper (apply max (q-list high-q)))
	      (set! lower (apply min (q-list low-q)))
	      (set! mid   (arithmetic-mean upper lower))))

	;; drop old point if full
	(if (>= points count)
	    (begin
	      (deq! high-q)
	      (deq! low-q))
	    ;; not full, so POINTS increases
	    (set! points (1+ points)))

	;; add new point
	(enq! high-q high)
	(enq! low-q  low)

	(values lower mid upper)))))


(define (donchian-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count count))
	 (array  (series-array parent start hi))
	 (proc   (donchian-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(proc (or (array-ref array i 1) close)    ;; high
	      (or (array-ref array i 2) close)))) ;; low

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(receive (lower mid upper)
	    (proc (or (array-ref array i 1) close)  ;; high
		  (or (array-ref array i 2) close)) ;; low
	  (array-set! new-array lower i 0)
	  (array-set! new-array mid   i 1)
	  (array-set! new-array upper i 2))))))

(define-public (donchian-series series count)
  (ohlcv-required "Donchian Channel" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Donchian") count))
    (series-datatype!  new-series donchian-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts donchian-fill
							   (1- count)
							   '(0 2)))
    (series-array!     new-series #f)

    new-series))

(define-public donchian-datatype
  (datatype-new!
   #:key       'donchian
   #:name      (_ "Donchian Channel")
   #:units     #f
   #:input     datatype-ohlcv
   #:linestyle 'bands
   #:manual    (_ "CONTEXT(manual-node): Donchian Channel")
   #:proc      donchian-series
   #:closes-apply-arg '(1) ;; midpoint
   #:parameter-list
   (list (datatype-param-new
	  #:key     'donchian-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
