;; Chart mainline.

;; Copyright 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart down-command)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart download)
  #:use-module (chart misc)
  #:use-module (chart symlists))


;;-----------------------------------------------------------------------------
;; download status to tty
;;
;; Status is printed like "foo " with no newline, then gets rewritten by
;; "\rbar " or whatever.  The space moves the cursor away from the message
;; and makes possible diagnostic output more readable.
;;
;; A message erases the status before printing, with "\r \r" or however many
;; spaces necessary.
;;
;; The length of the currently displaying status is obtained from
;; `port-column', so if some diagnostics have moved to the next line then no
;; erasing is attempted (as if there was no current message).

(define output-width #f)

(define (output-width-init)
  (sigaction SIGWINCH (lambda (signum)
			(set! output-width (output-port-width))))
  (set! output-width (output-port-width))

  ;; once only, and gc this code
  (set! output-width-init noop))

(define-public (down-command-status-output str)
  ;; progress messages only to a tty
  (if (isatty? (current-output-port))
      (begin
	;; limit to 79 chars, so as not to take more than one line
	(output-width-init)
	(set! str (utf8-string-take str (min (utf8-string-length str)
					     (1- output-width))))

	;; print spaces to overwrite anything exceeding STR on the current
	;; line (a previous status message)
	(let ((extra (max 0 (- (port-column (current-output-port))
			       (utf8-string-length str)))))
	  (display (utf8->locale
		    (format #f "\r~a~v_~a~!"
			    str extra (make-string extra #\bs)))))

	;; guile 1.6.7 and earlier doesn't count \b towards port-column, and
	;; in both 1.6 and 1.8 if the locale is multibyte the simple
	;; counting of bytes will be wrong too
	(set-port-column! (current-output-port) (utf8-string-length str)))))


;; the test col!=0 could be done with ~[, but guile 1.6.7 seems to have a
;; bit of a bug when "v" is used as a parameter for a directive inside a ~[,
;; an extra argument is discarded, somehow
;;
(define-public (down-command-message str)
  (let ((col (port-column (current-output-port))))
    (if (positive? col)
	(format #t "\r~v_\r" col))
    (display (utf8->locale (format #f "~a\n~!" str)))))


;;-----------------------------------------------------------------------------

(define charset:star-question (char-set #\* #\?))

;; For a shell style wildcard PATTERN string, return a regexp string.
;; A trailing dot like "*." is taken to mean no ".FOO" extension at all,
;; ie. US symbols.
(define (shell-pattern->regexp pattern)
  (let ((any      ".")
	(any-star ".*"))
    (if (string-suffix? "." pattern)
	(begin
	  (set! pattern (string-drop-right pattern 1))
	  (set! any "[^.]")
	  (set! any-star "[^.]*")))

    (string-append "^"
		   (regexp-substitute/global
		    #f "[^A-Za-z0-9 ,]" pattern
		    'pre (lambda (m)
			   (let ((c (string-ref pattern (match:start m))))
			     (case c
			       ((#\*) any-star)
			       ((#\?) any)
			       (else  (string #\[ c #\])))))
		    'post)
		   "$")))

;; return a list of elements from CANDIDATES list which match the string
;; PATTERN
(define (glob pattern candidates)
  (if (string-any charset:star-question pattern)
      (let* ((regexp      (make-regexp (shell-pattern->regexp pattern)))
	     (symbol-list (filter (lambda (candidate)
				    (regexp-exec regexp candidate))
				  candidates)))
	(if (null? symbol-list)
	    (format #t (_ "chart: warning, no matches for ~s\n") pattern))
	symbol-list)
      ;; avoid creating a regexp if no wildcards
      (list pattern)))

(define-public (down-command symbol-list option-backto)

  (if (null? symbol-list)
      (begin
	(display (utf8->locale (_ "chart: no symbols to download
use --all, --favourites, or a list of desired symbols\n"))
		 (current-error-port))
	(exit 1)))

  (let ((candidates (symlist-list all-symlist)))
    (set! symbol-list (append-map (lambda (symbol)
				    (glob symbol candidates))
				  symbol-list)))

  (c-main-enq! download
	       #:symbols (delete-duplicates symbol-list)
	       #:backto option-backto)
  (c-main)
  (exit 0))
