;; Download cost-based strategies.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart down-cost)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart http)
  #:use-module (chart database)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; going by day or by symbol

;; Decide whether to use daily downloads or individual symbol downloads to
;; update all of SYMBOL-LIST.  Download will be from the
;; `download-start-tdate' of each symbol, up to AVAIL-TDATE.
;;
;; DAILY-COST is the cost (in bytes) to download a day.
;; (SYMBOL-COST-PROC tdate) should return the cost of downloading from TDATE
;; to now (ie. AVAIL-TDATE) for a symbol.
;; Both costs will get `http-get-cost' added to them automatically.
;;
;; The return is two values (USE-SYMBOL-LIST TDATE), which is a list of
;; symbols to do individually, and a tdate to start from for daily
;; downloads.  That tdate is (1+ avail-tdate) when no dailies should be
;; done.
;;
;; As an example, if there's a few symbols quite a bit behind but the rest
;; needing just a day or two then the result can be to do the laggards
;; individually, then whole days for the most recent.
;;

(define-public (download-by-day-or-by-symbol avail-tdate symbol-list
					     symbol-cost-proc daily-cost)
  (set! daily-cost (+ daily-cost http-get-cost))

  (let* ((by-date           '())
	 (best-tdate        #f)
	 (best-cost         #f)
	 (best-symbol-list  '())
	 (total-symbol-cost 0)
	 (use-symbol-list   '()))

    (define (consider tdate cost)
      (if (or (not best-tdate)
	      (< cost best-cost ))
	  (begin
	    (set! best-tdate tdate)
	    (set! best-cost cost)
	    (set! best-symbol-list use-symbol-list))))

    ;; build by-tdate alist of (TDATE . SYMBOL-LIST)
    ;;
    (for-each (lambda (symbol)
		(let ((start-tdate (download-start-tdate symbol)))
		  (set! by-date
			(assv-set! by-date start-tdate
				   (cons symbol
					 (or (assv-ref by-date start-tdate)
					     '()))))))
	      symbol-list)
    (set! by-date (sort! by-date car<))  ;; oldest dates first

    (for-each (lambda (elem)
		(let* ((tdate       (car elem))
		       (symbol-list (cdr elem)))
		  (consider tdate (+ total-symbol-cost
				     (* (- avail-tdate tdate -1) daily-cost)))

		  ;; move the ELEM symbol-list onto total-symbol-cost
		  (set! total-symbol-cost (+ total-symbol-cost
					     (* (length symbol-list)
						(+ (symbol-cost-proc tdate)
						   http-get-cost))))
		  (set! use-symbol-list (append symbol-list use-symbol-list))))
	      by-date)

    ;; perhaps do everything with hist
    (consider (1+ avail-tdate) total-symbol-cost)

    ;; same order as original symbol-list
    (set! best-symbol-list (lset-intersection string=?
					      symbol-list best-symbol-list))

    (values best-symbol-list best-tdate)))


;;-----------------------------------------------------------------------------
;; download cost adaption
;;
;; Download sizes are recorded in a cache file, to adapt to new sizes if
;; downloads are bigger than the defaults, over time.
;;
;; Costs recorded are not decreased immediately, only if the previous
;; recording is over 7 days old will a drop be allowed.  This avoids
;; temporary drops from getting in, but still allows an anomalous big value
;; to get replaced after a while.
;;
;; When retrieving a value, values less than half the default are ignored,
;; as a protection against empty sizes resulting from server errors etc.
;;

(define download-cost-file (g-build-filename cache-directory
					     "download-costs"))

(define-public (download-cost-store key body)
  (define time (current-time))
  (if (string? body)
      (set! body (string-length body)))

  (database-call-with-lock cache-directory
    (lambda ()
      (let ((lst (call-with-input-file-or '() download-cost-file read)))

	;; only keep entries newer than 7 days
	(let ((old-time (- time (* 7 86400))))
	  (set! lst (filter! (lambda (elem)
			       (> (third elem) old-time))
			     lst)))

	(let ((val (assq-ref lst key)))
	  (if (or (not val)              ;; no prev entry
		  (> body (first val)))  ;; bigger than prev recorded
	      (begin
		(set! lst (assq-set! lst key (list body time)))

		(call-with-output-file download-cost-file
		  (lambda (port)
		    (display ";; Download costs, generated by Chart
;; Format: (KEY COST-IN-BYTES UNIX-TIME_T-SEEN)
" port)
		    (write-list-lines lst port)
		    (newline port))))))))))

(define-public (download-cost-get key default)
  (let ((val (assq-ref (call-with-input-file-or '() download-cost-file read)
		       key)))
    (if val
	(max (first val) (quotient default 2))
	(begin
	  (download-cost-store key default)
	  default))))


;;-----------------------------------------------------------------------------
;; selecting among possibly overlapping files


;; return true if range (alo,ahi) overlaps range (blo,bhi)
;; each endpoint is taken as inclusive, so say (1,4) and (4,7) do overlap
(define (overlap-inclusive? alo ahi blo bhi)
  (not (or (< ahi blo)
	   (> alo bhi))))

(define (downranges-cover? lst want-lo want-hi)
  (let outer ((upto want-lo))
    (if (> upto want-hi)
	#t
	(let inner ((lst lst))
	  (if (null? lst)
	      #f
	      (let-keywords (first lst) #t ((lo #f) (hi #f))
		(if (within-range? lo hi upto)
		    (outer (1+ hi))
		    (inner (cdr lst)))))))))

(define (for-each-combination proc lst)
  (define len (length lst))
  (do ((bits 0 (1+ bits))
       (this '() '()))
      ((>= bits (ash 1 len)))
    (do ((i (1- len) (1- i)))
	((negative? i))
      (if (logbit? i bits)
	  (set! this (cons (list-ref lst i) this))))
    (proc this)))
  
(define (downranges-cost lst)
  (fold (lambda (elem total)
	  (+ (or (plist-get elem #:cost) 0)
	     http-get-cost
	     total))
	0 lst))

(define-public (downranges-select lst want-lo want-hi)
  (if (> want-lo want-hi)
      '()

      (begin
	;; restrict wanted range to what's available
	(let ((got-lo #f)
	      (got-hi #f))
	  (for-each (lambda (elem)
		      (let-keywords elem #t ((lo #f) (hi #f))
			(set! got-lo (min-maybe got-lo lo))
			(set! got-hi (max-maybe got-hi hi))))
		    lst)
	  (set! want-lo (max want-lo got-lo))
	  (set! want-hi (min want-hi got-hi)))

	;; only look at elements which cover some of the wanted range
	(set! lst (filter (lambda (elem)
			    (let-keywords elem #t ((lo #f) (hi #f))
			      (overlap-inclusive? lo hi want-lo want-hi)))
			  lst))

	(let ((best-cost #f)
	      (best-lst  #f))
	  (for-each-combination (lambda (lst)
				  (if (downranges-cover? lst want-lo want-hi)
				      (let ((cost (downranges-cost lst)))
					(if (or (not best-lst)
						(< cost best-cost))
					    (begin
					      (set! best-cost cost)
					      (set! best-lst  lst))))))
				lst)
	  best-lst))))
