;; Download helpers.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart download)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 regex)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:autoload   (chart down-command)  (down-command-status-output
				      down-command-message)
  #:use-module (chart http)
  #:use-module (chart database)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart symlists))


;;-----------------------------------------------------------------------------
;; list stuff

;; return key matching VAL in ALIST
(define-public (rassoc-ref alist val)
  (and (not (null? alist))
       (if (equal? val (cdar alist))
	   (caar alist)
	   (rassoc-ref (cdr alist) val))))

;; return true if any predicate in PRED-LIST returns true on OBJ
(define-public (any-pred pred-list obj)
  (any (lambda (pred)
	 (pred obj))
       pred-list))

(define-public (map-rows! proc row-list)
  (filter-map-in-order! (lambda (row)
			  (apply proc row))
			row-list))

(define (filter-map-in-order! proc lst)
  (delq #f (map-in-order! proc lst)))

(define map-in-order! map!)


;;-----------------------------------------------------------------------------
;; string stuff

;; return list of non-empty lines in STR
;; CR, NL or CRLF line endings all work
(define-public (string-split-lines str)
  (string-tokenize str (char-set-complement char-set:crlf)))

;; trim PART from the right of STR
(define-public (string-trim-right-str str part)
  (if (string-suffix? part str)
      (string-drop-right str (string-length part))
      str))

;; return position after PART in STR, or #f if PART not in STR
(define-public (string-contains-after-ci str part . rest)
  (let ((pos (apply string-contains-ci str part rest)))
    (and pos
	 (+ pos (string-length part)))))

;; Return STR with the first line dropped.
;; The first line is as per `string-first-line' and `string-split-lines',
;; ie. CR or NL or any combination is the end of the line.  This means any
;; blank lines after the first are dropped too.
;; If STR doesn't have any CR or NL then it's just one line and "" is
;; returned.
;;
(define-public (string-drop-first-line str)
  (let ((idx (string-index str char-set:crlf)))
    (if idx
	(string-trim str char-set:crlf idx)
	"")))

;; return the last non-empty line of STR, excluding CR or NL
(define-public (string-last-line str)
  (let* ((endlast   (string-skip-right str char-set:crlf))
	 (startlast (or (string-index-right str char-set:crlf 0 endlast)
			-1)))
    (substring str (1+ startlast) (1+ endlast))))

;; return the initial portion of STR up to but excluding CHAR
(define-public (string-take-to-char str char)
  (string-take str (string-index str char)))

;; turn runs of whitespace in STR into single space each
(define-public (string-collapse-whitespace str)
  (let ((prev #\x)
	(ret  '()))
    (string-for-each (lambda (c)
		       (if (char-whitespace? c)
			   (if (not (char-whitespace? prev))
			       (set! ret (cons #\space ret)))
			   (set! ret (cons c ret)))
		       (set! prev c))
		     str)
    (list->string (reverse! ret))))


;;-----------------------------------------------------------------------------
;; date stuff

(define-public (mdate-today . zone)
  (apply ymd->mdate (apply today-ymd zone)))

;; Return a year to go with MONTH, within -5 to +6 months of today.
;; The dates considered are the 12 months from today's month less 5, through
;; to today's month plus 6.  Whichever one of those is the given MONTH is
;; then taken to be the date desired, and its year is returned.
(define-public (month->nearest-year month)
  (let* ((base-mdate (- (mdate-today) 5)))
    (mdate->year (+ base-mdate (modulo (- month 1 base-mdate) 12)))))

;; return code character for MONTH (1 to 12)
(define-public (month->M-code month)
  (string-ref "FGHJKMNQUVXZ" (1- month)))

;; return a string like "Z05" for the given MDATE
(define-public (mdate->MYY-str mdate)
  (receive-list (year month day)
      (mdate->ymd mdate)
    (format #f "~a~2,'0d" (month->M-code month) (modulo year 100))))

;; return list (TDATE TIME) where TIME is seconds since midnight
(define-public (tm->tdate-time tm)
  (list (tm->tdate tm)
	(tm->time tm)))

;; convert TM to list (TDATE TIME), but lock to within weekday times LO to HI
;; LO and HI are stretched by 2 minutes in case our system time is out
(define-public (tm->tdate-time-within tm lo hi)
  (set! lo (- lo #,(hms->seconds 0 2 0)))
  (set! hi (+ hi #,(hms->seconds 0 2 0)))
  (let* ((tdate (tm->tdate tm))
	 (time  (tm->time tm)))
    (cond ((tm:weekend? tm)
	   ;; weekend lock to Friday HI
	   (set! time hi))
	  ((< time lo)
	   ;; weekday before LO, use HI previous day
	   (set! tdate (1- tdate))
	   (set! time  hi))
	  (else
	   ;; weekday after LO, lock to HI
	   (set! time (min time hi))))
    (list tdate time)))

;; convert TM to list (ADATE TIME), but lock to within weekday times LO to HI
;; LO are stretched by 2 minutes in case our system time is out
(define-public (tm->adate-time-within tm lo hi)
  (set! lo (- lo #,(hms->seconds 0 2 0)))
  (let ((adate (tm->adate tm))
	(time  (tm->time tm)))
    (cond ((tm:weekend? tm)
	   ;; weekend lock to Friday HI
	   (set! adate (- adate (modulo adate 7) -4))
	   (set! time hi))
	  ((< time lo)
	   ;; weekday before LO, use HI previous day (Fri for Mon)
	   (set! adate (- adate (if (= 1 (tm:wday tm))
				    3 1)))
	   (set! time  hi))
	  (else
	   ;; weekday after LO, lock to HI
	   (set! time (min time hi))))
    (list adate time)))

;; return list (ADATE TIME) where TIME is seconds since midnight
(define-public (tm->adate-time tm)
  (list (tm->adate tm)
	(tm->time tm)))

;; return the last tdate in MDATE
(define-public (mdate->last-tdate mdate)
  (1- (mdate->tdate (1+ mdate))))

;; return the last tdate in YEAR
;; YEAR is a 4 digit value like 2007
(define-public (year->last-tdate year)
  (1- (year->tdate (1+ year))))

;; return the first adate prior to the given ADATE which is a valid tdate
;; for a weekday this means the previous weekday (Monday goes back to Friday)
;; for a weekend this means the previous Friday
(define-public (adate-prev-tdate adate)
  (if (adate-weekend? adate)
      (tdate->adate (adate->tdate adate))
      (tdate->adate (1- (adate->tdate adate)))))


;;-----------------------------------------------------------------------------
;; date parsing stuff

;; separator "." is for casablanca
(define-public date-part-char-set (char-set-complement
				   (char-set #\/ #\- #\, #\. #\space)))

(define-public (month-str->month str)
  (or (Mmm-str->month str)
      (begin
	(or (string-every char-set:digit str)
	    (error "Unrecognised month:" str))
	(let ((month (string->number str)))
	  (or (within-range? 1 12 month)
	      (error "Month out of range:" month))
	  month))))

(define-public (mday-str->day str)
  (or (string-every char-set:digit str)
      (error "Unrecognised day of month:" str))
  (let ((day (string->number-err str)))
    (or (within-range? 1 31 day)
	(error "Day of month out of range:" day))
    day))

;; return list (YEAR MONTH DAY) for STR like "30/6/04", "30/6/2004",
;; "28.07.2006", "28 July 2006"
(define-public (d/m/y-str->ymd str)
  (receive-list (day month year)
      (string-tokenize str date-part-char-set)
    (list (yyy-str->year year)
	  (month-str->month month)
	  (mday-str->day day))))

;; return adate for STR like "30/6/04", "30/6/2004", "28.07.2006", "28 July
;; 2006"
(define-public (d/m/y-str->adate str)
  (apply ymd->adate (d/m/y-str->ymd str)))

;; return tdate for STR like "30/6/04" or "30/6/2004"
(define-public (d/m/y-str->tdate str)
  (adate->tdate (d/m/y-str->adate str)))


;; return list (YEAR MONTH DAY) for STR like "6/30/04", "6/30/2004",
;; "Jul 13. 2006"
(define-public (m/d/y-str->ymd str)
  (receive-list (month day year)
      (string-tokenize str date-part-char-set)
    (list (yyy-str->year year)
	  (month-str->month month)
	  (mday-str->day day))))

;; return adate for STR like "6/30/04" or "6/30/2004"
(define-public (m/d/y-str->adate str)
  (apply ymd->adate (m/d/y-str->ymd str)))

;; return tdate for STR like "6/30/04" or "6/30/2004"
(define-public (m/d/y-str->tdate str)
  (adate->tdate (m/d/y-str->adate str)))


(define-public (y/m/d-str->ymd str)
  (receive-list (year month day)
      (string-tokenize str date-part-char-set)
    (list (yyy-str->year year)
	  (month-str->month month)
	  (mday-str->day day))))

;; return adate for STR like "2004/6/30" or  "2000/01/04" or "2004-13-31"
(define-public (y/m/d-str->adate str)
  (apply ymd->adate (y/m/d-str->ymd str)))

;; return tdate for STR like "2004/6/30" or  "2000/01/04" or "2004-13-31"
(define-public (y/m/d-str->tdate str)
  (adate->tdate (y/m/d-str->adate str)))


;; return tdate for STR like "20050422"
(define-public (yyyymmdd-str->tdate str)
  (ymd->tdate (string->number-err (substring str 0 4))
	      (month-str->month (substring str 4 6))
	      (mday-str->day (substring str 6 8))))


;; return adate for a string like "20 April", using whatever year puts that
;; day/month closest to today
(define-public (d/m-str->adate-nearest str)
  (receive-list (day month)
      (string-tokenize str date-part-char-set)
    (set! month (month-str->month month))
    (ymd->adate (month->nearest-year month)
		month
		(mday-str->day day))))

;; return tdate for a string like "20 April", using whatever year puts that
;; day/month closest to today
(define-public (d/m-str->tdate-nearest str)
  (adate->tdate (d/m-str->adate-nearest str)))


;;-----------------------------------------------------------------------------
;; time parsing stuff

;; return seconds since midnight for STR like "16:43", "16:43:00", "7:30pm"
(define-public (h:m:s-str->hms str)
  ;;                                1        2           3 4               56    7
  (let* ((m    (or (regexp-exec #,(regexp "^([0-9]+):([0-9][0-9])(:([0-9][0-9]))? *((AM)|(PM))?$" regexp/icase) str)
		   (error "Bad H:M:S time string:" str)))
	 (hour (string->number-err (match:substring m 1))))
    (list (+ hour
	     (if (and (match:start m 6)  ;; "AM", adj 12am -> 0
		      (= hour 12))
		 -12 0)
	     (if (and (match:start m 7)  ;; "PM", add 12 except at 12pm
		      (< hour 12))
		 12 0))
	  (string->number-err (match:substring m 2))
	  (if (match:start m 4)
	      (string->number-err (match:substring m 4))
	      0))))

(define-public (h:m:s-str->seconds str)
  (apply hms->seconds (h:m:s-str->hms str)))


;;-----------------------------------------------------------------------------
;; regexp stuff

;; this is designed for use as "(must-match (string-match ...))", it gives
;; an immediate error message if there was no match, such an error is
;; clearer than a subsequent error from match:substring etc getting #f
;;
;; a list from list-matches is allowed, but it must be non-empty
;;
(define-public (must-match m)
  (or (regexp-match? m)
      (and (list? m)
	   (not (null? m))
	   (every regexp-match? m))
      (error "Regexp didn't match"))
  m)


;;-----------------------------------------------------------------------------
;; data available tdates

;; return today's tdate in ZONE, but only at or after HOUR:MINUTE, before
;; that time the return is the previous tdate
;; on a weekend (and until HOUR:MINUTE on Monday), the return is Friday
;;
(define-public (tdate-today-after hour minute zone)
  (let* ((tm (localtime (current-time) zone)))
    (- (tm->tdate tm)
       (if (and (tm:weekday? tm)
		(two< (tm:hour tm) (tm:min tm) hour minute))
	   1 0))))


;;-----------------------------------------------------------------------------
;; source help dispatch

;; list of elements which are lists (PRED NODENAME PRIORITY)
;; higher values of PRIORITY come first
;;
(define-public source-help-list '())

(define*-public (source-help! pred node #:optional (priority 0))
  (set! source-help-list
	(merge! source-help-list (list (list pred node priority))
		(lambda (x y)
		  (> (third x) (third y))))))


;;-----------------------------------------------------------------------------
;; intraday handler dispatch

;; list of (SYMBOL-PRED URL-PROC MODES-PROC)
;;
;; Calls (SYMBOL-PRED symbol) find the handler for SYMBOL.
;;
;; Call (MODES-PROC) in `intraday-modes' returns list of mode strings, like
;; ("1 Day" "5 Day"), or empty list for none.
;;
;; Call (URL-PROC symbol mode) in `intraday-url', where MODE is one of the
;; MODES-PROC strings, returns the url of an image (possibly doing other
;; downloads to find it, if necessary).
;;
(define-public intraday-handler-list '())

(define*-public (intraday-url-proc! #:key symbol-pred url-proc modes-proc)
  (set! intraday-handler-list (cons (list symbol-pred url-proc modes-proc)
				    intraday-handler-list)))


;;-----------------------------------------------------------------------------
;; download status and messages
;;
;; `download-message' takes a utf8 string to print as a message.
;;
;; `download-status' takes a list of utf8 strings to print, separated by
;; spaces, as a current status.  This might be truncated to 80 columns on a
;; tty, and in any case is just informational.
;;
;; When a status is printed, further new status is held but not printed.  At
;; the end of 0.5 seconds the most recent status is printed.  This is
;; designed not to hammer either a terminal or a gtk label display.
;;
;; download-message and the internal download-status-output functions are
;; `set!' to gui versions by down-dialog.scm.  The defaults here are
;; autoloads of code in down-command.scm, designed to do some good in a
;; script calling `(download ...)'.
;;

(define status-str "")

;; print STR as a download message, an error, warning, whatever
;; STR should not have a newline at the end
(define-public (download-message str)
  (down-command-message str))

(let-toplevel ((paused     #f)
	       (paused-str #f)
	       (main-str   ""))

  (define (unpause)
    (set! paused #f)
    (if paused-str
	(begin
	  (download-status-output status-str)
	  (set! paused-str #f))))

  (define (show-status str)
    (set! status-str str)
    (if paused
	(set! paused-str status-str)
	(begin
	  (download-status-output status-str)
	  (set! paused #t)
	  (c-gtk-timeout-add 500 unpause))))

  ;; display the http status STR, on top of the main status set by
  ;; `download-status'
  (define-toplevel (download-http-status str)
    (show-status (string-append main-str " [" str "]")))

  ;; display STR now as the download status
  ;; the default here `down-command-status-output' is for command line
  ;; display to a tty, and this function is overridden by the GUI setups
  (define-public (download-status-output str)
    (down-command-status-output str))

  (define-public (download-status-active . str-list)
    (set! main-str (string-join str-list))
    (show-status main-str)))

;; print or display STR-LIST as the current status of downloading
;; the status is not necessarily printed immediately, a delay is applied so
;; frequent status changes don't get bogged down in terminal output
;;
(define-public (download-status . str-list)
  (apply (fluid-ref download-status-proc) str-list))

(define download-status-proc (make-fluid))
(fluid-set! download-status-proc noop)


;;-----------------------------------------------------------------------------
;; download status helpers

(define-public (tdate-range-string lo hi)
  (if (= lo hi)
      (strftime-utf8 (locale-d-fmt) (tdate->tm lo))
      (simple-format #f (_ "CONTEXT(date-range): ~a to ~a")
		     (strftime-utf8 (locale-d-fmt) (tdate->tm lo))
		     (strftime-utf8 (locale-d-fmt) (tdate->tm hi)))))

(define-public (symbol-list-range-string symbol-list)
  (if (null? (cdr symbol-list))
      (first symbol-list)
      (simple-format #f (_ "CONTEXT(symbol-range): ~a to ~a")
		     (first symbol-list) (last symbol-list))))


;;-----------------------------------------------------------------------------
;; download handler dispatch

(let-toplevel ((rec-type (make-record-type "download-handler"
					   '(selector update backto
						      priority avail-proc))))
  (define rec-new        (record-constructor rec-type))
  (define rec-selector   (rec-accessor rec-type 'selector))
  (define rec-update     (rec-accessor rec-type 'update))
  (define rec-backto     (rec-accessor rec-type 'backto))
  (define rec-priority   (rec-accessor rec-type 'priority))
  (define rec-avail-proc (rec-accessor rec-type 'avail-proc))

  (define handler-list '())

  ;; return true if record X > Y, compared by priority
  (define (rec-priority> x y)
    (> (rec-priority x) (rec-priority y)))

  (define*-public (download-handler!
		   #:key
		   selector
		   update
		   (backto (error "download-handler!: no #:backto"))
		   (priority 0)
		   avail-proc)
    (set! handler-list
	  (merge handler-list
		 (list (rec-new selector update backto priority avail-proc))
		 rec-priority>)))

  (define*-public (download #:key
			    symbols  ;; list of symbols
			    backto)  ;; tdate, or default #f for update

    (for-each symbol-setups symbols)
    (apply database-new symbols)

    ;; elevated out of historical list
    (if (not backto)
	(historical-remove symbols))

    (download-status (_ "Selecting"))
    (with-fluids ((http-status-callback download-http-status)
		  (download-status-proc download-status-active))
      (for-each
       (lambda (rec)
	 (let ((this-list (filter (rec-selector rec) symbols)))
	   (if (not (null? this-list))
	       ;; errors shown as messages, but continue with other
	       ;; handlers
	       (c-lazy-catch #t
		 (lambda ()
		   (download-status "")
		   (if backto
		       ((rec-backto rec) backto this-list)

		       (begin
			 ;; exclude any already up to date
			 (and-let* ((avail-proc  (rec-avail-proc rec))
				    (avail-tdate (avail-proc)))
			   (set! this-list
				 (filter (lambda (symbol)
					   (>= avail-tdate
					       (download-start-tdate symbol)))
					 this-list)))
			 (if (not (null? this-list))
			     (begin
			       ((rec-update rec) this-list)
			       (historical-check this-list))))))

		 (lambda args
		   (if (eq? (first args) 'job-stop)
		       (apply throw args))

		   ;; a bit of protection here in case a source has
		   ;; forgotten to call `download-status' to identify itself
		   ;; before a `http-request
		   (let ((base (if (string-prefix? " " status-str)
				   (simple-format #f "Oops, unknown source operating on `~a'"
						  (procedure-name (rec-selector rec)))
				   status-str)))
		     (download-message
		      (string-join-notnull (list base
						 (string-escape-unprintable
						  (error-message args)))
					   ": ")))

		   ;; for connection or http protocol problems
		   ;; (including http timeout), just the error
		   ;; message is enough, for anything else dump a
		   ;; backtrace
		   (if (not (error-connection-related? args))
		       (let ((stack (make-stack #t 2)))
			 (cond (stack
				(download-message
				 (call-with-output-string
				  (lambda (port)
				    (c-display-backtrace stack port)))))
			       ((memq 'debug (debug-options))
				(download-message (_ "No backtrace available.\n")))
			       (else
				(download-message (_ "No backtrace available, run with --debug to get one.\n")))))))))))

       handler-list)

      (download-status "")))

  ;; return true if there's some data to update for SYMBOL (or might be)
  (define-public (download-update? symbol)
    (let* ((rec (find (lambda (rec)
			((rec-selector rec) symbol))
		      handler-list))
	   (avail-proc (and rec (rec-avail-proc rec)))
	   (avail-tdate (and avail-proc (avail-proc))))
      (if avail-tdate
	  (>= avail-tdate (download-start-tdate symbol))
	  ;; assume worth trying if don't know for sure it isn't
	  #t))))

(define (string-escape-unprintable str)
  (string-concatenate-reverse
   (string-fold (lambda (c prev)
		  (cons (if (char-set-contains? char-set:printing c)
			    (string c)
			    (format #f "\\x~:@(~2,'0x~)" (char->integer c)))
			prev))
		'() str)))


;;-----------------------------------------------------------------------------
;; download-now -- checking conditions for download

;; list of procedures called (PROC symbol-list) returning string message, or #f
(define-public download-now-list '())

(define-public (download-now-handler! proc)
  (set! download-now-list (append! download-now-list (list proc))))

(define-public (download-now-commodity-all-months)
  (_ "\
Data download from this source gives all months for a commodity in one go.  \
If you're adding several contract months of the same commodity then you can \
download just once after making all desired additions."))

(define-public (download-now-all-commodities-and-months)
  (_ "\
Data download from this source gives all commodities and all months in one \
go.  If you're adding several contract months or several commodities to the \
database then you can download just once after making all desired additions."))


;;-----------------------------------------------------------------------------
;; download names using latest records

;; if SYMBOL has no name, or the name is old
(define-public (download-want-name? symbol)
  (or (not (database-read-info symbol 'name))
      (let ((t (database-read-info symbol 'name-time)))
	(or (not t)
	    (< t (- (current-time) (* 30 86400)))))))

;; download updated names for SYMBOL-LIST
;; SOURCE is a string, the name of the exchange or source
(define-public (download-name-from-latest source symbol-list)
  (for-each (lambda (symbol)
	      (if (download-want-name? symbol)
		  (begin
		    (download-status source (_ "name") symbol)
		    (let ((latest (latest-get-today symbol)))
		      (database-call-modify-data symbol
			(lambda (series)
			  (if latest
			      (series-name! series (latest-name latest)))
			  (series-info-add! series 'name-time
					    (current-time))))))))
	    symbol-list))


;;-----------------------------------------------------------------------------
;; download ETag / Last-Modified

;; return a `series-info' list element for tag of URL from data in HEADERS
;; URL and HEADERS can be #f when no tag info
(define-public (download-tag-info url headers)
  (cons 'download-tag
	(download-tag-value url headers)))

(define-public (download-tag-value url headers)
  (list url
	(and headers (http-headers-get headers "ETag"))
	(and headers (http-headers-get headers "Last-Modified"))))

;; URL and HEADERS can be #f when no tag info
(define-public (download-series-tag-info! series url headers)
  (series-info-add! series 'download-tag
		    (download-tag-value url headers)))

;; return a list of args for `http-request', #:if-none-match and
;; #:if-modified-since args from the database for SYMBOL-LIST
;; all of SYMBOL-LIST must have the same recorded URL and tag data to use
;; those request options
(define-public (download-tag-args url symbol-list)
  (let ((val (database-read-info (first symbol-list) 'download-tag)))
    (if (and val
	     (equal? url (first val))
	     (every (lambda (symbol)
		      (equal? val (database-read-info symbol 'download-tag)))
		    (cdr symbol-list)))
	(list #:if-none-match     (second val)
	      #:if-modified-since (third val))
	'())))


;;-----------------------------------------------------------------------------
;; misc helpers

;; modify LST to insert ELEM at POS
;; POS counts from 1 for just after the first element
(define-public (list-insert-at! elem pos lst)
  (set! lst (list-tail lst (1- pos)))
  (set-cdr! lst (cons elem (cdr lst))))

;; return true if SERIES is from an empty database file
(define (database-empty-series? series)
  (and (= (series-lo series) (series-hi series))
       (every not (first (array->list (series-array series
						    (series-lo series)
						    (series-hi series)))))))

;; tdate at which to start downloading for SYMBOL
(define*-public (download-start-tdate symbol #:key initial)
  (let ((init (- (tdate-today (symbol-timezone symbol))
		 (or initial initial-download-days))))

    (and-let* ((series (database-read-series symbol)))
      ;; where the data goes up to
      (if (not (database-empty-series? series))
	  (set! init (max init (series-hi series))))

      ;; or where the download-done marker says was done
      (and-let* ((done (assq-ref (series-info series) 'download-done)))
	(set! init (max init done))))

    (1+ init)))

;; set the `download-done' info in each of SYMBOL-LIST to TDATE
(define-public (download-done symbol-list suffix tdate)
  (for-each (lambda (symbol)
	      (set! symbol (string-append symbol suffix))
	      (c-idle) ;; pause between writes
	      (database-call-modify-data symbol
		(lambda (series)
		  (series-info-add! series 'download-done tdate))))
	    symbol-list))

;; tdate at which to start downloading for SYMBOL going backwards, ie. one
;; before earliest data
(define-public (download-back-tdate symbol)
  (let ((series (database-read-series symbol)))
    (if (or (not series)
	    (database-empty-series? series))
	(1+ (tdate-today (symbol-timezone symbol)))

	;; where there's data, or where download-done-start marker says covered
	(1- (min-maybe (series-lo series)
		       (series-info-get series 'download-done-start))))))

;; tdate maximum among SYMBOL-LIST at which to start downloading going
;; backwards, ie. one before earliest data
(define-public (download-back-tdate-max symbol-list)
  (apply max (map download-back-tdate symbol-list)))


;; return SYMBOL-LIST but adding any other symbols which satisfy PRED
(define*-public (download-also symbol-list
			       #:key
			       (selector (error "download-also: no selector"))
			       start-tdate
			       end-tdate
			       commodity    ;; true to get only same commodity
			       (pred noop)  ;; (PRED symbol) true or false
			       expiry-proc) ;; (EXPIRY-PROC symbol) #f or tdate
  (set! commodity (and commodity
		       (chart-symbol-commodity (first symbol-list))))
  (append symbol-list
	  (filter (lambda (symbol)
		    ;; avoid repeated file reads
		    (define d-start-tdate
		      (make-procedure-with-cache-0
		       (lambda ()
			 (download-start-tdate symbol))))

		    (and (selector symbol)
			 (or (not commodity)
			     (string=? commodity
				       (chart-symbol-commodity symbol)))
			 (pred symbol)
			 (or (not start-tdate)
			     (>= (d-start-tdate) start-tdate))
			 (or (not end-tdate)
			     (<= (d-start-tdate) end-tdate))
			 (or (not expiry-proc)
			     (let ((expiry-tdate (expiry-proc symbol)))
			       (or (not expiry-tdate)
				   (<= (d-start-tdate) expiry-tdate))))))
		  (lset-difference string=? (symlist-list all-symlist)
				   symbol-list))))

;; return SYMBOL with just commodity and suffix, ie. no specific month
(define-public (chart-symbol-commodity+suffix symbol)
  (string-append (chart-symbol-commodity symbol)
		 (chart-symbol-suffix symbol)))

;; Call (PROC elem) for each element in LST.
;; Return the elem for which PROC returns smallest number.
;; PROC can return #f, such elements are ignored.
;;
(define-public (find-minimum proc lst)
  (let ((result #f)
	(val    #f))
    (for-each (lambda (elem)
		(let ((v (proc elem)))
		  (if (and v
			   (or (not val)
			       (< v val)))
		      (begin
			(set! result elem)
			(set! val v)))))
	      lst)
    result))


;;-----------------------------------------------------------------------------
;; misc parsing helpers

;; undo java string backslash quoting in STR
(define (java-string-unquote str)
  (regexp-substitute/global #f #,(regexp "\\\\[\"'n]") str
			    'pre
			    (lambda (m)
			      (let ((c (string-ref (match:substring m) 1)))
				(case c
				  ((#\n) "\n")
				  (else (string c))))) ;; \' and \"
			    'post))

;; return output produced by java document.write('foo') calls in STR
(define-public (java-write str)
  (java-string-unquote
   (string-concatenate-reverse/shared
    (fold-matches #,(regexp "document.write\\('((\\\\.|[^\\'])*)'\\)") str '()
		  (lambda (m prev)
		    (cons (match:substring m 1) prev))))))

;; return a list of strings which is STR broken up into pieces of WIDTH-LIST
(define-public (string-split-widths str width-list)
  (let ((pos 0))
    (map-in-order (lambda (width)
		    (set! pos (+ pos width))
		    (string-trim-both (substring str (- pos width) pos)))
		  width-list)))

;; call (PROC str) for each line in DATA, STR doesn't include newlines etc
;; can cope with LF, CRLF or just CR
;;
;; iterating like this is designed to cope with the slightly large DATA for
;; a csv month file (about 250 kbytes), a simple string-split would use a
;; lot of memory duplicating the data (esp in 1.6 without shared strings)
;;
(define-public (for-each-line proc data)
  (let more ((upto 0))
    (if (< upto (string-length data))
	(let ((end (or (string-index data char-set:crlf upto)
		       (string-length data))))
	  (proc (substring data upto end))
	  (more (or (string-skip data char-set:crlf end)
		    (string-length data)))))))

;; split STR as comma separated values, return a list of strings
;; quotes "" are stripped from quoted parts of STR
(define-public (csv-split str)
  (let more ((lst '())
	     (pos 0))
    (if (string-prefix? "\"" str 0 1 pos)
	(let ((quot (string-index str #\" (1+ pos))))
	  (if quot
	      (if (string-prefix? "," str 0 1 (1+ quot))
		  (more (cons (substring str (1+ pos) quot) lst) (+ 2 quot))
		  (reverse! (cons (substring str (1+ pos) quot) lst)))
	      ;; oops no closing quote, take to end of string
	      (reverse! (cons (substring str (1+ pos)) lst))))
	(let ((comma (string-index str #\, pos)))
	  (if comma
	      (more (cons (substring str pos comma) lst) (1+ comma))
	      (reverse! (cons (substring str pos) lst)))))))


;;-----------------------------------------------------------------------------
;; number parsing stuff

;; convert STR to a number, returning #f for empty "" but otherwise not
;; allowing non-numeric strings
(define-public (string->number-or-empty str)
  (and (not (string-null? str))
       (string->number-err str)))

;; `strings->numbers+decimals' but reduce to PREFER decimals if the values
;; all have enough zeros that nothing is lost by doing so
;;
(define-public (strings->numbers+decimals-prefer prefer . str-list)
  (receive-list (decimals . value-list)
      (apply strings->numbers+decimals str-list)

    (and-let* ((       (> decimals prefer))
	       (factor (integer-expt 10 (- decimals prefer)))
	       (       (every (lambda (value)
				(or (not value)
				    (= 0 (modulo value factor))))
			      value-list)))
      (set! decimals prefer)
      (set! value-list (map! (lambda (value)
			       (and value
				    (quotient value factor)))
			     value-list)))

    (cons decimals value-list)))

;; array is an OHLCV containing strings and #f's in the OHLC parts
;; convert the strings to numbers and return how many decimals were used
(define-public (array-strings->numbers+decimals array)
  ;; OHLC parts only
  (set! array (make-shared-array array list
				 (first (array-shape array)) '(0 3)))
  (let ((decimals 0))
    (array-for-each (lambda (elem)
		      (if (string? elem)
			  (set! decimals (max decimals
					      (string-decimals elem)))))
		    array)
    (array-map! array (lambda (elem)
			(if (string? elem)
			    (string->number+decimals elem decimals)
			    elem))
		array)
    decimals))

;; return number OBJ, with 0 converted to #f
(define-public (0->f obj)
  (and (not (eqv? 0 obj))
       (not (eqv? 0.0 obj))
       obj))

;; Return STR with trailing zeros trimmed to give just two decimal places,
;; if possible.
;;
;; Only trailing zeros are trimmed, and at least two decimals are left
;; (assuming there's two decimals to start with).  If STR is not a number
;; then it's returned unchanged.
;;
(define-public (string-trim-to-2-decimals str)
  (let ((m (regexp-exec #,(regexp "^([0-9]*[.]..(0*[1-9]+)*)0+$") str)))
    (if m
	(match:substring m 1)
	str)))


;; normalize a number in OBJ
;; OBJ is usually a string, but can be a number or #f
;; some transformations are made, but stripping "," commas, whitespace, and
;; taking "-" and "N/A" to be #f
;;
(define-public (crunch-number obj)
  (cond ((number? obj)      ;; already a number
	 obj)
	((not obj)          ;; #f
	 #f)
	((string? obj)
	 ;; ignore commas
	 (set! obj (string-delete obj #\,))

	 ;; ignore leading and trailing whitespace
	 (set! obj (string-trim-both obj))

	 ;; empty, "-" or "N/A" is #f
	 (if (or (string-null? obj)
		 (string=? "-" obj)
		 (string-ci=? "n/a" obj))
	     #f
	     obj))
	(else
	 (error "Bad amount:" obj))))

;; normalize a price in OBJ
;; OBJ is usually a string, but can be a number or #f
;; the transforms of `crunch-number' above are made, plus "0" or "0.0" are
;; considered to be missing values and are returned as #f
;;
(define-public (crunch-price obj)
  (set! obj (crunch-number obj))
  (if (and (string? obj)
	   (0-string? obj))
      #f
      obj))


;; normalize a volume or open interest in OBJ
;; OBJ is usually a string, but can be a number or #f
;; the transforms of `crunch-number' above are made, plus "-" is considered 0
;;
(define-public (crunch-volume obj)
  (if (equal? obj "-")
      0
      (begin
	(set! obj (crunch-number obj))
	(if (and (string? obj)
		 (0-string? obj))
	    #f
	    obj))))

;; return true if STR is a zero like "0", "0.", ".0" or "0.0" (with any
;; number of zeros)
(define-public (0-string? str)
  (regexp-exec #,(regexp "^[+-]?0+(\\.0*)?$|^[+-]?\\.0+$") str))

(define-public (string-comma->dot str)
  (string-map (lambda (c)
		(if (eqv? c #\,) #\. c))
	      str))


;;-----------------------------------------------------------------------------
;; dividend stuff

;; parse a number string like "1.50" into a dividend amount object like
;;     (amount 150 (decimals . 2))
;; or with optional imputation,
;;     (amount 15000 (decimals . 4) (imputation 6428))
;;
;; if STR is not a number, or if imputation is given and it's not a number
;; or #f, then return #f
;;
;; #:imputation is another number string, like "0.6428"
;;
;; #:extra-decimals is added to the decimals value
;;
;; #:decimal-point is what to expect to find in STR and IMPUTATION, eg. ","
;; for European style
;;
(define*-public (string->dividend-amount str #:key
					 imputation
					 (decimal-point ".")
					 (extra-decimals 0))
  (let ((str    (string-change-decimal-point str decimal-point))
	(impstr (and imputation
		     (string-change-decimal-point imputation decimal-point))))
    (and str                   ;; STR must have ok decimal point
	 (or (not imputation)  ;; IMPUTATION if given must have ok decimal
	     impstr)

	 (receive-list (decimals amount imp)
	     (strings->numbers+decimals str impstr)
	   (cons* 'amount amount
		  (cons 'decimals (+ decimals extra-decimals))
		  (if imputation
		      (list (cons 'imputation imp))
		      '()))))))

;; return STR with DECIMAL-POINT changed to a "."
;; if STR isn't a number string like "12" "12x" "x12" "12x34", where x is
;; the given DECIMAL-POINT, then return #f
(define (string-change-decimal-point str decimal-point)
  (set! decimal-point (regexp-quote decimal-point))
  (and-let* ((m (or (string-match (string-append "^([0-9]+)(" decimal-point
						 "([0-9]*))?$") str)
		    (string-match (string-append "^([0-9]*)(" decimal-point
						 "([0-9]+))?$") str))))
    (string-append (match:substring m 1)
		   "."
		   (or (match:substring m 3) ""))))


;;-----------------------------------------------------------------------------
;; language helpers

;; TARGETS is a list of language + value combinations, like
;;
;;     (lang-select '(("en" 123)
;;                    ("ja" 456)))
;;
;; The return is the value from TARGETS corresponding to the preferred
;; language, or if none are in the $LANGUAGE, $LANG etc preferences, then
;; the value in the first entry is returned.
;;
;; An optional #:charset can be given in an entry, to use that entry only if
;; the charset is available from iconv
;;
;;     (lang-select '(("en" "english")
;;                    ("ja" "japanese" #:charset "EUC-JP")))
;;
(define-public (lang-select targets)
  (second
   (or (and-let* (;; priority sequence as per "Using gettextized software" in
		  ;; glibc manual; LANGUAGE is a ":" separated list, the
		  ;; others aren't but `string-split' on them won't hurt
		  (lang-pref-str (or (getenv "LANGUAGE")
				     (getenv "LC_ALL")
				     (getenv "LC_MESSAGES")
				     (getenv "LANG"))))
	 (any (lambda (lang)
		;; strip "_territory.codeset@modifier" parts for comparison
		(set! lang (string-take-while lang char-set:letter))

		(any (lambda (elem)
		       (and (equal? lang (first elem))
			    (let-keywords (cddr elem) #f ((charset #f))
			      (or (not charset)
				  (iconv-charset-available? charset)))
			    elem))
		     targets))

	      (string-split lang-pref-str #\:)))

       (first targets))))


;;-----------------------------------------------------------------------------
;; timezones used in several places

(define-public timezone-chicago
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Chicago")
		      '("America/Chicago"
			;; US rule: second Sun in Mar, to first Sun in Nov
			"EST+6EDT,M3.2.0,M11.1.0")
		      ;; traditional unix was US daylight savings rules
		      "EST+6EDT"))))

(define-public timezone-europe-cet ;; represented by Berlin
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Berlin")
		      '("Europe/Berlin"
			;; EU rule: last Sun in Mar, to last Sun in Oct
			"CET-1CEDT,M3.5.0,M10.5.0/3")
		      "CET-1"))))

(define-public timezone-london
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "London")
		      '("Europe/London"
			;; EU rule: last Sun in Mar, to last Sun in Oct
			"GMT+0BST,M3.5.0/1,M10.5.0")
		      "GMT"))))

(define-public timezone-newyork
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "New York")
		      '("America/New_York"
			;; US rule: second Sun in Mar, to first Sun in Nov
			;; (commencing 2007)
			"EST+5EDT,M3.2.0,M11.1.0")
		      ;; traditional unix was US daylight savings rules
		      "EST+5EDT"))))

(define-public timezone-singapore
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Singapore")
		      '("Asia/Singapore"
			;; no daylight savings
			"SGT-8")
		      "SGT-8"))))

(define-public timezone-sydney
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Sydney")
		      '("Australia/Sydney"
			;; first Sun in Oct, to first Sun in Apr, for 2008 on
			"EST-10EDT,M10.1.0,M4.1.0/3")
		      "EST-10"))))

(define-public timezone-tokyo
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Tokyo")
		      '("Asia/Tokyo"
			;; no daylight savings
			"JST-9")
		      "JST-9"))))


;;-----------------------------------------------------------------------------
;; database downloads - keyword style processing

(define*-public (download-process
		 #:key
		 (module            ;; string or list of strings, for messages
		  (error "download-process: no #:module"))
		 (symbol-list       ;; list of symbol strings
		  (error "download-process: no #:symbol-list"))
		 row-list
		 name          	    ;; string or #f
		 url headers   	    ;; strings or both #f
		 copyright     	    ;; string or #f
		 lo hi         	    ;; tdate or #f
		 currency      	    ;; string or #f
		 expiry-proc        ;; called (PROC symbol), or #f
		 (extra-decimals 0) ;; positive integer
		 last-download      ;; #t or #f
		 (dividends '())    ;; alist (TDATE . ...)
		 (info-alist '())   ;; alist (KEY . VALUE)
		 prefer-decimals    ;; integer or #f
		 )

  ;; flatten to string if a list of strings
  (if (pair? module)
      (set! module (string-join module)))

  (for-each (lambda (row)
	      (let ((tdate (plist-get row #:tdate)))
		(set! lo (min-maybe lo tdate))
		(set! hi (max-maybe hi tdate))))
	    row-list)

  (if (not lo)
      (begin
	(set! lo (tdate-today))
	(set! hi lo)))

  (for-each
   (lambda (symbol)
     (define this-extra-decimals extra-decimals)
     (download-status module (_ "processing")
		      (tdate-range-string lo hi)
		      symbol)
     ;; pause between writes
     (c-idle)

     (database-call-with-lock data-directory
       (lambda ()
	 ;; array 0,1,2,3,4,5 = OHLCVI
	 ;;       6 = mdate for that data
	 ;;       7 = close of prev contract
	 ;;       8 = mdate of prev contract
	 (let* ((series         (or (database-read-series symbol)
				    (series-new #:symbol symbol
						#:array (make-array #f (list lo lo) '(0 4)))))
		(covered-lo     lo)
		(covered-hi     hi)
		(lo             (min lo (series-lo series)))
		(hi             (max hi (series-hi series)))
		(want-commodity (chart-symbol-commodity symbol))
		(want-mdate     (chart-symbol-mdate symbol))
		(array          (make-array #f (list lo hi) '(0 8)))
		(this-currency  currency)
		(this-expiry    (and want-mdate   ;; specifics only
				     expiry-proc
				     (expiry-proc symbol)))
		(notes-write?   #f))

	   (define (second-last tdate mdate close)
	     (if (and mdate
		      (or (not (array-ref array tdate 8))
			  (< mdate (array-ref array tdate 8))))
		 (begin
		   (set! close (crunch-price close))
		   (array-set! array close tdate 7)
		   (array-set! array mdate tdate 8))))

	   (array-copy! (series-array series
				      (series-lo series)
				      (series-hi series))
			array)
	   (if name
	       (series-name! series name))

	   ;; fill existing mdates
	   (let ((months (filter (lambda (elem)
				   (and (list>=? elem 2)
					(eq? 'month (second elem))))
				 (series-dividends series)))
		 (pos    (series-lo series)))
	     (for-each (lambda (elem)
			 (set! pos (car elem))
			 (array-fill! (make-shared-array
				       array
				       (lambda (i) (list i 6))
				       (list pos (series-hi series)))
				      (third elem)))
		       (sort months car<)))

	   (and-let* ((val   (series-info-get series 'second-last))
		      (tdate (first val)))
	     (array-set! array (second val) tdate 7)
	     (array-set! array (third val)  tdate 8))

	   (series-dividends! series (append! (series-dividends series)
					      dividends))

	   (for-each (lambda (elem)
		       (series-info-add! series (car elem) (cdr elem)))
		     info-alist)

	   (for-each
	    (lambda (row)
	      (if (equal? want-commodity (plist-get row #:commodity))
		  (let-keywords row #f
				((commodity      (error "download-process: missing commodity" row))
				 (name           #f)
				 (mdate          #f)
				 (tdate          #f)
				 (expiry-tdate   this-expiry)
				 (open           #f)
				 (high           #f)
				 (low            #f)
				 (close          #f)
				 (sessions       #f)
				 (volume         #f)
				 (open-interest  #f)
				 (currency       #f)
				 (extra-decimals #f))

		    (and
		     ;; for specific month, must match mdate
		     (or (not want-mdate)
			 (equal? mdate want-mdate))

		     ;; for front month, ignore anything past expiry
		     ;; (but include it for specific)
		     (or want-mdate
			 (not expiry-tdate)
			 (<= tdate expiry-tdate))

		     ;; for front month, only take earliest mdate
		     (or want-mdate
			 (if (or (not (array-ref array tdate 3)) ;; close
				 (not (array-ref array tdate 6))
				 (not mdate)
				 (<= mdate (array-ref array tdate 6)))
			     (begin
			       (if (and (array-ref array tdate 6)
					mdate
					(< mdate (array-ref array tdate 6)))
				   (second-last tdate
						(array-ref array tdate 6)
						(array-ref array tdate 3)))
			       #t)
			     (begin
			       (if sessions
				   (begin
				     (set! sessions (filter-map crunch-price
								sessions))
				     (if (not (null? sessions))
					 (set! close (last sessions)))))
			       (second-last tdate mdate close)
			       #f)))

		     (begin
		       (if name
			   (series-name! series name))
		       (if currency
			   (set! this-currency currency))
		       (if expiry-tdate
			   (set! this-expiry expiry-tdate))
		       (if extra-decimals
			   (set! this-extra-decimals extra-decimals))

		       (if sessions
			   (begin
			     (set! sessions (filter-map crunch-price sessions))
			     (if (not (null? sessions))
				 (begin
				   (set! open  (first sessions))
				   (set! close (last sessions))
				   (set! high  (find-maximum string->number-err
							     sessions))
				   (set! low   (find-minimum string->number-err
							     sessions)))))
			   (begin
			     (set! open  (crunch-price open))
			     (set! high  (crunch-price high))
			     (set! low   (crunch-price low))
			     (set! close (crunch-price close))))

		       ;; don't overwrite existing data, so can process
		       ;; prices separate from volumes, etc
		       (if open
			   (array-set! array open    tdate 0))
		       (if high
			   (array-set! array high    tdate 1))
		       (if low
			   (array-set! array low     tdate 2))
		       (if close
			   (begin
			     (array-set! array close tdate 3)
			     (array-set! array mdate tdate 6)))

		       (if volume
			   (array-set! array volume  tdate 4))
		       (if open-interest
			   (array-set! array open-interest tdate 5)))))))
	    row-list)

	   (let ((conv-decimals (- (series-decimals series)
				   this-extra-decimals))
		 (new-decimals  #f)
		 (openint?      #f))

	     ;; find how many decimals on new data
	     (do ((tdate lo (1+ tdate)))
		 ((> tdate hi))
	       (set! conv-decimals
		     (max-maybe conv-decimals
				(string-decimals-maybe
				 (array-ref array tdate 0) ;; open
				 prefer-decimals)
				(string-decimals-maybe
				 (array-ref array tdate 1) ;; high
				 prefer-decimals)
				(string-decimals-maybe
				 (array-ref array tdate 2) ;; low
				 prefer-decimals)
				(string-decimals-maybe
				 (array-ref array tdate 3) ;; close
				 prefer-decimals)
				(string-decimals-maybe
				 (array-ref array tdate 7)
				 prefer-decimals))))       ;; second close
	     (set! new-decimals (+ conv-decimals this-extra-decimals))

	     ;; possibly increase decimals in existing values
	     (if (> new-decimals (series-decimals series))
		 (let ((factor (integer-expt 10 (- new-decimals
						   (series-decimals series)))))
		   (do ((tdate lo (1+ tdate)))
		       ((> tdate hi))
		     (if (number? (array-ref array tdate 0))  ;; open
			 (array-set! array (* factor (array-ref array tdate 0))
				     tdate 0))
		     (if (number? (array-ref array tdate 1))  ;; high
			 (array-set! array (* factor (array-ref array tdate 1))
				     tdate 1))
		     (if (number? (array-ref array tdate 2))  ;; low
			 (array-set! array (* factor (array-ref array tdate 2))
				     tdate 2))
		     (if (number? (array-ref array tdate 3))  ;; close
			 (array-set! array (* factor (array-ref array tdate 3))
				     tdate 3))
		     (if (number? (array-ref array tdate 7))  ;; second last
			 (array-set! array (* factor (array-ref array tdate 7))
				     tdate 7)))

		   (series-notes! series (notes-change-decimals
					  (series-notes series)
					  (series-decimals series)
					  new-decimals))
		   (set! notes-write? #t)

		   (series-decimals! series new-decimals)))

	     ;; convert strings to numbers
	     (do ((tdate lo (1+ tdate)))
		 ((> tdate hi))
	       (if (string? (array-ref array tdate 0)) ;; open
		   (array-set! array (string->number+decimals
				      (array-ref array tdate 0) conv-decimals)
			       tdate 0))
	       (if (string? (array-ref array tdate 1)) ;; high
		   (array-set! array (string->number+decimals
				      (array-ref array tdate 1) conv-decimals)
			       tdate 1))
	       (if (string? (array-ref array tdate 2)) ;; low
		   (array-set! array (string->number+decimals
				      (array-ref array tdate 2) conv-decimals)
			       tdate 2))
	       (if (string? (array-ref array tdate 3)) ;; close
		   (array-set! array (string->number+decimals
				      (array-ref array tdate 3) conv-decimals)
			       tdate 3))
	       (if (string? (array-ref array tdate 7)) ;; second last
		   (array-set! array (string->number+decimals
				      (array-ref array tdate 7) conv-decimals)
			       tdate 7))

	       ;; check for high/low right way around
	       (if (and (array-ref array tdate 1) ;; high
			(array-ref array tdate 2) ;; low
			(< (array-ref array tdate 1)
			   (array-ref array tdate 2)))
		   (download-message
		    (simple-format #f "~a: ~a high < low: ~a ~a ~a"
				   module symbol
				   (strftime-utf8 (locale-d-fmt)
						  (tdate->tm tdate))
				   (array-ref array tdate 1)
				   (array-ref array tdate 2))))

	       (let ((volume (array-ref array tdate 4)))
		 (set! volume (crunch-volume volume))
		 (if (string? volume)
		     (set! volume (string->number-err volume)))
		 (array-set! array volume tdate 4))

	       (let ((openint (array-ref array tdate 5)))
		 (set! openint (crunch-volume openint))
		 (if (string? openint)
		     (set! openint (string->number-err openint)))
		 (array-set! array openint tdate 5)
		 (if openint
		     (set! openint? #t))))

	     (series-array! series (make-shared-array-columns
				    array 0 (if openint? 5 4)))
	     (series-lo! series lo)
	     (series-hi! series hi)
	     (series-require! series noop))

	   ;; dividend entries for mdate rollovers
	   (if (not want-mdate)
	       (let ((prev-mdate #f)
		     (tlast      #f))

		 (do ((tdate (1- lo) (1- tdate)))
		     ((or tlast
			  (< tdate hi)))
		   (if (array-ref array tdate 3)
		       (set! tlast tdate)))

		 (do ((tdate lo (1+ tdate)))
		     ((> tdate hi))
		   (if (and (array-ref array tdate 6)
			    (not (eqv? prev-mdate (array-ref array tdate 6))))
		       (begin
			 (series-month-rollover!
			  series tdate (array-ref array tdate 6)
			  (and tlast (array-ref array tlast 3))
			  (and tlast (array-ref array tlast 7)))
			 (set! prev-mdate (array-ref array tdate 6))))
		   (if (array-ref array tdate 3)
		       (set! tlast tdate)))))

	   ;; note second-last at highest tdate
	   (do ((tdate hi (1- tdate)))
	       ((< tdate lo))
	     (if (array-ref array tdate 3)
		 (begin
		   (if (array-ref array tdate 8)
		       (series-info-add! series 'second-last
					 (list tdate
					       (array-ref array tdate 7)
					       (array-ref array tdate 8)))
		       (series-info-del! series 'second-last))
		   (set! tdate (1- lo)))))

	   (if copyright
	       (series-info-add! series 'copyright copyright))

	   (if this-currency
	       (series-info-add! series 'currency this-currency))

	   (if url
	       (download-series-tag-info! series url headers))

	   (if (and want-mdate
		    this-expiry)
	       (series-info-add! series 'expiry-date this-expiry))

	   ;; all symbols mentioned in the file have the full date range
	   ;; covered, even if some have no data at the very start or very
	   ;; end
	   (series-info-add!
	    series 'download-done-start
	    (min-maybe covered-lo
		       (series-info-get series 'download-done-start)))

	   (let ((done (max-maybe covered-hi
				  (series-info-get series
						   'download-done))))
	     ;; suppress info tag if got data to that point anyway, mainly
	     ;; for ease of manual editing of database files
	     (if (and (within-range? lo hi done)
		      (or (array-ref array done 0)   ;; open
			  (array-ref array done 1)   ;; high
			  (array-ref array done 2)   ;; low
			  (array-ref array done 3)   ;; close
			  (array-ref array done 4)   ;; volume
			  (array-ref array done 5))) ;; openint
		 (series-info-del! series 'download-done)
		 (series-info-add! series 'download-done done)))

	   (if last-download
	       (series-info-add! series 'last-download
				 (tdate-today (symbol-timezone symbol))))

	   (database-write-series series)
	   (if notes-write?
	       (database-write-notes series))))))

   symbol-list))


(define (string-decimals-maybe obj prefer)
  (and (string? obj)
       (let ((pos (string-index obj #\.)))
	 (if (not pos)
	     0
	     (let ((decimals (- (string-length obj) pos 1)))
	       (if (or (not prefer)
		       (<= decimals prefer))
		   decimals
		   (max prefer
			(- (string-skip-right obj #\0) pos))))))))


(define (series-month-rollover! series tdate new-mdate
				close second-close)
  (let ((old-step (and-let* ((val (assv-ref (series-dividends series)
					    tdate))
			     (    (eqv? new-mdate (second val))))
		    (cddr val))))

    ;; delete possible NEW-MDATE at the start of data (being replaced by new
    ;; earlier data)
    (let ((val (list 'month new-mdate)))
      (series-dividends! series (remove! (lambda (elem)
					   (equal? (cdr elem) val))
					 (series-dividends series))))

    (series-dividend-add!
     series tdate
     (cons* 'month new-mdate
	    (or (and close second-close
		     (list (number+decimals->string (- second-close close)
						    (series-decimals series))))
		old-step
		'())))))

;; return string for N, shifted down by DECIMALS
;; for example if N is 123 and DECIMALS is 2 it means "1.23"
(define-public (number+decimals->string n decimals)
  (if (zero? decimals)
      (number->string n)
      (let ((str (number->string (abs n))))
	(begin
	  (set! str (string-pad str (max (1+ decimals) (string-length str))
				#\0))
	  (string-append (if (negative? n) "-" "")
			 (string-drop-right str decimals)
			 "."
			 (string-take-right str decimals))))))

