;; Detrended Price Oscillator (DPO).

;; Copyright 2003, 2004, 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart dpo)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart sma))


;; http://www.incrediblecharts.com/technical/detrended_price_oscillator.htm
;; http://www.incrediblecharts.com/technical/dpo_construction.htm
;;     Formula, usage guidelines.
;;
;; http://www.marketscreen.com/help/AtoZ/default.asp?hideHF=&Num=41
;;     Formula and description.
;;


;; `drop-right!' but limiting N to the LST length
(define (drop-right-maybe! lst n)
  (if (length>=? lst n)
      (drop-right! lst n)
      '()))

(define (dpo-delta count)
  (1+ (ash count -1)))

(define (dpo-fill series new-array lo hi)
  (let* ((count    (series-data series))
	 (delta    (dpo-delta count))
	 (parent   (series-parent series))
	 (array    (series-array-closes parent (- lo delta) hi))
	 (lst      (series-prev-n-closes parent
					 #:before (- lo delta)
					 #:count  count))
	 (val      (if (null? lst) #f (last lst))) ;; last prev val
	 (sma-proc (sma-calc-proc count (drop-right-maybe! lst 1)))
	 (sma      (and val (sma-proc val))))  ;; with last prev val

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((prev (array-ref array (- i delta))))
	(set! sma (sma-proc prev)))

      (and-let* ((this (array-ref array i))
		 sma)
	(array-set! new-array (- this sma) i)))))

(define-public (dpo-series series count)
  (let ((new-series (series-copy series))
	(delta      (dpo-delta count)))
    (series-name!      new-series (format #f "~a ~a" (_ "DPO") count))
    (series-datatype!  new-series dpo-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   dpo-fill (+ delta (1- count))))
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    ;; DELTA fewer elements
    (let ((lo (+ (series-lo series) delta))
	  (hi (series-hi series)))
      (series-lo! new-series lo)

      (if (> lo hi)
	  (begin  ;; less than DELTA points in parent series
	    (series-lo! new-series hi)
	    (series-array! new-series (make-array #f (list hi hi))))))

    new-series))

(define-public dpo-datatype
  (datatype-new!
   #:key       'dpo
   #:name      (_ "DPO - Detrended Price Oscillator")
   #:units     'price
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Detrended Price Oscillator")
   #:proc      dpo-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'dpo-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
