;; Ease of movement indicator.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ease)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.incrediblecharts.com/technical/ease_of_movement.htm
;; http://www.incrediblecharts.com/technical/emv_construction.htm
;;
;; http://www.paritech.com.au/education/technical/indicators/strength/ease.asp
;;
;; http://www.linnsoft.com/tour/techind/arms.htm
;;     Sample Apple (AAPL) from April 2000, but using SMA instead of EMA.
;;
;; http://kb.worden.com/default.asp?id=49&Lang=1&SID=
;;     Sample Amazon (AMZN) from 2006, 14-day, maybe SMA instead of EMA.
;;


;; Return a procedure which calculates the ease of movement index, with an
;; exponential moving average of COUNT days.
;;
;; Each call (PROC high low volume) enters a new point into the window, and
;; the return is the EOM up to (and including) that point, or #f if not
;; enough values yet.
;;
;; To prime initiall, call PROC with (1+ (ema-warmup-count count)) many
;; points before the first desired, to get that next call based on a full
;; window.
;;
(define-public (ease-calc-proc count)
  (let ((prev-mid #f)
	(ema-proc (ema-calc-proc count)))

    (lambda (high low volume)
      (let* ((mid  (/ (+ high low) 2))
	     (ease (and prev-mid
			(/ (* (- mid prev-mid) (- high low))
			   (* 1e-6 volume)))))
	(set! prev-mid mid)
	(and ease
	     (ema-proc ease))))))


(define (ease-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (warmup (1+ (ema-warmup-count count)))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo
					   #:count  warmup))
	 (array  (series-array parent start hi))
	 (ease   (ease-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(ease (or (array-ref array i 1) close) ;; high
	      (or (array-ref array i 2) close) ;; low
	      (array-ref array i 4))))         ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))

	(array-set! new-array (ease (or (array-ref array i 1) close) ;; high
				    (or (array-ref array i 2) close) ;; low
				    (array-ref array i 4))           ;; volume
		    i)))))

(define-public (ease-series series count)
  (ohlcv-required "Ease of Movement" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Ease") count))
    (series-datatype!  new-series ease-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
		       		  ease-fill (1+ (ema-warmup-count count))))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define ease-datatype
  (datatype-new!
   #:key    'ease
   #:name   (_ "Ease of Movement")
   #:input  datatype-ohlcv
   #:units  'ease
   #:hlines '(0)
   #:manual (_ "CONTEXT(manual-node): Ease of Movement")
   #:proc   ease-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'ease-days
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
