;; EMA of EMA.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ema-2)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))



;; Return a procedure which calculates an EMA of EMA with given COUNT period
;; smoothing (on both).
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the EMAofEMA up to (and including) that value, or #f if not enough
;; points yet.
;;
;; An EMA of EMA is in theory influenced by all preceding data, but
;; `ema-2-warmup-count' below is designed to determine a warmup count.
;;
;;
;; An implementation in terms of `ema-calc-proc' would be as follows,
;;
;;     (define-public (ema-2-calc-proc count)
;;       (define ema-A-proc (ema-calc-proc count))
;;       (define ema-B-proc (ema-calc-proc count))
;;       (lambda (x)
;;         (and-let* ((eA (ema-A-proc x)))
;;           (ema-B-proc eA))))
;;
;; but this behaves poorly near the start of data, since the "A" proc
;; doesn't start giving out values until 0.75 weighting and only then does
;; the "B" proc start accumulating.  With the combined function the weight
;; of a partial value can be carried through.
;;
(define-public (ema-2-calc-proc count)
  (let* ((alpha (ema-count->alpha count))
	 (f     (ema-count->f count))
	 (a-sum 0.0)
	 (a-div 0.0)
	 (b-sum 0.0)
	 (b-div 0.0))

    (lambda (x)
      ;; reduce past data
      (set! a-sum (* a-sum f))
      (set! a-div (* a-div f))
      ;; add new data
      (set! a-sum (+ a-sum (* x alpha)))
      (set! a-div (+ a-div alpha))

      (set! b-sum (* b-sum f))
      (set! b-div (* b-div f))
      (set! b-sum (+ b-sum (* a-sum alpha)))
      (set! b-div (+ b-div (* a-div alpha)))

      ;; average, provided enough total weight
      (and (>= b-div indicator-minimum-fraction)
	   (/ b-sum b-div)))))

;; By priming an EMA-2 accumulator PROC above with (ema-2-warmup-count COUNT)
;; many values, the next call will have an omitted weight of no more than
;; 0.1% of the total.  Omitting 0.1% should be negligable, unless past
;; values are ridiculously bigger than recent ones.
;;
;; The implementation here does a binary search for the first i satisfying
;; R(i)<=0.001, so it's not very fast.  Perhaps there'd be a direct
;; closed-form solution to the equation R(i)=0.001 below, but i*f^i doesn't
;; look like it can rearrange.
;;
(define-public (ema-2-warmup-count count)
  (define f (ema-count->f count))
  (bsearch-first-true (lambda (i)
			(<= (ema-2-omitted f i) 0.001))
		      count))

;; `ema-2-omitted' returns the fraction (between 0 and 1) of weight omitted
;; by stopping an EMA of EMA at the f^k term, which means the first k+1
;; terms.
;;
;; The total weight up to that term is,
;;
;;    W(k) = (1-f)^2 * ( 1 + 2f + 3f^2 + 4f^3 + ... + (k+1)*f^k )
;;
;; Multiplying (1-f)^2 through leads to the middle terms cancelling, leaving
;; just the 1 at the start, and two terms at the end,
;;
;;    W(k) = 1 + (-2*(k+1) + k) * f^(k+1)
;;             + (k+1)          * f^(k+2)
;;
;; The omitted part is 1 - W(k), so the "1 +" is dropped and the rest
;; negated.  The terms simplify to
;;
;;    R(k) = f^(k+1) * (k+2 - f * (k+1))
;;
;; See devel/closed-form.scm for automated checking of this calculation.
;;
(define-public (ema-2-omitted f k)
  (* (expt f (1+ k))
     (- k -2
	(* f (1+ k)))))


(define (ema-2-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (ema-2-calc-proc count)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (proc this) i))))))

(define-public (ema-2-series series count)
  (let ((new-series (series-copy series))
	(warmup     (ema-2-warmup-count count)))
    (series-name!     new-series (format #f "~a ~a" (_ "EMAofEMA") count))
    (series-datatype! new-series ema-2-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count warmup))
    (series-require!  new-series (indicator-require-parts ema-2-fill warmup))
    (series-array!    new-series #f)

    new-series))


(define ema-2-datatype
  (datatype-new!
   #:key      'ema-2
   #:name     (_ "EMA of EMA")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): EMA of EMA")
   #:proc     ema-2-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'ema-2-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
