;; EMA of EMA of #MA.

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ema-3)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))



;; Return a procedure which calculates an EMA of EMA of EMA with given COUNT
;; period smoothing (on each).
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the EMAofEMAofEMA up to (and including) that value, or #f if not
;; enough points yet.
;;
;; An EMAofEMAofEMA is in theory influenced by all preceding data, but
;; `ema-3-warmup-count' below is designed to determine a warmup count.
;;
;;
;; An implementation in terms of `ema-calc-proc' would be as follows,
;;
;;     (define-public (ema-3-calc-proc count)
;;       (define ema-A-proc (ema-calc-proc count))
;;       (define ema-B-proc (ema-calc-proc count))
;;       (define ema-C-proc (ema-calc-proc count))
;;       (lambda (x)
;;         (and-let* ((eA (ema-A-proc x))
;;                    (eB (ema-B-proc x)))
;;           (ema-C-proc eB))))
;;
;; but this behaves very badly near the start of data, since the "A" proc
;; doesn't start giving out values until 0.75 weighting and only then does
;; the "B" proc start accumulating, and then the "C" proc has to wait for
;; it.  With the combined function the weight of a partial value can be
;; carried through.
;;
(define-public (ema-3-calc-proc count)
  (let* ((f     (ema-count->f count))
	 (alpha (ema-count->alpha count))

	 (a-sum 0.0)
	 (a-div 0.0)
	 (b-sum 0.0)
	 (b-div 0.0)
	 (c-sum 0.0)
	 (c-div 0.0))

    (lambda (x)
      ;; reduce past data
      (set! a-sum (* a-sum f))
      (set! a-div (* a-div f))
      ;; add new data
      (set! a-sum (+ a-sum (* x alpha)))
      (set! a-div (+ a-div alpha))

      (set! b-sum (* b-sum f))
      (set! b-div (* b-div f))
      (set! b-sum (+ b-sum (* a-sum alpha)))
      (set! b-div (+ b-div (* a-div alpha)))

      (set! c-sum (* c-sum f))
      (set! c-div (* c-div f))
      (set! c-sum (+ c-sum (* b-sum alpha)))
      (set! c-div (+ c-div (* b-div alpha)))

      ;; average, provided enough total weight
      (and (>= c-div indicator-minimum-fraction)
	   (/ c-sum c-div)))))


;; `ema-3-omitted' returns the fraction, between 0 and 1, of weight omitted
;; by stopping an EMA of EMA of EMA at the f^k term, which means taking the
;; first k+1 terms.
;;
;; The total weight up to that term is,
;;
;;    W(k) = (1-f)^3 * ( 1 + 3f + 6f^2 + 10f^3 + ... + T(k+1)*f^k )
;;
;; where T(k)=k*(k+1)/2 is a triangle number.  Multiplying the (1-f)^3
;; through leads to the middle terms cancelling, because
;;
;;    T(k+1) - 3*T(k) + 3*T(k-1) - T(k-2) = 0
;;
;; Which leaves just three at the end,
;;
;;    W(k) = 1 + (-3*T(k+1) + 3*T(k)   - T(k-1)) * f^(k+1)
;;             + (            3*T(k+1) - T(k)  ) * f^(k+2)
;;             + (                     - T(k+1)) * f^(k+3)
;;
;; The omitted part is 1 - W(k), so the "1 +" is dropped and the rest
;; negated.  The triangle number terms simplify to quadratics in i,
;;
;;    R(k) = f^(k+1) * 1/2 * (               k^2 + 5*k + 6
;;                            + f * (     -2*k^2 - 8*k - 6
;;                                   + f * (k+1)*(k+2)))
;;
;; See devel/closed-form.scm for automated checking of this calculation.
;;
(define-public (ema-3-omitted f k)
  (* (expt f (1+ k))
     0.5
     (+ 6 (* k (+ 5 k))  ;; k^2 + 5*k + 6
	(* f
	   (+ -6 (* k (+ -8 (* k -2))) ;; -2*k^2 - 8*k - 6
	      (* f
		 (1+ k) (+ 2 k)))))))  ;; (k+1)*(k+2)

;; By priming an EMA-3 accumulator PROC above with (ema-3-warmup-count
;; COUNT) many values, the next call will have an omitted weight of no more
;; than 0.1% of the total.  Omitting 0.1% should be negligable, unless past
;; values are ridiculously bigger than recent ones.
;;
;; The implementation here does a binary search for the first i satisfying
;; R(i)<=0.001, so it's not very fast.  Perhaps there'd be a direct
;; closed-form solution to that equation, but f^i*(quadratic in i) doesn't
;; look like it can rearrange.
;;
(define-public (ema-3-warmup-count count)
  (ema-3-omitted-search count 0.001))

(define-public (ema-3-omitted-search count target)
  (define f (ema-count->f count))
  (bsearch-first-true (lambda (i)
			(<= (ema-3-omitted f i) target))
		      count))



(define (ema-3-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (ema-3-calc-proc count)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (proc this) i))))))

(define-public (ema-3-series series count)
  (let ((new-series (series-copy series))
	(warmup     (ema-3-warmup-count count)))
    (series-name!     new-series (format #f "~a ~a" (_ "EMAofEMAofEMA") count))
    (series-datatype! new-series ema-3-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count warmup))
    (series-require!  new-series (indicator-require-parts ema-3-fill warmup))
    (series-array!    new-series #f)

    new-series))


(define ema-3-datatype
  (datatype-new!
   #:key      'ema-3
   #:name     (_ "EMA of EMA of EMA")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): EMA of EMA of EMA")
   #:proc     ema-3-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'ema-3-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
