;; Exponential moving average (EMA).

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ema)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; In the manual it's noted that the first n days weight make up 86.5% of
;; the total weight in an EMA.  That amount is x = 1 + f + f^2 + ... +
;; f^(n-1), and for total weight t
;;
;;     t = x + f^n*(1 + f + f^2 + ...)
;;
;;     t = x + f^n*t
;;
;; so the fraction of the total is
;;
;;     x/t = 1 - f^n
;;
;;               /      2  \ n
;;         = 1 - | 1 - --- |
;;               \     n+1 /
;;
;;               /     -2  \ n+1
;;               | 1 + --- |
;;               \     n+1 /
;;         = 1 - -----------
;;               /      2  \
;;               | 1 - --- |
;;               \     n+1 /
;;
;; As n increases, the numerator approaches e^-2 from the limit (1+x/n)^n
;; --> e^x by Euler, and the numerator approaches 1.  So the result is
;;
;;                  1
;;     x/t --> 1 - ---  = 0.8646647...
;;                 e^2
;;


;; f=1-2/(N+1), rearranged to f=(N-1)/(N+1).
(define-public (ema-count->f count)
  (exact->inexact (/ (1- count) (1+ count))))

;; convert a COUNT in J. Welles Wilder's reckoning to one in the standard form
;; Wilder alpha=1/W, alpha=2/(N+1), so N=2*W-1
(define-public (ema-count-wilder count)
  (1- (* 2 count)))

;; convert a COUNT in standard form to  J. Welles Wilder's reckoning
;; W = (N+1)/2
(define-public (ema-count-to-wilder count)
  (* 0.5 (1+ count)))

;; N = 2/alpha - 1
(define-public (ema-alpha->count alpha)
  (1- (/ 2.0 alpha)))

;; alpha=2/(N+1)
(define-public (ema-count->alpha count)
  (/ 2.0 (1+ count)))


;; Return a procedure which calculates an exponential moving average over an
;; accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the EMA up to (and including) that value, or #f if not enough points
;; yet.
;;
;; An EMA is in theory influenced by all preceding data, but
;; `ema-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (ema-warmup-count COUNT) many values, the next call
;; will have an omitted weight of no more than 0.1% of the total.  Omitting
;; 0.1% should be negligable, unless past values are ridiculously bigger
;; than recent ones.
;;
(define-public (ema-calc-proc count)

  (if (<= count 1)
      noop

      ;; SUM is x0 + x1*f + x2*f^2 + x3*f^3 + ... + xk*f^k, for as many x's as
      ;; so far entered
      ;;
      ;; DIV is the corresponding 1 + f + f^2 + ... + f^k.  This approaches
      ;; 1/(1-f), but on the first few outputs it's as little as 75% of that
      ;; total, so we need to calculate it explicitly.
      ;;
      (let* ((f       (ema-count->f count))
	     (min-div (* indicator-minimum-fraction (/ (- 1.0 f))))
	     (sum     0.0)
	     (div     0.0))

	(lambda (x)
	  (set! sum (+ x   (* sum f)))
	  (set! div (+ 1.0 (* div f)))

	  (and (>= div min-div)
	       (/ sum div))))))


;; By priming an EMA accumulator PROC above with (ema-warmup-count COUNT)
;; many values, the next call will have an omitted weight of no more than
;; 0.1% of the total.  Omitting 0.1% should be negligable, unless past
;; values are ridiculously bigger than recent ones.
;;
;; The implementation is quite fast, per the `ema-omitted-search' below.
;;
;; Knowing that log(f) approaches -2/count as count increases, the result
;; from `ema-omitted-search' is roughly log(0.001)/(-2/count) = 3.45*count.
;;
(define-public (ema-warmup-count count)
  (if (<= count 1)
      0
      (1- (ema-omitted-search (ema-count->f count) 0.001))))


;; `ema-omitted' returns the fraction (between 0 and 1) of weight omitted by
;; stopping an EMA at the f^k term, which means the first k+1 terms.
;;
;; The weight, out of a total 1, in those first terms
;;
;;     W(k) = (1-f) (1 + f + f^2 + ... + f^k)
;;
;; multiplying through makes the middle terms cancel, leaving
;;
;;     W(k) = 1 - f^(k+1)
;;
;; The omitted part is then O = 1-W,
;;
;;     Omitted(k) = f^(k+1)
;;
(define-public (ema-omitted f k)
  (expt f (1+ k)))


;; `ema-omitted-search' returns the number of terms t needed in an EMA to
;; have an omitted part <= TARGET, where target is a proportion between 0
;; and 1.  This means
;;
;;     Omitted(t-1) <= target
;;     f^t <= target
;;     t >= log(target) / log(f)
;;
;; Can have f==0 when count==1 (a degenerate EMA, which just follows the
;; given points exactly).  log(0) isn't supported on guile 1.6, hence the
;; special case.
;;
;; Actually log(f) approaches -2/N as N increases, but it's easy enough to
;; do the calculation exactly.
;;
(define-public (ema-omitted-search f target)
  (if (zero? f)
      0
      (ceiling-exact (/ (log target) (log f)))))



(define (ema-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (ema    (ema-calc-proc count)))

    ;; warmup
    (for-each ema (series-prev-n-closes parent
					#:before lo
					#:count (ema-warmup-count count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	(array-set! new-array (ema (array-ref array i))
		    i)))))

(define-public (ema-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "EMA") count))
    (series-datatype! new-series ema-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-array!    new-series #f)
    (series-require!  new-series (indicator-require-parts
				  ema-fill (ema-warmup-count count)))

    new-series))


(define ema-datatype
  (datatype-new!
   #:key      'ema
   #:name     (_ "EMA - Exponential MA")
   #:priority 10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Exponential Moving Average")
   #:proc     ema-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'ema-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))



;;-----------------------------------------------------------------------------
;; adaptive EMA

;; Return a procedure which calculates an adaptive exponential moving
;; average.
;;
;; Each call (PROC val alpha) enters a new value into the calculation, with
;; ALPHA as the smoothing factor to apply at that point.  ALPHA should range
;; from 0 for no weight on VAL, up to 1 for VAL itself to become the new
;; average.
;;
;; The return from each PROC call is the new calculated average, or #f if
;; not enough values yet (the usual `indicator-minimum-fraction').
;;
(define-public (adaptive-ema-calc-proc)
  (define avg    0.0)
  (define weight 0.0)

  (lambda (x alpha)
    (set! avg    (+ avg    (* alpha (- x avg))))
    (set! weight (+ weight (* alpha (- 1 weight))))
    (and (>= weight indicator-minimum-fraction)
	 (/ avg weight))))

;; Look at SERIES from LO downwards and return a starting position for an
;; adaptive moving average which calculates alpha values using PROC.
;;
;; PROC is applied backwards, ie. starting from LO, so this is only really
;; suitable for algorithms that are essentially the same going backwards.
;;
(define-public (adaptive-ema-find-start series lo proc)
  (define arg (series-closes-apply-arg series))
  (define slo (series-lo series))
  (define prod 1.0)
  (let more ((lo lo))
    (cond ((<= lo slo)
	   slo)
	  ((and-let* ((close (apply array-ref (series-array series lo lo)
				    lo arg))
		      (alpha (proc close)))
	     (set! prod (* prod (- 1 alpha)))
	     (< prod 0.001))  ;; 0.1% rule per ema.scm
	   lo)
	  (else
	   (more (1- lo))))))
