;; Emacs interaction, supporting chart.el.

;; Copyright 2005, 2006, 2007, 2008, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart emacs)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 rw)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart database)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart symlists))


;; This code talks to emacs in utf-8, the same as the GUI uses.  This is
;; easiest on the whole, currently.  Some Japanese characters are missing
;; from utf-8 in Emacs 21, but tThe mule-ucs package can provide them and
;; you probably want that anyway if you're in an affected locale.
;;


;; version protocol
;;   51       11      same
;;   52       11      same
;;   53       11      same
;;
(define protocol-version 11)


;; return the utf8 string STR truncated or padded (with spaces) on the right
;; to be TARGET many characters (utf8 characters)
(define (utf8-string-pad-right str target)
  (let ((len (utf8-string-length str)))
    (if (> len target)
	(utf8-string-take str target)  ;; truncate
	(string-append str (make-string (- target len) #\space))))) ;; extend

(define (layout width-list str-list)
  (let ((target-right 0)
	(minimum-pad  0))
    (let more ((width-list width-list)
	       (str-list   str-list))
      (if (not (null? width-list))
	  (let ((width (first width-list))
		(str   (first str-list)))
	    (set! target-right (+ target-right (abs width)))
	    ;; track column but don't actually output if only whitespace
	    (if (string-every char-whitespace? str)
		(if (not (string-null? str))
		    (set! minimum-pad 1))
		(begin
		  ;; pad on the left for positives
		  (if (positive? width)
		      (let ((pad (max minimum-pad
				      (- target-right
					 (port-column (current-output-port))
					 (utf8-string-length str)))))
			(display (make-string pad #\space))))
		  (display str)
		  (set! minimum-pad 0)))
	    (more (cdr width-list) (cdr str-list)))))))

;; return string for LATEST
(define-public (emacs-latest-line latest)
  (with-output-to-string
    (lambda ()
      (let ((note (latest-form-note latest)))
	(layout `(9  ;; symbol
		  7  ;; bid
		  1  ;; slash
		  -7 ;; offer
		  1
		  7  ;; last
		  1
		  6  ;; change
		  1
		  7  ;; low
		  1
		  7  ;; high
		  1
		  8  ;; volume
		  1
		  7  ;; date/time
		  1
		  ;; the full width of whatever's required
		  ,(utf8-string-length note))
		(list
		 ;; truncate symbol
		 (utf8-string-pad-right (latest-symbol latest) 9)
		 (latest-form-price latest (latest-bid latest))
		 (latest-form-quote-slash latest)
		 (latest-form-price latest (latest-offer latest))
		 " "
		 (latest-form-price latest (latest-last latest))
		 " "
		 (latest-form-price latest (latest-change latest))
		 " "
		 (latest-form-price latest (latest-low latest))
		 " "
		 (latest-form-price latest (latest-high latest))
		 " "
		 (latest-form-volume-km latest)
		 " "
		 (latest-form-date/time latest)
		 " "
		 note))))))

;; like `split-at' but restrict N to length of LST
(define (split-at-maybe lst n)
  (if (length>=? lst n)
      (split-at lst n)
      (values lst '())))

(define (seconds->hms time)
  (list (quotient time 3600)
	(quotient (remainder time 3600) 60)
	(remainder time 60)))

(define (full-adate-time adate time)
  (let ((tm (adate->tm adate)))
    (if time
	(begin
	  (apply set-tm:hms tm (seconds->hms time))
	  (strftime-utf8 (string-append "%H:%M:%S  %a " (locale-d-fmt)) tm))
	(strftime-utf8 (string-append "%a " (locale-d-fmt)) tm))))

;; return element list (SYMBOL STRING FACE HELP)
(define (latest-elem symbol)
  (let ((latest (latest-get symbol)))
    (list symbol
	  (emacs-latest-line latest)
	  (let ((change (latest-change latest)))
	    (cond ((latest-in-progress? symbol)
		   'chart-in-progress)
		  ((and change (positive? change))
		   'chart-up)
		  ((and change (negative? change))
		   'chart-down)
		  (else
		   'nil)))
	  (string-join-notnull
	   (list (let ((name (or (latest-name (latest-get symbol))
				 (database-read-info symbol 'name))))
		   (string-join-notnull (list symbol name) " - "))
		 (and (latest-quote-adate latest)
		      (format #f (_ "Quote: ~a")
			      (full-adate-time (latest-quote-adate latest)
					       (latest-quote-time latest))))
		 (and (latest-last-adate latest)
		      (format #f (_ "Last:  ~a")
			      (full-adate-time (latest-last-adate latest)
					       (latest-last-time latest))))
		 (format #f (_ "(~a time; source \"~a\")")
			 (symbol-location symbol)
			 (latest-source latest)))
	   "\n"))))


;;-----------------------------------------------------------------------------
;; commands

;; send latest lines for SYMBOL-LIST
(define-public (emacs-command-latest-get-list symbol-list)
  (if (not (null? symbol-list))
      (receive (this-list rest-list)
	  (split-at-maybe symbol-list 20)  ;; write in chunks of 20 symbols
	(emacs-write 'latest-line-list
		     (filter-map latest-elem this-list))
	(emacs-command-latest-get-list rest-list))))

(define-public (emacs-command-request-symbols symbol-list)
  (latest-request-symbols symbol-list))

(define-public (emacs-command-request-explicit symbol-list)
  (latest-request-explicit symbol-list))

;; send symlist
(define-public (emacs-command-get-symlist key)
  (emacs-write 'symlist-list key (symlist-list (emacs-symlist-find key))))

(define-public (emacs-command-synchronize n)
  (emacs-write 'synchronize n))

(define-public (emacs-command-synchronous seq func . args)
  (c-main-enq! (lambda ()
		 (emacs-write 'synchronous seq
			      (apply (eval (sym-append 'emacs-command- func)
					   (interaction-environment))
				     args)))))

(define-public (emacs-command-noop)
  'nil)


;;-----------------------------------------------------------------------------
;; symlist manipulations

;; return symlist from KEY (a scheme symbol)
;; keys coming from chart.el should always be valid, so throw an error if not
(define-public (emacs-symlist-find key)
  (let ((symlist (symlist-find key)))
    (or symlist
	(error "cannot find symlist:" key))
    symlist))

(define-public (emacs-command-request-symlist key)
  (latest-request-symbols (symlist-interesting (emacs-symlist-find key))))

(define-public (emacs-command-request-explicit-symlist key)
  (latest-request-explicit (symlist-interesting (emacs-symlist-find key))))

(define-public (emacs-command-symlist-delete key pos count)
  (symlist-delete (symlist-find key) pos count))

(define-public (emacs-command-symlist-insert key pos symbol-list)
  (symlist-insert (symlist-find key) pos symbol-list))



;;-----------------------------------------------------------------------------
;; completion - symbols

;; `merge' applied to LST and lists in REST, comparing with LESS
(define (merge-many less lst . rest)
  (if (null? rest)
      lst
      (apply merge-many less (merge lst (first rest) less) (cdr rest))))

;; return a list of latest record symbols on disk
(define (latest-cache-symbols)
  (or (directory-files latest-directory #f data-filename?) '()))

(define-public (emacs-command-get-completion-symbols)
  (map list (merge-many chart-symbol<?
			(sort (symlist-list favourites-symlist) chart-symbol<?)
			(sort (latest-cache-symbols) chart-symbol<?)
			(database-symbols))))

(define (completions-update . args)
  (emacs-write 'completion-symbols-update))

(notify-connect 'database-new       completions-update)
(notify-connect 'database-delete    completions-update)
(notify-connect 'latest-cache-added completions-update)

(notify-connect 'symlist-update
  (lambda (symlist-key-list)
    (if (memq 'favourites symlist-key-list)
	(completions-update))))


;;-----------------------------------------------------------------------------
;; symlist alist

;; return symbol `nil' or `t' according as obj is #f or non-#f
(define (flag->nil/t obj)
  (if obj 't 'nil))

;; the `chart-symlist-alist' data, being a list of elements (NAME KEY
;; EDITABLE), is sent up as an asynch, and with a dummy return value
;;
;; this suits chart-watchlist startup which doesn't want to wait, and
;; chart-symlist-alist forced initialization, which does want to wait
;;
(define-public (emacs-command-get-symlist-alist)
  (emacs-write 'symlist-alist
	       (map (lambda (symlist)
		      (list (symlist-name symlist)
			    (symlist-key symlist)
			    (flag->nil/t (symlist-editable symlist))))
		    symlist-all-lists))
  'nil)


;;-----------------------------------------------------------------------------
;; individual quotes

(define (string-append-sep s1 sep s2)
  (cond ((string-null? s1)
	 s2)
	((string-null? s2)
	 s1)
	(else
	 (string-append s1 sep s2))))

;; return when SYMBOL has been retrieved
(define (latest-explicit-synchronous symbol)
  (call-with-current-continuation
   (lambda (cont)
     (define (callback symbol-list)
       (if (and (not (latest-in-progress? symbol))
		(member symbol symbol-list))
	   (begin
	     (notify-disconnect 'latest-update callback)
	     (cont #f))))
     (notify-connect 'latest-update callback)
     (c-main-enq! latest-request-explicit (list symbol))
     (c-main-goto-top))))

(define-public (emacs-command-quote-one symbol)
  (latest-explicit-synchronous symbol)

  (let ((elem (latest-elem symbol)))
    (let ((extra ""))

      ;; full symbol if it's truncated in the elem line
      (if (> (utf8-string-length symbol) 9)
	  (set! extra symbol))

      ;; " - name", when the name is available
      (and-let* ((name (or (latest-name (latest-get symbol))
			   (database-read-info symbol 'name))))
	(set! extra (string-append-sep extra " - " name)))

      (and-let* ((location (symbol-location-maybe symbol)))
	(set! location (string-append "[" location "]"))
	(set! extra (string-append-sep extra "  " location)))

      (set-second! elem (string-append-sep (second elem) "\n" extra)))

    elem))


;;-----------------------------------------------------------------------------
;; latest records

(define (mung obj)
  (cond ((not obj) 'nil)
	((string? obj) obj)
	(else obj)))

(define-public (emacs-command-get-latest-record symbol)
  (let ((latest (latest-get symbol)))
    (list 'symbol   (latest-symbol latest)
	  'name     (mung (latest-name   latest))
	  'bid      (mung (latest-bid    latest))
	  'offer    (mung (latest-offer  latest))
	  'open     (mung (latest-open   latest))
	  'high     (mung (latest-high   latest))
	  'low      (mung (latest-low    latest))
	  'last     (mung (latest-last   latest))
	  'change   (mung (latest-change latest))
	  'volume   (mung (latest-volume latest))
	  'decimals (latest-decimals latest)
	  'note     (mung (latest-note   latest))
	  'source   (latest-source   latest))))


;;-----------------------------------------------------------------------------
;; spreadsheet support

;; return when SYMBOL-LIST has been downloaded
(define-public (emacs-command-request-symbols-synchronous symbol-list)
  (call-with-current-continuation
   (lambda (cont)
     (define (callback this-symbol-list)
       (set! this-symbol-list (remove latest-in-progress? this-symbol-list))
       (set! symbol-list (lset-difference string=?
					  symbol-list this-symbol-list))
       (if (null? symbol-list)
	   (begin
	     (notify-disconnect 'latest-update callback)
	     (cont #f))))
     (notify-connect 'latest-update callback)
     (c-main-enq! latest-request-symbols symbol-list)
     (c-main-goto-top)))
  'nil)


;;-----------------------------------------------------------------------------
;; interaction

;; cf `symbol-append' in boot-9.scm
(define (sym-append sym . rest)
  (set! rest (cons sym rest))
  (string->symbol (string-concatenate (map! symbol->string rest))))

(define emacs-port #f)

(define char-set:dquote-backslash
  (char-set #\" #\\))

;; print STR to PORT, with any double quote " or backslash \ escaped by a
;; precding backslash
;;
(define (display-string-escape-dquote str port)
  (let more ((pos 0))
    (let ((end (string-index str char-set:dquote-backslash pos)))
      (if end
	  (begin
	    (display (substring/shared str pos end) port)
	    (display "\\" port)
	    (display (string-ref str end) port)
	    (more (1+ end)))
	  (display (substring/shared str pos) port)))))

;; print OBJ on PORT, in a style ready for the elisp reader
;; this differs from plain scheme `write' in that it doesn't escape 8-bit chars
;; (we need them to go through unchanged for utf8)
;;
(define (el-write obj port)
  (cond ((pair? obj)
	 (display "(" port)
	 (let more ((obj obj))
	   (el-write (car obj) port)
	   (cond ((null? (cdr obj))
		  ;; nothing further
		  )
		 ((pair? (cdr obj))
		  (display " " port)
		  (more (cdr obj)))
		 (else
		  (display " . " port)
		  (el-write (cdr obj) port))))
	 (display ")" port))

	((string? obj)
	 (display "\"" port)
	 (display-string-escape-dquote obj port)
	 (display "\"" port))

	(else
	 (write obj port))))

;; form output in string and write in one block
(define-public (emacs-write proc . args)
  (el-write (cons proc args) emacs-port)
  (force-output emacs-port))

;; return OBJ with any `nil' symbols changed to ()
;; lists are processed recursively
;;
(define (nil->empty obj)
  (cond ((eq? obj 'nil)
	 '())
	((pair? obj)
	 (map nil->empty obj))
	(else
	 obj)))

(define-public (emacs-run-command form)
  (apply (eval (sym-append 'emacs-command- (first form))
	       (interaction-environment))
	 (cdr form)))

(define-public (emacs-error stack key args)
  (let* ((err-str   (call-with-output-string
		     (lambda (port)
		       (display (_ "Chart") port)
		       (display " " port)
		       (if (= 4 (length args))
			   (apply display-error #f port args)
			   (format port "~a: ~a\n" key args)))))
	 (backtrace (if stack
			(call-with-output-string
			 (lambda (port)
			   (display-backtrace stack port)
			   (display err-str)))
			'nil)))
    (set! err-str (string-trim-right err-str))
    (if stack
	(set! err-str
	      (string-append err-str "\n"
			     (_ "See *chart-process-backtrace* buffer"))))
    (emacs-write 'error err-str backtrace)))

(define-public (emacs-interaction)

  ;; emacs is on stdout
  (set! emacs-port (current-output-port))
  (setvbuf emacs-port _IOFBF 8192)

  ;; discard output (diagnostics or whatever)
  (set-current-output-port (%make-void-port "w"))
  (set-current-error-port  (current-output-port))

  (notify-connect 'latest-update
    (lambda (symbol-list)
      (emacs-write 'update symbol-list)))

  (notify-connect 'database-update (lambda (symbol-list)
				     (emacs-write 'update symbol-list)))

  (notify-connect 'symlist-update
    (lambda (symlist-key-list)
      (emacs-write 'symlist-update symlist-key-list)))

  (fcntl (current-input-port) F_SETFL
	 (logior (fcntl (current-input-port) F_GETFL) O_NONBLOCK))

  ;; initial messages
  (emacs-write 'init "UTF-8" protocol-version)

  (let ((buf ""))
    (c-gtk-input-add (current-input-port) '(read)
      (lambda (fd conds)
	;; read as many bytes as available
	(let more ()
	  (let* ((part  (make-string 256))
		 (nread (read-string!/partial part (current-input-port))))
	    (cond ((not nread) ;; eof
		   (exit 0))
		  ((positive? nread)
		   (set! buf (string-append buf (string-take part nread)))
		   (more)))))

	;; process bytes read, or the complete forms within it
	(let more ()
	  (let* ((port (open-input-string buf))
		 (form (catch #t
			 (lambda () (read port))
			 (lambda args eof-object))))
	    (if (not (eof-object? form))
		(begin
		  (set! form (nil->empty form))
		  (set! buf (string-drop buf (ftell port)))
		  (emacs-run-command form)
		  (more))))))))

  (set! c-main-run-q
	(lambda ()
	  (if (not (q-empty? c-main-queue))
	      (let ((lst (deq! c-main-queue)))

		(catch 'chart-continue
		  (lambda ()
		    (lazy-catch #t
		      (lambda ()
			(apply (car lst) (cdr lst)))
		      (lambda (key . args)
			(if (eq? key 'quit) ;; `exit' call
			    (apply throw key args))
			(let ((stack (make-stack #t 5)))
			  (emacs-error stack key args))
			(throw 'chart-continue))))
		  (lambda (key . args)
		    #f))
		(c-main-run-q)))))

  (notify-listen)
  (c-main))
