;; Extra Gdk functions.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart extra-gdk)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gdk)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; misc

;; make-procedure-with-cache-eq returns a procedure which calls (FUNC obj),
;; and caches the results.  Any subsequent call with the same OBJ, as tested
;; with `eq?' just returns the cached value.
;;
(define-public (make-procedure-with-cache-eq func)
  (or (procedure? func)
      (error "make-procedure-with-cache-eq: bad function:" func))
  (let ((prop (primitive-make-property
	       (lambda (prop x)
		 (func x)))))
    (lambda (x)
      (primitive-property-ref prop x))))


;;-----------------------------------------------------------------------------
;; PangoLayout

;; set STR into LAYOUT and return the width in pixels of that string
(define-public (c-pango-layout-width layout str)
  (pango-layout-set-text layout str)
  (first (pango-layout-get-pixel-size layout)))

;; return the width of pixels of an "M" in LAYOUT
(define-public (c-pango-layout-em layout)
  (c-pango-layout-width layout "M"))

(define-public (c-pango-layout-font-metrics layout)
  (let* ((context  (pango-layout-get-context layout))
	 (fontdesc (or (pango-layout-get-font-description layout)
		       (pango-context-get-font-description context)))
	 (lang     (pango-context-get-language context)))
    (pango-context-get-metrics context fontdesc lang)))

;; return the line height in pixels for LAYOUT
(define-public (c-pango-layout-line-height layout)
  (let ((metrics (c-pango-layout-font-metrics layout)))
    (quotient-ceiling (+ (pango-font-metrics-get-ascent metrics)
			 (pango-font-metrics-get-descent metrics))
		      pango-scale)))


;;-----------------------------------------------------------------------------
;; gdk drawing primitives

;; Clear an outlined rectangle, given as width/height like gdk-draw-rectangle.
;; Draw only if X1,Y1 - X2,Y2 crosses the window of WIDTH and HEIGHT, and
;; only draw that portion which intersects (so as to avoid big values of x
;; or y.
;;
;; FIXME: find and draw only the intersecting bit
;;
(define-public (c-gdk-draw-line-restrict drawable gc x1 y1 x2 y2 width height)
  (if (and (>= (max x1 x2) 0)
	   (<  (min x1 x2) width)
	   (>= (max y1 y2) 0)
	   (<  (min y1 y2) height))
      (gdk-draw-line drawable gc x1 y1 x2 y2)))


;;-----------------------------------------------------------------------------
;; GdkCursor

;; gdk-cursor-new, as of gtk 2.10.12, does a new CreateGlyphCursor for every
;; call, so cache
(define-public c-gdk-cursor-watch
  (make-procedure-with-cache-0 (lambda ()
				 (gdk-cursor-new 'watch))))


;;-----------------------------------------------------------------------------
;; GdkDrawable

;; not wrapped in guile-gtk 2.0
(if (not (defined? 'gdk-drawable-get-size))
    (define-public gdk-drawable-get-size gdk-window-get-size))

(define-public (c-gdk-drawable-get-width drawable)
  (car (gdk-drawable-get-size drawable)))

(define-public (c-gdk-drawable-get-height drawable)
  (cdr (gdk-drawable-get-size drawable)))


;;-----------------------------------------------------------------------------
;; GdkEvent

;; diagnostic print, taking string or events
(define-public (d-event . args)
  (for-each (lambda (arg)
	      (if (string? arg)
		  (begin
		    (display arg)
		    (display " "))
		  (format #t "~a button=~a x=~a y=~a state=~a\n"
			  (gdk-event-type arg)
			  (gdk-event-button arg)
			  (gdk-event-x arg)
			  (gdk-event-y arg)
			  (gdk-event-state arg))))
	    args))


;;-----------------------------------------------------------------------------
;; GdkFont

(define-public (c-gdk-font-height font)
  (+ (gdk-font-ascent font)
     (gdk-font-descent font)))

;; the width in pixels of the widest digit in LAYOUT
(define-public c-pango-layout-digit-width
  (make-procedure-with-cache-eq
   (lambda (layout)
     (apply max (map (lambda (str)
		       (c-pango-layout-width layout str))
		     '("0" "1" "2" "3" "4" "5" "6" "7" "8" "9"))))))


;;-----------------------------------------------------------------------------
;; GdkGc pool

(let-toplevel ((pool '()))  ;; alist (ARGS . GC)

  (define*-public (c-gdk-gc-get win #:key
				(dashed #f)
				(font "fixed")
				(foreground "white")
				(function 'copy))
    (let ((key (list dashed
		      font
		      (if (string? foreground)
			  foreground
			  (gdk-color-pixel foreground))
		      function)))
      (or (assoc-ref pool key)

	  (let ((gc (gdk-gc-new-with-values
		     win
		     #:graphics-exposures #f
		     #:foreground foreground
		     #:function   function
		     #:line-style (if dashed 'on-off-dash 'solid)
		     #:font       (gdk-font-intern font))))
	    (set! pool (acons key gc pool))
	    gc)))))


;;-----------------------------------------------------------------------------
;; Gdk keys

(define-public GDK_VoidSymbol (gdk-keyval-from-name "VoidSymbol"))
