;; Extra Gtk functions.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart extra-gtk)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gdk)
  #:use-module (chart gui-common)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; Idle priority levels:
;;
;; -100  gtk-priority-high     g-priority-high
;;    0  gdk-priority-events   g-priority-default         timeouts, X events
;;  100                        g-priority-high-idle
;;  110  gtk-priority-resize                              widget resize
;;  120  gtk-priority-redraw                              widget redraw
;;  200  gtk-priority-default  g-priority-default-idle    default idle
;;  300  gtk-priority-low      g-priority-low             background
;;
;; Redraws are at gtk-priority-redraw as ususal.  Busy indicator released at
;; gtk-priority-low.

;; not wrapped
(if (not (defined? 'g-priority-high-idle))
    (define-public g-priority-high-idle 100))


;;-----------------------------------------------------------------------------
;; gtk-idle

;; (c-gtk-idle-once THUNK) arranges for THUNK to be executed once when gtk
;; becomes idle.  THUNK should be a procedure of no arguments.
;;
;; Until THUNK has executed, further calls to c-gtk-idle-once for that same
;; THUNK do nothing, it'll still only be executed once.
;;
;; THUNK is allowed to call c-gtk-idle-once to get itself executed again, or
;; some other procedure executed.
;;
;; As usual for gtk+ functions, c-gtk-idle-once is not thread-safe and it
;; should be protected with gdk-threads-enter / gdk-threads-leave if two
;; threads might be doing gtk things simultaneously.
;;
;; Gtk idle functions are executed outside the gdk-threads-enter /
;; gdk-threads-leave which is applied to normal event dispatch, and the same
;; is true when THUNK is executed.  If THUNK does any gtk stuff and another
;; thread might also be doing so too then its up to THUNK to apply
;; gdk-threads-enter and gdk-threads-leave.
;;
(let-toplevel ((prop (primitive-make-property #f)))

  (define-public (c-gtk-idle-once thunk)
    (if (not (primitive-property-ref prop thunk))
	(primitive-property-set!
	 prop thunk
	 (gtk-idle-add (lambda ()
			 (primitive-property-del! prop thunk)
			 (thunk)
			 #f)))))

  ;; remove any queued idle call to THUNK
;;   (define-public (c-gtk-idle-once-remove thunk)
;;     (let ((id (primitive-property-ref prop thunk)))
;;       (if id
;; 	  (begin
;; 	    (gtk-idle-remove id)
;; 	    (primitive-property-del! prop thunk)))))

  ;; forcibly run THUNK now, and remove any queued idle call to it
;;   (define-public (c-gtk-idle-once-now thunk)
;;     (c-gtk-idle-once-remove thunk)
;;     (thunk))
  )

;;-----------------------------------------------------------------------------
;; gtk signals

;; connect PROC to SIGNAL from WIDGET, but run it only once
;; after the first run the procedure is disconnected
;; this can allow initialization code to be garbage collected
;;
(define-public (c-gtk-signal-connect-once widget signal proc)
  (letrec ((id (gtk-signal-connect widget signal
				   (lambda args
				     (gtk-signal-disconnect widget id)
				     (apply proc args)))))
    id))


;;-----------------------------------------------------------------------------
;; GtkAdjustment

(define*-public (c-gtk-adjustment-change adj
					 #:key
					 (lower #f)
					 (upper #f)
					 (page-size #f)
					 (step-increment #f)
					 (page-increment #f))
  (define (adj-values)
    (list (gtk-adjustment-lower adj)
	  (gtk-adjustment-upper adj)
	  (gtk-adjustment-page-size adj)
	  (gtk-adjustment-step-increment adj)
	  (gtk-adjustment-page-increment adj)))
  (let ((old-values (adj-values)))
    (if lower
	(gtk-adjustment-set-lower adj lower))
    (if upper
	(gtk-adjustment-set-upper adj upper))
    (if page-size
	(gtk-adjustment-set-page-size adj page-size))
    (if step-increment
	(gtk-adjustment-set-step-increment adj step-increment))
    (if page-increment
	(gtk-adjustment-set-page-increment adj page-increment))
    (if (not (equal? old-values (adj-values)))
	(gtk-adjustment-changed adj))))

(define-public (c-gtk-adjustment-upper-inpage adj)
  (- (gtk-adjustment-upper     adj)
     (gtk-adjustment-page-size adj)))

;; set-value, but restricting to the valid range lower to upper-page
(define-public (c-gtk-adjustment-set-value-inpage adj val)
  (gtk-adjustment-set-value adj
			    (max (gtk-adjustment-lower adj)
				 (min val
				      (c-gtk-adjustment-upper-inpage adj)))))

;; move ADJ to its upper end
(define-public (c-gtk-adjustment-end-up adj)
  (gtk-adjustment-set-value adj (max (gtk-adjustment-lower          adj)
				     (c-gtk-adjustment-upper-inpage adj))))

;; move ADJ down by a page
(define-public (c-gtk-adjustment-page-down adj)
  (c-gtk-adjustment-set-value-inpage
   adj
   (- (gtk-adjustment-value adj)
      (gtk-adjustment-page-increment adj))))

;; move ADJ up by a page
(define-public (c-gtk-adjustment-page-up adj)
  (c-gtk-adjustment-set-value-inpage
   adj
   (+ (gtk-adjustment-value adj)
      (gtk-adjustment-page-increment adj))))

;; set ADJ to have VAL in the centre of the visible part
(define-public (c-gtk-adjustment-set-value-centre adj val)
  (c-gtk-adjustment-set-value-inpage
   adj (- val (* 0.5 (gtk-adjustment-page-size adj)))))


;;-----------------------------------------------------------------------------
;; GtkCheckMenuItem

;; not wrapped in guile-gtk 2.0
(if (not (defined? 'gtk-check-menu-item-set-active))
    (define-public (gtk-check-menu-item-set-active item state)
      (or (gtk-check-menu-item-active item)
	  (gtk-menu-item-activate item))))


;;-----------------------------------------------------------------------------
;; GtkContainer

(define-public (c-gtk-container-remove-all container)
  (for-each (lambda (child)
	      (gtk-container-remove container child))
	    (gtk-container-children container)))


;;-----------------------------------------------------------------------------
;; GtkDialog

;; setup DIALOG so that a delete_event (from the window manager) hides the
;; dialog, instead of the way the default GtkWidget signal handler does a
;; destroy
(define-public (c-gtk-dialog-hide-on-delete dialog)
  (gtk-signal-connect dialog "delete_event"
    (trampoline-return (lambda (event)
			 (gtk-widget-hide dialog))
		       #t))) ;; don't propagate to default destroy
   

;;-----------------------------------------------------------------------------
;; GtkDrawingArea

;; draw STR centred in WIDGET
(define-public (c-gtk-draw-string-centred widget str)
  (let* ((layout   (primitive-property-ref c-layout-prop widget))
	 (drawable (gtk-widget-window widget))
	 (style    (gtk-widget-get-style widget))
	 (gc       (gtk-style-text-gc style 'normal)))
    (pango-layout-set-text layout str)
    (receive-list (width height)
	(pango-layout-get-pixel-size layout)
      (gdk-draw-layout drawable gc
		       (quotient (- (c-gdk-drawable-get-width drawable)
				    width) 2)
		       (quotient (- (c-gdk-drawable-get-height drawable)
				    height) 2)
		       layout))))




;;-----------------------------------------------------------------------------
;; GtkEditable

;; C macro not wrapped
(define (gtk-widget-realized widget)
  (->bool (memq 'realized (gtk-widget-flags widget))))

;; like `gtk-editable-select-region' but don't put the region on the clipboard
;; this is done by pretending the widget is unrealized
(define-public (c-gtk-editable-select-region-noclip widget start end)
  (let ((flag (if (gtk-widget-realized widget) '(realized) '())))
    (gtk-widget-unset-flags widget flag)
    (gtk-editable-select-region widget start end)
    (gtk-widget-set-flags widget flag)))


;;-----------------------------------------------------------------------------
;; GtkLabel stuff

;; set LABEL to STR, if not already that text
;; (each set does a redraw, so avoiding a set of the same text stops flashing)
;;
(define-public (c-gtk-label-set-text-maybe label str)
  (if (not (string=? str (gtk-label-get label)))
      (gtk-label-set-text label str)))


;;-----------------------------------------------------------------------------
;; GtkList

;; setup for N labels in LST (a list widget)
;; call (PROC item) on each newly created item
;;
;; when using existing item widgets, those up to N are shown and those above
;; are hidden
;;
(define-public (c-gtk-list-require-n-labels lst n)
  (do ((i (length (gtk-container-children lst)) (1+ i)))
      ((>= i n))
    (let ((item (gtk-list-item-new-with-label "")))
      (gtk-list-append-item lst item)))
  (let ((children (gtk-container-children lst)))
    (for-each gtk-widget-show (take children n))
    (for-each gtk-widget-hide (drop children n))))


;;-----------------------------------------------------------------------------
;; GtkMenuItem

;; create a menu item with a mnemonic accelerator for use in OPTION-MENU
;;
(define-public (c-gtk-menu-item-new-with-mnemonic option-menu accel-group str)
  (let* ((item   (gtk-menu-item-new))
	 (label  (gtk-label-new-with-mnemonic str))
	 (keyval (gtk-label-get-mnemonic-keyval label)))
    (gtk-misc-set-alignment label 0.0 0.5)
    (or (= keyval GDK_VoidSymbol)
	(gtk-widget-add-accelerator item "activate" accel-group
				    keyval '(mod1-mask) '(locked)))
    (gtk-signal-connect item "activate"
      (lambda()
	(c-gtk-option-menu-set-history-item option-menu item)))
    (gtk-container-add item label)
    item))

  
;;-----------------------------------------------------------------------------
;; GtkOptionMenu

;; set the given ITEM as current on OPTION-MENU
(define (c-gtk-option-menu-set-history-item option-menu item)
  (let* ((menu     (gtk-option-menu-get-menu option-menu))
	 (children (gtk-container-children menu))
	 (index    (or (list-index-elem item children)
		       (error "c-gtk-option-menu-set-history-item: item not found:" option-menu item))))
    (gtk-option-menu-set-history option-menu index)))


;;-----------------------------------------------------------------------------
;; GtkStyle

;; gone from gtk 2.0
(if (not (defined? 'gtk-style-font))
    (define-public (gtk-style-font style)
      "fixed"))


;;-----------------------------------------------------------------------------
;; GtkTextBuffer

;; append STR to given GtkTextBuffer
(define-public (c-gtk-text-buffer-append buffer str)
  (let ((iter (gtk-text-iter-new)))
    (gtk-text-buffer-get-end-iter buffer iter)
    (gtk-text-buffer-insert buffer iter str)))
;; hack to show new text
;;  (for-each c-gtk-text-view-scroll-to-end (rec-viewlist buf)))


;;-----------------------------------------------------------------------------
;; GtkTextView

;; scroll to show end of text
(define-public (c-gtk-text-view-scroll-to-end textview)
  (let ((iter (gtk-text-iter-new))
	(textbuf (gtk-text-view-get-buffer textview)))
    (gtk-text-buffer-get-end-iter textbuf iter)
    (gtk-text-view-scroll-to-iter textview iter 0 #t 0 1.0)))


;;-----------------------------------------------------------------------------
;; GtkWidget stuff

;; not wrapped
(define-public (gtk-widget-mapped widget)
  (->bool (memq 'mapped (gtk-widget-flags widget))))


(define-public c-layout-prop
  (primitive-make-property (lambda (prop obj)
			     (gtk-widget-create-pango-layout obj))))


;;-----------------------------------------------------------------------------
;; GtkWidget extras

;; change the cursor to show busy, automatically reverted when idle
;;
(let-toplevel ((busy-property (primitive-make-property #f)))
  (define-public (c-gtk-widget-busy widget)
    (or (primitive-property-ref busy-property widget)
	(not (gtk-widget-realized widget))

	(begin
	  (gdk-window-set-cursor (gtk-widget-window widget)
				 (c-gdk-cursor-watch))
	  (gdk-flush)
	  (primitive-property-set! busy-property widget #t)

	  ;; priority after redrawing done
	  (gtk-idle-add-priority (1+ gtk-priority-redraw)
	    (lambda ()
	      (gdk-window-set-cursor (gtk-widget-window widget) #f)
	      (primitive-property-del! busy-property widget)
	      #f))))))  ;; remove handler

;; get font from widget style
(define-public (c-gtk-widget-get-font widget)
  (gtk-style-font (gtk-widget-get-style widget)))

;; raise the widget's window, or do nothing if not yet realized
(define-public (c-gtk-widget-raise widget)
  (if (gtk-widget-window widget)
      (gdk-window-raise (gtk-widget-window widget))))

;; do `gtk-widget-show-all' if not yet mapped, otherwise just raise
;; this is meant for use on a toplevel or dialog widget
(define-public (c-gtk-widget-show-all-or-raise w)
  ((if (gtk-widget-mapped w)
       c-gtk-widget-raise
       gtk-widget-show-all) w))

;; do `gtk-widget-show' if not yet mapped, otherwise just raise
;; this is meant for use on a toplevel or dialog widget
(define-public (c-gtk-widget-show-or-raise w)
  ((if (gtk-widget-mapped w)
       c-gtk-widget-raise
       gtk-widget-show) w))

;; set usize, if not currently that size
(define-public (c-gtk-widget-set-usize-maybe widget width height)
  (define (usize= x y)
    (or (< x 0)
	(= x y)))
  (let ((win (gtk-widget-window widget)))
    (if (not (and win
		  (usize= width  (c-gdk-drawable-get-width  win))
		  (usize= height (c-gdk-drawable-get-height win))))
	(gtk-widget-set-usize widget width height))))

;; WIDTH can be <int>             pixels
;;              (em . <num>)      "M" width
;;              (digit . <num>)   digit width
;;              (str . <string>)  string width
;;
;; The <num> values can be integers or floats, the result is an exact
;; integer, rounding if necessary.
;;
(define-public (c-gtk-widget-units->width widget width)
  (if (pair? width)
      (let ((layout (primitive-property-ref c-layout-prop widget)))
	(case (car width)
	  ((em)
	   (set! width (* (cdr width) (c-pango-layout-em layout))))
	  ((digit)
	   (set! width (* (cdr width) (c-pango-layout-digit-width layout))))
	  ((str)
	   (set! width (c-pango-layout-width layout (cdr width)))))))
  (round-exact width))

;; HEIGHT can be <int>             pixels
;;               (height . <int>)  font height
;;
(define-public (c-gtk-widget-units->height widget height)
  (if (pair? height)
      (let ((layout (primitive-property-ref c-layout-prop widget)))
	(cond ((eq? 'height (car height))
	       (set! height (* (cdr height)
			       (c-pango-layout-line-height layout)))))))
  (round-exact height))

;; WIDTH, HEIGHT in units per above
(define-public (c-gtk-widget-set-usize-units widget width height)
  (gtk-widget-set-usize widget
			(c-gtk-widget-units->width widget width)
			(c-gtk-widget-units->height widget height)))

;; the width of an "M" in WIDGET
(define-public (c-gtk-widget-em widget)
  (c-pango-layout-em (primitive-property-ref c-layout-prop widget)))


;;-----------------------------------------------------------------------------
;; GtkWindow - extra events

;; c-gtk-widget-add-events and c-gtk-widget-remove-events maintain event
;; masks under supplied keys, allowing them to be added or removed and
;; updating the widget to the union of those masks plus its base mask.
;;
;; These function can be used before the widget is realized, the "realize"
;; signal is hooked to install the events at that point.  The realize signal
;; remains hooked to cope with unrealize/re-realize, as occurs for instance
;; when unparenting then re-parenting to change GtkTable anchors.
;;
(let-toplevel ((prop (primitive-make-property (lambda (prop obj) '()))))

  (define (update-events widget)
    (and-let* ((win (gtk-widget-window widget)))
      (let ((alst (primitive-property-ref prop widget)))
	(or (assq-ref alst 'base)
	    (begin
	      (set! alst (acons 'base (gdk-window-get-events win) alst))
	      (primitive-property-set! prop widget alst)))
	(gdk-window-set-events win (append-map cdr alst)))))

  (define-public (c-gtk-widget-add-events widget key mask)
    (let ((alst (primitive-property-ref prop widget)))
      (or (assq 'realize-connected alst)
	  (begin
	    (gtk-signal-connect widget "realize"
	      (lambda ()
		(update-events widget)))
	    (set! alst (acons 'realize-connected '() alst))))
      (primitive-property-set! prop widget (assq-set! alst key mask)))
    (update-events widget))

  (define-public (c-gtk-widget-remove-events widget key)
    (let ((alst (primitive-property-ref prop widget)))
      (or (assq key alst)
	  (error "c-gtk-widget-remove-events: no such key" key widget))
      (primitive-property-set! prop widget (assq-remove! alst key)))
    (update-events widget)))


;;-----------------------------------------------------------------------------
;; GtkWindow extras

;; return STR with "Chart:" added
(define-public (c-make-window-title str)
  (if (or (not str)
	  (string-null? str))
      (_ "Chart")
      (string-append (_ "Chart") ": " str)))


;;-----------------------------------------------------------------------------
;; redrawing

(let-toplevel ((alst '())  ;; (WIDGET . CLEAR-FLAG)
	       (prop (primitive-make-property #f)))

  ;; install the idle handler if not already installed, ie. when alst is
  ;; currently empty
  (define (install-idle)
    (if (null? alst)
	(gtk-idle-add-priority gtk-priority-redraw do-idle)))

  (define (do-idle)
    (let ((al alst))
      (set! alst '()) ;; allow drawing to re-queue

      ;; run all clears first
      (for-each (lambda (pair)
		  (if (and (cdr pair)
			   ;; only if window hasn't been destroyed
			   (gtk-widget-window (car pair)))
		      (gdk-window-clear (gtk-widget-window (car pair)))))
		al)

      ;; then all drawing
      (for-each (lambda (pair)
		  (let* ((widget (car pair)))
		    ;; only if window hasn't been destroyed
		    (if (gtk-widget-window widget)
			(let ((procs  (primitive-property-ref prop widget)))
			  (if (not (cdr pair))
			      ((cdr procs)))  ;; un-xor, unless cleared
			  ((car procs))       ;; draw
			  ((cdr procs))))))   ;; xor
		al)))

  (define (queue-draw widget)
    (install-idle)
    (if (not (assq widget alst))
	(set! alst (acons widget #t alst))))

  ;; install redraw handlers for WIDGET, to call (DRAW) and (XOR)
  (define*-public (c-gtk-widget-drawing widget draw #:optional (xor noop))
    (primitive-property-set! prop widget (cons draw xor))
    (gtk-signal-connect widget "expose_event"
      (lambda (event)
	(if (= 0 (gdk-event-count event)) ;; final expose
	    (queue-draw widget)))))

  (define-public (c-gtk-widget-queue-redraw widget)
    (install-idle)
    (set! alst (assq-set! alst widget #t))
    #f) ;; propagate (can end up called before gui-common trampoline init)

  ;; return true is redraw is queued for WIDGET
  (define-public (c-gtk-widget-queued-redraw? widget)
    (assq widget alst)))
