;; Fatdog Exchange setups and data downloading.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; The site announces itself as down as of April 2008.  Leave the code here,
;; but remove it from the manual (see the end of this file for that text).


(define-module (chart fatdog)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (fatdog-symbol? symbol)
  (string-suffix? ".FAT" symbol))

(symbol-timezone! fatdog-symbol? (_ "Local") timezone-local)


;;-----------------------------------------------------------------------------
;; weblink - new page

(weblink-handler! fatdog-symbol?
  (lambda (symbol)
    (list (_ "Fatdog _Celebrity Page")
	  (_ "Open web browser at the Fatdog Exchange page for this celebrity")
	  (string-append "http://www.fatdogexchange.com/celebrity/news/"
			 (space->underscore (chart-symbol-sans-dot symbol))
			 ".htm"))))

(define (space->underscore str)
  (string-map (lambda (c)
		(if (eqv? c #\space) #\_ c))
	      str))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses quotes pages like
;;
;;     http://www.fatdogexchange.com/celebrity/news/Minnie_Driver.htm
;;
;; which have the price part in a frame like
;;
;;     http://www.fatdogexchange.com/celebrity/core/price.php?celeb=Minnie%20Driver
;;
;; This price part is only about 800 bytes.  There's no ETag or
;; Last-Modified.


(define (body->latest symbol body)

  (let* ((str    (string-collapse-whitespace (html-text-extract body)))
	 (m      (must-match (regexp-exec #,(regexp "^ *Price current: ([0-9.]+) change: ([0-9.-]+) high: ([0-9.]+); low: ([0-9.]+) volume: ([0-9,.]+)") str)))
	 (last   (match:substring m 1))
	 (change (match:substring m 2))
	 (high   (match:substring m 3))
	 (low    (match:substring m 4))
	 (volume (match:substring m 5))
	 (tm     (localtime (current-time))))

    ;; last empty for no such symbol
    (if (string-null? last)
	(latest-new-unknown symbol)

	(latest-new #:symbol      symbol
		    #:quote-adate (tm->adate tm)
		    #:quote-time  (tm->time tm)
		    #:last-adate  (tm->adate tm)  ;; probably not quite right
		    #:high        high
		    #:low         low
		    #:last        last
		    #:change      change
		    #:volume      volume
		    #:source      'fatdog))))

(define (fatdog-latest-get symbol-list extra-list proc)
  (for-each
   (lambda (symbol)
     (receive (headers body)
	 (http-request (string-append
			"http://www.fatdogexchange.com/celebrity/core/price.php?celeb="
			(url-quote (chart-symbol-sans-dot symbol)))
		       #:want-ok #t)
       (proc (list (body->latest symbol body)))))
   symbol-list))

(latest-handler! #:selector fatdog-symbol?  
		 #:handler  fatdog-latest-get)


;;-----------------------------------------------------------------------------
;; intraday
;;
;; This uses intraday graphs included on the quote page above, like
;;
;;     http://www.fatdogexchange.com/celebrity/chart/hourly.php?celeb=Adam%20Sandler
;;     http://www.fatdogexchange.com/celebrity/chart/daily.php?celeb=Adam%20Sandler

(define mode-month (_ "_Month"))

(define (fatdog-intraday-url symbol mode)
  (string-append "http://www.fatdogexchange.com/celebrity/chart/"
		 (cond ((string=? mode mode-month)
			"daily")
		       (else ;; Day
			"hourly"))
		 ".php?celeb="
		 (url-quote (chart-symbol-sans-dot symbol))))

(intraday-url-proc! #:symbol-pred fatdog-symbol?
		    #:url-proc    fatdog-intraday-url
		    #:modes-proc  (lambda ()
				    (list (_ "_Day")
					  mode-month)))



;; (source-help! fatdog-symbol? (_ "CONTEXT(manual-node): Fatdog Exchange"))

;; @node Fatdog Exchange
;; @section Fatdog Exchange
;; @cindex Fatdog Exchange

;; @uref{http://fatdogexchange.com}

;; Fatdog provides

;; @itemize
;; @item
;; Quotes.
;; @item
;; Intraday and past month graphs.
;; @end itemize

;; @cindex @code{.FAT}
;; In Chart Fatdog symbols are the celebrity name with a @samp{.FAT} suffix,
;; eg.@: @samp{Mark Webber.FAT}.  Lists of celebrities can be found under the
;; Market Table links on the home page.
;; @c SYMBOL: Mark Webber.FAT

;; Past data is only available as a graph (a PNG image), it can be viewed in the
;; intraday window (@pxref{Intraday}) by selecting ``Month'' in the option menu.

;; There's no obvious timezone for Fatdog, so the system local timezone is used
;; in the watchlist.
