;; Fisher Transform.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart fisher)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart midpoint)
  #:use-module (chart misc)
  #:use-module (chart williams-r))


;; http://www.mesasoftware.com/Papers/Using%20the%20Fisher%20Transform.exe
;;     Paper by John Elhers (self-extracting zip archive of MS word).
;;
;; http://www.linnsoft.com/tour/techind/fish.htm
;;     Sample NYSE QQQ, August 2002.


;; Return a procedure which calculates a Fisher transform over a window of
;; COUNT many points.
;;
;; Each call (PROC x) enters a new point into the window, and the return is
;; the FISH value up to (and including) that point, or #f if not enough
;; points yet.
;;
;; To prime the window initially, call PROC with (fisher-warmup-count COUNT)
;; many points preceding the first desired.
;;
(define-public (fisher-calc-proc count)

  ;; val-smooth is alpha=0.33, which is 5 days
  ;; fish-smooth is alpha=0.5, which is 3 days
  ;;
  (define highlow-proc (highlow-calc-proc count))
  (define val-smooth   (ema-calc-proc 5))
  (define fish-smooth  (ema-calc-proc 3))

  (lambda (x)
    (receive (high low)
	(highlow-proc x x)
      (and high
	   (let* ((range (- high low))
		  (raw   (if (zero? range)
			     0
			     (1- (/ (* 2.0 (- x low)) range)))))
	     (and-let* ((value (val-smooth raw)))
	       (set! value (max -0.999 (min 0.999 value)))
	       (fish-smooth (log (/ (+ 1 value)
				    (- 1 value))))))))))

(define (fisher-warmup-count count)
  (+ count -1 warmup-fixed))
(define warmup-fixed (+ (ema-warmup-count 5)   ;; value smoothing
			(ema-warmup-count 3))) ;; transformed fish smoothing


(define (fisher-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (fisher-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (for-each proc (series-prev-n-closes parent #:before lo
					 #:count (fisher-warmup-count count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i))
		      i)))))

(define-public (fisher-series series count)
  ;; could do the midpoint with explicit code, but midpoint-series is easier
  ;; and might be able to make it configurable the same way the moving
  ;; average input is
  (set! series (midpoint-series series))

  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Fisher") count))
    (series-datatype!  new-series fisher-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   fisher-fill (fisher-warmup-count count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public fisher-datatype
  (datatype-new!
   #:key      'fisher
   #:name     (_ "Fisher Transform")
   #:units    'fisher
   #:hlines   '(0)
   #:manual   (_ "CONTEXT(manual-node): Fisher Transform")
   #:proc     fisher-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'fisher-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
