;; Force index.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart force-index)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.linnsoft.com/tour/techind/efi.htm
;;     Formula, but sample only an intraday.
;;


;; Return a procedure which calculates a force index on successive closing
;; prices, with smoothing by COUNT periods.
;;
;; Each call (PROC close volume) enters a new point into the window, and the
;; return is the force index up to (and including) that point, or #f if not
;; enough points yet.
;;
;; The force index uses an EMA and is thus in theory influenced by all
;; preceding data, but `force-index-warmup-count' below is designed to
;; determine a warmup count.  By calling PROC with (force-index-warmup-count
;; COUNT) many values, the next call will have an omitted weight of no more
;; than 0.1% of the total.  Omitting 0.1% should be negligable, unless past
;; values are ridiculously bigger than recent ones.
;;
(define-public (force-index-calc-proc count)
  (let ((prev-close  #f)
	(smooth-proc (ema-calc-proc count)))

    (lambda (close volume)
      (let ((ret (and prev-close
		      (smooth-proc (* (or volume 1)
				      (- close prev-close))))))
	(set! prev-close close)
	ret))))

(define-public (force-index-warmup-count count)
  (1+ (ema-warmup-count count)))


(define (force-index-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (start  (series-prev-n-closes-pos
		  parent #:before lo #:count (force-index-warmup-count count)))
	 (array  (series-array parent start hi))
	 (proc   (force-index-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 3)    ;; close
		(array-ref array i 4)))) ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (proc (array-ref array i 3)  ;; close
				      (array-ref array i 4)) ;; volume
		      i)))))

(define-public (force-index-series series count)
  (ohlcv-required "Force Index" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Force") count))
    (series-datatype!  new-series force-index-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   force-index-fill
				   (force-index-warmup-count count)))
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public force-index-datatype
  (datatype-new!
   #:key     'force-index
   #:name    (_ "Force Index")
   #:input   datatype-ohlcv
   #:units   'force-index
   #:hlines  '(0)
   #:manual  (_ "CONTEXT(manual-node): Force Index")
   #:proc    force-index-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'force-index-days
	  #:name    (_ "Smooth Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 2))))
