;; Forecast Oscillator (FOSC).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart fosc)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart linreg)
  #:use-module (chart misc))


;; http://store.traders.com/-v10-c05-forcast-pdf.html
;;     Abstract of TASC May 1992 article, rest for sale.
;;
;; http://trader.online.pl/MSZ/e-w-Chandes_Forecast_Oscillator_I.html
;;     Formula, text of TASC article.
;;
;; http://www.fmlabs.com/reference/ForecastOscillator.htm
;;
;; https://www.stockworm.com/do/community/manual;jsessionid=8F865CC4F15C7E055E0A62A865AF7415?chapter=forecast-oscillator
;;     Sample Oracle (ORCL), year not marked but is 2003.
;;
;; http://www.prophet.net/analyze/popglossary.jsp?studyid=FOSC
;;     Sample chart of Peoplesoft (symbol PSFT maybe, since taken over).
;;
;; http://www.paritech.com/paritech-site/education/technical/indicators/trend/forecast.asp
;;     Sample chart of Dollar Sweets (since taken over, what symbol was it?).
;;
;; http://trader.online.pl/MSZ/e-st-Forecast_Oscillator.html
;;     Trading system by Steve Karnish.
;;


;; Return a procedure which calculates forecast oscillator values over a
;; window of COUNT many points.
;;
;; Each call (PROC x) enters a new point into the window, and the return is
;; the FOSC value up to (and including) that point, or #f if not enough
;; values yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many values
;; preceding the first desired point.  `fosc-warmup-count' below can
;; calculate that.
;;
(define-public (fosc-calc-proc count)
  (define linreg-proc (linreg-calc-proc count))
  (define prev-proc   (prevnth-calc-proc 1))

  (lambda (close)
    (and-let* ((prev-close (prev-proc close)))
      (receive (a b)
	  (linreg-proc prev-close)
	(and a
	     (not (zero? close))
	     (* 100.0 (/ (- close (+ a b)) close)))))))

(define-public fosc-warmup-count linreg-warmup-count)


(define (fosc-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (fosc-calc-proc count)))

    ;; warmup
    (for-each proc (series-prev-n-closes
		    parent #:before lo #:count (fosc-warmup-count count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i))
		      i)))))

(define-public (fosc-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "FOSC %F") count))
    (series-datatype!  new-series fosc-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   fosc-fill (fosc-warmup-count count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public fosc-datatype
  (datatype-new!
   #:key       'fosc
   #:name      (_ "Forecast Oscillator %F")
   #:units     'roc
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Forecast Oscillator")
   #:proc      fosc-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'fosc-days
	  #:type    'integer
	  #:minimum 1
	  #:default 5))))
