;; Fractal Adaptive Moving Average (FRAMA).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart frama)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (srfi srfi-1)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.mesasoftware.com/technicalpapers.htm
;; http://www.mesasoftware.com/Papers/FRAMA.exe
;;     John Ehler's paper on FRAMA (self-extracting zip of MS word document).
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/102005/TradersTips/TradersTips.html
;;     Traders Tips October 2005.
;;
;; http://trader.online.pl/MSZ/e-w-Moving_Average_Fractal_Adaptive_(FAMA).html
;;
;; http://www.tradingsolutions.com/download/tip0510.zip
;;     Formulas.
;;


;;-----------------------------------------------------------------------------
;; FRAMA calc


;; Return a procedure which calculates fractal adaptive moving average
;; dimension values on an accumulated window of COUNT values.
;;
;; Each call (PROC high low) enters a new value into the window, and the
;; return is the alpha up to (and including) that value, or #f if not enough
;; values yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (frama-dimension-calc-proc count)

  ;; HIGH-Q is the high values for each day
  ;; LOW-Q is the corresponding low values for each day
  ;; POINTS is how many entries in HIGH-Q (and LOW-Q likewise), up to COUNT
  ;;
  (define high-q     (make-q))
  (define low-q      (make-q))
  (define points     0)

  (define near-count (quotient count 2))
  (define far-count  (- count (quotient count 2)))

  (lambda (high low)
    (if (>= points count)
	;; drop old point if full
	(begin
	  (deq! high-q)
	  (deq! low-q))
	;; not full, so POINTS increases
	(set! points (1+ points)))

    ;; add new point
    (enq! high-q high)
    (enq! low-q  low)

    ;; only start when window full
    (and (= points count)

	 (let* ((long-avg  (/ (- (apply max (q-list high-q))
				 (apply min (q-list low-q)))
			      count))

		(near-avg  (/ (- (apply max (take (q-list high-q) near-count))
				 (apply min (take (q-list low-q)  near-count)))
			      near-count))

		(far-avg   (/ (- (apply max (drop (q-list high-q) near-count))
				 (apply min (drop (q-list low-q)  near-count)))
			      far-count)))

	   (if (or (zero? near-avg)
		   (zero? far-avg)
		   (zero? long-avg))
	       0
	       (* 1.442695040888963407359924681  ;; 1/log(2)
		  (log (/ (+ near-avg far-avg)
			  long-avg))))))))


;; Return a procedure which calculates fractal adaptive moving average alpha
;; values on an accumulated window of COUNT values.
;;
;; Each call (PROC high low) enters a new value into the window, and the
;; return is the alpha up to (and including) that value, or #f if not enough
;; values yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (frama-alpha-calc-proc count)
  (define dim-proc (frama-dimension-calc-proc count))

  (lambda (high low)
    (and-let* ((dim (dim-proc high low)))
      (minmax-restrict 0.01 1.0 (exp (* -4.6 (1- dim)))))))


;; Return a procedure which calculates a fractal adaptive moving average
;; based on an accumulated window of COUNT values.
;;
;; Each call (PROC high low) enters a new value into the window, and the
;; return is the FRAMA up to (and including) that value, or #f if not enough
;; values yet.
;;
;; A FRAMA is in theory influenced by all preceding data, but
;; `frama-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (frama-warmup-count COUNT) many values, the next call
;; will have an omitted weight of no more than 0.1% of the total (even at
;; the slowest adaptive smoothing).  Omitting 0.1% should be negligable,
;; unless past values are ridiculously bigger than recent ones.
;;

(define-public (frama-calc-proc count)

  (define alpha-proc (frama-alpha-calc-proc count))
  (define ama-proc   (adaptive-ema-calc-proc))

  (lambda (high low)
    (and-let* ((alpha (alpha-proc high low)))
      (ama-proc (arithmetic-mean high low) alpha))))


;; Could get a shorter count by looking at the actual alpha values, since
;; they're likely to be much bigger than the minimum, but doing that could
;; be as much work as just taking a worst case.
;;
(define-public (frama-warmup-count count)
  (+ count slowest-ema-warmup))

(define slowest-ema-warmup
  (ema-warmup-count (ema-alpha->count 0.01)))


;;-----------------------------------------------------------------------------
;; FRAMA dimension series

(define (frama-dimension-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count (1- count)))
	 (array  (series-array parent start hi))
	 (proc   (frama-dimension-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(proc (or (array-ref array i 1) close)    ;; high
	      (or (array-ref array i 2) close)))) ;; low

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array (proc (or (array-ref array i 1) close)  ;; high
				    (or (array-ref array i 2) close)) ;; low
		    i)))))

(define-public (frama-dimension-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "FRAMA dimension") count))
    (series-datatype!  new-series frama-dimension-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   frama-dimension-fill (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public frama-dimension-datatype
  (datatype-new!
   #:key      'frama-dimension
   #:name     (_ "FRAMA dimension")
   #:priority -10
   #:units    'frama-dim
   #:minimum  1
   #:maximum  2
   #:manual   (_ "CONTEXT(manual-node): Fractal Adaptive Moving Average")
   #:proc     frama-dimension-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'frama-days
	  #:type    'integer
	  #:minimum 2
	  #:default 16))))


;;-----------------------------------------------------------------------------
;; FRAMA alpha series

(define (frama-alpha-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count (1- count)))
	 (array  (series-array parent start hi))
	 (proc   (frama-alpha-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(proc (or (array-ref array i 1) close)    ;; high
	      (or (array-ref array i 2) close)))) ;; low

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array (proc (or (array-ref array i 1) close)  ;; high
				    (or (array-ref array i 2) close)) ;; low
		    i)))))

(define-public (frama-alpha-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "FRAMA alpha") count))
    (series-datatype!  new-series frama-alpha-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   frama-alpha-fill (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public frama-alpha-datatype
  (datatype-new!
   #:key      'frama-alpha
   #:name     (_ "FRAMA alpha")
   #:priority -10
   #:units    'ema-alpha
   #:minimum  0
   #:maximum  1
   #:manual   (_ "CONTEXT(manual-node): Fractal Adaptive Moving Average")
   #:proc     frama-alpha-series
   #:parameter-list
   (list     (datatype-param-new
	      #:name    (_ "Days")
	      #:key     'frama-days
	      #:type    'integer
	      #:minimum 2
	      #:default 16))))


;;-----------------------------------------------------------------------------
;; FRAMA series

(define (frama-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (warmup (frama-warmup-count count))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count warmup))
	 (array  (series-array parent start hi))
	 (proc   (frama-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(proc (or (array-ref array i 1) close)    ;; high
	      (or (array-ref array i 2) close)))) ;; low

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array (proc (or (array-ref array i 1) close)  ;; high
				    (or (array-ref array i 2) close)) ;; low
		    i)))))

(define-public (frama-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "FRAMA") count))
    (series-datatype! new-series frama-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts
				  frama-fill (frama-warmup-count count)))
    (series-array!    new-series #f)
    new-series))

(define frama-datatype
  (datatype-new!
   #:key    'frama
   #:name   (_ "FRAMA - Fractal Adaptive MA")
   #:units  #f
   #:manual (_ "CONTEXT(manual-node): Fractal Adaptive Moving Average")
   #:proc   frama-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'frama-days
	  #:type    'integer
	  #:minimum 2
	  #:default 16))))
