;; Futures data helpers.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart futures)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart latest)
  #:use-module (chart misc))



;;-----------------------------------------------------------------------------
;; list stuff

;; partition LST into sets of elements according to what (PROC elem) returns
;; each set has elements for which the PROC returns are all `equal?'
;; the sets are returned in the form of a list of lists
;;
(define-public (partition-equal proc lst)
  (let ((alst '()))  ;; (VAL . LIST)
    (for-each (lambda (elem)
		(let ((val (proc elem)))
		  (cond ((assoc val alst)
			 => (lambda (pair)
			      (set-cdr! pair (cons elem (cdr pair)))))
			(else
			 (set! alst (acons val (list elem) alst))))))
	      lst)
    (reverse! (map! (lambda (pair)
		      (reverse! (cdr pair)))
		    alst))))


;;-----------------------------------------------------------------------------
;; misc helpers

;; convert eighths like "357-6" to "357.75"
;;
;; two decimal places are given even for "-0" and "-4", since 1/4 cent is
;; the normal unit for MGEX wheat (1/8 is for options)
;;
(define-public (dash-eighths->decimals str)
  (if (and (> (string-length str) 2)
	   (eqv? #\- (string-ref str (- (string-length str) 2))))
      (string-append (string-drop-right str 2)
		     (vector-ref #(".00" ".125" ".25" ".375"
				   ".50" ".625" ".75" ".875")
				 (string->number-err
				  (string-take-right str 1))))
      str))


;; return string like "JAN 05" from MDATE
(define-public (mdate->MMM_YY-str mdate)
  (receive-list (year month day)
      (mdate->ymd mdate)
    (format #f "~a ~2,'0d" (vector-ref #(#f "JAN"
					    "FEB"
					    "MAR"
					    "APR"
					    "MAY"
					    "JUN"
					    "JUL"
					    "AUG"
					    "SEP"
					    "OCT"
					    "NOV"
					    "DEC") month)
	    (modulo year 100))))


;;-----------------------------------------------------------------------------
;; mdate parsing stuff

;; return mdate for STR like "05/02" or "2005/02"
(define-public (y/m-str->mdate str)
  (receive-list (year month)
      (string-split str #\/)
    (ymd->mdate (yyy-str->year year)
		(month-str->month month)
		1)))

;; return mdate for STR like "Apr-05"
(define-public (m/y-str->mdate str)
  (receive-list (month year)
      (string-tokenize str date-part-char-set)
    (ymd->mdate (yyy-str->year year)
		(month-str->month month)
		1)))

;; return mdate for STR like "200504"
(define-public (yyyymm-str->mdate str)
  (ymd->mdate (string->number-err (substring str 0 4))
	      (month-str->month (substring str 4 6))
	      1))


;;-----------------------------------------------------------------------------
;; latest records

;; LATEST-LIST is a list of latest records.
;; Return that list with new records for the front month of each commodity
;; consed onto the start.
;;
;; LATEST-LIST can include some non-futures records (with #f for
;; latest-contract-mdate), such entries are preserved, but ignored.
;;
(define-public (latest-list-add-front-months latest-list)
  (let ((alst '())) ;; elems (COMMODITY . LATEST-RECORD)
    (for-each (lambda (latest)
		(if (latest-contract-mdate latest)
		    (let* ((commodity (chart-symbol-commodity
				       (latest-symbol latest)))
			   (elem      (assoc commodity alst)))
		      (if elem
			  (if (< (latest-contract-mdate latest)
				 (latest-contract-mdate (cdr elem)))
			      (set-cdr! elem latest))
			  (set! alst (acons commodity latest alst))))))
	      latest-list)

    (append! (map! (lambda (elem)
		     (let* ((latest    (record-copy (cdr elem)))
			    (commodity (car elem))
			    (symbol    (latest-symbol latest)))
		       (latest-symbol! latest (string-append
					       commodity
					       (chart-symbol-suffix symbol)))
		       latest))
		   alst)
	     latest-list)))


;;-----------------------------------------------------------------------------

(define (chart-symbol-front-first<? x y)
  (and (equal? (chart-symbol-commodity x)
	       (chart-symbol-commodity y))
       (chart-symbol-mdate x)
       (not (chart-symbol-mdate y))))

(define (symbol-list-remove-specific symbol-list commodity mdate)
  (remove (lambda (symbol)
	    (and (equal? commodity
			 (chart-symbol-commodity symbol))
		 (equal? mdate
			 (chart-symbol-mdate symbol))))
	  symbol-list))

;; return the mdate of the latest month in the dividends for SYMBOL, or #f
;; if none
(define-public (database-dividend-mdate symbol)
  (and-let* ((series (database-read-series symbol)))
    (let ((elem (find-maximum (lambda (pair)
				(and (pair? (cdr pair)) ;; 'month key
				     (eq? 'month (second pair))
				     (car pair)))       ;; dividend date
			      (series-dividends series))))
      (and elem
	   (third elem)))))

(define*-public (futures-download-by-specific symbol-list
					      #:key
					      download-proc
					      avail-tdate-proc
					      (initial-proc (lambda (symbol)
							      #f)))

  (define (front symbol)
    (let* ((commodity   (chart-symbol-sans-dot symbol))
	   (avail-tdate (avail-tdate-proc symbol))
	   (avail-mdate (1+ (mdate-today (symbol-timezone symbol))))
	   (start-tdate (download-start-tdate symbol
					      #:initial (initial-proc symbol)))
	   (mdate       (or (database-dividend-mdate symbol)
			    (tdate->mdate start-tdate))))
      (if (>= avail-tdate start-tdate)
	  (begin
	    (set! symbol-list (symbol-list-remove-specific symbol-list
							   commodity mdate))
	    (download-proc (list symbol) commodity mdate)
	    (set! mdate (1+ mdate))))

      ;; when current front month is this month or next, look ahead for
      ;; possible rollover
      ;;
      ;; must download if front is next month because expiry can be in the
      ;; preceding month, eg. CL.NYMEX is 3 days before the 25th of the
      ;; preceding
      ;;
      (let more ((attempt 1))
	(if (and (<= mdate avail-mdate)
		 (>= avail-tdate (download-start-tdate
				  symbol #:initial (initial-proc symbol))))
	    (begin
	      (set! symbol-list (symbol-list-remove-specific symbol-list
							     commodity mdate))
	      (if (download-proc (list symbol) commodity mdate)
		  (set! attempt 0))
	      (set! mdate (1+ mdate))
	      (if (>= attempt 6)
		  (download-message (format #f (_ "Seems to be no data for ~a")
					    symbol))
		  (more (1+ attempt))))))))

  ;; download for a specific month SYMBOL
  (define (specific symbol)
    (if (>= (avail-tdate-proc symbol) (download-start-tdate symbol))
	(download-proc (list symbol)
		       (chart-symbol-commodity symbol)
		       (chart-symbol-mdate symbol))))

  (set! symbol-list (stable-sort symbol-list chart-symbol-front-first<?))
  (let more ()
    (if (not (null? symbol-list))
	(let ((symbol (first symbol-list)))
	  (set! symbol-list (cdr symbol-list))
	  (if (<= (download-start-tdate symbol)
		  (avail-tdate-proc symbol))
	      (if (chart-symbol-mdate symbol)
		  (specific symbol)
		  (front symbol)))
	  (more)))))
