;; Gopalakrishnan Range Index (GAPO).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart gapo)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://trader.online.pl/MSZ/e-w-Gopalakrishnan_Range_Index_GAPO.html
;;     Formula, sample chart for WIG something.
;;
;; http://www.tradingsolutions.com/download/tip0105.zip
;;     Formula.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/012001/TradersTips/TradersTips.html
;;     Traders tips formulas for several packages.  Small sample chart of
;;     ^IXIC nasdaq composite for 2000.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/012001/Abstracts_new/Jayanthi/Jayanthi.html
;;     First few paras of article.
;;
;; http://store.traders.com/v19114gopran.html
;;     Article for sale, first para only.


;; Return a procedure which calculates GAPO values over a window of COUNT
;; many points.
;;
;; Each call (PROC high low) enters a new point into the window, and the
;; return is the GAPO value up to (and including) that point, or #f if not
;; enough values yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many values
;; preceding the first desired.
;;
(define-public (gapo-calc-proc count)

  ;; HIGH-Q is the high values for each day
  ;; LOW-Q is the corresponding low values for each day
  ;; POINTS is how many entries in each of HIGH-Q and LOW-Q
  ;;
  (let ((high-q     (make-q))
	(low-q      (make-q))
	(points     0)
	(points-min (ceiling-exact (* indicator-minimum-fraction count)))
	(scale      1))   ;; 1/log(0) initially

    (lambda (high low)

      (if (>= points count)
	  ;; full, drop old point
	  (begin
	    (deq! high-q)
	    (deq! low-q))
	  ;; not full, so POINTS increases
	  (begin
	    (set! points (1+ points))
	    (set! scale (/ (log points)))))

      ;; add new point
      (enq! high-q high)
      (enq! low-q  low)

      (and (>= points points-min)
	   (let* ((high (apply max (q-list high-q)))
		  (low  (apply min (q-list low-q)))
		  (diff (- high low)))
	     (and (not (zero? diff)) ;; avoid -inf.0
		  (* scale (log diff))))))))

(define (gapo-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count (1- count)))
	 (array  (series-array parent start hi))
	 (gapo   (gapo-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(gapo (or (array-ref array i 1) close)     ;; high
	      (or (array-ref array i 2) close))))  ;; low

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array (gapo (or (array-ref array i 1) close)  ;; high
				    (or (array-ref array i 2) close)) ;; low
		    i)))))

(define-public (gapo-series series count)
  (ohlcv-required "GAPO" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "GAPO") count))
    (series-datatype!  new-series gapo-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts gapo-fill
							   (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public gapo-datatype
  (datatype-new!
   #:key       'gapo
   #:name      (_ "GAPO")
   #:input     datatype-ohlcv
   #:priority  -10
   #:units     'gapo
   #:minimum   0
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Gopalakrishnan Range Index")
   #:proc      gapo-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'gapo-days
	  #:type    'integer
	  #:minimum 1
	  #:default 5))))
