;; DNS gethost lookup.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart gethost)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (chart http)
  #:use-module (chart misc))


;; This module implements an asynchronous DNS lookup with `gethost'.  In
;; Guile 1.8 it uses a thread, in 1.6 it uses a child process.  In both
;; cases there's no parallelism, just one query is in progress at any time.
;; The important point is that it doesn't block the GUI during a query.
;;
;; The thread or child process is kept alive to run all queued queries, then
;; killed.  This is mainly for the benefit of the child process case, since
;; a fork could be expensive.  A few lookups can come together when latest
;; quote downloads are spawned for several symbols on different exchanges.
;;
;; Killing the thread or process when finished frees memory, and that could
;; be a lot for a subprocess.  Queries are cached, so the lookups of
;; interest will probably come early, then nothing further needed, so it'd
;; be wasteful to keep something around permanently.
;;
;; There are, of course, several C-level asynch DNS libraries, but don't
;; really want to make any of them a requirement.


;;-----------------------------------------------------------------------------
;; misc

;; close PORT if it's not already closed
(define-public (close-port-maybe port)
  (or (port-closed? port)
      (close-port port)))

;; close the ports of the PIPE pair, if they're not already closed
(define (pipe-close-maybe pipe)
  (close-port-maybe (car pipe))
  (close-port-maybe (cdr pipe)))


;;-----------------------------------------------------------------------------
;; choke
;;
;; A choke point lets multiple threads queue waiting on an arbitrary
;; condition.  The choose procedure selects the next which is able to run,
;; if any.

(define (delq-must elem lst)
  (or (memq elem lst)
      (format (current-error-port)
	      "chart: oops, choke element not in list: ~a ~a\n" elem lst))
  (delq elem lst))

(let-toplevel ((rec-type (make-record-type "choke" '(choose emptied
							    pending running))))
  (define rec-new      (record-constructor rec-type))
  (define rec-choose   (rec-accessor rec-type 'choose))
  (define rec-emptied  (rec-accessor rec-type 'emptied))
  (define rec-pending  (rec-accessor rec-type 'pending))
  (define rec-running  (rec-accessor rec-type 'running))
  (define rec-pending! (rec-modifier rec-type 'pending))
  (define rec-running! (rec-modifier rec-type 'running))

  (define (try-choose choke)
    (if (null? (rec-pending choke))
	(if (null? (rec-running choke))
	    ((rec-emptied choke)))

	(and-let* ((elem ((rec-choose choke)
			  (rec-pending choke) (rec-running choke))))

	  (rec-pending! choke (delq-must elem (rec-pending choke)))
	  (rec-running! choke (append (rec-running choke) (list elem)))
	  (c-main-enq! try-choose choke)
	  ((cdr elem) noop))))

  (define*-public (choke-new #:key choose (emptied noop))
    (rec-new choose emptied '() '()))

  (define-public (choke-call choke data thunk)

    (if c-main-goto-top
	(let ((elem (cons data #f)))
	  (job-pause
	   (lambda (cont)
	     (set-cdr! elem cont)
	     (rec-pending! choke (append (rec-pending choke) (list elem)))
	     (c-main-enq! try-choose choke))
	   (lambda ()
	     (rec-pending! choke (delq-must elem (rec-pending choke)))
	     (c-main-enq! try-choose choke)))
	  (set-cdr! elem #f) ;; gc continuation

	  (call-with-cleanup
	      (lambda ()
		(rec-running! choke (delq-must elem (rec-running choke)))
		(c-main-enq! try-choose choke))
	    thunk))

	(thunk))))


;;-----------------------------------------------------------------------------
;; subproc
;;
;; Variations between child process for 1.6 and threads for 1.8.  Mostly
;; expressed in the threads functions style, with the process case adapting
;; to that.

(define subproc-use-threads
  (and (provided? 'threads)
       (not (string=? "1.6" (effective-version)))
       ;; Disabled: threads are a bit dodgy as of guile 1.8.2, not worth
       ;; taking the chance until it improves
       #f))

;; fork a child process to run THUNK
;; in the parent this function returns the PID, in the child it doesn't return
(define (call-with-new-fork thunk)
  (let ((pid (primitive-fork)))
    (if (= 0 pid)
	(begin
	  ;; child
	  (thunk)
	  (primitive-_exit 0)))
    ;; parent
    pid))

(define call-with-new-subproc
  (if subproc-use-threads
      call-with-new-thread
      call-with-new-fork))

(define join-subproc
  (if subproc-use-threads
      join-thread
      waitpid))

;; `close-port', but only when using a child process
;; this allows the "wrong" ends of the pipes to be closed in the parent or
;; child, but left alone when using threads
(define subproc-close-port
  (if subproc-use-threads
      noop
      close-port))

      
;;-----------------------------------------------------------------------------
;; gethost
;;
;; A subprocess is used to do the lookup, so it's non-blocking, even in
;; guile 1.6 which doesn't have concurrent threads.  The info is cached once
;; found, so each name is only looked up once.
;;
;; The only need for a pipe in the thread case is something the Gtk main
;; loop can wait on, the actual data each way could go through a variable.
;; But it's easiest to let the thread do the same as the child process case.
;;
;; Everything still works when not using c-main.  c-gethost initiates the
;; query and blocks in the read until getting the result.  It's a bit
;; wasteful to go through the read/write hoops, but it works fine.


;; alist (HOSTNAME . HOST-OBJECT)
;; only successful returns are entered
(define cache '())

;; pipes each way to the child
(define child->parent #f)
(define parent->child #f)

;; #f when child not running, child pid number or thread object when running
(define child-id #f)

(define (child-proc)

  ;; so our read sees eof if the parent dies
  (subproc-close-port (cdr parent->child))  ;; output

  ;; so our write gets SIGPIPE if the parent dies
  (subproc-close-port (car child->parent))  ;; input

  (do ((hostname (read (car parent->child))
		 (read (car parent->child))))
      ((eof-object? hostname))

    (catch #t
      (lambda ()
	(write (list 'ok (gethost hostname))
	       (cdr child->parent)))
      (lambda args
	(write (list 'error args)
	       (cdr child->parent))))
    (display "\r\n" (cdr child->parent))
    (force-output (cdr child->parent))))

(define (start-child)
  (if (not child-id)
      (begin
	(set! child->parent (pipe))
	(set! parent->child (pipe))

	(set! child-id (call-with-new-subproc child-proc))

	;; so our read sees eof if the child dies unexpectedly
	(subproc-close-port (cdr child->parent)) ;; output

	;; so our write gets sigpipe if the child dies unexpectedly (not
	;; sure if that's a good thing though)
	(subproc-close-port (car parent->child)) ;; input

	;; non-blocking in http-read-to-crlf below
	(fcntl-set-nonblock (car child->parent)))))

(define (stop-child)
  (if child-id
      (begin
	(close-port (cdr parent->child))  ;; output side
	(join-subproc child-id)           ;; child will notice eof
	(pipe-close-maybe parent->child)  (set! parent->child #f)
	(pipe-close-maybe child->parent)  (set! child->parent #f)
	(set! child-id #f))))

(define (run-one hostname)
  (if (not child-id)
      (start-child))

  ((fluid-ref http-status-callback) (string-append (_ "lookup")
						   " " hostname))
  (write hostname (cdr parent->child))
  (force-output (cdr parent->child))

  (let ((str (http-read-to-crlf (car child->parent))))

    ;; either from the cache or from the str just received
    (receive-list (key data)
	(read (open-input-string str))
      (if (eq? key 'ok)
	  (begin
	    (set! cache (assoc-set! cache hostname data))
	    data)
	  (apply throw data)))))

(define (gethost-choose pending running)
  (and (null? running)
       (first pending)))

(define gethost-choke (choke-new #:choose gethost-choose
				 #:emptied stop-child))

;; return `gethost' info for HOSTNAME, or throw an error
;;
(define-public (c-gethost hostname)
  (or (assoc-ref cache hostname)
      (choke-call gethost-choke hostname
	(lambda ()
	  ;; look at cache again in case earlier queued job got it
	  (or (assoc-ref cache hostname)
	      (run-one hostname))))))
