/* Chart gettext for Guile 1.6.

   Copyright 2005, 2006, 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>     /* for NULL */

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <libguile.h>

#include "gettext.h"

#include "misc.h"
#include "compat.h"


/*---------------------------------------------------------------------------*/
/* gettext in guile 1.8 style, made available in guile 1.6 */

#if ! HAVE_SCM_GETTEXT

SCM_DEFINE_PUBLIC (chart_gettext, "gettext", 1, 1, 0,
                   (SCM str, SCM domain),
                   "Return the translation of @var{str}.")
#define FUNC_NAME s_chart_gettext
{
  SCM_VALIDATE_STRING (SCM_ARG1, str);
  if (SCM_UNBNDP (domain))
    return scm_from_locale_string (gettext (SCM_STRING_CHARS (str)));
  else
    return scm_from_locale_string (dgettext (SCM_STRING_CHARS (domain),
                                             SCM_STRING_CHARS (str)));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_ngettext, "ngettext", 3, 0, 0,
            (SCM msg, SCM plural, SCM n),
	    "Return the translation of @var{msg}/@var{plural} according to\n"
	    "the integer @var{n}.")
#define FUNC_NAME s_chart_ngettext
{
  long nn;
  SCM_VALIDATE_STRING (SCM_ARG1, msg);
  SCM_VALIDATE_STRING (SCM_ARG2, plural);
  SCM_VALIDATE_INUM_COPY (SCM_ARG3, n, nn);
  return scm_from_locale_string (ngettext (SCM_STRING_CHARS (msg),
                                           SCM_STRING_CHARS (plural), nn));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_bindtextdomain, "bindtextdomain", 1, 1, 0,
            (SCM domain, SCM dir),
	    "With just a @var{domain} argument, return the current directory\n"
	    "searched for translations in that domain.\n"
	    "\n"
	    "With a @var{dir} argument, set the directory to search for\n"
	    "translations in @var{domain}.")
#define FUNC_NAME s_chart_bindtextdomain
{
  const char  *t;
  SCM_VALIDATE_STRING (SCM_ARG1, domain);

  if (SCM_UNBNDP (dir))
    {
      t = bindtextdomain (SCM_STRING_CHARS (domain), NULL);
      return (t == NULL ? SCM_BOOL_F : scm_from_locale_string (t));
    }
  else
    {
      SCM_VALIDATE_STRING (SCM_ARG2, dir);
      t = bindtextdomain (SCM_STRING_CHARS (domain), SCM_STRING_CHARS (dir));
      if (t == NULL)
        SCM_SYSERROR;
      return scm_from_locale_string (t);
    }
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_bind_textdomain_codeset, "bind-textdomain-codeset", 1, 1, 0,
            (SCM domain, SCM codeset),
	    "With just a @var{domain} argument, return the currrent codeset\n"
	    "used for translations in that domain.\n"
	    "\n"
	    "With a @var{codeset} argument, set the codeset to use for\n"
	    "translations in @var{domain}.")
#define FUNC_NAME s_chart_bind_textdomain_codeset
{
  const char  *t;
  SCM_VALIDATE_STRING (SCM_ARG1, domain);

  if (SCM_UNBNDP (codeset))
    {
      t = bind_textdomain_codeset (SCM_STRING_CHARS (domain), NULL);
      return (t == NULL ? SCM_BOOL_F : scm_from_locale_string (t));
    }
  else
    {
      SCM_VALIDATE_STRING (SCM_ARG2, codeset);
      t = bind_textdomain_codeset (SCM_STRING_CHARS (domain),
                                   SCM_STRING_CHARS (codeset));
      if (t == NULL)
        SCM_SYSERROR;
      return scm_from_locale_string (t);
    }
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_textdomain, "textdomain", 0, 1, 0,
            (SCM domain),
	    "With no arguments, return the current default translation\n"
	    "domain.\n"
	    "\n"
	    "With an argument, set the default translation domain to the\n"
	    "string @var{domain}.")
#define FUNC_NAME s_chart_textdomain
{
  const char *t;

  if (SCM_UNBNDP (domain))
    {
      t = textdomain (NULL);
      return (t == NULL ? SCM_BOOL_F : scm_from_locale_string (t));
    }
  else
    {
      SCM_VALIDATE_STRING (SCM_ARG1, domain);
      t = textdomain (SCM_STRING_CHARS (domain));
      if (t == NULL)
        SCM_SYSERROR;
      return scm_from_locale_string (t);
    }
}
#undef FUNC_NAME

#endif /* HAVE_SCM_GETTEXT */


/*---------------------------------------------------------------------------*/

void
chart_gettext_init (void)
{
#ifndef SCM_MAGIC_SNARFER
#include "gettext.x"
#endif
}
