;; Ghana Stock Exchange setups and data downloading.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ghana)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (ghana-symbol? symbol)
  (string-suffix? ".GHA" symbol))

(define-public timezone-accra
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Accra")
		      '("Africa/Accra"
			;; no daylight savings post 1942
			"GMT")
		      "GMT"))))

(symbol-timezone! ghana-symbol? (_ "Ghana") timezone-accra)
(source-help! ghana-symbol? (_ "CONTEXT(manual-node): Ghana Stock Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - company info

;; eg. http://www.gse.com.gh/listedir/detail.asp?symbol=ABL
;; This is normally in a frame, would like to show it that way, but how?
;;
(weblink-handler! ghana-symbol?
  (lambda (symbol)
    (list (_ "GSE _Company Information")
	  (_ "Open web browser at the Ghana Stock Exchange information page for this company")
	  (string-append "http://www.gse.com.gh/listedir/detail.asp?symbol="
			 (chart-symbol-sans-dot symbol)))))


;;-----------------------------------------------------------------------------
;; whole day parse

;; return values (TDATE ROW-LIST) with row-list elements (SYMBOL ISIN ...)
;;
(define (daily-parse body)

  (let* (;; eg. "24/04/2006\n\t\t</font>&nbsp;Trading Results"
	 (m        (must-match (string-match "([0-9]+/[0-9]+/[0-9][0-9][0-9][0-9])([ \t\n]*(<[^>]*>|&nbsp;))*[ \t\n]*Trading Results" body)))
	 (tdate  (d/m/y-str->tdate (match:substring m 1)))

	 ;; the last <table> before "Year High" heading
	 (m        (must-match (string-match ".*(<table).*Year High" body)))
	 (pos      (match:start m 1))
	 (row-list (html-table-rows body pos)))

    (let ((headings (first row-list)))
      (or (equal? headings
		  '(""
		    "ISIN"
		    "Year High (&cent;)"
		    "Year Low (&cent;)"
		    "Previous Closing Price (&cent;)"
		    "Opening Price (&cent;)"
		    "Closing Price (&cent;)"
		    "Price Change (&cent;)"
		    "Closing Bid (&cent;)"
		    "Closing Offer (&cent;)"
		    "Total Shares Traded"))
	  (error "GSE unrecognised headings:" headings)))
    (set! row-list (cdr row-list))

    ;; sub-heading
    (if (equal? (first row-list) '("First List"))
	(set! row-list (cdr row-list)))

    ;; 18jan07 had an empty row, delete that based on empty symbol field
    (set! row-list (remove! (lambda (row)
			      (string-null? (first row)))
			    row-list))

    (for-each (lambda (row)
		(let ((symbol (first row)))

		  ;; "TBL (The Gambia)" turned into "TBL"
		  (and-let* ((m (regexp-exec #,(regexp "^(.*) \\(.*\\)$")
					     symbol)))
		    (set! symbol (match:substring m 1)))

		  ;; trailing dot on "SBC PREF."
		  (set! symbol (string-trim-right symbol #\.))

		  (set-first! row symbol)))
	      row-list)

    (values tdate row-list)))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the daily data page
;;
;;     http://www.gse.com.gh/marketstat/equitiesdef.asp
;;
;; There's no etag or last-modified to avoid re-downloading if not yet
;; updated to the new day.
;;
;; An alternative would be the individual data pages like
;;
;;     http://www.gse.com.gh/marketstat/officiallist_details.asp?Symbol=ABL
;;
;; They're not available for all symbols, but those which have them give the
;; company name as well as the data.  The individual pages are about 13k
;; each where the whole is 23k, so if more than one symbol is of interest
;; then the whole is best.


(define (body->latest-list body)
  (receive (tdate row-list)
      (daily-parse body)
    (map-rows! (lambda (symbol isin year-high year-low prev open close
			       change bid offer volume)
		 (latest-new #:symbol      (string-append symbol ".GHA")
			     #:quote-tdate tdate
			     #:bid         bid
			     #:offer       offer
			     #:last-tdate  tdate
			     #:open        open
			     #:last        close
			     #:change      change
			     #:volume      volume
			     #:source      'ghana))
	       row-list)))

(define (ghana-latest-get symbol-list extra-list proc)
  (receive (headers body)
      (http-request "http://www.gse.com.gh/marketstat/equitiesdef.asp"
		    #:want-ok #t)
    (proc (body->latest-list body))

    ;; show any not covered
    (proc (map latest-new-unknown
	       (lset-difference string=? symbol-list
				(map latest-symbol (body->latest-list body)))))

    ;; and update database at the same time
    (ghana-process-download '() body)))

;; return list (ADATE TIME) for expected latest quote
(define (ghana-quote-adate-time . dummy-symbol)
  (list (tdate->adate (ghana-available-tdate)) #f))

(latest-handler! #:selector   ghana-symbol?  
		 #:handler    ghana-latest-get
		 #:adate-time ghana-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - names
;;
;; This uses the info pages like
;;
;;     http://www.gse.com.gh/marketstat/officiallist_details.asp?Symbol=ABL
;;
;; These aren't available for all stocks.

(define (process-officiallist-name symbol body)

  ;; eg. "    <td><strong> <span class="just">2241 <strong>&nbsp;Trading Session </strong></span>for:<font color="#D98713"> Accra Brewery Company Ltd.</font></strong></td>"

  (let* ((m    (must-match (regexp-exec (make-regexp "Trading Session.*for:(.*)" regexp/newline) body)))
	 (name (string-trim-both (html-text-extract (match:substring m 1)))))

    (database-call-modify-data symbol
      (lambda (series)
	(series-name! series name)
	(series-info-add! series 'name-time (current-time))))))

(define (ghana-download-names symbol-list)
  (for-each
   (lambda (symbol)
     (if (download-want-name? symbol)
	 (begin
	   (download-status (_ "GSE") (_ "name") symbol)
	   (receive (headers body)
	       (http-request (string-append "http://www.gse.com.gh/marketstat/officiallist_details.asp?Symbol=" (chart-symbol-sans-dot symbol)))

	     (let ((status-code (http-headers-status-code headers)))
	       (cond ((http-status-code-ok? status-code)
		      (process-officiallist-name symbol body))

		     ((= 404 status-code)
		      ;; note attempt
		      (database-call-modify-data symbol
			(lambda (series)
			  (series-info-add! series
					    'name-time (current-time)))))

		     (else
		      (http-headers-throw headers))))))))
   symbol-list))


;;-----------------------------------------------------------------------------
;; download by whole day files
;;

(define (ghana-process-download symbol-list body)
  (receive (tdate row-list)
      (daily-parse body)
    (set! symbol-list (download-also symbol-list
				     #:selector    ghana-symbol?
				     #:start-tdate tdate
				     #:end-tdate   tdate))
    (download-process
     #:module        (_ "GSE")
     #:symbol-list   symbol-list
     #:currency      "GHC"
     #:lo            tdate
     #:hi            tdate
     #:row-list
     (map-rows! (lambda (symbol isin year-high year-low prev open close
				change bid offer volume)
		  (list #:tdate     tdate
			#:commodity symbol
			#:open      open
			#:close     close
			#:volume    volume))
		row-list))

    symbol-list))

;; compare the cdrs of X and Y with <
(define-public (cdr< x y)
  (< (cdr x) (cdr y)))

;; parse out the session code numbers from the current data page
;; the return is a list of (CODE . TDATE), sorted by ascending tdate
(define (daily-parse-sessions body start-tdate)
  (sort!
   (filter-map (lambda (m)
		 (define tdate (d/m/y-str->tdate (match:substring m 2)))
		 (and (>= tdate start-tdate)
		      (cons (match:substring m 1) tdate)))
	       (must-match (list-matches "<option +value=\"([0-9]+)\"> *([0-9]+/[0-9]+/[0-9][0-9][0-9][0-9]) *</option>" body)))
   cdr<))

(define (ghana-download-data symbol-list)
  (define start-tdate (apply min (map download-start-tdate symbol-list)))

  (download-status (_ "GSE") (_ "latest data"))
  (receive (newest-headers newest-body)
      (http-request "http://www.gse.com.gh/marketstat/equitiesdef.asp"
		    #:want-ok #t)

    ;; each available day
    (for-each (lambda (elem)
		(download-status (_ "GSE") (_ "data")
				 (tdate-range-string (cdr elem) (cdr elem)))
		(receive (headers body)
		    (http-request (string-append "http://www.gse.com.gh/marketstat/equitiesdef.asp?session=" (car elem))
				  #:want-ok #t)
		  (set! symbol-list (ghana-process-download symbol-list
							    body))))

	      (daily-parse-sessions newest-body start-tdate))

    (ghana-process-download symbol-list newest-body)

    ;; and also record all as latest
    (latest-handler-callback (body->latest-list newest-body))))

(define (ghana-backto want-tdate symbol-list)
  (download-message
   (string-append
    (_ "GSE") ": "
    (_ "no further historical data available"))))

(define (ghana-download symbol-list)
  (ghana-download-data symbol-list)
  (ghana-download-names symbol-list))

;; Return tdate for available daily report.
;;
;; Trading is from 10am to about midday, not sure when the new day's data
;; will appear, have seen it still not available after midnight on the
;; trading day, so let's try from 9am the following day.
;; 
(define (ghana-available-tdate)
  (1- (tdate-today-after 9 0 (timezone-accra))))

(download-handler! #:selector   ghana-symbol?
		   #:update     ghana-download
		   #:avail-proc ghana-available-tdate
		   #:backto     ghana-backto)
