;; Emulated Gnome UI functions.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; The idea in this file is to emulate using Gtk those few bits of Gnome
;; that Chart would like to use.  If actual Gnome bindings exist (they don't
;; yet) then they can be used, otherwise the emulation.  GnomeApp and
;; gnome-help seem like good ways to work in with Gnome conventions.
;;
;; Not sure about the gnome-app-helper stuff.  Don't really know what
;; wrapped versions of those functions will look like.  Maybe GtkUIManager
;; does enough anyway, or do something direct to make menu widgets more
;; easily available for user manipulation.

(define-module (chart gnome-ui)
  #:use-module (ice-9 and-let-star)
  #:use-module (gtk gtk)
  #:use-module (chart gnome)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; GnomeApp

(gtk-class-new 'GtkWindow "GnomeApp")

(let-toplevel ((rec-type
		(make-record-type "gnome-app"
				  '(vbox contents-hbox
					 menubar toolbar toolbar-id contents
					 statusbar))))
  (define rec-new           (record-constructor rec-type))
  (define rec-vbox          (rec-accessor rec-type 'vbox))
  (define rec-contents-hbox (rec-accessor rec-type 'contents-hbox))
  (define rec-menubar       (rec-accessor rec-type 'menubar))
  (define rec-toolbar       (rec-accessor rec-type 'toolbar))
  (define rec-toolbar-id    (rec-accessor rec-type 'toolbar-id))
  (define rec-contents      (rec-accessor rec-type 'contents))
  (define rec-statusbar     (rec-accessor rec-type 'statusbar))
  (define rec-menubar!      (rec-modifier rec-type 'menubar))
  (define rec-toolbar!      (rec-modifier rec-type 'toolbar))
  (define rec-toolbar-id!   (rec-modifier rec-type 'toolbar-id))
  (define rec-contents!     (rec-modifier rec-type 'contents))
  (define rec-statusbar!    (rec-modifier rec-type 'statusbar))

  (define-public (gnome-app-new appname title)
    (let ((app           (gtk-widget-new 'GnomeApp #:type 'toplevel))
	  (vbox          (gtk-vbox-new #f 0))
	  (contents-hbox (gtk-hbox-new #f 0)))
      (set! (c-prop app)
	    (rec-new vbox contents-hbox #f #f #f #f #f))
      (gtk-window-set-title app title)
      (gtk-signal-connect app "delete_event" (lambda (event) (gtk-exit)))

      (gtk-container-add app vbox)
      (gtk-container-add vbox contents-hbox)
      app))

  (define-public (gnome-app-menubar app)
    (rec-menubar (c-prop app)))
  (define-public (gnome-app-toolbar app)
    (rec-toolbar (c-prop app)))
  (define-public (gnome-app-contents app)
    (rec-contents (c-prop app)))
  (define-public (gnome-app-statusbar app)
    (rec-statusbar (c-prop app)))

  (define-public (gnome-app-set-menus app menubar)
    (let* ((rec  (c-prop app))
	   (vbox (rec-vbox rec)))
      (and-let* ((old (rec-menubar rec)))
	(gtk-container-remove vbox old))
      (rec-menubar! rec menubar)
      (gtk-box-pack-start vbox menubar #f #f 0)
      (gtk-box-reorder-child vbox menubar 0)))

  (define-public (gnome-app-set-contents app contents)
    (let* ((rec  (c-prop app))
	   (hbox (rec-contents-hbox rec)))
      (and-let* ((old (rec-contents rec)))
	(gtk-container-remove hbox old))
      (rec-contents! rec contents)
      (gtk-box-pack-end hbox contents #t #t 0)))

  (define-public (gnome-app-set-statusbar app statusbar)
    (let* ((rec  (c-prop app))
	   (vbox (rec-vbox rec)))
      (and-let* ((old (rec-statusbar rec)))
	(gtk-container-remove vbox old))
      (rec-statusbar! rec statusbar)
      (gtk-box-pack-end vbox statusbar #f #f 0)))

  (define-public (gnome-app-set-toolbar app toolbar)
    (let* ((rec  (c-prop app))
	   (vbox (rec-vbox rec))
	   (hbox (rec-contents-hbox rec)))
      (and-let* ((old (rec-toolbar rec)))
	(gtk-signal-disconnect toolbar (rec-toolbar-id rec))
	(gtk-container-remove (gtk-widget-get-parent old) old))
      (rec-toolbar! rec toolbar)
      (case (gtk-object-get toolbar #:orientation)
	((horizontal)
	 (gtk-box-pack-start vbox toolbar #f #f 0)
	 (gtk-box-reorder-child vbox toolbar (if (rec-menubar rec) 1 0)))
	((vertical)
	 (gtk-box-pack-start hbox toolbar #f #f 0)))
      (rec-toolbar-id! rec (gtk-signal-connect toolbar "orientation-changed"
			     (lambda (orientation)
			       (gnome-app-set-toolbar app toolbar)))))))


;;-----------------------------------------------------------------------------
;; GnomeAppBar

(gtk-class-new 'GtkHBox "GnomeAppBar")

(let-toplevel ((rec-type (make-record-type "gnome-appbar"
					   '(prog stat
						  stat-gen-id
						  stat-set-id))))
  (define rec-new         (record-constructor rec-type))
  (define rec-stat        (rec-accessor rec-type 'stat))
  (define rec-stat-gen-id (rec-accessor rec-type 'stat-gen-id))
  (define rec-stat-set-id (rec-accessor rec-type 'stat-set-id))

  (define-public (gnome-appbar? widget)
    (gtk-check-type widget 'GnomeAppBar))

  (define-public (gnome-appbar-new has-progress has-status interactivity)
    ;; default box not homogenous, and spacing 0
    (let ((appbar      (gtk-widget-new 'GnomeAppBar))
	  (prog        #f)
	  (stat        #f)
	  (stat-gen-id #f)
	  (stat-set-id #f))
      (if has-progress
	  (begin
	    (set! prog (gtk-progress-bar-new))
	    (gtk-box-pack-start appbar prog #t #t 0)))
      (if has-status
	  (begin
	    (set! stat (gtk-statusbar-new))
	    (gtk-box-pack-start appbar stat #t #t 0)
	    (set! stat-gen-id
		  (gtk-statusbar-get-context-id stat "gnome-appbar-gen"))
	    (set! stat-set-id
		  (gtk-statusbar-get-context-id stat "gnome-appbar-set"))))
      (set! (c-prop appbar)
	    (rec-new prog stat stat-gen-id stat-set-id))
      appbar))

  (define-public gnome-appbar-get-progress
    (c-prop-rec-accessor rec-type 'prog))

  (define-public (gnome-appbar-push appbar string)
    (let* ((rec  (c-prop appbar))
	   (stat (rec-stat rec)))
      (gtk-statusbar-pop  stat (rec-stat-set-id rec))
      (gtk-statusbar-push stat (rec-stat-gen-id rec) string)))

  (define-public (gnome-appbar-pop appbar)
    (let* ((rec  (c-prop appbar))
	   (stat (rec-stat rec)))
      (gtk-statusbar-pop stat (rec-stat-set-id rec))
      (gtk-statusbar-pop stat (rec-stat-gen-id rec))))

  (define-public (gnome-appbar-set-status appbar string)
    (let* ((rec         (c-prop appbar))
	   (stat        (rec-stat        rec))
	   (stat-set-id (rec-stat-set-id rec)))
      (gtk-statusbar-pop  stat stat-set-id)
      (gtk-statusbar-push stat stat-set-id string))))
