;; Emulated Gnome library functions.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; The idea in this file is to emulate using Gtk those few bits of Gnome
;; that Chart would like to use.  If actual Gnome bindings exist (they don't
;; yet) then they can be used, otherwise the emulation.  GnomeApp and
;; gnome-help seem like good ways to work in with Gnome conventions.
;;
;; Not sure about the gnome-app-helper stuff.  Don't really know what
;; wrapped versions of those functions will look like.  Maybe GtkUIManager
;; does enough anyway, or do something direct to make menu widgets more
;; easily available for user manipulation.

(define-module (chart gnome)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 popen)  (close-pipe open-input-pipe)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart extra-glib)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; support

;; return a procedure to be called (PROC widget), which is a record accessor
;; for FIELD in REC-TYPE in a c-prop
(define-public (c-prop-rec-accessor rec-type field)
  (let ((rec-acc-proc (rec-accessor rec-type field)))
    (lambda (widget)
      (rec-acc-proc (c-prop widget)))))


;;-----------------------------------------------------------------------------
;; gconf

;; Same as gconf_client_get_string(), but running the gconftool program.
;; The CLIENT arg is unused.
;;
;; gconftool prints an error to stderr when the key is bad, but just look
;; for empty output for now, that's easier (the exit status is still 0
;; unfortunately)
;;
(define (gconf-client-get-string client key)
  (let* ((pipe (open-input-pipe (string-append
				 "(gconftool -g " key ") 2>/dev/null")))
	 (str  (read-string-to-eof pipe)))
    (if (or (not (eqv? 0 (status:exit-val (close-pipe pipe))))
	    (string-null? str))
	;; dunno what the real error keys should be, they're not used in
	;; chart yet so it doesn't matter
	(throw 'g-error 'gconf-error 'gconf-error ""))
    (string-trim-right str #\newline)))


;;-----------------------------------------------------------------------------
;; gnome-exec

;; set the close-on-exec option on PORT
(define (fcntl-set-cloexec port)
  (fcntl port F_SETFD (logior FD_CLOEXEC (fcntl port F_GETFD))))

;; return PID, or -1 if cannot run
;;
(define-public (gnome-execute-async dir . args)
  (or (every string? args)
      (error "gnome-execute-async: args not strings:" args))

  (let* ((pair (pipe))
	 (in   (car pair))
	 (out  (cdr pair))
	 (pid  (primitive-fork)))
    (if (= 0 pid)

	;; immediate child
	(catch #t
	  (lambda ()
	    (close in)

	    (let ((pid (primitive-fork)))
	      (if (not (= 0 pid))
		  (begin
		    ;; immediate child
		    (display pid out)
		    (force-output out)
		    (primitive-_exit 0)))

	      ;; sub child
	      (fcntl-set-cloexec out)
	      (if dir
		  (chdir dir))

	      ;; close all fds, except stdin, stdout, stderr and
	      ;; `out'
	      (port-for-each
	       (lambda (port)
		 (false-if-exception ;; non-fd ports & close err
		  (let ((fd (fileno port)))
		    (or (<= fd 2)
			(= fd (fileno out))
			(close-fdes fd))))))
	      (apply execlp (car args) args)))

	  (lambda (key . args)
	    (display "x" out)
	    ;;(display key out)
	    ;;(display args out)
	    (close out)
	    (primitive-_exit 127))))

    ;; parent
    (close out)
    (let* ((str    (read-string-to-eof in))
	   (status (cdr (waitpid pid)))
	   (result (if (and (eqv? 0 (status:exit-val status))
			    (string-every char-set:digit str))
		       (string->number str)
		       -1)))
      (close in)
      result)))


;;-----------------------------------------------------------------------------
;; gnome-help

;; PATH is a directory
;; DOCNAME is a basename, with or without ".html" or whatever
;; return filename including PATH, or #f if not found
;;
(define (locate-help-file path docname)
  (gnome-i18n-get-language-list LC_MESSAGES)
  (any (lambda (lang)
	 (any (lambda (ext)
		(let ((full (g-build-filename
			     path lang (string-append docname ext))))
		  (and (file-exists? full)
		       full)))
	      '(".xml" ".docbook" ".sgml" ".html" "")))
       (gnome-i18n-get-language-list #f)))

;; LINK-ID can be #f for top of manual.
(define-public (gnome-help-display filename link-id)
  (gnome-help-display-with-doc-id #f #f filename link-id))

;; LINK-ID can be #f for the top of FILE-NAME
(define-public (gnome-help-display-with-doc-id program doc-id file-name link-id)

  ;; PROGRAM can be #f for default
  (or program
      (set! program (gnome-program-get)))

  ;; DOC-ID can be #f for default
  (or doc-id
      (set! doc-id (gnome-program-app-id program)))

  ;; full name and path
  (let* ((file (or
		;; local help
		(locate-help-file (first (gnome-program-locate-file
					  program 'app-help "" #f))
				  file-name)

		;; global help
		(locate-help-file (first (gnome-program-locate-file
					  program 'help "" #f))
				  file-name)

		(throw 'g-error 'help-error 'help-error-not-found
		       (_ "Cannot find help file"))))

	 ;; Not sure if it's meant to be "#" or "?" introducing the link.
	 ;; Only "#" works properly with yelp 2.14, but thought it needed
	 ;; "?" in the past.
	 ;;
	 ;; The gnome docs manual shows "?"
	 ;; http://developer.gnome.org/projects/gdp/handbook/gdp-handbook/ar01s07.html
	 ;;
	 (url (if link-id
		  (string-append "ghelp:" file "#" link-id)
		  (string-append "ghelp:" file))))

    (gnome-help-display-uri url)))

(define-public (gnome-help-display-uri uri)
  (gnome-url-show uri))


;;-----------------------------------------------------------------------------
;; gnome-i18n

(define (gnome-i18n-get-language-list category)
  (list "C"))


;;-----------------------------------------------------------------------------
;; gnome-program

;; app-id is a string

(let-toplevel*
    ((rec-type        (make-record-type "gnome-program"
					'(app-datadir
					  app-id
					  app-version
					  gnome-path ;; colon sep
					  human-readable-name)))
     (rec-new         (record-constructor rec-type '()))
     (rec?            (record-predicate rec-type))
     (rec-app-id      (rec-accessor rec-type 'app-id))
     (rec-gnome-path  (rec-accessor rec-type 'gnome-path))
     (rec-human-readable-name (rec-accessor rec-type 'gnome-path))

     ;; (rec-gnome-path! (rec-modifier rec-type 'gnome-path))

     (global-program  #f))

  (define-public (gnome-program? program)
    (rec? program))

  (define-public (gnome-program-get)
    global-program)

  (define-public (gnome-program-app-id program)
    (rec-app-id program))

  (define-public gnome-program-app-version
    (c-prop-rec-accessor rec-type 'app-version))

  (define-public (gnome-program-get-human-readable-name program)
    (or (rec-human-readable-name (or program global-program))
	(car (program-arguments))))

  (define-public (gnome-program-init app-id app-version
				     module-info argv . properties)
    (let ((program (rec-new)))
      (gnome-program-set-property program 'app-id      app-id)
      (gnome-program-set-property program 'app-version app-version)
      (gnome-program-set-property program 'gnome-path
				  "/usr/share/gnome:/usr/local/share/gnome")
      (if (not global-program)
	  (set! global-program program))
      (for-each-two (lambda (property value)
		      (gnome-program-set-property program property value))
		    properties)
      program))

  (define-public (gnome-program-set-property program param-id value)
    ((rec-modifier rec-type param-id) program value))

  (define-public (gnome-program-get-property program param-id)
    ((rec-accessor rec-type param-id) program))

  ;; return list of paths for FILE-NAME in given DOMAIN
  ;; DOMAIN is a symbol 'app-help 'help
  (define-public (gnome-program-locate-file program domain file-name
					    only-if-exists)
    (set! program (or program global-program))
    (let ((attr-name  #f)
	  (attr-rel   #f)
	  (prefix-rel #f))
      (case domain
	((app-help)
	 (set! prefix-rel (g-build-filename "share/gnome/help"
					    (rec-app-id program)))
	 (set! attr-name 'app-datadir)
	 (set! attr-rel  (g-build-filename "gnome/help"
					   (rec-app-id program))))
	((help)
	 (set! prefix-rel "share/gnome/help"))
	(else
	 (error "gnome-program-locate-file: unrecognised GnomeFileDomain:"
		domain)))

      (let ((result '())
	    (dir-list (map (lambda (dir)
			     (g-build-filename dir prefix-rel))
			   (string-split (rec-gnome-path program) #\:))))
	(if attr-name
	    (let ((attr-dir (gnome-program-get-property program attr-name)))
	      (if attr-dir
		  (set! dir-list (cons (g-build-filename attr-dir attr-rel)
				       dir-list)))))

	(for-each (lambda (dir)
		    (let ((name (g-build-filename dir file-name)))
		      (if (or (not only-if-exists)
			      (access? name F_OK))
			  (set! result (cons name result)))))
		  dir-list)
	(reverse! result)))))


;;-----------------------------------------------------------------------------
;; gnome-url

(define (ghelp-url->file-url url)
  (if (string-prefix? "ghelp:" url)
      (begin
	(set! url (string-append "file://" (string-drop url 6)))
	;; ?LINKID becomes #LINKID
	(set! url (string-map (lambda (c)
				(if (eqv? c #\?)
				    #\#
				    c))
			      url))))
  url)


;; return STR with a "%s" expanded to REPL
;; at most one %s is expected
(define (string-substitute-%s str repl)
  (let ((pos (string-contains str "%s")))
    (if pos
	(string-append (string-take str pos)
		       repl
		       (string-drop str (+ 2 pos)))
	str)))

;; return the scheme part of a url, for instance "http://foo.com/index.html"
;; gives "http"; if there's no valid scheme the return is #f
;;
;; as per RFC 1738, only letters, digits + . and - are allowed, and the
;; scheme is downcased
;;
(define (url-scheme url)
  (let ((m (string-match "^([A-Za-z0-9+.-]+):" url)))
    (and m
	 (string-downcase (match:substring m 1)))))

;; The use of gconf here emulates what the real gnome_url_show does.  But we
;; fallback to gnome-help directly or a likely browser (using file: instead
;; of ghelp:).
;;
;; Not sure if looking at gconf is a great idea.  It's good for users
;; running Gnome, but if you've merely had gconf dragged in as a dependency
;; of some gnome app you might not want to use its setups.
;;
;; Yelp (gnome-help) 2.6 seemed to have problems going to the linkid
;; location, but 2.10 is ok.
;;
(define-public (gnome-url-show url)

  (define (attempt-gconf key)
    (and-let* ((command (catch 'g-error
			  (lambda ()
			    (gconf-client-get-string #f key))
			  (lambda args
			    #f))))
      (set! command (string-substitute-%s command url))

      ;; quotes around url on debian sensible-browser
      (set! command (string-delete command #\"))

      ;; probably supposed to /bin/sh -c, but then can't tell if command exists
      (not (= -1 (apply gnome-execute-async
			#f (string-split command #\space))))))

  (define (attempt-prog prog)
    (not (= -1 (gnome-execute-async #f prog url))))

  (or
   ;; per gnome-url.c, if no recognisable scheme then don't look for a
   ;; specific handler (just go on to the "unknown" one
   (and-let* ((scheme (url-scheme url)))
     (attempt-gconf (string-append "/desktop/gnome/url-handlers/"
				   scheme "/command")))

   (attempt-gconf "/desktop/gnome/url-handlers/unknown/command")

   (and (string-prefix? "ghelp:" url)
	(or
	 (attempt-prog "gnome-help")            ;; gnome 2 - yelp
	 (attempt-prog "gnome-help-browser")))  ;; gnome 2 - yelp

   ;; at this stage in ghelp spawning a ghelp url has a full path, so it can
   ;; be turned into a "file:" directly
   (begin
     (set! url (ghelp-url->file-url url))
     #t)

   (and-let* ((browser (getenv "BROWSER"))) ;; environ
     (attempt-prog browser))

   (attempt-prog "sensible-browser")   ;; debian
   (attempt-prog "gnome-moz-remote")   ;; usual gnome
   (attempt-prog "mozilla")
   (attempt-prog "navigator")

   (throw 'g-error 'spawn-error 'spawn-error-noexec
	  (_ "Cannot execute browser"))))


