;; Graph widgets.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart graph)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:autoload   (chart bar)         (bar-draw)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart gui-common)
  #:use-module (chart heading)
  #:autoload   (chart latest)      (latest-get)
  #:autoload   (chart line)        (line-draw bands-draw zigzag-draw)
  #:use-module (chart misc)
  #:autoload   (chart ohlc-draw)   (candles-draw hl-draw ohlc-draw)
  #:autoload   (chart ohlcv)       (ohlcv-initial-page)
  #:autoload   (chart points)      (points-draw)
  #:autoload   (chart solid)       (solid-draw)
  #:autoload   (chart stops)       (stops-draw)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; configs

(define-public scrollbar-page-fraction  0.9)
(define-public scrollbar-step-fraction  0.1)

(define-public viewstyle-default '((ohlc (none) (window-size 4))
				   ((volume) (none) (window-size 1))))


;;-----------------------------------------------------------------------------
;; GdkRectangle stuff

(define gdk-rectangle-x      caar)
(define gdk-rectangle-y      cdar)
(define gdk-rectangle-width  cadr)
(define gdk-rectangle-height cddr)


;;-----------------------------------------------------------------------------
;; GdkAdjustment stuff

;; get value at the upper end of the thumb
(define (c-gtk-adjustment-value-above adj)
  (+ (gtk-adjustment-value adj)
     (gtk-adjustment-page-size adj)))


;;-----------------------------------------------------------------------------
;; misc

(define-public (note-line? note)
  (and (pair? note)
       (memq (first note) '(line hline))))

(define-public (note-alert? note)
  (and (pair? note)
       (memq (first note) '(alert-above alert-below))))

(define-public graph-up-colour         "light green")
(define-public graph-down-colour       "pink")
(define-public graph-unchanged-colour  "white")
(define-public graph-band-colour       "blue")

;; '(line  <t1> <p1> <t2> <p2>)
;; '(hline <t1> <p1> <t2> <p2>)

(define (line-or-alert? note)
  (and (pair? note)
       (memq (first note) '(line hline alert-above alert-below))))

(define-public line-t1  second)
(define-public line-p1  third)
(define-public line-t2  fourth)
(define-public line-p2  fifth)


;;-----------------------------------------------------------------------------
;; ChartGraph elements

(let-toplevel ((rec-type (make-record-type "graph-element"
					   '(proc expand-proc))))
  (define rec-new  	  (record-constructor rec-type))
  (define rec-proc 	  (rec-accessor rec-type 'proc))
  (define rec-expand-proc (rec-accessor rec-type 'expand-proc))
  (define draw-list     '())
  (define xor-list '())

  (define*-public (graph-element-new #:key (xor #f)
				     (proc #f) (expand-proc noop))
    (let ((elem (rec-new proc expand-proc)))
      (if xor
	  (set! xor-list  (append! xor-list  (list elem)))
	  (set! draw-list (append! draw-list (list elem))))))

  (define-public (graph-element-draw elem graph)
    ((rec-proc elem) graph))

  ;; run all vadj expansion procedures on GRAPH
  (define-public (graph-element-expand-all graph)
    (for-each (lambda (elem)
		((rec-expand-proc elem) graph))
	      (append draw-list xor-list)))

  (define-public (graph-element-draw-all graph)
    (for-each (lambda (elem)
		((rec-proc elem) graph))
	      draw-list))

  (define-public (graph-element-xor-all graph)
    (for-each (lambda (elem)
		((rec-proc elem) graph))
	      xor-list)))

(graph-element-new #:proc (lambda args (apply heading-draw-in-graph args)))


;;-----------------------------------------------------------------------------
;; draw no data message

;; return true if SERIES is all #f's within the LO to HI range
(define (series-empty? series lo hi)
  (null? (delq! #f (array-contents->list
		    (make-shared-array-restrict (series-array series lo hi)
						lo hi)))))

(define (draw-nodata graph)
  (let ((series (graph-series graph)))
    (if (and series
	     (series-empty? series
			    (graph-draw-x-lo graph series)
			    (graph-draw-x-hi graph series)))
	(c-gtk-draw-string-centred graph (_ "No data")))))

(graph-element-new #:proc draw-nodata)


;;-----------------------------------------------------------------------------
;; draw dividends and annotations

(define (draw-notes-y-size graph)
  (let* ((layout (primitive-property-ref c-layout-prop graph))
	 (font-height  (c-pango-layout-line-height layout))
	 (y-margin     (round-exact (* 0.3 font-height)))
	 (y-top        (round-exact (* 1.5 font-height)))
	 (height       (+ y-margin font-height y-margin)))
    (+ y-top height font-height)))

;; return a utf-8 string for a dividend value DIV
(define-public (dividend->string div)
  (or (if (string? div)
	  div    ;; already utf-8

	  ;; list
	  (case (first div)
	    ((month)
	     ;; (month MDATE)
	     ;; (month MDATE STRINGNUMBER)
	     (let ((str (strftime-utf8 "%b %y" (mdate->tm (second div)))))
	       (if (length>=? div 3)
		   (set! str (string-append str "\n" (third div))))
	       str))

	    ((split)
	     (format #f (_ "~a for ~a") (second div) (third div)))

	    ((amount)
	     (let ((amount (second div))
		   (alist  (cddr div)))
	       (cond ((eq? amount 'unknown)
		      (_ "Unknown"))
		     ((eq? amount 'to-be-advised)
		      (_ "To be advised"))
		     ((number? amount)
		      (let* ((decimals (or (assq-ref alist 'decimals) 0))
			     (str (locale-number+decimals amount decimals)))
			(if (assq-ref alist 'estimated)
			    (set! str (format #f (_ "Est ~a") str)))
			(and-let* ((imputation (assq-ref alist 'imputation)))
			  (if (number? imputation)
			      (set! imputation
				    (locale-number+decimals imputation decimals)))
			  (set! str (format #f "~a + ~a" str imputation)))
			str))
		     (else
		      #f))))))

      (format #f "~s" div)))

;; LST is ((tdate . dividend) ...)
(define (draw-notes graph lst)
  (if (not (null? lst))
      (let* ((layout      (primitive-property-ref c-layout-prop graph))
	     (style       (gtk-widget-get-style graph))
	     (text-gc     (gtk-style-white-gc style))
	     (font-height (c-pango-layout-line-height layout))
	     (win         (gtk-widget-window graph))
	     (y-margin    (round-exact (* 0.3 font-height)))
	     (x-margin    (round-exact (* 0.8 (c-pango-layout-em layout))))
	     (y-top       (round-exact (* 1.5 font-height)))
	     (y-str       (+ y-top y-margin))
	     (scale-x     (graph-scale-x-proc graph))
	     (lo          (- (graph-x-lo graph) 50))
	     (hi          (+ (graph-x-hi graph) 50)))

	(for-each
	 (lambda (pair)
	   (let ((t (car pair)))
	     (if (within-range? lo hi t)
		 (let ((str (dividend->string (cdr pair))))
		   (pango-layout-set-text layout str)
		   (receive-list (str-width str-height)
		       (pango-layout-get-pixel-size layout)
		     (let* ((x-centre (+ (scale-x t)
					 (quotient scale-x-step 2)))
			    (x-str    (- x-centre (quotient str-width 2)))
			    (x-left   (- x-str x-margin))
			    (width    (+ str-width x-margin x-margin))
			    (height   (+ y-margin str-height y-margin)))
		       (gdk-draw-layout win text-gc x-str y-str layout)
		       (gdk-draw-rectangle win text-gc #f
					   x-left y-top width height)
		       (gdk-draw-line win text-gc
				      x-centre (+ y-top height)
				      x-centre (+ y-top height font-height))))))))
	 lst))))

(define (draw-dividends graph)
  (draw-notes graph (series-dividends (graph-series graph))))

(graph-element-new #:proc draw-dividends)

(define (draw-annotations graph)
  (let ((tdate->t (timebase-convert-proc timebase-days
					 (graph-timebase graph))))
    (draw-notes graph
		(map (lambda (note)
		       (cons (tdate->t (second note)) (third note)))
		     (filter annotation?
			     (series-notes (graph-series graph)))))))

(graph-element-new #:proc draw-annotations)


;;-----------------------------------------------------------------------------
;; draw lines and alerts

(define line-colour  "grey")
(define line-xor-gc #f)

(define alert-colour  "red")
(define alert-xor-gc #f)
(define-public graph-alert-width 20)

;; series element during edit, #f otherwise
(define-public line-edit-state #f)

(define-public (line-xor-list graph line-list)
  (let* ((win      (gtk-widget-window graph))
	 (width    (c-gdk-drawable-get-width win))
	 (height   (c-gdk-drawable-get-height win))
	 (x-mid    (quotient scale-x-step 2))
	 (scale-x  (graph-scale-x-proc graph))
	 (scale-y  (graph-scale-y-proc graph))
	 (tdate->t (timebase-convert-proc timebase-days
					  (graph-timebase graph)))
	 (series   (graph-series graph)))

    (if (not line-xor-gc)
	(let ((c  (gdk-color-new))
	      (bg (gtk-style-bg (gtk-widget-get-style graph) 'normal)))

	  (set! line-colour (gdk-color-intern line-colour))
	  (gdk-color-set-pixel c (logxor (gdk-color-pixel line-colour)
					 (gdk-color-pixel bg)))
	  (set! line-xor-gc (c-gdk-gc-get
			     win
			     #:foreground c
			     #:function 'xor))

	  (set! alert-colour (gdk-color-intern alert-colour))
	  (gdk-color-set-pixel c (logxor (gdk-color-pixel alert-colour)
					 (gdk-color-pixel bg)))
	  (set! alert-xor-gc (c-gdk-gc-get
			      win
			      #:foreground c
			      #:function 'xor))))

    (for-each (lambda (elem)
		(case (first elem)
		  ((alert-above alert-below)
		   (let ((p (second elem)))
		     (if (string? p)
			 (set! p (string->number+decimals
				  p (series-decimals series))))
		     (if (or (not (graph-logarithmic graph))
			     (positive? p))
			 (let* ((y (scale-y p))
				(t (1+ (min (graph-today graph)
					    (hadjustment-hi
					     (graph-hadjustment graph)))))
				(x (1- (scale-x t))))
			   (c-gdk-draw-line-restrict  ;; horiz
			    win alert-xor-gc (- x graph-alert-width) y (1- x) y
			    width height)
			   (c-gdk-draw-line-restrict  ;; vert
			    win alert-xor-gc
			    x y x (if (eq? 'alert-above (first elem))
				      (- y 8)
				      (+ y 8))
			    width height)))))
		  (else
		   (let ((p1 (line-p1 elem))
			 (p2 (line-p2 elem)))
		     (if (or (not (graph-logarithmic graph))
			     (and (positive? p1)
				  (positive? p2)))
			 (c-gdk-draw-line-restrict
			  win line-xor-gc
			  (+ x-mid (scale-x (tdate->t (line-t1 elem))))
			  (scale-y p1)
			  (+ x-mid (scale-x (tdate->t (line-t2 elem))))
			  (scale-y p2)
			  width height))))))
	      line-list)))

(define (lines-xor graph)
  (let ((series (graph-series graph)))
    (if series
	(let ((line-list (delete! line-edit-state
				  (filter line-or-alert?
					  (series-notes series)))))
	  (if (not (null? line-list))
	      (line-xor-list graph line-list))))))

(graph-element-new #:xor #t
		   #:proc lines-xor)


;;-----------------------------------------------------------------------------
;; draw latest

(define latest-colour "yellow")

;; return true if any element of row I in ARRAY is not #f
(define (array-row-any-non-f? array i)
  (let ((shape (array-shape array)))
    (if (= 1 (length shape))
	(array-ref array i)
	(let ((col-lo (first (second shape)))
	      (col-hi (second (second shape))))
	  (let more ((j col-lo))
	    (and (not (> j col-hi))
		 (or (array-ref array i j)
		     (more (1+ j)))))))))

;; return true if SERIES has some data at or beyond TDATE
;; this is meant for looking near the end of the series, with a possibly
;; all-#f day from an empty database series on the first day of trading
(define (series-covers-tdate? series tdate)
  (let more ((tdate (max tdate (series-lo series))))
    (and (<= tdate (series-hi series))
	 (let ((array (series-array series tdate tdate)))
	   (or (array-row-any-non-f? array tdate)
	       (more (1+ tdate)))))))

;; return LATEST adapted to the decimals of SERIES
(define (latest-for-series latest series)
  (if (= (series-decimals series)
	 (latest-decimals latest))
      latest

      (let ((factor (integer-expt-pn 10 (- (series-decimals series)
					   (latest-decimals latest)))))
	(define (conv p)
	  (and p (* factor p)))
	(let ((new (record-copy latest)))
	  (latest-open!     new (conv (latest-open latest)))
	  (latest-high!     new (conv (latest-high latest)))
	  (latest-low!      new (conv (latest-low  latest)))
	  (latest-last!     new (conv (latest-last latest)))
	  (latest-change!   new (conv (latest-change latest)))
	  (latest-bid!      new (conv (latest-bid latest)))
	  (latest-offer!    new (conv (latest-offer latest)))
	  (latest-decimals! new (series-decimals series))
	  new))))

(let-toplevel ((gc          #f)
	       (latest-prop (make-object-property)))

  (define (latest-xor-record graph latest)
    (set! (latest-prop graph) latest)
    (set! latest (latest-for-series latest (graph-series graph)))
    (let* ((scale-x (graph-scale-x-proc graph))
	   (series  (graph-series graph))
	   (scale-y (graph-scale-y-proc graph))
	   (win     (gtk-widget-window graph))
	   (height  (c-gdk-drawable-get-height win)))

      (if (not gc)
	  (let ((c  (gdk-color-new))
		(bg (gtk-style-bg (gtk-widget-get-style graph) 'normal)))
	    (set! latest-colour (gdk-color-intern latest-colour))
	    (gdk-color-set-pixel c (logxor (gdk-color-pixel latest-colour)
					   (gdk-color-pixel bg)))
	    (set! gc (c-gdk-gc-get win
				   #:foreground c
				   #:function 'xor))))

      (cond
       ((eq? (series-datatype series) datatype-ohlcv)
	(let* ((rect-list '())
	       (lw         (max 1 (round-exact (/ scale-x-step 3.5))))
	       ;; high/low position, centred in width, and rounded down so
	       ;; it's open (rather than close) which is overwritten when
	       ;; only 2 pixels
	       (hl-off     (quotient (- scale-x-step lw) 2)))

	  ;; draw rectangle x,y,w,h, and double draw any overlaps with
	  ;; previously drawn rectangles, so as to balance the xoring
	  (define (draw-rectangle x y w h)
	    (gdk-draw-rectangle win gc #t x y w h)
	    (let ((r (cons (cons x y) (cons w h))))
	      (for-each (lambda (rold)
			  (and-let* ((rint (gdk-rectangle-intersect r rold)))
			    (gdk-draw-rectangle win gc #t
						(gdk-rectangle-x rint)
						(gdk-rectangle-y rint)
						(gdk-rectangle-width rint)
						(gdk-rectangle-height rint))
			    (set! rect-list (cons rint rect-list))))
			rect-list)
	      (set! rect-list (cons r rect-list))))

	  ;; open, high, low, last
	  ;; round up adate to show weekend in slot for monday
	  (and-let* ((last-t (latest-last-tdate latest))
		     ;; must be past series data
		     (       (not (series-covers-tdate? (graph-series graph)
							last-t))))
	    (let ((last-x (scale-x last-t)))

	      ;; open
	      (and-let* ((open (latest-open latest)))
		(let ((o-off (max 0 (- hl-off lw)))) ;; open position
		  (draw-rectangle (+ last-x o-off)
				  (scale-y open)
				  (- (+ hl-off lw) o-off) ;; across hl bar
				  1)))

	      ;; high/low
	      (and-let* ((high (latest-high latest))
			 (low  (latest-low  latest)))
		(draw-rectangle (+ last-x hl-off)
				(scale-y high)
				lw
				(- (scale-y low) (scale-y high)
				   -1))) ;; inclusive

	      ;; last
	      (and-let* ((last (latest-last latest)))
		;; across hl bar and extended
		(draw-rectangle (+ last-x hl-off)
				(scale-y last)
				(* 4 lw)
				1))))

	  (and-let* ((quote-t (latest-quote-tdate latest))
		     ;; must be past series data
		     (       (not (series-covers-tdate? (graph-series graph)
							quote-t))))
	    (let ((quote-x (scale-x quote-t)))

	      ;; bid
	      (and-let* ((bid (latest-bid latest)))
		(draw-rectangle (+ quote-x hl-off)
				(scale-y bid)
				(* 2 lw)
				1))

	      (and-let* ((offer (latest-offer latest)))
		(draw-rectangle (+ quote-x hl-off)
				(scale-y offer)
				(* 2 lw)
				1))))))

       ((eq? 'volume (datatype-key (series-datatype series)))

	;; volume
	;; round up to show weekend in slot for monday
	(and-let* ((last-t (latest-last-tdate latest))
		   ;; must be past series data
		   (       (not (series-covers-tdate? (graph-series graph)
						      last-t)))
		   (volume (latest-volume latest))
		   (       (not (= 0 volume))))
	  (let* ((bar-width  (max 1 (quotient (* scale-x-step 2) 3)))
		 (bar-offset (quotient (- scale-x-step bar-width) 2))
		 (x          (+ bar-offset (scale-x last-t)))
		 ;; if stored data doesn't have volumes, then latest volume
		 ;; could make y a very big negative
		 (y          (max 0 (scale-y volume))))
	    (gdk-draw-rectangle win gc #t x y bar-width (- height y))))))))

  (define (latest-draw graph)
    (and-let* ((series (graph-series graph))
	       (symbol (series-symbol series))
	       (latest (latest-get symbol)))
      (latest-xor-record graph latest)))

  ;; SYMBOL-LIST is new latest records, update the drawing in GRAPH if one
  ;; of them is being displayed (erase old, draw new)
  ;;
  (define-toplevel (latest-draw-update-maybe graph symbol-list)
    (and-let* ((series (graph-series graph))
	       (symbol (series-symbol series))
	       (       (member symbol symbol-list)))
      ;; possibly expand vadj
      (graph-range-update-latest graph)

      ;; don't draw now if a redraw is already pending
      (or (c-gtk-widget-queued-redraw? graph)
	  (begin
	    (and-let* ((old-latest (latest-prop graph))) ;; old, if any
	      (latest-xor-record graph old-latest))
	    (and-let* ((new-latest (latest-get symbol))) ;; new, if any
	      (latest-xor-record graph new-latest))))))

  (graph-element-new #:proc latest-draw))


;;-----------------------------------------------------------------------------
;; today

(define today-colour "grey")

;; return today's date, in the timebase and symbol timezone of GRAPH
(define-public (graph-today graph)
  (series-today (graph-series graph)))

(define (today-draw graph)
  (let ((t (1+ (graph-today graph))))
    ;; only if in window
    (if (<= t (hadjustment-hi (graph-hadjustment graph)))
	(let* ((win (gtk-widget-window graph))
	       (gc  (c-gdk-gc-get win
				  #:foreground today-colour
				  #:dashed #t))
	       ;; centred in slot of day after today
	       (x   (+ (graph-scale-x graph t)
		       (ceiling-exact (* 0.5 scale-x-step)))))
	  (gdk-draw-line win gc x 0 x (c-gdk-drawable-get-height win))))))

(graph-element-new #:proc today-draw)


;;-----------------------------------------------------------------------------
;; hlines from datatype

(define (c-gdk-draw-hline win gc x1 x2 y)
  (gdk-draw-line win gc x1 y x2 y))

(define hline-colour "grey")

(define (hlines-draw graph)
  (and-let* ((series (graph-series graph))
	     (hlines (datatype-hlines (series-datatype series)))
	     (       (not (null? hlines))))
    (let* ((win     (gtk-widget-window graph))
	   (width   (c-gdk-drawable-get-width win))
	   (height  (c-gdk-drawable-get-height win))
	   (scale-y (graph-scale-y-proc graph))
	   (gc      (c-gdk-gc-get win
				  #:foreground hline-colour
				  #:dashed #t)))

      (for-each (lambda (level)
		  (let ((y (scale-y level)))
		    (if (and (>= y 0)
			     (< y height))
			(c-gdk-draw-hline win gc 0 width y))))
		hlines))))

(graph-element-new #:proc hlines-draw)


;;-----------------------------------------------------------------------------
;; ChartComponent widget

(gtk-class-new 'GtkDrawingArea "ChartComponent")

;; the ChartView widget holding this component
(define-public component-view gtk-widget-get-parent)



;;-----------------------------------------------------------------------------
;; graph range

(define graph-range-stretch-amount 0.1)

(define (graph-range-update-latest graph)
  (and-let* ((series   (graph-series graph))
	     (         (eq? timebase-days (series-timebase series)))
	     (symbol   (series-symbol series))
	     (latest   (latest-get symbol))
	     (datatype (series-datatype series)))

    ;; prices
    (if (eq? datatype-ohlcv datatype)
	(begin
	  (set! latest (latest-for-series latest series))
	  (graph-range-expand graph
			      (latest-bid latest)
			      (latest-offer latest)
			      (latest-last latest)
			      (latest-open latest)
			      (latest-high latest)
			      (latest-low latest))))

    ;; volume
    (if (eq? 'volume (datatype-key datatype))
	(graph-range-expand graph (latest-volume latest))))

  (graph-range-apply graph))

(define (graph-range-update-notes graph)
  (for-each (lambda (series)
	      (let ((datatype (series-datatype series)))
		
		;; datatype hlines
		(apply graph-range-expand graph (datatype-hlines datatype))

		;; datatype specified min and max
		(if (datatype-minimum datatype)
		    (graph-range-expand graph (datatype-minimum datatype)))
		(if (datatype-maximum datatype)
		    (graph-range-expand graph (datatype-maximum datatype))))

	      ;; alert levels and drawn lines
	      (for-each (lambda (elem)
			  (cond ((note-alert? elem)
				 (graph-range-expand graph (second elem)))
				((note-line? elem)
				 (graph-range-expand graph (line-p1 elem)
						     (line-p2 elem)))))
			(series-notes series)))

	    (graph-series-list graph)))

;; return the contents of ARRAY as a flat list
;; this is like (array->list (array-contents ARRAY)), but it copes with
;; non-contiguous arrays that array-contents doesn't support
(define (array-contents->list array)
  (let ((lst (array->list array)))
    (do ((i (array-rank array) (1- i)))
	((<= i 1))
      (set! lst (concatenate! lst)))
    lst))

(define (graph-range-update-data graph)
  (for-each (lambda (series)
	      (let ((datatype (series-datatype series))
		    (array    (series-array series
					    (graph-x-lo graph)
					    (graph-x-hi graph))))
		(if (eq? datatype datatype-ohlcv)
		    (set! array (make-shared-array-columns array 0 3)))
		(apply graph-range-expand graph (array-contents->list array))))
	    (graph-series-list graph)))

(define-public (positive-number? obj)
  (and (number? obj)
       (positive? obj)))


(define (graph-range-expand graph . value-list)
  (define rec (c-prop graph))
  (rec-upper! rec (apply max-maybe (rec-upper rec) value-list))
  (rec-lower! rec (apply min-maybe (rec-lower rec) value-list))

  (set! value-list (filter positive-number? value-list))
  (and-let* ((m (apply max-maybe value-list)))
    (rec-upperlog! rec (max-maybe (rec-upperlog rec) (log m))))
  (and-let* ((m (apply min-maybe value-list)))
    (rec-lowerlog! rec (min-maybe (rec-lowerlog rec) (log m)))))

(define (graph-range-apply graph)
  (define rec (c-prop graph))
  (let ((upper (if (graph-logarithmic graph)
		   (rec-upperlog rec)
		   (rec-upper rec)))
	(lower (if (graph-logarithmic graph)
		   (rec-lowerlog rec)
		   (rec-lower rec))))

    (if (and upper lower)
	(let ((extra (* graph-range-stretch-amount
			(gtk-adjustment-page-size (graph-vadjustment graph)))))
	  ;; extend upper
	  (set! upper (+ upper extra))

	  ;; extend lower, except when non-logarithmic don't let it go
	  ;; negative if it isn't already
	  (let ((new-lower (- lower extra)))
	    (if (and (not (graph-logarithmic graph))
		     (positive? lower))
		(set! new-lower (max 0 new-lower)))
	    (set! lower new-lower))

	  (c-gtk-adjustment-change (graph-vadjustment graph)
				   #:upper upper
				   #:lower lower)))))


;;-----------------------------------------------------------------------------
;; ChartGraph widget

(gtk-class-new 'ChartComponent "ChartGraph")
(c-gtk-rc-chart-style "ChartGraph")

(define rec-type (make-record-type "chartgraph"
				   '(series-list
				     vadjustment
				     scale-y-factor
				     logarithmic
				     upper lower upperlog lowerlog)))

(define rec-new  (record-constructor rec-type '(vadjustment
						series-list
						scale-y-factor)))
(define rec-vadjustment     (rec-accessor rec-type 'vadjustment))
(define rec-series-list     (rec-accessor rec-type 'series-list))
(define rec-series-list!    (rec-modifier rec-type 'series-list))
(define rec-scale-y-factor  (rec-accessor rec-type 'scale-y-factor))
(define rec-scale-y-factor! (rec-modifier rec-type 'scale-y-factor))
(define rec-logarithmic     (rec-accessor rec-type 'logarithmic))
(define rec-logarithmic!    (rec-modifier rec-type 'logarithmic))

(define rec-upper     (rec-accessor rec-type 'upper))
(define rec-upper!    (rec-modifier rec-type 'upper))
(define rec-lower     (rec-accessor rec-type 'lower))
(define rec-lower!    (rec-modifier rec-type 'lower))
(define rec-upperlog    (rec-accessor rec-type 'upperlog))
(define rec-upperlog!   (rec-modifier rec-type 'upperlog))
(define rec-lowerlog    (rec-accessor rec-type 'lowerlog))
(define rec-lowerlog!   (rec-modifier rec-type 'lowerlog))

(define-public (graph-vadjustment graph)
  (rec-vadjustment (c-prop graph)))

(define-public (graph-scale-y-factor graph)
  (rec-scale-y-factor (c-prop graph)))

;; For a new series-list in GRAPH, set the vadj initial range and page.
;;
;; The page size is a smart choice based on the apparent recent volatility
;; etc.  The range is from the data in the initial visible portion, and
;; extended if necessary to accomodate the desired page size.
;;
;; This is all done in one function so there's a single
;; c-gtk-adjustment-change and "changed" signal from the vadj.
;;
(define (scale-y-initial graph)

  (let* ((rec      (c-prop graph))
	 (series   (graph-series graph))
	 (datatype (series-datatype series))
	 (win      (gtk-widget-window graph))
	 (height   (c-gdk-drawable-get-height win))
	 (page     0)
	 (vadj     (graph-vadjustment graph)))

    (graph-range-update-data graph)
    (graph-range-update-notes graph)
    (graph-range-update-latest graph)

    (receive-list (l h)
	((or (datatype-initial-page-proc datatype)
	     ohlcv-initial-page)
	 series
	 (graph-draw-x-lo graph series)
         (graph-draw-x-hi graph series))
      (if l
	  (begin
	    (set! l (graph-log graph l))
	    (set! h (graph-log graph h))
	    (let* ((extend (* graph-range-stretch-amount (- h l))))
	      (set! h (+ h extend))
	      (if (or (graph-logarithmic graph)
		      (not (datatype-minimum datatype))
		      (< l (datatype-minimum datatype)))
		  (set! l (- l extend))
		  (set! l (max (datatype-minimum datatype)
			       (- l extend))))
	      (set! page (- h l))))))
    (if (zero? page)
	(set! page (graph-log graph 10)))

    ;; don't show more than the datatype specified min/max
    (and-let* ((         (datatype-maximum datatype))
	       (dmax     (graph-log graph (datatype-maximum datatype)))
	       (         (datatype-minimum datatype))
	       (dmin     (graph-log graph (datatype-minimum datatype)))
	       (max-page (* (1+ (* 2 graph-range-stretch-amount)) ;; eg. 1.2
			    (- dmax dmin))))
      (set! page (min page max-page)))

    (rec-scale-y-factor! rec (/ height page))
    (c-gtk-adjustment-change
     vadj
     #:page-size      page
     #:step-increment (* scrollbar-step-fraction page)
     #:page-increment (* scrollbar-page-fraction page))

    ;; volume start at zero, others centre on visible
    (let ((pos (if (memq (datatype-key datatype) '(volume open-interest))
		   0.0
		   (- (graph-centre-vertical-pos graph)
		      (* 0.5 page)))))
      (graph-range-expand graph pos (+ page pos))
      (graph-range-apply graph)
      (gtk-adjustment-set-value vadj pos))))

;; GRAPH window has resized, adjust scale-y-factor accordingly.
;;
;; An alternative would be to leave scale-y-factor alone and show more or
;; less of the chart.  Staying on the same part is easier for now.  When
;; an integer points-to-pixels comes about it will probably be better to
;; show more or less.
;;
(define (scale-y-window-resize graph)
  (let ((rec (c-prop graph)))
    (rec-scale-y-factor!
     rec
     (/ (c-gdk-drawable-get-height (gtk-widget-window graph))
	(gtk-adjustment-page-size (rec-vadjustment rec))))
    (c-gtk-widget-queue-redraw graph)))

(define-public (graph-new vadj)
  ;; (gtk-widget-new 'ChartGraph #:width -1 #:height -1)
  (let ((graph (gtk-widget-new 'ChartGraph))
	(rec   (rec-new vadj '() 1.0)))

    (define (draw)
      (if (not (null? (rec-series-list rec)))
	  (graph-element-draw-all graph)
	  (c-gtk-draw-string-centred
	   graph (_ "Use File/Open to open or add a symbol"))))
    (define (xor)
      (if (not (null? (rec-series-list rec)))
	  (graph-element-xor-all graph)))

    (define (queue-redraw . args)
      (c-gtk-widget-queue-redraw graph))

    (set! (c-prop graph) rec)
    (c-gtk-widget-drawing graph draw xor)

    ;; for line-edit and scroll-drag
    (c-gtk-widget-add-events graph 'graph-base-button-drags
			     '(button-motion-mask
			       pointer-motion-hint-mask
			       button-release-mask))

    ;; font changes
    (gtk-signal-connect graph "style_set" queue-redraw)

    ;; window size changes
    (gtk-signal-connect graph "configure-event"
      (lambda (event)
	(scale-y-window-resize graph)))

    ;; hadj movement, when we get a view parent with a hadj
    (c-gtk-signal-connect-once graph "parent-set"
      (lambda (old-parent)
	(let ((hadj (graph-hadjustment graph)))
	  ;; expand so can scroll to changed or wider hadj portion,
	  ;; and draw for that moved or wider portion
	  (define (update-for-hadj)
	    (graph-range-update-data graph)
	    (graph-range-apply graph)
	    (queue-redraw))
	  (gtk-signal-connect hadj "value-changed" update-for-hadj)
	  (gtk-signal-connect hadj "changed" update-for-hadj))))

    ;; vadj movement
    (gtk-signal-connect vadj "value-changed" queue-redraw)
    (gtk-signal-connect vadj "changed" queue-redraw)

    ;; new latest
    (notify-connect 'latest-update
      (lambda (symbol-list)
	(latest-draw-update-maybe graph symbol-list)))

    graph))

(define-public (graph-scale-y-factor! graph factor)

  (let* ((rec      (c-prop graph))
	 (win      (gtk-widget-window graph))
	 (height   (c-gdk-drawable-get-height win))
	 (new-page (/ height factor))
	 (vadj     (rec-vadjustment rec))
	 (upper    (gtk-adjustment-upper vadj))
	 (lower    (gtk-adjustment-lower vadj))
	 (old-page (gtk-adjustment-page-size vadj))
	 (old-bot  (gtk-adjustment-value vadj))
	 (old-top  (+ old-bot old-page))
	 (new-top  #f)
	 (new-bot  #f)
	 (glog?    (rec-logarithmic rec))
	 (glog     (graph-log-proc graph)))

    (rec-scale-y-factor! rec factor)

    ;; increase upper/lower if necessary for new-page
    ;; extend each end equally, except don't let lower go negative if it's
    ;; not already
    (if (> new-page (- upper lower))
	(let* ((extra (- new-page (- upper lower)))
	       (lower-extra (if (or glog?
				    (negative? lower))
				(/ extra 2)
				(min lower (/ extra 2))))
	       (upper-extra (- extra lower-extra)))
	  (set! upper (+ upper upper-extra))
	  (set! lower (- lower lower-extra))))

    ;; if we've got a series then try to make a smart choice of the Y
    ;; range to stay on when shrinking, or the new Y range when expanding
    (and-let* ((series (graph-series graph))
	       (lo     (graph-draw-x-lo graph series))
	       (hi     (graph-draw-x-hi graph series))
	       ;; possibly series entirely outside graph draw range
	       (       (>= hi lo)))
      (let* ((array    (series-array series lo hi))
	     (datatype (series-datatype series))
	     (lh       (make-array 0 (list lo hi) 2)))

	;; make lh an array of low and high for the series data
	;; the h part, ie. lh[i][1] is #f for no data
	;;
	(cond ((eq? datatype datatype-ohlcv)
	       ;; columns 1,2 for low/high
	       (set! lh (make-shared-array array
					   (lambda (t col)
					     (list t (- 2 col)))
					   (list lo hi)
					   '(0 1))))

	      ((= (array-rank array) 2)
	       (let ()
		 (do ((c-hi (array-last-column array))
		      (l    #f)
		      (h    #f)
		      (i lo (1+ i)))
		     ((> i hi))

		   (do ((j 0 (1+ j)))
		       ((> j c-hi))
		     (if l
			 (set! l (min l (or l (array-ref array i j))))
			 (set! l (array-ref array i j)))
		     (if h
			 (set! h (min h (or h (array-ref array i j))))
			 (set! h (array-ref array i j))))

		   (array-set! lh l i 0)
		   (array-set! lh h i 1))))

	      ((memq (datatype-key datatype) '(volume open-interst))
	       ;; treat volume or open interest as as 0 to N
	       (array-copy! (make-shared-array array list (list lo hi))
			    (make-shared-array-column lh 1)))

	      (else
	       (array-copy! (make-shared-array array list (list lo hi))
			    (make-shared-array-column lh 0))
	       (array-copy! (make-shared-array array list (list lo hi))
			    (make-shared-array-column lh 1))))

	(if (< new-page old-page)

	    ;; shrink

	    ;; work from right to left, keeping as much as possible of
	    ;; the portions currently visible
	    (do ((i hi (1- i)))
		((< i lo))
	      (and-let* ((h (array-ref lh i 1))
			 (h (glog h))
			 (l (glog (array-ref lh i 0)))
			 (  (overlap-exclusive? old-bot old-top l h)))
		(let ((bot (min-maybe new-bot (max old-bot l)))
		      (top (max-maybe new-top (min old-top h))))
		  (if (< (- top bot) new-page)
		      (begin
			(set! new-bot bot)
			(set! new-top top))))))

	    ;; expand
	    ;;
	    (let ((hi-visible #f))

	      ;; expand new-bot,new-top to include the portion lh[i]
	      ;; provided it doesn't make new-bot,new-top bigger than
	      ;; new-page
	      (define (accumulate i)
		(if (array-ref lh i 1)
		    (let ((bot (min-maybe new-bot (glog (array-ref lh i 0))))
			  (top (max-maybe new-top (glog (array-ref lh i 1)))))
		      (if (< (- top bot) new-page)
			  (begin
			    (set! new-bot bot)
			    (set! new-top top))))))

	      ;; must include current region
	      (set! new-bot old-bot)
	      (set! new-top old-top)

	      ;; work from right to left, accumulating those portions
	      ;; currently visible
	      (do ((i hi (1- i)))
		  ((< i lo))
		(if (and (array-ref lh i 1)
			 (overlap-exclusive? old-bot old-top
					     (glog (array-ref lh i 0))
					     (glog (array-ref lh i 1))))
		    (begin
		      ;; note rightmost visible portion
		      (set! hi-visible (or hi-visible i))
		      (accumulate i))))

	      ;; work from the rightmost visible portion towards the
	      ;; right, trying to make that region at the right visible
	      (if hi-visible
		  (do ((i hi-visible (1+ i)))
		      ((> i hi))
		    (accumulate i)))

	      ;; finally work from right to left on everything else
	      (do ((i (or hi-visible hi) (1- i)))
		  ((< i lo))
		(accumulate i))))))

    ;; if there was no data at all then just say centred at the
    ;; current position
    (if (not new-top)
	(begin
	  (set! new-bot old-bot)
	  (set! new-top old-bot)))

    ;; new page and possibly expanded upper,lower
    (c-gtk-adjustment-change
     vadj
     #:lower          lower
     #:upper          upper
     #:page-size      new-page
     #:step-increment (* scrollbar-step-fraction new-page)
     #:page-increment (* scrollbar-page-fraction new-page))

    ;; centre on chosen new-bot,new-top
    (c-gtk-adjustment-set-value-centre
     vadj
     (arithmetic-mean new-bot new-top))))

(define-public (graph-series graph)
  (first-or-f (graph-series-list graph)))

;; return the list of series GRAPH is showing, possibly empty
(define-public (graph-series-list graph)
  (rec-series-list (c-prop graph)))

;; set GRAPH to show SERIES-LIST, a list of series
;; SERIES-LIST can be empty to show nothing
(define-public (graph-series-list! graph series-list)
  (define rec (c-prop graph))
  (rec-series-list! rec series-list)
  (if (null? series-list)
      (begin
	(rec-upper! rec #f)
	(rec-lower! rec #f)
	(rec-upperlog! rec #f)
	(rec-lowerlog! rec #f))
      (begin
	(graph-range-update-data graph)
	(graph-range-update-notes graph)
	(scale-y-initial graph)
	(graph-element-expand-all graph)))
  (c-gtk-widget-queue-redraw graph))

;; return PROC to be called (PROC t) returning X pixel for start of time T
(define-public (graph-scale-x-proc graph)
  (let ((lo (round-exact (gtk-adjustment-value (graph-hadjustment graph)))))
    (lambda (t)
      (* scale-x-step (- t lo)))))

;; return X pixel for start of time T
(define-public (graph-scale-x graph t)
  ((graph-scale-x-proc graph) t))

;; return PROC to be called (PROC x) returning tdate T for pixel X
;; GRAPH can be in another timebase, conversion to tdate is made
(define-public (graph-x->tdate-proc graph)
  (let ((lo       (round-exact (gtk-adjustment-value
				(graph-hadjustment graph))))
	(t->tdate (timebase-convert-proc (series-timebase
					  (graph-series graph))
					 timebase-days)))
    (lambda (x)
      (t->tdate (+ lo (quotient x scale-x-step))))))

(define-public (graph-x->tdate graph x)
  ((graph-x->tdate-proc graph) x))

(define-public (graph-scale-y-proc graph)
  (let* ((rec            (c-prop graph))
	 (vadj           (rec-vadjustment rec))
	 (scale-y-base   (c-gtk-adjustment-value-above vadj))
	 (scale-y-factor (rec-scale-y-factor rec)))
    (if (rec-logarithmic rec)
	(lambda (price)
	  (round-exact (* scale-y-factor (- scale-y-base
					    (log (max 0.00001 price))))))
	(lambda (price)
	  (round-exact (* scale-y-factor (- scale-y-base price)))))))

(define-public (graph-scale-y graph price)
  ((graph-scale-y-proc graph) price))

(define-public (graph-y->price graph y)
  (round-exact (graph-y->price-inexact graph y)))

(define-public (graph-y->price-inexact graph y)
  (let* ((rec            (c-prop graph))
	 (vadj           (rec-vadjustment rec))
	 (scale-y-base   (c-gtk-adjustment-value-above vadj))
	 (scale-y-factor (rec-scale-y-factor rec)))
    (if (rec-logarithmic rec)
	(exp (- scale-y-base (/ y scale-y-factor)))
	(- scale-y-base (/ y scale-y-factor)))))

(define-public (graph-log graph x)
  (if (rec-logarithmic (c-prop graph))
      (log x)
      x))
(define-public (graph-exp graph x)
  (if (rec-logarithmic (c-prop graph))
      (exp x)
      x))
(define (graph-log-proc graph)
  (if (rec-logarithmic (c-prop graph))
      log
      noop))

;; return date in graph timebase, corresponding to the given X coordinate
(define-public (graph-x->date graph x)
  (+ (round-exact (gtk-adjustment-value (graph-hadjustment graph)))
     (quotient x scale-x-step)))

(define-public (graph-hadjustment graph)
  (view-hadjustment (component-view graph)))

(define-public (graph-timebase graph)
  (view-timebase (component-view graph)))

;; return true if range (alo,ahi) overlaps range (blo,bhi)
;; each endpoint is taken as exclusive, so say (1,4) and (4,7) don't overlap
(define (overlap-exclusive? alo ahi blo bhi)
  (and (> ahi blo)
       (> bhi alo)))

;; return true if range (alo,ahi) overlaps range (blo,bhi)
;; each endpoint is taken as inclusive, so say (1,4) and (4,7) do overlap
(define-public (overlap-inclusive? alo ahi blo bhi)
  (and (>= ahi blo)
       (>= bhi alo)))
       

;;-----------------------------------------------------------------------------
;; centre on visible part

;; as much from the right as will fit under the current scale
;; return #f if no series
(define (graph-centre-vertical-pos graph)
  (or (and-let* ((series (graph-series graph)))
	(let* ((datatype (series-datatype series))
	       (lo       (graph-draw-x-lo draw series))
	       (hi       (graph-draw-x-hi draw series))
	       (array    (series-array series lo hi))
	       (col-lo   (and (>= (array-rank array) 2)
			      (array-first-column array)))
	       (col-hi   (and (>= (array-rank array) 2)
			      (array-last-column array)))
	       (vadj     (graph-vadjustment graph))
	       ;; less 5% at each end for margin
	       (notes    (if (or (not (null? (series-notes series)))
				 (not (null? (series-dividends series))))
			     (/ (draw-notes-y-size graph)
				(graph-scale-y-factor graph))
			     0))
	       (page     (- (* 0.9 (gtk-adjustment-page-size vadj))
			    notes))
	       (top      #f)
	       (bot      #f)
	       (log-proc (graph-log-proc graph)))

	  ;; extend top or bot to VAL, if range then doesn't exceed page
	  (define (accumulate-value val)
	    (if val
		(begin
		  (set! val (log-proc val))
		  (set! top (or top val))
		  (set! bot (or bot val))
		  (let* ((new-top  (max top val))
			 (new-bot  (min bot val))
			 (new-size (- new-top new-bot)))
		    (if (<= new-size page)
			(begin
			  (set! top new-top)
			  (set! bot new-bot)))))))

	  (define (accumulate-ohlcv i)
	    (accumulate-value (array-ref array i 3))   ;; close
	    (accumulate-value (array-ref array i 0))   ;; open
	    (accumulate-value (array-ref array i 1))   ;; high
	    (accumulate-value (array-ref array i 2)))  ;; low

	  (define (accumulate-one i)
	    (accumulate-value (array-ref array i)))

	  (define (accumulate-multi i)
	    (do ((j col-lo (1+ j)))
		((> j col-hi))
	      (accumulate-value (array-ref array i j))))

	  ;; latest record
	  (and-let* ((       (eq? timebase-days (series-timebase series)))
		     (       (eq? datatype-ohlcv datatype))
		     (symbol (series-symbol series))
		     (latest (latest-get symbol)))
	    (and-let* ((quote-tdate (latest-quote-tdate latest))
		       (            (within-range? lo hi quote-tdate)))
	      (set! latest (latest-for-series latest series))
	      (accumulate-value (latest-bid latest))
	      (accumulate-value (latest-offer latest)))
	    (and-let* ((last-tdate (latest-last-tdate latest))
		       (           (within-range? lo hi last-tdate)))
	      (set! latest (latest-for-series latest series))
	      (accumulate-value (latest-last latest))
	      (accumulate-value (latest-open latest))
	      (accumulate-value (latest-high latest))
	      (accumulate-value (latest-low latest))))

	  ;; from right to left
	  (let ((accumulate (cond ((eq? datatype-ohlcv datatype)
				   accumulate-ohlcv)
				  (col-lo
				   accumulate-multi)
				  (else
				   accumulate-one))))
	    (do ((i hi (1- i)))
		((< i lo))
	      (accumulate i)))

	  (and top
	       (+ (arithmetic-mean top bot) (* 0.5 notes)))))
      0))

(define-public (graph-centre-vertical graph)
  (and-let* ((value (graph-centre-vertical-pos graph)))
    (c-gtk-adjustment-set-value-centre (graph-vadjustment graph) value)))


;;-----------------------------------------------------------------------------
;; logarithmic toggle

;; set ADJ so the value TOP is the upper end of the thumb
(define-public (c-gtk-adjustment-set-value-above adj top)
  (gtk-adjustment-set-value adj (- top (gtk-adjustment-page-size adj))))

(define-public (graph-logarithmic graph)
  (rec-logarithmic (c-prop graph)))

(define-public (graph-logarithmic! graph state)
  (set! state (->bool state))
  (let ((rec (c-prop graph)))
    (if (not (eq? state (rec-logarithmic rec)))
	(begin
	  (and-let* ((series (graph-series graph)))
	    (let* ((vadj        (rec-vadjustment rec))
		   (win         (gtk-widget-window graph))
		   (height      (c-gdk-drawable-get-height win))
		   (lo          (graph-x-lo graph))
		   (hi          (graph-x-hi graph))
		   (range       (series-range series lo hi))
		   (price-lower (or (first range) 0.1))
		   (price-upper (or (second range) 1))
		   (scale-y     (graph-scale-y-proc graph))
		   (lower-y     (scale-y price-lower))
		   (upper-y     (scale-y price-upper))
		   (conv-log    (if state log noop))
		   (conv-exp    (if state exp noop))
		   (new-lower   (conv-log price-lower))
		   (new-upper   (conv-log price-upper))
		   (factor      (exact->inexact (/ (- lower-y upper-y)
						   (- new-upper new-lower))))
		   (new-page    (/ height factor))
		   (new-top     (+ new-upper (/ upper-y factor)))

		   (new-bot     (- new-lower (/ (- height lower-y) factor))))

	      (rec-scale-y-factor! rec factor)
	      (c-gtk-adjustment-change
	       vadj
	       #:page-size      new-page
	       #:step-increment (* scrollbar-step-fraction new-page)
	       #:page-increment (* scrollbar-page-fraction new-page))

	      ;; perhaps bigger scroll range, before setting new value
	      (graph-range-expand graph (conv-exp new-top) (conv-exp new-bot))
	      (rec-logarithmic! rec state)
	      (graph-range-apply graph)
	      (gtk-adjustment-set-value vadj new-bot)

	      (c-gtk-widget-queue-redraw graph)
	      (c-gtk-widget-queue-redraw (view-heading ;; redraw heading
					  (component-view graph)))))

	  (rec-logarithmic! rec state)))))


;;-----------------------------------------------------------------------------
;; ChartGraph drawing

;; the highest visible in HADJ, inclusive
(define-public (hadjustment-hi hadj)
  (1- (round-exact (c-gtk-adjustment-value-above hadj))))

;; low and high visible timebase values
(define-public (graph-x-lo graph)
  (round-exact (gtk-adjustment-value (graph-hadjustment graph))))
(define-public (graph-x-hi graph)
  (hadjustment-hi (graph-hadjustment graph)))

;; return the low index to use for drawing SERIES in GRAPH, this is the
;; start of the window, restricted to the start of the series
;; optional ADJ can be -1 to go one past the start of the window (to draw a
;; line segment coming in for instance)
(define-public (graph-draw-x-lo graph series . adj)
  (max (apply + (graph-x-lo graph) adj)
       (series-lo series)))

;; return the high index to use for drawing SERIES in GRAPH, this is the
;; end of the window, restricted to the end of the series
;; optional ADJ can be 1 to go one past the end of the window (to draw a
;; line segment going out for instance)
(define-public (graph-draw-x-hi graph series . adj)
  (min (apply + (graph-x-hi graph) adj)
       (series-hi series)))

(define (draw-series-list graph)
  (let ((fg-gc (gtk-style-fg-gc (gtk-widget-get-style graph) 'normal)))
    (for-each (lambda (series)
		(let* ((datatype (series-datatype series))
		       (lstyle   (datatype-linestyle datatype)))
		  (if (eq? lstyle 'ohlc)
		      (set! lstyle
			    (first (first (preference-get 
					   'viewstyle viewstyle-default)))))
		  ((second (or (assq lstyle chartdraw-linestyle-list)
			       (list 'line line-draw "")))
		   graph fg-gc series)))
	      (graph-series-list graph))))
(graph-element-new #:proc draw-series-list)


;;-----------------------------------------------------------------------------
;; ChartGraph line styles

(define-public chartdraw-linestyle-list '())
(define (chartdraw-linestyle! key proc desc)
  (set! chartdraw-linestyle-list (cons (list key proc desc)
				       chartdraw-linestyle-list)))

(chartdraw-linestyle! 'bands
 		      (lambda args (apply bands-draw args))
		      #f)  ;; hidden
(chartdraw-linestyle! 'bar
		      (lambda args (apply bar-draw args))
		      (_ "Bars"))
(chartdraw-linestyle! 'candle
		      (lambda args (apply candles-draw args))
		      (_ "Candles"))
(chartdraw-linestyle! 'hl
		      (lambda args (apply hl-draw args))
		      (_ "HL Figures"))
(chartdraw-linestyle! 'line
 		      (lambda args (apply line-draw args))
		      (_ "Line"))
(chartdraw-linestyle! 'ohlc
		      (lambda args (apply ohlc-draw args))
		      (_ "OHLC Figures"))
(chartdraw-linestyle! 'points
		      (lambda args (apply points-draw args))
		      (_ "Points"))
(chartdraw-linestyle! 'solid
		      (lambda args (apply solid-draw args))
		      (_ "Line solid"))
(chartdraw-linestyle! 'stops
		      (lambda args (apply stops-draw args))
		      #f)  ;; hidden
(chartdraw-linestyle! 'zigzag
		      (lambda args (apply zigzag-draw args))
		      #f)  ;; hidden

(define (aroon-draw graph gc series)
  (line-draw graph gc series #:column 0 #:colour graph-up-colour)
  (line-draw graph gc series #:column 1 #:colour graph-down-colour)
  (line-draw graph gc series #:column 2 #:colour "white"))
(chartdraw-linestyle! 'aroon aroon-draw #f)

(define (dmi-draw graph gc series)
  (line-draw graph gc series #:column 0 #:colour graph-up-colour)
  (line-draw graph gc series #:column 1 #:colour graph-down-colour))
(chartdraw-linestyle! 'dmi dmi-draw #f)

(define (ichimoku-draw graph gc series)
  (line-draw graph gc series #:column 0 #:colour "orange")      ;; tenkan
  (line-draw graph gc series #:column 1 #:colour "dark red")    ;; kijun
  (line-draw graph gc series #:column 2 #:colour "blue")        ;; chikou
  (line-draw graph gc series #:column 3 #:colour "dark green")  ;; senkou A
  (line-draw graph gc series #:column 4 #:colour "dark green")) ;; senkou A
(chartdraw-linestyle! 'ichimoku ichimoku-draw #f)

(define (macd-draw graph gc series)
  (solid-draw graph gc series #:column 2 #:colour "grey")     ;; diff
  (line-draw graph gc series #:column 1 #:colour graph-down-colour) ;; smoothed
  (line-draw graph gc series #:column 0 #:colour graph-up-colour))  ;; macd
(chartdraw-linestyle! 'macd macd-draw #f)

(define (stochastics-draw graph gc series)
  (line-draw graph gc series #:column 0 #:colour graph-down-colour)
  (line-draw graph gc series #:column 1 #:colour graph-up-colour))
(chartdraw-linestyle! 'stochastics stochastics-draw #f)
