;; Common GUI setups.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart gui-common)
  #:autoload   (ice-9 format)       (format)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:autoload (chart error-dialog)   (error-dialog-show)
  #:autoload (chart gnome)          (gnome-help-display)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; main loop

;; setup `c-main' and friends so they use the gtk event loop etc
;;
(define-public (c-main-use-gtk)

  ;; INT is an internal time amount, per `get-internal-real-time', return
  ;; the equivalent in milliseconds
  (define (internal-time->milliseconds int)
    (quotient (* int 1000) internal-time-units-per-second))

  (set! c-gtk-input-add         gtk-input-add)
  (set! c-gtk-input-remove      gtk-input-remove)
  (set! c-gtk-timeout-add       gtk-timeout-add)
  (set! c-gtk-timeout-remove    gtk-timeout-remove)
  (set! c-gtk-idle-add-priority gtk-idle-add-priority)
  (set! c-gtk-idle-remove       gtk-idle-remove)
  ;; there's something fishy with the return from gtk-main-iteration,
  ;; pretend it always says false to keep running (no quit)
  (set! c-main-iteration        (lambda ()
				  (gtk-main-iteration)
				  #f))
  (set! c-main-stop?            (lambda () #f))
  (set! c-main-run-q c-main-run-q-with-gtk-catch)

  ;; bring any already established sources across

  (for-each (lambda (input)
	      (apply gtk-input-add input))
	    c-gtk-input-list)
  (set! c-gtk-input-list '())

  (for-each (lambda (timeout)
	      (receive-list (target-internal proc)
		  timeout
		(let ((remaining (- target-internal (get-internal-real-time))))
		  (if (negative? remaining)
		      (c-main-enq! proc)
		      (gtk-timeout-add (internal-time->milliseconds remaining)
			proc)))))
	    c-gtk-timeout-list)
  (set! c-gtk-timeout-list '())

  (for-each (lambda (idle)
	      (apply gtk-idle-add-priority idle))
	    c-gtk-idle-list)
  (set! c-gtk-idle-list '())

  (gtk-callback-trampoline
   (lambda (proc arglst)
     (apply c-main-enq! proc arglst)
     ;; default #f is once-only for idle or timeout, propagate for event, can
     ;; be overridden by `trampoline-return' setting
     (hashq-ref trampoline-return-hash proc)))

  ;; once only, and gc this code
  (set! c-main-use-gtk noop))


;; hash keyed by a PROC, with VALUE to return from trampoline
(define-public trampoline-return-hash (make-hash-table 17))

;; setup so that when PROC is called via the gtk trampoline, the return
;; value is VALUE, instead of the default #f
;;
(define-public (trampoline-return proc value)
  (hashq-set! trampoline-return-hash proc value)
  proc)


;; run any entries in `c-main-queue', each one run with an error trap which
;; opens an `error-dialog-show' to display the error
;;
(define (c-main-run-q-with-gtk-catch)
  (or (q-empty? c-main-queue)
      (let ((call-lst (deq! c-main-queue)))  ;; list of proc and its args

	(catch 'c-main-run-q-with-gtk-catch--continue
	  (lambda ()
	    (lazy-catch #t
	      (lambda ()
		(apply (car call-lst) (cdr call-lst)))
	      (lambda (key . args)
		(if (eq? key 'quit) ;; `exit' call
		    (apply throw key args))
		(let ((stack (make-stack #t 2)))
		  (c-main-enq! error-dialog-show stack key args)
		  (throw 'c-main-run-q-with-gtk-catch--continue)))))
	  (lambda (key . args)
	    #f))
	(c-main-run-q))))


;;-----------------------------------------------------------------------------
;; rc styles

;; font = \"fixed\"

(let-toplevel ((str "
style \"Chart_style\"
  {
    fg[ACTIVE]        = { 1.0, 1.0, 1.0 }
    fg[NORMAL]        = { 1.0, 1.0, 1.0 }
    fg[PRELIGHT]      = { 1.0, 1.0, 1.0 }
    fg[SELECTED]      = { 1.0, 1.0, 1.0 }
    fg[INSENSITIVE]   = { 1.0, 1.0, 1.0 }
    text[ACTIVE]      = { 1.0, 1.0, 1.0 }
    text[NORMAL]      = { 1.0, 1.0, 1.0 }
    text[PRELIGHT]    = { 1.0, 1.0, 1.0 }
    text[SELECTED]    = { 1.0, 1.0, 1.0 }
    text[INSENSITIVE] = { 1.0, 1.0, 1.0 }

    bg[ACTIVE]        = { 0, 0, 0 }
    bg[NORMAL]        = { 0, 0, 0 }
    bg[PRELIGHT]      = { 0, 0, 0 }
    bg[SELECTED]      = { 0, 0, 0 }
    bg[INSENSITIVE]   = { 0, 0, 0 }
    base[ACTIVE]      = { 0, 0, 0 }
    base[NORMAL]      = { 0, 0, 0 }
    base[PRELIGHT]    = { 0, 0, 0 }
    base[SELECTED]    = { 0, 0, 0 }
    base[INSENSITIVE] = { 0, 0, 0 }
  }
"))

  ;; install GtkRcStyle setups to make WIDGET-NAME use "Chart_style"
  (define-public (c-gtk-rc-chart-style widget-name)
    (if str
        (begin
	  (gtk-rc-parse-string str)
	  (set! str #f)))
    (gtk-rc-parse-string
     (simple-format #f "widget \"*.~a\" style \"Chart_style\"" widget-name))))


;;-----------------------------------------------------------------------------
;; error dialog

(define-public (error-dialog-popup parent message)
  (let ((dialog (gtk-message-dialog-new parent
					'(modal destroy-with-parent)
					'error 'ok message)))
    (gtk-signal-connect dialog "response"
      (lambda (id)
	(gtk-widget-destroy dialog)))  ;; ok button, or delete-event
    (gtk-widget-show-all dialog)))


;;-----------------------------------------------------------------------------
;; manual

;; NODE is a texinfo file node name, return the corresponding anchor emitted
;; by makeinfo in a HTML file.
;;
;; The transformation is described in the "HTML Xref Node Name Expansion"
;; section of the texinfo manual:
;;    A-Z 0-9 - unchanged (except 0-9 at start of name, which we don't use)
;;    space   - becomes dash "-"
;;    other   - becomes "_00XX" with XX the ascii in lower case hex
;;
(define (texinfo-node->html-anchor node)
  (regexp-substitute/global #f "[^A-Za-z0-9]" node
			    'pre
			    (lambda (m)
			      (let ((c (string-ref (match:substring m) 0)))
				(case c
				  ((#\space) "-")
				  (else (format #f "_00~2,'0x"
						(char->integer c))))))
			    'post))

;; Open the manual using the help browser.
;; NODE is a texinfo node name or anchor name string, or omitted or #f for
;; the start of the manual
;; PARENT-WIDGET is the originating main window or dialog window widget,
;; this is only used for an error display
;;
(define*-public (manual-popup #:optional node parent-widget)
  (if (string? node)
      (set! node (texinfo-node->html-anchor node)))
  (catch 'g-error
    (lambda ()
      (gnome-help-display "chart.html" node))
    (lambda (key domain code message)
      (error-dialog-popup parent-widget
			  (string-append (_ "Cannot display manual:\n")
					 message)))))
