;; Guppy multiple moving averages (GMMA).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart guppy-mma)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.guppytraders.com/gup329.shtml
;;     Summary.
;;
;; http://www.chartfilter.com/articles/movingaverage.htm
;;     2004 article by Darryl Guppy.
;;
;; http://www.incrediblecharts.com/technical/multiple_moving_averages.htm
;;     Sample of ASX.AX from 1999.
;;

(define (guppy-mma-warmup-count count-list)
  (ema-warmup-count (apply max count-list)))

(define (guppy-mma-fill series new-array lo hi)
  (let* ((count-list  (series-data series))
	 (parent      (series-parent series))
	 (array       (series-array-closes parent lo hi))
	 (proc-list   (map ema-calc-proc count-list))
	 (column-list (iota (length count-list))))

    ;; warmup
    (for-each (lambda (close)
		(for-each (lambda (proc)
			    (proc close))
			  proc-list))
	      (series-prev-n-closes
	       parent #:before lo
	       #:count (guppy-mma-warmup-count count-list)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i)))
	(for-each (lambda (proc column)
		    (array-set! new-array (proc close) i column))
		  proc-list column-list)))))

(define-public (guppy-mma-series series . count-list)
  (set! count-list (delv 0 count-list))
  (let ((new-series (series-copy series)))
    (series-name!     new-series
		      (format #f (_ "GMMA ~a")
			      (string-join (map number->string count-list)
					   (_ ","))))
    (series-datatype! new-series guppy-mma-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count-list)
    (series-require!  new-series
		      (indicator-require-parts
		       guppy-mma-fill
		       (guppy-mma-warmup-count count-list)
		       (list 0 (1- (length count-list)))))
    (series-array!    new-series #f)

    new-series))

(define-public guppy-mma-datatype
  (datatype-new!
   #:key    'guppy-mma
   #:name   (_ "Gupply Multiple MA")
   #:units  #f
   #:manual (_ "CONTEXT(manual-node): Guppy Multiple Moving Average")
   #:proc   guppy-mma-series
   #:closes-apply-arg '(0)
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "Days")
	  #:key      'guppy-mma-days-3
	  #:type     'integer
	  #:default  3
	  #:minimum  1)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-5
	  #:type     'integer
	  #:default  5
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-8
	  #:type     'integer
	  #:default  8
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-10
	  #:type     'integer
	  #:default  10
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-12
	  #:type     'integer
	  #:default  12
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-15
	  #:type     'integer
	  #:default  15
	  #:minimum  0)

	 (datatype-param-new
	  #:name     (_ "CONTEXT(long-moving-average): Long")
	  #:key      'guppy-mma-days-30
	  #:type     'integer
	  #:default  30
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-35
	  #:type     'integer
	  #:default  35
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-40
	  #:type     'integer
	  #:default  40
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-45
	  #:type     'integer
	  #:default  45
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-50
	  #:type     'integer
	  #:default  50
	  #:minimum  0)
	 (datatype-param-new
	  #:name     ""
	  #:key      'guppy-mma-days-60
	  #:type     'integer
	  #:default  60
	  #:minimum  0))))
