;; gzip program functions.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart gzip)
  #:use-module (ice-9 optargs)
  #:autoload   (ice-9 popen)   (open-input-pipe)
  #:use-module (srfi srfi-13)
  #:use-module (chart http)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; misc

;; Return a procedure which runs COMMAND (a string) to get a version number
;; string from a program.  The new procedure returns a string or #f.
;;
;; TRANSFORM is a procedure called (TRANSFORM str) with the output of
;; COMMAND, it should pick out the desired version number part.
;;
;; If COMMAND is available, the version string is cached so it's not run
;; again.  If COMMAND is not available then it's only attempted at most once
;; a second, in case the created procedure is getting tested repeatedly.
;;
(define*-public (program-version-proc command transform #:key (want-exit 0))
  (define last-attempt #f)
  (make-procedure-with-cache-0-true
   (lambda ()
     (if (eqv? last-attempt (current-time))
	 #f
	 (let* ((pipe   (open-input-pipe (string-append "(" command ") 2>&1")))
		(str    (read-string-to-eof pipe))
		(status (close-pipe pipe)))
	   (set! last-attempt (current-time))
	   (and (eqv? want-exit (status:exit-val status))
		(transform str)))))))

;; return a procedure which calls (THUNK)
;; if THUNK returns true, that value is cached and it's not called again
;; if THUNK returns #f, it's called each time
;;
(define-public (make-procedure-with-cache-0-true thunk)
  (or (thunk? thunk)
      (error "make-procedure-with-cache-0-true: bad function:" thunk))
  (set! thunk (let ((uncached-thunk thunk))
		(lambda ()
		  (let ((result (uncached-thunk)))
		    (if result
			(set! thunk (make-procedure-returning result)))
		    result))))
  (lambda ()
    (thunk)))

;; trim PART from the left of STR
(define-public (string-trim-str str part)
  (if (string-prefix? part str)
      (string-drop str (string-length part))
      str))

;; Run "COMMAND tempfile" where tempfile contains the given STR, and return
;; the output as a string.  The exit status from COMMAND should be 0.
;;
;; This uses a temporary file because it's easier than opening two pipes and
;; reading one at the same time as writing the other.
;;
(define*-public (program-filter-string
		 command str
		 #:key
		 (ext "")
		 (exit-test (lambda (status)
			      (eqv? 0 (status:exit-val status)))))
  (let ((command-proc (if (string? command)
			  (lambda (filename)
			    (string-append command " " filename))
			  command)))
    (call-with-tempfileport (mkstemp-ext ext)
      (lambda (fileport)
	(let ((filename (port-filename fileport)))
	  (display str fileport)
	  (close-port fileport)

	  (let* ((pipeport (open-input-pipe (command-proc filename)))
		 (out-str  (read-string-to-eof pipeport))
		 (status   (close-pipe pipeport)))
	    (or (exit-test status)
		(error "chart: subprocess non-zero exit:"
		       command (status:exit-val status)))

	    out-str))))))


;;-----------------------------------------------------------------------------
;; gzip

;; a call "(gzip-version)" returns a version number string like "1.3.12" for
;; the gzip program, or #f if not available
;;
(define-public gzip-version
  (program-version-proc "gzip --version"
    (lambda (str)
      (set! str (string-first-line str))
      (string-trim-str str "gzip "))))

;; gunzip-string takes a compressed DATA string and returns decompressed
(define-public (gunzip-string data)
  (program-filter-string "gzip -cd" data))

