;; ChartHeading widget and graph elements.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart heading)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart gui-common)
  #:use-module (chart misc)
  #:use-module (chart view))


(define-public heading-string-alist
  (list (cons timebase-days     (format #f (_ "to ~a") (locale-wd-fmt)))
	(cons timebase-weeks    (format #f (_ "to week ~a") (locale-d-fmt)))
	(cons timebase-months   (_ "to %b %Y"))
	(cons timebase-quarters (_ "to %b %Y"))
	(cons timebase-years    (_ "to %Y"))
	(cons timebase-decades  (_ "to decade %Y"))))


;;-----------------------------------------------------------------------------
;; misc

(define (dividend-adjustment-active?)
  (viewstyle-dividend (preference-get 'viewstyle viewstyle-default)))


;;-----------------------------------------------------------------------------
;; widget

(gtk-class-new 'GtkDrawingArea "ChartHeading")
(c-gtk-rc-chart-style "ChartHeading")

(define rec-type (make-record-type "chartheading" '(series)))
(define rec-new  (record-constructor rec-type '()))

(define (heading-string heading graph)
  (and-let* ((       (eq? graph draw))
	     (series (graph-series graph)))
    (let* ((timebase   (series-timebase series))
	   (done-tdate (series-info-get series 'download-done))
	   (hi-tdate   (timebase-convert timebase timebase-days (series-hi series)))
	   (to-tdate   (max-maybe done-tdate hi-tdate)))

      (string-append
       (string-join-notnull (list (series-symbol series)
				  (series-name series))
			    " - ")
       "     "
       (strftime-utf8 (or (assq-ref heading-string-alist timebase)
			  (locale-d-fmt))
		      (tdate->tm to-tdate))
       "       "))))

(define (heading-string-left graph)
  (let ((series-list (graph-series-list graph)))
    (if (and (not (null? series-list))
	     (eq? graph draw))
	(set! series-list (cdr series-list)))

    ;; lose top series, look only at derived ones
    (set! series-list (filter series-parent series-list))

    (let ((name-list (delq #f (map series-name series-list))))
      (string-join-notnull (append name-list
				   (list (and (graph-logarithmic graph)
					      (_ "Log"))
					 (and (eq? graph draw) ;; upper only
					      (dividend-adjustment-active?)
					      (_ "Div adj"))))
			   " - "))))

(define-public (heading-new graph)
  (let ((heading (gtk-widget-new 'ChartHeading))
	(rec     (rec-new)))

    (define (draw)
      (let* ((win       (gtk-widget-window heading))
	     (win-width (c-gdk-drawable-get-width win))
	     (style     (gtk-widget-get-style heading))
	     (text-gc   (gtk-style-text-gc style 'normal))
	     (layout    (primitive-property-ref c-layout-prop heading)))

	;; left
	(and-let* ((str (heading-string-left graph)))
	  (pango-layout-set-text layout str)
	  (gdk-draw-layout win text-gc 0 0 layout))

	;; right
	(and-let* ((str (heading-string heading graph)))
	  (gdk-draw-layout win text-gc
			   (- win-width (c-pango-layout-width layout str))
			   0 layout))))

    (define (update-usize . args)
      (let* ((layout (primitive-property-ref c-layout-prop heading))
	     (height (c-pango-layout-line-height layout)))
	(gtk-widget-set-usize heading -1 height)
	(queue-redraw)))

    (define (queue-redraw . args)
      (c-gtk-widget-queue-redraw heading))

    (set! (c-prop heading) rec)
    (c-gtk-widget-drawing heading draw)
    (gtk-signal-connect heading "style_set" update-usize)
    (gtk-signal-connect heading "configure-event" queue-redraw)
    (gtk-signal-connect heading "parent-set"
      (lambda (old-parent)
	(gtk-signal-connect (component-view heading)
	    "series-changed"
	  queue-redraw)))
    (update-usize)

    heading))


;;-----------------------------------------------------------------------------
;; drawing element

(define-public (heading-draw-in-graph graph)
  (if (not (eq? graph draw))
      (let* ((win    (gtk-widget-window graph))
	     (style  (gtk-widget-get-style graph))
	     (layout (primitive-property-ref c-layout-prop graph)))
	(pango-layout-set-text layout
			       (string-append " "
					      (heading-string-left graph)
					      " "))
	(let* ((text-gc     (gtk-style-text-gc style 'normal))
	       (font-height (c-pango-layout-line-height layout))
	       (y-margin    (round-exact (* 0.3 font-height))))
	  (gdk-draw-layout win text-gc 0 y-margin layout)))))
