;; ChartHScale widget.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart hscale)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart gui-common)
  #:use-module (chart misc)
  #:use-module (chart view))


;; maximum width in LAYOUT of any strftime string from FORMAT (in utf8)
(define (c-pango-layout-strftime-width layout format)
  (let ((tm       (tm-new))
	(maxwidth 0))
    (define (iterate setter lo hi)
      (do ((imax lo)
	   (i    lo (1+ i)))
	  ((> i hi)
	   (setter tm imax))
	(setter tm i)
	(let ((width (c-pango-layout-width layout (strftime-utf8 format tm))))
	  (if (> width maxwidth)
	      (begin
		(set! maxwidth width)
		(set! imax     i))))))
    (set-tm:hour tm 23)
    (set-tm:min  tm 59)
    (set-tm:sec  tm 59)
    (iterate set-tm:mday 1 30)
    (iterate set-tm:mon  0 11)
    (iterate set-tm:year 0 199)  ;; 1900 to 2099
    maxwidth))


;;-----------------------------------------------------------------------------

(gtk-class-new 'GtkDrawingArea "ChartHScale")
(c-gtk-rc-chart-style "ChartHScale")

;; trailing spaces force a gap to the next label drawn
(define-public hscale-month-formats '("%B %Y  "
				      "%b %Y  "
				      "%b %y  "))
(define-public hscale-year-formats '("%Y  "
				     "%y "))

(define hscale-mark-choices (list timebase-days
				  timebase-weeks
				  timebase-months
				  timebase-quarters
				  timebase-years
				  timebase-decades))

(define small-tick-fraction    0.2)
(define medium-tick-fraction   0.4)
(define medium-gap-fraction    0.1)
(define label-left-em-fraction 0.7)

(define format-choices
  (list (list timebase-months
	      timebase-quarters
	      timebase-years
	      timebase-decades)
	(list hscale-month-formats
	      hscale-month-formats
	      hscale-year-formats
	      hscale-year-formats)))

(define-public (hscale-new hadj)
  (define hscale (gtk-widget-new 'ChartHScale))

  (define (draw)
    (let* ((view   (component-view hscale))
	   (hadj   (view-hadjustment view))
	   (lo     (round-exact (gtk-adjustment-value hadj)))
	   (hi     (hadjustment-hi hadj)))
      (if (> hi lo)  ;; suppress for initial no display
	  (let* ((win            (gtk-widget-window hscale))
		 (layout         (primitive-property-ref c-layout-prop hscale))
		 ;; (width          (c-gdk-drawable-get-width win))
		 (height         (c-gdk-drawable-get-height win))
		 (style          (gtk-widget-get-style hscale))
		 (font-height    (c-pango-layout-line-height layout))
		 (em             (c-pango-layout-em layout))
		 (label-x-offset (ceiling-exact
				  (* label-left-em-fraction em)))
		 (fg-gc          (gtk-style-fg-gc style 'normal))
		 (text-gc        (gtk-style-text-gc style 'normal))
		 (small-y        (ceiling-exact
				  (* font-height small-tick-fraction)))
		 (medium-y       (ceiling-exact
				  (* font-height medium-tick-fraction)))
		 (label-y        (- height font-height))
		 (timebase       (view-timebase view)))

	    (define (scale-x t)
	      (* scale-x-step (- t lo)))

	    (define (tick-marks step-tb y)
	      (let ((s->d (timebase-convert-proc step-tb timebase))
		    (d->s (timebase-convert-proc timebase step-tb)))
		(gdk-draw-segments
		 win fg-gc
		 (map (lambda (x)
			(cons (cons x 0) (cons x y)))
		      (map (lambda (s) (scale-x (s->d s)))
			   (iota-range (d->s lo) (1+ (d->s hi))))))))

	    (define (tick-labels step-tb format)
	      (let ((s->d     (timebase-convert-proc step-tb timebase))
		    (d->s     (timebase-convert-proc timebase step-tb))
		    (s->tdate (timebase-convert-proc step-tb timebase-days)))
		(do ((shi (d->s hi))
		     (s   (d->s lo) (1+ s)))
		    ((> s shi))
		  (let ((x (scale-x (s->d s))))
		    (gdk-draw-line win fg-gc x 0 x height)
		    (pango-layout-set-text
		     layout (strftime-utf8 format (tdate->tm (s->tdate s))))
		    (gdk-draw-layout win text-gc
				     (+ x label-x-offset) label-y layout)))))

	    ;; return proc for drawing labels, or #f
	    (define (choose-one step-tb format-list)
	      (let ((width-available
		     (- (* (- (timebase-convert step-tb timebase 1)
			      (timebase-convert step-tb timebase 0))
			   scale-x-step)
			label-x-offset)))
		(and-let* ((format
			    (find (lambda (format)
				    (>= width-available
					(c-pango-layout-strftime-width
					 layout format)))
				  format-list)))
		  (lambda ()
		    (tick-labels step-tb format)))))

	    ;; return list of drawing procs
	    (define (choose-drawing)
	      (append (and-let* ((ch (memq timebase hscale-mark-choices)))
			(list (lambda ()
				(tick-marks timebase small-y))
			      (if (not (null? (cdr ch)))
				  (lambda ()
				    (tick-marks (second ch) medium-y))
				  noop)))
		      (list (or (apply any choose-one  format-choices)
				noop))))

	    (for-each (lambda (proc)
			(proc))
		      (choose-drawing))))))

  (define (update-usize . dummy-args)
    (let* ((layout (primitive-property-ref c-layout-prop hscale))
	   (font-height (c-pango-layout-line-height layout))
	   (height      (+ (ceiling-exact (* font-height
					     medium-tick-fraction))
			   (ceiling-exact (* font-height
					     medium-gap-fraction))
			   font-height)))
      (gtk-widget-set-usize hscale -1 height))
    (queue-redraw))

  (define (queue-redraw . dummy-args)
    (c-gtk-widget-queue-redraw hscale))

  (update-usize)
  (c-gtk-widget-drawing hscale draw)
  (gtk-signal-connect hscale "style_set" update-usize)
  (gtk-signal-connect hscale "configure-event" queue-redraw)

  (gtk-signal-connect hadj "value-changed" queue-redraw)
  (gtk-signal-connect hadj "changed"       queue-redraw)

  hscale)



