;; HTML input form functions.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart html-form)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-13)
  #:use-module (chart http)
  #:use-module (chart misc))


;; TAGS is a string like "<input name="foo" value="bar" />"
;; return an alist like (("name" . "foo") ("value" . "bar"))
;;
;; Values are either a bare name (which is supposed to be letters, digits,
;; periods and hyphens), or a string "foo" or 'foo' with the quote char not
;; occurring in the string.  "&foo;" entities can give those chars.  (Per
;; RFC 1866.)
;;
(define (html-tag-parameters-alist tag)
  (map (lambda (m)
	 (cons (match:substring m 1)
	       (or (match:substring m 3)
		   (match:substring m 4)
		   (match:substring m 5))))
       (list-matches
	"([^ \t\r\n=]+)=(([^\"' \t\r\n][^ \t\r\n]*)|\"([^\"]*)\"|'([^']*)')"
	tag)))
  
(define*-public (html-form-parse body #:optional (pos 0))

  (let* ((m     (or (string-match "<form[^>]*" body pos)
		    (error "html-form-parse: no <form>")))
	 (end   (or (string-contains-ci body "</form" (match:end m 0))
		    (error "html-form-parse: no </form>")))
	 (ret   (html-tag-parameters-alist (match:substring m 0))))
    (set! body (substring/shared body (match:start m 0) end))

    (append! ret
	     (map (lambda (m)
		    (cons 'input
			  (html-tag-parameters-alist (match:substring m 0))))
		  (list-matches (make-regexp "<(input|select)[^>]*>" regexp/icase)
				body)))))

(define-public (html-form-data form-alist values-alist)
  (string-join-notnull
   (map (lambda (elem)
	  (and (eq? (car elem) 'input)
	       (let* ((alist (cdr elem))
		      (type  (or (assoc-ref alist "type") "text")))
		 (cond ((string=? type "image")
			;; image gets "name.x" and "name.y" coords,
			;; optionally given in VALUES-ALIST
			(and-let* ((name (assoc-ref alist "name")))
			  (let ((name-x (string-append name ".x"))
				(name-y (string-append name ".y")))
			    (string-append
			     (url-quote name-x) "="
			     (or (assoc-ref values-alist name-x) "0") "&"
			     (url-quote name-y) "="
			     (or (assoc-ref values-alist name-y) "0")))))
		       (else
			;; ignore nameless submit buttons
			(and-let* ((name (assoc-ref alist "name")))
			  (let ((value (or (assoc-ref values-alist name)
					   (assoc-ref alist "value")
					   (error "html-form-data: no value for field:" name))))
			    (string-append (url-quote name) "="
					   (url-quote value)))))))))
	form-alist)
   "&"))

(define-public (html-form-submit-url base-url form-alist)
  (http-url-relative base-url (or (assoc-ref form-alist "action")
				  (error "html-form-submit-url: no form action field"))))
