;; HTML parsing functions.

;; Copyright 2003, 2004, 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart html)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart http)
  #:use-module (chart misc))


;; Crib notes:
;;     RFC1866  html 2.0
;;     RFC1942  html tables


;; return position after PART in STR, or #f if PART not in STR
(define-public (string-contains-after str part . rest)
  (let ((pos (apply string-contains str part rest)))
    (and pos
	 (+ pos (string-length part)))))


;; Return the charset for a html document STR, from the meta tag like
;;
;;   <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
;;
;; or from www.m-x.ca,
;;
;;   <meta http-equiv='content-Type' content='text/html; charset=iso-8859-1'/>
;;
;; Default to latin-1 as specified in RFC1945 (and subsequent) sect 3.6.1,
;; though the W3C notes this default has been pretty useless in practice
;; (http://www.w3.org/TR/html4/charset.html).
;;
;; Strictly speaking the tag args to meta are entity encoded (&foo;), but in
;; practice no need to worry about that for the charset indicator.
;;
(define-public (html-charset headers body)
  (or (and-let* ((ct (http-headers-get headers "Content-Type"))
		 (m  (regexp-exec #,(regexp "charset=([^\"';]+)") ct)))
	(match:substring m 1))

      (and-let* ((m (regexp-exec
		     #,(regexp "<meta [^>]*charset=([^\"'>]+)") body)))
	(match:substring m 1))

      "iso-8859-1")) ;; if not otherwise specified, but might be ascii-only


;; Extract the text from html STR, by dropping all <foo> </bar> forms.
;; An unclosed form like "<foo" at the end of the string is dropped too.
;;
;; &nbsp; and &ndash; are changed to plain characters
;; (not all entities are converted, just those that arise in practice)
;;
(define-public (html-text-extract str)
  ;; entity terminator ";" not needed if followed by non-alpha char
  (set! str (regexp-substitute/global
	     #f #,(regexp "&(ndash|nbsp)(;|([^A-Za-z]|$))") str
	     'pre
	     (lambda (m)
	       (string-append/shared (assoc-ref '(("ndash" . "-")
						  ("nbsp"  . " "))
						(match:substring m 1))
				     (or (match:substring m 3) "")))
	     'post))
  (regexp-substitute/global #f #,(regexp "<[^>]*>?") str 'pre 'post))


;;-----------------------------------------------------------------------------
;; table parsing
;;
;; This does enough to parse out the basic structure a table; don't need or
;; particularly want a full html parse.  Comments are recognised (since they
;; can suppress field markers).  Sub-tables are recognised and are included
;; in the field returned.


;; Return list of text row lists for a table.  Eg.
;;
;;     <table><tr> <td>one</td>   <td>two</td></tr>
;;            <tr> <td>three</td> <td>four</td></tr> </table>
;; gives
;;
;;     (("one"   "two")
;;      ("three" "four"))
;;
;; All html tags in the fields are stripped, and whitespace trimmed.
;;
(define*-public (html-table-rows body #:optional (start 0))
  (html-rows-text-extract (html-table-rows-raw body start)))

;; extract plain text from a list of row lists
(define-public (html-rows-text-extract row-list)
  (map! (lambda (row)
	  (map! string-trim-both
		(map! html-text-extract row)))
	row-list))

;; Return a raw list of rows of table elements.  Eg.
;;
;;     <table><tr> <td>one</td>   <td>two</td></tr>
;;            <tr> <td>three</td> <td>four</td></tr> </table>
;; gives
;;
;;     (("<td>one</td>"   "<td>two</td>")
;;      ("<td>three</td>" "<td>four</td>"))
;;
;; all the html elements remain in the fields returned, including <td>
;; markers.
;;
(define*-public (html-table-rows-raw body #:optional (start 0))
  (set! start (+ 6 (or (string-contains-ci body "<table" start)
		       (error "No <table> found"))))
  (call-with-C-locale
   (lambda ()
     (let ((ret   '())
	   (row   #f)
	   (field #f))
       (let more ()
	 (let ((m (regexp-exec
		   ;;           2       3        4    5     6       7                8
		   #,(regexp "<((table)|(/table)|(tr)|(/tr)|(t[hd])|(/t[hd]))[^>]*>?|(<!--)"
			     regexp/icase)
		   body start)))
	   (if m
	       (set! start (match:end m)))

	   (cond
	    ;; comment
	    ;; used by wtb.scm, the english quote page has an
	    ;; "&Auml;nderung:" row commented out, also past bits of
	    ;; hollywood.scm had a little comment in the middle of a row
	    ((and m
		  (match:start m 8)) ;; <!--
	     (set! start (or (string-contains-after body "-->" start)
			     (string-length body)))
	     (more))

	    ;; nested <table>
	    ((and m
		  (match:start m 2))      ;; table
	     (set! start (html-skip-to-</table> body start))
	     (more))

	    (else
	     ;; flush open field
	     (if field
		 (begin
		   (set! row (cons (substring body field
					      (or (and m (match:start m))
						  (string-length body)))
				   row))
		   (set! field #f)))

	     ;; flush open row
	     (and (or (not m)            ;; end of string
		      (match:start m 3)  ;; /table
		      (match:start m 4)  ;; tr
		      (match:start m 5)) ;; /tr
		  (if row
		      (begin
			(set! ret (cons (reverse! row) ret))
			(set! row #f))))

	     ;; start new row
	     (and m
		  (or (match:start m 4)      ;; tr
		      (and (match:start m 6) ;; th td, need open row
			   (not row)))
		  (set! row '()))

	     ;; start new field
	     (and m
		  (match:start m 6)      ;; th td
		  (set! field (match:start m)))

	     (if (or (not m)            ;; end of string
		     (match:start m 3)) ;; /table
		 (reverse! ret)

		 (more))))))))))

;; START is inside a <table>, return the position after the matching </table>
;; or after the last char if string ends first
;; nested tables are recognised
;;
(define (html-skip-to-</table> body start)
  (let nest ((level 1))
    (let ((m (regexp-exec #,(regexp "<((table)|(/table))[^>]*>?" regexp/icase)
			  body start)))
      (if m
	  (set! start (match:end m)))
      (cond ((not m) ;; end of string
	     (string-length body))
	    ((match:start m 2) ;; further nested <table>
	     (nest (1+ level)))
	    ((> level 1)       ;; end </table>, but still nested
	     (nest (1- level)))
	    (else              ;; end top </table>
	     start)))))
