;; HTTP functions.

;; Copyright 2003, 2004, 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart http)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 q)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (ice-9 rw)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:autoload   (chart cookies) (cookiejar-get-line cookiejar-add-headers)
  #:use-module (chart extra-glib)
  #:use-module (chart gethost)
  #:use-module (chart gzip)
  #:use-module (chart html)
  #:autoload   (chart https)   (https-open)
  #:use-module (chart misc)
  #:use-module (chart zlib))


;; Crib notes:
;;
;; RFC1945 -- http 1.0
;; RFC2068 -- http 1.1
;; RFC2616 -- http 1.1 clarified/revised


(define-public http-timeout 150) ;; seconds

;; total maximum simultaneous connections, and connections to one host
;;
;; the http proxy (if any) doesn't count here, these limits are on the final
;; destination
;;
(define http-max-connections 4)
(define http-max-to-one-host 2)

;; This is the cost to apply for doing an individual http GET.  Each GET has
;; some header bytes, but this cost is mainly about rating the slowdown
;; caused by separate round-trip requests.
;;
;; If total downloaded bytes are important to you (more than time), then set
;; this to 0.
;;
(define-public http-get-cost 3000)


;;-----------------------------------------------------------------------------

;; a procedure called as (PROC string)
(define-public http-status-callback (make-fluid))
(fluid-set! http-status-callback noop)


;;-----------------------------------------------------------------------------
;; waiting for arbitrary conditions
;;
;; A call (c-waitfor TESTPROC) blocks until the call (TESTPROC) returns
;; true.  When TESTPROC returns true the `c-waitfor' continuation runs
;; immediately, no other code runs in between.  The return from c-waitfor is
;; the return from TESTPROC.
;;
;; Multiple continuations can be waiting, they're kept in `c-waitfor-queue'
;; as pairs (TESTPROC . CONTINUATION).  `c-waitfor-check' should be called
;; when some resources have been released that might allow waiting entries
;; to run.  If a TESTPROC returns true then `c-waitfor-queue' jumps to its
;; continuation, if all are false (or there's none queued) then
;; c-waitfor-check returns normally.  Usually c-waitfor-check is for use as
;; `(c-main-enq!  c-waitfor-check)', and in fact when it jumps to an entry
;; run that's how it sets up to check for more that might run.

(define c-waitfor-queue (make-q))

(define-public (c-waitfor testproc msgproc)
  (if msgproc
      (let ((old-testproc testproc))
	(set! testproc (lambda ()
			 (or (old-testproc)
			     (begin
			       (msgproc)
			       (set! msgproc noop)
			       #f))))))
  ;; If nobody else is waiting then try TESTPROC without copying a whole
  ;; continuation.  In principle could also do that if all the other
  ;; testprocs return false, but if one of them returned true it's a bit
  ;; tricky to setup the job-pause for this continuation and then also jump
  ;; to the new one with no code running in between.
  (or (and (q-empty? c-waitfor-queue)
	   (testproc))
      (job-pause (lambda (cont)
		   (let ((elem (cons testproc cont)))
		     (enq! c-waitfor-queue elem)
		     (list elem)))
		 (lambda (elem)
		   (q-remove! c-waitfor-queue elem)))))

(define-public (c-waitfor-check)
  (any (lambda (elem)
	 (and-let* ((val ((car elem))))
	   (q-remove! c-waitfor-queue elem)
	   (c-main-enq! c-waitfor-check)
	   ((cdr elem) val)))
       (q-list c-waitfor-queue)))


;;-----------------------------------------------------------------------------
;; misc stuff

;; convert an AF constant to the corresponding PF
(define-public (af->pf af)
  (cond ((and (defined? 'AF_UNIX)  (= af AF_UNIX))  PF_UNIX)
	((and (defined? 'AF_INET)  (= af AF_INET))  PF_INET)
	((and (defined? 'AF_INET6) (= af AF_INET6)) PF_INET6)
	(else (error "Unrecognised AF address family" af))))

;; call (THUNK) with SIGPIPE set to SIG_IGN
;; this means a write to a pipe or socket where the other end has closed
;; will get an EPIPE error instead of raising SIGPIPE
(define (call-with-epipe thunk)
  (let ((pair #f))
    (dynamic-wind
	(lambda ()
	  (set! pair (sigaction SIGPIPE SIG_IGN)))
	thunk
	(lambda ()
	  (sigaction SIGPIPE (car pair) (cdr pair))))))

;; LST can contain strings, lists of strings, and #f
;; concatenate all those strings and return the result
(define (string-append-flatten . lst)
  (string-concatenate (delq #f (append-map (lambda (elem)
					     (if (pair? elem)
						 elem
						 (list elem)))
					   lst))))

(define-public (string-take-while str pred)
  (string-take str (or (string-skip str pred)
		       (string-length str))))

;; return #t if PREF is a prefix of LST, with elements compared using `equal?'
(define-public (list-prefix? pref lst)
  (or (null? pref)
      (and (not (null? lst))
	   (equal? (car pref) (car lst))
	   (list-prefix? (cdr pref) (cdr lst)))))

(let-toplevel ((econn-errno-list
		(delq! #f (list (and (defined? 'ENETUNREACH)  ENETUNREACH)
				(and (defined? 'ECONNREFUSED) ECONNREFUSED)
				(and (defined? 'ECONNRESET)   ECONNRESET)
				(and (defined? 'EHOSTUNREACH) EHOSTUNREACH)))))
  ;; ARGS is a list of arguments to an error catch handler, return true if
  ;; they're a system-error related to a network connection problem
  (define-public (system-error-econn? args)
    (memv (system-error-errno args) econn-errno-list)))

;; ARGS is a list of arguments to an error catch handler, return true if
;; they're a gethost try-again error
(define-public (gethost-error-try-again? args)
  (eq? (first args) 'try-again))

;; `c-lazy-catch' combines the features of `catch' and `lazy-catch'.
;;
;; Like catch (but unlike lazy-catch), the HANDLER is allowed to return, and
;; its return value is the return from c-lazy-catch.
;;
;; Like lazy-catch (but unlike catch), a re-`throw' within the handler
;; preserves the full stack, allowing a helpful backtrace in a higher catch.
;;
(define-public (c-lazy-catch key thunk handler)
  (catch 'c-lazy-catch
    (lambda ()
      (lazy-catch key thunk
		  (lambda args
		    (throw 'c-lazy-catch (apply handler args)))))
    (lambda (key val)
      val)))

(define-public char-set:crlf (char-set #\cr #\nl))

;; return the first line of STR, excluding any terminating CR/NL
;; either CR or NL or any combination of them is the end of a line
;; if there's no CR or NL then the whole string is considered the first line
(define-public (string-first-line str)
  (let ((pos (string-index str char-set:crlf)))
    (if pos
	(substring str 0 pos)
	str)))

;; call (THUNK), but with (CLEANUP) run after either a normal exit or under
;; an error throw (but not just when switching out with a continuation)
(define-public (call-with-cleanup cleanup thunk)
  (let ((ret (lazy-catch #t
	       thunk
	       (lambda args
		 (cleanup)
		 (apply throw args)))))
    (cleanup)
    ret))


;;-----------------------------------------------------------------------------
;; error stuff

(let-toplevel ((eagain-errno-list
		(delq! #f (list (and (defined? 'EAGAIN)      EAGAIN)
				(and (defined? 'EINPROGRESS) EINPROGRESS)
				(and (defined? 'EWOULDBLOCK) EWOULDBLOCK)))))
  ;; return true if ARGLIST is an EAGAIN style system-error
  (define-public (system-error-eagain? args)
    (memv (system-error-errno args) eagain-errno-list)))

;; return the error message from a 'system-error ARGLIST
;; this will be locale encoded
(define (system-error-message arglist)
  (apply simple-format #f (third arglist) (fourth arglist)))

;; return true if ARGLIST ia an error from `gethost' and friends
(define (gethost-error? arglist)
  (memq (first arglist) '(host-not-found try-again no-recovery no-data)))

;; return error message string from gethost error ARGLIST
;; this will be locale encoded
(define gethost-error-message third)

;; return a utf8 message string for ARGS from a `catch' handler
(define-public (error-message args)
  (cond ((eq? 'http (first args))
	 (second args))
	((gethost-error? args)
	 (locale->utf8 (gethost-error-message args)))
	((eq? 'system-error (first args))
	 (locale->utf8 (system-error-message args)))
	(else
	 (simple-format #f (_ "error: ~a") args))))


;;-----------------------------------------------------------------------------
;; program filtering

;; return a port open on a temporary file with the given EXT extension, like
;; ".exe", or "" for no extension
;;
;; O_EXCL is used to ensure no existing file is overwritten, even if
;; malicious symlinks have been setup, the same as the real mkstemp!
;;
(define-public (mkstemp-ext ext)
  (let more ((attempt 0))
    (if (> attempt 5000)
	(error "Oops, cannot create temporary file"))
    (let ((filename (g-build-filename (g-get-tmp-dir)
				      (simple-format #f "chart-temp-~a~a"
						     (random 1000000) ext))))
      (or (false-if-exception
	   (open filename (logior O_RDWR O_CREAT O_EXCL) #o600))
	  (more (1+ attempt))))))

;; call (PROC port) with a port which is a temporary file
;; the file is removed when PROC returns or throws an error, though not when
;; it leaves through a continuation
;; PROC can close PORT early if it wants
;;
(define-public (call-with-tempfileport port proc)
  (let ((filename (port-filename port))) ;; save this in case closed by PROC
    (call-with-cleanup (lambda ()
			 (or (port-closed? port)
			     (close-port port))
			 (delete-file filename))
		       (lambda ()
			 (proc port)))))


;;-----------------------------------------------------------------------------
;; url stuff
;;
;; Crib notes:
;;
;; Current RFC3986 -- URI spec
;;
;; Past RFC2396 -- previous URI spec
;;      RFC1808 -- relative URI spec
;;      RFC1738 -- previous again
;;      RFC1736, RFC1737 -- functional specs
;;

;; see RFC 1738 start of section 3.1, and section 5 bnf
(define (http-url-match url)
  (or (regexp-exec
       #,(regexp
	  ;;        user+password[2] host[3]   port[5]   path[6]
	  ;;        |                |         |         |
	  "^https?://(([^:@]+:[^@]+)@)?([^:/]+)(:([0-9]+))?(/(.*))?$") url)
      (error "Invalid URL:" url)))

(define-public (url-scheme url)
  (string-take url (or (string-index url #\:)
		       (error "Invalid URL:" url))))

;; return true if URL is a https://
(define (url-https? url)
  (string=? "https" (url-scheme url)))

;; return the host part of a http url string
(define-public (http-url-host url)
  (match:substring (http-url-match url) 3))

;; return the host and port part of a http url string
(define-public (http-url-host:port url)
  (let ((m (http-url-match url)))
    (substring url (match:start m 3) (or (match:start m 6)
					 (string-length url)))))

;; return the port number from a http url string as an integer, defaulting
;; to 80 for http or 443 for https
(define-public (http-url-port url)
  (cond ((match:substring (http-url-match url) 5) => string->number)
	((url-https? url) 443)
	(else 80)))

;; return the path from a http url string, defaulting to "/"
(define-public (http-url-path url)
  (or (match:substring (http-url-match url) 6)
      "/"))

;; return true if HOST is in the given DOMAIN
(define (host-in-domain? host domain)
  (or (string-ci=? host domain)
      (string-suffix-ci? (string-append "." (string-trim domain #\.))
			 host)))

;; return STR with characters unsafe for a url changed to %20 etc forms
;; see RFC 1738 end of section 2.2 for the safe chars
(define-public (url-quote str)
  (regexp-substitute/global #f #,(regexp "[^A-Za-z0-9_.!*'(),-]") str
			    'pre
			    (lambda (m)
			      (format #f "%~@:(~2,'0x~)"
				      (char->integer
				       (string-ref str (match:start m)))))
			    'post))

;; return a url which is LINK relative to a http URL
;;
;; The "Location:" header is supposed to be an absolute url (RFC 2068 sect
;; 14.30), but mediamammon.org was seen giving a relative one, and we want
;; relative for code in tge.scm, so accept both absolute and relative links
;; here.
;;
(define-public (http-url-relative url link)
  (if (or (string-prefix? "http://" link)
	  (string-prefix? "https://" link))
      link

      (let* ((m    (http-url-match url))
	     (pre  (string-take url (or (match:start m 6)
					(string-length url)))))

	(if (string-prefix? "/" link)
	    ;; absolute link
	    (string-append pre link)

	    ;; relative link
	    (let ((path (or (match:substring m 6)
			    "/")))
	      (if (not (string-suffix? "/" path))
		  (set! path (dirname path)))
	      (let more ()
		(cond ((string-prefix? "./" link)
		       (set! link (string-drop link 2))
		       (more))
		      ((string-prefix? "../" link)
		       (set! link (string-drop link 3))
		       (set! path (dirname path))
		       (more))))
	      (string-append pre path
			     (if (string-suffix? "/" path) "" "/")
			     link))))))


;;-----------------------------------------------------------------------------
;; proxy stuff
;;
;; References:
;;
;; /usr/share/doc/libwww-doc/html/Library/User/Using/Proxy.html
;;     http_proxy, no_proxy
;;     Eg. http_proxy=http://proxy.zipworld.com.au:8080/
;;
;; /usr/share/doc/lynx/lynx_help/keystrokes/environments.html
;;     shows spaces in no_proxy
;;
;; wget.info Proxies
;;

;; same as `getenv' but if the variable's value is the empty string return #f
(define (getenv-notempty var)
  (and-let* ((val (getenv var))
	     (    (not (string-null? val))))
    val))

;; return true if URL is matched by the no_proxy environment variable
;; no_proxy is a comma separated list of domain or domain:port
(define (http-no-proxy? url)
  (and-let* ((no-proxy (getenv-notempty "no_proxy"))
	     (url-host (http-url-host url))
	     (url-port (http-url-port url)))
    (any (lambda (entry)
	   ;; in case user leaves whitespace in the variable
	   (set! entry (string-trim-both entry))
	   (and-let* ((entry-url    (string-append "http://" entry))
		      (entry-domain (http-url-host entry-url))
		      (entry-port   (http-url-port entry-url)))
	     (and (= url-port entry-port)
		  (host-in-domain? url-host entry-domain))))
	 (string-split no-proxy #\,))))

;; return proxy url to use for the given URL, or #f to go direct
;; this is simply the http_proxy environment variable, if set and if URL
;; isn't in the host list in the no_proxy environment variable
(define-public (http-proxy url)
  (and (not (url-https? url))
       (not (http-no-proxy? url))
       (getenv-notempty "http_proxy")))


;;-----------------------------------------------------------------------------
;; http message stuff

;; return status code from HEADERS, or 0 if invalid or a locally detect error
(define-public (http-headers-status-code headers)
  (or (and-let* ((sp1 (string-index headers #\ ))
		 (sp2 (string-index headers #\  (1+ sp1))))
	(string->number (substring headers (1+ sp1) sp2)))
      0))

(define-public (http-status-code-ok? status)
  (and (>= status 200)
       (<= status 299)))

;; return true if HEADERS is an "ok" response from the server
(define-public (http-headers-ok? headers)
  (http-status-code-ok? (http-headers-status-code headers)))

;; return the value of field NAME from HEADERS, or #f if not present
;; leading and trailing whitespace is stripped, and the CRLFs between
;; continuation lines
;;
;; regexp matches value part of a header field up to CRLF and continuing on
;; further lines beginning with space or tab
;;
(define-public (http-headers-get headers name)
  (and-let* ((pos   (string-contains-ci headers
					(string-append "\r\n" name ":")))
	     (match (regexp-exec #,(regexp "[ \t]*(([^\r]*\r\n[ \t])*[^\r]*)\r\n")
				 headers (+ pos 3 (string-length name)))))
    (string-trim-right (string-delete (match:substring match 1)
				      (char-set #\cr #\nl)))))

;; return the content length from http HEADERS string, or #f if none
(define (http-headers-content-length headers)
  (and-let* ((val (http-headers-get headers "Content-Length")))
    (or (string->number val)
	(throw 'http (_ "Invalid response")))))

(define (http-headers-chunked? headers)
  (string-ci=? "chunked"
	       (or (http-headers-get headers "Transfer-Encoding") "")))

;; return true if server HEADERS mean the socket connection should be closed
;; HTTP/1.0 connections always close after the response
;; HTTP/1.1 connections remain open unless the server explicitly says close
;;
(define (http-headers-close? headers)
  (or (string-prefix? "HTTP/1.0" headers)
      (= 0 (http-headers-status-code headers)) ;; synthetic error
      (member "close" (http-headers-get-tokens headers "Connection")
	      string-ci=?)))

;; return the tokens from field NAME in HEADERS as a list of strings
;; if NAME is not present the empty list is returned
(define (http-headers-get-tokens headers name)
  (http-headers-split-tokens (or (http-headers-get headers name) "")))

;; STR is the value part of a http header line which is a set of tokens
;; return those tokens as a list of strings
;; tokens are separated by commas, and leading and trailing whitespace is
;; stripped
;; an empty or whitespace-only STR results in the empty list
;;
(define (http-headers-split-tokens str)
  (set! str (string-trim-both str))
  (if (string-null? str)
      '()
      (map! string-trim-both (string-split str #\,))))

;; return an error string constructed from HEADERS
(define-public (http-headers-error-str headers)
  (fold (lambda (prefix str)
	  (if (string-prefix? prefix str)
	      (substring str (string-length prefix))
	      str))
	(string-first-line headers)
	;; drop any of these prefixes from STR
	'("HTTP/1.0 " "HTTP/1.1 " "000 ")))

;; throw a 'http error for HEADERS
(define-public (http-headers-throw headers)
  (throw 'http (http-headers-error-str headers)))


;;-----------------------------------------------------------------------------
;; http plain stuff

(define (http-plain-open url)
  (let* ((proxy-url    (http-proxy url))
	 (dest-url     (or proxy-url url))
	 (dest-host    (http-url-host dest-url))
	 (dest-hostent (c-gethost dest-host))
	 (dest-af      (hostent:addrtype dest-hostent))
	 (dest-addr    (first (hostent:addr-list dest-hostent)))
	 (dest-port    (http-url-port dest-url))
	 (sock         (socket (af->pf dest-af) SOCK_STREAM 0)))

    ((fluid-ref http-status-callback) (if proxy-url
					  (_ "connect (proxy)")
					  (_ "connect")))

    ;; the linux kernel default for keepalive is 9 tries 75
    ;; seconds apart, which is fully 12.5 minutes, making this
    ;; option fairly useless, our own timeouts are shorter
    (setsockopt sock SOL_SOCKET SO_KEEPALIVE 1)

    (fcntl-set-nonblock sock)

    ;; connect normally gives EWOULDBLOCK here (even on
    ;; INADDR_LOOPBACK), but can also give an immediate error
    ;; like ENETUNREACH
    (catch 'system-error
      (lambda ()
	(connect sock dest-af dest-addr dest-port))
      (lambda args
	(if (not (system-error-eagain? args))
	    (begin
	      (close-port sock)
	      (apply throw args)))))

    sock))


;;-----------------------------------------------------------------------------
;; sockpool of keep-alive connected sockets

(define sockpool-timeout-ms  60000) ;; 1 minute in milliseconds

(let-toplevel
    ((rec-type (make-record-type "sockpool"
				 '(sock host port
					inuse input-id timeout-id))))
  (define-toplevel sockpool-new        (record-constructor rec-type))
  (define-toplevel sockpool-sock       (record-accessor rec-type 'sock))
  (define-toplevel sockpool-host       (record-accessor rec-type 'host))
  (define-toplevel sockpool-port       (record-accessor rec-type 'port))
  (define-toplevel sockpool-inuse      (record-accessor rec-type 'inuse))
  (define-toplevel sockpool-input-id   (record-accessor rec-type 'input-id))
  (define-toplevel sockpool-timeout-id (record-accessor rec-type 'timeout-id))

  (define-toplevel sockpool-sock!       (record-modifier rec-type 'sock))
  (define-toplevel sockpool-inuse!      (record-modifier rec-type 'inuse))
  (define-toplevel sockpool-input-id!   (record-modifier rec-type 'input-id))
  (define-toplevel sockpool-timeout-id! (record-modifier rec-type 'timeout-id)))

(define sockpool-list '())


(define (sockpool-not-inuse rec)
  (not (sockpool-inuse rec)))

;; close and remove ELEM from `sockpool-list' either due to us not needing
;; it any more or because the remote server has closed the socket
(define (sockpool-kill rec)
  (set! sockpool-list (delq rec sockpool-list))
  (sockpool-cancel-timers rec)
  (and-let* ((sock (sockpool-sock rec)))
    (http-close sock)))

(define (sockpool-cancel-timers rec)
  (and-let* ((input-id (sockpool-input-id rec)))
    (c-gtk-input-remove input-id)
    (sockpool-input-id! rec #f))
  (and-let* ((timeout-id (sockpool-timeout-id rec)))
    (c-gtk-timeout-remove timeout-id)
    (sockpool-timeout-id! rec #f)))

(define (sockpool-resource? host port)
  (let ((lst (filter sockpool-inuse sockpool-list)))
    (and (< (length lst) http-max-connections)
	 (< (count (lambda (rec)
		     (string=? host (sockpool-host rec)))
		   lst)
	    http-max-to-one-host))))

(define (sockpool-allocate url)
  (let ((host (http-url-host url))
	(port (http-url-port url)))
    (and (sockpool-resource? host port)
	 (let more ((lst sockpool-list))
	   (if (null? lst)
	       (let ((rec (sockpool-new #f host port #t #f #f)))
		 (set! sockpool-list (cons rec sockpool-list))
		 (sockpool-prune)
		 rec)
	       (let ((rec (first lst)))
		 (if (and (sockpool-not-inuse rec)
			  (string=? host (sockpool-host rec))
			  (=        port (sockpool-port rec)))
		     (begin
		       (sockpool-inuse! rec #t)
		       (sockpool-cancel-timers rec)
		       rec)
		     (more (cdr lst)))))))))

(define (sockpool-prune)
  (and-let* ((    (length>=? sockpool-list http-max-connections))
	     (rec (find-last sockpool-not-inuse (reverse sockpool-list))))
    (sockpool-kill rec)
    (sockpool-prune)))

;; return the last element of LST which satisfies PRED
(define (find-last pred lst)
  (find pred (reverse lst)))

(define (sockpool-release rec)
  (sockpool-inuse! rec #f)
  (if (or (not (sockpool-sock rec))
	  (port-closed? (sockpool-sock rec)))
      (sockpool-kill rec)

      (let ((handler (lambda args
		       ;; check not in use before killing, in case gets
		       ;; allocated in between the callback going off and
		       ;; this code running
		       (if (not (sockpool-inuse rec))
			   (sockpool-kill rec)))))
	(sockpool-input-id! rec (c-gtk-input-add (sockpool-sock rec) '(read)
				  handler))
	(sockpool-timeout-id! rec (c-gtk-timeout-add sockpool-timeout-ms
				    handler))
	(c-gtk-id-stoppable (sockpool-input-id rec))
	(c-gtk-id-stoppable (sockpool-timeout-id rec)))))

;; Remove any dead connections in `sockpool-list', ie. connections the
;; remote server has closed.
;; Closed is detected by attempting a read.  There shouldn't be anything
;; coming in so the expected return from read-string!/partial is 0 for
;; connected (but nothing to read) or #f for EOF.
;;
(define (sockpool-purge)
  (let ((buf (make-string 128)))
    (for-each (lambda (rec)
		(if (and (not (sockpool-inuse rec))
			 (not (read-string!/partial buf (sockpool-sock rec))))
		    (sockpool-kill rec)))
	      sockpool-list)))

;; like `c-main-enq!', but only put one copy of PROC into the queue
(define-public (c-main-enq-once! thunk)
  (or (memq thunk (q-list c-main-queue))
      (c-main-enq! thunk)))


(define (sockpool-call url proc)
  (sockpool-purge)
  (receive (headers body)
      (let* ((proxy-url (http-proxy url))
	     (sock-url  (or proxy-url url))
	     (rec       (c-waitfor (lambda ()
				     (sockpool-allocate sock-url))
				   (lambda ()
				     ((fluid-ref http-status-callback)
				      (_ "wait for socket"))))))
	(call-with-cleanup (lambda ()
			     ;; only enq one copy, in particular don't keep
			     ;; growing it when single threaded
			     (c-main-enq-once! c-waitfor-check)
			     (sockpool-release rec))
	  (lambda ()
	    (let* ((re-sock (sockpool-sock rec)) ;; port or #f re-used
		   (sock    re-sock))
	      (if sock
		  ;; re-use existing connection
		  ((fluid-ref http-status-callback)
		   (if proxy-url (_ "request (proxy)") (_ "request")))
		  ;; new connection
		  (begin
		    (set! sock (http-open url))
		    (sockpool-sock! rec sock)))

	      (receive (headers body)
		  (if re-sock
		      ;; if re-using a connection then retry if a
		      ;; connection problem is found
		      (c-lazy-catch #t
			(lambda ()
			  (proc sock))
			(lambda args
			  (if (error-connection-related? args)
			      (begin
				(http-close sock)
				(values 'repeat #f))
			      (apply throw args))))
		      ;; but a fresh connection gets only one attempt
		      (proc sock))

		(if (and (string? headers) ;; 'repeat stays open, by default
			 (http-headers-close? headers))
		    (http-close sock))
		(values headers body))))))

    (if (eq? headers 'repeat)
	(sockpool-call (or body url) proc)
	(values headers body))))

(define (http-open url)
  (if (url-https? url)
      (https-open url)
      (http-plain-open url)))

(define (http-close port)
  (if (not (port-closed? port))
      (close-port port))
  (and-let* ((pid (object-property port 'https-pid)))
    (kill pid SIGTERM)))

;; ARGS is a list of arguments to an error catch handler, return true if
;; they're network connection related.
;; Connection problems are reported as a message, other errors are taken to
;; be programming problems and go to a backtrace etc.
;;
(define-public (error-connection-related? args)
  (or (eq? 'http (first args))
      (gethost-error-try-again? args)
      (system-error-econn? args)))


;;-----------------------------------------------------------------------------
;; GET stuff

;; return values (HEADERS BODY)
;; (throw 'http desc-str) on error, eg. host look or connect fail
;;
;; Since If-None-Match is only a http 1.1 feature, it's worth using
;; If-Modified-Since on top of that, in case a request is passing through a
;; http 1.0 proxy on the way to the server.
;;
(define*-public (http-request
		 url
		 #:key
		 (method "GET")
		 (protocol-level "1.1")
		 if-modified-since ;; If-Modified-Since header with given str
		 if-none-match     ;; If-None-Match header with given str
		 want-compressed   ;; #t not to automatically uncompress
		 (user-agent 'default)
		 data
		 want-ok           ;; #t to throw 'http if not 200 Ok
		 (follow #t)       ;; #t to follow 301,302, #f not
		 (follow-warn #t)  ;; #t to print a warning about following
		 data-callback     ;; called (DATA-CALLBACK str) with pieces
		 cookiejar         ;; filename
		 headers)

  (sockpool-call url
    (lambda (sock)
      (if (eq? user-agent 'default)
	  (set! user-agent
		(string-join-notnull
		 (list (string-append "Chart/" chart-version)
		       (object-property sock 'https-product-token)))))
      (lazy-catch #t
	(lambda ()

	  ;; (or (c-input-timeout-wait sock '(write) http-timeout)
	  ;;     (throw 'http (_ "Timeout")))
	  ;; ((fluid-ref http-status-callback) (_ "request"))

	  (http-write-string
	   sock
	   (string-append-flatten
	    method " " (if (http-proxy url)
			   url
			   (http-url-path url))
	    " " "HTTP/" protocol-level "\r\n"
	    "Host: " (http-url-host:port url) "\r\n"
	    (and if-modified-since
		 (list "If-Modified-Since: " if-modified-since "\r\n"))
	    (and if-none-match
		 (list "If-None-Match: " if-none-match "\r\n"))
	    (accept-encoding-header) "\r\n"
	    (and (string<> "1.0" protocol-level)
		 "TE: trailers\r\n")
	    (and user-agent
		 (list "User-Agent: " user-agent "\r\n"))
	    (and cookiejar
		 (cookiejar-get-line cookiejar url))
	    headers
	    (and data
		 (list
		  "Content-Type: application/x-www-form-urlencoded\r\n"
		  (simple-format #f  "Content-Length: ~a\r\n"
				 (string-length data))))
	    "\r\n"
	    data))

	  (if (string=? method "HEAD")
	      (let ((headers (http-read-headers sock)))
		(values headers ""))

	      (receive (headers body)
		  (http-read-response sock data-callback)

		(if cookiejar
		    (cookiejar-add-headers cookiejar url headers))

		(let ((status-code (http-headers-status-code headers)))
		  (cond ((and follow (memv status-code '(301 302)))
			 ;; follow 301 Moved Permanently
			 ;; and    302 Moved Temporarily
			 ;; seen eg. from au.finance.yahoo.com

			 ;; if a POST results in a 302 then retrieve it with
			 ;; just a GET
			 (set! method "GET")
			 (set! data #f)
			 ;; new location can't use these
			 (set! if-modified-since #f)
			 (set! if-none-match #f)

			 ;; don't follow more than 5 times, in case
			 ;; some stupid server tries to send us into
			 ;; an infinite loop
			 (or (number? follow)
			     (set! follow 1))
			 (or (<= follow 5)
			     (throw 'http (_ "Too many 301/302 follows")))
			 (set! follow (1+ follow))

			 (let* ((new (http-url-relative
				      url (http-headers-get headers
							    "Location"))))
			   ;; print a message since normally want the right
			   ;; url in the code in the first place, to save a
			   ;; round trip
			   (if follow-warn
			       (format (current-error-port)
				       "chart: following ~s\n~a -> ~s\n"
				       (string-first-line headers) url new))
			   (values 'repeat new)))

			((and want-ok
			      (not (or (http-status-code-ok? status-code)
				       (and (not follow)
					    (memv status-code '(301 302)))
				       (and (or if-modified-since
						if-none-match)
					    (= 304 status-code)))))
			 ;; WANT-OK means demand a good status-code;
			 ;; for if-modified-since and if-none-match a "304
			 ;; Not Modified" is considered good
			 (http-headers-throw headers))

			(else
			 (if (not want-compressed)
			     (set! body (undo-content-encoding headers body)))
			 (values headers body)))))))

	(lambda args
	  (http-close sock)
	  (and-let* ((proc (object-property sock 'http-error-transform)))
	    (set! args (proc sock args)))
	  (apply throw args))))))

;; return an Accept-Encoding header string,
;;   with nothing:  "Accept-Encoding:"
;;   with gzip:     "Accept-Encoding: gzip, compress"
;;   with zlib:     "Accept-Encoding: gzip, deflate"
;;   with both:     "Accept-Encoding: gzip, compress, deflate"
;; only gzip does compress and only zlib does deflate, but both do gzip
;;
(define (accept-encoding-header)
  (string-join-notnull (list "Accept-Encoding:"
			     (string-join-notnull
			      (list (and (or (gzip-version)
					     (zlib-version)) "gzip")
				    (and (gzip-version)      "compress")
				    (and (zlib-version)      "deflate"))
			      ", "))
		       " "))

;; return decoded body
(define (undo-content-encoding headers body)
  (let ((encoding (http-headers-get headers "Content-Encoding")))
    (cond ((or (not encoding)
	       (string-null? encoding))
	   body)

	  ((member encoding '("compress" "x-compress") string-ci=?)
	   ;; zlib can't handle .Z compress format, must spawn gzip
	   (gunzip-string body))

	  ((member encoding '("gzip" "x-gzip") string-ci=?)
	   (if (zlib-version)
	       (zlib-inflate body)
	       (gunzip-string body)))

	  ((string-ci=? encoding "deflate")
	   (zlib-inflate body))

	  (else
	   (throw 'http
		  (simple-format #f (_ "Unrecognised Content-Encoding: ~s")
				 encoding))))))


;; return values (HEADERS BODY)
(define (http-read-response sock data-callback)
  (let ((headers (http-read-headers sock)))
    (cond ((eqv? 100 (http-headers-status-code headers))
	   ;; 100 Continue, real response follows.
	   ;; RFC 2616 allows the server to send a Continue for a POST even
	   ;; if you haven't asked "Expect: 100-continue".  An MS Orifice
	   ;; has been seen doing that.
	   (http-read-response sock data-callback))

	  ((http-headers-chunked? headers)
	   (let* ((body         (http-read-chunks sock data-callback))
		  (more-headers (http-read-headers sock #t)))
	     (values (string-append (string-drop-right headers 2) more-headers)
		     body)))

	  (else
	   ;; not chunked, so either a given "Content-Length" or just to EOF
	   (values headers
		   (let ((len (http-headers-content-length headers)))
		     (if len
			 (http-read-bytes sock len data-callback
					  (fluid-ref http-status-callback))
			 (let ((body (http-read-to-eof sock data-callback)))
			   (http-close sock)
			   body))))))))

;; Return headers string read from SOCK.
;; If EXTRAS is true then it's for extra headers at the end of a chunked
;; body and an empty headers (just CRLF) is allowed.
;;
(define* (http-read-headers sock #:optional extras)
  (define done-status #f)
  (http-read-to-match sock
		      (lambda (str)
			(if (not done-status)
			    (begin
			      ((fluid-ref http-status-callback) (_ "response"))
			      (set! done-status #t)))
			(or (and extras
				 (string-prefix? "\r\n" str)
				 2)
			    (string-contains-after str "\r\n\r\n")))))

;; Read body chunks from SOCK.  Stop reading after the 0 length terminator
;; chunk and return all the data as a string.
;;
;; There's no Content-Length in a chunked response (per RFC 2068 sect 4.4)
;; so the status report from http-read-bytes can only say N "or more" bytes
;; being read.
;;
(define (http-read-chunks sock data-callback)
  (let more ((lst '())
	     (total 0))
    (let* ((line (http-read-to-crlf sock))
	   (n    (string->number (string-take-while
				  line char-set:hex-digit) 16)))
      (cond ((not n)
	     (throw 'http (_ "Invalid response (chunk size)")))
	    ((= 0 n)
	     (string-concatenate-reverse/shared lst))
	    (else
	     (let ((part (http-read-bytes sock n
					  data-callback
					  (fluid-ref http-status-callback)
					  total)))
	       ;; skip CRLF after part data
	       (or (equal? "\r\n" (http-read-bytes sock 2))
		   (throw 'http
			  (_ "Invalid response (CRLF after chunk data)")))

	       (more (cons part lst) (+ total n))))))))


;;-----------------------------------------------------------------------------
;; i/o primitives

;; Return string read up to and including CRLF.
;;
(define-public (http-read-to-crlf sock)
  (http-read-to-match sock
		      (lambda (str)
			(string-contains-after str "\r\n"))))

;; Read a string from SOCK up to the point PROC says matches.
;; PROC is called (PROC str) and should return the number of chars which
;; match, or #f if not complete yet.  The balance past where PROC says is
;; putback with `unread-string'.
;;
(define (http-read-to-match port proc)
  (define buf (make-string 512))
  (define str "")
  (let more ()
    (let ((nread (read-string!/partial buf port)))
      (cond ((not nread)
	     ;; eof, incomplete headers
	     (throw 'http (_ "Incomplete response")))

	    ((= 0 nread)
	     ;; nothing ready yet
	     (or (c-input-timeout-wait port '(read) http-timeout)
		 (throw 'http (_ "Timeout")))
	     (more))

	    (else
	     ;; some data
	     (set! str (string-append str (substring/shared buf 0 nread)))
	     (let ((pos (proc str)))
	       (if pos
		   (begin
		     (unread-string (string-drop str pos) port)
		     (string-take str pos))
		   (more))))))))

;; return bytes read to eof on PORT
(define*-public (http-read-to-eof port #:optional data-callback)
  (let more ((lst   '())
	     (total 0))
    (let* ((str   (make-string 4096))
	   (nread (read-string!/partial str port)))
      (cond ((not nread)
	     ;; eof
	     (string-concatenate-reverse/shared lst))

	    ((= 0 nread)
	     ;; would block
	     (or (c-input-timeout-wait port '(read) http-timeout)
		 (throw 'http (_ "Timeout")))
	     (more lst total))

	    (else
	     ;; got some, allow for less than requested
	     (set! str (substring/shared str 0 nread)) ;; no copy if full STR
	     (if data-callback
		 (data-callback str))
	     ((fluid-ref http-status-callback) (simple-format #f (_ "~a bytes")
							      (+ total nread)))
	     (more (cons str lst) (+ total nread)))))))

;; Return string of N bytes read from PORT.
;;
;; DATA-CALLBACK, if not #f, is a procedure called (DATA-CALLBACK str) for
;; blocks of bytes as they arrive.
;;
;; STATUS-CALLBACK, if not #f, is a procedure called (STATUS-CALLBACK str)
;; with a progress description.
;;
;; START, if not #f, is a starting count for the STATUS-CALLBACK, for when
;; this http-read-bytes is one of multiple chunks read for a single body.
;; The "or more" in the message reflects this chunked reading use.
;;
(define*-public (http-read-bytes port n
				 #:optional
				 data-callback status-callback start)
  (let ((str (make-string n)))
    (let more ((pos 0))
      (if (>= pos n)
	  str
	  (let ((nread (read-string!/partial str port pos)))
	    (cond ((not nread)
		   (throw 'http (_ "Incomplete response")))

		  ((= 0 nread)
		   ;; nothing ready
		   (or (c-input-timeout-wait port '(read) http-timeout)
		       (throw 'http (_ "Timeout")))
		   (more pos))

		  (else
		   ;; some data
		   (if data-callback
		       (data-callback (substring/shared str pos
							(+ pos nread))))
		   (if status-callback
		       (status-callback
			(if start
			    (simple-format #f (_ "~a of ~a or more bytes")
					   (+ pos nread start)
					   (+ n start))
			    (simple-format #f (_ "~a of ~a bytes")
					   (+ pos nread) n))))

		   (more (+ pos nread)))))))))

;; write STR to PORT
;; ECONNREFUSED is a possible error after `connect' has initiated
(define-public (http-write-string port str)
  (let more ((upto 0))
    (if (< upto (string-length str))
	(begin
	  (or (c-input-timeout-wait port '(write) http-timeout)
	      (throw 'http (_ "Timeout")))
	  (let ((n (call-with-epipe (lambda ()
				      (write-string/partial str port upto)))))
	    ;; n!=0 is some written or n==0 is an unexpected EWOULDBLOCK,
	    ;; either way continue
	    (more (+ upto n)))))))
