;; Secure HTTP functions, supporting http.scm.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart https)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart config)
  #:use-module (chart extra-glib)
  #:use-module (chart gethost)
  #:use-module (chart gzip)
  #:use-module (chart http)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; gnutls

;; `chart-gnutls-program' is the full pathname to the "chart-gnutls" program.
;; This program doesn't exist if no gnutls in the build, and it won't run if
;; the gnutls library isn't available (the library is only a recommends in
;; the debian package).
;;
(define chart-gnutls-program
  (g-build-filename chart-config-libchartdir "chart-gnutls"))

;; return the version number string for GnuTLS, or #f if not available
(define-public gnutls-version
  (program-version-proc (string-append chart-gnutls-program " --gnutlsversion")
			string-trim-both))


;;-----------------------------------------------------------------------------
;; openssl

;; return the version number string for the openssl program, or #f if not
;; available
(define-public openssl-version
  (program-version-proc "openssl version"
			(lambda (str)
			  (set! str (string-first-line str))
			  (set! str (string-trim-str str "OpenSSL "))
			  (string-take str (or (string-index str #\space)
					       (string-length str))))))


;;-----------------------------------------------------------------------------
;; https

(define-public (https-available?)
  (or (gnutls-version)
      (openssl-version)))

;; hold socket ports which are open to child processes
(define https-guardian (make-guardian))

(define (https-reaper)
  (do ((sock (https-guardian) (https-guardian)))
      ((eq? #f sock))

    (if (not (port-closed? sock))
	(close-port sock))

    (and-let* ((errport (object-property sock 'https-errport)))
      (set-object-property! sock 'https-errport #f)
      (close-port errport))

    (and-let* ((pid (object-property sock 'https-pid)))
      (let ((pair (waitpid pid WNOHANG)))
	(if (zero? (car pair))
	    (begin
	      (kill pid SIGTERM)
	      (https-guardian sock))
	    (set-object-property! sock 'https-pid #f))))))

;; (define (https-close sock)
;;   (if (not (port-closed? sock))
;;       (close-port sock))
;;   (and-let* ((pid (object-property sock 'https-pid)))
;;     (kill pid SIGTERM)))
  

(define-public (https-open url)
  (define url-host (http-url-host url))
  (define url-port (http-url-port url))

  (add-hook! after-gc-hook https-reaper)

  ;; do hostname lookup here so as to throw a gethost error when offline
  ;; such an error is caught by `error-connection-related?', where a 'http
  ;; error from the child process is not
  (c-gethost url-host)

  (let* ((pair          (socketpair PF_UNIX SOCK_STREAM 0))
	 (parent-sock   (car pair))
	 (child-sock    (cdr pair))
	 (args
	  (cond ((gnutls-version)
		 => (lambda (version)
		      ((fluid-ref http-status-callback) (_ "gnutls connect"))
		      (set-object-property! parent-sock 'https-product-token
					    (string-append "GnuTLS/" version))
		      (list chart-gnutls-program
			    url-host (number->string url-port))))
		((openssl-version)
		 => (lambda (version)
		      ((fluid-ref http-status-callback) (_ "openssl connect"))
		      (set-object-property! parent-sock 'https-product-token
					    (string-append "OpenSSL/" version))
		      (list "openssl" "s_client" "-quiet"
			    "-connect" (format #f "~a:~a" url-host url-port))))
		(else
		 (error (_ "https connection requires GnuTLS or OpenSSL")))))

	 (errport       (mkstemp-ext ""))
	 (pid           (primitive-fork)))

    (if (eqv? 0 pid)  ;; child
	(catch #t
	  (lambda ()
	    (dup2 (fileno child-sock) 0)
	    (dup2 (fileno child-sock) 1)
	    (dup2 (fileno errport)    2)

	    (port-for-each
	     (lambda (port)
	       (false-if-exception ;; for non-fd ports and/or close errors
		(let ((fd (fileno port)))
		  (or (<= fd 2)
		      (close-fdes fd))))))

	    (apply execlp (first args) args))

	  (lambda errargs
	    (primitive-_exit 127))))

    ;; parent
    (close-port child-sock)

    ;; file is deleted but kept open
    (delete-file (port-filename errport))

    (set-object-property! parent-sock 'https-errport errport)
    (set-object-property! parent-sock 'https-pid pid)
    (set-object-property! parent-sock
			  'http-error-transform https-error-transform)
    (https-guardian parent-sock)

    (fcntl-set-nonblock parent-sock)
    parent-sock))


(define (https-error-transform sock args)
  (let ((errno (system-error-errno args)))
    (if (or (and (defined? 'ECONNRESET)
		 (eqv? errno ECONNRESET))
	    (and (defined? 'EPIPE)
		 (eqv? errno EPIPE)))

	(let ((errport (object-property sock 'https-errport)))
	  (if (not (port-closed? sock))
	      (close-port sock))

	  ;; wait child to ensure the error message is in the file
	  (and-let* ((pid (object-property sock 'https-pid)))
	    (set-object-property! sock 'https-pid #f)
	    (kill pid SIGTERM)
	    (waitpid pid))

	  ;; errport is a dup of the child's fd, so must seek back to read
	  (seek errport 0 SEEK_SET)
	  (list 'http (string-trim-right
		       (read-string-to-eof errport))))

	;; not a https child error
	args)))
