;; Hull moving average.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart hull-ma)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart wma))


;; http://www.alanhull.com.au/hma/hma.html
;;     Alan Hull's explanation and formula, samples of BHP Billiton (symbol
;;     BHP.AX) weekly 2001-2004.
;;


;; Return a procedure which calculates a Hull moving average over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the HMA up to (and including) that value, or #f if not enough values
;; yet.
;;
;;
(define-public (hull-ma-calc-proc count)

  (define wma1-proc  (wma-calc-proc (quotient count 2)))
  (define wma2-proc  (wma-calc-proc count))
  (define wsqrt-proc (wma-calc-proc (floor-exact (sqrt count))))

  (lambda (x)
    (let ((w1 (wma1-proc x))
	  (w2 (wma2-proc x)))
      (and w2
	   (wsqrt-proc (- (* 2 w1) w2))))))

(define-public (hull-ma-warmup-count count)
  (+ (wma-warmup-count count)
     1
     (wma-warmup-count (floor-exact (sqrt count)))))


(define (hull-ma-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (hull-ma-calc-proc count)))

    ;; warmup
    (for-each proc
	      (series-prev-n-closes parent #:before lo
				    #:count (hull-ma-warmup-count count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (hull-ma-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Hull MA") count))
    (series-datatype!  new-series hull-ma-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   hull-ma-fill (hull-ma-warmup-count count)))
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public hull-ma-datatype
  (datatype-new!
   #:key    'hull-ma
   #:name   (_ "Hull MA")
   #:units  #f
   #:manual (_ "CONTEXT(manual-node): Hull Moving Average")
   #:proc   hull-ma-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'hull-ma-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
