;; Hurst exponent, by detrended fluctuation analysis (DFA) method.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart hurst-dfa)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (srfi srfi-1)
  #:use-module (chart hurst)
  #:use-module (chart indicator)
  #:use-module (chart misc))



(define (hurst-dfa lst len)
  (if (< len 2)
      #f
      (let ((xy-lst '()))
	(do ((segment-count 1 (ash segment-count 1)))
	    ((and (length>=? xy-lst 2)
		  (< (/ len segment-count) 8)))
	  (let* ((segment-len (quotient len segment-count))
		 (dfa-avg (hurst-dfa-pass lst len segment-count segment-len)))
	    (set! xy-lst (acons (log segment-len)
				(log dfa-avg)
				xy-lst))))
	(linreg-slope xy-lst))))

;; Return a procedure which calculates a Hurst exponent using the rescaled
;; range method, over a window of COUNT many steps.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the Hurst exponent up to and including that point, or #f if not enough
;; points yet.
;;
;; The window can be primed by calling with COUNT-1 many points preceding
;; the first desired.
;;
(define*-public (hurst-dfa-calc-proc count #:optional (data '()))
  (set! data (take-right-maybe data (1- count)))
  (let* ((q          (list->q! (map log data)))
	 (points     (length (q-list q)))
	 (min-points (ceiling-exact (* indicator-minimum-fraction count))))
    (set! data #f) ;; gc

    (lambda (x)
      (set! x (log (abs x)))
      (if (< points count)
	  (set! points (1+ points))
	  (deq! q))
      (enq! q x)

      (and (>= points min-points)
	   (hurst-dfa (q-list q) points)))))


(define (hurst-dfa-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (hurst-dfa-calc-proc count
				      (series-prev-n-closes parent
							    #:before lo
							    #:count  count))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (hurst-dfa-series series count)
  (set! count (max count 16))
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Hurst DFA") count))
    (series-datatype!  new-series hurst-dfa-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts hurst-dfa-fill count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public hurst-dfa-datatype
  (datatype-new!
   #:key       'hurst-dfa
   #:name      (_ "Hurst Exponent - DFA")
   #:priority  -10
   #:units     'zero-to-one
   #:minimum   0
   #:maximum   1
   #:hlines    '(0.5)
   #:manual    (_ "CONTEXT(manual-node): Hurst Exponent")
   #:proc      hurst-dfa-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'hurst-dfa-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 128))))
