/* Chart Hurst exponent support.

   Copyright 2007, 2008 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define _GNU_SOURCE   /* for tgamma */

#include "config.h"
#include <math.h>
#include <libguile.h>

#include "compat.h"

#define DEBUG 0

#define MAX(x,y) ((x) > (y) ? (x) : (y))


/*---------------------------------------------------------------------------*/
/* Rescaled range (R/S) */

SCM_DEFINE_PUBLIC (chart_hurst_rs_pass, "hurst-r/s-pass", 4, 0, 0,
            (SCM lst, SCM s_total_len, SCM s_segment_count, SCM s_segment_len),
	    "Make a pass over the values in list @var{lst} calculating an\n"
	    "average Hurst @math{R/S} rescaled range for\n"
	    "@var{s_segment_count} many segments, each @var{s_segment_len}\n"
	    "long, within a total of @var{s_total_len} many steps.\n"
	    "\n"
	    "@var{s_total_len} represents ``steps'', meaning differences,\n"
	    "between successive data values, so @math{@var{s_total_len}+1}\n"
	    "many values from @var{lst} are used.  Likewise\n"
	    "@var{s_segment_len} is a count of steps between values.\n"
	    "\n"
	    "If the segment count and length don't completely cover the\n"
	    "total length, gaps in between segments are skipped, to spread\n"
	    "the segments out evenly across the total.")
#define FUNC_NAME s_chart_hurst_rs_pass
{
  unsigned total_len = scm_to_uint (s_total_len);
  unsigned segment_len = scm_to_uint (s_segment_len);
  unsigned segment_count = scm_to_uint (s_segment_count);
  unsigned gap_total = total_len - segment_len * segment_count;
  unsigned gap_done = 0;
  double rs_sum = 0.0;
  int rs_count = 0;
  unsigned segnum;

  for (segnum = 0; segnum < segment_count; segnum++)
    {
      double first = scm_to_double (scm_car (lst));
      double last = scm_to_double (scm_list_ref (lst, s_segment_len));
      double mean_diff = (last - first) / segment_len;
      double sum_squares = 0.0;
      double prev = first;
      double line = first;
      double hi = 0.0;
      double lo = 0.0;
      unsigned i;

      {
        unsigned gap_target
          = (segnum * gap_total + (segment_count >> 1)) / segment_count;
        while (gap_done < gap_target)
          {
            lst = scm_cdr (lst);
            gap_done++;
          }
      }

      for (i = 0; i < segment_len; i++)
        {
          double c, diff_to_line, diff_to_prev_detrended;

          lst = scm_cdr (lst);
          c = scm_to_double (scm_car (lst));
#if DEBUG
          printf ("%d %lg\n", i, c);
#endif
          /* range of deviations from endpoint-to-endpoint line */
          line += mean_diff;
          diff_to_line = c - line;
          if (diff_to_line > hi)
            hi = diff_to_line;
          if (diff_to_line < lo)
            lo = diff_to_line;

          /* this step difference, less the mean difference, for stddev */
          diff_to_prev_detrended = c - prev - mean_diff;
          sum_squares += diff_to_prev_detrended * diff_to_prev_detrended;

          prev = c;
        }

      if (sum_squares != 0)
        {
          double stddev = sqrt (sum_squares / segment_len);
          double range = hi - lo;
          double rs = range / stddev;
#if DEBUG
          printf ("rs %lg\n", rs);
#endif
          rs_sum += rs;
          rs_count++;
        }
    }

  if (rs_count == 0)
    return scm_from_double (0.0);
  else
    return scm_from_double (rs_sum / rs_count);
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* Detrended fluctuation analysis (DFA) */

SCM_DEFINE_PUBLIC (chart_hurst_dfa_pass, "hurst-dfa-pass", 4, 0, 0,
            (SCM lst, SCM s_total_len, SCM s_segment_count, SCM s_segment_len),
	    "Make a pass over the values in list @var{lst} calculating an\n"
	    "average DFA method F(n) standard error for\n"
	    "@var{s_segment_count} many segments, each @var{s_segment_len}\n"
	    "long, within a total of @var{s_total_len} many values.\n"
	    "\n"
	    "@var{s_total_len} represents data values.  @var{lst} can be\n"
	    "longer than @var{s_total_len}, any extra is ignored.\n"
	    "\n"
	    "If the segment count and length don't completely cover the\n"
	    "total length, gaps in between segments are skipped, to spread\n"
	    "the segments out evenly across the total.")
#define FUNC_NAME s_chart_hurst_dfa_pass
{
  unsigned total_len = scm_to_uint (s_total_len);
  unsigned segment_len = scm_to_uint (s_segment_len);
  unsigned segment_count = scm_to_uint (s_segment_count);
  unsigned gap_total = total_len - segment_len * segment_count;
  unsigned gap_done = 0;
  double stderr_sum = 0.0;
  int stderr_count = 0;
  unsigned segnum;
  double x_total = 0.5 * segment_len * (segment_len + 1);
  double segment_len_squared = segment_len * segment_len;
  double x_variance_fac = (segment_len < 2 ? 1.0
                           : 12.0 / (segment_len_squared
                                     * (segment_len_squared - 1.0)));
  double stderr_fac = 1.0 / segment_len;

  for (segnum = 0; segnum < segment_count; segnum++)
    {
      double y_total = 0.0;
      double yy_total = 0.0;
      double xy_total = 0.0;
      double variance_y, covariance_xy, stderr;
      unsigned i;

      {
        unsigned gap_target
          = (segnum * gap_total + (segment_count >> 1)) / segment_count;
        while (gap_done < gap_target)
          {
            lst = scm_cdr (lst);
            gap_done++;
          }
      }

      for (i = 1; i <= segment_len; i++)
        {
          double c = scm_to_double (scm_car (lst));
          lst = scm_cdr (lst);

          y_total += c;
          yy_total += c * c;

          xy_total += c * i;
        }

      variance_y = segment_len * yy_total - y_total * y_total;
      covariance_xy = segment_len * xy_total - x_total * y_total;

      stderr = stderr_fac
        * sqrt (MAX (0.0,
                     variance_y - x_variance_fac * covariance_xy * covariance_xy));

      stderr_sum += stderr;
      stderr_count++;
    }

  return scm_from_double (stderr_sum / stderr_count);
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* Misc extras */

#if HAVE_LGAMMA && ! HAVE_LGAMMA_R   /* replacement lgamma_r */
static inline double
lgamma_r (double x, int *signp)
{
  double lgam = lgamma (x);
  *signp = signgam;
  return lgam;
}
#undef HAVE_LGAMMA_R
#define HAVE_LGAMMA_R 1
#endif

#if HAVE_LGAMMA_R && ! HAVE_TGAMMA   /* replacement tgamma */
static inline double
tgamma (double x)
{
  int sign;
  double lgam = lgamma_r (x, &sign);
  double tgam = exp (lgam);
  return (tgam < 0 ? -tgam : tgam);
}
#undef HAVE_TGAMMA
#define HAVE_TGAMMA 1
#endif


#if HAVE_LGAMMA_R
SCM_DEFINE_PUBLIC (chart_log_gamma, "log-gamma", 1, 0, 0,
            (SCM x),
	    "Calculate the natural logarithm of the Gamma function.  The\n"
	    "return is two values, the logarithm and the sign that goes with\n"
	    "the powered value: @code{(values sign lgam)}.  @var{sign} is 1\n"
	    "or @math{-1}, so that @math{Gamma(@var{x}) = @var{sign} *\n"
	    "exp(@var{lgam})}.\n"
	    "\n"
	    "Because the Gamma function grows so rapidly, it's often\n"
	    "necessary to work with logarithms.")
#define FUNC_NAME s_chart_log_gamma
{
  int sign;
  double lgam = lgamma_r (scm_to_double (x), &sign);
  return scm_values (scm_list_2 (scm_from_int (sign),
                                 scm_from_double (lgam)));
}
#undef FUNC_NAME
#endif

#if HAVE_LGAMMA_R
SCM_DEFINE_PUBLIC (chart_log_abs_gamma, "log-abs-gamma", 1, 0, 0,
            (SCM x),
	    "Return the natural logarithm of the absolute value of the Gamma\n"
	    "function, ie.@: @math{log(|Gamma(@var{x})|)}.\n"
	    "\n"
	    "Because the Gamma function grows so rapidly, it's often\n"
	    "necessary to work with logarithms.")
#define FUNC_NAME s_chart_log_abs_gamma
{
  int sign;
  double lgam = lgamma_r (scm_to_double (x), &sign);
  return scm_from_double (lgam);
}
#undef FUNC_NAME
#endif

#if HAVE_TGAMMA
SCM_DEFINE_PUBLIC (chart_gamma, "gamma", 1, 0, 0,
            (SCM x),
	    "Return the gamma function @math{Gamma(@var{x})}.")
#define FUNC_NAME s_chart_gamma
{
  return scm_from_double (tgamma (scm_to_double (x)));
}
#undef FUNC_NAME
#endif


/*---------------------------------------------------------------------------*/

void
chart_hurst_init (void)
{
#ifndef SCM_MAGIC_SNARFER
#include "hurst.x"
#endif
}
