;; Ichimoku Kinko Hyo indicator.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ichimoku)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart williams-r))


;; http://www.fmlabs.com/reference/IchimokuKinkoHyo.htm
;;     Formulas, some description.


(define-public (range-midpoint-calc-proc count)
  (define highlow-proc (highlow-calc-proc count))

  (lambda (high low)
    (receive (hh ll)
	(highlow-proc high low)
      (and hh
	   (arithmetic-mean hh ll)))))


(define (ichimoku-fill series new-array lo hi)
  (receive-list (tenkan-count kijun-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos parent #:before lo
					     #:count (* 3 kijun-count)))
	   (array  (series-array parent lo hi))
	   (pre-lo (max (series-lo parent) (- lo kijun-count)))
	   (phi    (min (series-hi parent) hi)))

      (define tenkan-proc   (range-midpoint-calc-proc tenkan-count))
      (define kijun-proc    (range-midpoint-calc-proc kijun-count))
      (define senkou-B-proc (range-midpoint-calc-proc (* 2 kijun-count)))

      ;; chikou - closing prices drawn KIJUN-COUNT points behind
      (let ((chikou-hi (min hi (- (series-hi parent)
				  kijun-count))))
	(do ((i lo (1+ i)))
	    ((>= i chikou-hi))
	  (and-let* ((close (array-ref array (+ i kijun-count) 3)))
	    (array-set! new-array close i 2))))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i pre-lo))
	(and-let* ((close (array-ref array i 3)))
	  (let ((high (or (array-ref array i 1) close))
		(low  (or (array-ref array i 2) close)))
	    (tenkan-proc high low)
	    (kijun-proc high low)
	    (senkou-B-proc high low))))

      ;; points lo to hi inclusive
      (do ((i pre-lo (1+ i)))
	  ((> i phi))
	(and-let* ((close (array-ref array i 3)))
	  (let ((high (or (array-ref array i 1) close))
		(low  (or (array-ref array i 2) close)))
	    (let* ((tenkan   (tenkan-proc high low))
		   (kijun    (kijun-proc high low))
		   (senkou-A (and tenkan kijun
				  (arithmetic-mean tenkan kijun)))
		   (senkou-B (senkou-B-proc high low)))
	      (if (>= i lo)
		  (begin
		    (array-set! new-array tenkan i 0)
		    (array-set! new-array kijun  i 1)))
	      (let ((forward (+ i kijun-count)))
		(if (<= forward hi)
		    (begin
		      (array-set! new-array senkou-A forward 3)
		      (array-set! new-array senkou-B forward 4)))))))))))

(define-public (ichimoku-series series tenkan-count kijun-count)
  (ohlcv-required "Ichimoku" series)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a,~a" (_ "Ichimoku")
					 tenkan-count kijun-count))
    (series-datatype! new-series ichimoku-datatype)
    (series-lo!       new-series (- (series-lo series) kijun-count))
    (series-hi!       new-series (+ (series-hi series) kijun-count))
    (series-parent!   new-series series)
    (series-data!     new-series (list tenkan-count kijun-count))
    (series-require!  new-series (indicator-require-parts ichimoku-fill
							  (* 3 kijun-count)
							  '(0 4)))
    (series-array!    new-series #f)
    new-series))

(define ichimoku-datatype
  (datatype-new!
   #:key       'ichimoku
   #:name      (_ "Ichimoku Kinko Hyo")
   #:units     #f
   #:input     datatype-ohlcv
   #:linestyle 'ichimoku
   #:manual    (_ "CONTEXT(manual-node): Ichimoku Kinko Hyo")
   #:proc      ichimoku-series
   #:closes-apply-arg '(0)
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'tenkan-days
	  #:type    'integer
	  #:minimum 1
	  #:default 9)
	 (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'kijun-days
	  #:type    'integer
	  #:minimum 1
	  #:default 26))))
