/* Chart character conversions.

   Copyright 2003, 2005, 2006, 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <string.h>
#include <langinfo.h>
#include <nl_types.h>
#include <stdio.h>     /* for NULL */

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_ICONV
#include <iconv.h>
#endif

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <libguile.h>

#include "misc.h"
#include "compat.h"


/*---------------------------------------------------------------------------*/
/* fake iconv when not otherwise available */

#if ! HAVE_ICONV

struct chart_iconv_t_data {
  char *from;  /* strdup'ed */
  char *to;    /* strdup'ed */
};
typedef struct chart_iconv_t_data *iconv_t;

#define ICONV_CONST

#ifndef EILSEQ  /* system without iconv probably doesn't have EILSEQ */
#define EILSEQ 12345
#endif

static iconv_t
iconv_open (const char *tocode, const char *fromcode)
{
  iconv_t cd = malloc (sizeof (struct chart_iconv_t_data));
  if (cd == NULL)
    {
    nomem:
      errno = ENOMEM;
      return (iconv_t) -1;
    }
  cd->from = strdup (fromcode);
  if (cd->from == NULL)
    {
      free (cd);
      goto nomem;
    }
  cd->to = strdup (tocode);
  if (cd->to == NULL)
    {
      free (cd->from);
      free (cd);
      goto nomem;
    }
  return cd;
}

static int
iconv_close (iconv_t cd)
{
  free (cd->from);
  free (cd->to);
  free (cd);
  return 0;
}

/* return non-zero if PART is a suffix of STR */
static int
strsuffix (const char *part, const char *str)
{
  size_t plen = strlen (part);
  size_t slen = strlen (str);
  return (slen >= plen
          && memcmp (str + slen - plen, part, plen) == 0);
}

static size_t
iconv (iconv_t cd,
       char **inbuf, size_t *inleft, char **outbuf, size_t *outleft)
{
  int  c;
  int  latin1, translit;

  if (inbuf == NULL)
    return 0;

  latin1 = (strcmp (cd->from, "ISO-8859-1") == 0);
  translit = strsuffix ("//TRANSLIT", cd->to);

  for (;;)
    {
      if (*inleft == 0)
        return 0;

      if (*outleft == 0)
        {
          errno = E2BIG;
          return (size_t) -1;
        }

      c = (int) (unsigned char) **inbuf;

      /* latin1 e-acute -> plain e, to support casablanca SOTA and SOTB
         names (that char is in an otherwise English page) */
      if (latin1)
        {
          if (c == 0xE9)
            c = 'e';
        }

      if (translit)
        {
          /* non-ASCII sequences translated to '?'
             no attempt is made to tell how many chars a non-ascii run is
             supposed to be, it just becomes a single '?' */
          if (! isascii (c))
            {
              while ((*inleft) > 1)
                {
                  c = (int) (unsigned char) *((*inbuf)+1); /* next char */
                  if (isascii (c))
                    break;
                  (*inbuf)++;
                  (*inleft)--;
                }
              c = '?';
            }
        }
      else
        {
          /* non-ASCII provokes EILSEQ */
          if (! isascii (c))
            {
              errno = EILSEQ;
              return (size_t) -1;
            }
        }

      **outbuf = c;

      (*inbuf)++;
      (*inleft)--;

      (*outbuf)++;
      (*outleft)--;
    }
}

#endif /* ! HAVE_ICONV */


/*---------------------------------------------------------------------------*/
/* iconv stuff */

static iconv_t utf8_to_wchar = (iconv_t) -1;

void
utf8_to_wchar_init (void)
{
#define FUNC_NAME "utf8_to_wchar_init"
  if (utf8_to_wchar == (iconv_t) -1)
    {
      utf8_to_wchar = iconv_open ("WCHAR_T", "UTF-8");
      if (utf8_to_wchar == (iconv_t) -1)
        {
        syserr:
          SCM_SYSERROR;
        }
    }
  else
    {
      if (iconv (utf8_to_wchar, NULL, NULL, NULL, NULL) == -1)
        goto syserr;
    }
#undef FUNC_NAME
}


/* return STR converted using CD
   if CLOSE_CD is non-zero then CD is closed when done (or error)
   FUNC is the originating function to report in an error */
static SCM
chart_iconv (const char *func, iconv_t cd, SCM str, int close_cd)
{
  ICONV_CONST char *in_str;
  char   *in_buf, *out_buf, *out_ptr;
  size_t in_len, out_alloc, out_left;
  SCM    out;

  if (! scm_is_string (str))
    {
      if (close_cd)
        iconv_close (cd);
      scm_wrong_type_arg_msg(func, SCM_ARGn, str, "STRINGP");
    }

  in_buf = scm_to_locale_stringn (str, &in_len);
  in_str = in_buf;

  out_alloc = 6 * in_len;
  out_buf = scm_malloc (out_alloc);

  out_ptr = out_buf;
  out_left = out_alloc;
  if (iconv (cd, &in_str, &in_len, &out_ptr, &out_left) == -1)
    {
      int err = errno;
      free (in_buf);
      free (out_buf);
      if (close_cd)
        iconv_close (cd);
      errno = err;
      scm_syserror (func);
    }

  out = scm_from_locale_stringn (out_buf, out_alloc - out_left);
  free (in_buf);
  free (out_buf);

  if (close_cd)
    {
      if (iconv_close (cd) != 0)
        scm_syserror (func);
    }
  return out;
}

/* open an iconv_t for charset conversion FROM to TO
   if cannot open then throw an error, reporting FUNC as the originator */
static iconv_t
chart_iconv_open (const char *func, const char *to, const char *from)
{
  iconv_t cd = iconv_open (to, from);
  if (cd == (iconv_t) -1)
    scm_syserror (func);
  return cd;
}

SCM_DEFINE_PUBLIC (chart_iconv_charset_available_p, "iconv-charset-available?", 1, 0, 0,
            (SCM charset),
	    "Return @code{#t} if @var{charset} is known to @code{iconv}, or\n"
	    "@code{#f} if not.  The test is that @var{charset} can be\n"
	    "converted to UTF-8.")
#define FUNC_NAME s_chart_iconv_charset_available_p
{
  char *c_charset = scm_to_locale_string (charset);

#if HAVE_ICONV
  {
    iconv_t cd = iconv_open ("UTF-8", c_charset);
    free (c_charset);
    if (cd == (iconv_t) -1)
      return SCM_BOOL_F;
    if (iconv_close (cd) != 0)
      SCM_SYSERROR;
    return SCM_BOOL_T;
  }

#else
  {
    int cmp = strcmp (c_charset, "ASCII");
    free (c_charset);
    if (cmp == 0)
      return SCM_BOOL_T;
    else
      return SCM_BOOL_F;
  }
#endif
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_locale_to_utf8, "locale->utf8", 1, 0, 0,
            (SCM str),
	    "Return @var{str} converted from locale encoding to UTF-8.  An\n"
	    "error is thrown if @var{str} has invalid bytes.")
#define FUNC_NAME s_chart_locale_to_utf8
{
  static iconv_t locale_to_utf8 = (iconv_t) -1;
  const char * codeset;

  codeset = nl_langinfo (CODESET);
  if (strcmp (codeset, "UTF-8") == 0)
    return str; /* locale is already utf-8, no translation */

  if (locale_to_utf8 == (iconv_t) -1)
    locale_to_utf8 = chart_iconv_open (FUNC_NAME,
                                       "UTF-8", nl_langinfo (CODESET));
  else
    if (iconv (locale_to_utf8, NULL, NULL, NULL, NULL) == -1)
      SCM_SYSERROR;

  return chart_iconv (FUNC_NAME, locale_to_utf8, str, 0);
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_charset_to_utf8, "charset->utf8", 2, 0, 0,
            (SCM charset, SCM str),
	    "Return @var{str} converted from the given @var{charset} to\n"
	    "UTF-8.  @var{charset} is a string, the name of a charset.  An\n"
	    "error is thrown if @var{str} has invalid bytes.")
#define FUNC_NAME s_chart_charset_to_utf8
{
  iconv_t cd;
  char *c_charset = scm_to_locale_string (charset);
  int saved_errno;

  if (strcmp (c_charset, "UTF-8") == 0)
    {
      free (c_charset);
      return str; /* no conversion */
    }

  cd = iconv_open ("UTF-8", c_charset);
  saved_errno = errno;
  free (c_charset);
  if (cd == (iconv_t) -1)
    {
      errno = saved_errno;
      SCM_SYSERROR;
    }

  return chart_iconv (FUNC_NAME, cd, str, 1);
}
#undef FUNC_NAME


static char *
scm_strcatdup (const char *s, const char *t)
{
  char *r = scm_malloc (strlen (s) + strlen (t) + 1);
  strcpy (r, s);
  return strcat (r, t);
}

SCM_DEFINE_PUBLIC (chart_utf8_to_locale, "utf8->locale", 1, 0, 0,
            (SCM str),
	    "Return @var{str} converted from UTF-8 into the locale encoding.\n"
	    "\n"
	    "Characters which don't exist in the locale charset are\n"
	    "transliterated by @code{iconv} where possible, or are changed\n"
	    "to question marks \"?\" otherwise.")
#define FUNC_NAME s_chart_utf8_to_locale
{
  static iconv_t utf8_to_locale = (iconv_t) -1;

  ICONV_CONST char *in_str;
  char       *in_buf, *out_buf, *out_ptr;
  size_t     in_len, out_alloc, out_left;
  SCM        out;
  const char *codeset;

  codeset = nl_langinfo (CODESET);
  if (strcmp (codeset, "UTF-8") == 0)
    return str; /* locale is already utf-8, no translation */

  if (utf8_to_locale == (iconv_t) -1)
    {
      const char *codeset = nl_langinfo (CODESET);
      char *locale_translit = scm_strcatdup (codeset, "//TRANSLIT");
      utf8_to_locale = iconv_open (locale_translit, "UTF-8");
      free (locale_translit);

      if (utf8_to_locale == (iconv_t) -1)
        utf8_to_locale = chart_iconv_open (FUNC_NAME, codeset, "UTF-8");
    }
  else
    {
      if (iconv (utf8_to_locale, NULL, NULL, NULL, NULL) == -1)
        goto syserr;
    }

  in_buf = scm_to_locale_stringn (str, &in_len);
  in_str = in_buf;

  out_alloc = 4 * in_len;
  out_buf = scm_malloc (out_alloc);

  out_ptr = out_buf;
  out_left = out_alloc;

  if (iconv (utf8_to_locale, &in_str, &in_len, &out_ptr, &out_left) == -1)
    {
      if (errno != EILSEQ)
        {
          int saved_errno;
        free_syserror:
          saved_errno = errno;
          free (in_buf);
          free (out_buf);
          errno = saved_errno;
        syserr:
          SCM_SYSERROR;
        }

      /* EILSEQ, input char unrepresentable in locale, show "?" */

      utf8_to_wchar_init ();
      for (;;)
        {
          wchar_t buf;
          char *bufptr = (char *) &buf;
          size_t buflen = sizeof(buf);
          size_t old_in_len = in_len;

          /* success if this is the last char of the input string
             E2BIG if there's more input

             check in_len has decreased before allowing E2BIG, must have
             converted something or we go into an infinite loop  */

          if (iconv (utf8_to_wchar, &in_str, &in_len, &bufptr, &buflen) == -1
              && (errno != E2BIG || in_len == old_in_len))
            goto free_syserror;

          *out_ptr++ = '?';
          out_left--;

          /* convert rest of input, looping around for "_" on EILSEQ char */
          if (iconv (utf8_to_locale, &in_str, &in_len, &out_ptr, &out_left)
              != -1)
            break;
          if (errno != EILSEQ)
            goto free_syserror;
        }
    }

  out = scm_from_locale_stringn (out_buf, out_alloc - out_left);
  free (in_buf);
  free (out_buf);
  return out;
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_utf8_validate, "utf8-validate", 1, 0, 0,
            (SCM str),
	    "Check that @var{str} is a valid UTF-8 encoded string.  Throw an\n"
	    "error if it's not.\n"
	    "\n"
	    "If @code{iconv} is not available, this function does nothing.")
#define FUNC_NAME s_chart_utf8_validate
{
#if HAVE_ICONV
  static iconv_t utf8_to_utf8 = (iconv_t) -1;

  ICONV_CONST char *in_str;
  char *in_buf, *out_buf, *out_ptr;
  size_t in_len, out_alloc, out_left;
  int ret, saved_errno;

  if (utf8_to_utf8 == (iconv_t) -1)
    {
      utf8_to_utf8 = iconv_open ("UTF-8", "UTF-8");
      if (utf8_to_utf8 == (iconv_t) -1)
        {
        syserr:
          SCM_SYSERROR;
        }
    }
  else
    {
      if (iconv (utf8_to_utf8, NULL, NULL, NULL, NULL) == -1)
        goto syserr;
    }

  in_buf = scm_to_locale_stringn (str, &in_len);
  in_str = in_buf;

  out_alloc = in_len;
  out_buf = scm_malloc (out_alloc);

  out_ptr = out_buf;
  out_left = out_alloc;

  ret = iconv (utf8_to_utf8, &in_str, &in_len, &out_ptr, &out_left);

  saved_errno = errno;
  free (in_buf);
  free (out_buf);

  if (ret == -1)
    {
      errno = saved_errno;
      goto syserr;
    }
#else

  SCM_VALIDATE_STRING (SCM_ARG1, str);
#endif

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_utf8_string_length, "utf8-string-length", 1, 0, 0,
            (SCM str),
	    "Return the number of UTF-8 encoded characters in @var{str}.\n"
	    "An error is thrown if @var{str} is not properly encoded.")
#define FUNC_NAME s_chart_utf8_string_length
{
#if HAVE_ICONV
  ICONV_CONST char *in_str;
  char *in_buf, *out_buf, *out_ptr;
  size_t in_len, out_alloc, out_left;
  int ret, saved_errno;

  utf8_to_wchar_init ();

  in_buf = scm_to_locale_stringn (str, &in_len);
  in_str = in_buf;

  out_alloc = sizeof (wchar_t) * in_len;
  out_buf = scm_malloc (out_alloc);
  out_ptr = out_buf;
  out_left = out_alloc;

  ret = iconv (utf8_to_wchar, &in_str, &in_len, &out_ptr, &out_left);

  saved_errno = errno;
  free (in_buf);
  free (out_buf);
  if (ret == -1)
    {
      errno = saved_errno;
      SCM_SYSERROR;
    }

  return scm_from_size_t ((wchar_t *) out_ptr - (wchar_t *) out_buf);
#else

  return scm_string_length (str);
#endif
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_utf8_string_take, "utf8-string-take", 2, 0, 0,
            (SCM str, SCM len),
	    "Return a string comprising the first @var{len} UTF-8 encoded\n"
	    "characters in @var{str}.  An error is thrown if @var{str} is\n"
	    "not properly encoded.")
#define FUNC_NAME s_chart_utf8_string_take
{
#if HAVE_ICONV
  ICONV_CONST char *in_str;
  char *in_buf, *out_buf, *out_ptr;
  size_t want_len, in_len, out_alloc, out_left;
  int ret, saved_errno;

  want_len = scm_to_size_t (len);
  utf8_to_wchar_init ();

  in_buf = scm_to_locale_stringn (str, &in_len);
  in_str = in_buf;

  out_alloc = sizeof (wchar_t) * want_len;
  out_buf = scm_malloc (out_alloc);
  out_ptr = out_buf;
  out_left = out_alloc;

  ret = iconv (utf8_to_wchar, &in_str, &in_len, &out_ptr, &out_left);

  saved_errno = errno;
  free (in_buf);
  free (out_buf);
  if (ret == -1 && saved_errno != E2BIG)
    {
      errno = saved_errno;
      SCM_SYSERROR;
    }

  if ((wchar_t*) out_ptr - (wchar_t*) out_buf != want_len)
    scm_out_of_range (FUNC_NAME, len);

  len = scm_from_size_t (in_str - in_buf);
#endif

  /* In guile 1.6 scm_string_take isn't in the core and srfi-13.h is a bit
     hard to use, so stay with scm_substring to avoid trouble.  */
  return scm_substring (str, scm_from_int (0), len);
}
#undef FUNC_NAME



/*---------------------------------------------------------------------------*/

void
chart_iconv_init (void)
{
#if HAVE_ICONV
  scm_c_define ("iconv-available?", SCM_BOOL_T);
#else
  scm_c_define ("iconv-available?", SCM_BOOL_F);
#endif
  scm_c_export ("iconv-available?", NULL);

#ifndef SCM_MAGIC_SNARFER
#include "iconv.x"
#endif
}
