;; Intraday momentum index.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart imi)
  #:use-module (chart indicator)
  #:use-module (chart misc))

;; http://www.prophet.net/learn/taglossary.jsp?index=I
;;     Description.
;;
;; http://www.fmlabs.com/reference/IMI.htm
;;     Formula, brief description.
;;


;; Return a procedure which calculates an intraday momentum index over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC open close) enters a new point into the window, and the
;; return is the intraday momentum index up to (and including) that point.
;; If there's not enough points yet the return is #f.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (imi-calc-proc count)

  (define num-proc (sum-calc-proc count))
  (define den-proc (sum-calc-proc count))

  (lambda (open close)
    (and open
	 (let* ((num (num-proc (if (> close open)
				   (- close open)
				   0)))
		(den (den-proc (abs (- open close)))))
	   (and num ;; when enough data
		(not (zero? den))
		(/ (* 100.0 num) ;; force to inexact to avoid fractions
		   den))))))

(define-public imi-warmup-count 1-)


(define (imi-fill series new-array lo hi)
  (let* ((count    (series-data series))
	 (parent   (series-parent series))
	 (start    (series-prev-n-closes-pos
		    parent
		    #:before lo
		    #:count  (imi-warmup-count count)))
	 (array    (series-array parent start hi))
	 (imi-proc (imi-calc-proc count)))

    ;; warmup points preceding lo, if available
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (imi-proc (array-ref array i 0)    ;; open
		    (array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array
		      (imi-proc (array-ref array i 0)   ;; open
				(array-ref array i 3))  ;; close
		      i)))))

(define-public (imi-series series count)
  (ohlcv-required "IMI" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "IMI") count))
    (series-datatype!  new-series imi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-array!     new-series #f)
    (series-require!   new-series (indicator-require-parts
				   imi-fill 
				   (imi-warmup-count count)))
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public imi-datatype
  (datatype-new!
   #:key       'imi
   #:name      (_ "IMI - Intraday Momentum Index")
   #:input     datatype-ohlcv
   #:units     'percentage
   #:minimum   0
   #:maximum   100
   #:hlines    '(30 70)
   #:manual    (_ "CONTEXT(manual-node): Intraday Momentum Index")
   #:proc      imi-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'imi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
