;; Indicator popup menus.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart indicator-menu)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart keymap)
  #:use-module (chart misc)
  #:use-module (chart view)
  #:use-module (chart viewstyle))


;;-----------------------------------------------------------------------------
;; GtkMenu stuff

;; Append ITEM to MENU like `gtk-menu-append' but if MENU is getting big
;; start a submenu of "More..." at the end.  The return value is the new
;; menu to continue appending to.
;;
;; Gtk 2 has scrolling for big menus so doesn't need this stuff (though the
;; scrolling isn't fantastic).
;;
(define-public (c-gtk-menu-append-more menu item)
  (if (and (= 1 gtk-major-version)
	   (> (+ (cdr (gtk-widget-size-request menu)))
	      (cdr (gtk-widget-size-request item)))
	 (* 0.9 (gdk-screen-height)))
      (let ((subitem (gtk-menu-item-new-with-label (_ "More ...")))
	    (submenu (gtk-menu-new)))
	(gtk-menu-item-set-submenu subitem submenu)
	(gtk-widget-show subitem)
	(gtk-menu-append menu subitem)
	(set! menu submenu)))
  (gtk-menu-append menu item)
  menu)

;; Return all the items in MENU, and all items in submenus too.
(define (c-gtk-menu-all-children menu)
  (let ((children (gtk-container-children menu)))
    (append children
	    (append-map (lambda (item)
			  (let ((submenu (gtk-menu-item-submenu item)))
			    (if submenu
				(c-gtk-menu-all-children submenu)
				'())))
			children))))

;; turn off check mark for all GtkCheckMenuItem children of the GtkMenu MENU
(define (c-gtk-menu-uncheck-all menu)
  (for-each (lambda (item)
	      (if (gtk-check-menu-item? item)
		  (gtk-check-menu-item-set-active item #f)))
	    (c-gtk-menu-all-children menu)))


;;-----------------------------------------------------------------------------
;; indicator menu

(gtk-class-new 'GtkMenu "ChartIndicatorMenu")

(define indicator-menu-list '())

(define (select-item-upper datatype)
  (let ((viewstyle (copy-tree (preference-get 'viewstyle viewstyle-default))))
    (set-first! (second (first viewstyle)) (datatype-key datatype))
    (preference-set 'viewstyle viewstyle)))

(define (select-item-lower datatype)
  (let ((viewstyle (copy-tree (preference-get 'viewstyle viewstyle-default))))
    (set-first! (first (second viewstyle)) (datatype-key datatype))
    (preference-set 'viewstyle viewstyle)))

;; workaround for trouble creating radio menu items
(define (c-gtk-radio-menu-item-new-with-label str)
  (let ((item  (gtk-widget-new 'GtkRadioMenuItem))
	(label (gtk-label-new str)))
    (gtk-misc-set-alignment label 0 0.5)
    (gtk-container-add item label)
    item))

(define (indicator-menu-get graph)
  (or (object-property graph 'indicator-menu)
      (let ((menu (gtk-widget-new 'ChartIndicatorMenu)))
	;; dummy to avoid selecting the first entry "None" too easily
	(gtk-menu-shell-append menu (gtk-separator-menu-item-new))

	(gtk-menu-shell-append menu (gtk-tearoff-menu-item-new))

	(let* ((submenu       menu)
	       (datatype-list (sort (filter (if (eq? graph draw)
						datatype-parent-units?
						datatype-own-units?)
					    (datatype-list-all))
				    datatype<?))
	       (prev-priority (datatype-priority (first datatype-list)))
	       (group-item    #f)) ;; the first one created
	  (for-each
	   (lambda (datatype)
	     ;; separator between priority levels
	     (or (eqv? prev-priority (datatype-priority datatype))
		 (gtk-menu-shell-append submenu (gtk-separator-menu-item-new)))
	     (set! prev-priority (datatype-priority datatype))

	     (let ((item (c-gtk-radio-menu-item-new-with-label
			  (datatype-name datatype))))
	       (set-object-property! item 'datatype datatype)
	       (if group-item
		   (gtk-widget-set item #:group group-item)
		   (set! group-item item))
	       (gtk-widget-show item)
	       (set! submenu (c-gtk-menu-append-more submenu item))
	       (gtk-signal-connect item "activate"
		 (lambda ()
		   ;; also called when toggling off, so check for on
		   (if (gtk-check-menu-item-active item)
		       (if (eq? graph draw)
			   (select-item-upper datatype)
			   (select-item-lower datatype)))))))
	   datatype-list))

	(gtk-widget-show-all menu)
	(set-object-property! graph 'indicator-menu menu)
	(set-object-property! menu  'graph graph)
	(set! indicator-menu-list (cons menu indicator-menu-list))
	menu)))

(define (menu-upper-get-key)
  (first (second (first (preference-get 'viewstyle viewstyle-default)))))

(define (menu-lower-get-key)
  (first (first (second (preference-get 'viewstyle viewstyle-default)))))

;; return the `datatype-key' for a given ITEM (a GtkRadioMenuItem)
(define (item-datatype-key item)
  (and-let* ((datatype (object-property item 'datatype)))
    (datatype-key datatype)))

;; set the check mark in MENU for the item corresponding to the datatype
;; selected for its graph
;; the datatype might not have an item if it comes from a saved prefs file
;; when ids or something have changed, uncheck everything in that case
;;
(define (menu-freshen menu)
  (let ((key (if (eq? (object-property menu 'graph) draw)
		 (menu-upper-get-key)
		 (menu-lower-get-key))))
    (let ((item (find (lambda (item)
			(eq? key (item-datatype-key item)))
		      (c-gtk-menu-all-children menu))))
      (if item
	  (gtk-check-menu-item-set-active item #t)
	  (c-gtk-menu-uncheck-all menu)))))

(notify-connect 'preferences-save
  (lambda (dummy)
    (for-each (lambda (menu)
		(if (gtk-widget-mapped menu)
		    (menu-freshen menu)))
	      indicator-menu-list)))


(define-public (indicator-menu-popup)
  (let ((menu (indicator-menu-get keymap-widget)))
    (menu-freshen menu)
    (gtk-menu-popup menu #f #f (gdk-event-button keymap-event) 0)))


