;; Indicator helpers.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart indicator)
  #:use-module (ice-9 optargs)
  #:autoload   (ice-9 q)   (sync-q!)
  #:use-module (srfi srfi-1)
  #:use-module (chart misc))


(define-public indicator-minimum-fraction 0.75)


;;-----------------------------------------------------------------------------
;; ice-9 q stuff

;; return a new queue consisting of the elements of LST
;; LST is copied so that subsequent q operations don't modify it
(define-public (list->q lst)
  (sync-q! (cons (list-copy lst) #f)))


;;-----------------------------------------------------------------------------
;; misc stuff

;; compare the cars of X and Y with >
(define-public (car> x y)
  (> (car x) (car y)))

;; return the inexact number X formatted as a string with at least DECIMALS
;; many decimal places, and more if necessary for the value
;;
(define-public (locale-number-min-decimals x decimals)
  (locale-number+decimals (* x (integer-expt 10 decimals))
			  decimals))

;; `take-right' but limiting N to the LST length
(define-public (take-right-maybe lst n)
  (if (length>=? lst n)
      (take-right lst n)
      lst))

;; return a circular list comprising N copies of INIT
;; like `make-list', but a circular list
;;
(define-public (make-circular-list n init)
  (let ((lst (make-list n init)))
    (set-cdr! (last-pair lst) lst)
    lst))

;; Find the first integer >=0 for which (PRED i) returns true.  INCR is a
;; guess at what sort of increment to use for the initial upwards probing.
;;
(define-public (bsearch-first-true pred incr)
  (set! incr (ceiling-exact incr))
  (let ((lo 0)
	(hi incr))

    ;; search upwards
    (let upwards ()
      (or (pred hi)
	  (begin
	    (set! lo (1+ hi))
	    (set! hi (* 2 hi))
	    (upwards))))

    ;; at this point (PRED lo) is unknown, (PRED hi) is #t

    (if (pred lo)
	lo

	;; binary search loop, with (PRED lo) #f, (PRED hi) #t
	(do ()
	    ((<= (- hi lo) 1)
	     hi)
	  (let ((mid (ash (+ lo hi) -1)))
	    (if (pred mid)
		(set! hi mid)
		(set! lo mid)))))))

;; return the element of LST at the F'th fractile
;;
;; if k=length(lst) then the first element would be for 0<=f<1/k, the
;; second for 1/k<=f<2/k, etc, up to the last element for (k-1)/k<=f<=1
;;
(define-public (list-ref-fraction lst f)
  (let ((len (length lst)))
    (list-ref lst (min (1- len) (floor-exact (* f len))))))

(define-public (minmax-restrict lo hi x)
  (max lo (min hi x)))


;;-----------------------------------------------------------------------------
;; series stuff

(define-public (series-closes-apply-arg series)
  (let ((datatype (series-datatype series)))
    (if datatype
	(datatype-closes-apply-arg datatype)
	'())))

;; like `series-array', but pick out just the closing prices from ohlcv
(define-public (series-array-closes series lo hi)
  (define array (series-array series lo hi))
  (define arg   (series-closes-apply-arg series))
  (if (null? arg)
      array
      (apply make-shared-array-column array arg)))

;; Return the start position of the COUNT'th previous closing price in
;; SERIES prior to BEFORE (and not including that).
;;
;; BEFORE can be outside the SERIES range, either above or below.  When
;; above the search starts from the actual high.  When below the return will
;; be (series-lo SERIES), which is also the return if there are fewer than
;; COUNT closes before BEFORE.
;;
(define*-public (series-prev-n-closes-pos
		 series #:key
		 (before (error "series-prev-n-closes-pos: missing #:before"))
		 (count  (error "series-prev-n-closes-pos: missing #:count")))
  (set! before (min before (1+ (series-hi series))))
  (if (<= count 0)
      (max (series-lo series) (min before (series-hi series)))

      (let ((got         0)
	    (first-chunk (+ 3 (ceiling-exact (* 1.05 count))))
	    (arg         (series-closes-apply-arg series))
	    (pos         (1- before)))

	(define (examine-chunk n)
	  (let* ((chunk-stop  (- pos n -1))
		 (array (series-array series chunk-stop pos)))
	    (let more ()
	      (cond ((< pos (series-lo series))
		     (series-lo series)) ;; hit beginning of series

		    ((< pos chunk-stop)
		     #f)  ;; out and grab more array

		    ((apply array-ref array pos arg)
		     (set! got (1+ got))
		     (if (>= got count)
			 pos  ;; found desired number
			 (begin
			   (set! pos (1- pos))
			   (more))))

		    (else
		     (set! pos (1- pos))
		     (more))))))

	;; Guess 12 public holidays per year, say another 3 days trading
	;; halts is 15 out of 260 weekdays, is about a factor 1.05 to
	;; stretch.  Add a fixed 3 in case just hit christmas or easter.
	;;
	;; If that first estimate isn't enough, then probe back by expanding
	;; amounts.
	;;
	(or (examine-chunk first-chunk)
	    (let more ((chunk (+ 4 (ceiling-exact (* 1.2 (- count got))))))
	      (or (examine-chunk chunk)
		  (more (* 2 chunk))))))))


;; Return a list of COUNT many closing prices in SERIES prior to POS (and
;; not including POS).  The returned list may be shorter than COUNT if
;; there's fewer than that many before the start of SERIES.
;;
(define*-public (series-prev-n-closes series #:key before count extract)
  (let* ((start (min (1- before) (series-hi series)))
	 (stop  (series-prev-n-closes-pos series
					  #:before before #:count count))
	 (arg   (series-closes-apply-arg series))
	 (array (series-array series start stop))
	 (ret   '()))
    (or extract
	(set! extract (lambda (i)
			(apply array-ref array i arg))))
    (do ((i start (1- i)))
	((< i stop))
      (if (apply array-ref array i arg)
	  (set! ret (cons (extract i) ret))))
    ret))


;; check that SERIES is a datatype-ohlcv, throwing an error if not
;; NAME is given in the message to identify the module needing ohlcv
;; there's no message translation since it's only a diagnostic
;;
(define-public (ohlcv-required name series)
  (or (eq? datatype-ohlcv (series-datatype series))
      (error (string-append name ": need OHLCV series:")
	     (datatype-key (series-datatype series)))))


;;-----------------------------------------------------------------------------

;; Return a circular list of COUNT values, with the list INITIAL providing
;; the initial contents.
;;
;; If INITIAL is shorter than COUNT, it's padded at the start with #f's.  If
;; INITIAL is longer then just the last COUNT values of it are taken.
;;
(define-public (indicator-circular count initial)
  (let ((len (length initial)))
    (if (< len count)
	(set! initial (append! (make-list (- count len) #f) initial))
	(set! initial (take-right initial count))))
  (apply circular-list initial))


;;-----------------------------------------------------------------------------

;; Return a procedure which returns the COUNT'th previous value.  For
;; example COUNT=1 would on each call return the value from the immediately
;; preceding call.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the value entered COUNT many points prior to VAL, or #f if not that
;; many points entered yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from first to last).  COUNT many
;; values is enough to ensure the first PROC call gives out the COUNT'th
;; previous value.
;;
(define*-public (prevnth-calc-proc count #:optional (data '()))
  (set! data (indicator-circular (1+ count) data))
  (lambda (x)
    (set-car! data x)
    (set! data (cdr data))
    (car data)))

;; Return a procedure which calculates a moving sum over a window of COUNT
;; values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the sum up to (and including) that value, or #f if not enough values
;; yet (the usual indicator-minimum-fraction).
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from start to end).  COUNT-1 is
;; enough to ensure the first PROC call is based on a full window.
;;
(define*-public (sum-calc-proc count #:optional (data '()))

  ;; POINTS is the number of non-#f points in DATA
  ;; TOTAL is the sum of the point values in DATA

  (set! data (take-right-maybe data count))
  (let ((min-points (ceiling-exact (* indicator-minimum-fraction count)))
	(points     (length data))
	(total      (apply + data)))
    (set! data (indicator-circular count data))

    (lambda (x)

      (if (car data)
	  ;; drop oldest point
	  (set! total  (- total (car data)))
	  ;; otherwise gaining a point
	  (set! points (1+ points)))

      ;; replace oldest with new point
      (set-car! data x)
      (set! data (cdr data))

      ;; add this point
      (set! total (+ total x))

      (and (>= points min-points)
	   total))))


;;-----------------------------------------------------------------------------

(define*-public (indicator-require-parts fill warmup #:optional shape)
  ;; allow for some public holidays
  (set! warmup (ceiling-exact (* 1.1 warmup)))

  (set! shape (if shape (list shape) '()))

  (lambda (series lo hi)
    (indicator-require-parts-main series lo hi fill warmup shape)))

(define (indicator-require-parts-main series lo hi fill warmup shape)

  ;; extend to start of series if warmup could cover that anyway
  (if (<= (- lo warmup) (series-lo series))
      (set! lo (series-lo series)))

  ;; extend hi and/or lo to do at least 1/2 the warmup amount
  (let ((small (ceiling-exact (* 0.5 warmup))))
    (set! hi (min (series-hi series)
		  (max hi (+ lo small))))
    (set! lo (max (series-lo series)
		  (min lo (- hi small)))))

  ;; prefetch parent portion likely needed
  (series-array (series-parent series) (- lo warmup) hi)

  (let* ((new-array (apply make-array #f (list lo hi) shape))
	 (old-array (series-held-array series)))

    (if old-array
	(let ((old-lo (array-first-index old-array))
	      (old-hi (array-last-index old-array)))
	  (array-copy! old-array new-array)

	  (if (<= (- old-hi old-lo) warmup)
	      ;; warmup would cover whole existing part, so fill the lot in
	      ;; one call
	      (fill series new-array lo hi)

	      (begin
		(if (< lo old-lo)
		    (fill series new-array lo (1- old-lo)))
		(if (> hi old-hi)
		    (fill series new-array old-hi hi)))))

	;; no previous contents, one fill of everything
	(fill series new-array lo hi))

    new-array))


;;-----------------------------------------------------------------------------
;; accumulation style series

;; return the position of a non-#f close at or near T, or #f if none at all
(define-public (series-find-close series t)
  (let ((arg (series-closes-apply-arg series)))
    (or (let more ((t t))
	  (and (<= t (series-hi series))
	       (if (apply array-ref (series-array series t t) t arg)
		   t
		   (more (1+ t)))))
	(let more ((t t))
	  (and (>= t (series-lo series))
	       (if (apply array-ref (series-array series t t) t arg)
		   t
		   (more (1- t))))))))

(define (accumulation-series-require series lo hi)
  (receive-list (lo-post-pos hi-prev-pos func multiplicative)
      (series-data series)
    (define forward-acc  (if (zero? multiplicative) + *))
    (define backward-acc (if (zero? multiplicative) - /))

    (let* ((old-array (series-held-array series))
	   (old-lo    (and old-array (array-first-index old-array)))
	   (old-hi    (and old-array (array-last-index old-array)))
	   (new-array #f)
	   (parent    (series-parent series))
	   (array     (series-array parent lo hi)))

      (if old-array
	  (begin
	    ;; previous contents
	    (set! new-array (make-array #f (list lo hi)))
	    (array-copy! old-array new-array))

	  (begin
	    ;; or start from zero at or near LO
	    (set! old-lo (series-find-close parent lo))
	    (if old-lo
		(begin
		  (set! old-hi old-lo)
		  (set! lo (min lo old-lo)) ;; in case found outside requested
		  (set! hi (max hi old-hi))
		  (set! lo-post-pos old-lo)
		  (set! hi-prev-pos old-lo)
		  (set! new-array (make-array #f (list lo hi)))
		  (array-set! new-array multiplicative old-lo))

		;; oops, parent has all #f closes, fill ourself with #f's,
		;; using a shared array
		(begin
		  (set! lo (series-lo parent))
		  (set! hi (series-hi parent))
		  (set! new-array (series-array-closes parent lo hi))
		  (set! old-lo lo)
		  (set! old-hi hi)))))

      ;; extend forwards
      ;;
      (if (> hi old-hi)
	  (let ((acc (array-ref new-array hi-prev-pos)))
	    (do ((i (1+ old-hi) (1+ i)))
		((> i hi))
	      (if (array-ref array i 3)
		  (begin
		    (set! acc (forward-acc acc (func array hi-prev-pos i)))
		    (set! hi-prev-pos i)
		    (array-set! new-array acc i))))))
      (set-second! (series-data series) hi-prev-pos)

      ;; extend backwards
      ;;
      (if (< lo old-lo)
	  (let ((acc (array-ref new-array lo-post-pos)))
	    (do ((i (1- old-lo) (1- i)))
		((< i lo))
	      (if (array-ref array i 3)
		  (begin
		    (set! acc (backward-acc acc (func array i lo-post-pos)))
		    (set! lo-post-pos i)
		    (array-set! new-array acc i))))))
      (set-first!  (series-data series) lo-post-pos)

      new-array)))

(define*-public (accumulation-series series name datatype func
				     #:key
				     (multiplicative 0))
  (ohlcv-required name series)
  (let ((new-series (series-copy series)))
    (series-data!      new-series (list #f #f func multiplicative))
    (series-datatype!  new-series datatype)
    (series-name!      new-series name)
    (series-require!   new-series accumulation-series-require)
    (series-parent!    new-series series)
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))
