;; Inertia indicator.

;; Copyright 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart inertia)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart rvi)
  #:use-module (chart indicator)
  #:use-module (chart epma)
  #:use-module (chart misc))


;; http://trader.online.pl/MSZ/e-w-Inertia.html
;;     
;; http://www.fmlabs.com/reference/Inertia.htm
;;     Formula, some description.
;;
;; http://www.prophet.net/analyze/popglossary.jsp?studyid=INERT
;;     Sample DELL from 2001/2.
;;
;; http://store.traders.com/-v13-c09-refinin-pdf.html
;;     Dorsey TASC 1995 for sale.
;;



;; Return a procedure which calculates an inertia indicator.
;;
;; Each call (PROC high low) enters a new day's trading range into the
;; window, and the return is the inertia up to (and including) that value,
;; or #f if not enough points yet.
;;
;; HIGH and LOW must not be #f, close values can be substituted if
;; necessary, and in that case the underlying RVI calculation ends up
;; falling back to Dorsey's 1993 original definition.
;;
;; An inertia is in theory influenced by all preceding data, but
;; `inertia-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (inertia-warmup-count STDDEV-COUNT SMOOTH-COUNT
;; LSQMA-COUNT) many values, the next call will have an omitted weight of no
;; more than 0.1% of the total.  Omitting 0.1% should be negligable, unless
;; past values are ridiculously bigger than recent ones.
;;
(define-public (inertia-calc-proc stddev-count smooth-count lsqma-count)
  (let ((rvi-proc   (rvi-calc-proc stddev-count smooth-count))
	(lsqma-proc (epma-calc-proc lsqma-count)))

    (lambda (high low)
      (and-let* ((rvi (rvi-proc high low)))
	(lsqma-proc rvi)))))

(define-public (inertia-warmup-count stddev-count smooth-count lsqma-count)
  (+ (rvi-warmup-count stddev-count smooth-count)
     lsqma-count -1))




(define (inertia-fill series new-array lo hi)
  (receive-list (stddev-count smooth-count lsqma-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos
		    parent #:before lo
		    #:count (inertia-warmup-count stddev-count smooth-count
						  lsqma-count)))
	   (array  (series-array parent start hi))
	   (proc   (inertia-calc-proc stddev-count smooth-count lsqma-count)))

      ;; warmup
      (do ((i start (1+ i)))
	  ((>= i lo))
	(and-let* ((close (array-ref array i 3)))
	  (proc (or (array-ref array i 1) close)    ;; high
		(or (array-ref array i 2) close)))) ;; low

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((close (array-ref array i 3)))
	  (array-set! new-array (proc (or (array-ref array i 1) close)  ;; high
				      (or (array-ref array i 2) close)) ;; low
		      i))))))

(define-public (inertia-series series stddev-count smooth-count lsqma-count)
  (ohlcv-required "Inertia" series)
  (let ((new-series (series-copy series)))
    (series-name!    new-series (format #f "~a ~a, ~a, ~a" (_ "Inertia")
					stddev-count smooth-count lsqma-count))
    (series-datatype!  new-series inertia-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list stddev-count smooth-count lsqma-count))
    (series-require!   new-series
		       (indicator-require-parts
			inertia-fill (inertia-warmup-count
				      stddev-count smooth-count lsqma-count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public inertia-datatype
  (datatype-new!
   #:key     'inertia
   #:name    (_ "Inertia")
   #:input   datatype-ohlcv
   #:units   'percentage
   #:minimum 0
   #:maximum 100
   #:hlines  '(40 50 60)
   #:manual  (_ "CONTEXT(manual-node): Relative Volatility Index")
   #:proc    inertia-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'rvi-stddev-days
	  #:name    (_ "Stddev Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)
	 (datatype-param-new
	  #:key     'rvi-days
	  #:name    (_ "RVI Smooth Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14)
	 (datatype-param-new
	  #:key     'inertia-lsqma-days
	  #:name    (_ "LSQMA Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
