;; opint sign ?
;; 2*I can go negative





;; Herrick Payoff Index.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart hpi)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.marketscreen.com/help/AtoZ/default.asp?hideHF=&Num=54
;;
;; http://www.fmlabs.com/reference/HerrickPayoffIndex.htm
;;     Formula only, not sure about position of "1 +".



;; Return a procedure which calculates a Herrick payoff index over an
;; accumulated window of values.
;;
;; Each call (PROC high low volume openint) enters a new value into the
;; window, and the return is the HPI up to (and including) that, or #f if
;; not enough values yet.
;;
;; Because HPI uses an exponential moving average, it's in theory influenced
;; by all preceding data, but `hpi-warmup-count' below is designed to
;; determine a warmup count.
;;
(define-public (hpi-calc-proc smooth)

  (define openint-prev  #f)
  (define midpoint-prev #f)
  (define ema-proc      (ema-calc-proc smooth))

  (lambda (high low volume openint)
    (let* ((midpoint    (* 0.5 (+ high low)))
	   (openint-max (and openint openint-prev
			     (max openint openint-prev)))
	   (int-factor  (if (and openint-max
				 (positive? openint-max))
			    ((if (>= midpoint midpoint-prev) + -)
			     1 (/ (* 2.0 (abs (- openint openint-prev)))
				  openint-max))
			    1)) ;; no open interest
	   (raw         (and midpoint-prev
			     (* volume
				(- midpoint midpoint-prev)
				(or int-factor 1))))
	   (hpi         (and raw
			     (ema-proc raw))))
      (set! openint-prev  openint)
      (set! midpoint-prev midpoint)
      hpi)))

(define-public (hpi-warmup-count smooth)
  (1+ (ema-warmup-count smooth)))


(define (hpi-fill series new-array lo hi)
  (let* ((smooth (series-data series))
	 (parent (series-parent series))
	 (warmup (hpi-warmup-count smooth))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count warmup))
	 (array  (series-array parent start hi))
	 (proc   (hpi-calc-proc smooth))
	 (open-interest? (ohlcv-open-interest? series)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(proc (or (array-ref array i 1) close)   ;; high
	      (or (array-ref array i 2) close)   ;; low
	      (or (array-ref array i 4) 1)       ;; volume
	      (and open-interest?
		   (array-ref array i 5)))))     ;; open int

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array
		    (proc (or (array-ref array i 1) close)   ;; high
			  (or (array-ref array i 2) close)   ;; low
			  (or (array-ref array i 4) 1)       ;; volume
			  (and open-interest?
			       (array-ref array i 5)))       ;; open int
		    i)))))


(define-public (hpi-series series smooth)
  (ohlcv-required "Herrick Payoff Index" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "HPI") smooth))
    (series-datatype!  new-series hpi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series smooth)
    (series-require!   new-series (indicator-require-parts
				   hpi-fill (hpi-warmup-count smooth)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public hpi-datatype
  (datatype-new!
   #:key       'hpi
   #:units     'hpi
   #:input     datatype-ohlcv
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Herrick Payoff Index")
   #:proc      hpi-series
   #:parameter-list
   (list (datatype-param-new
	  #:key      'hpi-smooth
	  #:name     (_ "Smooth")
	  #:type     'float
	  #:decimals 1
	  #:minimum  1
	  #:default  5
	  #:step     1))))
