;; adding slopes leaves behind a weighted sum from the slope calculation,
;; which makes the sum higher or lower according to where the calculation
;; began





;; Integrated Linear Regression Slope (ILRS).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ilrs)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart linreg)
  #:use-module (chart misc)
  #:use-module (chart sma))


;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/0198/Abstracts_new/TILLSON/Tillson9801.html
;;     Excerpt from TASC January 1998.  Sample chart of ILRS (and EPMA and
;;     IE/2) on Hewlett Packard (symbol HPQ) from 1996/7.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/0298/TradersTips/Tips9802.html
;;     Traders Tips Feb 1998, implementation for tradestation, showing only
;;     as single slope+average.
;;
;; http://wwww.xeatrade.com/trading/2/S/1230.html
;; http://paritech.com/paritech-site/education/technical/custom/indicators/98jan2.asp
;; http://www.paritech.com/education/technical/custom/indicators/98jan2.asp
;;     Metastock implementation, cumulative from start of available data.

;; http://www.xeatrade.com/trading/14/I/740.html
;;     Metastock implementaiton, single slope+average.
;;
;; http://forum.esignalcentral.com/showthread.php?s=&threadid=12924
;;     EFS code, as SMA+LRS (linear regression slope).
;;
;; http://trader.online.pl/ELZ/t-i-IE_2.html
;; http://trader.online.pl/MSZ/e-w-Smoothing_Techniques.html
;;
;; http://www.tssupport.com/support/base/?action=article&id=1418
;;     IE/2 showing ILRS just as SMA+LRS.
;; http://www.tssupport.com/support/base/?action=article&id=1449
;;     Some description, about the same as TASC abstract above.
;; http://www.tssupport.com/support/base/?action=article&id=1771
;;     Easylanguage, as SMA+LRS.
;;


(define (linreg-slope-calc-proc count)
  (define linreg-proc (linreg-calc-proc count))
  (lambda (x)
    (receive (a b)
	(linreg-proc x)
      b)))


;; Return a procedure which calculates an integrated linear regression slope
;; (ILRS) average over an accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the ILRS up to (and including) that value, or #f if not enough values
;; yet.
;;
;;
(define-public (ilrs-calc-proc count)

  (define slope-proc (linreg-slope-calc-proc count))
  (define sma-proc   (sma-calc-proc count))
  (define points 0)
  (define sum #f)

  (lambda (x)
    (let ((slope (slope-proc x)))
      (if sum
	  (set! sum (+ sum slope))

	  (let ((sma (sma-proc x)))
	    (set! points (1+ points))
	    (if (= points count)
		(begin
		  (dv sma)
		  (set! sum sma)
		  (set! sma-proc #f)))))

      sum)))



(define-public (ilrs-calc-proc count)

  (define slope-proc (linreg-slope-calc-proc count))
  (define sma-proc   (sma-calc-proc count))

  (lambda (x)
    (let ((slope (slope-proc x))
	  (sma   (sma-proc x)))
      (and slope
	   sma
	   (+ sma slope)))))



(define-public (ilrs-warmup-count count)
  (+ 200 count))


(define (ilrs-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (ilrs-calc-proc count)))

    ;; warmup
    (for-each proc
	      (series-prev-n-closes parent #:before lo
				    #:count (ilrs-warmup-count count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))


(define-public (ilrs-series series count)
  (set! count (max 1 count))
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "ILRS") count))
    (series-datatype!  new-series ilrs-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   ilrs-fill (ilrs-warmup-count count)))
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public ilrs-datatype
  (datatype-new!
   #:key       'ilrs
   #:manual    (_ "CONTEXT(manual-node): Integrated Linear Regression Slope")
   #:proc      ilrs-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'ilrs-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 15))))
