;; Know Sure Thing (KST) indicator.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart kst)
;;  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart roc)
  #:use-module (chart sma))



;; http://www.pring.com/movieweb/daily_kst.htm
;;     Daily KST.
;; http://www.pring.com/articles/article28.htm
;;     Monthly KST.
;; http://www.pring.com/kstformulas.htm
;;
;;



;; Return a procedure which calculates a KST indicator over a window of
;; COUNT many points.
;;
;; Each call (PROC close) enters a new point into the window, and the return
;; is the KST up to (and including) that point, or #f if not enough points
;; yet.
;;
;; To prime the window initially, call PROC with COUNT many points preceding
;; the first desired.
;;
(define-public (kst-calc-proc  rc-1 mc-1  rc-2 mc-2  rc-3 mc-3  rc-4 mc-4)
  (define roc-1-proc (roc-calc-proc rc-1))
  (define roc-2-proc (roc-calc-proc rc-2))
  (define roc-3-proc (roc-calc-proc rc-3))
  (define roc-4-proc (roc-calc-proc rc-4))

  (define ma-1-proc (sma-calc-proc mc-1))
  (define ma-2-proc (sma-calc-proc mc-2))
  (define ma-3-proc (sma-calc-proc mc-3))
  (define ma-4-proc (sma-calc-proc mc-4))

  (lambda (x)

    (let ((r-1 (roc-1-proc x))
	  (r-2 (roc-2-proc x))
	  (r-3 (roc-3-proc x))
	  (r-4 (roc-4-proc x)))
      (and r-1 r-2 r-3 r-4
	   (let ((m-1 (ma-1-proc r-1))
		 (m-2 (ma-2-proc r-2))
		 (m-3 (ma-3-proc r-3))
		 (m-4 (ma-4-proc r-4)))
	     (and m-1 m-2 m-3 m-4
		  (* 0.1 (+ m-1 (* 2 m-2) (* 3 m-3) (* 4 m-4)))))))))

(define (sma-warmup-count count)
  (1- count))

(define (kst-warmup-count-one rc mc)
  (+ rc (sma-warmup-count mc)))

(define (kst-warmup-count  rc-1 mc-1  rc-2 mc-2  rc-3 mc-3  rc-4 mc-4)
  (max (kst-warmup-count-one rc-1 mc-1)
       (kst-warmup-count-one rc-2 mc-2)
       (kst-warmup-count-one rc-3 mc-3)
       (kst-warmup-count-one rc-4 mc-4)))
  

(define (kst-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (data   (series-data series))
	 (kst    (apply kst-calc-proc data)))

    (for-each kst 
	      (series-prev-n-closes parent
				    #:before lo
				    #:count  (apply kst-warmup-count data)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	(array-set! new-array (kst (array-ref array i)) i)))))

(define-public (kst-series series  rc-1 mc-1  rc-2 mc-2  rc-3 mc-3  rc-4 mc-4)

  (let ((new-series (series-copy series))
	(data       (list rc-1 mc-1  rc-2 mc-2  rc-3 mc-3  rc-4 mc-4)))
    (series-name!      new-series (format #f "~a ~a" (_ "KST") 0))
    (series-datatype!  new-series kst-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series data)
    (series-require!   new-series (indicator-require-parts
				   kst-fill (apply kst-warmup-count data)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public kst-datatype
  (datatype-new!
   #:key     'kst
   #:name    (_ "KST - Know Sure Thing")
   #:units   'roc
   #:hlines  '(0)
   #:manual  (_ "CONTEXT(manual-node): Know Sure Thing")
   #:proc    kst-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'kst-roc-1-days
	  #:name    (_ "ROC1")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)
	 (datatype-param-new
	  #:key     'kst-ma-1-days
	  #:name    (_ "MA1")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)

	 (datatype-param-new
	  #:key     'kst-roc-2-days
	  #:name    (_ "ROC2")
	  #:type    'integer
	  #:minimum 1
	  #:default 15)
	 (datatype-param-new
	  #:key     'kst-ma-2-days
	  #:name    (_ "MA2")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)

	 (datatype-param-new
	  #:key     'kst-roc-3-days
	  #:name    (_ "ROC3")
	  #:type    'integer
	  #:minimum 1
	  #:default 20)
	 (datatype-param-new
	  #:key     'kst-ma-3-days
	  #:name    (_ "MA3")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)

	 (datatype-param-new
	  #:key     'kst-roc-4-days
	  #:name    (_ "ROC4")
	  #:type    'integer
	  #:minimum 1
	  #:default 30)
	 (datatype-param-new
	  #:key     'kst-ma-4-days
	  #:name    (_ "MA4")
	  #:type    'integer
	  #:minimum 1
	  #:default 15)
	 )))
