;; MESA adaptive moving average (MAMA).

;; Copyright 2006, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mama)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart midpoint)
  #:use-module (chart misc)
  #:use-module (chart ema)
  #:use-module (chart wma))



;;-----------------------------------------------------------------------------
;; calc

(define (hilbert-transform-calc-proc)
  (define points     0)
  (define data (indicator-circular 7 '()))

  (lambda (x prev-period)

    (or (car data)
	(set! points (1+ points)))  ;; gaining a point

    ;; replace oldest with new point
    (set-car! data x)
    (set! data (cdr data))

    (and (>= points 7)
	 (* (+ (* .0962  (second data))
	       (* .5769  (fourth data))
	       (* -.5769 (sixth  data))
	       (* -.0962 (first  data))) ;; wrap around from eighth
	    (+ (* 0.075 prev-period)
	       0.54)))))

(define (homodyne-re-calc-proc)
  (define prev-i #f)
  (define prev-q #f)
  (lambda (i q)
    (let ((result (and prev-i
		       (+ (* i prev-i)
			  (* q prev-q)))))
      (set! prev-i i)
      (set! prev-q q)
      result)))

(define (homodyne-im-calc-proc)
  (define prev-i #f)
  (define prev-q #f)
  (lambda (i q)
    (let ((result (and prev-i
		       (- (* i prev-q)
			  (* q prev-i)))))
      (set! prev-i i)
      (set! prev-q q)
      result)))


(define (radians->degrees x)
  (* 57.29578 x))

(define speed 0.8)

(define-public (hilbert-period-calc-proc fast-limit slow-limit)
  (define smooth-wma-proc (wma-calc-proc 4))
  (define prev-period     0)
  (define detrender-proc  (hilbert-transform-calc-proc))
  (define Q1-proc         (hilbert-transform-calc-proc))
  (define I1-proc         (prevnth-calc-proc 3 '()))
  (define jI-proc         (hilbert-transform-calc-proc))
  (define jQ-proc         (hilbert-transform-calc-proc))
  (define I2-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define Q2-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define Re-proc         (homodyne-re-calc-proc))
  (define Im-proc         (homodyne-im-calc-proc))
  (define Re-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define Im-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))

  (lambda (x)
    (and-let* ((smooth    (smooth-wma-proc x))
	       (detrender (detrender-proc smooth prev-period))
	       (Q1        (Q1-proc detrender prev-period))
	       (I1        (I1-proc detrender))
	       (jI        (jI-proc I1 prev-period))
	       (jQ        (jQ-proc Q1 prev-period))
	       (I2        (- I1 jQ))
	       (Q2        (+ Q1 jI))
	       (I2        (I2-smooth-proc I2))
	       (Q2        (Q2-smooth-proc Q2))
	       (Re        (Re-proc I2 Q2))
	       (Im        (Im-proc I2 Q2))
	       (Re        (Re-smooth-proc Re))
	       (Im        (Im-smooth-proc Im))
	       (period    (and (not (zero? Re))
			       (not (zero? Im))
			       (/ 360 (radians->degrees (atan (/ Im Re)))))))
      (set! period (- period))
      (set! period (minmax-restrict (* .67 prev-period) (* 1.5 prev-period)
				    period))
      (set! period (minmax-restrict 6 50 period))
      (set! prev-period period)
      period)))


;; Return a procedure which calculates a stochastic RSI on successive
;; closing prices, with RSI smoothing of COUNT periods and high/low range of
;; the same COUNT periods.  For the smoothing COUNT is interpreted by
;; Wilder's reckoning, so alpha = 1/COUNT.
;;
;; Each call (PROC close) enters a new point into the window, and the return
;; is the stochastic RSI up to (and including) that value, or #f if not
;; enough points yet.
;;
;; To prime initially, call PROC with (mama-warmup-count COUNT) many
;; values, then the next call will be based on full data.
;;
(define-public (mama-alpha-calc-proc fast-limit slow-limit)
  (define prev-period     0)
  (define smooth-wma-proc (wma-calc-proc 4))
  (define detrender-proc  (hilbert-transform-calc-proc))
  (define Q1-proc         (hilbert-transform-calc-proc))
  (define I1-proc         (prevnth-calc-proc 3 '()))
  (define jI-proc         (hilbert-transform-calc-proc))
  (define jQ-proc         (hilbert-transform-calc-proc))
  (define I2-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define Q2-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define Re-proc         (homodyne-re-calc-proc))
  (define Im-proc         (homodyne-im-calc-proc))
  (define Re-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define Im-smooth-proc  (ema-calc-proc (ema-alpha->count .2)))
  (define prev-phase-proc (prevnth-calc-proc 1 '()))

  (lambda (x)
    (and-let* ((smooth    (smooth-wma-proc x))
	       prev-period
	       (detrender (detrender-proc smooth prev-period))
	       (Q1        (Q1-proc detrender prev-period))
	       (I1        (I1-proc detrender))
	       (jI        (jI-proc I1 prev-period))
	       (jQ        (jQ-proc Q1 prev-period))
	       (I2        (- I1 jQ))
	       (Q2        (+ Q1 jI))
	       (I2        (I2-smooth-proc I2))
	       (Q2        (Q2-smooth-proc Q2))
	       (Re        (Re-proc I2 Q2))
	       (Im        (Im-proc I2 Q2))
	       (Re        (Re-smooth-proc Re))
	       (Im        (Im-smooth-proc Im))
	       (period    (and (not (zero? Re))
			       (not (zero? Im))
			       (/ 360 (radians->degrees (atan (/ Im Re)))))))

      (set! period (- period))
      (set! period (minmax-restrict
		    6 50
		    (minmax-restrict (* .67 prev-period) (* 1.5 prev-period)
				     period)))
      (set! prev-period period)

      (and-let* ((phase       (and (not (zero? I1))
				   (radians->degrees (atan (/ Q1 I1)))))
		 (prev-phase  (prev-phase-proc phase)))
	(let* ((delta-phase (min 1 (- prev-phase phase))))
	  (minmax-restrict slow-limit fast-limit (/ speed delta-phase)))))))

(define-public (mama-calc-proc fast-limit slow-limit factor)
  (define alpha-proc (mama-alpha-calc-proc fast-limit slow-limit))
  (define ama-proc   (adaptive-ema-calc-proc))

  (lambda (x)
    (and-let* ((alpha (alpha-proc x)))
      (ama-proc x (* factor alpha)))))

(define-public (mama-warmup-count fast-limit slow-limit)
  (ema-warmup-count (ema-alpha->count slow-limit)))


;;-----------------------------------------------------------------------------
;; MAMA series

(define (mama-fill series new-array lo hi)
  (receive-list (fast-limit slow-limit factor)
      (series-data series)
    (let* ((parent (series-parent series))
	   (warmup (mama-warmup-count fast-limit slow-limit))
	   (start  (series-prev-n-closes-pos parent #:before lo
					     #:count warmup))
	   (array  (series-array parent start hi))
	   (mama   (mama-calc-proc fast-limit slow-limit factor)))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i lo))
	(if (array-ref array i)
	    (mama (array-ref array i))))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (mama (array-ref array i))
			i))))))

(define-public (mama-series series fast-limit slow-limit factor)
  (if (eq? datatype-ohlcv (series-datatype series))
      (set! series (midpoint-series series)))

  (let ((new-series (series-copy series)))
    (series-name! new-series
		  (if (= factor 1)
		      (format #f (_ "MAMA ~a, ~a")
			      (locale-number-min-decimals fast-limit 1)
			      (locale-number-min-decimals slow-limit 1))
		      (format #f (_ "MAMA ~a, ~a, ~a")
			      (locale-number-min-decimals fast-limit 1)
			      (locale-number-min-decimals slow-limit 1)
			      (locale-number-min-decimals factor 1))))
    (series-datatype! new-series mama-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list fast-limit slow-limit factor))
    (series-require!  new-series (indicator-require-parts
				  mama-fill
				  (mama-warmup-count fast-limit slow-limit)))
    (series-array!    new-series #f)

    new-series))

(define-public mama-datatype
  (datatype-new!
   #:key     'mama
   #:manual  (_ "CONTEXT(manual-node): MESA Adaptive Moving Average")
   #:proc    mama-series
   #:parameter-list
   (list (datatype-param-new
	  #:key      'mama-fast-limit
	  #:name     (_ "Fast limit")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.5)
	 (datatype-param-new
	  #:key      'mama-slow-limit
	  #:name     (_ "Slow limit")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.05)

	 (datatype-param-new
	  #:key      'mama-factor
	  #:name     (_ "Factor")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.5))))



;;-----------------------------------------------------------------------------
;; MAMA alpha series

(define (mama-alpha-fill series new-array lo hi)
  (receive-list (fast-limit slow-limit)
      (series-data series)
    (let* ((parent  (series-parent series))
	   (warmup  (mama-warmup-count fast-limit slow-limit))
	   (start   (series-prev-n-closes-pos parent #:before lo
					      #:count warmup))
	   (array   (series-array parent start hi))
	   (proc    (mama-alpha-calc-proc fast-limit slow-limit)))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i lo))
	(if (array-ref array i)
	    (proc (array-ref array i))))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i)) i))))))

(define-public (mama-alpha-series series fast-limit slow-limit)
  (if (eq? datatype-ohlcv (series-datatype series))
      (set! series (midpoint-series series)))

  (let ((new-series (series-copy series)))
    (series-name! new-series
		  (format #f (_ "MAMA alpha ~a, ~a")
			  (locale-number-min-decimals fast-limit 1)
			  (locale-number-min-decimals slow-limit 1)))
    (series-datatype!  new-series mama-alpha-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list fast-limit slow-limit))
    (series-require!   new-series (indicator-require-parts
				   mama-alpha-fill
				   (mama-warmup-count fast-limit slow-limit)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())

    new-series))

(define-public mama-alpha-datatype
  (datatype-new!
   #:key     'mama-alpha
   #:manual  (_ "CONTEXT(manual-node): MESA Adaptive Moving Average")
   #:proc    mama-alpha-series
   #:minimum 0
   #:maximum 1
   #:units   'ema-alpha
   #:parameter-list
   (list (datatype-param-new
	  #:key      'mama-fast-limit
	  #:name     (_ "Fast limit")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.5)
	 (datatype-param-new
	  #:key      'mama-slow-limit
	  #:name     (_ "Slow limit")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.05))))



;;-----------------------------------------------------------------------------
;; MAMA hilbert period series

(define (hilbert-period-fill series new-array lo hi)
  (receive-list (fast-limit slow-limit)
      (series-data series)
    (let* ((parent  (series-parent series))
	   (warmup  (mama-warmup-count fast-limit slow-limit))
	   (start   (series-prev-n-closes-pos parent #:before lo
					      #:count warmup))
	   (array   (series-array parent start hi))
	   (proc (hilbert-period-calc-proc fast-limit slow-limit)))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i lo))
	(if (array-ref array i)
	    (proc (array-ref array i))))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i)) i))))))

(define-public (hilbert-period-series series fast-limit slow-limit)
  (if (eq? datatype-ohlcv (series-datatype series))
      (set! series (midpoint-series series)))

  (let ((new-series (series-copy series)))
    (series-name! new-series
		  (format #f (_ "Hilb Period ~a, ~a")
			  (locale-number-min-decimals fast-limit 1)
			  (locale-number-min-decimals slow-limit 1)))
    (series-datatype!  new-series hilbert-period-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list fast-limit slow-limit))
    (series-require!   new-series (indicator-require-parts
				   hilbert-period-fill
				   (mama-warmup-count fast-limit slow-limit)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())

    new-series))

(define-public hilbert-period-datatype
  (datatype-new!
   #:key     'mama-hilbert-period
   #:manual  (_ "CONTEXT(manual-node): MESA Adaptive Moving Average")
   #:proc    hilbert-period-series
   #:parameter-list
   (list (datatype-param-new
	  #:key      'mama-fast-limit
	  #:name     (_ "Fast limit")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.5)
	 (datatype-param-new
	  #:key      'mama-slow-limit
	  #:name     (_ "Slow limit")
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  0.05))))


