;; Squelch.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart squelch)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart midpoint)
  #:use-module (chart misc)
  #:use-module (chart momentum)
  #:use-module (chart ema))


;; http://trader.online.pl/ELZ/t-i-Hilbert_Transform.html
;;     Sample chart Hilbert inphase and quadrature on S&P 500 (^GSPC) from
;;     2001.   Inphase looks close, quadrature doesn't.  Maybe unsmoothed?
;;


;;-----------------------------------------------------------------------------
;; calc


;; Value3 =.75*(Value1 - Value1[6]) + .25*(Value1[2] - Value1[4]);
;;
(define (v3-calc-proc)

  ;; data   v[6]  first
  ;;        v[5]  second
  ;;        v[4]  third
  ;;        v[3]  fourth
  ;;        v[2]  fifth
  ;;        v[1]  sixth
  ;;
  (define points     0)
  (define data (make-circular-list 6 #f))

  (lambda (x)
    (let ((ret (if (< points 6)
		   (begin
		     (set! points (1+ points))
		     #f)
		   (+ (* 0.75 x)                 ;; value1[0]
		      (* -0.75 (first data))     ;; value1[6]
		      (* 0.25 (fifth data))      ;; value1[2]
		      (* -0.25 (third data)))))) ;; value1[4]
      (set-car! data x)
      (set! data (cdr data))
      ret)))


(define (radians->degrees x)
  (* 57.29578 x))

(define (take-maybe! lst n)
  (if (> (length lst) n)
      (take! lst n)
      lst))

(define (period-calc-proc)
  (let ((lst '()))
    (lambda (delta-phase)
      (set! lst (cons delta-phase lst))
      (set! lst (take-maybe! lst 100))

      (let ((total 0))
	(let more ((count 0)
		   (lst   lst))
	  (if (null? lst)
	      #f
	      (begin
		(set! total (+ total (first lst)))
		(if (> total 360)
		    count
		    (more (1+ count) (cdr lst))))))))))
		   
    

(define (squelch-period-calc-proc)
  (let ((v1-proc              (momentum-calc-proc 6))
	(v2-proc              (prevnth-calc-proc 3))
	(v3-proc              (v3-calc-proc))
	(inphase-proc         (ema-calc-proc (ema-alpha->count .33)))
	(inphase-prev-proc    (prevnth-calc-proc 1))
	(quadrature-proc      (ema-calc-proc (ema-alpha->count .2)))
	(quadrature-prev-proc (prevnth-calc-proc 1))
	(phase-prev-proc      (prevnth-calc-proc 1))
	(period-proc          (period-calc-proc))
	(period-smooth-proc   (ema-calc-proc (ema-alpha->count .25))))

    (lambda (x)
      (and-let* ((v1 (v1-proc x)))
	(let* ((v2              (v2-proc v1))
	       (v3              (v3-proc v1))
	       (inphase         (and v2 (inphase-proc v2)))
	       (inphase-prev    (and inphase (inphase-prev-proc inphase)))
	       (quadrature      (and v3 (quadrature-proc v3)))
	       (quadrature-prev (and quadrature
				     (quadrature-prev-proc quadrature))))
	  (and quadrature-prev
	       (let ((phase (if (positive? (+ inphase inphase-prev))
				(radians->degrees
				 (atan (/ (abs (+ quadrature quadrature-prev))
					  (+ inphase inphase-prev))))
				0)))
		 (if (negative? inphase)
		     (if (positive? quadrature)
			 (set! phase (- 180 phase))
			 (set! phase (+ 180 phase)))
		     (if (negative? quadrature)
			 (set! phase (- 360 phase))))
		 (and-let* ((phase-prev (phase-prev-proc phase))
			    (delta-phase
			     (max 1
				  (min 60
				       (if (and (< phase 90)
						(> phase-prev 270))
					   (- phase-prev phase -360)
					   (- phase-prev phase)))))
			    (period (period-proc delta-phase)))
		   (period-smooth-proc period)))))))))


(define squelch-warmup-count 100)


;;-----------------------------------------------------------------------------
;; squelch period series

(define (squelch-period-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (warmup squelch-warmup-count)
	 (start  (series-prev-n-closes-pos parent #:before lo
					   #:count warmup))
	 (array  (series-array parent start hi))
	 (proc   (squelch-period-calc-proc)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i)
	  (proc (array-ref array i))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (squelch-period-series series)
  (if (eq? datatype-ohlcv (series-datatype series))
      (set! series (midpoint-series series)))

  (let ((new-series (series-copy series)))
    (series-name! new-series (_ "Squelch Period"))
    (series-datatype!  new-series squelch-period-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   squelch-period-fill
				   squelch-warmup-count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())

    new-series))

(define-public squelch-period-datatype
  (datatype-new!
   #:key     'squelch-period
   #:manual  (_ "CONTEXT(manual-node): Squelch")
   #:proc    squelch-period-series))












;;-----------------------------------------------------------------------------
;; inphase series

(define (inphase-calc-proc)
  (let ((v1-proc      (momentum-calc-proc 6))
	(v2-proc      (prevnth-calc-proc 3))
	(inphase-proc (ema-calc-proc (ema-alpha->count .33))))
    (lambda (x)
      (and-let* ((v1 (v1-proc x))
		 (v2 (v2-proc v1)))
	(inphase-proc v2)))))

(define inphase-warmup-count
  (+ 5 1 3 1 (ema-warmup-count (ema-alpha->count .33))))

(define (inphase-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (warmup squelch-warmup-count)
	 (start  (series-prev-n-closes-pos parent #:before lo
					   #:count warmup))
	 (array  (series-array parent start hi))
	 (proc   (inphase-calc-proc)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i)
	  (proc (array-ref array i))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (inphase-series series)
  (if (eq? datatype-ohlcv (series-datatype series))
      (set! series (midpoint-series series)))

  (let ((new-series (series-copy series)))
    (series-name! new-series (_ "Squelch Inphase"))
    (series-datatype!  new-series inphase-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   inphase-fill inphase-warmup-count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())

    new-series))

(define-public inphase-datatype
  (datatype-new!
   #:key     'squelch-inphase
   #:manual  (_ "CONTEXT(manual-node): Squelch")
   #:proc    inphase-series))





;;-----------------------------------------------------------------------------
;; quadrature series

(define (quadrature-calc-proc)
  (let ((v1-proc         (momentum-calc-proc 6))
	(v3-proc         (v3-calc-proc))
	(quadrature-proc (ema-calc-proc (ema-alpha->count 1))))

    (lambda (x)
      (and-let* ((v1 (v1-proc x))
		 (v3 (v3-proc v1)))
	(quadrature-proc v3)))))

(define quadrature-warmup-count
  (+ 5 1 6 1 (ema-warmup-count (ema-alpha->count 1))))

(define (quadrature-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (warmup squelch-warmup-count)
	 (start  (series-prev-n-closes-pos parent #:before lo
					   #:count warmup))
	 (array  (series-array parent start hi))
	 (proc   (quadrature-calc-proc)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i)
	  (proc (array-ref array i))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (quadrature-series series)
  (if (eq? datatype-ohlcv (series-datatype series))
      (set! series (midpoint-series series)))

  (let ((new-series (series-copy series)))
    (series-name! new-series (_ "Squelch Quadrature"))
    (series-datatype!  new-series quadrature-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   quadrature-fill quadrature-warmup-count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())

    new-series))

(define-public quadrature-datatype
  (datatype-new!
   #:key     'squelch-quadrature
   #:manual  (_ "CONTEXT(manual-node): Squelch")
   #:proc    quadrature-series))








;; (let ((proc (v3-calc-proc)))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 1))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0))
;;   (dv (proc 0)))
;; (exit 0)
