;; Trend Detection Index (TDI).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tdi)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart momentum)
  #:use-module (chart misc))


;; http://www.linnsoft.com/tour/techind/tdi.htm
;;     Sample MSFT from Oct 2001, but peak seems too high, or on a different
;;     scale, or something.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/062002/TradersTips/TradersTips.html
;;     TASC Trader's Tips October 2001, easylanguage code by M.H. Pee.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/112001/TradersTips/TradersTips.html
;;     TASC Trader's Tips November 2001, metastock code from Equis.
;;


;; Return a procedure which calculates the trend detection index over an
;; accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is two values (values TDI DI) the trend direction index and the direction
;; index up to (and including) that point, or #f if not enough values yet.
;;
;; The window can be primed by calling with 4*COUNT-1 many values preceding
;; the first desired.  `tdi-warmup-count' below can make that calculation.
;;
(define-public (tdi-calc-proc count)
  (let ((mom-proc  (momentum-calc-proc count))
	(td-proc   (sum-calc-proc count))
	(sum1-proc (sum-calc-proc count))
	(sum2-proc (sum-calc-proc (* 2 count))))

    (lambda (x)
      (let* ((mom  (mom-proc x))
	     (td   (and mom (td-proc mom)))

	     (ab   (and mom (abs mom)))
	     (sum1 (and ab (sum1-proc ab)))
	     (sum2 (and ab (sum2-proc ab))))

	'(and sum2
	     (pk (abs td) (- sum2 sum1)))
	(values (and sum2
		     (- (abs td) (- sum2 sum1))
		     ;; (+ (abs td) sum1 (- sum2))
		     )
		td)))))


(define (tdi-warmup-count count)
  (1- (* 4 count)))

(define (tdi-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (warmup (tdi-warmup-count count))
	 (array  (series-array-closes parent lo hi))
	 (proc   (tdi-calc-proc count)))

    ;; warmup
    (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (receive (tdi di)
	      (proc (array-ref array i))
	    (if tdi
		(begin
		  (array-set! new-array tdi i 0)
		  (array-set! new-array di  i 1))))))))

(define-public (tdi-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "TDI") count))
    (series-datatype!  new-series tdi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   tdi-fill (tdi-warmup-count count)
				   '(0 1)))
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public tdi-datatype
  (datatype-new!
   #:key       'tdi
   #:linestyle 'dmi
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Trend Detection Index")
   #:proc      tdi-series
   #:closes-apply-arg '(0) ;; TDI
   #:parameter-list
   (list (datatype-param-new
	  #:key     'tdi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
