;; Volatility SAR.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart volatility-sar)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart atr)
  #:use-module (chart misc)
  #:use-module (chart wma))


;; http://www.biz-analyst.com/library/stocks_commodities_trend.pdf
;;     Tushar Chande, "Rating Trend Strength", TASC V.16:9, September 1998,
;;     reproduced in a collection of trend related TASC articles.
;;     (Apparently http://store.traders.com/trarpap.html)
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/0998/Abstracts_new/Abraham/Abraham9809.html
;; http://store.traders.com/-v16-c09-069trnd-pdf.html
;;     Start of article, rest for sale.
;;
;; http://www.traders.com/Documentation/Feedbk_docs/Archive/1198/Letters/Letters9811.html
;;     TASC letters Nov 1998, clarifying the WMA weighted moving average is
;;     intended.
;;
;; http://store.traders.com/v175letters.html
;;     TASC letters May 1999, further clarifying calculation, and origin
;;     from Wilder, with multiple 2.8 to 3.1.
;;
;; http://trader.online.pl/MSZ/e-w-Trend_Trailing_Indicator.html
;;     Formula, sample on WIG20.
;;
;; http://www.traderclub.com/toolkit.htm#chandelier
;;


(define (atrwma-calc-proc count)
  (define tr-proc  (true-range-calc-proc))
  (define wma-proc (wma-calc-proc count))
  (lambda (high low close)
    (and-let* ((tr (tr-proc high low close)))
      (wma-proc tr))))

;; This does the whole series from start to end.
;;
(define (volatility-sar-fill series lo hi)

  (set! lo (series-lo series))
  (set! hi (series-hi series))

  (let* ((parent    (series-parent series))
	 (multiple  (series-data series))
	 (array     (series-array parent lo hi))
	 (new-array (make-array #f (list lo hi)))
	 (atr-proc  (atrwma-calc-proc 21))
	 (extreme   #f)
	 (sar       #f)
	 (proc      #f))

    (define (initial high low close)
      (atr-proc high low close)
      (set! extreme   close)
      (set! proc      rising))

    (define (rising high low close)
      (set! extreme (max extreme close))

      (and-let* ((atr (atr-proc high low close)))
	(set! sar (- extreme (* multiple atr)))

	;; if SAR penetrated, reverse for next day
	(if (<= close sar)
	    (begin
	      (set! extreme close)
	      (set! proc    falling)))))

    (define (falling high low close)
      (set! extreme (min extreme close))

      (and-let* ((atr (atr-proc high low close)))
	(set! sar (+ extreme (* multiple atr)))

	;; if SAR penetrated, reverse for next day
	(if (>= close sar)
	    (begin
	      (set! extreme close)
	      (set! proc    rising)))))

    (set! proc initial)

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((>= i hi))
      (and-let* ((close (array-ref array i 3)))
	(proc (or (array-ref array i 1) close)   ;; high
	      (or (array-ref array i 2) close)   ;; low
	      close))
      (array-set! new-array sar i))

    ;; prospective stop one beyond parent series
    (array-set! new-array sar hi)

    new-array))

(define-public (volatility-sar-series series multiple)
  (ohlcv-required "Volatility SAR" series)
  (let ((new-series (series-copy series)))
    (series-name! new-series (format #f "~a ~a" (_ "Volatility SAR")
				     (locale-number-min-decimals multiple 1)))
    (series-datatype! new-series volatility-sar-datatype)
    (series-parent!   new-series series)
    (series-hi!       new-series (1+ (series-hi series)))
    (series-data!     new-series multiple)
    (series-require!  new-series volatility-sar-fill)
    (series-array!    new-series #f)
    new-series))

(define volatility-sar-datatype
  (datatype-new!
   #:key       'volatility-sar
   #:name      (_ "Volatility SAR")
   #:units     #f
   #:input     datatype-ohlcv
   #:linestyle 'stops
   #:manual    (_ "CONTEXT(manual-node): Volatility SAR")
   #:proc      volatility-sar-series
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "ATR multiple")
	  #:key      'volatility-sar-multiple
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:default  3
	  #:step     0.1))))
