;; Intraday dialogs.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart intraday)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (gtk gtk)
  #:use-module (gtk-2.0 gdk-pixbuf)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart extra-gtk)
  #:use-module (chart gui-common)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


;; Intraday graphs are stored on disk as follows.
;;
;; Each SYMBOL + MODE combination results in two files, "FOO-1_Day.data" is
;; the graphics file (PNG, GIF, whatever), and "FOO-1_Day.info" contains an
;; alist like
;;
;;     ((url           . "url str")
;;      (etag          . "date str")
;;      (last-modified . "tag str"))
;;
;; which is used, where possible, to put If-None-Match and If-Modified-Since
;; headers when re-downloading.  The SFE (www.sfe.com.au) for example
;; provides ETag and Last-Modified on its year graph.
;;
;; Some thought was given to using .png, .gif or whatever extension instead
;; of always .data, but the pixbuf loader is smart enough to work out the
;; format and having one name to go to is easier than say having to look in
;; the .info first
;;


;; eg. "/home/username/Chart/cache/intraday"
(define intraday-directory 
  (g-build-filename cache-directory "intraday"))

;; return the base name for SYMBOL and MODE, eg. "^GSPC-1_Day"
(define (intraday-basename symbol mode)
  (set! mode (string-delete mode #\_)) ;; drop accelerator markers
  (string-append symbol
		 "-"
		 (string-map (lambda (c)
			       (case c
				 ((#\space #\/) #\_)
				 (else          c)))
			     mode)))

;; return filename for graph data file of SYMBOL + MODE, or #f
;; eg. "/home/username/Chart/cache/intraday/^GSPC-1_Day.data"
(define (intraday-filename symbol mode)
  (and mode
       (g-build-filename intraday-directory
			 (string-append (intraday-basename symbol mode)
					".data"))))

;; return filename for info file of SYMBOL + MODE
;; eg. "/home/username/Chart/cache/intraday/^GSPC-1_Day.info"
(define (intraday-info-filename symbol mode)
  (g-build-filename intraday-directory
		    (string-append (intraday-basename symbol mode)
				   ".info")))

;; write DATA to intraday file SYMBOL + MODE
;; FIXME: probably need O_BINARY here on dos
;;
(define (intraday-store symbol mode url headers data)
  (let* ((data-filename (intraday-filename symbol mode))
	 (info-filename (intraday-info-filename symbol mode))
	 (alist         `((url           . ,url)
			  (etag          . ,(http-headers-get headers "ETag"))
			  (last-modified . ,(http-headers-get
					     headers "Last-Modified")))))

    (database-call-with-lock intraday-directory
      (lambda ()
	(database-call-with-output-file info-filename
	  (lambda (port)
	    (display ";; Intraday graph info, generated by Chart\n" port)
	    (write-list-lines alist port)))

	(database-call-with-output-file data-filename
	  (lambda (port)
	    (display data port)))))))

;; return a list like (#:if-none-match "x" #:if-modified-since #f) of
;; arguments for `http-request', possibly empty '()
;;
(define (intraday-tag-args symbol mode url)
  (let ((alist (call-with-input-file-or
		   '() (intraday-info-filename symbol mode) read)))
    (if (equal? url (assoc-ref alist 'url))
	(list #:if-none-match     (assoc-ref alist 'etag)
	      #:if-modified-since (assoc-ref alist 'last-modified))
	'())))
    

;; return url string obtained from intraday handler for SYMBOL+MODE
;; will be #f if not supported for this SYMBOL
(define-public (intraday-url symbol mode)
  (symbol-setups symbol)
  (any (lambda (elem)
	 (receive-list (symbol-pred url-proc modes-proc)
	     elem
	   (and (symbol-pred symbol)
		(url-proc symbol mode))))
       intraday-handler-list))

;; return list of mode strings, empty if no intraday for symbol
(define-public (intraday-modes symbol)
  (symbol-setups symbol)
  (or (any (lambda (elem)
	     (receive-list (symbol-pred url-proc modes-proc)
		 elem
	       (and (symbol-pred symbol)
		    (modes-proc))))
	   intraday-handler-list)
      '()))


;;-----------------------------------------------------------------------------
;; refresh throttling
;;
;; If an intraday window is closed and then re-opened we don't want to
;; re-download.  The rule established here is that if SYMBOL+MODE was done
;; within the last 2 minutes then don't do it again.  (This doesn't affect
;; the Refresh dialog button, which always immediately attempts.)
;;

;; list ((SYMBOL MODE) . TIME_T)
(define refresh-time-alist '())

(define (refresh-record-now symbol mode)
  (set! refresh-time-alist
	(assoc-set! refresh-time-alist
		    (list symbol mode) (current-time))))

;; remove any old entries from `refresh-time-alist' (more than 2 minutes)
(define (refresh-purge)
  (let ((limit (- (current-time) #,(hms->seconds 0 2 0))))
    (set! refresh-time-alist
	  (remove! (lambda (pair)
		     (< (cdr pair) limit))
		   refresh-time-alist))))

;; return true if we should re-download the graph for SYMBOL+MODE
(define (refresh-symbol? symbol mode)
  (refresh-purge)
  (not (assoc (list symbol mode) refresh-time-alist)))


;;-----------------------------------------------------------------------------

;; list of dialog widgets
(define intraday-dialog-list '())

;; open an intraday window for SYMBOL
;; if SYMBOL is in some intraday, then just raise that one
(define-public (intraday-popup symbol)
  (cond ((find (lambda (dialog)
		 (equal? symbol (object-property dialog 'symbol)))
	       intraday-dialog-list)
	 => c-gtk-widget-show-or-raise)
	(else
	 (intraday-new symbol))))

;; return true if `gdk-pixbuf-loader-write' works with NUL bytes
;; in guile-gtk 0.5 and earlier it didn't work
;;
(define c-working-gdk-pixbuf-loader-write
  (make-procedure-with-cache-0
   (lambda ()
     (false-if-exception
      (begin
	(gdk-pixbuf-loader-write (gdk-pixbuf-loader-new)
				 (string (integer->char 0))
				 1)
	#t)))))

;; FILENAME is a graphics file, but possibly incomplete
;; read it and return a pixbuf of as much as possible, or #f if FILENAME
;; doesn't exist or doesn't have enough data to show anything
;;
(define (c-gdk-pixbuf-new-from-partial-file filename)
  (if (c-working-gdk-pixbuf-loader-write)
      (and-let* ((str    (call-with-input-file-or #f filename
			   read-string-to-eof))
		 (loader (gdk-pixbuf-loader-new)))
	(catch 'g-error
	  (lambda ()
	    (and (gdk-pixbuf-loader-write loader str (string-length str))
		 (gdk-pixbuf-loader-close loader)
		 (gdk-pixbuf-loader-get-pixbuf loader)))
	  (lambda args ;; scm-error (key sbur msg msgargs data)
	    #f)))

      (gdk-pixbuf-new-from-file filename)))

;; SYMBOL can be #f to open with nothing showing
(define (intraday-new symbol)

  (define dialog      (gtk-dialog-new-with-buttons "" #f '()))
  (define drawing     (gtk-drawing-area-new))
  (define title-label (gtk-label-new ""))
  (define label       (gtk-label-new ""))
  (define entry       (gtk-entry-new))
  (define menu        (gtk-menu-new))
  (define option      (gtk-option-menu-new))
  (define accel-group (gtk-accel-group-new))
  (define pixbuf      #f)
  (define mode        #f)

  (define (fill)
    (let ((filename (and symbol
			 (intraday-filename symbol mode))))
      (set! pixbuf (and filename
			(c-gdk-pixbuf-new-from-partial-file filename)))
      (gtk-label-set-text
       title-label
       (let ((st (and pixbuf (false-if-exception (stat filename)))))
	 (if st
	     (string-append symbol " - "
			    (apply form-adate-time symbol
				   (tm->adate-time
				    (localtime (stat:mtime st)
					       (symbol-timezone symbol)))))
	     "")))

      (if pixbuf
	  (gtk-widget-set-usize drawing
				(gdk-pixbuf-get-width pixbuf)
				(gdk-pixbuf-get-height pixbuf)))
      (c-gtk-widget-queue-redraw drawing)))

  ;; display STR as the downloading status, STR can be empty
  (define (status-downloading str)
    (gtk-label-set-text label (string-append  (_ "Downloading ...") " " str)))

  (define (refresh)
    (if symbol
	(begin
	  (refresh-record-now symbol mode)
	  (status-downloading "")
	  (fill)
	  (c-main-enq!
	   (let ((symbol symbol) ;; in case change while in progress
		 (mode   mode))
	     (lambda ()
	       (c-lazy-catch #t
		 (lambda ()
		   ;; `intraday-url' inside the catch in case it has to
		   ;; query the data source for the image url, eg. barchart
		   (let ((url (intraday-url symbol mode)))
		     (if (not url)
			 (gtk-label-set-text
			  label (_ "Intraday not available for this symbol"))

			 (let ((contents ""))
			   (define (data-callback str)
			     ;; could just append to the file, but that
			     ;; would be wrong if two processes are
			     ;; downloading at the same time
			     (set! contents (string-append contents str))
			     (intraday-store symbol mode "" "" contents)
			     (fill))
			   (receive (headers body)
			       (with-fluids ((http-status-callback
					      status-downloading))
				 (apply http-request url
					#:want-ok       #t
					#:data-callback data-callback
					(intraday-tag-args symbol mode url)))
			     (gtk-label-set-text label "")
			     (if (http-headers-ok? headers)
				 (begin
				   ;; if some new data
				   (intraday-store symbol mode
						   url headers body)
				   (fill))))))))

		 (lambda args
		   (gtk-label-set-text label (error-message args))
		   (fill)
		   ;; continue on connection or http protocol problems
		   ;; (including http timeout), throw full error otherwise
		   (if (not (or (eq? 'http (first args))
				(gethost-error-try-again? args)
				(system-error-econn? args)))
		       (apply throw args))))))))))

  ;; initiate a `refresh' download, but only if the last download was more
  ;; than 2 minutes ago
  (define (refresh-maybe)
    (if (refresh-symbol? symbol mode)
	(refresh)))

  ;; drawing from a pixbuf like this probably sends a big chunk of data to
  ;; the server every time, it'd save X traffic to draw once into a
  ;; background pixmap, but guile-gtk 0.41 doesn't handle pixmap
  ;; destruction properly (prints warnings and doesn't free the X object)
  (define (draw)
    (cond (pixbuf
	   (gdk-pixbuf-render-to-drawable
	    pixbuf (gtk-widget-window drawing)
	    (gtk-style-fg-gc (gtk-widget-get-style drawing) 'normal)
	    0 0 0 0
	    (gdk-pixbuf-get-width pixbuf)
	    (gdk-pixbuf-get-height pixbuf)
	    'normal 0 0))
	  ((not (string-prefix? (_ "Downloading ...") (gtk-label-get label)))
	   ;; don't show "no data" while downloading, to avoid confusion
	   (c-gtk-draw-string-centred drawing (_ "No data")))))

  (define (fill-for-symbol)
    (gtk-window-set-title dialog (c-make-window-title
				  (string-append (or symbol "")
						 (if symbol " " "")
						 (_ "Intraday"))))

    ;; remove any previous mode options
    ;; note: must destroy to get accelerators removed
    (for-each (lambda (child)
		(gtk-container-remove menu child)
		(gtk-widget-destroy child))
	      (gtk-container-children menu))

    (if symbol
	(let ((mode-list (intraday-modes symbol)))
	  (for-each (lambda (m num)
		      (let* ((item (c-gtk-menu-item-new-with-mnemonic
				    option accel-group m)))
			(gtk-widget-show-all item)
			(gtk-menu-append menu item)
			(gtk-signal-connect item "activate"
			  (lambda ()
			    (if (not (string=? m mode))
				(begin
				  (set! mode m)
				  (fill)
				  (refresh-maybe)))))))
		    mode-list (iota (length mode-list)))

	  (if (not (null? mode-list))
	      (set! mode (first mode-list)))
	  (gtk-option-menu-set-history option 0))))

  (define (entry-goto)
    (set! symbol (gtk-entry-get-text entry))
    (set-object-property! dialog 'symbol symbol)
    (if (string-null? symbol)
	(set! symbol #f))
    (fill-for-symbol)
    (fill)
    (refresh-maybe))

  (gtk-window-add-accel-group dialog accel-group)
  (gtk-box-pack-start (gtk-dialog-vbox dialog) title-label #f #f 0)

  (gtk-widget-set-usize drawing 512 288)
  (fill)
  (c-gtk-widget-drawing drawing draw)
  (gtk-box-pack-start (gtk-dialog-vbox dialog) drawing #t #t 0)

  (gtk-box-pack-start (gtk-dialog-vbox dialog) label #f #f 0)

  (let ((hbox   (gtk-hbox-new #f 0))
	(button (gtk-button-new-with-label (_ "Enter"))))
    (gtk-box-pack-start hbox (gtk-label-new (_ "Symbol")) #f #f 0)

    (if symbol
	(gtk-entry-set-text entry symbol))
    (gtk-box-pack-start hbox entry #t #t
			(c-gtk-widget-units->width hbox '(em . 0.5)))
    (gtk-signal-connect entry "activate" entry-goto)
    '(gtk-signal-connect entry "changed"
       (lambda ()
	 #f))

    (gtk-box-pack-start hbox button #f #f 0)
    (gtk-signal-connect button "clicked" entry-goto)

    (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0))

  (gtk-option-menu-set-menu option menu)
  (gtk-container-add (gtk-dialog-action-area dialog) option)

  (gtk-dialog-add-buttons dialog
			  "gtk-refresh" 0
			  "gtk-close"   gtk-response-close
			  "gtk-help"    gtk-response-help)

  (gtk-widget-grab-focus entry)

  (gtk-signal-connect dialog "response"
    (lambda (id)
      (cond ((= id 0) ;; refresh
	     (refresh))

	    ((or (= id gtk-response-close)
		 (= id gtk-response-delete-event))
	     ;; remove from list of existing dialogs
	     (set! intraday-dialog-list
		   (delete dialog intraday-dialog-list))
	     (gtk-widget-destroy dialog))

	    ((= id gtk-response-help)
	     (manual-popup (_ "CONTEXT(manual-node): Intraday") dialog)))))

  (set-object-property! dialog 'symbol symbol)
  (fill-for-symbol)
  (refresh)
  (gtk-widget-show-all dialog)

  ;; add to list of existing dialogs
  (set! intraday-dialog-list (cons dialog intraday-dialog-list)))
