;; Kaufman Adaptive Moving Average (KAMA).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart kama)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/0398/TradersTips/Tips9803.html
;;    Traders tips March 1998: metastock and easylanguage code
;;
;; http://www.perrykaufman.com
;;    Perry Kaufman's web site (no formulas).
;; 
;; http://www.tssupport.com/support/base/?action=article&id=1769
;;    Short tradestation article, incl EFS code.
;;


;;-----------------------------------------------------------------------------
;; calc

(define fastest (ema-count->alpha 2))
(define slowest (ema-count->alpha 30))
(define fastest-slowest (- fastest slowest))


;; Return a procedure which calculates alpha factors for a Kaufman adaptive
;; moving average on an accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the KAMA up to (and including) that value, or #f if not enough values
;; yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (kama-alpha-calc-proc count)

  (define prev1-proc   (prevnth-calc-proc 1))
  (define prevnth-proc (prevnth-calc-proc count))
  (define sum-proc     (sum-calc-proc count))

  (lambda (x)
    (let* ((prev       (prev1-proc x))
	   (total-move (and prev
			    (sum-proc (abs (- x prev)))))
	   (prevnth    (prevnth-proc x)))

      (and prevnth
	   (let* ((net-move (abs (- x prevnth)))
		  (efratio  (if (zero? total-move)  ;; would be 0/0
				1
				(/ net-move total-move))))
	     (square (+ slowest (* efratio fastest-slowest))))))))


;; Return a procedure which calculates a Kaufman adaptive moving average,
;; with alpha factors calculated on an accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the KAMA up to (and including) that value, or #f if not enough values
;; yet.
;;
(define-public (kama-calc-proc count)

  (define kama-alpha-proc (kama-alpha-calc-proc count))
  (define ama-proc        (adaptive-ema-calc-proc))

  (lambda (x)
    (and-let* ((alpha (kama-alpha-proc x)))
      (ama-proc x alpha))))


;; guessed warmup count, based on an assumed efficiency ratio of 0.25
(define guess-warmup
  (ema-warmup-count (ema-alpha->count
		     (square (+ slowest (* 0.25 fastest-slowest))))))

(define (kama-estimate-warmup count)
  (+ (1- count) guess-warmup))


;;-----------------------------------------------------------------------------
;; KAMA alpha series

(define (kama-alpha-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (kama   (kama-alpha-calc-proc count)))

    ;; warmup
    (for-each kama (series-prev-n-closes parent
					 #:before lo #:count (1- count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (kama (array-ref array i)) i)))))

(define-public (kama-alpha-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "KAMA alpha") count))
    (series-datatype!  new-series kama-alpha-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts kama-alpha-fill
							   (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public kama-alpha-datatype
  (datatype-new!
   #:key      'kama-alpha
   #:name     (_ "KAMA alpha")
   #:priority -10
   #:units    'ema-alpha
   #:minimum  (square slowest)
   #:maximum  (square fastest)
   #:manual   (_ "CONTEXT(manual-node): Kaufman Adaptive Moving Average")
   #:proc     kama-alpha-series
   #:parameter-list
   (list     (datatype-param-new
	      #:name    (_ "Days")
	      #:key     'kama-days
	      #:type    'integer
	      #:minimum 1
	      #:default 10))))


;;-----------------------------------------------------------------------------
;; KAMA series

(define (kama-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (adaptive-ema-find-start parent lo
					  (kama-alpha-calc-proc count)))
	 (array  (series-array-closes parent start hi))
	 (kama   (kama-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i)
	  (kama (array-ref array i))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (kama (array-ref array i)) i)))))

(define-public (kama-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "KAMA") count))
    (series-datatype! new-series kama-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts
				  kama-fill (kama-estimate-warmup count)))
    (series-array!    new-series #f)

    new-series))

(define kama-datatype
  (datatype-new!
   #:key      'kama
   #:name     (_ "KAMA - Kaufmann Adaptive MA")
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Kaufman Adaptive Moving Average")
   #:proc     kama-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'kama-days
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
