;; Kansas City Board of Trade (KCBT) setups.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart kcbt)
  #:autoload   (ice-9 format)    (format)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart barchart-customer)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (kcbt-symbol? symbol)
  (string-suffix? ".KCBT" symbol))

;; timezone-chicago representing US central
(symbol-timezone! kcbt-symbol? (_ "Kansas City") timezone-chicago)
(source-help! kcbt-symbol?
	      (_ "CONTEXT(manual-node): Kansas City Board of Trade"))


;;-----------------------------------------------------------------------------
;; weblink - contract specs

(define kcbt-weblinks-list
  '(("KW" . "http://www.kcbt.com/contract_wheat.html")
    ("KE" . "http://www.kcbt.com/symbols_trading_hours.html")
    ("MV" . "http://www.kcbt.com/symbols_trading_hours.html")))

(weblink-handler! kcbt-symbol?
  (lambda (symbol)
    (list (_ "KCBT _Contract Specifications")
	  (_ "Open web browser at the Kansas City Board of Trade contract specifications for this commodity")
	  (assoc-ref kcbt-weblinks-list (chart-symbol-commodity symbol)))))


;;-----------------------------------------------------------------------------
;; misc

;; return the tdate of the last record of SYMBOL which has data in COLUMN
(define (column-last-tdate symbol column)
  (let ((series (database-read-series symbol)))
    (if (not series)
	;; no data at all, shouldn't get here since yearly should have given
	;; something, go back only a moderate amount
	(- (tdate-today (timezone-chicago)) 30)

	(let more ((i (series-hi series)))
	  (cond ((< i (series-lo series))
		 ;; nothing at all, again shouldn't get here
		 (- (tdate-today (timezone-chicago)) 200))
		((array-ref (series-array series i i) i column)
		 i) ;; found
		(else
		 (more (1- i))))))))

(define-public (m/d/y-str->mdate str)
  (apply ymd->mdate (m/d/y-str->ymd str)))

(define (set-cddr! lst val)
  (set-cdr! (cdr lst) val))


;;-----------------------------------------------------------------------------
;; contract expiry dates

(define (kcbt-symbol-expiry symbol)
  (kcbt-commodity-mdate-expiry (chart-symbol-commodity symbol)
			       (chart-symbol-mdate symbol)))

;; http://www.kcbt.com/contract_wheat.html
;; http://www.kcbt.com/symbols_trading_hours.html
(define (kcbt-commodity-mdate-expiry commodity mdate)
  (if mdate
      (cond ((string=? commodity "MV")
	     ;; value line

	     ;; third thursday of the contract month
	     (receive-list (year month day)
		 (mdate->ymd mdate)
	       (let ((tdate (ymd->tdate-ceil year month 1)))
		 (+ tdate
		    ;; advance, if necessary, to make it a thursday
		    (modulo (- 4 tdate) 5)
		    ;; advance 2 weeks, making the third thursday
		    10))))

	    (else
	     ;; KW and KE wheat
	     ;; last trading is the business day preceding the 15th
	     ;; calendar day of the contract month
	     (receive-list (year month day)
		 (mdate->ymd mdate)
	       (1- (ymd->tdate-ceil year month 15)))))

      ;; rolling front month, no expiry
      (+ 100 (tdate-today))))


;;-----------------------------------------------------------------------------
;; intraday

(barchart-customer-intraday
 kcbt-symbol?
 (lambda (symbol)
   (let ((date-time (kcbt-quote-adate-time symbol)))
     (set-first! date-time (adate->tdate (first date-time)))
     date-time))
 "http://customer1.barchart.com/cgi-bin/mri/kcbtchart.htx?page=chart&code=mfo&org=com&crea=Y")


;;-----------------------------------------------------------------------------
;; latest
;;
;; The download here is the quotes pages
;;
;;     http://www.kcbt.com/wheat_quotes.asp?page=quote&sym=KW
;;     http://www.kcbt.com/wheat_quotes.asp?page=quote&sym=MV
;;
;; The operative part comes from barchart.com, about 6kbytes for KW (which
;; includes KE) and about 3kbytes for MV.  That's a lot smaller than
;; barchart.com gives on its own pages for the same thing.
;;
;; Sometimes "KE" electronic wheat has the last prices outside the high/low
;; range, maybe that's when it's only a settlement and there was no trading,
;; or something.


;; return current KCBT adate/time, with 10-minute delay
(define (kcbt-quote-adate-time symbol)
  (let ((tm (localtime (- (current-time) #,(hms->seconds 0 10 0))
		       (timezone-chicago))))
    (let ((commodity (chart-symbol-commodity symbol)))
      (cond ((string=? "KW" commodity)
	     ;; 9:30am to 1:15pm
	     (tm->adate-time-within tm
				    #,(hms->seconds 9 30 0)
				    #,(hms->seconds 13 15 0)))

	    (else
	     ;; MV, KE both electronic
	     ;; should be 7:15pm to 3:15pm Sun through Fri
	     (tm->adate-time tm))))))

;; convert SYMBOL from string "KWN5" to "KWN05", if it doesn't already have
;; 2 digits for the year
;;
(define (kcbt-symbol-ensure-yy symbol)
  (let ((m (regexp-exec #,(regexp "^([^0-9]+)([0-9]+)$") symbol)))
    (if m
	(format #f "~a~2,'0d"
		(match:substring m 1)
		(modulo (yyy-str->year (match:substring m 2)) 100))
	symbol)))

;; return list of latest records
(define (kcbt-body->latest-list commodity body)
  (define ret '())

  (receive-list (quote-adate quote-time)
      (kcbt-quote-adate-time commodity)

    ;; the name is in a java expression, so can't just `java-write' the
    ;; whole body at the start
    (for-each
     (lambda (m)
       (let* ((name      (match:substring m 1))
	      (table-str (java-write
			  (substring body (match:end m)
				     ;; table up to next bcfn heading
				     (or (string-contains body
							  "bcfn_PrintLinks"
							  (match:end m))
					 (string-length body)))))
	      (row-list     (html-table-rows table-str)))

	 ;; for MV the heading is just "Quick Links", not the contract name
	 (if (string=? name "Quick Links")
	     (set! name #f))

	 (let ((headings (first row-list)))
	   (or (equal? headings '("CONTRACT" "LAST" "CHANGE"
				  "OPEN" "HIGH" "LOW" "PREV SETTLE"
				  "VOLUME" "OPEN INT" "LAST TRADE TIME"))
	       (error "KCBT quote: unrecognised data columns:" headings)))
	 (set! row-list (cdr row-list))

	 (for-each
	  (lambda (row)
	    (receive-list (symbol last change open high low
				  prev-settle volume int last-time)
		row
	      (set! symbol (kcbt-symbol-ensure-yy symbol))

	      (let ((mdate      (chart-symbol-mdate symbol))
		    (last-adate quote-adate))

		;; "s" for settlement price
		;; and note LAST can be empty when no trade (on "KE" at
		;; least), which will fall back to PREV
		(set! last (string-trim-right last #\s))

		;; "unch" for no change
		(set! change (if (string=? "unch" change)
				 "0"
				 (dash-eighths->decimals change)))

		(set! open 	  (dash-eighths->decimals open))
		(set! high 	  (dash-eighths->decimals high))
		(set! low  	  (dash-eighths->decimals low))
		(set! last 	  (dash-eighths->decimals last))
		(set! prev-settle (dash-eighths->decimals prev-settle))

		;; last time is empty when no trade
		(set! last-time (and (not (string-null? last-time))
				     (h:m:s-str->seconds last-time)))

		;; if last time is after current, or no time value at all,
		;; then must be yesterday
		(if (or (not last-time)
			(> last-time quote-time))
		    (set! last-adate (1- last-adate)))

		(set! ret (cons (latest-new
				 #:symbol         (string-append symbol ".KCBT")
				 #:name           name
				 #:quote-adate    quote-adate
				 #:quote-time     quote-time
				 #:last-adate     last-adate
				 #:last-time      last-time
				 #:open           open
				 #:high           high
				 #:low            low
				 #:last           last
				 #:prev           prev-settle
				 #:change         change
				 #:volume         volume
				 #:contract-mdate mdate
				 #:source         'kcbt)
				ret)))))

	  row-list)))

     (must-match (list-matches #,(regexp "^bcfn_PrintLinks.*'([^']*)'\\);"
					 regexp/newline) body))))

  (latest-list-add-front-months ret))

(define (kcbt-latest-get symbol-list extra-list proc)
  (let ((commodity-list (delete-duplicates (map chart-symbol-commodity
						symbol-list))))
    ;; KW page includes KE
    (if (member "KW" commodity-list)
	(set! commodity-list (delete "KE" commodity-list)))

    (for-each (lambda (commodity)
		(receive (headers body)
		    (http-request (string-append "http://sites.barchart.com/pl/kcbt/quote.htx?sym=" commodity))
		  (if (http-headers-ok? headers)
		      (proc (kcbt-body->latest-list commodity body)))))
	      commodity-list)))

(latest-handler! #:selector kcbt-symbol?
		 #:handler  kcbt-latest-get)


;;-----------------------------------------------------------------------------
;; download - years
;;
;; This uses the yearly historical files at
;;
;;     http://www.kcbt.com/historical_data.asp
;;
;; In kw2003.txt some of the months are not in order (they're sorted
;; lexically so eg. on 3/26/03 the month 3/1/04 is before 5/1/03.  Also in
;; kw2003.txt the dates are not in order, eg. 4/1/03 is before 3/31/03.  And
;; in a couple of spots there's no record for what should be the front month
;; still, eg. on 2/19/04 there's no line for month 3/1/04.
;;
;; In mv2004.txt, 9 Aug 04 has a bad line, the close is "146" but probably
;; should be "1460" or something; could delete that or hard code what it
;; should be, but better leave that to kcbt themselves to get right.
;;

(define char-set:not-comma-tab (char-set-complement (char-set #\, #\tab)))

;; return STR with the last digit interpreted as eighths and changed to
;; decimals
(define (eighths str)
  (string-append (string-drop-right str 1)
		 (vector-ref #("" ".125" ".25" ".375"
			       ".5" ".625" ".75" ".875")
			     (string->number-err (string-take-right str 1)))))

;; if STR is 4 digits with no decimal point, apply `eighths' above
(define (eighths-4digit str)
  (if (and (not (string-any #\. str))
	   (= 4 (string-length str)))
      (eighths str)
      str))

(define (kcbt-yearly-process symbol-list year body)

  (define commodity (chart-symbol-commodity (first symbol-list)))

  ;; remove quotes around fields in old files like mv1999.txt
  (set! body (string-delete body #\"))

  ;; mv1999.txt is just NL, mv2000.txt is just CR, mv2004.txt is CRLF,
  ;; `string-split-lines' can cope with all, including with blank lines at
  ;; end of mv1999.txt
  ;;
  ;; field separator is comma in old like kw2004.txt, tabs in new kw2005.txt
  ;;
  (let ((row-list (map! (lambda (row-str)
			  (string-tokenize row-str char-set:not-comma-tab))
			(string-split-lines body))))

    ;; mv1999.txt and mv2003.txt have some dodgy lines with empty date and
    ;; month fields, ignore those
    (set! row-list (remove! (lambda (row)
			      (string-null? (first row)))
			    row-list))

    ;; KW has symbol as third field, MV doesn't; lose it from KW
    (for-each (lambda (row)
		(if (string-any char-set:letter (third row))
		    (set-cddr! row (cdddr row))))
	      row-list)

    (download-process
     #:module        (_ "KCBT")
     #:symbol-list   symbol-list
     ;; Mark whole year as covered, to avoid re-downloading if 31 Dec is a
     ;; non-trading day.  But don't do that for the current year since it
     ;; won't have the end of the year yet.  The rule is that we have to be
     ;; 15 trading days past the end of the year to mark it.
     #:lo            (year->tdate year)
     #:hi            (and (>= (tdate-today) (+ 15 (year->last-tdate year)))
			  (year->last-tdate year))
     #:currency      "USD"
     #:expiry-proc   kcbt-symbol-expiry
     #:row-list
     ;; old data like kw1970.txt doesn't have volume and open interest
     ;; fields (and some like "9/19/03,9/1/03,KW,0,0,0,3320,," can be empty,
     ;; though maybe that's only on final settlement day)
     (map-rows!
      (lambda* (tdate mdate open high low close #:optional volume openint)

	(cond ((and (equal? "KW" commodity)
		    (= year 2004))
	       ;; kw2004.txt has a nasty mix of eighths and decimals;
	       ;; as a hack, assume 4 digits is eighths
	       (set! open  (eighths-4digit open))
	       (set! high  (eighths-4digit high))
	       (set! low   (eighths-4digit low))
	       (set! close (eighths-4digit close)))

	      ((equal? "KW" commodity) ;; kw last digit is eighths
	       (set! open  (eighths open))
	       (set! high  (eighths high))
	       (set! low   (eighths low))
	       (set! close (eighths close))))

	;; current files have 2-digit year, old files like
	;; mv2000.txt have 4-digit
	(list #:tdate         (m/d/y-str->tdate tdate)
	      #:mdate         (m/d/y-str->mdate mdate)
	      #:commodity     commodity
	      #:open          open
	      #:high          high
	      #:low           low
	      #:close         close
	      #:volume        volume
	      #:open-interest openint))

      row-list))))

;; return the year to start from for updating SYMBOL
(define (yearly-start symbol)
  (tdate->year (download-start-tdate symbol)))

;; return the tdate which should be available in the yearly files
;;
;; these files are supposed to be updated monthly, though 2005 was seen
;; stuck at only January right through to 1Jun; guess 1st of this month
;; anyway
;;
;; saw kw2007.txt going up to 13apr07 as of 26may07, dunno what boundary
;; that's supposed to be (rolling past 4 weeks ago maybe)
;;
(define (yearly-available-tdate)
  (mdate->tdate (tdate->mdate (tdate-today (timezone-chicago)))))

(define (yearly-download-one symbol-list year)
  (let* ((commodity (chart-symbol-commodity (first symbol-list)))
	 (filename  (simple-format #f "~a~a.txt"
				   (string-downcase commodity) year)))
    (download-status (_ "KCBT") (_ "data") filename)

    (receive (headers body)
	(http-request (string-append "http://www.kcbt.com/histdata/"
				     filename))
      (let ((status-code (http-headers-status-code headers)))
	(cond ((http-headers-ok? headers)
	       (kcbt-yearly-process symbol-list year body)
	       #t)
	      ((= 404 status-code)
	       ;; no more years for this commodity
	       #f)
	      (else
	       (http-headers-throw headers)))))))

(define (yearly-download symbol-list)
  (define avail-year  (today-year (timezone-chicago)))

  (for-each
   (lambda (symbol-list) ;; of each commodity
     (let ((start-year (apply min (map yearly-start symbol-list))))

       ;; having decided to download, also do any others covered by this
       ;; data, ie. same commodity and at or beyond start-year
       (set! symbol-list
	     (download-also symbol-list
			    #:selector    kcbt-symbol?
			    #:commodity   #t
			    #:expiry-proc kcbt-symbol-expiry
			    #:pred        (lambda (symbol)
					    (>= (yearly-start symbol)
						start-year))))

       (do ((year start-year (1+ year)))
	   ((> year avail-year))
	 (yearly-download-one symbol-list year))))

   (partition-equal chart-symbol-commodity symbol-list)))


;;-----------------------------------------------------------------------------
;; download - wheat daily price
;;
;; This uses the daily wheat price csv files like
;;
;;     http://www.kcbt.com/download/kcprccsv_20050217.csv
;;
;; This only has KW wheat, it doesn't have MV value line.  Whitespace and
;; all the options contracts pad it out to about 8.8 kbytes for each day.
;;
;; Some reports like kcprccsv_20050517.csv start with a month that's already
;; expired, hence the use of `kcbt-symbol-expiry'.
;;

;; return the tdate of the available daily wheat price file
;; each day's prices are available from 2:30pm (US central)
(define (daily-prices-available-tdate)
  (tdate-today-after 14 30 (timezone-chicago)))

(define (daily-prices-process symbol-list body)

  ;; "" quotes on exchange, symbol and date fields
  ;; lots of leading and trailing spaces
  ;; trailing DOS style ^Z
  (set! body (string-delete body (char-set #\" #\sp #\sub)))

  (let ((row-list (map! (lambda (line)
			  (string-split line #\,))
			(string-split-lines body))))

    (let ((headings (first row-list)))
      (or (equal? headings '("EXCH" "COMID" "MONTH" "YEAR" "STRIKE"
			     "PREVIOUS" "OPEN" "OP_RANGE" "HIGH" "LOW"
			     "CLOSE" "CL_RANGE" "SETTLE" "DATE"))
	  (error "KCBT unrecognised daily wheat price format"))
      (set! row-list (cdr row-list)))

    (download-process
     #:module         (_ "KCBT")
     #:symbol-list    symbol-list
     #:currency       "USD"
     #:expiry-proc    kcbt-symbol-expiry
     #:extra-decimals 2
     #:row-list
     (map-rows!
      (lambda (exchange commodity month year strike prev
			open open-range high low close close-range
			settle tdate)

	(and (string=? strike "0") ;; only futures, not the options lines

	     ;; current files have 2-digit year, old files like
	     ;; mv2000.txt have 4-digit
	     (list #:tdate      (m/d/y-str->tdate tdate)
		   #:mdate      (ymd->mdate (yyy-str->year year)
					    (string->number-err month)
					    1)
		   #:commodity  commodity
		   #:open       open
		   #:high       high
		   #:low        low
		   ;; settlement is used for the close, that seems to be
		   ;; what's in the yearly files; close and close-range
		   ;; are a closing quote, or something
		   #:close      settle)))
      row-list))))

;; return tdate to start downloading to update SYMBOL
(define (daily-prices-start-tdate symbol)
  (1+ (column-last-tdate symbol 3))) ;; last close

(define (daily-prices-download symbol-list)
  (define avail-tdate (daily-prices-available-tdate))
  (define start-tdate (apply min (map daily-prices-start-tdate symbol-list)))

  ;; also update any other wheat in the database which can be covered by
  ;; this start-tdate
  (set! symbol-list
	(download-also symbol-list
		       #:selector    kcbt-symbol?
		       #:commodity   #t
		       #:start-tdate start-tdate
		       #:expiry-proc kcbt-symbol-expiry))

  (do ((t start-tdate (1+ t)))
      ((> t avail-tdate))
    (download-status (_ "KCBT") (_ "wheat prices") (tdate-range-string t t))
    (receive (headers body)
	(http-request (apply format #f "http://www.kcbt.com/download/kcprccsv_~a~2,'0d~2,'0d.csv" (tdate->ymd t)))
      ;; public holidays give 404
      (if (http-headers-ok? headers)
	  (daily-prices-process symbol-list body)))))


;;-----------------------------------------------------------------------------
;; download - daily volume
;;
;; This uses the daily volume report files like
;;
;;     http://www.kcbt.com/download/kcvoicsv_20050224.csv
;;
;; This has KW wheat and MV value line, but we only use it for KW.
;; Whitespace and all the options contracts pad it out to about 8.3 kbytes
;; for each day.


;; return the tdate of the available daily volume file
;; each day's volume is available from 6pm (US central)
;;
(define (daily-volume-available-tdate)
  (tdate-today-after 18 0 (timezone-chicago)))

;; update all in SYMBOL-LIST for volume csv data BODY of TDATE
;;
(define (daily-volume-process symbol-list tdate body)

  ;; "" quotes on exchange, symbol and date fields
  ;; lots of leading and trailing spaces
  ;; trailing DOS style ^Z
  (set! body (string-delete body (char-set #\" #\sp #\sub)))

  (let ((row-list (map (lambda (row-str)
			 (string-split row-str #\,))
		       (string-split-lines body))))

    (let ((headings (first row-list)))
      (or (equal? headings
		  '("COMID" "MONTH" "YEAR" "STRIKE" "VOLUME" "XCFO" "TRANSFER"
		    "GIVEUP" "OPENINT" "CHANGE" "EXR/DELV" "DATE"))
	  (error "KCBT unrecognised daily volume format:" headings))
      (set! row-list (cdr row-list)))

    (download-process
     #:module      (_ "KCBT")
     #:symbol-list symbol-list
     #:currency    "USD"
     #:expiry-proc kcbt-symbol-expiry
     #:row-list
     (map-rows!
      (lambda (commodity month year strike volume xcfo transfer
			 give-up openint change exer-deliv tdate)

	(and (string=? strike "0") ;; only futures, not the options lines

	     ;; current files have 2-digit year, old files like
	     ;; mv2000.txt have 4-digit
	     (list #:tdate      (m/d/y-str->tdate tdate)
		   #:mdate      (ymd->mdate (yyy-str->year year)
					    (month-str->month month)
					    1)
		   #:commodity     commodity
		   #:volume        volume
		   #:open-interest openint)))
      row-list))))

;; return tdate to start downloading to update SYMBOL
(define (daily-volume-start-tdate symbol)
  (1+ (column-last-tdate symbol 4))) ;; last volume

(define (daily-volume-download symbol-list)
  (define avail-tdate (daily-volume-available-tdate))
  (define start-tdate (apply min (map daily-volume-start-tdate symbol-list)))

  ;; also update any other wheat in the database which can be covered by
  ;; this start-tdate
  (set! symbol-list
	(download-also symbol-list
		       #:selector    kcbt-symbol?
		       #:commodity   #t
		       #:expiry-proc kcbt-symbol-expiry
		       #:pred        (lambda (symbol)
				       (>= (daily-volume-start-tdate symbol)
					   start-tdate))))

  (do ((t start-tdate (1+ t)))
      ((> t avail-tdate))
    (download-status (_ "KCBT") (_ "wheat volume") (tdate-range-string t t))
    (receive (headers body)
	(http-request (apply format #f "http://www.kcbt.com/download/kcvoicsv_~a~2,'0d~2,'0d.csv" (tdate->ymd t)))
      ;; 404 for public holidays
      (if (http-headers-ok? headers)
	  (daily-volume-process symbol-list t body)))))


;;-----------------------------------------------------------------------------
;; download - barchart daily
;;
;; This download uses the "Decimal File" option on the kcbt charts page.
;;
;;     http://www.kcbt.com/wheat_quotes.asp?page=ccharts
;;
;; which goes to
;;
;;     http://www.kcbt.com/wheat_quotes.asp?page=chart&cc=MV&mon=M&year=07&data=A&date=060307&size=b&den=med&jav=asc&expm=0&sly=N&ch1=011&arga=&argb=&argc=&ov1=&argd=&arge=&argf=&ch2=&argg=&argh=&argi=&ov2=&argj=&argk=&argl=&sym=MVM07&page=chart&code=mfo&org=com&crea=Y
;;
;; and the javascript directs to
;;
;;     http://customer1.barchart.com/cgi-bin/mri/kcbtchart.htx?data=A&page=chart&cc=MV&mon=M&year=07&data=A&date=060307&size=b&den=med&jav=asc&expm=0&sly=N&ch1=011&arga=&argb=&argc=&ov1=&argd=&arge=&argf=&ch2=&argg=&argh=&argi=&ov2=&argj=&argk=&argl=&sym=MVM07&page=chart&code=mfo&org=com&crea=Y
;;
;; which is html containing a link like
;;
;;     http://customer1.barchart.com//custom/stocks/4671.gif
;;
;; but it's meant to be .txt, not .gif.


;; return tdate of available download data
;;
;; not sure what time new data is available, wheat volume is only out at 6pm
;; so presumably it's no earlier than that, try 8pm
;;
(define (bar-available-tdate . dummy-symbol)
  (tdate-today-after 20 0 (timezone-chicago)))

;; process a barchart file body into the database
;; lines are like "03/16/06,425,425.5,420,423,0,0   "
;; (KW wheat is in decimals too)
;;
;; if a front symbol like just "KE" is asked for in the download there's an
;; extra month field after the date, like "Z05", but that doesn't arise for
;; us
;;
;; START is the tdate to begin from, or #f for everything
;;
(define* (bar-process symbol-list body-mdate body #:optional start)
  (define commodity (chart-symbol-commodity (first symbol-list)))

  (define row-list (map! (lambda (row-str)
			   (string-split row-str #\,))
			 (string-split-lines body)))

  (download-process
   #:module        (_ "KCBT")
   #:symbol-list   symbol-list
   #:last-download #t
   #:expiry-proc   kcbt-symbol-expiry
   #:currency      "USD"
   #:row-list
   (map-rows!
    (lambda (tdate open high low close volume openint)
      (set! tdate (m/d/y-str->tdate tdate))
      (and (or (not start)  ;; ignore before START
	       (>= tdate start))
	   (list #:tdate         tdate
		 #:mdate         body-mdate
		 #:commodity     commodity
		 #:open          open
		 #:high          high
		 #:low           low
		 #:close         close
		 #:volume        volume
		 #:open-interest openint)))
    row-list)))

;; return mdate which the earliest not yet expired on TDATE
;; expiry is some date in the contract month, so check if TDATE is before or
;; after that point in its month
;; (define (kcbt-tdate->front-mdate commodity tdate)
;;   (let ((mdate (tdate->mdate tdate)))
;;     (if (> tdate (kcbt-commodity-mdate-expiry commodity mdate))
;; 	(1+ mdate)
;; 	mdate)))

;; return #t body, or #f
;;
(define (bar-download-one symbol-list commodity mdate)
  (define specific (string-append commodity (mdate->MYY-str mdate)))

  (download-status (_ "KCBT") (_ "link") specific)
  (receive (headers body)
      (http-request (string-append
		     "http://customer1.barchart.com/cgi-bin/mri/kcbtchart.htx?data=A&page=chart&code=mfo&org=com&crea=Y&jav=ASC&vol=Y&sym=" specific)
		    #:want-ok #t)

    ;; When no such contract month the body is a little document.write()
    ;; with "There is no chart available."  Note "data=A" above, for daily,
    ;; is essential to get that error, without it various weird old months
    ;; are returned.
    ;;
    (and (not (string-contains-ci body "There is no chart available"))

	 (let* ((m (must-match (string-match "<img src=\"([^\"\r\n]+)gif\""
					     body))))
	   (download-status (_ "KCBT") (_ "data") specific)
	   (receive (headers body)
	       (http-request (string-append (match:substring m 1) "txt"))
	     (and (http-headers-ok? headers)
		  (begin
		    (bar-process symbol-list mdate body)
		    #t)))))))

;; return a count of initial download days for SYMBOL
;; KE electronic wheat only goes back to 2005, so limit it 
(define (kcbt-initial-proc symbol)
  (if (string=? "KE" (chart-symbol-commodity symbol))
      500
      #f))

(define (bar-download symbol-list)
  (futures-download-by-specific symbol-list
				#:download-proc    bar-download-one
				#:avail-tdate-proc bar-available-tdate
				#:initial-proc     kcbt-initial-proc))


;;-----------------------------------------------------------------------------
;; download
;;
;; This is the main update download, choosing between methods.
;;
;; KW wheat has yearly and daily files.  Dates between the database last and
;; where the yearly files go to can be done with either yearly or daily, the
;; remainder up to today can only be done with the daily.
;;
;; A cost is calculated for using yearly or daily on the portion with a
;; choice.  The daily files are a bit expensive at about 8k prices plus 8k
;; volumes, so near the start of the year when the yearly file is still
;; small it can work out best to get for instance the last few days of
;; January using the yearly instead of daily.


;; return the cost for downloading yearly files for tdates LO to HI
;; full year KW files are about 74k, and partial year is about 340 bytes per
;; day, and add 2k as a penalty for each separate http request
;;
(define (calc-yearly-cost lo hi)
  (let* ((lo-year (tdate->year lo))
	 (hi-year (tdate->year hi))
	 (hi-start (year->tdate hi-year)))
    (+ (* (+ 74000 2000) (- hi-year lo-year)) ;; full files size+penalty
       (* (* 340 (- hi hi-start -1)))         ;; partial file size
       2000)))                                ;; partial file penalty

;; return the cost for downloading daily files for tdates LO to HI
;; currently 8.8k for prices and 8.3k for volumes, and add 2k as a penalty
;; for each separate http request
;;
(define (calc-daily-cost lo hi)
  (* (+ 8800 2000 8300 2000) (- hi lo -1)))

;; return the cost for downloading barchart files for SYMBOL-LIST

;; currently 26k for the page with the link, then 40k for the actual data,
;; and add 2k as a penalty for each separate http request
;; ENHANCE-ME: this doesn't take into account that specific months and the
;; front month may overlap, so need only count for one file
;;
(define (calc-bar-cost symbol-list)
  (apply + (map (lambda (symbol)
		  (if (chart-symbol-mdate symbol)
		      ;; specific month file 40k plus penalty
		      (+ 26000 40000 2000)
		      ;; front month, guess new file for each 3 months
		      (* (+ 26000 40000 2000)
			 (max 1 (quotient (- (mdate-today)
					     (tdate->mdate
					      (download-start-tdate symbol)))
					  3)))))
		symbol-list)))

;; return the last tdate of data for SYMBOL-LIST
(define (symbols-last-tdate symbol-list)
  (apply min (map (lambda (symbol)
		    (or (database-last-tdate symbol)
			(- (tdate-today (timezone-chicago))
			   initial-download-days))  )
		  symbol-list)))

(define (kcbt-download symbol-list)
  (for-each
   (lambda (symbol-list) ;; of each commodity
     (let ((commodity (chart-symbol-commodity (first symbol-list))))
       (cond ((string=? commodity "KW")

	      ;; for KW the dates from start-tdate to yearly-tdate can be
	      ;; done either by downloading year files or day files, use the
	      ;; year files if that would be cheaper
	      ;;
	      (let* ((yearly-tdate (yearly-available-tdate))
		     (start-tdate  (1+ (symbols-last-tdate symbol-list))))
		;; if yearly can do something and it's less costly than daily
		(if (and (<= start-tdate yearly-tdate)
			 (< (calc-yearly-cost start-tdate yearly-tdate)
			    (calc-daily-cost start-tdate yearly-tdate)))
		    (yearly-download symbol-list)))

	      ;; now choose between barchart files or daily updates
	      ;;
	      ;; the daily updates at 16k per day are a bit costly, the
	      ;; barchart files work out better if it's more than a few days
	      ;;
	      (let* ((start-tdate (1+ (symbols-last-tdate symbol-list))))
		(if (< (calc-bar-cost symbol-list)
		       (calc-daily-cost start-tdate
					(daily-prices-available-tdate)))
		    (bar-download symbol-list)
		    (begin
		      (daily-prices-download symbol-list)
		      (daily-volume-download symbol-list)))))

	     ((string=? commodity "MV")
	      ;;
	      ;; barchart files need to be downloaded anyway to get the
	      ;; latest, but the yearly is good for a front month symbol
	      ;; where the last-tdate is fairly old and would therefore need
	      ;; several barchart files; use the yearly if more than 6
	      ;; months old
	      ;;
	      (if (any (lambda (symbol)
			 (and (not (chart-symbol-mdate symbol))
			      (or (not (database-last-tdate symbol))
				  (< (database-last-tdate symbol)
				     (- (tdate-today) 130)))))
		       symbol-list)
		  (yearly-download symbol-list))

	      (bar-download symbol-list))

	     (else ;; KE
	      ;;
	      ;; there's no yearly files for KE, so barchart is the only
	      ;; option
	      (bar-download symbol-list)))))

   (partition-equal chart-symbol-commodity symbol-list)))

(define-public (kcbt-backto want-tdate symbol-list)
  (define want-year (tdate->year want-tdate))

  (for-each
   (lambda (symbol-list) ;; of each commodity
     (do ((year (tdate->year (download-back-tdate-max symbol-list))
		(1- year)))
	 ((< year want-year))
       ;; stop at 404
       (if (not (yearly-download-one symbol-list year))
	   (set! year want-year))))

   (partition-equal chart-symbol-commodity symbol-list)))

(download-handler! #:selector kcbt-symbol?
		   #:update   kcbt-download
		   #:backto   kcbt-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any kcbt-symbol? symbol-list)
			      (download-now-commodity-all-months))))
