;; Keltner Channel.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart keltner)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart sma))


;; Return a procedure which calculates a Keltner style simple moving average
;; of typical price, and a simple moving average of daily trading range
;; (high to low), over an accumulated window of COUNT points.
;;
;; Each call (PROC high low close) enters a new value into the window, and
;; the return is 2 values (values TP R) which are the calculated values up
;; to (and including) that data point.  TP and R are #f if not enough points
;; yet.
;;
;; HIGH and/or LOW can be #f for no values for them.  If only CLOSE is
;; available then the calculation degenerates to an SMA of the closes.
;;
(define*-public (keltner-calc-proc count #:optional (width 1))
  (let ((typ-sma   (sma-calc-proc count))
	(range-sma (sma-calc-proc count)))

    (lambda (high low close)
      ;; use closes if no high/low data
      (or high (set! high close))
      (or low  (set! low  close))

      (values (typ-sma   (/ (+ high low close) 3.0))
	      (range-sma (* width (- high low)))))))


(define (keltner-fill series new-array lo hi)
  (receive-list (count width)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos parent #:before lo #:count count))
	   (array  (series-array parent start hi))
	   (proc   (keltner-calc-proc count width)))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i lo))
	(if (array-ref array i 3)
	    (proc (array-ref array i 1)    ;; high
		  (array-ref array i 2)    ;; low
		  (array-ref array i 3)))) ;; close

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i 3)
	    (receive (tp r)
		(proc (array-ref array i 1)  ;; high
		      (array-ref array i 2)  ;; low
		      (array-ref array i 3)) ;; close
	      (array-set! new-array tp i 1)
	      (and tp r
		   (begin
		     (array-set! new-array (- tp r) i 0)
		     (array-set! new-array (+ tp r) i 2)))))))))

(define*-public (keltner-series series count #:optional (width 1))
  (ohlcv-required "Keltner" series)
  (let ((new-series (series-copy series)))
    (series-name! new-series
		  (if (= 1 width)
		      (format #f "~a ~a" (_ "Keltner") count)
		      (format #f (_ "Keltner ~a, ~a") count
			      (locale-number-min-decimals width 0))))
    (series-datatype! new-series keltner-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count width))
    (series-require!  new-series (indicator-require-parts keltner-fill
							  (1- count)
							  '(0 2)))
    (series-array!    new-series #f)
    new-series))

(define-public keltner-datatype
  (datatype-new!
   #:key       'keltner
   #:name      (_ "Keltner Channel")
   #:units     #f
   #:input     datatype-ohlcv
   #:linestyle 'bands
   #:manual    (_ "CONTEXT(manual-node): Keltner Channel")
   #:proc      keltner-series
   #:closes-apply-arg '(1)  ;; middle MA line
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "Days")
	  #:key      'keltner-days
	  #:type     'integer
	  #:default  10
	  #:minimum  1)
	 (datatype-param-new
	  #:name     (_ "Width")
	  #:key      'keltner-width
	  #:type     'float
	  #:decimals 1
	  #:default  1
	  #:step     0.5
	  #:minimum  0))))
