;; Kansai Commodities Exchange (KEX) setups.     -*- coding: shift_jis -*-

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart kex)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format) (format)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart gzip)  ;; program-version-proc and program-filter-string
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (kex-symbol? symbol)
  (string-suffix? ".KEX" symbol))

(symbol-timezone! kex-symbol? (_ "Kansai") timezone-tokyo)
(source-help! kex-symbol?
	      (_ "CONTEXT(manual-node): Kansai Commodities Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - contract specifications

;; the english pages might be slightly old, but hopefully still accurate
(define kex-weblinks-list
  '(("EBI" ;; frozen shrimp
     "http://kanex.or.jp/english/youkou-ebi-eng.htm"
     "http://kanex.or.jp/yoko/yoko_ebi.html")
    ("CF"  ;; coffee
     "http://kanex.or.jp/english/youkou-cafe-eng.htm"
     "http://kanex.or.jp/yoko/yoko_coffee.html")
    ("KI"  ;; corn75 index
     "http://kanex.or.jp/english/youkou-corn-eng.htm"
     "http://kanex.or.jp/yoko/yoko_corn_index.html")
    ("CO"  ;; yellow corn
     #f
     "http://kanex.or.jp/yoko/yoko_corn75.html")
    ("RB"  ;; azuki beans
     "http://kanex.or.jp/english/youkou-azuki-eng.htm"
     "http://kanex.or.jp/yoko/yoko_azuki.html")
    ("SG"  ;; sugar
     "http://kanex.or.jp/english/youkou-sugar-eng.htm"
     "http://kanex.or.jp/yoko/yoko_sugar.html")
    ("SM"  ;; soybean meal ;; gone as of Jul07
     #f
     "http://kanex.or.jp/yoko/yoko_smeal.html")
    ("BR"  ;; broiler
     #f
     "http://kanex.or.jp/yoko/yoko_broiler.html")
    ("N" ;; non-gm soybeans
     "http://kanex.or.jp/english/youkou-non-eng.htm"
     "http://kanex.or.jp/yoko/yoko_daizu.html")))
    ;;     ("RS"  ;; raw silk
    ;;      #f
    ;;      #f)

(weblink-handler! kex-symbol?
  (lambda (symbol)
    (list (_ "KEX _Contract Specifications")
	  (_ "Open web browser at the Kansai Commodities Exchange contract specifications for this commodity")
	  ;; selected en/ja language, and fall back to ja if there's no
	  ;; english link at all
	  (and-let* ((elem (assoc (chart-symbol-commodity symbol)
				  kex-weblinks-list)))
	    (let ((lang-func (lang-select `(("ja" ,third)
					    ("en" ,second)))))
	      (or (lang-func elem)
		  (third elem)))))))


;;-----------------------------------------------------------------------------
;; lha archive stuff

;; return the version number of the lha program, or #f if not available
(define-public lha-version
  (program-version-proc "lha --version"
    (lambda (str)
      (set! str (string-trim-str str "lha for unix version "))
      (string-trim-right str))
    #:want-exit 1))

;; the ".exe" extension on the temporary file is essential, it's how lha
;; recognises a self-extracting archive versus an ordinary one
(define-public (unlha-exe-string str)
  (program-filter-string "lha pq" str #:ext ".exe"))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the quotes pages (about 10kbytes each) like
;;
;;     http://kanex.or.jp/market/price_coffee.html
;;     http://kanex.or.jp/market/price_corn.html
;;     http://kanex.or.jp/market/price_soybeans.html
;;
;; Which can be translated by google
;;
;;     http://translate.google.com/translate?u=http://kanex.or.jp/market/price_soybeans.html
;;
;; The server provides ETag and Last-Modified, which we use to avoid
;; re-downloading when unchanged.


;; return list (ADATE TIME) for latest quote
;; delayed by 15 mins
;;
;;     m1  9:00
;;     m2 10:00
;;     m3 11:00
;;     a1 13:00
;;     a2 14:00
;;     a3 15:00
;;
(define (kex-quote-adate-time . dummy-symbol)
  (tm->adate-time-within (localtime (current-time) (timezone-tokyo))
			 #,(hms->seconds 9 0 0)
			 #,(hms->seconds 15 30 0)))

(define (strip-ABHL str)
  (if (or (string-prefix? "A" str)
	  (string-prefix? "B" str)
	  (string-prefix? "H" str)
	  (string-prefix? "L" str))
      (string-trim str #\space 1)
      str))

(define (body->latest-list commodity headers body)
  (let* ((ret         '())
	 (charset     (html-charset headers body))
	 (pos         (string-contains body "mainContents"))
	 (name-list   (html-table-rows body pos))
	 (name-func   (lang-select `(("en" ,third)
				     ("ja" ,second #:charset ,charset))))
	 (name        (charset->utf8 charset (name-func (first name-list))))

	 (pos         (string-contains-after-ci body "<table" pos))
	 (row-list    (html-table-rows body pos))
	 (m           (must-match (regexp-exec #,(regexp "^([0-9][0-9][0-9][0-9])/([0-9]+)/([0-9]+) ([0-9]+):([0-9]+)$") (first (first row-list)))))
	 (year        (string->number (match:substring m 1)))
	 (month       (string->number (match:substring m 2)))
	 (day         (string->number (match:substring m 3)))
	 (hour        (string->number (match:substring m 4)))
	 (minute      (string->number (match:substring m 5)))
	 (quote-adate (ymd->adate year month day))
	 (quote-time  (hms->seconds hour minute 0)))
    (set! row-list (cdr row-list))

    ;; "morning" and "afternoon" rowspan entries
    (set! row-list (map! (lambda (row)
			   (if (member (first row) '("O" ""))
			       (cdr row)
			       row))
			 row-list))

    ;; EBI shrimp has only m2 m3 a2 a3, insert dummy m1
    (or (string-contains (first (third row-list)) "m.1st")
	(list-insert-at! '("m.1st" "" "" "" "" "" "" "") 2 row-list))

    ;; RB has only m1 a1 a3, insert dummy m2
    (or (string-contains (first (fourth row-list)) "m.2nd")
    	(list-insert-at! '("m.2nd" "" "" "" "" "" "" "") 3 row-list))

    ;; SG sugar has only m1 m2 a1 a2 a3, insert dummy m3
    (or (string-contains (first (fifth row-list)) "m.3d")
    	(list-insert-at! '("m.3d" "" "" "" "" "" "" "") 4 row-list))

    ;; EBI shrimp has only m2 m3 a2 a3, insert dummy a1
    (or (string-contains (first (sixth row-list)) "a.1st")
    	(list-insert-at! '("a.1st" "" "" "" "" "" "" "") 5 row-list))

    ;; RS has only m1 m3 a1 a3, insert dummy a2
    (or (string-contains (first (seventh row-list)) "a.2nd")
    	(list-insert-at! '("a.2nd" "" "" "" "" "" "" "") 6 row-list))

    (set! row-list (apply zip row-list))

    ;; rest have all m1 m2 m3 a1 a2 a3

    (let ((headings (first row-list)))
      (set! headings
	    (map (lambda (heading)
		   (let ((m (regexp-exec-C #,(regexp "([(]|i)(.*)([)]|j)")
					   heading)))
		     (if m
			 (match:substring m 2)
			 heading)))
		 headings))

      (or (every (lambda (field want)
		   (string-contains field want))
		 headings
		 ;; note multibyte "r" in "3rd"
		 '(".month" "pre.sett"
		   "m.1st" "m.2nd" "m.3d"
		   "a.1st" "a.2nd" "a.3d"
		   "prev.chg" "volume" "open.int" "daily price limits"))
	  (error "KEX unrecognised quote headings:" headings)))
    (set! row-list (cdr row-list))

    (let ((final (last row-list)))
      (or (string-contains (first final) "volume")
	  (error "KEX unrecognised quote total volume:" final)))
    (set! row-list (drop-right! row-list 1))

    (for-each
     (lambda (row)
       (receive-list (month prev m1 m2 m3 a1 a2 a3 change volume int limits)
	   row
	 (define mdate (y/m-str->mdate month))
	 (define note #f)

	 ;; commas in prices and volume
	 (set! prev   (string-delete prev   #\,))
	 (set! m1     (string-delete m1     #\,))
	 (set! m2     (string-delete m2     #\,))
	 (set! m3     (string-delete m3     #\,))
	 (set! a1     (string-delete a1     #\,))
	 (set! a2     (string-delete a2     #\,))
	 (set! a3     (string-delete a3     #\,))
	 (set! change (string-delete change #\,))

	 (and-let* ((last (find (lambda (field)
				  (not (string-null? field)))
				(list a3 a2 a1 m3 m2 m1))))
	   (cond ((string-prefix? "H" last)
		  (set! note (_ "limit up")))
		 ((string-prefix? "L" last)
		  (set! note (_ "limit down")))))

	 ;; possible "A", "H" or "L" prefix on prices, not sure what it is
	 (set! prev (strip-ABHL prev))
	 (set! m1   (strip-ABHL m1))
	 (set! m2   (strip-ABHL m2))
	 (set! m3   (strip-ABHL m3))
	 (set! a1   (strip-ABHL a1))
	 (set! a2   (strip-ABHL a2))
	 (set! a3   (strip-ABHL a3))

	 ;; volume can be "-", maybe before opening
	 (if (string=? "-" volume)
	     (set! volume #f))

	 (receive-list (decimals        prev m1 m2 m3 a1 a2 a3 change)
	     (strings->numbers+decimals prev m1 m2 m3 a1 a2 a3 change)

	   (let ((open       (or m1 m2))
		 (high       (max-maybe m1 m2 m3 a1 a2 a3))
		 (low        (min-maybe m1 m2 m3 a1 a2 a3))
		 (last       (or a3 a2 a1 m3 m2 m1))
		 (last-adate quote-adate)
		 (last-time  (cond (a3   #,(hms->seconds 15 0 0))
				   (a2   #,(hms->seconds 14 0 0))
				   (a1   #,(hms->seconds 13 0 0))
				   (m3   #,(hms->seconds 11 0 0))
				   (m2   #,(hms->seconds 10 0 0))
				   (m1   #,(hms->seconds  9 0 0))
				   (else #f))))

	     ;; if nothing in todays sessions m1 to a3 then show prev
	     ;; this happens in the morning, eg. on EBI before 10am
	     (if (not last)
		 (begin
		   (set! last-adate (adate-prev-tdate last-adate))
		   (set! last-time #f)
		   (set! open #f)
		   (set! high #f)
		   (set! low #f)
		   (set! last prev)
		   (set! change #f)
		   (set! volume #f)))

	     (set! ret
		   (cons (latest-new
			  #:symbol  (string-append commodity
						   " "
						   (mdate->MMM_YY-str mdate)
						   ".KEX")
			  #:name           (string-append name " " month)
			  #:quote-adate    quote-adate
			  #:quote-time     quote-time
			  #:last-adate     last-adate
			  #:last-time      last-time
			  #:open           open
			  #:high           high
			  #:low            low
			  #:last           last
			  #:change         change
			  #:volume         volume
			  #:decimals       decimals
			  #:contract-mdate mdate
			  #:source         'kex
			  #:headers        headers)
			 ret))))))

     row-list)
    (latest-list-add-front-months ret)))

(define (kex-latest-get symbol-list extra-list proc)

  (for-each (lambda (symbol-list) ;; of each commodity
	      (let* ((commodity (chart-symbol-commodity (first symbol-list)))
		     (code (assoc-ref '(("CF"    . "coffee") ;; coffee
					("CAFE"  . #f)      ;; coffee index val
					("INDEX" . #f)      ;; corn75 index val
					("EBI" 	 . "ebi")      ;; shrimp
					("KI"  	 . "corn75")   ;; corn 75
					("N"   	 . "soybeans") ;; non-gmo soy
					("RB"  	 . "azuki")  ;; azuki redbeans
					("RS"  	 . #f)       ;; raw silk
					("SG"    . "sugar")  ;; raw sugar
					("SB"    . #f)       ;; IOM soybeans
					("BR"    . "broiler") ;; broiler
					("CO"    . "corn")   ;; yellow corn
					;; ex-fukuoka, gone as of Jul07
					;; ("SM"    . "smeal")  ;; soy meal
					)
				      commodity)))
		(if code
		    (receive (headers body)
			(apply http-request (string-append
					     "http://kanex.or.jp/market/price_"
					     code ".html")
			       #:want-ok #t
			       (latest-tag-args 'kex symbol-list))
		      (if (http-headers-ok? headers) ;; not 304 Not Modified
			  (proc (body->latest-list commodity
						   headers body)))))))

	    (partition-equal chart-symbol-commodity symbol-list)))

(latest-handler! #:selector   kex-symbol?  
		 #:handler    kex-latest-get
		 #:adate-time kex-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - data
;;
;; This uses the historical data downloads at
;;
;;     http://kanex.or.jp/market/his_index.html
;; 
;; Which can be viewed translated by google
;;
;;     http://translate.google.com/translate?u=http://kanex.or.jp/market/his_index.html
;;
;; The files are a whole years trading, like the following for shrimp for
;; 2006,
;;
;;     http://kanex.or.jp/history/EBI06.exe
;;
;; The latest year file goes up to yesterday.  There's an ETag/Last-Modified
;; to avoid re-downloading before it changes.
;;


;; This is a nasty hack for the SB02.exe file.  It has nested lzh and exe
;; archives and three copies of the desired sb02.csv (the last has some
;; extra whitespace).  Do a special extract.
;;
(define (hack-SB02-body body)
  (set! body (program-filter-string (lambda (filename)
				      (string-append "lha pq " filename
						     " sb02.lzh"))
				    body #:ext ".exe"))
  (set! body (program-filter-string (lambda (filename)
				      (string-append "lha pq " filename
						     " sb02.lzh"))
				    body #:ext ".lzh"))
  (program-filter-string (lambda (filename)
			   (string-append "lha pq " filename
					  " sb02.csv"))
			 body #:ext ".lhz"))

(define (process-download symbol-list year url headers body)

  (define commodity (chart-symbol-commodity (first symbol-list)))

  (if (string-suffix? "SB02.exe" url)
      (set! body (hack-SB02-body body))
      (set! body (unlha-exe-string body)))

  (let ((row-list (string-split-lines body)))

    ;; only `string-prefix?' because EBI02.exe has repetitions of the field
    ;; names headings (and empty fields for them in the data)
    (let ((headings (first row-list)))
      (or (string-prefix? "yr_mo_dy,contract,contract_month,contract_price_m1,contract_price_m2,contract_price_m3,contract_price_a1,contract_price_a2,contract_price_a3,sett_price,volume,open_int,net_position" headings)

	  ;; this form "date" in BRO07.exe
	  (string-prefix? "date,contract,contract_month,contract_price_m1,contract_price_m2,contract_price_m3,contract_price_a1,contract_price_a2,contract_price_a3,sett_price,volume,open_int,net_position" headings)

	  (error "KEX data unrecognised headings:" headings))
      (set! row-list (cdr row-list)))

    (let ((headings (first row-list)))
      (or (string-contains headings ",OP,OQ,OR,P,Q,R,li,o,g,Ќ")
	  (error "KEX data unrecognised second headings:" headings))
      (set! row-list (cdr row-list)))

    ;; lose empty rows at end of old files (commas but nothing in the fields)
    (set! row-list (remove! (lambda (row-str)
			      (string-every #\, row-str))
			    row-list))

    ;; use `csv-split' because some old data in SB01.exe has commas and
    ;; quotes in fields
    (set! row-list (map! csv-split row-list))

    (download-process
     #:module        (_ "KEX")
     #:symbol-list   symbol-list
     #:lo            (year->tdate year)
     #:hi            (year->last-tdate year)
     #:url           url
     #:headers       headers
     #:row-list
     (map-rows!
      ;; ". REST" because SB02.exe has big runs of empty commas at end of lines
      ;;
      ;; NAME unused, instead `download-name-from-latest' gives english or
      ;; japanese according to preference
      ;;
      (lambda (tdate name mdate m1 m2 m3 a1 a2 a3 settle volume openint
		     net-position . rest)
	;; some rows of only commas at end
	(and (not (string-every #\sp tdate))
	     (list #:tdate         (yyyymmdd-str->tdate tdate)
		   #:mdate         (yyyymm-str->mdate mdate)
		   #:commodity     commodity
		   #:sessions      (list m1 m2 m3 a1 a2 a3 settle)
		   #:volume        volume
		   #:open-interest openint)))
      row-list))))

;; return code string, or #f if unknown commodity
(define (kex-commodity->download-code commodity)
  (assoc-ref '(("EBI" . "EBI")  ;; shimp
	       ("CF"  . "COF")  ;; coffee index
	       ;; COFINDEX  ;; cash value (?)
	       ("KI"  . "C75")  ;; corn75
	       ;; C75INDEX  ;; cash value (?)
	       ("N"   . "NON")  ;; non-gmo soy
	       ("RB"  . "RBN")  ;; azuki
	       ("SG"  . "SUG")  ;; raw sugar

	       ;; former fukuoka contracts
	       ("SM"  . "SML")  ;; soybean meal
	       ("CO"  . "YCN")  ;; yellow corn
	       ("BR"  . "BRO")) ;; broiler
	     commodity))

;; return #t if got data, #f if 404
(define (kex-download-year symbol-list year) ;; of one commodity
  (and-let* ((code (kex-commodity->download-code
		    (chart-symbol-commodity (first symbol-list)))))

    (let ((url (format #f "http://kanex.or.jp/history/~a~2,'0d.exe"
		       code (modulo year 100))))
      (download-status (_ "KEX") (_ "data") (basename url))

      (receive (headers body)
	  (apply http-request url (download-tag-args url symbol-list))
	(let ((status-code (http-headers-status-code headers)))
	  (cond ((http-status-code-ok? status-code)
		 (process-download symbol-list year url headers body)
		 #t)
		((= 404 status-code)
		 #f)
		((= 304 status-code)
		 ;; not modified
		 #t)
		(else
		 ;; anything else is bad
		 (http-headers-throw headers))))))))

;; seen Last-Modified
;;    EBI05.exe: Fri, 09 Sep 2005 00:05:24 GMT   = 9:05:24 tokyo
;;    CF05.exe:  Fri, 09 Sep 2005 00:05:22 GMT
;;    N05.exe:   Fri, 16 Sep 2005 00:08:18 GMT
;; each day available after 9:15am the following morning
;;
(define (kex-available-tdate)
  (1- (tdate-today-after 9 15 (timezone-tokyo))))

(define (kex-download symbol-list)
  (if (not (lha-version))
      (download-message (string-append (_ "KEX") ": "
				       (_ "LHa program not available")))

      (let* ((avail-tdate (kex-available-tdate))
	     (avail-year  (tdate->year avail-tdate)))

	(for-each
	 (lambda (symbol-list) ;; of the same commodity
	   (let ((start-year (tdate->year (apply min (map download-start-tdate
							  symbol-list)))))
	     (set! symbol-list
		   (download-also symbol-list
				  #:selector    kex-symbol?
				  #:start-tdate (year->tdate start-year)
				  #:commodity   #t))
	     (do ((year start-year (1+ year)))
		 ((> year avail-year))
	       (kex-download-year symbol-list year))))

	 (partition-equal chart-symbol-commodity symbol-list))))

  (download-name-from-latest (_ "KEX") symbol-list))

(define (kex-backto want-tdate symbol-list)
  (define want-year (tdate->year want-tdate))

  (for-each
   (lambda (symbol-list) ;; of the same commodity
     (let ((start-year (tdate->year (apply max (map download-back-tdate
						    symbol-list)))))
       (do ((year start-year (1- year)))
	   ((< year want-year))
	 ;; stop at 404
	 (or (kex-download-year symbol-list year)
	     (set! year want-year)))))

   (partition-equal chart-symbol-commodity symbol-list)))


;;-----------------------------------------------------------------------------
;; download

(download-handler! #:selector   kex-symbol?
		   #:update     kex-download
		   #:avail-proc kex-available-tdate
		   #:backto     kex-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any kex-symbol? symbol-list)
			      (download-now-commodity-all-months))))


;;-----------------------------------------------------------------------------
;; setups - intraday

;; Daily data chart images are on the page
;;
;;     u=http://kanex.or.jp/market/chart.html
;; translated:
;;     http://translate.google.com/translate?u=http://kanex.or.jp/market/chart.html
;;
;; which results in images like
;;
;;     http://kanex.or.jp/market/img/ebi_saki.jpg
;;
;; or for a particular month (YYMM)
;;
;;     http://kanex.or.jp/market/img/cof_0705.jpg
;;
;; The commodity code parts are hard coded here, they're not quite the same
;; as the download filenames.
;;
;; Not sure how much value there is in showing what's already available from
;; the main chart graphs.

(define (kex-intraday-url symbol mode)
  (and-let* ((code (assoc-ref '(("EBI" . "ebi")  ;; shimp
				("N"   . "non")  ;; non-gmo soy
				("RB"  . "rbn")  ;; azuki
				("CF"  . "cof")  ;; coffee index
				("KI"  . "c75")  ;; corn75
				("SG"  . "sug")  ;; raw sugar
				) (chart-symbol-commodity symbol))))
    (let ((mdate (chart-symbol-mdate symbol)))
      (if mdate
	  (receive-list (year month day)
	      (mdate->ymd mdate)
	    (format #f "http://kanex.or.jp/market/img/~a_~2,'0d~2,'0d.jpg"
		    code (modulo year 100) month ))
	  (string-append "http://kanex.or.jp/market/img/"
			 code
			 "_saki.jpg")))))

(intraday-url-proc! #:symbol-pred kex-symbol?
		    #:url-proc    kex-intraday-url
		    #:modes-proc  (lambda ()
				    (list (_ "Daily"))))
