;; Keymaps, definitions and dispatch.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart keymap)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (chart misc)
  #:autoload   (chart view)   (view-message))


;;-----------------------------------------------------------------------------
;; Gdk keyval extras

;; Compare two keyvals, by name, and with plain letters sorting before
;; function keys.  Believe these names are ascii, so plain `string<?' ok.
;;
(define-public (c-gdk-keyval<? x y)
  (let ((x (gdk-keyval-name x))
	(y (gdk-keyval-name y)))
    (or (and (= 1 (string-length x))
	     (not (= 1 (string-length y))))
	(string<? x y))))

;; return #t if KEYVAL is an alphabetic key
(define (c-gdk-keyval-alpha? keyval)
  (not (= (gdk-keyval-to-lower keyval)
	  (gdk-keyval-to-upper keyval))))

;; return keyval GDK_Pointer_Button1 etc for BUTTON values 1 to 5
(define (c-gdk-button-to-keyval button)
  (gdk-keyval-from-name (format #f "Pointer_Button~a" button)))


;;-----------------------------------------------------------------------------
;; GtkAccelGroup extras

;; not wrapped in guile-gtk 2.0
(if (not (defined? 'gtk-accelerator-get-default-mod-mask))
    (define (gtk-accelerator-get-default-mod-mask)
      '(control-mask shift-mask mod1-mask super-mask hyper-mask meta-mask)))


;;-----------------------------------------------------------------------------

;; keymap-set! parses as per gtk-accelerator-parse.  Alphabetic characters
;; are not case significant, so "X" or "x" are the same.  "<Shift>X" can be
;; used for an "upper case" key.
;;
;; (TEST-FUNC) returns #t or #f according to when the keymap should be used
;; NAME is a description of the state when the keymap is used

;; compare symbols, in an unspecified order
(define (symbol<? x y)
  (string<? (symbol->string x) (symbol->string y)))

;; `keymap-widget' is probably better handled with `gtk-get-event-widget'
;; if/when that's wrapped
(define-public keymap-widget #f)
(define-public keymap-event  #f)
(define-public keymap-keyval #f)
(define-public keymap-mods   #f)

;; all keymaps created
(define-public keymap-list '())

(define rec-type (make-record-type "keymap" '(name testfunc alist propagate)))
(define rec-new  (record-constructor rec-type))

;; return a pair (keyval . mods)
(define (keymap-parse str)
  (let* ((parse  (gtk-accelerator-parse str))
	 (keyval (first parse))
	 (mods   (second parse)))
    (if (= 0 keyval)
	(error "keymap-parse: cannot parse key: " str))
    ;; sorted for consistent comparison
    (cons keyval (sort! mods symbol<?))))

;; find KEYVAL/MODS in an active keymap, return procedure or #f if no binding
;; no propagate doesn't prevent accelerators, unfortunately
(define (keymap-dispatch-find keyval mods)
  (define target (cons keyval mods))
  (let more ((lst keymap-list))
    (and (not (null? lst))
	 (let ((keymap (first lst)))
	   (cond ((not ((keymap-testfunc keymap)))
		  (more (cdr lst)))  ;; not active
		 ((assoc target (keymap-alist keymap))
		  => cdr)            ;; found entry
		 ((keymap-propagate keymap)
		  (more (cdr lst)))  ;; continue to others
		 (else
		  #f))))))           ;; no propagate

(define        keymap-testfunc  (rec-accessor rec-type 'testfunc))
(define        keymap-propagate (rec-accessor rec-type 'propagate))
(define-public keymap-name      (rec-accessor rec-type 'name))
(define-public keymap-alist     (rec-accessor rec-type 'alist))
(define        keymap-alist!    (rec-modifier rec-type 'alist))

(define*-public (keymap-new name testfunc #:key (propagate #t))
  (let ((keymap (rec-new name testfunc '() propagate)))
    (set! keymap-list (cons keymap keymap-list))
    keymap))

;; keymap-set! KEYMAP STR PROC
(define-public (keymap-set! keymap str proc)
  (if (not (procedure-name proc))
      (format (current-error-port) "No procedure-name for key: ~a: ~a\n"
	      (keymap-name keymap) str))
  (keymap-alist! keymap (assoc-set! (keymap-alist keymap)
				    (keymap-parse str) proc)))

;; EVENT should be a key press
(define-public (keymap-dispatch-key event)
  (view-message #f)
  (set! keymap-event  event)
  ;; gtk-accelerator-parse forces the keyval to lower case, do the same
  (set! keymap-keyval (gdk-keyval-to-lower (gdk-event-keyval event)))
  (set! keymap-mods   (gdk-event-state event))
  (crunch-keymap-mods)

  (let ((proc (keymap-dispatch-find keymap-keyval keymap-mods)))
    (if proc
	(proc))
    ;; (if (not proc)
    ;; 	(format #t "unbound key: ~a ~a ~a\n"
    ;; 		keymap-keyval (gdk-keyval-name keymap-keyval) keymap-mods))
    ))

;; EVENT should be a button press
(define-public (keymap-dispatch-button widget event)
  ;; ignore 2button-press or 3button-press
  (if (eq? 'button-press (gdk-event-type event))
      (begin
	(view-message #f)
	(set! keymap-widget widget)
	(set! keymap-event  event)
	(set! keymap-keyval (c-gdk-button-to-keyval
			     (gdk-event-button event)))
	(set! keymap-mods   (gdk-event-state event))
	(crunch-keymap-mods)

	(and-let* ((proc (keymap-dispatch-find keymap-keyval keymap-mods)))
	  (proc)))))

(define (crunch-keymap-mods)
  (set! keymap-mods (list-copy keymap-mods))

  ;; lock-mask is ignored, except for alphabetic keys, in which case it
  ;; toggles shift-mask
  (if (memq 'lock-mask keymap-mods)
      (begin
	(set! keymap-mods (delq! 'lock-mask keymap-mods))
	(if (c-gdk-keyval-alpha? keymap-keyval)
	    (set! keymap-mods (lset-xor! eq? keymap-mods '(shift-mask))))))

  ;; look only at the usual accelerator modifiers
  ;; in particular ignore num-lock, which is probably mod2
  (set! keymap-mods
	(lset-intersection! eq? keymap-mods
			    (gtk-accelerator-get-default-mod-mask)))

  ;; same order as in the keymaps
  (set! keymap-mods (sort keymap-mods symbol<?)))

