;; Help/Keys dialog.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart keys)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (gtk gtk)
  #:use-module (chart cross)
  #:use-module (chart extra-gtk)
  #:use-module (chart keymap)
  #:use-module (chart line-edit)
  #:use-module (chart misc))

;; modules loaded for the side effect of defining keymaps, make them look
;; used to guile-lint
(display (list cross-keymap line-edit-keymap)
	 (%make-void-port "w"))


;;-----------------------------------------------------------------------------
;; GtkAccelGroup extras

;; compare two GtkModifierType lists, sorting them in the order the
;; modifiers are given by `gtk-accelerator-name'
(define (c-gtk-accelerator-mods<? x y)
  (let ((diff (find (lambda (mask)
		      (not (eq? (memq mask x) (memq mask y))))
		    '(shift-mask control-mask mod1-mask mod2-mask
				 mod3-mask mod4-mask mod5-mask))))
    (and diff
	 (memq diff y))))

(define (c-gtk-accelerator<? key1 mods1 key2 mods2)
  (if (not (= key1 key2))
      (c-gdk-keyval<? key1 key2)
      (c-gtk-accelerator-mods<? mods1 mods2)))


;;-----------------------------------------------------------------------------
;; descriptions

(define (procedure-description proc)

  (case (procedure-name proc)

    ((chart-redraw)
     (_ "Force a redraw."))
    ((download-update-current)
     (_ "Download data update for current chart."))
    ((zoom-in-x)
     (_ "Zoom in horizontally."))
    ((zoom-out-x)
     (_ "Zoom out horizontally."))
    ((zoom-in-y)
     (_ "Zoom in vertically."))
    ((zoom-out-y)
     (_ "Zoom out vertically."))
    ((zoom-in-y-lower)
     (_ "Zoom in just the lower window."))
    ((zoom-out-y-lower)
     (_ "Zoom out just the lower window."))

    ((scroll-step-left)
     (_ "Move the chart left by a step."))
    ((scroll-page-left)
     (_ "Move the chart left by a page."))
    ((scroll-step-right)
     (_ "Move the chart right by a step."))
    ((scroll-page-right)
     (_ "Move the chart right by a page."))
    ((scroll-end-left)
     (_ "Move to the left end of the chart."))
    ((scroll-end-right)
     (_ "Move to the right end of the chart."))
    ((scroll-step-up)
     (_ "Move the chart up by a step."))
    ((scroll-page-up)
     (_ "Move the chart up by a page."))
    ((scroll-step-down)
     (_ "Move the chart down by a step."))
    ((scroll-page-down)
     (_ "Move the chart down by a page."))
    ((scroll-drag-start-button)
     (_ "Move chart by dragging."))

    ((chart-next-list)
     (_ "Go to the start of the next symbol list."))
    ((chart-prev-list)
     (_ "Go to the start of the previous symbol list."))

    ((cross-disable)
     (_ "Disable cross display"))
    ((cross-drag-button)
     (_ "Enable the cross, while the button is held down."))
    ((cross-enable)
     (_ "Enable cross display"))
    ((cross-toggle)
     (_ "Toggle cross display"))

    ((indicator-menu-popup)
     (_ "Popup an indicator selection menu."))

    ((line-edit-delete)
     (_ "Delete the line being edited."))
    ((line-edit-abort)
     (_ "Abort a line edit."))
    ((line-edit-or-new-button)
     (_ "Edit a line, or start a new one."))
    ((line-edit-toggle-horizontal)
     (_ "Toggle the line being edited between horizontal and sloping."))
    ((line-edit-toggle-alert)
     (_ "Toggle a line to an alert level, or toggle above/below."))

    ((lasso-abort)
     (_ "Abort a lasso zoom."))
    ((lasso-start-button)
     (_ "Zoom to region selected with a lasso."))

    ((diagnostics-dialog-popup)
     (_ "Popup a dialog with some development diagnostics."))

    ((data-edit-abort)
     (_ "Abort a data edit."))

    (else
     "")))

(define-public (keymap-text)
  (define (keymap-entry-text elem)
    (let* ((keyval      (caar elem))
	   (modifiers   (cdar elem))
	   (proc        (cdr  elem))
	   (symbol      (procedure-name proc))
	   (symbol-name (if symbol
			    (symbol->string symbol)
			    (string-append "<" (_ "unknown") ">")))
	   (docstr      (procedure-description proc)))
      (string-append (gtk-accelerator-name keyval modifiers)
		     " : " symbol-name "\n"
		     "        " docstr "\n")))
  (define (keymap-entry<? x y)
    (c-gtk-accelerator<? (caar x) (cdar x)
			 (caar y) (cdar y)))

  (string-concatenate-reverse
   (map (lambda (keymap)
	  (string-append
	   (keymap-name keymap) ":\n\n"
	   (string-concatenate/shared (map! keymap-entry-text
					    (sort (keymap-alist keymap)
						  keymap-entry<?)))
	   "\n"))
	keymap-list)))


;;-----------------------------------------------------------------------------
;; keymap dialog, created on demand then re-used

(define dialog #f)

(define (init)
  (set! dialog (gtk-dialog-new-with-buttons (c-make-window-title (_ "Keys"))
					    #f '()
					    "gtk-close" gtk-response-close))

  (let* ((scrolled (gtk-scrolled-window-new))
	 (textbuf  (gtk-text-buffer-new #f))
	 (textview (gtk-text-view-new-with-buffer textbuf)))

    (define (update-textbuf)
      (gtk-text-buffer-set-text textbuf (keymap-text)))

    (c-gtk-dialog-hide-on-delete dialog)
    (gtk-signal-connect dialog "response"
      (lambda (id)
	(cond ((= id gtk-response-close)
	       (gtk-widget-hide dialog)))))

    ;; text is updated only when mapped (to save unnecessary work),
    ;; and only when idle (to save updates for multiple keymap changes)
    (let ((old-keymap-set! keymap-set!))
      (set! keymap-set!
	    (lambda args
	      (apply old-keymap-set! args)
	      (if (memq 'mapped (gtk-widget-flags dialog))
		  (c-gtk-idle-once update-textbuf)))))

    (gtk-signal-connect dialog "map_event"
      (lambda (event)
	(c-gtk-idle-once update-textbuf)))
    (gtk-text-buffer-set-text textbuf "")

    (gtk-scrolled-window-set-policy scrolled 'never 'always)
    (gtk-text-view-set-wrap-mode textview 'word)
    (c-gtk-widget-set-usize-units textview '(em . 50) '(height . 20))
    (gtk-container-add scrolled textview)
    (gtk-box-pack-start (gtk-dialog-vbox dialog) scrolled #t #t 0))

  ;; once-only, and gc
  (set! init noop))

(define-public (keys-popup)
  (init)
  (c-gtk-widget-show-all-or-raise dialog))
