;; Klinger volume oscillator.

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart kvo)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart macd)
  #:use-module (chart misc))


;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/1297/TradersTips/Tips9712.html
;;     Traders Tips December 1997 formulas.
;;
;; http://trader.online.pl/ELZ/t-i-Klinger_Volume_Oscillator.html
;;     Repeat of EasyChart section of traders tips above.
;;
;; http://www.prophet.net/learn/taglossary.jsp?index=K
;;     Sample chart of intel (INTC) Nov02-Feb03, KVO and also histogram.
;; http://www.prophet.net/analyze/popglossary.jsp?studyid=KVO
;;     Same chart of INTC, but just the KVO not the histogram.


;; Return a procedure which calculates the volume force for the Klinger
;; oscillator
;;
;; Each call (PROC high low close volume) enters a new point into the
;; window, and the return is the VF up to (and including) that point, or #f
;; if not enough points yet.  CLOSE must be a number, but HIGH and/or LOW
;; can be #f, in which case the CLOSE is used for those.
;;
;; A calculation needs to be warmed up by starting at a point where the
;; trend value (typical price direction) changes.
;; `klinger-volume-force-start-pos' will search back for such a point where
;; TARGET can be a correct fully calculated point.
;;
(define-public (klinger-volume-force-calc-proc)

  (define prev-hlc   #f)
  (define prev-trend #f)
  (define prev-dm    0)
  (define cm         0)

  (lambda (high low close volume)
    (or high   (set! high   close))
    (or low    (set! low    close))
    (or volume (set! volume 0))

    (let* ((hlc   (+ high low close))
	   (trend (if prev-hlc
		      (if (> hlc prev-hlc)
			  1
			  -1)
		      0))
	   (dm    (- high low)))
      (if (eqv? trend prev-trend)
	  (set! cm (+ cm dm))
	  (set! cm (+ dm prev-dm)))
      (set! prev-hlc   hlc)
      (set! prev-trend trend)
      (set! prev-dm    dm)
      (if (zero? cm)
	  0
	  (* volume trend (abs (- (* 2.0 (/ dm cm)) 1)))))))

(define-public (klinger-volume-force-start-pos series target)
  (let* ((after-tp    #f)
	 (after-trend #f))
    (let more ((pos (min target (series-hi series))))
      (if (<= pos (series-lo series))
	  (series-lo series)  ;; hit beginning of series

	  (let* ((array (series-array series pos pos))
		 (close (array-ref array pos 3)))
	    (if (not close)
		(more (1- pos))

		(let* ((tp    (+ (or (array-ref array pos 1) close)
				 (or (array-ref array pos 2) close)
				 close))
		       (trend (and after-tp
				   (if (> after-tp tp) 1 -1))))
		  (if (and trend after-trend
			   (not (eqv? trend after-trend)))
		      ;; and one before trend change
		      (series-prev-n-closes-pos series #:before pos #:count 1)

		      (begin
			(set! after-tp tp)
			(set! after-trend trend)
			(more (1- pos)))))))))))



(define-public (kvo-calc-proc fast slow)
  (define vf-proc      (klinger-volume-force-calc-proc))
  (define fast-proc    (ema-calc-proc fast))
  (define slow-proc    (ema-calc-proc slow))

  (lambda (high low close volume)
    (and-let* ((vf (vf-proc high low close volume)))
      (let ((fast (fast-proc vf))
	    (slow (slow-proc vf)))
	(and fast slow
	     (- fast slow))))))


;;-----------------------------------------------------------------------------
;; VF series

(define (klinger-volume-force-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (start  (klinger-volume-force-start-pos parent lo))
	 (array  (series-array parent start hi))
	 (proc   (klinger-volume-force-calc-proc)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 1)    ;; high
		(array-ref array i 2)    ;; low
		(array-ref array i 3)    ;; close
		(array-ref array i 4)))) ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (proc (array-ref array i 1)  ;; high
				      (array-ref array i 2)  ;; low
				      (array-ref array i 3)  ;; close
				      (array-ref array i 4)) ;; volume
		      i)))))

(define-public (klinger-volume-force-series series)
  (ohlcv-required "Klinger volume force" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "KVO volume force"))
    (series-datatype!  new-series klinger-volume-force-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   klinger-volume-force-fill
				   10)) ;; estimated warmup
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public klinger-volume-force-datatype
  (datatype-new!
   #:key      'klinger-volume-force
   #:name     (_ "KVO volume force")
   #:priority -10
   #:input    datatype-ohlcv
   #:units    'klinger-volume-force
   #:hlines   '(0)
   #:manual   (_ "CONTEXT(manual-node): Klinger Volume Oscillator")
   #:proc     klinger-volume-force-series))



;;-----------------------------------------------------------------------------
;; KVO series

(define-public (kvo-series series fast slow trigger histogram)
  (ohlcv-required "KVO" series)
  (let ((new-series (macd-series (klinger-volume-force-series series)
				 fast slow trigger histogram)))
    (series-name!      new-series (format #f (_ "KVO ~a,~a,~a")
					  fast slow trigger))
    (series-datatype!  new-series kvo-datatype)
    new-series))

(define-public kvo-datatype
  (datatype-new!
   #:key       'kvo
   #:name      (_ "KVO - Klinger Volume Oscillator")
   #:units     'kvo
   #:input     datatype-ohlcv
   #:linestyle 'macd
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Klinger Volume Oscillator")
   #:proc      kvo-series
   #:closes-apply-arg '(0)
   #:parameter-list
   (list (datatype-param-new
	  #:key     'kvo-fast
	  #:name    (_ "Fast days")
	  #:type    'integer
	  #:minimum 1
	  #:default 34)
	 (datatype-param-new
	  #:key     'kvo-slow
	  #:name    (_ "Slow days")
	  #:type    'integer
	  #:minimum 1
	  #:default 55)
	 (datatype-param-new
	  #:name    (_ "Trigger")
	  #:key     'kvo-trigger-days
	  #:default 13
	  #:type    'integer
	  #:minimum 1)
	 (datatype-param-new
	  #:name    (_ "Histogram")
	  #:key     'kvo-histogram
	  #:type    'boolean
	  #:default #f))))
