;; Laguerre Filter moving average.

;; Copyright 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart laguerre)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;;http://www.mesasoftware.com/Papers/Time%20Warp%20Without%20Space%20Travel.exe
;;     Paper by John Elhers (self-extracting zip archive of MS word).
;;
;; http://www.mesasoftware.com/Seminars/TSWorld05.ppt
;;     Powerpoint summary by John Ehlers of several of his and other averages.
;;     View in google,
;;     * A Laguerre filter warps time in the filter coefficients
;;       - Enables extreme smoothing with just a few filter terms
;;     * A NonLinear Laguerre filter measures the difference between the
;;       current price and the last computed filter output.
;;       - Objective is to drive this "error" to zero
;;       - The "error", normalized to the error range over a selected period
;;         is the alpha of the Laguerre filter
;;


;; Return a procedure which calculates a laguerre filter given ALPHA
;; smoothing.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the laguerre filter result up to (and including) that value, or #f if
;; not enough points yet.
;;
;; A laguerre filter is in theory influenced by all preceding data, but
;; `laguerre-warmup-count' below is designed to determine a warmup count.
;; After calling PROC with (laguerre-warmup-count ALPHA) many values, the
;; next call will have an omitted weight of no more than 0.1% of the total.
;; Omitting 0.1% should be negligable, unless past values are ridiculously
;; bigger than recent ones.
;;
(define-public (laguerre-calc-proc alpha)
  (let ((f         (- 1 alpha))
	(a-sum     0.0)
	(a-div     0.0)
	(a-prev    #f)
	(a-pdiv    #f)
	(b-sum     0.0)
	(b-div     0.0)
	(b-prev    #f)
	(b-pdiv    #f)
	(c-sum     0.0)
	(c-div     0.0)
	(c-prev    #f)
	(c-pdiv    #f)
	(d-sum     0.0)
	(d-div     0.0)
	(weight-ok #f))

    (lambda (x)
      (set! a-prev a-sum)
      (set! a-pdiv a-div)
      (set! a-sum (+ (* x alpha) (* a-sum f)))
      (set! a-div (+ alpha       (* a-div f)))

      (set! b-prev b-sum)
      (set! b-pdiv b-div)
      (set! b-sum (+ (- (* a-sum f)) a-prev (* b-sum f)))
      (set! b-div (+ (- (* a-div f)) a-pdiv (* b-div f)))

      (set! c-prev c-sum)
      (set! c-pdiv c-div)
      (set! c-sum (+ (- (* b-sum f)) b-prev (* c-sum f)))
      (set! c-div (+ (- (* b-div f)) b-pdiv (* c-div f)))

      (set! d-sum (+ (- (* c-sum f)) c-prev (* d-sum f)))
      (set! d-div (+ (- (* c-div f)) c-pdiv (* d-div f)))

      (or weight-ok
	  (set! weight-ok (> (/ (+ a-div b-div b-div c-div c-div d-div)
				6.0)
			     indicator-minimum-fraction)))

      (and weight-ok
	   (/ (+ (/ a-sum a-div)
		 (* 2 (/ b-sum b-div))
		 (* 2 (/ c-sum c-div))
		 (/ d-sum d-div))
	      6.0)))))

(define-public (laguerre-warmup-count alpha)
  (laguerre-omitted-search alpha 0.001))

(define-public (laguerre-omitted-search alpha target)
  (define f (- 1 alpha))
  (bsearch-first-true (lambda (i)
			(<= (laguerre-omitted f i) target))
		      (ema-alpha->count alpha)))

;; `laguerre-omitted' returns the fraction (between 0 and 1) of weight
;; omitted by stopping at the f^k term, ie. taking the first k+1 terms.
;;
;; This is hairy polynomials in k on factors f^k through f^(k+4).  The
;; coefficients are established in devel/closed-form.scm.  Maybe something
;; simpler is possible.
;;
(define-public (laguerre-omitted f k)
  (if (< k 2)
      1
(* (expt f (- k 2))
   (+ (* k (+ -1/36 (* k k 1/36)))
    (* f             ;; f^(k-1)
     (+ (* k (+ 1/4 (* k 1/4)))
      (* f             ;; f^(k+0)
       (+ (+ 5/6 (* k (+ 3/4 (* k (+ -1/6 (* k -1/12))))))
        (* f             ;; f^(k+1)
         (+ (+ 1/2 (* k (1- (* k -1/2))))
          (* f             ;; f^(k+2)
           (+ (+ -2/3 (* k (+ -5/12 (* k (+ 1/3 (* k 1/12))))))
            (* f             ;; f^(k+3)
             (+ (+ 1/2 (* k (+ 3/4 (* k 1/4))))
              (* f             ;; f^(k+4)
               (+ (+ -1/6 (* k (+ -11/36 (* k (+ -1/6 (* k -1/36))))))
              ))))))))))))))))


(define (laguerre-fill series new-array lo hi)
  (receive-list (alpha warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (laguerre-calc-proc alpha)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (proc this) i))))))

(define-public (laguerre-series series alpha)
  (or (within-range? 0 1 alpha)
      (error "Laguerre: alpha outside range 0 to 1:" alpha))
  (set! alpha (max alpha 0.00001))
  (set! alpha (min alpha 0.99999))

  (let ((new-series (series-copy series))
	(warmup     (laguerre-warmup-count alpha)))
    (series-name!     new-series (format #f "~a ~a" (_ "Laguerre")
					 (locale-number-min-decimals alpha 1)))
    (series-datatype! new-series laguerre-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list alpha warmup))
    (series-require!  new-series (indicator-require-parts laguerre-fill
							  warmup))
    (series-array!    new-series #f)
    new-series))

(define laguerre-datatype
  (datatype-new!
   #:key    'laguerre
   #:name   (_ "Laguerre Filter")
   #:manual (_ "CONTEXT(manual-node): Laguerre Filter")
   #:units  #f
   #:proc   laguerre-series
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "Alpha")
	  #:key      'laguerre-alpha
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:maximum  1
	  #:default  0.2
	  #:step     0.1))))


;;-----------------------------------------------------------------------------

;; Return a procedure which calculates a laguerre filter step number N of
;; the given ALPHA smoothing.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the laguerre filter result up to (and including) that value, or #f if
;; not enough points yet.
;;
(define-public (laguerre-LN-calc-proc N alpha)
  (let ((f        (- 1 alpha))
	(sum-vec  (make-vector (1+ N) 0.0))
	(div-vec  (make-vector (1+ N) 0.0))
	(prev-vec (make-vector (1+ N) #f))
	(pdiv-vec (make-vector (1+ N) #f)))

    (lambda (x)
      (vector-set! prev-vec 0 (vector-ref sum-vec 0))
      (vector-set! pdiv-vec 0 (vector-ref div-vec 0))
      (vector-set! sum-vec 0 (+ (* alpha x)
				(* f (vector-ref sum-vec 0))))
      (vector-set! div-vec 0 (+ alpha
				(* f (vector-ref div-vec 0))))

      (do ((i 1 (1+ i)))
	  ((> i N))
	(vector-set! prev-vec i (vector-ref sum-vec i))
	(vector-set! pdiv-vec i (vector-ref div-vec i))
	(vector-set! sum-vec i (+ (- (* f (vector-ref sum-vec (1- i))))
				  (vector-ref sum-vec (1- i))
				  (* f (vector-ref sum-vec i))))
	(vector-set! div-vec i (+ (- (* f (vector-ref div-vec (1- i))))
				  (vector-ref div-vec (1- i))
				  (* f (vector-ref div-vec i)))))

      (and (>= (vector-ref div-vec N) indicator-minimum-fraction)
	   (/ (vector-ref sum-vec N) 
	      (vector-ref div-vec N))))))

(define-public (laguerre-LN-warmup-count N alpha)
  (* 5 (1+ N) (ema-alpha->count alpha)))


(define (laguerre-LN-fill series new-array lo hi)
  (receive-list (N alpha warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (laguerre-LN-calc-proc N alpha)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (proc this) i))))))

(define-public (laguerre-LN-series series N alpha)
  (or (within-range? 0 1 alpha)
      (error "Laguerre: alpha outside range 0 to 1:" alpha))
  (set! alpha (max alpha 0.00001))
  (set! alpha (min alpha 0.99999))

  (let ((new-series (series-copy series))
	(warmup     (laguerre-LN-warmup-count N alpha)))
    (series-name!     new-series (format #f (_ "Laguerre L~a ~a")
					 N
					 (locale-number-min-decimals alpha 1)))
    (series-datatype! new-series laguerre-LN-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list N alpha warmup))
    (series-require!  new-series (indicator-require-parts laguerre-LN-fill
							  warmup))
    (series-array!    new-series #f)
    new-series))

(define laguerre-LN-datatype
  (datatype-new!
   #:key      'laguerre-LN
   #:name     (_ "Laguerre Filter Part")
   #:manual   (_ "CONTEXT(manual-node): Laguerre Filter")
   #:priority -10
   #:units    #f
   #:proc     laguerre-LN-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "N")
	  #:key     'laguerre-LN
	  #:type    'integer
	  #:minimum 0
	  #:default 1)
	 (datatype-param-new
	  #:name     (_ "Alpha")
	  #:key      'laguerre-alpha ;; shared
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:maximum  1
	  #:default  0.2
	  #:step     0.1))))
