;; Lasso for zoom.

;; Copyright 2003, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart lasso)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (srfi srfi-1)
  #:autoload   (chart cross)       (cross-disable)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart keymap)
  #:use-module (chart misc)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; Gdk drawing stuff

;; draw a rectangle outline with the given corner coordinates
(define-public (c-gdk-draw-rectangle-corners drawable gc x1 y1 x2 y2)
  (gdk-draw-rectangle drawable gc #f
		      (min x1 x2)
		      (min y1 y2)
		      (abs (- x1 x2))
		      (abs (- y1 y2))))


;;-----------------------------------------------------------------------------
;; GtkWidget stuff

;; return X restricted to the range of x coords in WIDGET
(define (c-gtk-widget-restrict-x widget x)
  (let ((win (gtk-widget-window lasso-graph)))
    (max 0 (min x (1- (c-gdk-drawable-get-width win))))))

;; return Y restricted to the range of y coords in WIDGET
(define (c-gtk-widget-restrict-y widget y)
  (let ((win (gtk-widget-window lasso-graph)))
    (max 0 (min y (1- (c-gdk-drawable-get-height win))))))


;;-----------------------------------------------------------------------------
;; cross state

(let-toplevel ((state           #f)
	       (old-cross-state #f))
  (notify-connect 'cross-update (lambda (new-state)
				  (set! state new-state)))

  ;; do a `cross-disable', but avoid loading the cross module if it isn't
  ;; already in use
  (define-toplevel (lasso-cross-disable)
    (set! old-cross-state state)
    (if state
	(cross-disable)))

  ;; restore the cross setting to what it was before `lasso-cross-disable'
  (define-toplevel (lasso-cross-restore)
    (if old-cross-state
	(cross-enable))))


;;-----------------------------------------------------------------------------
;; graph stuff

;; zoom GRAPH to show from dates D1 to D2 and prices P1 to P2
(define (graph-zoom-range graph d1 d2 p1 p2)
  (set! p1 (graph-log graph p1))
  (set! p2 (graph-log graph p2))
  (let* ((height (c-gdk-drawable-get-height (gtk-widget-window graph)))
	 (page   (abs (- p1 p2)))
	 (vadj   (graph-vadjustment graph)))
    (graph-scale-y-factor! graph (/ height page))
    (c-gtk-adjustment-set-value-centre vadj (arithmetic-mean p1 p2)))

  (let* ((width (c-gdk-drawable-get-width (gtk-widget-window graph)))
	 (page  (abs (- d1 d2 -1)))
	 (view  (component-view graph))
	 (hadj  (view-hadjustment view)))
    (view-scale-x-step! view (floor-exact (/ width page)))
    (c-gtk-adjustment-set-value-centre hadj (arithmetic-mean d1 d2))))


;;-----------------------------------------------------------------------------
;; lasso

(define lasso-colour "white")
(define lasso-cursor (gdk-cursor-new 'left-ptr))

(define lasso-graph  #f)
(define x1 #f)
(define y1 #f)
(define x2 #f)
(define y2 #f)
(define lasso-gc #f)

(define (lasso-xor)
  (if lasso-graph
      (let* ((win (gtk-widget-window lasso-graph)))

	(if (not lasso-gc)
	    (let ((c  (gdk-color-new))
		  (bg (gtk-style-bg (gtk-widget-get-style lasso-graph) 'normal)))
	      (set! lasso-colour (gdk-color-intern lasso-colour))
	      (gdk-color-set-pixel c (logxor (gdk-color-pixel lasso-colour)
					     (gdk-color-pixel bg)))
	      (set! lasso-gc (c-gdk-gc-get win
					   #:foreground c
					   #:dashed #t
					   #:function 'xor))))

	(c-gdk-draw-rectangle-corners win lasso-gc x1 y1 x2 y2))))

(define (lasso-move x y)
  (lasso-xor)   ;; erase old
  (set! x2 (c-gtk-widget-restrict-x lasso-graph x))
  (set! y2 (c-gtk-widget-restrict-y lasso-graph y))
  (lasso-xor))  ;; draw new

;; start a lasso zoom from a button press in `keymap-event'
(define-public (lasso-start-button)
  (set! lasso-graph draw)

  (or (object-property lasso-graph 'lasso-init-done)
      (begin
	(set-object-property! lasso-graph 'lasso-init-done #t)

	(gtk-signal-connect lasso-graph "motion_notify_event"
	  (lambda (event)
	    (if lasso-graph
		(let ((pointer (gdk-window-get-pointer
				(gtk-widget-window lasso-graph))))
		  (lasso-move (inexact->exact (second pointer))
			      (inexact->exact (third  pointer)))))))

	(gtk-signal-connect lasso-graph "button_release_event"
	  (lambda (event)
	    (if lasso-graph
		(begin
		  (lasso-move (inexact->exact (gdk-event-x event))
			      (inexact->exact (gdk-event-y event)))
		  (lasso-done)))))))

  (lasso-cross-disable)
  (gdk-window-set-cursor (gtk-widget-window lasso-graph) lasso-cursor)

  (set! x1 (inexact->exact (gdk-event-x keymap-event)))
  (set! y1 (inexact->exact (gdk-event-y keymap-event)))

  (set! x2 x1)
  (set! y2 y1)
  (lasso-xor)) ;; initial draw

(define (lasso-done)
  (gdk-window-set-cursor (gtk-widget-window lasso-graph) #f)
  (lasso-xor) ;; clear
  (lasso-cross-restore)
  (graph-zoom-range lasso-graph
		    (graph-x->date lasso-graph x1)
		    (graph-x->date lasso-graph x2)
		    (graph-y->price lasso-graph y1)
		    (graph-y->price lasso-graph y2))
  (set! lasso-graph #f))

(define (lasso-abort)
  (if lasso-graph
      (begin
	(gdk-window-set-cursor (gtk-widget-window lasso-graph) #f)
	(lasso-xor) ;; clear
	(lasso-cross-restore)
	(set! lasso-graph #f))))

(define-public lasso-keymap (keymap-new (_ "While lassoing to zoom")
					(lambda () lasso-graph)
					#:propagate #f))

(keymap-set! lasso-keymap "Escape" lasso-abort)
