;; Latest quotes.

;; Copyright 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart latest)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format) (format)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart http)
  #:use-module (chart misc))


;; eg. "/home/username/Chart/cache/latest"
(define-public latest-directory
  (g-build-filename cache-directory "latest"))

;; eg. "/home/username/Chart/cache/latest/^GSPC"
(define-public (latest-filename symbol)
  (or (filename-valid? symbol)
      (error "bad symbol:" symbol))
  (g-build-filename latest-directory symbol))


;; return how many decimal places on the number in STR
;; if LIMIT is not #f, it's a limit on the return, the return is no greater
;; than LIMIT unless there's non-zero digits beyond that point
;;
(define (string-decimals-limit str limit)
  (and str
       (let ((dot (string-index str #\.)))
	 (if (not dot)
	     0
	     (let* ((end  (string-length str))
		    (decs (- end dot 1)))
	       (if (and limit
			(> decs limit))
		   (let ((lend (+ (- end decs) limit)))
		     (set! end (string-skip-right str #\0 lend))
		     (set! end (if end (1+ end) lend))
		     (set! decs (- end dot 1))))
	       decs)))))

(define (adate->year a)
  (first (adate->ymd a)))


;;-----------------------------------------------------------------------------
;; latest records, including I/O

(let-toplevel ((rec-type (make-record-type
			  "latest"
			  '(fetch-time  ;; per current-time
			    symbol
			    name
			    last-adate
			    last-time   ;; seconds since midnight, or #f
			    open        ;;
			    high
			    low
			    last
			    change      ;; last - prevclose, #f if unknown
			    volume      ;; possibly zero, #f if unknown
			    quote-adate
			    quote-time  ;; seconds since midnight, or #f
			    bid         ;; #f if unknown or none
			    offer       ;; #f if unknown or none
			    note        ;; string or #f
			    source      ;; symbol, for now
			    decimals    ;; for prices
			    contract-mdate ;; futures
			    etag
			    last-modified
			    ))))
  (define rec-new (record-constructor rec-type))

  (define-public latest-fetch-time  (rec-accessor rec-type 'fetch-time))
  (define-public latest-symbol      (rec-accessor rec-type 'symbol))
  (define-public latest-name        (rec-accessor rec-type 'name))
  (define-public latest-quote-adate (rec-accessor rec-type 'quote-adate))
  (define-public latest-quote-time  (rec-accessor rec-type 'quote-time))
  (define-public latest-bid         (rec-accessor rec-type 'bid))
  (define-public latest-offer       (rec-accessor rec-type 'offer))
  (define-public latest-last-adate  (rec-accessor rec-type 'last-adate))
  (define-public latest-last-time   (rec-accessor rec-type 'last-time))
  (define-public latest-open        (rec-accessor rec-type 'open))
  (define-public latest-high        (rec-accessor rec-type 'high))
  (define-public latest-low         (rec-accessor rec-type 'low))
  (define-public latest-last        (rec-accessor rec-type 'last))
  (define-public latest-change      (rec-accessor rec-type 'change))
  (define-public latest-volume      (rec-accessor rec-type 'volume))
  (define-public latest-note        (rec-accessor rec-type 'note))
  (define-public latest-source      (rec-accessor rec-type 'source))
  (define-public latest-decimals    (rec-accessor rec-type 'decimals))
  (define-public latest-contract-mdate (rec-accessor rec-type 'contract-mdate))
  (define-public latest-etag        (rec-accessor rec-type 'etag))
  (define-public latest-last-modified (rec-accessor rec-type 'last-modified))

  (define-public latest-symbol!     (rec-modifier rec-type 'symbol))
  (define-public latest-fetch-time! (rec-modifier rec-type 'fetch-time))
  (define-public latest-open!       (rec-modifier rec-type 'open))
  (define-public latest-high!       (rec-modifier rec-type 'high))
  (define-public latest-low!        (rec-modifier rec-type 'low))
  (define-public latest-last!       (rec-modifier rec-type 'last))
  (define-public latest-change!     (rec-modifier rec-type 'change))
  (define-public latest-bid!        (rec-modifier rec-type 'bid))
  (define-public latest-offer!      (rec-modifier rec-type 'offer))
  (define-public latest-decimals!   (rec-modifier rec-type 'decimals))

  (define-public (latest-last-tdate latest)
    (and (latest-last-adate latest)
	 (adate->tdate-ceil (latest-last-adate latest))))
  (define-public (latest-quote-tdate latest)
    (and (latest-quote-adate latest)
	 (adate->tdate-ceil (latest-quote-adate latest))))

  (define*-public (latest-new #:key
			      (fetch-time (current-time))
			      symbol name

			      ;; accept quote-tdate in old cached records
			      quote-tdate
			      (quote-adate (and quote-tdate
						(tdate->adate quote-tdate)))
			      quote-time
			      bid offer

			      ;; accept last-tdate in old cached records
			      last-tdate
			      (last-adate (and last-tdate
					       (tdate->adate last-tdate)))
			      last-time
			      open high low
			      last
			      change prev
			      volume

			      note source
			      (decimals 0)
			      limit-decimals
			      contract-mdate

			      headers
			      (etag (and headers
					 (http-headers-get headers "ETag")))
			      (last-modified (and headers
						  (http-headers-get
						   headers "Last-Modified"))))

    (let ((lst (list change bid offer open high low last prev)))
      (if (any string? lst)
	  (begin
	    (set-car! lst (crunch-number change))
	    (set-cdr! lst (map! crunch-price (cdr lst)))

	    (set! decimals
		  (apply max-maybe
			 decimals
			 (map (lambda (str)
				(string-decimals-limit str limit-decimals))
			      lst)))

	    (set! lst (map! (lambda (str)
			      (and str
				   (string->number+decimals str decimals)))
			    lst))

	    (set! change (first   lst))
	    (set! bid    (second  lst))
	    (set! offer  (third   lst))
	    (set! open   (fourth  lst))
	    (set! high   (fifth   lst))
	    (set! low    (sixth   lst))
	    (set! last   (seventh lst))
	    (set! prev   (eighth  lst)))))

    (cond
     ;; if there's no LAST etc, then show PREV as previous tdate
     ((and prev
	   (not (or open high low last)))
      (set! last prev)
      (set! change #f)
      (set! volume #f)
      (set! last-time #f)
      (set! last-adate (and last-adate
			    (adate-prev-tdate last-adate))))

     ;; if no CHANGE, calculate it from PREV
     ((and (not change)
	   prev
	   last)
      (set! change (- last prev))))

    ;; various sources give 0 for no information
    (set! bid   (0->f bid))
    (set! offer (0->f offer))
    (set! open  (0->f open))
    (set! high  (0->f high))
    (set! low   (0->f low))
    (set! last  (0->f last))
    (if (not last)
	(set! change #f))

    (set! volume (crunch-volume volume))
    (if (string? volume)
	(set! volume (string->number-err volume)))

    ;; turn 0 volume to #f if no open/high/low/last/change
    ;; various sources have 0 when no information
    (if (and (or (eqv? 0 volume)
		 (eqv? 0.0 volume))
	     (not (or open high low last change)))
	(set! volume #f))

    ;; forget about last-adate/time if no open/high/low/last/change/volume
    (or open high low last change volume
	(begin
	  (set! last-adate #f)
	  (set! last-time  #f)))

    (let ((latest (rec-new fetch-time
			   symbol
			   name
			   last-adate
			   last-time
			   open
			   high
			   low
			   last
			   change
			   volume
			   quote-adate
			   quote-time
			   bid
			   offer
			   note
			   source
			   decimals
			   contract-mdate
			   etag
			   last-modified)))
      (latest-validate latest)
      latest))

  (define-toplevel (latest-write latest-list)
    (let ((added #f))
      (database-call-with-lock latest-directory
	(lambda ()
	  (for-each
	   (lambda (latest)
	     ;; don't save stuff from the database
	     (if (not (eq? 'database (latest-source latest)))

		 (let ((filename (latest-filename (latest-symbol latest))))
		   (if (not (file-exists? filename))
		       (set! added #t))

		   (database-call-with-output-file filename
		     (lambda (port)
		       (simple-format port ";; ~a latest, generated by Chart  -*- coding: utf-8 -*-\n\n"
				      (latest-symbol latest))
		       (write-list-lines
			(map (lambda (field)
			       (list field ((rec-accessor rec-type field) latest)))
			     (record-type-fields rec-type))
			port))))))

	   latest-list)))

      ;; this is broadcast for the benefit of emacs subprocesses, so they
      ;; can add new stuff to the symbol completions
      (if added
	  (notify-send 'latest-cache-added #f))))

  ;; return latest record, or #f
  (define-public (latest-read symbol)
    (call-with-input-file-or #f (latest-filename symbol)
      (lambda (port)
	(apply latest-new
	       (append-map (lambda (elem)
			     (list (symbol->keyword (first elem))
				   (second elem)))
			   (read port)))))))


;;-----------------------------------------------------------------------------
;; latest fetch, with cache

(define cache (make-weak-value-hash-table 97))

(define (database-last-adate symbol)
  (and-let* ((tdate (database-last-tdate symbol)))
    (tdate->adate tdate)))

;; return latest record
(define-public (latest-get symbol)
  (let ((latest (hash-ref cache symbol)))

    (and-let* ((rlat (latest-read symbol)))
      (if (or (not latest)
	      (> (latest-fetch-time rlat) (latest-fetch-time latest)))
	  (set! latest rlat)))

    (and-let* ((db-adate (database-last-adate symbol))
	       ;; an error record has precedence (otherwise errors would be
	       ;; unseen for anything in the database)
	       (         (not (and latest
				   (eq? 'error (latest-source latest))))))
      (let ((lt-adate (and latest
			   (max-maybe (latest-last-adate latest)
				      (latest-quote-adate latest)))))
	(if (or (not lt-adate)
		;; prefer database over an intermediate of the same date
		(>= db-adate lt-adate))
	    (set! latest (series->latest (database-read-series symbol))))))

    ;; dummy record
    (if (not latest)
	(set! latest (latest-new #:fetch-time 0
				 #:symbol     symbol
				 #:source     'dummy)))

    (hash-set! cache symbol latest)
    latest))

;; return a list of http #:if-none-match and #:if-modified-since args from
;; the latest records of SYMBOL-LIST
;; all of SYMBOL-LIST must have the same source and the same tag string to
;; use the headers
(define-public (latest-tag-args source symbol-list)
  (let* ((latest-list (map latest-get symbol-list)))
    (if (apply eq? source (map latest-source latest-list))
	(let ()
	  (define (list->arg lst)
	    (and (apply equal? lst)
		 (first lst)))
	  (list #:if-none-match     (list->arg (map latest-etag
						    latest-list))
		#:if-modified-since (list->arg (map latest-last-modified
						    latest-list))))
	;; not all same `latest-source'
	'())))


;;-----------------------------------------------------------------------------
;; latest from database series

;; Return the tdate in SERIES on or before START where the last trade
;; occurred, or #f if no such trades.
;;
;; The search is for a non-#f close, the volume doesn't matter.  This is
;; good for futures where a settlement is established even without any
;; trading.
;;
(define (find-last-trade series i)
  (and (>= i (series-lo series))
       (let ((array (series-array series i i)))
	 (if (array-ref array i 3) ;; close
	     i
	     (find-last-trade series (1- i))))))

;; return the mdate of the month indicated in the dividends of SERIES, on or
;; before TDATE
(define (series-dividend-mdate series tdate)
  (and-let* ((pair (find-maximum (lambda (pair)
				   (and (<= (car pair) tdate)
					(pair? (cdr pair))
					(eq? 'month (second pair))
					(car pair)))
				 (series-dividends series))))
    (third pair)))

;; make a latest record out of SERIES
(define (series->latest series)

  ;; prefetch last and prev, allowing for possible 2 public holidays between
  (series-array series (- (series-hi series) 3) (series-hi series))

  (let* ((last-tdate (find-last-trade series (series-hi series)))
	 (prev-tdate (and last-tdate (find-last-trade series (1- last-tdate))))
	 (array      (and last-tdate (series-array series
						   (or prev-tdate last-tdate)
						   last-tdate)))
	 (last       (and last-tdate (array-ref array last-tdate 3)))
	 (prev       (and prev-tdate (array-ref array prev-tdate 3)))
	 (last-mdate (or (and (series-symbol series)
			      (chart-symbol-mdate (series-symbol series)))
			 (series-dividend-mdate series last-tdate)))
	 (notes-list '()))

    ;; if prev-tdate was from a different contract month add "rollover" as a
    ;; note (since the change in this case includes the contango)
    (if (and prev-tdate last-mdate)
	(let ((prev-mdate (or (and (series-symbol series)
				   (chart-symbol-mdate (series-symbol series)))
			      (series-dividend-mdate series prev-tdate))))
	  (if (not (eqv? prev-mdate last-mdate))
	      (set! notes-list (cons (_ "rollover") notes-list)))))

    ;; ex dividend note
    (and-let* ((div (assq-ref (series-dividends series) last-tdate))
	       (    (not (pair? div))))
      (set! notes-list (cons (simple-format #f (_ "ex ~a") div) notes-list)))

    (latest-new #:fetch-time 0 ;; allow immediate retry from data sources
		#:symbol     (series-symbol series)
		#:name       (series-name series)
		#:last-tdate last-tdate
		#:last-time  #f
		#:open       (and last-tdate (array-ref array last-tdate 0))
		#:high       (and last-tdate (array-ref array last-tdate 1))
		#:low        (and last-tdate (array-ref array last-tdate 2))
		#:last       last
		#:change     (and last prev (- last prev))
		#:volume     (and array (array-ref array last-tdate 4))
		#:decimals   (series-decimals series)
		#:contract-mdate last-mdate
		;; ex-dividend, if today
		#:note       (and (not (null? notes-list))
				  (string-join notes-list (_ ", ")))
		#:source     'database)))

;; get a latest record for SYMBOL using only the database information
(define-public (latest-get-database symbol)
  (and-let* ((series (database-read-series symbol)))
    (series->latest series)))


;;-----------------------------------------------------------------------------
;; handlers

(let-toplevel ((rec-type (make-record-type
			  "latest-handler"
			  '(selector-proc
			    handler-proc
			    priority
			    adate-time-proc)))
	       ;; list of records, sorted by PRIORITY
	       (handler-list '()))

  ;; SELECTOR-PROC called (SELECTOR-PROC sym)
  ;;
  ;; HANDLER-PROC called (HANDLER-PROC required-list optional-list callback)
  ;; and it should call (CALLBACK latest-list) with a list of new latest
  ;; records

  (define rec-new (record-constructor rec-type))

  (define rec-selector-proc   (rec-accessor rec-type 'selector-proc))
  (define rec-handler-proc    (rec-accessor rec-type 'handler-proc))
  (define rec-priority        (rec-accessor rec-type 'priority))
  (define rec-adate-time-proc (rec-accessor rec-type 'adate-time-proc))

  ;; compare priorities of records X and Y
  (define (rec-priority> x y)
    (> (rec-priority x) (rec-priority y)))

  (define*-public (latest-handler! #:key
				   (selector (error "no latest selector proc"))
				   (handler  (error "no latest handler proc"))
				   (priority 0)
				   (adate-time #f))
    (set! handler-list
	  (stable-sort (append handler-list
			       (list (rec-new selector handler priority adate-time)))
		       rec-priority>)))


  ;; request symbols as the result of an explicit use command, so refetch no
  ;; matter when the last fetch was done
  ;;
  (define*-public (latest-request-explicit symbol-list
					   #:optional (extra-list '()))
    (set! symbol-list (remove latest-in-progress? symbol-list))
    (set! extra-list  (remove latest-in-progress? extra-list))

    (for-each symbol-setups symbol-list)
    (for-each symbol-setups extra-list)

    (latest-mark-in-progress symbol-list)

    (let ((skipped-list '()))

      (for-each
       (lambda (rec)
	 (c-idle)
	 (receive (this-symbol-list next-symbol-list)
	     (partition (rec-selector-proc rec) symbol-list)
	   (receive (this-extra-list next-extra-list)
	       (partition (rec-selector-proc rec) extra-list)

	     (if (rec-adate-time-proc rec)
		 (let ()
		   ;; return true if the latest-get record for SYMBOL is
		   ;; already the newest available date/time, with the
		   ;; available time from the handler ADATE-TIME-PROC
		   ;;
		   (define (already-newest? symbol)
		     (let* ((latest (latest-get symbol)))
		       (and (latest-last-adate latest)
			    ;; no time is the closing price for today, trumping
			    ;; any avail-time on the same date
			    (let ((last-time (or (latest-last-time latest)
						 #,(hms->seconds 24 0 0))))
			      (receive-list (avail-adate avail-time)
				  ((rec-adate-time-proc rec) symbol)
				(two<= avail-adate
				       (or avail-time
					   #,(hms->seconds 24 0 0))
				       (latest-last-adate latest)
				       last-time))))))

		   (receive (drop keep)
		       (partition already-newest? this-symbol-list)
		     (set! this-symbol-list keep)
		     (set! skipped-list (append skipped-list drop)))

		   (receive (drop keep)
		       (partition already-newest? this-extra-list)
		     (set! this-extra-list keep)
		     (set! skipped-list (append skipped-list drop)))))


	     ;; initiate all handlers in parallel
	     (if (not (null? this-symbol-list))
		 (c-main-enq! run-handler (rec-handler-proc rec)
			      this-symbol-list this-extra-list))
	     (set! symbol-list next-symbol-list)
	     (set! extra-list next-extra-list))))
       handler-list)

      (latest-fetch-fail skipped-list))

    ;; symbols with no handler
    (if (not (null? symbol-list))
	(latest-handler-callback (map latest-new-unknown symbol-list))))

  ;; return a record from the latest cache (not the database), from today,
  ;; or #f if unknown
  ;;
  ;; if there's no cached record or it's not from today (the current
  ;; avaiable day) then do a download, throwing errors for failure
  ;;
  (define-public (latest-get-today symbol)
    (let ((rec (find (lambda (rec)
		       ((rec-selector-proc rec) symbol))
		     handler-list)))
      (or rec (error "No latest handler for" symbol))

      (let ((latest (latest-read symbol)))
	(if (or (not latest)
		(not (latest-quote-adate latest)) ;; eg. an error record
		(receive-list (adate time)
		    ((rec-adate-time-proc rec) symbol)
		  (< (or (latest-quote-adate latest)
			 (latest-last-adate latest)) adate)))
	    ((rec-handler-proc rec) (list symbol) '()
	     latest-handler-callback)))
      (latest-read symbol))))

;; return a latest record for an unknown SYMBOL
(define-public (latest-new-unknown symbol)
  (latest-new #:symbol symbol
	      #:note   (_ "unknown symbol")
	      #:source 'dummy))

;; request symbols for general display, no need to fetch again if did so a
;; short time ago
;;
(define*-public (latest-request-symbols symbol-list
					#:optional (extra-list '()))
  (let ((target-time (- (current-time) 120)))
    (define (want? symbol)
      ;; want download if record is more than 2 minutes old
      (< (latest-fetch-time (latest-get symbol)) target-time))

    (receive (symbol-list drop-list)
	(partition want? symbol-list)
      (latest-fetch-fail drop-list)

      ;; drop new enough entries from extra-list so that it'll be old
      ;; entries which are freshened if possible
      (receive (extra-list drop-list)
	  (partition want? extra-list)

	(latest-request-explicit symbol-list extra-list)))))

(define (run-handler handler symbol-list extra-list)
  (c-lazy-catch #t
    (lambda ()
      (handler symbol-list extra-list
	       (lambda (latest-list)
		 (set! symbol-list
		       (lset-difference string=? symbol-list
					(map latest-symbol latest-list)))
		 (latest-handler-callback latest-list)))
      ;; anything HANDLER didn't return
      (latest-fetch-fail symbol-list))

    (lambda args
      (cond ((or (gethost-error-try-again? args)
		 (system-error-econn? args))
	     ;; offline, quietly ignore this error
 	     (latest-fetch-fail symbol-list))

	    (else
	     ;; some sort of parse error, all remaining SYMBOL-LIST fail,
	     ;; and re-throw the error to get a backtrace
	     (c-main-enq!
	      (lambda ()
		(let ((note (error-message args)))
		  (latest-handler-callback (map (lambda (symbol)
						  (latest-new #:symbol symbol
							      #:note   note
							      #:source 'error))
						symbol-list)))))
	     (apply throw args))))))

;; latest records fetched by a handler function
(define-public (latest-handler-callback latest-list)
  (latest-unmark-in-progress (map latest-symbol latest-list))
  (for-each (lambda (latest)
	      (hash-set! cache (latest-symbol latest) latest))
	    latest-list)
  (latest-write latest-list)
  (notify-send 'latest-update (map latest-symbol latest-list)))

;; set the fetch time to now, to suppress further attempts for a while
(define-public (latest-fetch-fail symbol-list)
  (if (not (null? symbol-list))
      (begin
	(latest-unmark-in-progress symbol-list)
	(let ((t (current-time)))
	  (for-each (lambda (symbol)
		      (and-let* ((latest (hash-ref cache symbol)))
			(latest-fetch-time! latest t)
			(hash-set! cache symbol latest)))
		    symbol-list)
	  (database-call-with-lock latest-directory
	    (lambda ()
	      (for-each (lambda (symbol)
			  (and-let* ((latest (latest-read symbol)))
			    (latest-fetch-time! latest t)
			    (latest-write (list latest))))
			symbol-list))))
	(notify-send 'latest-update symbol-list))))

(define latest-in-progress-list '())

(define-public (latest-in-progress? symbol)
  (member symbol latest-in-progress-list))

(define (latest-mark-in-progress symbol-list)
  ;; newly in progress
  (set! symbol-list (lset-difference string=? symbol-list latest-in-progress-list))
  (set! latest-in-progress-list (append symbol-list latest-in-progress-list))
  (notify-send-locally 'latest-update symbol-list))

(define (latest-unmark-in-progress symbol-list)
  (set! latest-in-progress-list
	(lset-difference string=? latest-in-progress-list symbol-list)))

;; return when SYMBOL has been retrieved
(define (latest-request-synchronous symbol)
  (call-with-current-continuation
   (lambda (cont)
     (define (callback symbol-list)
       (if (and (not (latest-in-progress? symbol))
		(member symbol symbol-list))
	   (begin
	     (notify-disconnect 'latest-update callback)
	     (cont #f))))
     (notify-connect 'latest-update callback)
     (c-main-enq! latest-request-symbols (list symbol))
     (c-main-goto-top))))

;; return mdate of SYMBOL
;; a specific "CLZ05.NYMEX" has its mdate returned directly
;; a front month "CL.NYMEX" is looked up from `latest-get', wanting TDATE or
;; later there, and doing a request if necessary to get that
;;
(define-public (latest-symbol-mdate symbol tdate)
  (or (latest-symbol-mdate-nodownload symbol tdate)

      ;; fetch new latest record now, and return from that (whatever it gives)
      (begin
	(latest-request-synchronous symbol)
	(latest-contract-mdate (latest-get symbol)))))

(define-public (latest-symbol-mdate-nodownload symbol tdate)
  (or (chart-symbol-mdate symbol)

      ;; from existing record, if date new enough
      (and-let* ((latest    (latest-get symbol))
		 (got-tdate (or (latest-quote-tdate latest)
				(latest-last-tdate latest)))
		 (          (>= got-tdate tdate)))
	(latest-contract-mdate latest))))


;;-----------------------------------------------------------------------------
;; formatting

;; "/" or "x" for bid/offer, the "x" being when crossed (bid>=offer)
(define-public (latest-form-quote-slash latest)
  (if (and (latest-bid latest)
	   (latest-offer latest)
	   (>= (latest-bid latest)
	       (latest-offer latest)))
      "x"
      "/"))

;; return a bid/offer string like "2.05/2.06"
(define-public (latest-form-bid/offer latest)
  (let ((bid   (latest-bid latest))
	(offer (latest-offer latest)))
    (if (or bid offer)
	(string-append (if bid (latest-form-price latest bid) "--")
		       (latest-form-quote-slash latest)
		       (if offer (latest-form-price latest offer) "--"))
	"")))

(define-public (latest-form-price latest val)
  (if val
      (locale-number+decimals val (latest-decimals latest))
      ""))

(define (seconds->h s)
  (quotient s 3600))

(define (seconds->m s)
  (quotient (remainder s 3600) 60))

;; return today's adate, in the current timezone or the given ZONE (a string)
(define-public (adate-today . zone)
  (tm->adate (apply localtime (current-time) zone)))

(define (adate->wdate a)
  (quotient-floor a 7))

;; return a short date/time string for LATEST
(define-public (latest-form-date/time latest)
  (let ((adate (latest-quote-adate latest))
	(time  (latest-quote-time latest)))

    ;; if there's no quote-adate then use last-adate
    ;; if there is a quote-adate but no bid/offer, then use last-adate, but
    ;; only if there's a value for last-adate
    (if (or (not adate)
	    (and (latest-last-adate latest)
		 (not (latest-bid latest))
		 (not (latest-offer latest))))
	(begin
	  (set! adate (latest-last-adate latest))
	  (set! time  (latest-last-time latest))))

    (form-adate-time (latest-symbol latest) adate time)))

;; return a utf8 string which is the date/time ADATE and TIME formatted for
;; friendly display, in the current time of SYMBOL
;;
(define-public (form-adate-time symbol adate time)
  (let ((today (adate-today (symbol-timezone symbol))))
    (cond ((not adate)
	   "")

	  ;; for today just show time
	  ((= adate today)
	   (if time
	       (format #f "~2,'0d:~2,'0d" (seconds->h time) (seconds->m time))
	       (_ "Today")))

	  ;; for same week, or on Mon the prev Thu,Fri or on Tue the prev
	  ;; Fri, just show day
	  ((or (= (adate->wdate adate) (adate->wdate today))
	       (<= (- (adate->tdate today) (adate->tdate adate)) 2))
	   (strftime-utf8 "%a" (adate->tm adate)))

	  ;; this year, or in Jan the prev Dec, just show mday+month
	  ((or (= (adate->year today) (adate->year adate))
	       (<= (- adate today) 30))
	   (strftime-utf8 "%d-%b" (adate->tm adate)))

	  ;; +/- 40 years of today, show abbreviated year
	  ((> 40 (abs (- (adate->year today) (adate->year adate))))
	   (strftime-utf8 "%d%b%y" (adate->tm adate)))

	  ;; otherwise full date
	  (else
	   (strftime-utf8 "%d%b%Y" (adate->tm adate))))))

;; return volume as a string, with k/m/b for thousands, millions, billions
(define-public (latest-form-volume-km latest)
  (let ((volume (latest-volume latest)))
    (if volume
	(let ((suffix  "")
	      (decimals 0))
	  (cond ((>= volume 10000000000)
		 (set! volume (quotient volume 1000000000))
		 (set! suffix (_ "CONTEXT(billions): b")))
		((>= volume 1000000000)
		 (set! volume (quotient volume 10000000))
		 (set! decimals 2)
		 (set! suffix (_ "CONTEXT(billions): b")))
		((>= volume 10000000)
		 (set! volume (quotient volume 1000000))
		 (set! suffix (_ "CONTEXT(millions): m")))
		((>= volume 1000000)
		 (set! volume (quotient volume 10000))
		 (set! decimals 2)
		 (set! suffix (_ "CONTEXT(millions): m")))
		((>= volume 10000)
		 (set! volume (quotient volume 1000))
		 (set! suffix (_ "CONTEXT(thousands): k"))))
	  (string-append (locale-number+decimals volume decimals) suffix))
	"")))

;; return a volume string for LATEST
(define-public (latest-form-volume latest)
  (if (latest-volume latest)
      (locale-number+decimals (latest-volume latest) 0)
      ""))

;; return `latest-contract-mdate' as a utf8 string
(define-public (latest-form-mdate latest)
  (if (latest-contract-mdate latest)
      (strftime-utf8 "[%b%y]" (mdate->tm (latest-contract-mdate latest)))
      ""))  

;; return note string, as utf8
(define-public (latest-form-note latest)
  (let ((note (latest-note latest)))
    (or note
	(latest-form-mdate latest))))


;;-----------------------------------------------------------------------------
;; validation
;;
;; Don't want to get too upset if some data source is a bit dodgy, but also
;; don't want a programming bug to let bad values silently come in.

(define (exact-integer? n)
  (and (number? n)
       (exact? n)
       (integer? n)))

(define (exact-non-negative-integer? n)
  (and (number? n)
       (exact? n)
       (integer? n)
       (not (negative? n))))

(define (time-in-seconds? n)
  (and (number? n)
       (exact? n)
       (integer? n)
       (within-range? 0 #,(hms->seconds 23 59 59) n)))

(define likely-adate-hi (+ (adate-today) 365))          ;; +1 year
(define likely-adate-lo (- (adate-today) (* 100 365)))  ;; -100 years
(define (likely-adate? n)
  (and (number? n)
       (exact? n)
       (integer? n)
       (within-range? likely-adate-lo likely-adate-hi n)))

(define likely-mdate-hi (+ (tdate->mdate (tdate-today)) 120)) ;; +10 years
(define likely-mdate-lo (- (tdate->mdate (tdate-today)) 1200)) ;; -100 years
(define (likely-mdate? n)
  (and (number? n)
       (exact? n)
       (integer? n)
       (within-range? likely-mdate-lo likely-mdate-hi n)))

(define (latest-validate latest)
  (define (check field cond)
    (let ((value (field latest)))
      (or (not value)
	  (cond value)
	  (simple-format (current-error-port) "latest-new: ~s ~s not ~s: ~s\n"
			 (latest-symbol latest) (c-procedure-name field)
			 cond value))))

  (or (string? (latest-symbol latest))
      (simple-format (current-error-port) "latest-new: latest-symbol not string: ~s\n"
		     (latest-symbol latest)))

  (check latest-fetch-time  exact-integer?)
  (check latest-bid   	    exact-integer?)
  (check latest-offer 	    exact-integer?)
  (check latest-quote-adate likely-adate?)
  (check latest-quote-time  time-in-seconds?)
  (check latest-last-adate  likely-adate?)
  (check latest-last-time   time-in-seconds?)
  (check latest-open 	    exact-non-negative-integer?)
  (check latest-high 	    exact-non-negative-integer?)
  (check latest-low  	    exact-non-negative-integer?)
  (check latest-last 	    exact-non-negative-integer?)
  (check latest-change 	    exact-integer?)
  (check latest-volume 	    exact-non-negative-integer?)
  (check latest-note 	    string?)
  (check latest-decimals    exact-non-negative-integer?)
  (check latest-contract-mdate likely-mdate?)

  (let ((open (latest-open latest))
	(high (latest-high latest))
	(low  (latest-low  latest))
	(last (latest-last  latest)))

    (if (and (or open high low)
	     (not (latest-last-adate latest)))
	(simple-format (current-error-port)
		       "latest-new: ~s open/high/low without last-adate\n"
		       (latest-symbol latest)))

    (if (and high low (< high low))
	(simple-format (current-error-port)
		       "latest-new: ~s high less than low: ~s ~s\n"
		       (latest-symbol latest)
		       (latest-high latest)
		       (latest-low latest)))

    (if (and last low (< last low))
	(simple-format (current-error-port)
		       "latest-new: ~s last less than low: ~s ~s\n"
		       (latest-symbol latest)
		       (latest-last latest)
		       (latest-low latest)))
    (if (and last high (> last high))
	(simple-format (current-error-port)
		       "latest-new: ~s last greater than high: ~s ~s\n"
		       (latest-symbol latest)
		       (latest-last latest)
		       (latest-high latest)))

    (if (and open low (< open low))
	(simple-format (current-error-port)
		       "latest-new: ~s open less than low: ~s ~s\n"
		       (latest-symbol latest)
		       (latest-open latest)
		       (latest-low latest)))
    (if (and open high (> open high))
	(simple-format (current-error-port)
		       "latest-new: ~s open greater than high: ~s ~s\n"
		       (latest-symbol latest)
		       (latest-open latest)
		       (latest-high latest))))

  (if (latest-name latest) (utf8-validate (latest-name latest)))
  (if (latest-note latest) (utf8-validate (latest-note latest))))
