;; Line annotation editing.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart line-edit)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (srfi srfi-1)
  #:autoload   (chart cross)       (cross-disable)
  #:autoload   (chart database)    (database-write-notes)
  #:use-module (chart download)
  #:use-module (chart graph)
  #:use-module (chart keymap)
  #:use-module (chart misc)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; gdk distances

;; convert x pixels to millimetres
(define c-gdk-x-pixels->mm
  (let ((factor (exact->inexact
		 (/ (gdk-screen-width-mm) (gdk-screen-width)))))
    (lambda (x)
      (* x factor))))

;; convert y pixels to millimetres
(define c-gdk-y-pixels->mm
  (let ((factor (exact->inexact
		 (/ (gdk-screen-height-mm) (gdk-screen-height)))))
    (lambda (y)
      (* y factor))))

;; sqrt(x^2+y^2)
(define (hypot x y)
  (sqrt (+ (* x x) (* y y))))

;; the distance in millimetres between pixel points x1,y1 and x2,y2
(define (c-gdk-distance-mm x1 y1 x2 y2)
  (hypot (c-gdk-x-pixels->mm (- x1 x2))
	 (c-gdk-y-pixels->mm (- y1 y2))))


;;-----------------------------------------------------------------------------
;; cross state

(let-toplevel ((state #f))
  (notify-connect 'cross-update (lambda (new-state)
				  (set! state new-state)))
  ;; do a `cross-disable', but avoid loading the cross module if it isn't
  ;; already in use
  (define-toplevel (line-edit-cross-disable)
    (if state
	(cross-disable))))


;;-----------------------------------------------------------------------------
;; lines

(define line-t1! set-second!)
(define line-p1! set-third!)
(define line-t2! set-fourth!)
(define line-p2! set-fifth!)

;; swap the endpoints of a line, in-place
(define (line-swap! line)
  (let ((t1 (line-t1 line))
	(p1 (line-p1 line))
	(t2 (line-t2 line))
	(p2 (line-p2 line)))
    (line-t1! line t2)
    (line-p1! line p2)
    (line-t2! line t1)
    (line-p2! line p1)))


;;-----------------------------------------------------------------------------
;; line editing

(define line-edit-cursor (gdk-cursor-new 'left-ptr))

(define new-line #f)
(define x-offset #f)  ;; x,y offset applied for diff between
(define y-offset #f)  ;;   initial pointer and initial t,p

(define (horiz?)
  (eq? 'hline (first new-line)))

;; create a new line at the pointer location given by the event
(define (line-create event)
  (let ((t (graph-x->tdate draw (round-exact (gdk-event-x event))))
	(p (graph-y->price draw (gdk-event-y event))))
    (list 'line t p t p)))

(define (line-edit-xor)
  (let ((line (list-copy new-line)))
    (if (horiz?)
	(line-p1! line (line-p2 line)))
    (line-xor-list draw (list line))))

(define (line-edit-move x y)
  (if line-edit-state
      (begin
	(line-edit-xor)
	(cond ((note-alert? new-line)
	       (set-second! new-line (graph-y->price draw (+ y-offset y))))
	      (else
	       (line-t2! new-line (graph-x->tdate draw (+ x-offset x)))
	       (line-p2! new-line (graph-y->price draw (+ y-offset y)))))
	(line-edit-xor))))

;; return a line from notes, or #f
;; the line ends are swapped if necessary to make t2,p2 the point near x,y
;;
(define (line-find x y)
  (let* ((graph    draw)
	 (tdate->t (timebase-convert-proc timebase-days
					  (graph-timebase graph)))
	 (scale-x  (graph-scale-x-proc graph))
	 (scale-y  (graph-scale-y-proc graph)))

    ;; distance in millimetres from tdate/price to X,Y
    (define (distance-tp tdate price)
      (c-gdk-distance-mm x y (scale-x (tdate->t tdate)) (scale-y price)))

    (define (distance-alert price)
      ;; x position as per `line-xor-list' drawing
      (let* ((alert-t (1+ (min (graph-today graph)
			       (hadjustment-hi
				(graph-hadjustment graph)))))
	     (alert-x (1- (scale-x alert-t))))
	(c-gdk-distance-mm x y alert-x (scale-y price))))

    (find-minimum (lambda (elem)
		    (or (and (note-line? elem)
			     (let ((d  (distance-tp (line-t2 elem)
						    (line-p2 elem)))
				   (ds (distance-tp (line-t1 elem)
						    (line-p1 elem))))
			       (if (< ds d)
				   (begin
				     (line-swap! elem)
				     (set! d ds)))
			       (and (<= d 5)  ;; no more than 5 mm
				    d)))
			(and (note-alert? elem)
			     (let ((d (distance-alert (second elem))))
			       (and (<= d 5)
				    d)))))
		  (series-notes (graph-series graph)))))



(define (init)
  (gtk-signal-connect draw "motion_notify_event"
    (lambda (event)
      (let ((pointer (gdk-window-get-pointer
		      (gtk-widget-window draw))))
	(line-edit-move (inexact->exact (second pointer))
			(inexact->exact (third  pointer))))))

  (gtk-signal-connect draw "button_release_event"
    (lambda (event)
      (line-edit-move (inexact->exact (gdk-event-x event))
		      (inexact->exact (gdk-event-y event)))
      (line-edit-done)))

  ;; once only, and gc code
  (set! init noop))

(define-public (line-edit-or-new-button)
  (init)
  (line-edit-cross-disable)
  (gdk-window-set-cursor (gtk-widget-window draw) line-edit-cursor)

  (let* ((x        (inexact->exact (gdk-event-x keymap-event)))
	 (y        (inexact->exact (gdk-event-y keymap-event)))
	 (tdate->t (timebase-convert-proc timebase-days
					  (graph-timebase draw)))
	 (scale-x  (graph-scale-x-proc draw))
	 (scale-y  (graph-scale-y-proc draw))
	 (found    (line-find x y)))
    (set! line-edit-state (or found
			      (line-create keymap-event)))
    (set! new-line (list-copy line-edit-state))

    (cond ((note-alert? new-line)
	   (set! x-offset 0)
	   (set! y-offset (- (scale-y (second new-line)) y)))
	  (else
	   (set! x-offset (- (scale-x (tdate->t (line-t2 new-line))) x))
	   (set! y-offset (- (scale-y (line-p2 new-line)) y))))

    ;; FIXME: this initial draw causes a pixel left drawn after an abort, why?
    ;; initial draw of new line
    '(if (not found)
	 (line-edit-xor))))

(define-public (line-edit-done)
  (if line-edit-state
      (begin
	(if (horiz?)
	    (line-p1! new-line (line-p2 new-line)))
	(if (note-alert? new-line)
	    (set-cdr! (cdr new-line) '()))  ;; chop to (alert-xxx p1)
	(let* ((series (series-parent-top (graph-series draw)))
	       (notes  (series-notes series)))
	  (set! notes (delete line-edit-state notes))
	  (set! notes (cons new-line notes))
	  (series-notes! series notes)
	  (database-write-notes series))
	(gdk-window-set-cursor (gtk-widget-window draw) #f)
	(set! line-edit-state #f))))

(define-public (line-edit-abort)
  (if line-edit-state
      (begin
	(line-edit-xor)
	(set! new-line line-edit-state)
	(if new-line
	    (line-edit-xor))
	(gdk-window-set-cursor (gtk-widget-window draw) #f)
	(set! line-edit-state #f))))

(define-public (line-edit-toggle-horizontal)
  (if (note-line? line-edit-state)
      (begin
	(line-edit-xor)
	(set-first! new-line (if (eq? 'line (first new-line))
				 'hline 'line))
	(line-edit-xor))))

(define-public (line-edit-toggle-alert)
  (line-edit-xor)
  (case (first new-line)
    ((alert-above)
     (set-first! new-line 'alert-below))
    ((alert-below)
     (set-first! new-line 'alert-above))
    ((line hline)
     (set-first! new-line 'alert-above)
     (set-second! new-line (line-p2 new-line))))
  (line-edit-xor))

(define-public (line-edit-delete)
  (if line-edit-state
      (begin
	(let* ((series (series-parent-top (graph-series draw)))
	       (notes  (series-notes series)))
	  (series-notes! series (delq line-edit-state notes))
	  (database-write-notes series))
	(line-edit-xor)
	(gdk-window-set-cursor (gtk-widget-window draw) #f)
	(set! line-edit-state #f))))


(define-public line-edit-keymap (keymap-new (_ "While moving a line")
					    (lambda () line-edit-state)))

(keymap-set! line-edit-keymap "Escape" line-edit-abort)
(keymap-set! line-edit-keymap "A"      line-edit-toggle-alert)
(keymap-set! line-edit-keymap "D"      line-edit-delete)
(keymap-set! line-edit-keymap "<Alt>D" line-edit-delete)
(keymap-set! line-edit-keymap "H"      line-edit-toggle-horizontal)
