;; Line segment graph drawing.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart line)
  #:use-module (ice-9 optargs)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gdk)
  #:use-module (chart graph)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart view))


;; LST is a list of pairs (X . Y), draw a line connecting those, possibly
;; just a single point, or nothing at all for an empty list
(define-public (c-gdk-draw-lines-or-point-maybe win gc lst)
  (if (not (null? lst))
      (begin
	(if (null? (cdr lst))
	    (set! lst (cons (car lst) lst)))
	(gdk-draw-lines win gc lst))))


;; Return the start position of COUNT many closing prices in SERIES prior to
;; POS (and not including POS).
;;
(define*-public (series-next-close-pos series #:key after)
  (define arg (series-closes-apply-arg series))

  (let more ((pos (1+ after)))
    (cond ((> pos (series-hi series))
	   (series-hi series)) ;; hit end of series

	  ((apply array-ref (series-array series pos pos) pos arg)
	   pos)  ;; found

	  (else
	   (more (1+ pos))))))

;; return x coordinate to go with Y-TARGET on a line X1,Y1 to X2,Y2
(define (interpolate x1 y1 x2 y2 y-target)
  (if (= y1 y2)
      x2
      (round-exact (/ (+ (* x1 (- y-target y2))
			 (* x2 (- y1 y-target)))
		      (exact->inexact ;; avoid taking gcd for a fraction
		       (- y1 y2))))))


(define*-public (line-draw graph gc series
			   #:key
			   column
			   (colour "white")
			   (solid #f))

  ;; LO/HI is taken to the next close outside the window, to show lines
  ;; extending away
  (let* ((lo      (series-prev-n-closes-pos
		   series #:before (graph-draw-x-lo graph series) #:count 1))
	 (hi      (series-next-close-pos
		   series #:after (graph-draw-x-hi graph series)))
	 (array   (series-array series lo hi))
	 (win     (gtk-widget-window graph))
	 (height  (c-gdk-drawable-get-height win))
	 (scale-y (graph-scale-y-proc graph))
	 (x-first (+ (quotient scale-x-step 2) ;; centre in step
		     (graph-scale-x graph lo)))
	 (gc      (c-gdk-gc-get win #:foreground colour #:dashed #f))
	 (dash-gc (c-gdk-gc-get win #:foreground colour #:dashed (not solid))))

    (define (draw-column . col)

      ;; NEXT-GC is the gc to use for the next line segment drawing, it's
      ;; either the plain GC, or DASH-GC
      ;;
      ;; LAST-X,LAST-Y are the coordinates of the previous point, without
      ;; any truncation
      ;;
      ;; POLY is a list of points (X . Y) to be drawn as a polygon using
      ;; POLY-GC.
      ;;
      ;; X1,Y1 and X2,Y2 calculated are LAST-X,LAST-Y and X,Y truncated to
      ;; the size of the window, so as to cope with huge Y values that might
      ;; overflow a "gint" required by the gdk functions if put directly
      ;; into POLY.  It's possibly the Y range is completely outside the
      ;; window, in that case nothing is added to poly at all.
      ;;
      (let* ((last-x  #f)
	     (last-y  #f)
	     (poly    '())
	     (poly-gc gc)
	     (next-gc gc))

	(define (flush-poly)
	  (if (not (null? poly))
	      (begin
		(c-gdk-draw-lines-or-point-maybe win poly-gc poly)
		(set! poly '()))))

	(do ((i lo (1+ i))
	     (x x-first (+ x scale-x-step)))
	    ((> i hi))
	  (let ((p (apply array-ref array i col)))
	    (if p
		(let* ((y  (scale-y p))
		       (x1 last-x)
		       (y1 last-y)
		       (x2 x)
		       (y2 y))

		  (if (and y1
			   (overlap-inclusive? (min y1 y2) (max y1 y2)
					       0 (1- height)))
		      (begin
			(cond ((negative? y1)
			       (set! x1 (interpolate last-x last-y x y -1))
			       (set! y1 -1)
			       (flush-poly))
			      ((> y1 height)
			       (set! x1 (interpolate last-x last-y x y height))
			       (set! y1 height)
			       (flush-poly)))

			(cond ((negative? y2)
			       (set! x2 (interpolate last-x last-y x y -1))
			       (set! y2 -1))
			      ((> y2 height)
			       (set! x2 (interpolate last-x last-y x y height))
			       (set! y2 height)))

			(if (not (eq? poly-gc next-gc))
			    (begin
			      (flush-poly)
			      (set! poly-gc next-gc)))

			(if (and y1
				 (null? poly))
			    (set! poly (list (cons x1 y1))))

			(set! poly (cons (cons x2 y2) poly)))

		      ;; if entirely off screen then don't continue polygon
		      (flush-poly))

		  (set! last-x x)
		  (set! last-y y)
		  (set! next-gc gc))

		(set! next-gc dash-gc))))

	;; possibly a single point only
	(if (and last-y
		 (null? poly)
		 (within-range? 0 (1- height) last-y))
	    (set! poly (list (cons last-x last-y))))

	(flush-poly)))

    (cond (column
	   ;; specific requested column
	   (draw-column column))

	  ((eq? 'ohlcv (datatype-key (series-datatype series)))
	   ;; closes for ohlcv
	   (draw-column 3))

	  ((= 1 (array-rank array))
	   ;; only one column
	   (draw-column))

	  (else
	   ;; draw lines for all columns
	   (do ((i (array-first-column array) (1+ i))
		(col-hi (array-last-column array)))
	       ((> i col-hi))
	     (draw-column i))))))

(define-public (bands-draw graph gc series)
  (line-draw graph gc series #:column 0 #:colour graph-band-colour)
  (line-draw graph gc series #:column 1 #:colour graph-up-colour)
  (line-draw graph gc series #:column 2 #:colour graph-band-colour))

(define-public (zigzag-draw graph gc series)
  (line-draw graph gc series #:solid #t))  ;; force to always solid
