;; Linear regression standard error.

;; Copyright 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart linreg-stderr)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://mathworld.wolfram.com/LeastSquaresFitting.html
;;     Showing how to calculate the variance in the error amounts (e[i])
;;     without calculating a,b parameters.
;;


;;               Sum (y - (a+b*x))^2
;; stderr = sqrt -------------------
;;                       N^2
;;
;;               Sum ((y - My) - b*x)^2
;;        = sqrt ----------------------     where My = Mean(Y)
;;                       N^2
;;
;;               Sum (y - My)^2 - 2*My*b*x + b^2*x^2
;;        = sqrt -----------------------------------
;;                               N^2
;;
;;                   (y - My)^2       2*My*b*x - b^2*x^2
;;        = sqrt Sum ---------- - Sum ------------------
;;                    N^2                    N^2
;;
;; The first term is Variance(Y), and with b = Mean(X*Y) - Mean(X) * Mean(Y)
;;
;;                            2*M(y)*Sum(x) - M(X*Y)*Sum(x^2)/M(X)*M(Y)
;;        = sqrt Var(Y) - b * -----------------------------------------
;;                                               N^2
;;
;; ....


;; Return the factor for 1/Variance(X) arising in the stderr formula.  With
;; X values -1.5,-0.5,0.5, 1.5 this means
;;
;;                         1
;;      --------------------------------------------
;;      n * ((-1.5)^2 + (-0.5)^2 + (0.5)^2 + (1.5)^2)
;;
;; and the denominator here is (n^4 - n^2)/12.  See devel/closed-form.scm
;; for checking this.  Same func in rsquared.scm too.
;;
(define (xfactor n)
  (if (< n 2)
      1.0
      (/ 12.0 (* n n (1- (* n n))))))


;; Return a procedure which calculates a linear regression standard error
;; over a window of COUNT many points.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the stderr up to and including that point, or #f if not enough points
;; yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;

(define-public (linreg-stderr-calc-proc count)

  ;; Y-DATA is a circular list of the last COUNT Y values, ie. prices.  The
  ;; value at Y-DATA itself is the oldest.
  ;;
  ;; POINTS is the number of non-#f values in Y-DATA.
  ;;
  ;; Y-TOTAL is the sum of the values in Y-DATA.
  ;;
  ;; YY-TOTAL is the sum of the squares of the values in Y-DATA.
  ;;
  ;; X-TOTAL is the sum of the X values corresponding to the values in
  ;; Y-DATA, which means 1+2+...+POINTS, which is a triangular number.
  ;; This varies until POINTS=COUNT is reached, and is then a constant.
  ;;
  ;; X-FAC is the variance factor in the denominator, see `xfactor' above.
  ;; This varies until POINTS=COUNT is reached and is then a constant.
  ;;
  ;; XY-TOTAL is the sum of the product x*y for each x,y pair, which means
  ;; 1*y[1] + 2*y[2] + ... + POINTS*y[POINTS].  When shifting out an old Y,
  ;; the weighting of each y in the sum is reduced by subtracting Y-TOTAL.
  ;;
  ;; The calculation ensures that if all Y values coming in are exact then
  ;; the sums Y-TOTAL, YY-TOTAL and XY-TOTAL are exact.  The value returned
  ;; is always inexact though.
  ;;

  (define y-data   (make-circular-list count #f))
  (define y-total  0)
  (define yy-total 0)
  (define points   0)
  (define x-fac    #f)
  (define xy-total 0)
  (define min-points (ceiling-exact (* indicator-minimum-fraction count)))

  (lambda (y)
    (if (car y-data)
	;; drop oldest point
	(begin
	  (set! yy-total (- yy-total (* (car y-data) (car y-data))))
	  (set! xy-total (+ xy-total (* 0.5 (1- points) (car y-data))))
	  (set! y-total  (- y-total (car y-data))))

	;; otherwise gaining a point
	(begin
	  (set! points   (1+ points))
	  (set! x-fac    (xfactor points))
	  (set! xy-total (+ xy-total (* 0.5 y-total))))) ;; adj so 0.5 less each x

    ;; shift xy products onto 1 less x each
    (set! xy-total (- xy-total y-total))

    ;; add this point
    (set! y-total  (+ y-total y))
    (set! yy-total (+ yy-total (* y y)))
    (set! xy-total (+ xy-total (* y (* 0.5 (1- points)))))

    ;; replace oldest with new point
    (set-car! y-data y)
    (set! y-data (cdr y-data))

    ;;     (let* ((ylst          (delq #f (take y-data count)))
    ;; 	   (xlst          (iota points (* 0.5 (- (1- points)))))
    ;; 	   (want-xy-total (apply + (map * xlst ylst))))
    ;;       (dv points xlst ylst y-total yy-total xy-total want-xy-total)
    ;;       (assert (= y-total (apply + ylst)))
    ;;       (assert (= yy-total (apply + (map square ylst))))
    ;;       (assert (= xy-total want-xy-total))
    ;;       (dv "cov" (* points xy-total)))

    (and (>= points min-points)
	 (/ (sqrt
	     (max 0 (- (* points yy-total) (* y-total y-total)  ;; Var(Y)
		       (* (square (* points xy-total))   ;; (Covar(X,Y))^2
			  x-fac))))  ;; 1 / (X variance)
	    points))))



(define (linreg-stderr-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count (1- count)))
	 (array  (series-array-closes parent start hi))
	 (proc   (linreg-stderr-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i)
	  (proc (array-ref array i))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i))
		      i)))))

(define-public (linreg-stderr-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Linreg Stderr") count))
    (series-datatype!  new-series linreg-stderr-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts linreg-stderr-fill
							   (1- count)))
    (series-array!     new-series #f)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public linreg-stderr-datatype
  (datatype-new!
   #:key      'linreg-stderr
   #:name     (_ "Linear Regression Stderr")
   #:priority -10
   #:units    'price
   #:minimum  0
   #:manual   (_ "CONTEXT(manual-node): Linear Regression Standard Error")
   #:proc     linreg-stderr-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'linreg-stderr-days
	  #:type    'integer
	  #:default 20
	  #:minimum 1))))
