;; Linear regression.

;; Copyright 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart linreg)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://mathworld.wolfram.com/LeastSquaresFitting.html
;;     Least squares generally, including deriving formula using
;;     derivative==0 as follows:
;;
;;     The sum of squares is
;;
;;         R^2(a,b) = Sum (y[i] - (a + b*x[i]))^2
;;
;;     Partial derivative with b is
;;
;;         d R^2(a,b)
;;         ---------- = Sum 2 * (y[i]-b*x[i]) * -x[i]
;;             db
;;
;;     And want that to be zero at the minimum, so
;;
;;         Sum -2*x[i]*y[i] + 2*b*Sum x[i]^2 = 0
;;
;;             Sum x[i]*y[i]
;;         b = -------------
;;               Sum x[i]^2
;;


;; `xx2-calc' returns 2*Sum(X^2) for the set of N points centred around
;; zero as described in linreg-slope-calc-proc below.  This means for
;; instance,
;;
;;     N=4:  2 * [ (-1.5)^2 + (-0.5)^2 + (0.5)^2 + (1.5)^2 ]
;;     N=5:  2 * [ (-2)^2   + (-1)^2   + 0^2     + (1)^2   + (2)^2 ]
;;
;; This is (N^3-N)/6, which can be established by taking successive
;; differences or verified by induction (done separately for odds and evens
;; is easiest).
;;
;; N^3-N is always a multiple of 6, since it can be written (N-1)*N*(N+1)
;; which is three consecutive numbers so one is certainly a multiple of 3
;; and another a multiple of 2.  The result is forced to inexact since
;; that's what's wanted for the linreg-slope-calc-proc returns.
;;
(define (xx2-calc n)
  (/ (* n (1- (* n n))) 6.0))


;; Return a procedure which calculates a linear regression line fit over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is two values (values A B) where the line is A+B*X.  The last point
;; entered is at X=0 and the preceding ones at X=-1, X=-2, etc.  A and B are
;; #f if not enough data yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.  `linreg-warmup-count' below can be used to
;; to calculate that.
;;
(define-public (linreg-calc-proc count)

  ;; The X values used are centred around 0,
  ;;     POINTS=4: -1.5, -0.5, 0.5, 1.5
  ;;     POINTS=5: -2, -1, 0, 1, 2
  ;;     POINTS=6: -2.5, -1.5, -0.5, 0.5, 1.5, 2.5
  ;;     etc
  ;;
  ;; DATA is a circular list of COUNT many values.  The one at DATA is the
  ;; oldest and is replaced by a new value to cycle that in.
  ;;
  ;; POINTS is how many points are in data (non-#f ones).
  ;;
  ;; Y-TOTAL is the total of the past COUNT many Y values, ie. the values in
  ;; DATA.  This is exact if those values are all exact.
  ;;
  ;; XY2-TOTAL is the sum of 2*X*Y for each Y value in DATA.  The factor of
  ;; 2 makes the X values integers, so XY-TOTAL is an integer if all the
  ;; input Y values are integers, preserving exactness.
  ;;
  ;; XX2-TOTAL is 2 * the sum of X*X for each X value used.  This is a
  ;; constant once POINTS stops at COUNT.  A minimum 1 is enforced for the
  ;; degenerate case of COUNT==0 (there's no slope to in that case but at
  ;; least avoid a divide by zero).  The factor 2 compensates for the same
  ;; factor in XY2-TOTAL.
  ;;
  (define data       (make-circular-list count #f))
  (define y-total    0)
  (define xy2-total  0)
  (define xx2-total  1.0)
  (define points     0)
  (define min-points (ceiling-exact (* indicator-minimum-fraction count)))

  (lambda (new-y)

    (if (car data)
	;; drop oldest point
	(begin
	  (set! y-total  (- y-total (car data)))
	  (set! xy2-total (+ xy2-total (* (1- points) (car data)))))

	;; otherwise gaining a point
	(begin
	  (set! points (1+ points))
	  (set! xy2-total (+ xy2-total y-total)) ;; adj so below is 1 less x
	  (set! xx2-total (max 1.0 (xx2-calc points)))))

    ;; shift xy products onto 2 less x each
    (set! xy2-total (- xy2-total y-total y-total))

    ;; add this point
    (set! y-total  (+ y-total new-y))
    (set! xy2-total (+ xy2-total (* (1- points) new-y)))

    ;; replace oldest with new point
    (set-car! data new-y)
    (set! data (cdr data))

    (if (>= points min-points)
	(let ((b (/ xy2-total xx2-total)))
	  (values (+ (/ y-total points)
		     (* b 0.5 (1- points)))
		  b))
	(values #f #f))))

(define-public linreg-warmup-count 1-)


;;-----------------------------------------------------------------------------
;; linear regression slope indicator

(define (linreg-slope-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (linreg-calc-proc count)))

    ;; warmup
    (for-each proc (series-prev-n-closes parent #:before lo
					 #:count (1- count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(receive (a b)
	    (proc this)
	  (array-set! new-array b i))))))

(define-public (linreg-slope-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "Linreg Slope") count))
    (series-datatype! new-series linreg-slope-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts linreg-slope-fill
							  (1- count)))
    (series-array!    new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define linreg-slope-datatype
  (datatype-new!
   #:key       'linreg-slope
   #:name      (_ "Linear Regression Slope")
   #:priority  -10
   #:units     'price-slope
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Linear Regression Slope")
   #:proc      linreg-slope-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'linreg-slope-days
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
