;; Ljubljana Stock Exchange (LJSE) setups.

;; Copyright 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ljubljana)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (ljubljana-symbol? symbol)
  (string-suffix? ".LJ" symbol))

(symbol-timezone! ljubljana-symbol? (_ "Ljubljana") timezone-europe-cet)
(source-help! ljubljana-symbol?
	      (_ "CONTEXT(manual-node): Ljubljana Stock Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - company info
;;
;; eg.
;; Slovenian: http://www.ljse.si/cgi-bin/jve.cgi?SecurityID=DRKR&doc=818
;; English:   http://www.ljse.si/cgi-bin/jve.cgi?SecurityID=DRKR&doc=3131
;;

(weblink-handler! ljubljana-symbol?
  (lambda (symbol)
    (list (_ "LJSE _Company Information")
	  (_ "Open web browser at the Ljubljana Stock Exchange information page for this company")
	  (string-append "http://www.ljse.si/cgi-bin/jve.cgi?SecurityID="
			 (chart-symbol-sans-dot symbol)
			 "&doc=" (lang-select '(("en" "3131")
						("sl" "818")))))))


;;-----------------------------------------------------------------------------
;; misc

;; Daily data available after 15:00 CET
;; per price list format SIT: http://www.ljse.si/cgi-bin/jve.cgi?att=2799
;;                   and EUR: http://www.ljse.si/cgi-bin/jve.cgi?att=4604
(define (ljubljana-available-tdate)
  (tdate-today-after 15 0 (timezone-europe-cet)))

;; pick out the "0002" trading day from a .txt file (same in BTS and TUVL)
(define (ljubljana-txt->tdate body)
  (let ((m (or (regexp-exec (make-regexp "^ 0002 ([0-9][0-9])\\.([0-9][0-9])\\.([0-9][0-9][0-9][0-9]) " regexp/newline) body)
	       (error "Ljubljana: txt file missing 0002 date line"))))
    (ymd->tdate (string->number (match:substring m 3))
		(string->number (match:substring m 2))
		(string->number (match:substring m 1)))))


;;-----------------------------------------------------------------------------
;; latest - BTS whole day
;;

;; VALUE is a string like "1234.50" which is a latest index value, and
;; change-% is a string like "-0.09" which is a percentage change from the
;; previous index value.  Return a change amount as a string in points
;; instead of percentage.
;;
(define (change%->change value change-%)
  (let ((decimals (string-decimals value)))
    (set! value    (string->number-err value))
    (set! change-% (string->number-err change-%))

    ;; value = prev * (1 + change%/100)
    ;; so       prev = value * 100/(100+change%)
    ;; want     change = value - prev
    ;; which is change = value * (1 - 100/(100+change%))
    ;;                 = value * -change% / (100+change%)
    ;;                 = value * change% / (100+change%)
    ;;
    (format #f "~,vf" decimals (* value (/ change-% (+ change-% 100.0))))))

(define (L0010->latest headers tdate line)
  (receive-list (line-code symbol name value change change-%)
      (string-split-widths line '(5 9 41 16 16 16))

    ;; charset not specified, but is codepage 1250
    (set! name (charset->utf8 "windows-1250" name))

    (set! value  (string-comma->dot value))
    (set! change (string-comma->dot change))

    ;; as of May 2007 the change field is empty, but the change % field is
    ;; supplied
    (if (string-null? change)
	(begin
	  (set! change-% (string-comma->dot change-%))
	  (set! change (change%->change value change-%))))

    (latest-new #:symbol     (string-append "^" symbol ".LJ")
		#:name       name
		#:last-tdate tdate
		#:last       value
		#:change     change
		#:limit-decimals 2
		#:source     'ljubljana
		#:headers    headers)))

(define (L0020->latest headers tdate line)
  (receive-list (line-code tier type symbol isin desc dividend
			   note-num average change change-% last-date
			   bid offer high low open close
			   volume volume-offmarket
			   turnover-1000 turnover-bas if-1 if-2 if-%
			   note shares-issued dividend-date
			   ;; p/e principal interest coupon-num discount
			   ;; name+city
			   ;; trading-mode market-maker-cont
			   ;; transactions NAV num-units
			   ;; turnover-nonblock-millions
			   )
      (string-split-widths line '(5 5 5 9 21 41 16
				    11 16 16 16 11
				    16 16 16 16 16 16
				    13 13
				    13 13 16 16 16
				    11 13 11
				    ;; 11 21 21 6 16 81
				    ))

    ;; charset not specified, but is codepage 1250
    (set! desc (charset->utf8 "windows-1250" desc))

    ;; bid/offer -1 for market order
    (if (string=? bid "-1")    (set! bid   #f))
    (if (string=? offer "-1")  (set! offer #f))

    ;; empty when no trades
    (if (string-null? volume)
	(set! volume 0))

    ;; dividend date is empty, apparently, otherwise would show "ex" as note
    ;;
    ;; other notes:
    ;; A - cross
    ;; B - block
    ;; o - utilised tax allowance
    ;; NP - data not received
    ;;
    (set! note (cond ((string=? note "S")  (_ "suspended"))
		     ((string=? note "Z")  (_ "halted"))
		     ((string=? note "*")  (_ "10% limit move"))
		     (else                 #f)))

    (latest-new #:symbol      (string-append symbol ".LJ")
		#:name        desc
		#:quote-tdate tdate
		#:bid         (string-comma->dot bid)
		#:offer       (string-comma->dot offer)
		#:last-tdate  tdate
		#:open        (string-comma->dot open)
		#:high        (string-comma->dot high)
		#:low         (string-comma->dot low)
		#:last        (string-comma->dot close)
		#:change      (string-comma->dot change)
		#:volume      volume
		#:limit-decimals 2
		#:note        note
		#:source      'ljubljana
		#:headers     headers)))

(define (bts->latest-list headers body)
  (let* ((line-list (string-split-lines body))
	 (tdate     (ljubljana-txt->tdate body)))

    ;; file format marker
    (or (find (lambda (line)
		(string-prefix? " 0001 110 " line))
	      line-list)
	(error "Ljubljana: BTS file missing 0001 id line"))

    (append
     ;; stocks - these first for latest-quote-tdate in cache below
     (filter-map (lambda (line)
		   (and (string-prefix? " 0020 " line)
			(L0020->latest headers tdate line)))
		 line-list)

     ;; indices
     (filter-map (lambda (line)
		   (and (string-prefix? " 0010 " line)
			(L0010->latest headers tdate line)))
		 line-list))))


;;-----------------------------------------------------------------------------
;; latest - TUVL whole day

(define (L0040->latest headers tdate line)
  (receive-list (line-code symbol type name average high low turnover volume
			   num-trades bid offer)
      (string-split-widths line '(5 9 13 81 16 16 16 13 13
				    13 16 16))

    ;; charset not specified, but is codepage 1250
    (set! name (charset->utf8 "windows-1250" name))

    ;; volume field empty when no trades
    (if (string-null? volume)
	(set! volume 0))

    (latest-new #:symbol      (string-append symbol ".LJ")
		#:name        name
		#:quote-tdate tdate
		#:bid         (string-comma->dot bid)
		#:offer       (string-comma->dot offer)
		#:last-tdate  tdate
		#:high        (string-comma->dot high)
		#:low         (string-comma->dot low)
		#:last        (string-comma->dot average)
		#:volume      volume
		#:limit-decimals 2
		#:source      'ljubljana
		#:headers     headers)))

(define (tuvl->latest-list headers body)
  (let* ((line-list (string-split-lines body))
	 (tdate     (ljubljana-txt->tdate body)))

    ;; file format marker
    (or (find (lambda (line)
		(string-prefix? " 0001 120 " line))
	      line-list)
	(error "Ljubljana: TUVL file missing 0001 id line"))

    (filter-map (lambda (line)
		  (and (string-prefix? " 0040 " line)
		       (L0040->latest headers tdate line)))
		line-list)))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the text files available at
;;
;;     http://www.ljse.si/cgi-bin/jve.cgi?doc=2195
;;
;; which are
;;
;;     http://www.ljse.si/datoteke/BTStecajEUR.txt
;;     http://www.ljse.si/datoteke/TUVLtecajEUR.txt
;;
;; They're 160k and 6k respectively.  The charset isn't specified in the
;; http headers, nor the file format specifications, but it's codepage 1250.
;; The format is described at
;;
;;     http://www.ljse.si/cgi-bin/jve.cgi?att=4604
;;
;; The server provides ETag and Last-Modified, which we use to avoid
;; re-downloading when unchanged.
;;
;; The symbols in each file are cached, so it's usually possible to go
;; straight to the right one.  If some symbols aren't where expected then
;; both are downloaded to recheck.
;;

(define bts-symbols-filename
  (g-build-filename cache-directory "ljubljana-bts"))

(define (bts-symbols-cache! tdate symbol-list)
  (database-call-with-output-file bts-symbols-filename
    (lambda (port)
      (apply format port
	     ";; Chart ljublana cached BTStecajEUR.txt symbols.
;;
;; Downloaded: ~a-~2,'0d-~2,'0d

"
	     (tdate->ymd tdate))
      (write-list-lines symbol-list port))))

(define (bts-symbols-cached)
  (call-with-input-file-or '(0) bts-symbols-filename read))


(define tuvl-symbols-filename
  (g-build-filename cache-directory "ljubljana-tuvl"))

(define (tuvl-symbols-cache! tdate symbol-list)
  (database-call-with-output-file tuvl-symbols-filename
    (lambda (port)
      (apply format port
	     ";; Chart ljublana cached TUVLtecajEUR.txt symbols.
;;
;; Downloaded: ~a-~2,'0d-~2,'0d

"
	     (tdate->ymd tdate))
      (write-list-lines symbol-list port))))

(define (tuvl-symbols-cached)
  (call-with-input-file-or '(0) tuvl-symbols-filename read))


;; return true if LST-X and LST-Y have any elements in common
(define (any-intersection lst-x lst-y)
  (any (lambda (x)
	 (member x lst-y))
       lst-x))  

(define (ljubljana-latest-get symbol-list extra-list proc)
  (define done-bts #f)
  (define done-tuvl #f)

  (define (attempt-tuvl)
    (receive (headers body)
	(apply http-request
	       "http://www.ljse.si/datoteke/TUVLtecajEUR.txt"
	       ;; file seems to be gone as of April 2008, so dummy this out
	       ;; #:want-ok #t
	       (latest-tag-args 'ljubljana symbol-list))
      (if (http-headers-ok? headers)
	  (let ((latest-list (tuvl->latest-list headers body)))
	    (if (not (null? latest-list))
		(let* ((tdate        (ljubljana-txt->tdate body))
		       (tuvl-symbols (cons tdate (map latest-symbol
						      latest-list))))
		  (tuvl-symbols-cache! tdate tuvl-symbols)
		  (proc latest-list)
		  (set! symbol-list (lset-difference equal? symbol-list
						     tuvl-symbols)))))))
    (set! done-tuvl #t))

  (define (attempt-bts)
    (receive (headers body)
	(apply http-request
	       "http://www.ljse.si/datoteke/BTStecajEUR.txt"
	       #:want-ok #t
	       (latest-tag-args 'ljubljana symbol-list))
      (if (http-headers-ok? headers)
	  (let* ((latest-list (bts->latest-list headers body))
		 (tdate       (ljubljana-txt->tdate body))
		 (bts-symbols (cons tdate (map latest-symbol latest-list))))
	    (bts-symbols-cache! tdate bts-symbols)
	    (proc latest-list)
	    (set! symbol-list (lset-difference equal?
					       symbol-list bts-symbols)))))
    (set! done-bts #t))

  (define cached-tuvl (tuvl-symbols-cached))
  (define cached-bts  (bts-symbols-cached))

  (if (any-intersection symbol-list cached-tuvl)
      (attempt-tuvl))
  (if (any-intersection symbol-list cached-bts)
      (attempt-bts))

  (if (and (not (null? symbol-list))
	   (or (not (= (ljubljana-available-tdate) (first cached-tuvl)))
	       (not (= (ljubljana-available-tdate) (first cached-bts)))))
      (begin
	(or done-tuvl
	    (attempt-tuvl))
	(or done-bts
	    (attempt-bts))))

  (proc (map latest-new-unknown symbol-list)))

(define (ljubljana-quote-adate-time symbol)
  (list (tdate->adate (ljubljana-available-tdate)) #,(hms->seconds 24 0 0)))

(latest-handler! #:selector   ljubljana-symbol?
		 #:handler    ljubljana-latest-get
		 #:adate-time ljubljana-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - individual
;;
;; This uses the data archives at
;;
;;     http://www.ljse.si/cgi-bin/jve.cgi?doc=2069
;;
;; which gives a csv download like
;;
;;     http://www.ljse.si/cgi-bin/jve.cgi?sid=cUg7wt8bhOKsEJiy&doc=2137&date1=02.07.2007&date2=04.07.2007&IndexCode=SBI20&SecurityId=ACLG&x=53&y=10
;;
;; "sid" is a session id of some sort, it works to drop it and the x and y.
;; The "csv" option isn't shown on the web page any more, but still works.
;; Thus,
;;
;;     http://www.ljse.si/cgi-bin/jve.cgi?doc=2137&date1=02.07.2007&date2=04.07.2007&IndexCode=SBI20&SecurityId=ACLG&csv=1
;;
;; When there's no trading days in the requested range (public holidays),
;; the file obtained is empty.

(define (podatki-process symbol body)
  (define commodity (chart-symbol-commodity symbol))
  (define index? (string-prefix? "^" symbol))

  (download-process
   #:module        (_ "LJSE")
   #:symbol-list   (list symbol)
   #:currency      "EUR"
   #:last-download #t
   #:row-list
   (map! (lambda (row-str)
	   (receive-list (tdate index-price stock-price)
	       (string-split row-str #\;)
	     (list #:tdate     (m/d/y-str->tdate tdate)
		   #:commodity commodity
		   #:close     (if index? index-price stock-price))))
	 (string-split-lines body))))

;; return a url string, as per the examples shown above
(define (podatki-download-url symbol lo hi)
  (set! symbol (chart-symbol-sans-dot symbol))
  (receive-list (lo-year lo-month lo-day)
      (tdate->ymd lo)
    (receive-list (hi-year hi-month hi-day)
	(tdate->ymd hi)
      (format #f "http://www.ljse.si/cgi-bin/jve.cgi?doc=2137&date1=~2,'0d.~2,'0d.~a&date2=~2,'0d.~2,'0d.~a&IndexCode=~a&SecurityId=~a&csv=1"

	      lo-day lo-month lo-year
	      hi-day hi-month hi-year
	      (if (string-prefix? "^" symbol)     ;; when an index
		  (chart-symbol-sans-hat symbol)
		  "")
	      (if (string-prefix? "^" symbol)     ;; when shares
		  "X"       ;; dummy
		  symbol)))))

;; return true if data found, false if not
(define (podatki-download-chunk symbol lo hi)
  (if (<= lo hi)  ;; unless "back-to" already complete
      (begin
	(download-status (_ "LJSE") (_ "data") symbol
			 (tdate-range-string lo hi))
	(receive (headers body)
	    (http-request (podatki-download-url symbol lo hi)
			  #:want-ok #t)
	  (podatki-process symbol body)))))


;;-----------------------------------------------------------------------------
;; download

(define (ljubljana-download symbol-list)
  (define avail-tdate (ljubljana-available-tdate))
  (for-each (lambda (symbol)
	      (podatki-download-chunk symbol
				      (download-start-tdate symbol)
				      avail-tdate))
	    symbol-list)
  (download-name-from-latest (_ "LJSE") symbol-list))

(define-public (ljubljana-backto want-tdate symbol-list)
  (for-each (lambda (symbol)
	      (podatki-download-chunk symbol
				      want-tdate
				      (download-back-tdate symbol)))
	    symbol-list))

(download-handler! #:selector ljubljana-symbol?
		   #:update   ljubljana-download
		   #:backto   ljubljana-backto)
