;; London Metal Exchange (LME) setups.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart lme)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)        
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart cookies)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart html-form)
  #:use-module (chart http)
  #:use-module (chart https)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart pagebits)
  #:use-module (chart xls))


;; Note: As of July 2007, in https requests to secure.lme.com for the daily
;; metals prices it seems essential to use http/1.1 persistent connections.
;; If "Connection: close" is requested by the client something fishy happens
;; and the connection hangs at about byte 48887 out of about 62110 (waiting
;; for the last 16kbyte tls packet).  This is with either gnutls or openssl
;; and a trace with gnutls shows it just stops sending, though the TCP
;; connection remains up.  Either the default http/1.1 persistence (no
;; Connection header at all) or the compatibility "Connection: keep-alive"
;; style seems to make it better.  Presumably it's something buggy in the
;; server (Microsoft-IIS 6.0).


(define-public (lme-symbol? symbol)
  (string-suffix? ".LME" symbol))

(symbol-timezone! lme-symbol? (_ "London") timezone-london)
(source-help! lme-symbol? (_ "CONTEXT(manual-node): London Metal Exchange"))

(define (plastics-symbol? symbol)
  (member (chart-symbol-commodity symbol) '("PP" "PA" "PE" "PN"
					    "LL" "LA" "LE" "LN")))

(define (lme-plastics-symbol? symbol)
  (and (lme-symbol? symbol)
       (plastics-symbol? symbol)))

(define (lme-metal-symbol? symbol)
  (and (lme-symbol? symbol)
       (not (plastics-symbol? symbol))))

(define lme-cookiejar-filename
  (g-build-filename cache-directory "lme-cookies"))


;;-----------------------------------------------------------------------------
;; weblink - commodity pages

(weblink-handler! lme-symbol?
  (lambda (symbol)
    (list (_ "LME _Commodity Page")
	  (_ "Open web browser at the London Metal Exchange page for this commodity")
	  (let ((str (chart-symbol-commodity symbol)))
	    ;; strip forward months count like "COPPER 3"
	    (set! str (string-trim-right str char-set:digit))

	    (if (member str '("PP" "PA" "PE" "PN"))
		(set! str "polypropylene"))
	    (if (member str '("LL" "LA" "LE" "LN"))
		(set! str "linear_low"))

	    (set! str (string-downcase str))

	    ;; no space for "aluminiumalloy"
	    (set! str (string-delete str #\space))

	    (string-append "http://www.lme.co.uk/" str ".asp")))))


;;-----------------------------------------------------------------------------
;; misc

;; return tdate for available daily report
;; 
;; https://secure.lme.com/Data/community/Dataprices_daily_metals.aspx
;; http://www.lme.co.uk/who_how_ringtimes.asp
;;     Prices after second ring session each trading day, which would be
;;     16:15 maybe, try at 16:30.
;; https://secure.lme.com/Data/community/Dataprices_daily_prices_plastics.aspx
;;     Prices at 14:05 each trading day.
;; As of jul07 there's no last-modified to show when the updates happen.
;;
(define (daily-available-tdate type)
  (case type
    ((metals)   (tdate-today-after 16 30 (timezone-london)))
    ((plastics) (tdate-today-after 14 5  (timezone-london)))
    (else       (error "oops"))))


;; LABEL is "Cash", "3-Month", "15-Month", "27-Month", or "Month 1" for PP/LL
;; add " 3", " 15", " 27" to COMMODITY for months, or nothing for Cash
(define (commodity+label->symbol commodity label)
  (cond ((string=? "Month 1" label)
	 ;; front month
	 (string-append commodity ".LME"))

	((Mmm-str->month (first (string-split label #\space)))
	 => (lambda (month)
	      (let ((mdate (ymd->mdate (month->nearest-year month) month 1)))
		(string-append commodity " " (mdate->MMM_YY-str mdate)
			       ".LME"))))

	((string-prefix? "Month" label)
	 ;; further months not yet handled
	 (string-append commodity "-something.LME"))

	(else
	 (string-append (string-join-notnull (list commodity
						   (string-take-while
						    label char-set:digit)))
			".LME"))))

;; return mdate for STR like "January_2007" or "Jan_07", or #f if not that
;; format
(define (Mmm_yyy-str->mdate str)
  ;; drop "(1)" part of "http://www.lme.co.uk/downloads/May_2007(1).xls"
  (and-let* ((pos (string-index str #\()))
    (set! str (string-take str pos)))
  (and-let* ((split (string-split str #\_))
	     (      (= 2 (length split)))
	     (month (Mmm-str->month (first split)))
	     (year  (yyy-str->year (second split))))
    (ymd->mdate year month 1)))


;;-----------------------------------------------------------------------------
;; secure login
;;
;; This logs in at the data service page
;;
;;     https://secure.lme.com/Data/Community/Login.aspx?ReturnUrl=%2fdata%2fcommunity%2findex.aspx
;;
;; The result is a cookie ".ASPXAUTH" recorded in lme-cookiejar-filename for
;; subsequent use.  An extra cookie with a dummy path is used to note success.


(define (lme-ensure-login)
  ;; lose any cruft
  (cookiejar-expire lme-cookiejar-filename (hms->seconds (* 7 24) 0 0))
  (if (not (lme-logged-in?))
      (lme-login)))

;; return true if we're still logged in
;; consider logged out after 1 hour
;; not sure if that's ok, the server doesn't give its own expiry
(define (lme-logged-in?)
  (any (lambda (cookie)
	 (and (equal? lme-logged-in-magic-path (cookie-path cookie))
	      (<= (cookie-age cookie) #,(hms->seconds 1 0 0))))
       (cookiejar-read lme-cookiejar-filename)))

(define lme-logged-in-magic-path "/chart-lme-logged-in")

;; return the age of a cookie in seconds, ie. how long since it was created
(define (cookie-age cookie)
  (- (current-time) (cookie-ctime cookie)))

(define (lme-login)
  (download-status (_ "LME") (_ "login"))
  (delete-file-maybe lme-cookiejar-filename)
  (let ((login-url "https://secure.lme.com/Data/Community/Login.aspx?ReturnUrl=%2fData%2fcommunity%2findex.aspx")
	(username  (preference-get 'lme-username ""))
	(password  (preference-get 'lme-password "")))

    (if (string-null? username)
	(error "No LME username set in preferences"))

    (receive (headers login-body)
	(http-request login-url
		      #:cookiejar lme-cookiejar-filename
		      #:want-ok #t)

      (let* ((form-alist (html-form-parse login-body))
	     (post-url   (html-form-submit-url login-url form-alist))
	     (data       (html-form-data form-alist
					 `(("_logIn$_userID"   . ,username)
					   ("_logIn$_password" . ,password)))))
	(receive (headers body)
	    (http-request post-url
			  #:method "POST"
			  #:cookiejar lme-cookiejar-filename
			  #:data data
			  #:follow #f)

	  ;; the POST is to the Login.aspx page and success is a redirect to
	  ;; the main data page (/Data/community/index.aspx)
	  ;; so failure is not 302, or no Location, or a Location but
	  ;; containing "Login"
	  ;;
	  (define location (http-headers-get headers "Location"))
	  (if (not (and (eqv? 302 (http-headers-status-code headers))
			location
			(not (string-contains-ci location "Login"))))
	      (error "LME: login failed"))))))

  (cookiejar-add-cookies
   lme-cookiejar-filename
   (list (cookie-new #:path lme-logged-in-magic-path))))


;;-----------------------------------------------------------------------------
;; daily page html parsing

;; Split a daily prices page heading into commodity code and name, the
;; return is a list (COMMODITY NAME).  Only plastics have name parts, eg.
;;     "PP\r\n\t\t\tGlobal" -> "PP" and "Global"
;; where metals are just the whole
;;     "ALUMINIUM ALLOY" -> "ALUMINIUM ALLOY" and #f
;;
(define (daily-heading->commodity+name heading)
  ;; lose tabs and newlines
  (set! heading (string-trim-both (string-collapse-whitespace heading)))
  (and (not (string-null? heading))
       (let* ((m         (string-match "^([A-Z][A-Z]+( [A-Z][A-Z]+)*)(.*)$"
				       heading))
	      (commodity (if m (string-trim-right (match:substring m 1))
			     heading))
	      (name      (and m
			      (not (string-null? (match:substring m 3)))
			      (string-trim (match:substring m 3)))))
	 (list commodity name))))


;; return list ((SYMBOL NAME TDATE BUY-STR SELL-STR MDATE) ...) from BODY
;; data, or #f
;;
(define (daily-html-parse body)
  (define ret '())

  ;; date is: "26 Jan 2005 (Data >1 day old)   </b></td>"
  ;; or:      "3 Feb 2005   </b></td>"
  (let* ((m        (must-match (string-match " Prices[ ,][^\n]*for +([0-9]+) ([A-Za-z]+) ([0-9][0-9][0-9][0-9])" body)))
	 (tdate    (ymd->tdate (string->number (match:substring m 3))
			       (Mmm-str->month (match:substring m 2))
			       (string->number (match:substring m 1))))
	 (row-list (html-table-rows body (match:end m))))

    ;; blank separator lines
    (set! row-list (remove! (lambda (row)
			      (every string-null? row))
			    row-list))

    (let ((commodity-list (map daily-heading->commodity+name
			       (first row-list))))
      (set! row-list (cdr row-list))

      (for-each-two
       (lambda (buyer-row seller-row)
	 ;; row like ("" "September Buyer" "" "932" "" "931" "")
	 ;;          ("" "Cash buyer" "" "1,555.00" "" "1,737.00" "" ...)
	 (for-each
	  (lambda (commodity+name buy sell)
	    (if commodity+name
		(receive-list (commodity name)
		    commodity+name

		  (define symbol (commodity+label->symbol
				  commodity (second buyer-row)))
		  (define (lat sym)
		    (set! ret (cons (list sym name tdate buy sell
					  (chart-symbol-mdate symbol))
				    ret)))

		  (set! buy  (crunch-price buy))
		  (set! sell (crunch-price sell))

		  ;; first row as front month
		  (if (and (eq? buyer-row (first row-list))
			   (chart-symbol-mdate symbol))
		      (lat (string-append commodity ".LME")))

		  ;; all rows with month in symbol
		  (lat symbol))))

	  commodity-list buyer-row seller-row))
       row-list)))

  (and-let* ((m (string-match "LMEX Index value [^0-9\n]*([0-9]+ [A-Za-z]+ [0-9][0-9][0-9][0-9])[^0-9.\n]+([0-9.]+)" body)))
    (set! ret (cons (list "LMEX.LME"
			  #f
			  (d/m/y-str->tdate (match:substring m 1))
			  #f ;; no separate buy price
			  (match:substring m 2)
			  #f)
		    ret)))

  ret)


;;-----------------------------------------------------------------------------
;; daily page
;;
;; The latest daily settlements are at
;;
;; https://secure.lme.com/Data/community/Dataprices_daily_metals.aspx
;; https://secure.lme.com/Data/community/Dataprices_daily_prices_plastics.aspx
;;

(define (daily-latest-parse body)
  (list
   (cons 'form   (html-form-parse body))
   (cons 'prices (daily-html-parse body))))

(define (daily-latest-info type)
  (lme-ensure-login)

  (pagebits-read #:filename  (case type
			       ((metals)   "lme-latest-metals")
			       ((plastics) "lme-latest-plastics"))
		 #:status    (list (_ "LME")
				   (case type
				     ((metals)   (_ "metals latest"))
				     ((plastics) (_ "plastics latest"))))
		 #:url       (list (case type
				     ((metals)   "https://secure.lme.com/Data/community/Dataprices_daily_metals.aspx")
				     ((plastics) "https://secure.lme.com/Data/community/Dataprices_daily_prices_plastics.aspx"))
				   #:cookiejar lme-cookiejar-filename
				   #:follow    #f)
		 #:timezone  (timezone-london)
		 #:parse     daily-latest-parse))

;; return `daily-html-parse'd contents
(define (lme-daily-download-tdate type tdate)

  (receive-list (year month day)
      (tdate->ymd tdate)

    (receive (headers body)
	(case type
	  ((metals)
	   (let* ((alist (daily-latest-info type))
		  (url   (assq-ref alist 'url))
		  (form  (assq-ref alist 'form))
		  (data  (html-form-data
			  form `(("_searchForm$_lstdate"  . ,(number->string day))
				 ("_searchForm$_lstmonth" . ,(number->string month))
				 ("_searchForm$_lstyear"  . ,(number->string year))
				 ("_searchForm$_search"   . "")))))

	     (lme-ensure-login)
	     (download-status (_ "LME") (_ "daily metals")
			      (tdate-range-string tdate tdate))
	     (http-request url
			   #:method "POST"
			   #:cookiejar lme-cookiejar-filename
			   #:want-ok #t
			   #:data data)))

	  ((plastics)
	   (lme-ensure-login)
	   (download-status (_ "LME") (_ "daily plastics")
			    (tdate-range-string tdate tdate))
	   (http-request (format #f "https://secure.lme.com/Data/community/Dataprices_daily_prices_plastics.aspx?sDate=~a&sMonth=~a&sYear=~a" day month year)
			 #:cookiejar lme-cookiejar-filename
			 #:want-ok #t))

	  (else
	   (error "oops, unknown type" type)))

      (or (http-headers-ok? headers)
	  (error "LME: oops, daily download failed:" headers))

      ;; for a public holiday there's a sorry message and the date and data
      ;; supplied is instead the latest available
      (if (string-contains body
			   "Sorry, there is no LME market data reported for")
	  '()
	  (let ((lst (daily-html-parse body)))
	    (if (and lst
		     (not (= tdate (third (first lst)))))
		(error "LME: oops, daily download wrong date:"
		       tdate (third (first lst))))
	    lst)))))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the daily prices in the login "free data service" from (login
;; requried)
;;
;;     https://secure.lme.com/Data/community/Dataprices_daily_metals.aspx
;;
;; This plain url gives the most recent prices, which we take as a quote for
;; the indicated day then work back with form-data fetching previous days to
;; find price change amounts.  Or the database is used if it covers the
;; desired symbol(s).
;;
;; Unfortunately there's no Last-Modified or ETag to save refetching if the
;; latest GET contents have not yet updated.  (???)


;; return the tdate from a parsed daily data list
(define (data-tdate data)
  (receive-list (symbol name tdate buy sell mdate)
      (first data)
    tdate))

(define (daily-from-database symbol-list)
  (define data
    (map (lambda (symbol)
	   (let* ((latest (latest-get symbol))
		  (buy    (latest-bid latest))
		  (sell   (or (latest-offer latest)
			      (latest-last latest))))
	     (list symbol
		   (latest-name latest)
		   (latest-last-tdate latest)
		   (and buy
			(number+decimals->string buy
						 (latest-decimals latest)))
		   (and sell
			(number+decimals->string sell
						 (latest-decimals latest)))
		   (latest-contract-mdate latest))))
	 symbol-list))

  (if (and (third (first data))
	   (apply eqv? (map third data)))
      data
      #f))

(define (lme-type->selector type)
  (case type
    ((metals)   lme-metal-symbol?)
    ((plastics) lme-plastics-symbol?)))


(define (lme-latest-update-database type newest-data prev-data)
  (let* ((end-tdate   (data-tdate newest-data))
	 (start-tdate (if prev-data
			  (data-tdate prev-data)
			  end-tdate))
	 (db-list     (download-also '() #:selector (lme-type->selector type)
				     #:start-tdate start-tdate
				     #:end-tdate end-tdate)))
    (if prev-data
	(daily-process db-list prev-data))
    (daily-process db-list newest-data)))

(define (lme-latest-process newest-data prev-data proc)
  (define lst '())

  (for-each
   (lambda (elem)
     (receive-list (symbol name tdate buy sell mdate)
	 elem

       (and-let* ((prev-elem (assoc symbol prev-data))) ;; match car
	 (let ((prev-sell (fifth prev-elem)))

	   (receive-list (decimals buy sell prev-sell)
	       (strings->numbers+decimals buy sell prev-sell)

	     ;; need both buy and sell to show as quote
	     (define bid         buy)
	     (define offer       (and buy sell))
	     (define quote-tdate (and buy sell tdate))

	     ;; sell is normally always present, but have seen entire page
	     ;; blank (empty fields "" which become #f) 31aug05 after
	     ;; 29aug05 bank holiday

	     (set! lst
		   (cons (latest-new #:symbol         symbol
				     #:name           name
				     #:quote-tdate    quote-tdate
				     #:bid            bid
				     #:offer          offer
				     #:last-tdate     tdate
				     #:last           sell
				     #:prev           prev-sell
				     #:decimals       decimals
				     #:contract-mdate mdate
				     #:source         'lme)
			 lst)))))))
   newest-data)

  (proc lst))

(define (lme-latest-type symbol-list type proc)

  (and-let* ((newest-data  (assq-ref (daily-latest-info type) 'prices)))

    ;; COVERED-TDATE is the data we already have for all of SYMBOL-LIST (or
    ;; rather for the worst among that list), default to a dummy 100 days
    ;; ago
    (let* ((newest-tdate  (data-tdate newest-data))
	   (covered-data  (daily-from-database symbol-list))
	   (covered-tdate (if covered-data
			      (data-tdate covered-data)
			      (- (daily-available-tdate type) 100))))
      (let more ((attempt 1))
	(if (> attempt 5)
	    (error "LME: can't find previous daily data"))

	(let ((prev-tdate (- newest-tdate attempt)))
	  (if (>= covered-tdate prev-tdate)
	      (begin
		(lme-latest-update-database type newest-data #f)
		(lme-latest-process newest-data covered-data proc))

	      (let ((prev-data (lme-daily-download-tdate type prev-tdate)))
		(if prev-data
		    (begin
		      (lme-latest-update-database type newest-data prev-data)
		      (lme-latest-process newest-data prev-data proc))

		    (more (1+ attempt))))))))))

(define (lme-symbol->type symbol)
  (if (lme-metal-symbol? symbol) 'metals 'plastics))

(define (lme-latest-get symbol-list extra-list proc)

  (if (string-null? (preference-get 'lme-username ""))
      (proc (map (lambda (symbol)
		   (latest-new #:symbol symbol
			       #:note   (_ "must register")
			       #:source 'lme))
		 (append symbol-list extra-list)))

      ;; look for one or both metal and plastics in symbol-list, do the two in
      ;; the order they appear in SYMBOL-LIST
      (for-each (lambda (type)
		  (lme-latest-type symbol-list type proc))
		(delete-duplicates (map lme-symbol->type symbol-list)))))

(define (lme-quote-adate-time symbol)
  (list (tdate->adate (daily-available-tdate (lme-symbol->type symbol))) #f))

(latest-handler! #:selector   lme-symbol?
		 #:handler    lme-latest-get
		 #:adate-time lme-quote-adate-time)


;;-----------------------------------------------------------------------------
;; historical download page
;;
;; This uses the historical data at
;;
;;     http://www.lme.co.uk/dataprices_historical.asp
;;
;; That page is downloaded to get urls of XLS files for prices and volumes
;; for each calendar month.  A price file is like
;;
;;     http://www.lme.co.uk/downloads/January_2007.xls
;;
;; and a volumes file
;;
;;     http://www.lme.co.uk/downloads/volumes_Jan_08.xls
;;

(define historical-xls-page "http://www.lme.co.uk/dataprices_historical.asp")

;; BODY is the "dataprices_historical.asp" page, return a single-entry alist
;; (("xls-urls" . URL-LIST)) which is the various xls file urls
;;
;; At the start of 2007 there was nothing available (the previous 2006 files
;; being made chargable items) so it's possible for URL-LIST to be empty.
;;
;; Files like stocks_2007.xls and Mar2007.xls are warehouse stocks, so
;; demand a "_" and not starting "stocks".
;;
(define (historical-xls-parse body)
  (list
   (cons "xls-urls"
	 (delete-duplicates!
	  (filter-map (lambda (m)
			(and-let* ((link (match:substring m 1))
				   (     (string-index link #\_))
				   (     (not (string-prefix? "stocks" link))))
			  (http-url-relative historical-xls-page link)))
		      (list-matches "href=\"([^\"]*\\.xls)\""
				    body))))))

;; return list of urls
;;
;; (The filename "lme-historical-xls-2" has a "-2" because the format
;; changed from an initial "lme-historical-xls".)
;;
(define (historical-xls-urls)
  (assoc-ref (pagebits-read #:filename  "lme-historical-xls-2"
			    #:status    (list (_ "LME")
					      (_ "historical downloads page"))
			    #:url       (list historical-xls-page
					      #:method "POST"
					      #:data "disclaimer=agreed")
			    #:timezone  (timezone-london)
			    #:parse     historical-xls-parse
			    ;; only check every two days
			    #:frequency 2)
	     "xls-urls"))


;;-----------------------------------------------------------------------------
;; download - volume
;;
;; This crunches files like
;;     http://www.lme.co.uk/downloads/volumes_Jan_08.xls


;; Call (PROC idx...) for each entry in ARRAY.
;; Currently only works for arrays of one dimension.
;;
(define (array-for-each-index proc array)
  (receive-list (lo hi)
      (first (array-shape array))
    (do ((i lo (1+ i)))
	((> i hi))
      (proc i))))

;; May 2008 has gone to 1-May-08 string date in column 0,
;; prior to that (eg. Jan 2008) column 0 was #f and column 1
;; was an msdate number
;;
;; And avoid blanks at end of month, and avoid "Total" row.
;;
(define (volume-row-tdate array)
  (cond ((not (array-ref array 0))
	 (msdate->tdate (array-ref array 1)))
	((regexp-exec #,(regexp "^[0-9].*[0-9]$") (array-ref array 0))
	 (d/m/y-str->tdate (array-ref array 0)))
	(else
	 #f)))

(define (volume-process-body symbol-list url body)
  (let ((commodities-row    #f)
	(type-row           #f)
	(column-list        '())
	(column-commod-list '())
	(data               '()))

    (define (find-futures-columns)
      ;; style up to Jun 2007, two rows, first commodity name then second
      ;; "FUT" etc for futures/options, etc
      (cond ((member "FUT" (array->list type-row))
	     (array-for-each-index
	      (lambda (i)
		(and-let* ((          (equal? "FUT" (array-ref type-row i)))
			   (commodity (array-ref commodities-row i)))
		  (set! commodity (string-upcase commodity))

		  (cond ((equal? commodity "ALLOY")
			 (set! commodity "ALUMINIUM ALLOY"))
			((equal? commodity "POLYPROPYLENE")
			 (set! commodity "PP"))
			((equal? commodity "LINEAR LOW")
			 (set! commodity "LL")))

		  (set! column-list        (cons i column-list))
		  (set! column-commod-list (cons commodity column-commod-list))))
	      commodities-row))

	    ;; style of Jul 2007, one row of "AAFUT" for aluminium alloy, etc
	    ((any (lambda (entry)
		    (and (string? entry)
			 (string-suffix? "FUT" entry)))
		  (array->list type-row))
	     (array-for-each-index
	      (lambda (i)
		(and-let* ((entry     (array-ref type-row i))
			   (          (string? entry)) ;; skip date headings
			   (          (string-suffix? "FUT" entry))
			   (commodity (string-drop-right entry 3)))
		  (set! commodity (or (assoc-ref '(("AA" . "ALUMINIUM ALLOY")
						   ("NA" . "NASAAC")
						   ("AH" . "ALUMINIUM")
						   ("CA" . "COPPER")
						   ("NI" . "NICKEL")
						   ("PB" . "LEAD")
						   ("SN" . "TIN")
						   ("ZS" . "ZINC"))
						 commodity)
				      commodity))

		  (set! column-list        (cons i column-list))
		  (set! column-commod-list (cons commodity column-commod-list))))
	      type-row))))

    (xls-for-each-row
     (lambda (array)
       (cond ((null? column-list)
	      (set! commodities-row type-row)
	      (set! type-row array)
	      (if commodities-row
		  (find-futures-columns)))

	     ((volume-row-tdate array)
	      => (lambda (tdate)
		   (for-each
		    (lambda (col commodity)
		      (set! data
			    (cons (list #:tdate     tdate
					#:commodity commodity
					#:volume    (floor-exact
						     (array-ref array col)))
				  data)))
		    column-list column-commod-list)))))
     body)

    (if (null? column-list)
	(error "LME volumes: no recognised columns"))

    (download-process
     #:module      (_ "LME")
     #:symbol-list symbol-list
     #:info-alist  (list (cons 'volumes-url url))
     #:row-list    data)))

(define (volumexls-url->mdate url)
  (let ((base (basename url ".xls")))
    (and (string-prefix? "volumes_" base)
	 (Mmm_yyy-str->mdate (string-drop base 8)))))

(define (volumexls-url<? x y)
  (< (or (volumexls-url->mdate x) 0)
     (or (volumexls-url->mdate y) 0)))

(define (volume-download symbol-list)
  (for-each (lambda (url)
	      (and-let* ((url-mdate (volumexls-url->mdate url))
			 (url-tdate (mdate->last-tdate url-mdate)))
		(if (any (lambda (symbol)
			   (let ((vol-tdate (database-last-volume symbol)))
			     (or (not vol-tdate)
				 (> url-tdate vol-tdate))))
			 symbol-list)
		    (begin
		      (download-status (_ "LME") (basename url))
		      (set! symbol-list
			    (download-also symbol-list
					   #:start-tdate (mdate->tdate
							  url-mdate)
					   #:end-tdate   url-tdate
					   #:selector lme-symbol?))
		      (receive (headers body)
			  (http-request url)
			(if (http-headers-ok? headers)
			    (volume-process-body symbol-list url body)))))))
	    (sort (historical-xls-urls) volumexls-url<?)))


;;-----------------------------------------------------------------------------
;; download - monthly
;;
;; This crunches files like
;;     http://www.lme.co.uk/downloads/January_2007.xls


;; same as `array-ref' but return #f if INDICES are out of bounds
(define-public (array-ref-or-f array . indices)
  (and (apply array-in-bounds? array indices)
       (apply array-ref array indices)))

;; Return tdate of anticipated available montly .xls download, that being
;; the end of the previous month.
;;
;; Don't know exactly when a new month full of data becomes available,
;; assume here midnight at the start of the second trading day of the new
;; month.
;;
(define (monthxls-available-tdate)
  ;; last trading day of prev month
  (1- (mdate->tdate
       ;; current month, but not until second day in
       (tdate->mdate (1- (tdate-today (timezone-london)))))))

(define (monthxls-line->commodity array)
  (and-let* ((key (find string? (array->list array))))
    (any (lambda (elem)
	   (and (string-contains-ci key (car elem))
		(cdr elem)))
	 '(("LME COPPER"                                 . "COPPER")
	   ("LME Aluminium Alloy"                        . "ALUMINIUM ALLOY")
	   ("LME North American Special Aluminium Alloy" . "NASAAC")
	   ("LME SHG Zinc"                               . "ZINC")
	   ("LME PRIMARY ALUMINIUM"                      . "ALUMINIUM")
	   ("LME LEAD"                                   . "LEAD")
	   ("LME TIN"                                    . "TIN")
	   ("LME PRIMARY NICKEL"                         . "NICKEL")
	   ("LME PP"                                     . "PP")
	   ("LME LL"                                     . "LL")))))

(define (monthxls-headings-metal? array)
  (let ((lst (array->list array)))

    ;; "Aluminium Alloy" headings have "27 MONTHS " with a space rather than
    ;; "-", and trailing space
    (set! lst (map (lambda (str)
		     (and (string? str)
			  (string-upcase
			   (string-filter str char-set:letter+digit))))
		   lst))

    (cond ((list-prefix? '(#f #f
			      "CASH" "CASH" "MEAN"
			      "3MONTHS" "3MONTHS" "MEAN"
			      "15MONTHS" "15MONTHS" "MEAN"
			      "27MONTHS" "27MONTHS" "MEAN")
			 lst)
	   '((2 "") (5 " 3") (8 " 15") (11 " 27")))

	  ;; only up to 15 months for TIN and LEAD
	  ((list-prefix? '(#f #f
			      "CASH" "CASH" "MEAN"
			      "3MONTHS" "3MONTHS" "MEAN"
			      "15MONTHS" "15MONTHS" "MEAN")
			 lst)
	   '((2 "") (5 " 3") (8 " 15")))

	  (else
	   #f))))

(define (month-month-mean? lst)
  (and (length>=? lst 3)
       (string? (first lst))
       (let ((month (Mmm-str->month (string-trim-both (first lst)))))
	 (and (Mmm-str->month (string-trim-both (second lst)))
	      (string-ci=? "Mean" (string-trim-both (third lst)))
	      month))))

(define (monthxls-headings-plastic? array)
  (define lst (array->list array))
  (let more ((col 2)
	     (ret '()))
    (cond ((month-month-mean? (drop-maybe lst col))
	   => (lambda (month)
		(more (+ col 3)
		      (cons (list col month)
			    ret))))
	  (else
	   (and (not (null? ret))
		ret)))))
    
(define (monthxls-price->str price)
  (and price ;; possibly #f from expired LL,PP futures
       (let* ((str  (number->string price))
	      (dpos (string-index str #\.)))
	 (if dpos
	     (set! str (string-take str (min (+ dpos 2)
					     (string-length str)))))
	 str)))    
  
;; return an MDATE for MONTH, on or after PREV-MDATE
(define (month->mdate-after prev-mdate month)
  (+ prev-mdate (modulo (- (1- month) prev-mdate) 12)))

;; BODY is a month .xls file, return a list of elements (#:commodity ...)
;; suitable for download-process
;;
(define (monthxls->keylist body)
  (let ((commodity #f)
	(cols      '())
	(process   noop)
	(ret       '()))

    (define (return-metal msuff tdate price)
      (set! ret (cons (list #:commodity (string-append commodity msuff)
			    #:tdate     tdate
			    #:close     (monthxls-price->str price))
		      ret)))

    (define (process-metal array)
      (let ((tdate (msdate->tdate (array-ref array 1))))
	(for-each (lambda (elem)
		    (receive-list (col msuff)
			elem
		      (return-metal msuff tdate (array-ref array col))))
		  cols)))

    (define (return-plastic month tdate price)
      ;; check close is not #f for empty rows in Jun07
      (and-let* ((close (monthxls-price->str price)))
	(let ((mdate (month->mdate-after (tdate->mdate tdate) month)))
	  (set! ret (cons (list #:commodity commodity
				#:tdate     tdate
				#:mdate     mdate
				#:close     close)
			  ret)))))

    (define (process-plastic array)
      (let ((tdate (msdate->tdate (array-ref array 1))))
	(for-each (lambda (elem)
		    (receive-list (col month)
			elem
		      (return-plastic month tdate (array-ref array col))))
		  cols)))

    (xls-for-each-row
     (lambda (array)
       (cond ((and (> (array-last-index array) 2)
		   (number? (array-ref array 1)))
	      (process array))

	     ((monthxls-line->commodity array)
	      => (lambda (c)
		   (set! commodity c)))

	     ((monthxls-headings-metal? array)
	      => (lambda (c)
		   (set! cols c)
		   (set! process process-metal)))

	     ((monthxls-headings-plastic? array)
	      => (lambda (c)
		   (set! cols c)
		   (set! process process-plastic)))

	     ((equal? "Average" (array-ref-or-f array 1))
	      (set! process noop))))
     body)

    ret))

(define (monthxls-process url-mdate symbol-list body)
  (download-process
   #:module      (_ "LME")
   #:symbol-list symbol-list
   #:lo          (mdate->tdate url-mdate)
   #:hi          (mdate->last-tdate url-mdate)
   #:currency    "USD"
   #:row-list    (monthxls->keylist body)))

(define (monthxls-url->mdate url)
  (Mmm_yyy-str->mdate (basename url ".xls")))

(define (monthxls-url<? x y)
  (< (or (monthxls-url->mdate x) 0)
     (or (monthxls-url->mdate y) 0)))

(define (monthxls-download symbol-list)
  (for-each (lambda (url)
	      (and-let* ((url-mdate (monthxls-url->mdate url))
			 (url-tdate (mdate->last-tdate url-mdate)))
		(if (any (lambda (symbol)
			   (>= url-tdate (download-start-tdate symbol)))
			 symbol-list)
		    (begin
		      (download-status (_ "LME") (_ "data") (basename url))
		      (set! symbol-list
			    (download-also symbol-list
					   #:start-tdate (mdate->tdate
							  url-mdate)
					   #:end-tdate   url-tdate
					   #:selector lme-symbol?))
		      (receive (headers body)
			  (http-request url)
			(if (http-headers-ok? headers)
			    (monthxls-process url-mdate symbol-list body)))))))
	    (sort (historical-xls-urls) monthxls-url<?)))


;;-----------------------------------------------------------------------------
;; download - historical prices and/or volumes

;; return tdate of last volume value recorded for SYMBOL, or #f if none ever
(define (database-last-volume symbol)
  (and-let* ((series (database-read-series symbol)))
    (series-array series   ;; initial request past month to look at
		  (- (series-hi series) 25)
		  (series-hi series))
    (let more ((i (series-hi series)))
      (and (>= i (series-lo series))
	   (if (array-ref (series-array series i i) i 4)
	       i
	       (more (1- i)))))))

(define (lme-historical-download symbol-list)

  (let* ((avail-tdate (monthxls-available-tdate)))

    ;; whether can update prices for SYMBOL using xls
    (define (want-monthxls? symbol)
      (>= avail-tdate (download-start-tdate symbol)))

    ;; whether can update volume for SYMBOL
    (define (want-volume? symbol)
      ;; no volumes for forward symbols like "ZINC 3.LME" or futures
      ;; specific symbols like "PP MAY 06.LME"
      (and (not (string-any char-numeric? symbol))
	   (let ((last-tdate (database-last-volume symbol)))
	     (or (not last-tdate)
		 (< last-tdate avail-tdate)))))

    ;; whether can update anything for SYMBOL
    (define (want-update? symbol)
      (or (want-monthxls? symbol)
	  (want-volume? symbol)))

    (if (any want-update? symbol-list)
	(begin
	  (if (any want-monthxls? symbol-list)
	      (monthxls-download symbol-list))
	  (if (any want-volume? symbol-list)
	      (volume-download symbol-list))))))


;;-----------------------------------------------------------------------------
;; download - daily
;;
;; This uses the metals and plastics settlement pages (login required) at
;;
;;     https://secure.lme.com/Data/community/Dataprices_daily_metals.aspx
;;     https://secure.lme.com/Data/community/Dataprices_daily_plastics.aspx


;; LST has elements (SYMBOL NAME TDATE BUY-STR SELL-STR MDATE) per
;; `daily-html-parse'
;;
;; The sell price is used.  The report for cash prices has the seller marked
;; as the settlement and for the forwards the historical files can be seen
;; with the seller price.
;;
(define (daily-process symbol-list lst)
  (download-process
   #:module      (_ "LME")
   #:symbol-list symbol-list
   #:currency    "USD"
   #:row-list
   (map (lambda (row)
	  (receive-list (symbol name tdate buy sell mdate)
	      row
	    (list #:tdate     tdate
		  #:mdate     mdate
		  #:commodity (chart-symbol-commodity symbol)
		  #:close     sell)))
	lst)))

(define (lme-daily-download symbol-list type)
  (define selector (case type
		     ((metals)   lme-metal-symbol?)
		     ((plastics) lme-plastics-symbol?)))

  (set! symbol-list (filter selector symbol-list))
  (if (not (null? symbol-list))

      (let* ((end-data  (assq-ref (daily-latest-info type) 'data))
	     (end-tdate (if end-data
			    (data-tdate end-data)
			    (daily-available-tdate type))))

	(set! symbol-list
	      (download-also symbol-list #:selector selector))

	;; only go back 25 days for LMEX or others without yearly data,
	;; since at 70kbytes per day it quickly becomes slow
	;;
	(do ((t (apply min (map (lambda (symbol)
				  (download-start-tdate symbol #:initial 25))
				symbol-list))
		(1+ t)))
	    ((>= t end-tdate))
	  (and-let* ((data (lme-daily-download-tdate type t)))
	    (daily-process symbol-list data)))

	(if end-data
	    (daily-process symbol-list end-data)))))


;;-----------------------------------------------------------------------------
;; download

(define (lme-download symbol-list)
  (lme-historical-download symbol-list)

  (if (https-available?)
      (begin
	(lme-daily-download symbol-list 'metals)
	(lme-daily-download symbol-list 'plastics))
      (download-message
       (string-append
	(_ "LME") ": " (_ "further data available to registered users")))))

(define (lme-backto want-tdate symbol-list)
  (download-message
   (string-append
    (_ "LME") ": "
    (_ "no further historical data available"))))

(define (lme-download-available-tdate)
  (if (https-available?)
      (daily-available-tdate 'metals)
      (monthxls-available-tdate)))

(download-handler! #:selector   lme-symbol?
		   #:update     lme-download
		   #:backto     lme-backto
		   #:avail-proc lme-download-available-tdate)

(download-now-handler! (lambda (symbol-list)
			 (and (any lme-symbol? symbol-list)
			      (download-now-all-commodities-and-months))))
