;; Main GUI.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart main)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart graph)
  #:autoload (chart about)          (about-dialog-popup)
  #:autoload (chart ann-edit)       (annotations-edit-popup)
  #:use-module (chart config)
  #:autoload (chart cross)          (cross-drag-button cross-set-state!)
  #:autoload (chart database)       (database-read-series)
  #:autoload (chart delete)         (delete-dialog-popup)
  #:autoload (chart diagnostics)    (diagnostics-dialog-popup)
  #:autoload (chart div-edit)       (dividends-edit-popup)
  #:autoload (chart download)       (source-help-list)
  #:autoload (chart down-dialog)    (download-dialog-popup download-dialog-one)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-glib)
  #:use-module (chart extra-gtk)
  #:use-module (chart gnome)
  #:use-module (chart gnome-ui)
  #:use-module (chart gui-common)
  #:autoload (chart indicator-menu) (indicator-menu-popup)
  #:autoload (chart intraday)       (intraday-popup)
  #:use-module (chart keymap)
  #:autoload (chart keys)           (keys-popup)
  #:autoload (chart lasso)          (lasso-start-button)
  #:autoload (chart latest)         (latest-request-symbols latest-purge-old)
  #:autoload (chart line-edit)      (line-edit-or-new-button)
  #:autoload (chart scroll-drag)    (scroll-drag-start-button)
  #:use-module (chart misc)
  #:autoload (chart open)           (open-dialog-popup)
  #:autoload (chart prefs-edit)     (pref-dialog-popup)
  #:use-module (chart symlists)
  #:use-module (chart view)
  #:autoload (chart viewstyle)      (viewstyle-dialog-popup)
  #:autoload (chart watchlist)      (watchlist-dialog-popup))


;;-----------------------------------------------------------------------------
;; initializations

(define-public program-object
  (gnome-program-init "chart" chart-version (program-arguments) '()))

(gnome-program-set-property program-object 'app-datadir chart-config-datadir)


;;-----------------------------------------------------------------------------
;; manual nodes

;; return the node name in the manual for the current symbol's data source,
;; or #f if none
(define (manual-data-source-node)
  (and-let* ((symbol (view-symbol outer)))
    (symbol-setups symbol)
    (cond ((sympred-assoc symbol source-help-list) => second)
	  (else #f))))

;; return the node name in the manual for the current moving average (the
;; upper graph), or #f if none
(define (manual-average-node)
  (and-let* ((          (view-symbol outer)) ;; when showing something
	     (viewstyle (preference-get 'viewstyle viewstyle-default))
	     (datatype  (datatype-find (first (second (first viewstyle))))))
    (datatype-manual datatype)))

;; return the node name in the manual for the current indicator (the lower
;; graph), or #f if none
(define (manual-indicator-node)
  (and-let* ((          (view-symbol outer)) ;; when showing something
	     (viewstyle (preference-get 'viewstyle viewstyle-default))
	     (datatype  (datatype-find (first (first (second viewstyle))))))
    (datatype-manual datatype)))


;;-----------------------------------------------------------------------------
;; zoom

(define (zoom-in-x)
  (view-scale-x-step! outer (1+ scale-x-step)))

(define (zoom-out-x)
  (if (> scale-x-step 1)
      (view-scale-x-step! outer (max 1 (1- scale-x-step)))))

(define (zoom-out-y)
  (for-each (lambda (graph)
	      (graph-scale-y-factor!
	       graph (* 0.5 (graph-scale-y-factor graph))))
	    (view-graph-list outer)))

(define (zoom-in-y)
  (for-each (lambda (graph)
	      (graph-scale-y-factor!
	       graph (* 2.0 (graph-scale-y-factor graph))))
	    (view-graph-list outer)))

(define-public (zoom-in)
  (zoom-in-x)
  (zoom-in-y))

(define-public (zoom-out)
  (zoom-out-x)
  (zoom-out-y))

(define (zoom-out-y-lower)
  (graph-scale-y-factor! drawvol (* 0.5 (graph-scale-y-factor drawvol))))

(define (zoom-in-y-lower)
  (graph-scale-y-factor! drawvol (* 2.0 (graph-scale-y-factor drawvol))))


;;-----------------------------------------------------------------------------
;; main gui

(define-public app-widget
  (gnome-app-new (gnome-program-app-id program-object) (_ "Chart")))
;; default initial size 9/10 of screen, and restrict to aspect ratio 4:3 so
;; as not to use whole width of a wide screen
(let ((width  (floor-exact (* 9/10 (gdk-screen-width-mm))))
      (height (floor-exact (* 9/10 (gdk-screen-height-mm)))))
  (set! width  (min width  (* 4/3 height)))
  (set! height (min height (* 3/4 width)))
  (set! width  (* width  (/ (gdk-screen-width) (gdk-screen-width-mm))))
  (set! height (* height (/ (gdk-screen-height) (gdk-screen-height-mm))))
  (gtk-window-set-default-size app-widget
			       (round-exact width) (round-exact height)))

(define app-widget-icon-keep #f)
(c-gtk-signal-connect-once app-widget "realize"
  (lambda ()
    (let* ((win        (gtk-widget-window app-widget))
	   (bitmap-loc (vector #f))
	   (pixmap     (gdk-pixmap-create-from-xpm
			win bitmap-loc #f
			(g-build-filename chart-config-pixmapsdir
					  "chart.xpm"))))
      (false-if-exception ;; guile-gtk 0.40 doesn't allow #f for icon window
       (gdk-window-set-icon win #f pixmap (vector-ref bitmap-loc 0)))
      ;; guile-gtk 0.41 releases pixmaps with gdk_window_unref, which
      ;; provokes gdk warnings about undestroyed window freed, avoid that
      (set! app-widget-icon-keep (list pixmap (vector-ref bitmap-loc 0))))))

(gtk-widget-add-events app-widget '(key-press-mask))
(gtk-signal-connect app-widget "key-press-event" keymap-dispatch-key)
(gtk-signal-connect app-widget "destroy" (lambda () (exit 0)))

(define-public app-accel-group (gtk-accel-group-new))
(gtk-window-add-accel-group app-widget app-accel-group)

(define-public app-tooltips (gtk-tooltips-new))

(define-public outer (view-new))
(gnome-app-set-contents app-widget outer)

(define appbar-widget (gnome-appbar-new #f #t 'never))
(gnome-app-set-statusbar app-widget appbar-widget)

(define-public (message fmt . args)
  (gnome-appbar-set-status appbar-widget (if fmt
					     (apply simple-format #f fmt args)
					     "")))
;; indented to hide from etags
 (set! view-message message)
 (set! weblink-message message)


;;-----------------------------------------------------------------------------
;; symlists

(define-public (chart-first)
  (chart-open (smarker-first (view-smarker outer))))

(define-public (chart-last)
  (chart-open (smarker-last (view-smarker outer))))

(define-public (chart-next?)
  (smarker-next? (view-smarker outer)))

(define-public (chart-next)
  (c-gtk-widget-busy app-widget)
  (let more ((symbol (smarker-next (view-smarker outer))))
    (if (and symbol
	     (database-read-series symbol))
	(chart-open symbol)
	(let ((next (smarker-next (view-smarker outer))))
	  (if (not (equal? next symbol))
	      (more next))))))

(define-public (chart-prev?)
  (smarker-prev? (view-smarker outer)))

(define-public (chart-prev)
  (c-gtk-widget-busy app-widget)
  (let more ((symbol (smarker-prev (view-smarker outer))))
    (if (and symbol
	     (database-read-series symbol))
	(chart-open symbol)
	(let ((prev (smarker-prev (view-smarker outer))))
	  (if (not (equal? prev symbol))
	      (more prev))))))

(define-public (chart-next-list)
  (chart-open (smarker-next-list (view-smarker outer))))

(define-public (chart-prev-list)
  (chart-open (smarker-prev-list (view-smarker outer))))


(define (favourites-add-symbol symbol)
  (let* ((lst (symlist-list favourites-symlist))
	 (pos (list-index-elem symbol lst)))
    (if pos
	(begin
	  (symlist-move favourites-symlist pos 1 0)
	  (message (_ "~a elevated in favourites list") symbol))
	(begin
	  (symlist-insert favourites-symlist (length lst) (list symbol))
	  (message (_ "~a added to favourites list") symbol)))))

(define (favourites-remove-symbol symbol)
  (let ((pos (list-index-elem symbol (symlist-list favourites-symlist))))
    (if pos
	(begin
	  (message (_ "~a removed from favourites list") symbol)
	  (symlist-delete favourites-symlist pos 1)
	  (if (eq? favourites-symlist (smarker-symlist (view-smarker outer)))
	      (chart-next)))
	(message (_ "~a not in favourites list") symbol))))

(define (favourites-add)
  (and-let* ((symbol (view-symbol outer)))
    (favourites-add-symbol symbol)))

(define (favourites-remove)
  (and-let* ((symbol (view-symbol outer)))
    (favourites-remove-symbol symbol)))


;;-----------------------------------------------------------------------------
;; time

(define (chart-timebase! new-timebase)
  (c-gtk-widget-busy app-widget)
  (view-timebase! outer new-timebase))

(define (chart-timebase-days)
  (chart-timebase! timebase-days))

(define (chart-timebase-weeks)
  (chart-timebase! timebase-weeks))

(define (chart-timebase-months)
  (chart-timebase! timebase-months))

(define (chart-logarithmic! state)
  (let ((graph (first (view-graph-list outer))))
    (graph-logarithmic! graph state)))

(define (chart-intraday)
  (intraday-popup (view-symbol outer)))


;;-----------------------------------------------------------------------------
;; data

;; If today or yesterday is visible, then request latest for SYMBOL.  Allow
;; yesterday since that's the latest before the start of trading.
;;
;; Look ahead and request the next too, for faster response to "next".  Up
;; to a further 8 of the symlist are included as "extras" in the request.
;;
(define (latest-for-graph symbol)
  (let ((adj (view-hadjustment outer)))
    (if (>= (gtk-adjustment-upper adj)
	    (1- (tdate-today (symbol-timezone symbol))))
	(let* ((symlist (smarker-symlist (view-smarker outer)))
	       (whole   (if symlist (symlist-list symlist) '()))
	       (upto    (or (member symbol whole) '()))
	       (must    (filter database-symbol?
				(take-maybe upto 2)))
	       (extra   (filter database-symbol?
				(take-maybe (drop-maybe upto 2) 8))))
	  (latest-request-symbols (cons symbol (drop-maybe must 1))
				  extra)))))

;; lists of elements each of which is a list (SYMBOL NAME SMARKER)
(define-public chart-open-history '())
(define-public chart-open-forward '())

(define chart-open-history-limit 30)

;; open SYMBOL, and which can be #f to clear the display
(define*-public (chart-open symbol
			    #:key (window app-widget) smarker)
  (c-gtk-widget-busy app-widget)
  (let ((series (and symbol (database-read-series symbol))))
    (if (and symbol
	     (not series))
	(error-dialog-popup window (simple-format #f (_ "No data for ~a")
						  symbol))
	(begin
	  (gtk-window-set-title app-widget (c-make-window-title symbol))
	  (if smarker
	      (view-smarker! outer smarker))
	  (view-open outer symbol)

	  (if (and symbol
		   (or (null? chart-open-history)
		       (not (equal? symbol
				    (first (first chart-open-history))))))
	      (set! chart-open-history
		    (take-maybe (cons (list symbol
					    (series-name series)
					    (smarker-copy (view-smarker
							   outer)))
				      chart-open-history)
				chart-open-history-limit)))
	  (set! chart-open-forward
		(take-maybe chart-open-forward chart-open-history-limit))
	  (notify-send-locally 'chart-open-history-changed #f)

	  (latest-for-graph symbol)))))

(define-public (chart-back?)
  (length>=? chart-open-history 2))

(define-public (chart-back)
  (if (chart-back?)
      (begin
	(set! chart-open-forward (cons (first chart-open-history)
				       chart-open-forward))
	(set! chart-open-history (cdr chart-open-history))
	(receive-list (symbol name smarker . rest)
	    (first chart-open-history)
	  (chart-open symbol #:smarker smarker)))))

(define-public (chart-forward?)
  (not (null? chart-open-forward)))

(define-public (chart-forward)
  (if (chart-forward?)
      (let ((symbol (first (first chart-open-forward))))
	(set! chart-open-forward (cdr chart-open-forward))
	(chart-open symbol))))

(define (download-update-current)
  (and-let* ((symbol (view-symbol outer)))
    (download-dialog-one symbol)))


;;-----------------------------------------------------------------------------
;; scrolling

;; move ADJ to its lower end
(define (c-gtk-adjustment-end-down adj)
  (gtk-adjustment-set-value adj (gtk-adjustment-lower adj)))

(define (c-gtk-adjustment-step-down adj)
  (c-gtk-adjustment-set-value-inpage
   adj
   (- (gtk-adjustment-value adj)
      (gtk-adjustment-step-increment adj))))
(define (c-gtk-adjustment-step-up adj)
  (c-gtk-adjustment-set-value-inpage
   adj
   (+ (gtk-adjustment-value adj)
      (gtk-adjustment-step-increment adj))))


(define-public (scroll-end-left)
  (c-gtk-adjustment-end-down (view-hadjustment outer)))

(define-public (scroll-end-right)
  (c-gtk-adjustment-end-up (view-hadjustment outer)))

(define-public (scroll-step-left)
  (c-gtk-adjustment-step-down (view-hadjustment outer)))

(define-public (scroll-step-right)
  (c-gtk-adjustment-step-up (view-hadjustment outer)))

(define-public (scroll-page-left)
  (c-gtk-adjustment-page-down (view-hadjustment outer)))

(define-public (scroll-page-right)
  (c-gtk-adjustment-page-up (view-hadjustment outer)))

(define-public (scroll-step-up)
  (c-gtk-adjustment-step-up (view-vadjustment outer)))

(define-public (scroll-step-down)
  (c-gtk-adjustment-step-down (view-vadjustment outer)))

(define-public (scroll-page-up)
  (c-gtk-adjustment-page-up (view-vadjustment outer)))

(define-public (scroll-page-down)
  (c-gtk-adjustment-page-down (view-vadjustment outer)))

(define-public (scroll-centre-vertical)
  (view-centre-vertical outer))

(define-public (chart-redraw)
  (define (redraw widget)
    (and-let* ((window (gtk-widget-window widget)))
      (gdk-window-clear-area-e window 0 0
			       (c-gdk-drawable-get-width window)
			       (c-gdk-drawable-get-height window)))
    (if (gtk-container? widget)
	(for-each redraw (gtk-container-children widget))))
  (redraw app-widget))


;;-----------------------------------------------------------------------------
;; menus helpers

(define (make-top-menu name)
  (let* ((menu (gtk-menu-new))
	 (item (gtk-menu-item-new-with-mnemonic name)))
    (gtk-menu-item-set-submenu item menu)
    (gtk-menu-shell-append menubar-widget item)
    menu))

(define* (make-sub-item menu name
			#:key creator accelerator callback tooltip)
  (let ((item (if creator
		  (creator name)
		  (gtk-image-menu-item-new-from-stock
		   name (if accelerator #f app-accel-group)))))
    ;; global accelerator
    (if accelerator
	(receive-list (keyval mods)
	    (gtk-accelerator-parse accelerator)
	  (gtk-widget-add-accelerator item "activate" app-accel-group
				      keyval mods '(visible))
	  ;; with <Shift>, this is essential for "!" which can usually
	  ;; only be typed with shift in the modifiers
	  (gtk-widget-add-accelerator item "activate" app-accel-group keyval
				      (cons 'shift-mask mods) '())))
    (if callback
	(gtk-signal-connect item "activate" callback))
    (if tooltip
	(gtk-tooltips-set-tip app-tooltips item tooltip ""))
    (gtk-menu-shell-append menu item)
    (gtk-widget-show item)
    item))

;; workaround for trouble creating radio menu items
(define (c-gtk-radio-menu-item-new-with-mnemonic str)
  (let ((item  (gtk-widget-new 'GtkRadioMenuItem))
	(label (gtk-label-new-with-mnemonic str)))
    (gtk-misc-set-alignment label 0 0.5)
    (gtk-container-add item label)
    item))


;;-----------------------------------------------------------------------------
;; weblinks menu (shared with watchlist.scm)

;; URL is a url string, or #f, or a procedure returning a url string or #f
(define (weblink-open url)
  (if (procedure? url)
      (set! url (url)))
  (if url
      (begin
	(message (_ "Opening web browser ..."))
	(gnome-url-show url))
      ;; don't force button to insensitive in this case, in case it's
      ;; only because temporarily offline
      (message (_ "Oh, in fact no web link for this symbol"))))

;; find weblink entries for SYMBOL, return a list where each list element is
;; (MENUNAME TOOLTIP URL/PROC)
(define (symbol-weblinks symbol)
  (symbol-setups symbol)
  (filter-map (lambda (elem)
		(and (sympred-test (car elem) symbol)
		     ((cdr elem) symbol)))
	      weblink-handler-list))

;; replace entries in the weblinks menu ITEM with those for SYMBOL
(define-public (weblinks-menuitem-symbol! item symbol)

  (let* ((menu (c-prop item))
	 (lst (if symbol
		  (sort (symbol-weblinks symbol) car-string<?)
		  '())))
    (c-gtk-container-remove-all menu)
    (gtk-widget-set-sensitive item (not (null? lst)))
    (for-each (lambda (elem)
		(receive-list (menuname tooltip url)
		    elem
		  (gtk-widget-set-sensitive
		   (make-sub-item menu menuname
				  #:tooltip tooltip
				  #:callback (lambda ()
					       (weblink-open url)))
		   ;; URL=#f for menu entry shown, but insensitive
		   url)))
	      lst)))

;; create a new menu item which opens a weblink sub-menu
(define-public (weblinks-menuitem-new)
  (let ((item (gtk-menu-item-new-with-mnemonic (_ "We_b")))
	(menu (gtk-menu-new)))
    (gtk-menu-item-set-submenu item menu)
    ;; gtk-menu-item-get-submenu not wrapped in guile-gtk 2.0, so record in
    ;; c-prop instead
    (set! (c-prop item) menu)
    item))


;;-----------------------------------------------------------------------------
;; menus

(define-public menubar-widget (gtk-menu-bar-new))

(define (car-string<? x y)
  (string<? (car x) (car y)))

(let-toplevel ()

  ;; return the symbol being viewed, or #f
  (define (viewing-symbol)
    (view-symbol outer))

  ;; ITEM is a GtkMenuItem, (PROC) is called each time the displayed series
  ;; changes and should return true if ITEM should be sensitive
  ;; PROC is called immediately to get an initial setting too
  (define (sensitive-series proc item)
    (gtk-widget-set-sensitive item (proc))
    (sensitive-series-later proc item))
  (define (sensitive-series-later proc item)
    (gtk-signal-connect outer "series-changed"
      (lambda ()
	(gtk-widget-set-sensitive item (proc)))))

  ;; ITEM is a GtkMenuItem, (PROC) is called each time the user preferences
  ;; change and should return true if ITEM should be sensitive
  (define (sensitive-preferences proc item)
    (gtk-widget-set-sensitive item (proc))
    (notify-connect 'preferences-save
      (lambda (dummy)
	(gtk-widget-set-sensitive item (proc)))))

  (define-public file-menu-widget (make-top-menu (_ "_File")))
  (make-sub-item file-menu-widget "gtk-open"
		 #:callback    (lambda () (open-dialog-popup))
		 #:accelerator (_ "CONTEXT(accelerator-open): O")
		 #:tooltip  (_ "View chart for a symbol, or add new symbol to the database"))
  (sensitive-series viewing-symbol
    (make-sub-item file-menu-widget "gtk-delete"
		   #:callback    (lambda () (delete-dialog-popup))))
  (sensitive-series-later chart-next?
    (make-sub-item file-menu-widget (_ "_Next")
		   #:callback    chart-next
		   #:accelerator (_ "CONTEXT(accelerator-next): N")))
  (sensitive-series-later chart-prev?
    (make-sub-item file-menu-widget (_ "_Prev")
		   #:callback    chart-prev
		   #:accelerator (_ "CONTEXT(accelerator-prev): P")))

  (sensitive-series chart-back?
    (make-sub-item file-menu-widget "gtk-go-back"
		   #:callback    chart-back
		   #:accelerator (_ "CONTEXT(accelerator-back): B")))
  (sensitive-series chart-forward?
    (make-sub-item file-menu-widget "gtk-go-forward"
		   #:callback    chart-forward
		   #:accelerator (_ "CONTEXT(accelerator-forward): F")))

  ;; showing "add" always doesn't make the elevate function clear, but want
  ;; to keep the Alt-A mnemonic the same
  (sensitive-series viewing-symbol
    (make-sub-item file-menu-widget (_ "_Add as Favourite")
		   #:callback    favourites-add
		   #:accelerator "exclam"
		   #:tooltip  (_ "Add this symbol to the favourites list, or elevate if already there")))

  (let ((item (make-sub-item file-menu-widget (_ "_Remove Favourite")
			     #:callback    favourites-remove
			     #:accelerator "<Ctrl>exclam"
			     #:tooltip  (_ "Remove this symbol from the favourites list"))))
    ;; return true if there's a current symbol being viewed and it's also in
    ;; the favourites list
    (define (viewing-favourites-symbol?)
      (and-let* ((symbol (viewing-symbol)))
	(member symbol (symlist-list favourites-symlist))))
    ;; sensitive check on different viewed series or on favourites list
    ;; change
    (sensitive-series viewing-favourites-symbol? item)
    (notify-connect 'symlist-update
      (lambda (symlist-key-list)
	(if (memq 'favourites symlist-key-list)
	    (gtk-widget-set-sensitive item (viewing-favourites-symbol?))))))

  (make-sub-item file-menu-widget "gtk-quit"
		 #:callback    (lambda () (exit 0)))

  (define-public edit-menu-widget (make-top-menu (_ "_Edit")))

  (sensitive-series viewing-symbol
    (make-sub-item edit-menu-widget (_ "_Annotations")
		   #:callback (lambda () (annotations-edit-popup))
		   #:tooltip  (_ "Edit user text annotations on this symbol")))

  (sensitive-series viewing-symbol
    (make-sub-item edit-menu-widget (_ "_Dividends")
		   #:callback (lambda () (dividends-edit-popup))
		   #:tooltip (_ "Edit downloaded dividend and split information for this symbol")))
  (make-sub-item edit-menu-widget (_ "_View style")
		 #:callback (lambda () (viewstyle-dialog-popup))
		 #:tooltip (_ "Edit the viewing options, such as averages and indicators shown"))
  (make-sub-item edit-menu-widget "gtk-preferences"
		 #:callback (lambda () (pref-dialog-popup)))

  (define-public view-menu-widget (make-top-menu (_ "_View")))
  (let* ((daily   (make-sub-item
		   view-menu-widget (_ "_Daily")
		   #:callback    chart-timebase-days
		   #:creator     c-gtk-radio-menu-item-new-with-mnemonic
		   #:accelerator (_ "CONTEXT(accelerator-daily): <Ctrl>D")))
	 (weekly  (make-sub-item
		   view-menu-widget (_ "_Weekly")
		   #:callback    chart-timebase-weeks
		   #:creator     c-gtk-radio-menu-item-new-with-mnemonic
		   #:accelerator (_ "CONTEXT(accelerator-weekly): <Ctrl>W")))
	 (monthly (make-sub-item
		   view-menu-widget (_ "_Monthly")
		   #:callback    chart-timebase-months
		   #:creator     c-gtk-radio-menu-item-new-with-mnemonic
		   #:accelerator (_ "CONTEXT(accelerator-monthly): <Ctrl>M"))))
    (gtk-widget-set weekly #:group daily)
    (gtk-widget-set monthly #:group daily)
    (gtk-check-menu-item-set-active daily #t))

  (gtk-menu-shell-append view-menu-widget (gtk-separator-menu-item-new))
  (sensitive-series viewing-symbol
    (make-sub-item view-menu-widget (_ "_Centre vertically")
		   #:callback    scroll-centre-vertical
		   #:accelerator (_ "CONTEXT(accelerator-centre): <Ctrl>C")))
  (make-sub-item view-menu-widget "gtk-zoom-in"
		 #:callback    zoom-in
		 #:accelerator "<Ctrl>plus")
  (make-sub-item view-menu-widget "gtk-zoom-out"
		 #:callback    zoom-out
		 #:accelerator "<Ctrl>minus")
  (gtk-menu-shell-append view-menu-widget (gtk-separator-menu-item-new))
  (letrec ((item (make-sub-item
		  view-menu-widget (_ "_Logarithmic")
		  #:callback    (lambda ()
				  (chart-logarithmic!
				   (gtk-check-menu-item-active item)))
		  #:creator     gtk-check-menu-item-new-with-mnemonic
		  #:accelerator (_ "CONTEXT(accelerator-logarithmic): <Ctrl>L"))))
    #f)
  (make-sub-item view-menu-widget (_ "_Intraday")
		 #:callback    chart-intraday
		 #:accelerator "<Ctrl>I")

  (let ((item (weblinks-menuitem-new)))
    (gtk-menu-append view-menu-widget item)
    (gtk-signal-connect outer "series-changed"
      (lambda ()
	(weblinks-menuitem-symbol! item (view-symbol outer)))))

  (define-public tools-menu-widget (make-top-menu (_ "_Tools")))
  (let ((item (make-sub-item
	       tools-menu-widget (_ "_Cross")
	       #:creator gtk-check-menu-item-new-with-mnemonic
	       #:accelerator (_ "CONTEXT(accelerator-cross): C"))))
    ;; the "toggled" callback occurs for the direct call to
    ;; `gtk-check-menu-item-set-state' too, so must look at
    ;; `gtk-check-menu-item-active' for the state, not just unconditionally
    ;; do `cross-toggle
    (gtk-signal-connect item "toggled"
      (lambda ()
	(cross-set-state! (gtk-check-menu-item-active item))))
    (notify-connect 'cross-update
      (lambda (state)
	(gtk-check-menu-item-set-state item state))))
  (make-sub-item tools-menu-widget (_ "_Download")
		 #:callback (lambda () (download-dialog-popup)))
  (make-sub-item tools-menu-widget (_ "_Watchlist")
		 #:callback (lambda () (watchlist-dialog-popup))
		 #:tooltip (_ "Open list of current quotes for favourites, or for all symbols)"))
  (let ((item (make-sub-item
	       tools-menu-widget (_ "_Ticker")
	       #:creator gtk-check-menu-item-new-with-mnemonic
	       #:tooltip (_ "Show a scrolling stock ticker at the bottom of the window)"))))
    (gtk-signal-connect item "toggled"
      (lambda ()
	(view-ticker-state! outer (gtk-check-menu-item-active item))))
    (notify-connect 'view-ticker-state-update
      (lambda (view)
	(gtk-check-menu-item-set-state item (view-ticker-state view)))))

  (define-public help-menu-widget (make-top-menu (_ "_Help")))
  (make-sub-item help-menu-widget "gtk-about"
		 #:callback (lambda () (about-dialog-popup)))
  (let ((item (make-sub-item help-menu-widget (_ "_Manual")
			     #:callback    (lambda ()
					     (manual-popup #f app-widget))
			     #:accelerator "F1")))
    (gtk-image-menu-item-set-image
     item (gtk-image-new-from-stock "gtk-help" 'menu)))

  (make-sub-item help-menu-widget (_ "_Keys")
		 #:callback (lambda () (keys-popup)))

  (sensitive-series manual-data-source-node
    (make-sub-item help-menu-widget (_ "This _Data Source")
		   #:callback (lambda ()
				(and-let* ((node (manual-data-source-node)))
				  (manual-popup node app-widget)))
		   #:tooltip (_ "Open the manual at the section about the current symbol data source")))

  (sensitive-preferences manual-average-node
    (make-sub-item help-menu-widget (_ "This _Moving Average")
		   #:callback (lambda ()
				(and-let* ((node (manual-average-node)))
				  (manual-popup node app-widget)))
		   #:tooltip (_ "Open the manual at the section about the moving average shown in the upper graph")))

  (sensitive-preferences manual-indicator-node
    (make-sub-item help-menu-widget (_ "This _Indicator")
		   #:callback (lambda ()
				(and-let* ((node (manual-indicator-node)))
				  (manual-popup node app-widget)))
		   #:tooltip (_ "Open the manual at the section about the indicator shown in the lower graph"))))

(gnome-app-set-menus app-widget menubar-widget)


;;-----------------------------------------------------------------------------
;; keys

(define (procedure-name! proc name)
  (set-procedure-property! proc 'name name)
  proc)

(define-public app-keymap (keymap-new (_ "Key map") (lambda () #t)))

(keymap-set! app-keymap "<Control>L"      chart-redraw)
(keymap-set! app-keymap "<Ctrl>U"         download-update-current)
(keymap-set! app-keymap "W"               zoom-in-x)
(keymap-set! app-keymap "<Shift>W"        zoom-out-x)
(keymap-set! app-keymap "Z"               zoom-in-y)
(keymap-set! app-keymap "<Shift>Z"        zoom-out-y)
(keymap-set! app-keymap "<Control>Z"      zoom-in-y-lower)
(keymap-set! app-keymap "<Control><Shift>Z" zoom-out-y-lower)

(keymap-set! app-keymap "Left"            scroll-step-left)
(keymap-set! app-keymap "<Control>Left"   scroll-page-left)
(keymap-set! app-keymap "Right"           scroll-step-right)
(keymap-set! app-keymap "<Control>Right"  scroll-page-right)
(keymap-set! app-keymap "Home"            scroll-end-left)
(keymap-set! app-keymap "End"             scroll-end-right)
(keymap-set! app-keymap "Up"              scroll-step-up)
(keymap-set! app-keymap "<Control>Up"     scroll-page-up)
(keymap-set! app-keymap "Down"            scroll-step-down)
(keymap-set! app-keymap "<Control>Down"   scroll-page-down)
(keymap-set! app-keymap "Page_Up"         scroll-page-up)
(keymap-set! app-keymap "Page_Down"       scroll-page-down)
(keymap-set! app-keymap "<Control>N"      chart-next-list)
(keymap-set! app-keymap "<Control>P"      chart-prev-list)

(keymap-set! app-keymap "Pointer_Button1"
	     (procedure-name! (lambda () (scroll-drag-start-button))
			      'scroll-drag-start-button))
(keymap-set! app-keymap "<Shift>Pointer_Button1"
	     (procedure-name! (lambda () (lasso-start-button))
			      'lasso-start-button))
(keymap-set! app-keymap "Pointer_Button2"
	     (procedure-name! (lambda () (line-edit-or-new-button))
			      'line-edit-or-new-button))
(keymap-set! app-keymap "Pointer_Button3"
	     (procedure-name! (lambda () (cross-drag-button))
			      'cross-drag-button))
(keymap-set! app-keymap "<Shift>Pointer_Button3"
	     (procedure-name! (lambda () (indicator-menu-popup))
			      'indicator-menu-popup))

(keymap-set! app-keymap "<Alt><Shift>D"
	     (procedure-name! (lambda () (diagnostics-dialog-popup))
			      'diagnostics-dialog-popup))


;;-----------------------------------------------------------------------------
;; purging old stuff

(gtk-idle-add-priority (+ 100 gtk-priority-low) ;; low priority
  (lambda ()
    ;; return true if PATH is an ordinary file last modified more than 14
    ;; days ago
    (define old-file?
      (let ((oldtime (- (current-time) (* 14 86400))))
	(lambda (path)
	  ;; allow for dangling symlink from emacs lock file
	  (let ((s (false-if-exception (stat path))))
	    (and s
		 (eq? 'regular (stat:type s))
		 (< (stat:mtime s) oldtime))))))

    ;; return true if NAME (without a directory part) is a chart temp file
    (define (temp-filename? name)
      (and (string-prefix? ".temp-" name)
	   (= 8 (string-length name))))

    ;; return true if PATH is a chart temp file
    (define (temp-pathname? path)
      (temp-filename? (basename path)))

    ;; return true for a chart temp file older than 14 days
    (define (old-temp? path)
      (and (temp-pathname? path)
	   (old-file? path)))

    ;; return true for any file older than 14 days, except .lockfile
    (define (old-any? path)
      (and (not (string=? ".lockfile" (basename path)))
	   (old-file? path)))

    ;; return true if PATH is a subdirectory, ie. directory and not . or ..
    (define (file-subdirectory? path)
      (and-let* ((  (not (member (basename path) '("." ".."))))
		 ;; allow for dangling symlink from emacs lock file
		 (s (false-if-exception (stat path))))
	(eq? 'directory (stat:type s))))

    ;; remove files satisfying (PRED path) from within DIR
    (define (purge-directory dir pred)
      (for-each delete-file-maybe (directory-files dir #t pred)))

    ;; remove files satisfying (PRED path) from the immediate
    ;; subdirectories of TOP-DIR
    (define (purge-subdirectories top-dir pred)
      (for-each (lambda (dir)
		  (purge-directory dir pred))
		(directory-files top-dir #t file-subdirectory?)))

    ;; old temporaries under database and user notes, plus in user notes dir
    ;; itself (possible leftovers from prefs.scm writes)
    (purge-subdirectories database-directory old-temp?)
    (if (not (string=? user-directory database-directory))
	(purge-subdirectories user-directory old-temp?))
    (purge-directory user-directory old-temp?)

    ;; anything old in and under cache
    (purge-directory      cache-directory old-any?)
    (purge-subdirectories cache-directory old-any?)

    #f)) ;; once only


;;-----------------------------------------------------------------------------

(notify-listen)
