;; Create datatype-setups.scm, symbol-setups.scm and yahoo-setups.scm.

;; Copyright 2007 Kevin Ryde
;;
;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: guile -s make-setups.scm filename.scm...
;;
;;

(use-modules (ice-9 format)
	     (ice-9 pretty-print)
	     (ice-9 regex)
	     (ice-9 rw)
	     (srfi srfi-1)
	     (srfi srfi-10)
	     (srfi srfi-13)
	     (srfi srfi-14))

;; dummies to allow code to be read
(define-reader-ctor 'regexp       (lambda args #f))
(define-reader-ctor 'hms->seconds (lambda args #f))


;;-----------------------------------------------------------------------------
;; misc

(define (list>=? lst n)
  (or (<= n 0)
      (and (pair? lst)
	   (list>=? (cdr lst) (1- n)))))

(define (car-string<? x y)
  (string<? (car x) (car y)))

(define (first-or-f obj)
  (and (pair? obj)
       (first obj)))

(define (plist-get plist key)
  (and (not (null? plist))
       (if (eq? key (first plist))
	   (second plist)
	   (plist-get (cddr plist) key))))

;; Like for-each, but take two elements at a time from LST, and call PROC
;; with those two as parameters.
;;
;; If LST has an odd number of elements, the last call to PROC will be with
;; one argument.  It's left up to PROC whether this is acceptable or an
;; error.
;;
(define-public (for-each-two proc lst)
  (if (not (null? lst))
      (if (null? (cdr lst))
	  (proc (car lst))
	  (begin
	    (proc (car lst) (cadr lst))
	    (for-each-two proc (cddr lst))))))

(define (pretty-print-no-newline obj port)
  (display (string-trim-right (pretty-print->string obj)) port))

;; like slib, but no width argument
(define (pretty-print->string obj)
  (call-with-output-string
   (lambda (port)
     (pretty-print obj port))))

(define (slurp filename)
  (call-with-input-file filename read-string-to-eof))

(define-public (read-string-to-eof port)
  (let more ((lst '()))
    (let* ((buf   (make-string 4096))
	   (nread (read-string!/partial buf port)))
      (cond ((not nread)
	     ;; eof, join parts
	     (string-concatenate-reverse/shared lst))
	    (else
	     ;; got some, allow for less than requested, no copy if full BUF
	     (more (cons (substring/shared buf 0 nread) lst)))))))


;;-----------------------------------------------------------------------------
;; read inputs

(define dir ".")
(define setup-prefix-alist '())
(define setup-suffix-alist '())
(define datatype-list '())
(define setup-prefix-copyrights (list (port-filename (current-load-port))))
(define setup-suffix-copyrights (list (port-filename (current-load-port))))
(define datatype-copyrights     (list (port-filename (current-load-port))))

(define (form-define-module? form)
  (and (list>=? form 3)
       (eq? (first form) 'define-module)))

(define (form-define? form)
  (and (list>=? form 3)
       (memq (first form) '(define define-public))))

(define (form-timezone!? form)
  (and (list>=? form 3)
       (eq? (first form) 'symbol-timezone!)))

(define (process-file filename)
  (define modulename #f)
  (define def-list '())

  (define (process-datatype-new! form)
    (cond ((and (list>=? form 1)
		(eq? 'datatype-new! (first form))
		(not (memq 'priority form))) ;; not func in misc.scm
	   (or modulename
	       (error "no module name in:" filename))
	   (let ((plist (cdr form)))
	     (set! datatype-copyrights (cons filename datatype-copyrights))
	     (set! datatype-list
		   (cons
		    (append (list 'datatype-new!
				  #:key      (plist-get plist #:key)
				  #:name     (plist-get plist #:name)
				  #:units    (plist-get plist #:units))
			    (if (plist-get plist #:priority)
				(list #:priority (plist-get plist #:priority))
				'())
			    (list #:module   (list 'quote modulename)))
		    datatype-list))))
	  ((list? form)
	   (for-each process-datatype-new! form))))

  (define (process-def form)
    (let ((match #f))
      (let recurse ((form form))
	(if (list>=? form 1)
	    (begin
	      ;; skip a few oddities
	      (if (memq (first form) '(member                    ;; cbot.scm
				       sfe-newzealand-commodity? ;; sfe.scm
				       yahoo-currency-symbol?    ;; usa.scm
				       ))
		  (set! match #t))

	      (if (memq (first form) '(string-prefix? string=?))
		  (begin
		    (or modulename
			(error "no module name in:" filename))
		    (set! match #t)
		    (set! setup-prefix-copyrights
			  (cons filename setup-prefix-copyrights))
		    (set! setup-prefix-alist
			  (acons (second form)
				 modulename
				 setup-prefix-alist))))

	      (if (eq? 'string-suffix? (first form))
		  (begin
		    (or modulename
			(error "no module name in:" filename))
		    (set! match #t)
		    (set! setup-suffix-copyrights
			  (cons filename setup-suffix-copyrights))
		    (set! setup-suffix-alist
			  (acons (second form)
				 modulename
				 setup-suffix-alist))))))
	(if (list? form)
	    (for-each recurse form)))
      (or match
	  (error "nothing found in define:" form))))

  (define (process-setup pred)
    (let ((def (or (find (lambda (def)
			   (eq? pred (first-or-f (second def))))
			 def-list)
		   (error "no definition for:" pred))))
      (process-def def)))

  (call-with-input-file (string-append dir "/" filename)
    (lambda (port)
      (do ((form (read port) (read port)))
	  ((eof-object? form))

	(cond ((form-define? form)
	       (set! def-list (cons form def-list)))
	      ((form-define-module? form)
	       (set! modulename (second form)))
	      ((form-timezone!? form)
	       (process-setup (second form))))

	(process-datatype-new! form)))))

(let more ((lst (cdr (program-arguments))))
  (cond ((null? lst)
	 )
	((string=? (first lst) "--dir")
	 (set! dir (second lst))
	 (more (cddr lst)))
	(else
	 (process-file (first lst))
	 (more (cdr lst)))))


;;-----------------------------------------------------------------------------
;; check prefix clashes

(let ((good #t))
  (for-each (lambda (e1)
	      (for-each (lambda (e2)
			  (let ((p1 (car e1))
				(p2 (car e2)))
			    (if (and (not (equal? e1 e2))
				     (string-prefix? p1 p2))
				(begin
				  (display "prefix clash:\n")
				  (format #t "   ~s ~s\n" p1 (cdr e1))
				  (format #t "   ~s ~s\n" p2 (cdr e2))
				  (set! good #f)))))
			setup-prefix-alist))
	    setup-prefix-alist)
  (or good
      (exit 1)))


;;-----------------------------------------------------------------------------
;; copyrights

(define copyright-regexp (make-regexp "Copyright ([0-9, ]+)"))

(define (filename->copyright-years filename)
  (let* ((body (slurp filename))
	 (m    (or (regexp-exec copyright-regexp body)
		   (error "No copyright line in file:" filename))))
    (string-tokenize (match:substring m 1) char-set:digit)))

(define (filenames->copyright-line filename-list)
  (let ((years-list (append-map filename->copyright-years filename-list)))
    (string-append
     ";; Copyright "
     (string-join (sort (delete-duplicates years-list string=?)
			string<?)
		  ", ")
     " Kevin Ryde\n")))

(define (display-copyright filename-list port)
  (display (filenames->copyright-line filename-list) port)
  (display 
";;
;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
\n\n" port))


;;-----------------------------------------------------------------------------
;; write symbol-setups.scm

(set! setup-suffix-alist
      (sort! setup-suffix-alist car-string<?))

(call-with-output-file "symbol-setups.scm"
  (lambda (port)
    (display ";; Symbol setups - generated by make-setups.scm - DO NOT EDIT\n\n" port)
    (display-copyright setup-suffix-copyrights port)

    (let ((width (+ 2 (apply max 0 (map! string-length
					 (map car setup-suffix-alist))))))
      (for-each (lambda (elem)
		  (format port "(symbol-setups! ~vs '~s)\n"
			  width (car elem) (cdr elem)))
		setup-suffix-alist))))

;;-----------------------------------------------------------------------------
;; write yahoo-setups.scm

(set! setup-prefix-alist
      (sort! setup-prefix-alist car-string<?))

(call-with-output-file "yahoo-setups.scm"
  (lambda (port)
    (display ";; Yahoo symbol setups - generated by make-setups.scm - DO NOT EDIT\n\n" port)
    (display-copyright setup-prefix-copyrights port)

    (let ((width (+ 2 (apply max 0 (map! string-length
					 (map car setup-prefix-alist))))))
      (for-each (lambda (elem)
		  (format port "(symbol-setups! ~vs '~s)\n"
			  width (car elem) (cdr elem)))
		setup-prefix-alist))))

;;-----------------------------------------------------------------------------
;; write datatype-setups.scm

(set! datatype-list
      (sort! datatype-list
	     (lambda (x y)
	       (string<? (symbol->string (second (plist-get (cdr x) #:key)))
			 (symbol->string (second (plist-get (cdr y) #:key)))))))

(call-with-output-file "datatype-setups.scm"
  (lambda (port)
    (display ";; Datatype setups - generated by make-setups.scm - DO NOT EDIT\n\n" port)
    (display-copyright datatype-copyrights port)

    (for-each (lambda (form)
		(define plist (cdr form))
		(format port "(or (datatype-find '~a)\n"
			(second (plist-get plist #:key)))
		(let ((indent "    (datatype-new! "))
		  (for-each-two (lambda (key value)
				  (format port "~a~11s " indent key)
				  (pretty-print-no-newline value port)
				  (set! indent "\n                   "))
				plist))
		(display "))\n\n" port))
	      datatype-list)))

(exit 0)
