;; Media Mammon setups and data downloading.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mammon)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart pagebits))


;; The site/game seems to be slightly dead (as of January 2008).  The list
;; of all words used to be at
;;     http://mediamammon.drunkmenworkhere.org/share.php?sub=all
;; but was empty for a while and now asks for username/password (for the
;; toplevel http://mediamammon.drunkmenworkhere.org too).  For now leave the
;; code here, but remove it from the manual (see the end of this file for
;; that text).


(define-public (mammon-symbol? symbol)
  (string-suffix? ".MAM" symbol))

(symbol-timezone! mammon-symbol? (_ "Mammon") timezone-local)
;; (source-help! mammon-symbol? (_ "CONTEXT(manual-node): Media Mammon"))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses quotes pages like
;;
;;     http://mediamammon.drunkmenworkhere.org/share.php?id=2508
;;
;; The IDs are on the "all" page of links to those quotes pages,
;;
;;     http://mediamammon.drunkmenworkhere.org/share.php?sub=all
;;
;; An unknown id, from a word we have a number cached for but has now gone,
;; results in a 302 redirect to the full list at "/share.php".
;;


;; return alist ((SYMBOL . CODE-STRING) ...)
(define (mammon-id-parse body)
  (map! (lambda (m)
	  (cons (string-append (match:substring m 2) ".MAM")
		(match:substring m 1)))
	;; As of March/Apri 2007 no entries at all in the all page.  Dunno
	;; if it's going away, but for now don't require a match.
	(noop ;; must-match
	 (list-matches
	  "<a href=\"/share.php\\?id=([^\">]+)\">([^<]+)</a>" body))))

(define (symbol->id symbol)
  (assoc-ref (pagebits-read #:filename "mammon-ids"
			    #:url      "http://mediamammon.drunkmenworkhere.org/share.php?sub=all"
			    #:timezone (timezone-local)
			    #:parse    mammon-id-parse)
	     symbol))

(define (body->latest body)

  (let* ((m        (must-match (regexp-exec #,(regexp "<div class=\"t\">([^<]+)<div") body)))
	 (symbol   (string-append (match:substring m 1) ".MAM"))

	 (str      (string-collapse-whitespace (html-text-extract body)))
	 (m        (must-match (regexp-exec #,(regexp "price: \\$([0-9.]+)change: \\+?([0-9.-]+)% in last 24 hoursvolume: ([0-9]+) in the last 24 hours")
					    str)))
	 (last     (match:substring m 1))
	 (decimals (string-decimals last))
	 (change   (match:substring m 2))
	 (volume   (match:substring m 3))
	 (tm       (localtime (current-time))))

    (set! last (string->number+decimals last decimals))

    ;; change "+0.39%" or "-1.27%" as an absolute amount
    ;; prev*(1+change)==last, so last-prev==l*(change/(1+change))
    (if change
	(begin
	  (set! change (* 0.01 (string->number change)))
	  (set! change (round-exact (* last (/ change
					       (1+ change)))))))

    (latest-new #:symbol      symbol
		#:quote-adate (tm->adate tm)
		#:quote-time  (tm->time tm)
		#:last-adate  (tm->adate tm) ;; probably not quite right
		#:last        last
		#:change      change
		#:volume      volume
		#:decimals    decimals
		#:source      'mammon)))

(define (mammon-latest-get symbol-list extra-list proc)
  (for-each
   (lambda (symbol)
     (let ((id (symbol->id symbol)))
       (if (not id)
	   (proc (list (latest-new-unknown symbol)))

	   (receive (headers body)
	       (http-request
		(string-append
		 "http://mediamammon.drunkmenworkhere.org/share.php?id=" id)
		#:follow #f)
	     (let ((status-code (http-headers-status-code headers)))
	       (cond ((http-status-code-ok? status-code)
		      (let ((latest (body->latest body)))
			(proc (list latest))
			(or (equal? symbol (latest-symbol latest))
			    (error "Mammon: oops, id didn't give expected symbol"))))
		     ((= 302 status-code)
		      ;; redirect to /share.php for an id now unknown
		      (proc (list (latest-new-unknown symbol))))

		     (else
		      (http-headers-throw headers))))))))
   symbol-list))

(latest-handler! #:selector mammon-symbol?  
		 #:handler  mammon-latest-get)


;;-----------------------------------------------------------------------------
;; intraday
;;
;; This uses intraday graphs included on the quote page above,
;;
;;     http://mediamammon.drunkmenworkhere.org/charts/quote/22454.png
;;     http://mediamammon.drunkmenworkhere.org/charts/volume/22454.png
;;

(define mode-volume (_ "3 Day _Volume"))

(define (mammon-intraday-url symbol mode)
  (and-let* ((id (symbol->id symbol)))
    (string-append "http://mediamammon.drunkmenworkhere.org/charts/"
		   (cond ((string=? mode mode-volume)
			  "volume/")
			 (else ;; 3 Day Price
			  "quote/"))
		   id ".png")))

(intraday-url-proc! #:symbol-pred mammon-symbol?
		    #:url-proc    mammon-intraday-url
		    #:modes-proc  (lambda ()
				    (list (_ "3 Day _Price")
					  mode-volume)))




;; @node Media Mammon
;; @section Media Mammon
;; @cindex Media Mammon
;; @cindex Mammon

;; @uref{http://mediamammon.drunkmenworkhere.org}

;; Media Mammon provides

;; @itemize
;; @item
;; Quotes.
;; @item
;; Intraday graphs of the past 3 days, updated hourly.
;; @end itemize

;; @cindex @code{.MAM}
;; In Chart, Mammon symbols are the word with a @samp{.MAM} suffix,
;; eg.@: @samp{BILLION.MAM}.  The full list of words can be found at
;; @c SYMBOL: BILLION.MAM

;; @quotation
;; @uref{http://mediamammon.drunkmenworkhere.org/share.php?sub=all}
;; @end quotation

;; The intraday graphs for price and volume are separate, volume can be
;; selected in the option menu.

;; There's no obvious timezone for Mammon, so the system local timezone is
;; used in the watchlist.


