;; MASS index.

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mass)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema-2)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))

;; "The Mass Index", Donald Dorsey, TASC June 1992
;;
;; http://www.incrediblecharts.com/technical/mass_index.htm
;; http://www.incrediblecharts.com/technical/mi_construction.htm
;;
;; http://www.wallstreettape.com/tutorials/ta/c21.asp
;;     Chartfilter description.
;;


;; Return a procedure which calculates MASS index values over a window of
;; EMA-COUNT and SUM-COUNT many points.
;;
;; Each call (PROC range) enters a new point into the window, and the return
;; is the MASS value up to (and including) that point, or #f if not enough
;; values yet.
;;
;; To prime the window initially, call PROC with `mass-warmup-count' many
;; values preceding the first desired point.
;;
(define-public (mass-calc-proc ema-count sum-count)

  (let ((ema-proc  (ema-calc-proc ema-count))
	(dema-proc (ema-2-calc-proc ema-count))
	(sum-proc  (sum-calc-proc sum-count '())))

    (lambda (range)
      (let ((e  (ema-proc  range))
	    (de (dema-proc range)))
	(and e de
	     (sum-proc (if (zero? de)
			   0
			   (/ e de))))))))

(define-public (mass-warmup-count ema-count sum-count)
  (+ (max (ema-warmup-count ema-count)
	  (ema-2-warmup-count ema-count))
     (1- sum-count)))



(define (mass-fill series new-array lo hi)
  (receive-list (ema-count sum-count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos parent
					     #:before lo #:count warmup))
	   (array  (series-array parent start hi))
	   (mass   (mass-calc-proc ema-count sum-count)))

      ;; warmup
      (do ((i start (1+ i)))
	  ((>= i lo))
	(and-let* ((close (array-ref array i 3)))
	  (mass (- (or (array-ref array i 1) close)      ;; high
		   (or (array-ref array i 2) close)))))  ;; low

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((close (array-ref array i 3)))
	  (array-set! new-array
		      (mass (- (or (array-ref array i 1) close)   ;; high
			       (or (array-ref array i 2) close))) ;; low
		      i))))))

(define-public (mass-series series ema-count sum-count)
  (ohlcv-required "MASS Index" series)
  (let ((new-series (series-copy series))
	(warmup     (mass-warmup-count ema-count sum-count)))
    (series-name!      new-series (format #f (_ "MASS ~a,~a")
					  ema-count sum-count))
    (series-datatype!  new-series mass-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list ema-count sum-count warmup))
    (series-require!   new-series (indicator-require-parts mass-fill warmup))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public mass-datatype
  (datatype-new!
   #:key       'mass
   #:name      (_ "MASS Index")
   #:input     datatype-ohlcv
   #:units     'mass
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): MASS Index")
   #:proc      mass-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "EMA days")
	  #:key     'mass-ema-days
	  #:type    'integer
	  #:minimum 1
	  #:default 9)
	 (datatype-param-new
	  #:name    (_ "Sum days")
	  #:key     'mass-sum-days
	  #:type    'integer
	  #:minimum 1
	  #:default 25))))
