;; Median-Average Adaptive Filter.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart median-average)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (srfi srfi-1)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart tma))


;; http://www.mesasoftware.com/technicalpapers.htm
;; http://www.mesasoftware.com/Papers/What's%20the%20Difference.exe
;;     John Ehler's paper.  Sample chart of something unspecified.
;;
;; http://share.esignal.com/download.jsp?groupid=84&amp;folder=Stocks_Commodities&amp;file=InstantaneousTrendLine.efs
;;     Sample code.
;;


(define threshold .002)

(define (median-average-calc-proc)

  ;; Q is a queue of the last 39 smoothed input values
  ;; POINTS is how many points in Q
  ;; SORTED is the points from Q sorted in numerical order
  ;; PREV is the previous median-average value calculated
  ;;
  ;; PREV is initialized to the first SMOOTHed input.  Ehler's code and some
  ;; of the Trader's tips show an initial zero, which makes the filter rise
  ;; up from zero at the start of the data.  Since each alpha determined
  ;; depends on the previous value there's no way to chop off an infinite
  ;; sequence like an ordinary EMA.  Using the first smoothed should be
  ;; reasonable, it won't take too long for the medians to start moving and
  ;; the average tracking towards that.
  ;;
  (let ((smooth-proc (tma-calc-proc 4))
	(q           (make-q))
	(sorted      '())
	(points      0)
	(prev        #f))

    (lambda (x)
      (and-let* ((smooth (smooth-proc x)))
	(or prev
	    (set! prev smooth))

	;; add new entry to window
	(enq! q smooth)
	(set! sorted (merge! sorted (list smooth) <))

	(if (< points 39)
	    (begin
	      (set! points (1+ points))
	      #f)
	    (begin
	      ;; remove oldest from window
	      (set! sorted (delv1! (deq! q) sorted))

	      (let more ((len    39)
			 (lst    (q-list q))
			 (sorted (list-copy sorted)))

		;; (or (lset= eqv? lst sorted)
		;;    (begin
		;;       (d len lst sorted)
		;;       (exit 1)))

		(let* ((alpha (ema-count->alpha len))
		       (med   (list-ref sorted (quotient len 2)))
		       (avg   (+ (* alpha smooth)
				 (* (- 1 alpha) prev)))
		       (rat   (/ (abs (- med avg)) med)))

		  (if (or (<= rat threshold)
			  (<= len 3))
		      (begin
			(set! prev avg)
			(list avg alpha))
		      (more (- len 2)
			    (cddr lst)
			    (delv1! (first lst)
				    (delv1! (second lst) sorted))))))))))))

(define median-average-warmup-count
  (+ 4   ;; TMA smoothing
     39  ;; lookback for median
     (ema-warmup-count 39)))  ;; worst case smooth of 39


;;-----------------------------------------------------------------------------
;; filter series

(define (median-average-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (median-average-calc-proc)))

    (for-each proc (series-prev-n-closes parent #:before lo
					 #:count median-average-warmup-count))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (let ((ma (proc (array-ref array i))))
	    (array-set! new-array (and ma (first ma)) i))))))

(define-public (median-average-series series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "Median-Average"))
    (series-datatype!  new-series median-average-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   median-average-fill
				   median-average-warmup-count))
    (series-array!     new-series #f)

    new-series))

(define-public median-average-datatype
  (datatype-new!
   #:key     'median-average
   #:name    (_ "Median-Average Adaptive")
   #:manual  (_ "CONTEXT(manual-node): Median-Average Adaptive Filter")
   #:units   #f
   #:proc    median-average-series))


;;-----------------------------------------------------------------------------
;; alpha series

(define (median-average-alpha-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (median-average-calc-proc)))

    (for-each proc (series-prev-n-closes parent #:before lo
					 #:count median-average-warmup-count))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (let ((ma (proc (array-ref array i))))
	    (array-set! new-array (and ma (second ma)) i))))))

(define-public (median-average-alpha-series series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "Median-Average Alpha"))
    (series-datatype!  new-series median-average-alpha-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   median-average-alpha-fill
				   median-average-warmup-count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public median-average-alpha-datatype
  (datatype-new!
   #:key      'median-average-alpha
   #:name     (_ "Median-Average Alpha")
   #:priority -10
   #:manual   (_ "CONTEXT(manual-node): Median-Average Adaptive Filter")
   #:units    'ema-alpha
   #:minimum  0 ;; actually only down to count==39, which is 0.05, but show 0
   #:maximum  (ema-count->alpha 3)
   #:proc     median-average-alpha-series))
