;; Median of past N days.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart median)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a moving median over a window of
;; COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the median up to (and including) that value, or #f if not enough
;; values yet.
;;
;; The window can be primed by calling with COUNT-1 many values preceding
;; the first desired.  `median-warmup-count' below returns that count.
;;
(define*-public (median-calc-proc count #:optional (data '()))

  ;; POINTS is how many values retained so far; it stops at COUNT
  ;;
  ;; SORTED is a sorted list of the values, in ascending order.
  ;;
  ;; Q is an ice-9 queue of the values in the order they arrived.  The next
  ;; to `deq!' is the oldest.
  ;;
  ;; MIDPOS is an index into SORTED of the median position, or that position
  ;; rounded down when POINTS is an even number.
  ;;
  (set! data (take-right-maybe data (median-warmup-count count)))
  (let ((sorted     (sort data <))
	(q          (list->q data))
	(points     (length data))
	(midpos     (ash (length data) -1))
	(min-points (ceiling-exact (* indicator-minimum-fraction count))))
    (set! data #f) ;; gc initial

    (lambda (x)
      (if (< points count)
	  (begin
	    ;; not full yet, don't remove anything, but median pos updates
	    (set! points (1+ points))
	    (set! midpos (ash points -1)))

	  ;; oldest item removed from queue and then that item removed from
	  ;; sorted list
	  (set! sorted (delv1! (deq! q) sorted)))

      ;; new item added to queue and to sorted list
      (enq! q x)
      (set! sorted (merge! sorted (list x) <))

      (and (>= points min-points)
	   (if (odd? points)
	       (list-ref sorted midpos)
	       (arithmetic-mean (list-ref sorted midpos)
				(list-ref sorted (1- midpos))))))))

(define-public median-warmup-count 1-)


(define (median-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (count  (series-data series))
	 (proc   (median-calc-proc count
				   (series-prev-n-closes
				    parent #:before lo
				    #:count (median-warmup-count count))))
	 (array  (series-array-closes parent lo hi)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (median-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Median") count))
    (series-datatype!  new-series median-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   median-fill
				   (median-warmup-count count)))
    (series-array!     new-series #f)

    new-series))

(define-public median-datatype
  (datatype-new!
   #:key      'median
   #:name     (_ "Moving Median")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Moving Median")
   #:proc     median-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'median-days
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
