;; Money Flow Index (MFI).

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mfi)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.linnsoft.com/tour/techind/mfi.htm
;;     Formula, and sample Apple (AAPL) from 1999/2000 as 9-day MFI smoothed
;;     by 7-day SMA.
;;


;; The MFI formula is sometimes written
;;
;;                      /       100       \
;;                     |  ---------------  |
;;         MFI = 100 - |      positive MF  |
;;                     |  1 + -----------  |
;;                      \     negative MF /
;;
;;
;; but it's much easier to rearrange that to what's shown in the manual,
;;
;;                     positive MF
;;         MFI = 100 * -----------
;;                      pos+neg MF



;; Return a procedure which calculates the percentage of positive numbers to
;; positive and negatives in a window of COUNT points.  Ie.
;;
;;           total pos
;;     100 * ----------
;;           total pos+neg
;;
;; For the total pos+neg, negatives are added with their absolute value, so
;; the divisor is not merely the net amount.
;;
;; Each call (PROC x) enters a new value into the window, and the return is
;; the percentage up to (and including) that value, or #f if not enough
;; points yet.  Initial calls for with COUNT-1 values will ensure the next
;; call is based on a full window.
;;
;; When all the points in the window are 0, the formula would be 100 * 0/0,
;; but we return a neutral value 50.
;;
(define-public (percent-posneg-calc-proc count)
  (let ((data       (make-circular-list count #f))
	(total      0)
	(pos-total  0)
	(points     0)
	;; at least 3/4 of points in window to give a value
	(min-points (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (x)

      ;; subtract old point from values
      (if (car data)
	  (begin
	    (set! total (- total (abs (car data))))
	    (if (positive? (car data))
		(set! pos-total (- pos-total (car data)))))

	  ;; no old point, so we're increasing the values in the window
	  (set! points (1+ points)))

      ;; add this point to totals
      (set! total (+ total (abs x)))
      (if (positive? x)
	  (set! pos-total (+ pos-total x)))

      ;; enter this point into window, replacing old
      (set-car! data x)
      (set! data (cdr data))

      ;; 100.0 forces result to flonum
      (and (>= points min-points)
	   (if (= 0 total)
	       50
	       (* 100.0 (/ pos-total total)))))))


;; Return a procedure which calculates an MFI over a window of COUNT many
;; values.
;;
;; Each call (PROC high low close volume) enters a new point into the
;; window, and the return is the MFI up to (and including) that value, or #f
;; if not enough points yet.
;;
;; The typical price TP is (high + low + close) / 3, but if HIGH and LOW are
;; #f then just CLOSE alone is used.  This adapts to data where daily
;; high/low is not available.  Volume must be available though.
;;
;; To prime the window initially, call PROC with COUNT+2 many values
;; preceding the first desired point.
;;
(define-public (mfi-calc-proc count)
  (let ((prev-tp   #f)
	(perc-proc (percent-posneg-calc-proc count)))

    (lambda (high low close volume)
      ;; use closes if no high/low data
      (set! high (or high close))
      (set! low  (or low  close))

      ;; if we have a close then want to work that day into the calculation,
      ;; use volume==0 if necessary
      (or volume
	  (set! volume 0))

      (let* ((tp (* 0.33333333333333333333333333 (+ high low close)))
	     ;; money flow is +tp*volume, -tp*volume, or 0
	     (mf (and prev-tp
		      (cond ((> tp prev-tp)
			     (* tp volume))
			    ((< tp prev-tp)
			     (- (* tp volume)))
			    (else
			     0)))))
	(set! prev-tp tp)
	(and mf
	     (perc-proc mf))))))


(define (mfi-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count count))
	 (array  (series-array parent start hi))
	 (mfi    (mfi-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (mfi (array-ref array i 1)    ;; high
	       (array-ref array i 2)    ;; low
	       (array-ref array i 3)    ;; close
	       (array-ref array i 4)))) ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (mfi (array-ref array i 1)  ;; high
				     (array-ref array i 2)  ;; low
				     (array-ref array i 3)  ;; close
				     (array-ref array i 4)) ;; volume
		      i)))))

(define-public (mfi-series series count)
  (ohlcv-required "MFI" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "MFI") count))
    (series-datatype!  new-series mfi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts mfi-fill count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public mfi-datatype
  (datatype-new!
   #:key     'mfi
   #:name    (_ "MFI - Money Flow Index")
   #:input   datatype-ohlcv
   #:units   'percentage
   #:minimum 0
   #:maximum 100
   #:hlines  '(20 80)
   #:manual  (_ "CONTEXT(manual-node): Money Flow Index")
   #:proc    mfi-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'mfi-days
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
