;; Minneapolis Grain Exchange (MGEX) setups.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mgex)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart barchart-customer)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (mgex-symbol? symbol)
  (string-suffix? ".MGEX" symbol))

;; timezone-chicago representing US central
(symbol-timezone! mgex-symbol? (_ "Minneapolis") timezone-chicago)
(source-help! mgex-symbol?
	      (_ "CONTEXT(manual-node): Minneapolis Grain Exchange"))



;;-----------------------------------------------------------------------------
;; weblink - product info
;;
;; Believe these overview pages are more helpful than the contracts specs
;; page.

(weblink-handler! mgex-symbol?
  (lambda (symbol)
    (list (_ "MGEX _Product Information")
	  (_ "Open web browser at the Minneapolis Grain Exchange product information page for this company")
	  (if (member (chart-symbol-commodity symbol) '("MW" "MWE"))
	      "http://www.mgex.com/spring_wheat.html"
	      "http://www.mgex.com/indexes_index.html"))))


;;-----------------------------------------------------------------------------
;; misc

;; Return the mdate which is the current front month for MGEX.
;; The last trading day for a given contract is the business day preceding
;; the 15th of the month.  So if we're after the 14th, rounded down if
;; that's a weekend, then the front month is not the current month but next.
;;
(define (mgex-front-mdate)
  (let* ((ymd              (today-ymd (timezone-chicago)))
	 (tdate-today      (apply ymd->tdate ymd))
	 (tdate-14th-floor (ymd->tdate (first ymd) (second ymd) 14))
	 (mdate            (apply ymd->mdate ymd)))
    (if (> tdate-today tdate-14th-floor)
	(1+ mdate)
	mdate)))


;;-----------------------------------------------------------------------------
;; latest quotes - delayed
;;
;; The pages like
;;
;;     http://www.mgex.com/quotes.html?page=quote&sym=MW
;;
;; use barchart.com like
;;
;;     http://sites.barchart.com/pl/mgex/quote.htx?&page=quote&sym=MW&mode=i
;;
;; for the actual data.  We go directly there and parse the javascript
;; output statements.  MW is about 7.5kbytes.


;; return list of latest records
(define (mgex-body->latest-list body)
  (set! body (java-write body))
  (let* ((row-list    (html-table-rows body))
	 ;; eg. "<B>SPRING WHEAT</B></B> Delayed Futures -15:40 - Friday, 18 March<BR>"
	 (m           (must-match (regexp-exec #,(regexp "(.*) Delayed Futures -([0-9]*:[0-9]*)[^0-9]*([0-9]+ [A-Za-z]*)") (first (first row-list)))))
	 (name        (match:substring m 1))
	 (quote-time  (h:m:s-str->seconds (match:substring m 2)))
	 (quote-adate (d/m-str->adate-nearest (match:substring m 3))))
    ;; drop name/date heading
    (set! row-list (cdr row-list))

    (let ((headings (first row-list)))
      (or (equal? headings '("Contract" "Last" "Change" "Bid" "Ask"
			     "Open" "High" "Low" "Prev. Stl." "Time"))
	  (error "MGEX delayed: unrecognised data columns:" headings)))
    (set! row-list (cdr row-list))

    (latest-list-add-front-months
     (map-rows!
      (lambda (month+symbol last change bid offer open high low
			    prev-settle last-time)

	;; month+symbol is eg. "May '05 (MWK05)"
	(let* ((m          (must-match (regexp-exec #,(regexp "\\((.*)\\)")
						    month+symbol)))
	       (symbol     (match:substring m 1))
	       (mdate      (chart-symbol-mdate symbol))
	       (last-adate quote-adate))

	  ;; exclude "KW" kcbt wheat and "W" cbot wheat on the "MW" page
	  (and (not (member (chart-symbol-commodity symbol) '("KW" "W")))
	       (begin

		 ;; a dodgy entry for ICZ16 on 3Jan07 had "no record"
		 (if (string=? last "no record")
		     (set! last ""))

		 ;; "s" for settlement price
		 (set! last (string-trim-right last #\s))

		 ;; "unch" for no change
		 (if (string=? "unch" change)
		     (set! change "0"))

		 (cond ((string=? ":" last-time)
			;; saw empty ":" on dodgy ICZ16
			(set! last-adate #f)
			(set! last-time #f))

		       ((string-any #\/ last-time)
			;; last-time is a date like "05/27/05"
			(set! last-adate (m/d/y-str->adate last-time))
			(set! last-time #f))

		       (else
			(set! last-time (h:m:s-str->seconds last-time))

			;; if time is after what's in the heading, then prices
			;; are from the day before; except on the weekend when
			;; they're always from Friday (quote-adate is already
			;; Friday)
			(if (and (not (adate-weekend? quote-adate))
				 (> last-time quote-time))
			    (set! last-adate (1- quote-adate)))))

		 (latest-new #:symbol (string-append symbol ".MGEX")
			     #:name           name
			     #:quote-adate    quote-adate
			     #:quote-time     quote-time
			     #:bid            (dash-eighths->decimals bid)
			     #:offer          (dash-eighths->decimals offer)
			     #:last-adate     last-adate
			     #:last-time      last-time
			     #:open           (dash-eighths->decimals open)
			     #:high           (dash-eighths->decimals high)
			     #:low            (dash-eighths->decimals low)
			     #:last           (dash-eighths->decimals last)
			     #:change         (dash-eighths->decimals change)
			     #:contract-mdate mdate
			     #:source         'mgex)))))
      row-list))))

;; get the delayed quote pages for SYMBOL-LIST (possibly an empty list)
(define (delayed-get symbol-list proc)
  (for-each (lambda (commodity)
	      (receive (headers body)
		  (http-request (string-append "http://sites.barchart.com/pl/mgex/quote.htx?&page=quote&mode=i&sym=" commodity)
				#:want-ok #t)
		(proc (mgex-body->latest-list body))))
	    (delete-duplicates! (map chart-symbol-commodity symbol-list))))


;;-----------------------------------------------------------------------------
;; latest - aquote index futures
;;
;; This uses the real-time agricultural index futures part of
;;
;;     http://www.mgex.com/data_charts.html
;;
;; The actual data comes from some javascript
;;
;;     http://sites.barchart.com/pl/mgex/aquotes.htx
;;
;; This is about 11kbytes, it updates every 60 seconds apparently, but
;; there's no ETag or Last-Modified to save re-downloading.


(define (aquote-name->commodity name)
  (or (assoc-ref '(("PIT NCI" . "IC")
		   ("NCI"     . "IC")
		   ("HRWI"    . "IH")
		   ("HRSI"    . "IP")
		   ("SRWI"    . "IW")
		   ("NSI"     . "IS")) name)
      (error "MGEX: Unrecognised aquote commodity:" name)))

;; turn say "5 x 195-2" into just "195-2"
(define (aquote-strip-x str)
  (let* ((m (regexp-exec #,(regexp "^[0-9]+ x ") str)))
    (if m
	(match:suffix m)
	str)))

;; return list of latest records, or #f
;;
(define (aquote-body->latest-list body)
  (define row-list (html-table-rows (java-write body)))

  ;; various empty rows
  (set! row-list (delete! '("") row-list))

  (let ((headings (first row-list)))
    (or (equal? headings '("Contract/Month" "Last" "Change" "Bid" "Ask"
			   "Open" "High" "Low" "PrevSettle" "Time"))
	(error "MGEX indexes: unrecognised data columns:" headings)))
  (set! row-list (cdr row-list))

  (let* ((quote-tm       (localtime (current-time) (timezone-chicago)))
	 (quote-adate    (tm->adate quote-tm))
	 (quote-time     (tm->time quote-tm)))

    (latest-list-add-front-months
     (map-rows! (lambda (name last change bid ask open
			      high low prev last-time)

		  ;; name like "SRWI Feb '06"
		  ;;           "PIT NCI Jan '06"
		  ;;           "NCI Mar '07 1900 Call"
		  (let ((m (must-match
			    (regexp-exec
			     #,(regexp "^((PIT )?[A-Z]+) ([A-Za-z]+) '([0-9][0-9])( [0-9]+ Call)?$") name))))

		    (and (not (match:start m 5)) ;; skip call options
			 (let* ((name      (match:substring m))
				(commodity (aquote-name->commodity
					    (match:substring m 1)))
				(month     (Mmm-str->month (match:substring m 3)))
				(year-str  (match:substring m 4))
				(mdate     (ymd->mdate (yyy-str->year year-str)
						       month
						       1))
				(last-adate quote-adate))

			   ;; "5 x " or whatever for count of bid/offers
			   (set! bid (aquote-strip-x bid))
			   (set! ask (aquote-strip-x ask))

			   ;; trailing "s" for settlement price
			   (set! last (string-trim-right last #\s))

			   ;; "unch" for no change
			   (if (string-contains change "unch")
			       (set! change "0"))

			   ;; have seen some far forward months empty, so
			   ;; watch out for empty string time
			   (set! last-time (if (string-null? last-time)
					       #f
					       (h:m:s-str->seconds last-time)))

			   (latest-new
			    #:symbol (format #f "~a~a~a.MGEX"
					     commodity
					     (month->M-code month)
					     year-str)
			    #:name           name
			    #:quote-adate    quote-adate
			    #:quote-time     quote-time
			    #:bid            (dash-eighths->decimals bid)
			    #:offer          (dash-eighths->decimals ask)
			    #:last-adate     last-adate
			    #:last-time      last-time
			    #:open           (dash-eighths->decimals open)
			    #:high           (dash-eighths->decimals high)
			    #:low            (dash-eighths->decimals low)
			    #:last           (dash-eighths->decimals last)
			    #:prev           (dash-eighths->decimals prev)
			    #:change         (dash-eighths->decimals change)
			    #:contract-mdate mdate
			    #:source         'mgex)))))
		row-list))))

;; return latest list, or #f
(define (aquote-get)
  (receive (headers body)
      (http-request "http://sites.barchart.com/pl/mgex/aquotes.htx")
    (and (http-headers-ok? headers)
	 (aquote-body->latest-list body))))


;;-----------------------------------------------------------------------------
;; latest quotes - all
;;
;; The strategy is to get the aquote realtime indexes if those near months
;; are wanted, then the rest from the delayed quotes.


(define (mgex-quote-adate-time symbol)
  (define tm (localtime (- (current-time)
			   ;; aquote front 2 index months real-time, others
			   ;; 10-minute delayed
			   (if (aquote-symbol? symbol)
			       0
			       #,(hms->seconds 0 10 0)))
			(timezone-chicago)))
  (define commodity (chart-symbol-commodity symbol))

  (if (string=? commodity "MW")
      ;; MW pit trading weekdays 9:30am to 1:15pm
      (tm->adate-time-within tm #,(hms->seconds 9 30 0) 
			     #,(hms->seconds 13 15 0))

      ;; MWE electronic trading Sun-Fri 7:32pm to 6:00am (CT)
      ;; other electronic trading Sun-Fri 7:30pm to 1:45 (CT)
      ;; (actually IC starts 7:30pm, IS at 7:31pm then rest 7:32pm, but
      ;; don't worry about that, just use 7:30pm for all
      ;; stretch start and end by 2 minutes like `tm->adate-time-within'
      ;; does, in case the system time is out
      ;;
      (let ((adate    (tm->adate tm))
	    (time     (tm->time tm))
	    (end-time (if (string=? commodity "MWE")
			     #,(hms->seconds 6 2 0)
			     #,(hms->seconds 13 47 0))))
	(cond ((= 5 (tm:wday tm))
	       ;; friday, limit to END-TIME
	       (set! time (min time end-time)))
	      ((= 6 (tm:wday tm))
	       ;; saturday, back to Fri END-TIME
	       (set! adate (1- adate))
	       (set! time  end-time))
	      ((positive? (tm:wday tm))
	       ;; mon-thu, until 7:30pm limit back to END-TIME
	       (if (< time #,(hms->seconds 19 28 0))
		   (set! time (min time end-time))))
	      ((< time #,(hms->seconds 19 28 0))
	       ;; sunday before 7:30pm, back to Fri END-TIME
	       (set! adate (- adate 2))
	       (set! time  end-time)))
	(list adate time))))

;; Index futures exist for all months, and the aquote download gives the
;; nearest two.  If mgex-front-mdate rolls to the next month a bit earlier
;; than the aquote page does then it won't matter, since the delayed quote
;; is used as a fallback.
;;
(define (aquote-symbol? symbol)
  (and (string-prefix? "I" symbol)
       (let ((mdate (chart-symbol-mdate symbol)))
	 (or (not mdate)
	     (<= mdate (1+ (mgex-front-mdate)))))))

(define (mgex-latest-get symbol-list extra-list proc)
  ;; do delayed-get if no aquotes, or if aquote was gotten successfully
  (if (or (not (any aquote-symbol? symbol-list))
	  (and-let* ((latest-list (aquote-get)))
	    (proc latest-list)
	    ;; remove those already fetched
	    (set! symbol-list (lset-difference string=? symbol-list
					       (map latest-symbol
						    latest-list)))
	    #t))
      (delayed-get symbol-list proc)))

(latest-handler! #:selector   mgex-symbol?
		 #:handler    mgex-latest-get
		 #:adate-time mgex-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - name
;;
;; Names are extracted from the little menu in the historical download form,
;;
;;     http://exchanges.barchart.com/pl/mgex/data.js
;;
;; and month/year appended.
;;
;; There's an ETag on this page, but it doesn't seem worth using that when
;; we check names only infrequently.


(define (mgex-process-name symbol-list body)
  (set! body (java-write body))
  (let* ((cc-pos  (string-contains body "<SELECT NAME=\"cc\""))
	 (end-pos (string-contains body "</SELECT>" cc-pos))
	 ;; list (COMMODITY . NAME)
	 (alist   (map! (lambda (m)
			  (cons (match:substring m 1) (match:substring m 2)))
			(must-match
			 (list-matches (make-regexp "<OPTION VALUE=\"[^\">]+\">([A-Z0-9]+) - ([^<]+)</OPTION>" regexp/icase)
				       (substring body cc-pos end-pos))))))
    (for-each
     (lambda (symbol)
       (download-status (_ "MGEX") (_ "name") symbol)
       (let ((name  (assoc-ref alist (chart-symbol-commodity symbol)))
	     (mdate (chart-symbol-mdate symbol)))
	 (if (and name mdate)
	     (set! name (string-append name " - "
				       (strftime-utf8 "%b %y"
						      (mdate->tm mdate)))))
	 (database-call-modify-data symbol
	   (lambda (series)
	     (series-name! series name)
	     (series-info-add! series 'name-time (current-time))))))
     symbol-list)))

(define (mgex-download-name symbol-list)
  (set! symbol-list (filter download-want-name? symbol-list))
  (if (not (null? symbol-list))
      (begin
	(download-status (_ "MGEX") (_ "names"))
	(receive (headers body)
	    (http-request "http://exchanges.barchart.com/pl/mgex/data.js"
			  #:want-ok #t)
	  ;; process all symbols while we're at it
	  (mgex-process-name (download-also symbol-list
					    #:selector mgex-symbol?)
			     body)))))


;;-----------------------------------------------------------------------------
;; download - data
;;
;; This uses the download form at
;;
;;     http://www.mgex.com/download.html
;;
;; which ends up at barchart, eg. for MWU07 wheat Sep07,
;;
;;     http://customer1.barchart.com/cgi-bin/mri/mgexdata.csv?sym=MWU07&vol=Y&cc=MW&mon=U&year=07&data=d.csv
;;


;; Eg. Symbol,Date,Open,High,Low,Last,Volume,Open Interest
;;     MWN05,02/05/2004,408,408,406.5,406.5,8886,33895
;;
(define* (mgex-process-csv symbol-list body #:optional start)

  (define row-list (map! (lambda (line)
			   (string-split line #\,))
			 (string-split-lines body)))

  (let ((headings (map! string-trim-right (first row-list))))
    (or (equal? headings
		'("Symbol" "Date" "Open" "High" "Low" "Last"
		  "Volume" "Open Interest"))
	(error "MGEX data: unrecognised csv headings:" headings)))
  (set! row-list (cdr row-list))

  (download-process
   #:module        (_ "MGEX")
   #:symbol-list   symbol-list
   #:last-download #t
   ;; there's entries for the delivery days immediately following expiry,
   ;; but the prices there aren't from any trading, so need expiry date
   #:expiry-proc   mgex-mdate-expiry
   #:currency      "USD"
   #:row-list
   (map-rows!
    (lambda (symbol tdate open high low last volume openint)
      (list #:tdate         (m/d/y-str->tdate tdate)
	    #:mdate         (chart-symbol-mdate symbol)
	    #:commodity     (chart-symbol-commodity symbol)
	    #:open          open
	    #:high          high
	    #:low           low
	    #:close         last
	    #:volume        volume
	    #:open-interest openint))
    row-list)))

;; return tdate of available download data
;; today not available until 5pm
(define (mgex-available-tdate . dummy-symbol)
  (tdate-today-after 17 0 (timezone-chicago)))

;; return tdate of last trading for MDATE
;; this is the business day before the 15th of the month
;; public holidays are ignored here
;;
(define (mgex-mdate-expiry mdate)
  (receive-list (year month day)
      (mdate->ymd mdate)
    (ymd->tdate year month 14)))

;; return tdate of last trading for SYMBOL
(define (mgex-symbol-expiry symbol)
  (let ((mdate (chart-symbol-mdate symbol)))
    (if mdate
	(mgex-mdate-expiry mdate)  ;; specific month
	(+ 100 (tdate-today)))))   ;; rolling front month, no expiry

;; return mdate which the earliest not yet expired on TDATE
(define (mgex-tdate->front-mdate tdate)
  (receive-list (year month day)
      (tdate->ymd tdate)
    (ymd->mdate year (if (< day 15)
			 month
			 (1+ month)) 1)))

(define (mgex-download-one symbol-list commodity mdate)
  (define specific (string-append commodity
				  (mdate->MYY-str mdate)))
  (download-status (_ "MGEX") (_ "data") specific)

  (receive (headers body)
      (http-request (string-append "http://customer1.barchart.com/cgi-bin/mri/mgexdata.csv?vol=Y&data=d.csv&sym=" specific)
		    #:want-ok #t)
    ;; second line " Error: No Data" for invalid month
    (and (not (string-contains body "Error: No Data"))
	 (mgex-process-csv symbol-list body))))

(define (mgex-download-data symbol-list)
  (futures-download-by-specific
   symbol-list
   #:download-proc    mgex-download-one
   #:avail-tdate-proc mgex-available-tdate))

(define (mgex-download symbol-list)
  (mgex-download-data symbol-list)
  (mgex-download-name symbol-list))

(download-handler! #:selector mgex-symbol?
		   #:update   mgex-download
		   #:backto
		   (lambda (want-tdate symbol-list)
		     (download-message "MGEX back data not yet supported")))


;;-----------------------------------------------------------------------------
;; intraday

(barchart-customer-intraday
 mgex-symbol?
 (lambda (symbol)
   (let ((date-time (mgex-quote-adate-time symbol)))
     (set-first! date-time (adate->tdate (first date-time)))))
 "http://customer1.barchart.com/cgi-bin/mri/mgexchart.htx?page=chart&code=mfo&org=com&crea=Y")
