/* Chart miscellaneous C code.

   Copyright 2003, 2005, 2006, 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <string.h>
#include <langinfo.h>
#include <nl_types.h>
#include <stdio.h>     /* for NULL */

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

/* TIOCGWINSZ might be in termios.h according to the linux kernel
   tty_ioctl(4) man page, suspect that's not quite right */
#if HAVE_TERMIOS_H
#include <termios.h>
#endif

/* TIOCGWINSZ is in sys/ioctl.h on glibc 2.3.5 */
#if HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#include <libguile.h>

#include "gettext.h"
#include "misc.h"
#include "compat.h"


/*---------------------------------------------------------------------------*/
/* compatibility */

#if ! HAVE_SCM_ONEMINUS  /* new in guile 1.8 */
#define scm_oneminus(x)  scm_difference (x, scm_from_int (1))
#endif

#if ! HAVE_SCM_ONEPLUS  /* new in guile 1.8 */
#define scm_oneplus(x)   scm_sum (x, scm_from_int (1))
#endif


#define MAX(x,y) ((x) > (y) ? (x) : (y))


/*---------------------------------------------------------------------------*/
/* locale info */

/* only used within misc.scm, so not made public */
SCM_DEFINE (chart_mb_cur_max, "mb-cur-max", 0, 0, 0,
            (),
	    "Return @code{MB_CUR_MAX}, the maximum number of bytes to encode\n"
	    "a character in the current locale charset.")
#define FUNC_NAME s_chart_mb_cur_max
{
  return scm_from_int (MB_CUR_MAX);
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_locale_codeset, "locale-codeset", 0, 0, 0,
            (),
	    "Return the locale charset name (a string).")
#define FUNC_NAME s_chart_locale_codeset
{
  return scm_from_locale_string (nl_langinfo (CODESET));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_locale_d_fmt, "locale-d-fmt", 0, 0, 0,
            (),
	    "Return the locale date format as a utf8 @code{strftime} format.")
#define FUNC_NAME s_chart_locale_d_fmt
{
  static SCM ret = SCM_BOOL_F;
  if (scm_is_false (ret))
    {
      char  *str = nl_langinfo (D_FMT);

      /* an empty string is not acceptable */
      if (str[0] == '\0')
        str = "%x";

      ret = scm_permanent_object
        (chart_locale_to_utf8 (scm_from_locale_string (str)));
    }
  return ret;
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_locale_decimal_point, "locale-decimal-point", 0, 0, 0,
            (),
	    "Return the locale decimal point as a utf8 string.")
#define FUNC_NAME s_chart_locale_decimal_point
{
  static SCM ret = SCM_BOOL_F;
  if (scm_is_false (ret))
    {
      char *str = localeconv()->decimal_point;

      /* an empty string is not acceptable */
      if (str[0] == '\0')
        str = ".";

      ret = scm_permanent_object
        (chart_locale_to_utf8 (scm_from_locale_string (str)));
    }
  return ret;
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_locale_thousands_sep, "locale-thousands-sep", 0, 0, 0,
            (),
	    "Return the locale thousands separator as a utf8 string.  This\n"
	    "is an empty string when no thousands separator should be shown.")
#define FUNC_NAME s_chart_locale_thousands_sep
{
  static SCM ret = SCM_BOOL_F;
  if (scm_is_false (ret))
    {
      ret =  scm_permanent_object
        (chart_locale_to_utf8
         (scm_from_locale_string (localeconv()->thousands_sep)));
    }
  return ret;
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* Rounding */

SCM_DEFINE_PUBLIC (chart_ceiling_exact, "ceiling-exact", 1, 0, 0,
            (SCM x),
	    "Same as @code{ceiling}, but returning an exact integer.")
#define FUNC_NAME s_chart_ceiling_exact
{
  return scm_inexact_to_exact (scm_ceiling (x));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_floor_exact, "floor-exact", 1, 0, 0,
            (SCM x),
	    "Same as @code{floor}, but returning an exact integer.")
#define FUNC_NAME s_chart_floor_exact
{
  return scm_inexact_to_exact (scm_floor (x));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_round_exact, "round-exact", 1, 0, 0,
            (SCM x),
	    "Same as @code{round}, but returning an exact integer.")
#define FUNC_NAME s_chart_round_exact
{
  return scm_inexact_to_exact (scm_round_number (x));
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* Numeric stuff */

SCM_DEFINE_PUBLIC (chart_quotient_floor, "quotient-floor", 2, 0, 0,
            (SCM n, SCM d),
	    "Return the quotient @var{n} divided by @var{d}, rounded\n"
	    "downwards, ie.@: towards negative infinity.\n"
	    "\n"
	    "Currently only positive integers @var{d} are supported.")
#define FUNC_NAME s_chart_quotient_floor
{
  if (scm_is_false (scm_positive_p (d)))
    SCM_OUT_OF_RANGE (SCM_ARG2, d);

  if (scm_is_true (scm_negative_p (n)))
    n = scm_difference (n, scm_oneminus (d));

  return scm_quotient (n, d);
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_square, "square", 1, 0, 0,
            (SCM x),
	    "Return the square of @var{x}, ie.@: @math{@var{x} * @var{x}}.")
#define FUNC_NAME s_chart_square
{
  return scm_product (x, x);
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_arithmetic_mean, "arithmetic-mean", 1, 0, 1,
            (SCM arg1, SCM rest),
	    "Return the arithmetic mean of the arguments, ie.@:\n"
	    "\n"
	    "@example\n"
	    "       arg1 + arg2 + ... + argN\n"
	    "mean = ------------------------\n"
	    "                  N\n"
	    "@end example\n"
	    "\n"
	    "This is always returned as a flonum.")
#define FUNC_NAME s_chart_arithmetic_mean
{
  double sum = scm_to_double (arg1);
  int count = 1;

  for ( ; ! scm_is_null (rest); rest = scm_cdr (rest))
    {
      sum += scm_to_double (scm_car (rest));
      count++;
    }
  return scm_from_double (sum / count);
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_max_maybe, "max-maybe", 0, 0, 1,
            (SCM rest),
	    "Return the biggest of the given arguments, like @code{max}\n"
	    "does, but ignore any @code{#f} arguments.  If there's no\n"
	    "non-@code{#f} arguments, or no arguments at all, the return is\n"
	    "@code{#f}.")
#define FUNC_NAME s_chart_max_maybe
{
  SCM biggest = SCM_BOOL_F;  /* default, for no args at all */
  int argnum;

  for (argnum = 1; ! scm_is_null (rest); rest = scm_cdr (rest), argnum++)
    {
      SCM this = scm_car (rest);
      if (! scm_is_false (this))
        {
          if (! scm_is_real (this))  /* if not #f then must be a number */
            SCM_WRONG_TYPE_ARG (argnum, this);

          if (scm_is_false (biggest))
            biggest = this;
          else
            biggest = scm_max (biggest, this);
        }
    }

  return biggest;
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_min_maybe, "min-maybe", 0, 0, 1,
            (SCM rest),
	    "Return the smalles of the given arguments, like @code{min}\n"
	    "does, but ignore any @code{#f} arguments.  If there's no\n"
	    "non-@code{#f} arguments, or no arguments at all, the return is\n"
	    "@code{#f}.")
#define FUNC_NAME s_chart_min_maybe
{
  SCM smallest = SCM_BOOL_F;  /* no args at all */
  int argnum;

  for (argnum = 1; ! scm_is_null (rest); rest = scm_cdr (rest), argnum++)
    {
      SCM this = scm_car (rest);
      if (! scm_is_false (this))
        {
          if (! scm_is_real (this))
            SCM_WRONG_TYPE_ARG (argnum, this);

          if (scm_is_false (smallest))
            smallest = this;
          else
            smallest = scm_min (smallest, this);
        }
    }

  return smallest;
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* List stuff */

SCM_DEFINE_PUBLIC (chart_length_ge_p, "length>=?", 2, 0, 0,
            (SCM lst, SCM n),
	    "Return true if @var{lst} length is >= @var{n}.")
#define FUNC_NAME s_chart_length_ge_p
{
  unsigned nn = scm_to_uint (n);

  for (;;)
    {
      if (nn == 0)
        return SCM_BOOL_T;

      if (scm_is_null (lst))
        return SCM_BOOL_F;

      lst = scm_cdr (lst);
      nn--;
    }
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_plist_get, "plist-get", 2, 0, 0,
                   (SCM plist, SCM key),
                   "Return the value associated with @var{key} in @var{plist}, or\n"
                   "@code{#f} if no entry for @var{key}.\n"
                   "\n"
                   "@var{plist} is a property list, meaning a list @code{(@var{key}\n"
                   "@var{value} @var{key} @var{value} @dots{})}.")
#define FUNC_NAME s_chart_plist_get
{
  /* (define (plist-get plist key) */
  /*   (and (not (null? plist)) */
  /*        (if (eq? key (first plist)) */
  /* 	   (second plist) */
  /* 	   (plist-get (cddr plist) key)))) */

  for ( ; ! scm_is_null (plist); plist = scm_cddr (plist))
    if (scm_is_eq (scm_car (plist), key))
      return scm_cadr (plist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* Misc stuff */

SCM_DEFINE_PUBLIC (chart_output_port_width, "output-port-width", 0, 1, 0,
            (SCM port),
	    "Return the number of columns of output available on @var{port},\n"
	    "or 79 if it cannot be determined.  @var{port} defaults to\n"
	    "@code{current-output-port}.\n"
	    "\n"
	    "The width of a tty is obtained from @code{TIOCGWINSZ}, if that\n"
	    "@code{ioctl} is available.  For other ports only the default 79\n"
	    "is returned.\n"
	    "\n"
	    "This is compatible with the Slib function of the same name.")
#define FUNC_NAME s_chart_output_port_width
{
  if (SCM_UNBNDP (port))
    port = scm_current_output_port ();

  SCM_VALIDATE_PORT (SCM_ARG1, port);

  /* A pseudo-tty under emacs has ws_col==0, give the default 79 in that
     case. */
#ifdef TIOCGWINSZ
  if (scm_is_true (scm_file_port_p (port)))
    {
      struct winsize w;
      if (ioctl (scm_to_int (scm_fileno (port)), TIOCGWINSZ, &w) == 0
          && w.ws_col != 0)
        return scm_from_int (w.ws_col);
    }
#endif

  return scm_from_int (79);
}
#undef FUNC_NAME


#if ! HAVE_SCM_PRIMITIVE__EXIT
SCM_DEFINE_PUBLIC (chart_primitive__exit, "primitive-_exit", 1, 0, 0,
            (SCM status),
	    "Call the @code{_exit} system function, to immediately terminate\n"
	    "the process, with no Scheme level unwinding and no C level\n"
	    "@code{atexit} routines and stdio flushing.")
#define FUNC_NAME s_chart_primitive__exit
{
  _exit (SCM_UNBNDP (status) ? 0 : scm_to_int (status));
}
#undef FUNC_NAME
#endif


#ifdef scm_tc3_struct /* new in guile 1.8 */
SCM_DEFINE_PUBLIC (chart_record_struct_copy, "record-struct-copy", 1, 0, 0,
            (SCM st),
	    "Return a copy of record structure @var{rec}.  This is a shallow\n"
	    "copy, there's no recursive copying of the objects in the\n"
	    "fields.\n"
	    "\n"
	    "There's no checking that @var{rec} is actually a record, only\n"
	    "that it's a structure, and structure types other than records\n"
	    "are not supported.")
#define FUNC_NAME s_chart_record_struct_copy
{
  scm_t_bits *vtable_data, *st_data, *new_st_data;
  size_t n_words;

  SCM_VALIDATE_STRUCT (SCM_ARG1, st);

  vtable_data = SCM_STRUCT_VTABLE_DATA (st);

  /* Only standard structs handled here, not goops entity or light forms.
     Believe that's enough for ordinary public uses of structures as created
     from `make-struct', `make-vtable', `make-vtable-vtable'.  */
  if (vtable_data[scm_struct_i_flags]
      & (SCM_STRUCTF_ENTITY | SCM_STRUCTF_LIGHT))
    SCM_WRONG_TYPE_ARG (SCM_ARG1, st);

  st_data = SCM_STRUCT_DATA (st);
  n_words = st_data[scm_struct_i_n_words];
  new_st_data = scm_alloc_struct (n_words,
                                  scm_struct_n_extra_words,
                                  "struct");

  memcpy (new_st_data, st_data, n_words * sizeof(scm_t_bits));
  scm_remember_upto_here_1 (st);

  return scm_double_cell ((scm_t_bits) vtable_data + scm_tc3_struct,
                          (scm_t_bits) new_st_data, 0, 0);
}
#undef FUNC_NAME
#endif


/*---------------------------------------------------------------------------*/
/* Diagnostics

   http://osr507doc.sco.com/en/man/html.S/malloc.S.html
       SCO malloc/mallinfo man page.

   http://www.ecoscentric.com/ecospro/doc/html/ref/memalloc-standard-api.html
       eCos, only fields arena, ordblks, uordblks, fordblks and maxfree.
   */

SCM_DEFINE_PUBLIC (chart_mallinfo, "mallinfo", 0, 0, 0,
            (),
            "")
#define FUNC_NAME s_chart_mallinfo
{
#if HAVE_MALLINFO
  struct mallinfo m = mallinfo();
  SCM ret = SCM_EOL;
  SCM *tail = &ret;

  /* this isn't used much, so prefer to create the symbols only when needed */
#define ADD(val,name)                                                         \
  do {                                                                        \
    *tail = scm_acons (scm_from_locale_symbol (name), scm_from_int (val),     \
                       SCM_EOL);                                              \
    tail = SCM_CDRLOC (*tail);                                                \
  } while (0)

  ADD (m.arena,    "arena");
  ADD (m.ordblks,  "ordblks");
  ADD (m.smblks,   "smblks");
  ADD (m.hblks,    "hblks");
  ADD (m.hblkhd,   "hblkhd");
  ADD (m.usmblks,  "usmblks");
  ADD (m.fsmblks,  "fsmblks");
  ADD (m.uordblks, "uordblks");
  ADD (m.fordblks, "fordblks");
  ADD (m.keepcost, "keepcost");
#undef ADD
  return ret;
#else
  /* mingw heapwalk() could give some similar information, if anyone's using
     that */
  return SCM_EOL;
#endif
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_mallinfo_used, "mallinfo-used", 0, 0, 0,
            (),
	    "Return the total memory space, in bytes, currently allocated\n"
	    "through @code{malloc} and friends.")
#define FUNC_NAME s_chart_mallinfo_used
{
#if HAVE_MALLINFO
  struct mallinfo m = mallinfo();

  return scm_from_int (m.hblkhd       /* mmap block space in use */
                       + m.usmblks    /* small block space in use */
                       + m.uordblks); /* ordinary block space in use */
#else
  return scm_from_int (0);
#endif
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* adates */

/* for positive d only */
#define QUOTIENT_FLOOR(n,d) \
  (((n) - ((n) < 0 ? (d)-1 : 0)) / (d))

/* for positive d only */
#define REMAINDER_FLOOR(rvar, n,d)       \
  do {                                   \
    (rvar) = (n) % (d);                  \
    if ((rvar) < 0)                      \
      (rvar) += d;                       \
  } while (0)

SCM_DEFINE_PUBLIC (chart_ymd_to_adate, "ymd->adate", 3, 0, 0,
            (SCM year, SCM month, SCM day),
	    "Return adate for the given @var{year}, @var{month} and\n"
	    "@var{day}.\n"
	    "\n"
	    "@var{year} is a full value, like 2003.\n"
	    "\n"
	    "@var{month} is 1 to 12, but can be outside that range to wrap\n"
	    "into an earlier or later year.\n"
	    "\n"
	    "@var{day} is 1 to 31, but can be outside that range to wrap to\n"
	    "an earlier or later date.")
#define FUNC_NAME s_chart_ymd_to_adate
{
  /* This explicit calculation is much faster than `mktime' (with "GMT"),
     but doesn't handle 1900, 2100 etc as non-leap years.  */

  static const int month_table[4][12] = {
    { -1, 30, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 },
    {  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 },
    {  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 },
    {  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 }
  };

  long c_year  = scm_to_long (year);
  long c_month = scm_to_long (month);
  long c_day   = scm_to_long (day);
  long adate;

  c_month--; /* 0 based instead of 1 based */

  /* bring month into range 0 to 11 */
  c_year += QUOTIENT_FLOOR (c_month, 12);
  REMAINDER_FLOOR (c_month, c_month, 12);

  adate =
    c_year * 365
    + (c_year >> 2)  /* leap day every 4 years */
    + month_table[c_year & 3][c_month]
    + c_day
    -719547L;
  return scm_from_long (adate);
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_adate_to_ymd, "adate->ymd", 1, 0, 0,
            (SCM adate),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given @var{adate}.")
#define FUNC_NAME s_chart_adate_to_ymd
{
  /* This table generated by the devel/ymd-to-adate.scm program.
     For a given day D starting from 0 for 1st of March, month_table[d][0]
     is the month (1 to 12) and month_table[d][1] is the day within that
     month (1 to 31).  */
  static const char month_table[366][2] = {
    /* March */
    {3,1},{3,2},{3,3},{3,4},{3,5},{3,6},{3,7},{3,8},{3,9},{3,10},{3,11},{3,12},
    {3,13},{3,14},{3,15},{3,16},{3,17},{3,18},{3,19},{3,20},{3,21},{3,22},
    {3,23},{3,24},{3,25},{3,26},{3,27},{3,28},{3,29},{3,30},{3,31},
    /* April */
    {4,1},{4,2},{4,3},{4,4},{4,5},{4,6},{4,7},{4,8},{4,9},{4,10},{4,11},{4,12},
    {4,13},{4,14},{4,15},{4,16},{4,17},{4,18},{4,19},{4,20},{4,21},{4,22},
    {4,23},{4,24},{4,25},{4,26},{4,27},{4,28},{4,29},{4,30},
    /* May */
    {5,1},{5,2},{5,3},{5,4},{5,5},{5,6},{5,7},{5,8},{5,9},{5,10},{5,11},{5,12},
    {5,13},{5,14},{5,15},{5,16},{5,17},{5,18},{5,19},{5,20},{5,21},{5,22},
    {5,23},{5,24},{5,25},{5,26},{5,27},{5,28},{5,29},{5,30},{5,31},
    /* June */
    {6,1},{6,2},{6,3},{6,4},{6,5},{6,6},{6,7},{6,8},{6,9},{6,10},{6,11},{6,12},
    {6,13},{6,14},{6,15},{6,16},{6,17},{6,18},{6,19},{6,20},{6,21},{6,22},
    {6,23},{6,24},{6,25},{6,26},{6,27},{6,28},{6,29},{6,30},
    /* July */
    {7,1},{7,2},{7,3},{7,4},{7,5},{7,6},{7,7},{7,8},{7,9},{7,10},{7,11},{7,12},
    {7,13},{7,14},{7,15},{7,16},{7,17},{7,18},{7,19},{7,20},{7,21},{7,22},
    {7,23},{7,24},{7,25},{7,26},{7,27},{7,28},{7,29},{7,30},{7,31},
    /* August */
    {8,1},{8,2},{8,3},{8,4},{8,5},{8,6},{8,7},{8,8},{8,9},{8,10},{8,11},{8,12},
    {8,13},{8,14},{8,15},{8,16},{8,17},{8,18},{8,19},{8,20},{8,21},{8,22},
    {8,23},{8,24},{8,25},{8,26},{8,27},{8,28},{8,29},{8,30},{8,31},
    /* September */
    {9,1},{9,2},{9,3},{9,4},{9,5},{9,6},{9,7},{9,8},{9,9},{9,10},{9,11},{9,12},
    {9,13},{9,14},{9,15},{9,16},{9,17},{9,18},{9,19},{9,20},{9,21},{9,22},
    {9,23},{9,24},{9,25},{9,26},{9,27},{9,28},{9,29},{9,30},
    /* October */
    {10,1},{10,2},{10,3},{10,4},{10,5},{10,6},{10,7},{10,8},{10,9},{10,10},
    {10,11},{10,12},{10,13},{10,14},{10,15},{10,16},{10,17},{10,18},{10,19},
    {10,20},{10,21},{10,22},{10,23},{10,24},{10,25},{10,26},{10,27},{10,28},
    {10,29},{10,30},{10,31},
    /* November */
    {11,1},{11,2},{11,3},{11,4},{11,5},{11,6},{11,7},{11,8},{11,9},{11,10},
    {11,11},{11,12},{11,13},{11,14},{11,15},{11,16},{11,17},{11,18},{11,19},
    {11,20},{11,21},{11,22},{11,23},{11,24},{11,25},{11,26},{11,27},{11,28},
    {11,29},{11,30},
    /* December */
    {12,1},{12,2},{12,3},{12,4},{12,5},{12,6},{12,7},{12,8},{12,9},{12,10},
    {12,11},{12,12},{12,13},{12,14},{12,15},{12,16},{12,17},{12,18},{12,19},
    {12,20},{12,21},{12,22},{12,23},{12,24},{12,25},{12,26},{12,27},{12,28},
    {12,29},{12,30},{12,31},
    /* January */
    {1,1},{1,2},{1,3},{1,4},{1,5},{1,6},{1,7},{1,8},{1,9},{1,10},{1,11},{1,12},
    {1,13},{1,14},{1,15},{1,16},{1,17},{1,18},{1,19},{1,20},{1,21},{1,22},
    {1,23},{1,24},{1,25},{1,26},{1,27},{1,28},{1,29},{1,30},{1,31},
    /* February */
    {2,1},{2,2},{2,3},{2,4},{2,5},{2,6},{2,7},{2,8},{2,9},{2,10},{2,11},{2,12},
    {2,13},{2,14},{2,15},{2,16},{2,17},{2,18},{2,19},{2,20},{2,21},{2,22},
    {2,23},{2,24},{2,25},{2,26},{2,27},{2,28},{2,29},
  };
  long c_adate = scm_to_long (adate);
  int year, month, day;

  c_adate -= 786;  /* change base so 0 is 1 Mar 1972 */

  /* 1461 days per 4-year leap cycle */
  year = 1972 + 4 * QUOTIENT_FLOOR (c_adate, 1461);
  REMAINDER_FLOOR (c_adate, c_adate, 1461);  /* from 1 Mar in YEAR */

  /* divide c_adate by 365 for years within the 4-year cycle, leaving it in
     the range 0 to 365 (counting from 1 Mar still) */
  {
    int sub_year = c_adate / 365;
    c_adate %= 365;

    if (sub_year == 4)
      {
        c_adate = 365;
        sub_year = 3;
      }
    year += sub_year;
  }

  month = month_table[c_adate][0];
  day = month_table[c_adate][1];
  year += (month < 3);

  return scm_list_3 (scm_from_int (year),
                     scm_from_int (month),
                     scm_from_int (day));
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* tdates */


SCM_DEFINE_PUBLIC (chart_tdate_to_adate, "tdate->adate", 1, 0, 0,
            (SCM tdate),
	    "Return the adate corresponding to @var{tdate}.")
#define FUNC_NAME s_chart_tdate_to_adate
{
  /* two extra days sat and sun for each full week in tdate */
  long c_tdate = scm_to_long (tdate);
  long weekends = QUOTIENT_FLOOR (c_tdate, 5);
  long c_adate = c_tdate + 2 * weekends;
  return scm_from_long (c_adate);
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_adate_to_tdate, "adate->tdate", 1, 0, 0,
            (SCM adate),
	    "Return the tdate corresponding to @var{adate}, rounding down to\n"
	    "Friday if @var{adate} is a Saturday or Sunday.")
#define FUNC_NAME s_chart_adate_to_tdate
{
  /* two extra days sat and sun for each full week in tdate */
  long c_adate = scm_to_long (adate);
  long weekends = QUOTIENT_FLOOR (c_adate, 7);
  long wday;
  REMAINDER_FLOOR (wday, c_adate, 7);
  return scm_from_long (c_adate - 2 * weekends - MAX (wday-4, 0));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_adate_to_tdate_ceil, "adate->tdate-ceil", 1, 0, 0,
            (SCM adate),
	    "Return the tdate corresponding to @var{adate}, rounding up to\n"
	    "next Monday if @var{adate} is a Saturday or Sunday.")
#define FUNC_NAME s_chart_adate_to_tdate_ceil
{
  /* two extra days sat and sun for each full week in tdate */
  long c_adate = scm_to_long (adate);
  long weekends = QUOTIENT_FLOOR (c_adate, 7);
  long wday;
  REMAINDER_FLOOR (wday, c_adate, 7);

  /* saturday has +1 for monday already, sunday +2 reduced to +1 for monday */
  return scm_from_long (c_adate - 2 * weekends - (wday == 6));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_ymd_to_tdate, "ymd->tdate", 3, 0, 0,
            (SCM year, SCM month, SCM day),
	    "Return tdate for the given @var{year}, @var{month} and\n"
	    "@var{day}.  Saturday or Sunday is rounded down to the previous\n"
	    "Friday.\n"
	    "\n"
	    "@var{year} is a full value, like 2003.\n"
	    "\n"
	    "@var{month} is 1 to 12, but can be outside that range to wrap\n"
	    "into an earlier or later year.\n"
	    "\n"
	    "@var{day} is 1 to 31, but can be outside that range to wrap to\n"
	    "an earlier or later date.")
#define FUNC_NAME s_chart_ymd_to_tdate
{
  return chart_adate_to_tdate (chart_ymd_to_adate (year, month, day));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_ymd_to_tdate_ceil, "ymd->tdate-ceil", 3, 0, 0,
            (SCM year, SCM month, SCM day),
	    "Return tdate for the given @var{year}, @var{month} and\n"
	    "@var{day}.  Saturday or Sunday is rounded up to the next\n"
	    "Monday.\n"
	    "\n"
	    "@var{year} is a full value, like 2003.\n"
	    "\n"
	    "@var{month} is 1 to 12, but can be outside that range to wrap\n"
	    "into an earlier or later year.\n"
	    "\n"
	    "@var{day} is 1 to 31, but can be outside that range to wrap to\n"
	    "an earlier or later date.")
#define FUNC_NAME s_chart_ymd_to_tdate_ceil
{
  return chart_adate_to_tdate_ceil (chart_ymd_to_adate (year, month, day));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_tdate_to_ymd, "tdate->ymd", 1, 0, 0,
            (SCM tdate),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given @var{tdate}.")
#define FUNC_NAME s_chart_tdate_to_ymd
{
  return chart_adate_to_ymd (chart_tdate_to_adate (tdate));
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* wdates */

SCM_DEFINE_PUBLIC (chart_wdate_to_tdate, "wdate->tdate", 1, 0, 0,
            (SCM wdate),
	    "Return tdate for the given @var{wdate}.")
#define FUNC_NAME s_chart_wdate_to_tdate
{
  return scm_product (wdate, scm_from_int (5));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_tdate_to_wdate, "tdate->wdate", 1, 0, 0,
            (SCM tdate),
	    "Return wdate for the week containing @var{tdate}.")
#define FUNC_NAME s_chart_tdate_to_wdate
{
  /* (define-public (tdate->wdate t) */
  /*   (quotient-floor t 5)) */

  return chart_quotient_floor (tdate, scm_from_int (5));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_wdate_to_ymd, "wdate->ymd", 1, 0, 0,
            (SCM wdate),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given wdate.")
#define FUNC_NAME s_chart_wdate_to_ymd
{
  return chart_tdate_to_ymd (chart_wdate_to_tdate (wdate));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_ymd_to_wdate, "ymd->wdate", 3, 0, 0,
            (SCM year, SCM month, SCM day),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given wdate.")
#define FUNC_NAME s_chart_ymd_to_wdate
{
  /* (define-public (ymd->wdate year month day) */
  /*   (tdate->wdate (ymd->tdate year month day))) */

  return chart_tdate_to_wdate (chart_ymd_to_tdate (year, month, day));
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* mdates */

SCM_DEFINE_PUBLIC (chart_mdate_to_year, "mdate->year", 1, 0, 0,
            (SCM mdate),
	    "Return the year (a value like 2007) containing @var{mdate}.")
#define FUNC_NAME s_chart_mdate_to_year
{
  /* (define-public (mdate->year m) */
  /*   (+ 1970 (quotient-floor m 12))) */

  return scm_sum (scm_from_int (1970),
                  chart_quotient_floor (mdate, scm_from_int (12)));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_year_to_mdate, "year->mdate", 1, 0, 0,
            (SCM year),
	    "Return the mdate for January of @var{year} (a value like 2007).")
#define FUNC_NAME s_chart_year_to_mdate
{
  /* (define-public (year->mdate y) */
  /*   (* 12 (- y 1970))) */

  return scm_product (scm_from_int (12),
                      scm_difference (year, scm_from_int (1970)));
}
#undef FUNC_NAME


/* return a month 1 to 12 */
static SCM
chart_mdate_to_month (SCM mdate)
{
  return scm_oneplus (scm_modulo (mdate, scm_from_int (12)));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_mdate_to_ymd, "mdate->ymd", 1, 0, 0,
            (SCM mdate),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given @var{mdate}.")
#define FUNC_NAME s_chart_mdate_to_ymd
{
  return scm_list_3 (chart_mdate_to_year (mdate),
                     chart_mdate_to_month (mdate),
                     scm_from_int (1));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_ymd_to_mdate, "ymd->mdate", 3, 0, 0,
            (SCM year, SCM month, SCM day),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given mdate.")
#define FUNC_NAME s_chart_ymd_to_mdate
{
  long c_year  = scm_to_long (year);
  long c_month = scm_to_long (month);
  long c_day   = scm_to_long (day);
  long mdate;

  /* no support for DAY outside normal 1 to 30, don't think that's needed */
  if (c_day < 1 || c_day > 31)
    SCM_OUT_OF_RANGE (SCM_ARG3, day);

  mdate = 12 * (c_year - 1970) + c_month-1;
  return scm_from_long (mdate);
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_mdate_to_tdate, "mdate->tdate", 1, 0, 0,
                   (SCM mdate),
                 "Return the tdate which is the first weekday in @var{mdate}.")
#define FUNC_NAME s_chart_mdate_to_tdate
{
  /* (define (mdate->tdate mdate) */
  /*   (apply ymd->tdate-ceil (mdate->ymd mdate))) */

  return chart_ymd_to_tdate_ceil (chart_mdate_to_year (mdate),
                                  chart_mdate_to_month (mdate),
                                  scm_from_int (1));
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* qdates */

SCM_DEFINE_PUBLIC (chart_qdate_to_year, "qdate->year", 1, 0, 0,
                   (SCM qdate),
                   "Return the year containing @var{qdate}.")
#define FUNC_NAME s_chart_qdate_to_year
{
  /* (define-public (qdate->year q) */
  /*   (+ 1970 (quotient-floor q 4))) */

  return scm_sum (scm_from_int (1970),
                  chart_quotient_floor (qdate, scm_from_int (4)));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_year_to_qdate, "year->qdate", 1, 0, 0,
                   (SCM year),
      "Return the qdate for January of @var{year} (a year value like 2007).")
#define FUNC_NAME s_chart_year_to_qdate
{
  /* (define-public (year->qdate y) */
  /*   (* 4 (- y 1970))) */

  return scm_product (scm_from_int (4),
                      scm_difference (year, scm_from_int (1970)));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_qdate_to_month, "qdate->month", 1, 0, 0,
                   (SCM qdate),
                   "Return the month containing @var{qdate}.")
#define FUNC_NAME s_chart_qdate_to_month
{
  /* (1+ (* 3 (modulo q 4))) */
  SCM quart = scm_modulo (qdate, scm_from_int (4));
  int c_quart = scm_to_int (quart);
  return scm_from_int (c_quart * 3 + 1);
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_qdate_to_ymd, "qdate->ymd", 1, 0, 0,
            (SCM qdate),
	    "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
	    "given @var{qdate}.")
#define FUNC_NAME s_chart_qdate_to_ymd
{
  return scm_list_3 (chart_qdate_to_year (qdate),
                     chart_qdate_to_month (qdate),
                     scm_from_int (1));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_ymd_to_qdate, "ymd->qdate", 3, 0, 0,
                   (SCM year, SCM month, SCM day),
                   "Return a list @code{(@var{year} @var{month} @var{day})} for the\n"
                   "given qdate.")
#define FUNC_NAME s_chart_ymd_to_qdate
{
  /* (define-public (ymd->qdate year month day) */
  /*   (+ (* 4 (- year 1970)) */
  /*      (quotient-floor (1- month) 3))) */

  long c_year  = scm_to_long (year);
  long c_month = scm_to_long (month);
  long c_day   = scm_to_long (day);

  /* no support for DAY outside normal 1 to 30, don't think that's needed */
  if (c_day < 1 || c_day > 31)
    SCM_OUT_OF_RANGE (SCM_ARG3, day);

  return scm_from_long (4 * (c_year - 1970)
                        + QUOTIENT_FLOOR (c_month - 1, 3));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_qdate_to_tdate, "qdate->tdate", 1, 0, 0,
                   (SCM qdate),
                   "Return the tdate which is the first in @var{qdate}.")
#define FUNC_NAME s_chart_qdate_to_tdate
{
  /* (define-public (qdate->tdate q) */
  /*   (mdate->tdate (qdate->mdate q))) */

  return chart_ymd_to_tdate (chart_qdate_to_year (qdate),
                             chart_qdate_to_month (qdate),
                             scm_from_int (1));
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_qdate_to_mdate, "qdate->mdate", 1, 0, 0,
                   (SCM qdate),
                   "Return the mdate which is the start of @var{qdate}.")
#define FUNC_NAME s_chart_qdate_to_mdate
{
  /* (define-public (qdate->mdate q) */
  /*   (* 3 q)) */

  return scm_product (scm_from_int (3), qdate);
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_mdate_to_qdate, "mdate->qdate", 1, 0, 0,
                   (SCM mdate),
                   "Return the qdate containing @var{mdate}.")
#define FUNC_NAME s_chart_mdate_to_qdate
{
/* (define-public (mdate->qdate m) */
/*   (quotient-floor m 3)) */

  return chart_quotient_floor (mdate, scm_from_int (3));
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/
/* years */

SCM_DEFINE_PUBLIC (chart_year_to_ymd, "year->ymd", 1, 0, 0,
            (SCM year),
	    "Return a list @code{(@var{year} 1 1)}, being the start of the\n"
	    "given @var{year}.  This function only exists for use in\n"
	    "@code{timebase-new}.")
#define FUNC_NAME s_chart_year_to_ymd
{
  /* (define-public (year->ymd year) */
  /*   (list year 1 1)) */

  return scm_list_3 (year, scm_from_int (1), scm_from_int (1));
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_year_to_tdate, "year->tdate", 1, 0, 0,
            (SCM year),
	    "Return the first tdate in @var{year} (a year value like 2007).")
#define FUNC_NAME s_chart_year_to_tdate
{
  /* (define-public (year->tdate year) */
  /*   (ymd->tdate-ceil year 1 1)) */

  return chart_ymd_to_tdate_ceil (year, scm_from_int (1), scm_from_int (1));
}
#undef FUNC_NAME



/*---------------------------------------------------------------------------*/

void
chart_misc_init (void)
{
#ifndef SCM_MAGIC_SNARFER
#include "misc.x"
#endif

  chart_gettext_init ();
  chart_iconv_init ();
}
