;; Miscellaneous functions.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart misc)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:autoload   (ice-9 pretty-print) (pretty-print)
  #:use-module (ice-9 q)
  #:use-module (ice-9 regex)
  #:autoload   (ice-9 rw)           (read-string!/partial)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-10)
  #:use-module (srfi srfi-13)
  #:use-module (chart config)
  #:autoload   (chart database)     (database-symbols
				     database-call-with-output-file
				     database-call-with-lock
				     alert-interest-read
				     historical-read
				     preferences-load)
  #:use-module (chart extra-glib)
  #:export-syntax (let-toplevel let-toplevel*))

(load-extension (g-build-filename chart-config-libchartdir "misc.la")
		"chart_misc_init")


(define-macro (receive-list vars expr . body)
  `(apply (lambda ,vars ,@body) ,expr))
(export receive-list)


;; various properties attached to objects
;;
(define-public c-prop (make-object-property))


;;-----------------------------------------------------------------------------
;; application

(define-public chart-version chart-config-version)


;;-----------------------------------------------------------------------------
;; debugging etc

;; Print the arguments.
(define-public (d . args)
  (for-each (lambda (elem)
	      (if (pair? elem)
		  (pretty-print-no-newline elem)
		  (write elem))
	      (display " "))
	    args)
  (display "\n"))

;; Usage: (dv expr ...)
;; Print names and values of expressions, for diagnostic purposes.
;; Literal strings are printed as prefixes to the following argument.
(define-macro (dv . args)
  (if (null? args)
      #f
      (let ((exprs (map (lambda (elem)
			  (if (string? elem)
			      `(,display (,string-append ,elem " "))
			      `(,d ',elem ,elem)))
			args)))
	(if (length>=? exprs 2)
	    (cons 'begin exprs)
	    (first exprs)))))
(export dv)

(define (pretty-print-no-newline obj)
  (display (string-trim-right (pretty-print->string obj))))

;; like slib, but no width argument
(define (pretty-print->string obj)
  (call-with-output-string
   (lambda (port)
     (pretty-print obj port))))

;; Call (PROC frame) for each frame in the STACK
;; The most recent (deepest) frame is the first call
(define-public (stack-for-each proc stack)
  (do ((frame (stack-ref stack 0)
	      (frame-previous frame)))
      ((not frame))
    (proc frame)))

;; `procedure-name' or synthesized from source code position for anonymous
(define-public (c-procedure-name proc)
  (or (procedure-name proc)
      (and-let* ((source (procedure-source proc)))
	(any (lambda (elem)
	       (and (pair? elem)
		    (let ((props (source-properties elem)))
		      (and (assq-ref props 'filename)
			   (string->symbol
			    (format #f "[~a:~a:~a]"
				    (assq-ref props 'filename)
				    (assq-ref props 'line)
				    (assq-ref props 'column)))))))
	     source))))

(define-public (c-display-frame frame port)
  (if (frame-procedure? frame)
      (let* ((source    (frame-source frame))
	     (proc      (frame-procedure frame))
	     (proc-name (and (procedure? proc)
			     (c-procedure-name proc))))
	(if source
	    (format port "~a:~a:~a: "
		    (source-property source 'filename)
		    (source-property source 'line)
		    (source-property source 'column)))
	(display proc-name port)
	;; (display (cons proc-name (frame-arguments frame)) port)
	(newline port))))

(define-public (c-display-backtrace stack port)
  (display (_ "Stack backtrace (innermost first):\n") port)
  (display-backtrace stack port)

  ;; display-backtrace is usually ok
  ;;   (if (= 0 (stack-length stack))
  ;;       (display (_ "  empty\n") port)
  ;;       (stack-for-each (lambda (frame)
  ;; 			(c-display-frame frame port))
  ;; 		      stack))
  (display (_ "End backtrace\n") port))


;;-----------------------------------------------------------------------------
;; let-toplevel

;; let-toplevel accepts define-toplevel forms in its body, to establish
;; definitions outside the scope of the let.
;;
;; This is useful at the top level for establishing variables to be shared
;; by a few functions, but not otherwise visible.
;;
;; For example,
;;
;;     (let-toplevel ((x something))
;;       (define-toplevel (foo ...) (... use x))
;;       (define-toplevel (bar ...) (... use x)))
;;
;; This is simply a shorthand for
;;
;;   (define foo #f)
;;   (define bar #f)
;;   (let ((x something))
;;     (set! foo (lambda (...) (... use x)))
;;     (set! bar (lambda (...) (... use x))))
;;
;; Further let or let* forms within let-toplevel are searched for
;; define-toplevel too, and again they're emited outside the let-toplevel.
;;
;; Ordinary defines (ie. local defines) can be used at the start of
;; let-toplevel.  define-toplevel forms can appear anywhere within the body.
;;
;; In a module, define-public can appear anywhere in the body too, and those
;; defines will be exported.

(define-public let-toplevel
  (procedure->memoizing-macro
   (lambda (expr env)
     (set! expr (copy-tree expr))
     (let ((let-symbol (assq-ref '((let-toplevel  . let)
				   (let-toplevel* . let*)) (car expr))))

       ;; copy source properties from SRC to DST (if any), and return DST
       ;; (`set-source-properties!' didn't work for this in guile 1.6.4, but
       ;; it's only for debugging, so no need to compensate)
       (define (source-properties-from src dst)
	 (set-source-properties! dst (source-properties src))
	 dst)

       ;; EXPR is (define-toplevel ...)
       ;; return a list (TYPE name #f) where name is the defined symbol
       (define (top-def type expr)
	 (source-properties-from expr
				 (list type
				       (if (pair? (cadr expr))
					   (caadr expr)
					   (cadr expr))
				       #f)))

       ;; EXPR is (define-toplevel ...)
       ;; Return a list like
       ;;    (begin
       ;;      (set! foo value)
       ;;      (if (procedure? foo)
       ;;        (set-procedure-property! foo 'name 'foo)))
       ;; where "value" will be a lambda for (define-toplevel (foo ...) ...),
       ;; or just the value given for (define-toplevel foo ...).
       ;;
       (define (top-set expr lambda-symbol)
	 (let ((name       (cadr expr))
	       (value-list (cddr expr)))
	   (if (pair? name)
	       (begin
		 (set! value-list
		       (list
			(source-properties-from
			 expr
			 (cons* lambda-symbol
				(if (pair? (cdr name))
				    (source-properties-from name (cdr name))
				    (cdr name))
				value-list))))
		 (set! name
		       (car name))))
	   (source-properties-from
	    expr
	    (list 'begin
		  (source-properties-from
		   expr
		   (cons* 'set!
			  name
			  value-list))
		  (source-properties-from
		   expr
		   (list 'if
			 (source-properties-from
			  expr
			  (list 'procedure? name))
			 (source-properties-from
			  expr
			  (list 'set-procedure-property!
				name
				(list 'quote 'name)
				(list 'quote name)))))))))

       (let ((defs '()))
	 (define (transform body)
	   (map (lambda (expr)
		  (cond ((eq? (car expr) 'define-toplevel)
			 (set! defs (cons (top-def 'define expr) defs))
			 (top-set expr 'lambda))
			((eq? (car expr) 'define*-toplevel)
			 (set! defs (cons (top-def 'define expr) defs))
			 (top-set expr 'lambda*))
			((eq? (car expr) 'define-public)
			 (set! defs (cons (top-def 'define-public expr) defs))
			 (top-set expr 'lambda))
			((eq? (car expr) 'define*-public)
			 (set! defs (cons (top-def 'define-public expr) defs))
			 (top-set expr 'lambda*))
			((memq (car expr) '(let let*))
			 (cons* (car expr) (cadr expr) (transform (cddr expr))))
			(else
			 expr)))
		body))

	 (let ((trans (transform (cddr expr))))
	   (source-properties-from
	    expr
	    (cons 'begin
		  (append! defs
			   (list (cons* let-symbol
					(cadr expr)
					trans)))))))))))

     
(define-public let-toplevel* let-toplevel)


;;-----------------------------------------------------------------------------
;; procedures with caching

;; return a procedure which returns VAL
(define-public (make-procedure-returning val)
  (lambda () val))

;; make-procedure-with-cache-0 returns a procedure which calls (FUNC) the
;; first time, then caches the result and just returns that.
;; The closures are arranged so FUNC can be garbage collected once run.
;;
(define-public (make-procedure-with-cache-0 func)
  (or (thunk? func)
      (error "make-procedure-with-cache-0: bad function:" func))
  (set! func (let ((uncached-func func))
	       (lambda ()
		 (set! func (make-procedure-returning (uncached-func)))
		 (func))))
  (lambda ()
    (func)))


;;-----------------------------------------------------------------------------
;; port stuff

;; return string read from PORT up to eof
;; read-string!/partial should make this quite efficient for big input
;;
;; http-read-to-eof is similar, but has timeouts and callbacks not wanted here
;;
(define-public (read-string-to-eof port)
  (let more ((lst '()))
    (let* ((buf   (make-string 4096))
	   (nread (read-string!/partial buf port)))
      (cond ((not nread)
	     ;; eof, join parts
	     (string-concatenate-reverse/shared lst))

	    ((= 0 nread)
	     ;; allow 0 meaning EWOULDBLOCK
	     ;; if 0 is in fact from a signal then this select does no harm
	     (c-input-wait port '(read))
	     (more lst))

	    (else
	     ;; got some, allow for less than requested, no copy if full BUF
	     (more (cons (substring/shared buf 0 nread) lst)))))))

;; `write' LST, with one element per line
(define-public (write-list-lines lst port)
  (display "(" port)
  (let ((padding (make-string (port-column port) #\space)))
    (newline port)
    (for-each (lambda (elem)
		(display padding port)
		(write elem port)
		(newline port))
	      lst)
    (display (string-drop padding 1) port)
    (display ")\n" port)))


;;-----------------------------------------------------------------------------
;; internationalization stuff

(setlocale LC_ALL "")
(textdomain "chart")
(bindtextdomain "chart" chart-config-localedir)
(bind-textdomain-codeset "chart" "UTF-8")  ;; utf8 from all gettext

(define (_-strip-context str)
  (if (string-prefix? "CONTEXT(" str)
      (substring str (+ 2 (string-index str #\:)))
      str))

;; peek at gtk 2.0 messages too
(define-public (_ str)
  (let ((trans (gettext str)))
    (if (string=? trans str)
	(gettext (_-strip-context str) "gtk20")
	(_-strip-context trans))))

;; return a utf8 strftime format for date and weekday
;; the default is %a with the `locale-d-fmt'
(define-public locale-wd-fmt
  (make-procedure-with-cache-0
   (lambda ()
     (simple-format #f (_ "%a ~a") (locale-d-fmt)))))

;; `locale-number+decimals' returns the number N converted to a utf8 string,
;; using locale-specific formatting.  DECIMALS is a power of 10 factor to
;; apply, so for instance with N=100 and DECIMALS=2 the return would be
;; "1.00" (with the `locale-decimal-point').
;;
;; The `locale-thousands-sep' is used, but with fixed 3-digit groups.
;; There's nothing for the locale "grouping" parameter because it's a bit
;; hard, and rarely needed.  (Only some locales for India, maybe.)
;;
;; There's nothing for the locale "negative_sign", it's just ascii "-"
;; (0x2D) in all GNU locales.
;;
(define-public (locale-number+decimals n decimals)
  ;; n is supposed to be an exact integer, but allow for a flonum just to
  ;; print something
  (if (not (exact? n))
      (begin
	(if (not (integer? n))
	    (let ((extra-decimals (string-decimals (number->string n))))
	      (set! n (round (* n (integer-expt 10 extra-decimals))))
	      (set! decimals (+ decimals extra-decimals))))
	(set! n (inexact->exact n))))

  (let* ((power (integer-expt 10 decimals))
	 (int   (quotient (abs n) power))
	 (frac  (remainder (abs n) power)))
    (string-append (if (negative? n)
		       "-"
		       "")
		   (thousands-insert (number->string int))
		   (if (= 0 decimals)
		       ""
		       (string-append (locale-decimal-point)
				      (string-pad (number->string frac)
						  decimals #\0))))))

;; return STR with locale-thousands-sep inserted every three characters
(define (thousands-insert str)
  (if (string-null? (locale-thousands-sep))
      str
      (string-join (thousands-split str) (locale-thousands-sep))))

(define (thousands-split str)
  (do ((lst '() (cons (substring str (- pos 3) pos) lst))
       (pos (string-length str) (- pos 3)))
      ((< pos 3)
       (if (= 0 pos)
	   lst
	   (cons (substring str 0 pos) lst)))))

;; `strftime' but with FMT as a utf8 string, and returning a utf8 string
(define-public (strftime-utf8 fmt tm)
  (locale->utf8 (strftime (utf8->locale fmt) tm)))


;;-----------------------------------------------------------------------------
;; regex stuff
;;
;; note the locale test bits must be after setlocale

;; call (THUNK) with the locale set to "C", then put it back to default
;; changing the locale like this seems reasonably fast on a GNU system
;;
(define-public (call-with-C-locale thunk)
  (dynamic-wind
      (lambda () (setlocale LC_ALL "C"))
      thunk
      (lambda () (setlocale LC_ALL ""))))

;; The following are variants of the normal guile regexp funcs, but forced
;; to operate in a unibyte locale.  If the current locale is multibyte, then
;; `call-with-C-locale' temporarily switches to unibyte.
;;
;; Unibyte seems the best way to handle regexps operating on strings with
;; encoding different from the current locale, and indeed with an encoding
;; that might not even exist in any locale available on the system.
;;
(let-toplevel ()
  ;; return true if the current locale is a unibyte charset
  (define (locale-unibyte?)
    (= 1 (mb-cur-max)))

  (define (make-unibyte-proc proc)
    (if (locale-unibyte?)
	proc

	(begin
	  (or (call-with-C-locale locale-unibyte?)
	      (error "Oops, \"C\" locale is multibyte"))

	  (lambda args
	    (call-with-C-locale
	     (lambda ()
	       (apply proc args)))))))

  (define-public make-regexp-C    (make-unibyte-proc make-regexp))
  (define-public string-match-C   (make-unibyte-proc string-match))
  (define-public regexp-exec-C    (make-unibyte-proc regexp-exec))
  (define-public list-matches-C   (make-unibyte-proc list-matches))

  ;; the fold PROC would be called with the unibyte locale in force
  ;; (define-public fold-matches-C   (make-unibyte-proc fold-matches))
  )

;; read syntax #,(regexp "foo" [flag...]) with flags per make-regexp
(define-reader-ctor 'regexp
  (lambda (str . args)
    (apply make-regexp-C str (map (lambda (arg)
				    (eval arg (current-module)))
				  args))))


;;-----------------------------------------------------------------------------
;; chart directories

;; eg. "/home/username/Chart"
(define-public user-directory
  (g-build-filename (or (g-get-home-dir)
			(error "chart: cannot find HOME directory"))
		    "Chart"))

;; eg. "/home/username/Chart"
(define-public database-directory
  (g-build-filename (g-get-home-dir) "Chart"))

;; eg. "/home/username/Chart/cache"
(define-public cache-directory
  (g-build-filename database-directory "cache"))


;;-----------------------------------------------------------------------------
;; numeric stuff

;; `quotient', but rounding upwards towards +infinity
;; note only positive D supported (that's all that's currently required)
(define-public (quotient-ceiling n d)
  (if (negative? n)
      (quotient n d)
      (quotient (+ n d -1) d)))

;; compare pairs of numbers xh:xl yh:yl (the "h"s being the high parts,
;; ie. most significant)
(define-public (two< xh xl yh yl)
  (or (< xh yh)
      (and (= xh yh)
	   (< xl yl))))

;; compare pairs of numbers xh:xl yh:yl (the "h"s being the high parts,
;; ie. most significant)
(define-public (two<= xh xl yh yl)
  (not (two< yh yl xh xl)))

(define-public (within-range? lo hi x)
  (and (>= x lo)
       (<= x hi)))

;; Return N rounded downwards to a multiple of D.
;; D can be an integer or a flonum
(define-public (round-down-multiple n d)
  (if (and (exact? d) (integer? d))
      (begin
	(set! n (floor-exact n))
	(- n (remainder n d)))
      (* d (floor (/ n d)))))

;; compare the cars of X and Y with <
(define-public (car< x y)
  (< (car x) (car y)))

;; `integer-expt' supporting negative exponents
(define-public integer-expt-pn
  (if (false-if-exception (integer-expt 2 -1))
      integer-expt  ;; guile 1.8 supports negative exponents already
      (lambda (base expon)
	(if (negative? expon)
	    (/ (integer-expt base (- expon)))
	    (integer-expt base expon)))))


;;-----------------------------------------------------------------------------
;; list stuff

(define-public (first-or-f lst)
  (and (not (null? lst))
       (first lst)))

;; return list of values LO to HI inclusive, incrementing by STEP
(define*-public (iota-range lo hi #:optional (step 1))
  (iota (floor-exact (/ (+ (- hi lo) step) step)) lo step))

;; return the tail of LST after ELEM, or empty list () if ELEM not present
(define-public (tail-after elem lst)
  (let ((pos (member elem lst)))
    (if pos
	(cdr pos)
	'())))

;; Like for-each, but take two elements at a time from LST, and call PROC
;; with those two as parameters.
;;
;; If LST has an odd number of elements, the last call to PROC will be with
;; one argument.  It's left up to PROC whether this is acceptable or an
;; error.
;;
(define-public (for-each-two proc lst)
  (if (not (null? lst))
      (if (null? (cdr lst))
	  (proc (car lst))
	  (begin
	    (proc (car lst) (cadr lst))
	    (for-each-two proc (cddr lst))))))

(define-public  set-first!             set-car!)
(define-public (set-second!  lst val)  (list-set! lst 1 val))
(define-public (set-third!   lst val)  (list-set! lst 2 val))
(define-public (set-fourth!  lst val)  (list-set! lst 3 val))
(define-public (set-fifth!   lst val)  (list-set! lst 4 val))
(define-public (set-sixth!   lst val)  (list-set! lst 5 val))

;; Call (PROC elem) for each element in LST.
;; Return the ELEM for which PROC gives the biggest number.
;; PROC can return #f, such elements are ignored.
;;
(define-public (find-maximum proc lst)
  (let ((result #f)
	(val    #f))
    (for-each (lambda (elem)
		(let ((v (proc elem)))
		  (if (and v
			   (or (not val)
			       (> v val)))
		      (begin
			(set! result elem)
			(set! val v)))))
	      lst)
    result))

;; return true if LST has at least N proper elements
(define-public (list>=? lst n)
  (or (<= n 0)
      (and (pair? lst)
	   (list>=? (cdr lst) (1- n)))))

;; `take' but limiting N to the LST length
(define-public (take-maybe lst n)
  (if (length>=? lst n)
      (take lst n)
      lst))

;; `drop' but limiting N to the LST length
(define-public (drop-maybe lst n)
  (if (length>=? lst n)
      (drop lst n)
      '()))

;; return a sublist of LST, from START (inclusive) to END (exclusive)
(define-public (sublist lst start end)
  (take (drop lst start) (- end start)))

;; return the position number where ELEM occurs in LST, or #f if doesn't
(define*-public (list-index-elem elem lst #:optional (= equal?))
  (let more ((n   0)
	     (lst lst))
    (cond ((null? lst)
	   #f)
	  ((= (first lst) elem)
	   n)
	  (else
	   (more (1+ n) (cdr lst))))))	 


;;-----------------------------------------------------------------------------
;; record stuff

(if (defined? 'record-struct-copy)
    (define-public (record-copy rec)
      (or (record? rec)
	  (error 'not-a-record-type rec))
      (record-struct-copy rec))

    (define-public (record-copy rec)
      (let ((type (record-type-descriptor rec)))
	(apply (record-constructor type)
	       (map (lambda (field)
		      ((record-accessor type field) rec))
		    (record-type-fields type))))))

;; set fields of REC to VALUE-LIST, skipping those in VALUE-LIST which are #f
(define-public (record-set-all-not-f! rec . value-list)
  (let ((type (record-type-descriptor rec)))
    (for-each (lambda (field value)
		(if value
		    ((rec-modifier type field) rec value)))
	      (record-type-fields type) value-list)))


;;-----------------------------------------------------------------------------
;; record type checking accessor

;; same as `record-accessor' but with an error if wrong record type
(define-public rec-accessor record-accessor)

;; same as `record-modifier' but with an error if wrong record type
(define-public rec-modifier record-modifier)

;; in guile 1.8.1 and later record-accessor and record-modifier have strict
;; type checks, in earlier versions set this up ourselves
;;
(if (let* ((type-A (make-record-type "type-A" '(field-A)))
	   (type-B (make-record-type "type-B" '(field-B)))
	   (new-A  (record-constructor type-A))
	   (get-B  (record-accessor type-B 'field-B))
	   (A      (new-A 123)))
      (catch 'wrong-type-arg
	(lambda ()
	  (get-B A)
	  #t)   ;; succeeds, replace with type checking versions
	(lambda args
	  #f))) ;; fails, record-accessor already checks

    (let ()
      ;; throw an error if OBJ is not of the given RTD record type
      (define (rec-demand-type obj rtd)
	(or (eq? rtd (record-type-descriptor obj))
	    (scm-error 'wrong-type-arg "rec-demand-check"
		       "Wrong type record (want `~S'): ~S"
		       (list (record-type-name rtd) obj)
		       #f)))

      (set! rec-accessor
	    (lambda (rtd field-name)
	      (let ((accessor (record-accessor rtd field-name)))
		(lambda (obj)
		  (rec-demand-type obj rtd)
		  (accessor obj)))))

      (set! rec-modifier
	    (lambda (rtd field-name)
	      (let ((modifier (record-modifier rtd field-name)))
		(lambda (obj val)
		  (rec-demand-type obj rtd)
		  (modifier obj val)))))))


;;-----------------------------------------------------------------------------
;; file stuff

;; Make a directory, and all components leading to it.
;; An error is thrown for any problem.
(define-public (mkdir-all dir)
  (let ((split (string-split dir #\/))
	(sofar "/"))
    (for-each
     (lambda (elem)
       (set! sofar (string-append sofar "/" elem))
       (if (not (access? sofar F_OK))
	   (mkdir sofar)))
     split)))

;; ignoring errors
(define-public (delete-file-maybe path)
  (false-if-exception (delete-file path)))

;; open FILENAME and call (PROC port), or return NOTFOUND if cannot open
(define-public (call-with-input-file-or notfound filename proc)
  (let ((port (false-if-exception (open-input-file filename))))
    (if port
	(let ((result (proc port)))
	  (close-input-port port)
	  result)
	notfound)))

;; return list of filenames in DIR, or empty if DIR doesn't exist at all
;; names are in no particular order
;; FULL can be #t to get full path names
;; PRED is called (PRED name) to select desired names
;;
(define*-public (directory-files dir #:optional full (pred noop))
  (let ((stream (catch 'system-error
		  (lambda () (opendir dir))
		  (lambda args #f))))
    (if (not stream)
	'()
	(let more ((lst '()))
	  (let ((name (readdir stream)))
	    (if (eof-object? name)
		(begin
		  (closedir stream)
		  lst)
		(begin
		  (if full
		      (set! name (g-build-filename dir name)))
		  (if (pred name)
		      (more (cons name lst))
		      (more lst)))))))))


;;-----------------------------------------------------------------------------
;; after load

;; as per `resolve-module' in boot-9.scm
;; an error can be thrown if the module exists but there's bad stuff in it
(define-public (module-loaded? spec)
  (let ((full-name (append '(%app modules) spec)))
    (nested-ref the-root-module full-name)))

(define after-load-hook (make-hook 1))
(let ((old-primitive-load primitive-load))
  (define (primitive-load-with-after-load-hook filename)
    (old-primitive-load filename)
    (run-hook after-load-hook filename))
  (set! primitive-load primitive-load-with-after-load-hook))

(define-public (after-module-load module thunk)
  (if (module-loaded? module)
      (thunk)
      (letrec ((target (string-append (string-join (map symbol->string
							module)
						   "/" 'prefix)
				      ".scm"))
	       (proc   (lambda (filename)
			 (if (string-suffix? target filename)
			     (begin
			       (remove-hook! after-load-hook proc)
			       (thunk))))))
	(add-hook! after-load-hook proc))))


;;-----------------------------------------------------------------------------
;; ice-9 q stuff

;; return the list of data values from an ice-9 q
(define-public q-list car)


;;-----------------------------------------------------------------------------
;; atexit

;; call (PROC) when exiting
(define-public (c-atexit proc)
  (let ((old-exit exit))
    (set! exit (lambda args
		 (proc)
		 (apply old-exit args)))))


;;-----------------------------------------------------------------------------
;; time stuff

(define-public (hms->seconds hour minute second)
  (+ (* 3600 hour)
     (* 60 minute)
     second))

;; read syntax #,(hms->seconds H M S)
(define-reader-ctor 'hms->seconds hms->seconds)

;; return the time in seconds since midnight in TM
(define-public (tm->time tm)
  (hms->seconds (tm:hour tm) (tm:min tm) (tm:sec tm)))


;;-----------------------------------------------------------------------------
;; date stuff

;; general policy for two digit years, use 1950 to 2049
(define-public (yy->year y)
  (+ y (if (>= y 50) 1900 2000)))

;; single digit year to four digits, according to nearest, ie. in the range
;; thisyear-5 to thisyear+4, inclusive
;;
(define (y->year y)
  (let* ((base-year (- (today-year) 5)))
    (+ base-year (modulo (- y base-year) 10))))


;;-----------------------------------------------------------------------------
;; date parsing stuff

;; return month 1 to 12 for STR like "Jan" or "January" etc, or #f
;; case is ignored, so "JAN" etc matches too
(define-public (Mmm-str->month str)
  (and-let* ((pair (assoc str '(("Jan" . 1)
				("Feb" . 2)
				("Mar" . 3)
				("Apr" . 4)
				("May" . 5)
				("Jun" . 6)
				("Jul" . 7)
				("Aug" . 8)
				("Sep" . 9)
				("Oct" . 10)
				("Nov" . 11)
				("Dec" . 12)
				("January"   . 1)
				("February"  . 2)
				("March"     . 3)
				("April"     . 4)
				("May"       . 5)
				("June"      . 6)
				("July"      . 7)
				("August"    . 8)
				("September" . 9)
				("October"   . 10)
				("November"  . 11)
				("December"  . 12)) string-ci=?)))
    (cdr pair)))


;;-----------------------------------------------------------------------------
;; tm and date stuff

;; return a new tm record, set to midnight 1 Jan 1970
(define-public (tm-new)
  (gmtime 0))

;; return year the current year, like 2003, in the given ZONE (default local
;; time)
(define-public (today-year . zone)
  (let ((tm (apply localtime (current-time) zone)))
    (+ (tm:year tm) 1900)))

;; return list (YEAR MONTH DAY) like (2003 12 31), in the given ZONE
;; (default local time)
(define-public (today-ymd . zone)
  (tm->ymd (apply localtime (current-time) zone)))

;; return list (YEAR MONTH DAY) like (2003 12 31) from TM object
(define-public (tm->ymd tm)
  (list (+ 1900 (tm:year tm))
	(1+ (tm:mon tm))
	(tm:mday tm)))

;; return true if TM is a weekday
(define-public (tm:weekday? tm)
  (within-range? 1 5 (tm:wday tm)))

;; return true if TM is the weekend
(define-public (tm:weekend? tm)
  (memv (tm:wday tm) '(0 6)))

(define-public (set-tm:ymd tm year month day)
  (set-tm:year tm (- year 1900))
  (set-tm:mon  tm (1- month))
  (set-tm:mday tm day))

(define-public (ymd->tm year month day)
  (let ((tm (tm-new)))
    (set-tm:ymd tm year month day)
    tm))

(define-public (set-tm:hms tm hour minute second)
  (set-tm:hour tm hour)
  (set-tm:min  tm minute)
  (set-tm:sec  tm second))


;;-----------------------------------------------------------------------------
;; timezones

;; Return true if timezone name ZONE is known.
;;
;; In glibc an unknown zone is silently treated as GMT, so our test is for
;; ZONE different from GMT, and tested in 12 different months to spot
;; daylight savings somewhere in the year.  "GMT" and "UTC" are recognised
;; explicitly.
;;
(define-public (timezone-known? name)
  (or (member name '("UTC" "GMT"))
      (any (lambda (delta)
	     (let* ((t  (+ (current-time) delta))
		    (tm (false-if-exception (localtime t name)))
		    (gm (gmtime t)))
	       (and tm
		    (or (not (= (tm:hour tm) (tm:hour gm)))
			(not (= (tm:min  tm) (tm:min  gm)))))))
	   (iota 12 0 (* 30 86400)))))

;; `timezone-choose' returns the first timezone from ZONE-LIST which is
;; known to the system.  The idea is to have alternatives, like glibc names,
;; POSIX format rule strings, or plain old style offset from GMT (even if
;; daylight savings will be wrong in that case).
;;
(define-public (timezone-choose name zone-list default-zone)
  (or (find timezone-known? zone-list)
      (begin
	(display (utf8->locale (format #f (_ "chart: no timezone known for ~a
       using ~a, daylight savings might be wrong\n")
				       name default-zone))
		 (current-error-port))
	default-zone)))

;; return the current local timezone
(define-public timezone-local
  (make-procedure-with-cache-0
   (lambda ()
     (let* ((tm   (localtime (current-time)))
	    (zone (tm:zone tm)))
       ;; zone is usually just a three or four letter code, and that won't
       ;; be unique, so append tm:gmtoff (in hours) to it
       (if (string-every char-alphabetic? zone)
	   (set! zone (format #f "~a~a" zone
			      (quotient (tm:gmtoff tm) 3600))))
       zone))))


;;-----------------------------------------------------------------------------
;; timezones by symbol

;; elements are lists (PRED LOCATION ZONE-STR-OR-PROC PRIORITY)
(define symbol-timezone-list '())

(define (fourth> x y)
  (> (fourth x) (fourth y)))

;; set the timezone for symbols passing PRED
;; ZONE is either a string or a procedure returning such a string
;;
;; The idea of a procedure returning a string is that we don't have to probe
;; for how to express the zone for a location until that location is
;; actually needed.  If there's a problem a warning can be given only when
;; it's needed.
;;
(define*-public (symbol-timezone! pred location zone #:optional (priority 0))
  (set! symbol-timezone-list
	(merge! symbol-timezone-list (list (list pred location zone priority))
		fourth>)))

;; return a location string for SYMBOL, or #f
(define-public (symbol-location-maybe symbol)
  (symbol-setups symbol)
  (cond ((sympred-assoc-ref symbol-timezone-list symbol)
	 => first)
	(else #f)))

;; return a location string for SYMBOL, in locale encoding
(define-public (symbol-location symbol)
  (or (symbol-location-maybe symbol)
      (_ "Unknown location")))

;; return a timezone string for SYMBOL
;;
;; For unknown SYMBOLs, the default is `timezone-local', which at least
;; has a chance of being right if it's local stocks which are mainly of
;; interest.  (This can arise from a new suffix yahoo knows but we don't.)
;;
(define-public (symbol-timezone symbol)
  (symbol-setups symbol)
  (cond ((sympred-assoc symbol symbol-timezone-list)
	 => (lambda (elem)
	      (let ((zone (third elem)))
		(if (procedure? zone)
		    (zone)
		    zone))))
	(else
	 (timezone-local))))


;;-----------------------------------------------------------------------------
;; timebase

(let-toplevel ((conv-alist '()))   ;; key (from . to)

  ;; set a conversion procedure
  (define-public (timebase-convert-proc! from-timebase to-timebase proc)
    (let ((key (cons from-timebase to-timebase)))
      (if (assoc-ref conv-alist key)
	  (error "already have timebase conversion:"
		 (timebase-name from-timebase)
		 (timebase-name to-timebase)))
      (set! conv-alist
	    (assoc-set! conv-alist key proc))))
  
  ;; get a conversion procedure
  (define-public (timebase-convert-proc from-timebase to-timebase)
    (or (assoc-ref conv-alist (cons from-timebase to-timebase))
	(error "no timebase conversion:"
	       (timebase-name from-timebase)
	       (timebase-name to-timebase)))))

(define-public (timebase-convert from-timebase to-timebase t)
  ((timebase-convert-proc from-timebase to-timebase) t))
  
(let-toplevel ((rec-type (make-record-type "timebase"
					   '(name      ;; string
					     adjective ;; string
					     ymd->t
					     t->ymd))))
  (define rec-new (record-constructor rec-type))

  (define-public timebase?            (record-predicate rec-type))
  (define-public timebase-name        (rec-accessor rec-type 'name))
  (define-public timebase-adjective   (rec-accessor rec-type 'adjective))
  (define-public timebase-ymd->t-proc (rec-accessor rec-type 'ymd->t))
  (define-public timebase-t->ymd-proc (rec-accessor rec-type 't->ymd))

  (define*-public (timebase-new #:key
				(name      "")
				(adjective "")
				(ymd->t    #f)
				(t->ymd    #f))
    (let ((timebase (rec-new name adjective ymd->t t->ymd)))
      (timebase-convert-proc! timebase timebase noop)
      timebase)))

(define-public (timebase-t->tm-proc timebase)
  (let ((t->ymd (timebase-t->ymd-proc timebase)))
    (lambda (t)
      (let ((tm (tm-new)))
	(t->ymd tm t)
	tm))))

;; return (YEAR MONTH DAY)
(define-public (timebase-t->ymd timebase t)
  ((timebase-t->ymd-proc timebase) t))

;; return a timebase value for YEAR,MONTH,DAY
(define-public (timebase-ymd->t timebase year month day)
  ((timebase-ymd->t-proc timebase) year month day))

;; return timebase value for today, in local timezone or ZONE
(define-public (timebase-today timebase . zone)
  (apply (timebase-ymd->t-proc timebase) (apply today-ymd zone)))

;; return a tm record for T
;; tdate->tm sets tm:wday etc too
(define-public (timebase-t->tm timebase t)
  (tdate->tm ((timebase-convert-proc timebase timebase-days) t)))


;;-----------------------------------------------------------------------------
;; adate is any date, ie. both weekdays and weekends, expressed as a count
;; of days starting from 0 for Mon 5 Jan 1970

;; adate to tm, midnight GMT
(define-public (adate->tm a)
  (let ((tm (tm-new)))
    (apply set-tm:ymd tm (adate->ymd a))
    (set-tm:wday tm (modulo (1+ a) 7))
    tm))

;; return adate for a TM record
(define-public (tm->adate tm)
  (ymd->adate (+ 1900 (tm:year tm))
	      (1+ (tm:mon tm))
	      (tm:mday tm)))

;; return true if ADATE is a weekend (Saturday or Sunday)
(define-public (adate-weekend? adate)
  (>= (modulo adate 7) 5))


;;-----------------------------------------------------------------------------
;; tdate is a trading date, meaning a weekday, expressed as a count of
;; weekdays since Mon 5 Jan 1970

(define-public timebase-days (timebase-new #:name      (_ "Days")
					   #:adjective (_ "Daily")
					   #:ymd->t    ymd->tdate
					   #:t->ymd    tdate->ymd))

;; tdate to tm, midnight
(define-public (tdate->tm t)
  (let ((tm  (tm-new))
	(ymd (tdate->ymd t)))
    (apply set-tm:ymd tm ymd)
    (set-tm:wday tm (1+ (modulo t 5)))
    tm))

(define-public (tdate-today . zone)
  (tm->tdate (apply localtime (current-time) zone)))

;; return tdate for a TM record
(define-public (tm->tdate tm)
  (ymd->tdate (+ 1900 (tm:year tm))
	      (1+ (tm:mon tm))
	      (tm:mday tm)))


;;-----------------------------------------------------------------------------
;; wdate is a week date, a count of weeks since Mon 5 Jan 1970


(define-public timebase-weeks (timebase-new #:name      (_ "Weeks")
					    #:adjective (_ "Weekly")
					    #:ymd->t    ymd->wdate
					    #:t->ymd    wdate->ymd))
(timebase-convert-proc! timebase-days timebase-weeks  tdate->wdate)
(timebase-convert-proc! timebase-weeks timebase-days  wdate->tdate)

;; rounded up, ie. the week beginning on or after tdate
(define-public (tdate->wdate-ceil t)
  (quotient-floor (+ t 4) 5))
  

;;-----------------------------------------------------------------------------
;; mdate is a month date, counting from 0 for Jan 1970

(define-public timebase-months (timebase-new #:name      (_ "Months")
					     #:adjective (_ "Monthly")
					     #:ymd->t    ymd->mdate
					     #:t->ymd    mdate->ymd))
(timebase-convert-proc! timebase-months timebase-days mdate->tdate)

(define-public (mdate->tm mdate)
  (apply ymd->tm (mdate->ymd mdate)))

;; the month containing T
(define-public (tdate->mdate t)
  (apply ymd->mdate (tdate->ymd t)))
(timebase-convert-proc! timebase-days timebase-months  tdate->mdate)

;; the first full week in M
(define-public (mdate->wdate m)
  (tdate->wdate-ceil (mdate->tdate m)))
(timebase-convert-proc! timebase-months timebase-weeks  mdate->wdate)

(define-public (wdate->mdate w)
  (tdate->mdate (wdate->tdate w)))
(timebase-convert-proc! timebase-weeks timebase-months  wdate->mdate)


;;-----------------------------------------------------------------------------
;; qdate is a quarter, counting from 0 for quarter Jan-Mar 1970

(define-public timebase-quarters (timebase-new #:name      (_ "Quarters")
					       #:adjective (_ "Quarterly")
					       #:ymd->t    ymd->qdate
					       #:t->ymd    qdate->ymd))

(define-public (tdate->qdate t)
  (apply ymd->qdate (tdate->ymd t)))
(timebase-convert-proc! timebase-quarters timebase-days qdate->tdate)
(timebase-convert-proc! timebase-days timebase-quarters tdate->qdate)

;; the first full week in Q
(define-public (qdate->wdate q)
  (tdate->wdate-ceil (qdate->tdate q)))
(timebase-convert-proc! timebase-quarters timebase-weeks qdate->wdate)
(define-public (wdate->qdate w)
  (tdate->qdate (wdate->tdate w)))
(timebase-convert-proc! timebase-weeks timebase-quarters wdate->qdate)

(timebase-convert-proc! timebase-months timebase-quarters mdate->qdate)
(timebase-convert-proc! timebase-quarters timebase-months qdate->mdate)



;;-----------------------------------------------------------------------------
;; years are a full value like 2005

;; doesn't handle DAY/MONTH outside normal ranges, don't think that's needed
(define-public (ymd->year year month day)
  year)

(define-public timebase-years (timebase-new #:name      (_ "Years")
					    #:adjective (_ "Yearly")
					    #:ymd->t    ymd->year
					    #:t->ymd    year->ymd))

(define-public (tdate->year t)
  (first (tdate->ymd t)))
(timebase-convert-proc! timebase-days timebase-years tdate->year)
(timebase-convert-proc! timebase-years timebase-days year->tdate)

;; first full week in Y
(define-public (year->wdate y)
  (tdate->wdate-ceil (year->tdate y)))
(define-public (wdate->year w)
  (tdate->year (wdate->tdate w)))
(timebase-convert-proc! timebase-years timebase-weeks year->wdate)
(timebase-convert-proc! timebase-weeks timebase-years wdate->year)

(timebase-convert-proc! timebase-months timebase-years mdate->year)
(timebase-convert-proc! timebase-years timebase-months year->mdate)

(timebase-convert-proc! timebase-years timebase-quarters year->qdate)
(timebase-convert-proc! timebase-quarters timebase-years qdate->year)


;;-----------------------------------------------------------------------------
;; ddate is a decade the starting year divided by 10, eg. 200 for the decade
;; 2000 to 2009

;; doesn't handle DAY/MONTH outside normal ranges, don't think that's needed
(define-public (ymd->ddate year month day)
  (quotient-floor year 10))

(define-public (ddate->ymd d)
  (list (* 10 d) 1 1))

(define-public timebase-decades (timebase-new #:name      (_ "Decades")
					      #:adjective (_ "Decadic")
					      #:ymd->t    ymd->ddate
					      #:t->ymd    ddate->ymd))

;; first tdate in the decade
(define-public (ddate->tdate d)
  (mdate->tdate (ddate->mdate d)))
(timebase-convert-proc! timebase-decades timebase-days ddate->tdate)

(define-public (tdate->ddate t)
  (quotient-floor (tdate->year t) 10))
(timebase-convert-proc! timebase-days timebase-decades tdate->ddate)


;; first full week in D
(define-public (ddate->wdate d)
  (tdate->wdate-ceil (ddate->tdate d)))
(timebase-convert-proc! timebase-decades timebase-weeks  ddate->wdate)

(define-public (wdate->ddate w)
  (tdate->ddate (wdate->tdate w)))
(timebase-convert-proc! timebase-weeks timebase-decades  wdate->ddate)


;; 120 months per decade
(define-public (ddate->mdate d)
  (* 120 (- d 197)))
(timebase-convert-proc! timebase-decades timebase-months  ddate->mdate)

(define-public (mdate->ddate m)
  (+ 197 (quotient-floor m 120)))
(timebase-convert-proc! timebase-months timebase-decades  mdate->ddate)


;; 40 quarters per decade
(define-public (ddate->qdate d)
  (* 40 (- d 197)))
(timebase-convert-proc! timebase-decades timebase-quarters  ddate->qdate)

(define-public (qdate->ddate q)
  (+ 197 (quotient-floor q 40)))
(timebase-convert-proc! timebase-quarters timebase-decades  qdate->ddate)


;; 10 years per decade
(define-public (ddate->year d)
  (* 10 d))
(timebase-convert-proc! timebase-decades timebase-years  ddate->year)

(define-public (year->ddate y)
  (quotient-floor y 10))
(timebase-convert-proc! timebase-years timebase-decades  year->ddate)


;;-----------------------------------------------------------------------------
;; datatype class

(define-public datatype-list-current '())

(let-toplevel
    ((rec-type (make-record-type "datatype"
				 '(key
				   name
				   priority      ;; number, default 0
				   units         ;; symbol or #f
				   input         ;; datatype or #f
				   module
				   reduce-proc   ;; (PROC array idxconv-proc)
				   min-proc      ;; (PROC elem)
				   max-proc      ;; (PROC elem)
				   linestyle     ;; symbol eg. 'bar
				   minimum       ;; minimum value taken
				   maximum       ;; maximum value taken
				   hlines        ;; horiz lines to draw
				   manual
				   proc
				   parameter-list
				   initial-page-proc
				   closes-apply-arg))))

  (define rec-new      (record-constructor rec-type '()))
  (define rec-module   (rec-accessor rec-type 'module))
  (define rec-module!  (rec-modifier rec-type 'module))
  (define rec-priority (rec-accessor rec-type 'priority))

  ;; return accessor func for FIELD, running setups first
  (define (setup-accessor field)
    (let ((accessor (rec-accessor rec-type field)))
      (lambda (rec)
	(if (rec-module rec)
	    (begin
	      ;; resolve-interface throws an error if cannot load
	      (resolve-interface (rec-module rec))
	      (rec-module! rec #f)))
	(accessor rec))))

  (define-public datatype?               (record-predicate rec-type))
  (define-public datatype-key         	 (rec-accessor rec-type 'key))
  (define-public datatype-name        	 (rec-accessor rec-type 'name))
  (define-public datatype-units       	 (rec-accessor rec-type 'units))
  (define-public datatype-input       	 (rec-accessor rec-type 'input))

  (define-public datatype-reduce-proc 	 (setup-accessor 'reduce-proc))
  (define-public datatype-min-proc    	 (setup-accessor 'min-proc))
  (define-public datatype-max-proc    	 (setup-accessor 'max-proc))
  (define-public datatype-linestyle   	 (setup-accessor 'linestyle))
  (define-public datatype-minimum     	 (setup-accessor 'minimum))
  (define-public datatype-maximum     	 (setup-accessor 'maximum))
  (define-public datatype-hlines      	 (setup-accessor 'hlines))
  (define-public datatype-manual      	 (setup-accessor 'manual))
  (define-public datatype-proc        	 (setup-accessor 'proc))
  (define-public datatype-parameter-list (setup-accessor 'parameter-list))
  (define-public datatype-initial-page-proc (setup-accessor 'initial-page-proc))
  (define-public datatype-closes-apply-arg (setup-accessor 'closes-apply-arg))

  (define-public (datatype-priority datatype)
    (or (rec-priority datatype) 0))

  ;; create datatype, or amend previously created
  (define*-public (datatype-new!
		   #:key
		   name
		   (key  (error "datatype-new!: no key"))
		   priority
		   units
		   input
		   module
		   reduce
		   (min        noop)
		   (max        noop)
		   (linestyle  'line)
		   minimum
		   maximum
		   (hlines     '())
		   manual
		   (proc (lambda args
			   (error "oops, no series procedure for" key name)))
		   (parameter-list '())
		   initial-page-proc
		   (closes-apply-arg '()))

    ;; #f values don't overwrite existing fields
    (let* ((old  (datatype-find key))
	   (rec  (or old (rec-new))))
      (record-set-all-not-f! rec
			     key name priority units input module reduce
			     min max linestyle minimum maximum hlines
			     manual proc parameter-list initial-page-proc 
			     closes-apply-arg)

      ;; names only from autoloads in `datatype-list-all'
      ;; (or (datatype-name rec) ;; must have a name for use by `datatype<?'
      ;;     (error "datatype-new!: no name for key" key))

      (if (not old)            ;; new record goes in list
	  (set! datatype-list-current (cons rec datatype-list-current)))

      rec)))

;; return datatype record for KEY
(define-public (datatype-find key)
  (find (lambda (rec)
	  (eq? key (datatype-key rec)))
	datatype-list-current))

(define-public (datatype-list-all)
  (datatypes-autoloads)
  datatype-list-current)

(define-public datatype-ohlcv
  (datatype-new! #:key      'ohlcv
		 #:name     (_ "OHLCV")
		 #:units    'price
		 #:priority -1000 ;; hidden
		 #:module   '(chart ohlcv)
		 #:closes-apply-arg '(3)))

(define (datatypes-autoloads)
  (set! datatypes-autoloads noop) ;; once only

  (datatype-new! #:key      'none
		 #:name     (string-append " " (_ "None") " ")
		 #:priority 100)

  (load-from-path "chart/datatype-setups.scm"))


;;-----------------------------------------------------------------------------
;; datatype parameters

(let-toplevel ((rec-type (make-record-type "datatype-param"
					   '(key name type decimals
						 minimum maximum
						 default step))))
  (define rec-new (record-constructor rec-type))
  (define-public datatype-param-key      (rec-accessor rec-type 'key))
  (define-public datatype-param-name     (rec-accessor rec-type 'name))
  (define-public datatype-param-type     (rec-accessor rec-type 'type))
  (define-public datatype-param-decimals (rec-accessor rec-type 'decimals))
  (define-public datatype-param-minimum  (rec-accessor rec-type 'minimum))
  (define-public datatype-param-maximum  (rec-accessor rec-type 'maximum))
  (define-public datatype-param-default  (rec-accessor rec-type 'default))
  (define-public datatype-param-step     (rec-accessor rec-type 'step))

  (define*-public (datatype-param-new
		   #:key name
		   (key
		    (error "datatype-param-new: no key for" name))
		   (type 'integer)
		   (decimals 0)
		   (minimum 1)
		   (maximum 9999)
		   (step 1)
		   (default
		     (error "datatype-param-new: no default for" name)))
    (rec-new key name type decimals minimum maximum default step)))


;;-----------------------------------------------------------------------------
;; series

(define (alist? obj)
  (and (list? obj)
       (every pair? obj)))

(define (alist-print lst port)
  (let ((indent (make-string (1+ (port-column port)) #\space)))
    (display "(" port)
    (if (not (null? lst))
	(begin
	  (display (first lst) port)
	  (for-each (lambda (elem)
		      (newline port)
		      (display indent port)
		      (display elem port))
		    (cdr lst))))
    (display ")" port)))

(define-public (record-print rec port)
  (define rtd (record-type-descriptor rec))
  (format port "~a record:\n" (record-type-name rtd))
  (for-each (lambda (field)
	      (let ((value ((record-accessor rtd field) rec)))
		(format port "  ~s: " field)
		(cond ((datatype? value)
		       (display (datatype-key value) port))
		      ((timebase? value)
		       (display (timebase-name value) port))
		      ((alist? value)
		       (alist-print value port))
		      (else
		       (write value port)))
		(newline port)))
	    (record-type-fields rtd)))

(let-toplevel
    ((rec-type (make-record-type "series"
				 '(symbol
				   name
				   datatype decimals array timebase
				   dividends
				   info lo hi require
				   parent data notes)
				 record-print)))

  (define rec-require (rec-accessor rec-type 'require))

  (define (proc-on-parent-top proc)
    (lambda (series . rest)
      (apply proc series rest)))

  (define-public series-copy       record-copy)
  (define-public series-symbol     (rec-accessor rec-type 'symbol))
  (define-public series-name       (rec-accessor rec-type 'name))
  (define-public series-datatype   (rec-accessor rec-type 'datatype))
  (define-public series-decimals   (rec-accessor rec-type 'decimals))
  (define-public series-timebase   (rec-accessor rec-type 'timebase))

  ;; alist (TDATE . AMOUNT)
  ;; AMOUNT is string
  ;;        or (list 'month MDATE)
  ;;        or (list 'split NEWNUM FOROLDNUM)
  ;;        or (list 'capital-return AMOUNT)
  (define-public series-dividends  (rec-accessor rec-type 'dividends))
  (define-public series-dividends! (rec-modifier rec-type 'dividends))

  ;; alist (KEY . VALUE), where KEY is a symbol
  ;;
  ;; (download-done . TDATE) is the date downloading has covered, which may
  ;; be later than the last actual data if for instance the most recent day
  ;; is a public holiday
  ;;
  (define-public series-info       (rec-accessor rec-type 'info))
  (define-public series-info!      (rec-modifier rec-type 'info))

  (define-public series-lo         (rec-accessor rec-type 'lo))
  (define-public series-hi         (rec-accessor rec-type 'hi))
  (define-public series-held-array (rec-accessor rec-type 'array))
  (define-public series-parent     (rec-accessor rec-type 'parent))
  (define-public series-data       (rec-accessor rec-type 'data))
  (define-public series-notes      (proc-on-parent-top
				    (rec-accessor rec-type 'notes)))

  (define-public series-name!      (rec-modifier rec-type 'name))
  (define-public series-datatype!  (rec-modifier rec-type 'datatype))
  (define-public series-decimals!  (rec-modifier rec-type 'decimals))
  (define-public series-timebase!  (rec-modifier rec-type 'timebase))
  (define-public series-array!     (rec-modifier rec-type 'array))
  (define-public series-lo!        (rec-modifier rec-type 'lo))
  (define-public series-hi!        (rec-modifier rec-type 'hi))
  (define-public series-require!   (rec-modifier rec-type 'require))
  (define-public series-parent!    (rec-modifier rec-type 'parent))
  (define-public series-data!      (rec-modifier rec-type 'data))
  (define-public series-notes!     (proc-on-parent-top
				    (rec-modifier rec-type 'notes)))

  (define-public (series-parent-top series)
    (if (series-parent series)
	(series-parent-top (series-parent series))
	series))

  (define-public series-new
    (let ((rec-new
	   (record-constructor
	    rec-type '(symbol name decimals datatype array timebase
			      dividends info lo hi require notes))))
      (lambda* (#:key (symbol    #f)
		      (name      #f)   ;; default no name
		      (decimals  0)
		      (array     (make-array #f '(0 0)))
		      datatype
		      (timebase  timebase-days)
		      (dividends '())
		      (info      '())
		      (require   noop)
		      (lo        (array-first-index array))
		      (hi        (array-last-index array)))

	;; 6 column ohlcv force format ohlcvi marker
	(if (and (eq? datatype datatype-ohlcv)
		 (= 5 (array-last-column array))
		 (not (assq 'format info)))
	    (set! info (acons 'format 'ohlcvi info)))

	(rec-new symbol name decimals datatype array timebase
		 dividends info lo hi require '()))))

  (define-public (series-array series lo hi)
    ;; restrict to actual range available
    (set! lo (min (series-hi series) (max (series-lo series) lo)))
    (set! hi (min (series-hi series) (max (series-lo series) hi)))

    ;; range in held array
    (let ((alo   lo)
	  (ahi   hi)
	  (array (series-held-array series)))
      (if array
	  (begin
	    (set! alo (array-first-index array))
	    (set! ahi (array-last-index array))))

      (if (or (not array)
	      (< lo alo)
	      (> hi ahi))
	  (begin
	    ;; need more data, either above or below

	    ;; add an extra few points to the end being extended, to try to
	    ;; avoid work if a subsequent operation wants an extra point or
	    ;; two to compare today with prev, or whatever
	    (if (< lo alo)
		(set! lo (max (series-lo series) (- lo 3))))
	    (if (> hi ahi)
		(set! hi (min (series-hi series) (+ hi 3))))

	    ;; extend requested range to what's already held, so as to keep
	    ;; all data in the held array when extending only one end
	    (set! lo (min lo alo))
	    (set! hi (max hi ahi))

	    (set! array ((rec-require series) series lo hi))
	    (series-array! series array)))
      array)))

;; in SERIES info, get value for KEY
(define-public (series-info-get series key)
  (assq-ref (series-info series) key))

;; in SERIES info, set KEY to VAL
(define-public (series-info-add! series key val)
  (series-info! series (assq-set! (series-info series) key val)))

(define-public (series-info-del! series key)
  (series-info! series (alist-delete! key (series-info series))))

(define-public (series-dividend-add! series tdate val)
  (series-dividends! series (assv-set! (series-dividends series) tdate val)))

;; Return a list (LOWVAL HIGHVAL) being the lowest and highest values in
;; SERIES between dates LO and HI (default #f for whole series).
;; If no data in SERIES then LOWVAL and HIGHVAL are #f.
(define*-public (series-range series #:optional lo hi)
  (set! lo (max-maybe lo (series-lo series)))
  (set! hi (min-maybe hi (series-hi series)))
  (let ((array (series-array series lo hi)))
    (if (eq? datatype-ohlcv (series-datatype series))
	(set! array (make-shared-array array list
				       (first (array-shape array))
				       '(0 3))))     ;; ohlc only
    (let ((lst (array->list-flat (make-shared-array-restrict array lo hi))))
      (set! lst (delq #f lst))
      (if (null? lst)
	  '(#f #f)
	  (list (apply min lst) (apply max lst))))))

(define-public (ohlcv-open-interest? series)
  (eq? 'ohlcvi (assq-ref (series-info series) 'format)))


;;-----------------------------------------------------------------------------
;; series timebase conversions

;; OPTIMIZE-ME: copy old parts of array
(define-public (series-convert-timebase-require series lo hi
						reduce-proc
						t-old->new
						t-new->old)
  (reduce-proc (series-array (series-parent series)
			     (t-new->old lo)
			     (1- (t-new->old (1+ hi))))
	       t-old->new))

;; return a new series, in the requested TO-TIMEBASE
(define-public (series-convert-timebase series to-timebase)
  (let ((from-timebase (series-timebase series)))
    (if (eq? from-timebase to-timebase)
	series

	(let ((new-series  (series-copy series))
	      (reduce-proc (datatype-reduce-proc  (series-datatype series)))
	      (t-old->new  (timebase-convert-proc from-timebase
						  to-timebase))
	      (t-new->old  (timebase-convert-proc to-timebase
						  from-timebase)))
	  (or reduce-proc
	      (error "no reduce proc on" (series-datatype series)))

	  (series-timebase! new-series to-timebase)
	  (series-name-append! new-series (timebase-adjective to-timebase))
	  (series-require! new-series
			   (lambda (series lo hi)
			     (series-convert-timebase-require series lo hi
							      reduce-proc
							      t-old->new
							      t-new->old)))
	  (series-parent! new-series series)
	  (series-array! new-series #f)
	  (series-lo! new-series (t-old->new (series-lo series)))
	  (series-hi! new-series (t-old->new (series-hi series)))

	  (series-dividends! new-series
			     (map (lambda (pair)
				    (cons (t-old->new (car pair))
					  (cdr pair)))
				  (series-dividends series)))

	  new-series))))

;; append " - STR" to the SERIES name
(define-public (series-name-append! series str)
  (series-name! series (let ((name (series-name series)))
			 (if name
			     (string-append name " - " str)
			     str))))

;; return the date of "today" for SERIES, in the series timebase and the
;; series symbol's timezone
(define-public (series-today series)
  (if (series-symbol series)
      (timebase-today (series-timebase series)
		      (symbol-timezone (series-symbol series)))
      (timebase-today (series-timebase series))))


;;-----------------------------------------------------------------------------
;; array stuff

;; first index of a one dimensional array
(define-public (array-first-index array)
  (caar (array-shape array)))

;; last index of a one dimensional array
(define-public (array-last-index array)
  (cadar (array-shape array)))

;; first column index of a two dimensional array
(define-public (array-first-column array)
  (caadr (array-shape array)))

;; last column index of a two dimensional array
(define-public (array-last-column array)
  (cadadr (array-shape array)))

;; Return a shared array which is column N of a two-dimensional array
(define-public (make-shared-array-column array n)
  (make-shared-array array
		     (lambda (i) (list i n))
		     (first (array-shape array))))

;; return a shared array which is columns X to Y of a two-dimensional array
(define-public (make-shared-array-columns array x y)
  (make-shared-array array list (first (array-shape array)) (list x y)))

;; return a shared array with the first index restricted to the given LO, HI
(define-public (make-shared-array-restrict array lo hi)
  (let* ((shape (array-shape array))
	 (alo   (caar shape))
	 (ahi   (cadar shape)))
    (apply make-shared-array
	   array list
	   (list (min ahi (max alo lo))
		 (max alo (min ahi hi)))
	   (cdr shape))))

;; return the last element of ARRAY satisfying (PRED elem), or #f if none
;; ARRAY must be 1-dimensional
(define-public (array-find-last array pred)
  (let ((lo (array-first-index array)))
    (let more ((i (array-last-index array)))
      (and (>= i lo)
	   (if (pred (array-ref array i))
	       (array-ref array i)  ;; found
	       (more (1- i)))))))

;; like `array->list' but return as a flat list of elements, not list of lists
(define-public (array->list-flat array)
  (if (= 1 (array-rank array))
      (array->list array)
      (concatenate! (array->list array))))


;;-----------------------------------------------------------------------------
;; string stuff

;; like `string-join' but ignore #f or "" empty strings in LST
(define-public (string-join-notnull lst . rest)
  (apply string-join (remove! string-null? (delq #f lst)) rest))


;;-----------------------------------------------------------------------------
;; string number stuff

;; return how many decimal places on the number in STR
(define-public (string-decimals str)
  (let ((pos (string-index str #\.)))
    (if pos
	(- (string-length str) pos 1)
	0)))

;; convert STR to a number, not allowing non-numeric strings
(define-public (string->number-err str)
  (or (string->number str)
      (error "not a number:" str)))

;; `string->number' with a factor 10^DECIMALS applied.
;; If STR is empty "", return #f, but otherwise must be a number.
;;
(define-public (string->number+decimals str decimals)
  (and (not (string-null? str))
       (begin
	 ;; strip out decimal point
	 (and-let* ((pos (string-index str #\.)))
	   (set! str (string-replace str "" pos (1+ pos)))
	   (set! decimals (- decimals (- (string-length str) pos))))

	 ;; trim zeros so "1.200" with decimals==2 stays exact in guile 1.6
	 (let ((newstr (string-trim-right str #\0)))
	   ;; leave "+0" or "-0" because just "+" or "-" is not accepted by
	   ;; `string->number'
	   (if (member newstr '("+" "-"))
	       (set! newstr (string-take str 2)))

	   (set! decimals (+ decimals (- (string-length str)
					 (string-length newstr))))
	   (set! str (if (string-null? newstr) "0" newstr)))

	 (* (string->number-err str)
	    (if (negative? decimals)
		(integer-expt 0.1 (- decimals))  ;; inexact not fraction
		(integer-expt 10 decimals))))))

;; Return list (DECIMALS VAL1 ... VALN).
;; Each VAL has a factor 10^DECIMALS applied, to make them all integers.
;; #f or "" are allowed in the given strings, they become #f in the return VALs
;;
(define-public (strings->numbers+decimals . str-lst)
  (let ((decimals (apply max 0 (map string-decimals (delq #f str-lst)))))
    (cons decimals (map (lambda (str)
			  (and str
			       (string->number+decimals str decimals)))
			str-lst))))


;;-----------------------------------------------------------------------------
;; chart symbol stuff

;; return SYMBOL without the ".ZZ" part, if any; eg. "BHP.AX" -> "BHP"
(define-public (chart-symbol-sans-dot symbol)
  (string-take symbol (or (string-rindex symbol #\.)
			  (string-length symbol))))

;; return the ".ZZ" part of SYMBOL, or "" if none
(define-public (chart-symbol-suffix symbol)
  (substring symbol (or (string-index symbol #\.)
			(string-length symbol))))

;; yahoo or similar index symbol like "^TNX", without the "^" part, if any
(define-public (chart-symbol-sans-hat symbol)
  (string-trim symbol #\^))

;; return month for code character
(define-public (M-code->month c)
  (and-let* ((idx (string-index "FGHJKMNQUVXZ" c)))
    (1+ idx)))

;; return year like 2005 for STR "5", "05", or "2005"
(define-public (yyy-str->year str)
  (case (string-length str)
    ((1) (y->year (string->number-err str)))
    ((2) (yy->year (string->number-err str)))
    ((4) (string->number-err str))
    (else #f)))

;;                                  12 3mon                                                4Mcode           5year     6num      7suffix
(let-toplevel ((regex (make-regexp "(( (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) |[^]?[^^]([FGHJKMNQUVXZ]))([0-9]+)| ([0-9]+))?((\\.[^.]+)?)$" regexp/icase)))

  ;; from a match of regex above, return the year as full 2005 etc, or the
  ;; number ending like 3 in "ZINC 3.LME", or 0 if neither
  (define (m-year-or-num-or-0 m)
    (or (and (match:start m 5)  ;; year
	     (yyy-str->year (match:substring m 5)))
	(and (match:start m 6)  ;; number
	     (string->number (match:substring m 6)))
	0))

  ;; from a match of regex above, return the month 1 to 12, or 0 if no month
  (define (m-month-or-0 m)
    (or (and (match:start m 3)  ;; Jan etc
	     (Mmm-str->month (match:substring m 3)))
	(and (match:start m 4)  ;; F etc M-code
	     (M-code->month (string-ref (match:string m) (match:start m 4))))
	0))

  ;; return commodity part of SYMBOL, or whole symbol (sans suffix) if no
  ;; month/year
  ;; eg. "GC" for "GCH05.COMEX"
  ;;     "TIN 3" for "TIN 3.LME"
  (define-public (chart-symbol-commodity symbol)
    (let ((m (regexp-exec regex symbol)))
      (string-take symbol (or (and (match:start m 6)  ;; if "TIN 3" style num
				   (match:start m 7)) ;;   stop at suffix
			      (match:start m 4)       ;; stop at M-code
			      (match:start m)))))     ;; stop at " JAN 06"

  ;; return mdate for SYMBOL, or #f if not a futures specific month
  (define-public (chart-symbol-mdate symbol)
    (let ((m (regexp-exec regex symbol)))
      (and (match:start m 5)      ;; year
	   (or (match:start m 3)  ;; Jan etc
	       (match:start m 4)) ;; F etc M-code
	   (ymd->mdate (yyy-str->year (match:substring m 5))
		       (if (match:start m 3)
			   (Mmm-str->month (match:substring m 3))
			   (M-code->month (string-ref symbol
						      (match:start m 4))))
		       1))))

  ;; the parts of the `or's here give a negative number for x<y, a positive
  ;; number for x>y, or #f if x==y and hence should go to the next term
  ;;
  (define-public (chart-symbol<? x y)
    (negative?
     (or
      ;; ^ indexes less than other symbols
      (number-order (string-prefix-length "^" y)
		    (string-prefix-length "^" x))

      (let* ((mx (regexp-exec regex x))
	     (my (regexp-exec regex y)))

	(or
	 ;; commodity parts
	 (string-order x y
		       0 (or (match:start mx 4)
			     (match:start mx))
		       0 (or (match:start my 4)
			     (match:start my)))

	 ;; suffix parts
	 (string-order x y
		       (match:start mx 7) (match:end mx 7)
		       (match:start my 7) (match:end my 7))

	 ;; year or number parts
	 (number-order (m-year-or-num-or-0 mx)
		       (m-year-or-num-or-0 my))

	 ;; month parts
	 (number-order (m-month-or-0 mx)
		       (m-month-or-0 my))

	 ;; equal
	 0))))))

;; X and Y are numbers, return positive if X>Y, negative if X<Y, #f if X==Y
(define (number-order x y)
  (set! x (- x y))
  (and (not (zero? x))
       x))

;; X and Y are strings, return positive if X>Y, negative if X<Y, #f if X==Y
(define (string-order x y x-start x-end y-start y-end)
  (or (string-compare-ci x y
			 lognot ;; less    0toN -> negative
			 not    ;; equal   0toN -> #f
			 1+     ;; greater 0toN -> positive
			 x-start x-end
			 y-start y-end)
      (string-compare x y
		      lognot ;; less    0toN -> negative
		      not    ;; equal   0toN -> #f
		      1+     ;; greater 0toN -> positive
		      x-start x-end
		      y-start y-end)))

(define-public (sympred-test pred symbol)
  (cond ((procedure? pred)
	 (pred symbol))
	((string? pred)
	 (string-suffix? pred symbol))
	((pair? pred)
	 (any (lambda (pred)
		(sympred-test pred symbol))
	      pred))
	(else
	 (error "Invalid symbol predicate:" pred))))

(define-public (sympred-assoc symbol alist)
  (find (lambda (elem)
	  (sympred-test (car elem) symbol))
	alist))

(define-public (sympred-assoc-ref alist symbol)
  (cond ((sympred-assoc symbol alist) => cdr)
	(else #f)))


;;-----------------------------------------------------------------------------
;; annotations

(define-public (annotation? note)
  (and (pair? note)
       (eq? 'annotation (first note))))


;;-----------------------------------------------------------------------------
;; waiting for idle/input/timeout
;;
;; The initial code here is for the single-threaded case.  `c-main' sets in
;; code for multi-continuation operation.

;; block until the main loop is idle, at PRIORITY level
(define*-public (c-idle #:optional (priority 200)) ;; gtk-priority-default
  #f)

;; block until one of CONDS is satisfied on PORT
;; CONDS is a list of symbols `read' and/or `write' as per `gtk-input-add',
;; eg. (c-input-wait sock '(read))
;;
(define-public (c-input-wait port conds)
  (select (if (memq 'read conds)
	      (list port)
	      '())
	  (if (memq 'write conds)
	      (list port)
	      '())
	  '()))
    
;; Block until one of CONDS is satisfied on PORT, or SECS time has elapsed,
;; whichever comes first.  Return #t for CONDS, or #f for timeout.
;;
;; SECS can be a flonum or fraction, but the resolution is only milliseconds
;; under Gtk, and probably only actually CLK_TCK anyway (normally 1/100 sec)
;;
(define-public (c-input-timeout-wait port conds secs)
  (receive-list (got-read got-write got-except)
      (select (if (memq 'read conds)
		  (list port)
		  '())
	      (if (memq 'write conds)
		  (list port)
		  '())
	      '()
	      secs)
    (or (not (null? got-read))
	(not (null? got-write)))))


;;-----------------------------------------------------------------------------
;; non-gui gtk style idle/input/timeout callbacks

(define-public c-gtk-input-list   '())  ;; list of (port conds proc)
(define-public c-gtk-timeout-list '())  ;; list of (target-internal proc)
(define-public c-gtk-idle-list    '())  ;; list of (priority proc)

(define-public (c-gtk-input-add port conds proc)
  (let ((id (list port conds proc)))
    (set! c-gtk-input-list (cons id c-gtk-input-list))
    id))
(define-public (c-gtk-input-remove id)
  (set! c-gtk-input-list (delq id c-gtk-input-list)))

(define (milliseconds->internal-time ms)
    (quotient (* ms internal-time-units-per-second) 1000))
(define-public (c-gtk-timeout-add milliseconds proc)
  (let ((id (list (+ (get-internal-real-time)
		     (milliseconds->internal-time milliseconds)) proc)))
    (set! c-gtk-timeout-list (cons id c-gtk-timeout-list))
    id))
;; timeouts are removed when called, this can remove them before called
(define-public (c-gtk-timeout-remove id)
  (set! c-gtk-timeout-list (delq id c-gtk-timeout-list)))

(define-public (c-gtk-idle-add-priority priority proc)
  (let ((id (list priority proc)))
    (set! c-gtk-idle-list (merge c-gtk-idle-list (list id) car<))
    id))
;; idles are removed when called, this can remove them before called
(define-public (c-gtk-idle-remove id)
  (set! c-gtk-idle-list (delq id c-gtk-idle-list)))


(define-public (c-main-stop?)
  (and (every c-gtk-id-stoppable? c-gtk-timeout-list)
       (every c-gtk-id-stoppable? c-gtk-input-list)
       (null? c-gtk-idle-list)
       (q-empty? c-main-queue)))

;; like `gtk-main-iteration' return true if should quit the main loop
(define-public (c-main-iteration)
  (or (c-main-stop?)
      (begin
	(let ((read-list '())
	      (write-list '()))
	  (for-each (lambda (input)
		      (let ((port (first input))
			    (conds (second input)))
			(or (port-closed? port)
			    (begin
			      (if (memq 'read conds)
				  (set! read-list (cons port read-list)))
			      (if (memq 'write conds)
				  (set! write-list (cons port write-list)))))))
		    c-gtk-input-list)

	  (receive-list (got-read got-write got-exception)
	      (apply select
		     read-list write-list '()
		     (cond ((not (null? c-gtk-idle-list))
			    '(0))
			   ((null? c-gtk-timeout-list)
			    '())
			   (else
			    (list (/ (max 0
					  (- (apply min
						    (map first
							 c-gtk-timeout-list))
					     (get-internal-real-time)))
				     internal-time-units-per-second)))))

	    (for-each (lambda (input)
			(receive-list (port conds proc)
			    input
			  (let ((readable (and (memq 'read conds)
					       (memq port got-read)))
				(writable (and (memq 'write conds)
					       (memq port got-write))))
			    (if (or readable writable)
				(c-main-enq! proc
					     (fileno port)
					     (append (if readable '(read) '())
						     (if writable '(write) '())))))))
		      c-gtk-input-list)))

	(let ((now (get-internal-real-time)))
	  (set! c-gtk-timeout-list
		(remove (lambda (timeout)
			  (receive-list (target-time proc)
			      timeout
			    (and (>= now target-time) ;; if reached
				 (begin
				   (c-main-enq! proc)
				   #t))))
			c-gtk-timeout-list)))

	(for-each (lambda (idle)
		    (receive-list (priority proc) idle
		      (c-main-enq! proc)))
		  c-gtk-idle-list)
	(set! c-gtk-idle-list '())

	#f))) ;; don't exit main loop


;;-----------------------------------------------------------------------------
;; continuation job control

(let-toplevel ((rec-type (make-record-type "job" '(cleanup))))
  (define rec-new      (record-constructor rec-type))
  (define rec-cleanup  (rec-accessor rec-type 'cleanup))
  (define rec-cleanup! (rec-modifier rec-type 'cleanup))
  (define job-current-fluid (make-fluid))

  (define-public (job-start thunk)
    (let ((job (rec-new #f)))
      (define (callback-initial)
	(rec-cleanup! job #f)
	(with-fluids* (list job-current-fluid) (list job)
	  (lambda ()
	    (catch 'job-stop thunk noop))))
      (rec-cleanup! job (list noop (lambda ()
				     (or (c-main-unqueue! callback-initial)
					 (error "job-start initial cleanup: oops callback not in queue")))))
      (c-main-enq! callback-initial)
      job))

  ;; SETUP is a procedure called (SETUP cont) with the continuation after
  ;; the job-pause call.  It should setup to goto cont when ready to
  ;; continue.  Its return is a list of arguments to the CLEANUP proc.
  ;; CLEANUP is called (CLEANUP ARGS...) if the job is stopped.
  ;;
  (define-public (job-pause setup cleanup)
    (define ret
      (call-with-current-continuation
       (lambda (cont)
	 (let ((cleanup-args (setup cont)))
	   (and-let* ((job (fluid-ref job-current-fluid)))
	     (if (rec-cleanup job)
		 (error "job-pause: oops, already paused"))
	     (rec-cleanup! job (list cont
				     (lambda ()
				       (apply cleanup cleanup-args))))))
	 (c-main-goto-top))))

    (and-let* ((job (fluid-ref job-current-fluid)))
      (or (rec-cleanup job)
	  (error "job-unpause: oops, not paused"))
      (rec-cleanup! job #f))

    (if (eq? ret 'throw-job-stop)
	(throw 'job-stop))
    ret)

  (define-public (job-stop job)
    (if (eq? job (fluid-ref job-current-fluid))
	(throw 'job-stop)

	(begin
	  (or (rec-cleanup job)
	      (error "job-stop: oops, no cleanup cont/proc"))
	  (receive-list (cont cleanup-thunk)
	      (rec-cleanup job)
	    (cleanup-thunk)
	    (c-main-enq! (lambda ()
			   (cont 'throw-job-stop))))))))


;;-----------------------------------------------------------------------------
;; main loop

(define-public c-main-queue (make-q))

;; (c-main-enq! PROC [ARGS...]) will call (PROC ARGS...)
(define-public (c-main-enq! . args)
  (enq! c-main-queue args))  
;;
;; some arg count checking, usually too late at this point
;;   (let ((got (1- (length args))))
;;       (receive-list (req more opt)
;; 	  (procedure-property (first args) 'arity)
;; 	(if (< got req)
;; 	    (error "c-main-queue: too few arguments" req args))))

;; run HOOK, but do so with separate c-main-queue jobs
(define-public (c-main-run-hook hook . args)
  (for-each (lambda (proc)
	      (enq! c-main-queue (cons proc args)))
	    (hook->list hook)))

;; remove PROC from `c-main-queue', return #t if removed, #f if wasn't present
(define (c-main-unqueue! proc)
  (and-let* ((elem (find (lambda (elem)
			   (eq? proc (first elem)))
			 (car c-main-queue))))
    (q-remove! c-main-queue elem)
    #t))

;; note that the given input or timeout ID should not prevent c-main-stop?
;; from returning true, ie. this handler is only and idle or cleanup event
;; rather than part of a protocol and therefore shouldn't prevent the
;; mainline from exiting
(define-public (c-gtk-id-stoppable id)
  (or (number? id) ;; do nothing for integer from real gtk
      (set-object-property! id 'c-main-stoppable #t)))

(define (c-gtk-id-stoppable? id)
  (object-property id 'c-main-stoppable))

(define-public (c-main-run-q)
  (if (not (q-empty? c-main-queue))
      (let ((lst (deq! c-main-queue)))
	(apply (car lst) (cdr lst))
	(c-main-run-q))))

(define-public c-main-goto-top #f)

(define-public (c-main)

  (set! c-idle
	(lambda* (#:optional (priority 200)) ;; gtk-priority-default
	  (job-pause (lambda (cont)
		       (define (callback)
			 (cont #f))
		       (list (c-gtk-idle-add-priority priority callback)
			     callback))
		     (lambda (id callback)
		       (c-gtk-idle-remove id)
		       (c-main-unqueue! callback)))))

  (set! c-input-wait
	(lambda (port conds)
	  (let ((id #f))
	    (job-pause (lambda (cont)
			 (define (callback fd conds)
			   (c-gtk-input-remove id)
			   (cont #f))
			 (set! id (c-gtk-input-add port conds callback))
			 (list callback))
		       (lambda (callback)
			 (c-gtk-input-remove id)
			 (c-main-unqueue! callback))))))


  ;; if the input and timeout both fire in one c-main-iteration then both
  ;; callbacks could be in c-main-queue, use `done' to only run one of them
  ;;
  (set! c-input-timeout-wait
	(lambda (port conds secs)
	  (let ((input-id #f)
		(time-id  #f))
	    (job-pause
	     (lambda (cont)
	       (define (input-callback fd conds)
		 (c-main-unqueue! time-callback)
		 (c-gtk-input-remove input-id)
		 (c-gtk-timeout-remove time-id)
		 (cont #t))
	       (define (time-callback)
		 (c-main-unqueue! input-callback)
		 (c-gtk-input-remove input-id)
		 (cont #f))
	       (set! input-id (c-gtk-input-add port conds input-callback))
	       (set! time-id  (c-gtk-timeout-add (ceiling-exact (* 1000 secs))
				time-callback))
	       (list input-callback time-callback))
	     (lambda (input-callback time-callback)
	       (c-main-unqueue! input-callback)
	       (c-main-unqueue! time-callback)
	       (c-gtk-input-remove input-id)
	       (c-gtk-timeout-remove time-id))))))

  (call-with-current-continuation (lambda (cont)
				    (set! c-main-goto-top cont)))

  ;; loop until c-main-iteration says to stop
  (do ()
      ((begin
	 (c-main-run-q)
	 (c-main-iteration)))))


;;-----------------------------------------------------------------------------
;; symbol weblink URLs

;; elements are pairs (SYMPRED . HANDLER-PROC)
(define-public weblink-handler-list '())

;; SYMPRED is passed to `sympred-test'
;; PROC is called as (PROC symbol) and should return a list (MENU-NAME
;; TOOLTIP URL), or #f if nothing for this symbol.  TOOLTIP can be #f if
;; MENU-NAME is self-explanatory.
;;
;; URL is either #f; a string url; or a procedure returning a string or #f.
;; The usual case is a string, which is generally formed from the symbol
;; argument to PROC.  #f lets MENU-NAME appear as a menu entry, but
;; insensitive; for example if symbols from a given source normally have
;; such a link but a few don't.  The procedure case allows a url to be found
;; by doing some downloading, if for instance a code number or similar is
;; needed (eg. bendigo.scm).
;;
(define-public (weblink-handler! pred proc)
  (set! weblink-handler-list
	(append weblink-handler-list (list (cons pred proc)))))

;; If a weblink url procedure needs to download something, it can print a
;; message with `weblink-message', to show what's happening.
;; setup by main.scm
(define-public (weblink-message fmt . args)
  #f)
''(guile-lint simple-format weblink-message)


;;-----------------------------------------------------------------------------
;; symbol setups

;; Calling (symbol-setups SYM) initializes download methods and timezone
;; information for SYM.  The idea is to load data sources and exchange
;; specific stuff only when a relevant symbol is used.

;; alist (PRED . SETUP)
;; this is top-level so it can be exercised in tests/misc.scm
(define symbol-setups-alist '())

;; register a call (PROC) to be made a symbol satisfying (PRED sym) is used
;; PRED can be a string to test for that suffix
;; PROC can be a module name list to load that module
(define-public (symbol-setups! pred setup)
  (set! symbol-setups-alist
	(acons (if (and (string? pred)
			(string-prefix? "^" pred))
		   (lambda (symbol)
		     (string-prefix? pred symbol))
		   pred)
	       setup
	       symbol-setups-alist)))

;; call setup procedures for SYMBOL
(define-public (symbol-setups symbol)
  (and-let* ((elem (sympred-assoc symbol symbol-setups-alist)))
    (set! symbol-setups-alist (delq! elem symbol-setups-alist))

    (let ((setup (cdr elem)))
      (if (list? setup)
	  (resolve-interface setup) ;; throws an error for not found
	  (setup))
      (symbol-setups symbol))))


;;-----------------------------------------------------------------------------
;; symbol setups - builtins

;; (symbol-setups! ".AX"    '(chart commsec))
;; (symbol-setups! ".AX"    '(chart float))
(symbol-setups! ".WCE"   '(chart barchart))  ;; intraday

(symbol-setups! (lambda (symbol) ;; no dot usa
		  (not (string-index symbol #\.)))
		'(chart usa))

(symbol-setups! (lambda (symbol) ;; various ^ indexes
		  (string-prefix? "^" symbol))
		'(chart yahoo))

(load "symbol-setups.scm")


;;-----------------------------------------------------------------------------
;; notifications

(define notify-directory 
  (g-build-filename database-directory "notify"))

;; set non-blocking mode on SOCK
(define-public (fcntl-set-nonblock sock)
  (fcntl sock F_SETFL
	 (logior (fcntl sock F_GETFL) O_NONBLOCK)))

;; host-prefix is eg. "foo.com.au:"
;; address is eg. "foo.com.au:1234" with pid
(let-toplevel* ((host-prefix (string-append (gethostname) ":"))
		(our-address (string-append host-prefix
					    (number->string (getpid))))
		;; (KEY . PROC-LIST)
		(listen-alist '()))

  ;; created when there's someone to talk to, since for a single user there
  ;; might be only one running chart
  (define out-sock
    (make-procedure-with-cache-0
     (lambda ()
       (let ((sock (socket PF_UNIX SOCK_DGRAM 0)))
	 (fcntl-set-nonblock sock)
	 sock))))

  (define-public (notify-disconnect key proc)
    (set! listen-alist
	  (assq-set! listen-alist key
		     (delq! proc (or (assq-ref listen-alist key) '())))))

  ;; call (PROC arg) when KEY is broadcast
  (define-public (notify-connect key proc)
    (set! listen-alist
	  (assq-set! listen-alist key
		     (append! (or (assq-ref listen-alist key) '())
			      (list proc)))))

  ;; initialize listening for notifications
  (define-public (notify-listen)
    (let ((sock (socket PF_UNIX SOCK_DGRAM 0))
	  (full (g-build-filename notify-directory our-address)))
      (mkdir-all notify-directory)

      ;; unlikely but possible leftover from previous chart under this pid
      (delete-file-maybe full)

      (bind sock AF_UNIX full)
      (c-atexit (lambda ()
		  (delete-file-maybe full)))

      ;; receive update notifications
      (c-gtk-input-add sock '(read)
	(lambda (fdes conds)
	  (let* ((buf (make-string 4096)))
	    (receive-list (key arg)
		(read (open-input-string
		       (substring buf 0 (recv! sock buf))))
	      (notify-send-locally key arg))))))

    ;; once only, and gc this code
    (set! notify-listen noop))

  (define-public (notify-send-locally key arg)
    (for-each (lambda (proc)
		(proc arg))
	      (or (assq-ref listen-alist key) '())))

  (define-public (notify-send key arg)
    (notify-send-locally key arg)
    (let ((elem-str (format #f "~s" (list key arg))))
      (for-each (lambda (dest)
		  (set! dest (g-build-filename notify-directory dest))
		  (catch 'system-error
		    (lambda ()
		      (sendto (out-sock) elem-str AF_UNIX dest))
		    (lambda args
		      ;; defunct notify destination, some previous run
		      ;; didn't remove its socket
		      (delete-file-maybe dest))))
		(or (directory-files
		     notify-directory #f
		     (lambda (name)
		       (and (string-prefix? host-prefix name)  ;; locals
			    (not (member name
					 (list our-address "." ".."))))))
		    '())))))


;;-----------------------------------------------------------------------------
;; preferences file

(define-public preferences-filename
  (g-build-filename user-directory "prefs.scm"))

;; call (PROC) until it returns an eof object
;; return the values in a list (excluding the eof object)
(define (list-call-to-eof proc)
  (let more ((lst '()))
    (let ((elem (proc)))
      (if (eof-object? elem)
	  (reverse! lst)
	  (more (cons elem lst))))))

;; return list of objects from PORT with `read' until EOF is reached
(define-public (list-read-to-eof port)
  (list-call-to-eof (lambda () (read port))))

;; return the preference value for a given KEY (a scheme symbol)
(define-public (preference-get key default)
  (call-with-input-file-or default preferences-filename
    (lambda (port)
      (let more ((data (read port)))
	(cond ((eof-object? data)
	       default)
	      ((eq? key (first data))
	       (second data))
	      (else
	       (more (read port))))))))

;; list values are written with one element per line, for readability,
;; eg. in the favourites list
(define-public (preference-set key value)
  (database-call-with-lock user-notes-directory
    (lambda ()
      (let ((lst (call-with-input-file-or '() preferences-filename
		   list-read-to-eof)))
	(set! lst (assq-set! lst key (list value)))
	(database-call-with-output-file preferences-filename
	  (lambda (port)
	    (display ";; User preferences file, generated by Chart\n\n" port)
	    (for-each (lambda (elem)
			(write-pref (first elem) (second elem) port))
		      lst))))))
  (notify-send 'preferences-save #f))

(define (write-pref key value port)
  (if (list? value)
      (begin
	(format port "(~s\n " key)
	(write-list-lines value port)
	(display  ")\n" port))
      (format port "~s\n" (list key value))))
