;; MLC Funds setups.

;; Copyright 2004, 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mlc)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart http)
  #:use-module (chart indicator)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (mlc-symbol? symbol)
  (string-suffix? ".MLC" symbol))

;; not sure exactly where MLC operates out of, but Sydney is close enough
(symbol-timezone! mlc-symbol? (_ "Sydney") timezone-sydney)
(source-help! mlc-symbol? (_ "CONTEXT(manual-node): MLC Funds"))

;; The home page has a link to fund descriptions, unfortunately it's in
;; flash format.


;;-----------------------------------------------------------------------------
;; download
;;
;; This uses the unit prices under
;;
;;     https://www.mlc.com.au/masterkeyWeb/execute/FramesetUnitPrices
;;
;; Filling in the boxes ends up with a url like the following, with full
;; fund and product name, and a requested date range (dd/mm/yyyy).
;;
;;     https://www.mlc.com.au/masterkeyWeb/execute/UnitPricesWQO?openAgent&reporttype=HistoricalDateRange&product=MasterKey%20Superannuation%20%28Gold%20Star%29&fund=Property%20Securities%20Fund&begindate=19/07/2003&enddate=19/07/2004&
;;
;; Old data has prices for weekends too, but the same as Friday.  ymd->tdate
;; rounds down to the friday,
;;


;; `body-regexp' matches a line like the following sample,
;;
;;     historicalProduct1funds[1]="Property Securities Fund,MasterKey Superannuation (Gold Star),15 July 2004,41.25436";
;;
;; or new style March 2007,
;;
;;     historicalProduct1funds[1]="MLC Property Securities Fund,MasterKey Superannuation (Gold Star),29 March 2007,64.71567,0.00000";
;;
;; matches: 1 - d m y
;;          2 - price
;;
(define body-regexp
  (make-regexp "historicalProduct1funds[^\"\r\n]*\"[^,]+,[^,]+,([^ ]+ [^ ]+ [^,]+),([^,\"]+)(\"|,[0.]+\")"))

(define (mlc-process-download symbol lo hi body)
  (define commodity (chart-symbol-sans-dot symbol))

  (download-process
   #:module        (_ "MLC")
   #:symbol-list   (list symbol)
   #:currency      "AUD"
   #:lo            lo
   #:hi            hi
   #:last-download #t
   #:row-list
   (map (lambda (m)
	  (list #:tdate     (d/m/y-str->tdate (match:substring m 1)) ;; date
		#:commodity commodity
		#:close     (match:substring m 2))) ;; price
	(list-matches body-regexp body))))

;; return fund part of SYMBOL
;; default to symbol without .MLC if no comma (which isn't a valid symbol)
(define (mlc-symbol-fund symbol)
  (string-take symbol (or (string-index symbol #\,)
			  (- (string-length symbol) 4))))

;; return product part of SYMBOL
;; default to empty if no comma (which isn't a valid symbol)
(define (mlc-symbol-product symbol)
  (substring symbol
	     (1+ (or (string-index symbol #\,)
		     -1))
	     (- (string-length symbol) 4)))

;; Sample url:
;;  https://www.mlc.com.au/masterkeyWeb/execute/UnitPricesWQO?openAgent&reporttype=HistoricalDateRange&product=MasterKey%20Superannuation%20%28Gold%20Star%29&fund=Property%20Securities%20Fund&begindate=19/07/2003&enddate=19/07/2004&
;;
(define (mlc-url-tdates symbol lo hi)
  (receive-list (lo-year lo-month lo-day)
      (tdate->ymd lo)
    (receive-list (hi-year hi-month hi-day)
	(tdate->ymd hi)
      (format #f "https://www.mlc.com.au/masterkeyWeb/execute/UnitPricesWQO?openAgent&reporttype=HistoricalDateRange&product=~a&fund=~a&begindate=~a/~a/~a&enddate=~a/~a/~a&"
	      (url-quote (mlc-symbol-product symbol))
	      (url-quote (mlc-symbol-fund symbol))
	      lo-day lo-month lo-year
	      hi-day hi-month hi-year))))

;; can request up to 12 months at a time
(define chunk-size 250)

;; Return the expected available tdate for data.
;;
;; This is only based on observation, in the morning it seems to be not the
;; previous weekday but the one before that, and Thursday on a weekend.
;; Don't know what time of day it ticks over, assume midnight for now.
;;
(define (mlc-available-tdate)
  (1- (tdate-today-after 24 0 (timezone-sydney))))

;; return true if data found, false if empty
(define (mlc-download-chunk symbol lo hi)
  (download-status (_ "MLC") (_ "data") symbol (tdate-range-string lo hi))
  (receive (headers body)
      (http-request (mlc-url-tdates symbol lo hi)
		    #:want-ok #t)
    ;; unknown funds or bad date ranges result in a message in the body
    (and (not (string-contains body "No unit prices available"))
	 (begin
	   (mlc-process-download symbol lo hi body)
	   #t))))

(define (mlc-update symbol-list)
  (define avail-tdate (mlc-available-tdate))
  (for-each (lambda (symbol)
	      ;; no more than CHUNK-SIZE in each download, and stop when
	      ;; AVAIL-TDATE is reached
	      (do ((lo (download-start-tdate symbol)))
		  ((> lo avail-tdate))
		(let ((hi (min avail-tdate (+ lo chunk-size))))
		  (mlc-download-chunk symbol lo hi)
		  (set! lo (1+ hi)))))
	    symbol-list))

(define (mlc-backto want-tdate symbol-list)
  (for-each (lambda (symbol)
	      ;; no more than CHUNK-SIZE in each download, and stop when
	      ;; WANT-TDATE or an empty chunk is reached
	      (do ((hi (download-back-tdate symbol)))
		  ((< hi want-tdate))
		(let ((lo (max want-tdate (- hi chunk-size))))
		  (if (mlc-download-chunk symbol lo hi)
		      (set! hi (1- lo))
		      (set! hi (1- want-tdate))))))
	    symbol-list))

(download-handler! #:selector   mlc-symbol?
		   #:update     mlc-update
		   #:avail-proc mlc-available-tdate
		   #:backto     mlc-backto)


;;-----------------------------------------------------------------------------
;; latest
;;
;; The approach here is to download the latest price the same as for the
;; database above.  Actually we get the latest and second latest, so as to
;; calculate a "change", and go back a few extra days to allow for public
;; holidays.
;;
;; There's a single download of latest prices for all funds and products.
;; If we had a lot in the SYMBOL-LIST then it might be worth doing that
;; instead of individual downloads, but not sure if a set of immediately
;; preceding prices would be available to make the "change" amount.  In any
;; case for now it's probably unlikely there'll be many funds in the
;; watchlist that are not in the database.
;;

;; return latest record
(define (mlc-body->latest symbol lo body)
  ;; list of (TDATE . PRICE) sorted with highest TDATE first
  (let ((lst (fold-matches body-regexp body '()
			   (lambda (m prev)
			     (acons (d/m/y-str->tdate (match:substring m 1))
				    (match:substring m 2)
				    prev)))))

    (if (null? lst)
	(error "MLC: no prices found for latest"))

    ;; sorting ensures we have the latest and second latest prices the right
    ;; way around, no matter how they appear in the body (there's only 4
    ;; days, so this sort is fast)
    (set! lst (sort! lst car>))

    (let* ((tdate      (car (first lst)))
	   (price      (cdr (first lst)))
	   (prev-price (and (not (null? (cdr lst)))
			    (cdr (second lst)))))

      ;; also add to the database, if we've got something new, and it
      ;; won't leave a gap
      (if (within-range? lo tdate (download-start-tdate symbol))
	  (mlc-process-download symbol lo tdate body))

      (latest-new #:symbol     symbol
		  #:last-tdate tdate
		  #:last       price
		  #:prev       prev-price
		  #:source     'mlc))))

(define (mlc-latest-get symbol-list extra-list proc)
  (define avail-tdate (mlc-available-tdate))

  (for-each (lambda (symbol)
	      (let ((lo (- avail-tdate 4)))
		(receive (headers body)
		    (http-request
		     ;; get newest day, -1 previous to calculate change, and
		     ;; also -2 and -3 in case up to two public holidays, and
		     ;; also get -4 in case our idea of newest is perhaps one
		     ;; too big
		     (mlc-url-tdates symbol lo avail-tdate)
		     #:want-ok #t)

		  ;; unknown funds or bad date ranges result in message in
		  ;; the body
		  (cond ((string-contains body "No unit prices available")
			 (latest-fetch-fail (list symbol)))
			(else
			 (proc (list (mlc-body->latest symbol lo body))))))))

	    symbol-list))

(define (mlc-quote-adate-time symbol)
  (list (tdate->adate (mlc-available-tdate)) #,(hms->seconds 24 0 0)))

(latest-handler! #:selector   mlc-symbol?
		 #:handler    mlc-latest-get
		 #:adate-time mlc-quote-adate-time)

