;; Montreal Exchange setups and data downloading.   -*- coding: latin-1 -*-

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart mx)
  #:autoload   (ice-9 format)   (format)   ;; not needed just for quotes
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (mx-symbol? symbol)
  (string-suffix? ".MON" symbol))

(define-public timezone-montreal
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Montreal")
		      '("America/Montreal"
			;; Canada: second Sun in Mar, to first Sun in Nov
			;; French abbreviations for the zones are HNE/HAE,
			;; but no need to translate that, we never print the
			;; zone
			"EST+5EDT,M3.2.0,M11.1.0")
		      "EST+5"))))

(symbol-timezone! mx-symbol? (_ "Montreal") timezone-montreal)
(source-help! mx-symbol? (_ "CONTEXT(manual-node): Montreal Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - contract specs
;;
;; eg. http://www.m-x.ca/produits_taux_int_cgb_en.php
;;     http://www.m-x.ca/produits_indices_sxf_en.php
;; The sector indexes SXA,SXB,SXH,SXY are all on the one sxa page.
;;

;; this table is the equity index contracts, the rest default to being an
;; interest rate contract (BAX, OBX, ONX, CTZ, CGB, OGB).
;;
(define mx-weblinks-alist
  '(("SXF" . "indices_sxf")
    ("SXA" . "indices_sxa")
    ("SXB" . "indices_sxa")
    ("SXH" . "indices_sxa")
    ("SXY" . "indices_sxa")))

(weblink-handler! mx-symbol?
  (lambda (symbol)
    (let* ((commodity (chart-symbol-commodity symbol))
	   (page (or (assoc-ref mx-weblinks-alist commodity)
		     (string-append "taux_int_"(string-downcase commodity)))))
      (list (_ "MX _Contract Specifications")
	    (_ "Open web browser at the Montreal Exchange contract specifications for this commodity")
	    (string-append "http://www.m-x.ca/produits_" page
			   "_" (lang-select '(("en" "en")
					      ("fr" "fr")))
			   ".php")))))


;;-----------------------------------------------------------------------------
;; latest front month
;;
;; The quotes summary pages like
;;
;;     http://www.m-x.ca/nego_cotes_en.php?symbol=CGB
;;
;; are read to determine the front month for the commodity.  This is cached
;; for re-use, and rechecked each new tdate.

(define mx-front-filename (g-build-filename cache-directory "mx-front-months"))

;; return the front month symbol like "CGBM06.MON" from a nego_cotes_en.php
;; page, or #f if unknown symbol
;;
(define (mx-nego->front-symbol body)

  (if (or (string-contains body "Nonvalid")    ;; en
	  (string-contains body "non valide")) ;; fr
      #f

      (let* ((pos      (string-contains body "<div id=\"titreElement\">"))
	     (row-list (html-table-rows-raw body pos)))

	;; blank row at start
	(set! row-list (delete '() row-list))

	(let ((headings (map html-text-extract (first row-list))))
	  (or (equal? headings '("Month / Strike"
				 "Bid Price"
				 "Ask Price"
				 "Settl. Price"
				 "Net Change"
				 "Vol."))
	      (error "MX: unrecognised quotes summary headings:" headings)))
	(set! row-list (cdr row-list))

	;; blank row
	(let ((row (map! string-trim-both
			 (map html-text-extract (first row-list)))))
	  (if (equal? row '("+" "" "" "" "" ""))
	      (set! row-list (cdr row-list))))

	(let* ((field (first (first row-list)))
	       (m     (must-match (string-match "insSymbol=([A-Z0-9]+)&" field))))
	  (string-append (match:substring m 1) ".MON")))))

;; for a front month symbol like "CGB.MON" return the current specific
;; symbol for that, like "CGBM06.MON", or #f if unknown symbol
;;
(define (mx-latest-front-month symbol)
  (let* ((today (tdate-today (timezone-montreal)))
	 (alist (call-with-input-file-or '() mx-front-filename read))
	 (elem  (or (assoc symbol alist)
		    (list symbol 0 ""))))
    (if (= (second elem) today)
	(third elem)

	(receive (headers body)
	    (http-request (string-append
			   "http://www.m-x.ca/nego_cotes_en.php?symbol="
			   (chart-symbol-sans-dot symbol))
			  #:want-ok #t)
	  (define front-symbol (mx-nego->front-symbol body))
	  (database-call-with-lock data-directory
	    (lambda ()
	      (define alist (call-with-input-file-or '() mx-front-filename
			      read))
	      (set! alist (assoc-set! alist symbol (list today front-symbol)))

	      (database-call-with-output-file mx-front-filename
		(lambda (port)
		  (display ";; Chart cached MX quotes front months.
;;

" port)
		  (write-list-lines alist port)))
	      front-symbol))))))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the quotes pages like
;;
;;     http://www.m-x.ca/nego_cotes_in_en.php?symbol=CGB&insSymbol=CGBH08
;;
;; which are reached from the summary pages like
;;
;;     http://www.m-x.ca/nego_cotes_en.php?symbol=CGB
;;
;; Unfortunately this is about 38k per symbol, and there's no ETag or
;; Last-Modified to save downloading when unchanged.
;;
;; FIXME: An expired or unknown contract seems to give a blank page, and
;; currently that results in "regexp didn't match", which is not a good
;; error indication.


;; return month 1 to 12 for STR like "Jan" or "January" etc
;; case is ignored, so "JAN" etc matches too
(define-public (Mmm-fr-str->month str)
  (assoc-ref '(("janvier" 	  . 1)
	       ("f&eacute;vrier"  . 2)
	       ("mars"    	  . 3)
	       ("avril"   	  . 4)
	       ("mai"     	  . 5)
	       ("juin"    	  . 6)
	       ("juillet" 	  . 7)
	       ("ao&ucirc;t"      . 8)
	       ("septembre"       . 9)
	       ("octobre"         . 10)
	       ("novembre"        . 11)
	       ("d&eacute;cembre" . 12)
	       ("fv"             . 2)
	       ("dc"             . 12)) str))

;; return STR with all occurrances of OLD changed to NEW
(define (string-substitute/global str old new)
  (let more ((upto 0)
	     (lst  '()))
    (let ((match (string-contains str old upto)))
      (if match
	  (more (+ match (string-length old))
		(cons* new (substring str upto match) lst))
	  (string-concatenate-reverse (cons (substring str upto) lst))))))

(define (mx-body->last-update-adate-time body)
  (cond
   ;; English page
   ;; "Last update: June 15, 2006 17:00 Montr&eacute;al time"
   ;; "Last update: Dec. 12, 2006 17:00 Montr&eacute;al time"
   ;;
   ((regexp-exec-C #,(regexp "Last update: ([A-Za-z.]+) ([0-9]+), ([0-9][0-9][0-9][0-9]) ([0-9]+:[0-9]+) Montr") body)
    => (lambda (m)
	 (list (ymd->adate (string->number (match:substring m 3))
			   (Mmm-str->month (string-delete (match:substring m 1)
							  #\.))
			   (string->number (match:substring m 2)))
	       (h:m:s-str->seconds (match:substring m 4)))))

   ;; French page
   ;; old: "Dernier mise  jour : 20 juin 2006 17 h 00 heure de Montral"
   ;; "Dernire mise  jour : 12 dc. 2006 17 h 00 heure de Montral"
   ((regexp-exec-C #,(regexp "Dernire mise  jour : ([0-9]+) ([A-Za-z].+) ([0-9][0-9][0-9][0-9]) ([0-9]+) h ([0-9]+) heure") body)
    => (lambda (m)
	 (list (ymd->adate (string->number (match:substring m 3))
			   (Mmm-fr-str->month
			    (string-delete (match:substring m 2) #\.))
			   (string->number (match:substring m 1)))
	       (hms->seconds (string->number (match:substring m 4))
			     (string->number (match:substring m 5))
			     0))))

   (else
    (error "MX: quote page: cannot find last update time"))))

;; return a `latest' record
;; SYMBOL is used for the commodity and suffix
;;
(define (mx-body->latest symbol headers body)

  (set! body (string-substitute/global body "&nbsp;" " "))

  (cond
   ;;    ((or (string-contains body "markets are closed")     ;; en
   ;; 	(string-contains body "la Bourse sont ferms")) ;; fr
   ;;     '())

   ((or (string-contains body "Nonvalid")    ;; en
   	(string-contains body "non valide")) ;; fr
    (latest-new-unknown symbol))

   (else
    (receive-list (adate time)
	(mx-body->last-update-adate-time body)
      (let* ((charset    (html-charset headers body))
	     ;; eg. "<div id="titre" ...>\nCGB - Ten-year Government of Canada Bond Futures</div>"
	     (m1         (must-match (regexp-exec-C #,(regexp "<div id=\"titre\"[^>]*>[ \t\n]*[A-Z0-9]+ - ([^<\n]+)") body)))
	     (name-com   (string-trim-both (match:substring m1 1)))
	     ;; eg. "<strong>June 06 ( CGB M 06 )</strong>"
	     (m2         (must-match (regexp-exec-C #,(regexp "<strong>([A-Za-z0-9 ]+)\\(") body (match:end m1))))
	     (name-mdate (string-trim-both (match:substring m2 1)))
	     (name       (string-append (charset->utf8 charset name-com)
					(_ ", ")
					(charset->utf8 charset name-mdate)))

	     (row-list   (html-table-rows body (match:end m2))))

	(define (field en-name fr-name)
	  (let ((row (or (assoc en-name row-list)
			 (assoc fr-name row-list))))
	    (or row
		(error "MX: quote page: cannot find field" name))
	    (second row)))

	;; two lots of three table elements on each row, flatten to rows of 3
	(set! row-list (append-map! (lambda (row)
				      (list (take row 3)
					    (drop row 3)))
				    row-list))

	(let ((bid    (field "Bid Price:"     "Cours achet. :"))
	      (ask    (field "Ask Price:"     "Cours vend. :"))
	      (last   (field "Last Price:"    "Dernier prix :"))
	      (change (field "Chang.:"        "Chang. :"))
	      (high   (field "High"           "Haut :"))
	      (low    (field "Low"            "Bas :"))
	      (volume (field "Vol.:"          "Vol. :"))
	      (close  (field "Closing Price:" "Prix de fermeture :"))
	      ;; (openint (field "Open Interest:" "Int. en cours :"))
	      (settle (field "Settl. Price:"  "Prix de rgl. :"))
	      (open   (field "Open:"          "Ouvert. :")))

	  (receive-list (decimals
			 bid ask last change high low close settle open)
	      ;; 0 for no trading prices, but there's always a settlement
	      (map! 0->f (strings->numbers+decimals
			  bid ask last change high low close settle open))

	    (latest-new #:symbol         symbol
			#:name           name
			#:quote-adate    adate
			#:quote-time     time
			#:bid            bid
			#:offer          ask
			#:last-adate     adate
			#:open           open
			#:high           high
			#:low            low
			#:last           (or settle close last)
			#:change         change
			#:volume         volume
			#:decimals       decimals
			#:contract-mdate (chart-symbol-mdate symbol)
			#:source         'mx))))))))

;; return latest record
(define (mx-latest-get-one symbol)
  (receive (headers body)
      (http-request
       (simple-format #f "http://www.m-x.ca/nego_cotes_in_~a.php?symbol=~a&insSymbol=~a"
		      (lang-select '(("en" "en")
				     ("fr" "fr" #:charset "ISO-8859-1")))
		      (chart-symbol-commodity symbol)
		      (chart-symbol-sans-dot symbol))
       #:want-ok #t)
    (mx-body->latest symbol headers body)))

(define (mx-latest-get symbol-list extra-list proc)
  (for-each
   (lambda (front-symbol)
     (set! symbol-list (delete front-symbol symbol-list))
     (let ((specific-symbol (mx-latest-front-month front-symbol)))
       (if (not specific-symbol)
	   (proc (list (latest-new-unknown front-symbol)))

	   (let ((latest (mx-latest-get-one specific-symbol)))
	     (proc (list latest)) ;; specific
	     (set! latest (record-copy latest))
	     (latest-symbol! latest front-symbol)
	     (proc (list latest)) ;; front
	     (set! symbol-list (delete specific-symbol symbol-list))))))

   (remove chart-symbol-mdate symbol-list))

  (for-each (lambda (symbol)
	      (proc (list (mx-latest-get-one symbol))))
	    symbol-list))

;; return current MX adate/time, with 15-minute delay
;;
;; interest rate products are early 6am to 8:05am, regular 8:20am to 3pm,
;; curb settle until 4pm
;;
;; index futures are 9:30am to 4:15pm
;;
;; don't bother trying to figure which symbol is which, just update 6am to
;; 4:15pm
;;
(define (mx-quote-adate-time symbol)
  (tm->adate-time-within (localtime (- (current-time) #,(hms->seconds 0 15 0))
				    (timezone-montreal))
			 #,(hms->seconds 6 0 0)
			 #,(hms->seconds 16 15 0)))

(latest-handler! #:selector   mx-symbol?
		 #:handler    mx-latest-get
		 #:adate-time mx-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - data
;;
;; This uses the end-of-day download page at
;;
;;     http://www.m-x.ca/nego_fin_jour_en.php
;;
;; which results in a download like the following for 12 Jun 06 to 20 Jun 06
;;
;;     http://www.m-x.ca/nego_cotes_csv.php?symbol=CGB&jj=12&mm=06&aa=06&jjF=23&mmF=06&aaF=06
;;

;; data only from 2005 onwards (as of Jan 2007 at least)
(define mx-download-start-tdate (year->tdate 2005))

;; return STR with html comments <!-- ... --> deleted
(define (html-delete-comments str)
  (let ((start (string-contains str "<!--")))
    (if start
	(let ((end (or (string-contains-after str "-->")
		       (string-length str))))
	  (html-delete-comments (string-append (string-take str start)
					       (string-drop str end))))
	str)))

(define want-heading "Date;ClassSymbol;ClassType;UnderlyingSymbol;ExtSymbol;StrikePrice;ExpiryDate;CallPut;RootSymbol;InsSymbol;InsType;UpdateTime;LastTradeTime;BidPrice;AskPrice;BidSize;AskSize;LastPrice;Volume;PrevClosePrice;NetChange;OpeningPrice;HighPrice;LowPrice;TotalValue;NbOfTrades;SettlementPrice;OpenInterest;ImpliedVolatility")

(define (mx-download-process symbol-list body)

  ;; html <!-- comments --> seen Aug 07
  (set! body (html-delete-comments body))

  ;; blank lines at start and end
  (set! body (string-trim-both body))

  ;; quotes in headings lines
  (set! body (string-delete body #\"))

  (let ((row-list (string-split-lines body)))

    ;; strip garbage "CGBfr12007-07-03" from headings lines seen in Jul07
    (set! row-list (map! (lambda (row)
			   (cond ((regexp-exec #,(regexp "^CGBfr1[0-9][0-9][0-9][0-9]-?[0-9][0-9]-?[0-9][0-9](Date.*)") row)
				  => (lambda (m)
				       (match:substring m 1)))
				 (else
				  row)))
			 row-list))

    (let ((headings (first row-list)))
      (or (string=? headings want-heading)
	  (error "MX data unrecognised headings:" headings)))
    (set! row-list (delete! want-heading row-list))

    (download-process
     #:module      (_ "MX")
     #:symbol-list symbol-list
     #:row-list
     (map! (lambda (row-str)
	     (receive-list (tdate class commodity underlying symbol
				  strike expiry-tdate call/put
				  root-symbol ins-symbol ins-type
				  update-time last-trade-time
				  bid ask bid-size ask-size last
				  volume prev-close change open high low
				  total-value num-trades settle openint
				  volatility
				  . dummy)  ;; trailing ";" extra field
		 (string-split row-str #\;)
	       (list #:commodity     commodity
		     #:tdate         (y/m/d-str->tdate tdate)
		     ;; contracts continue to appear for a few days after
		     ;; expiry, use this date to exclude
		     #:expiry-tdate  (y/m/d-str->tdate expiry-tdate)
		     #:mdate         (chart-symbol-mdate symbol)
		     #:open          open
		     #:high          high
		     #:low           low
		     #:close         settle
		     #:volume        volume
		     #:open-interest openint)))
	   row-list))))

;; latest download data available
;; today is available from 6pm montreal time (weekdays)
;;
(define (mx-available-tdate)
  (tdate-today-after 18 0 (timezone-montreal)))


(define (mx-download-url commodity start-tdate end-tdate)
  (receive-list (s-year s-month s-day)
      (tdate->ymd start-tdate)
    (receive-list (e-year e-month e-day)
	(tdate->ymd end-tdate)
      (format #f "http://www.m-x.ca/nego_cotes_csv.php?symbol=~a&jj=~2,'0d&mm=~2,'0d&aa=~2,'0d&jjF=~2,'0d&mmF=~2,'0d&aaF=~2,'0d"
		     commodity
		     s-day s-month (modulo s-year 100)
		     e-day e-month (modulo e-year 100)))))

;; of one commodity
(define (mx-download-hack symbol-list lo hi)
  (define commodity (chart-symbol-commodity (first symbol-list)))
  (download-status (_ "MX") (_ "data")
		   commodity (tdate-range-string lo hi))

  (receive (headers body)
      (http-request (mx-download-url commodity lo hi)
		    #:want-ok #t)
    (if (string-contains body "Java")
	#f

	(begin
	  ;; server indicates data not available by sending back "200 OK"
	  ;; with a Refresh to an error page
	  (or (http-headers-get headers "Refresh")
	      (mx-download-process symbol-list body))
	  #t))))

;; return the last tdate in the month containing TDATE
;; this can be the same as the given TDATE if it's already the last day
(define (tdate->end-month tdate)
  (mdate->last-tdate (tdate->mdate tdate)))

;; if a download fails with the Javascript error refresh then try again
;; going just from LO to the end of the month, followed by the balance
;;
(define (mx-download-chunk symbol-list lo hi)
  (or (mx-download-hack symbol-list lo hi)
      (let ((end-month-tdate (tdate->end-month lo)))
	(or (and (not (= end-month-tdate hi))
		 (mx-download-hack symbol-list lo end-month-tdate)
		 (mx-download-hack symbol-list (1+ end-month-tdate) hi))
	    (error "MX: date range download doesn't work properly")))))


;; 5 months of weekdays at worst 28 calendar days, is 20 weekdays
(define mx-chunk-size 100)

;; of one commodity
(define (mx-download-range symbol-list lo hi upwards)
  (do ()
      ((> lo hi))
    (if upwards
	(let ((end (min hi (+ lo (1- mx-chunk-size)))))
	  ;; don't include `mx-available-tdate' at the end of a multi-day
	  ;; range, there's something fishy with the server that this
	  ;; doesn't work
	  (if (and (= end (mx-available-tdate))
		   (not (= lo end)))
	      (set! end (1- end)))
	  (mx-download-chunk symbol-list lo end)
	  (set! lo (1+ end)))
	(let ((end (max lo (- hi (1- mx-chunk-size)))))
	  (mx-download-chunk symbol-list end hi)
	  (set! hi (1- end))))))

(define (mx-download symbol-list)
  (define avail-tdate (mx-available-tdate))
  (for-each
   (lambda (symbol-list) ;; of each commodity
     (let ((lo (max mx-download-start-tdate
		    (apply min (map download-start-tdate symbol-list))))
	   (hi avail-tdate))
       ;; others covered by this data
       (set! symbol-list (download-also symbol-list
					#:selector    mx-symbol?
					#:commodity   #t
					#:start-tdate lo))
       (mx-download-range symbol-list lo hi #t)))

   (partition-equal chart-symbol-commodity symbol-list))

  (download-name-from-latest (_ "MX") symbol-list))

(define (mx-backto want-tdate symbol-list)
  (set! want-tdate (max want-tdate (year->tdate 2005)))
  (for-each (lambda (symbol-list) ;; of each commodity
	      (mx-download-range symbol-list
				 want-tdate
				 (apply max (map download-back-tdate
						 symbol-list))
				 #f))
	    (partition-equal chart-symbol-commodity symbol-list)))

(download-handler! #:selector   mx-symbol?
		   #:update     mx-download
		   #:avail-proc mx-available-tdate
		   #:backto     mx-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any mx-symbol? symbol-list)
			      (download-now-commodity-all-months))))
