;; Normalized average true range (NATR).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart natr)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart atr)
  #:use-module (chart indicator)
  #:use-module (chart misc))

;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/052006/TradersTips/TradersTips.html
;;     TASC Trader's Tips May 2006, various formulas.
;;
;; http://www.theessentialsoftrading.com/Blog/index.php/category/technical-analysis/
;;     Blog by John Forman, sample of S&P500 back to 1980.
;;
;; http://www.trade2win.com/knowledge/articles/general_articles/average-true-range-indicator/
;;     Article by John Forman (2 pages), sample of monthly S&P 500 from 1998
;;     to 2007 comparing ATR to NATR (seems to be 14-period smoothing).
;; http://www.trade2win.com/knowledge/articles/general_articles/average-true-range-indicator/page2
;;     Continuing that article, sample of monthly S&P 500 from 1986 to 2006
;;     comparing ATR to NATR (again seems to be 14-period smoothing).
;;
;; http://www.theessentialsoftrading.com/
;;     John Forman's web site.
;;


;; Return a procedure which calculates a normalized average true range over
;; an accumulated exponential moving average of COUNT days.
;;
;; Each call (PROC high low close) enters a new day into the window, and the
;; return is the NATR for that day, or #f if not enough points yet.
;;
;; HIGH and/or LOW can be #f in each call, in which case the CLOSE is used.
;; (CLOSE cannot be #f.)
;;
;; A NATR is in theory influenced by all preceding data, but
;; `atr-warmup-count' (in atr.scm) is designed to determine a warmup count.
;;
(define-public (natr-calc-proc count)
  (define atr-proc (atr-calc-proc count))

  (lambda (high low close)
    (and-let* ((atr (atr-proc high low close)))
      (and (not (zero? close))
	   (* 100.0 (/ atr close))))))


(define (natr-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (warmup (atr-warmup-count count))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo
					   #:count  warmup))
	 (array  (series-array parent start hi))
	 (proc   (natr-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 1)    ;; high
		(array-ref array i 2)    ;; low
		(array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array
		      (proc (array-ref array i 1)  ;; high
			    (array-ref array i 2)  ;; low
			    (array-ref array i 3)) ;; close
		      i)))))

(define-public (natr-series series count)
  (ohlcv-required "NATR" series)
  (let ((new-series (series-copy series)))
    (series-datatype!  new-series natr-datatype)
    (series-name!      new-series (format #f "~a ~a" (_ "NATR") count))
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   natr-fill (atr-warmup-count count)))
    (series-parent!    new-series series)
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public natr-datatype
  (datatype-new!
   #:key     'natr
   #:name    (_ "NATR - Normalized ATR")
   #:input   datatype-ohlcv
   #:units   'natr ;; percentage, but only a small one
   #:minimum 0
   #:manual  (_ "CONTEXT(manual-node): Normalized ATR")
   #:proc    natr-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'atr-days  ;; shared with atr.scm
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
