;; New Zealand Stock Exchange (NZX) setups.

;; Copyright 2004, 2005, 2006, 2007, 2008, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart newzealand)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart misc)
  #:use-module (chart yahoo))


(define-public (newzealand-symbol? symbol)
  (string-suffix? ".NZ" symbol))

;; http://au.finance.yahoo.com/nzindices
;;     ^NZ50
;;     ^NZC50
;;     ^NZ10G
;;     ^NZMGC
;;     ^NZGI 
;;     ^NZSCG
;;     ^NZ50G
;;     ^NZ30G
(define (yahoo-index-symbol-newzealand? symbol)
  (string-prefix? "^NZ" symbol))

(define-public timezone-newzealand
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "New Zealand")
		      '("Pacific/Auckland"
			;; commencing 2007/8 summer: last Sun in Sep to 1st
			;; Sun in Apr
			"NZST-12NZDT,M9.5.0,M4.1.0/3")
		      "NST-12"))))

(symbol-timezone! newzealand-symbol?
		  (_ "New Zealand") timezone-newzealand)
(symbol-timezone! yahoo-index-symbol-newzealand?
		  (_ "New Zealand") timezone-newzealand)
(source-help! (list newzealand-symbol? yahoo-index-symbol-newzealand?)
	      (_ "CONTEXT(manual-node): New Zealand Stock Exchange"))

;; See http://www.nzx.com/nzxmarket/trading_hours
;; pre-open 9am-10am, trading 10am-5pm, adjust 5:30pm, then enquiry
;; NZAX shares are only 10am to 4:30pm, but don't worry about that
(yahoo-quote-lock! newzealand-symbol?
		   #,(hms->seconds 9 0 0) #,(hms->seconds 17 30 0))

;; Yahoo index values based on last trades, so they should only update
;; during trading 10am to 5pm.
(yahoo-quote-lock! yahoo-index-symbol-newzealand?
		   #,(hms->seconds 10 0 0) #,(hms->seconds 17 0 0))

(yahoo-quote-delay! newzealand-symbol? 20)


;;-----------------------------------------------------------------------------
;; weblink - NZX company info

(weblink-handler! newzealand-symbol?
  (lambda (symbol)
    (list (_ "NZX _Company Information")
	  (_ "Open web browser at the New Zealand Stock Exchange page for this stock")
	  (string-append "http://www.nzx.com/market/security_details/by_security?code=" (url-quote (chart-symbol-sans-dot symbol))))))


;;-----------------------------------------------------------------------------
;; dividends
;;
;; This uses the dividend page at
;;
;;     http://www.nzx.com/markets/upcoming_dividends
;;
;; There's no Eag or Last-Modified to avoid downloading when unchanged.  We
;; check every 6 calendar days (see below).


;; extract the symbol from the name and link in the dividend table entry
(define (dividend-symbol-extract str)
  (string-append (match:substring (must-match
				   (string-match "code=([A-Z]+)\"" str)) 1)
		 ".NZ"))

(define (newzealand-dividend-parse amount imputation)
  (define orig amount)
  (define orig-imputation imputation)

  (set! amount (string-collapse-whitespace amount))

  ;; lose currency if NZ dollars, but keep AUD or GBP
  (if (string-suffix? " NZD" amount)
      (set! amount (string-drop-right amount 4)))

  ;; lose imputation if zero
  (if (0-string? imputation)
      (set! imputation #f))

  (or (string->dividend-amount amount
			       #:imputation imputation
			       #:extra-decimals 2) ;; amounts are in cents

      ;; can't parse, just show strings, with imputation shown only if not
      ;; zero
      (if imputation
	  (string-append orig " + " orig-imputation)
	  orig)))

(define (process-dividends symbol-list body)
  (define row-list (html-table-rows
		    body (string-contains body "<h1>Upcoming Dividends</h1>")))

  (let ((headings (first (html-rows-text-extract (list (first row-list))))))
    (set! headings (map string-collapse-whitespace headings))
    (if (not (equal? '("Code" "Company" "Period" "Ex Date" "Amount*"
		       "Currency" "Imputation Credit*" "Payable")
		     headings))
	(error "NZX: dividends headings unrecognised" headings)))
  (set! row-list (cdr row-list))

  ;; drop footnote row
  (set! row-list (delete! '("* cents per share") row-list))

  (for-each
   (lambda (row)
     (receive-list (symbol name period ex amount currency imputation pay)
	 row
       (set! symbol (string-append symbol ".NZ"))
       (if (member symbol symbol-list)
	   (begin
	     (download-status (_ "NZX") (_ "dividends") symbol)

	     (set! symbol-list (delete symbol symbol-list))
	     (set! ex (d/m-str->tdate-nearest ex))

	     (if (not (equal? currency "NZD"))
		 (set! amount (string-append amount " " currency)))

	     (c-idle) ;; pause between writes
	     (database-call-modify-data symbol
	       (lambda (series)
		 (series-dividend-add!
		  series ex (newzealand-dividend-parse amount imputation))
		 (series-info-add! series 'newzealand-dividends-time
				   (current-time))
		 (series-info-add! series 'newzealand-dividends-copyright
				   "http://www.nzx.com/terms_of_use")))))))
   row-list)

  ;; record dividend coverage date for remaining symbols
  (for-each (lambda (symbol)
	      (download-status (_ "NZX") (_ "dividends") symbol)
	      (c-idle) ;; pause between writes
	      (database-call-modify-data symbol
		(lambda (series)
		  (series-info-add! series 'newzealand-dividends-time
				    (current-time)))))
	    symbol-list))

(define (dividends-update symbol-list)

  ;; update every 6 days, or if not previously fetched
  (define target-time (- (current-time) (* 6 86400)))
  (define (want-update? symbol)
    (let ((last-time (database-read-info symbol 'newzealand-dividends-time)))
      (or (not last-time)
	  (< last-time target-time))))

  (if (any want-update? symbol-list)
      (begin
	(download-status (_ "NZX") (_ "dividends"))

	(receive (headers body)
	    (http-request "http://www.nzx.com/markets/upcoming_dividends"
			  #:want-ok #t)
	  (process-dividends (download-also symbol-list
					    #:selector newzealand-symbol?)
			     body)))))


;;-----------------------------------------------------------------------------
;; setups

;; low priority so data is fetched first
(download-handler! #:selector newzealand-symbol?
		   #:update   dividends-update
		   #:backto   noop
		   #:priority -10) ;; after data downloads
