;; OHLC, HL and candle graph drawing.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ohlc-draw)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gdk)
  #:use-module (chart graph)
  #:use-module (chart indicator)
  #:use-module (chart line)        ;; series-next-close-pos
  #:use-module (chart misc)
  #:use-module (chart view))


;; when graph realized
(define up-gc        #f)
(define down-gc      #f)
(define unchanged-gc #f)

(define (init graph)
  (define win (gtk-widget-window graph))

  (set! up-gc (c-gdk-gc-get
	       win #:foreground graph-up-colour))
  (set! down-gc (c-gdk-gc-get
		 win #:foreground graph-down-colour))
  (set! unchanged-gc (c-gdk-gc-get
		      win #:foreground graph-unchanged-colour))

  ;; once only, and gc this code
  (set! init noop))


;; This is a combined candles, OHLC and HL figure drawing.  These go
;; together because for candles we draw an ohlc when there's no open, and
;; because for all of them the line segments drawing on close-only data (or
;; portions of the data) is the same.

(define (ohlc/candles-draw graph gc series want-candles hl-only)
  (init graph)

  ;; past window width to show partial lines extending away
  (let* ((lo            (series-prev-n-closes-pos
			 series
			 #:before (graph-draw-x-lo graph series)
			 #:count 1))
	 (hi            (series-next-close-pos
			 series #:after (graph-draw-x-hi graph series)))
	 (array         (series-array series lo hi))
	 (win           (gtk-widget-window graph))
	 (scale-x       (graph-scale-x-proc graph))
	 (scale-y       (graph-scale-y-proc graph))

	 ;; high/low prices of visible window
	 (win-p-hi   (graph-y->price-inexact graph -1))
	 (win-p-lo   (graph-y->price-inexact
		      graph (c-gdk-drawable-get-height win)))

	 ;; use 3/4 of step
	 (shadow-width  (max 1 (floor-exact (* 0.25 scale-x-step))))
	 (block-width   (min scale-x-step (* 3 shadow-width)))

	 ;; block centred in step
	 (block-offset  (quotient (- scale-x-step block-width) 2))

	 ;; shadow centred in block
	 (shadow-offset (+ block-offset
			   (quotient (- block-width shadow-width) 2)))

	 ;; half way into block, for when drawn alone (no high/low)
	 (open-width    (max 1 (quotient-ceiling block-width 2)))
	 (close-width   (max 1 (quotient-ceiling block-width 2)))

	 ;; left of block
	 (open-offset   block-offset)
	 ;; right of block
	 (close-offset  (+ block-offset (- block-width close-width)))

	 (centre-offset (quotient scale-x-step 2))

	 ;; placeholder for something to make the open/close horizontals
	 ;; more than 1 pixel, when a good notion of y scale comes about
	 (horiz-height  1)
	 (horiz-offset  (quotient horiz-height 2))

	 (poly          '())
	 (dash-x        #f)
	 (dash-y        #f)
	 (dash-gc       (c-gdk-gc-get win #:dashed #t)))

    (define (flush-poly)
      (cond ((null? poly))

	    ((null? (cdr poly))
	     ;; one point, draw as OHLC close
	     (gdk-draw-rectangle win gc #t
				 (+ (caar poly) close-offset (- centre-offset))
				 (- (cdar poly) horiz-offset)
				 close-width horiz-height)
	     (set! dash-x #f)
	     (set! poly '()))

	    (else ;; two or more points
	     (let ((end-x (caar poly))
		   (end-y (cdar poly)))
	       (set! poly (reverse! poly))
	       ;; dashed line from end of previous poly to start of this
	       (if dash-x
		   (gdk-draw-line win dash-gc dash-x dash-y
				  (caar poly) (cdar poly)))
	       (set! dash-x end-x)
	       (set! dash-y end-y))
	     (gdk-draw-lines win gc poly)
	     (set! poly '()))))

    ;; Draw a rectange X,WIDTH pixels, between prices P-LO and P-HI.
    ;; If the prices are completely outside the window then draw nothing.
    ;; Prices are truncated to the window range so huge values don't
    ;; overflow the "gint" parameters to gdk-draw-rectangle.
    ;;
    (define (rectangle gc x width p-lo p-hi)
      (if (overlap-inclusive? p-lo p-hi win-p-lo win-p-hi)
	  (let ((y-lo (scale-y (min win-p-hi p-hi)))
		(y-hi (scale-y (max win-p-lo p-lo))))
	    (gdk-draw-rectangle win gc #t x y-lo
				width (1+ (- y-hi y-lo))))))

    (do ((i lo           (1+ i))
	 (x (scale-x lo) (+ x scale-x-step)))
	((> i hi))
      (let ((o (array-ref array i 0))
	    (h (array-ref array i 1))
	    (l (array-ref array i 2))
	    (c (array-ref array i 3)))

	(cond
	 ((and want-candles
	       o
	       c)
	  ;; open and close available, draw candle
	  (flush-poly)
	  (set! dash-x #f)

	  (let ((gc (cond ((> c o) up-gc)
			  ((= c o) unchanged-gc)
			  (else    down-gc))))

	    ;; open to close real body
	    (rectangle gc (+ x block-offset) block-width (min o c) (max o c))

	    ;; high to low shadow, and if one of them missing then use
	    ;; open/close (if both missing then no need to draw at all, it's
	    ;; only the real body, but don't worry about that)
	    (rectangle gc (+ x shadow-offset) shadow-width
		       (or l (min o c))
		       (or h (max o c)))))

	 ((and (not o)
	       (not h)
	       (not l))
	  (if c
	      ;; close-only, extend poly to draw line segments
	      (set! poly (acons (+ x centre-offset) (scale-y c) poly))
	      ;; nothing at all, gap
	      (flush-poly)))

	 (hl-only
	  ;; HL blocks
	  (flush-poly)
	  (set! dash-x #f)

	  ;; Use min/max to cope with open/close only rather than high/low.
	  ;; One of open/high/low is not #f (though a close-only has been
	  ;; handled as a line).
	  (rectangle gc (+ x shadow-offset) shadow-width
		     (min-maybe o h l c)
		     (max-maybe o h l c)))

	 (else
	  ;; not both open/close, OHLC style figure
	  (flush-poly)
	  (set! dash-x #f)

	  (if (and h l)
	      (rectangle unchanged-gc (+ x shadow-offset) shadow-width
			 l h)
	      (begin
		;; cope with only one of high/low by drawing little blocks
		(if l
		    (rectangle unchanged-gc (+ x shadow-offset) shadow-width
			       l l))
		(if h
		    (rectangle unchanged-gc (+ x shadow-offset) shadow-width
			       h h))))
	  (if o
	      (rectangle gc (+ x open-offset) open-width o o))
	  (if c
	      (rectangle gc (+ x close-offset) open-width c c))))))

    (flush-poly)))


(define-public (ohlc-draw graph gc series)
  (ohlc/candles-draw graph gc series #f #f))

(define-public (candles-draw graph gc series)
  (ohlc/candles-draw graph gc series #t #f))

(define-public (hl-draw graph gc series)
  (ohlc/candles-draw graph gc series #f #t))
