;; OHLCV datatype.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ohlcv)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))



;; `ohlcv-initial-page' returns a list (LOWER UPPER) which is the initial
;; Y-axis page range to show for dates LO to HI.  This is for use both with
;; `datatype-ohlcv' SERIES and also any other series without its own
;; specific style.
;;
;; As described in "Main Window" in the manual, the aim is to scale
;; according to apparent volatility, so that daily range or daily
;; close-to-close change are some modest fraction of the initial page.  In
;; particular if the series is just going sideways it's not zoomed out to
;; try to fill the whole screen.  The absolute price level is not used, so
;; say bond prices which hover near 100 still get scaled out to make typical
;; daily changes of 0.1 or whatever visible.
;;
;;
(define-public (ohlcv-initial-page series lo hi)
  (let* ((datatype (series-datatype series))
	 (array    (series-array series lo hi))
	 (lst      '())
	 (page     0))

    ;; high to low ranges in ohlcv
    (if (eq? datatype-ohlcv datatype)
	(begin
	  (do ((i lo (1+ i)))
	      ((> i hi))
	    (if (and (array-ref array i 1)
		     (array-ref array i 2))
		(set! lst (cons (- (array-ref array i 1)
				   (array-ref array i 2))
				lst))))
	  (set! array (make-shared-array-column array 3))))

    ;; ENHANCE-ME: look at all parts of a multi-line like macd, bollinger,
    ;; guppy, etc
    (if (= 2 (array-rank array))
	;; macd, not quite right
	(set! array (make-shared-array-column array 0)))

    ;; close to close ranges
    (let ((prev #f))
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (begin
	      (if prev
		  (set! lst (cons (abs (- (array-ref array i) prev))
				  lst)))
	      (set! prev (array-ref array i))))))

    ;; no change or no range give us nothing
    (set! lst (delete 0 lst =))

    ;; page will show 25x the median range
    (if (not (null? lst))
	(begin
	  (set! lst (sort! lst <))
	  (set! page (* 25 (list-ref-fraction lst 0.5)))))

    ;; make page no more than twice upper,lower data range, so the
    ;; data is not less than half the window
    (receive-list (l h)
	(series-range series lo hi)
      (if l
	  (set! page (min page (* 2 (- h l))))))

    ;; make page no smaller than last 1/2 of data, so that's visible
    (receive-list (l h)
	(series-range series
		      (- hi (round-exact (* 0.5 (- hi lo)))) hi)
      (if l
	  (set! page (max page (- h l)))))

    (let* ((array-closes (series-array-closes series lo hi))
	   (lst-closes   (delq! #f (array->list (make-shared-array-restrict
						 array-closes lo hi)))))
      (if (null? lst-closes)
	  (list 0 1)

	  (let* ((med (list-ref-fraction lst-closes 0.5))
		 (lower (- med (* 0.5 page))))

	    ;; don't go below `datatype-minimum' (if present), so long as
	    ;; the actual data respects that minimum
	    (and-let* ((minimum    (datatype-minimum datatype))
		       (actual-min (apply min-maybe lst-closes))
		       (           (>= actual-min minimum)))
	      (set! lower (max lower minimum)))

	    (list lower
		  (+ lower page)))))))
    




;; return an empty array of the same dimensions as ARRAY, but with the first
;; index reduced with INDEX-PROC
;; INDEX-PROC is called (INDEX-PROC lo) and (INDEX-PROC hi) and the two
;; values it returns are the new lo/hi
;;
(define (array-reduction-new array index-proc)
  (let ((shape (array-shape array)))
    (apply make-array #f (list (index-proc (caar shape))
			       (index-proc (cadar shape))) (cdr shape))))

;; return (+ X Y), ignoring #f, return #f if both #f
(define (+-not-f x y)
  (if x
      (if y
	  (+ x y)
	  x)
      y))

(define (datatype-ohlcv-array-reduction array index-proc)
  (let ((open-interest? (= 5 (array-last-column array)))
	(hi             (array-last-index array))
	(new-array      (array-reduction-new array index-proc)))
    (do ((i (array-first-index array) (1+ i)))
	((> i hi))
      (let ((new-i (index-proc i)))

	;; open gets first open in period, or first close if no opens
	(if (not (array-ref new-array new-i 0))
	    (array-set! new-array (or (array-ref array i 0)  ;; open
				      (array-ref array i 3)) ;; close
			new-i 0))

	;; high is max high in period, or max close if no high data values
	(array-set! new-array (max-maybe (array-ref new-array new-i 1)
					 (array-ref array i 1)  ;; high
					 (array-ref array i 3)) ;; close
		    new-i 1)

	;; low is min low in period, or min close if no low data values
	(array-set! new-array (min-maybe (array-ref new-array new-i 2)
					 (array-ref array i 2)  ;; low
					 (array-ref array i 3)) ;; close
		    new-i 2)

	;; close is last non-#f close in period
	(if (array-ref array i 3)
	    (array-set! new-array (array-ref array i 3)
			new-i 3))

	;; volume is sum of all volumes in period, keeping #f if all #f there
	(array-set! new-array (+-not-f (array-ref new-array new-i 4)
				       (array-ref array i 4))
		    new-i 4)

	;; openint is  last non-#f openint
	(and open-interest?
	     (array-ref array i 5)
	     (array-set! new-array (array-ref array i 5)
			 new-i 5))))

    new-array))


(datatype-new! #:key       'ohlcv
	       #:linestyle 'ohlc
	       #:units     'price
	       #:minimum   0  ;; normally
	       #:reduce    datatype-ohlcv-array-reduction
	       #:closes-apply-arg '(3))
