;; OLE file reading.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Crib notes:
;;
;; catdoc xls2csv program includes ole reading
;; http://www.45.free.net/~vitus/ice/catdoc
;;
;; laola perl OLE::Storage module
;; http://wwwwbs.cs.tu-berlin.de/~schwartz/pmh/guide.html
;;
;; cole C OLE library in xlhtml
;; http://chicago.sourceforge.net/xlhtml
;;
;; libole2
;; http://arturo.directmail.org/filtersweb


(define-module (chart ole)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 rw)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; read-string!/partial on non-file ports

(let-toplevel ((handler-prop
		;; default to the normal `read-string!/partial'
		(primitive-make-property (lambda (prop port)
					   read-string!/partial))))

  ;; set a `c-read-string!/partial' handler PROC for PORT
  (define-public (c-read-string!/partial-handler! port proc)
    (primitive-property-set! handler-prop port proc))

  ;; like `read-string!/partial' but with handlers on extra port types
  (define*-public (c-read-string!/partial str #:optional
					  (port (current-input-port))
					  (start 0)
					  (end (string-length str)))
    ((primitive-property-ref handler-prop port) str port start end)))

;; Return an input string port for STR, with a `c-read-string!/partial'
;; handler.
;;
;; On an input string port there's no way to get at a whole block of data,
;; so read-str uses the original given string.  This probably means two
;; copies of the data is held (one in the port buffer and one in the string).
;;
(define (c-open-input-string str)
  (define (read-str buf port start end)
    (cond ((>= start end)
	   0)
	  ((>= (ftell port) (string-length str))
	   #f) ;; eof
	  (else
	   (let ((n (min (- end start)
			 (- (string-length str) (ftell port)))))
	     (string-copy! buf start str (ftell port) (+ (ftell port) n))
	     (seek port n SEEK_CUR)
	     n))))
  (let ((port (open-input-string str)))
    (c-read-string!/partial-handler! port read-str)
    port))


;;-----------------------------------------------------------------------------
;; misc

;; return a short (2-byte) signed integer from the bytes in STR at OFFSET
(define*-public (bytes->short str #:optional (offset 0))
  (+ (char->integer (string-ref str offset))
     (* 256 (char->integer (string-ref str (1+ offset))))
     (if (char>? (string-ref str (1+ offset)) #\del)
	 -65536
	 0)))

;; return a long (4-byte) signed integer from the bytes in STR at OFFSET
(define*-public (bytes->long str #:optional (offset 0))
  (+ (char->integer (string-ref str offset))
     (* 256 (char->integer (string-ref str (1+ offset))))
     (* 65536 (char->integer (string-ref str (+ 2 offset))))
     (* 16777216 (char->integer (string-ref str (+ 3 offset))))
     (if (char>? (string-ref str (+ 3 offset)) #\del)
	 -4294967296
	 0)))

;; Return a string of N bytes read from PORT.
;;
;; If eof is reached then the return may be less than N bytes.  If there are
;; no bytes left then an eof object is returned, except that if N==0 the
;; return is always just an empty string, without checking for eof.
;;
;; The implementation can cope with c-read-string!/partial returning less
;; than actually requested (like the OLE soft ports do).
;;
(define-public (read-bytes port n)
  (let ((str (make-string n)))
    (let more ((upto 0))
      (if (>= upto n)
	  str
	  (let ((got (c-read-string!/partial str port upto)))
	    (if got
		(more (+ upto got))
		(if (= 0 upto)
		    eof-object
		    (string-take str upto))))))))

;; read a 2-byte signed integer from PORT at file position POS
(define (read-short port pos)
  (seek port pos SEEK_SET)
  (bytes->short (read-bytes port 2)))

;; read a 4-byte signed integer from PORT at file position POS
(define (read-long port pos)
  (seek port pos SEEK_SET)
  (bytes->long (read-bytes port 4)))

;; return ascii from a utf16 string STR
(define (utf16->ascii str)
  (string-delete str #\nul))

;; return a procedure to be called (READCHAR) ready for `make-soft-port',
;; implemented using the `read-string!/partial' style PROC on PORT
;;
(define (read-string!/partial->read-char port proc)
  (lambda ()
    (let* ((str (make-string 1))
	   (got (proc str port 0 1)))
      (if got
	  (string-ref str 0)
	  eof-object))))


;;-----------------------------------------------------------------------------
;; properties

;; return a soft port which is the properties section of ole PORT
;;
(define (ole-open-properties port)

  (let* ((big-block-size  (ash 1 (read-short port #x1e)))
	 (blocknum        (read-short port #x30)) ;; first properties block
	 (offset          0))

    ;; return the file position of block N in the bbd
    (define (bbd-ref port n)
      (let* ((entries-per-block (quotient big-block-size 4))
	     (block             (quotient n entries-per-block))
	     (rem               (remainder n entries-per-block)))
	(read-long port (+ (* big-block-size
			      (1+ (read-long port (+ #x4c (* 4 block)))))
			   (* 4 rem)))))

    ;; read until a negative blocknum is reached, which is the end
    (define (prop-read-string!/partial str soft-port start end)
      (if (>= start end)
	  0
	  (begin
	    ;; new block if this one exhausted
	    (if (>= offset big-block-size)
		(begin
		  (set! offset 0)
		  (set! blocknum (bbd-ref port blocknum))
		  (if (positive? blocknum)
		      (seek port (* big-block-size (1+ blocknum)) SEEK_SET))))

	    (if (positive? blocknum)
		;; read bytes; no more than left in this block
		(let ((got (c-read-string!/partial
			    str port start (min end
						(+ start (- big-block-size offset))))))
		  ;; got won't be eof, assuming blocks are correct
		  (set! offset (+ offset got))
		  got)
		#f)))) ;; eof

    (if (positive? blocknum)
	(seek port (* big-block-size (1+ blocknum)) SEEK_SET))

    (let ((soft-port (make-soft-port
		      (vector #f ;; write-char
			      #f ;; write-string
			      #f ;; flush-output
			      (read-string!/partial->read-char
			       #f prop-read-string!/partial)
			      #f)
		      "r")))
      (c-read-string!/partial-handler! soft-port prop-read-string!/partial)
      soft-port)))


;;-----------------------------------------------------------------------------
;; part

(define ole-file-intro (list->string
			(map integer->char
			     '(#xD0 #xCF #x11 #xE0 #xA1 #xB1 #x1A #xE1))))

;; return a soft port which is PARTNAME of ole FILENAME
(define-public (ole-open-port port partname)

  (or (string=? (read-bytes port (string-length ole-file-intro))
		ole-file-intro)
      (error "not an ole file"))

  (let* ((big-block-size  (ash 1 (read-short port #x1e)))
	 (blocknum        #f)
	 (offset          0)
	 (remaining       #f))

    ;; return the file position of block N in the bbd
    (define (bbd-ref port n)
      (let* ((entries-per-block (quotient big-block-size 4))
	     (block             (quotient n entries-per-block))
	     (rem               (remainder n entries-per-block)))
	(read-long port (+ (* big-block-size
			      (1+ (read-long port (+ #x4c (* 4 block)))))
			   (* 4 rem)))))

    (define (part-read-string!/partial str soft-port start end)
      (cond ((>= start end)
	     0)
	    ((= 0 remaining)
	     #f) ;; eof
	    (else
	     ;; new block if this one exhausted
	     (if (>= offset big-block-size)
		 (begin
		   (set! blocknum (bbd-ref port blocknum))
		   (seek port (* big-block-size (1+ blocknum)) SEEK_SET)
		   (set! offset 0)))
	     ;; read bytes; no more than left in this block or total remaining
	     (let ((got (c-read-string!/partial
			 str port
			 start (min end
				    (+ start remaining)
				    (+ start (- big-block-size offset))))))
	       ;; got won't be eof, assuming `remaining' and blocks are correct
	       (set! remaining (- remaining got))
	       (set! offset (+ offset got))
	       got))))

    ;; find the PARTNAME desired
    (let ((prop (ole-open-properties port)))
      (let more ()
	(let* ((dirent (read-bytes prop 128)))
	  (cond ((eof-object? dirent)
		 (error "ole part not found:" partname))
		((string=? partname
			   (utf16->ascii
			    (substring dirent 0 (bytes->short dirent #x40))))
		 (set! blocknum  (bytes->long dirent #x74))
		 (set! remaining (bytes->long dirent #x78)))
		(else
		 (more))))))

    (if (positive? remaining)
	(seek port (* big-block-size (1+ blocknum)) SEEK_SET))

    (let ((soft-port (make-soft-port
		      (vector #f ;; write-char
			      #f ;; write-string
			      #f ;; flush-output
			      (read-string!/partial->read-char
			       #f part-read-string!/partial)
			      #f)
		      "r")))
      (c-read-string!/partial-handler! soft-port part-read-string!/partial)
      soft-port)))


;; return a soft port which is PARTNAME of ole STR
(define-public (ole-open-string str partname)
  (ole-open-port (c-open-input-string str) partname))
