;; Open dialog.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart open)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart database)
  #:autoload   (chart down-dialog)  (download-dialog-one-query)
  #:use-module (chart extra-gtk)
  #:use-module (chart gui-common)
  #:use-module (chart main)
  #:use-module (chart misc)
  #:use-module (chart symlists)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; misc

;; drop items from the front of LST until the first satisfies PRED
(define (drop-until pred lst)
  (if (or (null? lst)
	  (pred (first lst)))
      lst
      (drop-until pred (cdr lst))))

;; drop items on the right of LST until the last satisfies PRED, modifying LST
(define (drop-right-until! pred lst)
  (reverse! (drop-until pred (reverse! lst))))


;;-----------------------------------------------------------------------------
;; GtkTree

;; setup for N labels in LST (a list widget)
;; call (PROC item) on each newly created item
;;
;; when using existing item widgets, those up to N are shown and those above
;; are hidden
;;
(define (c-gtk-tree-require-n-items tree n proc)
  (do ((i (length (gtk-container-children tree)) (1+ i)))
      ((>= i n))
    (let ((item (gtk-tree-item-new-with-label "")))
      (proc item)
      (gtk-tree-append tree item)))
  (let ((children (gtk-container-children tree)))
    (for-each gtk-widget-show (take children n))
    (for-each gtk-widget-hide (drop children n))))


;;-----------------------------------------------------------------------------
;; GtkTreeItem extras

;; return the parent item of ITEM, ie. the item whose subtree contains ITEM
(define (c-gtk-tree-item-parent item)
  (let* ((subtree (gtk-widget-get-parent item))
	 (tree    (gtk-widget-get-parent subtree)))
    (find (lambda (tree-item)
	    (eq? subtree (gtk-tree-item-subtree tree-item)))
	  (gtk-container-children tree))))

;; toggle a GtkTreeItem between expanded and collapsed, and don't select it
(define (c-gtk-tree-item-toggle-expand item)
  (if (gtk-tree-item-expanded item)
      (gtk-tree-item-collapse item)
      (gtk-tree-item-expand item)))
;; (gtk-tree-unselect-child (gtk-widget-parent item) item)


;;-----------------------------------------------------------------------------
;; GtkWidget stuff

;; C macro not wrapped
(define (gtk-widget-visible widget)
  (->bool (memq 'visible (gtk-widget-flags widget))))


;;-----------------------------------------------------------------------------
;; file dialog

(define dialog   #f)
(define tree     #f)
(define entry    #f)
(define notfound #f)
(define vadj     #f)

;; used on top tree items, value is a symlist
(define top-item-symlist (make-object-property))

;; used on subtree items, a pair (SYMBOL . INITIALIZED)
(define tree-item-info    (make-object-property))

(define (handle-goto symbol-wanted symbol-got)
  (if symbol-got
      (begin
	(gtk-widget-hide dialog)
	(chart-open symbol-got))
      (gtk-widget-show notfound)))

(define (tree-select-child item)
  (let ((symbol (car (tree-item-info item))))
    (gtk-entry-set-text entry symbol)
    (handle-goto symbol 
		 (smarker-goto (view-smarker outer)
			       (top-item-symlist (c-gtk-tree-item-parent item))
			       symbol))))

(define (init)
  (set! dialog (gtk-dialog-new-with-buttons (c-make-window-title (_ "Open"))
					    #f '()))

  (gtk-dialog-add-buttons dialog
			  "gtk-open"   0
			  "gtk-new"    1
			  "gtk-cancel" gtk-response-cancel
			  "gtk-help"   gtk-response-help)
  (gtk-dialog-set-response-sensitive dialog 1 #f) ;; initial "New" insensitive

  (set! tree  (gtk-tree-new))
  (set! entry (gtk-entry-new))
  (set! notfound (gtk-label-new (_ "Not in database, click New to download.\n(Be sure capitalization is right for download.)")))
  (gtk-label-set-justify notfound 'center)

  (set! vadj (gtk-adjustment-new 0 0 0 0 0 0))
  (gtk-signal-connect vadj "changed" fill-window-items)
  (gtk-signal-connect vadj "value-changed" fill-window-items)

  ;; page-up and page-down keys
  ;; these are probably better done as GtkBindings, or maybe accelerators
  (gtk-widget-add-events dialog '(key-press-mask))
  (gtk-signal-connect dialog "key-press-event"
    (lambda (event)
      ;; no modifiers except caps-lock
      (if (member (gdk-event-state event) '(() (lock-mask)))
	  (cond ((= (gdk-event-keyval event)
		    (gdk-keyval-from-name "Page_Up"))
		 (c-gtk-adjustment-page-down vadj))
		((= (gdk-event-keyval event)
		    (gdk-keyval-from-name "Page_Down"))
		 (c-gtk-adjustment-page-up vadj))))))

  (let ((scrolled (gtk-scrolled-window-new #f vadj))
	(hbox     (gtk-hbox-new #f 4)))

    (define (entry-activate)
      (c-gtk-editable-select-region-noclip entry 0 -1)  ;; so can retype
      (let ((symbol (gtk-entry-get-text entry)))
	(handle-goto symbol
		     (smarker-goto-name (view-smarker outer)
					(gtk-entry-get-text entry)))))

    ;; buttons
    (c-gtk-dialog-hide-on-delete dialog)
    (gtk-signal-connect dialog "response"
      (lambda (id)
	(cond ((= id 0) ;; open
	       (entry-activate))

	      ((= id 1) ;; new
	       (gtk-widget-hide dialog)
	       (let ((symbol (gtk-entry-get-text entry)))
		 (set! symbol (string-trim-both symbol))
		 ;; supposed to have the "New" button inactive when no
		 ;; symbol entered, but check in case
		 (if (not (string-null? symbol))
		     (begin
		       (database-new symbol)
		       (chart-open symbol)
		       (download-dialog-one-query symbol)))))

	      ((= id gtk-response-cancel)
	       (gtk-widget-hide dialog))

	      ((= id gtk-response-help)
	       (manual-popup (_ "CONTEXT(manual-node): Open") dialog)))))

    (gtk-box-pack-start
     (gtk-dialog-vbox dialog)
     (gtk-label-new (_ "Select from the tree, or enter a symbol"))
     #f #f 0)

    ;; scrolled
    (gtk-box-pack-start (gtk-dialog-vbox dialog) scrolled #t #t 0)
    (gtk-scrolled-window-set-policy scrolled 'automatic 'always)
    (c-gtk-widget-set-usize-units scrolled '(em . 40) '(height . 15))

    ;; text entry
    (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0)
    (gtk-box-pack-start hbox (gtk-label-new (_ "Symbol")) #f #f 0)
    (gtk-box-pack-start hbox entry  #t #t 0)
    (gtk-signal-connect dialog "map_event"
      (lambda (event)
	(gtk-widget-grab-focus entry)
	;; so can retype
	(c-gtk-editable-select-region-noclip entry 0 -1)))
    (gtk-signal-connect entry "activate" entry-activate)
    (gtk-signal-connect entry "changed"
      (lambda ()
	(gtk-widget-hide notfound)
	(moveto (gtk-entry-get-text entry))

	(gtk-dialog-set-response-sensitive
	 dialog 1 ;; "New" button active when symbol entered
	 (not (string-every char-set:whitespace (gtk-entry-get-text entry))))))

    ;; show now, since tree filling might take time
    (gtk-widget-show-all dialog)
    (c-gtk-widget-busy dialog)

    ;; notfound, hidden initially
    (gtk-box-pack-start (gtk-dialog-vbox dialog) notfound #f #f 0)

    ;; tree
    (gtk-scrolled-window-add-with-viewport scrolled tree)
    (gtk-tree-set-selection-mode tree 'single)
    (gtk-signal-connect tree "select_child"
      c-gtk-tree-item-toggle-expand)

    (for-each (lambda (symlist)
		(let ((item (gtk-tree-item-new-with-label
			     (symlist-name symlist)))
		      (subtree (gtk-tree-new)))
		  (set! (top-item-symlist item) symlist)
		  (gtk-tree-append tree item)
		  (gtk-tree-item-set-subtree item subtree)
		  (gtk-tree-item-expand item)
		  (gtk-signal-connect subtree "select_child"
		    tree-select-child)
		  (gtk-widget-show subtree)))
	      symlist-all-lists)

    (gtk-widget-show-all tree))

  (let ()
    (define (freshen-once . args)
      (c-gtk-idle-once freshen))
    (notify-connect 'database-update freshen-once)  ;; name changes
    (notify-connect 'symlist-update  freshen-once)) ;; list movement

  ;; once only, and gc this code
  (set! init noop))

;; return list of items under TOP-ITEM, excluding those at the end hidden
;; after deleting etc
(define (subitem-list top-item)
  (drop-right-until! gtk-widget-visible
		     (gtk-container-children
		      (gtk-tree-item-subtree top-item))))

;; move to show and select the symbol match for STR
(define (moveto str)
  (and-let* ((      (not (string-null? str)))
	     (found (smarker-find-name (view-smarker outer) str)))
    (let ((pos 0))
      (find (lambda (top-item)
	      (let ((item-list (gtk-container-children
				(gtk-tree-item-subtree top-item))))
		(set! pos (1+ pos))
		(if (eq? (first found) (top-item-symlist top-item))
		    (find (lambda (item)
			    (set! pos (1+ pos))
			    (and (string=? (second found)
					   (car (tree-item-info item)))
				 (begin
				   (gtk-tree-unselect-child tree item)
				   (gtk-tree-select-child tree item)
				   (c-gtk-adjustment-set-value-centre
				    vadj (* pos (gtk-widget-allocation-height
						item)))
				   #t)))
			  item-list)
		    (begin
		      (if (gtk-tree-item-expanded top-item)
			  (set! pos (+ pos (length item-list))))
		      #f))))
	    (gtk-container-children tree)))))

;; set the label text for items within the window page
(define (fill-window-items)
  (let* ((top-item-list (gtk-container-children tree))
	 (pos           0)
	 (height        (gtk-widget-allocation-height (first top-item-list)))
	 (lo            (- (gtk-adjustment-value vadj) height))
	 (hi            (+ lo (gtk-adjustment-page-size vadj) height)))

    (find (lambda (top-item)
	    (set! pos (+ pos height))
	    (and (gtk-tree-item-expanded top-item)
		 (find (lambda (item)
			 ;; looking for item in adjustment window and not
			 ;; already initialized
			 (and-let* ((     (within-range? lo hi pos))
				    (pair (tree-item-info item))
				    (     (not (cdr pair))))
			   (let* ((symbol (car pair))
				  (name   (database-read-info symbol 'name)))
			     (gtk-label-set-text
			      (gtk-bin-get-child item)
			      (if name
				  (string-append symbol " - " name)
				  symbol)))
			   (set-cdr! pair #t))

			 (set! pos (+ pos height))
			 ;; top when past visible page
			 (> pos hi))

		       (subitem-list top-item))))
	  top-item-list)))

;; freshen all trees to match their symlists
(define (freshen)
  (for-each
   (lambda (top-item)
     (let* ((old-symbols (map (lambda (item)
				(car (tree-item-info item)))
			      (subitem-list top-item)))
	    (symlist     (top-item-symlist top-item))
	    (new-symbols (lset-intersection string=? (symlist-list symlist)
					    (database-symbols))))
       (if (not (equal? new-symbols old-symbols))
	   (begin
	     (c-gtk-tree-require-n-items (gtk-tree-item-subtree top-item)
		 (length new-symbols)
	       (lambda (item)
		 (set! (tree-item-info item) (cons #f #f))))
	     (for-each (lambda (item symbol)
			 ;; if the symbol for this item has changed (or new
			 ;; as #f) then set it and mark label uninitialized
			 (let ((pair (tree-item-info item)))
			   (if (not (equal? symbol (car pair)))
			       (begin
				 (set-car! pair symbol)
				 (set-cdr! pair #f)))))
		       (subitem-list top-item)
		       new-symbols)))))

   (gtk-container-children tree))
  (fill-window-items))

(define-public (open-dialog-popup)
  (init)
  (freshen)
  (c-gtk-widget-show-or-raise dialog))
