;; Pagebits caching things parsed from a web page

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart pagebits)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart extra-glib)
  #:use-module (chart http)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart misc))


;; URL is where to read from, either a string or a list of arguments for
;; `http-request' (including the url).  The latter allows "POST".
;;
;; FILENAME is within cache-directory and is where the info is held,
;; eg. "somesource-codes"
;;
;; PARSE is a funcation called (PARSE body) with the body string from URL,
;; it should return an alist of information
;;
;; TIMEZONE is the timezone to operate in, used to download just once each
;; tdate
;;
;; FREQUENCY is the number of tdates apart to recheck the info at URL, the
;; default is 1 meaning every new trading day (in TIMEZONE)
;;
;; There's versioning of the parse func, nor any record of the URL in the
;; cache file.  If there's a format change then FILENAME should be stepped
;; to something new.
;;

(define*-public (pagebits-read #:key url filename parse timezone status
			       (frequency 1) (want-download #t))

  (set! filename (g-build-filename cache-directory filename))
  (let* ((alist (call-with-input-file-or '() filename read))
	 (today (tdate-today timezone))
	 (tdate (assq-ref alist 'tdate)))

    (if (or (not tdate)                      ;; nothing cached
	    (>= (- today tdate) frequency))  ;; cached FREQ or more days ago

	;; either no old data, or it needs freshening
	(begin
	  (set! alist '())

	  (if want-download
	      (begin
		(if status
		    (apply download-status status))
		(receive (headers body)
		    (apply http-request
			   (append (if (string? url)
				       (list url)
				       url)
				   (list
				    #:want-ok           #t
				    #:if-none-match     (assq-ref alist 'etag)
				    #:if-modified-since (assq-ref alist
								  'last-modified))))

		  (if (http-headers-ok? headers)
		      (set! alist
			    `((tdate . ,today)
			      (url   . ,(if (string? url) url (first url)))
			      (etag  . ,(http-headers-get headers "ETag"))
			      (last-modified . ,(http-headers-get headers "Last-Modified"))
			      ,@(parse body)))

		      ;; not modified, record today's date
		      (set! alist (assq-set! alist 'tdate today)))

		  (database-call-with-output-file filename
		    (lambda (port)
		      (apply format port
			     ";; Chart cached information.
;;
;; From:       ~a
;; Downloaded: ~a-~2,'0d-~2,'0d

"
			     (write-str-one-line url)
			     (tdate->ymd today))
		      (write-list-lines alist port))))))))

    alist))

(define (write-str-one-line obj)
  (string-delete (format #f "~s" obj)
		 (char-set #\cr #\newline)))
