;; Parabolic SAR.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart parabolic-sar)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; This does the whole series from start to end.  It's not clear how to work
;; back to find where a previous SAR point should be.
;;
(define (parabolic-sar-fill series lo hi)
  (receive-list (initial-accel max-accel)
      (series-data series)
    (set! lo (series-lo series))
    (set! hi (series-hi series))

    (let* ((parent    (series-parent series))
	   (array     (series-array parent lo hi))
	   (new-array (make-array #f (list lo hi))))

      (define extreme   #f)
      (define sar       #f)
      (define accel     #f)
      (define proc      #f)

      (define (initial high low)
	(set! extreme   high)
	(set! accel     initial-accel)
	(set! sar       low)
	(set! proc      rising))

      (define (rising high low)
	(if (> high extreme)
	    (begin
	      (set! extreme high)
	      (set! accel (min max-accel (+ accel initial-accel)))))

	(set! sar (+ sar (* accel (- extreme sar))))

	;; if SAR penetrated, reverse
	(if (<= low sar)
	    (begin
	      (set! sar     extreme)
	      (set! extreme low)
	      (set! accel   initial-accel)
	      (set! proc    falling))))

      (define (falling high low)
	(if (< low extreme)
	    (begin
	      (set! extreme low)
	      (set! accel (min max-accel (+ accel initial-accel)))))

	(set! sar (+ sar (* accel (- extreme sar))))

	;; if SAR penetrated, reverse
	(if (>= high sar)
	    (begin
	      (set! sar     extreme)
	      (set! extreme high)
	      (set! accel   initial-accel)
	      (set! proc    rising))))

      (set! proc initial)

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((>= i hi))
	(and-let* ((close (array-ref array i 3)))
	  (proc (or (array-ref array i 1) close)   ;; high
		(or (array-ref array i 2) close))) ;; low
	(array-set! new-array sar i))

      ;; prospective stop one beyond parent series
      (set! sar (+ sar (* accel (- extreme sar))))
      (array-set! new-array sar hi)

      new-array)))

(define-public (parabolic-sar-series series initial-accel max-accel)
  (ohlcv-required "Parabolic SAR" series)
  (let ((new-series (series-copy series)))
    (series-name! new-series
		  (if (and (= initial-accel 0.02)
			   (= max-accel 0.2))
		      (_ "SAR")
		      (format #f (_ "SAR ~a, ~a")
			      (locale-number-min-decimals initial-accel 1)
			      (locale-number-min-decimals max-accel 1))))
    (series-datatype! new-series parabolic-sar-datatype)
    (series-parent!   new-series series)
    (series-hi!       new-series (1+ (series-hi series)))
    (series-data!     new-series (list initial-accel max-accel))
    (series-require!  new-series parabolic-sar-fill)
    (series-array!    new-series #f)
    new-series))

(define parabolic-sar-datatype
  (datatype-new!
   #:key       'parabolic-sar
   #:name      (_ "Parabolic SAR")
   #:units     #f
   #:input     datatype-ohlcv
   #:linestyle 'stops
   #:manual    (_ "CONTEXT(manual-node): Parabolic SAR")
   #:proc      parabolic-sar-series
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "Initial accel")
	  #:key      'parabolic-sar-initial-accel
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:maximum  1
	  #:default  0.02
	  #:step     0.01)
	 (datatype-param-new
	  #:name     (_ "Max accel")
	  #:key      'parabolic-sar-max-accel
	  #:type     'float
	  #:decimals 2
	  #:minimum  0
	  #:maximum  1
	  #:default  0.2
	  #:step     0.1))))
