;; Polarized fractal efficiency (PFE) indicator.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart pfe)
  #:use-module (srfi srfi-1)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart roc))


;; http://transcripts.fxstreet.com/2005/09/polarized_fract.html
;;     Presentation on use.
;;
;; http://trader.online.pl/MSZ/e-w-Polarized_Fractal_Efficiency.html
;;     Mov(If(C, >, Ref(C,-9),
;;     	    Sqr(Pwr(Roc(C,9,$),2) + Pwr(10,2))
;;     	      / Sum(Sqr( Pwr(Roc(C,1,$),2) + 1), 9),
;;          -Sqr(Pwr(Roc(C,9,$),2) + Pwr(10,2))
;;            / Sum(Sqr( Pwr(Roc(C,1,$),2) + 1), 9))
;;         *100, 5, E)
;;
;; http://www.traderslog.com/polarized-fractal-efficiency.htm
;;     Sample chart of IBM from 2002.  Year not shown, but data is 2002.
;;     Seems based on ROC as percentage, day step as 1 in hypot, and 5-day
;;     EMA smooth.
;;
;; http://store.traders.com/-v12-c01-polariz-pdf.html
;;     TASC article for sale.
;;


(define (hypot x y)
  (magnitude (make-rectangular x y)))


;; Return a procedure which calculates a polarized fractal efficiency over
;; an accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the PFE up to (and including) that value, or #f if not enough values
;; yet.
;;
;; Because PFE uses an exponential moving average, it's in theory influenced
;; by all preceding data, but `pfe-warmup-count' below is designed to
;; determine a warmup count.
;;
(define-public (pfe-calc-proc count smooth)

  ;; change to count as overall distance, rather than points used
  (set! count (1- count))

  (let ((rocN-proc (roc-calc-proc count))
	(roc1-proc (roc-calc-proc 1))
	(sum-proc  (sum-calc-proc count))
	(ema-proc  (ema-calc-proc smooth)))

    (lambda (x)
      (let* ((rocN (rocN-proc x))
	     (roc1 (roc1-proc x))
	     (sum  (and roc1
			(sum-proc (hypot roc1 1)))))
	(and rocN sum
	     (ema-proc (* (if (negative? rocN) -100.0 100.0)
			  (/ (hypot rocN count)
			     sum))))))))

(define-public (pfe-warmup-count count smooth)
  (+ (1- count)
     (ema-warmup-count smooth)))


(define (pfe-fill series new-array lo hi)
  (receive-list (count smooth)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (pfe-calc-proc count smooth)))

      ;; warmup
      (for-each proc (series-prev-n-closes
		      parent #:before lo 
		      #:count (pfe-warmup-count count smooth)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i)) i))))))


(define-public (pfe-series series count smooth)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f (_ "PFE ~a,~a") count smooth))
    (series-datatype!  new-series pfe-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list count smooth))
    (series-require!   new-series (indicator-require-parts 
				   pfe-fill (pfe-warmup-count count smooth)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public pfe-datatype
  (datatype-new!
   #:key       'pfe
   #:name      (_ "PFE - Polarized Fractal Efficiency")
   #:units     'pfe
   #:linestyle 'line
   #:minimum   -100
   #:maximum   100
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Polarized Fractal Efficiency")
   #:proc      pfe-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'pfe-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)
	 (datatype-param-new
	  #:key     'pfe-smooth
	  #:name    (_ "Smooth")
	  #:type    'integer
	  #:minimum 0
	  #:default 5))))
