;; Pretty Good Oscillator (PGO).

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart pgo)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart atr)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart sma))


;; http://www.purebytes.com/archives/omega/1999/msg12885.html
;;
;; http://trader.online.pl/MSZ/e-w-Pretty_Good_Oscillator.html
;;    Sample CSCO
;;


;; Return a procedure which calculates PGO values over a window of COUNT
;; many points.
;;
;; Each call (PROC high low close) enters a new point into the window, and
;; the return is the PGO value up to (and including) that point, or #f if
;; not enough values yet.
;;
;; To prime the window initially, call PROC with `pgo-warmup-count' many
;; values preceding the first desired point.
;;
(define-public (pgo-calc-proc count)

  (let ((sma-proc  (sma-calc-proc count))
	(atr-proc  (atr-calc-proc (ema-count-to-wilder count))))

    (lambda (high low close)
      (let ((avg (sma-proc close))
	    (atr (atr-proc high low close)))
	(and avg
	     atr
	     (not (zero? atr))
	     (/ (- close avg) atr))))))

;; return the number of points that should be entered into a `pgo-calc-proc'
;; to get the next based on a full window
;;
(define-public (pgo-warmup-count count)
  (max (1- count)                       ;; sma of closes
       (atr-warmup-count (ema-count-to-wilder count))))  ;; ema of true range


(define (pgo-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos parent
					     #:before lo #:count warmup))
	   (array  (series-array parent start hi))
	   (pgo    (pgo-calc-proc count)))

      ;; warmup
      (do ((i start (1+ i)))
	  ((>= i lo))
	(and-let* ((close (array-ref array i 3)))
	  (pgo (array-ref array i 1)  ;; high
	       (array-ref array i 2)  ;; low
	       close)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((close (array-ref array i 3)))
	  (array-set! new-array (pgo (array-ref array i 1)  ;; high
				     (array-ref array i 2)  ;; low
				     close)
		      i))))))

(define-public (pgo-series series count)
  (ohlcv-required "PGO" series)
  (let ((new-series (series-copy series))
	(warmup     (pgo-warmup-count count)))
    (series-name!      new-series (format #f "~a ~a" (_ "PGO") count))
    (series-datatype!  new-series pgo-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list count warmup))
    (series-require!   new-series (indicator-require-parts pgo-fill warmup))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public pgo-datatype
  (datatype-new!
   #:key       'pgo
   #:name      (_ "PGO - Pretty Good Oscillator")
   #:input     datatype-ohlcv
   #:units     'pgo
   #:linestyle 'line
   #:hlines    '(-3 0 3)
   #:manual    (_ "CONTEXT(manual-node): Pretty Good Oscillator")
   #:proc      pgo-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'pgo-days
	  #:type    'integer
	  #:minimum 1
	  #:default 89))))
