;; Preferences dialog.

;; Copyright 2004, 2005, 2006, 2007, 2008, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart prefs-edit)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (gtk gtk)
  #:use-module (chart download)
  #:use-module (chart extra-gtk)
  #:use-module (chart gui-common)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; preferences
    
(define rec-type (make-record-type "pref"
				   '(key name type data width default
					 validate)))
(define rec-new (record-constructor rec-type))

;; compare names in preference records X and Y
;; (define (pref-name<? x y)
;;   (string-ci<? (pref-name x) (pref-name y)))

(define pref-key      (rec-accessor rec-type 'key))
(define pref-name     (rec-accessor rec-type 'name))
(define pref-type     (rec-accessor rec-type 'type))
(define pref-data     (rec-accessor rec-type 'data))
(define pref-width    (rec-accessor rec-type 'width))
(define pref-default  (rec-accessor rec-type 'default))
;; PROC called as (PROC obj), or #f
(define pref-validate (rec-accessor rec-type 'validate))

;; list of all preference records
(define pref-record-list '())

(define* (pref-new!
	  #:key
	  (key  (error "pref-new! must have key"))
	  (name (error "pref-new! must have name"))
	  type data
	  (width 20)
	  default
	  validate)
  (set! pref-record-list
	(append pref-record-list
		(list (rec-new key name type data width default
			       validate)))))

(define (validate-timezone str)
  (if (timezone-known? str)
      #f
      (format #f (_ "Unknown timezone: ~s") str)))

;; (define (validate-http-url-or-empty str)
;;   (if (or (string-null? str)
;; 	  (string-prefix? "http://" str))
;;       #f
;;       (format #f (_ "Not a http URL: ~s") str)))

(define (validate-yes-or-no str)
  (if (or (string-ci=? str "yes")
	  (string-ci=? str "no"))
      #f
      (format #f (_ "Must be yes or no: ~s") str)))

(pref-new! #:key      'lme-username
	   #:name     (_ "LME username")
	   #:type     'string
	   #:width    20
	   #:default  "")

(pref-new! #:key      'lme-password
	   #:name     (_ "LME password")
	   #:type     'string
	   #:width    20
	   #:default  "")

(pref-new! #:key      'yahoo-quote-host
	   #:name     (_ "Yahoo quote host")
	   #:type     'string
	   #:width    20
	   #:default  "finance.yahoo.com")

(pref-new! #:key      'yahoo-quote-timezone
	   #:name     (_ "Yahoo quote timezone")
	   #:type     'string
	   #:validate validate-timezone
	   #:width    20
	   #:default  (timezone-newyork))

;; (pref-new! #:key      'commsec-enabled
;; 	   #:name     (_ "Use CommSec? (you must be a client)")
;; 	   #:type     'string
;; 	   #:validate validate-yes-or-no
;; 	   #:width    4
;; 	   #:default  "no")


;; (pref-new! #:key      'viewstyle
;; 		 #:name     (_ "View style")
;; 		 #:type     'hidden
;; 		 #:default  '((ohlc (none) (window-size 4))
;; 			      ((volume) (none) (window-size 1))))

;; (notify-connect 'symlist-update
;;   (lambda (symlist-key-list)
;;     (if (memq 'favourites symlist-key-list)
;; 	(preferences-load))))


;;-----------------------------------------------------------------------------
;; preferences dialog

(define pref-to-widget-prop (primitive-make-property #f))
(define dialog #f)

(define (init)
  (define (get-value pref)
    (let ((widget (primitive-property-ref pref-to-widget-prop pref)))
      (case (pref-type pref)
	((string)
	 (gtk-entry-get-text widget))
	((menu)
	 (cdr (list-ref (pref-data pref)
			(gtk-option-menu-get-history widget)))))))

  (set! dialog (gtk-dialog-new-with-buttons
		(c-make-window-title (_ "Preferences")) #f '()
		"gtk-ok"     gtk-response-ok
		"gtk-cancel" gtk-response-cancel))

  (c-gtk-dialog-hide-on-delete dialog)
  (gtk-signal-connect dialog "response"
    (lambda (id)
      (cond ((= id gtk-response-ok)
	     (cond ((any (lambda (pref)
			   (and-let* ((validate (pref-validate pref))
				      (errmsg  (validate (get-value pref))))
			     (format #f (_ "Field: ~a\n~a")
				     (pref-name pref) errmsg)))
			 pref-record-list)
		    => (lambda (message)
			 (error-dialog-popup dialog message)))
		   (else
		    (for-each (lambda (pref)
				(if (memq (pref-type pref) '(string menu))
				    (preference-set (pref-key pref)
						    (get-value pref))))
			      pref-record-list)
		    (gtk-widget-hide dialog))))

	    ((= id gtk-response-cancel)
	     (gtk-widget-hide dialog)))))

  (gtk-box-pack-start
   (gtk-dialog-vbox dialog)
   (gtk-label-new (_ "Edit preferences, then press Ok or Cancel"))
   #f #f 0)

  (for-each
   (lambda (pref)
     (if (not (eq? (pref-type pref) 'hidden))
	 (let ((hbox  (gtk-hbox-new #f 4))
	       (label (gtk-label-new (pref-name pref))))
	   (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0)
	   (gtk-box-pack-start hbox label #f #f 0)
	   (case (pref-type pref)
	     ((string)
	      (let ((entry (gtk-entry-new)))
		(c-gtk-widget-set-usize-units
		 entry (cons 'em (pref-width pref)) -1)
		(gtk-box-pack-start hbox entry #t #t 0)
		(primitive-property-set! pref-to-widget-prop
					 pref entry)))
	     ((menu)
	      (let ((option (gtk-option-menu-new))
		    (menu   (gtk-menu-new)))
		(for-each
		 (lambda (pair)
		   (let ((item (gtk-menu-item-new-with-label (car pair))))
		     (gtk-widget-show item)
		     (gtk-menu-append menu item)))
		 (pref-data pref))
		(gtk-option-menu-set-menu option menu)
		(gtk-box-pack-start hbox option #f #f 0)
		(primitive-property-set! pref-to-widget-prop
					 pref option))))
	   (gtk-widget-show-all hbox))))
   pref-record-list)

  ;; once only, and gc this code
  (set! init noop))

(define (fill)
  (for-each
   (lambda (pref)
     (and-let* ((widget (primitive-property-ref pref-to-widget-prop pref)))
       (case (pref-type pref)
	 ((string)
	  ;; Somehow #f got into `preferences-filename' for commsec data url
	  ;; in the past, so have an "or" here to cope with that.  It
	  ;; shouldn't happen with new uses though.
	  (gtk-entry-set-text widget
			      (or (preference-get (pref-key pref)
						  (pref-default pref))
				  (pref-default pref))))
	 ((menu)
	  (gtk-option-menu-set-history
	   widget
	   (list-index (lambda (pair)
			 (equal? (cdr pair)
				 (pref-default pref)))
		       (pref-data pref)))))))
   pref-record-list))

(define-public (pref-dialog-popup)
  (init)
  (if (memq 'mapped (gtk-widget-flags dialog))
      (c-gtk-widget-raise dialog)
      (begin
	(fill)
	(gtk-widget-show-all dialog))))

