;; Qstick oscillator.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart qstick)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart sma)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://trader.online.pl/MSZ/e-w-Chandes_QStick.html
;;


;; Return a procedure which calculates a Qstick oscillator over an
;; accumulated window of COUNT days.
;;
;; Each call (PROC open close) enters a new day into the window, and the
;; return is the Qstick for that day, or #f if not enough points yet.
;;
;; To prime initiall, call PROC with COUNT-1 many points before the first
;; desired, to get that next call based on a full window.
;;
(define-public (qstick-calc-proc count)
  (let ((sma-proc (sma-calc-proc count)))
    (lambda (open close)
      (sma-proc (- close open)))))


(define (qstick-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo
					   #:count  (1- count)))
	 (array  (series-array parent start hi))
	 (qstick (qstick-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((open  (array-ref array i 0))
		 (close (array-ref array i 3)))
	(qstick open close)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((open  (array-ref array i 0))
		 (close (array-ref array i 3)))
	(array-set! new-array (qstick open close)
		    i)))))

(define-public (qstick-series series count)
  (ohlcv-required "Qstick" series)
  (let ((new-series (series-copy series)))
    (series-datatype!  new-series qstick-datatype)
    (series-name!      new-series (format #f "~a ~a" (_ "Qstick") count))
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts qstick-fill
							   (1- count)))
    (series-parent!    new-series series)
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public qstick-datatype
  (datatype-new!
   #:key     'qstick
   #:name    (_ "Qstick")
   #:input   datatype-ohlcv
   #:units   'qstick
   #:hlines  '(0)
   #:manual  (_ "CONTEXT(manual-node): Qstick")
   #:proc    qstick-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'qstick-days
	  #:type    'integer
	  #:minimum 1
	  #:default 8))))
