;; Reserve Bank of Australia setups.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart rba)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart download)
  #:use-module (chart down-cost)
  #:use-module (chart gzip) ;; string-trim-str
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart pagebits)
  #:use-module (chart xls))


(define-public (rba-symbol? symbol)
  (string-suffix? ".RBA" symbol))

(symbol-timezone! rba-symbol? (_ "Sydney") timezone-sydney)
(source-help! rba-symbol?
	      (_ "CONTEXT(manual-node): Reserve Bank of Australia"))

(define rba-copyright-url "http://www.rba.gov.au/Copyright/index.html")


;;-----------------------------------------------------------------------------
;; misc

;; return MS date (days since midnight 1 Jan 1900) for adate
(define (adate->msdate adate)
  (+ adate 25573))

;; return MS date (days since midnight 1 Jan 1900) for tdate
(define (tdate->msdate tdate)
  (adate->msdate (tdate->adate tdate)))

;; return the month mdate containing the given MSDATE
(define (msdate->mdate msdate)
  (adate->mdate (msdate->adate msdate)))

;; return the month mdate containing the given ADATE
(define-public (adate->mdate adate)
  (apply ymd->mdate (adate->ymd adate)))

;; return a list of all the tdates within the month MDATE
(define (month-list-tdates mdate)
  (iota-range (mdate->tdate mdate) (mdate->last-tdate mdate)))

(define (xls->row-list data)
  (define ret '())
  (xls-for-each-row (lambda (array)
		      (set! ret (cons (array->list array) ret)))
		    data)
  (reverse! ret))


;;-----------------------------------------------------------------------------
;; latest exchange rates
;;
;;     http://www.rba.gov.au/Statistics/exchange_rates.html

(define rba-exchange-html-url
  "http://www.rba.gov.au/Statistics/exchange_rates.html")

;; return symbol, or #f for entries to exclude
(define (rba-name->symbol name)
  (cdr (or (assoc name
		  '(("United States dollar"  	   . "AUDUSD.RBA")
		    ("Japanese yen"          	   . "AUDJPY.RBA")
		    ("European euro"         	   . "AUDEUR.RBA")
		    ("South Korean won"      	   . "AUDKRW.RBA")
		    ("New Zealand dollar"    	   . "AUDNZD.RBA")
		    ("Chinese renminbi"      	   . "AUDCNY.RBA")
		    ("UK pound sterling"     	   . "AUDGBP.RBA")
		    ("New Taiwan dollar"     	   . "AUDTWD.RBA")
		    ("Singapore dollar"      	   . "AUDSGD.RBA")
		    ("Indonesian rupiah"     	   . "AUDIDR.RBA")
		    ("Hong Kong dollar"      	   . "AUDHKD.RBA")
		    ("Malaysian ringgit"     	   . "AUDMYR.RBA")
		    ("Swiss franc"           	   . "AUDCHF.RBA")
		    ("Special Drawing Right" 	   . "AUDSDR.RBA")
		    ("Trade-weighted Index (9am)"  . #f)
		    ("Trade-weighted Index (Noon)" . #f)
		    ("Trade-weighted Index (4pm)"  . "AUDTWI.RBA")))
	   (error "RBA: Unrecognised currency name:" name))))

;; return true if RATE is not value from a bank holiday column
;; bank holidays like 7 Sep 07 get "BANK HOLIDAY" written out in the column,
;; one letter per row
(define (rba-rate-good? rate)
  (not (string-every char-set:letter rate)))

;; return the index of the last element in LST which satisfies PRED
;; elements are numbered starting from 0 for the first
(define (list-index-last pred lst)
  (- (length lst) 1 (list-index pred (reverse lst))))

(define (rba-body->latest-list headers body)
  (let* ((rdpos    (string-contains-ci body "Release date:"))
	 (t1pos    (string-contains-after-ci body "<table" rdpos))
	 (row-list (html-table-rows body t1pos)))

    (or (string=? "Click for earlier rates" (first (first row-list)))
	(error "RBA: unrecognised latest exchange rate headings:"
	       (first row-list)))

    ;; look for the rightmost column of each row
    (filter-map (lambda (row)
		  (let* ((name   (first row))
			 (symbol (rba-name->symbol name)))
		    (and symbol ;; skip TWI 9am and Noon
			 (let* ((rate-pos   (list-index-last rba-rate-good?
							     row))
				(rate       (list-ref row rate-pos))
				(last-adate (list-ref (first row-list)
						      rate-pos)))
			   (set! name (string-trim-right-str name " (4pm)"))
			   (set! last-adate (d/m/y-str->adate last-adate))

			   (latest-new #:symbol      symbol
				       #:name        name
				       #:last-adate  last-adate
				       #:last-time   #,(hms->seconds 16 0 0)
				       #:last        rate
				       #:headers     headers)))))
		(cdr row-list))))

(define (rba-latest-get symbol-list extra-list proc)
  (receive (headers body)
      (http-request rba-exchange-html-url
		    #:want-ok #t)
    (proc (rba-body->latest-list headers body))))

(define (rba-quote-adate-time symbol)
  (tm->adate-time-within (localtime (current-time) (timezone-sydney))
			 #,(hms->seconds 16 0 0)
			 #,(hms->seconds 16 0 0)))

(latest-handler! #:selector   rba-symbol?
		 #:handler    rba-latest-get
		 #:adate-time rba-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - 5 day update
;;
;;     http://www.rba.gov.au/Statistics/exchange_rates.html

;; return true if processed into database, or #f if must go back further
;; (ie. to the xls files)
;;
(define (rba-5day-process symbol-list start-tdate headers body)
  (let* ((rdpos    (string-contains-ci body "Release date:"))
	 (t1pos    (string-contains-after-ci body "<table" rdpos))
	 (row-list (html-table-rows body t1pos)))

    (or (string=? "Click for earlier rates" (first (first row-list)))
	(error "RBA: unrecognised latest exchange rate headings:"
	       (first row-list)))

    ;; dates across top
    (let ((tdate-list (map d/m/y-str->tdate (cdr (first row-list)))))
      (set! row-list (cdr row-list))

      (and (<= (first tdate-list) start-tdate)
	   (begin
	     (download-process
	      #:module        (_ "RBA")
	      #:symbol-list   symbol-list
	      #:last-download #t
	      #:url           rba-exchange-html-url
	      #:headers       headers
	      #:copyright     rba-copyright-url
	      #:row-list
	      (append-map
	       (lambda (row)
		 (let* ((name      (first row))
			(symbol    (rba-name->symbol name))
			(commodity (and symbol
					(chart-symbol-commodity symbol))))
		   (set! name (string-trim-right-str name " (4pm)"))
		   (if commodity ;; skip extra TWI values as at 9am and Noon
		       (filter-map
			(lambda (tdate rate)
			  (and (rba-rate-good? rate)
			       (list #:commodity commodity
				     #:name      name
				     #:tdate     tdate
				     #:close     rate
				     #:currency  (string-drop commodity 3))))
			tdate-list (cdr row))
		       '())))
	       row-list))
	     #t)))))

;; return true if downloaded or #f if must use the xls historical files
;;
(define (rba-5day-download symbol-list)
  (define start-tdate (apply min (map download-start-tdate symbol-list)))

  ;; only if getting 5 days worth, from start to avail, inclusive
  (and (<= (- (rba-available-tdate) start-tdate) 4)
       (begin
	 (download-status (_ "RBA") (_ "past 5 days"))
	 (receive (headers body)
	     (http-request rba-exchange-html-url
			   #:want-ok #t)
	   (latest-handler-callback (rba-body->latest-list headers body))
	   (rba-5day-process symbol-list start-tdate headers body)))))


;;-----------------------------------------------------------------------------
;; download - historical page xls file urls

(define rba-historical-xls-page 
  "http://www.rba.gov.au/Statistics/HistoricalExchangeRates/index.html")

;; BODY is the historical exchange rates page, return a single-entry alist
;; (("xls-lst" . LST)) which is the various xls files. ; Each element of LST
;; is in downranges-select style, like
;;
;;     (#:lo 100 #:hi 200 #:cost 750000 #:url "http://...")
;;
(define (rba-historical-xls-parse body)
  (list
   (cons "xls-lst"
	 (map! (lambda (m)
		 (let* ((link       (match:substring m 1))
			(size       (* (string->number
					(or (match:substring m 4)
					    (error "RBA: oops, size not matched")))
				       (if (string=? (match:substring m 5) "K")
					   1000
					   1000000)))
			(url        (http-url-relative rba-historical-xls-page link))
			(m          (must-match (string-match "/([0-9][0-9][0-9][0-9])(to([0-9][0-9][0-9][0-9]))?[.]xls$" url)))
			(start-str  (match:substring m 1))
			(end-str    (or (match:substring m 3)
					(match:substring m 1)))
			(start-year (string->number start-str))
			(end-year   (string->number end-str)))
		   (list #:url  url
			 #:lo   (year->tdate start-year)
			 #:hi   (year->last-tdate end-year)
			 #:cost size)))

	       ;;                                 1                                   23           4        5
	       (list-matches (make-regexp "href=\"([^\"]*[0-9][0-9][0-9][0-9]\\.xls)\"((.*\n?).* - ([0-9.]+)([MK]))?" regexp/newline)
			     body)))))

(define (historical-xls-lst)
  (assoc-ref (pagebits-read #:filename  "rba-historical-xls"
			    #:status    (list (_ "RBA")
					      (_ "historical page"))
			    #:url       rba-historical-xls-page
			    #:timezone  (timezone-sydney)
			    #:parse     rba-historical-xls-parse)
	     "xls-lst"))


;;-----------------------------------------------------------------------------
;; download - historical monthly prices
;;
;; This uses the monthly rates file
;;
;;     http://www.rba.gov.au/Statistics/Bulletin/F11hist.xls
;;
;; but only from 1983 back, since there's daily files for 1983 onwards.

(define (rba-rate->string rate)
  (if (and (list>=? rate 2)
	   (eq? (second rate) 'cents))
      (set! rate (* 0.01 (first rate))))
  (and rate
       (number->string
	(round-exact (* 10000 rate)))))

;; mung the currency headings in the monthly file
(define (rba-monthly-fx->currency str)
  (or (assoc-ref '(("FXRJY"   . "JPY")  ;; japanese yen
		   ;; "FXRUSD"
		   ;; "FXREUR"
		   ("FXRSKW"  . "KRW")  ;; korean won
		   ;; "FXRNZD"
		   ("FXRCR"   . "CNY")  ;; chinese renminbi
		   ("FXRUKPS" . "GBP")  ;; british pound sterling
		   ("FXRNTD"  . "TWD")  ;; taiwan dollar
		   ("FXRSD"   . "SGD")  ;; singapore dollar
		   ("FXRIR"   . "IDR")  ;; indonesian rupiah
		   ;; "FXRHKD"
		   ("FXRMR"   . "MYR")  ;; malaysian ringgit
		   ;; "FXRSDR"          ;; special drawing right
		   ;; "FXRTWI"          ;; trade weighted index
		   )
		 str)
      (string-trim-str str "FXR")))

(define (rba-monthly-process symbol-list body)
  (define row-list (xls->row-list body))

  (set! row-list (or (find-tail (lambda (row)
				  (equal? "F11HIST.XLS" (first row)))
				row-list)
		     (error "RBA: historical data headings row not found")))
  (let ((currency-list (map rba-monthly-fx->currency (cdr (first row-list)))))
    (set! row-list (cdr row-list))

    ;; discard rows after the date the daily historical files begin, since
    ;; that data is better
    (let* ((stop-msdate (tdate->msdate (rba-historical-start-tdate))))
      (set! row-list (remove! (lambda (row)
				(>= (first row) stop-msdate))
			      row-list)))

    (download-process
     #:module         (_ "RBA")
     #:symbol-list    symbol-list
     #:copyright      rba-copyright-url
     #:extra-decimals 4
     #:row-list
     (append-map (lambda (row)
		   (let* ((tdate-list (month-list-tdates
				       (msdate->mdate (first row)))))
		     (append-map (lambda (currency rate)
				   (set! rate (rba-rate->string rate))
				   (map (lambda (tdate)
					  (list #:commodity (string-append
							     "AUD" currency)
						#:currency  currency
						#:tdate     tdate
						#:close     rate))
					tdate-list))
				 currency-list (cdr row))))
		 row-list))))

(define (rba-monthly-backto symbol-list)
  (define start-tdate (apply max (map download-back-tdate symbol-list)))

  ;; monthly goes back to July 1969 (could parse that from the web page
  ;; rather than hard-coding, maybe)
  (if (>= start-tdate (ymd->tdate-ceil 1969 7 1))
      (receive (headers body)
	  (http-request "http://www.rba.gov.au/Statistics/Bulletin/F11hist.xls"
			#:want-ok #t)
	(rba-monthly-process symbol-list body))

      (download-message
       (string-append
	(_ "RBA") ": " (_ "no further historical data available")))))


;;-----------------------------------------------------------------------------
;; download - historical

;; return the tdate where the historical xls files start
(define (rba-historical-start-tdate)
  (apply min (map (lambda (elem)
		    (plist-get elem #:lo))
		  (historical-xls-lst))))

(define (rba-historical-process symbol-list body)
  (define row-list (xls->row-list body))

  (set! row-list (or (find-tail (lambda (row)
				  (equal? "DAILY 4PM" (first row)))
				row-list)
		     (error "RBA: historical data headings row not found")))
  (let ((currency-list (cdr (first row-list)))
	(target-length (length (first row-list))))
    (set! row-list (filter! (lambda (row)
			      (= (length row) target-length))
			    row-list))

    (download-process
     #:module         (_ "RBA")
     #:symbol-list    symbol-list
     #:copyright      rba-copyright-url
     #:extra-decimals 4
     #:row-list
     (append-map (lambda (row)
		   (let ((tdate (msdate->tdate (first row))))
		     (map (lambda (currency rate)
			    (list #:commodity (string-append "AUD" currency)
				  #:currency  currency
				  #:tdate     tdate
				  #:close     (rba-rate->string rate)))
			  currency-list (cdr row))))
		 (cdr row-list)))))

(define (rba-historical-one symbol-list elem)
  (define url (plist-get elem #:url))
  (download-status (_ "RBA") (basename url))
  (receive (headers body)
      (http-request url
		    #:want-ok #t)
    (rba-historical-process symbol-list body)))

(define (plist-lo<? x y)
  (< (plist-get x #:lo)
     (plist-get y #:lo)))

(define (rba-historical-download symbol-list)
  (define start-tdate (apply min (map download-start-tdate symbol-list)))
  (for-each (lambda (elem)
	      (rba-historical-one symbol-list elem))
	    (sort (downranges-select (historical-xls-lst)
				     start-tdate (rba-available-tdate))
		  plist-lo<?)))

(define (rba-historical-backto want-tdate symbol-list)
  (define start-tdate (apply max (map download-back-tdate symbol-list)))

  (for-each (lambda (elem)
	      (rba-historical-one symbol-list elem))
	    (reverse
	     (sort (downranges-select (historical-xls-lst)
				      want-tdate start-tdate)
		   plist-lo<?))))


;;-----------------------------------------------------------------------------
;; download

(define (rba-available-tdate)
  (tdate-today-after 16 0 (timezone-sydney)))

(define (rba-download symbol-list)
  (or (rba-5day-download symbol-list)
      (rba-historical-download symbol-list))
  (download-name-from-latest (_ "RBA") symbol-list))

(define (rba-backto want-tdate symbol-list)
  (rba-historical-backto want-tdate symbol-list)
  (if (< want-tdate (rba-historical-start-tdate))
      (rba-monthly-backto symbol-list)))

(download-handler! #:selector   rba-symbol?
		   #:avail-proc rba-available-tdate
		   #:update     rba-download
		   #:backto     rba-backto)
