;; Regularized Exponential Moving Average.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart rema)
  #:use-module (ice-9 receive)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/072003/Abstracts_new/Satchwell/satchwell.html
;;     Start of TASC August 2003 article by Chris Satchwell.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/072003/TradersTips/TradersTips.html
;;     TASC Traders' Tips August 2003, regularization formulas.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/092003/Letters/Letters.html
;;     TASC letters September 2003, Chris Satchwell clarifying derivation of
;;     formula.
;;
;; http://www.equis.com/Customer/Resources/TASC/Article.aspx?Id=50
;;     Metastock code for regularized EMA and regularized momentum.
;;
;; http://trader.online.pl/ELZ/t-i-Regularization.html
;;     Easylanguage code, copy of trader's tips.
;;



;; Return a procedure which calculates a regularized EMA with given COUNT
;; days and L lambda factor.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the REMA up to (and including) that value, or #f if not enough points
;; yet.
;;
;; A REMA is in theory influenced by all preceding data, but
;; `rema-warmup-count' below is designed to determine a warmup count.  After
;; calling PROC with (rema-warmup-count COUNT L) many values, the next call
;; will have an omitted weight of no more than 0.1% of the total.  Omitting
;; 0.1% should be negligable, unless past values are ridiculously bigger
;; than recent ones.
;;
(define-public (rema-calc-proc count L)
  ;;
  ;; The formula
  ;;
  ;;     REMA = Rprev * (2L+1) + alpha*(close - Rprev) - L*Rprevprev
  ;;
  ;; is turned into (in the style of the AmiBroker code in Trader's Tips
  ;; above),
  ;;
  ;;     REMA = a*close + b*Rprev + c*Rprevprev
  ;;
  ;; with constants calculated by `rema-factors' below
  ;;
  ;;         alpha         2*L+1-alpha        -L
  ;;     a = -----     b = -----------    c = ---
  ;;          L+1              L+1            L+1
  ;;
  (receive (a b c)
      (rema-factors count L)
    (let ((r    0.0)
	  (rw   0.0)
	  (rP   0.0)
	  (rPw  0.0)
	  (rPP  0.0)
	  (rPPw 0.0))

      (lambda (x)
	(set! r  (+ (* x a) (* rP b)  (* rPP c)))
	(set! rw (+ a       (* rPw b) (* rPPw c)))

	(set! rPP  rP)
	(set! rPPw rPw)

	(set! rP  r)
	(set! rPw rw)

	(and (>= rw indicator-minimum-fraction)
	     (/ r rw))))))

(define (rema-factors count L)
  (let ((alpha (ema-count->alpha count)))
    (values (/ alpha                   (1+ L))
	    (/ (- (1+ (* 2 L)) alpha)  (1+ L))
	    (/ (- L)                   (1+ L)))))

(define-public (rema-warmup-count count L)
  (* 2 (ema-warmup-count (/ count L))))




(define (rema-fill series new-array lo hi)
  (receive-list (count L)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (rema-calc-proc count L)))

      (for-each proc (series-prev-n-closes
		      parent #:before lo #:count (rema-warmup-count count L)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i))
			i))))))

(define-public (rema-series series count L)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f (_ "REMA ~a, ~a") count L))
    (series-datatype! new-series rema-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count L))
    (series-require!  new-series (indicator-require-parts rema-fill
							  (rema-warmup-count count L)))
    (series-array!    new-series #f)
    new-series))

(define rema-datatype
  (datatype-new!
   #:key    'rema
   #:name   (_ "REMA - Regularized EMA")
   #:units  #f
   #:manual (_ "CONTEXT(manual-node): Regularized Exponential Moving Average")
   #:proc   rema-series
   #:parameter-list
   (list (datatype-param-new
	  #:name     (_ "Days")
	  #:key      'rema-days
	  #:type     'integer
	  #:minimum  1
	  #:default  21)
	 (datatype-param-new
	  #:name     (_ "Lambda")
	  #:key      'rema-lambda
	  #:type     'float
	  #:decimals 1
	  #:minimum  0
	  #:default  0.5
	  #:step     0.1))))
