;; Rate of Change (ROC) indicator.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart roc)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates rate-of-change over values COUNT
;; points apart.  For example COUNT=1 would give the change between today
;; and yesterday.
;;
;; Each call (PROC val) enters a new value into the calculation, and the
;; return is the ROC on that value, or #f if there hasn't been COUNT+1
;; points yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from first to last).  COUNT many
;; values is enough to ensure the first PROC call is based on a full window.
;;
(define*-public (roc-calc-proc count #:optional (data '()))
  (let ((prevnth-proc (prevnth-calc-proc count data)))
    (set! data #f) ;; gc

    (lambda (x)
      (let ((prev (prevnth-proc x)))
	(and prev
	     (not (zero? prev))
	     (* 100 (/ (- x prev)
		       (exact->inexact prev))))))))


(define (roc-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent (- lo count) hi))
	 (roc    (roc-calc-proc count (series-prev-n-closes parent
							    #:before lo
							    #:count count))))
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (roc this) i)))))

(define-public (roc-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "ROC") count))
    (series-datatype! new-series roc-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts roc-fill count))
    (series-array!    new-series #f)
    (series-decimals! new-series 0)

    ;; count fewer elements
    (let ((lo (+ count (series-lo series)))
	  (hi (series-hi new-series)))
      (series-lo! new-series lo)

      (if (> lo hi)
	  (begin  ;; less than COUNT points in parent series
	    (series-lo! new-series hi)
	    (series-array! new-series (make-array #f (list hi hi))))))

    new-series))

(define-public roc-datatype
  (datatype-new!
   #:key       'roc
   #:name      (_ "ROC - Rate of Change")
   #:units     'roc
   #:linestyle 'line
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Momentum and Rate of Change")
   #:proc      roc-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'momentum-days
	  #:type    'integer
	  #:minimum 1
	  #:default 40))))  ;; same as momentum.scm
