;; Relative Strength Index (RSI).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart rsi)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.incrediblecharts.com/technical/relative_strength_index.htm
;; http://www.incrediblecharts.com/technical/rsi_construction.htm
;;     Wilder's EMA reckoning, simple average for initial.
;;
;; http://www.stockcharts.com/education/IndicatorAnalysis/indic_RSI.html
;;     Numerical examples, graph of DELL.
;;
;; http://trader.online.pl/MSZ/e-w-Chandes_Momentum_Oscillator.html
;;     Simple moving average RSI.
;;


;; Return a procedure which calculates an RSI on successive closing prices,
;; with smoothing by COUNT periods.  COUNT is interpreted by Wilder's
;; reckoning, so alpha = 1/COUNT.
;;
;; Each call (PROC close) enters a new point into the window, and the return
;; is the RSI up to (and including) that value, or #f if not enough points
;; yet.
;;
;; An RSI using an EMA is in theory influenced by all preceding data, but
;; `rsi-warmup-count' below is designed to determine a warmup count, both
;; for the EMA or SMA cases.
;;
(define*-public (rsi-calc-proc count #:optional use-sma)

  (if (not use-sma)
      (set! count (ema-count-wilder count)))

  (let ((prev-close #f)
	(up-proc     (if use-sma
			 (sum-calc-proc count '())
			 (ema-calc-proc count)))
	(down-proc   (if use-sma
			 (sum-calc-proc count '())
			 (ema-calc-proc count))))
    (lambda (close)
      (define ret #f)
      (if prev-close
	  (let ((avg-gain (up-proc   (max 0 (- close prev-close))))
		(avg-loss (down-proc (max 0 (- prev-close close)))))
	    (and avg-gain
		 avg-loss
		 (let ((total (+ avg-loss avg-gain)))
		   (set! ret (if (zero? total)
				 50
				 (* 100.0 (/ avg-gain total))))))))
      (set! prev-close close)
      ret)))

(define*-public (rsi-warmup-count count #:optional use-sma)
  (if use-sma
      (1- count)
      (1+ (ema-warmup-count (ema-count-wilder count)))))


(define (rsi-fill series new-array lo hi)
  (receive-list (count use-sma)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos parent #:before lo
					     #:count (rsi-warmup-count
						      count use-sma)))
	   (array  (series-array-closes parent start hi))
	   (rsi    (rsi-calc-proc count use-sma)))

      ;; warmup
      (do ((i start (1+ i)))
	  ((>= i lo))
	(and-let* ((this (array-ref array i)))
	  (rsi this)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (rsi this) i))))))

(define*-public (rsi-series series count #:optional use-sma)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a"
					  (if use-sma
					      (_ "SMA RSI")
					      (_ "RSI"))
					  count))
    (series-datatype!  new-series rsi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list count use-sma))
    (series-require!   new-series (indicator-require-parts
				   rsi-fill (rsi-warmup-count count use-sma)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public rsi-datatype
  (datatype-new!
   #:key     'rsi
   #:name    (_ "RSI - Relative Strength Index")
   #:units   'percentage
   #:minimum 0
   #:maximum 100
   #:hlines  '(30 50 70)
   #:manual  (_ "CONTEXT(manual-node): Relative Strength Index")
   #:proc    rsi-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'rsi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14)
	 (datatype-param-new
	  #:key     'rsi-use-sma
	  #:name    (_ "SMA")
	  #:type    'boolean
	  #:default #f))))
