;; R-Squared indicator.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart rsquared)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.traderslog.com/r-squared.htm
;;     Sample Coors (NYSE symbol "TAP") from Feb 2002 (the year isn't shown,
;;     but is 2002).
;;
;; http://mathworld.wolfram.com/CorrelationCoefficient.html
;;     Explanation of general case, incl formulas.
;;



;; Return the triangular number of N, ie. 1+2+...+N, which is N*(N+1)/2.
;;
(define (triangular n)
  (ash (* n (1+ n)) -1))


;; Return the factor for r-squared arising from the denominator Y variance.
;; This is
;;
;;                            1
;;      -------------------------------------------------
;;      n * (1^2 + 2^2 + ... + n^2) - (1 + 2 + ... + n)^2
;;
;; and the denominator here is (n^4 - n^2)/12.  See devel/closed-form.scm
;; for checking this.  Same func in linreg-stderr.scm too.
;;
(define (yfactor n)
  (if (< n 2)
      1.0
      (/ 12.0 (* n n (1- (* n n))))))


;; Return a procedure which calculates an R-squared indicator over a window
;; of COUNT many points.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the R-squared up to and including that point, or #f if not enough
;; points yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;

(define-public (rsquared-calc-proc count)

  ;; This is the correlation coefficient of the last COUNT many X values,
  ;; against a set of Y values 1, 2, 3, ..., COUNT.  Or fewer than COUNT
  ;; until that many values have been seen.  The formula in the manual is
  ;;
  ;;                (Covariance X,Y)^2
  ;;     r^2 = ---------------------------
  ;;           (Variance X) * (Variance Y)
  ;;
  ;; which is
  ;;
  ;;                       (Sum X*Y)/N - (Sum X)/N * (Sum Y)/N
  ;;     r^2 = -------------------------------------------------------------
  ;;           ((Sum X^2)/N - (Sum X)^2/N^2) * ((Sum Y^2)/N - (Sum Y)^2/N^2)
  ;;
  ;; But in the code a factor of N^2 is put through numerator and
  ;; denominator to give
  ;;
  ;;                       N * Sum X*Y - Sum X * Sum Y
  ;;     r^2 = -----------------------------------------------------
  ;;           (N * Sum X^2 - (Sum X)^2) * (N * Sum Y^2 - (Sum Y)^2)
  ;;
  ;; In the degenerate case N==1, the variance of the Y values is 0, so
  ;; r^2=0/0.  We return 1 in that case, considering a single value is
  ;; perfectly correlated.
  ;;
  ;; If all the X values are the same then r^2 = 0/0.  We return 1 in that
  ;; case, because those X values are a perfect straight line, a horizontal
  ;; one.
  ;;
  ;;
  ;; X-DATA is a circular list of the last COUNT many X values.  The value
  ;; at X-DATA itself is the oldest.
  ;;
  ;; POINTS is the number of non-#f values in X-DATA.
  ;;
  ;; X-TOTAL is the sum of the values in X-DATA.
  ;;
  ;; XX-TOTAL is the sum of the squares of the values in X-DATA.
  ;;
  ;; Y-TOTAL is the sum of the Y values corresponding to the values in
  ;; X-DATA, which means 1+2+...+POINTS, which is a triangular number.
  ;; This varies until POINTS==COUNT is reached and is then a constant.
  ;;
  ;; Y-FAC is the variance factor in the denominator, see `yfactor' above.
  ;; This varies until POINTS==COUNT is reached and is then a constant.
  ;;
  ;; XY-TOTAL is the sum of the product x*y for each x,y pair, which means
  ;; x[1]*1 + x[2]*2 + ... + x[POINTS]*POINTS.  When shifting out an old X,
  ;; the weighting of each x in the sum is reduced by subtracting X-TOTAL.
  ;;
  ;; The calculation ensures that if all X values coming in are exact then
  ;; the sums X-TOTAL, XX-TOTAL and XY-TOTAL are exact.  The value returned
  ;; is always inexact though.
  ;;

  (define x-data   (make-circular-list count #f))
  (define points   0)
  (define x-total  0)
  (define xx-total 0)
  (define y-total  0)
  (define y-fac    #f)
  (define xy-total 0)
  (define min-points (ceiling-exact (* indicator-minimum-fraction count)))

  (lambda (x)

    (if (car x-data)
	;; drop oldest point
	(begin
	  (set! xx-total (- xx-total (* (car x-data) (car x-data))))
	  (set! xy-total (- xy-total x-total))
	  (set! x-total  (- x-total (car x-data))))

	;; otherwise gaining a point
	(begin
	  (set! points  (1+ points))
	  (set! y-total (triangular points))
	  (set! y-fac   (yfactor points))))

    (set! x-total  (+ x-total x))
    (set! xx-total (+ xx-total (* x x)))
    (set! xy-total (+ xy-total (* x points)))

    ;; replace oldest with new point
    (set-car! x-data x)
    (set! x-data (cdr x-data))

    (and (>= points min-points)
	 (let ((den (- (* xx-total points) (* x-total x-total))))
	   (if (zero? den)
	       1
	       (* y-fac (/ (square (- (* xy-total points) (* x-total y-total)))
			   den)))))))



(define (rsquared-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (rsquared-calc-proc count)))

    ;; warmup
    (for-each proc
	      (series-prev-n-closes parent
				    #:before lo
				    #:count (1- count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (proc this) i)))))

(define-public (rsquared-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "R-Squared") count))
    (series-datatype!  new-series rsquared-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   rsquared-fill
				   (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public rsquared-datatype
  (datatype-new!
   #:key       'rsquared
   #:name      (_ "R-Squared")
   #:units     'zero-to-one
   #:minimum   0
   #:maximum   1
   #:manual    (_ "CONTEXT(manual-node): R-Squared Index")
   #:proc      rsquared-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'rsquared-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
