;; Relative Volatility Index (RVI).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart rvi)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart stddev))


;; http://www.fmlabs.com/reference/RVI.htm
;;     Formula, some description.
;;
;; http://trader.online.pl/MSZ/e-w-Relative_Volatility_Index_RVI.html
;;     Reproducing equis.com original 1993 version, adding sample of Kredyt
;;     Bank (which is not in Yahoo, or only as .PK pink sheets).
;; http://trader.online.pl/MSZ/e-w-Inertia.html
;;     1995 revision.
;;
;; http://www.equis.com/customer/resources/formulas/formula.aspx?Id=52
;;     Original 1993.
;;
;; http://store.traders.com/-v11-c06-therela-pdf.html
;;     TASC 1993.
;;
;; http://store.traders.com/-v13-c09-refinin-pdf.html
;;     TASC 1995, high/low combination, and introducing Inertia.
;;
;; http://store.traders.com/v1340tradtip.html
;;     TASC traders tips for sale.
;;



;; Return a procedure which calculates a relative volatility index, using
;; Dorsey's original 1993 definition, over an accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the RVI up to (and including) that value, or #f if not enough points
;; yet.
;;
;; An RVI is in theory influenced by all preceding data, but
;; `rvi-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (rvi-warmup-count STDDEV-COUNT SMOOTH-COUNT) many
;; values, the next call will have an omitted weight of no more than 0.1% of
;; the total.  Omitting 0.1% should be negligable, unless past values are
;; ridiculously bigger than recent ones.
;;
(define-public (rvi-original-calc-proc stddev-count smooth-count)
  (let ((up-proc     (ema-calc-proc (ema-count-wilder smooth-count)))
	(down-proc   (ema-calc-proc (ema-count-wilder smooth-count)))
	(stddev-proc (stddev-calc-proc stddev-count))
	(prev-price  #f))

    (lambda (price)
      (let ((ret    #f)
	    (stddev (stddev-proc price)))
	(and stddev
	     prev-price
	     (let ((avg-up   (up-proc   (if (> price prev-price) stddev 0)))
		   (avg-down (down-proc (if (> price prev-price) 0 stddev))))
	       (and avg-up
		    avg-down
		    (let ((total (+ avg-down avg-up)))
		      (set! ret (if (zero? total)
				    50
				    (* 100.0 (/ avg-up total))))))))
	(set! prev-price price)
	ret))))


;; Return a procedure which calculates a relative volatility index, using
;; Dorsey's 1995 revised definition, over an accumulated window.
;;
;; Each call (PROC high low) enters a new day's trading range into the
;; window, and the return is the RVI up to (and including) that value, or #f
;; if not enough points yet.
;;
;; HIGH and LOW must not be #f, close values can be substituted if
;; necessary, and in that case the calculation ends up falling back to
;; Dorsey's 1993 original
;;
;; An RVI is in theory influenced by all preceding data, but
;; `rvi-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (rvi-warmup-count STDDEV-COUNT SMOOTH-COUNT) many
;; values, the next call will have an omitted weight of no more than 0.1% of
;; the total.  Omitting 0.1% should be negligable, unless past values are
;; ridiculously bigger than recent ones.
;;
(define-public (rvi-calc-proc stddev-count smooth-count)
  (let ((high-proc (rvi-original-calc-proc stddev-count smooth-count))
	(low-proc  (rvi-original-calc-proc stddev-count smooth-count)))

    (lambda (high low)
      (let ((h-rvi (high-proc high))
	    (l-rvi (low-proc  low)))
	(and h-rvi
	     l-rvi
	     (* 0.5 (+ h-rvi l-rvi)))))))

;; The first EMA warmup point will require an additional STDDEV-COUNT - 1
;; values to get the stddev for that first EMA.
;;
(define-public (rvi-warmup-count stddev-count smooth-count)
  (+ stddev-count -1 (ema-warmup-count (ema-count-wilder smooth-count))))



(define (rvi-fill series new-array lo hi)
  (receive-list (stddev-count smooth-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (series-prev-n-closes-pos
		    parent #:before lo
		    #:count (rvi-warmup-count stddev-count smooth-count)))
	   (array  (series-array parent start hi))
	   (proc   (rvi-calc-proc stddev-count smooth-count)))

      ;; warmup
      (do ((i start (1+ i)))
	  ((>= i lo))
	(and-let* ((close (array-ref array i 3)))
	  (proc (or (array-ref array i 1) close)    ;; high
		(or (array-ref array i 2) close)))) ;; low

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((close (array-ref array i 3)))
	  (array-set! new-array (proc (or (array-ref array i 1) close)  ;; high
				      (or (array-ref array i 2) close)) ;; low
		      i))))))

(define-public (rvi-series series stddev-count smooth-count)
  (ohlcv-required "RVI" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a, ~a" (_ "RVI")
					  stddev-count smooth-count))
    (series-datatype!  new-series rvi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list stddev-count smooth-count))
    (series-require!   new-series
		       (indicator-require-parts
			rvi-fill (rvi-warmup-count stddev-count smooth-count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public rvi-datatype
  (datatype-new!
   #:key     'rvi
   #:name    (_ "RVI - Relative Volatility Index")
   #:input   datatype-ohlcv
   #:units   'percentage
   #:minimum 0
   #:maximum 100
   #:hlines  '(40 50 60)
   #:manual  (_ "CONTEXT(manual-node): Relative Volatility Index")
   #:proc    rvi-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'rvi-stddev-days
	  #:name    (_ "Stddev Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 10)
	 (datatype-param-new
	  #:key     'rvi-days
	  #:name    (_ "Smooth Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
