;; Random Walk Index (RWI).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart rwi)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (chart atr)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://store.traders.com/-v10-c01-sideran-pdf.html
;;     Abstract of sidebar calculation, excess of channel height (scaled by
;;     one-day move) over square root.
;; http://store.traders.com/-v11-c11-sideran-pdf.html
;;     Similar.
;;     "Are There Persistent Cycles", E. Michael Poulos, September 1992
;;     TASC.
;;
;; http://www.prophet.net/analyze/popglossary.jsp?studyid=RWI
;;     Sample of McDonalds, year not shown but looks like May 2001.
;;
;; http://www.paritech.com/paritech-site/education/technical/indicators/momentum/random.asp
;;     Description, no formulas, sample of DJS.AX, year not shown but is
;;     1999 (RWI period not shown either).
;;
;; http://trader.online.pl/MSZ/e-w-Random_Walk_Index.html
;;     Formula.
;;
;; http://trader.online.pl/MSZ/e-w-Random_Walk_Index_II.html
;;     Long and short term high and low forms.  Chart of Telecom Polska
;;     (TPSA.WA, London international order book TPSD.IL).
;;
;; http://www.linnsoft.com/tour/techind/ranWalk.htm
;;     Formula.
;;
;; http://www.equis.com/customer/resources/formulas/formula.aspx?Id=48
;;     Metastock formula, but only the RWI High part
;;
;; http://xeatrade.com/trading/2/R/1160.html
;;     Metastock formula.
;;
;; http://www.tradecision.com/support/indicators/random_walk_index.htm
;;     Tradedecsion formula for 9-day, cumulative sum of 1-day ATRs.
;;
;; http://technical.traders.com/tradersonline/display.asp?art=606
;;     Dennis Peterson on a random walk trading system.  Sample Nasdaq
;;     composite (QQQ, yahoo now ^IXIC) from 1999/2000.
;;
;; http://www.sirtrade.com/serv03.htm
;;     Sample code of programming service by Pierre Orphelin, having
;;     struggled a bit with the correct definition.  (Rejected incorrect
;;     code using the highest high in the past N days, maybe.)
;;
;; http://www.biz-analyst.com/library/stocks_commodities_trend.pdf
;;     Copies of some TASC articles, including "Futures According to Trend
;;     Tendency" V.10:1 by Michael Poulos, which includes a sidebar on his
;;     RWI.  (Doesn't specify SMA or EMA for average.)
;;
;; http://ta-glossary.netfirms.com/taglossary-R.htm
;;     Description, sample Bradley Pharma (NYSE symbol BDY) from 2004.
;;
;; http://www.wallstreettape.com/tutorials/ta/c37.asp
;;     Chartfilter description.
;;
;; 


;; Return a procedure which calculates a random walk index over a window of
;; COUNT many points.
;;
;; Each call (PROC high low close) enters a new point into the window, and
;; the return is multiple values `(values RWI-HIGH RWI-LOW)' up to and
;; including that point, or `(values #f #f)' if not enough points yet.
;;
;; The window can be primed by calling with COUNT many points preceding the
;; first desired.
;;
(define-public (rwi-calc-proc count)

  ;; HIGH-LIST and LOW-LIST are the previous COUNT many high and low values
  ;; (or fewer than COUNT until that many received).  The first in each list
  ;; are yesterday's values.
  ;;
  ;; PREV-TR is yesterday's true range, ie. from the high/low/close of the
  ;; immediately previous call.
  ;;
  ;; TR-LIST is the previous COUNT many true range values before the PREV-TR
  ;; one (or fewer than COUNT until that many received).  The first in
  ;; TR-LIST is from the day before yesterday.
  ;;
  ;; FACTOR-LIST is sqrt(2), sqrt(3), ..., sqrt(COUNT), which are the
  ;; multiplying factors to apply.
  ;;
  ;; POINTS is how many entries are in each of HIGH-LIST, LOW-LIST and
  ;; TR-LIST, and MIN-POINTS is how many to demand before letting results go
  ;; out.
  ;;
  (define tr-proc     (true-range-calc-proc))
  (define high-list   '())
  (define low-list    '())
  (define tr-list     '())
  (define prev-tr     #f)
  (define factor-list (map! sqrt (iota count 2)))
  (define points      0)
  (define min-points  (ceiling-exact (* indicator-minimum-fraction count)))

  (lambda (high low close)
    ;; just use closes if no high/low
    (or high (set! high close))
    (or low  (set! low  close))

    (let ((rwi-low  #f)
	  (rwi-high #f))
      (if prev-tr
	  (let ((tr-sum prev-tr))
	    (for-each (lambda (prev-high prev-low tr factor)
			(set! tr-sum (+ tr-sum tr))
			(if (not (zero? tr-sum))
			    (begin
			      (set! rwi-high
				    (max-maybe rwi-high
					       (* factor (/ (- high prev-low)
							    tr-sum))))
			      (set! rwi-low
				    (max-maybe rwi-low
					       (* factor (/ (- prev-high low)
							    tr-sum)))))))
		      high-list low-list tr-list factor-list)

	    (if (>= points count)
		(begin
		  (set! high-list (take! high-list (1- count)))
		  (set! low-list  (take! low-list  (1- count)))
		  (set! tr-list   (take! tr-list   (1- count)))))
	    (set! points (min (1+ points) count))
	    (set! high-list (cons high    high-list))
	    (set! low-list  (cons low     low-list))
	    (set! tr-list   (cons prev-tr tr-list))))

      (set! prev-tr (tr-proc high low close))
      (if (>= points min-points)
	  (values rwi-high rwi-low)
	  (values #f #f)))))



(define-public (rwi-warmup-count count)
  count)


(define (rwi-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (warmup (rwi-warmup-count count))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count warmup))
	 (array  (series-array parent start hi))
	 (proc   (rwi-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 1)    ;; high
		(array-ref array i 2)    ;; low
		(array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (receive (rwi-high rwi-low)
	      (proc (array-ref array i 1)  ;; high
		    (array-ref array i 2)  ;; low
		    (array-ref array i 3)) ;; close
	    (array-set! new-array rwi-high i 0)
	    (array-set! new-array rwi-low  i 1))))))

(define-public (rwi-series series count)
  (ohlcv-required "Random Walk Index" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "RWI") count))
    (series-datatype!  new-series rwi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   rwi-fill (rwi-warmup-count count) '(0 1)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public rwi-datatype
  (datatype-new!
   #:key       'rwi
   #:name      (_ "RWI - Random Walk Index")
   #:input     datatype-ohlcv
   #:units     'rwi
   #:linestyle 'dmi
   #:minimum   0
   #:hlines    '(1)
   #:manual    (_ "CONTEXT(manual-node): Random Walk Index")
   #:proc      rwi-series
   #:closes-apply-arg '(0)
   #:parameter-list
   (list (datatype-param-new
	  #:key     'rwi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 9))))
