;; Sydney Futures Exchange (SFE) setups.

;; Copyright 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart sfe)
  #:autoload   (ice-9 format) (format)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart newzealand))



;; IR - last trading midday day before settle; settle second friday of
;; delivery month


(define-public (sfe-symbol? symbol)
  (string-suffix? ".SFE" symbol))

(define-public (sfe-newzealand-symbol? symbol)
  (and (sfe-symbol? symbol)
       (sfe-newzealand-commodity? (chart-symbol-commodity symbol))))

(define-public (sfe-sydney-symbol? symbol)
  (and (sfe-symbol? symbol)
       (not (sfe-newzealand-commodity? (chart-symbol-commodity symbol)))))

(define (sfe-newzealand-commodity? commodity)
  (->bool (member commodity
		  '("BB"     ;; NZ 90-day bank bills
		    "TN"     ;; NZ 3-year bonds
		    "TY"     ;; NZ 10-year bonds
		    "ZI"     ;; NZSX15 futures
		    "ZO")))) ;; NZ overnight cash

(symbol-timezone! sfe-newzealand-symbol? (_ "New Zealand") timezone-newzealand)
(symbol-timezone! sfe-symbol? (_ "Sydney") timezone-sydney)

;; This specs link is of fairly limited use, since it doesn't go to
;; something directly about the displayed symbol, but it's included for
;; consistency with other futures sources and because it's not easy to find
;; from the home page.
;;
(weblink-handler! sfe-symbol?
  (lambda (symbol)
    (list (_ "SFE _Contract Specifications")
	  (_ "Open web browser at the Sydney Futures Exchange contract specifications page (one PDF document for all commodities)")
	  "http://www.sfe.com.au/content/sfe/trading/contract_specs.htm")))


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the 15 minute delayed quotes for each commodity, such as for
;; Austrlian contracts,
;;
;;     http://www.sfe.com.au/content/prices/rtp15SFYT.html
;;
;; and "ZF" for New Zealand contracts,
;;
;;     http://www.sfe.com.au/content/prices/rtp15ZFBB.html
;;
;; Each are about 5 kbytes each.  The quotes include SYCOM overnight
;; trading, so it should in theory be worth refreshing any time day or
;; night.
;;
;; The server provides ETag and Last-Modified, which we use to avoid
;; re-downloading when unchanged.  Changes seem to occur once every minute
;; (every minute from some random starting point).
;;
;;
;; There's a whole market summary page of about 16 kbytes, with the front
;; few contracts.  It's not used at the moment.
;;
;;     http://www.sfe.com.au/content/prices/rtp15marketsummary.html
;;


;; drop the longest final suffix of LST whose elements all satisfy PRED
(define (drop-while-right! pred lst)
  (reverse! (drop-while pred (reverse lst))))

;; return list of `latest' records
(define (sfe-body->latest-list commodity source headers body)
  (define name #f)

  ;; on an individual commodity page there's a name in first table, beside
  ;; the REFRESH button
  (if (eq? source 'sfe-indiv)
      (let ((row-list (html-table-rows body)))
	(set! name (any (lambda (row)
			  (and (length>=? row 2)
			       (equal? "REFRESH" (second row))
			       (first row)))
			row-list))))

  (let* ((row-list    (html-table-rows ;; second table
		       body (string-contains-after-ci body "<table")))

	 ;; 15 minute delay
	 (quote-tm    (localtime (- (current-time) #,(hms->seconds 0 15 0))
				 (timezone-sydney)))
	 (quote-adate (tm->adate quote-tm))
	 (quote-time  (tm->time quote-tm)))

    ;; check table headings are what we expect
    (let ((headings (first row-list)))
      (if (string-null? (first headings))
	  (set! headings (cdr headings)))
      (or (equal? headings '("Expiry" "Bid" "Ask" "Open" "High" "Low"
			     "Last Trade" "Last Trade Date" "Last Trade Time"
			     "Change" "Traded Volume" "Previous Settlement"))
	  (error "SFE: unrecognised quote table columns" headings)))
    (set! row-list (cdr row-list))

    ;; remove "SFE Total" and "NZ Total" rows at end
    (set! row-list
	  (drop-while-right! (lambda (row)
			       (string-contains (first row) "Total"))
			     row-list))

    ;; html-table-rows doesn't handle rowspan, so replicate symbol into the
    ;; subsequent rows it applies to; and insert symbol into the data from a
    ;; single commodity body
    (set! row-list
	  (map-in-order (lambda (row)
			  (case (length row)
			    ((12)
			     (or commodity
				 (error "SFE: oops, no commodity in quote"))
			     (cons commodity row))
			    ((13)
			     (set! commodity (first row))
			     row)
			    (else
			     row)))
			row-list))

    (latest-list-add-front-months
     (map-rows! (lambda (commodity month bid offer open high low
				   last last-adate last-time
				   change volume prev)
		  (define mdate (m/y-str->mdate month))

		  ;; date and time can be empty when no trades yet
		  (set! last-adate (and (not (string-null? last-adate))
					(d/m/y-str->adate last-adate)))
		  (set! last-time  (and (not (string-null? last-time))
					(h:m:s-str->seconds last-time)))

		  (latest-new
		   #:symbol (string-append commodity
					   " " (mdate->MMM_YY-str mdate)
					   ".SFE")
		   #:name           name
		   #:quote-adate    quote-adate
		   #:quote-time     quote-time
		   #:bid            bid
		   #:offer          offer
		   #:last-adate     last-adate
		   #:last-time      last-time
		   #:open           open
		   #:high           high
		   #:low            low
		   #:last           last
		   #:change         change
		   #:prev           prev
		   #:volume         volume
		   #:contract-mdate mdate
		   #:source         source
		   #:headers        headers))
		row-list))))

(define (sfe-latest-get symbol-list extra-list proc)

;;   (define indiv-cost (* (+ 5000 http-get-cost)
;; 			;; how many different commodities in SYMBOL-LIST
;; 			(length (delete-duplicates (map chart-symbol-commodity
;; 							symbol-list)))))
;;   (define whole-cost (+ 16000 http-get-cost))

  (if #t ;; (<= whole-cost indiv-cost)

      (receive (headers body)
	  (apply http-request
		 "http://www.sfe.com.au/content/prices/rtp15marketsummary.html"
		 #:want-ok #t
		 (latest-tag-args 'sfe-whole symbol-list))
	(if (http-headers-ok? headers) ;; not 304 Not Modified
	    (proc (sfe-body->latest-list #f 'sfe-whole headers body))))

      (for-each
       (lambda (symbol-list) ;; of one commodity
	 (define commodity (chart-symbol-commodity (first symbol-list)))
	 (receive (headers body)
	     (apply http-request
		    (string-append
		     "http://www.sfe.com.au/content/prices/rtp15"
		     (if (sfe-newzealand-commodity? commodity) "ZF" "SF")
		     commodity ".html")
		    (latest-tag-args 'sfe-indiv symbol-list))
	   (let ((status-code (http-headers-status-code headers)))
	     (cond ((http-status-code-ok? status-code)
		    (proc (sfe-body->latest-list commodity 'sfe-indiv
						 headers body)))
		   ((= status-code 304)
		    ;; 304 Not Modified, no new data
		    )
		   ((= status-code 404)
		    ;; 404 Not Found, unknown symbol
		    (proc (map latest-new-unknown symbol-list)))
		   (else
		    (http-headers-throw headers))))))

       (partition-equal chart-symbol-commodity
			symbol-list))))

(latest-handler! #:selector sfe-symbol?
		 #:handler sfe-latest-get)


;;-----------------------------------------------------------------------------
;; database update
;;
;; This uses the daily report files at
;;
;;     http://www.sfe.com.au/content/sfe/marketdata/eod_sum.asp
;;
;; like (yymmdd)
;;
;;     http://www.sfe.com.au/Content/reports/EODWebMarketSummary061228SFD.htm
;;
;; and for New Zealand futures
;;
;;     http://www.sfe.com.au/Content/reports/EODWebMarketSummary061228ZFD.htm
;;
;; The summary page says reports are a "five day rolling history", but they
;; actually go right back, for at least a couple of years.  For now we just
;; follow the 5 days advised, mainly because at 200 kbytes per day the
;; amount downloaded soon becomes very big.


(define want-headings '("Expiry" "Open" "High" "Low" "Last" "Sett"
			 "Sett Chg" "Op Int" "Op Int Chg" "Volume"))

;; call (PROC series) for each series in BODY, which is tdate T
(define (sfe-process-download dataname symbol-list tdate body)

  (define row-list (html-table-rows body
				    ;; second table
				    (string-contains body "<TABLE ")))

  ;; have seen "No data available" in
  ;; http://www.sfe.com.au/Content/reports/EODWebMarketSummary070102ZFD.htm
  ;; instead of the usual 404 for non-trading days.
  ;;
  (if (member '("No data available") row-list)
      (set! row-list (list want-headings))) ;; dummy with headings only

  ;; heading rows at the start of each commodity section
  (or (member want-headings row-list)
      (error "SFE: unrecognised or missing daily report headings"))
  (set! row-list (delete! want-headings row-list))

  ;; empty rows
  (set! row-list (delete! '("") row-list))
  (set! row-list (delete! '("" "") row-list))

  ;; rows "Click here to view settlement price and volume graph"
  (set! row-list (remove! (lambda (row)
			    (string-prefix? "Click here" (first row)))
			  row-list))

  ;; split commodity+name heading row, and join onto applicable rows
  (let ((commodity  #f)
	(name       #f))
    (set! row-list
	  (filter-map (lambda (row)
			(cond ((regexp-exec #,(regexp "^([A-Z]+) - (.*)")
					    (first row))
			       => (lambda (m)
				    (set! commodity (match:substring m 1))
				    (set! name (match:substring m 2))
				    #f))
			      (else
			       (or commodity
				   (error "SFE: no symbol before first data:"
					  row))
			       (cons* commodity name row))))
		      row-list)))

  (download-process
   #:module      (_ "SFE")
   #:symbol-list symbol-list
   #:lo          tdate
   #:hi          tdate
   #:copyright "http://www.sfe.com.au/index.html?content/aboutsfe/disclaimer.htm"
   #:row-list
   (map-rows! (lambda (commodity name month open high low last
				 settle settle-change
				 openint openint-change volume)

		;; ignore row with settle of "-", it means settlement day
		;; which is the day after the final trading day
		(and (string<> "-" settle)
		     (list #:tdate         tdate
			   #:commodity     commodity
			   #:mdate         (m/y-str->mdate month)
			   #:name          name
			   #:open    	  open
			   #:high  	  high
			   #:low   	  low
			   #:close 	  settle
			   #:volume        volume
			   #:open-interest openint)))
	      row-list)))

;; Return the tdate we expect to have data available for now.
;; Each day is available from 7:30am Sydney time the following day,
;; including Saturday 7:30am for Friday's data.
;;
(define (sfe-available-tdate)
  (let* ((tm    (localtime (current-time) (timezone-sydney)))
	 (tdate (tm->tdate tm))) ;; today, or friday on weekend
    ;; on Mon to Fri prev tdate is available, on Weekend no change (Fri avail)
    (if (tm:weekday? tm)
	(set! tdate (1- tdate)))
    ;; on Tue to Sat before 7:30am, only the day before yesterday available
    (if (and (within-range? 2 6 (tm:wday tm))
	     (two< (tm:hour tm) (tm:min tm) 7 30))
	(set! tdate (1- tdate)))
    tdate))

(define (sfe-data-url tdate xf)
  (receive-list (year month day)
      (tdate->ymd tdate)
    (format #f "http://www.sfe.com.au/Content/reports/EODWebMarketSummary~2,'0d~2,'0d~2,'0d~aD.htm" (modulo year 100) month day xf)))

;; SYMBOL-LIST is all from the same Sydney or New Zealand, and xf is either
;; "SF" or "ZF" accordingly
;;
(define (sfe-download-xf symbol-list dataname xf selector)

  (let* ((avail-tdate  (sfe-available-tdate))
	 ;; oldest available from SFE
	 ;; avail-5 is the 6 latest days, but try 3 further back in case of
	 ;; holidays or our idea of the available is wrong somehow
	 (oldest-tdate (- avail-tdate 10)))

    ;; start from oldest in SYMBOL-LIST, but no more than OLDEST-TDATE
    (do ((t (max oldest-tdate
		 (apply min (map download-start-tdate symbol-list)))
	    (1+ t)))
	((> t avail-tdate))
      (download-status (_ "SFE") dataname (tdate-range-string t t))

      (receive (headers body)
	  (http-request (sfe-data-url t xf))
	(let ((status-code (http-headers-status-code headers)))
	  (cond ((http-status-code-ok? status-code)
		 (sfe-process-download dataname
				       (download-also symbol-list
						      #:selector    selector
						      #:start-tdate t
						      #:end-tdate   t)
				       t body))
		((not (= 404 status-code))
		 ;; 404 for public holidays, anything else is bad
		 (http-headers-throw headers))))))))

(define (sfe-download symbol-list)
  (for-each (lambda (symbol-list)
	      (if (sfe-newzealand-symbol? (first symbol-list))
		  (sfe-download-xf symbol-list (_ "New Zealand data")
				   "ZF" sfe-newzealand-symbol?)
		  (sfe-download-xf symbol-list (_ "data")
				   "SF" sfe-sydney-symbol?)))
	    (partition-equal sfe-newzealand-symbol? symbol-list)))

(define (sfe-backto want-tdate symbol-list)
  (download-message
   (string-append
    (_ "SFE") ": "
    (_ "no further historical data available"))))

(download-handler! #:selector   sfe-symbol?
		   #:update     sfe-download
		   #:avail-proc sfe-available-tdate
		   #:backto     sfe-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any sfe-symbol? symbol-list)
			      (download-now-all-commodities-and-months))))


;;-----------------------------------------------------------------------------
;; setups - intraday

;; The year-to-date graph shown at the bottom of a quotes page like
;;
;;     http://www.sfe.com.au/content/prices/rtp15sfYT.html
;; is
;;     http://www.sfe.com.au/content/graphs/YTDWebPriceGraphYT.png
;;
;; The images are about 5 kbytes and have etag and last-modified.

(define (sfe-intraday-url symbol mode)
  (string-append "http://www.sfe.com.au/content/graphs/YTDWebPriceGraph"
		 (url-quote (chart-symbol-sans-dot symbol))
		 ".png"))

(intraday-url-proc! #:symbol-pred sfe-symbol?
		    #:url-proc    sfe-intraday-url
		    #:modes-proc  (lambda ()
				    ;; no accelerator since only one entry
				    (list (_ "Year"))))
