;; Singapore Commodities Exchange (SICOM) setups.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart sicom)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (sicom-symbol? symbol)
  (string-suffix? ".SICOM" symbol))

(symbol-timezone! sicom-symbol? (_ "Singapore") timezone-singapore)
(source-help! sicom-symbol?
	      (_ "CONTEXT(manual-node): Singapore Commodities Exchange"))


;;-----------------------------------------------------------------------------
;; weblinks - contract specs
;;
;; These links are as per the home page and they open in a frame.  The
;; individual pages are http://www.sicom.com.sg/robusta.htm etc.
;;

(define sicom-weblinks-alist
  '(("RS" . "http://www.sicom.com.sg/index_sub.asp?content=rss1")
    ("RT" . "http://www.sicom.com.sg/index_sub.asp?content=rss3")
    ("TF" . "http://www.sicom.com.sg/index_sub.asp?content=tsr20")
    ("RI" . "http://www.sicom.com.sg/index_sub.asp?content=rss3index")
    ("CF" . "http://www.sicom.com.sg/index_sub.asp?content=robusta")))

(weblink-handler! sicom-symbol?
  (lambda (symbol)
    (list (_ "SICOM _Contract Specifications")
	  (_ "Open web browser at the Singapore Commodities Exchange contract specifications for this commodity")
	  (assoc-ref sicom-weblinks-alist (chart-symbol-commodity symbol)))))


;;-----------------------------------------------------------------------------
;; misc

;; return mdate for STR like "MAY 05", #f for quarter "J/M" "A/J" "J/S" "O/D"
(define (MMM_yy-str->mdate str)
  (receive-list (month year)
      (string-split str #\space)
    (and (not (member month '("J/M" "A/J" "J/S" "O/D"
			      "JFM" "AMJ" "JAS" "OND")))
	 (ymd->mdate (yyy-str->year year)
		     (Mmm-str->month month)
		     1))))

;; return the currency for a given COMMODITY
;; return #f for an unknown commodity
;;
;; contract specs:
;; http://www.sicom.com.sg/index_sub.asp?content=rss1
;; http://www.sicom.com.sg/index_sub.asp?content=rss3
;; http://www.sicom.com.sg/index_sub.asp?content=tsr20
;; http://www.sicom.com.sg/index_sub.asp?content=rss3index
;; http://www.sicom.com.sg/index_sub.asp?content=robusta
;;
(define (sicom-commodity->currency commodity)
  (assoc-ref '(("RS" . "SGD")  ;; RSS1
	       ("RT" . "USD")  ;; RSS3
	       ("TF" . "USD")  ;; TSR 20
	       ("RI" . "USD")  ;; RSS3 INDEX
	       ("CF" . "USD")) ;; Coffee
	     commodity))


;;-----------------------------------------------------------------------------
;; latest quotes - end of day prices
;;
;; This uses the following historical prices page (normally shown in a frame)
;;
;;     http://www.sicom.com.sg/dailyprice.asp
;;
;; which results in URLs like
;;
;;     http://www.sicom.com.sg/downloadHistoricalPrices.asp?downloaddate=9-20-2006
;;


;; return list of latest records
(define (sicom-hist->latest-list body)
  (define row-list (html-table-rows body ;; second <table>
				    (string-contains-after-ci body "<table")))

  ;; "Contract<br>Name" is collapsed to "ContractName" by html-table-rows,
  ;; maybe it should be <br> -> \n but at any rate
  (let ((headings (first row-list)))
    (or (equal? headings '("ContractName"
			   "ContractMonth"
			   "PreviousSettlement"
			   "High"
			   "Low"
			   "Close"
			   "Volume"
			   "OpenInterest"))
	(error "SICOM: unrecognised historical page headings:" headings)))
  (set! row-list (cdr row-list))

  (let* (;; eg. "<title>HISTORICAL PRICES ON 15 Aug 2006</title>"
	 (m        (must-match
		    (string-match "<title>HISTORICAL PRICES ON ([0-9]+) ([A-Za-z]+) ([0-9][0-9][0-9][0-9])</title>" body)))
	 (adate    (ymd->adate (string->number (match:substring m 3))
			       (Mmm-str->month (match:substring m 2))
			       (string->number (match:substring m 1))))
	 (ret      '()))

    (for-each (lambda (row)
		(receive-list (name month prev high low close volume int)
		    row

		  (and-let* ((commodity (daily-name->commodity name)))
		    (set! month (daily-mung-month-str month))

		    ;; commas seen in prices like coffee "2,257.00", strip
		    ;; them everywhere
		    (set! prev   (string-delete prev   #\,))
		    (set! high   (string-delete high   #\,))
		    (set! low    (string-delete low    #\,))
		    (set! close  (string-delete close  #\,))
		    (set! volume (string-delete volume #\,))
		    (set! int    (string-delete int    #\,))

		    (receive-list (decimals prev high low close)
			(strings->numbers+decimals prev high low close)

		      ;; mdate or #f for quarters
		      (define mdate (MMM_yy-str->mdate month))

		      (set! ret (cons (latest-new
				       #:symbol (string-append commodity
							       " " month
							       ".SICOM")
				       #:name       	name
				       #:last-adate 	adate
				       #:high       	high
				       #:low        	low
				       #:last       	close
				       #:change     	(- close prev)
				       #:volume     	volume
				       #:contract-mdate mdate
				       #:source         'sicom)
				      ret))))))

	      row-list)

    (latest-list-add-front-months ret)))

(define (sicom-csv-latest symbol-list proc)
  (define avail-tdate
    (first (sicom-tm->tdate-time-lock (localtime (current-time)
						 (timezone-singapore)))))

  ;; stop at the oldest date in existing records (latest-get here will look
  ;; at the database too)
  ;;
  ;; for an intraday, ie. has latest-last-time, then attempt that day
  ;; otherwise the record is already from a csv, attempt only the day after
  ;;
  (define lo-tdate (apply min-maybe
			  (map (lambda (symbol)
				 (let ((latest (latest-get symbol)))
				   (and (latest-last-tdate latest)
					(+ (latest-last-tdate latest)
					   (if (latest-last-time latest)
					       0 1)))))
			       symbol-list)))

  ;; limit how far back to look
  ;; allow for today (-0) and yesterday (-1) public holidays, so try
  ;; avail-2; and also an extraday avail-3 in case avail is one too big
  ;; perhaps
  (set! lo-tdate (max-maybe lo-tdate (- avail-tdate 3)))

  (let more ((tdate avail-tdate))
    (receive-list (year month day)
	(tdate->ymd tdate)
      (receive (headers body)
	  (http-request (format #f "http://www.sicom.com.sg/downloadHistoricalPrices.asp?downloaddate=~a-~a-~a" month day year)
			#:want-ok #t)
	(if (string-contains body "Data Not Available")
	    (if (> tdate lo-tdate) ;; message if no data this date, keep going
		(more (1- tdate)))
	    (proc (sicom-hist->latest-list body)))))))


;;-----------------------------------------------------------------------------
;; latest quotes - intraday html

;; return list (TDATE TIME) which is TM locked to within SICOM trading
;; hours, being weekdays 1000 to 1130, and 1530 to 1730.
;; the middle times are stretched by 2 minutes in case our system time is out
;;
(define (sicom-tm->tdate-time-lock tm)
  (let ((tdate-time (tm->tdate-time-within
		     tm #,(hms->seconds 10 0 0) #,(hms->seconds 17 30 0))))
    (if (within-range? #,(hms->seconds 11 32 0) #,(hms->seconds 15 28 0)
		       (second tdate-time))
	(set-second! tdate-time #,(hms->seconds 11 32 0)))
    tdate-time))

;; the contract specifications pages don't give symbols for RSS1 and RSS3
;; index, RS and RI were advised by email
(define html-name->commodity-alist
  '(("RSS1 (Singapore cents per kg)" . "RS")
    ("RSS3 (US cents per kg)"        . "RT")
    ("TSR 20 (US cents per kg)"      . "TF")
    ("RSS3 INDEX"                    . "RI")
    ("ROBUSTA COFFEE"                . "CF")))

(define (sicom-commodity->name commodity)
  (rassoc-ref html-name->commodity-alist commodity))

;; return list of latest records
(define (sicom-body->latest-list body)

  ;; for ease of regexp/newline
  (set! body (string-delete body #\cr))

  ;; Eg. "<tr><td>DELAYED PRICES</td><td align="right">13 Dec 2006 08:44AM</td></tr>"
  ;; the date/time is merely the current time, but using it makes certain of
  ;; showing the right thing even if our timezones are a bit off or whatever
  ;;
  (let* ((m     (must-match (regexp-exec #,(regexp "DELAYED PRICES[^\n]*>([0-9]+ [A-Za-z]+ [0-9][0-9][0-9][0-9]) ([0-9]+:[0-9][0-9][AP]M)" regexp/icase) body)))
	 (adate (d/m/y-str->adate (match:substring m 1)))
	 (time  (h:m:s-str->seconds (match:substring m 2)))
	 (ret   '()))

    (for-each
     (lambda (m)
       (let ((row-list (html-table-rows body (match:start m))))
	 ;; first table is empty
	 (and-let* ((          (not (null? row-list)))
		    (name      (first (first row-list)))
		    ;; only for commodity names we know
		    (commodity (assoc-ref html-name->commodity-alist name)))

	   ;; drop name row
	   (set! row-list (cdr row-list))

	   (let ((headings (first row-list)))
	     (or (equal? headings '("Contract Month" "Previous Sett." "High" "Low"
				    "Close" "Volume" "Open Interest"))
		 (error "SICOM: unrecognised data columns:"
			headings))
	     (set! row-list (cdr row-list)))

	   ;; remove final "Total" row
	   (set! row-list (alist-delete "Total" row-list))

	   (for-each
	    (lambda (row)
	      (receive-list (month prev high low close volume int)
		  row
		(define mdate (MMM_yy-str->mdate month))

		(if (string-prefix? "J/M" month)  ;; JFM
		    (string-set! month 1 #\F))
		(if (string-prefix? "A/J" month)  ;; AMJ
		    (string-set! month 1 #\M))
		(if (string-prefix? "J/S" month)  ;; JAS
		    (string-set! month 1 #\A))
		(if (string-prefix? "O/D" month)  ;; OND
		    (string-set! month 1 #\N))

		(set! ret (cons (latest-new
				 #:symbol (string-append commodity " "
							 month ".SICOM")
				 #:name           name
				 #:last-adate     adate
				 #:last-time      time

				 #:high           high
				 #:low            low
				 #:last           close
				 #:prev           prev
				 #:volume         volume
				 #:contract-mdate mdate
				 ;; no month note for months (with mdate), but
				 ;; keep it for quarters
				 #:note           (and (not mdate) month)
				 #:source         'sicom)
				ret))))

	    row-list))))

     (must-match (list-matches #,(regexp "^[ \t]*<table .*name=\"livechart\""
					 regexp/newline)
			       body)))

    (latest-list-add-front-months ret)))


;;-----------------------------------------------------------------------------
;; latest quotes
;;
;; This uses the intraday price page at
;;
;;     http://www.sicom.com.sg/delayedprices.asp
;;
;; and outside trading hours when that page is empty, the most recent daily
;; report the same as the data download below is used.
;;
;; The download and intraday give different commodity names but that doesn't
;; matter much.


(define (sicom-latest-get symbol-list extra-list proc)

  (receive (headers body)
      (http-request "http://www.sicom.com.sg/delayedprices.asp")
    (if (http-headers-ok? headers)
	(let ((latest-list (sicom-body->latest-list body)))
	  (if (not (null? latest-list))
	      (proc latest-list)

	      (sicom-csv-latest symbol-list proc))))))

(define (sicom-quote-adate-time symbol)
  (receive-list (avail-tdate avail-time)
      (sicom-tm->tdate-time-lock (localtime (current-time)
					    (timezone-singapore)))
    (list (tdate->adate avail-tdate) avail-time)))

(latest-handler! #:selector   sicom-symbol?  
		 #:handler    sicom-latest-get
		 #:adate-time sicom-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download - end of day html
;;
;; This uses the following historical prices page (normally shown in a frame)
;;
;;     http://www.sicom.com.sg/dailyprice.asp
;;
;; which results in URLs like
;;
;;     http://www.sicom.com.sg/downloadHistoricalPrices.asp?downloaddate=9-20-2006
;;
;; The headers that come back have "Content-Type: application/octet-stream"
;; and propose a filename like "HistoricalPrices817200674515AM.xls", but the
;; body is actually html.
;;
;; Each day is about 19 kbytes, unfortunately.  A body with "Data Not
;; Available" is given if there's no data for the requested date (a public
;; holiday for instance).
;;

;; return the tdate of the available sicom daily data
;; not sure what time new data is available, try after 14:30 each day
(define (sicom-available-tdate)
  (tdate-today-after 14 30 (timezone-singapore)))

;; the contract specifications pages don't give symbols for RSS1 and RSS3
;; index, but RS and RI were advised by email
(define (daily-name->commodity name)
  (assoc-ref '(("RSS1"  . "RS")
	       ("RSS3"  . "RT")
	       ("20FOB" . "TF")
	       ("RSS3I" . "RI")
	       ("COFFE" . "CF")) name))

;; MONTH is a string like "J/M 06" or "MAY 05"
;; mung the quarters to "JFM" etc
;; mung a "MAY05" to "MAY 05", once seen
;;
(define (daily-mung-month-str month)

  ;; quarter like "J/M 06" changed to "JFM 06", our preferred form
  (if (string-prefix? "J/M" month) ;; jan/feb/mar
      (string-set! month 1 #\F))
  (if (string-prefix? "A/J" month) ;; apr/may/jun
      (string-set! month 1 #\M))
  (if (string-prefix? "J/S" month) ;; jul/aug/sep
      (string-set! month 1 #\A))
  (if (string-prefix? "O/D" month) ;; oct/nov/dec
      (string-set! month 1 #\N))

  ;; report of 4-12-2005 has RSS1 with "MAY05", mung that to "MAY 05"
  (if (not (eqv? #\space (string-ref month (- (string-length month) 3))))
      (set! month (string-append (string-drop-right month 2)
				 " "
				 (string-take-right month 2))))
  month)

(define (sicom-process-download symbol-list body)
  (define row-list (html-table-rows body
				    ;; second <table>
				    (string-contains-after-ci body "<table")))

  ;; "Contract<br>Name" is collapsed to "ContractName" by html-table-rows,
  ;; maybe it should be <br> -> \n but at any rate
  (let ((headings (first row-list)))
    (or (equal? headings '("ContractName"
			   "ContractMonth"
			   "PreviousSettlement"
			   "High"
			   "Low"
			   "Close"
			   "Volume"
			   "OpenInterest"))
	(error "SICOM: unrecognised historical page headings:" headings)))
  (set! row-list (cdr row-list))

  (let* (;; eg. "<title>HISTORICAL PRICES ON 15 Aug 2006</title>"
	 (m        (must-match
		    (string-match "<title>HISTORICAL PRICES ON ([0-9]+ [A-Za-z]+ [0-9][0-9][0-9][0-9])</title>" body)))
	 (tdate    (d/m/y-str->tdate (match:substring m 1))))

    (download-process
     #:module      (_ "SICOM")
     #:symbol-list symbol-list
     #:lo          tdate
     #:hi          tdate
     #:row-list
     (map-rows! (lambda (name month prev high low close volume openint)
		  (set! month (daily-mung-month-str month))
		  (let* ((commodity  (daily-name->commodity name))
			 (mdate      (MMM_yy-str->mdate month))
			 ;; eg. "TF AMJ 07" for a quarter
			 (comm+quart (if mdate
					 commodity
					 (string-append commodity " " month))))
		    (list #:commodity  comm+quart
			  #:name       (sicom-commodity->name commodity)
			  #:currency   (sicom-commodity->currency commodity)
			  #:mdate         mdate
			  #:tdate         tdate
			  #:high          high
			  #:low           low
			  #:close         close
			  #:volume        volume
			  #:open-interest openint)))
		row-list))))

;; download from tdates START to END for SYMBOL-LIST
;; for an update START is before END, it's the other way around to work
;; backwards for a backto download
(define (sicom-download-range symbol-list start end)
  (do ((t start (if (<= start end) (1+ t) (1- t))))
      ((not (within-range? start end t)))
    (download-status (_ "SICOM") (_ "data") (tdate-range-string t t))
    (receive (headers body)
	(http-request
	 (receive-list (year month day)
	     (tdate->ymd t)
	   (format #f "http://www.sicom.com.sg/downloadHistoricalPrices.asp?downloaddate=~a-~a-~a" month day year))
	 #:want-ok #t)
      ;; message in the body if no data for this date
      (or (string-contains body "Data Not Available")
	  (sicom-process-download symbol-list body)))))

(define (sicom-download symbol-list)
  (define avail-tdate (sicom-available-tdate))

  ;; limit to last 30 days, since it's about 19k per day and a separate
  ;; download for each, which quickly becomes a lot of separate requests and
  ;; a largish total size
  (define start-tdate (max (- (tdate-today) 30)
			   (apply min (map download-start-tdate symbol-list))))

  ;; others covered by this data
  (set! symbol-list (download-also symbol-list
				   #:selector    sicom-symbol?
				   #:start-tdate start-tdate))

  (sicom-download-range symbol-list start-tdate avail-tdate))

(define (sicom-backto want-tdate symbol-list)
  (sicom-download-range symbol-list
			(apply max (map download-back-tdate
					symbol-list))
			want-tdate))

(download-handler! #:selector   sicom-symbol?
		   #:update     sicom-download
		   #:avail-proc sicom-available-tdate
		   #:backto     sicom-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any sicom-symbol? symbol-list)
			      (download-now-all-commodities-and-months))))
