;; Sine weighted moving average.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart sine-ma)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))



;; http://www.linnsoft.com/tour/techind/movAvg.htm
;;     Formula, and sample SineMA[20] of Nasdaq 100 (symbol QQQ, yahoo now
;;     ^IXIC) from 2001.
;;



(define pi 3.1415926535897932384626433832795029)


;; Return a procedure which calculates a sine weighted moving average over
;; an accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the Sine-MA up to (and including) that value, or #f if not enough
;; values yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from start to end).  COUNT-1 is
;; enough to ensure the first PROC call is based on a full window.
;;
(define*-public (sine-ma-calc-proc count #:optional (data '()))

  ;; DATA is changed to a queue of data points
  ;; POINTS is the length of DATA
  ;; FACTORS is the weights, sin(pi*1/(N+1)), ... sin(pi*N/(N+1))
  ;; WEIGHT is the sum of the first POINTS many from FACTORS

  (set! data (list->q (take-right-maybe data count)))
  (let* ((points     (q-length data))
	 (factors    (map! (lambda (i)
			     (sin (* pi (/ i (1+ count)))))
			   (iota count 1)))
	 (weight     (apply + (take factors points)))
	 (min-weight (* indicator-minimum-fraction
			(apply + factors))))

    (lambda (x)

      (if (>= points count)
	  ;; drop oldest point
	  (deq! data)
	  ;; otherwise gaining a point
	  (begin
	    (set! weight (+ weight (list-ref factors points)))
	    (set! points (1+ points))))

      ;; add this point
      (enq! data x)

      (and (>= weight min-weight)
	   (/ (apply + (map * (q-list data) factors))
	      weight)))))


(define (sine-ma-fill series new-array lo hi)
  (let* ((count   (series-data series))
	 (parent  (series-parent series))
	 (array   (series-array-closes parent lo hi))
	 (sine-ma (sine-ma-calc-proc
		   count (series-prev-n-closes parent
					       #:before lo
					       #:count  (1- count)))))
    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (sine-ma this) i)))))

(define-public (sine-ma-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "Sin-MA") count))
    (series-datatype! new-series sine-ma-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts sine-ma-fill (1- count)))
    (series-array!    new-series #f)

    new-series))

(define sine-ma-datatype
  (datatype-new!
   #:key      'sine-ma
   #:name     (_ "Sine Weighted MA")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Sine Weighted Moving Average")
   #:proc     sine-ma-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'sine-ma-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
