;; Simple moving average (SMA).

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart sma)
  #:use-module (ice-9 optargs)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a simple moving average over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the SMA up to (and including) that value, or #f if not enough values yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from start to end).  COUNT-1 is
;; enough to ensure the first PROC call is based on a full window.
;;
(define*-public (sma-calc-proc count #:optional (data '()))

  ;; POINTS is the number of non-#f points in DATA
  ;; TOTAL is the sum of the point values in DATA

  (set! data (take-right-maybe data count))
  (let ((min-points (ceiling-exact (* indicator-minimum-fraction count)))
	(points     (length data))
	(total      (apply + data)))
    (set! data (indicator-circular count data))

    (lambda (x)

      (if (car data)
	  ;; drop oldest point
	  (set! total  (- total (car data)))
	  ;; otherwise gaining a point
	  (set! points (1+ points)))

      ;; replace oldest with new point
      (set-car! data x)
      (set! data (cdr data))

      ;; add this point
      (set! total  (+ total x))

      (and (>= points min-points)
	   (exact->inexact (/ total points))))))

(define (sma-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (sma    (sma-calc-proc count
				(series-prev-n-closes parent
						      #:before lo
						      #:count  (1- count)))))
    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (sma (array-ref array i))
		      i)))))

(define-public (sma-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "SMA") count))
    (series-datatype! new-series sma-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts sma-fill (1- count)))
    (series-array!    new-series #f)

    new-series))

(define sma-datatype
  (datatype-new!
   #:key      'sma
   #:name     (_ "SMA - Simple MA")
   #:priority 10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Simple Moving Average")
   #:proc     sma-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'sma-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
