;; Standard deviation (STDDEV).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart stddev)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a standard deviation over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the stddev up to (and including) that VAL, or #f if not enough points
;; yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (stddev-calc-proc count)

  ;; If the inputs X are all exact then TOTAL and TOTAL-SQUARES are exact
  ;; and the only rounding is at the final sqrt and division.  But when
  ;; they're inexact the totals can round a little and there's a danger of
  ;; the subtraction making a negative sqrt, hence the max 0 there.
  ;;
  (let ((data          (make-circular-list count #f))
	(total         0)
	(total-squares 0)
	(points        0)
	(min-points    (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (x)

      (if (car data)
	  ;; drop oldest point
	  (begin
	    (set! total         (- total (car data)))
	    (set! total-squares (- total-squares (square (car data)))))
	  ;; introducing new point
	  (set! points (1+ points)))

      ;; replace oldest with new x
      (set-car! data x)
      (set! data (cdr data))

      ;; add new x
      (set! total         (+ total x))
      (set! total-squares (+ total-squares (* x x)))

      (and (>= points min-points)
	   (/ (sqrt (max 0 (- (* total-squares points)
			      (square total))))
	      points)))))


;;-----------------------------------------------------------------------------
;; Standard deviation series

(define (stddev-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (calc   (stddev-calc-proc count)))

    ;; warmup
    (for-each calc (series-prev-n-closes parent
					 #:before lo #:count (1- count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (calc (array-ref array i)) i)))))

(define-public (stddev-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Stddev") count))
    (series-datatype!  new-series stddev-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts stddev-fill
							   (1- count)))
    (series-array!     new-series #f)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public stddev-datatype
  (datatype-new!
   #:key      'stddev
   #:name     (_ "Standard Deviation")
   #:priority -10
   #:units    'price
   #:minimum  0
   #:manual   (_ "CONTEXT(manual-node): Standard Deviation")
   #:proc     stddev-series
   #:parameter-list
   (list     (datatype-param-new
	      #:name    (_ "Days")
	      #:key     'stddev-days
	      #:type    'integer
	      #:default 20
	      #:minimum 1))))
