;; Stochastic RSI.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart stoch-rsi)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart rsi)
  #:use-module (chart williams-r))


;; http://www.stockcharts.com/education/IndicatorAnalysis/indic_stochRSI.html
;;     Numerical examples, graph of DELL checked against the code here.
;;


;; Return a procedure which calculates a stochastic RSI on successive
;; closing prices, with RSI smoothing of COUNT periods and high/low range of
;; the same COUNT periods.  For the smoothing COUNT is interpreted by
;; Wilder's reckoning, so alpha = 1/COUNT.
;;
;; Each call (PROC close) enters a new point into the window, and the return
;; is the stochastic RSI up to (and including) that value, or #f if not
;; enough points yet.
;;
;; To prime initially, call PROC with (stoch-rsi-warmup-count COUNT) many
;; values, then the next call will be based on full data.
;;
(define-public (stoch-rsi-calc-proc count)

  (define rsi-proc (rsi-calc-proc count))
  (define %R-proc  (williams-%R-calc-proc count))

  (lambda (close)
    (and-let* ((rsi (rsi-proc close))
	       (%R  (%R-proc #f #f rsi)))
      (1+ (* 0.01 %R)))))

(define-public (stoch-rsi-warmup-count count)
  (+ (rsi-warmup-count count)
     (1- count)))


(define (stoch-rsi-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent    (series-parent series))
	   (start     (series-prev-n-closes-pos parent #:before lo
						#:count warmup))
	   (array     (series-array-closes parent start hi))
	   (stoch-rsi (stoch-rsi-calc-proc count)))

      ;; warmup
      (do ((i start (1+ i)))
	  ((>= i lo))
	(and-let* ((this (array-ref array i)))
	  (stoch-rsi this)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (stoch-rsi this) i))))))

(define-public (stoch-rsi-series series count)
  (let ((new-series (series-copy series))
	(warmup     (stoch-rsi-warmup-count count)))
    (series-name!      new-series (format #f "~a ~a" (_ "Stoch RSI") count))
    (series-datatype!  new-series stoch-rsi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list count warmup))
    (series-require!   new-series (indicator-require-parts stoch-rsi-fill
							   warmup))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public stoch-rsi-datatype
  (datatype-new!
   #:key     'stoch-rsi
   #:name    (_ "Stochastic RSI")
   #:units   'zero-to-one
   #:minimum 0
   #:maximum 1
   #:hlines  '(.20 .50 .80)
   #:manual  (_ "CONTEXT(manual-node): Relative Strength Index") ;; subsection
   #:proc    stoch-rsi-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'stoch-rsi-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14))))
