;; Stochastics %K and %D.

;; Copyright 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart stochastics)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart sma)
  #:use-module (chart williams-r))


;; Return a procedure which calculates a %K stochastic over a window of
;; COUNT points.
;;
;; Each call (PROC low high close) enters a new day's range and close into
;; the window.  The return is the %K up to (and including) that day.  LOW,
;; HIGH and CLOSE can be all #f for a holiday, the return is #f in that
;; case.  The return is also #f if there's not enough valid points in the
;; window.
;;
;; If data doesn't have LOW / HIGH daily ranges, just closes, then call
;; (PROC close close close).  This tends to make the indicator not look as
;; good (it sticks to 0 or 100 for new COUNT-day highs and lows), but that's
;; too bad if it's all that's available.

(define-public (stochastics-calc-proc K-count D-count slow-count)
  (let ((williams-proc (williams-%R-calc-proc K-count))
	(slow-proc     (if (<= slow-count 1)
			   noop
			   (sma-calc-proc slow-count)))
	(sma-proc      (sma-calc-proc D-count)))

    (lambda (high low close)
      (let ((K (williams-proc high low close))
	    (D #f))
	(if K
	    (begin
	      (set! K (slow-proc (+ K 100)))
	      (if K
		  (set! D (sma-proc K)))))
	(values K D)))))


(define (stochastics-fill series new-array lo hi)
  (receive-list (K-count D-count slow-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (warmup (stochastics-warmup-count K-count D-count slow-count))
	   (start  (series-prev-n-closes-pos parent
					     #:before lo #:count warmup))
	   (array  (series-array parent start hi))
	   (proc   (stochastics-calc-proc K-count D-count slow-count)))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i lo))
	(if (array-ref array i 3)
	    (proc (array-ref array i 1)    ;; high
		  (array-ref array i 2)    ;; low
		  (array-ref array i 3)))) ;; close

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i 3)
	    (receive (K D)
		(proc (array-ref array i 1)  ;; high
		      (array-ref array i 2)  ;; low
		      (array-ref array i 3)) ;; close
	      (array-set! new-array K i 0)
	      (array-set! new-array D i 1)))))))

(define (stochastics-warmup-count K-count D-count slow-count)
  (+ K-count
     (1- (max 1 slow-count))
     (1- D-count)))


(define-public (stochastics-series series K-count D-count slow-count)
  (ohlcv-required "Stochastics" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (if (zero? slow-count)
				      (format #f "%K ~a, %D ~a, no slow"
					      K-count D-count)
				      (format #f "%K ~a, %D ~a, slow ~a"
					      K-count D-count slow-count)))
    (series-datatype!  new-series stochastics-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list K-count D-count slow-count))
    (series-require!   new-series (indicator-require-parts
				   stochastics-fill
				   (stochastics-warmup-count K-count D-count
							     slow-count)
				   '(0 1)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public stochastics-datatype
  (datatype-new!
   #:key       'stochastics
   #:name      (_ "Stochastics %K/%D")
   #:input     datatype-ohlcv
   #:units     'percentage
   #:linestyle 'stochastics
   #:minimum   0
   #:maximum   100
   #:hlines    '(20 80)
   #:manual  (_ "CONTEXT(manual-node): Stochastics")
   #:proc      stochastics-series
   #:closes-apply-arg '(0)
   #:parameter-list
   (list (datatype-param-new
	  #:key     'stochastics-K
	  #:name    (_ "%K days")
	  #:type    'integer
	  #:minimum 1
	  #:default 14)
	 (datatype-param-new
	  #:key     'stochastics-D
	  #:name    (_ "%D days")
	  #:type    'integer
	  #:minimum 1
	  #:default 3)
	 (datatype-param-new
	  #:key     'stochastics-slow
	  #:name    (_ "Slow days")
	  #:type    'integer
	  #:minimum 0
	  #:default 0))))
