;; Symbol lists.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart symlists)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:autoload   (chart database)     (database-symbols
				     alert-interest-read
				     historical-read)
  #:autoload   (chart latest)       (latest-get)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; misc

;; return a sublist of LST, from START (inclusive) to END (exclusive)
;; allow those positions to be outside the length of LST
(define (sublist-maybe lst start end)
  (take-maybe (drop-maybe lst start) (- end start)))


;;-----------------------------------------------------------------------------
;; marker

(if (not (defined? 'hash-for-each))  ;; new in guile 1.8
    (define (hash-for-each proc table)
      (hash-fold (lambda (key value prev)
		   (proc key value))
		 #f table)))

(let-toplevel ((rec-type (make-record-type "smarker" '(list pos entry))))
  (define rec-new      (record-constructor rec-type '()))
  (define rec-symlist! (rec-modifier rec-type 'list))
  (define rec-pos!     (rec-modifier rec-type 'pos))
  (define rec-entry!   (rec-modifier rec-type 'entry))

  ;; track all markers, but allow them to gc
  (define all-smarkers (make-weak-key-hash-table 17))

  ;; call (PROC smarker) for all smarkers looking at symlist KEY
  (define (for-each-smarker key proc)
    (hash-for-each (lambda (smarker dummy)
		     (and-let* ((symlist (smarker-symlist smarker))
				(        (eq? key (symlist-key symlist))))
		       (proc smarker)))
		   all-smarkers))

  (define (return-pos smarker pos)
    (rec-pos! smarker pos)
    (let* ((lst   (symlist-list (smarker-symlist smarker)))
	   (entry (and (length>=? lst (1+ pos))
		       (list-ref lst pos))))
      (rec-entry! smarker entry)
      entry))

  (define-public smarker-symlist (rec-accessor rec-type 'list))

  ;; symbol, or #f if none
  (define-public smarker-entry   (rec-accessor rec-type 'entry))

  ;; position in symlist
  (define-public smarker-pos     (rec-accessor rec-type 'pos))

  (define-public (smarker-new)
    (define smarker (rec-new))
    (hashq-set! all-smarkers smarker #f)
    smarker)

  (define-public smarker-copy record-copy)

  (define-public (smarker-equal? x y)
    (and (eq?    (smarker-symlist x) (smarker-symlist y))
	 (equal? (smarker-entry x)   (smarker-entry y))
	 (eqv?   (smarker-pos x)     (smarker-pos y))))

  ;; goto given ELEM (a symbol string) in SYMLIST
  ;; the return is always ELEM, the marker is not moved if it isn't in SYMLIST
  (define-public (smarker-goto smarker symlist elem)
    (and-let* ((pos (list-index-elem elem (symlist-list symlist))))
      (rec-symlist! smarker symlist)
      (rec-pos!     smarker pos)
      (rec-entry!   smarker elem))
    elem)

  ;; return the symbol, or #f if SYMLIST is empty
  (define-public (smarker-goto-first-of-symlist smarker symlist)
    (rec-symlist! smarker symlist)
    (return-pos smarker 0))

  (define-public (smarker-goto-last-of-symlist smarker symlist)
    (rec-symlist! smarker symlist)
    (return-pos smarker (1- (length (symlist-list symlist)))))

  ;; goto the first entry in the first non-empty symlist
  ;; return the symbol, or #f if all symlists are empty
  (define-public (smarker-first smarker)
    (let* ((symlist (find symlist-non-null? symlist-all-lists)))
      (rec-symlist! smarker symlist)
      (and symlist
	   (return-pos smarker 0))))

  ;; goto the last entry in the last non-empty symlist
  (define-public (smarker-last smarker)
    (let* ((symlist (find symlist-non-null? (reverse symlist-all-lists))))
      (rec-symlist! smarker symlist)
      (and symlist
	   (return-pos smarker (1- (length (symlist-list symlist)))))))

  ;; goto the first element of the next non-empty list, if any
  (define-public (smarker-next-list smarker)
    (if (not (smarker-symlist smarker))
	(smarker-first smarker)  ;; nothing current, start at the start

	(let* ((symlist (find symlist-non-null?
			      (tail-after (smarker-symlist smarker)
					  symlist-all-lists))))
	  (if (not symlist)
	      (smarker-last smarker)

	      (begin
		(rec-symlist! smarker symlist)
		(return-pos smarker 0))))))

  ;; first element of prev non-empty list, if any
  (define-public (smarker-prev-list smarker)
    (and (smarker-symlist smarker)
	 (let ((symlist (find symlist-non-null?
			      (tail-after (smarker-symlist smarker)
					  (reverse symlist-all-lists)))))
	   (if (not symlist)
	       (smarker-first smarker)

	       (begin
		 (rec-symlist! smarker symlist)
		 (return-pos smarker 0))))))

  ;; goto the next element of current list, or first of next non-empty
  (define-public (smarker-next smarker)
    (if (not (smarker-symlist smarker))
	(smarker-first smarker) ;; nothing ever visited, start at the start

	(let ((lst (symlist-list (smarker-symlist smarker)))
	      (pos (smarker-pos smarker)))
	  (if (smarker-entry smarker)  ;; if looking at pos then skip it
	      (set! pos (1+ pos)))
	  (if (>= pos (length lst)) ;; if past end then next list
	      (smarker-next-list smarker)
	      (return-pos smarker pos)))))

  ;; return true if there's a next element after SMARKER (which
  ;; `smarker-next' could go to)
  (define-public (smarker-next? smarker)
    ;; at startup avoid looking into the alerts list, as that can be slow
    (if (not (smarker-symlist smarker))
	(not (null? (database-symbols)))

	(let ((s2 (smarker-copy smarker)))
	  (and (smarker-next s2)
	       (not (smarker-equal? smarker s2))))))

  ;; goto the next element of the current list, wrapping around at end
  (define-public (smarker-next-circular smarker)
    (and-let* ((symlist (smarker-symlist smarker)))
      (let ((lst (symlist-list symlist))
	    (pos (smarker-pos smarker)))
	(if (smarker-entry smarker)  ;; if looking at pos then skip it
	    (set! pos (1+ pos)))
	(if (>= pos (length lst)) ;; if past end then back to start
	    (smarker-goto-first-of-symlist smarker (smarker-symlist smarker))
	    (return-pos smarker pos)))))

  ;; goto the prev element of the current list, wrapping around at end
  (define-public (smarker-prev-circular smarker)
    (and-let* ((symlist (smarker-symlist smarker)))
      (let ((pos (smarker-pos smarker)))
	(if (positive? pos)
	    (return-pos smarker (min (1- (length (symlist-list symlist)))
				     (1- pos)))
	    (smarker-goto-last-of-symlist smarker
					  (smarker-symlist smarker))))))

  ;; prev element of current list, or last of prev non-empty
  (define-public (smarker-prev smarker)
    (and-let* ((symlist (smarker-symlist smarker)))
      (let ((pos (smarker-pos smarker)))
	(if (positive? pos)
	    (return-pos smarker (min (1- (length (symlist-list symlist)))
				     (1- pos)))

	    (begin ;; start of this list, go to end of previous list
	      (set! symlist (find symlist-non-null?
				  (tail-after symlist
					      (reverse symlist-all-lists))))
	      (if (not symlist) ;; no previous non-empty, got to beginning
		  (smarker-first smarker)

		  (begin
		    (rec-symlist! smarker symlist)
		    (return-pos smarker
				(1- (length (symlist-list symlist)))))))))))

  ;; return true if there's a previous element before SMARKER (which
  ;; `smarker-prev' could go to)
  (define-public (smarker-prev? smarker)
    ;; at startup avoid looking into the alerts list, as that can be slow
    (and (smarker-symlist smarker)
	 (let ((s2 (smarker-copy smarker)))
	   (and (smarker-prev s2)
		(not (smarker-equal? smarker s2))))))

  ;; return list (SYMLIST ELEM), or #f if no match
  (define-public (smarker-find-name smarker str)
    (define sls symlist-all-lists)

    (define (attempt matcher munger)
      (any (lambda (symlist)
	     (any (lambda (elem)
		    (and (matcher str (munger elem))
			 (list symlist elem)))
		  (symlist-list symlist)))
	   sls))

    ;; current list of SMARKER at front of list of symlists
    (and-let* ((curr (smarker-symlist smarker)))
      (set! sls (cons curr (delq curr sls))))

    ;; Make progressively looser matching attempts across the lists.
    ;; Note that an exact match anywhere is preferred over a partial match
    ;; in the current list, otherwise you could type an exact full symbol
    ;; like "FOO" and still be left on a "FOO.AX" in the current list.
    ;;
    (or (attempt string=?    noop)
	(attempt string-ci=? noop)
	(attempt string=?    	chart-symbol-sans-dot)
	(attempt string-ci=? 	chart-symbol-sans-dot)
	(attempt string-ci=? 	chart-symbol-sans-hat)
	(attempt string-prefix-ci? chart-symbol-sans-hat)))


  (notify-connect 'symlist-delete
    (lambda (arg)
      (receive-list (key pos count)
	  arg
	(for-each-smarker key
	  (lambda (smarker)
	    (if (member (smarker-entry smarker)
			(sublist-maybe (symlist-list (smarker-symlist smarker))
				       pos (+ pos count)))
		(rec-entry! smarker #f))  ;; delete of current entry
	    (if (> (smarker-pos smarker) pos)
		(rec-pos! smarker
			  (max pos (- (smarker-pos smarker) count)))))))))

  (notify-connect 'symlist-insert
    (lambda (arg)
      (receive-list (key pos addlst)
	  arg
	(for-each-smarker key
	  (lambda (smarker)
	    (let ((current-pos (smarker-pos smarker)))
	      (if (or (> current-pos pos)      ;; much before
		      (and (= current-pos pos) ;; just before current
			   (smarker-entry smarker)))
		  (rec-pos! smarker (+ current-pos (length addlst))))))))))


  (notify-connect 'symlist-move
    (lambda (arg)
      (receive-list (key from count to)
	  arg
	(for-each-smarker key
	  (lambda (smarker)
	    (let ((current-pos (smarker-pos smarker)))
	      (cond
	       ((if (smarker-entry smarker)
		    (within-range? from (+ from count -1) current-pos)
		    (within-range? (1+ from) (+ from count -2) current-pos))
		;; current within range being moved, follow to "to"
		(set! current-pos (+ to (- current-pos from))))

	       ((and (>= from current-pos)
		     (if (smarker-entry smarker)
			 (<= to current-pos)
			 (< to current-pos)))
		;; moving some after current to before, increase accordingly
		(rec-pos! smarker (+ current-pos count)))))))))))

;; return symbol, or #f if not found
(define-public (smarker-goto-name smarker str)
  (and-let* ((found (smarker-find-name smarker str)))
    (apply smarker-goto smarker found)))



;;-----------------------------------------------------------------------------
;; symlists

(define-public symlist-all-lists '())

(let-toplevel ((rec-type (make-record-type "symlist"
					   '(key name list refresh editable
						 interesting))))
  (define rec-new          (record-constructor rec-type))
  (define rec-list         (rec-accessor rec-type 'list))
  (define rec-refresh      (rec-accessor rec-type 'refresh))
  (define rec-interesting  (rec-accessor rec-type 'interesting))
  (define rec-interesting! (rec-modifier rec-type 'interesting))

  (define-public symlist-key       (rec-accessor rec-type 'key))
  (define-public symlist-name      (rec-accessor rec-type 'name))
  (define-public symlist-editable  (rec-accessor rec-type 'editable))
  ;; return list of symbols of interest to this symlist
  (define-public (symlist-interesting symlist)
    ((rec-interesting symlist)))
  (define-public symlist-rec-list! (rec-modifier rec-type 'list))


  (define*-public (symlist-new #:key name key (refresh noop) editable
			       interesting)
    (let ((symlist (rec-new key name '() refresh editable interesting)))
      (if (not interesting)
	  (rec-interesting! symlist (lambda ()
				      (symlist-list symlist))))
      (set! symlist-all-lists (append! symlist-all-lists (list symlist)))
      symlist))

  (define-public (symlist-list symlist)
    ((rec-refresh symlist))
    (rec-list symlist))

  (define-public (symlist-find key)
    (find (lambda (symlist)
	    (eq? key (symlist-key symlist)))
	  symlist-all-lists))


  (define-public (symlist-list! symlist lst)
    (if (not (equal? (rec-list symlist) lst))
	(begin
	  (symlist-rec-list! symlist lst)
	  (if (eq? symlist favourites-symlist)
	      (preference-set 'favourites lst))
	  (notify-send 'symlist-update (list (symlist-key symlist))))))


  ;; delete COUNT elements from POS (starting from 0 for first element)
  (define-public (symlist-delete symlist pos count)
    (notify-send 'symlist-delete (list (symlist-key symlist) pos count))
    (let ((lst (rec-list symlist)))
      (symlist-list! symlist
		     (append (take lst pos) (drop lst (+ pos count))))))

  ;; insert ADDLST symbols at POS (starting from 0 for first element)
  (define-public (symlist-insert symlist pos addlst)
    (notify-send 'symlist-delete (list (symlist-key symlist)
				       pos (length addlst)))
    (let ((lst (rec-list symlist)))
      (symlist-list! symlist
		     (append (take lst pos) addlst (drop lst pos)))))

  ;; move COUNT symbols at FROM to new position TO (both 0 for first element)
  ;;
  (define-public (symlist-move symlist from count to)
    (notify-send 'symlist-move (list (symlist-key symlist) from count to))
    (if (> to from)
	(set! to (- to count -1)))
    (let* ((lst     (rec-list symlist))
	   (chopped (append (take lst from) (drop lst (+ from count))))
	   (part    (sublist lst from (+ from count))))
      (symlist-list! symlist
		     (append (take chopped to)
			     part
			     (drop chopped to))))))

(define-public (symlist-null? symlist)
  (null? (symlist-list symlist)))

(define-public (symlist-non-null? symlist)
  (not (symlist-null? symlist)))

;; ENHANCE-ME: delete in blocks
(define-public (symlist-delete-symbols symlist symbol-list)
  (for-each (lambda (symbol)
	      (and-let* ((pos (list-index-elem symbol (symlist-list symlist))))
		(symlist-delete symlist pos 1)))
	    symbol-list))
		

;;-----------------------------------------------------------------------------
;; alert list

(define-public (alert-key? key)
  (memq key '(alert-above alert-below)))

(let-toplevel ((dirty-list #f))  ;; list (no duplicates) or #f if uninit

  (define (alert? symbol)
    (and-let* ((       (not (member symbol (symlist-list historical-symlist))))
	       ;; expect only a few alerts, so try to exclude the rest by
	       ;; reading just the notes file
	       (       (database-read-note symbol alert-key?))
	       (series (database-read-series symbol))
	       (latest (latest-get symbol)))
      (let* ((last    (latest-last latest))
	     (bid     (latest-bid latest))
	     (offer   (latest-offer latest))
	     (factor  (integer-expt-pn 10 (- (series-decimals series)
					     (latest-decimals latest)))))

	;; For testing "above" the last is tested, or the lesser of bid and
	;; offer if they're above the last.  The quote may have moved up
	;; without a trade yet, or in a crossed market the low offer is
	;; a minimum for the next trade.  Vice versa for "below" level.
	;;
	(or (and-let* ((bid   (max-maybe last (min-maybe bid offer)))
		       (above (assq 'alert-above (series-notes series))))
	      (>= (* bid factor) (second above)))

	    (and-let* ((offer (min-maybe last (max-maybe bid offer)))
		       (below (assq 'alert-below (series-notes series))))
	      (<= (* offer factor) (second below)))))))

  (define (alert-update-some symbol-list)
    (if dirty-list
	(set! dirty-list
	      (lset-union string=? dirty-list
			  (delete-duplicates symbol-list)))))

  (define (alert-refresh)
    (let ((dirty dirty-list))
      (set! dirty-list '())

      (cond ((not dirty)
	     ;; initial fill
	     (symlist-insert alert-symlist 0
			     (filter alert? (symlist-list all-symlist))))

	    ((not (null? dirty))
	     (receive (ins del)
		 (partition alert? dirty)
	       (symlist-delete-symbols alert-symlist del)
	       (symlist-insert-symbols-sorted alert-symlist ins))))))

  (define-public alert-symlist
    (symlist-new #:key         'alert
		 #:name        (_ "Alert")
		 #:refresh     alert-refresh
		 #:interesting (lambda () (alert-interest-read))))

  (notify-connect 'latest-update alert-update-some)
  (notify-connect 'database-update alert-update-some)

  (notify-connect 'alert-cache-update
    (lambda (dummy)
      (set! dirty-list #f)
      (notify-send-locally 'symlist-update '(alert)))))


(define-public (symlist-insert-symbols-sorted symlist symbol-list)
  (for-each (lambda (symbol)
	      (let ((lst (symlist-list symlist)))
		(if (not (member symbol lst))
		    (symlist-insert symlist
				    (merge-one-index lst symbol chart-symbol<?)
				    (list symbol)))))
	    symbol-list))

;; return the index in the sorted list LST at which ELEM should be inserted
;; LESS is the comparison procedure
(define (merge-one-index lst elem less)
  (or (list-index (lambda (lst-elem)
		    (less elem lst-elem))
		  lst)
      (length lst)))


;;-----------------------------------------------------------------------------
;; favourites list

(let-toplevel ((dirty #t))

  (define (favourites-refresh)
    (if dirty
	(begin
	  (set! dirty #f)
	  (symlist-rec-list! favourites-symlist
			     (preference-get 'favourites '())))))

  (define-public favourites-symlist (symlist-new #:key      'favourites
						 #:name     (_ "Favourites")
						 #:refresh  favourites-refresh
						 #:editable #t))

  (notify-connect 'preferences-save
    (lambda (symbol-list)
      (set! dirty #t)
      (notify-send-locally 'symlist-update '(favourites)))))


;;-----------------------------------------------------------------------------
;; all symlist

(let-toplevel ((dirty #t))

  (define (all-refresh)
    (if dirty
	(begin
	  (set! dirty #f)
	  (symlist-rec-list! all-symlist
			     (lset-difference
			      string=? (database-symbols)
			      (symlist-list historical-symlist))))))

  (define-public all-symlist
    (symlist-new #:key     'all
		 #:name    (_ "All")
		 #:refresh all-refresh))

  (notify-connect 'database-new
    (lambda (symbol-list)
      (set! dirty #t)
      (notify-send-locally 'symlist-update '(all))))

  (notify-connect 'database-delete
    (lambda (symbol-list)
      (set! dirty #t)
      (notify-send-locally 'symlist-update '(all))))

  (notify-connect 'symlist-update
    (lambda (symlist-key-list)
      (if (memq 'historical symlist-key-list)
	  (begin
	    (set! dirty #t)
	    (notify-send-locally 'symlist-update '(all)))))))


;;-----------------------------------------------------------------------------
;; historical symlist

(let-toplevel ((dirty #t))

  (define (historical-refresh)
    (if dirty
	(begin
	  (set! dirty #f)
	  (symlist-rec-list! historical-symlist
			     (sort (historical-read) chart-symbol<?)))))

  (define-public historical-symlist
    (symlist-new #:key     'historical
		 #:name    (_ "Historical")
		 #:refresh historical-refresh))

  (notify-connect 'historical-cache-update
    (lambda (dummy)
      (set! dirty #t)
      (notify-send-locally 'symlist-update '(historical)))))
