;; T3 moving average.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart t3)
  #:use-module (srfi srfi-1)
  #:use-module (chart ema)
  #:use-module (chart ema-3)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/0298/TradersTips/Tips9802.html
;;     Traders Tips Feb 1998, implementations of T3 and GD.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/0898/Letters9808/Letters9808.html
;;     TASC letters Oct 1998, George Herbert on GD as weighted average of
;;     EMA and DEMA.
;;
;; http://trader.online.pl/ELZ/t-i-T3.html
;;     Formula in terms of GD.  Sample S&P 500 (symbol ^GSPC) from 2001,
;;     with N=6 and v=0.7.
;;
;; http://www.linnsoft.com/tour/techind/t3.htm
;;



(define (cube x)
  (integer-expt x 3))


;; Return a procedure which calculates a T3 average over an accumulated
;; window of COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the T3 up to (and including) that value, or #f if not enough values
;; yet.
;;
;; Because T3 uses an exponential moving average, it's in theory influenced
;; by all preceding data, but `t3-warmup-count' below is designed to
;; determine a warmup count.
;;
(define-public (t3-calc-proc count vf)

  (let ((f     (ema-count->f count))
	(alpha (ema-count->alpha count))

	(a-sum 0.0)
	(a-div 0.0)
	(b-sum 0.0)
	(b-div 0.0)
	(c-sum 0.0)
	(c-div 0.0)
	(d-sum 0.0)
	(d-div 0.0)
	(e-sum 0.0)
	(e-div 0.0)
	(f-sum 0.0)
	(f-div 0.0)

	(c1    (cube (- vf)))
	(c2    (+ (* 3 (square vf))
		  (* 3 (cube vf))))
	(c3    (+ (* -6 (square vf))
		  (* -3 vf)
		  (* -3 (cube vf))))
	(c4    (+ 1
		  (* 3 vf)
		  (cube vf)
		  (* 3 (square vf)))))

    (lambda (x)
      ;; reduce past data
      (set! a-sum (* a-sum f))
      (set! a-div (* a-div f))
      ;; add new data
      (set! a-sum (+ a-sum (* x alpha)))
      (set! a-div (+ a-div alpha))

      (set! b-sum (* b-sum f))
      (set! b-div (* b-div f))
      (set! b-sum (+ b-sum (* a-sum alpha)))
      (set! b-div (+ b-div (* a-div alpha)))

      (set! c-sum (* c-sum f))
      (set! c-div (* c-div f))
      (set! c-sum (+ c-sum (* b-sum alpha)))
      (set! c-div (+ c-div (* b-div alpha)))

      (set! d-sum (* d-sum f))
      (set! d-div (* d-div f))
      (set! d-sum (+ d-sum (* c-sum alpha)))
      (set! d-div (+ d-div (* c-div alpha)))

      (set! e-sum (* e-sum f))
      (set! e-div (* e-div f))
      (set! e-sum (+ e-sum (* d-sum alpha)))
      (set! e-div (+ e-div (* d-div alpha)))

      (set! f-sum (* f-sum f))
      (set! f-div (* f-div f))
      (set! f-sum (+ f-sum (* e-sum alpha)))
      (set! f-div (+ f-div (* e-div alpha)))

      (and (>= c-div indicator-minimum-fraction)
	   (+ (* c4 (/ c-sum c-div))
	      (* c3 (/ d-sum d-div))
	      (* c2 (/ e-sum e-div))
	      (* c1 (/ f-sum f-div)))))))


(define-public (t3-warmup-count count vf)
  (ema-3-warmup-count count))



(define (t3-fill series new-array lo hi)
  (receive-list (count vf warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (t3-calc-proc count vf)))

      ;; warmup
      (for-each proc (series-prev-n-closes parent
					   #:before lo
					   #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i)) i))))))


(define-public (t3-series series count vf)
  (let ((new-series (series-copy series))
	(warmup     (t3-warmup-count count vf)))
    (series-name!      new-series (format #f (_ "T3 ~a, ~a" )
					  count
					  (locale-number-min-decimals
					   vf 1)))
    (series-datatype!  new-series t3-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list count vf warmup))
    (series-require!   new-series (indicator-require-parts t3-fill warmup))
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public t3-datatype
  (datatype-new!
   #:key    't3
   #:name   (_ "T3 Moving Average")
   #:units  #f
   #:manual (_ "CONTEXT(manual-node): T3 Moving Average")
   #:proc   t3-series
   #:parameter-list
   (list (datatype-param-new
	  #:key      't3-days
	  #:name     (_ "Days")
	  #:type     'integer
	  #:minimum  1
	  #:default  7)
	 (datatype-param-new
	  #:name     (_ "Factor")
	  #:key      't3-factor
	  #:type     'float
	  #:decimals 1
	  #:default  0.7
	  #:step     0.1
	  #:minimum  0
	  #:maximum  1))))
