;; Tom DeMark range expansion index.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tdrei)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://ta-glossary.netfirms.com/taglossary-R.htm
;;     Description quoting Tom DeMark on the calculation, sample BDY from
;;     2004.
;;
;; http://www.linnsoft.com/tour/techind/tdrei.htm
;;     Formula, shows both num and den terms both zeroed by condition.
;;     Sample on AOL call options.
;;
;; http://www.prophet.net/analyze/popglossary.jsp?studyid=TD_REI
;;     Description, sample of Sunoco Inc (NYSE symbol SUN), year not shown
;;     but is Jan/Feb 2003.  Note 6-day sums for the sample, not 5.
;;
;; http://trader.online.pl/MSZ/e-w-DeMark_Range_Expansion_Index.html
;;     Formula (parens in denominator look misplaced though).  TD3 and TD4
;;     work as an "or" of those two conditions.
;; http://trader.online.pl/MSZ/e-w-DeMark_Range_Expansion_Index_II.html
;;     Formula, similar but using momentum func, sample on WIG20.
;;
;; http://secure.aspenres.com/Documents/help/userguide/help/dmrkhelp/demarkRange_Expansion_Index.html
;;     Description, sample of CME S&P 500 futures SPU05.CME.
;;
;; http://www.traders.com/documentation/feedbk_docs/archive/0897/Abstracts_new/DeMark/DEMARK9708.html
;; http://store.traders.com/-v15-c08-thetdra-pdf.html
;;     Intro to Tom DeMark "The TD Range Expansion Index (TD REI)" TASC 15:8
;;     August 1997 (rest for sale), revising original from his book "The New
;;     Science of Technical Analysis".
;;
;; http://www.meta-formula.com/Metastock-Formulas-T.html
;;     Formula.
;;
;; http://www.amibroker.com/library/formula.php?id=15
;;     Formula, showing only numerator suppressed by condition.
;;
;; http://www.aiqsystems.com/PriceandVolumeBasedStrategies1.htm#down17
;;     Code.
;; http://www.aiqsystems.com/TDREI.gif
;;     Sample of Cabot Corp (NYSE symbol CBT) from Jan to Jun 2002.  Period
;;     not shown, looks like 5 days.
;;


(define (tdrei-calc-proc count)

  ;; HIGH-DATA and LOW-DATA are circular lists holding the 7 most recent
  ;; high/low prices.  After the current day is shifted in the values held
  ;; are (H6 H5 H4 H3 H2 H1 Htoday)
  ;;
  ;; CLOSE-DATA is a circular lists holding the 9 most recent close prices.
  ;; After the current day is shifted in the values held are (C8 C7 C6 C5 C4
  ;; C3 C2 C1 Ctoday)
  ;;
  (define high-data  (make-circular-list 7 #f))
  (define low-data   (make-circular-list 7 #f))
  (define close-data (make-circular-list 9 #f))
  (define sumval-proc (sum-calc-proc count))
  (define sumabs-proc (sum-calc-proc count))

  (lambda (high low close)

    ;; cope with no high/low in data
    (or high (set! high close))
    (or low  (set! low  close))

    ;; replace oldest with new point
    (set-car! high-data  high)
    (set-car! low-data   low)
    (set-car! close-data close)
    (set! high-data  (cdr high-data))
    (set! low-data   (cdr low-data))
    (set! close-data (cdr close-data))

    (and (first close-data) ;; once 9 closes have been seen
	 (let* ((h2     (fifth high-data))
		(l2     (fifth low-data))
		(use    (and  (or (>= h2 (second close-data))  ;; C7
				  (>= h2 (first close-data))   ;; C8
				  (>= high (second low-data))  ;; L5
				  (>= high (first  low-data))) ;; L6
			      (or (<= l2 (second close-data))    ;; C7
				  (<= l2 (first close-data))     ;; C8
				  (<= low (second high-data))    ;; H5
				  (<= low (first  high-data))))) ;; H6
		(val    (if use
			    (+ (- high h2) (- low l2))
			    0))
		(absval (+ (abs (- high h2)) (abs (- low l2))))
		(sumval (sumval-proc val))
		(sumabs (sumabs-proc absval)))

	   (and sumval
		(not (zero? sumabs))
		(/ (* 100.0 sumval) sumabs))))))

;; the first value goes into the sum at the 9th call, it's the first of
;; COUNT-1 wanted for the sums, hence 8+COUNT-1
(define (tdrei-warmup-count count)
  (+ count 7))


(define (tdrei-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos
		  parent #:before lo #:count (tdrei-warmup-count count)))
	 (array  (series-array parent start hi))
	 (proc   (tdrei-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 1)    ;; high
		(array-ref array i 2)    ;; low
		(array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (proc (array-ref array i 1)  ;; high
				      (array-ref array i 2)  ;; low
				      (array-ref array i 3)) ;; close
		      i)))))

(define-public (tdrei-series series count)
  (ohlcv-required "TDREI" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "TDREI") count))
    (series-datatype!  new-series tdrei-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts
				   tdrei-fill (tdrei-warmup-count count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public tdrei-datatype
  (datatype-new!
   #:key     'tdrei
   #:name    (_ "TD REI - Range Expansion Index")
   #:input   datatype-ohlcv
   #:units   'percentage-plusorminus100
   #:minimum -100
   #:maximum 100
   #:hlines  '(-45 45)
   #:manual  (_ "CONTEXT(manual-node): TD Range Expansion Index")
   #:proc    tdrei-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'tdrei-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 5))))
