;; TEMA Triple Exponential Moving Average.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tema)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart ema-2)
  #:use-module (chart ema-3)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;;
;;


;; Return a procedure which calculates a Triple EMA with given COUNT period
;; smoothing.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the Triple EMA up to (and including) that value, or #f if not enough
;; points yet.
;;
;; A Triple EMA is in theory influenced by all preceding data, but
;; `tema-warmup-count' below is designed to determine a warmup count.  After
;; calling PROC with (tema-warmup-count COUNT) many values, the next call
;; will have an omitted weight of no more than 0.1% of the total.  Omitting
;; 0.1% should be negligable, unless past values are ridiculously bigger
;; than recent ones.
;;
(define-public (tema-calc-proc count)
  (define ema-proc   (ema-calc-proc count))
  (define ema-2-proc (ema-2-calc-proc count))
  (define ema-3-proc (ema-3-calc-proc count))

  (lambda (x)
    (let ((e   (ema-proc x))
	  (ee  (ema-2-proc x))
	  (eee (ema-3-proc x)))
      (and e ee eee
	   (+ (* 3 e)
	      (* -3 ee)
	      eee)))))

;; ENHANCE-ME: this is almost certainly more than needed, but a calculation
;; like dema-warmup-count is a bit tricky
(define-public (tema-warmup-count count)
  (max (ema-warmup-count count)
       (ema-2-warmup-count count)
       (ema-3-warmup-count count)))


(define (tema-fill series new-array lo hi)
  (receive-list (count warmup)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (proc   (tema-calc-proc count)))

      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((this (array-ref array i)))
	  (array-set! new-array (proc this) i))))))

(define-public (tema-series series count)
  (let ((new-series (series-copy series))
	(warmup     (tema-warmup-count count)))
    (series-name!     new-series (format #f "~a ~a" (_ "TEMA") count))
    (series-datatype! new-series tema-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list count warmup))
    (series-require!  new-series (indicator-require-parts tema-fill warmup))
    (series-array!    new-series #f)
    new-series))


(define tema-datatype
  (datatype-new!
   #:key      'tema
   #:name     (_ "TEMA - Triple EMA")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Double and Triple Exponential Moving Average")
   #:proc     tema-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'tema-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
