;; Test cookies.scm functions.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(load "tests.scm")
(use-modules (chart cookies))


(define tempfile "cookies.tmp")

;; no record compares in guile 1.6
(define (cookie=? x y)
  (and (equal? (cookie-name x)   (cookie-name y))
       (equal? (cookie-value x)  (cookie-value y))
       (equal? (cookie-domain x) (cookie-domain y))
       (equal? (cookie-path x)   (cookie-path y))))

(define (cookie-list=? x y)
  (lset= cookie=? x y))

(define cookiejar-write (@@ (chart cookies) cookiejar-write))


;;
;; cookie-parse
;;

(let* ((str  "foo=bar")
       (url  "http://foo.com")
       (want (cookie-new #:name "foo" #:value "bar" #:domain #f #:path "/"))
       (got  (cookie-parse str url)))
  (test "cookie-parse" str want got #:compare cookie=?))

(let* ((str  "foo=bar;")
       (url  "http://foo.com")
       (want (cookie-new #:name "foo" #:value "bar" #:domain #f #:path "/"))
       (got  (cookie-parse str url)))
  (test "cookie-parse" str want got #:compare cookie=?))

(let* ((str  "foo=bar; path=/zz")
       (url  "http://foo.com")
       (want (cookie-new #:name "foo" #:value "bar" #:domain #f #:path "/zz"))
       (got  (cookie-parse str url)))
  (test "cookie-parse" str want got #:compare cookie=?))

(let* ((str  "foo=bar; path=/zz; domain=somewhere.com")
       (url  "http://foo.com")
       (want (cookie-new #:name "foo" #:value "bar" #:domain #f #:path "/zz"))
       (got  (cookie-parse str url)))
  (test "cookie-parse" str want got #:compare cookie=?))

(let* ((str  "foo=bar; expires=Sat, 05-May-2007 00:04:15 GMT")
       (url  "http://foo.com")
       (want (cookie-new #:name "foo" #:value "bar" #:domain #f #:path "/"))
       (got  (cookie-parse str url)))
  (test "cookie-parse" str want got #:compare cookie=?))


;;
;; cookie-send<?
;;

(define cookie-send<? (@@ (chart cookies) cookie-send<?))

(let* ((x    (cookie-new #:name "aaa" #:value "aaa" #:path "/"))
       (y    (cookie-new #:name "bbb" #:value "bbb" #:path "/zzz"))
       (want #f)
       (got  (cookie-send<? x y)))
  (test "cookie-send<?" (list x y) want got))

(let* ((x    (cookie-new #:name "bbb" #:value "bbb" #:path "/zzz"))
       (y    (cookie-new #:name "aaa" #:value "aaa" #:path "/"))
       (want #t)
       (got  (cookie-send<? x y)))
  (test "cookie-send<?" (list x y) want got))

(let* ((x    (cookie-new #:name "c" #:value "d" #:path "/quux"))
       (y    (cookie-new #:name "a" #:value "b" #:path "/"))
       (want #t)
       (got  (cookie-send<? x y)))
  (test "cookie-send<?" (list x y) want got))


;;
;; cookiejar-add-headers
;;

(delete-file-maybe tempfile)

(let* ((url     "http://somewhere.com/index.html")
       (headers "\
HTTP/1.x 200 OK\r
Set-Cookie: abc=def; expires=Sat, 05-May-2007 00:04:15 GMT; path=/\r
\r\n")
       (want    (list (cookie-new #:name "abc" #:value "def"
				  #:domain #f #:path "/")))
       (got     (begin
		  (cookiejar-add-headers tempfile url headers)
		  (cookiejar-read tempfile))))
  (test "cookiejar-add-headers" headers want got #:compare cookie-list=?))

(let* ((url     "http://somewhere.com/index.html")
       (headers "\
HTTP/1.x 200 OK\r
Set-Cookie: foo=bar; path=/quux\r
\r\n")
       (want    (list (cookie-new #:name "abc" #:value "def"
				  #:domain #f #:path "/")
		      (cookie-new #:name "foo" #:value "bar"
				  #:domain #f #:path "/quux")))
       (got     (begin
		  (cookiejar-add-headers tempfile url headers)
		  (cookiejar-read tempfile))))
  (test "cookiejar-add-headers" headers want got #:compare cookie-list=?))


;;
;; cookiejar-get-line
;;

(delete-file-maybe tempfile)
(cookiejar-write tempfile
		 (list (cookie-new #:name "abc" #:value "def"
				   #:domain #f #:path "/")))
(let* ((url     "http://somewhere.com/index.html")
       (headers "\
HTTP/1.x 200 OK\r
Set-Cookie: abc=def; expires=Sat, 05-May-2007 00:04:15 GMT; path=/\r
\r\n")
       (want    "Cookie: abc=def\r\n")
       (got     (cookiejar-get-line tempfile url)))
  (test "cookiejar-get-line" headers want got))

(cookiejar-write tempfile
		 (list (cookie-new #:name "abc" #:value "def"
				   #:domain #f #:path "/")
		       (cookie-new #:name "foo" #:value "bar"
				   #:domain #f #:path "/quux")))
(let* ((url     "http://somewhere.com/quux/index.html")
       (headers "\
HTTP/1.x 200 OK\r
Set-Cookie: abc=def; expires=Sat, 05-May-2007 00:04:15 GMT; path=/\r
\r\n")
       (want    "Cookie: foo=bar; abc=def\r\n")
       (got     (cookiejar-get-line tempfile url)))
  (test "cookiejar-get-line" headers want got))



(delete-file-maybe tempfile)
(delete-file-maybe ".lockfile")

(tests-end)
