;; Test download functions.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(load "tests.scm")
(use-modules (chart database)
	     (chart extra-glib))

(system "rm -rf database-temp")
(set! database-directory    (g-build-filename (getcwd) "database-temp"))
(set! data-directory        (g-build-filename database-directory "data"))
(set! cache-directory       (g-build-filename database-directory "cache"))
(set! user-notes-directory  (g-build-filename database-directory "notes"))



;;
;; string-escape-unprintable
;;

(define string-escape-unprintable
  (@@ (chart download) string-escape-unprintable))

(for-each (lambda (elem)
	    (let* ((str (first elem))
		   (want (second elem))
		   (got  (string-escape-unprintable str)))
	      (test "string-escape-unprintable" str want got)))

	  `((""    "")
	    ("a"   "a")
	    (" "   " ")
	    ("ab\ncd\n" "ab\ncd\n")
	    (,(string-ints 31)    "\\x1F")
	    (,(string-ints 0 1 2) "\\x00\\x01\\x02")))

;;
;; 0-string?
;;

(for-each (lambda (elem)
	    (receive-list (str want)
		elem
	      (let ((got  (->bool (0-string? str))))
		(test "0-string?" str want got))))

	  '(("0"      #t)
	    ("0."     #t)
	    (".0"     #t)
	    ("0.0"    #t)
	    ("0.0000" #t)
	    ("."      #f)
	    ("1"      #f)
	    ("10"     #f)
	    ("01"     #f)
	    ("-01"    #f)
	    ("0100"   #f)
	    ("0100.00" #f)
	    ("+0"     #t)
	    ("-0"     #t)
	    ("+0."    #t)
	    ("-0."    #t)
	    ("+.0"    #t)
	    ("-.0"    #t)
	    ("+0.0"   #t)
	    ("-0.0"   #t)))

;;
;; csv-split
;;

(let* ((str      "foo")
       (want     (list "foo"))
       (got      (csv-split str)))
  (test "csv-split" str want got))

(let* ((str      "foo,")
       (want     (list "foo" ""))
       (got      (csv-split str)))
  (test "csv-split" str want got))

(let* ((str      "foo,bar")
       (want     (list "foo" "bar"))
       (got      (csv-split str)))
  (test "csv-split" str want got))

(let* ((str      "\"foo\",bar")
       (want     (list "foo" "bar"))
       (got      (csv-split str)))
  (test "csv-split" str want got))

(let* ((str      "\"foo,bar\"")
       (want     (list "foo,bar"))
       (got      (csv-split str)))
  (test "csv-split" str want got))

(let* ((str      "\"foo,bar\",quux")
       (want     (list "foo,bar" "quux"))
       (got      (csv-split str)))
  (test "csv-split" str want got))

;;
;; d/m/y-str->tdate
;;

(for-each (lambda (elem)
	    (let* ((str  (first elem))
		   (want (second elem))
		   (got  (catch #t
			   (lambda ()
			     (d/m/y-str->tdate str))
			   (lambda args
			     'error))))
	      (test "d/m/y-str->tdate" str want got)))

	  `(("5 Jan 1970"      0)
	    ("5 January 1970"  0)
	    ("5/1/1970"        0)
	    ("5.1.1970"        0)
	    ("5-Jan-1970"      0)
	    ("5 Foo 1970"      error)
	    ("Jan.5.1970"      error)
	    ("5.30.1970"       error)
	    ("1970.5.30"       error)
	    ("1 Jan 1970"      ,(ymd->tdate 1970 1 1))
	    ("31 Dec 2000"     ,(ymd->tdate 2000 12 31))))

;;
;; download-message
;;

(let ((want "x\r \rabc\n")
      (got  (with-output-to-string
	      (lambda ()
		(display "x")
		(download-message "abc")))))
  (test "download-message" "x,abc" want got))

(let ((want "x\n")
      (got  (with-output-to-string
	      (lambda ()
		(download-message "x")))))
  (test "download-message" "x" want got))

(let ((want "xyz\n")
      (got  (with-output-to-string
	      (lambda ()
		(download-message "xyz")))))
  (test "download-message" "xyz" want got))

(let ((want "abc\r   \rx\n")
      (got  (with-output-to-string
	      (lambda ()
		(display "abc")
		(download-message "x")))))
  (test "download-message" "abc,x" want got))

;;
;; download-process
;;

(define (get-data symbol)
  (let* ((series (database-read-series symbol))
	 (lo     (series-lo series))
	 (hi     (series-hi series))
	 (array  (series-array series lo hi))
	 (ret    '()))
    (do ((i hi (1- i)))
	((< i lo))
      (set! ret (cons (list i
			    (array-ref array i 0)
			    (array-ref array i 1)
			    (array-ref array i 2)
			    (array-ref array i 3)
			    (array-ref array i 4))
		      ret)))
    ret))

(define (get-divs symbol)
  (let ((series (database-read-series symbol)))
    (sort (series-dividends series) car<)))

(define (get-decimals symbol)
  (series-decimals (database-read-series symbol)))

(let ((symbol "TESTSYMBOL"))
  (database-delete symbol)
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:row-list '((#:tdate 100 #:commodity "TESTSYMBOL" #:mdate 50
			 #:open 2 #:high 4 #:low 1 #:close 3)
		(#:tdate 101 #:commodity "TESTSYMBOL" #:mdate 50
			 #:open 6 #:high 8 #:low 5 #:close 7)))
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:row-list '((#:tdate 101 #:commodity "TESTSYMBOL" #:mdate 51
			 #:open 10 #:high 12 #:low 9 #:close 11)
		(#:tdate 102 #:commodity "TESTSYMBOL" #:mdate 51
			 #:open 14 #:high 16 #:low 13 #:close 15)))
  (let ((want-data '((100 2 4 1 3 #f)
		     (101 6 8 5 7 #f)
		     (102 14 16 13 15 #f)))
	(want-divs '((100 month 50)
		     (102 month 51 "4"))))
    (test "download-process sep months" symbol want-data (get-data symbol))
    (test "download-process sep months" symbol want-divs (get-divs symbol))))

(let ((symbol "TESTSYMBOL"))
  (database-delete symbol)
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:row-list '((#:tdate 100 #:commodity "TESTSYMBOL" #:mdate 50
			 #:open 2 #:high 4 #:low 1 #:close 3)))
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:row-list '((#:tdate 100 #:commodity "TESTSYMBOL" #:mdate 50
			 #:volume 999)))
  (let ((want-data '((100 2 4 1 3 999))))
    (test "download-process sep volume" symbol want-data (get-data symbol))))

(let ((symbol "TESTSYMBOL"))
  (database-delete symbol)
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:extra-decimals 2
   #:row-list '((#:tdate 100 #:commodity "TESTSYMBOL" #:close "123")))
  (let ((want-data     '((100 #f #f #f 123 #f)))
	(want-decimals 2))
    (test "download-process extra-decimals"
	  symbol want-data (get-data symbol))
    (test "download-process extra-decimals"
	  symbol want-decimals (get-decimals symbol))))

(let ((symbol "TESTSYMBOL"))
  (database-delete symbol)
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:row-list '((#:tdate 100 #:commodity "TESTSYMBOL" #:close "12.3")))
  (download-process
   #:module "Test"
   #:symbol-list (list symbol)
   #:extra-decimals 2
   #:row-list '((#:tdate 101 #:commodity "TESTSYMBOL" #:close "45.6")))
  (let ((want-data     '((100 #f #f #f 12300 #f)
			 (101 #f #f #f   456 #f)))
	(want-decimals 3))
    (test "download-process extra-decimals increase"
	  symbol want-data (get-data symbol))
    (test "download-process extra-decimals increase"
	  symbol want-decimals (get-decimals symbol))))

(database-delete "TESTSYMBOL")


;;
;; download-status-output
;;

;; force download-status-output to print, only normally goes to a tty
(let ((old-isatty? isatty?))
  (set! isatty? (lambda (port) #t))

  (let ((want "\rx")
	(got  (with-output-to-string
		(lambda ()
		  (download-status-output "x")))))
    (test "download-status-output" "x" want got))

  (let ((want "x\ry")
	(got  (with-output-to-string
		(lambda ()
		  (display "x")
		  (download-status-output "y")))))
    (test "download-status-output" "x,y" want got))

  (let ((want (string-append "xxx\ry  " (make-string 2 #\bs)))
	(got  (with-output-to-string
		(lambda ()
		  (display "xxx")
		  (download-status-output "y")))))
    (test "download-status-output" "xxx,y" want got))

  (let ((want (string-append "abcd\ry   " (make-string 3 #\bs)))
	(got  (with-output-to-string
		(lambda ()
		  (display "abcd")
		  (download-status-output "y")))))
    (test "download-status-output" "abcd,y" want got))

  (let ((want "a\rxyz")
	(got  (with-output-to-string
		(lambda ()
		  (display "a")
		  (download-status-output "xyz")))))
    (test "download-status-output" "x,y" want got))

  (set! isatty? old-isatty?))

;;
;; download-start-tdate
;;

(let* ((symbol "NOSUCHSYMBOL")
       (want   (- (tdate-today (symbol-timezone symbol))
		  initial-download-days -1))
       (got    (download-start-tdate symbol)))
  (test "download-status-output" symbol want got))

(let* ((symbol "EMPTYSYMBOL")
       (want   (- (tdate-today (symbol-timezone symbol))
		  initial-download-days -1))
       (got    (begin
		 (database-new symbol)
		 (download-start-tdate symbol))))
  (test "download-status-output" symbol want got))

(let* ((symbol "5DAYSAGOSYMBOL")
       (hi     (- (tdate-today (symbol-timezone symbol))
		  5))
       (array  (make-array #f (list hi hi) '(0 5)))
       (series (begin
		 (array-set! array 123 hi 3)
		 (series-new #:symbol symbol
			     #:array  array
			     #:datatype datatype-ohlcv)))
       (want   (1+ hi))
       (got    (begin
		 (database-write-series series)
		 (download-start-tdate symbol))))
  (test "download-status-output" symbol want got))

(let* ((symbol "DONEDOWNLOADSYMBOL")
       (hi     (- (tdate-today (symbol-timezone symbol))
		  5))
       (array  (make-array #f (list hi hi) '(0 5)))
       (series (begin
		 (array-set! array 123 hi 3)
		 (series-new #:symbol symbol
			     #:array  array
			     #:datatype datatype-ohlcv)))
       (want   (+ hi 3))
       (got    (begin
		 (series-info-add! series 'download-done (+ hi 2))
		 (database-write-series series)
		 (download-start-tdate symbol))))
  (test "download-status-output" symbol want got))

;;
;; map-rows!
;;

(let* ((lst      '(("a") ("b")))
       (counter  0)
       (want     '((1 "a") (2 "b")))
       (got      (map-rows! (lambda (x)
			      (set! counter (1+ counter))
			      (list counter x))
			    lst)))
  (test "map-rows!" lst want got))

;;
;; for-each-line
;;

(define (for-each-line-calls str)
  (let ((ret '()))
    (for-each-line (lambda (line)
		     (set! ret (cons line ret)))
		   str)
    (reverse! ret)))

(let* ((str  "")
       (want '())
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "\n")
       (want '(""))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "\r")
       (want '(""))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "\r\n")
       (want '(""))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "\r\n\r\n")
       (want '(""))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "foo")
       (want '("foo"))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "foo\r\n\r\n")
       (want '("foo"))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "foo\r\nbar\r\n")
       (want '("foo" "bar"))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

(let* ((str  "foo\r\nbar")
       (want '("foo" "bar"))
       (got  (for-each-line-calls str)))
  (test "for-each-line" str want got))

;;
;; h:m:s-str->hms
;;

(for-each (lambda (elem)
	    (let* ((str  (first elem))
		   (want (second elem))
		   (got  (h:m:s-str->hms str)))
	      (test "h:m:s-str->hms" str want got)))

	  `(("00:00"    (0 0 0))
	    ("1:00"     (1 0 0))
	    ("01:00"    (1 0 0))
	    ("01:00:00" (1 0 0))
	    ("12:00am"  (0 0 0))
	    ("1:00am"   (1 0 0))
	    ("1:01am"   (1 1 0))
	    ("12:00pm"  (12 0 0))
	    ("1:00pm"   (13 0 0))
	    ("11:59pm"  (23 59 0))
	    ("13:00"    (13 0 0))
	    ("23:59"    (23 59 0))))

;;
;; java-write
;;

(let* ((str  "blah.blah('');")
       (want "")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('');")
       (want "")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('abc');")
       (want "abc")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('abc');   document.write('def');")
       (want "abcdef")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('abc');   document.write('def');  blah();")
       (want "abcdef")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('\\'abc\\'');")
       (want "'abc'")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('\\'abc\\'def\\'');")
       (want "'abc'def'")
       (got  (java-write str)))
  (test "java-write" str want got))

(let* ((str  "document.write('abc\\ndef\\n');")
       (want "abc\ndef\n")
       (got  (java-write str)))
  (test "java-write" str want got))

;;
;; lang-select
;;

(define (call-with-env-var name value thunk)
  (let ((old-value #f))
    (dynamic-wind
	(lambda ()
	  (set! old-value (getenv name))
	  (setenv name value))
	thunk
	(lambda ()
	  (setenv name old-value)))))

(let* ((targets  '(("en" "foo")
		   ("ja" "bar")))
       (language "en:ja")
       (want     "foo")
       (got      (call-with-env-var "LANGUAGE" language
		   (lambda ()
		     (lang-select targets)))))
  (test "lang-select" (list language targets) want got))

(let* ((targets  '(("en" "foo")
		   ("ja" "bar")))
       (language "ja:en")
       (want     "bar")
       (got      (call-with-env-var "LANGUAGE" language
		   (lambda ()
		     (lang-select targets)))))
  (test "lang-select" (list language targets) want got))

(let* ((targets  '(("en" "foo")
		   ("ja" "bar")))
       (language "ja_JP")
       (want     "bar")
       (got      (call-with-env-var "LANGUAGE" language
		   (lambda ()
		     (lang-select targets)))))
  (test "lang-select" (list language targets) want got))

(let* ((targets  '(("en" "foo")
		   ("ja" "bar" #:charset "NOSUCHCHARSET")))
       (language "ja_JP")
       (want     "foo")
       (got      (call-with-env-var "LANGUAGE" language
		   (lambda ()
		     (lang-select targets)))))
  (test "lang-select" (list language targets) want got))

(let* ((targets  '(("en" "foo")
		   ("ja" "bar" #:charset "ASCII")))
       (language "ja_JP")
       (want     "bar")
       (got      (call-with-env-var "LANGUAGE" language
		   (lambda ()
		     (lang-select targets)))))
  (test "lang-select" (list language targets) want got))

(let* ((targets  '(("en" "foo")
		   ("ja" "bar")))
       (language "xx:yy")
       (want     "foo")
       (got      (call-with-env-var "LANGUAGE" language
		   (lambda ()
		     (lang-select targets)))))
  (test "lang-select" (list language targets) want got))

;;
;; m/d/y-str->tdate
;;

(for-each (lambda (elem)
	    (let* ((str  (first elem))
		   (want (second elem))
		   (got  (catch #t
			   (lambda ()
			     (m/d/y-str->tdate str))
			   (lambda args
			     'error))))
	      (test "m/d/y-str->tdate" str want got)))

	  `(("1/1/1970"        ,(ymd->tdate 1970 1 1))
	    ("12/31/2000"      ,(ymd->tdate 2000 12 31))
	    ("Jan 5 1970"      0)
	    ("January 5, 1970" 0)
	    ("1/5/1970"        0)
	    ("1.5.1970"        0)
	    ("Jan-5-1970"      0)
	    ("Foo 5 1970"      error)
	    ("5.Jan.1970"      error)
	    ("30.5.1970"       error)
	    ("1970-1-5"        error)
	    ("31 Dec 2000"     error)))


;;
;; mdate->last-tdate
;;

(for-each (lambda (elem)
	    (let* ((msdate (first elem))
		   (want   (second elem))
		   (got    (mdate->last-tdate msdate)))
	      (test "mdate->last-tdate" msdate want got)))

	  `((-1      -3) ;; December 1969
	    ))

;;
;; month->nearest-year
;;

(let ((old-mdate-today  mdate-today)
      (fake-today-year  #f)
      (fake-today-month #f))
  (set! mdate-today (lambda ()
		      (ymd->mdate fake-today-year fake-today-month 1)))
  (for-each
   (lambda (data)
     (receive-list (current-month current-year month want)
	 data
       (set! fake-today-year  current-year)
       (set! fake-today-month current-month)
       (let ((got (month->nearest-year month)))
	 (test "month->nearest-year" (list current-month current-year month)
	       want got))))
   '((1 2006   1  2006)
     (1 2006   2  2006)
     (1 2006   3  2006)
     (1 2006   4  2006)
     (1 2006   5  2006)
     (1 2006   6  2006)
     (1 2006   7  2006)
     (1 2006   8  2005)
     (1 2006   9  2005)
     (1 2006  10  2005)
     (1 2006  11  2005)
     (1 2006  12  2005)

     (3 2006   1  2006)
     (3 2006   2  2006)
     (3 2006   3  2006)
     (3 2006   4  2006)
     (3 2006   5  2006)
     (3 2006   6  2006)
     (3 2006   7  2006)
     (3 2006   8  2006)
     (3 2006   9  2006)
     (3 2006  10  2005)
     (3 2006  11  2005)
     (3 2006  12  2005)

     (8 2006   1  2007)
     (8 2006   2  2007)
     (8 2006   3  2006)
     (8 2006   4  2006)
     (8 2006   5  2006)
     (8 2006   6  2006)
     (8 2006   7  2006)
     (8 2006   8  2006)
     (8 2006   9  2006)
     (8 2006  10  2006)
     (8 2006  11  2006)
     (8 2006  12  2006)

     (12 2006   1  2007)
     (12 2006   2  2007)
     (12 2006   3  2007)
     (12 2006   4  2007)
     (12 2006   5  2007)
     (12 2006   6  2007)
     (12 2006   7  2006)
     (12 2006   8  2006)
     (12 2006   9  2006)
     (12 2006  10  2006)
     (12 2006  11  2006)
     (12 2006  12  2006)))

  (set! mdate-today old-mdate-today))

;;
;; number+decimals->string
;;

(for-each (lambda (elem)
	    (let* ((args (first elem))
		   (want (second elem))
		   (got  (apply number+decimals->string args)))
	      (test "number+decimals->string" args want got)))

	  `(((  0 0)    "0")
	    ((123 0)    "123")
	    ((123 2)    "1.23")
	    (( 23 2)    "0.23")
	    ((  1 2)    "0.01")
	    ((100 2)    "1.00")

	    ((-123 0)    "-123")
	    ((-123 2)    "-1.23")
	    (( -23 2)    "-0.23")
	    ((  -1 2)    "-0.01")
	    ((-100 2)    "-1.00")))

;;
;; rassoc-ref
;;

(let* ((alist '())
       (val   1)
       (want  #f)
       (got   (rassoc-ref alist val)))
  (test "rassoc-ref" (list val alist) want got))

(let* ((alist '((x . 1)
		(y . 2)
		(z . 3)))
       (val   1)
       (want  'x)
       (got   (rassoc-ref alist val)))
  (test "rassoc-ref" (list val alist) want got))

(let* ((alist '((x . 1)
		(y . 2)
		(z . 3)))
       (val   3)
       (want  'z)
       (got   (rassoc-ref alist val)))
  (test "rassoc-ref" (list val alist) want got))

(let* ((alist '((x . 1)
		(y . 2)
		(z . 3)))
       (val   999)
       (want  #f)
       (got   (rassoc-ref alist val)))
  (test "rassoc-ref" (list val alist) want got))

;;
;; source-help-list
;;

(symbol-setups-load-everything)
(for-each (lambda (elem)
	    (let* ((pred (first elem))
		   (want #t)
		   (got  (sympred? pred)))
	      (test "source-help-list -- sympred?" elem want got)))
	  source-help-list)


;;
;; string-collapse-whitespace
;;

(let* ((str  "")
       (want "")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

(let* ((str  " ")
       (want " ")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

(let* ((str  "\t\t\t")
       (want " ")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

(let* ((str  "a b")
       (want "a b")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

(let* ((str  "a  b")
       (want "a b")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

(let* ((str  "a  \t\r\n    b")
       (want "a b")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

(let* ((str  "  \t\r\n    a  b    ")
       (want " a b ")
       (got  (string-collapse-whitespace str)))
  (test "string-collapse-whitespace" str want got))

;;
;; string-change-decimal-point
;;

(define string-change-decimal-point
  (@@ (chart download) string-change-decimal-point))

(let* ((args    '(".23" "."))
       (want    ".23")
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))

(let* ((args    '("1." "."))
       (want    "1.")
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))

(let* ((args    '("1.23" "."))
       (want    "1.23")
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))


(let* ((args    '(",23" ","))
       (want    ".23")
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))

(let* ((args    '("1," ","))
       (want    "1.")
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))

(let* ((args    '("1,23" ","))
       (want    "1.23")
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))


(let* ((args    '("." "."))
       (want    #f)
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))

(let* ((args    '("1.23" ","))
       (want    #f)
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))

(let* ((args    '("1,234" "."))
       (want    #f)
       (got     (apply string-change-decimal-point args)))
  (test "string-change-decimal-point" args want got))


;;
;; string-decimals-maybe
;;

(define string-decimals-maybe (@@ (chart download) string-decimals-maybe))

(for-each (lambda (elem)
	    (let* ((args (first elem))
		   (want (second elem))
		   (got     (apply string-decimals-maybe args)))
	      (test "string-decimals-maybe" args want got)))
	  '((("0" #f)     0)
	    (("9" #f)     0)
	    (("9." #f)    0)
	    (("9.5" #f)   1)
	    (("9.50" #f)  2)
	    (("9.55" #f)  2)

	    (("0" 0)      0)
	    (("9" 0)      0)
	    (("9." 0)     0)
	    (("9.0" 0)    0)
	    (("9.00" 0)   0)
	    (("9.5" 0)    1)
	    (("9.50" 0)   1)
	    (("9.55" 0)   2)

	    (("0" 2)       0)
	    (("9" 2)       0)
	    (("9." 2)      0)
	    (("9.0" 2)     1)
	    (("9.00" 2)    2)
	    (("9.5" 2)     1)
	    (("9.50" 2)    2)
	    (("9.55" 2)    2)
	    (("9.550" 2)   2)
	    (("9.555" 2)   3)
	    (("9.0000" 2)  2)
	    (("9.9000" 2)  2)
	    (("9.0900" 2)  2)
	    (("9.0090" 2)  3)
	    (("9.0009" 2)  4)
	    ))
;;
;; string->dividend-amount
;;

(let* ((args    '("1.23"))
       (want    '(amount 123 (decimals . 2)))
       (got     (apply string->dividend-amount args)))
  (test "string->dividend-amount" args want got))

(let* ((args    '("1.230"))
       (want    '(amount 1230 (decimals . 3)))
       (got     (apply string->dividend-amount args)))
  (test "string->dividend-amount" args want got))

(let* ((args    '("1."))
       (want    '(amount 1 (decimals . 0)))
       (got     (apply string->dividend-amount args)))
  (test "string->dividend-amount" args want got))

(let* ((args    '(".9"))
       (want    '(amount 9 (decimals . 1)))
       (got     (apply string->dividend-amount args)))
  (test "string->dividend-amount" args want got))

(let* ((args    '("."))
       (want    #f)
       (got     (apply string->dividend-amount args)))
  (test "string->dividend-amount" args want got))

(let* ((args    '("foo"))
       (want    #f)
       (got     (apply string->dividend-amount args)))
  (test "string->dividend-amount" args want got))

;;
;; string-drop-first-line
;;

(let* ((str  "abc")
       (want "")
       (got  (string-drop-first-line str)))
  (test "string-drop-first-line" str want got))

(let* ((str  "abc\ndef")
       (want "def")
       (got  (string-drop-first-line str)))
  (test "string-drop-first-line" str want got))

(let* ((str  "abc\rdef")
       (want "def")
       (got  (string-drop-first-line str)))
  (test "string-drop-first-line" str want got))

(let* ((str  "abc\r\r\rdef")
       (want "def")
       (got  (string-drop-first-line str)))
  (test "string-drop-first-line" str want got))

(let* ((str  "abc\r\ndef")
       (want "def")
       (got  (string-drop-first-line str)))
  (test "string-drop-first-line" str want got))

(let* ((str  "abc\r\n")
       (want "")
       (got  (string-drop-first-line str)))
  (test "string-drop-first-line" str want got))

;;
;; string-last-line
;;

(for-each (lambda (elem)
	    (let* ((str  (first elem))
		   (want (second elem))
		   (got  (string-last-line str)))
	      (test "string-last-line" str want got)))
	  '(("abc\r\n"        	      "abc")
	    ("abc\n"          	      "abc")
	    ("xyz\r\nabc\r\n" 	      "abc")
	    ("xyz\nabc\n"     	      "abc")
	    ("abc\r\n\r\n"            "abc")
	    ("xyz\r\n\r\nabc\r\n\r\n" "abc")))


;;
;; string-trim-to-2-decimals
;;

(let* ((str  "0")
       (want "0")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "00")
       (want "00")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "000")
       (want "000")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "0000")
       (want "0000")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.0")
       (want "1.0")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.00")
       (want "1.00")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.000")
       (want "1.00")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.0000")
       (want "1.00")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.0009")
       (want "1.0009")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.0090")
       (want "1.009")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.009090")
       (want "1.00909")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "123")
       (want str)
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "abc")
       (want str)
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.23")
       (want str)
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.230")
       (want "1.23")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.2300")
       (want "1.23")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.2301")
       (want "1.2301")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.23010")
       (want "1.2301")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

(let* ((str  "1.230100")
       (want "1.2301")
       (got  (string-trim-to-2-decimals str)))
  (test "string-trim-to-2-decimals" str want got))

;;
;; y/m/d-str->tdate
;;

(let* ((str  "1970/1/5")
       (want 0)
       (got  (y/m/d-str->tdate str)))
  (test "y/m/d-str->tdate" str want got))

(let* ((str  "70/1/6")
       (want 1)
       (got  (y/m/d-str->tdate str)))
  (test "y/m/d-str->tdate" str want got))

(let* ((str  "1970/01/07")
       (want 2)
       (got  (y/m/d-str->tdate str)))
  (test "y/m/d-str->tdate" str want got))

(let* ((str  "1970-01-08")
       (want 3)
       (got  (y/m/d-str->tdate str)))
  (test "y/m/d-str->tdate" str want got))


;;
;; year->last-tdate
;;

(let* ((year 1969)
       (want -3)   ;; Wed 31 Dec 1969
       (got  (year->last-tdate year)))
  (test "year->last-tdate" year want got))



(system "rm -rf database-temp")
(tests-end)
