;; Test http functions.

;; Copyright 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(load "tests.scm")
(use-modules (ice-9 receive)
	     (chart http)
	     (chart gzip)
	     (chart zlib))


;;
;; accept-encoding-header
;;

(define accept-encoding-header (@@ (chart http) accept-encoding-header))

(let ((old-zlib-version zlib-version)
      (old-gzip-version gzip-version)
      (gzip #f)
      (zlib #f))
  (set! gzip-version (lambda () gzip))
  (set! zlib-version (lambda () zlib))

  (for-each (lambda (elem)
	      (set! gzip (first elem))
	      (set! zlib (second elem))
	      (let* ((want (third elem))
		     (got  (accept-encoding-header)))
		(test "accept-encoding-header" (list gzip zlib) want got)))
	    ;; gzip  zlib  want
	    '(( #f    #f   "Accept-Encoding:")
	      ( #t    #f   "Accept-Encoding: gzip, compress")
	      ( #f    #t   "Accept-Encoding: gzip, deflate")
	      ( #t    #t   "Accept-Encoding: gzip, compress, deflate")))

  (set! gzip-version old-gzip-version)
  (set! zlib-version old-zlib-version))

;;
;; call-with-cleanup
;;

(let* ((ran     #f)
       (cleaned #f)
       (want    #t)
       (got     (begin
		  (call-with-cleanup (lambda ()
				       (set! cleaned #t))
				     (lambda ()
				       (set! ran #t)))
		  (and ran
		       cleaned))))
  (test "call-with-cleanup success" #f want got))

(let* ((ran     #f)
       (cleaned #f)
       (caught  #f)
       (want    #t)
       (got     (begin
		  (catch 'blah
		    (lambda ()
		      (call-with-cleanup (lambda ()
					   (set! cleaned #t))
					 (lambda ()
					   (set! ran #t)))
		      (throw 'blah))
		    (lambda args
		      (set! caught #t)))
		  (and ran
		       cleaned
		       caught))))
  (test "call-with-cleanup throw" #f want got))

;;
;; call-with-tempfileport
;;

(define (next-fd)
  (let ((fd (dup 0)))
    (close fd)
    fd))

(let* ((filename     "http.tmp")
       (old-fd       (next-fd))
       (got-filename #f)
       (want         #t)
       (got          (begin
		       (call-with-tempfileport (open-output-file filename)
			 (lambda (port)
			   (set! got-filename (port-filename port))))
		       (and (equal? filename got-filename)
			    (= old-fd (next-fd))
			    (not (file-exists? filename))))))
  (test "call-with-tempfileport success" filename want got))

(let* ((filename     "http.tmp")
       (old-fd       (next-fd))
       (got-filename #f)
       (caught       #f)
       (want         #t)
       (got          (begin
		       (catch 'blah
			 (lambda ()
			   (call-with-tempfileport (open-output-file filename)
			     (lambda (port)
		   (set! got-filename (port-filename port))
			       (throw 'blah))))
			 (lambda args
			   (set! caught #t)))
		       (and caught
			    (equal? filename got-filename)
			    (= old-fd (next-fd))
			    (not (file-exists? filename))))))
  (test "call-with-tempfileport error" filename want got))

;;
;; gethost-error?
;;

(define gethost-error? (@@ (chart http) gethost-error?))

(let* ((want #t)
       (got  (catch #t
	       (lambda () (sink (gethost "nosuchhostname")) #f)
	       (lambda args (->bool (gethost-error? args))))))
  (test "gethost-error?" "gethost nosuchhostname" want got))

(let* ((want #f)
       (got  (catch #t
	       (lambda () (throw 'foo))
	       (lambda args (gethost-error? args)))))
  (test "gethost-error?" "throw foo" want got))

;;
;; gethost-error-message
;;

(define gethost-error-message (@@ (chart http) gethost-error-message))

(let* ((want #t)
       (got  (catch #t
	       (lambda () (sink (gethost "nosuchhostname")) #f)
	       (lambda args (string? (gethost-error-message args))))))
  (test "gethost-error-message" "nosuchhostname" want got))

;;
;; host-in-domain?
;;

(define host-in-domain? (@@ (chart http) host-in-domain?))

(for-each (lambda (elem)
	    (let* ((host   (first elem))
		   (domain (second elem))
		   (want   (third elem))
		   (got    (host-in-domain? host domain)))
	      (test "http-url-host" (list host domain) want got)))
	  '(("foo.com"        "foo.com"     #t)
	    ("foo.com"        "sub.foo.com" #f)
	    ("www.foo.com"    "foo.com"     #t)
	    ("www.foo.com"    ".foo.com"    #t)
	    ("www.foo.com"    "barfoo.com"  #f)
	    ("www.barfoo.com" "foo.com"     #f)))

;;
;; http-headers-close?
;;

(define http-headers-close? (@@ (chart http) http-headers-close?))

(for-each (lambda (elem)
	    (let* ((headers (first elem))
		   (want    (second elem))
		   (got     (->bool (http-headers-close? headers))))
	      (test "http-headers-close?" headers want got)))
	  '(("HTTP/1.0 200 Ok\r\nContent-length: 0\r\n\r\n" #t)
	    ("HTTP/1.1 200 Ok\r\nContent-length: 0\r\n\r\n" #f)
	    ("HTTP/1.1 200 Ok\r\nContent-length: 0\r\nConnection: CLOSE\r\n\r\n" #t)))

;;
;; http-headers-content-length
;;

(define http-headers-content-length
  (@@ (chart http) http-headers-content-length))

(for-each (lambda (elem)
	    (let* ((headers (first elem))
		   (want    (second elem))
		   (got     (catch #t
			      (lambda () (http-headers-content-length headers))
			      list)))
	      (test "http-headers-content-length" headers want got)))
	  `(("HTTP/1.0 200\r\nContent-Length: 123\r\n\r\n" 123)
	    ("HTTP/1.0 200\r\n\r\n" #f)
	    ("HTTP/1.0 200\r\nContent-Length: xyz\r\n\r\n"
	     ,(list 'http (_ "Invalid response")))))

;;
;; http-headers-get
;;

(for-each (lambda (elem)
	    (let* ((headers  (first elem))
		   (name     (second elem))
		   (want     (third elem))
		   (got      (http-headers-get headers name)))
	      (test "http-headers-get" (list headers name) want got)))
	  '(("HTTP/1.0 200 Ok\r\nFoo: bar\r\n\r\n"
	     "Connection" #f)
	    ("HTTP/1.0 200 Ok\r\nConnection: close\r\n\r\n"
	     "Connection" "close")
	    ("HTTP/1.0 200 Ok\r\nXonnection: xx\r\nConnection: close\r\n\r\n"
	     "Connection" "close")
	    ("HTTP/1.0 200 Ok\r\nFoo: a,\r\n b, c\r\n\r\n"
	     "foo" "a, b, c")))

;;
;; http-headers-chunked?
;;

(define http-headers-chunked? (@@ (chart http) http-headers-chunked?))

(for-each (lambda (elem)
	    (let* ((headers (first elem))
		   (want    (second elem))
		   (got     (http-headers-chunked? headers)))
	      (test "http-headers-chunked?" headers want got)))
	  '(("HTTP/1.0 200 Ok\r\nContent-length: 0\r\n\r\n"            #f)
	    ("HTTP/1.0 200 Ok\r\nTransfer-encoding: something\r\n\r\n" #f)
	    ("HTTP/1.0 200 Ok\r\nTransfer-encoding: ChUnKed\r\n\r\n"   #t)))


;;
;; http-headers-split-tokens
;;

(define http-headers-split-tokens (@@ (chart http) http-headers-split-tokens))

(for-each (lambda (elem)
	    (let* ((str      (first elem))
		   (want     (second elem))
		   (got      (http-headers-split-tokens str)))
	      (test "http-headers-split-tokens" str want got)))
	  '((""         ())
	    ("close"    ("close"))
	    (" close "  ("close"))
	    ("foo, bar" ("foo" "bar"))))

;;
;; http-read-headers
;;

(let ()
  (define http-read-headers (@@ (chart http) http-read-headers))

  (define (run str)
    (let* ((pair (pipe))
	   (in   (car pair))
	   (out  (cdr pair)))
      (display str out)
      (close-port out)
      (let* ((a (catch #t
		  (lambda ()
		    (http-read-headers in))
		  (lambda args
		    args)))
	     (b (read-string-to-eof in)))
	(close-port in)
	(list a b))))

  (let* ((str  "")
	 (want (list (list 'http (_ "Incomplete response")) ""))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  (let* ((str  "HTTP/1.0 200 OK\r\n\r\n")
	 (want '("HTTP/1.0 200 OK\r\n\r\n" ""))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  (let* ((str  "HTTP/1.0 200 OK\r\n\r\nsome body text")
	 (want '("HTTP/1.0 200 OK\r\n\r\n" "some body text"))
	 (got  (run str)))
    (test "http-read-headers" str want got)))


;;
;; http-read-response
;;

(let ()
  (define http-read-response (@@ (chart http) http-read-response))

  (define (run str)
    (let* ((pair (pipe))
	   (in   (car pair))
	   (out  (cdr pair)))
      (display str out)
      (close-port out)
      (receive (headers body)
	  (catch #t
	    (lambda ()
	      (http-read-response in #f))
	    (lambda args
	      (values args #f)))
	(close-port in)
	(list headers body))))

  (let* ((str  "")
	 (want (list (list 'http (_ "Incomplete response")) #f))
	 (got  (run str)))
    (test "http-read-response" str want got))

  ;; no length
  (let* ((str  "HTTP/1.0 200 OK\r\n\r\nsome body text")
	 (want '("HTTP/1.0 200 OK\r\n\r\n" "some body text"))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  ;; length satisfied
  (let* ((str  "HTTP/1.0 200 OK\r\nContent-Length: 10\r\n\r\n1234567890zz")
	 (want '("HTTP/1.0 200 OK\r\nContent-Length: 10\r\n\r\n"
		 "1234567890"))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  ;; length not satisfied
  (let* ((str  "HTTP/1.0 200 OK\r\nContent-Length: 10\r\n\r\n12345")
	 (want (list (list 'http (_ "Incomplete response")) #f))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  ;; chunked empty
  (let* ((str  "HTTP/1.0 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n\
0\r\n\
\r\n")
	 (want '("HTTP/1.0 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n"
		 ""))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  ;; chunked one
  (let* ((str  "HTTP/1.0 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n\
5\r\n\
hello\r\n\
0\r\n\
\r\n")
	 (want '("HTTP/1.0 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n"
		 "hello"))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  ;; chunked trailing header
  (let* ((str  "HTTP/1.0 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n\
5\r\n\
hello\r\n\
0\r\n\
Foo: bar\r\n\
\r\n")
	 (want '("HTTP/1.0 200 OK\r\nTransfer-Encoding: chunked\r\nFoo: bar\r\n\r\n"
		 "hello"))
	 (got  (run str)))
    (test "http-read-headers" str want got))

  )

;;
;; http-read-to-crlf
;;

(let ()
  (define (run str)
    (let* ((pair (pipe))
	   (in   (car pair))
	   (out  (cdr pair)))
      (display str out)
      (close-port out)
      (let* ((a (catch #t
		  (lambda ()
		    (http-read-to-crlf in))
		  (lambda args
		    args)))
	     (b (read-string-to-eof in)))
	(close-port in)
	(list a b))))

  (let* ((str  "\r\n")
	 (want '("\r\n" ""))
	 (got  (run str)))
    (test "http-read-to-crlf" str want got))

  (let* ((str  "hello\r\n")
	 (want '("hello\r\n" ""))
	 (got  (run str)))
    (test "http-read-to-crlf" str want got))

  (let* ((str  "hello\r\nworld\r\n")
	 (want '("hello\r\n" "world\r\n"))
	 (got  (run str)))
    (test "http-read-to-crlf" str want got))

  (let* ((str  "\r")
	 (want (list (list 'http (_ "Incomplete response")) ""))
	 (got  (run str)))
    (test "http-read-headers" str want got)))


;;
;; http-url-host
;;

(let* ((url  "http://foo.org")
       (want "foo.org")
       (got  (http-url-host url)))
  (test "http-url-host" url want got))

(let* ((url  "http://foo.org/")
       (want "foo.org")
       (got  (http-url-host url)))
  (test "http-url-host" url want got))

(let* ((url  "http://foo.org:8080")
       (want "foo.org")
       (got  (http-url-host url)))
  (test "http-url-host" url want got))

(let* ((url  "http://foo.org:8080/")
       (want "foo.org")
       (got  (http-url-host url)))
  (test "http-url-host" url want got))

(let* ((url  "bogus")
       (want #f)
       (got  (false-if-exception (http-url-host url))))
  (test "http-url-host" url want got))

(let* ((url  "https://foo.org:8080/")
       (want "foo.org")
       (got  (http-url-host url)))
  (test "http-url-host" url want got))

;;
;; http-url-host:port
;;

(let* ((url  "http://foo.org")
       (want "foo.org")
       (got  (http-url-host:port url)))
  (test "http-url-host:port" url want got))

(let* ((url  "http://foo.org/")
       (want "foo.org")
       (got  (http-url-host:port url)))
  (test "http-url-host:port" url want got))

(let* ((url  "http://foo.org:8080")
       (want "foo.org:8080")
       (got  (http-url-host:port url)))
  (test "http-url-host:port" url want got))

(let* ((url  "http://foo.org:8080/")
       (want "foo.org:8080")
       (got  (http-url-host:port url)))
  (test "http-url-host:port" url want got))

(let* ((url  "bogus")
       (want #f)
       (got  (false-if-exception (http-url-host:port url))))
  (test "http-url-host:port" url want got))

(let* ((url  "https://foo.org:8080/")
       (want "foo.org:8080")
       (got  (http-url-host:port url)))
  (test "http-url-host:port" url want got))

;;
;; http-url-path
;;

(let* ((url  "http://foo.org/index.html")
       (want "/index.html")
       (got  (http-url-path url)))
  (test "http-url-path" url want got))

(let* ((url  "http://foo.org/")
       (want "/")
       (got  (http-url-path url)))
  (test "http-url-path" url want got))

(let* ((url  "http://foo.org")
       (want "/")
       (got  (http-url-path url)))
  (test "http-url-path" url want got))

;;
;; http-url-port
;;

(let* ((url  "http://foo.org")
       (want 80)
       (got  (http-url-port url)))
  (test "http-url-port" url want got))

(let* ((url  "http://foo.org/")
       (want 80)
       (got  (http-url-port url)))
  (test "http-url-port" url want got))

(let* ((url  "https://foo.org/")
       (want 443)
       (got  (http-url-port url)))
  (test "http-url-port" url want got))

(let* ((url  "http://foo.org:8080")
       (want 8080)
       (got  (http-url-port url)))
  (test "http-url-port" url want got))

(let* ((url  "http://foo.org:8080/")
       (want 8080)
       (got  (http-url-port url)))
  (test "http-url-port" url want got))

(let* ((url  "https://foo.org:8080/")
       (want 8080)
       (got  (http-url-port url)))
  (test "http-url-port" url want got))

;;
;; http-url-relative
;;

(let* ((url  "http://foo.org")
       (link "/")
       (want "http://foo.org/")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/index.html")
       (link "/top.html")
       (want "http://foo.org/top.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/dir/index.html")
       (link "/top.html")
       (want "http://foo.org/top.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/dir/index.html")
       (link "../up.html")
       (want "http://foo.org/up.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/dir/subdir/index.html")
       (link "../up.html")
       (want "http://foo.org/dir/up.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/dir/subdir/index.html")
       (link "http://bar.net/index.html")
       (want "http://bar.net/index.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org")
       (link "index.html")
       (want "http://foo.org/index.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/one/two/foo.html")
       (link "../../x.html")
       (want "http://foo.org/x.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/one/two/foo.html")
       (link "./x.html")
       (want "http://foo.org/one/two/x.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/one/two/foo.html")
       (link "././x.html")
       (want "http://foo.org/one/two/x.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/one/two/foo.html")
       (link "././../x.html")
       (want "http://foo.org/one/x.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

(let* ((url  "http://foo.org/one/two/")
       (link "././../x.html")
       (want "http://foo.org/one/x.html")
       (got  (http-url-relative url link)))
  (test "http-url-relative" (list url link) want got))

;;
;; http-no-proxy?
;;

(define http-no-proxy? (@@ (chart http) http-no-proxy?))

(for-each (lambda (elem)
	    (let* ((url      (first elem))
		   (no-proxy (second elem))
		   (want     (third elem))
		   (got      (begin
			       (putenv no-proxy)
			       (http-no-proxy? url))))
	      (test "http-no-proxy?" (list url no-proxy) want got)))
	  '(("http://www.foo.com"      "no_proxy=foo.com"      #t)
	    ("http://www.foo.com:8080" "no_proxy=foo.com"      #f)
	    ("http://www.foo.com"      "no_proxy=foo.com:8001" #f)
	    ("http://www.foo.com:80"   "no_proxy=foo.com"      #t)
	    ("http://www.foo.com"      "no_proxy=foo.com:80"   #t)
	    ("http://www.foo.com:80"   "no_proxy=foo.com:80"   #t)
	    ("http://www.foo.com"      "no_proxy=bar.com,.foo.com" #t)
	    ("http://www.foo.com"      "no_proxy= bar.com , foo.com" #t)
	    ("http://www.bar.com"      "no_proxy= .bar.com , foo.com" #t)
	    ))

;;
;; string-first-line
;;

(for-each (lambda (elem)
	    
	    (let* ((str  (first elem))
		   (want (second elem))
		   (got  (string-first-line str)))
	      (test "string-first-line" str want got)))
	  '(("" "")
	    ("abc"          "abc")
	    ("abc\n"        "abc")
	    ("\n"           "")
	    ("abc\r\n"      "abc")
	    ("abc\ndef"     "abc")
	    ("abc\r\ndef"   "abc")
	    ("abc\ndef\n"   "abc")
	    ("abc\ndef\r\n" "abc")))

;;
;; system-error-message
;;

(define system-error-message (@@ (chart http) system-error-message))

(let* ((want (strerror EBADF))
       (got  (catch #t
	       (lambda ()
		 (dup (- 0 1)) ;; hide literal -1 from guile-lint
		 #f)
	       (lambda args (system-error-message args)))))
  (test "system-error-message" "EBADF" want got))

;;
;; system-error-eagain?
;;

(let* ((want #f)
       (got  (catch #t
	       (lambda ()
		 (dup (- 0 1)) ;; hide literal -1 from guile-lint
		 'oops)
	       (lambda args (system-error-eagain? args)))))
  (test "system-error-eagain?" "EBADF" want got))

(let* ((pair   (pipe))
       (inport (car pair))
       (want   #t)
       (got    (catch #t
	       (lambda ()
		 (fcntl inport F_SETFL
			(logior (fcntl inport F_GETFL) O_NONBLOCK))
		 (read-char inport)
		 #f)
	       (lambda args
		 (->bool (system-error-eagain? args))))))
  (test "system-error-eagain?" "read pipe" want got))

;;
;; url-quote
;;

(for-each (lambda (elem)
	    (let* ((str  (first elem))
		   (want (second elem))
		   (got  (url-quote str)))
	      (test "url-quote" str want got)))
	  '((" "   "%20")
	    ("a b" "a%20b")
	    ("  "  "%20%20")
	    ("a  "  "a%20%20")
	    ("a  b"  "a%20%20b")
	    ("a b "  "a%20b%20")))

;;
;; url-scheme
;;

(let* ((url  "http://foo.org")
       (want "http")
       (got  (url-scheme url)))
  (test "url-scheme" url want got))

(let* ((url  "https://foo.org")
       (want "https")
       (got  (url-scheme url)))
  (test "url-scheme" url want got))


(tests-end)
