;; Test iconv.c functions.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(load "tests.scm")
(use-modules (ice-9 rdelim))

(if (= 0 (mallinfo-used))
    (display "Warning: no mallinfo(), memory leak checking will be minimal\n"))


;; extract the HAVE_ICONV setting from config.h, as #t or #f
;;
(define HAVE_ICONV
  (call-with-input-file "../config.h"
    (lambda (port)
      (let more ()
	(let ((line (read-line port)))
	  (cond ((eof-object? line)
		 (error "no HAVE_ICONV in config.h"))
		((string=? line "/* #undef HAVE_ICONV */")
		 #f)
		((regexp-exec #,(regexp "^#[ \t]*define[ \t]+HAVE_ICONV[ \t]+0*[^0]") line)
		 #t)
		(else
		 (more))))))))


;;
;; charset->utf8
;;

(malloced-steady
 (lambda ()
   (let* ((args '(not-a-string "hello"))
	  (want #f)
	  (got  (false-if-exception (apply charset->utf8 args))))
     (test "charset->utf8" args want got))))

(malloced-steady
 (lambda ()
   (let* ((args '("ASCII" not-a-string))
	  (want #f)
	  (got  (false-if-exception (apply charset->utf8 args))))
     (test "charset->utf8" args want got))))

(malloced-steady
 (lambda ()
   (let* ((args '("UTF-8" "hello"))
	  (want "hello")
	  (got  (apply charset->utf8 args)))
     (test "charset->utf8" args want got))))

(malloced-steady
 (lambda ()
   (let* ((args (list "UTF-8" (string-ints #xE0))) ;; incomplete utf8
	  (want (string-ints #xE0))                ;; unchecked utf8->utf8
	  (got  (false-if-exception (apply charset->utf8 args))))
     (test "charset->utf8" args want got))))

(malloced-steady
 (lambda ()
   (let* ((args '("bogus charset name" (string-ints #xE0)))
	  (want #f)
	  (got  (false-if-exception (apply charset->utf8 args))))
     (test "charset->utf8" args want got))))


;;
;; iconv-charset-available?
;;

(malloced-steady
 (lambda ()
   (let* ((charset 'not-a-string)
	  (want    'wrong-type-arg)
	  (got     (catch #t
		     (lambda ()
		       (iconv-charset-available? charset))
		     (lambda (key . rest)
		       key))))
     (test "iconv-charset-available?" charset want got))))

(malloced-steady
 (lambda ()
   (let* ((charset "ASCII")
	  (want    #t)
	  (got     (iconv-charset-available? charset)))
     (test "iconv-charset-available?" charset want got))))

(malloced-steady
 (lambda ()
   (let* ((charset "UTF-8")
	  (want    (if HAVE_ICONV #t #f))
	  (got     (iconv-charset-available? charset)))
     (test "iconv-charset-available?" charset want got))))

(malloced-steady
 (lambda ()
   (let* ((charset "something not available")
	  (want    #f)
	  (got     (iconv-charset-available? charset)))
     (test "iconv-charset-available?" charset want got))))


;;
;; locale->utf8
;;

(setlocale LC_ALL "C")

(malloced-steady
 (lambda ()
   (let* ((str     'not-a-string)
	  (want    #f)
	  (got     (false-if-exception (locale->utf8 str))))
     (test "locale->utf8" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str     "")
	  (want    "")
	  (got     (locale->utf8 str)))
     (test "locale->utf8" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str     "hello")
	  (want    "hello")
	  (got     (locale->utf8 str)))
     (test "locale->utf8" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str     (string-ints #xE0)) ;; possibly bogus byte
	  (want    #f)
	  (got     (begin
		     (false-if-exception (locale->utf8 str))
		     #f)))
     (test "locale->utf8 malloc" str want got))))


;;
;; utf8->locale
;;

(setlocale LC_ALL "C")

(malloced-steady
 (lambda ()
   (let* ((str     'not-a-string)
	  (want    #f)
	  (got     (false-if-exception (utf8->locale str))))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "")
	  (want "")
	  (got  (utf8->locale str)))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "hello")
	  (want "hello")
	  (got  (utf8->locale str)))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  (string-ints #x20 #x38 #xE6 #x9C #x88 #x0A)) ;; jp month
	  (want " 8?\n")
	  (got  (utf8->locale str)))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  (string-ints #x38 #xE6 #x9C #x88)) ;; japanese month
	  (want "8?")
	  (got  (utf8->locale str)))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  (string-ints #xE6 #x9C #x88 #x38)) ;; japanese month
	  (want "?8")
	  (got  (utf8->locale str)))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  (string-ints #xE6 #x9C #x88)) ;; japanese month char only
	  (want "?")
	  (got  (utf8->locale str)))
     (test "utf8->locale" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str   (string-ints #xE0)) ;; incomplete utf8
	  ;; exception with real iconv, "?" with fake
	  (want  (if HAVE_ICONV #f "?"))
	  (got   (false-if-exception (utf8->locale str))))
     (test "utf8->locale bad" str want got))))


;;
;; utf8-string-length
;;

(define (try-utf8-string-length str)
  (catch #t
    (lambda ()
      (utf8-string-length str))
    (lambda (key . args)
      key)))

(malloced-steady
 (lambda ()
   (let* ((str  'not-a-string)
	  (want 'wrong-type-arg)
	  (got  (try-utf8-string-length str)))
     (test "utf8-string-length" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "hello")
	  (want 5)
	  (got  (try-utf8-string-length str)))
     (test "utf8-string-length" str want got))))

(if HAVE_ICONV
    (begin

      (malloced-steady
       (lambda ()
	 (let* ((str  (string-ints #xE6 #x9C #x88)) ;; japanese month char
		(want 1)
		(got  (try-utf8-string-length str)))
	   (test "utf8-string-length" str want got))))

      (malloced-steady
       (lambda ()
	 (let* ((str  (string-ints #xE0)) ;; incomplete utf8
		(want 'system-error)
		(got  (try-utf8-string-length str)))
	   (test "utf8-string-length" str want got))))))


;;
;; utf8-string-take
;;

(define (try-utf8-string-take str len)
  (catch #t
    (lambda ()
      (utf8-string-take str len))
    (lambda (key . args)
      key)))

(malloced-steady
 (lambda ()
   (let* ((str  'not-a-string)
	  (len  3)
	  (want 'wrong-type-arg)
	  (got  (try-utf8-string-take str len)))
     (test "utf8-string-take" (list str len) want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "hello")
	  (len  'not-a-number)
	  (want 'wrong-type-arg)
	  (got  (try-utf8-string-take str len)))
     (test "utf8-string-take" (list str len) want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "hello")
	  (len  3)
	  (want "hel")
	  (got  (try-utf8-string-take str len)))
     (test "utf8-string-take" (list str len) want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "hello")
	  (len  5)
	  (want "hello")
	  (got  (try-utf8-string-take str len)))
     (test "utf8-string-take" (list str len) want got))))

(malloced-steady
 (lambda ()
   (let* ((str  "hello")
	  (len  6)
	  (want 'out-of-range)
	  (got  (try-utf8-string-take str len)))
     (test "utf8-string-take" (list str len) want got))))

(if HAVE_ICONV
    (begin

      (malloced-steady
       (lambda ()
	 (let* ((str  (string-append
		       (string-ints #xE6 #x9C #x88) ;; japanese month char
		       "hello"))
		(len   2)
		(want (string-append (string-ints #xE6 #x9C #x88) "h"))
		(got  (try-utf8-string-take str len)))
	   (test "utf8-string-take" (list str len) want got))))

      (malloced-steady
       (lambda ()
	 (let* ((str  (string-ints #xE0)) ;; incomplete utf8
		(len  3)
		(want 'system-error)
		(got  (try-utf8-string-take str len)))
	   (test "utf8-string-take" (list str len) want got))))))

;;
;; utf8-validate
;;

(define (try-utf8-validate str)
  (catch #t
    (lambda ()
      (utf8-validate str)
      #t)
    (lambda (key . args)
      key)))

(malloced-steady
 (lambda ()
   (let* ((str  'not-a-string)
	  (want 'wrong-type-arg)
	  (got  (try-utf8-validate str)))
     (test "utf8-validate" str want got))))

(malloced-steady
 (lambda ()
   (let* ((str  (string-ints #xE6 #x9C #x88)) ;; japanese month char
	  (want #t)
	  (got  (try-utf8-validate str)))
     (test "utf8-validate" str want got))))

(if HAVE_ICONV
    (malloced-steady
     (lambda ()
       (let* ((str  (string-ints #xE0)) ;; incomplete utf8
	      (want 'system-error)
	      (got  (try-utf8-validate str)))
	 (test "utf8-validate" str want got)))))

(tests-end)
