;; Shared setups for tests.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; not really a module, this file is used with `(load "tests.scm")'
(use-modules (srfi srfi-1)
	     (srfi srfi-13)
	     (ice-9 and-let-star)
	     (ice-9 optargs))


(define option-trace #f)


;; call (PROC) when exiting
(define (c-atexit proc)
  (let ((old-exit exit))
    (set! exit (lambda args
		 (proc)
		 (apply old-exit args)))))


;;-----------------------------------------------------------------------------

(define builddir     (getcwd))
(define top-builddir (dirname builddir))

;; config.scm, must use getcwd since otherwise load is relative to current file
(load (string-append top-builddir "/config.scm"))
(use-modules (chart config))

;; override to look in builddir for libs
 (set! chart-config-libchartdir top-builddir)


;;-----------------------------------------------------------------------------
;; fake site directory, symlinking to sources

(define (filename->absolute filename)
  (if (string-prefix? "/" filename)
      filename
      (string-append (getcwd) "/" filename)))

(define fake-sitedir       (string-append (getcwd) "/fake-sitedir"))
(define fake-chart-sitedir (string-append fake-sitedir "/chart"))

(define srcdir (filename->absolute (getenv "srcdir")))
(define top-srcdir (string-append srcdir "/.."))

(if (not (file-exists? fake-sitedir))
    (let ()
      (mkdir fake-sitedir)
      (symlink top-srcdir fake-chart-sitedir)))

(c-atexit (lambda ()
 	    (delete-file fake-chart-sitedir)
	    (rmdir fake-sitedir)))

(set! %load-path (cons fake-sitedir %load-path))


;;-----------------------------------------------------------------------------

(use-modules (chart misc)
	     (chart download))

(define (flonum? obj)
  (and (number? obj)
       (real? obj)
       (inexact? obj)))

(define tests-count 0)
(define test-last-name #f)

(define*-public (test name data want got #:key (compare equal?))
  (set! test-last-name name)
  (if option-trace
      (format #t "~a ~a:\n" name data))
  (set! tests-count (1+ tests-count))
  (if (not (compare want got))
      (begin
	(format #t "~a:\n" name)
	(format #t "  data: ~s\n" data)
	(format #t "  want: ~s" want)
	(if (flonum? want)
	    (format #t "   (~a)" (inexact->exact want)))
	(newline)
	(format #t "  got:  ~s" got)
	(if (flonum? got)
	    (format #t "   (~a)" (inexact->exact got)))
	(newline)
	(exit 1))))

(define-public (=-close? want got)
  (within-range? (* want 0.9999) (* want 1.0001) got))

(define-public (tests-end)
  (format #t "~a tests ok\n" tests-count)
  ;; explicit exit to run c-atexit handlers
  (exit 0))

;; like Guile 1.8, made available in 1.6, but only for fetching
(if (not (defined? '@@))
    (begin
      (define-macro (@@ mod-name var-name)
	(let ((var (module-variable (resolve-module mod-name) var-name)))
	  (if (not var)
	      (error "no such variable" (list '@@ mod-name var-name)))
	  (list variable-ref var)))
      (export @@)))


;; like `string' but taking integer char values
(define-public (string-ints . int-list)
  (list->string (map! integer->char int-list)))

(define-public (seconds->hms time)
  (list (quotient time 3600)
	(quotient (remainder time 3600) 60)
	(remainder time 60)))

;; return a 1-dimensional array starting at LO and containing the DATA values
(define-public (array-from lo . data)
  (let* ((hi    (+ lo (length data) -1))
	 (array (make-array #f (list lo hi))))
    (do ((i lo (1+ i))
	 (data data (cdr data)))
	((null? data))
      (array-set! array (car data) i))
    array))

;; as per `resolve-interface' in boot-9.scm
;; an error can be thrown if the module exists but there's bad stuff in it
(define (module-exists? spec)
  (if (pair? (first spec))
      (set! spec (first spec)))
  (and-let* ((module (resolve-module spec)))
    (module-public-interface module)))

(define-public (need-gtk)
  (or (getenv "DISPLAY")
      (begin
	(display "Cannot run these tests without an X display\n")
	(exit 77)))
  (or (module-exists? '(gtk gtk))
      (begin
	(display "Cannot run these tests without guile-gtk\n")
	(exit 77))))

(define-public (source-help symbol)
  (symbol-setups symbol)
  (cond ((sympred-assoc symbol source-help-list) => second)
	(else #f)))

;; dummy to make given arg values look used to guile-lint
(define-public (sink . values)
  values)


;;-----------------------------------------------------------------------------
;; fake current-time

;; Set `current-time' to return the time corresponding to YMD/HMS in ZONE.
;;
(define*-public (fake-localtime-ymdhm! zone year month day hour minute
				       #:optional (second 0))
  (let ((tm (tm-new)))
    (set-tm:ymd tm year month day)
    (set-tm:hour tm hour)
    (set-tm:min  tm minute)
    (set-tm:sec  tm second)

    ;; establish usual tm:isdst for the given ymd/hms
    (set-tm:isdst tm (tm:isdst (cdr (mktime tm zone))))

    (set! fake-current-time-value (car (mktime tm zone)))))

(define fake-current-time-value #f)

(define real-current-time current-time)
(set! current-time
      (lambda ()
	(or fake-current-time-value
	    (real-current-time))))

(define-public (fake-time-end)
  (set! fake-current-time-value #f))


;;-----------------------------------------------------------------------------
;; malloc stuff

;; call (THUNK) a few times, looking for `gc-stats' bytes-malloced to be
;; unchanged, after a few tries at least
;;
;; this is designed to detect memory leaks or scm_done_malloc count leaks in
;; C code, a problem will show up as malloced increasing on every loop
;;
(define-public (malloced-steady thunk)
  (define old-malloced -1)
  (define new-malloced -1)
  (define old-used -1)
  (define new-used -1)
  (define saw-decrease #f)

  (let more ((attempt 0))
    (if (> attempt 30)
	(if saw-decrease
	    (begin
	      (format #t "Fluctuating: ~a\n~s\n"
		      test-last-name
		      (procedure-source thunk))
	      (dv old-malloced new-malloced
		  old-used new-used
		  (- new-malloced old-malloced)
		  (- new-used old-used))
	      #t)
	    (begin
	      (format #t "~a\n~s\n"
		      test-last-name
		      (procedure-source thunk))
	      (dv old-malloced new-malloced
		  old-used new-used
		  (- new-malloced old-malloced)
		  (- new-used old-used))
	      (error "Malloc leak")
	      #f))
	(begin
	  (set! old-malloced new-malloced)
	  (set! old-used new-used)

	  (gc)
	  (gc)
	  (thunk)
	  (thunk)
	  (gc)
	  (gc)
	  (set! new-malloced (assoc-ref (gc-stats) 'bytes-malloced))
	  (set! new-used (mallinfo-used))

	  (if (and (= old-malloced new-malloced)
		   (= old-used new-used))
	      'steady
	      (begin
		(if (and (positive? old-used)
			 (< old-used new-used))
		    (set! saw-decrease #t))
		(more (1+ attempt))))))))

;; a private object, not eq? to anything real
(define expect-exception-bad (cons 1 2))

(define-public (expect-exception thunk)
  (let ((ret (catch #t
	       (lambda ()
		 (thunk)
		 expect-exception-bad)
	       (lambda args
		 #f))))
    (if (eq? ret expect-exception-bad)
      (error "oops, should have got an exception"))
    ret))


;;-----------------------------------------------------------------------------
;; autoloads stuff

;; run all symbol-setups-alist autoloads, unconditionally
(define-public (symbol-setups-load-everything)
  (let ((done-lst '()))
    (let more ()
      (let ((lst (@@ (chart misc) symbol-setups-alist)))
	(set! lst (map cdr lst))
	(set! lst (delete-duplicates lst eq?))
	(set! lst (lset-difference eq? lst done-lst))
	(if (not (null? lst))
	    (begin
	      (for-each (lambda (setup)
			  (if (list? setup)
			      (resolve-interface setup) ;; module name
			      (setup))                  ;; setup func
			  (set! done-lst (lset-adjoin eq? done-lst setup)))
			lst)
	      (more)))))
    (length done-lst)))

;; return #t if PRED is a valid symbol predicate
(define-public (sympred? pred)
  (cond ((procedure? pred)
	 (pred "FOO.BAR")
	 #t)
	((string? pred)
	 (or (string-prefix? "." pred)   ;; normal
	     (string-prefix? "=" pred))) ;; yahoo "=X" currencies
	((list? pred)
	 (every sympred? pred))
	(else
	 #f)))
