;; Test timezone fallback data.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; This program reads finds all `timezone-choose' calls in the sources and
;; checks that the daylight savings in the posix style fallback (like
;; "EST-10EDT,M10.5.0,M3.5.0/3") match the name style like
;; "Australia/Sydney".


(load "tests.scm")
(use-modules (chart extra-glib))

;; time_t to adate, in GMT
(define-public (time->adate t)
  (- (quotient-floor t 86400) 4))

;; check from today, except apply a hack to skip to
;;
;;     - Nov 2006 which is after the new US/Canada daylight savings rules
;;       come in
;;
;;     - Oct 2008 which is after some special cases for Egypt ending DST
;;       early in 2006 and predicted 2007 and 2008
;; 
(define start-adate (max (ymd->adate 2006 11 8)
			 (ymd->adate 2008 10 1)
			 (time->adate (current-time))))

(define total-days (- (ymd->adate 2035 1 1) start-adate))  ;; to 2035
(define total-comparisons 0)
(define total-transitions 0)

(define print-filename #f)
(define print-zone-x #f)
(define print-zone-y #f)
(define all-good #t)

(define (form-tm tm)
  (format #f "~a ~a ~a  ~a:~a:~a  ~a ~a   ~a y~a"
	  (+ 1900 (tm:year tm))
	  (1+ (tm:mon tm))
	  (tm:mday tm)
	  (tm:hour tm)
	  (tm:min tm)
	  (tm:sec tm)
	  (tm:isdst tm)
	  (tm:zone tm)
	  (vector-ref '#("Sun" "Mon" "Tue" "Wed" "Thu" "Fri" "Sat")
		      (tm:wday tm))
	  (tm:yday tm)))

;; return time_t of midday on ADATE in ZONE
;; (actually, it can be an hour out depending on daylight savings, but that
;; doesn't matter for current uses)
(define (adate->time-midday adate zone)
  (let ((tm (apply ymd->tm (adate->ymd adate))))
    (set-tm:hour  tm 12)
    (set-tm:min   tm 0)
    (set-tm:sec   tm 0)
    (set-tm:isdst tm 0)
    (car (mktime tm zone))))

(define different-count 0)

(define (different t xtm ytm)
  (if print-filename
      (begin
	(format #t "~a:\n" print-filename)
	(set! print-filename #f)))
  (if print-zone-x
      (begin
	(format #t "~a versus ~a\n" print-zone-x print-zone-y)
	(set! print-zone-x #f)))

  (format #t "  diff ~a\n" t)
  (format #t "      x  ~a\n" (form-tm xtm))
  (format #t "      y  ~a\n" (form-tm ytm))

  (set! different-count (1+ different-count ))
  (if (> different-count 1000)
      (begin
	(display "too many errors, aborting\n")
	(exit 1))))


;; T-LIST is a list of time_t values which are midday with a daylight
;; savings transition taking place at some point in the following 24 hours.
;; Check timezones X and Y are the same on those transitions by checking
;; their tm:isdst each hour, on the half hour.
;;
(define (timezone-days t-list x y)
  (set! total-transitions (+ total-transitions (length t-list)))

  (let ((point-list (append-map! (lambda (t)
				   (iota 24
					 (+ t #,(hms->seconds 0 30 0))
					 #,(hms->seconds 0 30 0)))
				 t-list)))

    (setenv "TZ" x)
    (let ((xt-list (map localtime point-list)))
      (setenv "TZ" y)
      (let ((yt-list (map localtime point-list)))

	(for-each (lambda (t xtm ytm)
		    (set-tm:zone xtm "")
		    (set-tm:zone ytm "")
		    (if (not (equal? xtm ytm))
			(different t xtm ytm)))
		  point-list xt-list yt-list)))))
	
;; Check timezones X and Y are the same.
;;
(define (timezone-equality x y)
  (cond
   ((and (string<> "GMT" y)   ;; allow Africa/Accra against GMT
	 (not (timezone-known? x)))
    (format #t "  timezone ~a unknown, cannot check\n" x))
   ((not (timezone-known? y))
    (format #t "  timezone ~a unknown, cannot check\n" y))
   (else

    (set! total-comparisons (1+ total-comparisons))
    (set! print-zone-x x)
    (set! print-zone-y y)

    (let ((point-list (iota total-days
			    (adate->time-midday start-adate x)
			    86400))
	  (this-good  #t))

      (setenv "TZ" x)
      (let ((xt-list (map localtime point-list)))
	(setenv "TZ" y)
	(let ((yt-list (map localtime point-list)))

	  (for-each (lambda (t xtm ytm)
		      (set-tm:zone xtm "")
		      (set-tm:zone ytm "")
		      (if (not (equal? xtm ytm))
			  (begin
			    (set! this-good #f)
			    (different t xtm ytm))))
		    point-list xt-list yt-list))

	(if this-good
	    (let ((prev-dst (tm:isdst (first xt-list)))
		  (t-list   '()))
	      (for-each (lambda (t xtm)
			  (if (not (= prev-dst (tm:isdst xtm)))
			      (begin
				(set! t-list (cons t t-list))
				(set! prev-dst (tm:isdst xtm)))))
			point-list xt-list)
	      (timezone-days t-list x y))))

      (set! all-good (and all-good this-good))))))

(define (process-timezone expr)
  (receive-list (func name zone-list default-zone)
      expr
    (if (list? zone-list) ;; should be (quote ("foo" ...))
	(begin
	  (set! zone-list (second zone-list))

	  (while (length>=? zone-list 2)
	    (timezone-equality (first zone-list) (second zone-list))
	    (set! zone-list (cdr zone-list)))))))

(define (process-expr expr)
  (if (and (list? expr)
	   (not (null? expr)))
      (cond ((eq? 'timezone-choose (first expr))
	     (process-timezone expr))
	    ((not (null? expr))
	     (process-expr (car expr))
	     (process-expr (cdr expr))))))

(define (process-file filename)
  (set! print-filename filename)
  (call-with-input-file filename
    (lambda (port)
      (do ((obj (read port) (read port)))
	  ((eof-object? obj))
	(process-expr obj)))))


(for-each process-file
	  (directory-files (g-build-filename (getenv "srcdir") "..")
			   #t
			   (lambda (path)
			     (string-suffix? ".scm" path))))

(format #t "  ~a zone comparisons across ~a days, ~a total DST transitions\n"
	total-comparisons total-days total-transitions)
(exit (if all-good 0 1))
