;; Test Zlib functions.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(load "tests.scm")
(use-modules (chart zlib))


(or (gzopen-available?)
    (begin
      (display "gzopen not available (neither zlib library nor gzip program)\n")
      (exit 77)))

(if (= 0 (mallinfo-used))
    (display "Warning: no mallinfo(), memory leak checking will be minimal\n"))

(define (string-or-f? obj)
  (or (string? obj)
      (eq? #f obj)))


(define filename "zlib-temp.gz")
(define port (open-file filename "w+b"))
(define (store str)
  (seek port SEEK_SET 0)
  (truncate-file port 0)
  (display str port)
  (force-output port))

;; Run malloced-steady only with the real zlib, not with the pipe to gzip.
;; Ports in Guile 1.6.7 have an mtrigger leak which we would trip.
;;
(define (malloced-steady-with-zlib thunk)
  (if (defined? 'zlib-inflate)
      (malloced-steady thunk)
      (thunk)))

;;
;; gzopen
;;

(malloced-steady
 (lambda ()
   (let* ((args '(999 "rb"))  ;; bad filename
	  (want #f)
	  (got  (false-if-exception (apply gzopen args))))
     (if got
	 (gzclose got))
     (test "gzopen" args want got))))

(malloced-steady
 (lambda ()
   (let* ((args '("nosuchfile" 999))  ;; bad mode
	  (want #f)
	  (got  (false-if-exception (apply gzopen args))))
     (if got
	 (gzclose got))
     (test "gzopen" args want got))))

(malloced-steady
 (lambda ()
   (let* ((filename "nosuchfile")
	  (want     #f)
	  (got      (false-if-exception (gzopen filename "rb"))))
     (if got
	 (gzclose got))
     (test "gzopen" filename want got))))

;;
;; gzfile_free
;;

(malloced-steady-with-zlib
 (lambda ()
   (let ((str (string-ints 31 139 8 8 227 193 248 65 2 3 120 0 3
			   0 0 0 0 0 0 0 0 0)))
     (store str)
     (gzopen filename "rb")
     (test "gzopen gc" str #t #t))))

;;
;; gzclose
;;

(malloced-steady-with-zlib
 (lambda ()
   (let ((str (string-ints 31 139 8 8 227 193 248 65 2 3 120 0 3
			   0 0 0 0 0 0 0 0 0)))
     (store str)
     (gzclose (gzopen filename "rb"))
     (test "gzclose gc" str #t #t))))

;;
;; gzopen-available?
;;

(malloced-steady
 (lambda ()
   ;; no actual result checked, just availability
   (let* ((got (gzopen-available?))
	  (want got))
     (test "gzopen-available?" #f want got))))

;;
;; gzread
;;

(malloced-steady-with-zlib
 (lambda ()
   (let ((str (string-ints 31 139 8 8 227 193 248 65 2 3 120 0 3
			   0 0 0 0 0 0 0 0 0)))
     (store str)
     (let* ((gz   (gzopen filename "rb"))
	    (want #t)
	    (got  (eof-object? (gzread gz 999))))
       (gzclose gz)
       (test "gzread an empty file" str want got)))))

(malloced-steady-with-zlib
 (lambda ()
   (let ((str (string-ints 31 139 8 8 231 194 248 65 2 3 120 0 203
			   72 205 201 201 231 2 0 32 48 58 54 6 0 0 0)))
     (store str)
     (let* ((gz   (gzopen filename "rb"))
	    (want "hello\n")
	    (got  (gzread gz 999)))
       (gzclose gz)
       (test "gzread" str want got)))))

(malloced-steady-with-zlib
 (lambda ()
   (let ((str (string-ints 31 139 8 8 231 194 248 65 2 3 120 0 203
			   72 205 201 201 231 2 0 32 48 58 54 6 0 0 0)))
     (store str)
     (let ((gz (gzopen filename "rb")))

       (let ((want "hel")
	     (got  (gzread gz 3)))
	 (test "gzread" "hello 1st" want got))

       (let ((want "lo\n")
	     (got  (gzread gz 3)))
	 (test "gzread" "hello 2nd" want got))

       (let ((want #t)
	     (got  (eof-object? (gzread gz 999))))
	 (gzclose gz)
	 (test "gzread" "hello 3rd" want got))))))

;; bogus data
(malloced-steady-with-zlib
 (lambda ()
   (let ((str (string-ints 31 139)))
     (store str)
     (let* ((gz   (gzopen filename "rb"))
	    (want #f)
	    (got  (false-if-exception
		   (begin
		     ;; zlib gives the error from gzread, the gzip pipe
		     ;; implementation gets the error in gzclose
		     (gzread gz 999)
		     (gzclose gz)))))
       (test "gzread bogus data" str want got)))))


;;
;; zlib-inflate
;;

(if (defined? 'zlib-inflate)
    (let ()
      (define (try str)
	(catch #t
	  (lambda ()
	    (zlib-inflate str))
	  (lambda args
	    'error)))

      ;; gzip format
      (let* ((str  (list->string
		    (map integer->char
			 '(31 139 8 8 227 193 248 65 2 3 120 0 3
			      0 0 0 0 0 0 0 0 0))))
	     (want "")
	     (got  (zlib-inflate str)))
	(test "zlib-inflate" str want got))

      ;; gzip format
      (let* ((str  (list->string
		    (map integer->char
			 '(31 139 8 8 231 194 248 65 2 3 120 0 203 72 205 201
			      201 231 2 0 32 48 58 54 6 0 0 0))))
	     (want "hello\n")
	     (got  (zlib-inflate str)))
	(test "zlib-inflate" str want got))

      ;; gzip with garbage after
      (let* ((str  (list->string
		    (map integer->char
			 '(31 139 8 8 231 194 248 65 2 3 120 0 203 72 205 201
			      201 231 2 0 32 48 58 54 6 0 0 0
			      99))))
	     (want 'error)
	     (got  (try str)))
	(test "zlib-inflate" str want got))))


;;
;; zlib-version
;;

(let* ((version (zlib-version))
       (want    #t)
       (got     (string-or-f? version)))
  (test "zlib-version" version want got))


(delete-file filename)
(tests-end)
