;; Tokyo Grain Exchange (TGE) setups.                -*- coding: euc-jp -*-

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tge)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format)  (format)  ;; only for download of zip file
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart gzip)  ;; program-filter-string
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart unzip))

(define-public (tge-symbol? symbol)
  (string-suffix? ".TGE" symbol))

(symbol-timezone! tge-symbol? (_ "Tokyo") timezone-tokyo)
(source-help! tge-symbol?
	      (_ "CONTEXT(manual-node): Tokyo Grain Exchange"))


;;-----------------------------------------------------------------------------
;; weblinks - contract specs
;;
;; eg. http://www.tge.or.jp/japanese/contract/cont_s_corn.shtml
;;     http://www.tge.or.jp/english/contract/cont_s_soym.shtml
;;

(define tge-weblinks-alist
  '(("CO" . "corn")   ;; Corn
    ("SM" . "soym")   ;; Soybean Meal
    ("SB" . "soy")    ;; Soybean
    ("NG" . "gmo")    ;; Non-GMO Soybean
    ("RB" . "azuki")  ;; Azuki
    ("AC" . "ara")    ;; Arabica Coffee
    ("RC" . "rob")    ;; Robusta Coffee
    ("SG" . "raw")    ;; Raw Sugar
    ("SL" . "silk")   ;; Raw Silk
    ("VG" . "vege"))) ;; Vegetables

(weblink-handler! tge-symbol?
  (lambda (symbol)
    (list (_ "TGE _Contract Specifications")
	  (_ "Open web browser at the Tokyo Grain Exchange contract specifications for this commodity")
	  (and-let* ((link (assoc-ref tge-weblinks-alist
				      (chart-symbol-commodity symbol))))
	    (string-append "http://www.tge.or.jp/"
			   (lang-select '(("ja" "japanese")
					  ("en" "english")))
			   "/contract/cont_s_"
			   link
			   ".shtml")))))


;;-----------------------------------------------------------------------------
;; misc

;; return tdate of last trading for SYMBOL, or possibly a day or two after
;; (because public holidays are not accounted for)
;;
(define (tge-symbol-expiry symbol)
  (let ((mdate (chart-symbol-mdate symbol)))
    (if mdate
	(let ((commodity (chart-symbol-commodity symbol)))
	  (cond ((string=? commodity "CO")
		 ;; corn
		 ;; 15th calendar day of the month preceding the delivery month
		 (receive-list (year month day)
		     (mdate->ymd mdate)
		   (ymd->tdate year (1- month) 15)))

		((member commodity '("NG" "RB" "SB"))
		 ;; non-gm soybeans, azuki
		 ;; 2 business days prior to delivery day; delivery day
		 ;; is the business day prior to the last business day of
		 ;; the delivery month or 24th for December
		 (receive-list (year month day)
		     (mdate->ymd mdate)
		   (- (if (= 12 month)
			  (ymd->tdate year month 24)
			  (- (mdate->tdate (1+ mdate)) 2))
		      2)))

		((member commodity '("AC" "RC" "SM"))
		 ;; arabica, robusta, soybean meal
		 ;; 10 business days prior to the last business day of
		 ;; the delivery month
		 (- (mdate->tdate (1+ mdate)) 11))

		((string=? commodity "SG")
		 ;; raw sugar
		 ;; last business day two months prior to delivery month
		 (1- (mdate->tdate (1- mdate))))

		(else
		 ;; unknown, lets guess the last day of the contract month,
		 ;; this won't be right but at least won't be early
		 (1- (mdate->tdate (1+ mdate))))))

	;; rolling front month, no expiry
	(+ 100 (tdate-today)))))

;; see table in http://www.tge.or.jp/japanese/price/prisel.js.php
;; Enhance-me: maybe download and parse that page
(define commodity->nav-alist
  '(("CO" . "tou")   ;; corn
    ("SM" . "dai")   ;; soybean meal
    ("SB" . "ipd")   ;; soybeans
    ("NG" . "ngd")   ;; non-gmo soybeans
    ("RB" . "azu")   ;; azuki
    ("AC" . "ara")   ;; arabica coffee
    ("RC" . "rob")   ;; robusta coffee
    ("SG" . "sot")   ;; raw sugar
    ("SL" . "rsl")   ;; raw silk
    ("VG" . "veg"))) ;; vegetables

;; (define (commodity->nav commodity)
;;   (assoc-ref commodity->nav-alist commodity))

(define (nav->commodity nav)
  (or (rassoc-ref commodity->nav-alist nav)
      (error "TGE: unrecognised nav code:" nav)))


;;-----------------------------------------------------------------------------
;; latest quotes
;;
;; This uses the intraday price pages, either all commodities (about 130
;; kbytes) at
;;
;;     http://www.tge.or.jp/english/price/pri_all_01.shtml
;;
;; or just one (about 55 kbytes) at
;;
;;     http://www.tge.or.jp/english/price/pri_sel_01.shtml
;;
;; Unfortunately there's no ETag or Last-Modified on either to save
;; downloading when it hasn't got new session figures.

;; drop leading A, B, C or N marker from STR
(define (strip-ABCN str)
  (if (and (not (string-null? str))
	   (string-any (string-ref str 0) "ABCN"))
      (string-drop str 1)
      str))

;; drop leading ^ or v marker from front of STR like "v 9050"
(define (strip-limit str)
  (if (and (not (string-null? str))
	   (string-any (string-ref str 0) "^v"))
      (string-trim (string-drop str 1))
      str))

;; English: Jan-05
;; Japanese: "05ǯ09"
(define (tge-month->mdate month)
  (let ((m (must-match
	    ;;                         1month      2year    3year     4month
	    (regexp-exec-C #,(regexp "^([A-Za-z]+)-([0-9]+)|([0-9]+)ǯ([0-9]+)$") month))))
    (ymd->mdate (yyy-str->year (or (match:substring m 2)
				   (match:substring m 3)))
		(if (match:start m 2)
		    (Mmm-str->month (match:substring m 1))
		    (string->number(match:substring m 4)))
		1)))

;; return list of latest records
(define (tge-body->latest-list jp headers body)

  (define ret '())

  ;; for the japanese page we're expecting euc-jp (for the english page the
  ;; bits we care about are all ascii so it doesn't matter)
  (if jp
      (let ((charset (html-charset headers body)))
	(or (string-ci=? charset "euc-jp")
	    (error "TGE quote unrecognised HTML charset:" charset))))

  (for-each
   (lambda (m)
     (let* ((commodity  (nav->commodity (match:substring m 1)))
	    (outer-list (html-table-rows-raw body (match:end m)))

	    ;; English:
	    ;; "<b>04/07/2006 17:45&nbspCorn Futures</b>"
	    ;; "<b>09/20/2006 17:55&amp;nbspAzuki (Red Bean) Futures</b>"
	    ;;
	    ;; Japanese:
	    ;; 2005ǯ715 17:45Ȥʪ
	    ;;
	    (m
	     ;;                        2month   3day     4year    5year     6month    7day            8hour:minute   9                     10name     11
	     (must-match
	      (regexp-exec #,(regexp "(([0-9]+)/([0-9]+)/([0-9]+)|([0-9]+)ǯ([0-9]+)([0-9]+))[^0-9]+([0-9]+:[0-9]+)(&nbsp;?|&amp;nbsp|)([^<\r\n]+)( Futures|ʪ)")
			   (first (first outer-list)))))
	    (year       (or (match:substring m 4)   ;; english
			    (match:substring m 5))) ;; japanese
	    (month      (or (match:substring m 2)   ;; english
			    (match:substring m 6))) ;; japanese
	    (day        (or (match:substring m 3)   ;; english
			    (match:substring m 7))) ;; japanese
	    (tdate      (ymd->tdate (yyy-str->year year)
				    (string->number month)
				    (string->number day)))
	    (time       (h:m:s-str->seconds (match:substring m 8)))
	    (name       (charset->utf8 "EUC-JP" (match:substring m 10)))

	    ;; displays as columns, process as rows
	    (row-list   (apply zip (html-table-rows
				    (first (second outer-list)))))

	    (headings   (map! (lambda (str)
				(string-delete str char-set:whitespace))
			      (first row-list))))
       (set! row-list (cdr row-list))

       ;; last row is volume by session (of all months)
       (if (member (first (last row-list)) '("Volume" ""))
	   (set! row-list (drop-right! row-list 1)))

       (cond
	((member headings '(("" ""
			     "죱" "죲" "죳"
			     "죱" "죲" "죳"
			     "" "Ģ"
			     "" "")
			    ("" "PrevClosingPrice"
			     "Morning1st" "Morning2nd" "Morning3rd"
			     "Afternoon1st" "Afternoon2nd" "Afternoon3rd"
			     "Volume" "SettlementPrice"
			     "Changefromprevday" "OpenInterest")))
	 (for-each
	  (lambda (row)
	    (receive-list (month prev m1 m2 m3 a1 a2 a3 volume settle change int)
		row
	      (define mdate (tge-month->mdate month))
	      (set! month (charset->utf8 "EUC-JP" month))

	      ;; on prices: A=Ask B=Bid C=Lottery N=Nominal
	      ;; and after that: ^=Limit up v=Limit down
	      (let* ((seq  (map! strip-ABCN
				 (remove string-null?
					 (list settle a3 a2 a1 m3 m2 m1))))
		     (note (and (not (null? seq))
				(cond ((string-prefix? "^" (first seq))
				       (_ "limit up"))
				      ((string-prefix? "v" (first seq))
				       (_ "limit down"))
				      (else
				       #f)))))
		(set! seq (map! strip-limit seq))

		;; prev also has limit markers (not sure about ABCN)
		(set! prev (strip-limit (strip-ABCN prev)))

		(receive-list (decimals prev change . seq)
		    (apply strings->numbers+decimals prev change seq)

		  (define open  (and (not (null? seq))  (last seq)))
		  (define high  (apply max-maybe seq))
		  (define low   (apply min-maybe seq))
		  (define final (and (not (null? seq))  (first seq)))

		  (set! ret
			(cons (latest-new
			       #:symbol (string-append commodity " "
						       (mdate->MMM_YY-str mdate)
						       ".TGE")
			       #:name       (string-append name " " month)
			       #:last-tdate tdate
			       #:last-time  time
			       #:open       open
			       #:high       high
			       #:low        low
			       #:last       final
			       #:change     change
			       #:volume     volume
			       #:decimals   decimals
			       #:contract-mdate mdate
			       #:note       note
			       #:source     'tge)
			      ret))))))

	  row-list))

	((member headings
		 ;; "Current" during trading, "Closing" afterwards
		 '(("" "Ģ" "" "" "" "" ""
		    "Ģ" "" "")
		   ("" "Ģ" "" "" "" "" ""
		    "Ģ" "" "")
		   ("" "PreviousSettlementPrice"
		    "Open" "High" "Low" "Current" "Volume"
		    "SettlementPrice" "Changefromprevday" "OpenInterest")
		   ("" "PreviousSettlementPrice"
		    "Open" "High" "Low" "Closing" "Volume"
		    "SettlementPrice" "Changefromprevday" "OpenInterest")))
	 (for-each
	  (lambda (row)
	    (receive-list (month prev open high low last volume settle change int)
		row
	      (define mdate (tge-month->mdate month))
	      (set! month (charset->utf8 "EUC-JP" month))

	      ;; on prices: A=Ask B=Bid C=Lottery N=Nominal, maybe
	      (set! open (strip-ABCN open))
	      (set! last (strip-ABCN last))

	      ;; and after that: ^=Limit up v=Limit down
	      (let ((note (cond ((string-prefix? "^" last)
				 (_ "limit up"))
				((string-prefix? "v" last)
				 (_ "limit down"))
				(else
				 #f))))
		(set! open (strip-limit open))
		(set! low  (strip-limit low))
		(set! high (strip-limit high))
		(set! last (strip-limit last))

		;; prev also has limit markers (not sure about ABCN)
		(set! prev (strip-limit (strip-ABCN prev)))

		(set! ret
		      (cons (latest-new
			     #:symbol (string-append commodity " "
						     (mdate->MMM_YY-str mdate)
						     ".TGE")
			     #:name           (string-append name " " month)
			     #:last-tdate     tdate
			     #:last-time      time
			     #:open           open
			     #:high           high
			     #:low            low
			     #:last           last
			     #:change         change
			     #:volume         volume
			     #:contract-mdate mdate
			     #:note           note
			     #:source         'tge)
			    ret)))))
	  row-list))

	(else
	 (error "TGE: unrecognised data headings:" headings)))))

   (must-match (list-matches-C #,(regexp "<a name=\"nav_([a-z]+)\"></a>")
			       body)))

  (latest-list-add-front-months ret))

;; return #t to use japanese quotes pages, #f to use english
(define (lang-jp?)
  (lang-select '(("en" #f)
		 ("ja" #t #:charset "EUC-JP"))))

(define (tge-latest-get symbol-list extra-list proc)
  (let* ((jp   (lang-jp?))
	 (lang (if (lang-jp?) "japanese" "english")))

    ;; If all of SYMBOL-LIST is for the same commodity, then use the single
    ;; commodity page.
    ;;
    (if (apply equal? (map chart-symbol-commodity symbol-list))
	(let ((commodity (chart-symbol-commodity (first symbol-list))))
	  (receive (headers body)
	      (http-request (string-append "http://www.tge.or.jp/" lang
					   "/price/pri_sel_01.shtml")
			    #:method  "POST"
			    #:data    (string-append
				       "LANGUAGE=" (if jp "JP" "EN")
				       "&COMM_ID=" commodity
				       "&DATE_NUM=1&FTYPE=IT&SUBMIT_NUM=1")
			    #:want-ok #t)

	    (if (string-contains body "Not Found Data")
		(proc (map latest-new-unknown symbol-list))

		(proc (tge-body->latest-list jp headers body)))))

	(receive (headers body)
	    (http-request (string-append "http://www.tge.or.jp/" lang
					 "/price/pri_all_01.shtml")
			  #:want-ok #t)
	  (proc (tge-body->latest-list jp headers body))))))



;; return (ADATE TIME) of prospective quote
;; trading starts at 9:00 (corn) and final session starts at 15:45 (sugar),
;; not sure how long that final one lasts, guess until 16:00
;;
(define (tge-quote-adate-time symbol)
  (tm->adate-time-within (localtime (current-time)
				    (timezone-tokyo))
			 #,(hms->seconds 9 0 0) #,(hms->seconds 16 0 0)))

(latest-handler! #:selector   tge-symbol?  
		 #:handler    tge-latest-get
		 #:adate-time tge-quote-adate-time)


;;-----------------------------------------------------------------------------
;; download
;;
;; This uses the "download" under the price data page
;;
;;     http://www.tge.or.jp/english/price/pri_sel_01.shtml
;;
;; There's two files for each commodity,
;;
;;     http://www.tge.or.jp/data/down_load/co01.csv
;;     http://www.tge.or.jp/data/down_load/co01040610050609.zip
;;
;; The csv is only about 600 bytes, it has just the most recent day and is
;; used when that's all that's needed to update (which will be the case if
;; you're updating every day).
;;
;; The zip is about 35kbytes and has the same most recent day as the csv,
;; plus the past 12 months.  The filename is a date range
;; (XX01yymmddyymmdd.zip), but there's only one file, you can't get an
;; arbitrary range.
;;
;; The zip filename is formed from today's date, or a day or two earlier,
;; but if we don't hit it that way there's a fallback to fetch and parse the
;; actual download page.  The pattern in the filename is pretty clear, so
;; perhaps that fallback is unnecessary.
;;
;; "gunzip" can't be used for archive extraction, only the "unzip" program,
;; because the archive has more than one file (a separate file for each
;; calendar month).
;;
;; There's an ETag / Last-Modified on the year file, but it's hardly needed
;; since the dates are in the name.


(define (tge-process-zip symbol-list body)
  (tge-process-csv symbol-list #f #f (program-filter-string "unzip -p" body)))

(define (tge-process-csv symbol-list url headers body)
  (define row-list (map! (lambda (line)
			   (string-split line #\,))
			 (string-split-lines body)))

  (let ((want   '("yr_mo_dy" "contract" "contract_month"
		  "contract_price_m1" "contract_price_m2" "contract_price_m3"
		  "contract_price_a1" "contract_price_a2" "contract_price_a3"
		  "sett_price" "volume" "open_int" "net_position"))
	(headings (first row-list)))
    (or (equal? headings want)
	(error "TGE data: unrecognised headings:" headings))
    (set! row-list (delete! want row-list)))

  (download-process
   #:module        (_ "TGE")
   #:symbol-list   symbol-list
   #:url           url
   #:headers       headers
   #:currency      "JPY"
   #:expiry-proc   tge-symbol-expiry
   #:last-download #t
   #:row-list
   (map-rows!
    (lambda (tdate commodity mdate m1 m2 m3 a1 a2 a3 settle volume openint net)
      (list #:tdate         (yyyymmdd-str->tdate tdate)
	    #:mdate         (yyyymm-str->mdate mdate)
	    #:commodity     commodity
	    #:sessions      (list m1 m2 m3 a1 a2 a3 settle)
	    #:volume        volume
	    #:open-interest openint))
    row-list)))

;; return tdate for available download data
;; the .csv downloads have been seen with a Last-Modified headers
;;     Wed, 24 Aug 2005 01:00:20 GMT      == 10am tokyo
;;     Thu, 01 Sep 2005 22:33:05 GMT      ==  7am tokyo
;; so try at 10:05am tokyo
;;
(define (tge-available-tdate)
  (1- (tdate-today-after 10 5 (timezone-tokyo))))

;; return the html link for COMMODITY zip file download, by  or #f if unknown
;; commodity
(define (tge-download-page commodity)
  (receive (headers body)
      (http-request
       "http://www.tge.or.jp/english/price/pri_sel_01.shtml"
       #:want-ok #t
       #:method "POST"
       #:data (simple-format #f "LANGUAGE=EN&COMM_ID=~a&DATE_NUM=4&YYYY=&MM=&DD=&SUBMIT_NUM=1" commodity))

    (let ((m (must-match (string-match "alt=\"Download\"[^\r\n]*location\\.href='([^']+\\.zip)'" body))))
      (or (string-prefix-ci? commodity (basename (match:substring m 1)))
	  (error "TGE: oops, wrong commodity in download page"))
      (http-url-relative
       "http://www.tge.or.jp/english/price/pri_sel_01.shtml"
       (match:substring m 1)))))

;; return list (YEAR MONTH DAY) of values in usual range
(define (ymd-normalize year month day)
  (adate->ymd (ymd->adate year month day)))

;; eg. http://www.tge.or.jp/data/down_load/co01040610050609.zip
;;     for Jun/10/2004 - Jun/09/2005
(define (tge-download-tdate-url commodity tdate)
  (receive-list (end-year end-month end-day)
      (tdate->ymd tdate)
    ;; a year ago, plus one day, and normalize that +1 on the day in case it
    ;; rolls over the month
    (receive-list (start-year start-month start-day)
	(ymd-normalize (1- end-year) end-month (1+ end-day))
      (format #f "http://www.tge.or.jp/data/down_load/~a01~2,'0d~2,'0d~2,'0d~2,'0d~2,'0d~2,'0d.zip"
	      (string-downcase commodity)
	      (modulo start-year 100) start-month start-day
	      (modulo end-year   100) end-month   end-day))))

(define (tge-csv-tdate body)
  (yyyymmdd-str->tdate (string-take-to-char (string-last-line body) #\,)))

;; do a download of the last day .csv file for SYMBOL-LIST, if that would
;; update all those
;;
;; return #t if updated successful or #f if not (should use zip instead)
;; 
(define (tge-download-csv symbol-list) ;; of one commodity
  (define avail-tdate (tge-available-tdate))
  ;; use csv if all of SYMBOL-LIST just wanting AVAIL-TDATE
  (and (every (lambda (symbol)
		(= avail-tdate (download-start-tdate symbol)))
	      symbol-list)
       (let ((url (simple-format #f
				 "http://www.tge.or.jp/data/down_load/~a01.csv"
				 (string-downcase
				  (chart-symbol-commodity
				   (first symbol-list))))))
	 (download-status (_ "TGE") (_ "data") (basename url))
	 (receive (headers body)
	     (apply http-request url #:want-ok #t
		    (download-tag-args url symbol-list))
	   (if (http-headers-ok? headers)
	       (let ((got-tdate (tge-csv-tdate body)))
		 (cond ((= got-tdate avail-tdate)
			;; got the expected data, process it
			(tge-process-csv
			 ;; also processs for any others covered by this data
			 (download-also symbol-list
					#:selector    tge-symbol?
					#:commodity   #t
					#:start-tdate avail-tdate
					#:end-tdate   avail-tdate
					#:expiry-proc tge-symbol-expiry)
			 url headers body)
			#t)
		       ((< got-tdate avail-tdate)
			;; got something older, there's no new data
			#t)
		       (else
			#f)))
	       ;; 304 Not Modified, no new data (yet)
	       #t)))))

(define (tge-download-zip symbol-list)
  (define avail-tdate (tge-available-tdate))

  ;; nothing to download for symbols with data already up to what's
  ;; available, or already up to contract expiry
  (set! symbol-list
	(remove (lambda (symbol)
		  (let ((start-tdate (download-start-tdate symbol)))
		    (or (> start-tdate avail-tdate)
			(> start-tdate (tge-symbol-expiry symbol)))))
		symbol-list))

  (for-each
   (lambda (symbol-list) ;; of each commodity
     ;; also processs for any others covered by this data
     (set! symbol-list
	   (download-also symbol-list
			  #:selector    tge-symbol?
			  #:commodity   #t
			  #:expiry-proc tge-symbol-expiry))

     (let ((commodity   (chart-symbol-commodity (first symbol-list))))

       ;; return true if good
       (define (attempt url)
	 (download-status (_ "TGE") (_ "data") (basename url))
	 (receive (headers body)
	     (http-request url)
	   (and (http-headers-ok? headers)
		(begin
		  (tge-process-zip symbol-list body)
		  #t))))

       ;; try 0, -1, -2 direct filenames, then finally download page
       ;; (allowing for new COMM_NUM values too)
       ;;
       (or (attempt (tge-download-tdate-url commodity avail-tdate))
	   (attempt (tge-download-tdate-url commodity (- avail-tdate 1)))
	   (attempt (tge-download-tdate-url commodity (- avail-tdate 2)))
	   (attempt (tge-download-page commodity)))))

   (partition-equal chart-symbol-commodity symbol-list)))

(define (tge-download symbol-list)
  ;; don't download data at all unless have unzip program
  (if (unzip-version)
      (or (tge-download-csv symbol-list)
	  (tge-download-zip symbol-list)))
  (download-name-from-latest (_ "TGE") symbol-list))

(define (tge-backto want-tdate symbol-list)
  (download-message
   (string-append
    (_ "TGE") ": "
    (_ "no further historical data available"))))

(download-handler! #:selector   tge-symbol?
		   #:update     tge-download
		   #:avail-proc tge-available-tdate
		   #:backto     tge-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any tge-symbol? symbol-list)
			      (download-now-commodity-all-months))))

(download-now-handler! (lambda (symbol-list)
			 (and (any tge-symbol? symbol-list)
			      (download-now-need-unzip))))
