;; Ticker widget.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ticker)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart gui-common)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart symlists))


(define-public ticker-speed          4.5)      ;; in "M"s per second
(define-public ticker-scroll-period  300)      ;; milliseconds between moves
(define-public ticker-spacing        3)        ;; in "M"s
(define-public ticker-refresh-period (* 5 60)) ;; in seconds, 5 minutes

;; "down" is a brighter red, for contrast against a black background
;;
(define chart-up-colour         "green")
(define chart-down-colour       "#FF7070")
(define chart-inprogress-colour "light blue")


;;-----------------------------------------------------------------------------
;; misc

(define second-of-first cadar)

;; Start or stop a gtk timeout callback.
;; ID is either a currently running id (an integer) or #f
;; PERIOD is the desired timeout in milliseconds
;; HANDLER is the thunk for `gtk-timeout-add' to call when it fires
;;
;; CONDITION is true or false according to whether we want the timer to be
;; running.  If true and ID is #f then a new timer is started.  If #f and ID
;; is true then that running timer is stopped.
;;
;; The return is an id or #f.
;;
(define (c-gtk-timeout-start-or-stop id period handler condition)
  (if condition
      (if (not id)
	  (set! id (gtk-timeout-add period handler)))
      (if id
	  (begin
	    (gtk-timeout-remove id)
	    (set! id #f))))
  id)

;; return true if there's any intersection, ie. common elements, between
;; LST1 and LST2
;; element comparisons are made with the given `=' function
;;
(define (lset-intersection? = lst1 lst2)
  (any (lambda (elem1)
	 (member elem1 lst2 =))
       lst1))


;;-----------------------------------------------------------------------------
;; displayed ticker text

;; the same as `latest-form-price', but always include a + or - sign
;; (normally the + is omitted)
(define (latest-form-price-plusminus latest value)
  (string-append (if (> value 0) "+" "")
		 (latest-form-price latest value)))

;; return list (STR COLOUR) which is how to display SYMBOL
;;
(define (ticker-format symbol)
  (let* ((latest (latest-get symbol))
	 (last   (latest-last latest))
	 (change (latest-change latest))
	 (colour "white"))

    (cond ((not change)
	   (set! change ""))
	  ((not (eqv? (latest-last-adate latest)
		      (adate-today (symbol-timezone symbol))))
	   (set! change (latest-form-date/time latest)))
	  ((= 0 change)
	   (set! change (_ "unch")))
	  (else
	   (set! colour (if (positive? change)
			    chart-up-colour chart-down-colour))
	   (set! change (latest-form-price-plusminus latest change))))

    (if (latest-in-progress? symbol)
	(set! colour chart-inprogress-colour))

    (list (string-append
	   (latest-symbol latest)
	   " "
	   (or (and last (latest-form-price latest last))
	       (latest-note latest)
	       (_ "no data"))
	   " "
	   change)
	  colour)))


;;-----------------------------------------------------------------------------
;; ticker widget

(define rec-type (make-record-type "ChartTicker"
				   '(hash smarker offset drawn run menu)))

(define rec-new      (record-constructor rec-type))
(define rec-hash     (rec-accessor rec-type 'hash))
(define rec-smarker  (rec-accessor rec-type 'smarker))
(define rec-offset   (rec-accessor rec-type 'offset))
(define rec-offset!  (rec-modifier rec-type 'offset))

;; list of (SYMBOL POS STR-WIDTH NEXT-WIDTH)
(define rec-drawn    (rec-accessor rec-type 'drawn))
(define rec-drawn!   (rec-modifier rec-type 'drawn))

(define rec-run      (rec-accessor rec-type 'run))
(define rec-run!     (rec-modifier rec-type 'run))
(define rec-menu     (rec-accessor rec-type 'menu))
(define rec-menu!    (rec-modifier rec-type 'menu))

(gtk-class-new 'GtkDrawingArea "ChartTicker")
(c-gtk-rc-chart-style "ChartTicker")

(define*-public (ticker-new #:optional (menu-additions noop))
  (let* ((ticker  (gtk-widget-new 'ChartTicker))
	 (hash    (make-weak-key-hash-table 37))
	 (smarker (smarker-new))
	 (rec     (rec-new hash
			   smarker
			   0     ;; offset
			   '()   ;; drawn
			   #t    ;; run
			   menu-additions)) ;; menu
	 (scroll-timer-id  #f)
	 (refresh-timer-id #f)
	 (visibility-state 'unobscured) ;; initial map
	 (dragging         #f))

    (define (draw-this)
      (draw ticker))

    ;; timers stopped either by request or when hidden, or when empty
    ;; stops both scrolling and refreshes
    (define (want-timers?)
      (and (rec-run rec)
	   (gtk-widget-mapped ticker)
	   (not (eq? visibility-state 'fully-obscured))
	   (not (null? (symlist-list (smarker-symlist smarker))))))

    (define (want-scroll?)
      (and (not dragging)
	   (want-timers?)))

    (define (review-timers)
      (let ((run (want-timers?)))
	(set! scroll-timer-id
	      (c-gtk-timeout-start-or-stop
	       scroll-timer-id ticker-scroll-period
	       scroll-timer-handler run))
	(set! refresh-timer-id
	      (c-gtk-timeout-start-or-stop
	       refresh-timer-id (* 1000 ticker-refresh-period)
	       refresh-timer-handler run))))

    (define (scroll-timer-handler)
      (set! scroll-timer-id #f)
      (if (want-scroll?)
	  (ticker-scroll-chars ticker
			       (* ticker-speed               ;; Ms per second
				  ticker-scroll-period 0.001))) ;; time in secs
      (review-timers))

    (define (refresh-timer-handler)
      (set! refresh-timer-id #f)
      (if (want-timers?)
	  (let ((lst (symlist-list (smarker-symlist smarker)))
		(pos (smarker-pos smarker)))
	    ;; takes a moment to crunch desired requests
	    (c-gtk-widget-busy ticker)
	    ;; currently viewed symbols first in request, to get them
	    ;; retrieved first
	    (latest-request-explicit (append (drop lst pos)
					     (take lst pos)))))
      (review-timers))

    (c-gtk-widget-set-usize-units ticker -1 '(height . 1.2))

    (set! (c-prop ticker) rec)
    (set-object-property! ticker 'draw-this draw-this)
    (smarker-goto-first-of-symlist smarker favourites-symlist)

    (gtk-signal-connect ticker "expose_event"
      (lambda (event)
	(if (= 0 (gdk-event-count event)) ;; final expose
	    (begin
	      (rec-drawn! rec '())
	      (c-gtk-idle-once draw-this)))))

    (gtk-signal-connect ticker "visibility_notify_event"
      (lambda (event)
	(set! visibility-state (gdk-event-visibility-state event))
	(review-timers)))

    ;; freshen prices when initially displayed, but only the 2 minute rule
    (gtk-signal-connect ticker "show"
      (lambda ()
	(review-timers)
	(c-idle gtk-priority-low)
	(latest-request-symbols (symlist-list (smarker-symlist smarker)))))

    ;; menu popup
    (gtk-signal-connect ticker "button_press_event"
      (lambda (event)
	(if (= 3 (gdk-event-button event))
	    (ticker-menu-popup ticker))))

    (notify-connect 'symlist-update
      (lambda (symlist-key-list)
	(if (member (symlist-key (smarker-symlist smarker)) symlist-key-list)
	    ;; list may have gone from empty to non-empty
	    (begin
	      (c-gtk-idle-once draw-this)
	      (review-timers)))))

    (notify-connect 'latest-update
      (lambda (symbol-list)
	;; clear cached strings for changed symbols
	(for-each (lambda (symbol)
		    (hash-remove! hash symbol))
		  symbol-list)
	;; if anything currently drawn has changed then must redraw
	(if (lset-intersection? equal?
				(map first (rec-drawn rec)) symbol-list)
	    (c-gtk-idle-once draw-this))))

    ;; use pointer-motion-hint-mask as usual to avoid getting many movement
    ;; events, instead make a server round-trip when ready
    ;;
    (c-gtk-widget-add-events ticker 'ticker-base
			     '(button-press-mask
			       button-motion-mask
			       pointer-motion-hint-mask
			       button-release-mask
			       visibility-notify-mask))
    (let ((base-x #f))
      (define (drag event)
	(if dragging
	    (let ((x (second (gdk-window-get-pointer
			      (gdk-event-window event)))))
	      (and base-x
		   (ticker-scroll-pixels ticker (- base-x x)))
	      (set! base-x x))))
      (gtk-signal-connect ticker "button_press_event"
	(lambda (event)
	  (if (= 1 (gdk-event-button event))
	      (begin
		(set! base-x (round-exact (gdk-event-x event)))
		(set! dragging #t)))))
      (gtk-signal-connect ticker "motion_notify_event" drag)
      (gtk-signal-connect ticker "button_release_event"
	(lambda (event)
	  (if (= 1 (gdk-event-button event))
	      (begin
		(drag event)
		(set! dragging #f))))))

    (review-timers)
    ticker))

;; set the run state of TICKER, true to scroll, #f to stop
(define-public (ticker-run! ticker state)
  (let ((rec (c-prop ticker)))
    (rec-run! rec state)))

(define-public (ticker-run ticker)
  (rec-run (c-prop ticker)))

;; return the current symlist being displayed
(define-public (ticker-symlist ticker)
  (let ((rec (c-prop ticker)))
    (smarker-symlist (rec-smarker rec))))

(define-public (ticker-symlist! ticker symlist)
  (let* ((rec     (c-prop ticker))
	 (smarker (rec-smarker rec)))
    (if (not (eq? symlist (smarker-symlist smarker)))
	(begin
	  (c-gtk-widget-busy ticker)
	  (smarker-goto-first-of-symlist smarker symlist)
	  (c-gtk-idle-once (object-property ticker 'draw-this))))))

;; return list (STR STR-WIDTH GC)
;;
(define (form ticker rec symbol)
  (let* ((hash (rec-hash rec))
	 (info (hash-ref hash symbol))
	 (win  (gtk-widget-window ticker)))
    (if (not info)
	(receive-list (str colour)
	    (ticker-format symbol)
	  (let* ((gc        (c-gdk-gc-get win #:foreground colour))
		 (layout    (primitive-property-ref c-layout-prop ticker))
		 (str-width (c-pango-layout-width layout str)))
	    (set! info (list str str-width gc))
	    (hash-set! hash symbol info))))
    info))

(define (draw ticker)
  (and-let* ((win (gtk-widget-window ticker))) ;; only draw when realized
    (let* ((rec         (c-prop ticker))
	   (smarker     (rec-smarker rec))
	   (symlist     (smarker-symlist smarker))
	   (lst         (symlist-list symlist))
	   (layout      (primitive-property-ref c-layout-prop ticker))
	   (height      (c-gdk-drawable-get-height win))
	   (font-height (c-pango-layout-line-height layout))
	   (y           (quotient (- height font-height) 2))
	   (em          (c-pango-layout-em layout)))
      (if (null? lst)
	  (let* ((style     (gtk-widget-get-style ticker))
		 (gc        (gtk-style-text-gc style 'normal))
		 (str       (format #f (_ "(~a list empty)")
				    (symlist-name symlist)))
		 (str-width (c-pango-layout-width layout str)))
	    (gdk-window-clear win)
	    (gdk-draw-layout win gc em y layout)
	    (rec-drawn! rec (list (list #f em str-width str-width))))

	  (let* ((width          (c-gdk-drawable-get-width win))
		 (pos            (rec-offset rec))
		 (old-drawn      (rec-drawn rec))
		 (spacing-pixels (ceiling-exact (* ticker-spacing em)))
		 (drawn          '()))
	    (set! lst (drop (apply circular-list lst) (smarker-pos smarker)))

	    ;; The strategy here is to work across the window clearing old
	    ;; text and drawing new.  Doing the clear progressively helps
	    ;; the scrolling look smooth on a slow video.
	    (do ()
		((>= pos width))
	      (let ((symbol (first lst)))
		(receive-list (str str-width gc)
		    (form ticker rec symbol)
		  (let ((next-width (+ str-width spacing-pixels))
			(next-pos   (+ pos str-width spacing-pixels)))

		    ;; clear any OLD-DRAWN up to NEXT-POS
		    (do ()
			((or (null? old-drawn)
			     (> (second-of-first old-drawn) next-pos)))
		      (receive-list (symbol x str-width next-width)
			  (first old-drawn)
			(gdk-window-clear-area win x y str-width font-height)
			(set! old-drawn (cdr old-drawn))))

		    (pango-layout-set-text layout str)
		    (gdk-draw-layout win gc pos y layout)
		    (set! drawn (cons (list symbol pos str-width next-width)
				      drawn))
		    (set! pos next-pos)
		    (set! lst (cdr lst))))))
	    (rec-drawn! rec (reverse! drawn)))))))

(define-public (ticker-scroll-chars ticker chars)
  (ticker-scroll-pixels ticker 
			(ceiling-exact (* chars (c-gtk-widget-em ticker)))))

(define-public (ticker-scroll-pixels ticker pixels)
  (let* ((rec            (c-prop ticker))
	 (offset         (- (rec-offset rec) pixels))
	 (smarker        (rec-smarker rec))
	 (em             (c-gtk-widget-em ticker))
	 (spacing-pixels (ceiling-exact (* ticker-spacing em))))

    (if (positive? pixels)

	;; scrolling further along
	(let more ()
	  (and-let* ((symbol (or (smarker-entry smarker)
				 (smarker-next-circular smarker))))
	    (receive-list (str str-width gc)
		(form ticker rec symbol)
	      (if (not (positive? (+ offset str-width)))
		  (begin
		    (set! offset (+ offset str-width spacing-pixels))
		    (smarker-next-circular smarker)
		    (more))))))

	;; scrolling back
	(let more ()
	  (and-let* ((       (positive? offset))
		     (symbol (smarker-prev-circular smarker)))
	    (receive-list (str str-width gc)
		(form ticker rec symbol)
	      (set! offset (- offset str-width spacing-pixels))
	      (more)))))

    (rec-offset! rec offset))

  (c-gtk-idle-once (object-property ticker 'draw-this)))


;;-----------------------------------------------------------------------------
;; menu

(define (ticker-menu-popup ticker)
  (let* ((rec  (c-prop ticker))
	 (menu (rec-menu rec)))

    ;; created when needed
    (if (not (gtk-menu? menu))
	(let ((menu-additions menu))
	  (set! menu (gtk-menu-new))
	  (rec-menu! rec menu)

	  ;; separator
	  (gtk-menu-shell-append menu (gtk-separator-menu-item-new))

	  ;; refresh
	  (let ((item (gtk-image-menu-item-new-from-stock "gtk-refresh" #f)))
	    (gtk-menu-shell-append menu item)
	    (gtk-signal-connect item "activate"
	      (lambda ()
		(latest-request-explicit (symlist-list
					  (ticker-symlist ticker))))))

	  ;; run
	  (let ((item (gtk-check-menu-item-new-with-label (_ "Run"))))
	    (gtk-check-menu-item-set-state item #t)
	    (gtk-menu-shell-append menu item)
	    (gtk-signal-connect item "activate"
	      (lambda ()
		(ticker-run! ticker (gtk-check-menu-item-active item)))))

	  ;; symbol list
	  (let ((item (gtk-menu-item-new-with-mnemonic (_ "_Symbol list")))
		(submenu (gtk-menu-new))
		(group   #f))
	    (gtk-menu-item-set-submenu item submenu)
	    (gtk-menu-append menu item)
	    (for-each (lambda (symlist)
			(let ((item  (gtk-widget-new 'GtkRadioMenuItem))
			      (label (gtk-label-new (symlist-name symlist))))
			  (if group
			      (gtk-widget-set item #:group group)
			      (set! group item))
			  (gtk-misc-set-alignment label 0 0.5)
			  (gtk-container-add item label)
			  (gtk-menu-shell-append submenu item)
			  (if (eq? symlist (ticker-symlist ticker))
			      (gtk-check-menu-item-set-active item #t))
			  (gtk-signal-connect item "activate"
			    (lambda ()
			      (ticker-symlist! ticker symlist)))))
		      symlist-all-lists)
	    (gtk-widget-show-all submenu))

	  (menu-additions menu)
	  (gtk-widget-show-all menu)))

    (gtk-menu-popup menu #f #f 3 0)))
