;; Trend Intensity Index (TII).

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tii)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart sma))


;; http://www.linnsoft.com/tour/techind/tii.htm
;;     Sample Intel (INTC) 2001/2001, seems different to what get from the
;;     code here.  Maybe not 60/30 parameters?
;;
;; http://working-money.com/Documentation/FEEDbk_docs/Archive/062002/Abstracts_new/Pee/pee.html
;;     Start of article, rest for sale.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/062002/TradersTips/TradersTips.html
;;     TASC Trader's Tips June 2002.
;;     Linnsoft sample MSFT from Sep 2001.
;;


;; Return a procedure which calculates the trend intensity index over an
;; accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the TII up to (and including) that point, or #f if not enough values
;; yet.
;;
;; The window can be primed by calling with MACOUNT-1 or DEVCOUNT-1,
;; whichever is larger, many points preceding the first desired.
;; `tii-warmup-count' below can make that calculation.
;;
(define-public (tii-calc-proc ma-count dev-count)
  (let ((ma-proc    (sma-calc-proc ma-count))
	(q          (make-q))
	(points     0)
	(min-points (ceiling-exact (* indicator-minimum-fraction dev-count))))

    (lambda (x)
      (if (>= points dev-count)
	  (deq! q)                   ;; full, drop old point
	  (set! points (1+ points))) ;; not full, POINTS increases

      ;; add new point
      (enq! q x)

      (and-let* ((ma (ma-proc x))
		 (   (>= points min-points)))
	(let* ((up    0)
	       (total 0))
	  (for-each (lambda (x)
		      (let ((dev (- x ma)))
			(if (positive? dev)
			    (set! up (+ up dev)))
			(set! total (+ total (abs dev)))))
		    (q-list q))
	  (if (zero? total)
	      50
	      (* 100.0 (/ up total))))))))

(define (tii-warmup-count ma-count dev-count)
  (1- (max ma-count dev-count)))

(define (tii-fill series new-array lo hi)
  (receive-list (ma-count dev-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (warmup (tii-warmup-count ma-count dev-count))
	   (array  (series-array-closes parent lo hi))
	   (proc   (tii-calc-proc ma-count dev-count)))

      ;; warmup
      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i))
			i))))))

(define-public (tii-series series ma-count dev-count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f (_ "TII ~a, ~a")
					  ma-count dev-count))
    (series-datatype!  new-series tii-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list ma-count dev-count))
    (series-require!   new-series
		       (indicator-require-parts
			tii-fill (tii-warmup-count ma-count dev-count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public tii-datatype
  (datatype-new!
   #:key     'tii
   #:name    (_ "Trend Intensity Index")
   #:units   'percentage
   #:minimum 0
   #:maximum 100
   #:hlines  '(20 50 80)
   #:manual  (_ "CONTEXT(manual-node): Trend Intensity Index")
   #:proc    tii-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'tii-ma-days
	  #:name    (_ "MA Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 60)
	 (datatype-param-new
	  #:key     'tii-days
	  #:name    (_ "Dev Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 30))))
