;; Triangular moving average (TMA).

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tma)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.linnsoft.com/tour/techind/movAvg.htm
;;     Formula, and sample TMA N=20 on Nasdaq 100 (symbol QQQ, yahoo now in
;;     ^IXIC) from 2001.
;;


;; return a list of TMA weighting factors, like '(1 2 3 2 1)
(define (tma-factors count)
  (let ((c-count (quotient (1+ count) 2)))
    (append! (iota (quotient count 2) 1)
	     (iota c-count c-count -1))))


;; Return a procedure which calculates a triangular moving average over an
;; accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the TMA up to (and including) that value, or #f if not enough points
;; yet.
;;
;; (Thought about and even attempted taking an initial data values
;; parameter, but it got too messy to be worthwhile.)
;;

(define-public (tma-calc-proc count)

  ;; FACTORS is the weight factors, like '(1 2 3 2 1), stepped along to
  ;;         build the divisor while the window isn't yet full
  ;;
  ;; DATA is a circular list of the accumulated window
  ;;
  ;; TOTAL is the sum of the DATA values with weighting factors applied
  ;;
  ;; STEP is how much to add to TOTAL to adjust for moving the DATA one
  ;;      place along
  ;;
  ;; F-DATA and C-DATA are positions in DATA where the value stops adding
  ;;        and starts subtracting, as it reaches its peak weighting then
  ;;        starts decreasing
  ;; 
  (let* ((c-count (quotient (1+ count) 2))  ;; ceil(N/2)
	 (factors (tma-factors count))
	 (div-min (ceiling-exact (* indicator-minimum-fraction
				    (apply + factors))))
	 (div     0)
	 (total   0)
	 (step    0)
	 (data    (make-circular-list count #f))
	 (f-data  (list-cdr-ref data c-count)) ;; measured from other end
	 (c-data  (if (odd? count)
		      f-data
		      (cdr f-data))))

    (lambda  (x)
      ;; shift up past points, and add in new point
      (set! total (+ total step x))
      (set! step (+ step x))

      (if (car data)
	  ;; window full, drop oldest point out of stepping
	  (set! step (+ step (car data)))

	  ;; window not yet full, increase divisor for new point
	  (begin
	    (set! div (+ div (first factors)))
	    (set! factors (cdr factors))))

      ;; replace oldest point with new point
      (set-car! data x)
      (set! data (cdr data))

      ;; at floor(N/2) point stops adding
      (if (car f-data)
	  (set! step (- step (car f-data))))
      (set! f-data (cdr f-data))

      ;; at ceil(N/2) point starts subtracting
      (if (car c-data)
	  (set! step (- step (car c-data))))
      (set! c-data (cdr c-data))

      (and (>= div div-min)
	   (exact->inexact (/ total div))))))


(define (tma-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (tma    (tma-calc-proc count)))

    (for-each tma (series-prev-n-closes parent #:before lo #:count (1- count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (tma this) i)))))

(define-public (tma-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "TMA") count))
    (series-datatype! new-series tma-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts tma-fill (1- count)))
    (series-array!    new-series #f)

    new-series))

(define-public tma-datatype
  (datatype-new!
   #:key      'tma
   #:name     (_ "TMA - Triangular MA")
   #:priority -10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Triangular Moving Average")
   #:proc     tma-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'tma-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
