;; Tokyo Commodities Exchange (TOCOM) setups.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tocom)
  #:use-module (ice-9 and-let-star)
  #:autoload   (ice-9 format) (format)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart unzip))

(define-public (tocom-symbol? symbol)
  (string-suffix? ".TOCOM" symbol))

(symbol-timezone! tocom-symbol? (_ "Tokyo") timezone-tokyo)
(source-help! tocom-symbol?
	      (_ "CONTEXT(manual-node): Tokyo Commodities Exchange"))


;;-----------------------------------------------------------------------------
;; weblink - contract specs

;; eg. http://www.tocom.or.jp/guide/youkou/gold/index.html
;;     http://www.tocom.or.jp/jp/guide/youkou/crude_oil/index.html
;;
(weblink-handler! tocom-symbol?
  (lambda (symbol)
    (list (_ "TOCOM _Contract Specifications")
	  (_ "Open web browser at the Tokyo Commodities Exchange contract specifications for this commodity")
	  (string-append "http://www.tocom.or.jp/"
			 (lang-select '(("en" "")
					("ja" "jp/")))
			 "guide/youkou/"
			 (space->underscore
			  (string-downcase (chart-symbol-commodity symbol)))
			 "/index.html"))))

(define (space->underscore str)
  (string-map (lambda (c)
		(if (eqv? c #\space) #\_ c))
	      str))


;;-----------------------------------------------------------------------------
;; misc

;; return mdate for STR like "0504"
(define (yymm-str->mdate str)
  (ymd->mdate (yy->year (string->number-err (substring str 0 2)))
	      (string->number-err (substring str 2 4))
	      1))

;; return STR with any trailing zero decimal places stripped
;; eg. "100.00" -> "100."
;;     "100.50" -> "100.5"
;;     "100"    -> "100"
(define (string-trim-zero-decimals str)
  (if (string-any #\. str)
      (string-trim-right str #\0)
      str))


;;-----------------------------------------------------------------------------
;; csv misc

;; return tdate of last trading for SYMBOL, or possibly the day after
;;
(define (tocom-symbol-expiry symbol)
  (let ((mdate (chart-symbol-mdate symbol)))
    (cond ((not mdate)
	   ;; rolling front month, no expiry
	   (+ 100 (tdate-today)))

	  ((string-prefix? "Rubber" symbol)
	   ;; Rubber: Delivery is the last day of the month, except 28th for
	   ;; December, and if that's not a full business day then the next
	   ;; business day.  Last trading is the 4th business day before
	   ;; delivery.
	   (receive-list (year month day)
	       (mdate->ymd mdate)
	     ;; use -3 not -4 since don't know if the 28th/31st will be a
	     ;; public holiday
	     (- (if (= 12 month)
		    (ymd->tdate-ceil year 12 28)    ;; 28th or next weekday
		    (1- (mdate->tdate (1+ mdate)))) ;; last of mdate
		4)))

	  ((or (string-prefix? "Aluminum"  symbol)
	       (string-prefix? "Gold"      symbol)
	       (string-prefix? "Palladium" symbol)
	       (string-prefix? "Platinum"  symbol)
	       (string-prefix? "Silver"    symbol))
	   ;; Aluminium, Gold, Palladium, Platinum, Silver: Delivery is the
	   ;; last day of the month, except 24th for December, and if that's
	   ;; not a full business day then the next business day.  Last
	   ;; trading is the 3rd business day before delivery.
	   (receive-list (year month day)
	       (mdate->ymd mdate)
	     ;; use -2 not -3 since don't know if the 28th/31st will be a
	     ;; public holiday
	     (- (if (= 12 month)
		    (ymd->tdate-ceil year 12 24)    ;; 24th or next weekday
		    (1- (mdate->tdate (1+ mdate)))) ;; last of mdate
		3)))

	  ((or (string-prefix? "Gas"      symbol)
	       (string-prefix? "Kerosene" symbol))
	   ;; Gas Oil, Gasoline, Kerosene: 25th of the month, or next
	   ;; trading day.
	   (receive-list (year month day)
	       (mdate->ymd mdate)
	     ;; use 1+ since we don't track holidays and aren't sure if the
	     ;; 25th will be a trading day
	     (+ 0 (ymd->tdate year month 25))))

	  (else ;; (string-prefix? "Crude" symbol)
	   ;; Crude: Last business day of the month is end of normal
	   ;; trading, then first business day of next month all liquidate
	   ;; on the open at settlement price.  Also use this rule for
	   ;; anything unknown.
	   (mdate->tdate (1+ mdate))))))


;;-----------------------------------------------------------------------------
;; csv - old format (to Dec 2002)

(define (tocom-old-code->commodity code)
  (assoc-ref '(("01" . "Gold")
	       ("02" . "Silver")
	       ("03" . "Platinum")
	       ("04" . "Palladium")
	       ("05" . "Aluminum")
	       ("06" . "Gasoline")
	       ("07" . "Kerosene")
	       ("08" . "Crude Oil")
	       ("60" . "Cotton Yarn")
	       ("65" . "Wool Yarn")
	       ("80" . "Rubber"))
	     code))

(define (tocom-process-old-csv symbol-list body-mdate url headers body)
  (define row-list (map! (lambda (line)
			   (string-split line #\,))
			 (string-split-lines body)))

  (download-process
   #:module      (_ "TOCOM")
   #:symbol-list symbol-list
   #:lo          (and body-mdate (mdate->tdate body-mdate))
   #:hi          (and body-mdate (mdate->last-tdate body-mdate))
   #:url         url
   #:headers     headers
   #:expiry-proc tocom-symbol-expiry
   #:row-list
   (map-rows!
    (lambda (date code month open high low close settle
		  market volume openint)
      (list #:tdate         (yyyymmdd-str->tdate date)
	    #:mdate         (yymm-str->mdate month)
	    #:commodity     (tocom-old-code->commodity code)
	    #:open          open
	    #:high  	    high
	    #:low   	    low
	    #:close 	    settle
	    #:volume        volume
	    #:open-interest openint))
    row-list)))


;;-----------------------------------------------------------------------------
;; csv - new format (post Dec 2002)

(define (tocom-new-code->commodity code)
  (assoc-ref '(("11" . "Gold")
	       ("12" . "Silver")
	       ("13" . "Platinum")
	       ("14" . "Palladium")
	       ("21" . "Aluminum")
	       ("31" . "Gasoline")
	       ("32" . "Kerosene")
	       ("33" . "Crude Oil")
	       ("34" . "Gas Oil")
	       ("81" . "Rubber"))
	     code))

(define (tocom-process-new-csv symbol-list body-mdate url headers body)
  (define row-list (map! (lambda (line)
			   (string-split line #\,))
			 (string-split-lines body)))
  (download-process
   #:module      (_ "TOCOM")
   #:symbol-list symbol-list
   #:lo          (and body-mdate (mdate->tdate body-mdate))
   #:hi          (and body-mdate (mdate->last-tdate body-mdate))
   #:url         url
   #:headers     headers
   #:expiry-proc tocom-symbol-expiry
   #:row-list
   (map-rows!
    (lambda (date type code month strike
		  open high low close settle
		  market volume openint openintvalue)
      (and (string=? type "1") ;; futures
	   (list #:tdate         (yyyymmdd-str->tdate date)
		 #:mdate         (yyyymm-str->mdate month)
		 #:commodity     (tocom-new-code->commodity code)
		 ;; always 4 decimals, strip zeros so as to have only as
		 ;; many as really needed
		 #:open          (string-trim-zero-decimals open)
		 #:high  	 (string-trim-zero-decimals high)
		 #:low   	 (string-trim-zero-decimals low)
		 #:close 	 (string-trim-zero-decimals settle)
		 #:volume        volume
		 #:open-interest openint)))
    row-list)))


;;-----------------------------------------------------------------------------
;; data download - year
;;
;; The historical page at
;;
;;     http://www.tocom.or.jp/historical/index.html
;;
;; has year files like
;;
;;     http://www.tocom.or.jp/data/yakujou/comp/2002.zip
;;
;; containing trade data for that year, with all commodities and months.
;; Currently these only go up to 2004.  There's also .tgz files up to 2002,
;; which are about the same size, prefer the .zip since they go further.


;; return year to start from for SYMBOL
(define (download-start-year symbol)
  (first (tdate->ymd (download-start-tdate symbol))))

(define (zip-member->mdate filename)
  (let ((m (must-match (string-match "^([0-9][0-9][0-9][0-9]-[0-9][0-9])[.]csv$" filename))))
    (y/m-str->mdate (match:substring m 1))))

;; process a .zip in FILENAME
(define (process-zip-file symbol-list year filename)
  (for-each (lambda (member)
	      (tocom-process-csv symbol-list (zip-member->mdate member)
				 #f #f (unzip-extract-member filename member)))
	    (unzip-list-files filename)))

(define (yearly-download-one symbol-list year)
  (download-status (_ "TOCOM") (_ "data") (number->string year))

  (receive (headers body)
      (http-request (simple-format #f "http://www.tocom.or.jp/data/yakujou/comp/~a.zip" year))

    (let ((status-code (http-headers-status-code headers)))
      (cond ((http-headers-ok? headers)
	     (call-with-string-in-file body
	       (lambda (filename)
		 (process-zip-file symbol-list year filename)
		 #t)))

	    ((= 404 status-code)
	     ;; 404, no more year files, terminate loop
	     #f)

	    (else
	     (http-headers-throw headers))))))


(define (tocom-download-years symbol-list)
  (if (unzip-version)
      ;; assume each year file available on 1 Jan, though actually they only
      ;; go up to 2004 currently (ie. 2005 hasn't become available on 1 Jan
      ;; 2006)
      (let* ((avail-year  (1- (today-year (timezone-tokyo))))
	     (avail-tdate (ymd->tdate avail-year 12 31)))

	;; nothing to download for symbols with data and with data already
	;; up to what's available, or already up to contract expiry
	(set! symbol-list
	      (remove (lambda (symbol)
			(and-let* ((last-tdate (database-last-tdate symbol)))
			  (or (>= last-tdate avail-tdate)
			      (>= last-tdate (tocom-symbol-expiry symbol)))))
		      symbol-list))

	;; start at the lowest of those selected to download, and a dummy
	;; avail+1 if nothing
	(do ((year (apply min (1+ avail-year)
			  (map download-start-year symbol-list))
		   (1+ year)))
	    ((> year avail-year))

	  ;; stop at 404
	  (if (not (yearly-download-one symbol-list year))
	      (set! year avail-year))))))


;;-----------------------------------------------------------------------------
;; data download - month
;;
;; The historical page at
;;
;;     http://www.tocom.or.jp/historical/index.html
;;
;; has month files like
;;
;;     http://www.tocom.or.jp/data/yakujou/2005-06.csv
;;
;; which is trading for that month, all commodities and contract months.
;; The latest month file updates daily with additional data.
;;
;; The server provides an ETag and Last-Modified which we use to avoid
;; getting another copy of data we've already seen and processed (in case
;; our idea of when it's updated is a bit out).


(define (tocom-process-csv symbol-list body-mdate url headers body)
  (if (= 11 (length (string-split (string-first-line body)
				  #\,)))
      (tocom-process-old-csv symbol-list body-mdate url headers body)
      (tocom-process-new-csv symbol-list body-mdate url headers body)))

;; return tdate for available download data
;; seen Last-Modified
;;     2005-09.csv:  "Wed, 21 Sep 2005 23:24:33 GMT"  = 8:24 tokyo
;;                   "Wed, 28 Sep 2005 23:06:32 GMT"
;; so each day after 8:30am Tokyo
;;
(define (tocom-available-tdate)
  (1- (tdate-today-after 8 30 (timezone-tokyo))))

(define (tocom-download-months symbol-list)
  (define avail-tdate (tocom-available-tdate))

  ;; don't download symbols which already have data upto what's available,
  ;; or already to expiry
  (set! symbol-list
	(remove (lambda (symbol)
		  (let ((start-tdate (download-start-tdate symbol)))
		    (or (> start-tdate avail-tdate)
			(> start-tdate (tocom-symbol-expiry symbol)))))
		symbol-list))

  (if (not (null? symbol-list))
      (let ((start-tdate (apply min (map download-start-tdate symbol-list)))
	    (avail-mdate (tdate->mdate avail-tdate)))

	;; other symbols covered by this data
	(set! symbol-list (download-also symbol-list
					 #:selector    tocom-symbol?
					 #:start-tdate start-tdate))

	(do ((mdate (tdate->mdate start-tdate) (1+ mdate)))
	    ((> mdate avail-mdate))

	  (let* ((url (apply
		       format #f
		       "http://www.tocom.or.jp/data/yakujou/~a-~2,'0d.csv"
		       (mdate->ymd mdate))))
	    (download-status (_ "TOCOM") (_ "data") (basename url))
	    (receive (headers body)
		(apply http-request url
		       (download-tag-args url symbol-list))

	      (let ((status-code (http-headers-status-code headers)))
		(cond ((http-status-code-ok? status-code)
		       ;; no BODY-MDATE passed here, since latest month file
		       ;; doesn't cover whole month
		       (tocom-process-csv symbol-list #f url headers body))

		      ((memv status-code '(304 404))
		       ;; 304 not modified if nothing new, or 404 not found
		       ;; if looking too soon for new month file
		       )

		      (else
		       ;; anything else is bad
		       (http-headers-throw headers))))))))))


;;-----------------------------------------------------------------------------
;; data download

(define (tocom-download symbol-list)
  (tocom-download-years symbol-list)
  (tocom-download-months symbol-list))

(define-public (tocom-backto want-tdate symbol-list)
  (define want-year (tdate->year want-tdate))

  (for-each
   (lambda (symbol-list) ;; of each commodity
     (do ((year (tdate->year (download-back-tdate-max symbol-list))
		(1- year)))
	 ((< year want-year))
       ;; stop at 404
       (if (not (yearly-download-one symbol-list year))
	   (set! year want-year))))

   (partition-equal chart-symbol-commodity symbol-list)))

(download-handler! #:selector   tocom-symbol?
		   #:update     tocom-download
		   #:avail-proc tocom-available-tdate
		   #:backto     tocom-backto)

(download-now-handler! (lambda (symbol-list)
			 (and (any tocom-symbol? symbol-list)
			      (download-now-all-commodities-and-months))))

(download-now-handler! (lambda (symbol-list)
			 (and (any tocom-symbol? symbol-list)
			      (download-now-need-unzip))))


;;-----------------------------------------------------------------------------
;; latest quotes
;;
;; This uses the intraday quotes pages like
;;
;;     http://www.tocom.or.jp/souba/silver/index.html
;;

;; no date on delisted Gas Oil page
(define (tocom-latest-empty? body)
  (and-let* ((m (regexp-exec #,(regexp "As of(.*\n.*)Minimum" regexp/newline)
			     body)))
    (string-every char-set:whitespace 
		  (html-text-extract (match:substring m 1)))))

;; return list of latest records
(define (tocom-body->latest-list commodity headers body)
  ;; eg. "<td class="font12">As of Apr.28, 2005 18:00JST</td>"
  (let* ((m        (must-match (regexp-exec #,(regexp "As of ([A-Za-z]+\\.[0-9]+, [0-9]+) ([0-9]+:[0-9]+)JST") body)))
	 (adate    (m/d/y-str->adate (match:substring m 1)))
	 (time     (h:m:s-str->seconds (match:substring m 2)))
	 (row-list (html-table-rows-raw body (string-contains body "As of "))))

    ;; heading "Current" while trading, "Close" after
    (let ((headings (map! html-text-extract (first row-list))))
      (or (or (equal? headings '("Month" "Last Settlement Price" "Open"
				 "High" "Low" "Current" "Change" "Volume"))
	      (equal? headings '("Month" "Last Settlement Price" "Open"
				 "High" "Low" "Close" "Change" "Volume")))
	  (error "TOCOM: unrecognised data columns:" headings))
      (set! row-list (cdr row-list)))

    ;; "Total" volume at end
    (set! row-list (remove! (lambda (row)
			      (string-contains (first row) "Total"))
			    row-list))

    (latest-list-add-front-months
     (map! (lambda (row)
	     (let ((note (cond ((string-contains (sixth row) "up.gif")
				(_ "limit up"))
			       ((string-contains (sixth row) "down.gif")
				(_ "limit down"))
			       (else
				#f))))
	       (receive-list (month prev open high low current change volume)
		   (map! html-text-extract row)
		 (latest-new #:symbol   (string-append commodity " " month
						       ".TOCOM")
			     #:last-adate     adate
			     #:last-time      time
			     #:open           open
			     #:high           high
			     #:low            low
			     #:last           current
			     #:change         change
			     #:prev           prev
			     #:volume         volume
			     #:contract-mdate (m/y-str->mdate month)
			     #:note           note
			     #:source         'tocom
			     #:headers        headers))))
	   row-list))))

;; convert commodity like "Gas Oil" to page name like "gas_oil"
(define (commodity->page commodity)
  (string-map (lambda (c)
		(if (char=? c #\space)
		    #\_
		    c))
	      (string-downcase commodity)))

;; return list (TDATE TIME) which is TM locked to within TOCOM trading
;; hours, being weekdays 9:00am to 11:00am and 12:30pm to 3:30pm.
;; the middle times are stretched by 2 minutes in case our system time is out
;;
(define (tocom-tm->tdate-time-lock tm)
  (let ((tdate-time (tm->tdate-time-within
		     tm #,(hms->seconds 9 0 0) #,(hms->seconds 15 30 0))))
    (if (within-range? #,(hms->seconds 11 2 0) #,(hms->seconds 12 28 0)
		       (second tdate-time))
	(set-second! tdate-time #,(hms->seconds 11 2 0)))
    tdate-time))

(define (tocom-latest-get symbol-list extra-list proc)

  ;; Only download those symbols for which we have a newer prospective quote
  ;; time.  `latest-last-time' from the database is #f, treat that as the
  ;; close of trade 3:30pm.
  (receive-list (tdate time)
      (tocom-tm->tdate-time-lock (localtime (current-time) (timezone-tokyo)))
    (receive (drop keep)
	(partition (lambda (symbol)
		     (and-let* ((latest (latest-get symbol))
				(       (latest-last-tdate latest)))
		       (two<= tdate time
			      (latest-last-tdate latest)
			      (or (latest-last-time latest)
				  #,(hms->seconds 15 30 0)))))
		   symbol-list)
      (set! symbol-list keep)
      (latest-fetch-fail drop)))

  (for-each (lambda (symbol-list) ;; of each commodity
	      (let ((commodity (chart-symbol-commodity (first symbol-list))))
		(receive (headers body)
		    (apply http-request
			   (string-append "http://www.tocom.or.jp/souba/"
					  (commodity->page commodity)
					  "/index.html")
			   (latest-tag-args 'tocom symbol-list))

		  (if (and (http-headers-ok? headers)
			   (not (tocom-latest-empty? body)))
		      (proc (tocom-body->latest-list commodity headers body))

		      ;; 304 page unmodified for no new data, or perhaps 404
		      ;; for unknown commodity
		      (latest-fetch-fail symbol-list)))))

	    (partition-equal chart-symbol-commodity symbol-list)))

(latest-handler! #:selector tocom-symbol?
		 #:handler tocom-latest-get)
