;; Trendscore indicator.

;; Copyright 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart trendscore)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))



;; http://www.biz-analyst.com/library/stocks_commodities_trend.pdf [gone]
;;     Tushar Chande, "Rating Trend Strength", TASC V.11:9 September 1993,
;;     reproduced in a collection of trend related TASC articles.
;;     (Apparently http://store.traders.com/trarpap.html)
;;
;; http://trader.online.pl/MSZ/e-w-Chandes_Trendscore.html
;;     Formula, sample on polish WIG20 from 2000.
;;


;; Return a procedure which calculates a trendscore over a window of COUNT
;; many points.
;;
;; Each call (PROC close) enters a new point into the window, and the return
;; is the trendscore up to (and including) that point, or #f if not enough
;; points yet.
;;
;; To prime the window initially, call PROC with 19 points preceding the
;; first desired.
;;
(define*-public (trendscore-calc-proc #:optional (data '()))

  ;; KEEP-COUNT is how many past closes to keep, including today's close.
  ;; The target values are 11 to 20 days ago, so it's the 20 values from 1
  ;; to 20 days ago, plus today's, is 21.
  ;;
  ;; DATA is kept as a circular list of the previous KEEP-COUNT many values,
  ;; with #f's until that many seen.
  ;;
  (set! data (indicator-circular keep-count data))

  (lambda (close)
    ;; replace oldest with new point
    (set-car! data close)
    (set! data (cdr data))

    (and (first data)
	 (fold (lambda (prev-close score)
		 (+ score (if (>= close prev-close) 1 -1)))
	       0 (take data 10)))))

(define keep-count 21)
(define trendscore-warmup-count (1- keep-count))


(define (trendscore-fill series new-array lo hi)
  (let* ((parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (trendscore-calc-proc
		  (series-prev-n-closes parent
					#:before lo
					#:count  trendscore-warmup-count))))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (trendscore-series series)

  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "Trendscore"))
    (series-datatype!  new-series trendscore-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts
				   trendscore-fill trendscore-warmup-count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public trendscore-datatype
  (datatype-new!
   #:key     'trendscore
   #:name    (_ "Trendscore")
   #:minimum -10
   #:maximum 10
   #:units   'plus-or-minus-10
   #:hlines  '(0)
   #:manual  (_ "CONTEXT(manual-node): Trendscore")
   #:proc    trendscore-series))
