;; True Strength Index.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart tsi)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart momentum))


;; "The True Strength Index", William Blau, TASC Nov 1991.
;;
;; http://store.traders.com/-v10-c05-trading-pdf.html
;;     "Trading With the True Strength Index", William Blau, TASC V.10:5
;;     May 1992 by William Blau.  Intro, rest for sale.
;;
;; http://store.traders.com/-v11-c01-sidetru-pdf.html
;;     Sidebar included of "Stochastic Momentum" article January 1993.
;;     Formula not shown, but 1-day momentum described.
;;
;; http://www.traders.com/Documentation/FEEDbk_docs/Archive/112005/Letters/Letters.html
;;     TASC letters Nov 2005, Perry Kaufman giving formula.
;;
;; http://www.linnsoft.com/tour/techind/tsi.htm
;;     Formula, parameters 9 and 3, sample intraday AOL.
;;
;; http://www.meta-formula.com/Metastock-Formulas-T.html
;;     Formula.
;;
;; http://trader.online.pl/MSZ/e-w-True_Strength_Index.html
;;     Formula with ROC percentage and programmable period.
;; http://trader.online.pl/MSZ/e-w-True_Strength_Index_II.html
;;     Formula, plain mom, periods 25 and 13.
;; http://trader.online.pl/MSZ/e-w-True_Strength_Index_III.html
;;     Formula, plain momentum, periods 15 and 5.
;; http://trader.online.pl/MSZ/e-w-True_Strength_Index_IV.html
;;     Formula, plain momentum, periods 25 and 13.
;; http://trader.online.pl/MSZ/e-w-True_Strength_Index_V.html
;;     Formula, showing EMA periods 40 and 20, sample on WIG20.
;;
;; http://www.fmlabs.com/reference/default.htm?url=TSI.htm
;;     Formula.
;;
;; http://www.prophet.net/analyze/popglossary.jsp?studyid=TSI
;;     Formula, sample on Maxtor Corp (symbol MXO, taken over by Seagate, no
;;     data in Yahoo).
;;


;; Return a procedure which calculates the true strength index, with the
;; given smoothing periods.
;;
;; Each call (PROC close) enters a new value into the window, and the return
;; is the TSI up to (and including) that point, or #f if not enough values
;; yet.
;;
;; The TSI uses EMAs and is thus in theory influenced by all preceding data,
;; but `tsi-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (tsi-warmup-count EMA1-COUNT EMA2-COUNT) many values,
;; the next call will have an omitted weight of no more than 0.1% of the
;; total.  Omitting 0.1% should be negligable, unless past values are
;; ridiculously bigger than recent ones.
;;
(define-public (tsi-calc-proc ema1-count ema2-count)
  (let ((num-ema1-proc (ema-calc-proc ema1-count))
	(num-ema2-proc (ema-calc-proc ema2-count))
	(den-ema1-proc (ema-calc-proc ema1-count))
	(den-ema2-proc (ema-calc-proc ema2-count))
	(mom-proc  (momentum-calc-proc 1)))

    (lambda (close)
      (and-let* ((mom (mom-proc close)))
	(let ((num-ema1 (num-ema1-proc mom))
	      (den-ema1 (den-ema1-proc (abs mom))))
	  (and num-ema1
	       (let ((num-ema2 (num-ema2-proc num-ema1))
		     (den-ema2 (den-ema2-proc den-ema1)))
		 (and num-ema2
		      (* 100 (/ num-ema2 den-ema2))))))))))

(define (tsi-warmup-count ema1-count ema2-count)
  ;; one before the momentum difference comes out, then warmup for the ema1
  ;; to get it coming out, then likewise for the ema2
  (+ 1
     (ema-warmup-count ema1-count)
     (ema-warmup-count ema2-count)))


(define (tsi-fill series new-array lo hi)
  (receive-list (ema1-count ema2-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (warmup (tsi-warmup-count ema1-count ema2-count))
	   (array  (series-array-closes parent lo hi))
	   (proc   (tsi-calc-proc ema1-count ema2-count)))

      ;; warmup
      (for-each proc (series-prev-n-closes parent #:before lo #:count warmup))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (proc (array-ref array i))
			i))))))

(define-public (tsi-series series ema1-count ema2-count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f (_ "TSI ~a, ~a")
					  ema1-count ema2-count))
    (series-datatype!  new-series tsi-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list ema1-count ema2-count))
    (series-require!   new-series
		       (indicator-require-parts
			tsi-fill (tsi-warmup-count ema1-count ema2-count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public tsi-datatype
  (datatype-new!
   #:key     'tsi
   #:name    (_ "TSI - True Strength Index")
   #:units   'percentage-plusorminus100
   #:minimum -100
   #:maximum 100
   #:hlines  '(-25 25)
   #:manual  (_ "CONTEXT(manual-node): True Strength Index")
   #:proc    tsi-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'tsi-ema1-days
	  #:name    (_ "EMA-1 Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 25)
	 (datatype-param-new
	  #:key     'tsi-ema2-days
	  #:name    (_ "EMA-2 Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 13))))
