;; Twiggs Money Flow.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart twiggs-mf)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.incrediblecharts.com/technical/twiggs_money_flow.htm
;;


;; Return a procedure which calculates a Twiggs Money Flow index over an
;; accumulated window of COUNT values.
;;
;; Each call (PROC high low close volume) enters a new value into the
;; window, and the return is the Twiggs Money Flow up to (and including)
;; that value.  The values can be #f when data is not available.  If there's
;; not enough valid points the return will be #f.
;;
;; If all the VOLUME amounts in the window are zero, then the return is 0
;; too.
;;
;; This windowed parts/total calculation is similar to
;; `percent-posneg-calc-proc', but we scale to -1 to +1.
;;
(define-public (twiggs-mf-calc-proc count)

  (define num-ema-proc (ema-calc-proc (ema-count-wilder count)))
  (define den-ema-proc (ema-calc-proc (ema-count-wilder count)))
  (define prev-close   #f)

  (lambda (high low close volume)

    ;; dummy out if not high/low and/or volume
    (set! high   (or high close))
    (set! low    (or low  close))
    (set! volume (or volume 0))

    (if prev-close
	(begin
	  ;; possibly extend to previous close, as per true range
	  (set! high (max high prev-close))
	  (set! low  (min low  prev-close))
	  (set! prev-close close)

	  (let* ((diff (- high low))
		 (x    (if (zero? diff)
			   0
			   ;; 2.0 forces to a flonum, not a frac in guile 1.8
			   (* volume (1- (* 2.0 (/ (- close low)
						   diff))))))
		 (num  (num-ema-proc x))
		 (den  (den-ema-proc volume)))
	    (and num ;; when enough data
		 (if (= 0 den)
		     0
		     (/ num den)))))
	(begin
	  (set! prev-close close)
	  #f))))

;; the ema procs, plus 1 for prev-close
(define-public (twiggs-mf-warmup-count count)
  (1+ (ema-warmup-count (ema-count-wilder count))))


(define (twiggs-mf-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos
		  parent
		  #:before lo
		  #:count  (twiggs-mf-warmup-count count)))
	 (array  (series-array parent start hi))
	 (cmf    (twiggs-mf-calc-proc count)))

    ;; warmup points preceding lo, if available
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (cmf (array-ref array i 1)    ;; high
	       (array-ref array i 2)    ;; low
	       (array-ref array i 3)    ;; close
	       (array-ref array i 4)))) ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array
		      (cmf (array-ref array i 1)   ;; high
			   (array-ref array i 2)   ;; low
			   (array-ref array i 3)   ;; close
			   (array-ref array i 4))  ;; volume
		      i)))))

(define-public (twiggs-mf-series series count)
  (ohlcv-required "Twiggs MF" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Twiggs MF") count))
    (series-datatype!  new-series twiggs-mf-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-array!     new-series #f)
    (series-require!   new-series (indicator-require-parts
				   twiggs-mf-fill 
				   (twiggs-mf-warmup-count count)))
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public twiggs-mf-datatype
  (datatype-new!
   #:key       'twiggs-mf
   #:name      (_ "Twiggs Money Flow")
   #:input     datatype-ohlcv
   #:units     'plus-or-minus-one
   #:minimum   -1
   #:maximum   1
   #:hlines    '(0)
   #:manual    (_ "CONTEXT(manual-node): Twiggs Money Flow")
   #:proc      twiggs-mf-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'twiggs-mf-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 21))))
