;; Ulcer Index.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ulcer)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.tangotools.com/ui/ui.htm
;;     Description by creator Peter Martin.
;;
;; http://www.tangotools.com/misc/index.html
;;     Peter Martin's home page.
;;
;; http://www.investopedia.com/articles/technical/03/030403.asp
;;     Sample chart of XOM.
;;
;; http://trader.online.pl/MSZ/e-w-Ulcer_Index.html
;;     Believe HHV described is not right, want highest up to each point in
;;     the calculation, not of whole period.


;; Return the Ulcer Index calculated over the given LST of price points.
;;
(define-public (ulcer-calc lst)
  (if (null? lst)
      0
      (let ((high  (first lst))
	    (sumsq 0.0))
	(for-each (lambda (price)
		    (cond ((>= price high)
			   (set! high price))
			  ((not (zero? high)) ;; protect againt negative input
			   (set! sumsq (+ sumsq
					  (square (* 100.0 (/ (- high price)
							      high))))))))
		  (cdr lst))
	(sqrt (/ sumsq (length lst))))))


;; Return a procedure which calculates the Ulcer Index over an accumulated
;; window of COUNT points.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the ulcer index up to (and including) that VAL, or #f if not enough
;; points yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from start to end).  COUNT-1 is
;; enough to ensure the first PROC call is based on a full window.
;;
(define*-public (ulcer-calc-proc count #:optional (data '()))
  (set! data (take-right-maybe data count))
  (let ((points-min (ceiling-exact (* indicator-minimum-fraction count)))
	(q          (list->q data))
	(points     (length data)))
    (set! data #f) ;; gc

    (lambda (x)

      ;; drop old point if full
      (if (>= points count)
	  (deq! q)
	  ;; not full, so POINTS increases
	  (set! points (1+ points)))

      ;; add new point
      (enq! q x)

      (and (>= points points-min)
	   (ulcer-calc (q-list q))))))


(define (ulcer-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (ulcer  (ulcer-calc-proc count
				  (series-prev-n-closes parent
							#:before lo
							#:count  (1- count)))))
    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (ulcer this) i)))))

(define-public (ulcer-series series count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "Ulcer") count))
    (series-datatype!  new-series ulcer-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts ulcer-fill
							   (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public ulcer-datatype
  (datatype-new!
   #:key     'ulcer
   #:name    (_ "Ulcer Index")
   #:units   'ulcer
   #:manual  (_ "CONTEXT(manual-node): Ulcer Index")
   #:proc    ulcer-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'ulcer-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 60))))  ;; don't know what's good here, this is arbitrary
