;; Larry Williams' Ultimate oscillator.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart ultimate)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart sma))


;; http://www.stockcharts.com/education/IndicatorAnalysis/indic_ultimate.html



(define period1  7)
(define period2 14)
(define period3 28)


(define (ultimate-fill series new-array lo hi)
  (let* ((parent     (series-parent series))
	 (prev-close #f)
	 (start      (series-prev-n-closes-pos parent
					       #:before lo #:count period3))
	 (array      (series-array parent start hi))
	 (b1-proc    (sma-calc-proc period1))
	 (t1-proc    (sma-calc-proc period1))
	 (b2-proc    (sma-calc-proc period2))
	 (t2-proc    (sma-calc-proc period2))
	 (b3-proc    (sma-calc-proc period3))
	 (t3-proc    (sma-calc-proc period3)))

    (define (ult i)
      (and-let* ((close (array-ref array i 3)))

	;; use closes if no high/low data
	(let ((high (or (array-ref array i 1) close))
	      (low  (or (array-ref array i 2) close)))

	  ;; extend to true high/low
	  (if prev-close
	      (begin
		(set! high (max high prev-close))
		(set! low  (min low  prev-close))))

	  ;; new previous close position
	  (set! prev-close close)

	  (let ((bp (- close low))  ;; buying pressure
		(tr (- high low)))  ;; true range

	    (let ((b1 (b1-proc bp))
		  (t1 (t1-proc tr))
		  (b2 (b2-proc bp))
		  (t2 (t2-proc tr))
		  (b3 (b3-proc bp))
		  (t3 (t3-proc tr)))

	      (and b1 t1 b2 t2 b3 t3  ;; wait for all to fill up
		   (not (zero? t1))
		   (not (zero? t2))
		   (not (zero? t3))

		   (* 14.28571428571428571428571428   ;;  100 / 7
		      (+ (* 4 (/ b1 (exact->inexact t1)))
			 (* 2 (/ b2 (exact->inexact t2)))
			 (/ b3 (exact->inexact t3))))))))))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (ult i))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (array-set! new-array (ult i) i))))

(define-public (ultimate-series series)
  (ohlcv-required "Ultimate Oscillator" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (_ "Ultimate"))
    (series-datatype!  new-series ultimate-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series #f)
    (series-require!   new-series (indicator-require-parts ultimate-fill
							   period3))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!    new-series '())

    new-series))

(define-public ultimate-datatype
  (datatype-new!
   #:key       'ultimate
   #:name      (_ "Ultimate Oscillator")
   #:input     datatype-ohlcv
   #:units     'percentage
   #:minimum   0
   #:maximum   100
   #:hlines    '(30 50 70)
   #:manual    (_ "CONTEXT(manual-node): Ultimate Oscillator")
   #:proc      ultimate-series))
