#!/bin/sh
exec chart -s "$0" "$@"
!#

;; Import data files.

;; Copyright 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: chart -s brown72.scm
;;


(use-modules (ice-9 and-let-star)
	     (srfi srfi-13)
	     (srfi srfi-14)
	     (chart database)
	     (chart download)
	     (chart misc))

(define-public (array-num-rows array)
  (- (array-last-index array)
     (array-first-index array)
     -1))

(define (slurp filename)
  (call-with-input-file filename read-string-to-eof))

(define char-set:eE (char-set #\e #\E ))

(define (string-shift-decimal-point str exp)
  (let ((decimals (string-decimals str)))
    (set! str (string-delete str #\.))
    (set! decimals (- decimals exp))
    (cond ((negative? decimals)
	   (string-append str (make-string (- decimals) #\0)))
	  ((zero? decimals)
	   str)
	  (else
	   (let ((neg ""))
	     (if (string-prefix? "-" str)
		 (begin
		   (set! neg "-")
		   (set! str (string-drop str 1))))
	     (if (>= (string-length str) decimals)
		 (string-append neg
				(string-drop-right str decimals)
				"."
				(string-take-right str decimals))
		 (string-append neg
				"."
				(make-string (- decimals (string-length str))
					     #\0)
				str)))))))

(define (string-exponent->decimals str)
  (let ((pos (string-index str char-set:eE)))
    (if pos
	(let ((mant (string-take str pos))
	      (exp (string->number (string-drop str (1+ pos)))))
	  (string-shift-decimal-point mant exp))
	str)))

(define (string-list->series symbol lst)
  (let ((array (make-array #f (list 0 (1- (length lst))) '(0 4))))

    (do ((i 0 (1+ i))
	 (lst lst (cdr lst)))
	((null? lst))
      (array-set! array (string-exponent->decimals (car lst))
		  i 3))

    (let ((decimals (array-strings->numbers+decimals array)))
      (series-new #:symbol    symbol
		  #:decimals  decimals
		  #:array     array
		  #:datatype  datatype-ohlcv))))
      
(define (commasep-file->series symbol filename)
  (let* ((str (slurp filename))
	 (lst (string-tokenize str (char-set-complement
				    (char-set-adjoin char-set:whitespace
						     #\,)))))
    (string-list->series symbol lst)))


(define (series-cumulative series new-symbol)
  (let* ((lo        (1- (series-lo series)))
	 (hi        (series-hi series))
	 (array     (series-array series lo hi))
	 (m         (apply arithmetic-mean
			   (array->list (make-shared-array-column array 3))))
	 (new-array (make-array #f (list lo hi) '(0 4)))
	 (sum       0))
    (format #t "mean ~a\n" m)
    (set! m (* 40.0 (integer-expt 10 (series-decimals series))))
    (format #t "but use ~a\n" m)

    (array-set! new-array 0 lo 3)

    (do ((i (1+ lo) (1+ i)))
	((> i hi))
      (and-let* ((x (array-ref array i 3)))
	(set! sum (+ sum x (- m)))
	(array-set! new-array sum i 3)))

    (series-new #:symbol    new-symbol
		#:decimals  (series-decimals series)
		#:array     new-array
		#:datatype  datatype-ohlcv)))


(let* ((filename "/down/charting/fdi/hurst/hurst/data/brown72.h")
       (series   (commasep-file->series "BROWN72-DIFFS.DATA" filename)))
  (database-write-series series)
  (let ((cumul-series (series-cumulative series "BROWN72-CUMUL.DATA")))
    (database-write-series cumul-series))

  (format #t "total ~a data points\n" (- (series-hi series)
					 (series-lo series)
					 -1)))
(exit 0)
