;; Commonwealth Securities data downloading.

;; Copyright 2004, 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart commsec)
  #:autoload   (ice-9 format) (format)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart australia)
  #:use-module (chart download)
  #:use-module (chart down-cost)
  #:use-module (chart float)
  #:use-module (chart http)
  #:use-module (chart misc))



;;-----------------------------------------------------------------------------
;; misc

;; commsec is used only if the URL is set in the Edit/Preferences dialog
(define-public (commsec-available?)
  (string-ci=? "yes" (preference-get 'commsec-enabled "no")))

;; all ASX shares, but only when commsec enabled
(define (commsec-symbol? symbol)
  (and (commsec-available?)
       (australia-symbol? symbol)))

;; tdate of latest data available from commsec
;; today is available each day at 10:30pm Sydney time
;;
(define-public (commsec-available-tdate)
  (tdate-today-after 22 30 (timezone-sydney)))


;;-----------------------------------------------------------------------------
;; historical processing

;; Sample line: AEZ,"01 Jun 2007",1.35,1.37,1.305,1.325,4008329
;;
(define (commsec-process-download symbol body)
  (define row-list (if (string-contains-ci body "server error")
		       '() ;; unknown symbol, mark the last-download
		       (map! csv-split (string-split-lines body))))

  (download-process
   #:module          (_ "CommSec")
   #:symbol-list     (list symbol)
   #:last-download   #t
   #:prefer-decimals 2
   #:row-list
   (map-rows!
    (lambda (commodity tdate open high low close volume)
      (list #:tdate     (d/m/y-str->tdate tdate)
	    #:commodity commodity
	    #:open      open
	    #:high      high
	    #:low       low
	    #:close     close
	    #:volume    volume))
    row-list)))


;;-----------------------------------------------------------------------------
;; download - by historical files
;;

(define-public commsec-hist-months-alist
  '((1   . "1mo")
    (2   . "2mo")
    (3   . "3mo")
    (6   . "6mo")
    (12  . "1yr")
    (24  . "2yr")
    (36  . "3yr")
    (48  . "4yr")
    (60  . "5yr")
    (120 . "10yr")))

;; return period string like "6mo" for MONTHS
;; MONTHS can be any number, the next higher possible period is returned
(define (months->period-str months)
  (or (any (lambda (elem)
	     (and (>= (car elem) months)
		  (cdr elem)))
	   commsec-hist-months-alist)
      "10yr"))

;; return number of months needed to cover back to TDATE
(define-public (commsec-hist-tdate->months tdate)
  (set! tdate (- tdate 5)) ;; bit of leeway
  (receive-list (now-year now-month now-day)
      (today-ymd (timezone-sydney))
    (receive-list (td-year td-month td-day)
	(tdate->ymd tdate)
      (+ (* 12 (- now-year td-year))
	 (- now-month td-month)
	 (if (>= now-day td-day)
	     1
	     0)))))

(define-public (commsec-historical-one symbol months)
  (let* ((period-str (months->period-str months)))
    (download-status (_ "CommSec") (_ "data") symbol period-str)
    (receive (headers body)
	(http-request (string-append "http://charts.commsec.com.au/HistoryData/HistoryData.dll/GetData?Symbol="
				     (chart-symbol-sans-dot symbol)
				     "&TimePeriod=" period-str "&.csv"))

      (or (string-contains-ci body "server error")
	  (download-cost-store 'commsec-bytes-per-month
			       (quotient (string-length body)
					 months)))

      (if (http-headers-ok? headers)
	  (commsec-process-download symbol body)))))

(define-public (commsec-historical symbol-list)
  (for-each (lambda (symbol)
	      (commsec-historical-one symbol
				      (commsec-hist-tdate->months
				       (download-start-tdate symbol))))
	    symbol-list))

;; return the limit of data available from commsec, that being 10 years back
;; from today
(define (commsec-backto-limit)
  (receive-list (year month day)
      (today-ymd (timezone-sydney))
    (ymd->tdate year
		(- month
		   (car (last commsec-hist-months-alist))
		   -1) ;; bit of margin for error
		day)))

(define (commsec-backto want-tdate symbol-list)
  (let ((limit (commsec-backto-limit)))
    (for-each (lambda (symbol)
		;; only if within the limit of what commsec has
		(if (> (download-back-tdate symbol) limit)
		  (commsec-historical-one
		   symbol (commsec-hist-tdate->months want-tdate))))
	      symbol-list)))


;; Return a procedure which is to be called (PROC tdate), returning the cost
;; of using an historical download to get from TDATE to today.
;;
;; One month's historical is about 1300 bytes, but only multiples in
;; `commsec-hist-months-alist' can be taken.
;;
(define-public (commsec-make-hist-cost-proc)
  (define bytes-per-month (download-cost-get 'commsec-bytes-per-month 1300))
  (define avail-tdate (commsec-available-tdate))

  (receive-list (year month day)
      (tdate->ymd avail-tdate)
    (let ((cost-alist (map (lambda (elem)
			     (cons (+ 4 (ymd->tdate-ceil
					 year (- month (car elem)) day))
				   (* bytes-per-month (car elem))))
			   commsec-hist-months-alist)))
      (lambda (tdate)
	(cdr (or (assoc tdate cost-alist >=)
		 '(#f . 9999999999999999999)))))))


;;-----------------------------------------------------------------------------
;; download - by whole day files
;;
;; Commsec offers the following formats,
;;
;;     metastock - prices in dollars, date yymmdd
;;     metastock - prices in dollars, date yymmdd, volume in 100s
;;     ezychart  - prices in cents, date yymmdd
;;     insight   - prices in cents, date mm/dd/yy, space separated
;;     stockeasy - prices in dollars, date yyyymmdd
;;
;; Ezychart is used because it's the most compact -- there's no decimal
;; points in most prices, and no century on the date.
;;
;; The web page says only the past 20 days are available and that's all the
;; little menu presents, but the server actually goes back beyond that,
;; apparently unlimited.


(define (yymmdd-str->tdate str)
  (ymd->tdate (yy->year (string->number-err (substring str 0 2)))
	      (string->number-err (substring str 2 4))
	      (string->number-err (substring str 4 6))))

(define (daily-process-download symbol-list body)
  ;; public holidays and data not available yet gives:
  ;; "<h3>Server Error - [Exception - Server Error - [Error
  ;; Accessing SnapShot: 20040409]]</h3>"
  ;;
  (if (not (string-contains-ci body "server error"))
      (begin
	(download-cost-store 'commsec-daily-ezychart body)

	(download-process
	 #:module          (_ "CommSec")
	 #:symbol-list     symbol-list
	 #:currency        "AUD"
	 #:extra-decimals  2
	 #:prefer-decimals 0
	 #:row-list
	 (map! (lambda (row-str)
		 (receive-list (symbol tdate open high low close volume)
		     (string-split row-str #\,)
		   (list #:tdate     (yymmdd-str->tdate tdate)
			 #:commodity symbol
			 #:open      open
			 #:high      high
			 #:low       low
			 #:close     close
			 #:volume    volume)))
	       (string-split-lines body))))))

(define (commsec-daily-url tdate)
  (receive-list (year month day)
      (tdate->ymd tdate)
    (format #f "http://charts.commsec.com.au/HistoryData/HistoryData.dll/EzyChart-~d~2,'0d~2,'0d?DownloadDate=~d~2,'0d~2,'0d&DownloadFormat=EzyChart&.txt"
	    year month day year month day)))

(define-public (commsec-daily symbol-list start-tdate avail-tdate)
  (do ((tdate start-tdate (1+ tdate)))
      ((> tdate avail-tdate))
    (download-status (_ "CommSec") (_ "data") (tdate-range-string tdate tdate))

    (set! symbol-list (download-also symbol-list
				     #:selector    australia-symbol?
				     #:start-tdate tdate
				     #:end-tdate   tdate))
    (receive (headers body)
	(http-request (commsec-daily-url tdate)
		      #:want-ok #t)
      ;; when public holiday, date too old, etc, still get a successful
      ;; download, with an error message in the body
      (daily-process-download symbol-list body))))


;;-----------------------------------------------------------------------------
;; download
;;
;; This chooses between the whole-day files or the individual symbol
;; downloads, and lets float.scm have a go too.  The order ends up as
;;
;;     - commsec individuals
;;     - float individuals and daily
;;     - commsec daily (only used if float is down)
;;     - yahoo individuals
;;
;; The float daily files are preferred over the commsec ones since they're
;; zipped and are about 1/5 the size to download.
;;
;; The main problem with the whole-day file is how many entries it has,
;; about 4500 as of Jan 2007.  There's about 1600 companies (a lot of them
;; small caps), the rest is warrants on the majors, and a few prefs or bonds
;; on various.
;;
;; For the "backto" download the order ends up as follows.  Float is
;; preferred because it's zipped, but otherwise it's not obvious which one
;; will have the most old data.
;;
;;     - float individual
;;     - commsec individual
;;     - yahoo individual
;; 

(define (commsec-download-historical symbol-list)
  (download-status (_ "CommSec") (_ "strategy"))
  (let ((avail-tdate    (commsec-available-tdate))
	(hist-cost-proc (commsec-make-hist-cost-proc))
	(daily-cost     (min (float-daily-download-cost)
			     ;; 256k as of Aug 2007
			     (download-cost-get 'commsec-daily 256000))))
    (receive (hist-list start-tdate)
	(download-by-day-or-by-symbol avail-tdate symbol-list
				      hist-cost-proc daily-cost)
      ;; only the bits that can't better be done by float
      (commsec-historical hist-list))))

(download-handler! #:selector 	commsec-symbol?
		   #:update   	commsec-download-historical
		   #:backto   	noop
		   #:avail-proc commsec-available-tdate
		   #:priority   30) ;; before float

(define (commsec-download symbol-list)
  (download-status (_ "CommSec") (_ "strategy"))
  (let ((avail-tdate    (commsec-available-tdate))
	(hist-cost-proc (commsec-make-hist-cost-proc))
	(daily-cost     (download-cost-get 'commsec-daily
					   256000))) ;; as of Aug 2007
    (receive (hist-list start-tdate)
	(download-by-day-or-by-symbol avail-tdate symbol-list
				      hist-cost-proc daily-cost)

      (commsec-historical hist-list)
      (commsec-daily symbol-list start-tdate avail-tdate))))

(download-handler! #:selector 	commsec-symbol?
		   #:update   	commsec-download
		   #:backto   	commsec-backto
		   #:avail-proc commsec-available-tdate
		   #:priority   10) ;; after float, before yahoo
