/* Copyright 2007, 2008 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define HAVE_DB 1


#include <db_185.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include <libguile.h>


static scm_t_bits dbsmob_type;

#define DBSMOB_PTR(db)  ((DB *) SCM_SMOB_DATA (db))

#define DBSMOB_P(obj) \
  (SCM_SMOB_PREDICATE (dbsmob_type, obj) && DBSMOB_PTR(db) != NULL)

#define DBSMOB_VALIDATE(pos, scm)                                           \
  do {                                                                  \
    SCM_ASSERT_TYPE (DBSMOB_P (scm), scm, pos, FUNC_NAME, "open db");       \
  } while (0)


static size_t
dbsmob_free (SCM db)
{
  DB *p = DBSMOB_PTR (db);
  if (p != NULL)
    (*p->close) (p);
  return 0;
}

SCM_DEFINE (chart_dbopen, "dbopen", 3, 0, 0,
            (SCM filename, SCM flags, SCM mode),
	    "Open database @var{filename} and return a database object for\n"
	    "it.  @var{flags} are as per @code{open}, ie.@: a combination of\n"
	    "@code{O_RDWR}, @code{O_CREAT} etc.\n"
	    "\n"
	    "@var{mode} is a permissions mode used if @var{filename} is\n"
	    "newly created.  It's a set of RWX bits as per @code{open},\n"
	    "eg.@: @code{#o666} for read/write for everyone.")
#define FUNC_NAME s_chart_dbopen
{
  int c_flags = scm_to_int (flags);
  int c_mode = scm_to_int (mode);
  char *c_filename = scm_to_locale_string (filename);
  DB *p;

  p = dbopen (c_filename, c_flags, c_mode, DB_HASH, NULL);
  free (c_filename);
  if (p == NULL)
    SCM_SYSERROR;
  SCM_RETURN_NEWSMOB (dbsmob_type, p);
}
#undef FUNC_NAME


SCM_DEFINE (chart_db_put, "db-put", 3, 0, 0,
            (SCM db, SCM key, SCM value),
	    "Set @var{value} to be associated with @var{key} in @var{db}.\n"
	    "@var{key} and @var{value} are both strings.")
#define FUNC_NAME s_chart_db_put
{
  DBT c_key, c_value;
  int ret;
  DB *p;
  scm_dynwind_begin (0);

  DBSMOB_VALIDATE (SCM_ARG1, db);
  p = DBSMOB_PTR (db);
  
  c_key.data = scm_to_locale_stringn (key, &c_key.size);
  scm_dynwind_free (c_key.data);

  c_value.data = scm_to_locale_stringn (value, &c_value.size);
  scm_dynwind_free (c_value.data);

  ret = (*p->put) (p, &c_key, &c_value, 0);
  if (ret != 0)
    SCM_SYSERROR;

  scm_dynwind_end ();
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (chart_db_get, "db-get", 1, 0, 0,
            (SCM db, SCM key),
	    "Return the value associated with @var{key} in @var{db}.\n"
	    "@var{key} is a string, and the value returned likewise is a\n"
	    "string.")
#define FUNC_NAME s_chart_db_get
{
  DBT c_key, c_value;
  int ret;
  DB *p;

  DBSMOB_VALIDATE (SCM_ARG1, db);
  p = DBSMOB_PTR (db);

  c_key.data = scm_to_locale_stringn (key, &c_key.size);
  ret = (*p->get) (p, &c_key, &c_value, 0);
  free (c_key.data);

  if (ret != 0)
    SCM_SYSERROR;

  return scm_from_locale_stringn (c_value.data, c_value.size);
}
#undef FUNC_NAME


SCM_SYMBOL (sym_first, "first");
SCM_SYMBOL (sym_last,  "last");
SCM_SYMBOL (sym_next,  "next");
SCM_SYMBOL (sym_prev,  "prev");

SCM_DEFINE (chart_db_seq_move, "db-seq-move", 2, 0, 0,
            (SCM db, SCM where),
	    "Move the cursor in @var{db}.  @var{where} is a symbol:\n"
	    "@code{first} to go to the first record, or @code{next} to go to\n"
	    "the next record.\n"
	    "\n"
	    "The return value is a pair @code{(@var{key} . @var{value})}\n"
	    "which is the record moved to, or @code{#f} when no more records\n"
	    "(or none at all in the case of @code{first}).")
#define FUNC_NAME s_chart_db_seq_move
{
  DBT c_key, c_value;
  int ret, flag;
  DB *p;

  DBSMOB_VALIDATE (SCM_ARG1, db);
  p = DBSMOB_PTR (db);

  if (scm_is_eq (sym_first, where))
    flag = R_FIRST;
  else if (scm_is_eq (sym_last, where))
    flag = R_LAST;
  else if (scm_is_eq (sym_next, where))
    flag = R_NEXT;
  else if (scm_is_eq (sym_prev, where))
    flag = R_PREV;
  else
    SCM_MISC_ERROR ("invalid position: ~a", scm_list_1 (where));

  ret = (*p->seq) (p, &c_key, &c_value, flag);
  if (ret == 1)
    return SCM_BOOL_F;  /* not found */
  if (ret != 0)
    SCM_SYSERROR;

  return scm_cons (scm_from_locale_stringn (c_key.data, c_key.size),
                   scm_from_locale_stringn (c_value.data, c_value.size));
}
#undef FUNC_NAME


SCM_DEFINE (chart_db_del_current, "db-del-current", 1, 0, 0,
            (SCM db),
	    "Delete the current cursor record from @var{db}.")
#define FUNC_NAME s_chart_db_del_current
{
  static const DBT dummy_key = { NULL, 0 };
  int ret;
  DB *p;

  DBSMOB_VALIDATE (SCM_ARG1, db);
  p = DBSMOB_PTR (db);

  ret = (*p->del) (p, &dummy_key, R_CURSOR);
  if (ret == 1)
    SCM_MISC_ERROR ("no current record", SCM_EOL);
  if (ret != 0)
    SCM_SYSERROR;

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (chart_db_close, "db-close", 1, 0, 0,
            (SCM db),
	    "Close the database @var{db}.  Any data cached in memory is\n"
	    "written out.")
#define FUNC_NAME s_chart_db_close
{
  DB *p;
  int ret;
  DBSMOB_VALIDATE (SCM_ARG1, db);
  p = DBSMOB_PTR (db);

  ret = (*p->close) (p);
  SCM_SET_SMOB_DATA (db, (scm_t_bits) NULL);

  if (ret != 0)
    SCM_SYSERROR;

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (chart_db_sync, "db-sync", 1, 0, 0,
            (SCM db),
	    "Sync the database @var{db}.  Any data cached in memory is\n"
	    "written out.")
#define FUNC_NAME s_chart_db_sync
{
  DB *p;
  DBSMOB_VALIDATE (SCM_ARG1, db);
  p = DBSMOB_PTR (db);

  if ((*p->sync) (p, 0) != 0)
    SCM_SYSERROR;

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


/*---------------------------------------------------------------------------*/

void
chart_db_init (void)
{
#if HAVE_DB
  dbsmob_type = scm_make_smob_type ("db", 0);
  scm_set_smob_free (dbsmob_type, dbsmob_free);
#endif

/* #ifndef SCM_MAGIC_SNARFER */
/* #include "db.x" */
/* #endif */
}
