;; float.com.au setups and data downloading.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart float)
  #:use-module (ice-9 receive)
  #:autoload   (ice-9 format)  (format)  ;; only for daily download
  #:use-module (srfi srfi-1)
  #:use-module (chart australia)
  #:use-module (chart download)
  #:use-module (chart down-cost)
  #:use-module (chart http)
  #:use-module (chart misc)
  #:autoload   (chart unzip)  (unzip-version))



;; In early 2008 float.com.au was slightly stuck for a time, with files
;; later than 14 Feb 08 coming back empty.  That worked out ok actually, as
;; the download fallbacks went to commsec (when enabled) or yahoo.



;; Float offers the following formats,
;;
;; traded=on     - only symbols which have trades
;; csv=on        - zip file in csv format
;; ticker=XX,YY  - symbols to get
;; format=metastock     - prices in dollars, date yymmdd [default]
;;     	  fcharts       - prices in dollars, date yymmdd
;;     	  stockeasy     - prices in dollars, date yyyymmdd
;;     	  ezychart      - prices in cents, date yymmdd
;;     	  insighttrader - prices in dollars, date mm/dd/yy, space separated,
;;                        volume in 100s
;;     	  maus          - prices in cents, date mm/dd/yy, space separated
;; volume=1             - full volume value
;;        10            - in tens
;;        100           - in hundreds
;;        1000          - in thousands
;;
;; It doesn't seem to work to set options for a .zip download, it ends up
;; with the default.  That default might be meant to be "metastock", though
;; it has a 4-digit year like stockeasy.  Ezychart would be the most compact
;; (no decimal points in most prices, and no century on the date).
;;
;; The .zip files contain only one archive element, so in theory gzip is
;; supposed to be able to handle them, but it complains about CRC error, so
;; only using unzip for now.  Zlib is another possiblity, if the zip header
;; can be understood.


;;-----------------------------------------------------------------------------
;; download - common bits

(define float-use-zip?  unzip-version)

;; today's data is finalized at 9:15pm sydney time (weekdays)
(define-public (float-available-tdate)
  (tdate-today-after 21 15 (timezone-sydney)))

(define (float-process-download symbol-list single-tdate body)
  (define avail-tdate (float-available-tdate))

  ;; In Jul/Aug 2007 float was down or slightly broken for a while and sent
  ;; empty results.  Don't want that to interpret that as a public holiday
  ;; for a whole-day file, or an unknown symbol for an individual, so do
  ;; nothing and let the yahoo fallbacks have a go.
  ;;
  (if (not (string-null? body))
      (download-process
       #:module          (_ "Float")
       #:symbol-list     symbol-list
       #:lo              single-tdate
       #:hi              single-tdate  ;; covered even if empty file
       #:last-download   (not single-tdate)
       #:currency        "AUD"
       #:prefer-decimals 2
       #:row-list
       (map! (lambda (row-str)
	       (receive-list (symbol tdate open high low close volume)
		   (string-split row-str #\,)
		 (set! tdate (yyyymmdd-str->tdate tdate))

		 ;; don't go past avail-tdate, so as not to record data
		 ;; before finalized at 9pm
		 (and (<= tdate avail-tdate)
		      (list #:tdate     tdate
			    #:commodity symbol
			    #:open      open
			    #:high      high
			    #:low       low
			    #:close     close
			    #:volume    volume))))

	     (string-split-lines body)))))


;;-----------------------------------------------------------------------------
;; download - by symbol
;;
;; This uses the download of all data for a symbol like
;;
;;     http://www.float.com.au/download/BHP.txt
;; or
;;     http://www.float.com.au/download/BHP.zip
;;
;; The .zip is the .txt compressed.  A .csv is also available, but it's
;; bigger and the only difference is "" quotes on the fields.
;;
;; An unknown symbol produces an empty file, or a .zip containing an empty
;; file.
;;
;; Some old data, like 1997 in http://www.float.com.au/download/SGW.zip, is
;; not in date order.

(define-public (float-historical use-zip symbol-list)
  (for-each (lambda (symbol)
	      (let ((url (string-append "http://www.float.com.au/download/"
					(chart-symbol-sans-dot symbol)
					(if use-zip ".zip" ".txt"))))
		(download-status (_ "CONTEXT(web-site name): Float")
				 (_ "data") (basename url))
		(receive (headers body)
		    (http-request url #:want-ok #t)
		  (if use-zip
		      (begin
			(download-cost-store 'float-hist-zip body)
			(set! body (unzip-string body))))
		  (download-cost-store 'float-hist-txt body)

		  (float-process-download symbol-list #f body))))

	    symbol-list))


;;-----------------------------------------------------------------------------
;; download - by whole day files
;;
;; This uses the day files like
;;
;;     http://www.float.com.au/download/20050909.zip
;;     http://www.float.com.au/download/20050909.txt
;;
;; The .zip is the .txt compressed.  A .csv is also available, but it's
;; bigger and the only difference is "" quotes on the fields.
;;
;; On a public holiday like http://www.float.com.au/download/20060414.zip
;; the txt file is empty, or the zip file contains no files (no files at
;; all, not even an empty txt file).


;; return a url string like "http://www.float.com.au/download/20050909.txt"
;; or with .zip if USE-ZIP is true
(define (daily-url use-zip tdate)
  (receive-list (year month day)
      (tdate->ymd tdate)
    (format #f "http://www.float.com.au/download/~4,'0d~2,'0d~2,'0d.~a"
	    year month day
	    (if use-zip "zip" "txt"))))

;; download from LO to HI inclusive
(define-public (float-daily use-zip symbol-list lo hi)

  (do ((tdate lo (1+ tdate)))
      ((> tdate hi))
    (download-status (_ "CONTEXT(web-site name): Float") (_ "data")
		     (tdate-range-string tdate tdate))

    (receive (headers body)
	(http-request (daily-url use-zip tdate) #:want-ok #t)
      (if use-zip
	  (begin
	    (download-cost-store 'float-daily-zip body)
	    (set! body (unzip-string body))))
      (download-cost-store 'float-daily-txt body)

      (set! symbol-list (download-also symbol-list
				       #:selector    australia-symbol?
				       #:start-tdate tdate
				       #:end-tdate   tdate))
      (float-process-download symbol-list tdate body))))


;;-----------------------------------------------------------------------------
;; download
;;
;; The download chooses between
;;    - float whole-day update files (possible zipped)
;;    - float individual historical files (possibly zipped)
;;
;; To update many ASX shares the whole-day files are best, or for just a few
;; then the individual files are best.  A mixture is used too, if there's a
;; few symbols that are quite a bit behind then they'll be done
;; individually, and the balance with whole-day.
;;
;; The main problem with the whole-day file is how many entries it has,
;; about 4500 as of Jan 2007.  There's about 1600 companies (a lot of them
;; small caps), the rest is warrants on the majors, and a few prefs or bonds
;; on various.


(define-public (float-daily-download-cost)
  (if (float-use-zip?)
      (download-cost-get 'float-daily-zip 45000)    ;; Sep07
      (download-cost-get 'float-daily-txt 196000))) ;; Mar06

(define (float-download symbol-list)
  (download-status (_ "Float") (_ "strategy"))
  (let* ((use-zip     (float-use-zip?))
	 (hist-cost   (if use-zip
			  (download-cost-get 'float-hist-zip 30000)
			  (download-cost-get 'float-hist-txt 90000)))
	 (avail-tdate (float-available-tdate)))

    (define (hist-cost-proc tdate)
      hist-cost)

    (receive (hist-list start-tdate)
	(download-by-day-or-by-symbol avail-tdate symbol-list
				      hist-cost-proc
				      (float-daily-download-cost))
      (float-historical use-zip hist-list)
      (float-daily use-zip symbol-list start-tdate avail-tdate))))

(download-handler! #:selector australia-symbol?
		   #:update   float-download
		   #:backto   noop
		   #:priority 20) ;; before 2nd commsec, and before yahoo
