;; ...



;; FTP protocol.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 2, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License along
;; with Chart; see the file COPYING.  If not, write to the Free Software
;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
;; USA.


(define-module (chart ftp)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart download)
  #:use-module (chart gethost)
  #:use-module (chart http)
  #:use-module (chart misc))


;; Crib notes:
;;
;; RFC959 -- spec
;;


;; return the FTP server port number
;; this is 21, but look it up in /etc/services
(define ftp-connect-port
  (make-procedure-with-cache-0
   (lambda ()
     (servent:port (getservbyname "ftp" "tcp")))))


(define-public (ftp-open hostname)
  (let* ((hostent  (c-gethost hostname))
	 (hostaddr (first (hostent:addr-list hostent)))
	 (af       (hostent:addrtype hostent))
	 (sock     (socket (af->pf af) SOCK_STREAM 0)))

    ;; the linux kernel default for keepalive is 9 tries 75 seconds apart,
    ;; which is fully 12.5 minutes, making this option fairly useless, our
    ;; own timeouts are shorter
    (setsockopt sock SOL_SOCKET SO_KEEPALIVE 1)

    (fcntl-set-nonblock sock)

    ((fluid-ref http-status-callback) (_ "connect"))
    (catch 'system-error
      (lambda ()
	(connect sock af hostaddr (ftp-connect-port)))
      (lambda args
	(if (not (system-error-eagain? args))
	    (apply throw args))))

    (let ((resp (ftp-response-read sock)))
      (or (eqv? 220 (ftp-response-code resp))
	  (throw 'ftp resp)))

    sock))

;; login as an anonymous user, with a dummy email address as password
;;
(define-public (ftp-login-anonymous sock)
  ((fluid-ref http-status-callback) (_ "login anonymous"))
  (let ((command "USER anonymous\r\n"))
    (http-write-string sock command)
    (let ((resp (ftp-response-read sock)))
      (case (ftp-response-code resp)
	((230)   ;; ok
	 #f)
	((331)   ;; need password
	 ((fluid-ref http-status-callback) (_ "dummy password"))
	 (set! command "PASS foo@bar.com\r\n")
	 (http-write-string sock command)
	 (let ((resp (ftp-response-read sock)))
	   (case (ftp-response-code resp)
	     ((230)  ;; ok
	      #f)
	     (else
	      (throw 'ftp command resp)))))
	(else
	 (throw 'ftp command resp))))))

(define-public (ftp-type-image sock)
  ((fluid-ref http-status-callback) (_ "binary transfer mode"))
  (http-write-string sock "TYPE I\r\n")
  (let ((resp (ftp-response-read sock)))
    (or (eqv? 200 (ftp-response-code resp))
	(throw 'ftp resp))))

(define (ftp-response-line-complete? line)
  (regexp-exec #,(regexp "^[0-9][0-9][0-9] ") line))

(define (ftp-response-read sock)
  (let more ((ret '()))
    (let ((line (http-read-to-crlf sock)))
      (dv line)
      (if (ftp-response-line-complete? line)
	  (string-concatenate-reverse/shared ret line)
	  (more (cons line ret))))))

(define (ftp-response-code str)
  (string->number (string-take (string-last-line str) 3)))

;; ;; codes "1xx" are preliminary, a further response is expected
;; (define (ftp-response-code-preliminary? code)
;;   (within-range? 100 199 code))

(define-public (ftp-pasv sock)
  (or (object-property sock 'ftp-pasv-parameters)
      (let ((command "PASV\r\n"))
	((fluid-ref http-status-callback) (_ "passive socket mode"))
	(http-write-string sock command)
	(let ((resp (ftp-response-read sock)))
	  (or (eqv? 227 (ftp-response-code resp))
	      (throw 'ftp command resp))

	  (let ((m (regexp-exec #,(regexp ".*\\(([0-9]+,[0-9]+,[0-9]+,[0-9]+),([0-9]+),([0-9]+)\\)") resp)))
	    (or m
		(throw 'ftp "PASV parse" resp))
	    (let ((host (match:substring m 1))
		  (p1   (match:substring m 2))
		  (p2   (match:substring m 3)))
	      (let ((params (list (inet-aton
				   (string-map (lambda (c)
						 (if (eqv? c #\,) #\. c))
					       host))
				  (+ (* 256 (string->number p1))
				     (string->number p2)))))
		(set-object-property! sock 'ftp-pasv-parameters params)
		params)))))))

(define (ftp-data-read sock command hostaddr port)
  ((fluid-ref http-status-callback)
   (simple-format #f (_ "connect for data ~a:~a") (inet-ntoa hostaddr) port))
  (let ((data-sock (socket PF_INET SOCK_STREAM 0)))
    ;; the linux kernel default for keepalive is 9 tries 75 seconds apart,
    ;; which is fully 12.5 minutes, making this option fairly useless, our
    ;; own timeouts are shorter
    (setsockopt data-sock SOL_SOCKET SO_KEEPALIVE 1)
    (fcntl-set-nonblock data-sock)

    (catch 'system-error
      (lambda ()
	(connect data-sock AF_INET hostaddr port))
      (lambda args
	(if (not (system-error-eagain? args))
	    (apply throw args))))

    (let* ((resp (ftp-response-read sock)))
      (dv resp (ftp-response-code resp))
      (or (eqv? 150 (ftp-response-code resp))
	  (throw 'ftp command resp))

      (let* ((m    (string-match "\\(([0-9]+) bytes\\)" resp))
	     (size (and m (string->number (match:substring m 1))))
	     (data (if size
		       (http-read-bytes data-sock size #f
					(fluid-ref http-status-callback))
		       (http-read-to-eof data-sock))))
	(close data-sock)
	data))))


(define (ftp-data-command sock command)
  (define pasv-list (ftp-pasv sock))

  (http-write-string sock command)
  (let ((data (apply ftp-data-read sock command pasv-list))
	(resp (ftp-response-read sock)))
    (or (eqv? 226 (ftp-response-code resp))
	(throw 'ftp command resp))
    data))
      
(define-public (ftp-list sock filename)
  ((fluid-ref http-status-callback) (_ "file listing"))
  (ftp-data-command sock (simple-format #f "LIST ~a\n" filename)))

(define-public (ftp-retr sock filename)
  ((fluid-ref http-status-callback) (_ "file download"))
  (ftp-data-command sock (simple-format #f "RETR ~a\r\n" filename)))

;; Download FILENAME from HOSTNAME if it's "list" command has changed from
;; OLD-LISTING.  OLD-LISTING can be #f to always download.
;; The return is two values (LISTING DATA).
;; If unchanged then LISTING is the same as OLD-LISTING and DATA is #f.
;; If downloaded then LISTING is the new info and DATA is a string.
;;
(define-public (ftp-get-if-modified hostname filename old-listing)
  (let ((sock (ftp-open hostname)))
    (call-with-cleanup (lambda () (close sock))
      (lambda ()
	(ftp-login-anonymous sock)
	(ftp-type-image sock)
	(ftp-pasv sock)
	(let ((new-listing (ftp-list sock filename)))
	  (if (equal? old-listing new-listing)
	      (values new-listing #f)

	      (begin
		(set-object-property! sock 'ftp-pasv-parameters #f)
		(ftp-pasv sock)
		(values new-listing (ftp-retr sock filename)))))))))



;; (dv (ftp-connect-port))
;; (dv (ftp-response-complete? "230 Ok\r\n"))
;; (dv (ftp-get-if-modified "localhost" "/foo/bar.txt" #f))
