;; unzip program functions.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart unzip)
  #:use-module (ice-9 popen)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart gzip)  ;; for program-version-proc
  #:use-module (chart http)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; misc

;; return a string which is the output from running COMMAND
(define (read-input-pipe command)
  (let* ((port (open-input-pipe command))
	 (str  (read-string-to-eof port)))
    (or (eqv? 0 (status:exit-val (close-pipe port)))
	(error "command exited non-zero:" command))
    str))

;; Put STR into a temporary file and call (PROC filename) with the name.
;; The file is deleted when proc exits, either normally or from an error
;; throw (but not when escaped through a continuation).
;;
(define-public (call-with-string-in-file str proc)
  (let* ((filename (g-build-filename (g-get-tmp-dir) "chart-tempfile-XXXXXX"))
	 (port     (mkstemp! filename)))
    (call-with-cleanup (lambda () (delete-file filename))
      (lambda ()
	(display str port)
	(set! str #f) ;; gc string contents, maybe
	(close-port port)
	(proc filename)))))


;;-----------------------------------------------------------------------------
;; availability

;; return the version number of the unzip program, or #f if not available
(define-public unzip-version
  (program-version-proc "unzip"
    (lambda (str)
      (set! str (string-first-line str))
      (set! str (string-trim-str str "UnZip "))
      (string-take-to-char str #\space))))

(define-public (download-now-need-unzip)
  (and (not (unzip-version))
       (_ "\
Data download from this source requires the \"unzip\" program, but that \
program is not available.  You must install it before downloading.")))


;;-----------------------------------------------------------------------------
;; single decompress

;; DATA is the contents of a .zip file, return its contents decompressed
;; (all the files in it concatenated, if there's more than one).
;;
;; If there's no files at all in the .zip then the return is an empty
;; string.  The exit status from unzip is 1 and a message "warning
;; [foo.zip]: zipfile is empty" is printed.  That's handled and the message
;; suppressed.
;;
(define-public (unzip-string data)
  (call-with-tempfileport (mkstemp-ext ".zip")
    (lambda (errport)
      (let* ((out (program-filter-string
		   (lambda (filename)
		     (string-append "unzip -p " filename
				    " 2>" (port-filename errport)))
		   data
		   #:exit-test status:0-or-1?))
	     (err (read-string-to-eof errport)))
	(or (string-match "^warning[^\n]*zipfile is empty[\r\n]*" err)
	    (display err (current-error-port)))
	out))))

;; STATUS is an exit status per `waitpid' and friends, return true if it's a
;; normal exit with code either 0 or 1
(define (status:0-or-1? status)
  (memv (status:exit-val status) '(0 1)))


;;-----------------------------------------------------------------------------
;; multi decompress

;; return a list of the filenames within the zip file FILENAME
(define-public (unzip-list-files filename)
  (let ((str (read-input-pipe (string-append "unzip -l " filename))))
    (delete "files"
	    (map (lambda (m)
		   (match:substring m 2))
		 (list-matches
		  (make-regexp "^[ \t]*[0-9]([^ \t\r\n]*[ \t]+)*([^ \t\r\n]+)$"
			       regexp/newline)
		  str)))))

;; return the contents of MEMBER in zip FILENAME as a string
(define-public (unzip-extract-member filename member)
  (read-input-pipe (string-append "unzip -p " filename " " member)))
