;; USA setups - NYSE, AMEX, NASDAQ.

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart usa)
  #:use-module (srfi srfi-13)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart misc)
  #:use-module (chart yahoo-currencies)
  #:use-module (chart yahoo))


;; US stock symbol is no "^", no "=X" and no suffix
(define-public (usa-symbol? symbol)
  (not (or (yahoo-index-symbol? symbol)
	   (yahoo-currency-symbol? symbol)
	   (string-any #\. symbol))))

;; ^TV.N - total volume NYSE
(define (nyse-symbol? symbol)
  (string-suffix? ".N" symbol))

;; ^TV.O - total volume NASDAQ
(define (nasdaq-symbol? symbol)
  (string-suffix? ".O" symbol))

(define (yahoo-index-symbol-usa? symbol)
  (or
   ;; http://finance.yahoo.com/indices?e=nasdaq -- and more below
   ;; ^IXBK - nasdaq banking
   ;; ^IXIC - nasdaq composite
   ;; ^IXF  - nasdaq financials
   ;; ^IXID - nasdaq industrials
   ;; ^IXIS - nasdaq insurance
   ;; ^IXQ  - nasdaq nnm composite
   ;; ^IXK  - nasdaq computer
   ;; ^IXFN - nasdaq other finances
   ;; ^IXUT - nasdaq telecommunications
   ;; ^IXTR - nasdaq transports
   (string-prefix? "^IX" symbol)

   (member symbol
	   '(
	     "^VIX"  ;; volatility
	     "^VXO"  ;; CBOE options volatility

	     ;; http://finance.yahoo.com/indices?e=dow_jones
	     "^DJA"  ;; Dow average
	     "^DJI"  ;; Dow
	     "^DJT"  ;; Dow transports
	     "^DJU"  ;; Dow utilities

	     ;; http://finance.yahoo.com/indices?e=new_york
	     "^NYA"  ;; NYSE composite
	     "^NIN"  ;; NYSE international
	     "^NTM"  ;; NYSE TMT
	     "^NUS"  ;; NYSE US 100
	     "^NWL"  ;; NYSE world leaders
	    
	     ;; http://finance.yahoo.com/indices?e=nasdaq
	     "^NBI"  ;; nasdaq biotech
	     "^NDX"  ;; nasdaq 100 drm

	     ;; http://finance.yahoo.com/indices?e=sp
	     "^OEX"    ;; S&P 100
	     "^MID"    ;; S&P 400 midcap
	     "^GSPC"   ;; S&P 500
	     "^SPSUPX" ;; S&P composite 1500
	     "^SML"    ;; S&P small 600

	     ;; http://finance.yahoo.com/indices?e=other
	     "^XAX"    ;; AMEX composite
	     "^IIX"    ;; AMEX internet
	     "^NWX"    ;; AMEX networking
	     "^DWC"    ;; DJ wilshire 5000 tot
	     "^XMI"    ;; major market index
	     "^PSE"    ;; NYSE arca tech 100
	     "^SOXX"   ;; Philadelphia semiconductor
	     "^DOT"    ;; Philadelphia/thestreet.com internet
	     "^RUI"    ;; Russell 1000
	     "^RUT"    ;; Russell 2000
	     "^RUA"    ;; Russell 3000

	     ;; http://finance.yahoo.com/indices?e=treasury
	     "^TNX"    ;; 10-year treasury bond interest rate
	     "^IRX"    ;; 13-week treasury note interest rate
	     "^TYX"    ;; 30-year treasury bond interest rate
	     "^FVX"    ;; 5-year treasury bond interest rate

	     ;; http://finance.yahoo.com/indices?e=commodities
	     "^DJC"    ;; Dow commodities
	     "^DJS2"   ;; Dow industrial avg settle
	     "^XAU"    ;; Philadelphia gold and silver index

	     ;; old bits
	     "^WIL5" ;; Wiltshire 5000 - gone, now ^DWC maybe
	     ))))

(symbol-timezone! usa-symbol?             (_ "New York") timezone-newyork)
(symbol-timezone! nyse-symbol?            (_ "New York") timezone-newyork)
(symbol-timezone! nasdaq-symbol?          (_ "New York") timezone-newyork)
(symbol-timezone! yahoo-index-symbol-usa? (_ "New York") timezone-newyork)
(yahoo-symbol-add! nyse-symbol?)
(yahoo-symbol-add! nasdaq-symbol?)

(yahoo-quote-delay! usa-symbol?
  (lambda (symbol exchange)
    ;; exchange is NasdaqSC, NasdaqNM, etc, not sure what the suffix bit is
    (if (string-prefix-ci? "nasdaq" exchange)
	15    ;; NASDAQ
	20))) ;; AMEX, NYSE


;; return `exchange' field from database info, or #f if not available
(define (usa-symbol-stock-exchange symbol)
  (and-let* ((series (database-read-series symbol)))
    (assq-ref (series-info series) 'stock-exchange)))

(weblink-handler! usa-symbol?
  (lambda (symbol)
    (let ((exchange (usa-symbol-stock-exchange symbol)))
      (cond ((equal? exchange "NYSE")
	     (list (_ "NYSE _Company Information")
		   (_ "Open web browser at the New York Stock Exchange page for this company")
		   (string-append "http://www.nyse.com/about/listed/"
				  (chart-symbol-sans-dot symbol) ".html")))

	    ((equal? exchange "AMEX")
	     (list (_ "AMEX _Company Information")
		   (_ "Open web browser at the American Stock Exchange page for this company")
		   (string-append "http://www.amex.com/?href=/equities/listCmp/EqLCCmpDesc.jsp?Product_Symbol=" (chart-symbol-sans-dot symbol))))

	    ((and exchange
		  ;; allow NasdaqNM, NasdaqSC etc
		  (string-prefix? "Nasdaq" exchange))
	     (list (_ "NASDAQ _Company Information")
		   (_ "Open web browser at the NASDAQ page for this company")
		   (string-append "http://quotes.nasdaq.com/asp/summaryquote.asp?symbol=" (chart-symbol-sans-dot symbol))))
	    (else
	     #f)))))
