;; Vertical Horizontal Filter (VHF) oscillator.

;; Copyright 2006, 2007, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart vhf)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 q)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a VHF oscillator over a window of
;; COUNT many points.
;;
;; Each call (PROC close) enters a new point into the window, and the return
;; is the VHF up to (and including) that point, or #f if not enough points
;; yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define*-public (vhf-calc-proc count)

  ;; Q is the close values, `q-front' being the oldest
  ;; POINTS is how many entries in Q
  ;; SUM-PROC adds up the N-1 successive differences in Q
  ;;
  (let ((q          (make-q))
	(points     0)
	(sum-proc   (sum-calc-proc (1- count)))
	(points-min (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (close)
      (define abs-diffs (if (q-empty? q)
			    0.0
			    (sum-proc (abs (- close (q-rear q))))))

      ;; drop old point from queue if full
      (if (>= points count)
	  (deq! q)
	  ;; not full, so POINTS increases
	  (set! points (1+ points)))

      ;; add new point
      (enq! q close)

      (and (>= points points-min)
	   abs-diffs
	   (not (zero? abs-diffs))

	   (/ (- (apply max (q-list q))
		 (apply min (q-list q)))
	      (exact->inexact abs-diffs))))))

(define (vhf-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count count))
	 (array  (series-array-closes parent lo hi))
	 (vhf    (vhf-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((this (array-ref array i)))
	(vhf this)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (vhf this) i)))))

(define-public (vhf-series series count)

  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "VHF") count))
    (series-datatype!  new-series vhf-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts vhf-fill count))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define-public vhf-datatype
  (datatype-new!
   #:key     'vhf
   #:name    (_ "VHF - Vertical Horizontal Filter")
   #:units   'vhf
   #:minimum 0
   #:maximum 1
   #:manual  (_ "CONTEXT(manual-node): Vertical Horizontal Filter")
   #:proc    vhf-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'vhf-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 2
	  #:default 20))))
