;; VIDYA - Variable Index Dynamic Average.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart vidya)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc)
  #:use-module (chart stddev))


;; http://trader.online.pl/ELZ/t-i-VIDYA.html
;;     Formulas and S&P500 sample chart.
;;
;; http://www.fmlabs.com/reference/VIDYA.htm
;;
;; http://www.working-money.com/documentation/FEEDbk_docs/Archive/0298/TradersTips/Tips9802.html
;;     TASC trader's tips February 1998, formulas.
;;


;;-----------------------------------------------------------------------------
;; calc

;; Return a procedure which calculates alpha factors for a VIDYA moving
;; average on an accumulated window of SLOW-COUNT and FAST-COUNT values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the VIDYA up to (and including) that value, or #f if not enough values
;; yet.
;;
;; To prime the window initially, call PROC with SLOW-COUNT-1 many points
;; preceding the first desired.
;;
;;
;; Not sure if the ratio could go above 5, to make an alpha above 1,
;; probably yes with the right data and a small FAST-COUNT, so cap it at 1.
;;
(define-public (vidya-alpha-calc-proc fast-count slow-count)

  (define slow-proc (stddev-calc-proc slow-count))
  (define fast-proc (stddev-calc-proc fast-count))

  (lambda (x)
    (let ((slow (slow-proc x))
	  (fast (fast-proc x)))
      (and slow
	   (if (zero? slow)
	       0
	       (min 1 (* 0.2 (/ fast slow))))))))


;; Return a procedure which calculates a VIDYA moving average, with alpha
;; factors calculated on an accumulated window of SLOW-COUNT and FAST-COUNT
;; values.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the VIDYA up to (and including) that value, or #f if not enough values
;; yet.
;;
(define-public (vidya-calc-proc fast-count slow-count)

  (define vidya-alpha-proc (vidya-alpha-calc-proc fast-count slow-count))
  (define ama-proc         (adaptive-ema-calc-proc))

  (lambda (x)
    (and-let* ((alpha (vidya-alpha-proc x)))
      (ama-proc x alpha))))


(define guess-warmup
  (ema-warmup-count (ema-alpha->count 0.1)))

(define (vidya-estimate-warmup fast-count slow-count)
  (+ (1- slow-count) guess-warmup))


;;-----------------------------------------------------------------------------
;; VIDYA alpha series

(define (vidya-alpha-fill series new-array lo hi)
  (receive-list (fast-count slow-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (array  (series-array-closes parent lo hi))
	   (vidya   (vidya-alpha-calc-proc fast-count slow-count)))

      ;; warmup
      (for-each vidya (series-prev-n-closes parent
					    #:before lo
					    #:count (1- slow-count)))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (vidya (array-ref array i)) i))))))

(define-public (vidya-alpha-series series fast-count slow-count)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a,~a" (_ "VIDYA alpha")
					  fast-count slow-count))
    (series-datatype!  new-series vidya-alpha-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list fast-count slow-count))
    (series-require!   new-series (indicator-require-parts vidya-alpha-fill
							   (1- slow-count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-notes!     new-series '())
    (series-dividends! new-series '())
    new-series))

(define-public vidya-alpha-datatype
  (datatype-new!
   #:key      'vidya-alpha
   #:name     (_ "VIDYA alpha")
   #:priority -10
   #:units    'ema-alpha
   #:minimum  0
   #:maximum  1.0
   #:manual   (_ "CONTEXT(manual-node): Variable Index Dynamic Average")
   #:proc     vidya-alpha-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Fast days")
	  #:key     'vidya-fast-days
	  #:type    'integer
	  #:minimum 1
	  #:default 9)
	 (datatype-param-new
	  #:name    (_ "Slow days")
	  #:key     'vidya-slow-days
	  #:type    'integer
	  #:minimum 1
	  #:default 30))))


;;-----------------------------------------------------------------------------
;; VIDYA series

(define (vidya-fill series new-array lo hi)
  (receive-list (fast-count slow-count)
      (series-data series)
    (let* ((parent (series-parent series))
	   (start  (adaptive-ema-find-start
		    parent lo (vidya-alpha-calc-proc fast-count slow-count)))
	   (array  (series-array-closes parent start hi))
	   (vidya  (vidya-calc-proc fast-count slow-count)))

      ;; warmup start (inclusive) to lo (exclusive)
      (do ((i start (1+ i)))
	  ((>= i lo))
	(if (array-ref array i)
	    (vidya (array-ref array i))))

      ;; points lo to hi inclusive
      (do ((i lo (1+ i)))
	  ((> i hi))
	(if (array-ref array i)
	    (array-set! new-array (vidya (array-ref array i)) i))))))

(define-public (vidya-series series fast-count slow-count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a,~a" (_ "VIDYA")
					 fast-count slow-count))
    (series-datatype! new-series vidya-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series (list fast-count slow-count))
    (series-require!  new-series
		      (indicator-require-parts
		       vidya-fill (vidya-estimate-warmup slow-count
							 fast-count)))
    (series-array!    new-series #f)

    new-series))

(define vidya-datatype
  (datatype-new!
   #:key    'vidya
   #:name   (_ "VIDYA - Variable Index Dynamic")
   #:units  #f
   #:manual (_ "CONTEXT(manual-node): Variable Index Dynamic Average")
   #:proc   vidya-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Fast days")
	  #:key     'vidya-fast-days
	  #:type    'integer
	  #:minimum 1
	  #:default 9)
	 (datatype-param-new
	  #:name    (_ "Slow days")
	  #:key     'vidya-slow-days
	  #:type    'integer
	  #:minimum 1
	  #:default 30))))
