;; ChartView widget.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart view)
  #:use-module (gtk gtk)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:autoload   (chart adjust)      (adjust-series)
  #:autoload   (chart database)    (database-read-series)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart gui-common)
  #:use-module (chart heading)
  #:use-module (chart hscale)
  #:use-module (chart keymap)
  #:use-module (chart misc)
  #:use-module (chart symlists)
  #:autoload   (chart ticker)      (ticker-new)
  #:use-module (chart vscale))


;; set by main
(define-public (view-message fmt . args)
  #f)
''(guile-lint simple-format view-message)


;;-----------------------------------------------------------------------------
;; GtkTable stuff

(define (c-gtk-table-reattach table child . args)
  (if (gtk-widget-get-parent child)
      (gtk-container-remove table child))
  (apply gtk-table-attach table child args))


;;-----------------------------------------------------------------------------
;; GtkAdjustment inversion
;;
;; `c-gtk-adjustment-derive-invert' returns a new GtkAdjustment which tracks
;; the given ADJ but inverts the value.  Value changes in ADJ are propagated
;; to INV, and vice versa.  Range changes should be applied to ADJ, range
;; changes are propagated only fro ADJ to INV, not back.
;;
;; A GtkVScrollbar on ADJ would show lower at the top and upper at the
;; bottom.  Basing it on an inverse lets it work upper at the top and lower
;; at the bottom, as far as the value propagated into ADJ is concerned.
;;
;; This ends up with some flashing as the scrollbar changes its slider
;; window size immediately based on an update.

(define (c-gtk-adjustment-get-value-inverted adj)
  (+ (gtk-adjustment-lower adj)
     (- (gtk-adjustment-upper adj)
	(gtk-adjustment-value adj)
	(gtk-adjustment-page-size adj))))

(define (c-gtk-adjustment-derive-invert adj)
  (let ((inv    (gtk-adjustment-new (c-gtk-adjustment-get-value-inverted adj)
				    (gtk-adjustment-lower adj)
				    (gtk-adjustment-upper adj)
				    (gtk-adjustment-step-increment adj)
				    (gtk-adjustment-page-increment adj)
				    (gtk-adjustment-page-size adj)))
	(inv-id #f)
	(adj-id #f))

    (define (adj-value-changed)
      (gtk-signal-disconnect inv inv-id)
      (gtk-adjustment-set-value inv (c-gtk-adjustment-get-value-inverted adj))
      (set! inv-id (gtk-signal-connect inv "value-changed"
		     inv-value-changed)))

    (define (inv-value-changed)
      (gtk-signal-disconnect adj adj-id)
      (gtk-adjustment-set-value adj (c-gtk-adjustment-get-value-inverted inv))
      (set! adj-id (gtk-signal-connect adj "value-changed"
		     adj-value-changed)))

    (set! inv-id (gtk-signal-connect inv "value-changed" inv-value-changed))
    (set! adj-id (gtk-signal-connect adj "value-changed" adj-value-changed))

    ;; when ADJ changes, propagate same to INV, including update value there
    (gtk-signal-connect adj "changed"
      (lambda ()
	(c-gtk-adjustment-change
	 inv
	 #:lower 	  (gtk-adjustment-lower          adj)
	 #:upper 	  (gtk-adjustment-upper          adj)
	 #:step-increment (gtk-adjustment-step-increment adj)
	 #:page-increment (gtk-adjustment-page-increment adj)
	 #:page-size      (gtk-adjustment-page-size      adj))
	(adj-value-changed)))

    inv))
    

;;-----------------------------------------------------------------------------
;; GtkWidget stuff

;; not wrapped in guile-gtk 2.0
(if (not (defined? 'gtk-widget-set-direction))
    (define (gtk-widget-set-direction widget dir)
      #f))


;;-----------------------------------------------------------------------------
;; series stuff

;; return a list (LO HI) which is the date range to display for all the
;; series in SERIES-LIST
;;
(define (series-list-scroll-date-range series-list)
  (let ((range-list (map series-scroll-date-range series-list)))
    (list (apply min (map first range-list))
	  (apply max (map second range-list)))))

;; return a list (LO HI) which is the date range to display for SERIES
;; this covers the trading data, plus nearby past or future dividends, text
;; notes and lines
;;
(define (series-scroll-date-range series)
  (let* ((graph         draw)
	 (layout        (primitive-property-ref c-layout-prop graph))
	 (scale-x       (graph-scale-x-proc graph))
	 (x-margin      (round-exact (* 0.8 (c-pango-layout-em layout))))
	 (lo            (round-exact (gtk-adjustment-value
				      (graph-hadjustment graph))))
	 (tdate->t      (timebase-convert-proc timebase-days
					       (series-timebase series)))
	 (range-lo      (series-lo series))
	 (range-hi      (series-hi series))
	 (note-lo-limit (- (series-lo series) 20))
	 (note-hi-limit (+ (series-hi series) 20)))

    (define (x->date x)
      (+ lo (/ x (exact->inexact scale-x-step))))

    ;; expand RANGE-LO,RANGE-HI for an object drawn between dates LO to HI
    ;; objects like dividends too far outside the actual data are ignored
    (define (note-range lo hi)
      (if (and (>= hi note-lo-limit)
	       (<= lo note-hi-limit))
	  (begin
	    (set! range-lo (min range-lo lo))
	    (set! range-hi (max range-hi hi)))))

    ;; expand RANGE-LO,RANGE-HI for a string STR drawn at date T
    ;; width calculation as per `draw-notes'
    (define (note-str t str)
      (let* ((str-width (c-pango-layout-width layout str))
	     (x-centre  (+ (scale-x t)
			   (quotient scale-x-step 2)))
	     (radius    (+ (quotient str-width 2) x-margin)))
	(note-range (floor-exact   (x->date (- x-centre radius)))
		    (ceiling-exact (x->date (+ x-centre radius))))))

    (for-each (lambda (div)
		(note-str (car div) (dividend->string (cdr div))))
	      (series-dividends series))

    (for-each (lambda (note)
		(cond ((annotation? note)
		       (note-str (tdate->t (second note)) (third note)))
		      ((note-line? note)
		       (note-range (line-t1 note)
				   (line-t2 note)))))
	      (series-notes series))

    (list range-lo range-hi)))

    
;;-----------------------------------------------------------------------------
;;
;;   0              1  2  3
;; 0 +--------------+--+--+
;;   | heading            |
;; 1 +--------------+--+--+
;;   |              |v |v |
;;   | upper        |s |s |
;;   |              |c |c |
;;   |              |a |r |
;;   |              |l |o |
;;   |              |e |l |
;;   |              |  |l |
;; 4 +--------------+--+--+
;;   |              |v |v |
;;   | lower        |s |s |
;;   |              |c |b |
;; 5 +--------------+--+--+
;;   | hscale       |     |
;; 6 +--------------+     |
;;   | hscroll      |     |
;; 7 +--------------+--+--+
;;   | ticker             |
;; 8 +--------------------+

(define-public draw #f)
(define-public drawvol #f)

(define-public initial-scale-x-step 6)
(define-public scale-x-step initial-scale-x-step)


(gtk-class-new 'GtkTable "ChartView")
(gtk-signal-new-generic "series-changed" '() 'ChartView 'void '())

(let-toplevel ((rec-type (make-record-type "chartview"
					   '(hadjustment
					     vol-widgets
					     vol-state
					     symbol
					     series
					     graph-list
					     timebase
					     smarker
					     heading
					     ticker ticker-state))))
  (define rec-new     	   (record-constructor rec-type '(graph-list smarker)))
  (define rec-series  	   (rec-accessor rec-type 'series))
  (define rec-series! 	   (rec-modifier rec-type 'series))
  (define rec-symbol  	   (rec-accessor rec-type 'symbol))
  (define rec-symbol! 	   (rec-modifier rec-type 'symbol))
  (define rec-vol-widgets  (rec-accessor rec-type 'vol-widgets))
  (define rec-vol-widgets! (rec-modifier rec-type 'vol-widgets))
  (define rec-hadjustment  (rec-accessor rec-type 'hadjustment))
  (define rec-hadjustment! (rec-modifier rec-type 'hadjustment))
  (define rec-graph-list   (rec-accessor rec-type 'graph-list))
  (define rec-graph-list!  (rec-modifier rec-type 'graph-list))
  (define rec-timebase     (rec-accessor rec-type 'timebase))
  (define rec-timebase!    (rec-modifier rec-type 'timebase))
  (define rec-smarker      (rec-accessor rec-type 'smarker))
  (define rec-smarker!     (rec-modifier rec-type 'smarker))
  (define rec-heading      (rec-accessor rec-type 'heading))
  (define rec-heading!     (rec-modifier rec-type 'heading))
  (define rec-ticker       (rec-accessor rec-type 'ticker))
  (define rec-ticker!      (rec-modifier rec-type 'ticker))
  (define rec-ticker-state  (rec-accessor rec-type 'ticker-state))
  (define rec-ticker-state! (rec-modifier rec-type 'ticker-state))

  ;; choose and set an initial X coordinate range and position
  (define (x-range-initial view)
    (let* ((series (rec-series (c-prop view)))
	   (x      (and series (series-hi series))))
      (x-range-update-from view x x #t)))

  (define* (x-range-update view #:optional (old-timebase (view-timebase view)))
    (let* ((hadj     (view-hadjustment view))
	   (upper    (gtk-adjustment-upper hadj))
	   (page     (gtk-adjustment-page-size hadj))
	   (value    (gtk-adjustment-value hadj))
	   (at-end   (>= (+ value page) upper))
	   (timebase (view-timebase view)))
      (define (conv val)
	(timebase-convert old-timebase timebase (round-exact val)))
      (x-range-update-from view (conv page) (conv value) at-end)))

  (define (x-range-update-from view old-page old-value old-end)
    (and-let* ((top-series (rec-series (c-prop view))))
      (let* ((graph-list  (view-graph-list view))
	     (graph       (first graph-list))
	     (win         (gtk-widget-window graph))
	     (width       (c-gdk-drawable-get-width win))
	     (page        (quotient width scale-x-step))
	     (value       (+ old-value (/ (- old-page page) 2))) ;; centred
	     (hadj        (view-hadjustment view))
	     (series-list (cons top-series
				(append-map graph-series-list graph-list))))
	(receive-list (lo hi)
	    (series-list-scroll-date-range series-list)

	  ;; bit of margin
	  (set! lo (- lo 3))
	  (set! hi (+ hi 3))

	  ;; if hi is close to today then extend to there
	  (let ((today (timebase-today (view-timebase view))))
	    (if (< (- today hi) 10)
		(set! hi (max hi (1+ today)))))

	  ;; stretch to page if necessary by extending left (if page grown)
	  (set! lo (min lo (- hi page -1)))

	  ;; if value was rightmost, then stay there when shrinking
	  (if old-end
	      (set! value hi))

	  (c-gtk-adjustment-change
	   hadj
	   #:lower          lo
	   #:upper          hi
	   #:page-size      page
	   #:step-increment (* scrollbar-step-fraction page)
	   #:page-increment (* scrollbar-page-fraction page))
	  (c-gtk-adjustment-set-value-inpage hadj value)))))

  (define-public (view-new)
    (let ((view      (gtk-widget-new 'ChartView
				     #:n_columns 3 #:n_rows 8
				     #:column_spacing 0 #:row_spacing 0))
      	  (rec       (rec-new '() (smarker-new)))
	  (size-list '()))

      (define (refresh-this)
	(view-refresh-series view))

      (set! (c-prop view) rec)
      (rec-timebase! rec timebase-days)

      ;; only setup for left-to-right at the moment
      (gtk-widget-set-direction view 'ltr)

      (let ((hadj (gtk-adjustment-new 0 0 0 0 0 0)))
	(rec-hadjustment! rec hadj))

      (let ((vadj (gtk-adjustment-new 0 0 0 0 0 0)))
	(set! draw (graph-new vadj))
	(let ((vscale     (vscale-new draw))
	      (vscrollbar (gtk-widget-new
			   'GtkVScrollbar
			   #:adjustment (c-gtk-adjustment-derive-invert vadj)
			   #:can_focus #f
			   #:update_policy 'continuous)))
	  (set! size-list (list (list draw vscale vscrollbar)))))

      (c-gtk-widget-add-events draw 'view-keymap '(button-press-mask))
      (gtk-signal-connect draw "button-press-event"
	(lambda (event)
	  (keymap-dispatch-button draw event)))
      (gtk-signal-connect draw "configure-event"
	(lambda (event)
	  (x-range-update view)))

      (let ((heading (heading-new draw)))
	(rec-heading! rec heading)
	(gtk-table-attach view heading 0 3 0 1 '(fill shrink expand) '())
	(gtk-widget-show heading))

      (let ((vadj (gtk-adjustment-new 0 0 0 0 0 0)))
	(set! drawvol (graph-new vadj))

	(let ((vscale  (vscale-new drawvol))
	      (vscroll (gtk-widget-new
			'GtkVScrollbar
			#:adjustment (c-gtk-adjustment-derive-invert vadj)
			#:can_focus #f
			#:update_policy 'continuous)))

	  (rec-graph-list! rec (list draw drawvol))

	  (c-gtk-widget-add-events drawvol 'view-keymap '(button-press-mask))
	  (gtk-signal-connect drawvol "button-press-event"
	    (lambda (event)
	      (keymap-dispatch-button drawvol event)))

	  (set! size-list (append size-list
				  (list (list drawvol vscale vscroll))))
	  (rec-vol-widgets! rec (list drawvol vscale vscroll))))

      (let ((hscale     (hscale-new (rec-hadjustment rec)))
	    (hscrollbar (gtk-widget-new 'GtkHScrollbar
					#:adjustment (rec-hadjustment rec)
					#:update_policy 'continuous
					#:can_focus #f))
	    (fill       (gtk-drawing-area-new)))
	;; (gtk-widget-new 'GtkDrawingArea #:width -1 #:height -1))

	(let ((old-sizes #f))
	  (define (window-size lst default)
	    (let ((size (assq-ref lst 'window-size)))
	      (if size
		  (max 1 (first size))
		  default)))
	  (define (triplet lst pos size pad)
	    (c-gtk-table-reattach view (first lst) 0 1
				  pos (+ pos size)
				  '(fill shrink expand) '(fill shrink expand)
				  0 pad)
	    (c-gtk-table-reattach view (second lst) 1 2
				  pos (+ pos size)
				  '(fill shrink) '(fill expand)
				  0 pad)
	    (c-gtk-table-reattach view (third lst) 2 3
				  pos (+ pos size)
				  '(fill shrink) '(expand fill shrink)
				  0 pad))
	  (define (graph-reattach . dummy)
	    (let* ((viewstyle  (preference-get 'viewstyle viewstyle-default))
		   (upper-size (window-size (cdr (first viewstyle)) 3))
		   (lower-size (window-size (second viewstyle) 1)))

	      (if (not (equal? old-sizes (list upper-size lower-size)))
		  (let ((pos 1))
		    (set! old-sizes (list upper-size lower-size))

		    '(for-each (lambda (child)
				 (c-gtk-container-remove-maybe view child))
			       (concatenate size-list))
		    '(c-gtk-container-remove-maybe view hscale)
		    '(c-gtk-container-remove-maybe view hscrollbar)
		    '(c-gtk-container-remove-maybe view fill)

		    (triplet (first size-list) pos upper-size 2)
		    (set! pos (+ pos upper-size))
		    (triplet (second size-list) pos lower-size 0)
		    (set! pos (+ pos lower-size))

		    (c-gtk-table-reattach view hscale 0 1 pos (1+ pos)
					  '(fill shrink expand) '(fill shrink))
		    (c-gtk-table-reattach view fill 1 3 pos (+ 2 pos)
					  '(fill shrink) '(fill shrink))
		    (set! pos (1+ pos))

		    (c-gtk-table-reattach view hscrollbar 0 1 pos (1+ pos)
					  '(expand fill shrink) '(fill shrink))
		    (set! pos (1+ pos))

		    (if (rec-ticker rec)
			(c-gtk-table-reattach view (rec-ticker rec)
					      0 3 pos (1+ pos)
					      '(expand fill shrink) '(fill shrink)))))))

	  (graph-reattach)
	  (notify-connect 'preferences-save
	    (lambda (dummy)
	      (c-gtk-idle-once graph-reattach)))
	  (notify-connect 'view-ticker-creation
	    (lambda (affected-view)
	      (set! old-sizes #f)
	      (c-gtk-idle-once graph-reattach)))))

      (notify-connect 'database-update
	(lambda (symbol-list)
	  (if (member (view-symbol view) symbol-list)
	      (c-gtk-idle-once refresh-this))))

      (notify-connect 'preferences-save
	(lambda (dummy)
	  (c-gtk-idle-once refresh-this)))

      view))

  (define-public (view-timebase view)
    (rec-timebase (c-prop view)))
  (define-public (view-timebase! view timebase)
    (rec-timebase! (c-prop view) timebase)
    (view-refresh-series view))

  ;; return stock symbol being displayed by VIEW, or #f if none
  (define-public (view-symbol view)
    (rec-symbol (c-prop view)))


  (define*-public (view-open view symbol #:key smarker)
    (define rec (c-prop view))
    (graph-series-list! draw '())
    (graph-series-list! drawvol '())
    (rec-symbol! (c-prop view) symbol)
    (rec-series! rec (database-read-series symbol))
    (x-range-initial view)
    (c-gtk-adjustment-end-up (rec-hadjustment rec))
    (view-refresh-series view)
    (x-range-update view)) ;; in case bigger derived series

  (define-toplevel (view-refresh-series view)
    (and-let* ((symbol (view-symbol view)))
      (let* ((series       (database-read-series symbol))
	     (rec          (c-prop view))
	     (old-series   (rec-series rec))
	     (old-timebase (and old-series (series-timebase old-series))))
	(graph-series-list! draw '())
	(graph-series-list! drawvol '())

	(if series
	    (let* ((viewstyle    (preference-get 'viewstyle viewstyle-default))
		   (adj-splits     (viewstyle-split viewstyle))
		   (adj-dividends  (viewstyle-dividend viewstyle))
		   (adj-imputation (and adj-dividends
					(viewstyle-imputation viewstyle)))
		   (adj-rollovers  (viewstyle-rollover viewstyle)))
	      (set! series (adjust-series series adj-splits adj-dividends
					  adj-imputation adj-rollovers))

	      (set! series (series-convert-timebase series
						    (view-timebase view)))
	      (rec-series! rec series)

	      (if old-timebase
		  (x-range-update view old-timebase)
		  (x-range-initial view))

	      ;; upper
	      (graph-series-list! draw (cons series
					     (dspec-list->series-list
					      (cdr (first viewstyle))
					      series)))
	      ;; lower
	      (and-let* ((dspec-list (second viewstyle)))
		(let ((volseries (dspec->series (first dspec-list)
						series)))
		  (for-each (if volseries gtk-widget-show gtk-widget-hide)
			    (rec-vol-widgets rec))
		  (if volseries
		      (graph-series-list! drawvol
					  (cons volseries
						(dspec-list->series-list
						 (cdr dspec-list) volseries)))))))))
      (gtk-signal-emit view "series-changed")))

  (define-public (view-hadjustment view)
    (rec-hadjustment (c-prop view)))

  (define-public (view-heading view)
    (rec-heading (c-prop view)))

  (define-public (view-graph-list view)
    (rec-graph-list (c-prop view)))

  (define-public (view-smarker view)
    (rec-smarker (c-prop view)))
  (define-public (view-smarker! view smarker)
    (rec-smarker! (c-prop view) (smarker-copy smarker)))

  (define-public (view-scale-x-step! view val)
    (set! scale-x-step val)
    (x-range-update view))

  (define-public (gtk-table-resize table rows cols)
    (gtk-object-set table #:n_rows rows)
    (gtk-object-set table #:n_columns cols))

  (define-public (view-ticker-state view)
    (rec-ticker-state (c-prop view)))

  (define-public (view-ticker-state! view state)
    (set! state (->bool state))
    (let ((rec (c-prop view)))
      (if (not (eq? state (rec-ticker-state rec)))
	  (let ((ticker (rec-ticker rec)))
	    (if (not ticker)
		(let ()
		  (define (menu-additions menu)
		    (let ((item (gtk-menu-item-new-with-mnemonic (_ "Hide"))))
		      (gtk-menu-shell-append menu item)
		      (gtk-signal-connect item "activate"
			(lambda ()
			  (view-ticker-state! view #f))))
		    (let ((item (gtk-image-menu-item-new-from-stock "gtk-help"
								    #f)))
		      (gtk-menu-shell-append menu item)
		      (gtk-signal-connect item "activate"
			(lambda ()
			  (manual-popup (_ "CONTEXT(manual-node): Ticker")
					view)))))
		  (set! ticker (ticker-new menu-additions))
		  (gtk-widget-set ticker #:can_focus #f) ;; no steal keys etc
		  (gtk-widget-grab-focus view)
		  (rec-ticker! rec ticker)
		  (notify-send-locally 'view-ticker-creation view)))

	    ;; get it attached at the right position
	    (rec-ticker-state! rec state)
	    (notify-send-locally 'view-ticker-state-update view)
	    (if state
		(gtk-widget-show ticker)
		(gtk-widget-hide ticker)))))))

;; return true if the given VIEWSTYLE has the adjust split option set
(define-public (viewstyle-split viewstyle)
  (let ((elem (assq 'split (cdr (first viewstyle)))))
    (if elem
	(second elem)
	#t)))

;; return true if the given VIEWSTYLE has the adjust dividend option set
(define-public (viewstyle-dividend viewstyle)
  (let ((elem (assq 'dividend (cdr (first viewstyle)))))
    (if elem
	(second elem)
	#f)))  ;; default no

;; return true if the given VIEWSTYLE has the adjust imputation option set
(define-public (viewstyle-imputation viewstyle)
  (let ((elem (assq 'imputation (cdr (first viewstyle)))))
    (if elem
	(second elem)
	#f)))  ;; default no

;; return true if the given VIEWSTYLE has the adjust rollovers option set
(define-public (viewstyle-rollover viewstyle)
  (let ((elem (assq 'rollover (cdr (first viewstyle)))))
    (if elem
	(second elem)
	#t)))  ;; default yes


;; return the vadjustment for the main graph in VIEW
(define-public (view-vadjustment view)
  (graph-vadjustment (first (view-graph-list view))))

(define-public (view-centre-vertical view)
  (for-each graph-centre-vertical (view-graph-list view)))

;; return a series derived from SERIES according to DSPEC, or #f if no such
(define (dspec->series dspec series)
  (datatype-list-all)
  (let ((dtkey      (first dspec))
	(param-list (cdr dspec)))
    (and-let* ((         (not (eq? 'none dtkey)))
	       (datatype (datatype-find dtkey)))

      (and-let* ((applyto (assq 'applyto param-list))
		 (datatype (datatype-find (second applyto))))
	(set! series ((datatype-proc datatype) series)))

      (apply (datatype-proc datatype)
	     series
	     (map (lambda (param)
		    ;; parameter value from dspec, or default
		    (or (any (lambda (elem)
			       (and (eq? (first elem)
					 (datatype-param-key param))
				    (second elem)))
			     param-list)
			(datatype-param-default param)))
		  (datatype-parameter-list datatype))))))

;; return a list of series derived from SERIES according to DSPEC-LIST
;; DSPEC-LIST entries which are unknown are ignored
(define (dspec-list->series-list dspec-list series)
  (filter-map (lambda (dspec)
		(dspec->series dspec series))
	      dspec-list))
