;; View style dialog.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart viewstyle)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart gui-common)
  #:use-module (chart misc)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; misc

;; Placeholder for when there's some actual utf-8 in datatype names.  Don't
;; really want to depend on glib g_utf8_collate (currently can build C code
;; without gtk etc).  And jamming the C library into utf8 is tricky.
;;
(define utf8-string-ci<? string-ci<?)

;; return key matching VAL in ALIST, compared with `eqv?'
(define-public (rassv-ref alist val)
  (and (not (null? alist))
       (if (eqv? val (cdar alist))
	   (caar alist)
	   (rassv-ref (cdr alist) val))))


;;-----------------------------------------------------------------------------
;; GtkOptionMenu extras

;; do `gtk-option-menu-set-history' if IDX isn't already current in OPTMENU
(define (c-gtk-option-menu-set-history-maybe optmenu idx)
  (or (= idx (gtk-option-menu-get-history optmenu))
      (gtk-option-menu-set-history optmenu idx)))


;;-----------------------------------------------------------------------------
;; datatype stuff

;; sort datatype records first by priority, then name
(define-public (datatype<? x y)
  (or (< (datatype-priority y) (datatype-priority x))
      (and (= (datatype-priority x) (datatype-priority y))
	   (utf8-string-ci<? (datatype-name x) (datatype-name y)))))

(define (datatype-hidden? datatype)
  (<= (datatype-priority datatype) -1000))

;; return true if DATATYPE has its own units
(define-public (datatype-own-units? datatype)
  (or (eq? (datatype-key datatype) 'none)
      (and (not (datatype-hidden? datatype))
	   (datatype-units datatype))))

;; return true if DATATYPE uses the units of its parent (like an MA)
(define-public (datatype-parent-units? datatype)
  (or (eq? (datatype-key datatype) 'none)
      (and (not (datatype-hidden? datatype))
	   (eq? #f (datatype-units datatype)))))

;; return true if DATATYPE uses the units of its parent (like an MA), and
;; can cope with a non-OHLCV input
;; in practice this means all the moving averages, and none of the channels
;;
(define-public (datatype-parent-units-non-ohlcv? datatype)
  (and (datatype-parent-units? datatype)
       (not (eq? datatype-ohlcv (datatype-input datatype)))))


;;-----------------------------------------------------------------------------
;; datatypes option menu

;; create and return a GtkOptionMenu containing DATATYPE-LIST
(define (dtoptmenu-new datatype-list)
  (let ((dtoptmenu     (gtk-option-menu-new))
	(menu          (gtk-menu-new))
	(prev-priority (datatype-priority (first datatype-list)))
	(idx           0)
	(alist         '()))  ;; elems (IDX . DATATYPE)

    (for-each (lambda (datatype)
		;; separator between priority groups
		(or (eqv? prev-priority (datatype-priority datatype))
		    (begin
		      (gtk-menu-shell-append menu
					     (gtk-separator-menu-item-new))
		      (set! idx (1+ idx))))
		(set! prev-priority (datatype-priority datatype))

		(let ((item (gtk-menu-item-new-with-label
			     (datatype-name datatype))))
		  (gtk-widget-show item)
		  (gtk-menu-append menu item)
		  (set! alist (acons idx datatype alist))
		  (set! idx (1+ idx))))

	      datatype-list)

    (gtk-option-menu-set-menu dtoptmenu menu)
    (gtk-option-menu-set-history dtoptmenu 0)
    (set! (c-prop dtoptmenu) alist)
    dtoptmenu))

;; return the currently selected datatype in OPTMENU
(define (dtoptmenu-datatype dtoptmenu)
  (assv-ref (c-prop dtoptmenu)
	    (gtk-option-menu-get-history dtoptmenu)))

;; set the currently selected datatype in DTOPTMENU, by its datatype key
;; if there's no such datatype then quietly do nothing
;;
(define (dtoptmenu-datatype-key! dtoptmenu key)
  (let ((idx (any (lambda (pair)
		    (and (eq? key (datatype-key (cdr pair)))
			 (car pair)))
		  (c-prop dtoptmenu))))
    (if idx
	(gtk-option-menu-set-history dtoptmenu idx)
	(format (current-error-port)
		"viewstyle: datatype `~a' unknown, ignoring\n" key))))


;;-----------------------------------------------------------------------------
;; linestyles option menu

;; fields of a linestyle element in chartdraw-linestyle-list
(define linestyle-key  first)
(define linestyle-name third)

;; return true if X < Y for two linestyle elements from
;; chartdraw-linestyle-list, compared by name
(define (linestyle-name<? x y)
  (utf8-string-ci<? (linestyle-name x) (linestyle-name y)))

(define (lsoptmenu-new)
  (let ((lsoptmenu (gtk-option-menu-new))
	(menu      (gtk-menu-new))
	(lslst     (sort (filter linestyle-name ;; only elements with names
				 chartdraw-linestyle-list)
			 linestyle-name<?))
	(alist     '()) ;; elements (IDX . LINESTYLE-KEY)
	(idx       0))

    (for-each (lambda (linestyle)
		(let ((item (gtk-menu-item-new-with-label
			     (linestyle-name linestyle))))
		  (gtk-widget-show item)
		  (gtk-menu-append menu item)
		  (set! alist (acons idx (linestyle-key linestyle) alist))
		  (set! idx (1+ idx))))
	      lslst)

    (gtk-option-menu-set-menu lsoptmenu menu)
    (gtk-option-menu-set-history lsoptmenu 0)
    (set! (c-prop lsoptmenu) alist)
    lsoptmenu))

;; return the currently selected linestyle key (a scheme symbol) in LSOPTMENU
(define (lsoptmenu-linestyle lsoptmenu)
  (assv-ref (c-prop lsoptmenu)
	    (gtk-option-menu-get-history lsoptmenu)))

;; set the currently selected linestyle in LSOPTMENU, by its linestyle key
;; (a scheme symbol); if there's no such datatype then quietly do nothing
(define (lsoptmenu-linestyle-key! lsoptmenu key)
  (let ((idx (rassv-ref (c-prop lsoptmenu) key)))
    (if idx
	(c-gtk-option-menu-set-history-maybe lsoptmenu idx)
	(format (current-error-port)
		"viewstyle: linestyle `~a' unknown, ignoring\n" key))))


;;-----------------------------------------------------------------------------
;; datatype info button

;; create and return a "gtk-info" button which opens the manual at the
;; datatype selected in DTOPTMENU
;;
(define (infobutton-new dtoptmenu)
  (let ((info (gtk-button-new-from-stock "gtk-info")))

    ;; button is sensitive when the DTOPTMENU datatype has a `datatype-manual'
    (define (update-sensitive)
      (gtk-widget-set-sensitive
       info (datatype-manual (dtoptmenu-datatype dtoptmenu))))

    (gtk-tooltips-set-tip
     tooltips info
     (_ "Open the manual at the section describing this indicator")
     "")

    ;; initial sensitive, then when menu choice changes
    (update-sensitive)
    (gtk-signal-connect dtoptmenu "changed" update-sensitive)

    (gtk-signal-connect info "clicked"
      (lambda ()
	(manual-popup (datatype-manual (dtoptmenu-datatype dtoptmenu))
		      dialog))) ;; global dialog variable
    info))


;;-----------------------------------------------------------------------------
;; ChartEditNum widget
;;
;; An editable number, done as a GtkSpinButton with some tweaks.
;;

(gtk-signal-new-generic "user-changed" '() 'GtkSpinButton 'void '())

(define (=-close? x y)
  (<= (abs (- x y)) 0.001))

(define edit-programmatic-prop (make-object-property))

(define (number-num-digits n)
  (string-length (number->string n)))

(define (edit-num-new lo hi step page decimals)
  (let* ((adj  (gtk-adjustment-new lo lo hi step page 0))
	 (edit (gtk-spin-button-new adj 0 decimals)))

    (gtk-spin-button-set-numeric edit #t)
    (c-gtk-widget-set-usize-units edit
				  (cons 'em (+ (number-num-digits hi)
					       decimals
					       2))
				  -1)

    (gtk-signal-connect edit "changed"
      (lambda ()
	(or (edit-programmatic-prop edit)
	    (gtk-signal-emit edit "user-changed"))))

    edit))

(define (edit-num-get edit)
  (let* ((int? (= 0 (gtk-widget-get edit #:digits)))
	 (n    (if int?
		   (gtk-spin-button-get-value-as-int edit)
		   (gtk-spin-button-get-value-as-float edit)))
	 (adj  (gtk-widget-get edit #:adjustment)))
    (set! n (max (gtk-adjustment-lower adj)
		 (min (gtk-adjustment-upper adj) n)))
    (if int? (set! n (round-exact n)))
    n))

(define (edit-num-call-with-programmatic edit thunk)
  (let ((old-val #f))
    (dynamic-wind
	(lambda ()
	  (set! old-val (edit-programmatic-prop edit))
	  (set! (edit-programmatic-prop edit) #t))
	thunk
	(lambda ()
	  (set! (edit-programmatic-prop edit) old-val)))))

(define (edit-num-set edit val)
  (or (=-close? val (gtk-spin-button-get-value-as-float edit))
      (edit-num-call-with-programmatic
       edit (lambda ()
	      (gtk-spin-button-set-value edit val)))))


;;-----------------------------------------------------------------------------
;; ChartDerivedChoice widget
;;
;; This is a widget for selecting a derived series, such as one of the
;; averages or an indicator.
;;

(gtk-class-new 'GtkVBox "ChartDerivedChoice")

(define closechoice-list
  '(closes midpoint typical-price weighted-close opens highs lows))

(let-toplevel ((rec-type (make-record-type "derivedchoice"
					   '(option
					     param-list
					     field-alist
					     field-hbox
					     applyto))))
  (define rec-new          (record-constructor rec-type))
  (define rec-option       (rec-accessor rec-type 'option))
  (define rec-param-list   (rec-accessor rec-type 'param-list))
  (define rec-param-list!  (rec-modifier rec-type 'param-list))
  (define rec-field-hbox   (rec-accessor rec-type 'field-hbox))
  (define rec-field-hbox!  (rec-modifier rec-type 'field-hbox))
  (define rec-applyto      (rec-accessor rec-type 'applyto))
  (define rec-applyto!     (rec-modifier rec-type 'applyto))

  ;; (KEY . (LABEL EDIT-NUM-WIDGET))
  (define rec-field-alist  (rec-accessor rec-type 'field-alist))
  (define rec-field-alist! (rec-modifier rec-type 'field-alist))

  (define (rec-param-find rec key)
    (find (lambda (elem)
	    (eq? key (first elem)))
	  (rec-param-list rec)))

  (define (rec-param-cons rec key val)
    (rec-param-list! rec (cons (list key val)
			       (rec-param-list rec))))

  (define (rec-param-default rec key val)
    (if (not (rec-param-find rec key))
	(rec-param-cons rec key val)))

  (define (rec-param-get rec key)
    (and-let* ((elem (rec-param-find rec key)))
      (second elem)))

  (define (rec-param-set! rec key val default)
    (if (equal? val default)
	(rec-param-list! rec (alist-delete! key (rec-param-list rec)))

	(let ((elem (rec-param-find rec key)))
	  (if elem
	      (set-second! elem val)
	      (rec-param-cons rec key val)))))

  (define (field-changed rec edit param)
    (case (datatype-param-type param)
      ((boolean)
       (let ((value (gtk-toggle-button-get-active edit))
	     (key   (datatype-param-key param)))
	 (if (not (eq? value (rec-param-get rec key)))
	     (begin
	       (rec-param-set! rec key value (datatype-param-default param))
	       (widgets->viewstyle)))))
      (else
       ;; must do this update to get a text entered number coming back from
       ;; value-as-int
       (and-let* ((value (string->number (gtk-entry-get-text edit))))
	 (if (not (eq? 'float (datatype-param-type param)))
	     (set! value (round-exact value)))
	 (if (datatype-param-minimum param)
	     (set! value (max value (datatype-param-minimum param))))
	 (if (datatype-param-maximum param)
	     (set! value (min value (datatype-param-maximum param))))

	 ;; update if no existing value or the existing value isn't close
	 ;; enough
	 (let ((key (datatype-param-key param)))
	   (if (or (not (rec-param-get rec key))
		   (not (=-close? value (rec-param-get rec key))))
	       (begin
		 (edit-num-set edit value)
		 (rec-param-set! rec key value (datatype-param-default param))
		 (widgets->viewstyle))))))))

  (define (select-item dc option datatype)

    (let* ((rec        (c-prop dc))
	   (field-hbox (rec-field-hbox rec)))

      ;; lose all fields of old datatype
      (c-gtk-container-remove-all field-hbox)

      (for-each
       (lambda (param)
	 (let* ((key      (datatype-param-key param))
		(existing (assoc-ref (rec-field-alist rec) key))
		(label    #f)
		(field    #f))
	   (if existing
	       (begin
		 (set! label (first  existing))
		 (set! field (second existing)))

	       (begin
		 (case (datatype-param-type param)
		   ((boolean)
		    (set! field (gtk-check-button-new-with-label
				 (datatype-param-name param)))
		    (gtk-signal-connect field "toggled"
		      (lambda ()
			(field-changed rec field param))))
		   (else
		    (set! label (gtk-label-new (string-append
						" "
						(datatype-param-name param))))
		    (set! field
			  (edit-num-new (datatype-param-minimum param)
					(datatype-param-maximum param)
					(datatype-param-step param)
					(* 10 (datatype-param-step param))
					(datatype-param-decimals param)))
		    (gtk-signal-connect field "user-changed"
		      (lambda ()
			(field-changed rec field param)))))
		 (rec-field-alist! rec (acons key (list label field)
					      (rec-field-alist rec)))

		 (set! (c-prop field) param)
		 (rec-param-default rec key (datatype-param-default param))

		 (if label
		     (gtk-widget-show label))
		 (gtk-widget-show field)))

	   (if label
	       (gtk-box-pack-start field-hbox label #f #f 0))
	   (gtk-box-pack-start field-hbox field #f #f 0)

	   (case (datatype-param-type param)
	     ((boolean)
	      (gtk-toggle-button-set-active field (rec-param-get rec key)))
	     (else
	      (edit-num-set field (rec-param-get rec key))))))

       (datatype-parameter-list datatype)))

    (widgets->viewstyle))

  (define*-toplevel (derivedchoice-new pred #:optional use-applyto)

    (let* ((dc     (gtk-widget-new 'ChartDerivedChoice))
	   (option (dtoptmenu-new (sort! (filter pred (datatype-list-all))
					 datatype<?)))
	   (rec    (rec-new option '() '() #f #f)))
      (set! (c-prop dc) rec)

      (let ((hbox   (gtk-hbox-new #f 0)))
	(gtk-box-pack-start dc hbox #f #f 0)
	(gtk-box-set-spacing hbox (c-gtk-widget-units->width dc '(em . 0.5)))

	(gtk-signal-connect option "changed"
	  (lambda ()
	    (select-item dc option (dtoptmenu-datatype option))))
	(gtk-box-pack-start hbox option #f #f 0)

	'''(let* ((hadj            (gtk-adjustment-new 0 0 0 0 0 0))
		  (vadj            (gtk-adjustment-new 0 0 0 0 0 0))
		  (field-scrollwin (gtk-scrolled-window-new hadj vadj))
		  (field-viewp     (gtk-viewport-new hadj vadj))
		  (field-hbox      (gtk-hbox-new #f 0)))
	     (gtk-scrolled-window-set-policy field-scrollwin 'automatic 'never)
	     (gtk-viewport-set-shadow-type field-viewp 'none)
	     (gtk-box-set-spacing field-hbox (c-gtk-widget-units->width
					      field-hbox '(em . 0.5)))
	     (rec-field-hbox! rec field-hbox)
	     (gtk-container-add field-viewp field-hbox)
	     (gtk-container-add field-scrollwin field-viewp)
	     (gtk-container-add hbox field-scrollwin))

	;; indicator fields
	(let* ((field-hbox (gtk-hbox-new #f 0)))
	  (gtk-box-set-spacing field-hbox (c-gtk-widget-units->width
					   field-hbox '(em . 0.5)))
	  (rec-field-hbox! rec field-hbox)
	  (gtk-box-pack-start hbox field-hbox #t #t 0))

	;; info button
	(let ((info (infobutton-new option)))
	  (gtk-box-pack-end hbox info #f #t 0)))

      (if use-applyto
	  (let ((hbox    (gtk-hbox-new #f 0))
		(label   (gtk-label-new (string-append "        "
						       (_ "Apply to"))))
		(optmenu (dtoptmenu-new (map datatype-find closechoice-list))))
	    (rec-applyto! rec optmenu)
	    (gtk-box-pack-start dc hbox #f #f 0)
	    (gtk-box-pack-start hbox label #f #f (c-gtk-widget-units->width
						  hbox '(em . 0.5)))
	    (gtk-box-pack-start hbox optmenu #f #f 0)

	    (gtk-signal-connect optmenu "changed"
	      (lambda ()
		(rec-param-set! rec 'applyto
				(datatype-key (dtoptmenu-datatype optmenu))
				'closes)
		(widgets->viewstyle)))))

      dc))

  (define-toplevel (derivedchoice-get-viewstyle dc)
    (cons (datatype-key (dtoptmenu-datatype (rec-option (c-prop dc))))
	  (rec-param-list (c-prop dc))))

  (define-toplevel (derivedchoice-set-viewstyle dc viewstyle)
    (let ((rec        (c-prop dc))
	  (dtkey      (first viewstyle))
	  (param-list (cdr viewstyle)))
      (rec-param-list! (c-prop dc) param-list)
      (dtoptmenu-datatype-key! (rec-option rec) dtkey)
      (if (rec-applyto rec)
	  (dtoptmenu-datatype-key! (rec-applyto rec)
				   (second (or (assq 'applyto param-list)
					       '(applyto closes))))))))


;;-----------------------------------------------------------------------------
;; view style dialog

(define dialog #f)
(define upper-linestyle #f)
(define upper-derivedchoice-list #f)
(define upper-size-spin #f)
(define upper-split-toggle #f)
(define upper-dividend-toggle #f)
(define upper-imputation-toggle #f)
(define upper-rollover-toggle #f)
(define lower-indicator #f)
(define lower-derivedchoice-list #f)
(define lower-size-spin #f)
(define tooltips #f)

(define (widgets->viewstyle)
  (preference-set
   'viewstyle
   (list
    (append
     (list (lsoptmenu-linestyle upper-linestyle))
     (map derivedchoice-get-viewstyle upper-derivedchoice-list)
     `((window-size ,(edit-num-get upper-size-spin))
       (split       ,(gtk-toggle-button-get-active upper-split-toggle))
       ,@(if (gtk-toggle-button-get-active upper-dividend-toggle)
	     '((dividend #t))
	     '())
       ,@(if (gtk-toggle-button-get-active upper-imputation-toggle)
	     '((imputation #t))
	     '())
       ,@(if (not (gtk-toggle-button-get-active upper-rollover-toggle))
	     '((rollover #f))
	     '())))

    (append
     (list (derivedchoice-get-viewstyle lower-indicator))
     (map derivedchoice-get-viewstyle lower-derivedchoice-list)
     (list (list 'window-size
		 ;; #f if empty text
		 (or (string->number (gtk-entry-get-text lower-size-spin))
		     1)))))))

(define-public (viewstyle->widgets)
  (if dialog
      (let ((viewstyle (preference-get 'viewstyle viewstyle-default)))

	;; upper
	(lsoptmenu-linestyle-key! upper-linestyle (first (first viewstyle)))
	(for-each derivedchoice-set-viewstyle
		  upper-derivedchoice-list (cdr (first viewstyle)))

	(and-let* ((size (assq-ref (cdr (first viewstyle)) 'window-size)))
	  (edit-num-set upper-size-spin (first size)))

	(gtk-toggle-button-set-active upper-split-toggle
				      (viewstyle-split viewstyle))
	(gtk-toggle-button-set-active upper-dividend-toggle
				      (viewstyle-dividend viewstyle))
	(gtk-toggle-button-set-active upper-imputation-toggle
				      (viewstyle-imputation viewstyle))
	(gtk-toggle-button-set-active upper-rollover-toggle
				      (viewstyle-rollover viewstyle))

	;; lower
	(derivedchoice-set-viewstyle lower-indicator
				     (first (second viewstyle)))
	(for-each derivedchoice-set-viewstyle
		  lower-derivedchoice-list (cdr (second viewstyle)))

	(and-let* ((size (assq-ref (cdr (second viewstyle)) 'window-size)))
	  (edit-num-set lower-size-spin (first size))))))



(define (init)
  (set! dialog (gtk-dialog-new-with-buttons
		(c-make-window-title (_ "View Style")) #f '()))
  (c-gtk-widget-set-usize-units (gtk-dialog-vbox dialog) '(em . 68) -1)
  (gtk-dialog-add-buttons dialog
			  "gtk-close" gtk-response-close
			  "gtk-help"  gtk-response-help)
  (set! tooltips (gtk-tooltips-new))

  (c-gtk-dialog-hide-on-delete dialog)
  (gtk-signal-connect dialog "response"
    (lambda (id)
      (cond ((= id gtk-response-close)
	     (gtk-widget-hide dialog))

	    ((= id gtk-response-help)
	     (manual-popup (_ "CONTEXT(manual-node): Main Window") dialog)))))

  (let ((hbox  (gtk-hbox-new #f 0))
	(label (gtk-widget-new 'GtkLabel
			       #:label (_ "Upper")
			       #:xalign 0)))
    (gtk-misc-set-padding label (* 2 (c-gtk-widget-em label)) 0)
    (gtk-box-pack-start hbox label)

    (gtk-box-pack-start hbox (gtk-widget-new 'GtkLabel
					     #:label (_ "Size")
					     #:xalign 0)
			#f #f (c-gtk-widget-em label))
    (set! upper-size-spin (edit-num-new 1 999 1 5 0))
    (gtk-box-pack-start hbox upper-size-spin)
    (gtk-signal-connect upper-size-spin "user-changed" widgets->viewstyle)

    (set! upper-split-toggle
	  (gtk-check-button-new-with-label (_ "Adj splits")))
    (gtk-tooltips-set-tip tooltips upper-split-toggle
			  (_ "Whether to adjust past prices to the current share capital basis")
			  "")
    (gtk-box-pack-start hbox upper-split-toggle #f #f (c-gtk-widget-em label))
    (gtk-signal-connect upper-split-toggle "toggled"
      (lambda ()
	(or (eq? (viewstyle-split (preference-get 'viewstyle
						  viewstyle-default))
		 (gtk-toggle-button-get-active upper-split-toggle))
	    (widgets->viewstyle))))

    (set! upper-dividend-toggle
	  (gtk-check-button-new-with-label (_ "Adj dividends")))
    (gtk-tooltips-set-tip tooltips upper-dividend-toggle
			  (_ "Whether to adjust past prices down for dividends paid out")
			  "")
    (gtk-box-pack-start hbox upper-dividend-toggle #f #f
			(c-gtk-widget-em label))
    (gtk-signal-connect upper-dividend-toggle "toggled"
      (lambda ()
	(gtk-widget-set-sensitive
	 upper-imputation-toggle ;; imputation only with dividends
	 (gtk-toggle-button-get-active upper-dividend-toggle))
	(or (eq? (viewstyle-dividend (preference-get 'viewstyle
						     viewstyle-default))
		 (gtk-toggle-button-get-active upper-dividend-toggle))
	    (widgets->viewstyle))))

    (set! upper-imputation-toggle
	  (gtk-check-button-new-with-label (_ "Adj imputation")))
    (gtk-widget-set-sensitive upper-imputation-toggle #f) ;; only with divs
    (gtk-tooltips-set-tip tooltips upper-imputation-toggle
			  (_ "Whether to adjust past prices down for imputations credits attached to dividends")
			  "")
    (gtk-box-pack-start hbox upper-imputation-toggle #f #f
			(c-gtk-widget-em label))
    (gtk-signal-connect upper-imputation-toggle "toggled"
      (lambda ()
	(or (eq? (viewstyle-imputation (preference-get 'viewstyle
						       viewstyle-default))
		 (gtk-toggle-button-get-active upper-imputation-toggle))
	    (widgets->viewstyle))))

    (set! upper-rollover-toggle
	  (gtk-check-button-new-with-label (_ "Adj rollovers")))
    (gtk-tooltips-set-tip tooltips upper-rollover-toggle
			  (_ "Whether to adjust past prices up for futures rollovers")
			  "")
    (gtk-box-pack-start hbox upper-rollover-toggle #f #f
			(c-gtk-widget-em label))
    (gtk-signal-connect upper-rollover-toggle "toggled"
      (lambda ()
	(or (eq? (viewstyle-rollover (preference-get 'viewstyle
						       viewstyle-default))
		 (gtk-toggle-button-get-active upper-rollover-toggle))
	    (widgets->viewstyle))))

    (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox))

  ;; within a hbox so it doesn't stretch to full vbox width
  (let ((hbox (gtk-hbox-new #f 0)))
    (set! upper-linestyle (lsoptmenu-new))
    (gtk-signal-connect upper-linestyle "changed" widgets->viewstyle)
    (gtk-box-pack-start hbox upper-linestyle #f #f 0)
    (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0))

  (set! upper-derivedchoice-list (list (derivedchoice-new
					datatype-parent-units?
					#t)))
  (for-each (lambda (dc)
	      (gtk-box-pack-start (gtk-dialog-vbox dialog) dc))
	    upper-derivedchoice-list)


  (gtk-box-pack-start (gtk-dialog-vbox dialog) (gtk-hseparator-new)
		      #f #f (gdk-string-height
			     (c-gtk-widget-get-font dialog) "x"))

  (let ((hbox  (gtk-hbox-new #f 0))
	(label (gtk-widget-new 'GtkLabel
			       #:label (_ "Lower")
			       #:xalign 0)))
    (gtk-misc-set-padding label (c-gtk-widget-em label) 0)
    (gtk-box-pack-start hbox label #f #f 0)

    (gtk-box-pack-start hbox (gtk-widget-new 'GtkLabel
					     #:label (_ "Size")
					     #:xalign 0)
			#f #f (c-gtk-widget-em label))
    (set! lower-size-spin (edit-num-new 1 999 1 5 0))
    (gtk-box-pack-start hbox lower-size-spin #f #f 0)
    (gtk-signal-connect lower-size-spin "user-changed" widgets->viewstyle)

    (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox))

  (set! lower-indicator (derivedchoice-new datatype-own-units?))
  (gtk-box-pack-start (gtk-dialog-vbox dialog) lower-indicator)

  (set! lower-derivedchoice-list
	(list (derivedchoice-new datatype-parent-units-non-ohlcv?)))
  (for-each (lambda (dc)
	      (gtk-box-pack-start (gtk-dialog-vbox dialog) dc))
	    lower-derivedchoice-list)

  (viewstyle->widgets)

  (notify-connect 'preferences-save
    (let ((recurse #f))
      (lambda (dummy)
	;; don't re-enter when already doing a viewstyle->widgets from here
	(or recurse
	    (dynamic-wind
		(lambda ()
		  (set! recurse #t))
		(lambda ()
		  (if (gtk-widget-mapped dialog)
		      (viewstyle->widgets)))
		(lambda ()
		  (set! recurse #f)))))))

  (set! init noop))


(define-public (viewstyle-dialog-popup)
  (init)
  (if (gtk-widget-mapped dialog)
      (c-gtk-widget-raise dialog)
      (begin
	(viewstyle->widgets)
	(gtk-widget-show-all dialog))))
