;; Volatility Ratio.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart volr)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 q)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.incrediblecharts.com/technical/volatility_ratio.htm


;; Return a procedure which calculates VOLR values over a window of COUNT
;; many points.
;;
;; Each call (PROC high low close) enters a new point into the window, and
;; the return is the VOLR value up to (and including) that point, or #f if
;; not enough values yet.
;;
;; To prime the window initially, call PROC with COUNT many values before
;; the first desired point.  (On the very first call only the close value is
;; actually used, not the high/low there.)
;;
(define-public (volr-calc-proc count)

  ;; PREV-CLOSE is the last close value passed in
  ;; HIGH-Q is the high values for each day
  ;; LOW-Q is the corresponding low values for each day
  ;; POINTS is how many entries in HIGH-Q (and LOW-Q likewise)
  ;;
  ;; the high/low queue values have had their appropriate previous close
  ;; worked in, thus representing "true high" and "true low"
  ;;
  (let ((high-q     (make-q))
	(low-q      (make-q))
	(points     0)
	(points-min (ceiling-exact (* indicator-minimum-fraction count)))
	(prev-close #f))

    (lambda (high low close)
      (if prev-close
	  (begin
	    ;; use closes if no high/low data
	    (set! high (or high close))
	    (set! low  (or low  close))

	    ;; drop old point if full
	    (if (>= points count)
		(begin
		  (deq! high-q)
		  (deq! low-q))
		;; not full, so POINTS increases
		(set! points (1+ points)))

	    (let ((this-high (max high prev-close))
		  (this-low  (min low  prev-close)))

	      ;; add new point
	      (enq! high-q this-high)
	      (enq! low-q  this-low)
	      (set! prev-close close)

	      (and (>= points points-min)
		   (let* ((period-high  (apply max (q-list high-q)))
			  (period-low   (apply min (q-list low-q)))
			  (period-range (- period-high period-low)))
		     (if (zero? period-range)
			 0.5  ;; no range at all
			 (/ (- this-high this-low) period-range))))))

	  (begin
	    (set! prev-close close)
	    #f)))))

(define (volr-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count count))
	 (array  (series-array parent start hi))
	 (calc   (volr-calc-proc count)))

    ;; warmup
    (do ((i start (1+ i)))
	((>= i lo))
      (and-let* ((close (array-ref array i 3)))
	(calc (array-ref array i 1)  ;; high
	      (array-ref array i 2)  ;; low
	      close)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((close (array-ref array i 3)))
	(array-set! new-array (calc (array-ref array i 1)  ;; high
				    (array-ref array i 2)  ;; low
				    close)
		    i)))))

(define-public (volr-series series count)
  (ohlcv-required "VOLR" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "VOLR") count))
    (series-datatype!  new-series volr-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts volr-fill (1+ count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public volr-datatype
  (datatype-new!
   #:key       'volr
   #:name      (_ "Volatility Ratio")
   #:input     datatype-ohlcv
   #:units     'volr
   #:linestyle 'line
   #:minimum   0
   #:maximum   1
   #:hlines    '(0.5)
   #:manual    (_ "CONTEXT(manual-node): Volatility Ratio")
   #:proc      volr-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'volr-days
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
