;; Volume series.

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart volume)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; return a series which is the volumes from a given ohlcv SERIES
(define (volume-series series)
  (let ((new-series (series-copy series)))
    (series-parent!    new-series series)
    (series-require!   new-series volume-require)
    (series-datatype!  new-series volume-datatype)
    (series-name!      new-series (_ "Volume"))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())
    new-series))

(define (volume-require series lo hi)
  (make-shared-array-column (series-array (series-parent series) lo hi) 4))


;; `volume-initial-page' returns a list (LOWER UPPER) which is the initial
;; Y-axis page range to show for dates LO to HI.  This is for use with
;; trading volumes and also open-interest.
;;
;; The page size is meant to keep average to high volume days visible in the
;; window, but leave abnormally high days off the top of the screen so as
;; not to make the average ones come out tiny.
;;
;; The rule is to look for the 98% fractile, meaning a level which covers
;; 98% of the values in the LO to HI range, then an extra 1.5x that much.
;; Making UPPER just the 0.98 fractile alone would guarantee two or three
;; points off the top of the window even if they're only modestly bigger
;; than the average, hence going up to 1.5x that point.
;;
;;
;; ENHANCE-ME: If LO to HI is small then maybe the 98% should be reduced a
;; bit.  If the width is less than 50 (or is it 25?) dates then 98% ends up
;; meaning every data point.
;;
;;
(define-public (volume-initial-page series lo hi)
  (let* ((array (series-array series lo hi))
	 (lst   (delq! #f (array->list
			   (make-shared-array-restrict array lo hi)))))
    (if (null? lst)
	(list #f #f)  ;; no data at all

	(begin
	  (set! lst (delete 0 lst =))
	  (if (null? lst)
	      (list 0 0)

	      (let ((latest (last lst)))
		(set! lst (sort! lst <))
		(list 0 
		      (max latest
			   (list-ref-fraction lst 0.98)
			   (* 1.5 (apply arithmetic-mean lst))))))))))


(define-public volume-datatype
  (datatype-new!
   #:key       'volume
   #:name      (_ "Volume")
   #:units     'volume
   #:linestyle 'bar
   #:minimum   0
   #:manual    (_ "CONTEXT(manual-node): Volume and Open Interest")
   #:proc      volume-series
   #:initial-page-proc volume-initial-page))
