;; ChartVScale vertical scale widget.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart vscale)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gtk)
  #:use-module (gtk gdk)
  #:use-module (chart extra-gdk)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart gui-common)
  #:use-module (chart misc))


;;-----------------------------------------------------------------------------
;; misc

;; log10, rounded upwards to an exact integer
(define (log10-ceiling n)
  (ceiling-exact (log10 n)))

;; Return N rounded upwards to a multiple of D.
;; N can be any real number, D must be an exact integer.
(define (round-up-multiple n d)
  (set! n (ceiling-exact n))
  (+ n (modulo (- n) d)))


;;-----------------------------------------------------------------------------
;; text sizes

;; return the width of STR in pango units
(define (c-pango-layout-pangwidth layout str)
  (pango-layout-set-text layout str)
  (first (pango-layout-get-size layout)))

(define-public (c-pango-layout-height layout str)
  (pango-layout-set-text layout str)
  (second (pango-layout-get-pixel-size layout)))

;; height in pixels of the highest digit in LAYOUT
(define c-pango-layout-digit-height
  (make-procedure-with-cache-eq
   (lambda (layout)
     (c-pango-layout-height layout "0123456789"))))

;; the pango units width of a comma in LAYOUT
(define c-pango-layout-comma-pangwidth
  (make-procedure-with-cache-eq
   (lambda (layout)
     (c-pango-layout-pangwidth layout (locale-thousands-sep)))))

;; the pango units width of a decimal point in LAYOUT
(define c-pango-layout-decimal-pangwidth
  (make-procedure-with-cache-eq
   (lambda (layout)
     (c-pango-layout-pangwidth layout (locale-decimal-point)))))

;; the pango units width of a "-" sign in LAYOUT
(define c-pango-layout-minus-pangwidth
  (make-procedure-with-cache-eq
   (lambda (layout)
     (c-pango-layout-pangwidth layout "-"))))

;; the width of the widest digit in LAYOUT
(define c-pango-layout-digit-pangwidth
  (make-procedure-with-cache-eq
   (lambda (layout)
     (apply max (map (lambda (str)
		       (c-pango-layout-pangwidth layout str))
		     '("0" "1" "2" "3" "4" "5" "6" "7" "8" "9"))))))

;; Return the width in pixels for a value N/2^DECIMALS.
;; A decimal point is included if DECIMALS >= 1.
;; Thousands separators are included per `locale-thousands-sep'.
;;
(define (c-pango-layout-number-width layout n decimals)
  (let ((negwidth (if (negative? n)
		      (c-pango-layout-minus-pangwidth layout)
		      0))
	(int-digs (max 1 (if (zero? n)
			     1
			     (- (log10-ceiling (abs n)) decimals)))))
    (pango-pixels-ceil
     (+ negwidth
	(* (+ int-digs decimals)
	   (c-pango-layout-digit-pangwidth layout))
	(if (positive? decimals)
	    (c-pango-layout-decimal-pangwidth layout)
	    0)
	(* (quotient (1- int-digs) 3)
	   (c-pango-layout-comma-pangwidth layout))))))


;;-----------------------------------------------------------------------------

(gtk-class-new 'GtkDrawingArea "ChartVScale")
(c-gtk-rc-chart-style "ChartVScale")

;; all in multiples of digit-width
(define tick-width   0.8)  ;; hard against left edge of window
(define tick-margin  0.5)  ;; between tick and number
(define right-margin 0.2)  ;; between number and right edge of window

;; return a new vscale widget, tracking GRAPH
(define-public (vscale-new graph)
  (let ((vscale (gtk-widget-new 'ChartVScale))
	(width  -1)
	(vadj   (graph-vadjustment graph)))

    ;; return values (BOT TOP FACTOR UNIT DECIMALS) which represent the
    ;; drawing steps
    ;;
    ;; UNIT+DECIMALS is the amount to step by.
    ;; DECIMALS can be `series-decimals', or it can be a higher value.
    ;;
    ;; BOT and TOP are the highest and lowest values to be drawn.  BOT is an
    ;; integer, the actual starting point.  But TOP is only a limit, a
    ;; flonum value just off the top of the window.  Both are in the
    ;; returned DECIMALS scale.
    ;;
    ;; FACTOR is a multiplier to convert a UNIT value into the usual
    ;; `series-decimals' so that `graph-scale-y' can be applied to it.
    ;; FACTOR is 1 when there's no additional decimals being drawn.
    ;;
    (define (choose-step)
      (define series (graph-series graph))
      (if (or (not series)
	      (= 0 (gtk-adjustment-page-size vadj)))
	  ;; no info
	  (values #f #f #f #f #f)

	  (let* ((layout       (primitive-property-ref c-layout-prop graph))
		 (digit-height (c-pango-layout-digit-height layout))
		 (digit-half   (quotient digit-height 2))

		 (win          (gtk-widget-window vscale))
		 (height       (c-gdk-drawable-get-height win))
		 (bot     (graph-y->price-inexact graph (+ height digit-half)))
		 (top     (graph-y->price-inexact graph (- digit-half)))

		 ;; price amount which is DIGIT-HEIGHT, calculated near the
		 ;; top of the window (in log scale it's less near the
		 ;; bottom)
		 (price-height (- top
				  (graph-y->price-inexact graph digit-half)))

		 ;; minimum unit of 2*DIGIT-HEIGHT, as a price amount
		 (unit         (* 2 price-height))

		 (decimals     (series-decimals series))
		 (factor       1))

	    ;; now round up UNIT to the next higher value of the form 10^k,
	    ;; 2*10^k or 5*10^k for some integer K (possibly negative)
	    ;;
	    ;; start at rounding up to a value 10^k, then see if 5*10^(k-1)
	    ;; or 2*10^(k-1) would be bigger than the inexact UNIT value too
	    (let* ((k (log10-ceiling unit))
		   (r-unit (integer-expt-pn 10.0 k))) ;; 10^k rounded unit
	      (cond ((> (* r-unit 0.2) unit)
		     (set! unit 2)
		     (set! k (1- k)))
		    ((> (* r-unit 0.5) unit)
		     (set! unit 5)
		     (set! k (1- k)))
		    (else
		     (set! unit 1)))

	      ;; At this point we've decided to step by UNIT*10^K, for
	      ;; an integer UNIT and positive or negative K.
	      ;;
	      ;; If K is positive then work it into the UNIT value, giving
	      ;; perhaps 20 or 50 as a step amount, with DECIMALS left at
	      ;; `series-decimals'.
	      ;;
	      ;; If K is negative then we're going to display more than
	      ;; `series-decimals', so increase DECIMALS to return, make a
	      ;; scale FACTOR and adjust TOP accordingly.
	      ;;
	      (if (positive? k)
		  (set! unit (* unit (integer-expt 10 k)))

		  (begin
		    (set! factor (integer-expt 10 (- k)))
		    (set! bot (* bot factor))
		    (set! top (* top factor))
		    (set! factor (/ 1.0 factor))
		    (set! decimals (- decimals k)))))

	    (set! bot (round-up-multiple bot unit))
	    (values bot top factor unit decimals))))

    (define (draw)
      (receive (bot top factor unit decimals)
	  (choose-step)
	(if top
	    (let* ((layout        (primitive-property-ref c-layout-prop graph))
		   (win           (gtk-widget-window vscale))
		   (scale-y       (graph-scale-y-proc graph))
		   (style         (gtk-widget-get-style vscale))
		   (fg-gc         (gtk-style-fg-gc style 'normal))
		   (text-gc       (gtk-style-text-gc style 'normal))
		   (digit-height  (c-pango-layout-digit-height layout))
		   (digit-vcentre (quotient digit-height 2))
		   (digit-width   (c-pango-layout-digit-width layout))

		   ;; ticks drawn from 0 to tick-x
		   (tick-x        (floor-exact (* tick-width digit-width)))
		   ;; text drawn at text-x
		   (text-x        (+ tick-x (ceiling-exact
					     (* tick-margin digit-width)))))

	      (do ((i bot (+ unit i)))
		  ((>= i top))
		(let ((y (scale-y (* i factor))))
		  (pango-layout-set-text layout
					 (locale-number+decimals i decimals))
		  (gdk-draw-layout win text-gc
				   text-x (- y digit-vcentre) layout)
		  (gdk-draw-line win fg-gc 0 y tick-x y)))))))

    ;; The desired width never reduces from a previously desired value.
    ;; This prevents window resizing on every series with a different range.
    ;;
    ;; Usually we get to say 100,000,000 for a heavy volume stock, or even
    ;; 1,000,000,000 for an index which adds all shares traded in that
    ;; index, then stick at that size.
    ;;
    (define (update-usize)
      (receive (bot top factor unit decimals)
	  (choose-step)
	;; if not setup yet then size for "1.00"
	(if bot
	    (set! top (round-down-multiple (floor-exact top) unit))
	    (begin
	      (set! bot 100)
	      (set! top 100)
	      (set! decimals 2)))
	(let* ((layout    (primitive-property-ref c-layout-prop graph))
	       (new-width
		(ceiling-exact
		 (+ (max (c-pango-layout-number-width layout bot decimals)
			 (c-pango-layout-number-width layout top decimals))
		    (* (+ tick-width tick-margin right-margin)
		       (c-pango-layout-digit-width layout))))))
	  (set! width (max width new-width))

	  ;; increase window if necessary for desired width
	  ;; if increased then the expose from the server will get us
	  ;; redrawing, no need to do that here
	  (let ((win (gtk-widget-window vscale)))
	    (if (or (not win)
		    (< (c-gdk-drawable-get-width win) width))
		(gtk-widget-set-usize vscale width -1))))))

    (define (queue-redraw . args)
      (c-gtk-widget-queue-redraw vscale))

    (c-gtk-widget-drawing vscale draw)

    ;; possible new font, reset sticky minimum width and recalc all
    (gtk-signal-connect vscale "style_set"
      (lambda (prev)
	(set! width -1)
	(update-usize)
	(queue-redraw)))

    ;; window resize (doesn't change range of values shown, so only redraw)
    (gtk-signal-connect vscale "configure-event" queue-redraw)

    ;; moved
    (gtk-signal-connect vadj "value-changed"
      (lambda ()
	(update-usize)   ;; possibly bigger numbers shown
	(queue-redraw)))

    (gtk-signal-connect vadj "changed"
      (lambda ()
	(update-usize)   ;; possibly bigger numbers shown
	(queue-redraw)))

    ;; initial size
    (update-usize)
    vscale))
