;; Volume weighted moving average (VWMA).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart vwma)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; Return a procedure which calculates a volume weighted moving average over
;; an accumulated window of COUNT values.
;;
;; Each call (PROC close volume) enters a new point into the window, and the
;; return is the VWMA up to (and including) that point, or #f if not enough
;; values yet.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired point.
;;
(define-public (vwma-calc-proc count)

  ;; POINTS is the number of non-#f points in DATA
  ;; TOTAL-PRODUCT is the sum of the point values in DATA
  ;; TOTAL-VOLUME is the sum of the point values in DATA

  (let ((min-points    (ceiling-exact (* indicator-minimum-fraction count)))
	(points        0)
	(total-product 0)
	(total-volume  0)
	(data          (make-circular-list (* 2 count) #f)))

    (lambda (close volume)
      ;; cope with no volume data
      (or volume
	  (set! volume 0))

      (if (car data)
	  ;; drop oldest point
	  (let ((old-close (first data))
		(old-volume (second data)))
	    (set! total-volume  (- total-volume old-volume))
	    (set! total-product (- total-product (* old-close old-volume))))
	  ;; getting a new point
	  (set! points (1+ points)))

      ;; replace oldest with new point
      (set-first!  data close)
      (set-second! data volume)
      (set! data (cddr data))

      ;; add this point
      (set! total-volume  (+ total-volume volume))
      (set! total-product (+ total-product (* close volume)))

      (and (>= points min-points)
	   (not (zero? total-volume))
	   (exact->inexact (/ total-product total-volume))))))

(define (vwma-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent
					   #:before lo #:count (1- count)))
	 (array  (series-array parent start hi))
	 (vwma   (vwma-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (vwma (array-ref array i 3)    ;; close
		(array-ref array i 4)))) ;; volume

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (vwma (array-ref array i 3)  ;; close
				      (array-ref array i 4)) ;; volume
		      i)))))

(define-public (vwma-series series count)
  (ohlcv-required "VWMA" series)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "VWMA") count))
    (series-datatype! new-series vwma-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts vwma-fill (1- count)))
    (series-array!    new-series #f)

    new-series))

(define vwma-datatype
  (datatype-new!
   #:key      'vwma
   #:name     (_ "VWMA - Volume Weighted")
   #:units    #f
   #:input    datatype-ohlcv
   #:manual   (_ "CONTEXT(manual-node): Volume Weighted Moving Average")
   #:proc     vwma-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'vwma-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
