;; Watchlist dialog.

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart watchlist)
  #:use-module (ice-9 and-let-star)
  #:use-module (srfi srfi-1)
  #:use-module (gtk gdk)
  #:use-module (gtk gtk)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart main)
  #:use-module (chart extra-gtk)
  #:use-module (chart graph)
  #:use-module (chart gui-common)
  #:autoload   (chart intraday)  (intraday-popup)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart symlists)
  #:use-module (chart view))


;;-----------------------------------------------------------------------------
;; misc

(define (seconds->hms time)
  (list (quotient time 3600)
	(quotient (remainder time 3600) 60)
	(remainder time 60)))

(define (negative-number? obj)
  (and (number? obj)
       (negative? obj)))

(define (list-ref-or-f lst n)
  (if (null? lst)
      #f
      (if (= 0 n)
	  (first lst)
	  (list-ref-or-f (cdr lst) (1- n)))))


;;-----------------------------------------------------------------------------
;; Gdk extras

;; return true if MOD-LST of GdkModifierType symbols has no buttons
(define (c-gdk-modifier-type-no-buttons? mod-lst)
  (null? (lset-intersection eq? mod-lst '(button1-mask
					  button2-mask
					  button3-mask
					  button4-mask
					  button5-mask))))


;;-----------------------------------------------------------------------------
;; GtkCList extras

;; return text or #f
(define (c-gtk-clist-get-text clist row column)
  (let ((vec (vector 1)))
    (and (= 1 (gtk-clist-get-text clist row 0 vec))
	 (vector-ref vec 0))))

;; call (THUNK) with CLIST frozen, and unfreeze afterwards (using dynamic-wind)
(define (c-gtk-clist-call-with-freeze clist thunk)
  (dynamic-wind
      (lambda () (gtk-clist-freeze clist))
      thunk
      (lambda () (gtk-clist-thaw clist))))

;; change CLIST from OLD rows to NEW rows
(define (c-gtk-clist-change-rows clist old new)
  (let ((empty-list (make-list (gtk-object-get clist #:GtkCList::n_columns)
			       "")))
    (c-gtk-clist-call-with-freeze clist
      (lambda ()
	;; add extras
	(do ((i (- new old) (1- i)))
	    ((<= i 0))
	  (gtk-clist-append clist empty-list))
	;; remove excess
	(do ((i old (1- i)))
	    ((<= i new))
	  (gtk-clist-remove clist (1- i)))))))

;; return a PangoFontMetrics object for the given WIDGET
(define-public (c-pango-widget-font-metrics widget)
  (let* ((context  (gtk-widget-get-pango-context widget))
	 (fontdesc (pango-context-get-font-description context))
	 (lang     (pango-context-get-language context)))
    (pango-context-get-metrics context fontdesc lang)))

;; Return the height of each row.
;;
;; The #:row_height property can't be used to get back the default height,
;; only to force a particular height (and read that back).  With the default
;; height #:row_height just reads back as 0.
;;
;; The height calculation can be found in gtk_clist_set_row_height.  In Gtk
;; 1.2 it's font ascent + font descent + 1.  In Gtk 2 it's pango ascent +
;; descent converted to pixels using PANGO_PIXELS.  In both cases a further
;; 1 pixel CELL_SPACING is between each row.
;;
(define (c-gtk-clist-row-height clist)
  (let ((metrics (c-pango-widget-font-metrics clist)))
    (+ (pango-pixels (+ (pango-font-metrics-get-ascent metrics)
			(pango-font-metrics-get-descent metrics)))
       1)))

;; return list (LO HI) which is the range of rows visible
;; GtkCList::row_height reads back as 0 when the row height hasn't been
;; explicitly set, so need num-rows to derive the actual height
;;
;; upper is num-rows * (row-height + CELL_SPACING) + CELL_SPACING, where
;; CELL_SPACING is 1, hence the `1-'
;;
(define (c-gtk-clist-visible-rows clist num-rows)
  (if (zero? num-rows)
      '(0 -1) ;; no rows at all
      (let* ((vadj           (gtk-clist-get-vadjustment clist))
	     (pixels-per-row (/ (1- (gtk-adjustment-upper vadj)) num-rows)))
	(if (zero? pixels-per-row)
	    '(0 -1) ;; if upper not yet set for NUM-ROWS
	    (list (floor-exact (/ (gtk-adjustment-value vadj)
				  pixels-per-row))
		  (ceiling-exact (/ (+ (gtk-adjustment-value vadj)
				       (gtk-adjustment-page-size vadj))
				    pixels-per-row)))))))

;; return the row number for pixel position Y, where Y is from an event
;; registered on the CLIST
(define (c-gtk-clist-event-y->row clist y)
  (let* ((row-height (c-gtk-clist-row-height clist)))
    (floor-exact
     (/ (+ (gtk-adjustment-value (gtk-clist-get-vadjustment clist))
	   y)
	row-height))))

;; ;; return the height of a row in pixels, including any spacing between rows
;; ;; based on ROW_TOP_YPIXEL in gtkclist.c
;; ;;
;; (define (c-gtk-clist-row-height clist)
;;   (gtk-object-get clist #:row_height)

;; #define ROW_TOP_YPIXEL(clist, row) (((clist)->row_height * (row)) + \
;; 				    (((row) + 1) * CELL_SPACING) + \

;; return the row number for pixel position Y, where Y is from the CLIST
;; window (gtk-widget-window CLIST)
(define (c-gtk-clist-y->row clist y)
  (let* ((row-height  (c-gtk-clist-row-height clist))
	 (head-height (+ row-height 8)))
    (floor-exact
     (/ (+ (gtk-adjustment-value (gtk-clist-get-vadjustment clist))
	   y
	   (- head-height))
	row-height))))

;; return root window Y coordinate of top-left of ROW
;; ROW can be a float to get a fractional position
(define (c-gtk-clist-row->root-y clist row)
  (let* ((win         (gtk-widget-window clist))
	 (row-height  (c-gtk-clist-row-height clist))
	 (head-height (+ row-height 8)))
    (round-exact (+ (cdr (gdk-window-get-origin win))
		    head-height
		    (* row row-height)
		    (- (gtk-adjustment-value
			(gtk-clist-get-vadjustment clist)))))))


;;-----------------------------------------------------------------------------
;; GtkWidget stuff

;; C macro not wrapped
(define (gtk-widget-has-focus widget)
  (->bool (memq 'has-focus (gtk-widget-flags widget))))


;;-----------------------------------------------------------------------------
;; row menu (button3 popup)

(let-toplevel ((menu          #f)
	       (title-label   #f)
	       (intraday-item #f)
	       (menu-symbol   #f)
	       (weblinks-item #f))

  (define (init)
    (define (add-item name callback)
      (let* ((item (gtk-menu-item-new-with-mnemonic name)))
	(gtk-menu-shell-append menu item)
	(gtk-signal-connect item "activate" callback)
	item))

    (set! menu (gtk-menu-new))

    ;; title
    (set! title-label (gtk-label-new "X"))
    (let ((item (gtk-menu-item-new)))
      (gtk-misc-set-alignment title-label 0.5 0)
      (gtk-container-add item title-label)
      (gtk-menu-shell-append menu item))

    ;; separator
    (gtk-menu-shell-append menu (gtk-separator-menu-item-new))

    ;; refresh
    (add-item (_ "CONTEXT(stock): _Refresh") ;; share msgid with gnome.scm
	      (lambda ()
		;; takes a moment to crunch desired requests
		(c-gtk-widget-busy dialog)
		(latest-request-explicit (list menu-symbol))))

    ;; chart goto
    (add-item (_ "_Chart")
	      (lambda ()
		(c-gtk-widget-raise app-widget)
		(chart-open (smarker-goto (view-smarker outer)
					  watch-symlist menu-symbol)
			    #:window dialog)))

    ;; intraday
    (set! intraday-item (add-item (_ "_Intraday")
				  (lambda ()
				    (intraday-popup menu-symbol))))

    ;; weblinks
    (set! weblinks-item (weblinks-menuitem-new))
    (gtk-menu-shell-append menu weblinks-item)

    (gtk-widget-show-all menu)
    (set! init noop)) ;; once-only, and gc this code

  (define-toplevel (rowmenu-popup symbol)
    (init)
    (set! menu-symbol symbol)
    (gtk-label-set-text title-label symbol)
    (gtk-widget-set-sensitive intraday-item (intraday-handler? symbol))
    (weblinks-menuitem-symbol! weblinks-item symbol)
    (gtk-menu-popup menu #f #f 3 0)))


;;-----------------------------------------------------------------------------
;; clist tooltip

(define tip-timer #f)   ;; gtk timer id, or #f
(define tip-window #f)  ;; created when required
(define tip-label #f)
(define tip-symbol #f)

;; return tip popup string for SYMBOL
(define (tip-help symbol)
  (let* ((latest (latest-get symbol)))
    (string-join-notnull
     (list (let ((name (or (latest-name latest)
			   (database-read-info symbol 'name))))
	     (string-join-notnull (list symbol name) " - "))
	   (and (latest-quote-adate latest)
		(format #f (_ "Quote: ~a")
			(full-adate-time (latest-quote-adate latest)
					 (latest-quote-time latest))))
	   (and (latest-last-adate latest)
		(format #f (_ "Last:  ~a")
			(full-adate-time (latest-last-adate latest)
					 (latest-last-time latest))))
	   (format #f (_ "~a time; source \"~a\"")
		   (symbol-location symbol) (latest-source latest)))
     "\n")))

(define (tip-move event)
  ;; always get pointer since using hint
  (let* ((pointer (gdk-window-get-pointer (gdk-event-window event))))
    (tip-hide)
    (if (c-gdk-modifier-type-no-buttons? (gdk-event-state event))
	(let ((row (c-gtk-clist-y->row clist (third pointer))))

	  (define (tip-show)
	    (if tip-symbol
		(begin
		  (if (not tip-window)
		      (begin
			(set! tip-window (gtk-window-new 'popup))
			(gtk-window-set-policy tip-window #f #f #f)
			;; title only to identify in "xwininfo" output, it's
			;; not (normally) visible
			(gtk-window-set-title
			 tip-window (c-make-window-title "Watchlist tooltip"))
			(gtk-container-border-width tip-window 5)

			(set! tip-label  (gtk-label-new ""))
			(gtk-label-set-justify tip-label 'left)
			(gtk-widget-show tip-label)
			(gtk-container-add tip-window tip-label)))

		  (c-gtk-label-set-text-maybe tip-label (tip-help tip-symbol))
		  (gtk-widget-realize tip-window)

		  (gtk-window-move tip-window
				   (max 0 (+ (car (gdk-window-get-origin
						   (gtk-widget-window clist)))
					     (* 3 (c-gtk-widget-em clist))))
				   (c-gtk-clist-row->root-y clist (+ 1.5 row)))
		  (gtk-widget-show tip-window))))

	  (set! tip-symbol (row->symbol row))
	  (if tip-symbol
	      (set! tip-timer (gtk-timeout-add 500 tip-show)))))))

(define (tip-hide . dummy-event)
  (if tip-window
      (gtk-widget-hide tip-window))
  (if tip-timer
      (begin
	(gtk-timeout-remove tip-timer)
	(set! tip-timer #f))))

(define (tip-init)
  (c-gtk-widget-add-events clist 'tip
			   '(pointer-motion-mask
			     pointer-motion-hint-mask
			     leave-notify-mask))

  (gtk-signal-connect clist "motion_notify_event"  tip-move)
  (gtk-signal-connect clist "leave_notify_event"   tip-hide)
  (gtk-signal-connect clist "button_press_event"   tip-hide)
  (gtk-signal-connect clist "button_release_event" tip-move)

  (notify-connect 'latest-update
    (lambda (symbol-list)
      (if (and tip-label
	       (member tip-symbol symbol-list))
	  (c-gtk-label-set-text-maybe tip-label (tip-help tip-symbol)))))

  ;; gc this code
  (set! tip-init noop))


;;-----------------------------------------------------------------------------
;; colour styles
;;
;; red/green widgets must be realized before `gtk-widget-get-style' gives
;; the proper style (otherwise just get the default style).  Some GtkStyle
;; setter functions would be much cleaner.

(define (style-proc typesym rc)
  (gtk-rc-parse-string rc)
  (gtk-class-new 'GtkDrawingArea (symbol->string typesym))
  (let ((widget (gtk-widget-new typesym)))
    (gtk-box-pack-start (gtk-dialog-vbox dialog) widget #f #f 0)
    (gtk-widget-realize widget)
    (gtk-widget-get-style widget)))

(define style-red
  (make-procedure-with-cache-0
   (lambda ()
     (style-proc 'ChartRed "\
style \"Chart_red\" { fg[NORMAL] = { 0.5, 0.0, 0.0 } }
widget \"*.ChartRed\" style \"Chart_red\"
"))))

(define style-green
  (make-procedure-with-cache-0
   (lambda ()
     (style-proc 'ChartGreen "\
style \"Chart_green\" { fg[NORMAL] = { 0.0, 0.5, 0.0 } }
widget \"*.ChartGreen\" style \"Chart_green\"
"))))

(define style-lightblue
  (make-procedure-with-cache-0
   (lambda ()
     (style-proc 'ChartLightblue "\
style \"Chart_lightblue\" { fg[NORMAL] = { 0.0, 0.0, 0.5 } }
widget \"*.ChartLightblue\" style \"Chart_lightblue\"
"))))


;;-----------------------------------------------------------------------------
;; watchlist dialog

(define dialog        #f)
(define clist         #f)
(define insert-button #f)
(define watch-symlist favourites-symlist)

(define (price accessor)
  (lambda (latest)
    (and-let* ((n (accessor latest)))
      (latest-form-price latest n))))

(define want-fill #t)
(define (call-without-fill thunk)
  (dynamic-wind
      (lambda () (set! want-fill #f))
      thunk
      (lambda () (set! want-fill #t))))

(define (row->symbol row)
  (list-ref-or-f (symlist-list watch-symlist) row))

;; return true if there's an intraday handler for SYMBOL
(define (intraday-handler? symbol)
  (any (lambda (elem)
	 ((first elem) symbol))
       intraday-handler-list))

(define (full-adate-time adate time)
  (let ((tm (adate->tm adate)))
    (if time
	(begin
	  (apply set-tm:hms tm (seconds->hms time))
	  (strftime-utf8 (string-append "%H:%M:%S  %a " (locale-d-fmt))
			 tm))
	(strftime-utf8 (string-append "%a " (locale-d-fmt)) tm))))

(define column-data
  `((,(_ "Symbol")    left  (str . "MMMMM.MMM")     #f ,latest-symbol)
    (,(_ "Bid/Offer") right (str . "99.99/99.99")   #t ,latest-form-bid/offer)
    (,(_ "Last")      right (str . "99999.99")      #t ,(price latest-last))
    (,(_ "Change")    right (str . "+999.99")       #t ,(price latest-change))
    (,(_ "High")      right (str . "99999.99")      #t ,(price latest-high))
    (,(_ "Low")       right (str . "99999.99")      #t ,(price latest-low))
    (,(_ "Volume")    right (str . "999m")          #t ,latest-form-volume-km)
    (,(_ "When")      left  (str . "16:00 ")        #t ,latest-form-date/time)
    (,(_ "Notes")     left  (em . 10)               #f ,latest-form-note)))

(define (init)
  (set! dialog (gtk-dialog-new-with-buttons
		(c-make-window-title (_ "Watchlist")) #f '()))
  (set! clist (gtk-clist-new-with-titles (map first column-data)))

  (gtk-dialog-add-buttons dialog
			  "gtk-refresh"   0
			  "gtk-delete"    1
			  (_ "_Intraday") 2
			  "gtk-close"     gtk-response-close
			  "gtk-help"      gtk-response-help)

  (let ((selected-row #f)
	(column-width (make-vector (length column-data)))
	(show-list    '())
	(show-rows    0)
	(show-dirty   (make-uniform-vector 0 #t #t)))

    (define (draw-row row symbol)
      (let ((latest (latest-get symbol)))
	(for-each
	 (lambda (col data)
	   (let* ((str       (or ((fifth data) latest) ""))
		  (new-width (gdk-string-width
			      (c-gtk-widget-get-font clist) str)))
	     (gtk-clist-set-text clist row col str)
	     ;; expand column, if column-data says want resize
	     (if (and (fourth (list-ref column-data col))
		      (> new-width (vector-ref column-width col)))
		 (begin
		   (gtk-clist-set-column-width clist col new-width)
		   (gtk-clist-set-column-min-width clist col new-width)
		   (vector-set! column-width col new-width)))))
	 (iota (length column-data))
	 column-data)

	(gtk-clist-set-row-style
	 clist row
	 (cond ((latest-in-progress? symbol)
		(style-lightblue))
	       ((positive-number? (latest-change latest))
		(style-green))
	       ((negative-number? (latest-change latest))
		(style-red))
	       (else
		#f)))))

    (define (draw-visible)
      (receive-list (lo hi)
	  (c-gtk-clist-visible-rows clist show-rows)
	(set! hi (min hi (array-last-index show-dirty)))
	(if (>= hi lo)
	    (let ((row-list (filter (lambda (row)
				      (array-ref show-dirty row))
				    (iota-range lo hi))))
	      (if (not (null? row-list))
		  (c-gtk-clist-call-with-freeze clist
		    (lambda ()
		      (for-each (lambda (row)
				  (array-set! show-dirty #f row)
				  (draw-row row (list-ref show-list row)))
				row-list))))))))

    ;; latest records for SYMBOL-LIST have changed
    (define (update-some symbol-list)
      (let ((draw noop))
	(do ((row 0         (1+ row))
	     (lst show-list (cdr lst)))
	    ((null? lst))
	  (if (member (car lst) symbol-list)
	      (begin
		(array-set! show-dirty #t row)
		(set! draw draw-visible))))
	(draw)))

    ;; fill or amend to display current watch-symlist
    (define (fill)
      (c-gtk-clist-call-with-freeze clist
	(lambda ()
	  (let ((editable (symlist-editable watch-symlist)))

	    (gtk-clist-set-reorderable clist editable)
	    (gtk-widget-set-sensitive insert-button editable)
	    (gtk-dialog-set-response-sensitive dialog 1 editable)  ;; delete
	    (update-intraday-sensitive)

	    (let* ((new-list  (symlist-list watch-symlist))
		   (new-rows  (length new-list))
		   (new-dirty (make-uniform-vector new-rows #t #t)))
	      (c-gtk-clist-change-rows clist show-rows new-rows)

	      ;; row still good if same symbol
	      (do ((row 0         (1+ row))
		   (old show-list (cdr old))
		   (new new-list  (cdr new)))
		  ((or (null? old)
		       (null? new)))

		(array-set! new-dirty (or (not (string=? (car old) (car new)))
					  (array-ref show-dirty row))
			    row))

	      (set! show-list  new-list)
	      (set! show-rows  new-rows)
	      (set! show-dirty new-dirty))
	    (draw-visible)))))

    (define (refresh)
      (latest-request-explicit (symlist-interesting watch-symlist)))

    (define (update-intraday-sensitive)
      (gtk-dialog-set-response-sensitive
       dialog 2 ;; intraday
       (and-let* (selected-row
		  ;; can be blank after deleting last in list
		  (symbol (c-gtk-clist-get-text clist selected-row 0)))
	 (intraday-handler? symbol))))

    (c-gtk-dialog-hide-on-delete dialog)
    (gtk-signal-connect dialog "response"
      (lambda (id)
	(cond ((= id 0)
	       (refresh))

	      ((= id 1) ;; delete
	       (cond (selected-row
		      (gtk-clist-remove clist selected-row)
		      (symlist-delete watch-symlist selected-row 1)
		      (set! selected-row #f)
		      (fill))
		     (else
		      (gdk-beep))))

	      ((= id 2) ;; intraday
	       (if selected-row
		   (intraday-popup (c-gtk-clist-get-text clist selected-row 0))
		   (gdk-beep)))

	      ((= id gtk-response-close)
	       (gtk-widget-hide dialog))

	      ((= id gtk-response-help)
	       (manual-popup (_ "CONTEXT(manual-node): Watchlist") dialog)))))

    (gtk-signal-connect clist "select_row"
      (lambda (row column event)
	(gtk-dialog-set-response-sensitive
	 dialog 1 ;; delete
	 (symlist-editable watch-symlist))
	(set! selected-row row)
	(update-intraday-sensitive)
	(if (and event ;; event is #f from keyboard
		 (eq? '2button-press (gdk-event-type event)))
	    (begin
	      (c-gtk-widget-raise app-widget)
	      (chart-open (smarker-goto (view-smarker outer)
					watch-symlist
					(c-gtk-clist-get-text clist row 0))
			  #:window dialog)))))

    (gtk-signal-connect clist "unselect_row"
      (lambda (row column event)
	(gtk-dialog-set-response-sensitive dialog 1 #f)   ;; delete
	(gtk-dialog-set-response-sensitive dialog 2 #f))) ;; intraday

    (gtk-signal-connect clist "row-move"
      (lambda (row new-row)
	;; move has been made in the clist, do the same in the symlist,
	;; ie. the symbol at new-row is to move to that position
	(symlist-move watch-symlist row 1 new-row)
	(fill)))

    (gtk-dialog-set-response-sensitive dialog 1 #f) ;; delete
    (gtk-dialog-set-response-sensitive dialog 2 #f) ;; intraday
    (notify-connect 'latest-update update-some)

    (let ((hbox   (gtk-hbox-new #f 4))  ;; don't stretch to width
	  (option (gtk-option-menu-new))
	  (menu   (gtk-menu-new)))

      (for-each (lambda (symlist)
		  (let ((item (gtk-menu-item-new-with-label
			       (symlist-name symlist))))
		    (gtk-widget-show item)
		    (gtk-menu-append menu item)
		    (gtk-signal-connect item "activate"
		      (lambda ()
			;; can be slow with a big list like "All"
			(c-gtk-widget-busy dialog)
			(set! watch-symlist symlist)
			(fill)))))
		symlist-all-lists)

      (gtk-option-menu-set-menu option menu)
      (gtk-option-menu-set-history
       option (list-index-elem watch-symlist symlist-all-lists eq?))

      (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0)
      (gtk-box-pack-start hbox option #f #f 0))

    (let* ((hbox    (gtk-hbox-new #f 4))
	   (vadj    (gtk-adjustment-new 0 0 0 0 0 0))
	   (vscroll (gtk-widget-new 'GtkVScrollbar
				    #:adjustment vadj
				    #:can_focus #f
				    #:update_policy 'continuous)))
      (gtk-clist-column-titles-passive clist)
      (gtk-clist-set-vadjustment clist vadj)
      (c-gtk-widget-set-usize-units clist -1 '(height . 18))
      (tip-init)
      (gtk-box-pack-start hbox clist #t #t 0)

      (gtk-box-pack-start hbox vscroll #f #f 0)
      (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #t #t 0)

      ;; FIXME: the "changed" signal comes out quite a few times when
      ;; setting a new symlist, but don't want to wait until idle,
      ;; should set row contents immediately when scrolling
      (gtk-signal-connect vadj "changed" draw-visible)
      (gtk-signal-connect vadj "value-changed" draw-visible)

      ;; page-up and page-down outside of clist
      ;; this is probably better done as a GtkBinding, or maybe as an
      ;; accelerator onto the scrollbar
      (gtk-widget-add-events dialog '(key-press-mask))
      (gtk-signal-connect dialog "key-press-event"
	(lambda (event)
	  ;; no modifiers except caps-lock, and outside of clist
	  (if (and (not (gtk-widget-has-focus clist))
		   (member (gdk-event-state event) '(() (lock-mask))))
	      (cond ((= (gdk-event-keyval event)
			(gdk-keyval-from-name "Page_Up"))
		     (c-gtk-adjustment-page-down vadj))
		    ((= (gdk-event-keyval event)
			(gdk-keyval-from-name "Page_Down"))
		     (c-gtk-adjustment-page-up vadj)))))))

    (set! insert-button (gtk-button-new-with-label (_ "Insert")))
    (let ((hbox  (gtk-hbox-new #f 4))
	  (entry (gtk-entry-new)))

      (define (entry-insert)
	(let ((symbol (gtk-entry-get-text entry)))
	  (cond
	   ((list-index-elem symbol show-list)
	    ;; already exists, move to it
	    => (lambda (row)
		 (gtk-clist-moveto clist row 0 0.5 0)
		 (gtk-clist-select-row clist row 0)))

	   ((symlist-editable watch-symlist) ;; only change if editable

	    ;; insert after selected row, if that's visible, otherwise at
	    ;; the top of the current visible portion
	    (let* ((visible (c-gtk-clist-visible-rows clist show-rows))
		   (pos (if (and selected-row
				 (within-range? (first visible) (second visible)
						selected-row))
			    (1+ selected-row)
			    (1+ (first visible)))))
	      (set! pos (min pos show-rows))

	      (gtk-clist-insert clist pos
				(cons symbol (make-list
					      (1- (length column-data)) "")))
	      ;; initial for this symbol, if anything known yet
	      (update-some (list symbol))
	      (gtk-clist-select-row clist pos 0)
	      (or (gtk-clist-row-is-visible clist pos)
		  (gtk-clist-moveto clist pos 0 0.5 0))

	      (call-without-fill
	       (lambda ()
		 (symlist-insert watch-symlist pos (list symbol))))

	      ;; request this symbol, with everything else as extras
	      (latest-request-symbols
	       (list symbol) (delete symbol (symlist-list watch-symlist)))

	      (fill))))))

      (gtk-box-pack-start (gtk-dialog-vbox dialog) hbox #f #f 0)
      (gtk-box-pack-start hbox (gtk-label-new (_ "Symbol")) #f #f 0)
      (gtk-signal-connect entry "activate" entry-insert)
      (c-gtk-widget-set-usize-units entry '(digit . 10) '(height . 1.5))
      (gtk-box-pack-start hbox entry  #t #t 0)
      (gtk-box-pack-start hbox insert-button #f #f 0)

      (gtk-signal-connect insert-button "clicked" entry-insert)

      ;; rowmenu popup
      (gtk-signal-connect clist "button_press_event"
	(lambda (event)
	  (if (= 3 (gdk-event-button event))
	      (and-let* ((symbol (row->symbol
				  (c-gtk-clist-event-y->row
				   clist (gdk-event-y event)))))
		(rowmenu-popup symbol))))))

    (for-each (lambda (col data)
		(let ((pixels
		       (max (c-gtk-widget-units->width clist (third data))
			    (gtk-clist-optimal-column-width clist col))))
		  (vector-set! column-width col pixels)
		  (gtk-clist-set-column-width clist col pixels)
		  (gtk-clist-set-column-min-width clist col pixels)))
	      (iota (length column-data)) column-data)

    (for-each (lambda (col data)
		(gtk-clist-set-column-justification clist
						    col (second data)))
	      (iota (length column-data)) column-data)

    (gtk-widget-grab-focus clist)

    ;; wait for dialog to map before filling, since it's not fast
    (gtk-signal-connect dialog "map-event"
      (lambda (event)
	(fill)
	(notify-connect 'symlist-update
	  (lambda (symlist-key-list)
	    (if (and want-fill
		     (memq (symlist-key watch-symlist)
			   symlist-key-list))
		(fill))))
	;; initial refresh, but after first contents sized and displayed,
	;; and only a plain `latest-request-symbols', don't need to force
	;; a re-download
	(gtk-idle-add-priority (1+ gtk-priority-redraw)
	  (lambda ()
	    (latest-request-symbols
	     (symlist-interesting watch-symlist)))))))

  ;; only once, and gc this code
  (set! init noop))


(define-public (watchlist-dialog-popup)
  (init)
  (c-gtk-widget-show-all-or-raise dialog))
