;; Winnipeg Commodity Exchange (WCE) setups.

;; Copyright 2005, 2006, 2007, 2008, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart wce)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:autoload   (chart barchart)  (barchart-five-download)
  #:use-module (chart download)
  #:use-module (chart futures)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc)
  #:use-module (chart pagebits)
  #:use-module (chart unzip))


;; Note: As of May 2007, the server at www.wce.ca (Microsoft-IIS/6.0)
;; appears to give invalid responses to HTTP/1.1 requests.  It sends back
;; chunked data, but without a 0-length chunk as an ending.  Believe that
;; ending is mandatory.  As a workaround pretend in requests that we only
;; know HTTP/1.0, to prevent chunked.
;;
;; This is apparently a bug in the server:
;;  http://blogs.msdn.com/wndp/archive/2006/08/14/a-chunked-encoding-story.aspx
;; and maybe:
;;  http://support.microsoft.com/?id=834027
;;


(define-public (wce-symbol? symbol)
  (string-suffix? ".WCE" symbol))

(define-public timezone-winnipeg
  (make-procedure-with-cache-0
   (lambda ()
     (timezone-choose (_ "Winnipeg")
		      '("America/Winnipeg"
			;; Canada: second Sun in Mar, to first Sun in Nov
			;; French abbreviations for the zones are HNC/HAC,
			;; but no need to translate that, we never print the
			;; zone
			"CST+6CDT,M3.2.0,M11.1.0/3")
		      "CST+6"))))

(symbol-timezone! wce-symbol? (_ "Winnipeg") timezone-winnipeg)
;; commented out in the manual for now
;; (source-help! wce-symbol?
;; 	      (_ "CONTEXT(manual-node): Winnipeg Commodity Exchange"))

;; return last trading tdate for given MDATE contract month
;; last trading is the business day before the 15th of the month
;; see specs http://www.wce.ca/education/contractspecs/canola_f.asp etc
;;     
(define (wce-mdate-expiry mdate)
  (receive-list (year month day)
      (mdate->ymd mdate)
    (1- (ymd->tdate-ceil year month 15))))

(define (wce-symbol-expiry symbol)
  (and-let* ((mdate (chart-symbol-mdate symbol)))
    (wce-mdate-expiry mdate)))


;;-----------------------------------------------------------------------------
;; weblink - contract specifications
;;
;; When there's no specs for a given symbol the page served is the main
;; list of available contracts,
;;
;;   http://www.wce.ca/ContractsMarketInfo.aspx?first=contractspecifications
;;

(weblink-handler! wce-symbol?
  (lambda (symbol)
    (list (_ "WCE _Contract Specifications")
	  (_ "Open web browser at the Winnipeg Commodity Exchange contract specifications for this commodity")
	  (string-append "http://www.wce.ca/ContractsMarketInfo.aspx?first=contractspecifications&Commodity=" (chart-symbol-commodity symbol)))))


;;-----------------------------------------------------------------------------
;; download - historical pagebits
;;
;; The list returned is
;;
;;     ((name lo hi url)
;;      (name lo hi url)
;;      ...)
;;

(define wce-hist-links-url
  "http://www.wce.ca/ContractsMarketInfo.aspx?first=settlementprices&second=historicaldata")

;; names to symbols hard coded here, the 
(define wce-hist-links-names
  '(;; update handled separately
    ("All Commodities Update"       . ())
    ("Canola Futures"               . ("RS"))
    ("Canola Meal Futures"          . ("RM"))
    ("Feed Pea & Field Pea Futures" . ("WP"   ;; feed peas
				       "WQ")) ;; field peas
    ("Feed Wheat Futures"           . ("WW"))
    ("Flaxseed Futures"             . ("WF"))
    ("Oats Futures"                 . ("WO"))
    ("Western Barley Futures"       . ("AB"))))

;; For date ranges like "Aug 1, 2005 - present"
;;                      "Nov 1980 - Jul 1993"
;; return a list (START-TDATE END-TDATE)
(define (wce-hist-date-range str)
  (let* ((m (must-match
	     ;;             1                 23           4                        56                 7
	     (string-match "([A-Z][a-z][a-z]) (([0-9]+), )?([0-9][0-9][0-9][0-9]) - (([A-Z][a-z][a-z]) ([0-9][0-9][0-9][0-9])|present)" str)))
	 (start-month (Mmm-str->month (match:substring m 1)))
	 (start-day   (if (match:start m 3)
			  (string->number (match:substring m 3))
			  1))
	 (start-year  (string->number (match:substring m 4)))
	 (end-month   (and (match:start m 6)
			   (Mmm-str->month (match:substring m 6))))
	 (end-year    (and (match:start m 7)
			   (string->number (match:substring m 7)))))
    (list (ymd->tdate start-year start-month start-day)
	  (if end-year
	      (ym-last->tdate end-year end-month)
	      (1+ (tdate-today (timezone-winnipeg)))))))

(define (wce-hist-links-parse body)
  (define row-list (html-table-rows-raw
		    body (string-contains body "Historical Futures Prices")))
  (list
   (cons "hist-list"
	 (map-rows!
	  (lambda (empty name dates notes)
	    (set! name (html-text-extract name))
	    (let ((commodity-list (assoc-ref wce-hist-links-names name)))
	      (or commodity-list
		  (begin
		    (set! commodity-list '())
		    (format (current-error-port)
			    "chart: WCE: warning, unrecognised historical futures: ~a\n" name)))
	      (receive-list (lo hi)
		  (wce-hist-date-range (html-text-extract dates))
		(let ((link (match:substring
			     (must-match (string-match "href=\"([^\"]+)\""
						       dates)) 1)))
		  (list #:commodity-list commodity-list
			#:lo   lo
			#:hi   hi
			#:name name
			#:url  (http-url-relative wce-hist-links-url link))))))
	  row-list))))

(define (wce-hist-links-list)
  (assoc-ref
   (pagebits-read #:filename  "wce-historical-2" ;; new hist-list keywords
		  #:status    (list (_ "WCE") (_ "historical downloads page"))
		  #:url       wce-hist-links-url
		  #:timezone  (timezone-winnipeg)
		  #:parse     wce-hist-links-parse
		  #:frequency 5) ;; once a week
   "hist-list"))


;;-----------------------------------------------------------------------------
;; download - historical
;;
;; This uses the zipped csv files listed at
;;
;;     http://www.wce.ca/ContractsMarketInfo.aspx?first=settlementprices&second=historicaldata
;;
;; Firstly the historical files are downloaded (by commodity) and then the
;; update file (with all commodities) brings it up to the most recent week.
;; That update file is
;;
;;     http://www.wce.ca/WebContent.aspx?Web_Content_Id=309
;;
;; The server used to give an ETag, but unfortunately no longer, so we can't
;; avoid a redownload if our idea of when it's updated is a bit out.
;;
;;
;; There's three csv formats, one handled by style-many and the other two by
;; style-few.  The two formats in "few" differ only in fields past the
;; volume, which are not used.
;;
;; The data from the historical (by commodity) zips is about 10 years worth,
;; and a total of about 1.5 Mbytes.  That's a bit too much to hold in
;; memory, especially since twice or perhaps three times that could be used
;; after string splitting etc.  So "unzip" is run multiple times and the
;; data processed into arrays as its read.
;;
;; "unzip" must be used for archive extraction, "gunzip" can't handle the
;; archives because they contain more than one file.


;; return mdate for STR like either "1-Aug-1997" or "2004/6/1"
(define (dd-Mmm-yyyy-or-yyyy/mm/dd-str->mdate str)
  (apply ymd->mdate
	 (if (eqv? #\/ (string-ref str 4))
	     (y/m/d-str->ymd str)
	     (d/m/y-str->ymd str))))

;; BODY is a zip file covering LO to HI, process it for SYMBOL-LIST
;; LO and/or HI can be #f to just use the range within the data
;;
(define (wce-process-zip symbol-list lo hi url headers body)
  (call-with-string-in-file body
    (lambda (zipname)
      (for-each (lambda (memname)
		  (if (string-suffix? ".csv" memname)
		      (wce-process-csv symbol-list lo hi url headers
				       (unzip-extract-member zipname
							     memname))))
		(sort! (unzip-list-files zipname)
		       string>?)))))

(define (wce-process-csv symbol-list lo hi url headers body)

  (define row-list (map! (lambda (row-str)
			   (string-split row-str #\,))
			 (string-split-lines body)))

  (let ((headings (first row-list)))
    (set! row-list (cdr row-list))
    (cond ((member headings '(;; some historical files
			      ("Date" "Symbol" "Commodity" "Contract"
			       "Open" "High" "Low" "Settlement"
			       "Volume" "Open Interest")
			      ;; FuturesUpdate.zip update files
			      ("Date" "Symbol" "Commodity" "Contract"
			       "Open" "High" "Low" "Settlement"
			       "Trade Volume" "EFPs+EFRs" "Total Volume"
			       "Deliveries" "Open Interest")))
	   (wce-process-csv-few symbol-list lo hi url headers row-list))

	  ((equal? headings '("Date" "Symbol" "Commodity" "Contract"
			      "Open Range 1" "Open Range 2" "Open"
			      "High" "High BAT" "Low" "Low BAT"
			      "Close Range 1" "Close Range 2" "Settlement"
			      "Trade Volume" "EFPs" "EFRs" "Deliveries"
			      "Total Volume" "Open Interest"))
	   (wce-process-csv-many symbol-list lo hi url headers row-list))

	  (else
	   (error "WCE data unrecognised headings:" headings)))))

;; for files with few fields, eg. ABfut198905-199307.csv
(define (wce-process-csv-few symbol-list lo hi url headers row-list)
  (download-process
   #:module        (_ "WCE")
   #:symbol-list   symbol-list
   #:copyright     "http://www.wce.ca/aboutus.aspx?first=legal"
   #:url           url
   #:headers       headers
   #:last-download #f
   #:row-list
   (map-rows! (lambda (tdate commodity+year+M name mdate open high low close
			     volume . rest)
		(define openint (last rest))

		;; mdate yyyy/mm/dd in old files, or dd-Mmm-yyyy in
		;; FuturesUpdate.zip
		(list #:tdate (y/m/d-str->tdate tdate)
		      #:mdate (dd-Mmm-yyyy-or-yyyy/mm/dd-str->mdate mdate)
		      #:commodity     (commodity+year+M->commodity
				       commodity+year+M)
		      #:name          name
		      #:open          open
		      #:high          high
		      #:low           low
		      #:close         close
		      #:volume        volume
		      #:open-interest openint))
	      row-list)))

;; for files with many fields, eg. ABfut199708-200107.csv
(define (wce-process-csv-many symbol-list lo hi url headers row-list)
  (download-process
   #:module      (_ "WCE")
   #:symbol-list symbol-list
   #:expiry-proc wce-symbol-expiry
   #:copyright   "http://www.wce.ca/aboutus.aspx?first=legal"
   #:url         url
   #:headers     headers
   #:row-list
   (map-rows! (lambda (tdate commodity+year+M name mdate
			     open1 open2 open high high-bat low low-bat
			     close1 close2 settle
			     volume efp efr deliveries total-volume openint)

		;; when no actual open use open1 which is a bid or something
		(if (string-null? open)
		    (set! open open1))

		(list #:tdate (y/m/d-str->tdate tdate)
		      #:mdate (dd-Mmm-yyyy-or-yyyy/mm/dd-str->mdate mdate)
		      #:commodity     (commodity+year+M->commodity
				       commodity+year+M)
		      #:name          name
		      #:open          open
		      #:high          high
		      #:low           low
		      #:close         settle
		      #:volume        volume
		      #:open-interest openint))
	      row-list)))

(define (commodity+year+M->commodity str)
  (string-take str (string-index str char-set:digit)))

;; return the tdate for the last trading day in YEAR, MONTH
(define (ym-last->tdate year month)
  (1- (ymd->tdate-ceil year (1+ month) 1)))

;; Process the historical downloads page BODY, downloading links from it
;; which can add to any in SYMBOL-LIST.
;;
(define (wce-download-hist symbol-list)
  ;; historical zip files go to 1 Aug 2005, inclusive
  ;; don't really want to hard code this, but also don't want to download
  ;; the page every day to check
  (if (any (lambda (symbol)
	     (>= (ymd->tdate 2005 8 1) (download-start-tdate symbol)))
	   symbol-list)

      (for-each
       (lambda (elem)
	 (let-keywords elem #f ((commodity-list #f)
				(lo #f)
				(hi #f)
				(name #f)
				(url #f))
	   (let ((symbol-list
		  (filter (lambda (symbol)
			    (and (member (chart-symbol-commodity symbol)
					 commodity-list)
				 (>= hi (download-start-tdate symbol))))
			  symbol-list)))
	     (if (not (null? symbol-list))
		 (begin
		   (download-status (_ "WCE") (_ "data")
				    (format #f "~s ~a" name commodity-list)
				    (tdate-range-string lo hi))
		   (receive (headers body)
		       (http-request url
				     #:protocol-level "1.0"
				     #:want-ok #t)
		     (for-each (lambda (symbol-list)
				 (wce-process-zip symbol-list
						  lo hi #f #f body))
			       (partition-equal chart-symbol-commodity
						symbol-list))))))))

       (wce-hist-links-list))))

;; Update zip goes up to end of previous week, the Last-Modified back when
;; it was FuturesUpdate.zip was header has been seen as
;;     Mon, 22 Aug 2005 19:41:46 GMT  = 12:40 winnipeg (dst)
;;     Mon, 29 Aug 2005 19:26:38 GMT
;; so try for new data at 1pm
;; 
(define (current-available-tdate)
  (let* ((tm    (localtime (current-time) (timezone-winnipeg)))
	 (tdate (tm->tdate tm)))
    ;; Monday before 1pm treated as prev Friday
    (if (and (= (tm:wday tm) 1)
	     (< (tm:hour tm) 13))
	(set! tdate (1- tdate)))
    ;; data goes up to Friday of week before TDATE
    (1- (round-down-multiple tdate 5))))

(define (wce-download-current symbol-list)
  (define avail-tdate (current-available-tdate))
  (if (any (lambda (symbol)
	     (>= avail-tdate (download-start-tdate symbol)))
	   symbol-list)
      (let ((url "http://www.wce.ca/WebContent.aspx?Web_Content_Id=309"))
	(download-status (_ "WCE") (_ "data") (_ "update"))
	(receive (headers body)
	    (apply http-request url
		   #:protocol-level "1.0"
		   #:want-ok #t
		   (download-tag-args url symbol-list))
	  (if (http-headers-ok? headers)
	      (for-each (lambda (symbol-list)
			  (wce-process-zip symbol-list #f #f
					   url headers body))
			(partition-equal chart-symbol-commodity
					 symbol-list)))))))

(define (wce-download symbol-list)
  (if (unzip-version)
      (begin
	(wce-download-hist symbol-list)
	(wce-download-current symbol-list)))
  (barchart-five-download symbol-list))

(download-handler! #:selector   wce-symbol?
		   #:update     wce-download
		   #:backto
		   (lambda (want-tdate symbol-list)
		     (download-message "WCE past data not yet supported")))

(download-now-handler! (lambda (symbol-list)
			 (and (any wce-symbol? symbol-list)
			      (download-now-commodity-all-months))))
(download-now-handler! (lambda (symbol-list)
			 (and (any wce-symbol? symbol-list)
			      (download-now-need-unzip))))


;;-----------------------------------------------------------------------------
;; latest - all months parse

;; return list of latest records
(define (all-body->latest-list body)

  ;; "Quote Time" seems gone from the web page as of Jan 08
  ;;
  ;; <b>Quote Time:&nbsp;</b></span><span id="lblQuoteTime">8/3/2005 1:15:00 PM
  ;;   (let* ((m           (must-match (regexp-exec #,(regexp "Quote Time:[^\r\n0-9]*([0-9]+/[0-9]+/[0-9]+) ([0-9]+):([0-9]+):([0-9]+) ([AP]M)") body)))
  ;; 	 (quote-adate (m/d/y-str->adate (match:substring m 1)))
  ;; 	 (hour        (string->number-err (match:substring m 2)))
  ;; 	 (quote-time  (hms->seconds
  ;; 		       (+ (modulo hour 12) ;; 12,1-11 becomes 0-11
  ;; 			  (if (string=? (match:substring m 5) "AM") 0 12))
  ;; 		       (string->number-err (match:substring m 3))
  ;; 		       (string->number-err (match:substring m 4))))

  (receive-list (quote-adate quote-time)
      (wce-quote-adate-time #f)
    (define outer-list (html-table-rows-raw
			body
			;; first <table> after this heading
			(string-contains body "Delayed Futures Quotes")))
    (define ret '())

    (let ((headings (map html-text-extract (first outer-list))))
      (or (equal? headings '("Commodity" "Commodity Market Details"))
	  (error "WCE: unrecognised market headings:" headings)))
    (set! outer-list (cdr outer-list))

    (for-each
     (lambda (outer)
       (receive-list (name table)
	   outer
	 ;; every second outer row is total volume by commodity
	 (if (and (not (string-contains-ci table "Volume Total"))
		  (string-contains-ci table "<table"))

	     (let ((row-list   (html-table-rows table)))
	       (set! name (html-text-extract name))

	       (let ((headings (first row-list)))
		 (or (equal? headings '("Contract" "Code" "Status"
					"Open" "High"
					"" "Low" "" "Quote" "" "Chg" "Bid"
					"" "Ask" "" "Vol"))
		     (error "WCE: unrecognised market headings:" headings))
		 (set! row-list (cdr row-list)))

	       ;; What values can the "Status" column take?  Have seen
	       ;; "Close".

	       (for-each
		(lambda (row)
		  (receive-list (month symbol status open high high-s low low-s
				       last last-s change
				       bid bid-b ask ask-a volume)
		      row
		    (define commodity (chart-symbol-commodity symbol))

		    ;; lose commodity "AB" etc from end of name (there's a <br>
		    ;; separator, but that's dropped by `html-text-extract')
		    (set! name (string-trim-right-str name commodity))

		    (set! ret (cons (latest-new
				     #:symbol (string-append symbol ".WCE")
				     #:name           name
				     #:quote-adate    quote-adate
				     #:quote-time     quote-time
				     #:bid            bid
				     #:offer          ask
				     #:last-adate     quote-adate
				     #:open           open
				     #:high           high
				     #:low            low
				     #:last           last
				     #:change         change
				     #:volume         volume
				     #:contract-mdate (m/y-str->mdate month)
				     #:source         'wce)
				    ret))))
		row-list)))))

     outer-list)

    (latest-list-add-front-months ret)))


;;-----------------------------------------------------------------------------
;; latest
;;
;; The all months page of 15 minute delayed quotes (difference between quote
;; time and current time during trading hours can be seen) is used.
;; Unfortunately it's about about 110 kbytes.
;;
;;     http://www.wce.ca/ContractsMarketInfo.aspx?first=delayedquotes&second=nearbymonths
;;
;; There's also a near months page at about 104.5 kbytes, or the same on the
;; home page at about 109 kbytes,
;;
;;     http://www.wce.ca/ContractsMarketInfo.aspx?first=delayedquotes&second=allmonths
;;     http://www.wce.ca
;;
;; But the saving for those hardly seems worthwhile.
;;

;; Intraday price links which used to be on the home page went to
;;
;;     http://exchanges.barchart.com/intra/wpg/wpg.htm
;;     http://exchanges.barchart.com/intra/wpg/wpgdww.htm
;;
;; which are smaller, at about 5k for one or 17k for all commodities.  But
;; those pages don't have volume figures.  Since they're much smaller maybe
;; they'd be good enough.  (The main barchart ifut pages at 35k per
;; commodity are also available, but they also don't have volume.)


(define (wce-latest-get symbol-list extra-list proc)
  (receive (headers body)
      (http-request "http://www.wce.ca/ContractsMarketInfo.aspx?first=delayedquotes&second=allmonths"
		    #:protocol-level "1.0"
		    #:want-ok #t)
    (proc (all-body->latest-list body))))

;; 9:30am to 1:15pm
(define (wce-quote-adate-time symbol)
  (tm->adate-time-within (localtime (- (current-time) #,(hms->seconds 0 10 0))
				    (timezone-winnipeg))
			 #,(hms->seconds 9 30 0)
			 #,(hms->seconds 13 15 0)))

(latest-handler! #:selector   wce-symbol?
		 #:handler    wce-latest-get
		 #:adate-time wce-quote-adate-time)
