;; Williams %R oscillator.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart williams-r)
  #:use-module (ice-9 q)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc))


(define-public (highlow-calc-proc count)

  ;; HIGH-Q is the high values for each day
  ;; LOW-Q is the corresponding low values for each day
  ;; POINTS is how many entries in HIGH-Q (and LOW-Q likewise)
  ;;
  (let ((high-q     (make-q))
	(low-q      (make-q))
	(points     0)
	(points-min (ceiling-exact (* indicator-minimum-fraction count))))

    (lambda (high low)
      (if (>= points count)
	  (begin
	    ;; full, drop old point
	    (deq! high-q)
	    (deq! low-q))
	  ;; not full, so POINTS increases
	  (set! points (1+ points)))

      ;; add new point
      (enq! high-q high)
      (enq! low-q  low)

      (if (>= points points-min)
	  (values (apply max (q-list high-q))
		  (apply min (q-list low-q)))
	  (values #f #f)))))


;; Return a procedure which calculates a Williams %R over a window of COUNT
;; many points.
;;
;; Each call (PROC high low close) enters a new point into the window, and
;; the return is the %R value up to (and including) that point, or #f if not
;; enough points yet.  CLOSE must be a number, but HIGH and/or LOW can be
;; #f, in which case the CLOSE is used for those.
;;
;; To prime the window initially, call PROC with COUNT-1 many points
;; preceding the first desired.
;;
(define-public (williams-%R-calc-proc count)
  (define highlow-proc (highlow-calc-proc count))

  (lambda (high low close)
    (receive (hh ll)
	;; use closes if no high/low data
	(highlow-proc (or high close) (or low close))
      (and hh
	   (let* ((diff (- hh ll)))
	     (if (zero? diff)
		 -50
		 (* 100.0 (/ (- close hh) diff))))))))


(define (williams-%R-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (start  (series-prev-n-closes-pos parent #:before lo #:count count))
	 (array  (series-array parent start hi))
	 (proc   (williams-%R-calc-proc count)))

    ;; warmup start (inclusive) to lo (exclusive)
    (do ((i start (1+ i)))
	((>= i lo))
      (if (array-ref array i 3)
	  (proc (array-ref array i 1)    ;; high
		(array-ref array i 2)    ;; low
		(array-ref array i 3)))) ;; close

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i 3)
	  (array-set! new-array (proc (array-ref array i 1)  ;; high
				      (array-ref array i 2)  ;; low
				      (array-ref array i 3)) ;; close
		      i)))))

(define-public (williams-%R-series series count)
  (ohlcv-required "Williams %R" series)
  (let ((new-series (series-copy series)))
    (series-name!      new-series (format #f "~a ~a" (_ "%R") count))
    (series-datatype!  new-series williams-%R-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series count)
    (series-require!   new-series (indicator-require-parts williams-%R-fill
							   (1- count)))
    (series-array!     new-series #f)
    (series-decimals!  new-series 0)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public williams-%R-datatype
  (datatype-new!
   #:key     'williams-r
   #:name    (_ "Williams %R")
   #:units   'negative-percentage
   #:input   datatype-ohlcv
   #:minimum -100
   #:maximum 0
   #:hlines  '(-20 -80)
   #:manual  (_ "CONTEXT(manual-node): Williams %R")
   #:proc    williams-%R-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'williams-%R-days
	  #:name    (_ "Days")
	  #:type    'integer
	  #:minimum 1
	  #:default 10))))
