;; Weighted moving average (WMA).

;; Copyright 2002, 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart wma)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; return the triangular number of N, ie. N*(N+1)/2
(define (triangular n)
  (ash (* n (1+ n)) -1))

;; return the dot product of two lists, eg. (a b c) . (d e f) = a*d+b*e+c*f
(define (dot-product x y)
  (apply + (map * x y)))

;; Return a procedure which calculates a weighted moving average over an
;; accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the WMA up to (and including) that value, or #f if there's not enough
;; points yet.
;;
;; DATA is a list of initial values to enter into the window, the same as if
;; PROC were called on them successively (from start to end).  COUNT-1 is
;; enough to ensure the first PROC call is based on a full window.
;;
(define*-public (wma-calc-proc count #:optional (data '()))
  (set! data (take-right-maybe data count))

  ;; SUM is plain sum of DATA
  ;; WSUM is weighted sum of DATA
  ;; POINTS is number of non-#f in DATA
  ;; DIV is divisor for weighting in DATA
  ;; DIV-MIN is minimum DIV for valid result

  (let* ((points  (length data))
	 (sum     (apply + data))
	 (wsum    (dot-product data (iota points (1+ (- count points)))))
	 (div     (exact->inexact ;; avoid exact fraction in guile 1.8
		   (- (triangular count) (triangular (- count points)))))
	 (div-min (ceiling-exact (* indicator-minimum-fraction
				    (triangular count)))))
    (set! data (indicator-circular count data))

    (lambda  (x)

      ;; drop old point, including one less of each in DATA
      (set! wsum (- wsum sum))
      (if (car data)
	  (set! sum (- sum (car data)))
	  (begin
	    (set! div    (+ div (- count points)))
	    (set! points (1+ points))))

      ;; hold this point
      (set-car! data x)
      (set! data (cdr data))

      ;; add this point
      (set! sum  (+ sum x))
      (set! wsum (+ wsum (* x count)))
      
      (and (>= div div-min)
	   (/ wsum div)))))

(define-public (wma-warmup-count count)
  (1- count))


(define (wma-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (wma    (wma-calc-proc count
				(series-prev-n-closes parent
						      #:before lo
						      #:count  (1- count)))))
    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (and-let* ((this (array-ref array i)))
	(array-set! new-array (wma this) i)))))

(define-public (wma-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "WMA") count))
    (series-datatype! new-series wma-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-require!  new-series (indicator-require-parts wma-fill (1- count)))
    (series-array!    new-series #f)

    new-series))

(define-public wma-datatype
  (datatype-new!
   #:key      'wma
   #:name     (_ "WMA - Weighted MA")
   #:priority 10
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Weighted Moving Average")
   #:proc     wma-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'wma-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
