;; Commodity Exchange Hannover (Warenterminbrse).  -*- coding: latin-1 -*-

;; Copyright 2005, 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart wtb)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (srfi srfi-14)
  #:autoload   (gtk gdk)  (gdk-screen-width gdk-screen-height)
  #:use-module (chart download)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (wtb-symbol? symbol)
  (string-suffix? ".WTB" symbol))

(symbol-timezone! wtb-symbol? (_ "Hannover") timezone-europe-cet)
(source-help! wtb-symbol?
	      (_ "CONTEXT(manual-node): Risk Management Exchange"))

;; eg. hogs in English and German,
;; http://www.wtb-hannover.de/content/maerkte/schweine/kontraktspezifikationen/en/index.shtml?b21&en
;; http://www.wtb-hannover.de/content/maerkte/schweine/kontraktspezifikationen/en/index.shtml?b21&de
;;
(define wtb-weblinks-alist
  '(("H"  ;; hogs
    . "http://www.wtb-hannover.de/content/maerkte/schweine/kontraktspezifikationen/en/index.shtml?b21&")
   ("F"  ;; piglets
    . "http://www.wtb-hannover.de/content/maerkte/ferkel/kontraktspezifikationen/en/index.shtml?b22&")
   ("P"  ;; table potatoes
    . "http://www.wtb-hannover.de/content/maerkte/kartoffeln/kontraktspezifikationen/en/index.shtml?b22&")
   ("XP" ;; european processing potatoes
    . "http://www.wtb-hannover.de/content/maerkte/veredelungskartoffeln/kontraktspezifikationen/en/index.shtml?b23&")
   ("LP" ;; london potatoes
    . "http://www.wtb-hannover.de/content/maerkte/londonpotatoes/kontraktspezifikationen/en/index.shtml?b23&")
   ("W"  ;; wheat
    . "http://www.wtb-hannover.de/content/maerkte/weizen2001/kontraktspezifikationen/en/index.shtml?b24&")
   ("B"  ;;  brewing barley
    . "http://www.wtb-hannover.de/content/maerkte/braugerste/kontraktspezifikationen/en/index.shtml?b25&")))

(weblink-handler! wtb-symbol?
  (lambda (symbol)
    (list (_ "RMX _Contract Specifications")
	  (_ "Open web browser at the Risk Management Exchange Hanover contract specifications for this commodity")
	  (and-let* ((elem (assoc (chart-symbol-commodity symbol)
				  wtb-weblinks-alist)))
	    (string-append (cdr elem)
			   (lang-select '(("de" "de")
					  ("en" "en"))))))))
 

;;-----------------------------------------------------------------------------
;; misc

;; find the last occurrance of PART in the portion of STR between START and END
;; return the index if found, or #f if not
;;
(define (string-rcontains str part start end)
  (let more ((end end))
    (and-let* ((idx (string-rindex str (string-ref part 0) start end)))
      (if (string-prefix? part str
			  0 (string-length part)
			  idx end)
	  idx
	  (more idx)))))

(define (table-before str pos)
  (string-rcontains str "<table" 0 pos))


;;-----------------------------------------------------------------------------
;; latest quotes - delayed
;;
;; The pages like the front month
;;
;;     http://www.wtb-hannover.de/content/maerkte/charts/charts.php?b24&en&produkt=3
;;
;; or for a specific month
;;
;;     http://www.wtb-hannover.de/content/maerkte/charts/charts.php?b23&en&produkt=12&bild=LPJ06&type=_LIN&zeit=monat


;; return list of latest records
(define (wtb-body->latest-list front headers body)
  (let* ((charset   (html-charset headers body))
	 (name-list (html-table-rows
		     body (table-before
			   body (or (string-contains ;; english
				     body ">Charts/Quotes<")
				    (string-contains ;; german
				     body ">Charts<")))))
	 ;; doesn't seem to be any latin1 in the names currently, but apply
	 ;; the conversion just in case
	 (name      (charset->utf8 charset (first (first name-list))))
	 (row-list  (apply zip (html-table-rows
				body (or (string-contains ;; english
					  body "Details")
					 (string-contains ;; german
					  body "Kursinformationen")))))
	 (ret       '()))

    (let ((headings (first row-list)))
      (or (member headings '(("Contract Code:"
			      "Last Trading Day:"
			      "Last Update:"
			      "Last Price:"
			      "Open:"
			      "High:"
			      "Low:"
			      "Previous Day Close:"
			      "Bid:"
			      "Bid Volume:"
			      "Ask:"
			      "Ask Volume:"
			      "Trend:"
			      "Close:"
			      "Open Interest:"
			      "Traded Volume:")
			     ("Kontrakt Code:"
			      "Letzter Handelstag:"
			      "Kurszeit:"
			      "Aktueller Kurs:"
			      "Erffnungskurs:"
			      "Tagesh&ouml;chstkurs:"
			      "Tagestiefstkurs:"
			      "Vortagesschlu&szlig;kurs:"
			      "Geldkurs:"
			      "Geld-Volumen:"
			      "Briefkurs:"
			      "Brief-Volumen:"
			      "Tendenz:"
			      "Heutiger Schlu&szlig;kurs:"
			      "Open Interest:"
			      "Heutiger Umsatz:")))
	  (error "WTB delayed: unrecognised data columns:" headings)))
    (set! row-list (cdr row-list))

    (receive-list (symbol expiry date/time last open high low prev
			  bid bid-volume ask ask-volume trend close
			  openint volume)
	(first row-list)

      ;; decide whether this is the front month by looking at what's listed
      ;; under "Current Futures", unless we explicitly downloaded for the
      ;; front (ie. FRONT already true)
      (if (not front)
	  (let ((current-list (html-table-rows-raw
			       body (table-before
				     body (or (string-contains ;; english
					       body "Current Futures:")
					      (string-contains ;; german
					       body "Laufende Futures:"))))))
	    (let ((headings (map string-trim-both (map html-text-extract
						       (first current-list)))))
	      (or (member headings '(("Current Futures:")    ;; english
				     ("Laufende Futures:"))) ;; german
		  (error "WTB delayed: unrecognised data columns:" headings)))
	    (set! current-list (cdr current-list))

	    (let ((m (must-match (regexp-exec #,(regexp "bild=([^&\"]+)")
					      (first (first current-list))))))
	      (set! front (string=? symbol (match:substring m 1))))))

      (let* ((adate #f)
	     (time  #f))
	;; date/time empty for expired contract
	(if (not (string-null? date/time))
	    (let ((m (must-match (regexp-exec #,(regexp "^([0-9][0-9][0-9][0-9]-[0-9]+-[0-9]+) ([0-9]+:[0-9][0-9]:[0-9][0-9])$") date/time))))
	      (set! adate (y/m/d-str->adate (match:substring m 1)))
	      (set! time  (h:m:s-str->seconds (match:substring m 2)))))

	(let ((mdate (chart-symbol-mdate symbol)))
	  (define (lat this-symbol)
	    (set! ret
		  (cons (latest-new #:symbol (string-append this-symbol ".WTB")
				    #:name           name
				    #:quote-adate    adate
				    #:quote-time     time
				    #:bid            bid
				    #:offer          ask
				    #:last-adate     adate
				    #:last-time      time
				    #:open           open
				    #:high           high
				    #:low            low
				    #:last           last
				    #:prev           prev
				    #:volume         volume
				    #:contract-mdate mdate
				    #:source         'wtb)
			ret)))
	  (lat symbol)

	  ;; for the front month we strip the weeks digit, so say 4-week
	  ;; hogs "H4V05.WTB" becomase just "H.WTB" not "H4.WTB" (hence not
	  ;; using `latest-list-add-front-months')
	  (if front
	      (lat (string-trim-right (chart-symbol-commodity symbol)
				      char-set:digit))))))

    ret))


;; hard coding this isn't good, it might be possible to parse this out of
;; the javascript code that makes the menus, or that might be even worse
;;
(define (commodity->b/prod commodity)
  (or (assoc-ref '(("H"       "21" "0")   ;; hogs
		   ("F"       "22" "11")  ;; piglets
		   ("PIGLETS" "22" "11")
		   ("P"       "22" "1")   ;; table potatoes
		   ("TAPX"    "22" "1")
		   ("XP"      "23" "10")  ;; processing potatoes
		   ("POTX"    "23" "10")
		   ("LP"      "23" "12")  ;; london potatoes
		   ("WTBBPC"  "23" "12")
		   ("B"       "21" "13")  ;; brewing barley
		   ("W"       "24" "3"))  ;; wheat
		 commodity)
      '(#f #f)))

(define (wtb-latest-get symbol-list extra-list proc)
  (let more ()
    (or (null? symbol-list)
	(let* ((symbol    (first symbol-list))
	       (commodity (chart-symbol-commodity symbol)))
	  (set! symbol-list (cdr symbol-list))

	  (receive-list (b prod)
	      (commodity->b/prod (string-trim-right commodity char-set:digit))
	    (if b
		(receive (headers body)
		    (http-request
		     (string-append
		      "http://www.wtb-hannover.de/content/maerkte/charts/charts.php?b" b
		      "&" (lang-select '(("en" "en")
					 ("de" "de" #:charset "ISO-8859-1")))
		      "&produkt=" prod

		      (if (or (> (string-length commodity) 2)
			      (chart-symbol-mdate symbol))
			  (string-append "&bild="
					 (chart-symbol-sans-dot symbol))
			  ""))
		     #:want-ok #t)

		  (let* ((latest-list (wtb-body->latest-list
				       (not (chart-symbol-mdate symbol))
				       headers body))
			 (this-symbol-list (map latest-symbol latest-list)))
		    (proc latest-list)

		    ;; front symbol download might cover a specific, or vice
		    ;; versa
		    (set! symbol-list
			  (lset-difference string=? symbol-list this-symbol-list))

		    (or (member symbol this-symbol-list)
			(error "WTB: quote page didn't have symbol:" symbol))))

		;; not in our b/prod table
		(proc (list (latest-new-unknown symbol)))))

	  (more)))))

(latest-handler! #:selector   wtb-symbol?
		 #:handler    wtb-latest-get
		 ;; #:adate-time wtb-quote-adate-time
		 )


;;-----------------------------------------------------------------------------
;; intraday
;;
;; This uses intraday and other graphs like
;;
;;     http://www.wtb-hannover.de/content/maerkte/charts/charts.php?b21&en&produkt=0&bild=H5Q07&type=_LIN&zeit=intraday
;;
;; which is an image
;;
;;     http://www.wtb-hannover.de/bilder/charts/gross/intraday/H5Q07_LIN.gif?1440-900_1077121124
;;     http://www.wtb-hannover.de/bilder/charts/gross/intraday/F5U05_LIN.gif?1024-768_1058201145
;;
;; The number at the end is a date/time generated by writeImg() in the
;; quotes page, dunno what it's used for though.

(define mode-5day (_ "_5 Days"))
(define mode-month (_ "1 _Month"))
(define mode-whole (_ "_Whole Contract"))

;; SYMBOL is a front month like "H.WTB", return a specific symbol like
;; "H5Q07.WTB"
;; this function copes with the weeks count on hogs (H) and piglets (F)
;; which varies with the contract month
;;
(define (wtb-front-symbol->specific symbol)
  (let* ((commodity (chart-symbol-commodity symbol))
	 (mdate     (latest-symbol-mdate symbol (tdate-today
						 (timezone-europe-cet))))
	 (lst       (directory-files latest-directory #f
		      (lambda (symbol)
			(and (equal? ".WTB" (chart-symbol-suffix symbol))
			     (equal? commodity (string-trim-right
						(chart-symbol-commodity symbol)
						char-set:digit))
			     (equal? mdate (chart-symbol-mdate symbol)))))))
    (first-or-f lst)))

(define (wtb-intraday-url symbol mode)
  (define commodity (chart-symbol-sans-dot symbol))

  ;; for front month symbol (and not an index symbol) go to what that front
  ;; month actually is
  (and-let* ((       (<= (string-length commodity) 2))
	     (       (not (chart-symbol-mdate symbol)))
	     (symbol (wtb-front-symbol->specific symbol)))
    (set! commodity (chart-symbol-sans-dot symbol)))

  (let ((tm (localtime (current-time))))
    (format #f "http://www.wtb-hannover.de/bilder/charts/gross/~a/~a_~a.gif?~a-~a_~a~a~a~a~a"
	    (cond ((string=? mode mode-5day)
		   "woche")
		  ((string=? mode mode-month)
		   "monat")
		  ((string=? mode mode-whole)
		   "gesamt")
		  (else ;; 1 Day
		   "intraday"))
	    commodity
	    ;; only line chart available for day and week
	    (if (or (string=? mode (_ "1 Month"))
		    (string=? mode (_ "Whole Contract")))
		"BAR"
		"LIN")
	    (gdk-screen-width) (gdk-screen-height)
	    (tm:year tm) (tm:mon tm) (tm:mday tm) (tm:hour tm) (tm:min tm))))

(intraday-url-proc! #:symbol-pred wtb-symbol?
		    #:url-proc    wtb-intraday-url
		    #:modes-proc  (lambda ()
				    (list (_ "_1 Day")
					  mode-5day
					  mode-month
					  mode-whole)))
