;; XLS file reading.

;; Copyright 2005, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Cribs:
;;
;; http://www.45.free.net/~vitus/ice/catdoc
;;     catdoc xls2csv program
;;
;; http://chicago.sourceforge.net/xlhtml
;;     xlhtml
;;
;; Spreadsheet::ParseExcel
;; on cpan http://www.perl.com/CPAN/modules/by-modules/Spreadsheet/ and mirrors


(define-module (chart xls)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart misc)
  #:use-module (chart ole))



;;-----------------------------------------------------------------------------
;; microsoft dates
;;
;; XLS file dates are integers or floats counting days starting from 0 at
;; midnight 1 Jan 1900, except that 1900 is (incorrectly) considered a leap
;; year (so the basis is 0 at midnight 31 Dec 1899).  Excel for Macintosh
;; sets an option in the file to count from 1 Jan 1904 instead, but
;; fortunately haven't encountered that.

;; http://support.microsoft.com/kb/214326/en-us
;;     Excel following lotus123 wrong leap year, and kept that way even in
;;     new versions.
;; http://support.microsoft.com/kb/214330/en-us
;;     1900 vs 1904 date styles.

;; return adate for MSDATE
(define-public (msdate->adate d)
  (- (floor-exact d) 25573))

;; return tdate for MSDATE
(define-public (msdate->tdate d)
  (adate->tdate (msdate->adate d)))


;;-----------------------------------------------------------------------------

;; return an unsigned number extracted from byte data in STR, START and LEN
;; or whole STR
(define* (unum str #:optional (offset 0) (len (string-length str)))
  (let ((val 0))
    (do ((i   0 (1+ i))
	 (fac 1 (* 256 fac)))
	((>= i len))
      (set! val (+ val (* fac (char->integer (string-ref str (+ offset i)))))))
    val))

;; return a signed number extracted from byte data in STR, START and LEN or
;; whole STR
(define* (num str #:optional (offset 0) (len (string-length str)))
  (let ((val (unum str offset len)))
    (let ((power (ash 1 (1- (* 8 len)))))
      (if (>= val power)
	  (set! val (- val (ash power 1)))))
    val))

;; return the bits of exact integer N interpreted as 64-bits of IEEE double
(define (ieee-bits->number n)
  (let* ((mantissa (bit-extract n 0 52))   ;; low 52 bits
	 (exponent (bit-extract n 52 63))  ;; next 11 bits
	 (sign     (bit-extract n 63 64))) ;; high bit

    ;; implicit 1 bit above mantissa, except for denorm
    (if (not (= 0 exponent))
	(set! mantissa (+ mantissa #x10000000000000 )))

    ;; possible negative
    (if (= 1 sign)
	(set! mantissa (- mantissa)))

    ;; exponent bias +1023, and point is above 52 bits of mantissa
    (* mantissa (expt 2.0 (- exponent 1023 52)))))

;; `substring' but taking START and LEN (instead of END)
(define (substring-len str start len)
  (substring str start (+ start len)))

;; Convert 32-bit N for an RK record.
;; Low 2 bits are a tag, &2 for int (instead of float), &1 for factor 0.01.
;; For a float the 30 further bits are the high 30 bits of an ieee double.
;; For an int the 30 further bits are the value directly.
;;
;; FIXME: don't know if an int is meant to be interpreted as a signed 30
;; bits, it hasn't arisen so far.
;;
(define (rk n)
  (let ((r (if (logtest n 2)
	       (ash n -2)
	       (ieee-bits->number (ash (logand n -4) 32)))))
    (if (logtest n 1)
	(list r 'cents)
	r)))

;; return LEN bytes at OFFSET in STR converted to an unsigned integer
;; (little endian)
(define (bytes->n str offset len)
  (let ((factor 1)
	(ret    0))
    (string-for-each (lambda (c)
		       (set! ret (+ ret (* factor (char->integer c))))
		       (set! factor (* factor 256)))
		     str offset (+ offset len))
    ret))

;; interpret STR as XLS OLE file contents and call (PROC typesymbol data...)
;; for certain records (those with actual cell contents)
;;
(define-public (xls-for-each proc str)

  (define sst-array (make-array #f '(0 0)))

  (let ((port (ole-open-string str "Workbook")))
    (let more ((head (read-bytes port 4)))
      (if (not (eof-object? head))
	  (let* ((rectype (bytes->short head 0))
		 (reclen  (bytes->short head 2))
		 (rec     (read-bytes port reclen)))

	    (case rectype
	      ((#x06) ;; FORMULA
	       (proc 'formula
		     (bytes->short rec 0) ;; row
		     (bytes->short rec 2) ;; col
		     (bytes->short rec 4) ;; format code
		     (ieee-bits->number (bytes->n rec 6 8))))

	      ((#xbd) ;; MULRK
	       (let ((row    (bytes->short rec 0))
		     (col    (bytes->short rec 2))
		     (endcol (bytes->short rec (- reclen 2))))
		 (do ((i   0 (1+ i))  ;; col to endcol inclusive
		      (pos 4 (+ pos 6)))
		     ((> i (- endcol col)))
		   (proc 'rk
			 row
			 (+ col i)
			 (bytes->short rec pos) ;; format code
			 (rk (bytes->long rec (+ pos 2)))))))

	      ((#xfc) ;; SST
	       (let ((total-ss (bytes->long rec 4))
		     (sstsize  (bytes->long rec 4))
		     (pos      8))
		 (set! sst-array (make-array #f (list 0 (1- total-ss))))
		 (do ((i 0 (1+ i)))
		     ((>= i sstsize))
		   (let* ((len     (bytes->short rec pos))
			  (flags   (num rec (+ pos 2) 1))
			  (richcnt 0))
		     (set! pos (+ pos 3))

		     ;; flags:
		     ;; 0x01  0=unibyte, 1=utf16
		     ;; 0x08  rich text
		     ;;
		     (or (memv flags '(0 8))
			 (error "xls: unrecognised SST flags:" flags))
		     (if (logtest flags 8)
			 (begin
			   ;; 2-byte count before string, then 4 bytes for
			   ;; each rich count after the string
			   (set! richcnt (bytes->short rec pos))
			   (set! pos (+ pos 2))))

		     (array-set! sst-array (substring-len rec pos len) i)
		     (set! pos (+ pos len (* 4 richcnt)))))))

	      ((#xfd) ;; LABELSST
	       (let ((string-num (bytes->short rec 6)))
		 (proc 'string
		       (bytes->short rec 0) ;; row
		       (bytes->short rec 2) ;; col
		       (and (array-in-bounds? sst-array string-num)
			    (array-ref sst-array string-num)))))

	      ((#x203) ;; NUMBER
	       (proc 'number
		     (bytes->short rec 0) ;; row
		     (bytes->short rec 2) ;; col
		     (bytes->short rec 4) ;; format code
		     (ieee-bits->number (num rec 6 8))))

	      ((#x27e) ;; RK
	       (proc 'rk
		     (bytes->short rec 0) ;; row
		     (bytes->short rec 2) ;; col
		     (bytes->short rec 4) ;; format code
		     (rk (bytes->long rec 6)))))

	    (more (read-bytes port 4)))))))

(define (xls-value elem)
  (case (first elem)
    ((formula number rk)
     (fifth elem))
    ((string)
     (fourth elem))
    (else
     #f)))

;; call (PROC array) for each row
;; ARRAY has a low index 0 and high whatever the row needs
;;
(define-public (xls-for-each-row proc str)
  (let ((prev-row #f)
	(lst      '()))
    (xls-for-each (lambda args
		    (if (equal? prev-row (second args))
			(set! lst (cons args lst))
			(begin
			  (if (not (null? lst))
			      (proc (row->array lst)))
			  (set! lst (list args))))
		    (set! prev-row (second args)))
		  str)
    (if (not (null? lst))
	(proc (row->array lst)))))

(define (row->array lst)
  (let* ((array (make-array #f (list 0 (apply max (map third lst))))))
    (for-each (lambda (elem)
		(array-set! array (xls-value elem) (third elem)))
	      lst)
    array))
