;; Yahoo Finance setups and data downloading.

;; Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart yahoo)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart database)
  #:use-module (chart download)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


;; The field keys for the quote URLs can be found at the following (open an
;; account to get to them).
;;
;;         http://edit.my.yahoo.com/config/edit_pfview?.vk=v1
;;         http://edit.finance.yahoo.com/e9?.intl=au
;;
;; http://quote.yahoo.com/d?
;;   s=  ;; symbol
;;   f=  ;; format, concat of the following
;;     s   ;; symbol
;;     n   ;; company name
;;     l1  ;; last price
;;     d1  ;; last trade date (in yahoo server timezone)
;;     t1  ;; last trade time (in yahoo server timezone)
;;     c1  ;; change
;;     p2  ;; percent change
;;     v   ;; volume
;;     a2  ;; average daily volume
;;     b   ;; bid
;;     b6  ;; bid size
;;     a   ;; ask
;;     a5  ;; ask size
;;     k1  ;; "time - last" (ECN), with <b> and <i> markup
;;     c6  ;; change (ECN)
;;     m2  ;; day's range (ECN)
;;     b3  ;; bid (ECN)
;;     b2  ;; ask (ECN)
;;     p   ;; previous close
;;     o   ;; today's open
;;     m   ;; day's range, eg. "1.23 - 4.56"
;;     w   ;; 52-week range, eg. "1.23 - 4.56"
;;     e   ;; earnings per share
;;     r   ;; p/e ratio
;;     d   ;; div per share
;;     q   ;; ex div date, eg. "Mar 31" or "N/A"
;;     r1  ;; div pay date
;;     y   ;; div yield
;;     j1  ;; market cap
;;     x   ;; stock exchange
;;     c4  ;; currency, eg. "AUD"
;;     i   ;; more info links, letters
;;         ;;    c=chart, n=news, p=profile, r=research, i=insider,
;;         ;;    m=message board (yahoo)
;;     k   ;; 52-week high
;;
;; Don't know what the distinction between b,a and b3,b2 quotes are actually
;; meant to be.
;;
;; For the Australian Stock Exchange, b,a are "N/A", and b3,b2 is the SEATS
;; best quote (which is usually pretty sensible, even overnight).
;;
;; For US stocks b,a seems to be "N/A", and b3,b2 an ecn quote.  The latter
;; has been seen a long way away from from recent trades though, eg. in BRK-A.
;;

;;-----------------------------------------------------------------------------
;; sub-setups

(symbol-setups! "^"     '(chart sandp))     ;; various worldwide
(symbol-setups! "^"     '(chart usa))       ;; too many to list explicitly
(symbol-setups! "^AFLI" '(chart australia))
(symbol-setups! "^AORD" '(chart australia))
(symbol-setups! "^ATLI" '(chart australia))
(symbol-setups! "^ATOI" '(chart australia))

(load "yahoo-setups.scm")


;;-----------------------------------------------------------------------------
;; yahoo handled symbols

(define yahoo-symbol-predlist '())
(define-public yahoo-futures-symbol-predlist '())

;; add PRED to `yahoo-symbol-predlist', if not already there
(define-public (yahoo-symbol-add! pred)
  (set! yahoo-symbol-predlist
	(lset-adjoin eq? yahoo-symbol-predlist pred)))

(define (yahoo-symbol? symbol)
  (or (yahoo-index-symbol? symbol)
      (any-pred yahoo-symbol-predlist symbol)))

(define-public (yahoo-index-symbol? symbol)
  (string-prefix? "^" symbol))

(define (yahoo-futures-symbol? symbol)
  (any-pred yahoo-futures-symbol-predlist symbol))

(define (yahoo-non-futures-symbol? symbol)
  (and (yahoo-symbol? symbol)
       (not (yahoo-futures-symbol? symbol))))

(source-help! (list yahoo-symbol? yahoo-index-symbol?)
	      (_ "CONTEXT(manual-node): Yahoo Finance")
	      -10) ;; priority below others like ASX


;;-----------------------------------------------------------------------------
;; URLs etc

;; this is the "download spreadsheet format" links under say
;;
;;     http://table.finance.yahoo.com/k?s=BHP.AX&g=d
;;
;; these have table.finance.yahoo.com, but that's a 301 Redirect to
;; itable.finance.yahoo.com, so go there directly.
;;
;; Eg. for 16Nov02 to 17Feb03 of GWT.AX
;;
;;     http://itable.finance.yahoo.com/table.csv?a=10&b=16&c=2002&d=1&e=17&f=2003&s=GWT.AX&y=0&g=d&ignore=.csv
;;
;; Note month is 0 to 11.  There's no apparent local servers for this, links
;; from say au.finance.yahoo.com just to table.finance.yahoo.com.
;;
;; If there's no data at all in the given range then a 404 error is
;; returned.
;;
(define (yahoo-data-url-ymd symbol s-tdate e-tdate)
  (receive-list (s-year s-month s-day)
      (tdate->ymd s-tdate)
    (receive-list (e-year e-month e-day)
	(tdate->ymd e-tdate)
      (format #f "http://itable.finance.yahoo.com/table.csv?s=~a&a=~a&b=~a&c=~a&d=~a&e=~a&f=~a&y=0&g=d&ignore=.csv"
	      (url-quote symbol)
	      (1- s-month) s-day s-year
	      (1- e-month) e-day e-year))))

;; `yahoo-quote-limit' is the max symbols in each request for quotes and info.
;; Finance::Quote::Yahoo uses a limit of 40 to stop the url getting too
;; long, which apparently some servers or proxies can't handle
;;
;; This isn't enforced exactly, the futures month probing can go about 3
;; symbols past it.
;;
(define yahoo-quote-limit 40)


;;-----------------------------------------------------------------------------
;; misc

;; return tdate for a dividend date STR from a quote, or #f if none
;; examples: "Jan  7"  	    ;; finance.yahoo.com
;;           " 5 Jan"  	    ;; au.finance, uk.finance
;;           "24-Sep-04"    ;; ABB.AX on finance.yahoo.com
;;           "24 Sep, 2004" ;; ABB.AX on au.finance
;;           "Sep 24, 2004" ;; ABB.AX on ca.finance
;;
;; an error is thrown for an unrecognised string, don't want some new form
;; to end up with dividends silently forgotten
;;
(define (div-date-str->tdate str)
  (and (string<> str "N/A") ;; not available, return #f
       (string<> str "")    ;; empty for unknown or something, return #f

       (or
	;; "Jan  7"
	(and (regexp-exec #,(regexp "^[A-Za-z]+ +[0-9]+$") str)
	     (m/d-str->tdate-nearest str))

	;; " 5 Jan"
	(and-let* ((m (regexp-exec #,(regexp "^ *([0-9]+ +[A-Za-z]+)$") str)))
	  (d/m-str->tdate-nearest (match:substring m 1)))

	;; "24-Sep-04"
	(and-let* ((m (regexp-exec #,(regexp "^ *([0-9]+-[A-Za-z]+-[0-9][0-9])$") str)))
	  (d/m/y-str->tdate (match:substring m 1)))

	;; "24 Sep, 2004"
	(and-let* ((m (regexp-exec #,(regexp "^ *([0-9]+ +[A-Za-z]+,? +[0-9][0-9][0-9][0-9])$") str)))
	  (d/m/y-str->tdate (match:substring m 1)))

	;; "Sep 24, 2004"
	(and-let* ((m (regexp-exec #,(regexp "^([A-Za-z]+ +[0-9]+,? +[0-9][0-9][0-9][0-9])$") str)))
	  (m/d/y-str->tdate (match:substring m 1)))

	(error "unrecognised dividend date:" str))))

;; return adate for a string like "April 20", using whatever year puts that
;; day/month closest to today
(define-public (m/d-str->tdate-nearest str)
  (receive-list (month day)
      (string-tokenize str date-part-char-set)
    (set! month (month-str->month month))
    (ymd->tdate (month->nearest-year month)
		month
		(mday-str->day day))))

;; return tdate for a date STR from historical data, or throw an error
;; examples: 
;;     "2005-03-07"   ;; AGK.AX seen in jan07, maybe transient
;;     "20-Aug-02"    ;; past format
;;
(define (hist-date-str->tdate str)
  (receive-list (a month b)
      (string-split str #\-)
    (let ((m (string->number month)))
      (if m
	  (ymd->tdate (yyy-str->year a)   ;; "2005-03-07"
		      m
		      (mday-str->day b))
	  (ymd->tdate (yyy-str->year b)   ;; "20-Aug-02"
		      (Mmm-str->month month)
		      (mday-str->day a))))))


;;-----------------------------------------------------------------------------
;; update
;;
;; some 0 open/high/low on oldish data, eg. 000010.SS


;; return true if X is an integer, or near enough to one
(define (nearly-integer? x)
  (< (abs (- x (round x))) 0.02))

;; return a list (NUM DEN) which is a fraction close to the flonum X
;; for example 1.4999 returns (3 2)
(define (close-fraction x)
  (or (any (lambda (den)
	     (let ((num (* x den)))
	       (and (nearly-integer? num)
		    (list (round-exact num) den))))
	   '(1 2 3 4 5 6 7 8 9 10))
      (list x 1)))

;; FORWARD is #t if downloading forwards through time, ie. for an update, or
;; #f if not, ie. download back through time
;;
;; LAST-DOWNLOAD is true if this chunk is the last being downloaded, which
;; means we can set the last-download info
;;
(define (yahoo-process-download symbol lo hi forward last-download body)
  (define commodity (chart-symbol-sans-dot symbol))
  (define row-list (map! (lambda (row-str)
			   (string-split row-str #\,))
			 (string-split-lines body)))
  (define prev-adj-factor #f)
  (define dividends       '())

  (let ((headings (first row-list)))
    (or (equal? headings '("Date" "Open" "High" "Low" "Close" "Volume"
			   "Adj. Close*"))
	(equal? headings '("Date" "Open" "High" "Low" "Close" "Volume"
			   "Adj Close"))  ;; new style Jan07
	(error "Yahoo data headings unrecognised:" headings)))
  (set! row-list (cdr row-list))

  (set! row-list
	(map-rows! (lambda (tdate open high low close volume adj-close)
		     (set! tdate (hist-date-str->tdate tdate))

		     ;; index with zero volume and open, high, low, close all
		     ;; the same means just a single close value, eg. ^WIL5
		     ;;
		     ;; stock with zero volume means no trades, set open,
		     ;; high, low, close to all #f; but before zapping them
		     ;; check they're all equal, in case the volume is
		     ;; somehow wrong
		     ;;
		     (if (and (equal? open high low close)
			      (0-string? volume))
			 (begin
			   (set! open #f)
			   (set! high #f)
			   (set! low #f)
			   (if (not (yahoo-index-symbol? symbol))
			       (set! close #f))))

		     ;; turn changes in the adjusted close to split markers
		     (if close
			 (let* ((adj-factor (/ (string->number close)
					       (string->number adj-close))))
			   (if prev-adj-factor
			       (let ((split (/ adj-factor prev-adj-factor)))
				 ;; 1.0 if no change, and allow some rounding
				 ;; margin
				 (if (> (abs (- split 1)) 0.05)
				     (set! dividends
					   (acons (1+ tdate)
						  (cons 'split
							(close-fraction split))
						  dividends)))))
			   (set! prev-adj-factor adj-factor)))

		     (list #:tdate     tdate
			   #:commodity commodity
			   #:open      open
			   #:high      high
			   #:low       low
			   #:close     close
			   #:volume    volume))
		   row-list))

  ;; When downloading forwards, ie. updating, consider that the HI requested
  ;; is covered, irrespective of what data is actually obtained.  This
  ;; avoids repeatedly attempting to download for a stock with no data at
  ;; all.  As a protection against data merely not yet available, the
  ;; download-done is limited to 2 weeks ago.
  ;;
  (download-process
   #:module        (_ "Yahoo")
   #:symbol-list   (list symbol)
   #:last-download last-download
   #:hi            (and forward
			(min hi (- (tdate-today (symbol-timezone symbol)) 10)))
   #:row-list      row-list
   #:dividends     dividends))

(define chunk-size 150)

;; Return the expected available tdate for SYMBOL data.
;;
;; This is only based on observation, it seems the previous day's trading
;; only becomes "historical" when the exchange opens for the next day.  So
;; that means historical goes to the day before yesterday, until after
;; 10:30am on a weekday in which case it's yesterday.
;;
(define (data-available-tdate symbol)
  (1- (tdate-today-after 10 30 (symbol-timezone symbol))))

;; return true if data found, false if not
;; FORWARD is #t if downloading forwards through time, ie. for an update, or
;; #f if not, ie. download back through time
;;
(define (yahoo-download-chunk symbol lo hi forward last-download)
  (download-status (_ "Yahoo") (_ "data") symbol (tdate-range-string lo hi))
  (receive (headers body)
      (http-request (yahoo-data-url-ymd symbol lo hi))
    (let ((status-code (http-headers-status-code headers)))
      (cond ((http-status-code-ok? status-code)
	     (yahoo-process-download symbol lo hi forward last-download body)
	     #t)
	    ((= 404 status-code)
	     ;; 404 when no data available for requested range
	     ;; dummy to date covered when going forwards
	     (if forward
		 (yahoo-process-download
		  symbol lo hi forward last-download
		  "Date,Open,High,Low,Close,Volume,Adj Close\r\n"))
	     #f)
	    (else
	     ;; anything else is bad
	     (http-headers-throw headers))))))

;; Update symbols, downloading from their current last data to today.
;;
(define (yahoo-update-data symbol-list)
  (for-each (lambda (symbol)
	      (define avail-tdate (data-available-tdate symbol))

	      ;; no more than CHUNK-SIZE in each download, and stop when
	      ;; AVAIL-TDATE is reached
	      (do ((lo (download-start-tdate symbol)))
		  ((> lo avail-tdate))
		(let ((hi (min avail-tdate (+ lo chunk-size))))
		  (yahoo-download-chunk symbol lo hi #t (eqv? hi avail-tdate))
		  (set! lo (1+ hi)))))

	    symbol-list))

(define-public (yahoo-backto-data want-tdate symbol-list)
  (for-each (lambda (symbol)
	      ;; no more than CHUNK-SIZE in each download, and stop when
	      ;; WANT-TDATE or an empty chunk is reached
	      (do ((hi (download-back-tdate symbol)))
		  ((< hi want-tdate))
		(let ((lo (max want-tdate (- hi chunk-size))))
		  (if (yahoo-download-chunk symbol lo hi #f #f)
		      (set! hi (1- lo))
		      (set! hi (1- want-tdate))))))

	    symbol-list))

(define-public (yahoo-backto want-tdate symbol-list)
  (yahoo-backto-data want-tdate symbol-list)
  (yahoo-backto-dividends want-tdate symbol-list))

;; data and dividends together, they both come from itable, so we can stop
;; if that server is down
;; no historical data for futures (yet)
;;
(download-handler! #:selector yahoo-non-futures-symbol?
		   #:update (lambda (symbol-list)
			      (yahoo-update-data symbol-list)
			      (yahoo-update-dividends symbol-list))
		   #:backto yahoo-backto)


;;-----------------------------------------------------------------------------
;; info

;; list of symbol predicate functions
(define not-info-list '())

;; PROC is a symbol test function, symbols passing it will not get yahoo
;; info downloaded.
;; This is used by modules like australia.scm which have their own better
;; info download.
;;
(define-public (yahoo-info-symbol-not! proc)
  (set! not-info-list (cons proc not-info-list)))

;; symbols for which to use `yahoo-update-info' to get name, currency, etc
(define (yahoo-info-symbol? symbol)
  (and (yahoo-symbol? symbol)
       (not (any (lambda (proc)
		   (proc symbol))
		 not-info-list))))

(define (process-info body)
  (for-each-line
   (lambda (line)
     (receive-list (symbol name exchange currency div-tdate div-amount)
	 (csv-split line)
       (download-status (_ "Yahoo") (_ "info") symbol)

       (c-idle)
       (database-call-modify-data symbol
	 (lambda (series)

	   (series-info-add! series 'yahoo-info (current-time))

	   ;; unknown stocks have name==symbol, and exchange=="N/A"
	   ;; don't change anything in that case, in particular leave the
	   ;; currency alone, since it comes back as garbage (different
	   ;; garbage each time)
	   (if (not (and (string=? symbol name)
			 (string=? exchange "N/A")))
	       (begin
		 (series-name! series name)

		 ;; exchange is "N/A" for an unknown stock
		 ;; not sure if that arises in other cases
		 (if (and (not (string-null? exchange))
			  (string<> "N/A" exchange))
		     (series-info-add! series 'stock-exchange exchange))

		 (and (not (string-null? currency))
		      (series-info-add! series 'currency currency))

		 (set! div-tdate (div-date-str->tdate div-tdate))
		 (if div-tdate ;; if not "N/A"
		     (begin
		       ;; 0.00 for amounts not yet known, seen for instance
		       ;; on some ASX trusts
		       (if (string=? "0.00" div-amount)
			   (set! div-amount (_ "unknown")))
		       (series-dividend-add! series
					     div-tdate div-amount)))))))))
   body))
  
;; update every 14 days, or if not previously fetched
;;
(define (info-want-update? symbol)
  (let ((last-time (database-read-info symbol 'yahoo-info)))
    (or (not last-time)
	(< last-time (- (current-time) (* 14 86400))))))

(define (yahoo-update-info symbol-list)
  (set! symbol-list (filter info-want-update? symbol-list))

  (let more ()
    (if (not (null? symbol-list))
	(let* ((this-list (take-maybe symbol-list yahoo-quote-limit)))
	  (set! symbol-list (drop-maybe symbol-list yahoo-quote-limit))

	  (download-status (_ "Yahoo") (_ "info")
			   (symbol-list-range-string this-list))

	  (receive (headers body)
	      (http-request
	       (string-append "http://" (preference-get
					 'yahoo-quote-host
					 "download.finance.yahoo.com")
			      "/d?f=snxc4qd&s="
			      (string-join (map url-quote this-list) ","))
	       #:want-ok #t)
	    (process-info body)
	    (more))))))

(download-handler! #:selector yahoo-info-symbol?
		   #:update   yahoo-update-info
		   #:backto   noop
		   #:priority -10) ;; after data downloads

;; name etc from barchart
(yahoo-info-symbol-not! yahoo-futures-symbol?)


;;-----------------------------------------------------------------------------
;; dividends

;; parse a yahoo dividend string into an "amount" object (or just return the
;; string unchanged if not recognised)
;;
(define (yahoo-dividend-parse str)
  (define orig str)
  (define extra-decimals 0)

  ;; have seen values like "0.048300" or "0.040000", trim zeros down to
  ;; cents, ie. 2 decimals
  (set! str (string-trim-to-2-decimals str))

  ;; "c" for cents
  (if (string-suffix-ci? "c" str)
      (begin
	(set! extra-decimals 2)
	(set! str (string-drop-right str 1))))

  (cond ((string=? "0.00" str)
	 '(amount unknown))

	((string-match "^([0-9]*)(.([0-9]+))?$|^([0-9]+)(.([0-9]*))$" str)
	 => (lambda (m)
	      (let* ((decimals (string-decimals str))
		     (amount   (string->number+decimals str decimals)))
		`(amount ,amount (decimals . ,(+ decimals extra-decimals))))))

	(else
	 orig)))

(define (process-dividends symbol body)
  (let ((row-list (map! (lambda (row-str)
			  (string-split row-str #\,))
			(string-split (string-trim-right body) #\nl))))

    (let ((headings (first row-list)))
      (or (equal? headings '("Date" "Dividends"))
	  (error "Yahoo dividends unrecognised format:" headings)))
    (set! row-list (cdr row-list))

    (database-call-modify-data symbol
      (lambda (series)
	(let ((old-dividends (series-dividends series)))
	  (for-each (lambda (row)
		      (receive-list (tdate amount)
			  row
			(set! tdate (hist-date-str->tdate tdate))
			(if (member (assq-ref tdate old-dividends)
				    (list #f (_ "unknown")))
			    (series-dividend-add!
			     series
			     tdate
			     (yahoo-dividend-parse amount)))))
		    row-list)
	  (series-info-add! series 'yahoo-dividends-time (current-time)))))))

;; update every 14 days, or if not previously fetched
;;
(define (dividends-want-update? symbol)
  (let ((last-time (database-read-info symbol 'yahoo-dividends-time)))
    (or (not last-time)
	(< last-time (- (current-time) (* 14 86400))))))

;; this is the "download spreadsheet format" links under
;;     http://table.finance.yahoo.com/k?s=BHP.AX&g=d
;; these have table.finance.yahoo.com, but that's a 301 Redirect to
;; itable.finance.yahoo.com, so go there directly
;;
(define (yahoo-dividends-url want-tdate symbol)
  (let* ((e-tdate (+ (tdate-today) 60))
	 (s-tdate (or want-tdate
		      (- e-tdate 60 initial-download-days))))
    (receive-list (s-year s-month s-day)
	(tdate->ymd s-tdate)
      (receive-list (e-year e-month e-day)
	  (tdate->ymd e-tdate)
	(format #f "http://itable.finance.yahoo.com/table.csv?s=~a&a=~a&b=~a&c=~a&d=~a&e=~a&f=~a&y=0&g=v&ignore=.csv"
		(url-quote symbol)
		(1- s-month) s-day s-year
		(1- e-month) e-day e-year)))))

(define (yahoo-update-dividends symbol-list)
  ;; no dividends for indexes
  (set! symbol-list (remove yahoo-index-symbol? symbol-list))

  (yahoo-backto-dividends #f (filter dividends-want-update? symbol-list)))

(define (yahoo-backto-dividends want-tdate symbol-list)
  ;; no dividends for indexes
  (set! symbol-list (remove yahoo-index-symbol? symbol-list))

  (for-each (lambda (symbol)
	      (download-status (_ "Yahoo") (_ "dividends") symbol)
	      (receive (headers body)
		  (http-request (yahoo-dividends-url want-tdate symbol))
		(let ((status-code (http-headers-status-code headers)))
		  (cond ((http-status-code-ok? status-code)
			 (process-dividends symbol body))
			((= 404 status-code)
			 ;; 404 for stocks with no info, eg. NABHA.AX
			 ;; write time in database to stop repeated attempts
			 (database-call-modify-data symbol
			   (lambda (series)
			     (series-info-add! series 'yahoo-dividends-time
					       (current-time)))))
			(else
			 ;; anything else is bad
			 (http-headers-throw headers))))))
	    symbol-list))


;;-----------------------------------------------------------------------------
;; latest - quote adate/time within trading hours

;; list of elems (PRED . MINUTES-OR-PROC)
(define yahoo-quote-delay-alist
  (list (cons yahoo-index-symbol? 0))) ;; believe all indices are real-time

;; set delay for PRED symbols
;; default is 15 if not otherwise setup, that being most exchanges
(define-public (yahoo-quote-delay! pred minutes)
  (yahoo-symbol-add! pred)
  (set! yahoo-quote-delay-alist
	(append! yahoo-quote-delay-alist (list (cons pred minutes)))))

;; return delay in seconds
(define (yahoo-quote-delay symbol exchange)
  (let ((period (sympred-assoc-ref yahoo-quote-delay-alist symbol)))
    (cond ((procedure? period)
	   (set! period (period symbol exchange)))
	  ((not period)
	   ;; default 15 mins for the majority of exchanges
	   (set! period 15)))
    (hms->seconds 0 period 0)))

;; list of elems (PRED . (LO HI))
(define yahoo-quote-lock-alist '())

(define-public (yahoo-quote-lock! pred lo hi)
  (yahoo-symbol-add! pred)
  (set! yahoo-quote-lock-alist (acons pred (list lo hi)
				      yahoo-quote-lock-alist)))

;; return a list (ADATE TIME) for a latest bid/offer quote on SYMBOL
;; this is the current time in the symbol's timezone with a delay and
;; possibly a lock to business hours
;;
;; EXCHANGE is used to distinguish US exchanges since the delay for these
;; differs, but the exchange is not actually indicated in the symbol
;;
(define (yahoo-quote-adate-time symbol exchange)
  (let ((tm   (localtime (- (current-time) (yahoo-quote-delay symbol exchange))
			 (symbol-timezone symbol)))
	(lock (sympred-assoc-ref yahoo-quote-lock-alist symbol)))
    (if lock
	(apply tm->adate-time-within tm lock)
	(tm->adate-time tm))))


;;-----------------------------------------------------------------------------
;; futures helpers
;;
;; The front month is found by requesting 4 months starting from the current
;; mdate, and checking which has a "name" field (something other than a copy
;; of the symbol, which is zapped by body->latest).

;; chart suffix versus yahoo suffix
(define-public yahoo-suffix-transform-alist '())

;; return SYMBOL with suffix converted from chart to yahoo
(define (yahoo-suffix-chart->yahoo symbol)
  (or (and-let* ((yahoo (assoc-ref yahoo-suffix-transform-alist
				   (chart-symbol-suffix symbol))))
	(string-append (chart-symbol-sans-dot symbol) yahoo))
      symbol))

;; return SYMBOL with suffix converted from yahoo to chart
(define (yahoo-suffix-yahoo->chart symbol)
  (or (and-let* ((chart (rassoc-ref yahoo-suffix-transform-alist
				    (chart-symbol-suffix symbol))))
	(string-append (chart-symbol-sans-dot symbol) chart))
      symbol))

;; return the mdate for front month SYMBOL, or #f if unknown
;; this looks at a current latest record
;;
(define (yahoo-futures-front-mdate symbol)
  (and-let* ((latest (latest-get symbol))
	     (       (eqv? (latest-quote-adate latest)
			   (adate-today (symbol-timezone symbol)))))
    (latest-contract-mdate latest)))

(define (chart-symbol-with-MYY-mdate symbol mdate)
  (string-append (chart-symbol-commodity symbol)
		 (mdate->MYY-str mdate)
		 (chart-symbol-suffix symbol)))

;; return a list of specific symbols to attempt for SYMBOL
(define (yahoo-futures-attempt-symbols symbol)
  (cond ((or (not (yahoo-futures-symbol? symbol)) ;; stock symbol
	     (chart-symbol-mdate symbol))         ;; or specific month
	 (list symbol))

	((yahoo-futures-front-mdate symbol)  ;; front is known
	 => (lambda (mdate)
	      (list (chart-symbol-with-MYY-mdate
		     (yahoo-suffix-chart->yahoo symbol) mdate))))

	(else
	 (set! symbol (yahoo-suffix-chart->yahoo symbol))
	 (map (lambda (mdate)
		(chart-symbol-with-MYY-mdate symbol mdate))
	      (iota 4 (mdate-today (symbol-timezone symbol)))))))

(define (yahoo-futures-add-front symbol-list latest-list)
  (define (match symbol latest)
    (and (string=? symbol
		   (chart-symbol-commodity+suffix (latest-symbol latest)))
	 (latest-name latest)))
  (for-each
   (lambda (symbol)
     (if (yahoo-futures-symbol? symbol)
	 (and-let* ((latest
		     (find-minimum (lambda (latest)
				     (and (match symbol latest)
					  (latest-contract-mdate latest)))
				   latest-list)))
	   (set! latest (record-copy latest))
	   (latest-symbol! latest symbol)
	   (set! latest-list (cons latest latest-list)))))
   symbol-list)

  latest-list)


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the csv quote download like
;;
;;     http://finance.yahoo.com/d/quotes.csv?f=snoml1d1t1c1vb3b2qdx&e=.csv&s=GM
;;
;; See notes at the start of this file for the f= format code letters.


;; for tm records X and Y, return true if date X < date Y
(define (tm-date< x y)
  (or (< (tm:year x) (tm:year y))
      (and (= (tm:year x) (tm:year y))
	   (or (< (tm:mon x) (tm:mon y))
	       (and (= (tm:mon x) (tm:mon y))
		    (< (tm:mday x) (tm:mday y)))))))

;; return a list (ADATE TIME) in TARGET-ZONE for yahoo "last trade" DATE-STR
;; and TIME-STR
;;
;; DATE-STR and TIME-STR are in YAHOO-ZONE (the server's zone)
;;
(define (ymd-hmx-strs->adate-time date-str time-str yahoo-zone target-zone)

  ;; date/time can be "N/A" for unknown
  ;; date seen "N/A" with time varying garbage like "-23:23am"
  ;; date seen as garbage like 8/210/20537
  ;;
  (if (or (string=? date-str "N/A")
	  (string=? time-str "N/A")
	  (not (= 5 (- (string-length date-str)
		       (string-index-right date-str #\/)))))
      (list #f #f)

      (let ((last-ymd (m/d/y-str->ymd date-str))
	    (last-hms (h:m:s-str->hms time-str))
	    (tm       (tm-new)))

	;; tm is ymd-hms and yahoo-zone dst
	(apply set-tm:ymd tm last-ymd)
	(apply set-tm:hms tm last-hms)
	(set-tm:isdst tm (tm:isdst (cdr (mktime tm yahoo-zone))))

	;; tm is now the desired date/time in yahoo-zone,
	;; convert to target-zone
	(tm->adate-time (localtime (car (mktime tm yahoo-zone))
				   target-zone)))))

;; return a dividend amount string (utf-8 encoding) for use in a latest
;; record note on QUOTE-TDATE or LAST-TDATE
;;
(define (database-latest-dividend-amount symbol quote-tdate last-tdate)
  (and-let* ((series (database-read-series symbol))
	     (div    (any (lambda (elem)
			    (and (or (eqv? (car elem) quote-tdate)
				     (eqv? (car elem) last-tdate))
				 (let ((div (cdr elem)))
				   (and (or (string? div)
					    (and (pair? div)
						 (eq? 'amount (first div))))
					div))))
			  (series-dividends series))))
    (if (string? div)
	div ;; utf-8 from database
	(let ((amount (second div))
	      (alist  (cddr div)))
	  (cond ((eq? amount 'unknown)
		 (_ "unknown"))
		((eq? amount 'to-be-advised)
		 (_ "T.B.A."))
		((number? amount)
		 (let* ((decimals (or (assq-ref alist 'decimals) 0))
			(str      (locale-number+decimals amount decimals)))
		   (if (assq-ref alist 'estimated)
		       (set! str (format #f (_ "est ~a") str)))
		   ;; not sure about showing this
		   (and-let* ((imputation (assq-ref alist 'imputation)))
		     (if (number? imputation)
			 (set! imputation
			       (locale-number+decimals imputation decimals)))
		     (set! str (format #f "~a + ~a" str imputation)))
		   str))
		(else
		 ;; utf-8 for any strings in database object, and the list
		 ;; layout bits will be ascii
		 (format #f "~s" div)))))))


;; Return a note string (in utf-8) showing a possible ex-dividend on the
;; given QUOTE-ADATE or LAST-ADATE.  Return #f if not ex today.
;;
;; Quite often yahoo seems to have only old dividends, so also look in the
;; database for information (eg. from newzealand.scm).
;;
(define (yahoo-latest-dividend-note symbol quote-adate last-adate
				    div-tdate div-amount)
  (define quote-tdate (and quote-adate (adate->tdate quote-adate)))
  (define last-tdate  (and last-adate  (adate->tdate last-adate)))

  ;; "0.00" for various unknown or estimated dividends, eg. from ASX trusts
  (if (string=? div-amount "0.00")
      (set! div-amount (_ "unknown")))

  ;; zap if not today
  (if (or (not div-tdate)
	  (and (not (eqv? div-tdate quote-tdate))
	       (not (eqv? div-tdate last-tdate))))
      (set! div-amount #f))

  ;; Look at the database too, because yahoo can be out of date and for
  ;; instance newzealand.scm may have better.
  (and-let* ((db-amount (database-latest-dividend-amount
			 symbol quote-tdate last-tdate)))
    (set! div-amount db-amount))

  (and div-amount
       (format #f (_ "ex ~a") div-amount)))

(define (yahoo-body->latest-list body)
  (let ((yahoo-zone (preference-get 'yahoo-quote-timezone (timezone-newyork))))
    (map
     (lambda (line)
       (receive-list (symbol name open range last last-adate last-time change
			     volume bid offer div-tdate div-amount exchange)
	   (csv-split (string-trim-right line)) ;; strip trailing CR

	 ;; on unknown stocks the symbol is repeated, which is pretty
	 ;; useless so zap that
	 (if (string=? name symbol)
	     (set! name #f))

	 (set! symbol (yahoo-suffix-yahoo->chart symbol))

	 (set! div-tdate (div-date-str->tdate div-tdate))

	 (receive-list (low high)
	     (map string-trim-both (string-split range #\-))

	   (receive-list (quote-adate quote-time)
	       (yahoo-quote-adate-time symbol exchange)

	     ;; convert date/time, plus mung timezone
	     (let ((adate-time (ymd-hmx-strs->adate-time
				last-adate last-time
				yahoo-zone (symbol-timezone symbol))))
	       (set! last-adate (first adate-time))
	       (set! last-time  (second adate-time)))

	     ;; note on ex-dividend day
	     (let ((note (yahoo-latest-dividend-note symbol
						     quote-adate last-adate
						     div-tdate div-amount)))
	       (latest-new
		#:symbol      	 symbol
		#:name        	 name
		#:open        	 open
		#:low         	 low
		#:high        	 high
		#:last        	 last
		#:last-adate  	 last-adate
		#:last-time   	 last-time
		#:change      	 change
		#:volume      	 volume
		#:bid         	 bid
		#:offer       	 offer
		#:quote-adate 	 quote-adate
		#:quote-time  	 quote-time
		#:limit-decimals 2  ;; mostly for australia ...
		#:contract-mdate (and (yahoo-futures-symbol? symbol)
				      (chart-symbol-mdate symbol))
		#:note           note
		#:source         'yahoo))))))

     (string-split-lines body))))

;; call (PROC latest) with latest records for each in SYMBOL-LIST, and
;; possibly some from of EXTRA-LIST if they can be done at the same time
;;
(define (yahoo-latest-get symbol-list extra-list proc)

  ;; Pick some symbols from the front of SYMBOL-LIST and possibly EXTRA-LIST.
  ;; Both are set to new values, with the taken symbols peeled off.
  ;;
  ;; The return is two values (RAW-LIST REQ-LIST), where the RAW-LIST is
  ;; symbols as from SYMBOL-LIST and EXTRA-LIST, and REQ-LIST is an expanded
  ;; list with extra symbols to request in order to cope with futures front
  ;; months.
  ;;
  (define (chunk)
    (let more ((raw '())
	       (req '()))
      (if (or (length>=? req yahoo-quote-limit)
	      (and (null? symbol-list)
		   (null? extra-list)))
	  (values raw req)

	  (if (not (null? symbol-list))
	      (let ((symbol (first symbol-list)))
		(set! symbol-list (cdr symbol-list))
		(more (append! raw (list symbol))
		      (append! req (yahoo-futures-attempt-symbols symbol))))

	      (let ((symbol (first extra-list)))
		(set! extra-list (cdr extra-list))
		(more (append! raw (list symbol))
		      (append! req (yahoo-futures-attempt-symbols symbol))))))))

  (let more ()
    (if (not (null? symbol-list))  ;; stop when all requested are done
	(receive (raw-list req-list)
	    (chunk)

	  (receive (headers body)
	      (http-request (string-append
			     "http://"
			     (preference-get 'yahoo-quote-host
					     "download.finance.yahoo.com")
			     "/d/quotes.csv?f=snoml1d1t1c1vb3b2qdx&e=.csv&s="
			     (string-join (map! url-quote
						(map yahoo-suffix-chart->yahoo
						     req-list))
					  ","))
			    #:want-ok #t)
	    (proc (yahoo-futures-add-front raw-list
					   (yahoo-body->latest-list body)))
	    (more))))))

(latest-handler! #:selector   yahoo-symbol?
		 #:handler    yahoo-latest-get
		 #:priority   -10
		 ;; the exchange in this call doesn't matter, there's no
		 ;; lock to business hours for nasdaq vs amex/nyse yet
		 #:adate-time (lambda (symbol)
				(yahoo-quote-adate-time symbol "")))


;;-----------------------------------------------------------------------------
;; setups - europe CET

;; not sure if these indexes are based on a particular exchange
(define (germany-symbol? symbol)
  (or (string-prefix? "^SX" symbol)   ;; stoxx various
      (and (yahoo-index-symbol? symbol)
	   (string-contains symbol "DAX"))))  ;; ^GDAXI germany
(symbol-timezone! germany-symbol? (_ "Germany") timezone-europe-cet)

(define (bremen-symbol? symbol)      (string-suffix? ".BM" symbol))
(symbol-timezone! bremen-symbol?     (_ "Bremen")      timezone-europe-cet)

(define (frankfurt-symbol? symbol)   (string-suffix? ".F" symbol))
(symbol-timezone! frankfurt-symbol?  (_ "Frankfurt")   timezone-europe-cet)

(define (hamburg-symbol? symbol)     (string-suffix? ".HM" symbol))
(symbol-timezone! hamburg-symbol?    (_ "Hamburg")     timezone-europe-cet)

(define (hanover-symbol? symbol)     (string-suffix? ".HA" symbol))
(symbol-timezone! hanover-symbol?    (_ "Hanover")     timezone-europe-cet)

(define (xetra-symbol? symbol)       (string-suffix? ".DE" symbol))
(symbol-timezone! xetra-symbol?      (_ "Xetra")       timezone-europe-cet)


;;-----------------------------------------------------------------------------
;; web link - basic quote page
;;
;; Eg. http://finance.yahoo.com/q?s=BHP.AX
;;

(weblink-handler! yahoo-symbol?
  (lambda (symbol)

    ;; accelerator "Y" so as not to clash with "S" for stock on an
    ;; exchange-specific links like "ASX IRM _Stock Information"
    ;;
    (list (_ "_Yahoo Stock Page")
	  (_ "Open web browser at the Yahoo quote page for this stock")
	  (lambda ()
	    (if (and (yahoo-futures-symbol? symbol)
		     (not (chart-symbol-mdate symbol)))
		(let* ((want-tdate (adate->tdate
				    (first
				     (yahoo-quote-adate-time symbol ""))))
		       (mdate (or (latest-symbol-mdate-nodownload symbol
								  want-tdate)
				  (begin
				    (weblink-message
				     (_ "Finding front month ..."))
				    (latest-symbol-mdate symbol want-tdate)))))
		  (if mdate ;; might still be unknown
		      (set! symbol
			    (string-append (chart-symbol-commodity symbol)
					   (mdate->MYY-str mdate)
					   (chart-symbol-suffix symbol))))))

	    ;; mung suffix
	    (set! symbol (yahoo-suffix-chart->yahoo symbol))

	    ;; with a handler procedure the current preference is followed
	    ;; even when changed during viewing a symbol
	    (string-append "http://" (preference-get 'yahoo-quote-host
						     "finance.yahoo.com")
			   "/q?s=" symbol)))))


;;-----------------------------------------------------------------------------
;; setups - intraday

;; Images are fetched from the yahoo charts section, gifs like
;;
;;     http://ichart.finance.yahoo.com/z?s=%5EGSPC&t=1d&q=l&l=off&z=l&p=s
;;
;; or the link from au.finance.yahoo.com is
;;
;;     http://cchart.yahoo.com/z?s=cml.ax&t=5d&l=off&z=l&q=l&i=au
;;
;; Those two hostnames resolve to the same IP, don't know which one is
;; really meant to be used.
;;
;; The parts are
;;
;;     s=SYMBOL
;;     t=1d   1 day
;;       5d   5 days
;;     q=l    line
;;       b    bar
;;       c    candle
;;     l=on   logarithmic
;;       off  linear
;;     z=m    medium size
;;       l    large size
;;     a=   comma separated list of indicators
;;       v    volume
;;       vm   volume moving average
;;       r14  RSI
;;
;; Unfortunately there's no last-modified or etag to indicate when the image
;; has nothing new yet, or is unchanged outside trading hours.

(define mode-5day (_ "_5 Days"))

(define (yahoo-intraday-url symbol mode)
  (format #f "http://ichart.finance.yahoo.com/z?s=~a&t=~a&l=off&z=m&q=l&a=v"
	  (url-quote symbol)
	  (cond ((string=? mode mode-5day)
		 "5d")
		(else ;; 1 Day
		 "1d"))))

;; the futures charts from yahoo don't look too good, eg OU06.CBOT, so stay
;; with barchart for them
;;
(intraday-url-proc! #:symbol-pred yahoo-non-futures-symbol?
		    #:url-proc    yahoo-intraday-url
		    #:modes-proc  (lambda ()
				    (list (_ "_1 Day")
					  mode-5day)))
