;; Zagreb Stock Exchange (ZSE) setups.     -*- coding: utf-8 -*-

;; Copyright 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart zagreb)
  #:autoload   (ice-9 format)  (format)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13)
  #:use-module (chart cookies)
  #:use-module (chart download)
  #:use-module (chart extra-glib)
  #:use-module (chart html)
  #:use-module (chart http)
  #:use-module (chart latest)
  #:use-module (chart misc))


(define-public (zagreb-symbol? symbol)
  (string-suffix? ".ZAG" symbol))

(symbol-timezone! zagreb-symbol? (_ "Zagreb") timezone-europe-cet)
(source-help! zagreb-symbol?
	      (_ "CONTEXT(manual-node): Zagreb Stock Exchange"))

(define zagreb-cookiejar-filename
  (g-build-filename cache-directory "zagreb-cookies"))


;;-----------------------------------------------------------------------------
;; weblink - ZSE company info

;; eg. in Croatian and English,
;; http://www.zse.hr/security.php?ticker=DLKV-M-704A&languageId=HR
;; http://www.zse.hr/security.php?ticker=DLKV-M-704A&languageId=EN
;;
(weblink-handler! zagreb-symbol?
  (lambda (symbol)
    (list (_ "ZSE _Company Information")
	  (_ "Open web browser at the Zagreb Stock Exchange information page for this company")
	  (string-append "http://www.zse.hr/security.php?ticker="
			 (chart-symbol-sans-dot symbol)
			 "&languageId="
			 (lang-select '(("hr" "HR")
					("en" "EN")))))))


;;-----------------------------------------------------------------------------
;; download - data
;;
;; This uses the historical data under individual shares like
;;
;;     http://www.zse.hr/default.aspx?id=17560&dionica=ISTR-R-A
;;
;; The "Save as Excel" link like
;;
;;     http://www.zse.hr/export.aspx?ticker=ISTR-R-A&reporttype=security&DateTo=9.1.2008&DateFrom=10.11.2007&range=-1
;;
;; gives some rough HTML with date, close, high, low, change, volume.
;; Headings are utf-8.

(define (zagreb-process-download symbol body last-download)
  (define row-list (html-table-rows body))
  (define commodity (chart-symbol-commodity symbol))

  (let ((headings (first row-list)))
    (or (equal? headings '("Datum" "Vrsta prometa" "Zadnja"
			   "Najviša" "Najniža" "Promjena" "Količina"
			   "Promet"))
	(error "ZSE unrecognised data headings:" headings)))
  (set! row-list (cdr row-list))

  (download-process
   #:module          (_ "ZSE")
   #:symbol-list     (list symbol)
   #:currency        "HRK"  ;; Kuna
   #:last-download   last-download
   #:prefer-decimals 2
   #:row-list
   (map-rows! (lambda (tdate something close high low change-percent
			     volume dollar-volume)

		;; date like "3.1.2008 0:00:00" for 3jan08, lose time part
		(set! tdate (string-trim-right-str tdate " 0:00:00"))

		(list #:tdate     (d/m/y-str->tdate tdate)
		      #:commodity commodity
		      #:high      (string-comma->dot high)
		      #:low       (string-comma->dot low)
		      #:close     (string-comma->dot close)
		      #:volume    volume))
	      row-list)))

(define (zagreb-download-url symbol start-tdate end-tdate)
  (receive-list (s-year s-month s-day)
      (tdate->ymd start-tdate)
    (receive-list (e-year e-month e-day)
	(tdate->ymd end-tdate)
      (format #f "http://www.zse.hr/export.aspx?ticker=~a&reporttype=security&DateTo=~d.~d.~4,'0d&DateFrom=~d.~d.~4,'0d&range=-1"
	      (chart-symbol-sans-dot symbol)
	      e-day e-month e-year
	      s-day s-month s-year))))

(define (zagreb-download-range symbol lo hi last-download)
  (download-status (_ "ZSE") (_ "data") symbol (tdate-range-string lo hi))
  (receive (headers body)
      (http-request (zagreb-download-url symbol lo hi)
		    #:cookiejar zagreb-cookiejar-filename
		    #:want-ok #t)
    (zagreb-process-download symbol body last-download)))

;; There's no Last-Modified on http://www.zse.hr/dailyreport.php?languageId=EN
;; to indicate when new data might be available, so guess midnight.
;;
(define (zagreb-available-tdate)
  (1- (tdate-today-after 0 0 (timezone-europe-cet))))

(define (zagreb-download symbol-list)
  (define avail-tdate (zagreb-available-tdate))
  (for-each (lambda (symbol)
	      (zagreb-download-range symbol
				     (download-start-tdate symbol)
				     avail-tdate
				     #t))
	    symbol-list)
  (download-name-from-latest (_ "ZSE") symbol-list))

(define-public (zagreb-backto want-tdate symbol-list)
  (for-each (lambda (symbol)
	      (zagreb-download-range symbol
				     want-tdate
				     (download-back-tdate symbol)
				     #f))
	    symbol-list))

(download-handler! #:selector   zagreb-symbol?
		   #:update     zagreb-download
		   #:avail-proc zagreb-available-tdate
		   #:backto     zagreb-backto)


;;-----------------------------------------------------------------------------
;; latest
;;
;; This uses the "about share" pages like
;;
;;     http://www.zse.hr/default.aspx?id=17560&dionica=ADRS-P-A
;;
;; The change is only a percentage, so might end up rounded a bit, but
;; should be close enough for a quote.


(define (change%->change value change-%)
  (let ((decimals (string-decimals value)))
    (set! value    (string->number-err value))
    (set! change-% (string->number-err change-%))

    ;; value = prev * (1 + change%/100)
    ;; so       prev = value * 100/(100+change%)
    ;; want     change = value - prev
    ;; which is change = value * (1 - 100/(100+change%))
    ;;                 = value * -change% / (100+change%)
    ;;                 = value * change% / (100+change%)
    ;;
    (format #f "~,vf" decimals (* value (/ change-% (+ change-% 100.0))))))

(define (zagreb-span body key)
  (let* ((beg (or (string-contains-after body
					 (string-append "<span id='" key "'>"))
		  (error "ZSE: quote field not found:" key)))
	 (end (string-contains body "</span>" beg)))
    (html-text-extract (substring body beg end))))

(define (zagreb-body->latest body)

  ;; Use this to print all the span bits
  ;;   (fold-matches "<span id=\"[^\"]+\">[^<]*" body #f
  ;; 		(lambda (m prev)
  ;; 		  (display (match:substring m 0)) (newline)))
  
  (let ((charset (html-charset "" body)))
    (or (string-ci=? charset "utf-8")
	(error "ZSE: oops, expecting UTF-8")))

  (let* (;; eg. <h3>PLVA-R-A</h3>&nbsp;&nbsp;(Pliva farmaceutika, dioničko društvo)
	 (m      (regexp-exec #,(regexp "<h3>([-A-Z]+)</h3>(&nbsp;)*[(]?([^<]*)") body))
	 (symbol (string-append (match:substring m 1) ".ZAG"))
	 (name   (match:substring m 3))

	 ;; date like "8.1.2008" for 8jan08
	 (adate       (d/m/y-str->adate (zagreb-span body "Date_Time")))

	 (bidask      (zagreb-span body "BidAsk"))
	 (bidask-list (string-split bidask #\/))
	 (bid         (first bidask-list))
	 (offer       (second bidask-list))

	 (open        (zagreb-span body "Open"))
	 (high        (zagreb-span body "High"))
	 (low         (zagreb-span body "Low"))
	 (last        (zagreb-span body "Last_Price"))
	 (change      (zagreb-span body "Last_Price_Change1"))
	 (volume      (zagreb-span body "Last_Volume")))

    (set! name (string-trim-right name #\)))

    ;; when no data there's a header and footer but nothing in between
    ;;     (if (null? row-list)
    ;; 	(latest-new #:symbol     symbol
    ;; 		    #:note       (_ "No data")
    ;; 		    #:source     'zagreb)

    ;; dot for thousands separator, commas for decimal point, even in
    ;; english pages
    (set! bid       (string-comma->dot (string-delete bid    #\.)))
    (set! offer     (string-comma->dot (string-delete offer  #\.)))
    (set! open      (string-comma->dot (string-delete open   #\.)))
    (set! high      (string-comma->dot (string-delete high   #\.)))
    (set! low       (string-comma->dot (string-delete low    #\.)))
    (set! last      (string-comma->dot (string-delete last   #\.)))
    (set! change    (string-comma->dot (string-delete change #\.)))
    (set! volume    (string-comma->dot (string-delete volume #\.)))

    (set! change (string-trim-right change #\%))
    (set! change (change%->change last change))

    (latest-new #:symbol      symbol
		#:name        name
		#:quote-adate adate
		#:bid         bid
		#:offer       offer
		#:last-adate  adate
		#:open        open
		#:high        high
		#:low         low
		#:last        last
		#:change      change
		#:volume      volume
		#:source      'zagreb)))

;; return true if data found, false if not
(define (zagreb-latest-get symbol-list extra-list proc)
  (for-each (lambda (symbol)
	      (receive (headers body)
		  (http-request
		   (string-append "http://www.zse.hr/default.aspx?id=17560&dionica="
				  (chart-symbol-sans-dot symbol))
		   #:cookiejar zagreb-cookiejar-filename
		   #:want-ok #t)
		(proc (list (zagreb-body->latest body)))))
	    symbol-list))

(define (zagreb-quote-adate-time symbol)
  (list (tdate->adate (zagreb-available-tdate))
	#,(hms->seconds 17 0 0)))

(latest-handler! #:selector   zagreb-symbol?  
		 #:handler    zagreb-latest-get
		 #:adate-time zagreb-quote-adate-time)
