;; Zig Zag indicator.

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart zigzag)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 receive)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://stockcharts.com/education/IndicatorAnalysis/indic_ZigZag.html
;;     Sample HPQ chart 1999/2000.
;;


;; This does the whole series from start to end.  It should be possible to
;; work back looking for a PERCENT move which would establish the direction
;; and hence a starting point in the middle of the data.
;;
(define (zigzag-fill series lo hi)

  (set! lo (series-lo series))
  (set! hi (series-hi series))

  (receive-list (percent use-closes)
      (series-data series)

    (let* ((parent          (series-parent series))
	   (array           (series-array parent lo hi))
	   (close-arg       (series-closes-apply-arg parent))
	   (high-arg        close-arg)
	   (low-arg         close-arg)
	   (new-array       (make-array #f (list lo hi)))
	   (factor-increase (1+ (* 0.01 percent)))
	   (factor-decrease (/ factor-increase))
	   (direction       noop)
	   (extreme         #f)
	   (target          #f)
	   (extreme-pos     #f))

      (define (rising pos high low)
	(cond ((or (not extreme)
		   (> high extreme))
	       (set! extreme     high)
	       (set! extreme-pos pos)
	       (set! target      (* extreme factor-decrease))
	       (values #f #f))

	      ((<= low target)
	       (let ((ret-pos extreme-pos)
		     (ret-val extreme))
		 (set! direction   falling)
		 (set! extreme     low)
		 (set! extreme-pos pos)
		 (set! target      (* extreme factor-increase))
		 (values ret-pos ret-val)))

	      (else
	       (values #f #f))))

      (define (falling pos high low)
	(cond ((or (not extreme)
		   (< low extreme))
	       (set! extreme     low)
	       (set! extreme-pos pos)
	       (set! target      (* extreme factor-increase))
	       (values #f #f))

	      ((>= high target)
	       (let ((ret-pos extreme-pos)
		     (ret-val extreme))
		 (set! direction   rising)
		 (set! extreme     high)
		 (set! extreme-pos pos)
		 (set! target      (* extreme factor-decrease))
		 (values ret-pos ret-val)))

	      (else
	       (values #f #f))))

      (if (and (eq? datatype-ohlcv (series-datatype parent))
	       (not use-closes))
	  (begin
	    (set! high-arg '(1))
	    (set! low-arg  '(2))))

      (let ((high     #f)
	    (high-pos #f)
	    (low      #f)
	    (low-pos  #f))
	(do ((i lo (1+ i)))
	    ((> i hi))
	  (and-let* ((close (apply array-ref array i close-arg)))

	    (let ((this-high (or (apply array-ref array i high-arg) close)))
	      (if (or (not high)
		      (> this-high high))
		  (begin
		    (set! high this-high)
		    (set! high-pos i))))

	    (let ((this-low (or (apply array-ref array i low-arg) close)))
	      (if (or (not low)
		      (< this-low low))
		  (begin
		    (set! low this-low)
		    (set! low-pos i))))

	    (if (>= high (* low factor-increase))
		(begin
		  (if (> high-pos low-pos)
		      (begin
			(set! direction rising)
			(array-set! new-array low low-pos)
			(array-set! new-array low lo)
			(set! i hi)))

		  (if (> low-pos high-pos)
		      (begin
			(set! direction falling)
			(array-set! new-array high high-pos)
			(array-set! new-array high lo)
			(set! i hi))))))))

      (do ((i lo (1+ i)))
	  ((> i hi))
	(and-let* ((close (apply array-ref array i close-arg)))
	  (let ((high (or (apply array-ref array i high-arg) close))
		(low  (or (apply array-ref array i low-arg)  close)))
	    (receive (pos val)
		(direction i high low)
	      (if pos
		  (array-set! new-array val pos))))))

      (if extreme-pos
	  (begin
	    (array-set! new-array extreme extreme-pos)
	    (array-set! new-array extreme hi)))

      new-array)))

(define-public (zigzag-series series percent use-closes)
  (let ((new-series (series-copy series)))
    (series-name! new-series (format #f (_ "Zig Zag ~a%")
				     (locale-number-min-decimals percent 0)))
    (series-datatype!  new-series zigzag-datatype)
    (series-parent!    new-series series)
    (series-data!      new-series (list percent use-closes))
    (series-require!   new-series zigzag-fill)
    (series-array!     new-series #f)
    (series-dividends! new-series '())
    (series-notes!     new-series '())

    new-series))

(define-public zigzag-datatype
  (datatype-new!
   #:key       'zigzag
   #:name      (_ "Zig Zag")
   #:units     #f
   #:linestyle 'zigzag
   #:manual    (_ "CONTEXT(manual-node): Zig Zag Indicator")
   #:proc      zigzag-series
   #:parameter-list
   (list (datatype-param-new
	  #:key     'zigzag-days
	  #:name    (_ "% Change")
	  #:type    'float
	  #:minimum 0
	  #:default 5
	  #:step    1)
	 (datatype-param-new
	  #:key     'zigzag-hl
	  #:name    (_ "Closes")
	  #:type    'boolean
	  #:default #f))))
