;; Zero lag exponential moving average (ZLEMA).

;; Copyright 2006, 2007 Kevin Ryde

;; This file is part of Chart.
;;
;; Chart is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Chart is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (chart zlema)
  #:use-module (ice-9 and-let-star)
  #:use-module (chart ema)
  #:use-module (chart indicator)
  #:use-module (chart misc))


;; http://www.linnsoft.com/tour/techind/movAvg.htm
;;     Showing as EMA of "2*price-price[lag]", where lag=(n-1)/2
;;
;; http://www.mesasoftware.com/technicalpapers.htm
;; http://www.mesasoftware.com/Papers/Zero%20Lag%20Data%20Smoothers.exe  (Zip)
;;     John Ehlers on zero lag, with graphs of frequency response.
;;


;; Lag calculation:
;;
;; Taking a decreasing sequence to today is 0, yesterday is 1, the day
;; before 2, etc, then the EMA today, using the power formula, is
;;
;;     EMA = (1-f) * (0 + 1*f + 2*f^2 + 3*f^3 + 4*f^4 + ...)
;;
;; Multiplying through gives
;;
;;     EMA = 0 + 1*f + 2*f^2 + 3*f^3 + 4*f^4 + ...
;;             - 0*f - 1*f^2 - 2*f^2 - 3*f^4 - ...
;;
;;         = f + f^2 + f^3 + f^4 + ...
;;
;;         = f * 1/(1-f)
;;
;; And with f=1-2/(N+1) meaning 1-f=2/(N+1), and also f=(N-1)/(N+1),
;;
;;           N-1   N+1
;;     EMA = --- * ---
;;           N+1    2
;;
;; So EMA = (N-1)/2.  Ie. the EMA is the value as at (N-1)/2 days ago, which
;; is the lag.
;;


;; Return a procedure which calculates a zero-lag exponential moving average
;; over an accumulated window.
;;
;; Each call (PROC val) enters a new value into the window, and the return
;; is the ZLEMA up to (and including) that value, or #f if not enough points
;; yet.
;;
;; An ZLEMA is in theory influenced by all preceding data, but
;; `zlema-warmup-count' below is designed to determine a warmup count.  By
;; calling PROC with (zlema-warmup-count COUNT) many values, the next call
;; will have an omitted weight of no more than 0.1% of the total.  Omitting
;; 0.1% should be negligable, unless past values are ridiculously bigger
;; than recent ones.
;;
(define-public (zlema-calc-proc count)

  (define ema-proc  (ema-calc-proc count))
  (define prev-proc (prevnth-calc-proc (quotient (1- count) 2))) ;; lag (n-1)/2

  (lambda (x)
    (and-let* ((ema  (ema-proc x))
	       (prev (prev-proc ema)))
      (- (* 2 ema) prev))))

(define-public (zlema-warmup-count count)
  (+ (quotient (1- count) 2)
     (ema-warmup-count count)))


(define (zlema-fill series new-array lo hi)
  (let* ((count  (series-data series))
	 (parent (series-parent series))
	 (array  (series-array-closes parent lo hi))
	 (proc   (zlema-calc-proc count)))

    ;; warmup
    (for-each proc (series-prev-n-closes parent
					 #:before lo
					 #:count (zlema-warmup-count count)))

    ;; points lo to hi inclusive
    (do ((i lo (1+ i)))
	((> i hi))
      (if (array-ref array i)
	  (array-set! new-array (proc (array-ref array i)) i)))))

(define-public (zlema-series series count)
  (let ((new-series (series-copy series)))
    (series-name!     new-series (format #f "~a ~a" (_ "ZLEMA") count))
    (series-datatype! new-series zlema-datatype)
    (series-parent!   new-series series)
    (series-data!     new-series count)
    (series-array!    new-series #f)
    (series-require!  new-series (indicator-require-parts
				  zlema-fill (zlema-warmup-count count)))

    new-series))

(define zlema-datatype
  (datatype-new!
   #:key      'zlema
   #:name     (_ "ZLEMA - Zero Lag EMA")
   #:units    #f
   #:manual   (_ "CONTEXT(manual-node): Zero-Lag Exponential Moving Average")
   #:proc     zlema-series
   #:parameter-list
   (list (datatype-param-new
	  #:name    (_ "Days")
	  #:key     'zlema-days
	  #:type    'integer
	  #:minimum 1
	  #:default 20))))
