/* Zlib interface code.

   Copyright 2005, 2006, 2007 Kevin Ryde

   This file is part of Chart.

   Chart is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   Chart is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"

#include <errno.h>
#include <stdio.h>     /* for NULL */
#include <string.h>

#include <libguile.h>

#if HAVE_LIBZ
#include <zlib.h>
#endif

#include "compat.h"


/*---------------------------------------------------------------------------*/
/* guile 1.8 C stuff make available in guile 1.6 */

#if SCM_MAJOR_VERSION == 1 && SCM_MINOR_VERSION == 6

/* not a full implementation, CHR is ignored */
#define scm_c_make_string(len, chr)  scm_allocate_string (len)

/* Guile 1.6 scm_take_str() is not equivalent to scm_take_locale_stringn.
   scm_take_str assumes there's already a terminating '\0', but
   scm_take_locale_stringn doesn't need that and will realloc to put
   one.  */
static SCM
scm_take_locale_stringn (char *str, size_t len)
{
  str = realloc (str, len + 1);
  if (str == NULL)
    scm_memory_error ("scm_take_locale_stringn");
  str[len] = '\0';
  return scm_take_str (str, len);
}

#endif


/*---------------------------------------------------------------------------*/
/* gzfile */

#if HAVE_LIBZ

static scm_t_bits gzfile_type;

#define GZFILE_PTR(gzfile)   ((gzFile *) SCM_SMOB_DATA (gzfile))

#define GZFILE_P(obj) \
  (SCM_SMOB_PREDICATE (gzfile_type, obj) && GZFILE_PTR(gzfile) != NULL)

#define GZFILE_VALIDATE(pos, scm)                                             \
  do {                                                                        \
    SCM_ASSERT_TYPE (GZFILE_P (scm), scm, pos, FUNC_NAME, "open gzfile");     \
  } while (0)

static size_t
gzfile_free (SCM gzfile)
{
  gzFile *gz = GZFILE_PTR (gzfile);
  if (gz != NULL)
    gzclose (gz);
  return 0;
}

SCM_DEFINE_PUBLIC (chart_gzopen, "gzopen", 2, 0, 0,
            (SCM filename, SCM mode),
	    "Open the gzipped file @var{filename} using Zlib @code{gzopen},\n"
	    "returning a gzfile object.\n"
	    "\n"
	    "@var{mode} is the open mode, like \"rb\" for read binary.\n"
	    "Currently only reading is supported in Chart.")
#define FUNC_NAME s_chart_gzopen
{
  gzFile *gzfile;
  SCM smob;
  char *c_filename, *c_mode;
  scm_dynwind_begin (0);

  c_filename = scm_to_locale_string (filename);
  scm_dynwind_free (c_filename);

  c_mode = scm_to_locale_string (mode);
  scm_dynwind_free (c_mode);

  errno = 0;
  gzfile = gzopen (c_filename, c_mode);
  if (gzfile == NULL)
    {
      if (errno == 0) /* Z_MEM_ERROR */
        SCM_MEMORY_ERROR;
      else
        SCM_SYSERROR;
    }

  SCM_NEWSMOB (smob, gzfile_type, gzfile);
  scm_dynwind_end ();
  return smob;
}
#undef FUNC_NAME

SCM_DEFINE_PUBLIC (chart_gzread, "gzread", 2, 0, 0,
            (SCM gzfile, SCM n),
	    "Read a string of @var{n} bytes read from the given @var{gzfile}\n"
	    "object.\n"
	    "\n"
	    "If eof is reached then the return may be less than @var{N}\n"
	    "bytes.  If there are no bytes left then an eof object is\n"
	    "returned, except that if @var{N}==0 the return is always just\n"
	    "an empty string, without checking for eof.")
#define FUNC_NAME s_chart_gzread
{
  gzFile   *gz;
  unsigned want;
  int      got;
  char     *buf;

  GZFILE_VALIDATE (SCM_ARG1, gzfile);
  gz = GZFILE_PTR (gzfile);
  want = scm_to_uint (n);

  /* degenerate case, nothing to read */
  if (want == 0)
    return scm_c_make_string (0, SCM_UNDEFINED);

  buf = scm_malloc (want);

  got = gzread (gz, buf, want);
  scm_remember_upto_here_1 (gzfile);

  if (got == -1)
    {
      const char *msg;
      int zerr, saved_errno;

      saved_errno = errno;
      free (buf);
      msg = gzerror (gz, &zerr);

      if (zerr == Z_ERRNO)
        {
          errno = saved_errno;
          SCM_SYSERROR;
        }

      scm_misc_error (FUNC_NAME, "~S",
                      scm_list_1 (scm_from_locale_string (msg)));
    }

  if (got == 0)
    {
      free (buf);
      return SCM_EOF_VAL;
    }

  return scm_take_locale_stringn (buf, got);
}
#undef FUNC_NAME


SCM_DEFINE_PUBLIC (chart_gzclose, "gzclose", 1, 0, 0,
            (SCM gzfile),
	    "Close the given @var{gzfile} object.")
#define FUNC_NAME s_chart_gzclose
{
  gzFile *gz;
  int zerr;

  GZFILE_VALIDATE (SCM_ARG1, gzfile);

  gz = GZFILE_PTR (gzfile);
  zerr = gzclose (gz);
  SCM_SET_SMOB_DATA (gzfile, (scm_t_bits) NULL);

  if (zerr == Z_ERRNO)
    SCM_SYSERROR;

  /* note: cannot use gzerror after closing */
  if (zerr != Z_OK)
    scm_misc_error (FUNC_NAME, "Zlib error code ~S",
                    scm_list_1 (scm_from_int (zerr)));

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#endif /* HAVE_LIBZ */


/*---------------------------------------------------------------------------*/
/* other procs */


#if HAVE_LIBZ

/* Return zp->msg as a scheme string, or the ret code as a string if msg is
   NULL.

   Note use of scm_string_append() here instead of scm_string_concatenate()
   since in guile 1.6 the latter is in srfi-13 and we're not really linking
   properly to that (though it ends up working on an elf system). */

SCM
chart_z_stream_msg (z_stream *zp, int ret)
{
  if (zp->msg != NULL)
    return scm_from_locale_string (zp->msg);
  else
    return scm_string_append
      (scm_list_2 (scm_from_locale_string ("Zlib error code "),
                   scm_number_to_string (scm_from_int (ret), SCM_UNDEFINED)));
}

SCM_DEFINE_PUBLIC (chart_zlib_inflate, "zlib-inflate", 1, 0, 0,
                   (SCM str),
                   "@var{str} is compressed data, either Zlib RFC 1950 format or\n"
                   "gzip RFC 1952 format.  Return an uncompressed string.")
#define FUNC_NAME s_chart_zlib_inflate
{
  z_stream z;
  int ret;
  char buf[4096];
  SCM lst = SCM_EOL, *p = &lst;
  char *c_str;
  size_t c_len;

  scm_dynwind_begin (0);

  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  c_str = scm_to_locale_stringn (str, &c_len);
  scm_dynwind_free (c_str);
  z.next_in = (Bytef *) c_str;
  z.avail_in = c_len;

  z.next_out = (Bytef *) buf;
  z.avail_out = sizeof (buf);

  /* ask to accept either gzip or zlib header formats */
  ret = inflateInit2 (&z, 32 + 15);
  if (ret != Z_OK)
    scm_misc_error (FUNC_NAME, "inflateInit2: ~S",
                    scm_list_1 (chart_z_stream_msg (&z, ret)));

  for (;;)
    {
      ret = inflate (&z, Z_NO_FLUSH);
      if (ret != Z_OK && ret != Z_STREAM_END)
        {
          SCM msg = chart_z_stream_msg (&z, ret);
          inflateEnd (&z);
          scm_misc_error (FUNC_NAME, "inflate: ~S", scm_list_1 (msg));
        }

      if (z.avail_out == 0 || ret == Z_STREAM_END)
        {
          *p = scm_cons (scm_from_locale_stringn (buf, sizeof (buf)
                                                       - z.avail_out),
                         SCM_EOL);
          p = SCM_CDRLOC (*p);

          z.next_out = (Bytef *) buf;
          z.avail_out = sizeof (buf);
        }

      if (ret == Z_STREAM_END)
        break;
    }

  if (z.avail_in != 0)
    {
      ret = inflateEnd (&z);
      scm_misc_error (FUNC_NAME, "garbage after compressed data",
                      SCM_EOL);
    }


  ret = inflateEnd (&z);
  if (ret != Z_OK)
    scm_misc_error (FUNC_NAME, "inflateEnd: ~S",
                    scm_list_1 (chart_z_stream_msg (&z, ret)));

  scm_dynwind_end ();
  return scm_string_append (lst);
}
#undef FUNC_NAME

#endif /* HAVE_LIBZ */


SCM_DEFINE_PUBLIC (chart_zlib_version, "zlib-version", 0, 0, 0,
            (),
	    "Return the Zlib version string, as obtained from the\n"
	    "@code{zlibVersion} function (not merely the header compiled\n"
	    "against).  If Zlib is not available, return @code{#f}.")
#define FUNC_NAME s_chart_zlib_version
{
#if HAVE_LIBZ
  return scm_from_locale_string (zlibVersion ());
#else
  return SCM_BOOL_F;
#endif
}


/*---------------------------------------------------------------------------*/

void
chart_zlib_init (void)
{
#if HAVE_LIBZ
  gzfile_type = scm_make_smob_type ("gzfile", 0);
  scm_set_smob_free (gzfile_type, gzfile_free);
#endif

#ifndef SCM_MAGIC_SNARFER
#include "zlib.x"
#endif
}
